#include "Python.h"
#include "internal/pycore_gc.h"
#include "internal/pycore_code.h"
#include "internal/pycore_frame.h"
#include "internal/pycore_long.h"

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[340];
    }
importlib__bootstrap_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 339,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x72\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x69\x6d\x70\x6f\x72\x74\x2e\x0a\x0a\x54\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x4e\x4f\x54\x20\x6d\x65\x61\x6e\x74\x20\x74\x6f\x20\x62\x65\x20\x64\x69\x72\x65\x63\x74\x6c\x79\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x21\x20\x49\x74\x20\x68\x61\x73\x20\x62\x65\x65\x6e\x20\x64\x65\x73\x69\x67\x6e\x65\x64\x20\x73\x75\x63\x68\x0a\x74\x68\x61\x74\x20\x69\x74\x20\x63\x61\x6e\x20\x62\x65\x20\x62\x6f\x6f\x74\x73\x74\x72\x61\x70\x70\x65\x64\x20\x69\x6e\x74\x6f\x20\x50\x79\x74\x68\x6f\x6e\x20\x61\x73\x20\x74\x68\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x69\x6d\x70\x6f\x72\x74\x2e\x20\x41\x73\x0a\x73\x75\x63\x68\x20\x69\x74\x20\x72\x65\x71\x75\x69\x72\x65\x73\x20\x74\x68\x65\x20\x69\x6e\x6a\x65\x63\x74\x69\x6f\x6e\x20\x6f\x66\x20\x73\x70\x65\x63\x69\x66\x69\x63\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x61\x6e\x64\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x73\x20\x69\x6e\x20\x6f\x72\x64\x65\x72\x20\x74\x6f\x0a\x77\x6f\x72\x6b\x2e\x20\x4f\x6e\x65\x20\x73\x68\x6f\x75\x6c\x64\x20\x75\x73\x65\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x20\x61\x73\x20\x74\x68\x65\x20\x70\x75\x62\x6c\x69\x63\x2d\x66\x61\x63\x69\x6e\x67\x20\x76\x65\x72\x73\x69\x6f\x6e\x20\x6f\x66\x20\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_AttributeError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AttributeError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(__qualname__),
            & const_str_AttributeError._ascii.ob_base,
            &_Py_ID(type),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_toplevel_consts_1_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen importlib._bootstrap>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__object_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_object_name",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[51];
    }
importlib__bootstrap_toplevel_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 50,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x03\x05\x26\xd8\x0f\x12\xd7\x0f\x1f\xd1\x0f\x1f\xd0\x08\x1f\xf8\xdc\x0b\x19\xf2\x00\x01\x05\x26\xdc\x0f\x13\x90\x43\x8b\x79\xd7\x0f\x25\xd1\x0f\x25\xd2\x08\x25\xf0\x03\x01\x05\x26\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_toplevel_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x0b\x0e\x00\x8e\x1e\x2f\x03\xae\x01\x2f\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(obj),
        },
    },
};
static
    struct _PyCode_DEF(100)
importlib__bootstrap_toplevel_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 50,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 23,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 1,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__object_name._ascii.ob_base,
    .co_qualname = & const_str__object_name._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x18\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[48];
    }
importlib__bootstrap_toplevel_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 47,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Simple substitute for functools.update_wrapper.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_3_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(__module__),
            &_Py_ID(__name__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_3_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_3_consts_1._object.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_hasattr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hasattr",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_setattr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "setattr",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_update = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "update",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_hasattr._ascii.ob_base,
            & const_str_setattr._ascii.ob_base,
            &_Py_ID(getattr),
            &_Py_ID(__dict__),
            & const_str_update._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__wrap = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
importlib__bootstrap_toplevel_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe3\x13\x48\x88\x07\xdc\x0b\x12\x90\x33\x98\x07\xd5\x0b\x20\xdc\x0c\x13\x90\x43\x98\x17\xa4\x27\xa8\x23\xa8\x77\xd3\x22\x37\xd5\x0c\x38\xf0\x05\x00\x14\x49\x01\xf0\x06\x00\x05\x08\x87\x4c\x81\x4c\xd7\x04\x17\xd1\x04\x17\x98\x03\x9f\x0c\x99\x0c\xd5\x04\x25",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_new = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "new",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_old = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "old",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_new._ascii.ob_base,
            & const_str_old._ascii.ob_base,
            &_Py_ID(replace),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[4];
    }
importlib__bootstrap_toplevel_consts_3_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 3,
    },
    .ob_shash = -1,
    .ob_sval = "   ",
};
static
    struct _PyCode_DEF(164)
importlib__bootstrap_toplevel_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 82,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 40,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 2,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__wrap._ascii.ob_base,
    .co_qualname = & const_str__wrap._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x44\x00\x5d\x26\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x10\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x28\x04\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_sys = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(type),
            & const_str_sys._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__new_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_new_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_toplevel_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x14\x8c\x34\x94\x03\x8b\x39\x90\x54\x8b\x3f\xd0\x04\x1a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(name),
        },
    },
};
static
    struct _PyCode_DEF(44)
importlib__bootstrap_toplevel_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 48,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 3,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__new_module._ascii.ob_base,
    .co_qualname = & const_str__new_module._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__List = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_List",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__List._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd8\x04\x08",
};
static
    struct _PyCode_DEF(12)
importlib__bootstrap_toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 55,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 4,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__List._ascii.ob_base,
    .co_qualname = & const_str__List._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str__WeakValueDictionary = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[48];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 47,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.__init__.<locals>.KeyedRef",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(key),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_super = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "super",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_remove = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "remove",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_super._ascii.ob_base,
            &_Py_ID(__new__),
            & const_str_remove._ascii.ob_base,
            &_Py_ID(key),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[56];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 55,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.__init__.<locals>.KeyedRef.__new__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[38];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 37,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x17\x1c\x91\x77\x91\x7f\xa0\x74\xa8\x52\xb0\x14\xb7\x1b\xb1\x1b\xd3\x17\x3d\x90\x04\xd8\x1b\x1e\x90\x04\x94\x08\xd8\x17\x1b\x90\x0b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_ob = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ob",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(type),
            & const_str_ob._ascii.ob_base,
            &_Py_ID(key),
            &_Py_ID(self),
            &_Py_ID(__class__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x20\x20\x80",
};
static
    struct _PyCode_DEF(76)
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 38,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 74,
    .co_nlocalsplus = 5,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 5,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__new__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x04\x7c\x00\x8d\x05\x00\x00\x7c\x00\x7c\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x02\x7c\x03\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_super._ascii.ob_base,
            &_Py_ID(__init__),
            & const_str_remove._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[57];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 56,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.__init__.<locals>.KeyedRef.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[23];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 22,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x10\x15\x91\x07\xd1\x10\x20\xa0\x12\xa0\x54\xa7\x5b\xa1\x5b\xd5\x10\x31",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_ob._ascii.ob_base,
            &_Py_ID(key),
            &_Py_ID(__class__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x20\x80",
};
static
    struct _PyCode_DEF(58)
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 29,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 79,
    .co_nlocalsplus = 4,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 6,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x03\x7c\x00\x8d\x05\x00\x00\x7c\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__iterating = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_iterating",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__pending_removals = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_pending_removals",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__weakref = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_weakref",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str__remove_dead_weakref = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_remove_dead_weakref",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__iterating._ascii.ob_base,
            & const_str__pending_removals._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(key),
            & const_str__weakref._ascii.ob_base,
            & const_str__remove_dead_weakref._ascii.ob_base,
            &_Py_ID(data),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.__init__.<locals>.KeyedRef.remove",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[79];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 78,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf1\x08\x00\x18\x24\x93\x7e\x90\x04\xd8\x13\x17\xd0\x13\x23\xd8\x17\x1b\x97\x7f\x92\x7f\xd8\x18\x1c\xd7\x18\x2e\xd1\x18\x2e\xd7\x18\x35\xd1\x18\x35\xb0\x62\xb7\x66\xb1\x66\xd5\x18\x3d\xe4\x18\x20\xd7\x18\x35\xd1\x18\x35\xb0\x64\xb7\x69\xb1\x69\xc0\x12\xc7\x16\xc1\x16\xd5\x18\x48\xf0\x09\x00\x14\x24",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_wr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "wr",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_self_weakref = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "self_weakref",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_wr._ascii.ob_base,
            &_Py_ID(self),
            & const_str_self_weakref._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[4];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 3,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x80",
};
static
    struct _PyCode_DEF(210)
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 105,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 82,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 7,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_remove._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x02\x00\x89\x02\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x81\x5d\x7c\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x26\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x79\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_1._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_staticmethod = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "staticmethod",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            &_Py_ID(__new__),
            &_Py_ID(__init__),
            & const_str_staticmethod._ascii.ob_base,
            & const_str_remove._ascii.ob_base,
            &_Py_ID(__classcell__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_KeyedRef = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "KeyedRef",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[41];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 40,
    },
    .ob_shash = -1,
    .ob_sval = "\xf9\x84\x00\xe0\x18\x1e\x88\x49\xf4\x04\x03\x0d\x1c\xf4\x0a\x01\x0d\x32\xf0\x06\x00\x0e\x1a\xf3\x02\x08\x0d\x49\x01\xf3\x03\x00\x0e\x1a\xf4\x02\x08\x0d\x49\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(__class__),
            & const_str_self_weakref._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\x40\x80",
};
static
    struct _PyCode_DEF(66)
importlib__bootstrap_toplevel_consts_7_consts_1_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 33,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 70,
    .co_nlocalsplus = 2,
    .co_nlocals = 0,
    .co_ncellvars = 1,
    .co_nfreevars = 1,
    .co_version = 8,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_KeyedRef._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_0._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x87\x00\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x88\x00\x66\x01\x64\x02\x84\x08\x5a\x04\x88\x00\x66\x01\x64\x03\x84\x08\x5a\x05\x65\x06\x88\x01\x66\x01\x64\x04\x84\x08\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x88\x00\x78\x01\x5a\x08\x53\x00",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1.ob_base.ob_base,
            & const_str_KeyedRef._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_ref = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ref",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__KeyedRef = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_KeyedRef",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__weakref._ascii.ob_base,
            & const_str_ref._ascii.ob_base,
            & const_str__KeyedRef._ascii.ob_base,
            &_Py_ID(clear),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[54];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 53,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x17\x1f\x97\x7c\x91\x7c\xa0\x44\xd3\x17\x29\x88\x0c\xf6\x0a\x15\x09\x49\x01\x94\x78\x97\x7c\x91\x7c\xf4\x00\x15\x09\x49\x01\xf0\x2e\x00\x1a\x22\x88\x04\x8c\x0e\xd8\x08\x0c\x8f\x0a\x89\x0a\x8d\x0c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_KeyedRef._ascii.ob_base,
            & const_str_self_weakref._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[4];
    }
importlib__bootstrap_toplevel_consts_7_consts_1_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 3,
    },
    .ob_shash = -1,
    .ob_sval = "  @",
};
static
    struct _PyCode_DEF(148)
importlib__bootstrap_toplevel_consts_7_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 74,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_7_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 64,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 9,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x02\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x8a\x02\x02\x00\x47\x00\x88\x02\x66\x01\x64\x01\x84\x08\x64\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_set = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "set",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__pending_removals._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            & const_str__iterating._ascii.ob_base,
            &_Py_ID(data),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_toplevel_consts_7_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.clear",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[27];
    }
importlib__bootstrap_toplevel_consts_7_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 26,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x21\x23\x88\x04\xd4\x08\x1e\xdc\x1a\x1d\x9b\x25\x88\x04\x8c\x0f\xd8\x14\x16\x88\x04\x8d\x09",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(self),
        },
    },
};
static
    struct _PyCode_DEF(62)
importlib__bootstrap_toplevel_consts_7_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 96,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 10,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(clear),
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x67\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x69\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_pop = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pop",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_IndexError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IndexError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__pending_removals._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            &_Py_ID(data),
            & const_str_IndexError._ascii.ob_base,
            & const_str__weakref._ascii.ob_base,
            & const_str__remove_dead_weakref._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__commit_removals = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_commit_removals",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
importlib__bootstrap_toplevel_consts_7_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary._commit_removals",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[87];
    }
importlib__bootstrap_toplevel_consts_7_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 86,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0e\x12\xd7\x0e\x24\xd1\x0e\x24\xd7\x0e\x28\xd1\x0e\x28\x88\x03\xd8\x0c\x10\x8f\x49\x89\x49\x88\x01\xd8\x0e\x12\xf0\x02\x03\x0d\x17\xd9\x16\x19\x93\x65\x90\x03\xf4\x06\x00\x0d\x15\xd7\x0c\x29\xd1\x0c\x29\xa8\x21\xa8\x53\xd4\x0c\x31\xf0\x0b\x00\x0f\x13\xf8\xf4\x06\x00\x14\x1e\xf2\x00\x01\x0d\x17\xd9\x10\x16\xf0\x03\x01\x0d\x17\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
importlib__bootstrap_toplevel_consts_7_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\xa5\x07\x41\x03\x00\xc1\x03\x09\x41\x0f\x03\xc1\x0e\x01\x41\x0f\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_pop._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[100],
            &_Py_ID(key),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "    ",
};
static
    struct _PyCode_DEF(164)
importlib__bootstrap_toplevel_consts_7_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 82,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_7_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 101,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 11,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__commit_removals._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x09\x00\x09\x00\x02\x00\x7c\x01\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x1f\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_KeyError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "KeyError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__pending_removals._ascii.ob_base,
            & const_str__commit_removals._ascii.ob_base,
            &_Py_ID(data),
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_7_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.get",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[88];
    }
importlib__bootstrap_toplevel_consts_7_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 87,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\xd7\x0b\x21\xd2\x0b\x21\xd8\x0c\x10\xd7\x0c\x21\xd1\x0c\x21\xd4\x0c\x23\xf0\x02\x08\x09\x19\xd8\x11\x15\x97\x19\x91\x19\x98\x33\x91\x1e\x88\x42\xf1\x08\x00\x16\x18\x93\x54\x90\x09\x90\x01\xd0\x0f\x22\xd8\x17\x1e\x90\x0e\xe0\x17\x18\x90\x08\xf8\xf4\x0d\x00\x10\x18\xf2\x00\x01\x09\x1b\xd8\x13\x1a\x8a\x4e\xf0\x03\x01\x09\x1b\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib__bootstrap_toplevel_consts_7_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x9e\x0f\x3a\x00\xba\x0b\x41\x08\x03\xc1\x07\x01\x41\x08\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(key),
            &_Py_ID(default),
            & const_str_wr._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[111],
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "     ",
};
static
    struct _PyCode_DEF(150)
importlib__bootstrap_toplevel_consts_7_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 75,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_7_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 111,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 12,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(get),
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x10\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x03\x02\x00\x7c\x03\xab\x00\x00\x00\x00\x00\x00\x00\x78\x01\x7d\x04\x80\x02\x7c\x02\x53\x00\x7c\x04\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x7c\x02\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(data),
            & const_str_KeyError._ascii.ob_base,
            & const_str__pending_removals._ascii.ob_base,
            & const_str__commit_removals._ascii.ob_base,
            & const_str__KeyedRef._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_setdefault = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "setdefault",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_toplevel_consts_7_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_WeakValueDictionary.setdefault",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[110];
    }
importlib__bootstrap_toplevel_consts_7_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 109,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x03\x09\x15\xd8\x10\x1e\x90\x04\x97\x09\x91\x09\x98\x23\x91\x0e\xd3\x10\x20\x88\x41\xf0\x06\x00\x0c\x0d\x88\x39\xd8\x0f\x13\xd7\x0f\x25\xd2\x0f\x25\xd8\x10\x14\xd7\x10\x25\xd1\x10\x25\xd4\x10\x27\xd8\x1d\x21\x9f\x5e\x99\x5e\xa8\x47\xb0\x53\xd3\x1d\x39\x88\x44\x8f\x49\x89\x49\x90\x63\x89\x4e\xd8\x13\x1a\x88\x4e\xe0\x13\x14\x88\x48\xf8\xf4\x11\x00\x10\x18\xf2\x00\x01\x09\x15\xd8\x10\x14\x8a\x41\xf0\x03\x01\x09\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
importlib__bootstrap_toplevel_consts_7_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x14\x41\x17\x00\xc1\x17\x0b\x41\x25\x03\xc1\x24\x01\x41\x25\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(key),
            &_Py_ID(default),
            (PyObject *)&_Py_SINGLETON(strings).ascii[111],
        },
    },
};
static
    struct _PyCode_DEF(208)
importlib__bootstrap_toplevel_consts_7_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 104,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_7_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 124,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 13,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_setdefault._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_7_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x02\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x80\x3d\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x10\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x7c\x02\x53\x00\x7c\x03\x53\x00\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x00\x7d\x03\x59\x00\x8c\x4e\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str__WeakValueDictionary._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_1.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_3.ob_base.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_7_consts_5.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_7_consts_6.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            &_Py_ID(clear),
            & const_str__commit_removals._ascii.ob_base,
            &_Py_ID(get),
            & const_str_setdefault._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
importlib__bootstrap_toplevel_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x04\x1e\x05\x15\xf2\x40\x01\x03\x05\x17\xf2\x0a\x08\x05\x32\xf3\x14\x0b\x05\x19\xf4\x1a\x0b\x05\x15",
};
static
    struct _PyCode_DEF(46)
importlib__bootstrap_toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 62,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 14,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__WeakValueDictionary._ascii.ob_base,
    .co_qualname = & const_str__WeakValueDictionary._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x07\x64\x05\x84\x01\x5a\x06\x64\x07\x64\x06\x84\x01\x5a\x07\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__BlockingOnManager = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_BlockingOnManager",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[60];
    }
importlib__bootstrap_toplevel_consts_9_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 59,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "A context manager responsible to updating ``_blocking_on``.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_thread_id = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "thread_id",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_lock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lock",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_thread_id._ascii.ob_base,
            & const_str_lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_9_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_BlockingOnManager.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib__bootstrap_toplevel_consts_9_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x19\x22\x88\x04\x8c\x0e\xd8\x14\x18\x88\x04\x8d\x09",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_thread_id._ascii.ob_base,
            & const_str_lock._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(32)
importlib__bootstrap_toplevel_consts_9_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_9_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 158,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 15,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_9_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_9_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_9_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[68];
    }
importlib__bootstrap_toplevel_consts_9_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 67,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mark the running thread as waiting for self.lock. via _blocking_on.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_9_consts_3_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__blocking_on = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_blocking_on",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_blocked_on = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "blocked_on",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__blocking_on._ascii.ob_base,
            & const_str_setdefault._ascii.ob_base,
            & const_str_thread_id._ascii.ob_base,
            & const_str__List._ascii.ob_base,
            & const_str_blocked_on._ascii.ob_base,
            &_Py_ID(append),
            & const_str_lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_toplevel_consts_9_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_BlockingOnManager.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[53];
    }
importlib__bootstrap_toplevel_consts_9_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 52,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x1b\x27\xd7\x1a\x31\xd1\x1a\x31\xb0\x24\xb7\x2e\xb1\x2e\xc4\x25\xc3\x27\xd3\x1a\x4a\x88\x04\x8c\x0f\xd8\x08\x0c\x8f\x0f\x89\x0f\xd7\x08\x1e\xd1\x08\x1e\x98\x74\x9f\x79\x99\x79\xd5\x08\x29",
};
static
    struct _PyCode_DEF(168)
importlib__bootstrap_toplevel_consts_9_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 84,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_9_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_9_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 162,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 16,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_9_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_9_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
importlib__bootstrap_toplevel_consts_9_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Remove self.lock from this thread's _blocking_on list.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_9_consts_4_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_blocked_on._ascii.ob_base,
            & const_str_remove._ascii.ob_base,
            & const_str_lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_9_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_BlockingOnManager.__exit__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
importlib__bootstrap_toplevel_consts_9_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x08\x0c\x8f\x0f\x89\x0f\xd7\x08\x1e\xd1\x08\x1e\x98\x74\x9f\x79\x99\x79\xd5\x08\x29",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_kwargs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "kwargs",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(args),
            & const_str_kwargs._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(78)
importlib__bootstrap_toplevel_consts_9_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_9_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_9_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 15,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 173,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 17,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_9_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_9_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_9_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__BlockingOnManager._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_9_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_9_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_9_consts_3.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_9_consts_4.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib__bootstrap_toplevel_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd9\x04\x45\xf2\x02\x02\x05\x19\xf2\x08\x09\x05\x2a\xf3\x16\x02\x05\x2a",
};
static
    struct _PyCode_DEF(34)
importlib__bootstrap_toplevel_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 17,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_9_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 156,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 18,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__BlockingOnManager._ascii.ob_base,
    .co_qualname = & const_str__BlockingOnManager._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__DeadlockError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DeadlockError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__DeadlockError._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct _PyCode_DEF(12)
importlib__bootstrap_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 178,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 19,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__DeadlockError._ascii.ob_base,
    .co_qualname = & const_str__DeadlockError._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[755];
    }
importlib__bootstrap_toplevel_consts_13_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 754,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x68\x65\x63\x6b\x20\x69\x66\x20\x27\x74\x61\x72\x67\x65\x74\x5f\x69\x64\x27\x20\x69\x73\x20\x68\x6f\x6c\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x6c\x6f\x63\x6b\x20\x61\x73\x20\x61\x6e\x6f\x74\x68\x65\x72\x20\x74\x68\x72\x65\x61\x64\x28\x73\x29\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x73\x65\x61\x72\x63\x68\x20\x77\x69\x74\x68\x69\x6e\x20\x27\x62\x6c\x6f\x63\x6b\x69\x6e\x67\x5f\x6f\x6e\x27\x20\x73\x74\x61\x72\x74\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x74\x68\x72\x65\x61\x64\x73\x20\x6c\x69\x73\x74\x65\x64\x20\x69\x6e\x0a\x20\x20\x20\x20\x27\x63\x61\x6e\x64\x69\x64\x61\x74\x65\x5f\x69\x64\x73\x27\x2e\x20\x20\x27\x73\x65\x65\x6e\x5f\x69\x64\x73\x27\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x20\x61\x6e\x79\x20\x74\x68\x72\x65\x61\x64\x73\x20\x74\x68\x61\x74\x20\x61\x72\x65\x20\x63\x6f\x6e\x73\x69\x64\x65\x72\x65\x64\x0a\x20\x20\x20\x20\x61\x6c\x72\x65\x61\x64\x79\x20\x74\x72\x61\x76\x65\x72\x73\x65\x64\x20\x69\x6e\x20\x74\x68\x65\x20\x73\x65\x61\x72\x63\x68\x2e\x0a\x0a\x20\x20\x20\x20\x4b\x65\x79\x77\x6f\x72\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x3a\x0a\x20\x20\x20\x20\x74\x61\x72\x67\x65\x74\x5f\x69\x64\x20\x20\x20\x20\x20\x2d\x2d\x20\x54\x68\x65\x20\x74\x68\x72\x65\x61\x64\x20\x69\x64\x20\x74\x6f\x20\x74\x72\x79\x20\x74\x6f\x20\x72\x65\x61\x63\x68\x2e\x0a\x20\x20\x20\x20\x73\x65\x65\x6e\x5f\x69\x64\x73\x20\x20\x20\x20\x20\x20\x2d\x2d\x20\x41\x20\x73\x65\x74\x20\x6f\x66\x20\x74\x68\x72\x65\x61\x64\x73\x20\x74\x68\x61\x74\x20\x68\x61\x76\x65\x20\x61\x6c\x72\x65\x61\x64\x79\x20\x62\x65\x65\x6e\x20\x76\x69\x73\x69\x74\x65\x64\x2e\x0a\x20\x20\x20\x20\x63\x61\x6e\x64\x69\x64\x61\x74\x65\x5f\x69\x64\x73\x20\x2d\x2d\x20\x54\x68\x65\x20\x74\x68\x72\x65\x61\x64\x20\x69\x64\x73\x20\x66\x72\x6f\x6d\x20\x77\x68\x69\x63\x68\x20\x74\x6f\x20\x62\x65\x67\x69\x6e\x2e\x0a\x20\x20\x20\x20\x62\x6c\x6f\x63\x6b\x69\x6e\x67\x5f\x6f\x6e\x20\x20\x20\x2d\x2d\x20\x41\x20\x64\x69\x63\x74\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x65\x20\x74\x68\x72\x65\x61\x64\x2f\x62\x6c\x6f\x63\x6b\x69\x6e\x67\x2d\x6f\x6e\x20\x67\x72\x61\x70\x68\x2e\x20\x20\x54\x68\x69\x73\x20\x6d\x61\x79\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x62\x65\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x61\x73\x20\x74\x68\x65\x20\x67\x6c\x6f\x62\x61\x6c\x20\x27\x5f\x62\x6c\x6f\x63\x6b\x69\x6e\x67\x5f\x6f\x6e\x27\x20\x62\x75\x74\x20\x69\x74\x20\x69\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x61\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x20\x74\x6f\x20\x72\x65\x64\x75\x63\x65\x20\x74\x68\x65\x20\x69\x6d\x70\x61\x63\x74\x20\x74\x68\x61\x74\x20\x67\x6c\x6f\x62\x61\x6c\x20\x6d\x75\x74\x61\x62\x6c\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x61\x74\x65\x20\x68\x61\x73\x20\x6f\x6e\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x20\x6f\x66\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_seen_ids = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "seen_ids",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_candidate_ids = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "candidate_ids",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_blocking_on = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "blocking_on",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_13_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_seen_ids._ascii.ob_base,
            & const_str_candidate_ids._ascii.ob_base,
            & const_str_blocking_on._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_13_consts_0._ascii.ob_base,
            Py_True,
            Py_False,
            & importlib__bootstrap_toplevel_consts_13_consts_3._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__has_deadlocked = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_has_deadlocked",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(get),
            &_Py_ID(add),
            &_Py_ID(owner),
            & const_str__has_deadlocked._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[136];
    }
importlib__bootstrap_toplevel_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 135,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x20\x00\x08\x11\x90\x4d\xd1\x07\x21\xf0\x06\x00\x10\x14\xf3\x06\x00\x10\x1d\x88\x03\xd8\x29\x34\xaf\x1f\xa9\x1f\xb8\x13\xd3\x29\x3d\xd0\x10\x3d\xd0\x10\x25\xd0\x10\x3d\xe0\x0c\x14\xd8\x0d\x10\x90\x48\x89\x5f\xf1\x0a\x00\x14\x19\xd8\x08\x10\x8f\x0c\x89\x0c\x90\x53\xd4\x08\x19\xf1\x06\x00\x29\x3e\xd3\x10\x3e\xd1\x28\x3d\xa0\x04\x90\x14\x97\x1a\x93\x1a\xd0\x28\x3d\x88\x05\xd0\x10\x3e\xdc\x0b\x1a\x98\x39\xa8\x78\xc0\x75\xd8\x1c\x27\xf6\x03\x01\x0c\x29\xe1\x13\x17\xf0\x21\x00\x10\x1d\xf0\x24\x00\x0c\x11\xf9\xf2\x0b\x00\x11\x3f",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_toplevel_consts_13_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\xba\x13\x41\x23\x06",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_target_id = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "target_id",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_tid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tid",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_candidate_blocking_on = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "candidate_blocking_on",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_edges = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "edges",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_13_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_target_id._ascii.ob_base,
            & const_str_seen_ids._ascii.ob_base,
            & const_str_candidate_ids._ascii.ob_base,
            & const_str_blocking_on._ascii.ob_base,
            & const_str_tid._ascii.ob_base,
            & const_str_candidate_blocking_on._ascii.ob_base,
            & const_str_lock._ascii.ob_base,
            & const_str_edges._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
importlib__bootstrap_toplevel_consts_13_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "        ",
};
static
    struct _PyCode_DEF(208)
importlib__bootstrap_toplevel_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 104,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_13_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_13_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 3,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 183,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 20,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_13_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__has_deadlocked._ascii.ob_base,
    .co_qualname = & const_str__has_deadlocked._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x7c\x02\x76\x00\x72\x01\x79\x01\x7c\x02\x44\x00\x5d\x57\x00\x00\x7d\x04\x7c\x03\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x78\x01\x7d\x05\x73\x01\x8c\x17\x7c\x04\x7c\x01\x76\x00\x72\x02\x01\x00\x79\x02\x7c\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x44\x00\x8f\x06\x63\x02\x67\x00\x63\x02\x5d\x0e\x00\x00\x7d\x06\x7c\x06\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x91\x02\x8c\x10\x04\x00\x7d\x07\x7d\x06\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x07\x7c\x03\xac\x03\xab\x04\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x57\x01\x00\x79\x01\x04\x00\x79\x02\x63\x02\x01\x00\x63\x02\x7d\x06\x77\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__ModuleLock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLock",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[170];
    }
importlib__bootstrap_toplevel_consts_14_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 169,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x72\x65\x63\x75\x72\x73\x69\x76\x65\x20\x6c\x6f\x63\x6b\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x77\x68\x69\x63\x68\x20\x69\x73\x20\x61\x62\x6c\x65\x20\x74\x6f\x20\x64\x65\x74\x65\x63\x74\x20\x64\x65\x61\x64\x6c\x6f\x63\x6b\x73\x0a\x20\x20\x20\x20\x28\x65\x2e\x67\x2e\x20\x74\x68\x72\x65\x61\x64\x20\x31\x20\x74\x72\x79\x69\x6e\x67\x20\x74\x6f\x20\x74\x61\x6b\x65\x20\x6c\x6f\x63\x6b\x73\x20\x41\x20\x74\x68\x65\x6e\x20\x42\x2c\x20\x61\x6e\x64\x20\x74\x68\x72\x65\x61\x64\x20\x32\x20\x74\x72\x79\x69\x6e\x67\x20\x74\x6f\x0a\x20\x20\x20\x20\x74\x61\x6b\x65\x20\x6c\x6f\x63\x6b\x73\x20\x42\x20\x74\x68\x65\x6e\x20\x41\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__thread = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_thread",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_RLock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "RLock",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_allocate_lock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "allocate_lock",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_wakeup = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "wakeup",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_waiters = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "waiters",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str__thread._ascii.ob_base,
            & const_str_RLock._ascii.ob_base,
            & const_str_lock._ascii.ob_base,
            & const_str_allocate_lock._ascii.ob_base,
            & const_str_wakeup._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(owner),
            &_Py_ID(count),
            & const_str_waiters._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_toplevel_consts_14_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLock.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[70];
    }
importlib__bootstrap_toplevel_consts_14_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 69,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x2a\x00\x15\x1c\x97\x4d\x91\x4d\x93\x4f\x88\x04\x8c\x09\xdc\x16\x1d\xd7\x16\x2b\xd1\x16\x2b\xd3\x16\x2d\x88\x04\x8c\x0b\xf0\x06\x00\x15\x19\x88\x04\x8c\x09\xf0\x08\x00\x16\x1a\x88\x04\x8c\x0a\xf0\x16\x00\x16\x18\x88\x04\x8c\x0a\xf0\x1c\x00\x18\x1a\x88\x04\x8d\x0c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "  ",
};
static
    struct _PyCode_DEF(160)
importlib__bootstrap_toplevel_consts_14_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 80,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 232,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 21,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_14_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_14_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_14_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x7c\x00\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x7c\x00\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x7c\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_3_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_target_id._ascii.ob_base,
            & const_str_seen_ids._ascii.ob_base,
            & const_str_candidate_ids._ascii.ob_base,
            & const_str_blocking_on._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_14_consts_3_consts_1._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_get_ident = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_ident",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__has_deadlocked._ascii.ob_base,
            & const_str__thread._ascii.ob_base,
            & const_str_get_ident._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            &_Py_ID(owner),
            & const_str__blocking_on._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_has_deadlock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "has_deadlock",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_14_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLock.has_deadlock",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
importlib__bootstrap_toplevel_consts_14_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x10\x1f\xe4\x16\x1d\xd7\x16\x27\xd1\x16\x27\xd3\x16\x29\xdc\x15\x18\x93\x55\xf0\x06\x00\x1c\x20\x9f\x3a\x99\x3a\x98\x2c\xe4\x18\x24\xf4\x11\x09\x10\x0a\xf0\x00\x09\x09\x0a",
};
static
    struct _PyCode_DEF(114)
importlib__bootstrap_toplevel_consts_14_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 57,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_14_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_14_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 288,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 22,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_has_deadlock._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_14_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_14_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x01\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xac\x01\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[186];
    }
importlib__bootstrap_toplevel_consts_14_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 185,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x41\x63\x71\x75\x69\x72\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6c\x6f\x63\x6b\x2e\x20\x20\x49\x66\x20\x61\x20\x70\x6f\x74\x65\x6e\x74\x69\x61\x6c\x20\x64\x65\x61\x64\x6c\x6f\x63\x6b\x20\x69\x73\x20\x64\x65\x74\x65\x63\x74\x65\x64\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x20\x5f\x44\x65\x61\x64\x6c\x6f\x63\x6b\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4f\x74\x68\x65\x72\x77\x69\x73\x65\x2c\x20\x74\x68\x65\x20\x6c\x6f\x63\x6b\x20\x69\x73\x20\x61\x6c\x77\x61\x79\x73\x20\x61\x63\x71\x75\x69\x72\x65\x64\x20\x61\x6e\x64\x20\x54\x72\x75\x65\x20\x69\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
importlib__bootstrap_toplevel_consts_14_consts_4_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "deadlock detected by ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_14_consts_4_consts_0._ascii.ob_base,
            Py_True,
            Py_None,
            & importlib__bootstrap_toplevel_consts_14_consts_4_consts_3._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_acquire = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "acquire",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str__thread._ascii.ob_base,
            & const_str_get_ident._ascii.ob_base,
            & const_str__BlockingOnManager._ascii.ob_base,
            & const_str_lock._ascii.ob_base,
            &_Py_ID(count),
            &_Py_ID(owner),
            &_Py_ID(append),
            & const_str_has_deadlock._ascii.ob_base,
            & const_str__DeadlockError._ascii.ob_base,
            & const_str_wakeup._ascii.ob_base,
            & const_str_acquire._ascii.ob_base,
            & const_str_waiters._ascii.ob_base,
            &_Py_ID(release),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_toplevel_consts_14_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLock.acquire",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[245];
    }
importlib__bootstrap_toplevel_consts_14_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 244,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x0f\x16\xd7\x0e\x1f\xd1\x0e\x1f\xd3\x0e\x21\x88\x03\xdc\x0d\x1f\xa0\x03\xa0\x54\xd5\x0d\x2a\xd8\x12\x16\xf0\x08\x00\x16\x1a\x97\x59\x93\x59\xd8\x17\x1b\x97\x7a\x91\x7a\xa0\x52\xd2\x17\x27\xa8\x34\xaf\x3a\xa9\x3a\xb8\x13\xd2\x2b\x3c\xf0\x0e\x00\x26\x29\x98\x04\x9c\x0a\xd8\x18\x1c\x9f\x0a\x99\x0a\xd7\x18\x29\xd1\x18\x29\xa8\x24\xd4\x18\x2f\xd8\x1f\x23\xf7\x15\x00\x16\x1f\xf7\x0b\x00\x0e\x2b\xd0\x0d\x2a\xf0\x44\x01\x00\x18\x1c\xd7\x17\x28\xd1\x17\x28\xd4\x17\x2a\xdc\x1e\x2c\xd0\x2f\x44\xc0\x54\xc0\x48\xd0\x2d\x4d\xd3\x1e\x4e\xd0\x18\x4e\xf0\x1a\x00\x18\x1c\x97\x7b\x91\x7b\xd7\x17\x2a\xd1\x17\x2a\xa8\x35\xd4\x17\x31\xd8\x18\x1c\x9f\x0c\x99\x0c\xd7\x18\x2b\xd1\x18\x2b\xa8\x44\xd4\x18\x31\xf7\x59\x01\x00\x16\x1f\xf0\x62\x01\x00\x11\x15\x97\x0b\x91\x0b\xd7\x10\x23\xd1\x10\x23\xd4\x10\x25\xf0\x0a\x00\x11\x15\x97\x0b\x91\x0b\xd7\x10\x23\xd1\x10\x23\xd4\x10\x25\xf0\x75\x01\x00\x13\x17\xf7\x08\x00\x16\x1f\x90\x59\xfa\xf7\x0b\x00\x0e\x2b\xd0\x0d\x2a\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
importlib__bootstrap_toplevel_consts_14_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\xa1\x0e\x44\x1f\x03\xaf\x41\x02\x44\x13\x05\xc1\x31\x08\x44\x1f\x03\xc2\x02\x41\x14\x44\x13\x05\xc3\x16\x3d\x44\x1f\x03\xc4\x13\x05\x44\x1c\x09\xc4\x18\x07\x44\x1f\x03\xc4\x1f\x05\x44\x28\x07",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_tid._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(598)
importlib__bootstrap_toplevel_consts_14_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 299,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_14_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_14_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_14_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 304,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 23,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_14_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_acquire._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_14_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_14_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x09\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x6b\x28\x00\x00\x73\x0f\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6b\x28\x00\x00\x72\x34\x7c\x01\x7c\x00\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x64\x02\x64\x02\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x64\x02\x64\x02\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01\x7c\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x72\x0e\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x72\x1b\x7c\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x02\x64\x02\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x8c\xf0\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x3e\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x79\x02\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_toplevel_consts_14_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cannot release un-acquired lock",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_14_consts_5_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_RuntimeError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "RuntimeError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str__thread._ascii.ob_base,
            & const_str_get_ident._ascii.ob_base,
            & const_str_lock._ascii.ob_base,
            &_Py_ID(owner),
            & const_str_RuntimeError._ascii.ob_base,
            &_Py_ID(len),
            &_Py_ID(count),
            & const_str_pop._ascii.ob_base,
            & const_str_waiters._ascii.ob_base,
            & const_str_wakeup._ascii.ob_base,
            &_Py_ID(release),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_toplevel_consts_14_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLock.release",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[161];
    }
importlib__bootstrap_toplevel_consts_14_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 160,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0e\x15\xd7\x0e\x1f\xd1\x0e\x1f\xd3\x0e\x21\x88\x03\xd8\x0d\x11\x8f\x59\x8b\x59\xd8\x0f\x13\x8f\x7a\x89\x7a\x98\x53\xd2\x0f\x20\xdc\x16\x22\xd0\x23\x44\xd3\x16\x45\xd0\x10\x45\xdc\x13\x16\x90\x74\x97\x7a\x91\x7a\x93\x3f\xa0\x51\xd2\x13\x26\xd0\x0c\x26\xd0\x13\x26\xd8\x0c\x10\x8f\x4a\x89\x4a\x8f\x4e\x89\x4e\xd4\x0c\x1c\xdc\x13\x16\x90\x74\x97\x7a\x91\x7a\x94\x3f\xd8\x1d\x21\x90\x04\x94\x0a\xdc\x13\x16\x90\x74\x97\x7c\x91\x7c\xd3\x13\x24\xa0\x71\xd2\x13\x28\xd8\x14\x18\x97\x4c\x91\x4c\xd7\x14\x24\xd1\x14\x24\xd4\x14\x26\xd8\x14\x18\x97\x4b\x91\x4b\xd7\x14\x27\xd1\x14\x27\xd4\x14\x29\xf7\x13\x00\x0e\x17\x8f\x59\x89\x59\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_toplevel_consts_14_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\xa1\x42\x37\x43\x21\x03\xc3\x21\x05\x43\x2a\x07",
};
static
    struct _PyCode_DEF(474)
importlib__bootstrap_toplevel_consts_14_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 237,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_14_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_14_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_14_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 372,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 24,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_14_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(release),
    .co_qualname = & importlib__bootstrap_toplevel_consts_14_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_14_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6b\x37\x00\x00\x72\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x44\x00\x00\x73\x02\x4a\x00\x82\x01\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x53\x64\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x44\x00\x00\x72\x34\x7c\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x79\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
importlib__bootstrap_toplevel_consts_14_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLock(",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
importlib__bootstrap_toplevel_consts_14_consts_6_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ") at ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_14_consts_6_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_6_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(id),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_toplevel_consts_14_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLock.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib__bootstrap_toplevel_consts_14_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x11\x1d\x98\x64\x9f\x69\x99\x69\x98\x5d\xa8\x25\xb4\x02\xb0\x34\xb3\x08\xa8\x7a\xd0\x0f\x3a\xd0\x08\x3a",
};
static
    struct _PyCode_DEF(56)
importlib__bootstrap_toplevel_consts_14_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_14_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_14_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 385,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 25,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_14_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_14_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x02\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str__ModuleLock._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_3.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_4.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_5.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_6.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            & const_str_has_deadlock._ascii.ob_base,
            & const_str_acquire._ascii.ob_base,
            &_Py_ID(release),
            &_Py_ID(__repr__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
importlib__bootstrap_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x03\x05\x08\xf2\x0a\x36\x05\x1a\xf2\x70\x01\x0e\x05\x0a\xf2\x20\x42\x01\x05\x26\xf2\x48\x02\x0b\x05\x2a\xf3\x1a\x01\x05\x3b",
};
static
    struct _PyCode_DEF(46)
importlib__bootstrap_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 226,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 26,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__ModuleLock._ascii.ob_base,
    .co_qualname = & const_str__ModuleLock._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__DummyModuleLock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DummyModuleLock",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[87];
    }
importlib__bootstrap_toplevel_consts_16_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 86,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x73\x69\x6d\x70\x6c\x65\x20\x5f\x4d\x6f\x64\x75\x6c\x65\x4c\x6f\x63\x6b\x20\x65\x71\x75\x69\x76\x61\x6c\x65\x6e\x74\x20\x66\x6f\x72\x20\x50\x79\x74\x68\x6f\x6e\x20\x62\x75\x69\x6c\x64\x73\x20\x77\x69\x74\x68\x6f\x75\x74\x0a\x20\x20\x20\x20\x6d\x75\x6c\x74\x69\x2d\x74\x68\x72\x65\x61\x64\x69\x6e\x67\x20\x73\x75\x70\x70\x6f\x72\x74\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(count),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_16_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DummyModuleLock.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib__bootstrap_toplevel_consts_16_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x14\x18\x88\x04\x8c\x09\xd8\x15\x16\x88\x04\x8d\x0a",
};
static
    struct _PyCode_DEF(32)
importlib__bootstrap_toplevel_consts_16_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_16_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 393,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 27,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_14_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_16_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_16_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(count),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_16_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DummyModuleLock.acquire",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_toplevel_consts_16_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0a\x8a\x0a\x90\x61\x89\x0f\x8d\x0a\xd8\x0f\x13",
};
static
    struct _PyCode_DEF(46)
importlib__bootstrap_toplevel_consts_16_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_16_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 397,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 28,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_acquire._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_16_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_16_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x78\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7a\x0d\x00\x00\x63\x02\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_toplevel_consts_14_consts_5_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(count),
            & const_str_RuntimeError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_16_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DummyModuleLock.release",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[39];
    }
importlib__bootstrap_toplevel_consts_16_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 38,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\x8f\x3a\x89\x3a\x98\x11\x8a\x3f\xdc\x12\x1e\xd0\x1f\x40\xd3\x12\x41\xd0\x0c\x41\xd8\x08\x0c\x8f\x0a\x8a\x0a\x90\x61\x89\x0f\x8e\x0a",
};
static
    struct _PyCode_DEF(98)
importlib__bootstrap_toplevel_consts_16_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_16_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 401,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 29,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(release),
    .co_qualname = & importlib__bootstrap_toplevel_consts_16_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_16_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x0b\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x78\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7a\x17\x00\x00\x63\x02\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_toplevel_consts_16_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DummyModuleLock(",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_16_consts_5_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_14_consts_6_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_16_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DummyModuleLock.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib__bootstrap_toplevel_consts_16_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x11\x22\xa0\x34\xa7\x39\xa1\x39\xa0\x2d\xa8\x75\xb4\x52\xb8\x04\xb3\x58\xb0\x4a\xd0\x0f\x3f\xd0\x08\x3f",
};
static
    struct _PyCode_DEF(56)
importlib__bootstrap_toplevel_consts_16_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_14_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 406,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 30,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_16_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_16_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x02\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__DummyModuleLock._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_16_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_16_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_16_consts_3.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_16_consts_4.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_16_consts_5.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            & const_str_acquire._ascii.ob_base,
            &_Py_ID(release),
            &_Py_ID(__repr__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
importlib__bootstrap_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x01\x05\x20\xf2\x06\x02\x05\x17\xf2\x08\x02\x05\x14\xf2\x08\x03\x05\x18\xf3\x0a\x01\x05\x40\x01",
};
static
    struct _PyCode_DEF(40)
importlib__bootstrap_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 389,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 31,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__DummyModuleLock._ascii.ob_base,
    .co_qualname = & const_str__DummyModuleLock._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x79\x06",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__ModuleLockManager = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLockManager",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__lock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_lock",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_18_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__name._ascii.ob_base,
            & const_str__lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_18_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLockManager.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib__bootstrap_toplevel_consts_18_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x15\x19\x88\x04\x8c\x0a\xd8\x15\x19\x88\x04\x8d\x0a",
};
static
    struct _PyCode_DEF(32)
importlib__bootstrap_toplevel_consts_18_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_18_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 412,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 32,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_14_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_18_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_18_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__get_module_lock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_module_lock",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_18_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__get_module_lock._ascii.ob_base,
            & const_str__name._ascii.ob_base,
            & const_str__lock._ascii.ob_base,
            & const_str_acquire._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_toplevel_consts_18_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLockManager.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[35];
    }
importlib__bootstrap_toplevel_consts_18_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 34,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x15\x25\xa0\x64\xa7\x6a\xa1\x6a\xd3\x15\x31\x88\x04\x8c\x0a\xd8\x08\x0c\x8f\x0a\x89\x0a\xd7\x08\x1a\xd1\x08\x1a\xd5\x08\x1c",
};
static
    struct _PyCode_DEF(108)
importlib__bootstrap_toplevel_consts_18_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 54,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_18_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 416,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 33,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_18_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_18_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_18_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__lock._ascii.ob_base,
            &_Py_ID(release),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_18_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModuleLockManager.__exit__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_toplevel_consts_18_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0a\x89\x0a\xd7\x08\x1a\xd1\x08\x1a\xd5\x08\x1c",
};
static
    struct _PyCode_DEF(56)
importlib__bootstrap_toplevel_consts_18_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_18_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 15,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 420,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 34,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_9_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_18_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_18_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__ModuleLockManager._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_18_consts_1.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_18_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_18_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
importlib__bootstrap_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x04\x02\x05\x1a\xf2\x08\x02\x05\x1d\xf3\x08\x01\x05\x1d",
};
static
    struct _PyCode_DEF(30)
importlib__bootstrap_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 410,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 35,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__ModuleLockManager._ascii.ob_base,
    .co_qualname = & const_str__ModuleLockManager._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[140];
    }
importlib__bootstrap_toplevel_consts_20_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 139,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x47\x65\x74\x20\x6f\x72\x20\x63\x72\x65\x61\x74\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6c\x6f\x63\x6b\x20\x66\x6f\x72\x20\x61\x20\x67\x69\x76\x65\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x41\x63\x71\x75\x69\x72\x65\x2f\x72\x65\x6c\x65\x61\x73\x65\x20\x69\x6e\x74\x65\x72\x6e\x61\x6c\x6c\x79\x20\x74\x68\x65\x20\x67\x6c\x6f\x62\x61\x6c\x20\x69\x6d\x70\x6f\x72\x74\x20\x6c\x6f\x63\x6b\x20\x74\x6f\x20\x70\x72\x6f\x74\x65\x63\x74\x0a\x20\x20\x20\x20\x5f\x6d\x6f\x64\x75\x6c\x65\x5f\x6c\x6f\x63\x6b\x73\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str__imp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_imp",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_acquire_lock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "acquire_lock",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__module_locks = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_module_locks",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_release_lock = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "release_lock",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_20_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_acquire_lock._ascii.ob_base,
            & const_str__module_locks._ascii.ob_base,
            &_Py_ID(get),
            & const_str_release_lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_cb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cb",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_toplevel_consts_20_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_module_lock.<locals>.cb",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[74];
    }
importlib__bootstrap_toplevel_consts_20_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 73,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x10\x14\xd7\x10\x21\xd1\x10\x21\xd4\x10\x23\xf0\x02\x07\x11\x28\xf4\x08\x00\x18\x25\xd7\x17\x28\xd1\x17\x28\xa8\x14\xd3\x17\x2e\xb0\x23\xd1\x17\x35\xdc\x1c\x29\xa8\x24\xd0\x1c\x2f\xe4\x14\x18\xd7\x14\x25\xd1\x14\x25\xd5\x14\x27\xf8\x94\x44\xd7\x14\x25\xd1\x14\x25\xd5\x14\x27\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
importlib__bootstrap_toplevel_consts_20_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "\x96\x1e\x41\x09\x00\xc1\x09\x16\x41\x1f\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_20_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_ref._ascii.ob_base,
            &_Py_ID(name),
        },
    },
};
static
    struct _PyCode_DEF(196)
importlib__bootstrap_toplevel_consts_20_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 98,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_20_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_20_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 19,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 445,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 36,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_20_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_cb._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_20_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_20_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x75\x00\x72\x07\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3d\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_20_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_20_consts_0._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_20_consts_2.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_acquire_lock._ascii.ob_base,
            & const_str__module_locks._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str__thread._ascii.ob_base,
            & const_str__DummyModuleLock._ascii.ob_base,
            & const_str__ModuleLock._ascii.ob_base,
            & const_str__weakref._ascii.ob_base,
            & const_str_ref._ascii.ob_base,
            & const_str_release_lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[157];
    }
importlib__bootstrap_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 156,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x05\x09\xd7\x04\x15\xd1\x04\x15\xd4\x04\x17\xf0\x02\x19\x05\x1c\xf0\x02\x03\x09\x18\xdc\x13\x20\xa0\x14\xd1\x13\x26\xd3\x13\x28\x88\x44\xf0\x08\x00\x0c\x10\x88\x3c\xdc\x0f\x16\x88\x7f\xdc\x17\x27\xa8\x04\xd3\x17\x2d\x91\x04\xe4\x17\x22\xa0\x34\xd3\x17\x28\x90\x04\xe0\x1d\x21\xf3\x00\x09\x0d\x28\xf4\x16\x00\x23\x2b\xa7\x2c\xa1\x2c\xa8\x74\xb0\x52\xd3\x22\x38\x8c\x4d\x98\x24\xd1\x0c\x1f\xe4\x08\x0c\xd7\x08\x19\xd1\x08\x19\xd4\x08\x1b\xe0\x0b\x0f\x80\x4b\xf8\xf4\x31\x00\x10\x18\xf2\x00\x01\x09\x18\xd8\x13\x17\x8a\x44\xf0\x03\x01\x09\x18\xfb\xf4\x2c\x00\x09\x0d\xd7\x08\x19\xd1\x08\x19\xd5\x08\x1b\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[42];
    }
importlib__bootstrap_toplevel_consts_20_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 41,
    },
    .ob_shash = -1,
    .ob_sval = "\x97\x0d\x41\x3b\x00\xa4\x41\x01\x42\x0c\x00\xc1\x3b\x0b\x42\x09\x03\xc2\x06\x02\x42\x0c\x00\xc2\x08\x01\x42\x09\x03\xc2\x09\x03\x42\x0c\x00\xc2\x0c\x16\x42\x22\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_20_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_lock._ascii.ob_base,
            & const_str_cb._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(330)
importlib__bootstrap_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 165,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_20_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_20_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 426,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 37,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_20_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__get_module_lock._ascii.ob_base,
    .co_qualname = & const_str__get_module_lock._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x09\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x19\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x80\x3f\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x80\x0c\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x6e\x0b\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x66\x01\x64\x02\x84\x01\x7d\x02\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x3c\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x01\x7d\x01\x59\x00\x8c\x64\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[190];
    }
importlib__bootstrap_toplevel_consts_21_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 189,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x63\x71\x75\x69\x72\x65\x73\x20\x74\x68\x65\x6e\x20\x72\x65\x6c\x65\x61\x73\x65\x73\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6c\x6f\x63\x6b\x20\x66\x6f\x72\x20\x61\x20\x67\x69\x76\x65\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x69\x73\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x65\x6e\x73\x75\x72\x65\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x63\x6f\x6d\x70\x6c\x65\x74\x65\x6c\x79\x20\x69\x6e\x69\x74\x69\x61\x6c\x69\x7a\x65\x64\x2c\x20\x69\x6e\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x65\x76\x65\x6e\x74\x20\x69\x74\x20\x69\x73\x20\x62\x65\x69\x6e\x67\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x20\x62\x79\x20\x61\x6e\x6f\x74\x68\x65\x72\x20\x74\x68\x72\x65\x61\x64\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_21_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_21_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__get_module_lock._ascii.ob_base,
            & const_str_acquire._ascii.ob_base,
            &_Py_ID(release),
            & const_str__DeadlockError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[62];
    }
importlib__bootstrap_toplevel_consts_21_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 61,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x0c\x1c\x98\x44\xd3\x0b\x21\x80\x44\xf0\x02\x07\x05\x17\xd8\x08\x0c\x8f\x0c\x89\x0c\x8c\x0e\xf0\x0c\x00\x09\x0d\x8f\x0c\x89\x0c\x8d\x0e\xf8\xf4\x0b\x00\x0c\x1a\xf2\x00\x03\x05\x0d\xf1\x06\x00\x09\x0d\xf0\x07\x03\x05\x0d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_toplevel_consts_21_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x8d\x10\x2e\x00\xae\x09\x3a\x03\xb9\x01\x3a\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_21_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_lock._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(122)
importlib__bootstrap_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 61,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_21_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 463,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 38,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_21_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(_lock_unlock_module),
    .co_qualname = &_Py_ID(_lock_unlock_module),
    .co_linetable = & importlib__bootstrap_toplevel_consts_21_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x7c\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[303];
    }
importlib__bootstrap_toplevel_consts_22_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 302,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x72\x65\x6d\x6f\x76\x65\x5f\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x5f\x66\x72\x61\x6d\x65\x73\x20\x69\x6e\x20\x69\x6d\x70\x6f\x72\x74\x2e\x63\x20\x77\x69\x6c\x6c\x20\x61\x6c\x77\x61\x79\x73\x20\x72\x65\x6d\x6f\x76\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x73\x0a\x20\x20\x20\x20\x6f\x66\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x20\x66\x72\x61\x6d\x65\x73\x20\x74\x68\x61\x74\x20\x65\x6e\x64\x20\x77\x69\x74\x68\x20\x61\x20\x63\x61\x6c\x6c\x20\x74\x6f\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x0a\x0a\x20\x20\x20\x20\x55\x73\x65\x20\x69\x74\x20\x69\x6e\x73\x74\x65\x61\x64\x20\x6f\x66\x20\x61\x20\x6e\x6f\x72\x6d\x61\x6c\x20\x63\x61\x6c\x6c\x20\x69\x6e\x20\x70\x6c\x61\x63\x65\x73\x20\x77\x68\x65\x72\x65\x20\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x0a\x20\x20\x20\x20\x66\x72\x61\x6d\x65\x73\x20\x69\x6e\x74\x72\x6f\x64\x75\x63\x65\x73\x20\x75\x6e\x77\x61\x6e\x74\x65\x64\x20\x6e\x6f\x69\x73\x65\x20\x69\x6e\x74\x6f\x20\x74\x68\x65\x20\x74\x72\x61\x63\x65\x62\x61\x63\x6b\x20\x28\x65\x2e\x67\x2e\x20\x77\x68\x65\x6e\x20\x65\x78\x65\x63\x75\x74\x69\x6e\x67\x0a\x20\x20\x20\x20\x6d\x6f\x64\x75\x6c\x65\x20\x63\x6f\x64\x65\x29\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_22_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_22_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
const_str__call_with_frames_removed = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_call_with_frames_removed",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib__bootstrap_toplevel_consts_22_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf1\x10\x00\x0c\x0d\x88\x64\xd0\x0b\x1b\x90\x64\xd1\x0b\x1b\xd0\x04\x1b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_kwds = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "kwds",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_22_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[102],
            &_Py_ID(args),
            & const_str_kwds._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(18)
importlib__bootstrap_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 9,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_22_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 15,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 480,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 39,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_22_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__call_with_frames_removed._ascii.ob_base,
    .co_qualname = & const_str__call_with_frames_removed._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x7c\x00\x7c\x01\x69\x00\x7c\x02\xa4\x01\x8e\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_verbosity = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "verbosity",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_24 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_verbosity._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[62];
    }
importlib__bootstrap_toplevel_consts_25_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 61,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Print the message to stderr if -v/PYTHONVERBOSE is turned on.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
importlib__bootstrap_toplevel_consts_25_consts_1_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "import ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_25_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[35],
            & importlib__bootstrap_toplevel_consts_25_consts_1_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_toplevel_consts_25_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "# ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_25_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(file),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_25_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_25_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_25_consts_1._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_25_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_25_consts_3._object.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_verbose = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "verbose",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_startswith = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "startswith",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_print = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "print",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_25_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_verbose._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_print._ascii.ob_base,
            &_Py_ID(format),
            &_Py_ID(stderr),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__verbose_message = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_verbose_message",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[75];
    }
importlib__bootstrap_toplevel_consts_25_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 74,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x07\x0a\x87\x79\x81\x79\xd7\x07\x18\xd1\x07\x18\x98\x49\xd2\x07\x25\xd8\x0f\x16\xd7\x0f\x21\xd1\x0f\x21\xd0\x22\x32\xd4\x0f\x33\xd8\x16\x1a\x98\x57\x91\x6e\x88\x47\xdc\x08\x0d\x88\x6e\x88\x67\x8f\x6e\x89\x6e\x98\x64\xd0\x0e\x23\xac\x23\xaf\x2a\xa9\x2a\xd6\x08\x35\xf0\x07\x00\x08\x26",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_25_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(message),
            & const_str_verbosity._ascii.ob_base,
            &_Py_ID(args),
        },
    },
};
static
    struct _PyCode_DEF(188)
importlib__bootstrap_toplevel_consts_25 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 94,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_25_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 491,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 40,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_25_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__verbose_message._ascii.ob_base,
    .co_qualname = & const_str__verbose_message._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_25_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6b\x5c\x00\x00\x72\x3f\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x73\x05\x64\x02\x7c\x00\x7a\x00\x00\x00\x7d\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x8e\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x04\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[50];
    }
importlib__bootstrap_toplevel_consts_26_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 49,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Decorator to verify the named module is built-in.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_26_consts_1_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " is not a built-in module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_26_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_26_consts_1_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_builtin_module_names = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "builtin_module_names",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_ImportError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ImportError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_26_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_builtin_module_names._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
const_str__requires_builtin_wrapper = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_requires_builtin_wrapper",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[53];
    }
importlib__bootstrap_toplevel_consts_26_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 52,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_requires_builtin.<locals>._requires_builtin_wrapper",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[57];
    }
importlib__bootstrap_toplevel_consts_26_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 56,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x0b\x13\x9c\x33\xd7\x1b\x33\xd1\x1b\x33\xd1\x0b\x33\xdc\x12\x1d\xa0\x18\xa0\x0c\xd0\x2c\x45\xd0\x1e\x46\xd8\x23\x2b\xf4\x03\x01\x13\x2d\xf0\x00\x01\x0d\x2d\xe1\x0f\x12\x90\x34\x98\x18\xd3\x0f\x22\xd0\x08\x22",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_fullname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fullname",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_fxn = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fxn",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_26_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            & const_str_fxn._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(90)
importlib__bootstrap_toplevel_consts_26_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 45,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_26_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_26_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 501,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 41,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_26_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__requires_builtin_wrapper._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_26_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_26_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x7c\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x10\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x9b\x02\x64\x01\x9d\x02\x7c\x01\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x02\x00\x89\x02\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_26_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_26_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_26_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_26_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__wrap._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__requires_builtin = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_requires_builtin",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
importlib__bootstrap_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf4\x04\x04\x05\x23\xf4\x0a\x00\x05\x0a\xd0\x0a\x23\xa0\x53\xd4\x04\x29\xd8\x0b\x24\xd0\x04\x24",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_26_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_fxn._ascii.ob_base,
            & const_str__requires_builtin_wrapper._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
importlib__bootstrap_toplevel_consts_26_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "` ",
};
static
    struct _PyCode_DEF(42)
importlib__bootstrap_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 21,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_26_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 499,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 42,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_26_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__requires_builtin._ascii.ob_base,
    .co_qualname = & const_str__requires_builtin._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x88\x00\x66\x01\x64\x01\x84\x08\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x89\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[48];
    }
importlib__bootstrap_toplevel_consts_27_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 47,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Decorator to verify the named module is frozen.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_toplevel_consts_27_consts_1_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " is not a frozen module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_27_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_27_consts_1_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_is_frozen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_frozen",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_27_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_is_frozen._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str__requires_frozen_wrapper = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_requires_frozen_wrapper",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[51];
    }
importlib__bootstrap_toplevel_consts_27_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 50,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_requires_frozen.<locals>._requires_frozen_wrapper",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[55];
    }
importlib__bootstrap_toplevel_consts_27_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 54,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x0f\x13\x8f\x7e\x89\x7e\x98\x68\xd4\x0f\x27\xdc\x12\x1d\xa0\x18\xa0\x0c\xd0\x2c\x43\xd0\x1e\x44\xd8\x23\x2b\xf4\x03\x01\x13\x2d\xf0\x00\x01\x0d\x2d\xe1\x0f\x12\x90\x34\x98\x18\xd3\x0f\x22\xd0\x08\x22",
};
static
    struct _PyCode_DEF(96)
importlib__bootstrap_toplevel_consts_27_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 48,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_27_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_27_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 512,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 43,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_26_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__requires_frozen_wrapper._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_27_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_27_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x73\x10\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x9b\x02\x64\x01\x9d\x02\x7c\x01\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x02\x00\x89\x02\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_27_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_27_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_27_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__requires_frozen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_requires_frozen",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
importlib__bootstrap_toplevel_consts_27_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf4\x04\x04\x05\x23\xf4\x0a\x00\x05\x0a\xd0\x0a\x22\xa0\x43\xd4\x04\x28\xd8\x0b\x23\xd0\x04\x23",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_27_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_fxn._ascii.ob_base,
            & const_str__requires_frozen_wrapper._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(42)
importlib__bootstrap_toplevel_consts_27 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 21,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_27_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 510,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 44,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_27_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__requires_frozen._ascii.ob_base,
    .co_qualname = & const_str__requires_frozen._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_27_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x88\x00\x66\x01\x64\x01\x84\x08\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x89\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[131];
    }
importlib__bootstrap_toplevel_consts_28_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 130,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4c\x6f\x61\x64\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x6e\x74\x6f\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x69\x74\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x69\x73\x20\x64\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2e\x20\x20\x55\x73\x65\x20\x6c\x6f\x61\x64\x65\x72\x2e\x65\x78\x65\x63\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x29\x20\x69\x6e\x73\x74\x65\x61\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[104];
    }
importlib__bootstrap_toplevel_consts_28_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 103,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "the load_module() method is deprecated and slated for removal in Python 3.15; use exec_module() instead",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_28_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_28_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_28_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__warnings = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_warnings",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_warn = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "warn",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_DeprecationWarning = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "DeprecationWarning",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_spec_from_loader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spec_from_loader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__exec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_exec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__load = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_load",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_28_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_DeprecationWarning._ascii.ob_base,
            & const_str_spec_from_loader._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str__exec._ascii.ob_base,
            & const_str__load._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__load_module_shim = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_load_module_shim",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[98];
    }
importlib__bootstrap_toplevel_consts_28_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 97,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0c\x01\x0c\x34\x80\x43\xe4\x04\x0d\x87\x4e\x81\x4e\x90\x33\xd4\x18\x2a\xd4\x04\x2b\xdc\x0b\x1b\x98\x48\xa0\x64\xd3\x0b\x2b\x80\x44\xd8\x07\x0f\x94\x33\x97\x3b\x91\x3b\xd1\x07\x1e\xdc\x11\x14\x97\x1b\x91\x1b\x98\x58\xd1\x11\x26\x88\x06\xdc\x08\x0d\x88\x64\x90\x46\xd4\x08\x1b\xdc\x0f\x12\x8f\x7b\x89\x7b\x98\x38\xd1\x0f\x24\xd0\x08\x24\xe4\x0f\x14\x90\x54\x8b\x7b\xd0\x08\x1a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spec",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_28_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(msg),
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
        },
    },
};
static
    struct _PyCode_DEF(240)
importlib__bootstrap_toplevel_consts_28 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 120,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_28_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_28_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 522,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 45,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_28_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__load_module_shim._ascii.ob_base,
    .co_qualname = & const_str__load_module_shim._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_28_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x01\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x32\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x04\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x53\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
importlib__bootstrap_toplevel_consts_29_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "The implementation of ModuleType.__repr__().",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
importlib__bootstrap_toplevel_consts_29_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<module ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
importlib__bootstrap_toplevel_consts_29_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " from ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_toplevel_consts_29_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " (",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_toplevel_consts_29_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ")>",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_toplevel_consts_29_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_29_consts_0._ascii.ob_base,
            &_Py_ID(__loader__),
            Py_None,
            &_Py_ID(__spec__),
            (PyObject *)&_Py_SINGLETON(strings).ascii[63],
            & importlib__bootstrap_toplevel_consts_29_consts_5._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_29_consts_6._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[62],
            & importlib__bootstrap_toplevel_consts_29_consts_8._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_29_consts_9._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str__module_repr_from_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_module_repr_from_spec",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_29_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(getattr),
            & const_str__module_repr_from_spec._ascii.ob_base,
            &_Py_ID(__name__),
            & const_str_AttributeError._ascii.ob_base,
            &_Py_ID(__file__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__module_repr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_module_repr",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[178];
    }
importlib__bootstrap_toplevel_consts_29_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 177,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0d\x14\x90\x56\x98\x5c\xa8\x34\xd3\x0d\x30\x80\x46\xdc\x0f\x16\x90\x76\x98\x7a\xa8\x34\xd3\x0f\x30\xd0\x07\x30\x80\x74\xd0\x07\x30\xdc\x0f\x25\xa0\x64\xd3\x0f\x2b\xd0\x08\x2b\xf0\x04\x03\x05\x13\xd8\x0f\x15\x8f\x7f\x89\x7f\x88\x04\xf0\x06\x08\x05\x36\xd8\x13\x19\x97\x3f\x91\x3f\x88\x08\xf0\x0e\x00\x12\x1a\x98\x24\x98\x18\xa0\x16\xa8\x08\xa0\x7c\xb0\x31\xd0\x0f\x35\xd0\x08\x35\xf8\xf4\x15\x00\x0c\x1a\xf2\x00\x01\x05\x13\xd8\x0f\x12\x8a\x04\xf0\x03\x01\x05\x13\xfb\xf4\x08\x00\x0c\x1a\xf2\x00\x04\x05\x35\xd8\x0b\x11\x88\x3e\xd8\x15\x1d\x98\x64\x98\x58\xa0\x51\xd0\x13\x27\xd2\x0c\x27\xe0\x15\x1d\x98\x64\x98\x58\xa0\x52\xa8\x06\xa0\x7a\xb0\x12\xd0\x13\x34\xd2\x0c\x34\xf0\x09\x04\x05\x35\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[41];
    }
importlib__bootstrap_toplevel_consts_29_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 40,
    },
    .ob_shash = -1,
    .ob_sval = "\xa9\x0c\x41\x0b\x00\xb6\x0c\x41\x1c\x00\xc1\x0b\x0b\x41\x19\x03\xc1\x18\x01\x41\x19\x03\xc1\x1c\x11\x41\x3b\x03\xc1\x2f\x09\x41\x3b\x03\xc1\x3a\x01\x41\x3b\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_loader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_29_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(module),
            & const_str_loader._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(filename),
        },
    },
};
static
    struct _PyCode_DEF(252)
importlib__bootstrap_toplevel_consts_29 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 126,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_29_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_29_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_29_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 541,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 46,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_29_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__module_repr._ascii.ob_base,
    .co_qualname = & const_str__module_repr._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_29_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\x64\x02\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x03\x64\x02\xab\x03\x00\x00\x00\x00\x00\x00\x78\x01\x7d\x02\x72\x0b\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x09\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x09\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x64\x05\x7c\x03\x9b\x02\x64\x06\x7c\x04\x9b\x02\x64\x07\x9d\x05\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x04\x7d\x03\x59\x00\x8c\x23\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x16\x01\x00\x7c\x01\x80\x08\x64\x05\x7c\x03\x9b\x02\x64\x07\x9d\x03\x63\x02\x59\x00\x53\x00\x64\x05\x7c\x03\x9b\x02\x64\x08\x7c\x01\x9b\x02\x64\x09\x9d\x05\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_ModuleSpec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleSpec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1489];
    }
importlib__bootstrap_toplevel_consts_30_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1488,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x63\x61\x74\x69\x6f\x6e\x20\x66\x6f\x72\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x2c\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x6c\x6f\x61\x64\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x41\x20\x6d\x6f\x64\x75\x6c\x65\x27\x73\x20\x73\x70\x65\x63\x20\x69\x73\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x66\x6f\x72\x20\x69\x6e\x66\x6f\x72\x6d\x61\x74\x69\x6f\x6e\x20\x61\x62\x6f\x75\x74\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x20\x20\x46\x6f\x72\x0a\x20\x20\x20\x20\x64\x61\x74\x61\x20\x61\x73\x73\x6f\x63\x69\x61\x74\x65\x64\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2c\x20\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x73\x6f\x75\x72\x63\x65\x2c\x20\x75\x73\x65\x20\x74\x68\x65\x20\x73\x70\x65\x63\x27\x73\x0a\x20\x20\x20\x20\x6c\x6f\x61\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x60\x6e\x61\x6d\x65\x60\x20\x69\x73\x20\x74\x68\x65\x20\x61\x62\x73\x6f\x6c\x75\x74\x65\x20\x6e\x61\x6d\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x20\x20\x60\x6c\x6f\x61\x64\x65\x72\x60\x20\x69\x73\x20\x74\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x0a\x20\x20\x20\x20\x74\x6f\x20\x75\x73\x65\x20\x77\x68\x65\x6e\x20\x6c\x6f\x61\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x20\x20\x60\x70\x61\x72\x65\x6e\x74\x60\x20\x69\x73\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x20\x6f\x66\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x69\x6e\x2e\x20\x20\x54\x68\x65\x20\x70\x61\x72\x65\x6e\x74\x20\x69\x73\x20\x64\x65\x72\x69\x76\x65\x64\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x60\x69\x73\x5f\x70\x61\x63\x6b\x61\x67\x65\x60\x20\x64\x65\x74\x65\x72\x6d\x69\x6e\x65\x73\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x63\x6f\x6e\x73\x69\x64\x65\x72\x65\x64\x20\x61\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x6f\x72\x0a\x20\x20\x20\x20\x6e\x6f\x74\x2e\x20\x20\x4f\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x74\x68\x69\x73\x20\x69\x73\x20\x72\x65\x66\x6c\x65\x63\x74\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x60\x5f\x5f\x70\x61\x74\x68\x5f\x5f\x60\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x2e\x0a\x0a\x20\x20\x20\x20\x60\x6f\x72\x69\x67\x69\x6e\x60\x20\x69\x73\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x63\x20\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x20\x75\x73\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x20\x66\x72\x6f\x6d\x20\x77\x68\x69\x63\x68\x20\x74\x6f\x0a\x20\x20\x20\x20\x6c\x6f\x61\x64\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2c\x20\x69\x66\x20\x74\x68\x61\x74\x20\x69\x6e\x66\x6f\x72\x6d\x61\x74\x69\x6f\x6e\x20\x69\x73\x20\x61\x76\x61\x69\x6c\x61\x62\x6c\x65\x2e\x20\x20\x57\x68\x65\x6e\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x69\x73\x0a\x20\x20\x20\x20\x73\x65\x74\x2c\x20\x6f\x72\x69\x67\x69\x6e\x20\x77\x69\x6c\x6c\x20\x6d\x61\x74\x63\x68\x2e\x0a\x0a\x20\x20\x20\x20\x60\x68\x61\x73\x5f\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x60\x20\x69\x6e\x64\x69\x63\x61\x74\x65\x73\x20\x74\x68\x61\x74\x20\x61\x20\x73\x70\x65\x63\x27\x73\x20\x22\x6f\x72\x69\x67\x69\x6e\x22\x20\x72\x65\x66\x6c\x65\x63\x74\x73\x20\x61\x20\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x57\x68\x65\x6e\x20\x74\x68\x69\x73\x20\x69\x73\x20\x54\x72\x75\x65\x2c\x20\x60\x5f\x5f\x66\x69\x6c\x65\x5f\x5f\x60\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x73\x65\x74\x2e\x0a\x0a\x20\x20\x20\x20\x60\x63\x61\x63\x68\x65\x64\x60\x20\x69\x73\x20\x74\x68\x65\x20\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x74\x68\x65\x20\x63\x61\x63\x68\x65\x64\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x2c\x20\x69\x66\x20\x61\x6e\x79\x2e\x20\x20\x49\x74\x0a\x20\x20\x20\x20\x63\x6f\x72\x72\x65\x73\x70\x6f\x6e\x64\x73\x20\x74\x6f\x20\x74\x68\x65\x20\x60\x5f\x5f\x63\x61\x63\x68\x65\x64\x5f\x5f\x60\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x2e\x0a\x0a\x20\x20\x20\x20\x60\x73\x75\x62\x6d\x6f\x64\x75\x6c\x65\x5f\x73\x65\x61\x72\x63\x68\x5f\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x73\x60\x20\x69\x73\x20\x74\x68\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x20\x6f\x66\x20\x70\x61\x74\x68\x20\x65\x6e\x74\x72\x69\x65\x73\x20\x74\x6f\x0a\x20\x20\x20\x20\x73\x65\x61\x72\x63\x68\x20\x77\x68\x65\x6e\x20\x69\x6d\x70\x6f\x72\x74\x69\x6e\x67\x20\x73\x75\x62\x6d\x6f\x64\x75\x6c\x65\x73\x2e\x20\x20\x49\x66\x20\x73\x65\x74\x2c\x20\x69\x73\x5f\x70\x61\x63\x6b\x61\x67\x65\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x0a\x20\x20\x20\x20\x54\x72\x75\x65\x2d\x2d\x61\x6e\x64\x20\x46\x61\x6c\x73\x65\x20\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x2e\x0a\x0a\x20\x20\x20\x20\x50\x61\x63\x6b\x61\x67\x65\x73\x20\x61\x72\x65\x20\x73\x69\x6d\x70\x6c\x79\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x74\x68\x61\x74\x20\x28\x6d\x61\x79\x29\x20\x68\x61\x76\x65\x20\x73\x75\x62\x6d\x6f\x64\x75\x6c\x65\x73\x2e\x20\x20\x49\x66\x20\x61\x20\x73\x70\x65\x63\x0a\x20\x20\x20\x20\x68\x61\x73\x20\x61\x20\x6e\x6f\x6e\x2d\x4e\x6f\x6e\x65\x20\x76\x61\x6c\x75\x65\x20\x69\x6e\x20\x60\x73\x75\x62\x6d\x6f\x64\x75\x6c\x65\x5f\x73\x65\x61\x72\x63\x68\x5f\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x73\x60\x2c\x20\x74\x68\x65\x20\x69\x6d\x70\x6f\x72\x74\x0a\x20\x20\x20\x20\x73\x79\x73\x74\x65\x6d\x20\x77\x69\x6c\x6c\x20\x63\x6f\x6e\x73\x69\x64\x65\x72\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x6c\x6f\x61\x64\x65\x64\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x73\x70\x65\x63\x20\x61\x73\x20\x70\x61\x63\x6b\x61\x67\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x4f\x6e\x6c\x79\x20\x66\x69\x6e\x64\x65\x72\x73\x20\x28\x73\x65\x65\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x2e\x61\x62\x63\x2e\x4d\x65\x74\x61\x50\x61\x74\x68\x46\x69\x6e\x64\x65\x72\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x2e\x61\x62\x63\x2e\x50\x61\x74\x68\x45\x6e\x74\x72\x79\x46\x69\x6e\x64\x65\x72\x29\x20\x73\x68\x6f\x75\x6c\x64\x20\x6d\x6f\x64\x69\x66\x79\x20\x4d\x6f\x64\x75\x6c\x65\x53\x70\x65\x63\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_loader_state = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader_state",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_is_package = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_package",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(origin),
            & const_str_loader_state._ascii.ob_base,
            & const_str_is_package._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str_submodule_search_locations = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "submodule_search_locations",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__set_fileattr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_set_fileattr",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__cached = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_cached",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_loader._ascii.ob_base,
            &_Py_ID(origin),
            & const_str_loader_state._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            &_Py_ID(_uninitialized_submodules),
            & const_str__set_fileattr._ascii.ob_base,
            & const_str__cached._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_toplevel_consts_30_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleSpec.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[69];
    }
importlib__bootstrap_toplevel_consts_30_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 68,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x14\x18\x88\x04\x8c\x09\xd8\x16\x1c\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8c\x0b\xd8\x1c\x28\x88\x04\xd4\x08\x19\xd9\x30\x3a\xa9\x22\xc0\x04\x88\x04\xd4\x08\x27\xd8\x29\x2b\x88\x04\xd4\x08\x26\xf0\x06\x00\x1e\x23\x88\x04\xd4\x08\x1a\xd8\x17\x1b\x88\x04\x8d\x0c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
            & const_str_loader._ascii.ob_base,
            &_Py_ID(origin),
            & const_str_loader_state._ascii.ob_base,
            & const_str_is_package._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[7];
    }
importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 6,
    },
    .ob_shash = -1,
    .ob_sval = "      ",
};
static
    struct _PyCode_DEF(124)
importlib__bootstrap_toplevel_consts_30_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 62,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_30_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 3,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 599,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 47,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x72\x02\x67\x00\x6e\x01\x64\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x7c\x00\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x7c\x00\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
importlib__bootstrap_toplevel_consts_30_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "name=",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
importlib__bootstrap_toplevel_consts_30_consts_5_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader=",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
importlib__bootstrap_toplevel_consts_30_consts_5_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "origin=",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_30_consts_5_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "submodule_search_locations=",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_toplevel_consts_30_consts_5_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ", ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_30_consts_5_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_5_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_5_consts_3._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_5_consts_4._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[40],
            & importlib__bootstrap_toplevel_consts_30_consts_5_consts_6._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[41],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_loader._ascii.ob_base,
            &_Py_ID(origin),
            &_Py_ID(append),
            & const_str_submodule_search_locations._ascii.ob_base,
            &_Py_ID(__class__),
            &_Py_ID(__name__),
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_toplevel_consts_30_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleSpec.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[146];
    }
importlib__bootstrap_toplevel_consts_30_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 145,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x12\x17\x98\x04\x9f\x09\x99\x09\x90\x7d\xd0\x10\x25\xa8\x17\xb0\x14\xb7\x1b\xb1\x1b\xb0\x0f\xd0\x27\x40\xd0\x0f\x41\x88\x04\xd8\x0b\x0f\x8f\x3b\x89\x3b\xd0\x0b\x22\xd8\x0c\x10\x8f\x4b\x89\x4b\x98\x27\xa0\x24\xa7\x2b\xa1\x2b\xa0\x1f\xd0\x18\x31\xd4\x0c\x32\xd8\x0b\x0f\xd7\x0b\x2a\xd1\x0b\x2a\xd0\x0b\x36\xd8\x0c\x10\x8f\x4b\x89\x4b\xd0\x1a\x35\xb0\x64\xd7\x36\x55\xd1\x36\x55\xd0\x35\x56\xd0\x18\x57\xd4\x0c\x58\xd8\x12\x16\x97\x2e\x91\x2e\xd7\x12\x29\xd1\x12\x29\xd0\x11\x2a\xa8\x21\xa8\x44\xaf\x49\xa9\x49\xb0\x64\xab\x4f\xd0\x2b\x3c\xb8\x41\xd0\x0f\x3e\xd0\x08\x3e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(args),
        },
    },
};
static
    struct _PyCode_DEF(316)
importlib__bootstrap_toplevel_consts_30_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 158,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_30_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 612,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 48,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x02\x64\x02\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x02\x67\x02\x7d\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x1e\x7c\x01\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x1e\x7c\x01\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x05\x64\x06\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x07\x9d\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_cached = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cached",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_has_location = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "has_location",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_NotImplemented = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NotImplemented",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_submodule_search_locations._ascii.ob_base,
            &_Py_ID(name),
            & const_str_loader._ascii.ob_base,
            &_Py_ID(origin),
            & const_str_cached._ascii.ob_base,
            & const_str_has_location._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_toplevel_consts_30_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleSpec.__eq__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[176];
    }
importlib__bootstrap_toplevel_consts_30_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 175,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\xd7\x0f\x2e\xd1\x0f\x2e\x88\x04\xf0\x02\x08\x09\x22\xd8\x14\x18\x97\x49\x91\x49\xa0\x15\xa7\x1a\xa1\x1a\xd1\x14\x2b\xf2\x00\x05\x15\x3c\xd8\x14\x18\x97\x4b\x91\x4b\xa0\x35\xa7\x3c\xa1\x3c\xd1\x14\x2f\xf2\x03\x05\x15\x3c\xe0\x14\x18\x97\x4b\x91\x4b\xa0\x35\xa7\x3c\xa1\x3c\xd1\x14\x2f\xf2\x05\x05\x15\x3c\xf0\x06\x00\x15\x19\x98\x45\xd7\x1c\x3c\xd1\x1c\x3c\xd1\x14\x3c\xf2\x07\x05\x15\x3c\xf0\x08\x00\x15\x19\x97\x4b\x91\x4b\xa0\x35\xa7\x3c\xa1\x3c\xd1\x14\x2f\xf2\x09\x05\x15\x3c\xf0\x0a\x00\x15\x19\xd7\x14\x25\xd1\x14\x25\xa8\x15\xd7\x29\x3b\xd1\x29\x3b\xd1\x14\x3b\xf0\x0b\x05\x0d\x3d\xf8\xf4\x0c\x00\x10\x1e\xf2\x00\x01\x09\x22\xdc\x13\x21\xd2\x0c\x21\xf0\x03\x01\x09\x22\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_toplevel_consts_30_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x8e\x42\x15\x42\x24\x00\xc2\x24\x0f\x42\x36\x03\xc2\x35\x01\x42\x36\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_other = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "other",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_smsl = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "smsl",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_other._ascii.ob_base,
            & const_str_smsl._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(370)
importlib__bootstrap_toplevel_consts_30_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 185,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_30_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 620,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 49,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__eq__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x09\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x7b\x01\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x60\x01\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x45\x01\x00\x7c\x02\x7c\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x34\x01\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x19\x01\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x09\x01\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__bootstrap_external = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_bootstrap_external",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_NotImplementedError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NotImplementedError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__get_cached = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_cached",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__cached._ascii.ob_base,
            &_Py_ID(origin),
            & const_str__set_fileattr._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            & const_str__get_cached._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_toplevel_consts_30_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleSpec.cached",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[75];
    }
importlib__bootstrap_toplevel_consts_30_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 74,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0f\x8f\x3c\x89\x3c\xd0\x0b\x1f\xd8\x0f\x13\x8f\x7b\x89\x7b\xd0\x0f\x26\xa8\x34\xd7\x2b\x3d\xd2\x2b\x3d\xdc\x13\x26\xd0\x13\x2e\xdc\x1a\x2d\xd0\x14\x2d\xdc\x1f\x32\xd7\x1f\x3e\xd1\x1f\x3e\xb8\x74\xbf\x7b\xb9\x7b\xd3\x1f\x4b\x90\x04\x94\x0c\xd8\x0f\x13\x8f\x7c\x89\x7c\xd0\x08\x1b",
};
static
    struct _PyCode_DEF(194)
importlib__bootstrap_toplevel_consts_30_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 97,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 632,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 50,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_cached._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x48\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x3c\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x30\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x80\x06\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__cached._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
importlib__bootstrap_toplevel_consts_30_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x17\x1d\x88\x04\x8d\x0c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_cached._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(18)
importlib__bootstrap_toplevel_consts_30_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 9,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 641,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 51,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_cached._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib__bootstrap_toplevel_consts_30_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "The name of the module's parent.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_30_consts_9_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_rpartition = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rpartition",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_submodule_search_locations._ascii.ob_base,
            &_Py_ID(name),
            & const_str_rpartition._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_toplevel_consts_30_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleSpec.parent",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[53];
    }
importlib__bootstrap_toplevel_consts_30_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 52,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x0c\x10\xd7\x0b\x2a\xd1\x0b\x2a\xd0\x0b\x32\xd8\x13\x17\x97\x39\x91\x39\xd7\x13\x27\xd1\x13\x27\xa8\x03\xd3\x13\x2c\xa8\x51\xd1\x13\x2f\xd0\x0c\x2f\xe0\x13\x17\x97\x39\x91\x39\xd0\x0c\x1c",
};
static
    struct _PyCode_DEF(110)
importlib__bootstrap_toplevel_consts_30_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_30_consts_9_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 645,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 52,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(parent),
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_9_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x1e\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\x53\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__set_fileattr._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_toplevel_consts_30_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleSpec.has_location",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
importlib__bootstrap_toplevel_consts_30_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\xd7\x0f\x21\xd1\x0f\x21\xd0\x08\x21",
};
static
    struct _PyCode_DEF(26)
importlib__bootstrap_toplevel_consts_30_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 653,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 53,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_has_location._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_10_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_bool = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bool",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_bool._ascii.ob_base,
            & const_str__set_fileattr._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
importlib__bootstrap_toplevel_consts_30_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x1d\x21\xa0\x25\x9b\x5b\x88\x04\xd5\x08\x1a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(36)
importlib__bootstrap_toplevel_consts_30_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 18,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 657,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 54,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_has_location._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_30_consts_10_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str_ModuleSpec._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_1._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_30_consts_3._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_4.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_5.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_6.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_7.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_8.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_9.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_10.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_11.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_property = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "property",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_setter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "setter",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_toplevel_consts_30_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            &_Py_ID(__repr__),
            &_Py_ID(__eq__),
            & const_str_property._ascii.ob_base,
            & const_str_cached._ascii.ob_base,
            & const_str_setter._ascii.ob_base,
            &_Py_ID(parent),
            & const_str_has_location._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[144];
    }
importlib__bootstrap_toplevel_consts_30_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 143,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x22\x05\x08\xf0\x48\x01\x00\x30\x34\xc0\x24\xd8\x1c\x20\xf4\x03\x0b\x05\x1c\xf2\x1a\x06\x05\x3f\xf2\x10\x0a\x05\x22\xf0\x18\x00\x06\x0e\xf1\x02\x06\x05\x1c\xf3\x03\x00\x06\x0e\xf0\x02\x06\x05\x1c\xf0\x10\x00\x06\x0c\x87\x5d\x81\x5d\xf1\x02\x01\x05\x1e\xf3\x03\x00\x06\x13\xf0\x02\x01\x05\x1e\xf0\x06\x00\x06\x0e\xf1\x02\x05\x05\x1d\xf3\x03\x00\x06\x0e\xf0\x02\x05\x05\x1d\xf0\x0e\x00\x06\x0e\xf1\x02\x01\x05\x22\xf3\x03\x00\x06\x0e\xf0\x02\x01\x05\x22\xf0\x06\x00\x06\x12\xd7\x05\x18\xd1\x05\x18\xf1\x02\x01\x05\x29\xf3\x03\x00\x06\x19\xf1\x02\x01\x05\x29",
};
static
    struct _PyCode_DEF(164)
importlib__bootstrap_toplevel_consts_30 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 82,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_30_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_30_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 562,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 55,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_ModuleSpec._ascii.ob_base,
    .co_qualname = & const_str_ModuleSpec._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_30_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x64\x02\x64\x02\x64\x03\x9c\x03\x64\x04\x84\x02\x5a\x04\x64\x05\x84\x00\x5a\x05\x64\x06\x84\x00\x5a\x06\x65\x07\x64\x07\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x65\x08\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x65\x07\x64\x09\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0a\x65\x07\x64\x0a\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0b\x65\x0b\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0b\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_32 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(origin),
            & const_str_is_package._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
importlib__bootstrap_toplevel_consts_33_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return a module spec based on various loader methods.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__ORIGIN = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ORIGIN",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_get_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_filename",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_33_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_33_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_loader._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_33_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_33_consts_0._ascii.ob_base,
            Py_None,
            & const_str__ORIGIN._ascii.ob_base,
            & const_str_get_filename._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_33_consts_4._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_33_consts_5._object.ob_base.ob_base,
            & const_str_is_package._ascii.ob_base,
            Py_False,
            & importlib__bootstrap_toplevel_consts_32._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str_spec_from_file_location = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spec_from_file_location",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_33_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(getattr),
            & const_str_hasattr._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            & const_str_spec_from_file_location._ascii.ob_base,
            & const_str_is_package._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_ModuleSpec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[193];
    }
importlib__bootstrap_toplevel_consts_33_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 192,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x07\x0d\x80\x7e\xdc\x11\x18\x98\x16\xa0\x19\xa8\x44\xd3\x11\x31\x88\x06\xe1\x0b\x11\x94\x67\x98\x66\xa0\x6e\xd4\x16\x35\xdc\x0b\x1e\xd0\x0b\x26\xdc\x12\x25\xd0\x0c\x25\xdc\x22\x35\xd7\x22\x4d\xd1\x22\x4d\xd0\x08\x1f\xe0\x0b\x15\xd0\x0b\x1d\xd9\x13\x2a\xa8\x34\xb8\x06\xd4\x13\x3f\xd0\x0c\x3f\xd9\x17\x21\x91\x12\xa0\x74\x88\x06\xd9\x0f\x26\xa0\x74\xb0\x46\xd8\x42\x48\xf4\x03\x01\x10\x4a\x01\xf0\x00\x01\x09\x4a\x01\xf0\x06\x00\x08\x12\xd0\x07\x19\xdc\x0b\x12\x90\x36\x98\x3c\xd4\x0b\x28\xf0\x02\x03\x0d\x22\xd8\x1d\x23\xd7\x1d\x2e\xd1\x1d\x2e\xa8\x74\xd3\x1d\x34\x91\x0a\xf0\x0a\x00\x1a\x1f\x88\x4a\xe4\x0b\x15\x90\x64\x98\x46\xa8\x36\xb8\x6a\xd4\x0b\x49\xd0\x04\x49\xf8\xf4\x0d\x00\x14\x1f\xf2\x00\x01\x0d\x22\xd8\x1d\x21\x92\x0a\xf0\x03\x01\x0d\x22\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_toplevel_consts_33_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x26\x11\x42\x09\x00\xc2\x09\x0b\x42\x17\x03\xc2\x16\x01\x42\x17\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_search = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "search",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_33_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_loader._ascii.ob_base,
            &_Py_ID(origin),
            & const_str_is_package._ascii.ob_base,
            & const_str_spec_from_file_location._ascii.ob_base,
            & const_str_search._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(308)
importlib__bootstrap_toplevel_consts_33 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 154,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_33_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_33_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_33_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 2,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 662,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 56,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_33_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_spec_from_loader._ascii.ob_base,
    .co_qualname = & const_str_spec_from_loader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_33_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x80\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x02\x64\x01\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x73\x45\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x72\x39\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x80\x06\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x03\x80\x0a\x02\x00\x7c\x04\x7c\x00\x7c\x01\xac\x04\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x03\x72\x02\x67\x00\x6e\x01\x64\x01\x7d\x05\x02\x00\x7c\x04\x7c\x00\x7c\x01\x7c\x05\xac\x05\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x03\x80\x21\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x72\x13\x09\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x6e\x02\x64\x07\x7d\x03\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7c\x03\xac\x08\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x01\x7d\x03\x59\x00\x8c\x1c\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_34_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(origin),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_34_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & const_str__ORIGIN._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_34_consts_2._object.ob_base.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str___cached__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__cached__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_list = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "list",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
importlib__bootstrap_toplevel_consts_34_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            &_Py_ID(__spec__),
            & const_str_AttributeError._ascii.ob_base,
            &_Py_ID(__name__),
            &_Py_ID(__loader__),
            &_Py_ID(__file__),
            &_Py_ID(getattr),
            & const_str___cached__._ascii.ob_base,
            & const_str_list._ascii.ob_base,
            &_Py_ID(__path__),
            & const_str_ModuleSpec._ascii.ob_base,
            & const_str__set_fileattr._ascii.ob_base,
            & const_str_cached._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__spec_from_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_spec_from_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[310];
    }
importlib__bootstrap_toplevel_consts_34_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 309,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x06\x05\x18\xd8\x0f\x15\x8f\x7f\x89\x7f\x88\x04\xf0\x08\x00\x0c\x10\xd0\x0b\x1b\xd8\x13\x17\x88\x4b\xe0\x0b\x11\x8f\x3f\x89\x3f\x80\x44\xd8\x07\x0d\x80\x7e\xf0\x02\x04\x09\x11\xd8\x15\x1b\xd7\x15\x26\xd1\x15\x26\x88\x46\xf0\x08\x03\x05\x18\xd8\x13\x19\x97\x3f\x91\x3f\x88\x08\xf0\x06\x00\x08\x0e\x80\x7e\xd8\x0b\x11\xd0\x0b\x1d\xdc\x15\x1c\x98\x56\xa0\x59\xb0\x04\xd3\x15\x35\x88\x46\xd9\x0f\x15\x98\x28\xd0\x1a\x2e\xd8\x15\x1d\x88\x46\xf0\x02\x03\x05\x16\xd8\x11\x17\xd7\x11\x22\xd1\x11\x22\x88\x06\xf0\x06\x03\x05\x2a\xdc\x25\x29\xa8\x26\xaf\x2f\xa9\x2f\xd3\x25\x3a\xd0\x08\x22\xf4\x08\x00\x0c\x16\x90\x64\x98\x46\xa8\x36\xd4\x0b\x32\x80\x44\xd8\x22\x2a\xd0\x22\x32\x99\x15\xb8\x16\xc0\x38\xd1\x39\x4b\x80\x44\xd4\x04\x16\xd8\x12\x18\x80\x44\x84\x4b\xd8\x26\x40\x80\x44\xd4\x04\x23\xd8\x0b\x0f\x80\x4b\xf8\xf4\x47\x01\x00\x0c\x1a\xf2\x00\x01\x05\x0d\xd9\x08\x0c\xf0\x03\x01\x05\x0d\xfb\xf4\x14\x00\x10\x1e\xf2\x00\x02\x09\x11\xe1\x0c\x10\xf0\x05\x02\x09\x11\xfb\xf4\x0a\x00\x0c\x1a\xf2\x00\x01\x05\x18\xd8\x13\x17\x8a\x08\xf0\x03\x01\x05\x18\xfb\xf4\x12\x00\x0c\x1a\xf2\x00\x01\x05\x16\xd8\x11\x15\x8a\x06\xf0\x03\x01\x05\x16\xfb\xf4\x08\x00\x0c\x1a\xf2\x00\x01\x05\x2a\xd8\x25\x29\xd2\x08\x22\xf0\x03\x01\x05\x2a\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[88];
    }
importlib__bootstrap_toplevel_consts_34_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 87,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x0c\x42\x20\x00\xa1\x0c\x42\x2f\x00\xae\x0c\x42\x3e\x00\xc1\x12\x0c\x43\x0f\x00\xc1\x1f\x15\x43\x20\x00\xc2\x20\x09\x42\x2c\x03\xc2\x2b\x01\x42\x2c\x03\xc2\x2f\x09\x42\x3b\x03\xc2\x3a\x01\x42\x3b\x03\xc2\x3e\x0b\x43\x0c\x03\xc3\x0b\x01\x43\x0c\x03\xc3\x0f\x0b\x43\x1d\x03\xc3\x1c\x01\x43\x1d\x03\xc3\x20\x0b\x43\x2e\x03\xc3\x2d\x01\x43\x2e\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_location = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "location",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_34_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(module),
            & const_str_loader._ascii.ob_base,
            &_Py_ID(origin),
            & const_str_spec._ascii.ob_base,
            &_Py_ID(name),
            & const_str_location._ascii.ob_base,
            & const_str_cached._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(482)
importlib__bootstrap_toplevel_consts_34 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 241,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_34_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_34_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_34_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 691,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 57,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_34_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__spec_from_module._ascii.ob_base,
    .co_qualname = & const_str__spec_from_module._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_34_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x81\x02\x7c\x03\x53\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x01\x80\x0d\x09\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x02\x80\x15\x7c\x01\x81\x0d\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x73\x04\x7c\x05\x81\x02\x7c\x05\x7d\x02\x09\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x09\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x01\x7c\x02\xac\x02\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x05\x80\x02\x64\x03\x6e\x04\x7c\x02\x7c\x05\x6b\x28\x00\x00\x7c\x03\x5f\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x03\x5f\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x7c\x03\x5f\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x53\x00\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x99\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x8d\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x00\x7d\x05\x59\x00\x8c\x91\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x00\x7d\x06\x59\x00\x8c\x7e\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x00\x7d\x07\x59\x00\x8c\x79\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_override = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "override",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_36 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_override._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_37_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__name__),
            &_Py_ID(__loader__),
            &_Py_ID(__package__),
            &_Py_ID(__path__),
            &_Py_ID(__file__),
            & const_str___cached__._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_NamespaceLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[21];
        }_object;
    }
importlib__bootstrap_toplevel_consts_37_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 21,
        },
        .ob_item = {
            &_Py_ID(getattr),
            &_Py_ID(name),
            &_Py_ID(__name__),
            & const_str_AttributeError._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            & const_str_NamespaceLoader._ascii.ob_base,
            &_Py_ID(__new__),
            & const_str__path._ascii.ob_base,
            &_Py_ID(__file__),
            &_Py_ID(__loader__),
            &_Py_ID(parent),
            &_Py_ID(__package__),
            &_Py_ID(__spec__),
            &_Py_ID(__path__),
            & const_str_has_location._ascii.ob_base,
            &_Py_ID(origin),
            & const_str_cached._ascii.ob_base,
            & const_str___cached__._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__init_module_attrs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_init_module_attrs",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[502];
    }
importlib__bootstrap_toplevel_consts_37_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 501,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf1\x08\x00\x09\x11\x94\x47\x98\x46\xa0\x4a\xb0\x04\xd3\x14\x35\xd0\x14\x3d\xf0\x02\x03\x09\x11\xd8\x1e\x22\x9f\x69\x99\x69\x88\x46\x8c\x4f\xf1\x08\x00\x08\x10\x94\x37\x98\x36\xa0\x3c\xb0\x14\xd3\x13\x36\xd0\x13\x3e\xd8\x11\x15\x97\x1b\x91\x1b\x88\x06\xd8\x0b\x11\x88\x3e\xe0\x0f\x13\xd7\x0f\x2e\xd1\x0f\x2e\xd0\x0f\x3a\xdc\x13\x26\xd0\x13\x2e\xdc\x1a\x2d\xd0\x14\x2d\xdc\x22\x35\xd7\x22\x45\xd1\x22\x45\x90\x0f\xe0\x19\x28\xd7\x19\x30\xd1\x19\x30\xb0\x1f\xd3\x19\x41\x90\x06\xd8\x1f\x23\xd7\x1f\x3e\xd1\x1f\x3e\x90\x06\x94\x0c\xd8\x1e\x24\x90\x04\x94\x0b\xf0\x16\x00\x23\x27\x90\x06\x94\x0f\xf0\x02\x03\x09\x11\xd8\x20\x26\x88\x46\xd4\x0c\x1d\xf1\x08\x00\x08\x10\x94\x37\x98\x36\xa0\x3d\xb0\x24\xd3\x13\x37\xd0\x13\x3f\xf0\x02\x03\x09\x11\xd8\x21\x25\xa7\x1b\xa1\x1b\x88\x46\xd4\x0c\x1e\xf0\x08\x03\x05\x0d\xd8\x1a\x1e\x88\x06\x8c\x0f\xf1\x08\x00\x08\x10\x94\x37\x98\x36\xa0\x3a\xa8\x74\xd3\x13\x34\xd0\x13\x3c\xd8\x0b\x0f\xd7\x0b\x2a\xd1\x0b\x2a\xd0\x0b\x36\xf0\x04\x03\x0d\x15\xd8\x22\x26\xd7\x22\x41\xd1\x22\x41\x90\x06\x94\x0f\xf0\x08\x00\x08\x0c\xd7\x07\x18\xd2\x07\x18\xd9\x0b\x13\x94\x77\x98\x76\xa0\x7a\xb0\x34\xd3\x17\x38\xd0\x17\x40\xf0\x02\x03\x0d\x15\xd8\x22\x26\xa7\x2b\xa1\x2b\x90\x06\x94\x0f\xf1\x08\x00\x0c\x14\x94\x77\x98\x76\xa0\x7c\xb0\x54\xd3\x17\x3a\xd0\x17\x42\xd8\x0f\x13\x8f\x7b\x89\x7b\xd0\x0f\x26\xf0\x02\x03\x11\x19\xd8\x28\x2c\xaf\x0b\xa9\x0b\x90\x46\xd4\x14\x25\xf0\x06\x00\x0c\x12\x80\x4d\x88\x36\x80\x4d\xf8\xf4\x7f\x01\x00\x10\x1e\xf2\x00\x01\x09\x11\xda\x0c\x10\xf0\x03\x01\x09\x11\xfb\xf4\x38\x00\x10\x1e\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfb\xf4\x0c\x00\x10\x1e\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfb\xf4\x0a\x00\x0c\x1a\xf2\x00\x01\x05\x0d\xd9\x08\x0c\xf0\x03\x01\x05\x0d\xfb\xf4\x10\x00\x14\x22\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfb\xf4\x0e\x00\x14\x22\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfb\xf4\x0e\x00\x18\x26\xf2\x00\x01\x11\x19\xd8\x14\x18\xd8\x0b\x11\x80\x4d\xf0\x05\x01\x11\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[126];
    }
importlib__bootstrap_toplevel_consts_37_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 125,
    },
    .ob_shash = -1,
    .ob_sval = "\x91\x11\x45\x13\x00\xc2\x18\x07\x45\x23\x00\xc2\x2f\x11\x45\x32\x00\xc3\x01\x07\x46\x01\x00\xc3\x24\x11\x46\x10\x00\xc4\x11\x11\x46\x1f\x00\xc4\x3e\x11\x46\x2e\x00\xc5\x13\x09\x45\x20\x03\xc5\x1f\x01\x45\x20\x03\xc5\x23\x09\x45\x2f\x03\xc5\x2e\x01\x45\x2f\x03\xc5\x32\x09\x45\x3e\x03\xc5\x3d\x01\x45\x3e\x03\xc6\x01\x09\x46\x0d\x03\xc6\x0c\x01\x46\x0d\x03\xc6\x10\x09\x46\x1c\x03\xc6\x1b\x01\x46\x1c\x03\xc6\x1f\x09\x46\x2b\x03\xc6\x2a\x01\x46\x2b\x03\xc6\x2e\x09\x46\x3b\x03\xc6\x3a\x01\x46\x3b\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_37_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
            & const_str_override._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_NamespaceLoader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(892)
importlib__bootstrap_toplevel_consts_37 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 446,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_37_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_37_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_37_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 733,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 58,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_37_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__init_module_attrs._ascii.ob_base,
    .co_qualname = & const_str__init_module_attrs._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_37_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x73\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x12\x09\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x73\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x02\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x6e\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x80\x58\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x4c\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x80\x06\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x5f\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x7c\x01\x5f\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x09\x00\x7c\x03\x7c\x01\x5f\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x73\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x12\x09\x00\x7c\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x09\x00\x7c\x00\x7c\x01\x5f\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x73\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x04\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x1e\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x12\x09\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x10\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x50\x7c\x02\x73\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x05\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x12\x09\x00\x7c\x00\x6a\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x73\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x06\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x20\x7c\x00\x6a\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x14\x09\x00\x7c\x00\x6a\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x14\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x53\x00\x7c\x01\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x90\x01\x8c\x3d\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xcf\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xbd\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xc4\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xa6\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x88\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x7c\x01\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[44];
    }
importlib__bootstrap_toplevel_consts_38_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 43,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Create a module based on the provided spec.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_create_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "create_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_exec_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exec_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[67];
    }
importlib__bootstrap_toplevel_consts_38_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 66,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loaders that define exec_module() must also define create_module()",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_38_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_38_consts_0._ascii.ob_base,
            Py_None,
            & const_str_create_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_38_consts_4._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_38_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_hasattr._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_create_module._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__new_module._ascii.ob_base,
            &_Py_ID(name),
            & const_str__init_module_attrs._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_module_from_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "module_from_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[112];
    }
importlib__bootstrap_toplevel_consts_38_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 111,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x0e\x12\x80\x46\xdc\x07\x0e\x88\x74\x8f\x7b\x89\x7b\x98\x4f\xd4\x07\x2c\xf0\x06\x00\x12\x16\x97\x1b\x91\x1b\xd7\x11\x2a\xd1\x11\x2a\xa8\x34\xd3\x11\x30\x89\x06\xdc\x09\x10\x90\x14\x97\x1b\x91\x1b\x98\x6d\xd4\x09\x2c\xdc\x0e\x19\xf0\x00\x01\x1b\x3d\xf3\x00\x01\x0f\x3e\xf0\x00\x01\x09\x3e\xe0\x07\x0d\x80\x7e\xdc\x11\x1c\x98\x54\x9f\x59\x99\x59\xd3\x11\x27\x88\x06\xdc\x04\x16\x90\x74\x98\x56\xd4\x04\x24\xd8\x0b\x11\x80\x4d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_38_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
        },
    },
};
static
    struct _PyCode_DEF(246)
importlib__bootstrap_toplevel_consts_38 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 123,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_38_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_38_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 806,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 59,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_38_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_module_from_spec._ascii.ob_base,
    .co_qualname = & const_str_module_from_spec._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_38_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x72\x1c\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x6e\x21\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x72\x0b\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x80\x15\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_toplevel_consts_39_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the repr to use for the module.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib__bootstrap_toplevel_consts_39_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " (namespace) from ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_39_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_39_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[63],
            & importlib__bootstrap_toplevel_consts_29_consts_5._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[62],
            & importlib__bootstrap_toplevel_consts_39_consts_4._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_29_consts_8._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_29_consts_9._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_29_consts_6._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_39_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(origin),
            & const_str_loader._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            &_Py_ID(isinstance),
            & const_str_NamespaceLoader._ascii.ob_base,
            & const_str_list._ascii.ob_base,
            & const_str__path._ascii.ob_base,
            & const_str_has_location._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[201];
    }
importlib__bootstrap_toplevel_consts_39_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 200,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x12\x16\x97\x29\x91\x29\xd0\x12\x23\x89\x33\xa8\x14\xaf\x19\xa9\x19\x80\x44\xd8\x07\x0b\x87\x7b\x81\x7b\xd0\x07\x1a\xd8\x11\x15\x97\x1b\x91\x1b\x88\x06\xd8\x0b\x11\x88\x3e\xd8\x15\x1d\x98\x64\x98\x58\xa0\x51\xd0\x13\x27\xd0\x0c\x27\xe4\x0c\x1f\xd0\x0c\x2b\xdc\x10\x1a\x98\x36\xd4\x23\x36\xd7\x23\x46\xd1\x23\x46\xd4\x10\x47\xe0\x15\x1d\x98\x64\x98\x58\xd0\x25\x37\xbc\x04\xb8\x56\xbf\x5c\xb9\x5c\xd3\x38\x4a\xd0\x37\x4b\xc8\x31\xd0\x13\x4d\xd0\x0c\x4d\xe0\x15\x1d\x98\x64\x98\x58\xa0\x52\xa8\x06\xa0\x7a\xb0\x12\xd0\x13\x34\xd0\x0c\x34\xe0\x0b\x0f\xd7\x0b\x1c\xd2\x0b\x1c\xd8\x15\x1d\x98\x64\x98\x58\xa0\x56\xa8\x44\xaf\x4b\xa9\x4b\xa8\x3f\xb8\x21\xd0\x13\x3c\xd0\x0c\x3c\xe0\x15\x1d\x98\x64\x9f\x69\x99\x69\x98\x5d\xa8\x22\xa8\x54\xaf\x5b\xa9\x5b\xa8\x4d\xb8\x12\xd0\x13\x3c\xd0\x0c\x3c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_39_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_spec._ascii.ob_base,
            &_Py_ID(name),
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(376)
importlib__bootstrap_toplevel_consts_39 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 188,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_39_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_39_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 823,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 60,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_39_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__module_repr_from_spec._ascii.ob_base,
    .co_qualname = & const_str__module_repr_from_spec._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_39_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x02\x64\x01\x6e\x0b\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x59\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x80\x06\x64\x02\x7c\x01\x9b\x02\x64\x03\x9d\x03\x53\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x81\x36\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x1c\x64\x02\x7c\x01\x9b\x02\x64\x04\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x03\x9d\x05\x53\x00\x64\x02\x7c\x01\x9b\x02\x64\x05\x7c\x02\x9b\x02\x64\x06\x9d\x05\x53\x00\x7c\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x13\x64\x02\x7c\x01\x9b\x02\x64\x07\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x03\x9d\x05\x53\x00\x64\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x05\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x06\x9d\x05\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[71];
    }
importlib__bootstrap_toplevel_consts_40_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 70,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Execute the spec's specified module in an existing module's namespace.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
importlib__bootstrap_toplevel_consts_40_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "module ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_toplevel_consts_40_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " not in sys.modules",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
importlib__bootstrap_toplevel_consts_40_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "missing loader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[56];
    }
importlib__bootstrap_toplevel_consts_40_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 55,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".exec_module() not found; falling back to load_module()",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_toplevel_consts_40_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_40_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_40_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_40_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_40_consts_5._ascii.ob_base,
            Py_True,
            & importlib__bootstrap_toplevel_consts_36._object.ob_base.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_40_consts_9._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_ImportWarning = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ImportWarning",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_load_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "load_module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
importlib__bootstrap_toplevel_consts_40_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str__ModuleLockManager._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(get),
            & const_str_ImportError._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str__init_module_attrs._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            & const_str__object_name._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_ImportWarning._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[340];
    }
importlib__bootstrap_toplevel_consts_40_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 339,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0f\x8f\x39\x89\x39\x80\x44\xdc\x09\x1b\x98\x44\xd5\x09\x21\xdc\x0b\x0e\x8f\x3b\x89\x3b\x8f\x3f\x89\x3f\x98\x34\xd3\x0b\x20\xa8\x06\xd1\x0b\x2e\xd8\x14\x1b\x98\x44\x98\x38\xd0\x23\x36\xd0\x12\x37\x88\x43\xdc\x12\x1d\x98\x63\xa8\x04\xd4\x12\x2d\xd0\x0c\x2d\xf0\x02\x13\x09\x2c\xd8\x0f\x13\x8f\x7b\x89\x7b\xd0\x0f\x22\xd8\x13\x17\xd7\x13\x32\xd1\x13\x32\xd0\x13\x3a\xdc\x1a\x25\xd0\x26\x36\xb8\x54\xbf\x59\xb9\x59\xd4\x1a\x47\xd0\x14\x47\xe4\x10\x22\xa0\x34\xa8\x16\xb8\x24\xd6\x10\x3f\xe4\x10\x22\xa0\x34\xa8\x16\xb8\x24\xd5\x10\x3f\xdc\x17\x1e\x98\x74\x9f\x7b\x99\x7b\xa8\x4d\xd4\x17\x3a\xdc\x1e\x2a\xa8\x34\xaf\x3b\xa9\x3b\xd3\x1e\x37\xd0\x1d\x38\xf0\x00\x01\x39\x3b\xf0\x00\x01\x1c\x3b\x90\x43\xe4\x14\x1d\x97\x4e\x91\x4e\xa0\x33\xac\x0d\xd4\x14\x36\xd8\x14\x18\x97\x4b\x91\x4b\xd7\x14\x2b\xd1\x14\x2b\xa8\x44\xd5\x14\x31\xe0\x14\x18\x97\x4b\x91\x4b\xd7\x14\x2b\xd1\x14\x2b\xa8\x46\xd4\x14\x33\xf4\x08\x00\x16\x19\x97\x5b\x91\x5b\x97\x5f\x91\x5f\xa0\x54\xa7\x59\xa1\x59\xd3\x15\x2f\x88\x46\xd8\x25\x2b\x8c\x43\x8f\x4b\x89\x4b\x98\x04\x9f\x09\x99\x09\xd2\x0c\x22\xf7\x2f\x00\x0a\x22\xf0\x30\x00\x0c\x12\x80\x4d\xf8\xf4\x05\x00\x16\x19\x97\x5b\x91\x5b\x97\x5f\x91\x5f\xa0\x54\xa7\x59\xa1\x59\xd3\x15\x2f\x88\x46\xd8\x25\x2b\x8c\x43\x8f\x4b\x89\x4b\x98\x04\x9f\x09\x99\x09\xd2\x0c\x22\xfa\xf7\x2f\x00\x0a\x22\xf0\x30\x00\x0c\x12\x80\x4d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[39];
    }
importlib__bootstrap_toplevel_consts_40_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 38,
    },
    .ob_shash = -1,
    .ob_sval = "\x98\x35\x46\x35\x03\xc1\x0e\x43\x0b\x45\x2a\x02\xc4\x19\x41\x06\x46\x35\x03\xc5\x2a\x41\x08\x46\x32\x05\xc6\x32\x03\x46\x35\x03\xc6\x35\x05\x46\x3f\x07",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_40_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
            &_Py_ID(name),
            &_Py_ID(msg),
        },
    },
};
static
    struct _PyCode_DEF(900)
importlib__bootstrap_toplevel_consts_40 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 450,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_40_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_40_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_40_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 845,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 61,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_40_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__exec._ascii.ob_base,
    .co_qualname = & const_str__exec._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_40_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x01\x75\x01\x72\x13\x64\x01\x7c\x02\x9b\x02\x64\x02\x9d\x03\x7d\x03\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x02\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x09\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x32\x7c\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x17\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x64\x06\xac\x07\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x8d\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x64\x06\xac\x07\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\xab\x02\x00\x00\x00\x00\x00\x00\x73\x4e\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x09\x9d\x02\x7d\x03\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x1b\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x09\x00\x64\x04\x64\x04\x64\x04\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x7c\x01\x53\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_41_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__loader__),
            &_Py_ID(__package__),
            &_Py_ID(__path__),
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_ID(__spec__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
importlib__bootstrap_toplevel_consts_41_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & const_str_loader._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
            &_Py_ID(name),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str_pop._ascii.ob_base,
            &_Py_ID(getattr),
            &_Py_ID(__loader__),
            & const_str_AttributeError._ascii.ob_base,
            &_Py_ID(__name__),
            &_Py_ID(__package__),
            & const_str_hasattr._ascii.ob_base,
            & const_str_rpartition._ascii.ob_base,
            &_Py_ID(__spec__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
const_str__load_backward_compatible = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_load_backward_compatible",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[346];
    }
importlib__bootstrap_toplevel_consts_41_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 345,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x06\x05\x0e\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x1f\xd1\x08\x1f\xa0\x04\xa7\x09\xa1\x09\xd4\x08\x2a\xf4\x10\x00\x0e\x11\x8f\x5b\x89\x5b\x8f\x5f\x89\x5f\x98\x54\x9f\x59\x99\x59\xd3\x0d\x27\x80\x46\xd8\x1d\x23\x84\x43\x87\x4b\x81\x4b\x90\x04\x97\x09\x91\x09\xd1\x04\x1a\xdc\x07\x0e\x88\x76\x90\x7c\xa0\x54\xd3\x07\x2a\xd0\x07\x32\xf0\x02\x03\x09\x11\xd8\x20\x24\xa7\x0b\xa1\x0b\x88\x46\xd4\x0c\x1d\xf4\x06\x00\x08\x0f\x88\x76\x90\x7d\xa0\x64\xd3\x07\x2b\xd0\x07\x33\xf0\x02\x08\x09\x11\xf0\x08\x00\x22\x28\xa7\x1f\xa1\x1f\x88\x46\xd4\x0c\x1e\xdc\x13\x1a\x98\x36\xa0\x3a\xd4\x13\x2e\xd8\x25\x29\xa7\x59\xa1\x59\xd7\x25\x39\xd1\x25\x39\xb8\x23\xd3\x25\x3e\xb8\x71\xd1\x25\x41\x90\x06\xd4\x10\x22\xf4\x06\x00\x08\x0f\x88\x76\x90\x7a\xa0\x34\xd3\x07\x28\xd0\x07\x30\xf0\x02\x03\x09\x11\xd8\x1e\x22\x88\x46\x8c\x4f\xf0\x06\x00\x0c\x12\x80\x4d\x88\x36\x80\x4d\xf8\xf0\x3b\x04\x05\x0e\xd8\x0b\x0f\x8f\x39\x89\x39\x9c\x03\x9f\x0b\x99\x0b\xd1\x0b\x23\xdc\x15\x18\x97\x5b\x91\x5b\x97\x5f\x91\x5f\xa0\x54\xa7\x59\xa1\x59\xd3\x15\x2f\x88\x46\xd8\x25\x2b\x8c\x43\x8f\x4b\x89\x4b\x98\x04\x9f\x09\x99\x09\xd1\x0c\x22\xd8\x08\x0d\xfb\xf4\x10\x00\x10\x1e\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfb\xf4\x14\x00\x10\x1e\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfb\xf4\x0a\x00\x10\x1e\xf2\x00\x01\x09\x11\xd8\x0c\x10\xd8\x0b\x11\x80\x4d\xf0\x05\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[68];
    }
importlib__bootstrap_toplevel_consts_41_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 67,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x25\x43\x33\x00\xc1\x3b\x11\x45\x1b\x00\xc2\x1a\x41\x00\x45\x2a\x00\xc3\x28\x07\x45\x39\x00\xc3\x33\x41\x25\x45\x18\x03\xc5\x1b\x09\x45\x27\x03\xc5\x26\x01\x45\x27\x03\xc5\x2a\x09\x45\x36\x03\xc5\x35\x01\x45\x36\x03\xc5\x39\x09\x46\x06\x03\xc6\x05\x01\x46\x06\x03",
};
static
    struct _PyCode_DEF(786)
importlib__bootstrap_toplevel_consts_41 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 393,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_41_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_41_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_41_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 875,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 62,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_38_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__load_backward_compatible._ascii.ob_base,
    .co_qualname = & const_str__load_backward_compatible._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_41_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x12\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x02\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x41\x09\x00\x7c\x01\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x73\x23\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x64\x05\x19\x00\x00\x00\x7c\x01\x5f\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x06\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x80\x0a\x09\x00\x7c\x00\x7c\x01\x5f\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x53\x00\x7c\x01\x53\x00\x23\x00\x01\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x46\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x82\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xda\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x9b\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x7c\x01\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib__bootstrap_toplevel_consts_42_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "import {!r} # {!r}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_42_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            Py_None,
            & const_str_exec_module._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_40_consts_9._ascii.ob_base,
            Py_True,
            & importlib__bootstrap_toplevel_consts_40_consts_5._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_42_consts_6._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[18];
        }_object;
    }
importlib__bootstrap_toplevel_consts_42_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 18,
        },
        .ob_item = {
            & const_str_loader._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            & const_str__object_name._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_ImportWarning._ascii.ob_base,
            & const_str__load_backward_compatible._ascii.ob_base,
            & const_str_module_from_spec._ascii.ob_base,
            &_Py_ID(_initializing),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(name),
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            & const_str__verbose_message._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__load_unlocked = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_load_unlocked",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[329];
    }
importlib__bootstrap_toplevel_consts_42_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 328,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x07\x0b\x87\x7b\x81\x7b\xd0\x07\x1e\xe4\x0f\x16\x90\x74\x97\x7b\x91\x7b\xa0\x4d\xd4\x0f\x32\xdc\x16\x22\xa0\x34\xa7\x3b\xa1\x3b\xd3\x16\x2f\xd0\x15\x30\xf0\x00\x01\x31\x34\xf0\x00\x01\x14\x34\x88\x43\xe4\x0c\x15\x8f\x4e\x89\x4e\x98\x33\xa4\x0d\xd4\x0c\x2e\xdc\x13\x2c\xa8\x54\xd3\x13\x32\xd0\x0c\x32\xe4\x0d\x1d\x98\x64\xd3\x0d\x23\x80\x46\xf0\x0a\x00\x1a\x1e\x80\x44\xd4\x04\x16\xf0\x02\x17\x05\x23\xd8\x21\x27\x8c\x03\x8f\x0b\x89\x0b\x90\x44\x97\x49\x91\x49\xd1\x08\x1e\xf0\x02\x0c\x09\x12\xd8\x0f\x13\x8f\x7b\x89\x7b\xd0\x0f\x22\xd8\x13\x17\xd7\x13\x32\xd1\x13\x32\xd0\x13\x3a\xdc\x1a\x25\xd0\x26\x36\xb8\x54\xbf\x59\xb9\x59\xd4\x1a\x47\xd0\x14\x47\xf0\x06\x00\x11\x15\x97\x0b\x91\x0b\xd7\x10\x27\xd1\x10\x27\xa8\x06\xd4\x10\x2f\xf4\x16\x00\x12\x15\x97\x1b\x91\x1b\x97\x1f\x91\x1f\xa0\x14\xa7\x19\xa1\x19\xd3\x11\x2b\x88\x06\xd8\x21\x27\x8c\x03\x8f\x0b\x89\x0b\x90\x44\x97\x49\x91\x49\xd1\x08\x1e\xdc\x08\x18\xd0\x19\x2d\xa8\x74\xaf\x79\xa9\x79\xb8\x24\xbf\x2b\xb9\x2b\xd4\x08\x46\xe0\x1d\x22\x88\x04\xd4\x08\x1a\xe0\x0b\x11\x80\x4d\xf8\xf0\x21\x05\x09\x12\xf0\x02\x03\x0d\x15\xdc\x14\x17\x97\x4b\x91\x4b\xa0\x04\xa7\x09\xa1\x09\xd0\x14\x2a\xf0\x06\x00\x0d\x12\xf8\xf4\x05\x00\x14\x1c\xf2\x00\x01\x0d\x15\xd8\x10\x14\xd8\x0c\x11\xf0\x05\x01\x0d\x15\xfe\xf0\x16\x00\x1e\x23\x88\x04\xd5\x08\x1a\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[75];
    }
importlib__bootstrap_toplevel_consts_42_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 74,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x33\x1d\x45\x3c\x00\xc2\x11\x41\x0a\x45\x0b\x00\xc3\x1b\x41\x27\x45\x3c\x00\xc5\x0b\x02\x45\x39\x03\xc5\x0e\x1b\x45\x2a\x02\xc5\x29\x01\x45\x39\x03\xc5\x2a\x09\x45\x36\x05\xc5\x33\x02\x45\x39\x03\xc5\x35\x01\x45\x36\x05\xc5\x36\x03\x45\x39\x03\xc5\x39\x03\x45\x3c\x00\xc5\x3c\x09\x46\x05\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_42_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_spec._ascii.ob_base,
            &_Py_ID(msg),
            &_Py_ID(module),
        },
    },
};
static
    struct _PyCode_DEF(784)
importlib__bootstrap_toplevel_consts_42 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 392,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_42_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_42_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_42_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 911,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 63,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_42_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__load_unlocked._ascii.ob_base,
    .co_qualname = & const_str__load_unlocked._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_42_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x53\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x73\x3d\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x02\x9d\x02\x7d\x01\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x03\x7c\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x09\x00\x7c\x02\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x23\x7c\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x32\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x05\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x7c\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x64\x07\x7c\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x53\x00\x23\x00\x01\x00\x09\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x82\x00\x23\x00\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01\x78\x03\x59\x00\x77\x01\x23\x00\x64\x07\x7c\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[192];
    }
importlib__bootstrap_toplevel_consts_43_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 191,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x61\x20\x6e\x65\x77\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6f\x62\x6a\x65\x63\x74\x2c\x20\x6c\x6f\x61\x64\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x73\x70\x65\x63\x27\x73\x20\x6c\x6f\x61\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x61\x64\x64\x65\x64\x20\x74\x6f\x20\x69\x74\x73\x20\x70\x61\x72\x65\x6e\x74\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x61\x6c\x72\x65\x61\x64\x79\x20\x69\x6e\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x2c\x20\x74\x68\x61\x74\x20\x65\x78\x69\x73\x74\x69\x6e\x67\x20\x6d\x6f\x64\x75\x6c\x65\x20\x67\x65\x74\x73\x0a\x20\x20\x20\x20\x63\x6c\x6f\x62\x62\x65\x72\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_43_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_43_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_43_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__ModuleLockManager._ascii.ob_base,
            &_Py_ID(name),
            & const_str__load_unlocked._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
importlib__bootstrap_toplevel_consts_43_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x12\x00\x0a\x1c\x98\x44\x9f\x49\x99\x49\xd5\x09\x26\xdc\x0f\x1d\x98\x64\xd3\x0f\x23\xf7\x03\x00\x0a\x27\xd7\x09\x26\xd2\x09\x26\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
importlib__bootstrap_toplevel_consts_43_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x96\x0b\x2b\x03\xab\x05\x34\x07",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_43_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(110)
importlib__bootstrap_toplevel_consts_43 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_43_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_43_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_43_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 956,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 64,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_43_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__load._ascii.ob_base,
    .co_qualname = & const_str__load._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_43_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x63\x02\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x79\x01\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_BuiltinImporter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BuiltinImporter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[145];
    }
importlib__bootstrap_toplevel_consts_44_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 144,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4d\x65\x74\x61\x20\x70\x61\x74\x68\x20\x69\x6d\x70\x6f\x72\x74\x20\x66\x6f\x72\x20\x62\x75\x69\x6c\x74\x2d\x69\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x41\x6c\x6c\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x61\x72\x65\x20\x65\x69\x74\x68\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x6f\x72\x20\x73\x74\x61\x74\x69\x63\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x74\x6f\x20\x61\x76\x6f\x69\x64\x20\x74\x68\x65\x20\x6e\x65\x65\x64\x20\x74\x6f\x0a\x20\x20\x20\x20\x69\x6e\x73\x74\x61\x6e\x74\x69\x61\x74\x65\x20\x74\x68\x65\x20\x63\x6c\x61\x73\x73\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
importlib__bootstrap_toplevel_consts_44_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "built-in",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_34_consts_2._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_is_builtin = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_builtin",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_is_builtin._ascii.ob_base,
            & const_str_spec_from_loader._ascii.ob_base,
            & const_str__ORIGIN._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_find_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "find_spec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_44_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BuiltinImporter.find_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
importlib__bootstrap_toplevel_consts_44_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0f\x8f\x3f\x89\x3f\x98\x38\xd4\x0b\x24\xdc\x13\x23\xa0\x48\xa8\x63\xb8\x23\xbf\x2b\xb9\x2b\xd4\x13\x46\xd0\x0c\x46\xe0\x13\x17",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_cls = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cls",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(target),
        },
    },
};
static
    struct _PyCode_DEF(94)
importlib__bootstrap_toplevel_consts_44_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 47,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_44_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 982,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 65,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_find_spec._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_44_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x72\x18\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x01\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_44_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Create a built-in module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_44_consts_5_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_26_consts_1_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_create_builtin = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "create_builtin",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_sys._ascii.ob_base,
            & const_str_builtin_module_names._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_create_builtin._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_toplevel_consts_44_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BuiltinImporter.create_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[76];
    }
importlib__bootstrap_toplevel_consts_44_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 75,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x0c\x10\x8f\x39\x89\x39\x9c\x43\xd7\x1c\x34\xd1\x1c\x34\xd1\x0b\x34\xdc\x12\x1d\xa0\x14\xa7\x19\xa1\x19\xa0\x0d\xd0\x2d\x46\xd0\x1e\x47\xd8\x23\x27\xa7\x39\xa1\x39\xf4\x03\x01\x13\x2e\xf0\x00\x01\x0d\x2e\xe4\x0f\x28\xac\x14\xd7\x29\x3c\xd1\x29\x3c\xb8\x64\xd3\x0f\x43\xd0\x08\x43",
};
static
    struct _PyCode_DEF(182)
importlib__bootstrap_toplevel_consts_44_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 91,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_44_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 989,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 66,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_43_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_create_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_44_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x24\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x01\x9d\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_toplevel_consts_44_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Exec a built-in module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_44_consts_6_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_exec_builtin = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exec_builtin",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_exec_builtin._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_44_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BuiltinImporter.exec_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib__bootstrap_toplevel_consts_44_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x09\x22\xa4\x24\xd7\x22\x33\xd1\x22\x33\xb0\x56\xd5\x08\x3c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(module),
        },
    },
};
static
    struct _PyCode_DEF(56)
importlib__bootstrap_toplevel_consts_44_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_44_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 997,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 67,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_exec_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_44_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[58];
    }
importlib__bootstrap_toplevel_consts_44_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 57,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return None as built-in modules do not have code objects.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_44_consts_7_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_get_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_code",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_44_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BuiltinImporter.get_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
importlib__bootstrap_toplevel_consts_44_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x10\x14",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_toplevel_consts_44_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts_7_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1002,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 68,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_get_code._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_44_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[57];
    }
importlib__bootstrap_toplevel_consts_44_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 56,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return None as built-in modules do not have source code.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_44_consts_8_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_toplevel_consts_44_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BuiltinImporter.get_source",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_toplevel_consts_44_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts_8_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1008,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 69,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(get_source),
    .co_qualname = & importlib__bootstrap_toplevel_consts_44_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[53];
    }
importlib__bootstrap_toplevel_consts_44_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 52,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return False as built-in modules are never packages.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_44_consts_9_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_toplevel_consts_44_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BuiltinImporter.is_package",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
importlib__bootstrap_toplevel_consts_44_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x10\x15",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_toplevel_consts_44_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts_9_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1014,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 70,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_is_package._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_44_consts_9_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts_10 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_BuiltinImporter._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_2._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_44_consts_4.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_5.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_6.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_7.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_8.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_9.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_classmethod = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "classmethod",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
importlib__bootstrap_toplevel_consts_44_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str__ORIGIN._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_staticmethod._ascii.ob_base,
            & const_str_create_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str__requires_builtin._ascii.ob_base,
            & const_str_get_code._ascii.ob_base,
            &_Py_ID(get_source),
            & const_str_is_package._ascii.ob_base,
            & const_str__load_module_shim._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[172];
    }
importlib__bootstrap_toplevel_consts_44_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 171,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x05\x05\x08\xf0\x0e\x00\x0f\x19\x80\x47\xe0\x05\x10\xf2\x02\x04\x05\x18\xf3\x03\x00\x06\x11\xf0\x02\x04\x05\x18\xf0\x0c\x00\x06\x12\xf1\x02\x05\x05\x44\x01\xf3\x03\x00\x06\x12\xf0\x02\x05\x05\x44\x01\xf0\x0e\x00\x06\x12\xf1\x02\x02\x05\x3d\xf3\x03\x00\x06\x12\xf0\x02\x02\x05\x3d\xf0\x08\x00\x06\x11\xd8\x05\x16\xf1\x02\x02\x05\x14\xf3\x03\x00\x06\x17\xf3\x03\x00\x06\x11\xf0\x04\x02\x05\x14\xf0\x08\x00\x06\x11\xd8\x05\x16\xf1\x02\x02\x05\x14\xf3\x03\x00\x06\x17\xf3\x03\x00\x06\x11\xf0\x04\x02\x05\x14\xf0\x08\x00\x06\x11\xd8\x05\x16\xf1\x02\x02\x05\x15\xf3\x03\x00\x06\x17\xf3\x03\x00\x06\x11\xf0\x04\x02\x05\x15\xf1\x08\x00\x13\x1e\xd0\x1e\x2f\xd3\x12\x30\x81\x4b",
};
static
    struct _PyCode_DEF(164)
importlib__bootstrap_toplevel_consts_44 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 82,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_44_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 971,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 71,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_BuiltinImporter._ascii.ob_base,
    .co_qualname = & const_str_BuiltinImporter._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x65\x05\x64\x0a\x64\x04\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x07\x64\x05\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x65\x07\x64\x06\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x09\x65\x05\x65\x0a\x64\x07\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0b\x65\x05\x65\x0a\x64\x08\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0c\x65\x05\x65\x0a\x64\x09\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0d\x02\x00\x65\x05\x65\x0e\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x0f\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_FrozenImporter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[143];
    }
importlib__bootstrap_toplevel_consts_46_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 142,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4d\x65\x74\x61\x20\x70\x61\x74\x68\x20\x69\x6d\x70\x6f\x72\x74\x20\x66\x6f\x72\x20\x66\x72\x6f\x7a\x65\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x41\x6c\x6c\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x61\x72\x65\x20\x65\x69\x74\x68\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x6f\x72\x20\x73\x74\x61\x74\x69\x63\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x74\x6f\x20\x61\x76\x6f\x69\x64\x20\x74\x68\x65\x20\x6e\x65\x65\x64\x20\x74\x6f\x0a\x20\x20\x20\x20\x69\x6e\x73\x74\x61\x6e\x74\x69\x61\x74\x65\x20\x74\x68\x65\x20\x63\x6c\x61\x73\x73\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_frozen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "frozen",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str___origname__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__origname__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
importlib__bootstrap_toplevel_consts_46_consts_3_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "see PyImport_ImportFrozenModuleObject()",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_origname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "origname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_3_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(filename),
            & const_str_origname._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            Py_None,
            & const_str___origname__._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_3_consts_2._ascii.ob_base,
            &_Py_ID(__path__),
            & importlib__bootstrap_toplevel_consts_46_consts_3_consts_4._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_ID(__file__),
            &_Py_ID(filename),
            & const_str_origname._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_vars = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "vars",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_is_frozen_package = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_frozen_package",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__resolve_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_resolve_filename",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_implementation = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "implementation",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_insert = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "insert",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_sorted = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sorted",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[23];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 23,
        },
        .ob_item = {
            &_Py_ID(__spec__),
            & const_str_loader_state._ascii.ob_base,
            & const_str_vars._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_is_frozen_package._ascii.ob_base,
            &_Py_ID(__name__),
            & const_str__resolve_filename._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(type),
            & const_str_sys._ascii.ob_base,
            & const_str_implementation._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str_insert._ascii.ob_base,
            &_Py_ID(__file__),
            & const_str_AttributeError._ascii.ob_base,
            &_Py_ID(__path__),
            &_Py_ID(extend),
            & const_str_sorted._ascii.ob_base,
            & const_str_origname._ascii.ob_base,
            &_Py_ID(filename),
            & const_str_has_location._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_toplevel_consts_46_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter._fix_up_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[780];
    }
importlib__bootstrap_toplevel_consts_46_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 779,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x15\x8f\x7f\x89\x7f\x88\x04\xd8\x10\x14\xd7\x10\x21\xd1\x10\x21\x88\x05\xd8\x0b\x10\x89\x3d\xf4\x08\x00\x18\x1c\x98\x46\x93\x7c\xd7\x17\x27\xd1\x17\x27\xa8\x0e\xb8\x04\xd3\x17\x3d\x88\x48\xd9\x13\x1b\xd0\x0c\x46\xd0\x1d\x46\xd3\x0c\x46\x90\x38\xdc\x14\x1b\x98\x46\xa0\x4a\xd3\x14\x2f\x88\x45\xdc\x13\x17\xd7\x13\x29\xd1\x13\x29\xa8\x26\xaf\x2f\xa9\x2f\xd3\x13\x3a\xb8\x65\xd2\x13\x43\xd0\x0c\x4a\xc0\x55\xd3\x0c\x4a\xd0\x13\x43\xd8\x1f\x22\xd7\x1f\x34\xd1\x1f\x34\xb0\x58\xb8\x74\xbf\x79\xb9\x79\xc8\x25\xd3\x1f\x50\xd1\x0c\x1c\x88\x48\x90\x66\xd8\x20\x38\xa4\x04\xa4\x53\xd7\x25\x37\xd1\x25\x37\xd3\x20\x38\xd8\x19\x21\xd8\x19\x21\xf4\x05\x03\x21\x0e\x88\x44\xd4\x0c\x1d\xf0\x08\x00\x18\x1c\xd7\x17\x36\xd1\x17\x36\x88\x48\xd9\x0f\x14\xd8\x17\x1f\xa0\x32\x92\x7e\xd0\x10\x2f\xa0\x78\xd3\x10\x2f\x90\x7e\xd9\x13\x19\xd8\x14\x18\xd7\x14\x33\xd1\x14\x33\xd7\x14\x3a\xd1\x14\x3a\xb8\x31\xb8\x66\xd5\x14\x45\xe0\x17\x1f\xd0\x17\x27\xd0\x10\x31\xa8\x18\xd3\x10\x31\xd0\x17\x27\xf4\x06\x00\x18\x1f\x98\x76\xa0\x7a\xd4\x17\x32\xd0\x0c\x43\xb0\x46\xb7\x4f\xb1\x4f\xd3\x0c\x43\xd0\x13\x32\xd9\x0f\x17\xf0\x02\x03\x11\x19\xd8\x26\x2e\x90\x46\x94\x4f\xf2\x06\x00\x10\x15\xd8\x13\x19\x97\x3f\x91\x3f\xa0\x68\xd3\x13\x2e\xd8\x1b\x21\x9f\x3f\x99\x3f\xa8\x62\xd2\x1b\x30\xd0\x14\x41\xb0\x26\xb7\x2f\xb1\x2f\xd3\x14\x41\xd0\x1b\x30\xd8\x14\x1a\x97\x4f\x91\x4f\xd7\x14\x2a\xd1\x14\x2a\xa8\x38\xd6\x14\x34\xf0\x08\x00\x18\x1c\xd7\x17\x36\xd1\x17\x36\x88\x48\xd8\x14\x1c\xa0\x44\xd0\x14\x28\x88\x45\xe4\x13\x19\x9c\x24\x98\x75\x9b\x2b\xd3\x13\x26\xa8\x3a\xb0\x7a\xd0\x2a\x42\xd2\x13\x42\xd0\x0c\x49\xc0\x45\xd3\x0c\x49\xd0\x13\x42\xd8\x0f\x14\x8f\x7e\x8a\x7e\xf0\x06\x00\x16\x19\xd7\x15\x2a\xd1\x15\x2a\xa8\x35\xaf\x3e\xa9\x3e\xb8\x34\xbf\x39\xb9\x39\xc0\x65\xd3\x15\x4c\xf1\x03\x01\x11\x13\x90\x18\x98\x36\xe0\x17\x1c\x97\x7e\x91\x7e\xa8\x18\xd2\x17\x31\xd0\x10\x4d\xb0\x45\xb7\x4e\xb1\x4e\xc0\x48\xd0\x33\x4d\xd3\x10\x4d\xd0\x17\x31\xd9\x13\x19\xd8\x1b\x23\xa8\x06\xa0\x78\xd2\x1b\x2f\xd0\x14\x43\xb0\x28\xb8\x46\xd0\x31\x43\xd3\x14\x43\xd0\x1b\x2f\xe0\x1b\x23\xa9\x65\xa9\x02\xb8\x14\xd2\x1b\x3e\xd0\x14\x48\xc0\x08\xd3\x14\x48\xd0\x1b\x3e\xe0\x1b\x1f\x90\x08\xd8\x17\x1c\x97\x7e\x91\x7e\xd0\x17\x2d\xd0\x10\x3d\xa8\x75\xaf\x7e\xa9\x7e\xd3\x10\x3d\xd0\x17\x2d\xd8\x17\x1f\xa9\x25\xa1\x42\xb0\x54\xd2\x17\x3a\xd0\x10\x44\xb8\x48\xd3\x10\x44\xd0\x17\x3a\xe1\x0f\x17\xdc\x17\x1e\x98\x76\xa0\x7a\xd4\x17\x32\xd0\x10\x32\xd0\x17\x32\xd8\x17\x1d\x97\x7f\x91\x7f\xa8\x28\xd2\x17\x32\xd0\x10\x4f\xb0\x56\xb7\x5f\xb1\x5f\xc0\x68\xd0\x34\x4f\xd3\x10\x4f\xd0\x17\x32\xe4\x1b\x22\xa0\x36\xa8\x3a\xd4\x1b\x36\xd0\x10\x47\xb8\x06\xbf\x0f\xb9\x0f\xd3\x10\x47\xd0\x17\x36\xd9\x0f\x14\xdc\x17\x1e\x98\x76\xa0\x7a\xd4\x17\x32\xd0\x10\x32\xd0\x17\x32\xd8\x17\x1d\x97\x7f\x91\x7f\xa8\x28\xd2\x17\x32\xd0\x10\x4f\xb0\x56\xb7\x5f\xb1\x5f\xc0\x68\xd0\x34\x4f\xd3\x10\x4f\xd0\x17\x32\xe4\x1b\x22\xa0\x36\xa8\x3a\xd4\x1b\x36\xd0\x10\x47\xb8\x06\xbf\x0f\xb9\x0f\xd3\x10\x47\xd0\x17\x36\xd8\x13\x17\xd7\x13\x24\xd2\x13\x24\xd0\x08\x24\xd0\x0f\x24\xd0\x13\x24\xf8\xf4\x4b\x01\x00\x18\x26\xf2\x00\x01\x11\x19\xda\x14\x18\xf0\x03\x01\x11\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_toplevel_consts_46_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc4\x1d\x07\x4b\x39\x00\xcb\x39\x09\x4c\x06\x03\xcc\x05\x01\x4c\x06\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_state = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "state",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_ispkg = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ispkg",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_pkgdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pkgdir",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            &_Py_ID(module),
            & const_str_spec._ascii.ob_base,
            & const_str_state._ascii.ob_base,
            & const_str_origname._ascii.ob_base,
            & const_str_ispkg._ascii.ob_base,
            &_Py_ID(filename),
            & const_str_pkgdir._ascii.ob_base,
            &_Py_ID(__path__),
            &_Py_ID(__file__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "          ",
};
static
    struct _PyCode_DEF(1554)
importlib__bootstrap_toplevel_consts_46_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 777,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_46_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1034,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 72,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_46_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(_fix_up_module),
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x90\x01\x80\x58\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x73\x07\x4a\x00\x64\x02\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x05\x6b\x28\x00\x00\x73\x07\x4a\x00\x7c\x05\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x02\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x03\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x06\x7d\x07\x02\x00\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x04\xac\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x02\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x05\x72\x2b\x7c\x08\x67\x00\x6b\x28\x00\x00\x73\x07\x4a\x00\x7c\x08\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x07\x72\x26\x7c\x02\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x07\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x09\x7c\x08\x81\x07\x4a\x00\x7c\x08\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x72\x11\x4a\x00\x7c\x01\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x06\x72\x08\x09\x00\x7c\x06\x7c\x01\x5f\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x90\x01\x72\xc3\x7c\x01\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\x6b\x37\x00\x00\x90\x01\x72\xb3\x7c\x01\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x6b\x28\x00\x00\x73\x11\x4a\x00\x7c\x01\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x90\x01\x6e\x76\x7c\x02\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x64\x00\x75\x01\x7d\x05\x74\x27\x00\x00\x00\x00\x00\x00\x00\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x07\x64\x08\x67\x02\x6b\x28\x00\x00\x73\x07\x4a\x00\x7c\x03\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x03\x6a\x28\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x6d\x7c\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x6a\x28\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x03\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x09\x7d\x07\x7c\x03\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x6b\x28\x00\x00\x73\x13\x4a\x00\x7c\x03\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x66\x02\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x07\x72\x0f\x7c\x08\x7c\x07\x67\x01\x6b\x28\x00\x00\x73\x48\x4a\x00\x7c\x08\x7c\x07\x66\x02\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x08\x7c\x05\x72\x02\x67\x00\x6e\x01\x64\x00\x6b\x28\x00\x00\x73\x36\x4a\x00\x7c\x08\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x64\x00\x7d\x09\x7c\x03\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x11\x4a\x00\x7c\x03\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x08\x7c\x05\x72\x02\x67\x00\x6e\x01\x64\x00\x6b\x28\x00\x00\x73\x07\x4a\x00\x7c\x08\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x09\x72\x30\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x73\x02\x4a\x00\x82\x01\x7c\x01\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x6b\x28\x00\x00\x73\x30\x4a\x00\x7c\x01\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x66\x02\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x72\x11\x4a\x00\x7c\x01\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x05\x72\x30\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x73\x02\x4a\x00\x82\x01\x7c\x01\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\x6b\x28\x00\x00\x73\x30\x4a\x00\x7c\x01\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\x66\x02\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x72\x11\x4a\x00\x7c\x01\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x6a\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x02\x4a\x00\x82\x01\x79\x00\x23\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x90\x01\x8c\xe1\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__stdlib_dir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_stdlib_dir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_win32 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "win32",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
importlib__bootstrap_toplevel_consts_46_consts_5_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".__init__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
importlib__bootstrap_toplevel_consts_46_consts_5_consts_11 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__init__.py",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
importlib__bootstrap_toplevel_consts_46_consts_5_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".py",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            Py_None,
            & const_str__stdlib_dir._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
            & const_str_win32._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[60],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_8._ascii.ob_base,
            Py_False,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_11._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_12._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str__SEP = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_SEP",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_platform = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "platform",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(getattr),
            & const_str_sys._ascii.ob_base,
            & const_str__SEP._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            &_Py_ID(replace),
            & const_str__stdlib_dir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib__bootstrap_toplevel_consts_46_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter._resolve_filename",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[250];
    }
importlib__bootstrap_toplevel_consts_46_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 249,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe1\x0f\x17\x9c\x77\xa4\x73\xa8\x4d\xb8\x34\xd4\x1f\x40\xd8\x13\x1d\xf0\x02\x03\x09\x46\x01\xd8\x12\x15\x97\x28\x91\x28\x88\x43\xf0\x08\x00\x0c\x14\x90\x75\xd2\x0b\x1c\xd8\x0f\x17\xd7\x0f\x22\xd1\x0f\x22\xa0\x33\xd4\x0f\x27\xd8\x1b\x23\xa0\x41\xa0\x42\x98\x3c\x90\x08\xd9\x17\x1c\xd8\x22\x2a\xa0\x1a\xa8\x39\xd0\x1f\x35\x91\x48\xe0\x18\x1d\x90\x05\xd8\x12\x1a\xd7\x12\x22\xd1\x12\x22\xa0\x33\xa8\x03\xd3\x12\x2c\x88\x07\xd9\x0b\x10\xdc\x18\x1b\x9f\x0f\x99\x0f\xd0\x17\x28\xa8\x13\xa8\x05\xa8\x67\xa8\x59\xd0\x15\x37\x88\x46\xd8\x1a\x20\x98\x18\xa0\x23\xa0\x15\xa0\x6b\xd0\x17\x32\x88\x48\xf0\x08\x00\x10\x18\x98\x16\xd0\x0f\x1f\xd0\x08\x1f\xf0\x05\x00\x16\x1a\x88\x46\xdc\x1a\x1d\x9f\x2f\x99\x2f\xd0\x19\x2a\xa8\x33\xa8\x25\xb0\x07\xa8\x79\xb8\x03\xd0\x17\x3c\x88\x48\xd8\x0f\x17\x98\x16\xd0\x0f\x1f\xd0\x08\x1f\xf8\xf4\x23\x00\x10\x1e\xf2\x00\x01\x09\x46\x01\xdc\x25\x28\xa7\x5c\xa1\x5c\xb0\x57\xd2\x25\x3c\x99\x54\xc0\x23\xd0\x0c\x45\x88\x43\x90\x23\x96\x28\xf0\x03\x01\x09\x46\x01\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
importlib__bootstrap_toplevel_consts_46_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x96\x0c\x42\x19\x00\xc2\x19\x27\x43\x03\x03\xc3\x02\x01\x43\x03\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_relfile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "relfile",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(alias),
            & const_str_ispkg._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_relfile._ascii.ob_base,
            & const_str_pkgdir._ascii.ob_base,
            &_Py_ID(filename),
        },
    },
};
static
    struct _PyCode_DEF(396)
importlib__bootstrap_toplevel_consts_46_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 198,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_46_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1103,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 73,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_46_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__resolve_filename._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x72\x11\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x00\xab\x03\x00\x00\x00\x00\x00\x00\x73\x01\x79\x02\x09\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x01\x7c\x02\x6b\x37\x00\x00\x72\x20\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0d\x7c\x01\x64\x07\x64\x00\x1a\x00\x7d\x01\x7c\x03\x73\x08\x7c\x01\x9b\x00\x64\x08\x9d\x02\x7d\x01\x6e\x02\x64\x09\x7d\x03\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x03\x72\x21\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x7c\x04\x9b\x00\x7c\x05\x9b\x00\x9d\x03\x7d\x06\x7c\x06\x9b\x00\x7c\x04\x9b\x00\x64\x0b\x9d\x03\x7d\x07\x7c\x07\x7c\x06\x66\x02\x53\x00\x64\x00\x7d\x06\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x7c\x04\x9b\x00\x7c\x05\x9b\x00\x64\x0c\x9d\x04\x7d\x07\x7c\x07\x7c\x06\x66\x02\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x21\x01\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x6b\x28\x00\x00\x72\x02\x64\x04\x6e\x01\x64\x05\x78\x01\x7d\x04\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x8c\xa0\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_32._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_3_consts_4._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_find_frozen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "find_frozen",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_find_frozen._ascii.ob_base,
            & const_str_spec_from_loader._ascii.ob_base,
            & const_str__ORIGIN._ascii.ob_base,
            & const_str__resolve_filename._ascii.ob_base,
            &_Py_ID(type),
            & const_str_sys._ascii.ob_base,
            & const_str_implementation._ascii.ob_base,
            & const_str_loader_state._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str_insert._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_46_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter.find_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[158];
    }
importlib__bootstrap_toplevel_consts_46_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 157,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x28\xac\x14\xd7\x29\x39\xd1\x29\x39\xb8\x38\xd3\x0f\x44\x88\x04\xd8\x0b\x0f\x88\x3c\xd8\x13\x17\xf0\x1a\x00\x1e\x22\xd1\x08\x1a\x88\x01\x88\x35\x90\x28\xdc\x0f\x1f\xa0\x08\xa8\x23\xd8\x27\x2a\xa7\x7b\xa1\x7b\xd8\x2b\x30\xf4\x05\x02\x10\x32\x88\x04\xf0\x06\x00\x1c\x1f\xd7\x1b\x30\xd1\x1b\x30\xb0\x18\xb8\x38\xc0\x55\xd3\x1b\x4b\xd1\x08\x18\x88\x08\x90\x26\xd8\x1c\x34\x9c\x44\xa4\x13\xd7\x21\x33\xd1\x21\x33\xd3\x1c\x34\xd8\x15\x1d\xd8\x15\x1d\xf4\x05\x03\x1d\x0a\x88\x04\xd4\x08\x19\xf1\x08\x00\x0c\x12\xd8\x0c\x10\xd7\x0c\x2b\xd1\x0c\x2b\xd7\x0c\x32\xd1\x0c\x32\xb0\x31\xb0\x66\xd4\x0c\x3d\xd8\x0f\x13\x88\x0b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "info",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(target),
            & const_str_info._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            & const_str_ispkg._ascii.ob_base,
            & const_str_origname._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            &_Py_ID(filename),
            & const_str_pkgdir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
importlib__bootstrap_toplevel_consts_46_consts_6_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "           ",
};
static
    struct _PyCode_DEF(306)
importlib__bootstrap_toplevel_consts_46_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 153,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1128,
    .co_nlocalsplus = 11,
    .co_nlocals = 11,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 74,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_46_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_find_spec._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x80\x01\x79\x00\x7c\x04\x5c\x03\x00\x00\x7d\x05\x7d\x06\x7d\x07\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xac\x01\xab\x04\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x7c\x01\x7c\x06\xab\x03\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x09\x7d\x0a\x02\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x09\x7c\x07\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x08\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x72\x1c\x7c\x08\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x0a\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x08\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_toplevel_consts_46_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set __file__, if able.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_46_consts_7_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__new_module._ascii.ob_base,
            &_Py_ID(name),
            & const_str_loader_state._ascii.ob_base,
            &_Py_ID(filename),
            &_Py_ID(__file__),
            & const_str_AttributeError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_toplevel_consts_46_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter.create_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[85];
    }
importlib__bootstrap_toplevel_consts_46_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 84,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x12\x1d\x98\x54\x9f\x59\x99\x59\xd3\x11\x27\x88\x06\xf0\x02\x06\x09\x2b\xd8\x17\x1b\xd7\x17\x28\xd1\x17\x28\xd7\x17\x31\xd1\x17\x31\x88\x48\xf1\x08\x00\x10\x18\xd8\x22\x2a\x90\x06\x94\x0f\xd8\x0f\x15\x88\x0d\xf8\xf4\x0b\x00\x10\x1e\xf2\x00\x01\x09\x11\xd8\x0c\x10\xf0\x08\x00\x10\x16\x88\x0d\xf0\x0b\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib__bootstrap_toplevel_consts_46_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x97\x16\x38\x00\xb8\x09\x41\x05\x03\xc1\x04\x01\x41\x05\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
            &_Py_ID(filename),
        },
    },
};
static
    struct _PyCode_DEF(144)
importlib__bootstrap_toplevel_consts_46_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 72,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_46_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1158,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 75,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_46_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_create_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x72\x07\x7c\x02\x7c\x01\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x53\x00\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x7c\x01\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_get_frozen_object = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_frozen_object",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_exec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exec",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(__spec__),
            &_Py_ID(name),
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_get_frozen_object._ascii.ob_base,
            & const_str_exec._ascii.ob_base,
            &_Py_ID(__dict__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_toplevel_consts_46_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter.exec_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[53];
    }
importlib__bootstrap_toplevel_consts_46_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 52,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x15\x8f\x7f\x89\x7f\x88\x04\xd8\x0f\x13\x8f\x79\x89\x79\x88\x04\xdc\x0f\x28\xac\x14\xd7\x29\x3f\xd1\x29\x3f\xc0\x14\xd3\x0f\x46\x88\x04\xdc\x08\x0c\x88\x54\x90\x36\x97\x3f\x91\x3f\xd5\x08\x23",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(module),
            & const_str_spec._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(code),
        },
    },
};
static
    struct _PyCode_DEF(148)
importlib__bootstrap_toplevel_consts_46_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 74,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1171,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 76,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_46_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_exec_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[96];
    }
importlib__bootstrap_toplevel_consts_46_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 95,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4c\x6f\x61\x64\x20\x61\x20\x66\x72\x6f\x7a\x65\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x69\x73\x20\x64\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2e\x20\x20\x55\x73\x65\x20\x65\x78\x65\x63\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x29\x20\x69\x6e\x73\x74\x65\x61\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_46_consts_9_consts_0._ascii.ob_base,
            Py_None,
            &_Py_ID(__file__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str__load_module_shim._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_find_frozen._ascii.ob_base,
            & const_str___origname__._ascii.ob_base,
            & const_str_vars._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            &_Py_ID(__path__),
            &_Py_ID(_fix_up_module),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_toplevel_consts_46_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter.load_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[114];
    }
importlib__bootstrap_toplevel_consts_46_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 113,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x12\x23\xa0\x33\xa8\x08\xd3\x11\x31\x88\x06\xdc\x0f\x13\xd7\x0f\x1f\xd1\x0f\x1f\xa0\x08\xd3\x0f\x29\x88\x04\xd8\x0f\x13\xd0\x0f\x1f\xd0\x08\x1f\xd0\x0f\x1f\xd8\x1d\x21\xd1\x08\x1a\x88\x01\x88\x35\x90\x28\xd8\x1e\x26\x88\x06\xd4\x08\x1b\xdc\x08\x0c\x88\x56\x8b\x0c\xd7\x08\x18\xd1\x08\x18\x98\x1a\xa0\x54\xd4\x08\x2a\xd9\x0b\x10\xd8\x1e\x20\x88\x46\x8c\x4f\xd8\x08\x0b\xd7\x08\x1a\xd1\x08\x1a\x98\x36\xd4\x08\x22\xd8\x0f\x15\x88\x0d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(module),
            & const_str_info._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            & const_str_ispkg._ascii.ob_base,
            & const_str_origname._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "       ",
};
static
    struct _PyCode_DEF(212)
importlib__bootstrap_toplevel_consts_46_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 106,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_9_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1178,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 77,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_46_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_load_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_9_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x80\x02\x4a\x00\x82\x01\x7c\x03\x5c\x03\x00\x00\x7d\x04\x7d\x05\x7d\x06\x7c\x06\x7c\x02\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x72\x07\x67\x00\x7c\x02\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[46];
    }
importlib__bootstrap_toplevel_consts_46_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 45,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the code object for the frozen module.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_46_consts_10_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_get_frozen_object._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_toplevel_consts_46_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter.get_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
importlib__bootstrap_toplevel_consts_46_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x08\x00\x10\x14\xd7\x0f\x25\xd1\x0f\x25\xa0\x68\xd3\x0f\x2f\xd0\x08\x2f",
};
static
    struct _PyCode_DEF(44)
importlib__bootstrap_toplevel_consts_46_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_10_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1197,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 78,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_get_code._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_10_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
importlib__bootstrap_toplevel_consts_46_consts_11_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return None as frozen modules do not have source code.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_46_consts_11_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_46_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter.get_source",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_toplevel_consts_46_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_11_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1203,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 79,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(get_source),
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_11_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_44_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
importlib__bootstrap_toplevel_consts_46_consts_12_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if the frozen module is a package.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_46_consts_12_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_is_frozen_package._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_46_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FrozenImporter.is_package",
};
static
    struct _PyCode_DEF(44)
importlib__bootstrap_toplevel_consts_46_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts_12_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1209,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 80,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_44_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_is_package._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_toplevel_consts_46_consts_12_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & const_str_FrozenImporter._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_1._ascii.ob_base,
            & const_str_frozen._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_3.ob_base.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_46_consts_5.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_6.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_7.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_8.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_9.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_10.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_11.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_12.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_4_consts._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
importlib__bootstrap_toplevel_consts_46_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str__ORIGIN._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(_fix_up_module),
            & const_str__resolve_filename._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_staticmethod._ascii.ob_base,
            & const_str_create_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
            & const_str__requires_frozen._ascii.ob_base,
            & const_str_get_code._ascii.ob_base,
            &_Py_ID(get_source),
            & const_str_is_package._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[220];
    }
importlib__bootstrap_toplevel_consts_46_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 219,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x05\x05\x08\xf0\x0e\x00\x0f\x17\x80\x47\xe0\x05\x10\xf1\x02\x42\x01\x05\x25\xf3\x03\x00\x06\x11\xf0\x02\x42\x01\x05\x25\xf0\x48\x02\x00\x06\x11\xf2\x02\x16\x05\x20\xf3\x03\x00\x06\x11\xf0\x02\x16\x05\x20\xf0\x30\x00\x06\x11\xf2\x02\x1b\x05\x14\xf3\x03\x00\x06\x11\xf0\x02\x1b\x05\x14\xf0\x3a\x00\x06\x12\xf1\x02\x0a\x05\x16\xf3\x03\x00\x06\x12\xf0\x02\x0a\x05\x16\xf0\x18\x00\x06\x12\xf1\x02\x04\x05\x24\xf3\x03\x00\x06\x12\xf0\x02\x04\x05\x24\xf0\x0c\x00\x06\x11\xf1\x02\x10\x05\x16\xf3\x03\x00\x06\x11\xf0\x02\x10\x05\x16\xf0\x24\x00\x06\x11\xd8\x05\x15\xf1\x02\x02\x05\x30\xf3\x03\x00\x06\x16\xf3\x03\x00\x06\x11\xf0\x04\x02\x05\x30\xf0\x08\x00\x06\x11\xd8\x05\x15\xf1\x02\x02\x05\x14\xf3\x03\x00\x06\x16\xf3\x03\x00\x06\x11\xf0\x04\x02\x05\x14\xf0\x08\x00\x06\x11\xd8\x05\x15\xf1\x02\x02\x05\x30\xf3\x03\x00\x06\x16\xf3\x03\x00\x06\x11\xf1\x04\x02\x05\x30",
};
static
    struct _PyCode_DEF(198)
importlib__bootstrap_toplevel_consts_46 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 99,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_46_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_46_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1023,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 81,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_FrozenImporter._ascii.ob_base,
    .co_qualname = & const_str_FrozenImporter._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_46_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x65\x05\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x05\x64\x0d\x64\x05\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x65\x05\x64\x0e\x64\x06\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x65\x09\x64\x07\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0a\x65\x09\x64\x08\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0b\x65\x05\x64\x09\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0c\x65\x05\x65\x0d\x64\x0a\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0e\x65\x05\x65\x0d\x64\x0b\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0f\x65\x05\x65\x0d\x64\x0c\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x10\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__ImportLockContext = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ImportLockContext",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[37];
    }
importlib__bootstrap_toplevel_consts_48_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 36,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Context manager for the import lock.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_toplevel_consts_48_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Acquire the import lock.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_48_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_48_consts_2_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_48_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_acquire_lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_toplevel_consts_48_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ImportLockContext.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
importlib__bootstrap_toplevel_consts_48_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x0c\xd7\x08\x19\xd1\x08\x19\xd5\x08\x1b",
};
static
    struct _PyCode_DEF(44)
importlib__bootstrap_toplevel_consts_48_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_48_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_48_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1222,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 82,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_48_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_48_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[61];
    }
importlib__bootstrap_toplevel_consts_48_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 60,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Release the import lock regardless of any raised exceptions.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_48_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_48_consts_3_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_48_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_release_lock._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_48_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ImportLockContext.__exit__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_exc_traceback = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exc_traceback",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_48_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(exc_type),
            &_Py_ID(exc_value),
            & const_str_exc_traceback._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(44)
importlib__bootstrap_toplevel_consts_48_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_48_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_48_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1226,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 83,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_48_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & importlib__bootstrap_toplevel_consts_48_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_48_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_48_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__ImportLockContext._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_48_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_48_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_48_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_48_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib__bootstrap_toplevel_consts_48_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x2e\xf2\x04\x02\x05\x1c\xf3\x08\x02\x05\x1c",
};
static
    struct _PyCode_DEF(28)
importlib__bootstrap_toplevel_consts_48 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_48_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_48_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1218,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 84,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__ImportLockContext._ascii.ob_base,
    .co_qualname = & const_str__ImportLockContext._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_48_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[51];
    }
importlib__bootstrap_toplevel_consts_50_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 50,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Resolve a relative module name to an absolute one.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[51];
    }
importlib__bootstrap_toplevel_consts_50_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 50,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "attempted relative import beyond top-level package",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_50_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_50_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & importlib__bootstrap_toplevel_consts_50_consts_3._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_rsplit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rsplit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_50_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_rsplit._ascii.ob_base,
            &_Py_ID(len),
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__resolve_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_resolve_name",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[80];
    }
importlib__bootstrap_toplevel_consts_50_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 79,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x12\x8f\x3e\x89\x3e\x98\x23\x98\x75\xa0\x71\x99\x79\xd3\x0b\x29\x80\x44\xdc\x07\x0a\x88\x34\x83\x79\x90\x35\xd2\x07\x18\xdc\x0e\x19\xd0\x1a\x4e\xd3\x0e\x4f\xd0\x08\x4f\xd8\x0b\x0f\x90\x01\x89\x37\x80\x44\xd9\x1f\x23\x88\x64\x88\x56\x90\x31\x90\x54\x90\x46\xd0\x0b\x1b\xd0\x04\x2d\xa8\x14\xd0\x04\x2d",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_package = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "package",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_bits = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bits",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_50_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_package._ascii.ob_base,
            &_Py_ID(level),
            & const_str_bits._ascii.ob_base,
            &_Py_ID(base),
        },
    },
};
static
    struct _PyCode_DEF(126)
importlib__bootstrap_toplevel_consts_50 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 63,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_50_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_50_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1231,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 85,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_50_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__resolve_name._ascii.ob_base,
    .co_qualname = & const_str__resolve_name._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_50_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x02\x64\x02\x7a\x0a\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x6b\x02\x00\x00\x72\x0b\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x03\x64\x04\x19\x00\x00\x00\x7d\x04\x7c\x00\x72\x07\x7c\x04\x9b\x00\x64\x01\x7c\x00\x9b\x00\x9d\x03\x53\x00\x7c\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
importlib__bootstrap_toplevel_consts_51_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Find a module's spec.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
importlib__bootstrap_toplevel_consts_51_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys.meta_path is None, Python is likely shutting down",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_toplevel_consts_51_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys.meta_path is empty",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_51_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_51_consts_0._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_51_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_51_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_meta_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "meta_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_toplevel_consts_51_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_meta_path._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_ImportWarning._ascii.ob_base,
            &_Py_ID(modules),
            & const_str__ImportLockContext._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            &_Py_ID(__spec__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__find_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_find_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[263];
    }
importlib__bootstrap_toplevel_consts_51_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 262,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x10\x13\x97\x0d\x91\x0d\x80\x49\xd8\x07\x10\xd0\x07\x18\xe4\x0e\x19\xf0\x00\x01\x1b\x2a\xf3\x00\x01\x0f\x2b\xf0\x00\x01\x09\x2b\xf1\x06\x00\x0c\x15\xdc\x08\x11\x8f\x0e\x89\x0e\xd0\x17\x2f\xb4\x1d\xd4\x08\x3f\xf0\x0a\x00\x11\x15\x9c\x03\x9f\x0b\x99\x0b\xd0\x10\x23\x80\x49\xdb\x12\x1b\x88\x06\xdc\x0d\x1f\xd5\x0d\x21\xf0\x02\x05\x0d\x35\xd8\x1c\x22\xd7\x1c\x2c\xd1\x1c\x2c\x90\x09\xf1\x08\x00\x18\x21\xa0\x14\xa0\x74\xa8\x56\xd3\x17\x34\x91\x04\xf7\x0d\x00\x0e\x22\xf0\x0e\x00\x0c\x10\xd1\x0b\x1b\xe1\x13\x1c\xa0\x14\xac\x13\xaf\x1b\xa9\x1b\xd1\x21\x34\xdc\x19\x1c\x9f\x1b\x99\x1b\xa0\x54\xd1\x19\x2a\x90\x06\xf0\x02\x0b\x11\x28\xd8\x1f\x25\x9f\x7f\x99\x7f\x90\x48\xf0\x0e\x00\x18\x20\xd0\x17\x27\xd8\x1f\x23\x9a\x0b\xe0\x1f\x27\x9a\x0f\xe0\x17\x1b\x92\x0b\xf0\x33\x00\x13\x1c\xf0\x36\x00\x10\x14\xf8\xf4\x2f\x00\x14\x22\xf2\x00\x01\x0d\x19\xd8\x10\x18\xf7\x09\x00\x0e\x22\xd0\x0d\x21\xf0\x06\x01\x0d\x19\xfa\xf7\x07\x00\x0e\x22\xd0\x0d\x21\xfb\xf4\x1a\x00\x18\x26\xf2\x00\x04\x11\x20\xf0\x08\x00\x1c\x20\x94\x4b\xf0\x09\x04\x11\x20\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
importlib__bootstrap_toplevel_consts_51_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x1b\x01\x43\x1a\x05\xc1\x1d\x0c\x43\x03\x04\xc1\x29\x0a\x43\x1a\x05\xc2\x27\x0c\x43\x26\x02\xc3\x03\x09\x43\x17\x07\xc3\x0c\x01\x43\x1a\x05\xc3\x16\x01\x43\x17\x07\xc3\x17\x03\x43\x1a\x05\xc3\x1a\x05\x43\x23\x09\xc3\x26\x0b\x43\x36\x05\xc3\x35\x01\x43\x36\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_is_reload = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_reload",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_finder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "finder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_toplevel_consts_51_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(path),
            &_Py_ID(target),
            & const_str_meta_path._ascii.ob_base,
            & const_str_is_reload._ascii.ob_base,
            & const_str_finder._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
            &_Py_ID(__spec__),
        },
    },
};
static
    struct _PyCode_DEF(498)
importlib__bootstrap_toplevel_consts_51 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 249,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_51_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_51_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_51_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1240,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 86,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_51_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__find_spec._ascii.ob_base,
    .co_qualname = & const_str__find_spec._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_51_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x80\x0b\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x03\x73\x1a\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x7d\x04\x7c\x03\x44\x00\x5d\x71\x00\x00\x7d\x05\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x09\x00\x7c\x05\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x02\x00\x7c\x06\x7c\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x07\x09\x00\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7f\x07\x80\x01\x8c\x31\x7c\x04\x73\x3c\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x2a\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x19\x00\x00\x00\x7d\x08\x09\x00\x7c\x08\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x09\x7c\x09\x80\x04\x7c\x07\x63\x02\x01\x00\x53\x00\x7c\x09\x63\x02\x01\x00\x53\x00\x7c\x07\x63\x02\x01\x00\x53\x00\x04\x00\x79\x01\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0b\x01\x00\x59\x00\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x88\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x67\x78\x03\x59\x00\x77\x01\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x07\x01\x00\x7c\x07\x63\x02\x59\x00\x63\x02\x01\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_toplevel_consts_52_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Verify arguments are \"sane\".",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_toplevel_consts_52_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "module name must be str, not ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib__bootstrap_toplevel_consts_52_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "level must be >= 0",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_toplevel_consts_52_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__package__ not set to a string",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
importlib__bootstrap_toplevel_consts_52_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "attempted relative import with no known parent package",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_toplevel_consts_52_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Empty module name",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_52_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_52_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_52_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_toplevel_consts_52_consts_3._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_52_consts_4._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_52_consts_5._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_52_consts_6._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_str = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "str",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_TypeError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "TypeError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_ValueError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ValueError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_52_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            & const_str_ValueError._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__sanity_check = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_sanity_check",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[133];
    }
importlib__bootstrap_toplevel_consts_52_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 132,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x15\x90\x64\x9c\x43\xd4\x0b\x20\xdc\x0e\x17\xd0\x1a\x37\xbc\x04\xb8\x54\xbb\x0a\xb0\x7c\xd0\x18\x44\xd3\x0e\x45\xd0\x08\x45\xd8\x07\x0c\x88\x71\x82\x79\xdc\x0e\x18\xd0\x19\x2d\xd3\x0e\x2e\xd0\x08\x2e\xd8\x07\x0c\x88\x71\x82\x79\xdc\x0f\x19\x98\x27\xa4\x33\xd4\x0f\x27\xdc\x12\x1b\xd0\x1c\x3d\xd3\x12\x3e\xd0\x0c\x3e\xd9\x11\x18\xdc\x12\x1d\xf0\x00\x01\x1f\x28\xf3\x00\x01\x13\x29\xf0\x00\x01\x0d\x29\xe1\x0b\x0f\x90\x45\x98\x51\x92\x4a\xdc\x0e\x18\xd0\x19\x2c\xd3\x0e\x2d\xd0\x08\x2d\xf0\x03\x00\x15\x1f\x88\x34",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_52_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_package._ascii.ob_base,
            &_Py_ID(level),
        },
    },
};
static
    struct _PyCode_DEF(242)
importlib__bootstrap_toplevel_consts_52 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 121,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_52_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_52_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1285,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 87,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_52_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__sanity_check._ascii.ob_base,
    .co_qualname = & const_str__sanity_check._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_52_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x17\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x64\x02\x6b\x02\x00\x00\x72\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x64\x02\x6b\x44\x00\x00\x72\x28\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x0b\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x73\x0b\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x73\x11\x7c\x02\x64\x02\x6b\x28\x00\x00\x72\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x79\x07\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
importlib__bootstrap_toplevel_consts_53 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "No module named ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
importlib__bootstrap_toplevel_consts_54 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "{!r}",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_toplevel_consts_55_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "; ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_toplevel_consts_55_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " is not a package",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_toplevel_consts_55_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Cannot set an attribute on ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib__bootstrap_toplevel_consts_55_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " for child module ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_55_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_toplevel_consts_55_consts_3._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_55_consts_4._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & importlib__bootstrap_toplevel_consts_55_consts_7._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_55_consts_8._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__ERR_MSG_PREFIX = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ERR_MSG_PREFIX",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_ModuleNotFoundError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleNotFoundError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[18];
        }_object;
    }
importlib__bootstrap_toplevel_consts_55_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 18,
        },
        .ob_item = {
            & const_str_rpartition._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str__call_with_frames_removed._ascii.ob_base,
            &_Py_ID(__path__),
            & const_str_AttributeError._ascii.ob_base,
            & const_str__ERR_MSG_PREFIX._ascii.ob_base,
            & const_str_ModuleNotFoundError._ascii.ob_base,
            &_Py_ID(__spec__),
            & const_str__find_spec._ascii.ob_base,
            &_Py_ID(_uninitialized_submodules),
            &_Py_ID(append),
            & const_str__load_unlocked._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            & const_str_setattr._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_ImportWarning._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str__find_and_load_unlocked = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_find_and_load_unlocked",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[434];
    }
importlib__bootstrap_toplevel_consts_55_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 433,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\x80\x44\xd8\x0d\x11\x8f\x5f\x89\x5f\x98\x53\xd3\x0d\x21\xa0\x21\xd1\x0d\x24\x80\x46\xd8\x12\x16\x80\x4b\xd9\x07\x0d\xd8\x0b\x11\x9c\x13\x9f\x1b\x99\x1b\xd1\x0b\x24\xdc\x0c\x25\xa0\x67\xa8\x76\xd4\x0c\x36\xe0\x0b\x0f\x94\x33\x97\x3b\x91\x3b\xd1\x0b\x1e\xdc\x13\x16\x97\x3b\x91\x3b\x98\x74\xd1\x13\x24\xd0\x0c\x24\xdc\x18\x1b\x9f\x0b\x99\x0b\xa0\x46\xd1\x18\x2b\x88\x0d\xf0\x02\x04\x09\x40\x01\xd8\x13\x20\xd7\x13\x29\xd1\x13\x29\x88\x44\xf0\x08\x00\x17\x24\xd7\x16\x2c\xd1\x16\x2c\x88\x0b\xd8\x10\x14\x97\x0f\x91\x0f\xa0\x03\xd3\x10\x24\xa0\x51\xd1\x10\x27\x88\x05\xdc\x0b\x15\x90\x64\x98\x44\xd3\x0b\x21\x80\x44\xd8\x07\x0b\x80\x7c\xdc\x0e\x21\xa4\x5f\xd0\x24\x35\xb0\x64\xb0\x58\xd0\x22\x3e\xc0\x54\xd4\x0e\x4a\xd0\x08\x4a\xe1\x0b\x16\xf0\x06\x00\x0d\x18\xd7\x0c\x31\xd1\x0c\x31\xd7\x0c\x38\xd1\x0c\x38\xb8\x15\xd4\x0c\x3f\xf0\x02\x04\x09\x3c\xdc\x15\x23\xa0\x44\xd3\x15\x29\x88\x46\xe1\x0f\x1a\xd8\x10\x1b\xd7\x10\x35\xd1\x10\x35\xd7\x10\x39\xd1\x10\x39\xd5\x10\x3b\xd9\x07\x0d\xe4\x18\x1b\x9f\x0b\x99\x0b\xa0\x46\xd1\x18\x2b\x88\x0d\xf0\x02\x04\x09\x2f\xdc\x0c\x13\x90\x4d\xa0\x35\xa8\x26\xd4\x0c\x31\xf0\x08\x00\x0c\x12\x80\x4d\x88\x36\x80\x4d\xf8\xf4\x35\x00\x10\x1e\xf2\x00\x02\x09\x40\x01\xdc\x15\x24\xd0\x14\x25\xa0\x64\xa0\x58\xa8\x52\xb0\x06\xa8\x7a\xd0\x39\x4a\xd0\x12\x4b\x88\x43\xdc\x12\x25\xa0\x63\xb0\x04\xd4\x12\x35\xb8\x34\xd0\x0c\x3f\xf0\x05\x02\x09\x40\x01\xfb\xf1\x20\x00\x10\x1b\xd8\x10\x1b\xd7\x10\x35\xd1\x10\x35\xd7\x10\x39\xd1\x10\x39\xd5\x10\x3b\xf0\x03\x00\x10\x1b\xfb\xf4\x0e\x00\x10\x1e\xf2\x00\x02\x09\x2f\xd8\x14\x2f\xb0\x06\xa8\x7a\xd0\x39\x4b\xc8\x45\xc8\x39\xd0\x12\x55\x88\x43\xdc\x0c\x15\x8f\x4e\x89\x4e\x98\x33\xa4\x0d\xd5\x0c\x2e\xd8\x0b\x11\x80\x4d\xf0\x07\x02\x09\x2f\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
importlib__bootstrap_toplevel_consts_55_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x32\x0c\x44\x2e\x00\xc3\x1f\x0b\x45\x17\x00\xc4\x1d\x0d\x45\x39\x00\xc4\x2e\x26\x45\x14\x03\xc5\x17\x1f\x45\x36\x03\xc5\x39\x2b\x46\x28\x03\xc6\x27\x01\x46\x28\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_import_ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "import_",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_parent_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "parent_spec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_parent_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "parent_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_child = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "child",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_toplevel_consts_55_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_import_._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(parent),
            & const_str_parent_spec._ascii.ob_base,
            & const_str_parent_module._ascii.ob_base,
            &_Py_ID(msg),
            & const_str_child._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
        },
    },
};
static
    struct _PyCode_DEF(854)
importlib__bootstrap_toplevel_consts_55 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 427,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_55_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_55_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_55_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1304,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 88,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_55_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__find_and_load_unlocked._ascii.ob_base,
    .co_qualname = & const_str__find_and_load_unlocked._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_55_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x7d\x02\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\x7d\x03\x64\x00\x7d\x04\x7c\x03\x72\x83\x7c\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x0c\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x13\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x19\x00\x00\x00\x53\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x19\x00\x00\x00\x7d\x05\x09\x00\x7c\x05\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x05\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x06\x19\x00\x00\x00\x7d\x07\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x80\x15\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x05\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x04\x72\x1b\x7c\x04\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7f\x07\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x09\x7c\x04\x72\x1b\x7c\x04\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7c\x03\x72\x23\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x19\x00\x00\x00\x7d\x05\x09\x00\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7f\x07\x7c\x09\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x09\x53\x00\x7c\x09\x53\x00\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x1d\x01\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x7c\x00\x9b\x02\x64\x03\x7c\x03\x9b\x02\x64\x04\x9d\x05\x7d\x06\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x00\xac\x05\xab\x02\x00\x00\x00\x00\x00\x00\x64\x00\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x7c\x04\x72\x1b\x7c\x04\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x77\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x26\x01\x00\x64\x07\x7c\x03\x9b\x02\x64\x08\x7f\x07\x9b\x02\x9d\x04\x7d\x06\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x7c\x09\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_56_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Find and load the module.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
importlib__bootstrap_toplevel_consts_56_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "import of ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_toplevel_consts_56_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " halted; None in sys.modules",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_56_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_56_consts_0._ascii.ob_base,
            &_Py_ID(__spec__),
            Py_None,
            &_Py_ID(_initializing),
            Py_False,
            & importlib__bootstrap_toplevel_consts_56_consts_5._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_56_consts_6._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__NEEDS_LOADING = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NEEDS_LOADING",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_56_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(get),
            & const_str__NEEDS_LOADING._ascii.ob_base,
            &_Py_ID(getattr),
            & const_str__ModuleLockManager._ascii.ob_base,
            & const_str__find_and_load_unlocked._ascii.ob_base,
            &_Py_ID(_lock_unlock_module),
            & const_str_ModuleNotFoundError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[173];
    }
importlib__bootstrap_toplevel_consts_56_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 172,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x0e\x11\x8f\x5b\x89\x5b\x8f\x5f\x89\x5f\x98\x54\xa4\x3e\xd3\x0d\x32\x80\x46\xd8\x08\x0e\x94\x2e\xd1\x08\x20\xdc\x08\x0f\x94\x07\x98\x06\xa0\x0a\xa8\x44\xd3\x10\x31\xb0\x3f\xc0\x45\xd4\x08\x4a\xdc\x0d\x1f\xa0\x04\xd5\x0d\x25\xdc\x15\x18\x97\x5b\x91\x5b\x97\x5f\x91\x5f\xa0\x54\xac\x3e\xd3\x15\x3a\x88\x46\xd8\x0f\x15\x9c\x1e\xd1\x0f\x27\xdc\x17\x2e\xa8\x74\xb0\x57\xd3\x17\x3d\xf7\x07\x00\x0e\x26\xd1\x0d\x25\xe0\x0f\x27\xf7\x05\x00\x0e\x26\xf4\x12\x00\x09\x1c\x98\x44\xd4\x08\x21\xe0\x07\x0d\x80\x7e\xd8\x14\x1e\x98\x74\x98\x66\xd0\x24\x40\xd0\x12\x41\x88\x07\xdc\x0e\x21\xa0\x27\xb0\x04\xd4\x0e\x35\xd0\x08\x35\xe0\x0b\x11\x80\x4d\xf7\x1f\x00\x0e\x26\xd0\x0d\x25\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_toplevel_consts_56_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x10\x38\x42\x3d\x03\xc2\x3d\x05\x43\x06\x07",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_56_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_import_._ascii.ob_base,
            &_Py_ID(module),
            &_Py_ID(message),
        },
    },
};
static
    struct _PyCode_DEF(402)
importlib__bootstrap_toplevel_consts_56 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 201,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_56_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_56_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_56_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1349,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 89,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_56_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(_find_and_load),
    .co_qualname = &_Py_ID(_find_and_load),
    .co_linetable = & importlib__bootstrap_toplevel_consts_56_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x73\x18\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x01\x64\x02\xab\x03\x00\x00\x00\x00\x00\x00\x64\x03\x64\x04\xab\x03\x00\x00\x00\x00\x00\x00\x72\x61\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x15\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x63\x02\x64\x02\x64\x02\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x53\x00\x09\x00\x64\x02\x64\x02\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x80\x13\x64\x05\x7c\x00\x9b\x00\x64\x06\x9d\x03\x7d\x03\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\xac\x07\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x2b\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[307];
    }
importlib__bootstrap_toplevel_consts_57_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 306,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x49\x6d\x70\x6f\x72\x74\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x62\x61\x73\x65\x64\x20\x6f\x6e\x20\x69\x74\x73\x20\x6e\x61\x6d\x65\x2c\x20\x74\x68\x65\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x74\x68\x65\x20\x63\x61\x6c\x6c\x20\x69\x73\x0a\x20\x20\x20\x20\x62\x65\x69\x6e\x67\x20\x6d\x61\x64\x65\x20\x66\x72\x6f\x6d\x2c\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x6c\x65\x76\x65\x6c\x20\x61\x64\x6a\x75\x73\x74\x6d\x65\x6e\x74\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x73\x20\x74\x68\x65\x20\x67\x72\x65\x61\x74\x65\x73\x74\x20\x63\x6f\x6d\x6d\x6f\x6e\x20\x64\x65\x6e\x6f\x6d\x69\x6e\x61\x74\x6f\x72\x20\x6f\x66\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x61\x6c\x69\x74\x79\x0a\x20\x20\x20\x20\x62\x65\x74\x77\x65\x65\x6e\x20\x69\x6d\x70\x6f\x72\x74\x5f\x6d\x6f\x64\x75\x6c\x65\x20\x61\x6e\x64\x20\x5f\x5f\x69\x6d\x70\x6f\x72\x74\x5f\x5f\x2e\x20\x54\x68\x69\x73\x20\x69\x6e\x63\x6c\x75\x64\x65\x73\x20\x73\x65\x74\x74\x69\x6e\x67\x20\x5f\x5f\x70\x61\x63\x6b\x61\x67\x65\x5f\x5f\x20\x69\x66\x0a\x20\x20\x20\x20\x74\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x20\x64\x69\x64\x20\x6e\x6f\x74\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_57_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_57_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__gcd_import = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_gcd_import",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_57_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__sanity_check._ascii.ob_base,
            & const_str__resolve_name._ascii.ob_base,
            &_Py_ID(_find_and_load),
            & const_str__gcd_import._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[51];
    }
importlib__bootstrap_toplevel_consts_57_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 50,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x12\x00\x05\x12\x90\x24\x98\x07\xa0\x15\xd4\x04\x27\xd8\x07\x0c\x88\x71\x82\x79\xdc\x0f\x1c\x98\x54\xa0\x37\xa8\x45\xd3\x0f\x32\x88\x04\xdc\x0b\x19\x98\x24\xa4\x0b\xd3\x0b\x2c\xd0\x04\x2c",
};
static
    struct _PyCode_DEF(96)
importlib__bootstrap_toplevel_consts_57 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 48,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_57_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_57_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1375,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 90,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_52_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__gcd_import._ascii.ob_base,
    .co_qualname = & const_str__gcd_import._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_57_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x64\x01\x6b\x44\x00\x00\x72\x0d\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_recursive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "recursive",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_58 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_recursive._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[239];
    }
importlib__bootstrap_toplevel_consts_59_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 238,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x46\x69\x67\x75\x72\x65\x20\x6f\x75\x74\x20\x77\x68\x61\x74\x20\x5f\x5f\x69\x6d\x70\x6f\x72\x74\x5f\x5f\x20\x73\x68\x6f\x75\x6c\x64\x20\x72\x65\x74\x75\x72\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x69\x6d\x70\x6f\x72\x74\x5f\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x20\x69\x73\x20\x61\x20\x63\x61\x6c\x6c\x61\x62\x6c\x65\x20\x77\x68\x69\x63\x68\x20\x74\x61\x6b\x65\x73\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x20\x6f\x66\x20\x6d\x6f\x64\x75\x6c\x65\x20\x74\x6f\x0a\x20\x20\x20\x20\x69\x6d\x70\x6f\x72\x74\x2e\x20\x49\x74\x20\x69\x73\x20\x72\x65\x71\x75\x69\x72\x65\x64\x20\x74\x6f\x20\x64\x65\x63\x6f\x75\x70\x6c\x65\x20\x74\x68\x65\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x66\x72\x6f\x6d\x20\x61\x73\x73\x75\x6d\x69\x6e\x67\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x27\x73\x0a\x20\x20\x20\x20\x69\x6d\x70\x6f\x72\x74\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x69\x73\x20\x64\x65\x73\x69\x72\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
importlib__bootstrap_toplevel_consts_59_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".__all__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
importlib__bootstrap_toplevel_consts_59_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "``from list''",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
importlib__bootstrap_toplevel_consts_59_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Item in ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib__bootstrap_toplevel_consts_59_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " must be str, not ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_toplevel_consts_59_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_59_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_59_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_59_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_59_consts_3._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_59_consts_4._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[42],
            &_Py_ID(__all__),
            Py_True,
            & importlib__bootstrap_toplevel_consts_58._object.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
importlib__bootstrap_toplevel_consts_59_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            &_Py_ID(__name__),
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            & const_str_hasattr._ascii.ob_base,
            &_Py_ID(_handle_fromlist),
            &_Py_ID(__all__),
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str_ModuleNotFoundError._ascii.ob_base,
            &_Py_ID(name),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(get),
            & const_str__NEEDS_LOADING._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[244];
    }
importlib__bootstrap_toplevel_consts_59_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 243,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf3\x14\x00\x0e\x16\x88\x01\xdc\x0f\x19\x98\x21\x9c\x53\xd4\x0f\x21\xd9\x0f\x18\xd8\x18\x1e\x9f\x0f\x99\x0f\xa8\x2a\xd1\x18\x34\x91\x05\xe0\x18\x27\x90\x05\xdc\x12\x1b\x98\x68\xa0\x75\xa0\x67\xf0\x00\x01\x2e\x23\xdc\x23\x27\xa8\x01\xa3\x37\xd7\x23\x33\xd1\x23\x33\xd0\x22\x34\xf0\x03\x01\x1d\x36\xf3\x00\x01\x13\x37\xf0\x00\x01\x0d\x37\xe0\x0d\x0e\x90\x23\x8a\x58\xda\x13\x1c\xa4\x17\xa8\x16\xb0\x19\xd5\x21\x3b\xdc\x10\x20\xa0\x16\xa8\x16\xaf\x1e\xa9\x1e\xb8\x17\xd8\x2b\x2f\xf6\x03\x01\x11\x31\xe4\x11\x18\x98\x16\xa0\x11\xd5\x11\x23\xd8\x1b\x21\x9f\x3f\x99\x3f\xd0\x1a\x2b\xa8\x31\xa8\x51\xa8\x43\xd0\x18\x30\x88\x49\xf0\x02\x09\x0d\x16\xdc\x10\x29\xa8\x27\xb0\x39\xd5\x10\x3d\xf0\x1f\x00\x0e\x16\xf0\x30\x00\x0c\x12\x80\x4d\xf8\xf4\x11\x00\x14\x27\xf2\x00\x07\x0d\x16\xf0\x08\x00\x15\x18\x97\x48\x91\x48\xa0\x09\xd2\x14\x29\xdc\x14\x17\x97\x4b\x91\x4b\x97\x4f\x91\x4f\xa0\x49\xac\x7e\xd3\x14\x3e\xd0\x14\x4a\xdc\x14\x1c\xd8\x10\x15\xfb\xf0\x0f\x07\x0d\x16\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
importlib__bootstrap_toplevel_consts_59_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x1c\x0c\x42\x2c\x02\xc2\x2c\x09\x43\x33\x05\xc2\x35\x33\x43\x2e\x05\xc3\x2d\x01\x43\x2e\x05\xc3\x2e\x05\x43\x33\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_where = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "where",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_from_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "from_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_exc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exc",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_toplevel_consts_59_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(module),
            &_Py_ID(fromlist),
            & const_str_import_._ascii.ob_base,
            & const_str_recursive._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
            & const_str_where._ascii.ob_base,
            & const_str_from_name._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(492)
importlib__bootstrap_toplevel_consts_59 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 246,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_59_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_59_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_toplevel_consts_59_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 1390,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 91,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_59_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(_handle_fromlist),
    .co_qualname = &_Py_ID(_handle_fromlist),
    .co_linetable = & importlib__bootstrap_toplevel_consts_59_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x44\x00\x5d\xa4\x00\x00\x7d\x04\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x38\x7c\x03\x72\x10\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7a\x00\x00\x00\x7d\x05\x6e\x02\x64\x02\x7d\x05\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x05\x9b\x00\x64\x04\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x04\x64\x05\x6b\x28\x00\x00\x72\x2a\x7c\x03\x72\x01\x8c\x53\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x60\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x07\xac\x08\xab\x04\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x7a\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x72\x01\x8c\x87\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x09\x7c\x04\x9b\x00\x9d\x03\x7d\x06\x09\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x06\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x8c\xa6\x04\x00\x7c\x00\x53\x00\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x3e\x7d\x07\x7c\x07\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x6b\x28\x00\x00\x72\x29\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x81\x05\x59\x00\x64\x0a\x7d\x07\x7e\x07\x8c\xea\x82\x00\x64\x0a\x7d\x07\x7e\x07\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[168];
    }
importlib__bootstrap_toplevel_consts_60_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 167,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x61\x6c\x63\x75\x6c\x61\x74\x65\x20\x77\x68\x61\x74\x20\x5f\x5f\x70\x61\x63\x6b\x61\x67\x65\x5f\x5f\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x2e\x0a\x0a\x20\x20\x20\x20\x5f\x5f\x70\x61\x63\x6b\x61\x67\x65\x5f\x5f\x20\x69\x73\x20\x6e\x6f\x74\x20\x67\x75\x61\x72\x61\x6e\x74\x65\x65\x64\x20\x74\x6f\x20\x62\x65\x20\x64\x65\x66\x69\x6e\x65\x64\x20\x6f\x72\x20\x63\x6f\x75\x6c\x64\x20\x62\x65\x20\x73\x65\x74\x20\x74\x6f\x20\x4e\x6f\x6e\x65\x0a\x20\x20\x20\x20\x74\x6f\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x20\x74\x68\x61\x74\x20\x69\x74\x73\x20\x70\x72\x6f\x70\x65\x72\x20\x76\x61\x6c\x75\x65\x20\x69\x73\x20\x75\x6e\x6b\x6e\x6f\x77\x6e\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib__bootstrap_toplevel_consts_60_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__package__ != __spec__.parent (",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
importlib__bootstrap_toplevel_consts_60_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " != ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_60_consts_7 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(stacklevel),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[90];
    }
importlib__bootstrap_toplevel_consts_60_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 89,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "can't resolve package from __spec__ or __package__, falling back on __name__ and __path__",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
importlib__bootstrap_toplevel_consts_60_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_60_consts_0._ascii.ob_base,
            &_Py_ID(__package__),
            &_Py_ID(__spec__),
            & importlib__bootstrap_toplevel_consts_60_consts_3._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_60_consts_4._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[41],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            & importlib__bootstrap_toplevel_consts_60_consts_7._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_60_consts_8._ascii.ob_base,
            &_Py_ID(__name__),
            &_Py_ID(__path__),
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_toplevel_consts_60_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(get),
            &_Py_ID(parent),
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_DeprecationWarning._ascii.ob_base,
            & const_str_ImportWarning._ascii.ob_base,
            & const_str_rpartition._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__calc___package__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_calc___package__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[201];
    }
importlib__bootstrap_toplevel_consts_60_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 200,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0e\x00\x0f\x16\x8f\x6b\x89\x6b\x98\x2d\xd3\x0e\x28\x80\x47\xd8\x0b\x12\x8f\x3b\x89\x3b\x90\x7a\xd3\x0b\x22\x80\x44\xd8\x07\x0e\xd0\x07\x1a\xd8\x0b\x0f\xd0\x0b\x1b\xa0\x07\xa8\x34\xaf\x3b\xa9\x3b\xd2\x20\x36\xdc\x0c\x15\x8f\x4e\x89\x4e\xf0\x00\x01\x1c\x1f\xd8\x1f\x26\x98\x6b\xa8\x14\xa8\x64\xaf\x6b\xa9\x6b\xa8\x5f\xb8\x41\xf0\x03\x01\x1c\x3f\xe4\x1b\x2d\xb8\x21\xf0\x05\x00\x0d\x1b\xf4\x00\x02\x0d\x3d\xf0\x06\x00\x10\x17\x88\x0e\xd8\x09\x0d\xd0\x09\x19\xd8\x0f\x13\x8f\x7b\x89\x7b\xd0\x08\x1a\xe4\x08\x11\x8f\x0e\x89\x0e\xf0\x00\x01\x18\x3f\xe4\x17\x24\xb0\x11\xf0\x05\x00\x09\x17\xf4\x00\x02\x09\x34\xf0\x06\x00\x13\x1a\x98\x2a\xd1\x12\x25\x88\x07\xd8\x0b\x15\x98\x57\xd1\x0b\x24\xd8\x16\x1d\xd7\x16\x28\xd1\x16\x28\xa8\x13\xd3\x16\x2d\xa8\x61\xd1\x16\x30\x88\x47\xd8\x0b\x12\x80\x4e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_60_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(globals),
            & const_str_package._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(348)
importlib__bootstrap_toplevel_consts_60 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 174,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_60_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_60_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1427,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 92,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_60_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__calc___package__._ascii.ob_base,
    .co_qualname = & const_str__calc___package__._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_60_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x01\x81\x40\x7c\x02\x81\x3c\x7c\x01\x7c\x02\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x2d\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x01\x9b\x02\x64\x04\x7c\x02\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x05\x9d\x05\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xac\x07\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00\x7c\x02\x81\x0c\x7c\x02\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xac\x07\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x64\x09\x19\x00\x00\x00\x7d\x01\x64\x0a\x7c\x00\x76\x01\x72\x14\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x64\x0c\x19\x00\x00\x00\x7d\x01\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[472];
    }
importlib__bootstrap_toplevel_consts_61_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 471,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x49\x6d\x70\x6f\x72\x74\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x27\x67\x6c\x6f\x62\x61\x6c\x73\x27\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x69\x73\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x69\x6e\x66\x65\x72\x20\x77\x68\x65\x72\x65\x20\x74\x68\x65\x20\x69\x6d\x70\x6f\x72\x74\x20\x69\x73\x20\x6f\x63\x63\x75\x72\x72\x69\x6e\x67\x20\x66\x72\x6f\x6d\x0a\x20\x20\x20\x20\x74\x6f\x20\x68\x61\x6e\x64\x6c\x65\x20\x72\x65\x6c\x61\x74\x69\x76\x65\x20\x69\x6d\x70\x6f\x72\x74\x73\x2e\x20\x54\x68\x65\x20\x27\x6c\x6f\x63\x61\x6c\x73\x27\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x69\x73\x20\x69\x67\x6e\x6f\x72\x65\x64\x2e\x20\x54\x68\x65\x0a\x20\x20\x20\x20\x27\x66\x72\x6f\x6d\x6c\x69\x73\x74\x27\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x73\x70\x65\x63\x69\x66\x69\x65\x73\x20\x77\x68\x61\x74\x20\x73\x68\x6f\x75\x6c\x64\x20\x65\x78\x69\x73\x74\x20\x61\x73\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x73\x20\x6f\x6e\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x0a\x20\x20\x20\x20\x62\x65\x69\x6e\x67\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x20\x28\x65\x2e\x67\x2e\x20\x60\x60\x66\x72\x6f\x6d\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x6d\x70\x6f\x72\x74\x20\x3c\x66\x72\x6f\x6d\x6c\x69\x73\x74\x3e\x60\x60\x29\x2e\x20\x20\x54\x68\x65\x20\x27\x6c\x65\x76\x65\x6c\x27\x0a\x20\x20\x20\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x73\x20\x74\x68\x65\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x20\x74\x6f\x20\x69\x6d\x70\x6f\x72\x74\x20\x66\x72\x6f\x6d\x20\x69\x6e\x20\x61\x20\x72\x65\x6c\x61\x74\x69\x76\x65\x0a\x20\x20\x20\x20\x69\x6d\x70\x6f\x72\x74\x20\x28\x65\x2e\x67\x2e\x20\x60\x60\x66\x72\x6f\x6d\x20\x2e\x2e\x70\x6b\x67\x20\x69\x6d\x70\x6f\x72\x74\x20\x6d\x6f\x64\x60\x60\x20\x77\x6f\x75\x6c\x64\x20\x68\x61\x76\x65\x20\x61\x20\x27\x6c\x65\x76\x65\x6c\x27\x20\x6f\x66\x20\x32\x29\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_toplevel_consts_61_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_61_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            &_Py_ID(__path__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_partition = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "partition",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_61_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str__gcd_import._ascii.ob_base,
            & const_str__calc___package__._ascii.ob_base,
            & const_str_partition._ascii.ob_base,
            &_Py_ID(len),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(__name__),
            & const_str_hasattr._ascii.ob_base,
            &_Py_ID(_handle_fromlist),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[211];
    }
importlib__bootstrap_toplevel_consts_61_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 210,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x16\x00\x08\x0d\x90\x01\x82\x7a\xdc\x11\x1c\x98\x54\xd3\x11\x22\x89\x06\xe0\x1e\x25\xd0\x1e\x31\x91\x37\xb0\x72\x88\x08\xdc\x12\x23\xa0\x48\xd3\x12\x2d\x88\x07\xdc\x11\x1c\x98\x54\xa0\x37\xa8\x45\xd3\x11\x32\x88\x06\xd9\x0b\x13\xf0\x06\x00\x0c\x11\x90\x41\x8a\x3a\xdc\x13\x1e\x98\x74\x9f\x7e\x99\x7e\xa8\x63\xd3\x1f\x32\xb0\x31\xd1\x1f\x35\xd3\x13\x36\xd0\x0c\x36\xd9\x11\x15\xd8\x13\x19\x88\x4d\xf4\x08\x00\x17\x1a\x98\x24\x93\x69\xa4\x23\xa0\x64\xa7\x6e\xa1\x6e\xb0\x53\xd3\x26\x39\xb8\x21\xd1\x26\x3c\xd3\x22\x3d\xd1\x16\x3d\x88\x47\xf4\x06\x00\x14\x17\x97\x3b\x91\x3b\x98\x76\x9f\x7f\x99\x7f\xd0\x2f\x4c\xb4\x03\xb0\x46\xb7\x4f\xb1\x4f\xd3\x30\x44\xc0\x57\xd1\x30\x4c\xd0\x1f\x4d\xd1\x13\x4e\xd0\x0c\x4e\xdc\x09\x10\x90\x16\x98\x1a\xd4\x09\x24\xdc\x0f\x1f\xa0\x06\xa8\x08\xb4\x2b\xd3\x0f\x3e\xd0\x08\x3e\xe0\x0f\x15\x88\x0d",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_globals_ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "globals_",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_cut_off = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cut_off",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_toplevel_consts_61_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(globals),
            &_Py_ID(locals),
            &_Py_ID(fromlist),
            &_Py_ID(level),
            &_Py_ID(module),
            & const_str_globals_._ascii.ob_base,
            & const_str_package._ascii.ob_base,
            & const_str_cut_off._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
importlib__bootstrap_toplevel_consts_61_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "         ",
};
static
    struct _PyCode_DEF(428)
importlib__bootstrap_toplevel_consts_61 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 214,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_61_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_61_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1454,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 93,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_61_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__import__),
    .co_qualname = &_Py_ID(__import__),
    .co_linetable = & importlib__bootstrap_toplevel_consts_61_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x04\x64\x01\x6b\x28\x00\x00\x72\x0c\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x6e\x1e\x7c\x01\x81\x02\x7c\x01\x6e\x01\x69\x00\x7d\x06\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x07\x7c\x04\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x03\x73\x85\x7c\x04\x64\x01\x6b\x28\x00\x00\x72\x1d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x73\x02\x7c\x05\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x0a\x00\x00\x7d\x08\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x08\x7a\x0a\x00\x00\x1a\x00\x19\x00\x00\x00\x53\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x64\x04\xab\x02\x00\x00\x00\x00\x00\x00\x72\x11\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x03\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x05\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_toplevel_consts_62_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "no built-in module named ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_62_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_62_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_62_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_BuiltinImporter._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__load_unlocked._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__builtin_from_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_builtin_from_name",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
importlib__bootstrap_toplevel_consts_62_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0b\x1a\xd7\x0b\x24\xd1\x0b\x24\xa0\x54\xd3\x0b\x2a\x80\x44\xd8\x07\x0b\x80\x7c\xdc\x0e\x19\xd0\x1a\x35\xb8\x04\xd1\x1a\x3c\xd3\x0e\x3d\xd0\x08\x3d\xdc\x0b\x19\x98\x24\xd3\x0b\x1f\xd0\x04\x1f",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_62_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(98)
importlib__bootstrap_toplevel_consts_62 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_62_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_62_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1491,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 94,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_62_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__builtin_from_name._ascii.ob_base,
    .co_qualname = & const_str__builtin_from_name._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_62_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x80\x0e\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x7a\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[251];
    }
importlib__bootstrap_toplevel_consts_63_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 250,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x65\x74\x75\x70\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x20\x62\x79\x20\x69\x6d\x70\x6f\x72\x74\x69\x6e\x67\x20\x6e\x65\x65\x64\x65\x64\x20\x62\x75\x69\x6c\x74\x2d\x69\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x61\x6e\x64\x20\x69\x6e\x6a\x65\x63\x74\x69\x6e\x67\x20\x74\x68\x65\x6d\x0a\x20\x20\x20\x20\x69\x6e\x74\x6f\x20\x74\x68\x65\x20\x67\x6c\x6f\x62\x61\x6c\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x41\x73\x20\x73\x79\x73\x20\x69\x73\x20\x6e\x65\x65\x64\x65\x64\x20\x66\x6f\x72\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x20\x61\x63\x63\x65\x73\x73\x20\x61\x6e\x64\x20\x5f\x69\x6d\x70\x20\x69\x73\x20\x6e\x65\x65\x64\x65\x64\x20\x74\x6f\x20\x6c\x6f\x61\x64\x20\x62\x75\x69\x6c\x74\x2d\x69\x6e\x0a\x20\x20\x20\x20\x6d\x6f\x64\x75\x6c\x65\x73\x2c\x20\x74\x68\x6f\x73\x65\x20\x74\x77\x6f\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x65\x78\x70\x6c\x69\x63\x69\x74\x6c\x79\x20\x70\x61\x73\x73\x65\x64\x20\x69\x6e\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_63_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__thread._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str__weakref._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_63_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_63_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_63_consts_1._object.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[18];
        }_object;
    }
importlib__bootstrap_toplevel_consts_63_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 18,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(type),
            &_Py_ID(modules),
            &_Py_ID(items),
            &_Py_ID(isinstance),
            & const_str_builtin_module_names._ascii.ob_base,
            & const_str_BuiltinImporter._ascii.ob_base,
            & const_str_is_frozen._ascii.ob_base,
            & const_str_FrozenImporter._ascii.ob_base,
            & const_str__spec_from_module._ascii.ob_base,
            & const_str__init_module_attrs._ascii.ob_base,
            &_Py_ID(_fix_up_module),
            &_Py_ID(__name__),
            & const_str__builtin_from_name._ascii.ob_base,
            & const_str_setattr._ascii.ob_base,
            & const_str__WeakValueDictionary._ascii.ob_base,
            & const_str__blocking_on._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__setup = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_setup",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[230];
    }
importlib__bootstrap_toplevel_consts_63_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 229,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x12\x00\x0c\x17\x80\x44\xd8\x0a\x14\x80\x43\xf4\x06\x00\x13\x17\x94\x73\x93\x29\x80\x4b\xdc\x18\x1b\x9f\x0b\x99\x0b\xd7\x18\x29\xd1\x18\x29\xd6\x18\x2b\x89\x0c\x88\x04\x88\x66\xdc\x0b\x15\x90\x66\x98\x6b\xd5\x0b\x2a\xd8\x0f\x13\x94\x73\xd7\x17\x2f\xd1\x17\x2f\xd1\x0f\x2f\xdc\x19\x28\x91\x06\xdc\x11\x15\x97\x1e\x91\x1e\xa0\x04\xd4\x11\x25\xdc\x19\x27\x91\x06\xe0\x10\x18\xdc\x13\x24\xa0\x56\xa8\x56\xd3\x13\x34\x88\x44\xdc\x0c\x1e\x98\x74\xa0\x56\xd4\x0c\x2c\xd8\x0f\x15\x9c\x1e\xd2\x0f\x27\xd8\x10\x16\xd7\x10\x25\xd1\x10\x25\xa0\x66\xd5\x10\x2d\xf0\x17\x00\x19\x2c\xf4\x1c\x00\x13\x16\x97\x2b\x91\x2b\x9c\x68\xd1\x12\x27\x80\x4b\xdb\x18\x3c\x88\x0c\xd8\x0b\x17\x9c\x73\x9f\x7b\x99\x7b\xd1\x0b\x2a\xdc\x1d\x2f\xb0\x0c\xd3\x1d\x3d\x89\x4e\xe4\x1d\x20\x9f\x5b\x99\x5b\xa8\x1c\xd1\x1d\x36\x88\x4e\xdc\x08\x0f\x90\x0b\x98\x5c\xa8\x3e\xd5\x08\x3a\xf0\x0b\x00\x19\x3d\xf4\x10\x00\x14\x28\xd3\x13\x29\x81\x4c",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_sys_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__imp_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_imp_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_module_type = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "module_type",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_self_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "self_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_builtin_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "builtin_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_builtin_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "builtin_module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_toplevel_consts_63_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_sys_module._ascii.ob_base,
            & const_str__imp_module._ascii.ob_base,
            & const_str_module_type._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(module),
            & const_str_loader._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            & const_str_self_module._ascii.ob_base,
            & const_str_builtin_name._ascii.ob_base,
            & const_str_builtin_module._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(556)
importlib__bootstrap_toplevel_consts_63 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 278,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_63_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_63_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1498,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 95,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_63_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__setup._ascii.ob_base,
    .co_qualname = & const_str__setup._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_63_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x61\x00\x7c\x00\x61\x01\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x7a\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x13\x7c\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x07\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x6e\x1d\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x72\x07\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x6e\x01\x8c\x49\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x05\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x73\x01\x8c\x6a\x7c\x05\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x7c\x04\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x7d\x07\x64\x01\x44\x00\x5d\x40\x00\x00\x7d\x08\x7c\x08\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x0c\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x09\x6e\x13\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\x19\x00\x00\x00\x7d\x09\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x7c\x08\x7c\x09\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x42\x04\x00\x74\x21\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x61\x11\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[49];
    }
importlib__bootstrap_toplevel_consts_64_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 48,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Install importers for builtin and frozen modules",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_64_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_64_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_64_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__setup._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_meta_path._ascii.ob_base,
            &_Py_ID(append),
            & const_str_BuiltinImporter._ascii.ob_base,
            & const_str_FrozenImporter._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__install = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_install",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
importlib__bootstrap_toplevel_consts_64_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x04\x0a\x88\x3a\x90\x7b\xd4\x04\x23\xe4\x04\x07\x87\x4d\x81\x4d\xd7\x04\x18\xd1\x04\x18\x9c\x1f\xd4\x04\x29\xdc\x04\x07\x87\x4d\x81\x4d\xd7\x04\x18\xd1\x04\x18\x9c\x1e\xd5\x04\x28",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_toplevel_consts_64_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_sys_module._ascii.ob_base,
            & const_str__imp_module._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(168)
importlib__bootstrap_toplevel_consts_64 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 84,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_64_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_64_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1538,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 96,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_64_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__install._ascii.ob_base,
    .co_qualname = & const_str__install._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_64_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[58];
    }
importlib__bootstrap_toplevel_consts_65_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 57,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Install importers that require external filesystem access",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_toplevel_consts_65_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_65_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str__frozen_importlib_external = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_frozen_importlib_external",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_toplevel_consts_65_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__frozen_importlib_external._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__install._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
const_str__install_external_importers = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_install_external_importers",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
importlib__bootstrap_toplevel_consts_65_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf3\x06\x00\x05\x26\xd8\x1a\x34\xd0\x04\x17\xd8\x04\x1e\xd7\x04\x27\xd1\x04\x27\xac\x03\xaf\x0b\xa9\x0b\xb4\x48\xd1\x28\x3d\xd5\x04\x3e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_toplevel_consts_65_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__frozen_importlib_external._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(92)
importlib__bootstrap_toplevel_consts_65 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 46,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_65_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_65_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1546,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 97,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_65_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__install_external_importers._ascii.ob_base,
    .co_qualname = & const_str__install_external_importers._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_65_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x7d\x00\x7c\x00\x61\x01\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_toplevel_consts_69 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            Py_None,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[70];
        }_object;
    }
importlib__bootstrap_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 70,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_0._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_1.ob_base.ob_base,
            Py_None,
            & importlib__bootstrap_toplevel_consts_3.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_4.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_5.ob_base.ob_base,
            & const_str__List._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_7.ob_base.ob_base,
            & const_str__WeakValueDictionary._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_9.ob_base.ob_base,
            & const_str__BlockingOnManager._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_11.ob_base.ob_base,
            & const_str__DeadlockError._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_13.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_14.ob_base.ob_base,
            & const_str__ModuleLock._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_16.ob_base.ob_base,
            & const_str__DummyModuleLock._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_18.ob_base.ob_base,
            & const_str__ModuleLockManager._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_20.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_21.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_22.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & importlib__bootstrap_toplevel_consts_24._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_25.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_26.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_27.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_28.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_29.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_30.ob_base.ob_base,
            & const_str_ModuleSpec._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_32._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_33.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_34.ob_base.ob_base,
            Py_False,
            & importlib__bootstrap_toplevel_consts_36._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_37.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_38.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_39.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_40.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_41.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_42.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_43.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44.ob_base.ob_base,
            & const_str_BuiltinImporter._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46.ob_base.ob_base,
            & const_str_FrozenImporter._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_48.ob_base.ob_base,
            & const_str__ImportLockContext._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_50.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_51.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_52.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_53._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_54._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_55.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_56.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_57.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_58._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_59.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_60.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_61.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_62.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_63.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_64.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_65.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_69._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__ERR_MSG = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ERR_MSG",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[58];
        }_object;
    }
importlib__bootstrap_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 58,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str__object_name._ascii.ob_base,
            & const_str__thread._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str__weakref._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__wrap._ascii.ob_base,
            & const_str__new_module._ascii.ob_base,
            & const_str_list._ascii.ob_base,
            & const_str__List._ascii.ob_base,
            & const_str__WeakValueDictionary._ascii.ob_base,
            & const_str__module_locks._ascii.ob_base,
            & const_str__blocking_on._ascii.ob_base,
            & const_str__BlockingOnManager._ascii.ob_base,
            & const_str_RuntimeError._ascii.ob_base,
            & const_str__DeadlockError._ascii.ob_base,
            & const_str__has_deadlocked._ascii.ob_base,
            & const_str__ModuleLock._ascii.ob_base,
            & const_str__DummyModuleLock._ascii.ob_base,
            & const_str__ModuleLockManager._ascii.ob_base,
            & const_str__get_module_lock._ascii.ob_base,
            &_Py_ID(_lock_unlock_module),
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str__verbose_message._ascii.ob_base,
            & const_str__requires_builtin._ascii.ob_base,
            & const_str__requires_frozen._ascii.ob_base,
            & const_str__load_module_shim._ascii.ob_base,
            & const_str__module_repr._ascii.ob_base,
            & const_str_ModuleSpec._ascii.ob_base,
            & const_str_spec_from_loader._ascii.ob_base,
            & const_str__spec_from_module._ascii.ob_base,
            & const_str__init_module_attrs._ascii.ob_base,
            & const_str_module_from_spec._ascii.ob_base,
            & const_str__module_repr_from_spec._ascii.ob_base,
            & const_str__exec._ascii.ob_base,
            & const_str__load_backward_compatible._ascii.ob_base,
            & const_str__load_unlocked._ascii.ob_base,
            & const_str__load._ascii.ob_base,
            & const_str_BuiltinImporter._ascii.ob_base,
            & const_str_FrozenImporter._ascii.ob_base,
            & const_str__ImportLockContext._ascii.ob_base,
            & const_str__resolve_name._ascii.ob_base,
            & const_str__find_spec._ascii.ob_base,
            & const_str__sanity_check._ascii.ob_base,
            & const_str__ERR_MSG_PREFIX._ascii.ob_base,
            & const_str__ERR_MSG._ascii.ob_base,
            & const_str__find_and_load_unlocked._ascii.ob_base,
            &_Py_ID(object),
            & const_str__NEEDS_LOADING._ascii.ob_base,
            &_Py_ID(_find_and_load),
            & const_str__gcd_import._ascii.ob_base,
            &_Py_ID(_handle_fromlist),
            & const_str__calc___package__._ascii.ob_base,
            &_Py_ID(__import__),
            & const_str__builtin_from_name._ascii.ob_base,
            & const_str__setup._ascii.ob_base,
            & const_str__install._ascii.ob_base,
            & const_str__install_external_importers._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[411];
    }
importlib__bootstrap_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 410,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x07\x01\x04\xf2\x2c\x04\x01\x26\xf0\x12\x00\x0b\x0f\x80\x07\xd8\x0c\x10\x80\x09\xd8\x0b\x0f\x80\x08\xf0\x06\x00\x17\x1b\xd0\x00\x13\xf2\x06\x05\x01\x26\xf2\x10\x01\x01\x1b\xf4\x0e\x01\x01\x09\x88\x44\xf4\x00\x01\x01\x09\xf7\x0e\x49\x01\x01\x15\xf1\x00\x49\x01\x01\x15\xf0\x5c\x02\x00\x11\x13\x80\x0d\xf0\x1a\x00\x10\x14\x80\x0c\xf7\x06\x13\x01\x2a\xf1\x00\x13\x01\x2a\xf4\x2c\x01\x01\x09\x90\x5c\xf4\x00\x01\x01\x09\xf2\x0a\x28\x01\x11\xf7\x56\x01\x60\x02\x01\x3b\xf1\x00\x60\x02\x01\x3b\xf7\x46\x05\x12\x01\x40\x01\xf1\x00\x12\x01\x40\x01\xf7\x2a\x0b\x01\x1d\xf1\x00\x0b\x01\x1d\xf2\x20\x22\x01\x10\xf2\x4a\x01\x0e\x01\x17\xf2\x22\x08\x01\x1c\xf0\x16\x00\x30\x31\xf4\x00\x05\x01\x36\xf2\x10\x08\x01\x25\xf2\x16\x08\x01\x24\xf2\x18\x0f\x01\x1b\xf2\x26\x12\x01\x36\xf7\x2a\x61\x01\x01\x29\xf1\x00\x61\x01\x01\x29\xf0\x48\x03\x00\x2e\x32\xb8\x64\xf4\x00\x1a\x01\x4a\x01\xf3\x3a\x27\x01\x10\xf0\x54\x01\x00\x32\x37\xf4\x00\x46\x01\x01\x12\xf2\x52\x02\x0e\x01\x12\xf2\x22\x12\x01\x3d\xf2\x2c\x1b\x01\x12\xf2\x3c\x22\x01\x12\xf2\x48\x01\x29\x01\x12\xf2\x5a\x01\x0a\x01\x24\xf7\x1e\x31\x01\x31\xf1\x00\x31\x01\x31\xf7\x68\x01\x7e\x02\x01\x30\xf1\x00\x7e\x02\x01\x30\xf7\x46\x06\x0a\x01\x1c\xf1\x00\x0a\x01\x1c\xf2\x1a\x06\x01\x2e\xf3\x12\x2a\x01\x14\xf2\x5a\x01\x0d\x01\x2e\xf0\x20\x00\x13\x25\x80\x0f\xd8\x0b\x1a\x98\x56\xd1\x0b\x23\x80\x08\xf2\x04\x27\x01\x12\xf1\x54\x01\x00\x12\x18\x93\x18\x80\x0e\xf2\x06\x17\x01\x12\xf3\x34\x0c\x01\x2d\xf0\x1e\x00\x3e\x43\x01\xf4\x00\x22\x01\x12\xf2\x4a\x01\x18\x01\x13\xf3\x36\x22\x01\x16\xf2\x4a\x01\x04\x01\x20\xf2\x0e\x25\x01\x2a\xf2\x50\x01\x05\x01\x29\xf3\x10\x05\x01\x3f",
};
static
    struct _PyCode_DEF(522)
importlib__bootstrap_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 261,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 98,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & importlib__bootstrap_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x84\x00\x5a\x01\x64\x02\x5a\x02\x64\x02\x5a\x03\x64\x02\x5a\x04\x64\x02\x61\x05\x64\x03\x84\x00\x5a\x06\x64\x04\x84\x00\x5a\x07\x02\x00\x47\x00\x64\x05\x84\x00\x64\x06\x65\x08\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x09\x02\x00\x47\x00\x64\x07\x84\x00\x64\x08\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x0a\x69\x00\x5a\x0b\x64\x02\x61\x0c\x02\x00\x47\x00\x64\x09\x84\x00\x64\x0a\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x0d\x02\x00\x47\x00\x64\x0b\x84\x00\x64\x0c\x65\x0e\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x0f\x64\x0d\x84\x00\x5a\x10\x02\x00\x47\x00\x64\x0e\x84\x00\x64\x0f\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x11\x02\x00\x47\x00\x64\x10\x84\x00\x64\x11\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x12\x02\x00\x47\x00\x64\x12\x84\x00\x64\x13\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x13\x64\x14\x84\x00\x5a\x14\x64\x15\x84\x00\x5a\x15\x64\x16\x84\x00\x5a\x16\x64\x17\x64\x18\x9c\x01\x64\x19\x84\x02\x5a\x17\x64\x1a\x84\x00\x5a\x18\x64\x1b\x84\x00\x5a\x19\x64\x1c\x84\x00\x5a\x1a\x64\x1d\x84\x00\x5a\x1b\x02\x00\x47\x00\x64\x1e\x84\x00\x64\x1f\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x1c\x64\x02\x64\x02\x64\x20\x9c\x02\x64\x21\x84\x02\x5a\x1d\x64\x42\x64\x22\x84\x01\x5a\x1e\x64\x23\x64\x24\x9c\x01\x64\x25\x84\x02\x5a\x1f\x64\x26\x84\x00\x5a\x20\x64\x27\x84\x00\x5a\x21\x64\x28\x84\x00\x5a\x22\x64\x29\x84\x00\x5a\x23\x64\x2a\x84\x00\x5a\x24\x64\x2b\x84\x00\x5a\x25\x02\x00\x47\x00\x64\x2c\x84\x00\x64\x2d\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x26\x02\x00\x47\x00\x64\x2e\x84\x00\x64\x2f\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x27\x02\x00\x47\x00\x64\x30\x84\x00\x64\x31\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x28\x64\x32\x84\x00\x5a\x29\x64\x43\x64\x33\x84\x01\x5a\x2a\x64\x34\x84\x00\x5a\x2b\x64\x35\x5a\x2c\x65\x2c\x64\x36\x7a\x00\x00\x00\x5a\x2d\x64\x37\x84\x00\x5a\x2e\x02\x00\x65\x2f\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x30\x64\x38\x84\x00\x5a\x31\x64\x44\x64\x39\x84\x01\x5a\x32\x64\x23\x64\x3a\x9c\x01\x64\x3b\x84\x02\x5a\x33\x64\x3c\x84\x00\x5a\x34\x64\x45\x64\x3d\x84\x01\x5a\x35\x64\x3e\x84\x00\x5a\x36\x64\x3f\x84\x00\x5a\x37\x64\x40\x84\x00\x5a\x38\x64\x41\x84\x00\x5a\x39\x79\x02",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_importlib__bootstrap_toplevel(void)
{
    return Py_NewRef((PyObject *) &importlib__bootstrap_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[351];
    }
importlib__bootstrap_external_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 350,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x72\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x70\x61\x74\x68\x2d\x62\x61\x73\x65\x64\x20\x69\x6d\x70\x6f\x72\x74\x2e\x0a\x0a\x54\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x4e\x4f\x54\x20\x6d\x65\x61\x6e\x74\x20\x74\x6f\x20\x62\x65\x20\x64\x69\x72\x65\x63\x74\x6c\x79\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x21\x20\x49\x74\x20\x68\x61\x73\x20\x62\x65\x65\x6e\x20\x64\x65\x73\x69\x67\x6e\x65\x64\x20\x73\x75\x63\x68\x0a\x74\x68\x61\x74\x20\x69\x74\x20\x63\x61\x6e\x20\x62\x65\x20\x62\x6f\x6f\x74\x73\x74\x72\x61\x70\x70\x65\x64\x20\x69\x6e\x74\x6f\x20\x50\x79\x74\x68\x6f\x6e\x20\x61\x73\x20\x74\x68\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x69\x6d\x70\x6f\x72\x74\x2e\x20\x41\x73\x0a\x73\x75\x63\x68\x20\x69\x74\x20\x72\x65\x71\x75\x69\x72\x65\x73\x20\x74\x68\x65\x20\x69\x6e\x6a\x65\x63\x74\x69\x6f\x6e\x20\x6f\x66\x20\x73\x70\x65\x63\x69\x66\x69\x63\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x61\x6e\x64\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x73\x20\x69\x6e\x20\x6f\x72\x64\x65\x72\x20\x74\x6f\x0a\x77\x6f\x72\x6b\x2e\x20\x4f\x6e\x65\x20\x73\x68\x6f\x75\x6c\x64\x20\x75\x73\x65\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x20\x61\x73\x20\x74\x68\x65\x20\x70\x75\x62\x6c\x69\x63\x2d\x66\x61\x63\x69\x6e\x67\x20\x76\x65\x72\x73\x69\x6f\x6e\x20\x6f\x66\x20\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(len),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_external_toplevel_consts_6_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen importlib._bootstrap_external>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
importlib__bootstrap_external_toplevel_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xd0\x0a\x34\xa1\x4f\x98\x53\x8c\x33\x88\x73\x8b\x38\x90\x71\x8d\x3d\xa1\x4f\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
importlib__bootstrap_external_toplevel_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x18\x1a\x01",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".0",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            &_Py_ID(sep),
        },
    },
};
static
    struct _PyCode_DEF(56)
importlib__bootstrap_external_toplevel_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 46,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 99,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = &_Py_STR(anon_genexpr),
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x12\x00\x00\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x6b\x28\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x14\x04\x00\x79\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_win = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "win",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_9 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_win._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_cygwin = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cygwin",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_darwin = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "darwin",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_10 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_cygwin._ascii.ob_base,
            & const_str_darwin._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_PYTHONCASEOK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PYTHONCASEOK",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_external_toplevel_consts_11_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "PYTHONCASEOK",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[95];
    }
importlib__bootstrap_external_toplevel_consts_11_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 94,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "True if filenames must be checked case-insensitively and ignore environment flags are not set.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_11_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_11_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_ignore_environment = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ignore_environment",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str__os = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_os",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_environ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "environ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_11_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_ignore_environment._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__relax_case = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_relax_case",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
importlib__bootstrap_external_toplevel_consts_11_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_make_relax_case.<locals>._relax_case",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
importlib__bootstrap_external_toplevel_consts_11_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xe4\x17\x1a\x97\x79\x91\x79\xd7\x17\x33\xd1\x17\x33\xd0\x13\x33\xd2\x13\x4a\xb8\x03\xbc\x73\xbf\x7b\xb9\x7b\xd0\x38\x4a\xd0\x0c\x4a",
};
static
    struct _PyCode_DEF(98)
importlib__bootstrap_external_toplevel_consts_11_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_11_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_11_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 67,
    .co_nlocalsplus = 1,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 100,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_1._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[128]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__relax_case._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_11_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_11_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0c\x00\x78\x01\x72\x12\x01\x00\x89\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
importlib__bootstrap_external_toplevel_consts_11_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "True if filenames must be checked case-insensitively.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_11_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_11_consts_4_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_external_toplevel_consts_11_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x13\x18",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_11_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_11_consts_4_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 0 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 71,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 101,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__relax_case._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_11_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_11_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            & const_str_PYTHONCASEOK._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_11_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_11_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_11_consts_4.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
const_str__CASE_INSENSITIVE_PLATFORMS = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CASE_INSENSITIVE_PLATFORMS",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
const_str__CASE_INSENSITIVE_PLATFORMS_STR_KEY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CASE_INSENSITIVE_PLATFORMS_STR_KEY",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str__CASE_INSENSITIVE_PLATFORMS._ascii.ob_base,
            & const_str__CASE_INSENSITIVE_PLATFORMS_STR_KEY._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__make_relax_case = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_make_relax_case",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[79];
    }
importlib__bootstrap_external_toplevel_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 78,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x07\x0a\x87\x7c\x81\x7c\xd7\x07\x1e\xd1\x07\x1e\xd4\x1f\x3a\xd4\x07\x3b\xdc\x0b\x0e\x8f\x3c\x89\x3c\xd7\x0b\x22\xd1\x0b\x22\xd4\x23\x46\xd4\x0b\x47\xd8\x12\x20\x89\x43\xe0\x12\x21\x88\x43\xf4\x04\x02\x09\x4b\x01\xf0\x0e\x00\x0c\x17\xd0\x04\x16\xf2\x07\x02\x09\x19\xf0\x06\x00\x0c\x17\xd0\x04\x16",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__relax_case._ascii.ob_base,
            &_Py_ID(key),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
importlib__bootstrap_external_toplevel_consts_11_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = " @",
};
static
    struct _PyCode_DEF(178)
importlib__bootstrap_external_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 89,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 60,
    .co_nlocalsplus = 2,
    .co_nlocals = 1,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 102,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_11_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__make_relax_case._ascii.ob_base,
    .co_qualname = & const_str__make_relax_case._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x01\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x2f\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x03\x64\x01\x8a\x01\x6e\x02\x64\x02\x8a\x01\x88\x01\x66\x01\x64\x03\x84\x08\x7d\x00\x7c\x00\x53\x00\x64\x04\x84\x00\x7d\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[43];
    }
importlib__bootstrap_external_toplevel_consts_12_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 42,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Convert a 32-bit integer to little-endian.",
};
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[3];
    }
const_int_4294967295 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 3),
    .ob_digit = { 32767, 32767, 3 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_4294967295 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 1073741823, 3 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_12_consts_0._ascii.ob_base,
            & const_int_4294967295.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 4],
            &_Py_ID(little),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_int = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "int",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_to_bytes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "to_bytes",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_int._ascii.ob_base,
            & const_str_to_bytes._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__pack_uint32 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_pack_uint32",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
importlib__bootstrap_external_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0c\x0f\x90\x01\x8b\x46\x90\x5a\xd1\x0c\x1f\xd7\x0b\x29\xd1\x0b\x29\xa8\x21\xa8\x58\xd3\x0b\x36\xd0\x04\x36",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_12_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
        },
    },
};
static
    struct _PyCode_DEF(62)
importlib__bootstrap_external_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 79,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 103,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_12_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__pack_uint32._ascii.ob_base,
    .co_qualname = & const_str__pack_uint32._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x7a\x01\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[48];
    }
importlib__bootstrap_external_toplevel_consts_13_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 47,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Convert 4 bytes in little-endian to an integer.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_13_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 4],
            &_Py_ID(little),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_from_bytes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "from_bytes",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(len),
            & const_str_int._ascii.ob_base,
            & const_str_from_bytes._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__unpack_uint32 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_unpack_uint32",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
importlib__bootstrap_external_toplevel_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0e\x88\x74\x8b\x39\x98\x01\x8a\x3e\xd0\x04\x19\x88\x3e\xdc\x0b\x0e\x8f\x3e\x89\x3e\x98\x24\xa0\x08\xd3\x0b\x29\xd0\x04\x29",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_13_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(data),
        },
    },
};
static
    struct _PyCode_DEF(78)
importlib__bootstrap_external_toplevel_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_13_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 84,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 104,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_13_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__unpack_uint32._ascii.ob_base,
    .co_qualname = & const_str__unpack_uint32._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x73\x02\x4a\x00\x82\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[48];
    }
importlib__bootstrap_external_toplevel_consts_14_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 47,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Convert 2 bytes in little-endian to an integer.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_14_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            &_Py_ID(little),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__unpack_uint16 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_unpack_uint16",
};
static
    struct _PyCode_DEF(78)
importlib__bootstrap_external_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 89,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 105,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_13_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__unpack_uint16._ascii.ob_base,
    .co_qualname = & const_str__unpack_uint16._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x73\x02\x4a\x00\x82\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_15_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Replacement for os.path.join().",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_15_consts_0._ascii.ob_base,
            &_Py_STR(empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_Py_SINGLETON(strings).ascii[58],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_map = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "map",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__path_splitroot = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_splitroot",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_path_sep_tuple = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_sep_tuple",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_endswith = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "endswith",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_rstrip = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rstrip",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_path_separators = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_separators",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_path_sep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_sep",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_casefold = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "casefold",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            &_Py_ID(len),
            & const_str_map._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str__path_splitroot._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_path_sep_tuple._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            & const_str_rstrip._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
            & const_str_path_sep._ascii.ob_base,
            & const_str_casefold._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__path_join = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_join",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[297];
    }
importlib__bootstrap_external_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 296,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe1\x0f\x19\xd8\x13\x15\xdc\x0b\x0e\x88\x7a\x8b\x3f\x98\x61\xd2\x0b\x1f\xd8\x13\x1d\x98\x61\x91\x3d\xd0\x0c\x20\xd8\x0f\x11\x88\x04\xd8\x0f\x11\x88\x04\xdc\x1e\x21\xa4\x23\xd7\x22\x35\xd1\x22\x35\xb0\x7a\xd6\x1e\x42\x89\x4e\x88\x48\x90\x64\xd8\x0f\x17\xd7\x0f\x22\xd1\x0f\x22\xa4\x3e\xd4\x0f\x32\xb0\x68\xd7\x36\x47\xd1\x36\x47\xcc\x0e\xd4\x36\x57\xd8\x17\x1f\x97\x7f\x91\x7f\xa4\x7f\xd3\x17\x37\xd2\x17\x3f\xb8\x34\x90\x04\xdc\x18\x20\xa0\x34\x99\x0f\xd0\x17\x28\x91\x04\xd8\x11\x19\xd7\x11\x22\xd1\x11\x22\xa0\x33\xd4\x11\x27\xd8\x13\x17\x97\x3d\x91\x3d\x93\x3f\xa0\x68\xd7\x26\x37\xd1\x26\x37\xd3\x26\x39\xd2\x13\x39\xf0\x06\x00\x1c\x24\x90\x44\xd8\x1c\x20\x98\x36\x91\x44\xe0\x14\x18\x97\x4b\x91\x4b\xa0\x04\xd5\x14\x25\xe0\x17\x1f\xd2\x17\x27\xa0\x34\x90\x04\xd8\x10\x14\x97\x0b\x91\x0b\x98\x44\xd5\x10\x21\xf0\x1d\x00\x1f\x43\x01\xf1\x1e\x00\x34\x38\xd3\x0f\x3d\xb1\x34\xa8\x61\xba\x31\x90\x01\x97\x08\x91\x08\x9c\x1f\xd5\x10\x29\xb0\x34\x88\x04\xd0\x0f\x3d\xdc\x0b\x0e\x88\x74\x8b\x39\x98\x01\x8a\x3e\xa0\x24\xa0\x71\xa2\x27\xe0\x13\x17\x9c\x28\x91\x3f\xd0\x0c\x22\xd8\x0f\x13\x94\x68\x97\x6d\x91\x6d\xa0\x44\xd3\x16\x29\xd1\x0f\x29\xd0\x08\x29\xf9\xf2\x09\x00\x10\x3e",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_external_toplevel_consts_15_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\xc3\x30\x07\x45\x05\x04\xc3\x38\x17\x45\x05\x04",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_path_parts = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_parts",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_root = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "root",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_new_root = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "new_root",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_tail = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tail",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_15_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_path_parts._ascii.ob_base,
            & const_str_root._ascii.ob_base,
            &_Py_ID(path),
            & const_str_new_root._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
        },
    },
};
static
    struct _PyCode_DEF(660)
importlib__bootstrap_external_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 330,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_15_exceptiontable.ob_base.ob_base,
    .co_flags = 7,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 96,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 106,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_15_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_join._ascii.ob_base,
    .co_qualname = & const_str__path_join._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x73\x01\x79\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x28\x00\x00\x72\x05\x7c\x00\x64\x03\x19\x00\x00\x00\x53\x00\x64\x01\x7d\x01\x67\x00\x7d\x02\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x44\x00\x5d\xb4\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x7c\x03\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x15\x7c\x03\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x24\x7c\x03\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x78\x01\x73\x02\x01\x00\x7c\x01\x7d\x01\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7a\x00\x00\x00\x67\x01\x7d\x02\x8c\x54\x7c\x03\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x72\x39\x7c\x01\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x06\x7c\x03\x7d\x01\x7c\x04\x67\x01\x7d\x02\x8c\x8c\x7c\x02\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x9e\x7c\x03\x78\x01\x73\x02\x01\x00\x7c\x01\x7d\x01\x7c\x02\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\xb6\x04\x00\x7c\x02\x44\x00\x8f\x05\x63\x02\x67\x00\x63\x02\x5d\x1a\x00\x00\x7d\x05\x7c\x05\x73\x01\x8c\x06\x7c\x05\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x91\x02\x8c\x1c\x04\x00\x7d\x02\x7d\x05\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x28\x00\x00\x72\x0e\x7c\x02\x64\x03\x19\x00\x00\x00\x73\x09\x7c\x01\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x53\x00\x7c\x01\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x05\x77\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_15_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_path_sep._ascii.ob_base,
            &_Py_ID(join),
            & const_str_rstrip._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
importlib__bootstrap_external_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x17\x8f\x7d\x89\x7d\xd9\x2a\x34\xf3\x03\x01\x1e\x3e\xd9\x2a\x34\xa0\x24\xb9\x04\xf0\x03\x00\x1f\x23\x9f\x6b\x99\x6b\xac\x2f\xd5\x1e\x3a\xd8\x2a\x34\xf1\x03\x01\x1e\x3e\xf3\x00\x01\x10\x3f\xf0\x00\x01\x09\x3f\xf9\xf2\x00\x01\x1e\x3e",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
importlib__bootstrap_external_toplevel_consts_16_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x94\x1e\x39\x08",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_part = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "part",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_path_parts._ascii.ob_base,
            & const_str_part._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(124)
importlib__bootstrap_external_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 62,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_16_exceptiontable.ob_base.ob_base,
    .co_flags = 7,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 126,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 107,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_join._ascii.ob_base,
    .co_qualname = & const_str__path_join._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x44\x00\x8f\x01\x63\x02\x67\x00\x63\x02\x5d\x19\x00\x00\x7d\x01\x7c\x01\x72\x15\x7c\x01\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x91\x02\x8c\x1b\x04\x00\x63\x02\x7d\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x01\x77\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib__bootstrap_external_toplevel_consts_17_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Replacement for os.path.split().",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_rfind = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rfind",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_17_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_rfind._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_17_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_split.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[27];
    }
importlib__bootstrap_external_toplevel_consts_17_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 26,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xd0\x0b\x33\xa1\x3f\x98\x61\x88\x44\x8f\x4a\x89\x4a\x90\x71\x8f\x4d\xa1\x3f\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
importlib__bootstrap_external_toplevel_consts_17_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x1b\x1e\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_17_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(path),
        },
    },
};
static
    struct _PyCode_DEF(64)
importlib__bootstrap_external_toplevel_consts_17_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_17_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_17_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 134,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 108,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_17_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_17_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_17_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x15\x00\x00\x7d\x01\x89\x02\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x17\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_17_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_17_consts_1.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_STR(empty),
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_max = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "max",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_17_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_max._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__path_split = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_split",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
importlib__bootstrap_external_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xe4\x08\x0b\xd3\x0b\x33\xa5\x3f\xd3\x0b\x33\xd3\x08\x33\x80\x41\xd8\x07\x08\x88\x31\x82\x75\xd8\x0f\x11\x90\x34\x88\x78\x88\x0f\xd8\x0b\x0f\x90\x02\x90\x11\x88\x38\x90\x54\x98\x21\x98\x61\x99\x25\x98\x26\x90\x5c\xd0\x0b\x21\xd0\x04\x21",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_17_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
        },
    },
};
static
    struct _PyCode_DEF(96)
importlib__bootstrap_external_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 48,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_17_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 132,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 109,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_17_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_split._ascii.ob_base,
    .co_qualname = & const_str__path_split._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x88\x00\x66\x01\x64\x01\x84\x08\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x64\x02\x6b\x02\x00\x00\x72\x04\x64\x03\x89\x00\x66\x02\x53\x00\x89\x00\x64\x04\x7c\x01\x1a\x00\x89\x00\x7c\x01\x64\x05\x7a\x00\x00\x00\x64\x04\x1a\x00\x66\x02\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[127];
    }
importlib__bootstrap_external_toplevel_consts_18_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 126,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x74\x61\x74\x20\x74\x68\x65\x20\x70\x61\x74\x68\x2e\x0a\x0a\x20\x20\x20\x20\x4d\x61\x64\x65\x20\x61\x20\x73\x65\x70\x61\x72\x61\x74\x65\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x74\x6f\x20\x6d\x61\x6b\x65\x20\x69\x74\x20\x65\x61\x73\x69\x65\x72\x20\x74\x6f\x20\x6f\x76\x65\x72\x72\x69\x64\x65\x20\x69\x6e\x20\x65\x78\x70\x65\x72\x69\x6d\x65\x6e\x74\x73\x0a\x20\x20\x20\x20\x28\x65\x2e\x67\x2e\x20\x63\x61\x63\x68\x65\x20\x73\x74\x61\x74\x20\x72\x65\x73\x75\x6c\x74\x73\x29\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_18_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_stat = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "stat",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__path_stat = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_stat",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_external_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x0c\x0f\x8f\x38\x89\x38\x90\x44\x8b\x3e\xd0\x04\x19",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_18_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(path),
        },
    },
};
static
    struct _PyCode_DEF(44)
importlib__bootstrap_external_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 140,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 110,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_stat._ascii.ob_base,
    .co_qualname = & const_str__path_stat._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[50];
    }
importlib__bootstrap_external_toplevel_consts_19_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 49,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether the path is the specified mode type.",
};
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_61440 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 28672, 1 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_61440 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 61440 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_19_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_19_consts_0._ascii.ob_base,
            Py_False,
            & const_int_61440.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_OSError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "OSError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_st_mode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_mode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_19_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__path_stat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__path_is_mode_type = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_is_mode_type",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[62];
    }
importlib__bootstrap_external_toplevel_consts_19_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 61,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x14\x1e\x98\x74\xd3\x14\x24\x88\x09\xf0\x06\x00\x0d\x16\xd7\x0c\x1d\xd1\x0c\x1d\xa0\x08\xd1\x0c\x28\xa8\x54\xd1\x0b\x31\xd0\x04\x31\xf8\xf4\x05\x00\x0c\x13\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_external_toplevel_consts_19_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x0b\x1f\x00\x9f\x09\x2b\x03\xaa\x01\x2b\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_stat_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "stat_info",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_19_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(mode),
            & const_str_stat_info._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(92)
importlib__bootstrap_external_toplevel_consts_19 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 46,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_19_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_19_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_19_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 150,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 111,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_19_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_is_mode_type._ascii.ob_base,
    .co_qualname = & const_str__path_is_mode_type._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_19_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7a\x01\x00\x00\x7c\x01\x6b\x28\x00\x00\x53\x00\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_20_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Replacement for os.path.isfile.",
};
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_32768 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 0, 1 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_32768 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 32768 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_20_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_20_consts_0._ascii.ob_base,
            & const_int_32768.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_is_mode_type._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__path_isfile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_isfile",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib__bootstrap_external_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x1d\x98\x64\xa0\x48\xd3\x0b\x2d\xd0\x04\x2d",
};
static
    struct _PyCode_DEF(26)
importlib__bootstrap_external_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_20_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 159,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 112,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_isfile._ascii.ob_base,
    .co_qualname = & const_str__path_isfile._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_21_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Replacement for os.path.isdir.",
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_16384 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 16384 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_21_consts_0._ascii.ob_base,
            & const_int_16384.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_getcwd = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getcwd",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_21_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__os._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str__path_is_mode_type._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__path_isdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_isdir",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib__bootstrap_external_toplevel_consts_21_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe1\x0b\x0f\xdc\x0f\x12\x8f\x7a\x89\x7a\x8b\x7c\x88\x04\xdc\x0b\x1d\x98\x64\xa0\x48\xd3\x0b\x2d\xd0\x04\x2d",
};
static
    struct _PyCode_DEF(70)
importlib__bootstrap_external_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 164,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 113,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_isdir._ascii.ob_base,
    .co_qualname = & const_str__path_isdir._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_21_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x73\x14\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_22_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Replacement for os.path.isabs.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_external_toplevel_consts_22_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\\\",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_22_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_22_consts_0._ascii.ob_base,
            Py_False,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & importlib__bootstrap_external_toplevel_consts_22_consts_6._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_22_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__os._ascii.ob_base,
            & const_str__path_splitroot._ascii.ob_base,
            &_Py_ID(replace),
            &_Py_ID(len),
            & const_str_startswith._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__path_isabs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_isabs",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[85];
    }
importlib__bootstrap_external_toplevel_consts_22_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 84,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe1\x0f\x13\xd8\x13\x18\xdc\x0f\x12\xd7\x0f\x22\xd1\x0f\x22\xa0\x34\xd3\x0f\x28\xa8\x11\xd1\x0f\x2b\xd7\x0f\x33\xd1\x0f\x33\xb0\x43\xb8\x14\xd3\x0f\x3e\x88\x04\xdc\x0f\x12\x90\x34\x8b\x79\x98\x31\x89\x7d\xd2\x0f\x51\xa0\x24\xa7\x2f\xa1\x2f\xb0\x26\xd3\x22\x39\xd2\x22\x50\xb8\x54\xbf\x5d\xb9\x5d\xc8\x34\xd3\x3d\x50\xd0\x08\x51",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_22_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_root._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(192)
importlib__bootstrap_external_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 96,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_22_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_22_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 172,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 114,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_22_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_isabs._ascii.ob_base,
    .co_qualname = & const_str__path_isabs._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x73\x01\x79\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x64\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x05\x6b\x44\x00\x00\x78\x01\x72\x24\x01\x00\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x78\x01\x73\x11\x01\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_23_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_22_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_23_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_startswith._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
importlib__bootstrap_external_toplevel_consts_23_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7f\x89\x7f\x9c\x7f\xd3\x0f\x2f\xd0\x08\x2f",
};
static
    struct _PyCode_DEF(44)
importlib__bootstrap_external_toplevel_consts_23 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_23_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_23_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 180,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 115,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_isabs._ascii.ob_base,
    .co_qualname = & const_str__path_isabs._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_23_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib__bootstrap_external_toplevel_consts_24_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Replacement for os.path.abspath.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_24_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_24_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_removeprefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "removeprefix",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_24_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__path_isabs._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
            & const_str_removeprefix._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__path_abspath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_abspath",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
importlib__bootstrap_external_toplevel_consts_24_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x16\x90\x74\xd4\x0b\x1c\xdf\x13\x22\x88\x43\xd8\x13\x17\xd7\x13\x24\xd1\x13\x24\xa0\x71\xa8\x13\xa8\x05\xa0\x59\xd3\x13\x2f\x89\x44\xf0\x03\x00\x14\x23\xe4\x0f\x19\x9c\x23\x9f\x2a\x99\x2a\x9b\x2c\xa8\x04\xd3\x0f\x2d\xd0\x08\x2d\xe0\x0f\x13\x88\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_24_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(sep),
        },
    },
};
static
    struct _PyCode_DEF(150)
importlib__bootstrap_external_toplevel_consts_24 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 75,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_24_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_24_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 185,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 116,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_24_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_abspath._ascii.ob_base,
    .co_qualname = & const_str__path_abspath._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_24_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x3d\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x16\x00\x00\x7d\x01\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x01\x9b\x00\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x8c\x18\x04\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[163];
    }
importlib__bootstrap_external_toplevel_consts_25_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 162,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x42\x65\x73\x74\x2d\x65\x66\x66\x6f\x72\x74\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x74\x6f\x20\x77\x72\x69\x74\x65\x20\x64\x61\x74\x61\x20\x74\x6f\x20\x61\x20\x70\x61\x74\x68\x20\x61\x74\x6f\x6d\x69\x63\x61\x6c\x6c\x79\x2e\x0a\x20\x20\x20\x20\x42\x65\x20\x70\x72\x65\x70\x61\x72\x65\x64\x20\x74\x6f\x20\x68\x61\x6e\x64\x6c\x65\x20\x61\x20\x46\x69\x6c\x65\x45\x78\x69\x73\x74\x73\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x63\x6f\x6e\x63\x75\x72\x72\x65\x6e\x74\x20\x77\x72\x69\x74\x69\x6e\x67\x20\x6f\x66\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x74\x65\x6d\x70\x6f\x72\x61\x72\x79\x20\x66\x69\x6c\x65\x20\x69\x73\x20\x61\x74\x74\x65\x6d\x70\x74\x65\x64\x2e",
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_438 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 438 },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_wb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "wb",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
importlib__bootstrap_external_toplevel_consts_25_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "os.write() didn't write the full pyc file",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_25_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_25_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & const_int_438.ob_base,
            & const_str_wb._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_25_consts_5._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_O_EXCL = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "O_EXCL",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_O_CREAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "O_CREAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_O_WRONLY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "O_WRONLY",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_FileIO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileIO",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_25_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            &_Py_ID(id),
            & const_str__os._ascii.ob_base,
            &_Py_ID(open),
            & const_str_O_EXCL._ascii.ob_base,
            & const_str_O_CREAT._ascii.ob_base,
            & const_str_O_WRONLY._ascii.ob_base,
            &_Py_ID(_io),
            & const_str_FileIO._ascii.ob_base,
            &_Py_ID(write),
            &_Py_ID(len),
            & const_str_OSError._ascii.ob_base,
            &_Py_ID(replace),
            &_Py_ID(unlink),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__write_atomic = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_write_atomic",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[225];
    }
importlib__bootstrap_external_toplevel_consts_25_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 224,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x13\x17\x90\x16\x90\x71\x9c\x12\x98\x44\x9b\x18\x98\x0a\xd0\x0f\x23\x80\x48\xdc\x09\x0c\x8f\x18\x89\x18\x90\x28\xdc\x12\x15\x97\x2a\x91\x2a\x9c\x73\x9f\x7b\x99\x7b\xd1\x12\x2a\xac\x53\xaf\x5c\xa9\x5c\xd1\x12\x39\xb8\x34\xc0\x25\xb9\x3c\xf3\x03\x01\x0a\x49\x01\x80\x42\xf0\x04\x0f\x05\x0e\xf4\x06\x00\x0e\x11\x8f\x5a\x89\x5a\x98\x02\x98\x44\xd4\x0d\x21\xa0\x54\xd8\x1c\x20\x9f\x4a\x99\x4a\xa0\x74\xd3\x1c\x2c\x88\x4d\xf7\x03\x00\x0e\x22\xe0\x0b\x18\x9c\x43\xa0\x04\x9b\x49\xd2\x0b\x25\xf4\x06\x00\x13\x1a\xd0\x1a\x45\xd3\x12\x46\xd0\x0c\x46\xdc\x08\x0b\x8f\x0b\x89\x0b\x90\x48\x98\x64\xd5\x08\x23\xf7\x0d\x00\x0e\x22\xd0\x0d\x21\xfb\xf4\x0e\x00\x0c\x13\xf2\x00\x05\x05\x0e\xf0\x02\x03\x09\x11\xdc\x0c\x0f\x8f\x4a\x89\x4a\x90\x78\xd4\x0c\x20\xf0\x06\x00\x09\x0e\xf8\xf4\x05\x00\x10\x17\xf2\x00\x01\x09\x11\xd8\x0c\x10\xd8\x08\x0d\xf0\x05\x01\x09\x11\xfa\xf0\x07\x05\x05\x0e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[73];
    }
importlib__bootstrap_external_toplevel_consts_25_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 72,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x1c\x16\x43\x08\x00\xc1\x32\x12\x42\x3c\x03\xc2\x04\x37\x43\x08\x00\xc2\x3c\x05\x43\x05\x07\xc3\x01\x07\x43\x08\x00\xc3\x08\x09\x43\x38\x03\xc3\x12\x15\x43\x28\x02\xc3\x27\x01\x43\x38\x03\xc3\x28\x09\x43\x34\x05\xc3\x31\x02\x43\x38\x03\xc3\x33\x01\x43\x34\x05\xc3\x34\x04\x43\x38\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_path_tmp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_tmp",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_bytes_written = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytes_written",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_25_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(data),
            &_Py_ID(mode),
            & const_str_path_tmp._ascii.ob_base,
            &_Py_ID(fd),
            &_Py_ID(file),
            & const_str_bytes_written._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(502)
importlib__bootstrap_external_toplevel_consts_25 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 251,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_25_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_25_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 195,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 117,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_25_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__write_atomic._ascii.ob_base,
    .co_qualname = & const_str__write_atomic._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_25_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x9b\x00\x64\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x03\x7d\x03\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x07\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x07\x00\x00\x7c\x02\x64\x02\x7a\x01\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x04\x09\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x05\x7c\x05\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x64\x04\x64\x04\x64\x04\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7f\x06\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x0b\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x04\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x39\x78\x03\x59\x00\x77\x01\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x27\x01\x00\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x82\x00\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_3531 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 3531 },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
importlib__bootstrap_external_toplevel_consts_29 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\x0d\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str___pycache__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__pycache__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
importlib__bootstrap_external_toplevel_consts_31 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "opt-",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
importlib__bootstrap_external_toplevel_consts_33 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".pyw",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
importlib__bootstrap_external_toplevel_consts_34 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".pyc",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_optimization = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "optimization",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_35 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_optimization._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[767];
    }
importlib__bootstrap_external_toplevel_consts_36_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 766,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x47\x69\x76\x65\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x61\x20\x2e\x70\x79\x20\x66\x69\x6c\x65\x2c\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x69\x74\x73\x20\x2e\x70\x79\x63\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x2e\x70\x79\x20\x66\x69\x6c\x65\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x20\x6e\x65\x65\x64\x20\x74\x6f\x20\x65\x78\x69\x73\x74\x3b\x20\x74\x68\x69\x73\x20\x73\x69\x6d\x70\x6c\x79\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x2e\x70\x79\x63\x20\x66\x69\x6c\x65\x20\x63\x61\x6c\x63\x75\x6c\x61\x74\x65\x64\x20\x61\x73\x20\x69\x66\x20\x74\x68\x65\x20\x2e\x70\x79\x20\x66\x69\x6c\x65\x20\x77\x65\x72\x65\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x27\x6f\x70\x74\x69\x6d\x69\x7a\x61\x74\x69\x6f\x6e\x27\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x20\x63\x6f\x6e\x74\x72\x6f\x6c\x73\x20\x74\x68\x65\x20\x70\x72\x65\x73\x75\x6d\x65\x64\x20\x6f\x70\x74\x69\x6d\x69\x7a\x61\x74\x69\x6f\x6e\x20\x6c\x65\x76\x65\x6c\x20\x6f\x66\x0a\x20\x20\x20\x20\x74\x68\x65\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x2e\x20\x49\x66\x20\x27\x6f\x70\x74\x69\x6d\x69\x7a\x61\x74\x69\x6f\x6e\x27\x20\x69\x73\x20\x6e\x6f\x74\x20\x4e\x6f\x6e\x65\x2c\x20\x74\x68\x65\x20\x73\x74\x72\x69\x6e\x67\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x61\x74\x69\x6f\x6e\x0a\x20\x20\x20\x20\x6f\x66\x20\x74\x68\x65\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x69\x73\x20\x74\x61\x6b\x65\x6e\x20\x61\x6e\x64\x20\x76\x65\x72\x69\x66\x69\x65\x64\x20\x74\x6f\x20\x62\x65\x20\x61\x6c\x70\x68\x61\x6e\x75\x6d\x65\x72\x69\x63\x20\x28\x65\x6c\x73\x65\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x0a\x20\x20\x20\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x29\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x64\x65\x62\x75\x67\x5f\x6f\x76\x65\x72\x72\x69\x64\x65\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x20\x69\x73\x20\x64\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2e\x20\x49\x66\x20\x64\x65\x62\x75\x67\x5f\x6f\x76\x65\x72\x72\x69\x64\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x4e\x6f\x6e\x65\x2c\x0a\x20\x20\x20\x20\x61\x20\x54\x72\x75\x65\x20\x76\x61\x6c\x75\x65\x20\x69\x73\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x61\x73\x20\x73\x65\x74\x74\x69\x6e\x67\x20\x27\x6f\x70\x74\x69\x6d\x69\x7a\x61\x74\x69\x6f\x6e\x27\x20\x74\x6f\x20\x74\x68\x65\x20\x65\x6d\x70\x74\x79\x20\x73\x74\x72\x69\x6e\x67\x0a\x20\x20\x20\x20\x77\x68\x69\x6c\x65\x20\x61\x20\x46\x61\x6c\x73\x65\x20\x76\x61\x6c\x75\x65\x20\x69\x73\x20\x65\x71\x75\x69\x76\x61\x6c\x65\x6e\x74\x20\x74\x6f\x20\x73\x65\x74\x74\x69\x6e\x67\x20\x27\x6f\x70\x74\x69\x6d\x69\x7a\x61\x74\x69\x6f\x6e\x27\x20\x74\x6f\x20\x27\x31\x27\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x73\x79\x73\x2e\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x2e\x63\x61\x63\x68\x65\x5f\x74\x61\x67\x20\x69\x73\x20\x4e\x6f\x6e\x65\x20\x74\x68\x65\x6e\x20\x4e\x6f\x74\x49\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[71];
    }
importlib__bootstrap_external_toplevel_consts_36_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 70,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "the debug_override parameter is deprecated; use 'optimization' instead",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[51];
    }
importlib__bootstrap_external_toplevel_consts_36_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 50,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "debug_override or optimization must be set to None",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[37];
    }
importlib__bootstrap_external_toplevel_consts_36_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 36,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys.implementation.cache_tag is None",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_external_toplevel_consts_36_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " is not alphanumeric",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_36_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_36_consts_0._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_36_consts_2._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_36_consts_3._ascii.ob_base,
            &_Py_STR(empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & importlib__bootstrap_external_toplevel_consts_36_consts_7._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_external_toplevel_consts_36_consts_9._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[58],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_fspath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fspath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_cache_tag = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cache_tag",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_isalnum = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isalnum",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str__OPT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_OPT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_BYTECODE_SUFFIXES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BYTECODE_SUFFIXES",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_pycache_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pycache_prefix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_lstrip = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lstrip",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__PYCACHE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_PYCACHE",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[26];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_36_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 26,
        },
        .ob_item = {
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_DeprecationWarning._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__path_split._ascii.ob_base,
            & const_str_rpartition._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_implementation._ascii.ob_base,
            & const_str_cache_tag._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            &_Py_ID(join),
            &_Py_ID(flags),
            &_Py_ID(optimize),
            & const_str_str._ascii.ob_base,
            & const_str_isalnum._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str__OPT._ascii.ob_base,
            & const_str_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_pycache_prefix._ascii.ob_base,
            & const_str__path_abspath._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str_lstrip._ascii.ob_base,
            & const_str__PYCACHE._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_cache_from_source = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cache_from_source",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[416];
    }
importlib__bootstrap_external_toplevel_consts_36_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 415,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x24\x00\x08\x16\xd0\x07\x21\xdc\x08\x11\x8f\x0e\x89\x0e\xf0\x00\x01\x18\x30\xdc\x31\x43\xf4\x03\x01\x09\x45\x01\xe0\x0b\x17\xd0\x0b\x23\xd8\x16\x4a\x88\x47\xdc\x12\x1b\x98\x47\xd3\x12\x24\xd0\x0c\x24\xd9\x1d\x2b\x91\x72\xb0\x11\x88\x0c\xdc\x0b\x0e\x8f\x3a\x89\x3a\x90\x64\xd3\x0b\x1b\x80\x44\xdc\x11\x1c\x98\x54\xd3\x11\x22\x81\x4a\x80\x44\x88\x24\xd8\x16\x1a\x97\x6f\x91\x6f\xa0\x63\xd3\x16\x2a\x81\x4f\x80\x44\x88\x23\x88\x74\xdc\x0a\x0d\xd7\x0a\x1c\xd1\x0a\x1c\xd7\x0a\x26\xd1\x0a\x26\x80\x43\xd8\x07\x0a\x80\x7b\xdc\x0e\x21\xd0\x22\x48\xd3\x0e\x49\xd0\x08\x49\xd8\x16\x18\x97\x67\x91\x67\xa9\x04\xa1\x04\xb0\x24\xb8\x13\xb8\x63\xd0\x1e\x42\xd3\x16\x43\x80\x4f\xd8\x07\x13\xd0\x07\x1b\xdc\x0b\x0e\x8f\x39\x89\x39\xd7\x0b\x1d\xd1\x0b\x1d\xa0\x11\xd2\x0b\x22\xd8\x1b\x1d\x89\x4c\xe4\x1b\x1e\x9f\x39\x99\x39\xd7\x1b\x2d\xd1\x1b\x2d\x88\x4c\xdc\x13\x16\x90\x7c\xd3\x13\x24\x80\x4c\xd8\x07\x13\x90\x72\xd2\x07\x19\xd8\x0f\x1b\xd7\x0f\x23\xd1\x0f\x23\xd4\x0f\x25\xdc\x12\x1c\xa0\x0c\xd0\x1f\x2f\xd0\x2f\x43\xd0\x1d\x44\xd3\x12\x45\xd0\x0c\x45\xd8\x1d\x2c\xd0\x1c\x2d\xa8\x51\xac\x74\xa8\x66\xb0\x5c\xb0\x4e\xd0\x1a\x43\x88\x0f\xd8\x0f\x1e\xd4\x21\x32\xb0\x31\xd1\x21\x35\xd1\x0f\x35\x80\x48\xdc\x07\x0a\xd7\x07\x19\xd1\x07\x19\xd0\x07\x25\xf4\x12\x00\x10\x1d\x98\x54\xd3\x0f\x22\x88\x04\xf0\x0a\x00\x0c\x10\x90\x01\x89\x37\x90\x63\x8a\x3e\x98\x64\xa0\x31\x99\x67\xac\x5f\xd1\x1e\x3c\xd8\x13\x17\x98\x01\x98\x02\x90\x38\x88\x44\xf4\x08\x00\x10\x1a\xdc\x0c\x0f\xd7\x0c\x1e\xd1\x0c\x1e\xd8\x0c\x10\x8f\x4b\x89\x4b\x9c\x0f\xd3\x0c\x28\xd8\x0c\x14\xf3\x07\x04\x10\x0a\xf0\x00\x04\x09\x0a\xf4\x0a\x00\x0c\x16\x90\x64\x9c\x48\xa0\x68\xd3\x0b\x2f\xd0\x04\x2f",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_debug_override = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "debug_override",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_head = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "head",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_rest = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rest",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_almost_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "almost_filename",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_36_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_debug_override._ascii.ob_base,
            & const_str_optimization._ascii.ob_base,
            &_Py_ID(message),
            & const_str_head._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
            &_Py_ID(base),
            &_Py_ID(sep),
            & const_str_rest._ascii.ob_base,
            &_Py_ID(tag),
            & const_str_almost_filename._ascii.ob_base,
            &_Py_ID(filename),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_external_toplevel_consts_36_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "            ",
};
static
    struct _PyCode_DEF(828)
importlib__bootstrap_external_toplevel_consts_36 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 414,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_36_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_36_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 18 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 486,
    .co_nlocalsplus = 12,
    .co_nlocals = 12,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 118,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_36_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_36_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_cache_from_source._ascii.ob_base,
    .co_qualname = & const_str_cache_from_source._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_36_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x81\x2f\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x81\x0d\x64\x03\x7d\x03\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x72\x02\x64\x04\x6e\x01\x64\x05\x7d\x02\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x04\x7d\x05\x7c\x05\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x06\x7d\x07\x7d\x08\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x09\x7c\x09\x80\x0b\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x64\x04\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x72\x02\x7c\x06\x6e\x01\x7c\x08\x7c\x07\x7c\x09\x67\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0a\x7c\x02\x80\x3a\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x6b\x28\x00\x00\x72\x03\x64\x04\x7d\x02\x6e\x1a\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x64\x04\x6b\x37\x00\x00\x72\x2b\x7c\x02\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x73\x0e\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x9b\x02\x64\x09\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x0a\x9b\x00\x64\x06\x74\x24\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x7c\x02\x9b\x00\x9d\x04\x7d\x0a\x7c\x0a\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x19\x00\x00\x00\x7a\x00\x00\x00\x7d\x0b\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x28\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x51\x74\x2b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x64\x05\x19\x00\x00\x00\x64\x0a\x6b\x28\x00\x00\x72\x10\x7c\x04\x64\x08\x19\x00\x00\x00\x74\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x05\x7c\x04\x64\x0b\x64\x01\x1a\x00\x7d\x04\x74\x2f\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x28\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6a\x31\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x2c\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x0b\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x74\x2f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x74\x32\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0b\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[367];
    }
importlib__bootstrap_external_toplevel_consts_37_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 366,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x47\x69\x76\x65\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x61\x20\x2e\x70\x79\x63\x2e\x20\x66\x69\x6c\x65\x2c\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x69\x74\x73\x20\x2e\x70\x79\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x2e\x70\x79\x63\x20\x66\x69\x6c\x65\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x20\x6e\x65\x65\x64\x20\x74\x6f\x20\x65\x78\x69\x73\x74\x3b\x20\x74\x68\x69\x73\x20\x73\x69\x6d\x70\x6c\x79\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x0a\x20\x20\x20\x20\x74\x68\x65\x20\x2e\x70\x79\x20\x66\x69\x6c\x65\x20\x63\x61\x6c\x63\x75\x6c\x61\x74\x65\x64\x20\x74\x6f\x20\x63\x6f\x72\x72\x65\x73\x70\x6f\x6e\x64\x20\x74\x6f\x20\x74\x68\x65\x20\x2e\x70\x79\x63\x20\x66\x69\x6c\x65\x2e\x20\x20\x49\x66\x20\x70\x61\x74\x68\x20\x64\x6f\x65\x73\x0a\x20\x20\x20\x20\x6e\x6f\x74\x20\x63\x6f\x6e\x66\x6f\x72\x6d\x20\x74\x6f\x20\x50\x45\x50\x20\x33\x31\x34\x37\x2f\x34\x38\x38\x20\x66\x6f\x72\x6d\x61\x74\x2c\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x72\x61\x69\x73\x65\x64\x2e\x20\x49\x66\x0a\x20\x20\x20\x20\x73\x79\x73\x2e\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x2e\x63\x61\x63\x68\x65\x5f\x74\x61\x67\x20\x69\x73\x20\x4e\x6f\x6e\x65\x20\x74\x68\x65\x6e\x20\x4e\x6f\x74\x49\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_37_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " not bottom-level directory in ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_37_consts_7 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
        },
    },
};
// TODO: The above tuple should be a frozenset
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_external_toplevel_consts_37_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "expected only 2 or 3 dots in ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
importlib__bootstrap_external_toplevel_consts_37_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "optimization portion of filename does not start with ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_37_consts_13 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "optimization level ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_external_toplevel_consts_37_consts_14 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " is not an alphanumeric value",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_37_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_37_consts_0._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_36_consts_7._ascii.ob_base,
            Py_False,
            Py_True,
            & importlib__bootstrap_external_toplevel_consts_37_consts_5._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & importlib__bootstrap_external_toplevel_consts_37_consts_7._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_37_consts_8._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -2],
            & importlib__bootstrap_external_toplevel_consts_37_consts_12._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_37_consts_13._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_37_consts_14._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_SOURCE_SUFFIXES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SOURCE_SUFFIXES",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_37_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_implementation._ascii.ob_base,
            & const_str_cache_tag._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__path_split._ascii.ob_base,
            & const_str_pycache_prefix._ascii.ob_base,
            & const_str_rstrip._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_path_sep._ascii.ob_base,
            &_Py_ID(len),
            & const_str__PYCACHE._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            &_Py_ID(count),
            & const_str_rsplit._ascii.ob_base,
            & const_str__OPT._ascii.ob_base,
            & const_str_isalnum._ascii.ob_base,
            & const_str_partition._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str_SOURCE_SUFFIXES._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_source_from_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_from_cache",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[422];
    }
importlib__bootstrap_external_toplevel_consts_37_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 421,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x12\x00\x08\x0b\xd7\x07\x19\xd1\x07\x19\xd7\x07\x23\xd1\x07\x23\xd0\x07\x2b\xdc\x0e\x21\xd0\x22\x48\xd3\x0e\x49\xd0\x08\x49\xdc\x0b\x0e\x8f\x3a\x89\x3a\x90\x64\xd3\x0b\x1b\x80\x44\xdc\x1d\x28\xa8\x14\xd3\x1d\x2e\xd1\x04\x1a\x80\x44\xd0\x0a\x1a\xd8\x1e\x23\xd0\x04\x1b\xdc\x07\x0a\xd7\x07\x19\xd1\x07\x19\xd0\x07\x25\xdc\x18\x1b\xd7\x18\x2a\xd1\x18\x2a\xd7\x18\x31\xd1\x18\x31\xb4\x2f\xd3\x18\x42\x88\x0d\xd8\x0b\x0f\x8f\x3f\x89\x3f\x98\x3d\xac\x38\xd1\x1b\x33\xd4\x0b\x34\xd8\x13\x17\x9c\x03\x98\x4d\xd3\x18\x2a\xd0\x18\x2b\xd0\x13\x2c\x88\x44\xd8\x26\x2a\xd0\x0c\x23\xd9\x0b\x22\xdc\x18\x23\xa0\x44\xd3\x18\x29\x89\x0d\x88\x04\x88\x67\xd8\x0b\x12\x94\x68\xd2\x0b\x1e\xdc\x12\x1c\xa4\x08\x98\x7a\xd0\x29\x48\xd8\x20\x24\x98\x78\xf0\x03\x01\x1e\x29\xf3\x00\x01\x13\x2a\xf0\x00\x01\x0d\x2a\xe0\x10\x20\xd7\x10\x26\xd1\x10\x26\xa0\x73\xd3\x10\x2b\x80\x49\xd8\x07\x10\x98\x06\xd1\x07\x1e\xdc\x0e\x18\xd0\x1b\x38\xd0\x39\x49\xd0\x38\x4c\xd0\x19\x4d\xd3\x0e\x4e\xd0\x08\x4e\xd8\x09\x12\x90\x61\x8a\x1e\xd8\x17\x27\xd7\x17\x2e\xd1\x17\x2e\xa8\x73\xb0\x41\xd3\x17\x36\xb0\x72\xd1\x17\x3a\x88\x0c\xd8\x0f\x1b\xd7\x0f\x26\xd1\x0f\x26\xa4\x74\xd4\x0f\x2c\xdc\x12\x1c\xf0\x00\x01\x1e\x25\xdc\x25\x29\xa0\x48\xf0\x03\x01\x1e\x2e\xf3\x00\x01\x13\x2f\xf0\x00\x01\x0d\x2f\xe0\x14\x20\xa4\x13\xa4\x54\xa3\x19\xa0\x1a\xd0\x14\x2c\x88\x09\xd8\x0f\x18\xd7\x0f\x20\xd1\x0f\x20\xd4\x0f\x22\xdc\x12\x1c\xd0\x1f\x32\xb0\x3c\xd0\x32\x42\xf0\x00\x01\x43\x01\x32\xf0\x00\x01\x1e\x32\xf3\x00\x01\x13\x33\xf0\x00\x01\x0d\x33\xe0\x14\x24\xd7\x14\x2e\xd1\x14\x2e\xa8\x73\xd3\x14\x33\xb0\x41\xd1\x14\x36\x80\x4d\xdc\x0b\x15\x90\x64\x98\x4d\xac\x4f\xb8\x41\xd1\x2c\x3e\xd1\x1c\x3e\xd3\x0b\x3f\xd0\x04\x3f",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_pycache_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pycache_filename",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str_found_in_pycache_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "found_in_pycache_prefix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_stripped_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "stripped_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_pycache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pycache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_dot_count = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dot_count",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_opt_level = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "opt_level",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_base_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "base_filename",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_37_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_head._ascii.ob_base,
            & const_str_pycache_filename._ascii.ob_base,
            & const_str_found_in_pycache_prefix._ascii.ob_base,
            & const_str_stripped_path._ascii.ob_base,
            & const_str_pycache._ascii.ob_base,
            & const_str_dot_count._ascii.ob_base,
            & const_str_optimization._ascii.ob_base,
            & const_str_opt_level._ascii.ob_base,
            & const_str_base_filename._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(804)
importlib__bootstrap_external_toplevel_consts_37 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 402,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_37_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_37_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 556,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 119,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_37_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_source_from_cache._ascii.ob_base,
    .co_qualname = & const_str_source_from_cache._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_37_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x0b\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x64\x03\x7d\x03\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x4b\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x01\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x10\x7c\x01\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x1a\x00\x7d\x01\x64\x04\x7d\x03\x7c\x03\x73\x2b\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x05\x7c\x05\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x14\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x05\x7c\x00\x9b\x02\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x64\x07\x76\x01\x72\x0e\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x7c\x02\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x06\x64\x09\x6b\x28\x00\x00\x72\x6d\x7c\x02\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x64\x0a\xab\x02\x00\x00\x00\x00\x00\x00\x64\x0b\x19\x00\x00\x00\x7d\x07\x7c\x07\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x12\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0c\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x07\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x1a\x00\x7d\x08\x7c\x08\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x73\x0f\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\x7c\x07\x9b\x02\x64\x0e\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x64\x0f\x19\x00\x00\x00\x7d\x09\x74\x29\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x09\x74\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0f\x19\x00\x00\x00\x7a\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[189];
    }
importlib__bootstrap_external_toplevel_consts_38_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 188,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6e\x76\x65\x72\x74\x20\x61\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x61\x20\x73\x6f\x75\x72\x63\x65\x20\x70\x61\x74\x68\x20\x28\x69\x66\x20\x70\x6f\x73\x73\x69\x62\x6c\x65\x29\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x65\x78\x69\x73\x74\x73\x20\x70\x75\x72\x65\x6c\x79\x20\x66\x6f\x72\x20\x62\x61\x63\x6b\x77\x61\x72\x64\x73\x2d\x63\x6f\x6d\x70\x61\x74\x69\x62\x69\x6c\x69\x74\x79\x20\x66\x6f\x72\x0a\x20\x20\x20\x20\x50\x79\x49\x6d\x70\x6f\x72\x74\x5f\x45\x78\x65\x63\x43\x6f\x64\x65\x4d\x6f\x64\x75\x6c\x65\x57\x69\x74\x68\x46\x69\x6c\x65\x6e\x61\x6d\x65\x73\x28\x29\x20\x69\x6e\x20\x74\x68\x65\x20\x43\x20\x41\x50\x49\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_py = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "py",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_38_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_38_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & const_str_py._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_lower = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lower",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_38_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(len),
            & const_str_rpartition._ascii.ob_base,
            & const_str_lower._ascii.ob_base,
            & const_str_source_from_cache._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str__path_isfile._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[143];
    }
importlib__bootstrap_external_toplevel_consts_38_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 142,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x08\x0b\x88\x3d\xd3\x07\x19\x98\x51\xd2\x07\x1e\xd8\x0f\x13\xd8\x19\x26\xd7\x19\x31\xd1\x19\x31\xb0\x23\xd3\x19\x36\xd1\x04\x16\x80\x44\x88\x21\x88\x59\xd9\x0b\x0f\x90\x39\x97\x3f\x91\x3f\xd3\x13\x24\xa0\x52\xa8\x02\xd0\x13\x2b\xa8\x74\xd2\x13\x33\xd8\x0f\x1c\xd0\x08\x1c\xf0\x02\x03\x05\x29\xdc\x16\x27\xa8\x0d\xd3\x16\x36\x88\x0b\xf4\x06\x00\x1b\x27\xa0\x7b\xd4\x1a\x33\x88\x3b\xd0\x04\x46\xb8\x1d\xd0\x04\x46\xf8\xf4\x05\x00\x0d\x20\xa4\x1a\xd0\x0b\x2c\xf2\x00\x01\x05\x29\xd8\x16\x23\xa0\x43\xa0\x52\xd0\x16\x28\x8a\x0b\xf0\x03\x01\x05\x29\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_external_toplevel_consts_38_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x00\x0b\x41\x1a\x00\xc1\x1a\x14\x41\x31\x03\xc1\x30\x01\x41\x31\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_bytecode_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytecode_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_extension = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "extension",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_source_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_38_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_bytecode_path._ascii.ob_base,
            & const_str_rest._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            & const_str_extension._ascii.ob_base,
            & const_str_source_path._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(232)
importlib__bootstrap_external_toplevel_consts_38 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 116,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_38_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_38_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_38_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 596,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 120,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_38_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(_get_sourcefile),
    .co_qualname = &_Py_ID(_get_sourcefile),
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_38_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x01\x79\x02\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x01\x7d\x02\x7d\x03\x7c\x01\x72\x16\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x64\x04\x64\x05\x1a\x00\x64\x06\x6b\x37\x00\x00\x72\x02\x7c\x00\x53\x00\x09\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x72\x02\x7c\x04\x53\x00\x7c\x00\x53\x00\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x08\x01\x00\x7c\x00\x64\x02\x64\x05\x1a\x00\x7d\x04\x59\x00\x8c\x25\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_tuple = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tuple",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_39_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_endswith._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_SOURCE_SUFFIXES._ascii.ob_base,
            & const_str_cache_from_source._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            & const_str_BYTECODE_SUFFIXES._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[88];
    }
importlib__bootstrap_external_toplevel_consts_39_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 87,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x07\x0f\xd7\x07\x18\xd1\x07\x18\x9c\x15\x9c\x7f\xd3\x19\x2f\xd4\x07\x30\xf0\x02\x03\x09\x11\xdc\x13\x24\xa0\x58\xd3\x13\x2e\xd0\x0c\x2e\xf0\x06\x00\x0a\x12\xd7\x09\x1a\xd1\x09\x1a\x9c\x35\xd4\x21\x32\xd3\x1b\x33\xd4\x09\x34\xd8\x0f\x17\x88\x0f\xe0\x0f\x13\xf8\xf4\x0b\x00\x10\x23\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
importlib__bootstrap_external_toplevel_consts_39_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\xa0\x0a\x41\x0c\x00\xc1\x0c\x09\x41\x18\x03\xc1\x17\x01\x41\x18\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_39_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(filename),
        },
    },
};
static
    struct _PyCode_DEF(182)
importlib__bootstrap_external_toplevel_consts_39 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 91,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_39_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_39_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 615,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 121,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_39_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__get_cached._ascii.ob_base,
    .co_qualname = & const_str__get_cached._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_39_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0c\x09\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x02\x7c\x00\x53\x00\x79\x00\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[52];
    }
importlib__bootstrap_external_toplevel_consts_40_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 51,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Calculate the mode permissions for a bytecode file.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_40_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_40_consts_0._ascii.ob_base,
            & const_int_438.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 128],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_40_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__path_stat._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__calc_mode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_calc_mode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[62];
    }
importlib__bootstrap_external_toplevel_consts_40_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 61,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x0f\x19\x98\x24\xd3\x0f\x1f\xd7\x0f\x27\xd1\x0f\x27\x88\x04\xf0\x0a\x00\x05\x09\x88\x45\x81\x4d\x80\x44\xd8\x0b\x0f\x80\x4b\xf8\xf4\x0b\x00\x0c\x13\xf2\x00\x01\x05\x15\xd8\x0f\x14\x8a\x04\xf0\x03\x01\x05\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
importlib__bootstrap_external_toplevel_consts_40_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x1e\x00\x9e\x0b\x2c\x03\xab\x01\x2c\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_40_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(mode),
        },
    },
};
static
    struct _PyCode_DEF(94)
importlib__bootstrap_external_toplevel_consts_40 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 47,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_40_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_40_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_40_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 627,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 122,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_40_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__calc_mode._ascii.ob_base,
    .co_qualname = & const_str__calc_mode._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_40_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x64\x02\x7a\x14\x00\x00\x7d\x01\x7c\x01\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x01\x7d\x01\x59\x00\x8c\x14\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[253];
    }
importlib__bootstrap_external_toplevel_consts_41_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 252,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x65\x63\x6f\x72\x61\x74\x6f\x72\x20\x74\x6f\x20\x76\x65\x72\x69\x66\x79\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x62\x65\x69\x6e\x67\x20\x72\x65\x71\x75\x65\x73\x74\x65\x64\x20\x6d\x61\x74\x63\x68\x65\x73\x20\x74\x68\x65\x20\x6f\x6e\x65\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x6c\x6f\x61\x64\x65\x72\x20\x63\x61\x6e\x20\x68\x61\x6e\x64\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x66\x69\x72\x73\x74\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x28\x73\x65\x6c\x66\x29\x20\x6d\x75\x73\x74\x20\x64\x65\x66\x69\x6e\x65\x20\x5f\x6e\x61\x6d\x65\x20\x77\x68\x69\x63\x68\x20\x74\x68\x65\x20\x73\x65\x63\x6f\x6e\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x69\x73\x0a\x20\x20\x20\x20\x63\x6f\x6d\x70\x61\x72\x65\x64\x20\x61\x67\x61\x69\x6e\x73\x74\x2e\x20\x49\x66\x20\x74\x68\x65\x20\x63\x6f\x6d\x70\x61\x72\x69\x73\x6f\x6e\x20\x66\x61\x69\x6c\x73\x20\x74\x68\x65\x6e\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
importlib__bootstrap_external_toplevel_consts_41_consts_1_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader for ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
importlib__bootstrap_external_toplevel_consts_41_consts_1_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " cannot handle ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_41_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_41_consts_1_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_41_consts_1_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_41_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__check_name_wrapper = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_check_name_wrapper",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[41];
    }
importlib__bootstrap_external_toplevel_consts_41_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 40,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_check_name.<locals>._check_name_wrapper",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[83];
    }
importlib__bootstrap_external_toplevel_consts_41_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 82,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x0b\x0f\x88\x3c\xd8\x13\x17\x97\x39\x91\x39\x89\x44\xd8\x0d\x11\x8f\x59\x89\x59\x98\x24\xd2\x0d\x1e\xdd\x12\x1d\xd8\x21\x25\xa7\x19\xa3\x19\xa9\x44\xf0\x03\x01\x1f\x32\xd8\x38\x3c\xf4\x03\x01\x13\x3e\xf0\x00\x01\x0d\x3e\xe1\x0f\x15\x90\x64\x98\x44\xd0\x0f\x32\xa0\x34\xd2\x0f\x32\xa8\x36\xd1\x0f\x32\xd0\x08\x32",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_41_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
            &_Py_ID(args),
            & const_str_kwargs._ascii.ob_base,
            &_Py_ID(method),
        },
    },
};
static
    struct _PyCode_DEF(148)
importlib__bootstrap_external_toplevel_consts_41_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 74,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_41_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_41_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 31,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 647,
    .co_nlocalsplus = 5,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 123,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_41_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__check_name_wrapper._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_41_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_41_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x7c\x01\x80\x0d\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x6e\x2c\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6b\x37\x00\x00\x72\x1d\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x01\x64\x02\x7c\x01\x9b\x01\x9d\x04\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x02\x00\x89\x04\x7c\x00\x7c\x01\x67\x02\x7c\x02\xa2\x01\xad\x06\x69\x00\x7c\x03\xa4\x01\x8e\x01\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_41_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_3_consts_1._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_external_toplevel_consts_41_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_check_name.<locals>._wrap",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
importlib__bootstrap_external_toplevel_consts_41_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdb\x1b\x50\x90\x07\xdc\x13\x1a\x98\x33\xa0\x07\xd5\x13\x28\xdc\x14\x1b\x98\x43\xa0\x17\xac\x27\xb0\x23\xb0\x77\xd3\x2a\x3f\xd5\x14\x40\xf0\x05\x00\x1c\x51\x01\xf0\x06\x00\x0d\x10\x8f\x4c\x89\x4c\xd7\x0c\x1f\xd1\x0c\x1f\xa0\x03\xa7\x0c\xa1\x0c\xd5\x0c\x2d",
};
static
    struct _PyCode_DEF(164)
importlib__bootstrap_external_toplevel_consts_41_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 82,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_41_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 660,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 124,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__wrap._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_41_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_41_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x44\x00\x5d\x26\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x10\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x28\x04\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_41_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_41_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_41_consts_1.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_41_consts_2.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_41_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(_bootstrap),
            & const_str__wrap._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__check_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_check_name",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
importlib__bootstrap_external_toplevel_consts_41_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf5\x10\x06\x05\x33\xf4\x14\x00\x08\x12\xd0\x07\x1d\xdc\x10\x1a\xd7\x10\x20\xd1\x10\x20\x89\x05\xf2\x04\x04\x09\x2e\xf1\x0c\x00\x05\x0a\xd0\x0a\x1d\x98\x76\xd4\x04\x26\xd8\x0b\x1e\xd0\x04\x1e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_41_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(method),
            & const_str__check_name_wrapper._ascii.ob_base,
            & const_str__wrap._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[4];
    }
importlib__bootstrap_external_toplevel_consts_41_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 3,
    },
    .ob_shash = -1,
    .ob_sval = "`  ",
};
static
    struct _PyCode_DEF(90)
importlib__bootstrap_external_toplevel_consts_41 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 45,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_41_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_41_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 639,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 125,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_41_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_41_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__check_name._ascii.ob_base,
    .co_qualname = & const_str__check_name._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_41_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x64\x03\x88\x00\x66\x01\x64\x01\x84\x09\x7d\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x11\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x6e\x03\x64\x02\x84\x00\x7d\x02\x02\x00\x7c\x02\x7c\x01\x89\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[597];
    }
importlib__bootstrap_external_toplevel_consts_42_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 596,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x50\x65\x72\x66\x6f\x72\x6d\x20\x62\x61\x73\x69\x63\x20\x76\x61\x6c\x69\x64\x69\x74\x79\x20\x63\x68\x65\x63\x6b\x69\x6e\x67\x20\x6f\x66\x20\x61\x20\x70\x79\x63\x20\x68\x65\x61\x64\x65\x72\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x66\x6c\x61\x67\x73\x20\x66\x69\x65\x6c\x64\x2c\x0a\x20\x20\x20\x20\x77\x68\x69\x63\x68\x20\x64\x65\x74\x65\x72\x6d\x69\x6e\x65\x73\x20\x68\x6f\x77\x20\x74\x68\x65\x20\x70\x79\x63\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x20\x66\x75\x72\x74\x68\x65\x72\x20\x76\x61\x6c\x69\x64\x61\x74\x65\x64\x20\x61\x67\x61\x69\x6e\x73\x74\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x64\x61\x74\x61\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x63\x6f\x6e\x74\x65\x6e\x74\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x79\x63\x20\x66\x69\x6c\x65\x2e\x20\x28\x4f\x6e\x6c\x79\x20\x74\x68\x65\x20\x66\x69\x72\x73\x74\x20\x31\x36\x20\x62\x79\x74\x65\x73\x20\x61\x72\x65\x0a\x20\x20\x20\x20\x72\x65\x71\x75\x69\x72\x65\x64\x2c\x20\x74\x68\x6f\x75\x67\x68\x2e\x29\x0a\x0a\x20\x20\x20\x20\x2a\x6e\x61\x6d\x65\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x62\x65\x69\x6e\x67\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x20\x49\x74\x20\x69\x73\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x6c\x6f\x67\x67\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x65\x78\x63\x5f\x64\x65\x74\x61\x69\x6c\x73\x2a\x20\x69\x73\x20\x61\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x70\x61\x73\x73\x65\x64\x20\x74\x6f\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x69\x74\x20\x72\x61\x69\x73\x65\x64\x20\x66\x6f\x72\x0a\x20\x20\x20\x20\x69\x6d\x70\x72\x6f\x76\x65\x64\x20\x64\x65\x62\x75\x67\x67\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x20\x77\x68\x65\x6e\x20\x74\x68\x65\x20\x6d\x61\x67\x69\x63\x20\x6e\x75\x6d\x62\x65\x72\x20\x69\x73\x20\x69\x6e\x63\x6f\x72\x72\x65\x63\x74\x20\x6f\x72\x20\x77\x68\x65\x6e\x20\x74\x68\x65\x20\x66\x6c\x61\x67\x73\x0a\x20\x20\x20\x20\x66\x69\x65\x6c\x64\x20\x69\x73\x20\x69\x6e\x76\x61\x6c\x69\x64\x2e\x20\x45\x4f\x46\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x20\x77\x68\x65\x6e\x20\x74\x68\x65\x20\x64\x61\x74\x61\x20\x69\x73\x20\x66\x6f\x75\x6e\x64\x20\x74\x6f\x20\x62\x65\x20\x74\x72\x75\x6e\x63\x61\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_external_toplevel_consts_42_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bad magic number in ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_external_toplevel_consts_42_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ": ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
importlib__bootstrap_external_toplevel_consts_42_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "{}",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[41];
    }
importlib__bootstrap_external_toplevel_consts_42_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 40,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "reached EOF while reading pyc header of ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
importlib__bootstrap_external_toplevel_consts_42_consts_10 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "invalid flags ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
importlib__bootstrap_external_toplevel_consts_42_consts_11 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " in ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_42_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_42_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 4],
            & importlib__bootstrap_external_toplevel_consts_42_consts_3._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_42_consts_4._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_42_consts_5._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            & importlib__bootstrap_external_toplevel_consts_42_consts_7._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -4],
            & importlib__bootstrap_external_toplevel_consts_42_consts_10._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_42_consts_11._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_MAGIC_NUMBER = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MAGIC_NUMBER",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_EOFError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "EOFError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_42_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_MAGIC_NUMBER._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(len),
            & const_str_EOFError._ascii.ob_base,
            & const_str__unpack_uint32._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__classify_pyc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_classify_pyc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[186];
    }
importlib__bootstrap_external_toplevel_consts_42_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 185,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x20\x00\x0d\x11\x90\x12\x90\x21\x88\x48\x80\x45\xd8\x07\x0c\x94\x0c\xd2\x07\x1c\xd8\x14\x28\xa8\x14\xa8\x08\xb0\x02\xb0\x35\xb0\x29\xd0\x12\x3c\x88\x07\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xa0\x44\xa8\x27\xd4\x08\x32\xdc\x0e\x19\x98\x27\xd1\x0e\x31\xa0\x5b\xd1\x0e\x31\xd0\x08\x31\xdc\x07\x0a\x88\x34\x83\x79\x90\x32\x82\x7e\xd8\x14\x3c\xb8\x54\xb8\x48\xd0\x12\x45\x88\x07\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xa0\x44\xa8\x27\xd4\x08\x32\xdc\x0e\x16\x90\x77\xd3\x0e\x1f\xd0\x08\x1f\xdc\x0c\x1a\x98\x34\xa0\x01\xa0\x21\x98\x39\xd3\x0c\x25\x80\x45\xe0\x07\x0c\x88\x75\x82\x7d\xd8\x14\x22\xa0\x35\xa0\x29\xa8\x34\xb0\x04\xa8\x78\xd0\x12\x38\x88\x07\xdc\x0e\x19\x98\x27\xd1\x0e\x31\xa0\x5b\xd1\x0e\x31\xd0\x08\x31\xd8\x0b\x10\x80\x4c",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_exc_details = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exc_details",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_magic = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "magic",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_42_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(data),
            &_Py_ID(name),
            & const_str_exc_details._ascii.ob_base,
            & const_str_magic._ascii.ob_base,
            &_Py_ID(message),
            &_Py_ID(flags),
        },
    },
};
static
    struct _PyCode_DEF(300)
importlib__bootstrap_external_toplevel_consts_42 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 150,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_42_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_42_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 670,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 126,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_42_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__classify_pyc._ascii.ob_base,
    .co_qualname = & const_str__classify_pyc._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_42_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x64\x01\x64\x02\x1a\x00\x7d\x03\x7c\x03\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x2a\x64\x03\x7c\x01\x9b\x02\x64\x04\x7c\x03\x9b\x02\x9d\x04\x7d\x04\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x66\x01\x69\x00\x7c\x02\xa4\x01\x8e\x01\x82\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x06\x6b\x02\x00\x00\x72\x26\x64\x07\x7c\x01\x9b\x02\x9d\x02\x7d\x04\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\x64\x08\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x64\x09\x7a\x01\x00\x00\x72\x14\x64\x0a\x7c\x05\x9b\x02\x64\x0b\x7c\x01\x9b\x02\x9d\x04\x7d\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x66\x01\x69\x00\x7c\x02\xa4\x01\x8e\x01\x82\x01\x7c\x05\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[520];
    }
importlib__bootstrap_external_toplevel_consts_43_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 519,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x56\x61\x6c\x69\x64\x61\x74\x65\x20\x61\x20\x70\x79\x63\x20\x61\x67\x61\x69\x6e\x73\x74\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x6c\x61\x73\x74\x2d\x6d\x6f\x64\x69\x66\x69\x65\x64\x20\x74\x69\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x64\x61\x74\x61\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x63\x6f\x6e\x74\x65\x6e\x74\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x79\x63\x20\x66\x69\x6c\x65\x2e\x20\x28\x4f\x6e\x6c\x79\x20\x74\x68\x65\x20\x66\x69\x72\x73\x74\x20\x31\x36\x20\x62\x79\x74\x65\x73\x20\x61\x72\x65\x0a\x20\x20\x20\x20\x72\x65\x71\x75\x69\x72\x65\x64\x2e\x29\x0a\x0a\x20\x20\x20\x20\x2a\x73\x6f\x75\x72\x63\x65\x5f\x6d\x74\x69\x6d\x65\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x6c\x61\x73\x74\x20\x6d\x6f\x64\x69\x66\x69\x65\x64\x20\x74\x69\x6d\x65\x73\x74\x61\x6d\x70\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x73\x6f\x75\x72\x63\x65\x5f\x73\x69\x7a\x65\x2a\x20\x69\x73\x20\x4e\x6f\x6e\x65\x20\x6f\x72\x20\x74\x68\x65\x20\x73\x69\x7a\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x66\x69\x6c\x65\x20\x69\x6e\x20\x62\x79\x74\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x6e\x61\x6d\x65\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x62\x65\x69\x6e\x67\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x20\x49\x74\x20\x69\x73\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x6c\x6f\x67\x67\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x65\x78\x63\x5f\x64\x65\x74\x61\x69\x6c\x73\x2a\x20\x69\x73\x20\x61\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x70\x61\x73\x73\x65\x64\x20\x74\x6f\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x69\x74\x20\x72\x61\x69\x73\x65\x64\x20\x66\x6f\x72\x0a\x20\x20\x20\x20\x69\x6d\x70\x72\x6f\x76\x65\x64\x20\x64\x65\x62\x75\x67\x67\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x41\x6e\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x20\x69\x66\x20\x74\x68\x65\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x69\x73\x20\x73\x74\x61\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_external_toplevel_consts_43_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytecode is stale for ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_43_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_43_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 12],
            & const_int_4294967295.ob_base,
            & importlib__bootstrap_external_toplevel_consts_43_consts_4._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_42_consts_5._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_43_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__unpack_uint32._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str__validate_timestamp_pyc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_validate_timestamp_pyc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[138];
    }
importlib__bootstrap_external_toplevel_consts_43_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 137,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x26\x00\x08\x16\x90\x64\x98\x31\x98\x52\x90\x6a\xd3\x07\x21\xa0\x6c\xb0\x5a\xd1\x26\x3f\xd2\x07\x40\xd8\x14\x2a\xa8\x34\xa8\x28\xd0\x12\x33\x88\x07\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xa0\x44\xa8\x27\xd4\x08\x32\xdc\x0e\x19\x98\x27\xd1\x0e\x31\xa0\x5b\xd1\x0e\x31\xd0\x08\x31\xd8\x08\x13\xd0\x08\x1f\xdc\x08\x16\x90\x74\x98\x42\x98\x72\x90\x7b\xd3\x08\x23\xa8\x0b\xb0\x6a\xd1\x28\x40\xd2\x08\x41\xdc\x0e\x19\xd0\x1c\x32\xb0\x34\xb0\x28\xd0\x1a\x3b\xd1\x0e\x4b\xb8\x7b\xd1\x0e\x4b\xd0\x08\x4b\xf0\x03\x00\x09\x42\x01\xf0\x03\x00\x09\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_source_mtime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_mtime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_source_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_size",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_43_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(data),
            & const_str_source_mtime._ascii.ob_base,
            & const_str_source_size._ascii.ob_base,
            &_Py_ID(name),
            & const_str_exc_details._ascii.ob_base,
            &_Py_ID(message),
        },
    },
};
static
    struct _PyCode_DEF(198)
importlib__bootstrap_external_toplevel_consts_43 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 99,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_43_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_43_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 703,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 127,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_43_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__validate_timestamp_pyc._ascii.ob_base,
    .co_qualname = & const_str__validate_timestamp_pyc._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_43_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\x64\x02\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\x7a\x01\x00\x00\x6b\x37\x00\x00\x72\x27\x64\x04\x7c\x03\x9b\x02\x9d\x02\x7d\x05\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x05\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x66\x01\x69\x00\x7c\x04\xa4\x01\x8e\x01\x82\x01\x7c\x02\x81\x24\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\x64\x07\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x03\x7a\x01\x00\x00\x6b\x37\x00\x00\x72\x0f\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x03\x9b\x02\x9d\x02\x66\x01\x69\x00\x7c\x04\xa4\x01\x8e\x01\x82\x01\x79\x06\x79\x06",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[500];
    }
importlib__bootstrap_external_toplevel_consts_44_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 499,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x56\x61\x6c\x69\x64\x61\x74\x65\x20\x61\x20\x68\x61\x73\x68\x2d\x62\x61\x73\x65\x64\x20\x70\x79\x63\x20\x62\x79\x20\x63\x68\x65\x63\x6b\x69\x6e\x67\x20\x74\x68\x65\x20\x72\x65\x61\x6c\x20\x73\x6f\x75\x72\x63\x65\x20\x68\x61\x73\x68\x20\x61\x67\x61\x69\x6e\x73\x74\x20\x74\x68\x65\x20\x6f\x6e\x65\x20\x69\x6e\x0a\x20\x20\x20\x20\x74\x68\x65\x20\x70\x79\x63\x20\x68\x65\x61\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x64\x61\x74\x61\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x63\x6f\x6e\x74\x65\x6e\x74\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x79\x63\x20\x66\x69\x6c\x65\x2e\x20\x28\x4f\x6e\x6c\x79\x20\x74\x68\x65\x20\x66\x69\x72\x73\x74\x20\x31\x36\x20\x62\x79\x74\x65\x73\x20\x61\x72\x65\x0a\x20\x20\x20\x20\x72\x65\x71\x75\x69\x72\x65\x64\x2e\x29\x0a\x0a\x20\x20\x20\x20\x2a\x73\x6f\x75\x72\x63\x65\x5f\x68\x61\x73\x68\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x2e\x75\x74\x69\x6c\x2e\x73\x6f\x75\x72\x63\x65\x5f\x68\x61\x73\x68\x28\x29\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x6e\x61\x6d\x65\x2a\x20\x69\x73\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x62\x65\x69\x6e\x67\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x20\x49\x74\x20\x69\x73\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x6c\x6f\x67\x67\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x65\x78\x63\x5f\x64\x65\x74\x61\x69\x6c\x73\x2a\x20\x69\x73\x20\x61\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x70\x61\x73\x73\x65\x64\x20\x74\x6f\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x69\x74\x20\x72\x61\x69\x73\x65\x64\x20\x66\x6f\x72\x0a\x20\x20\x20\x20\x69\x6d\x70\x72\x6f\x76\x65\x64\x20\x64\x65\x62\x75\x67\x67\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x41\x6e\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x20\x69\x66\x20\x74\x68\x65\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x69\x73\x20\x73\x74\x61\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
importlib__bootstrap_external_toplevel_consts_44_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hash in bytecode doesn't match hash of source ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_44_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_44_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            & importlib__bootstrap_external_toplevel_consts_44_consts_3._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_44_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__validate_hash_pyc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_validate_hash_pyc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[55];
    }
importlib__bootstrap_external_toplevel_consts_44_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 54,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x22\x00\x08\x0c\x88\x41\x88\x62\x80\x7a\x90\x5b\xd2\x07\x20\xdc\x0e\x19\xd8\x0e\x3d\xb8\x64\xb8\x58\xd0\x0c\x46\xf1\x03\x03\x0f\x0a\xe0\x0e\x19\xf1\x05\x03\x0f\x0a\xf0\x00\x03\x09\x0a\xf0\x03\x00\x08\x21",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_source_hash = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_hash",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_44_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(data),
            & const_str_source_hash._ascii.ob_base,
            &_Py_ID(name),
            & const_str_exc_details._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(50)
importlib__bootstrap_external_toplevel_consts_44 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_44_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_44_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 731,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 128,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_44_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__validate_hash_pyc._ascii.ob_base,
    .co_qualname = & const_str__validate_hash_pyc._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_44_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x64\x01\x64\x02\x1a\x00\x7c\x01\x6b\x37\x00\x00\x72\x0f\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x02\x9b\x02\x9d\x02\x66\x01\x69\x00\x7c\x03\xa4\x01\x8e\x01\x82\x01\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
importlib__bootstrap_external_toplevel_consts_45_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Compile bytecode as found in a pyc.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
importlib__bootstrap_external_toplevel_consts_45_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "code object from {!r}",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_45_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Non-code object in ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_45_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(path),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_45_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_45_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_45_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_45_consts_2._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_45_consts_3._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_marshal = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "marshal",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_loads = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loads",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__code_type = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_code_type",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__fix_co_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fix_co_filename",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_45_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_marshal._ascii.ob_base,
            & const_str_loads._ascii.ob_base,
            &_Py_ID(isinstance),
            & const_str__code_type._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str__fix_co_filename._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__compile_bytecode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_compile_bytecode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[100];
    }
importlib__bootstrap_external_toplevel_consts_45_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 99,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x12\x8f\x3d\x89\x3d\x98\x14\xd3\x0b\x1e\x80\x44\xdc\x07\x11\x90\x24\x9c\x0a\xd4\x07\x23\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xd0\x24\x3b\xb8\x5d\xd4\x08\x4b\xd8\x0b\x16\xd0\x0b\x22\xdc\x0c\x10\xd7\x0c\x21\xd1\x0c\x21\xa0\x24\xa8\x0b\xd4\x0c\x34\xd8\x0f\x13\x88\x0b\xe4\x0e\x19\xd0\x1c\x2f\xb0\x0d\xd0\x2f\x40\xd0\x1a\x41\xd8\x1f\x23\xa8\x2d\xf4\x03\x01\x0f\x39\xf0\x00\x01\x09\x39",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_45_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(data),
            &_Py_ID(name),
            & const_str_bytecode_path._ascii.ob_base,
            & const_str_source_path._ascii.ob_base,
            &_Py_ID(code),
        },
    },
};
static
    struct _PyCode_DEF(206)
importlib__bootstrap_external_toplevel_consts_45 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 103,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_45_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_45_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 755,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 129,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_45_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__compile_bytecode._ascii.ob_base,
    .co_qualname = & const_str__compile_bytecode._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_45_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x30\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x81\x16\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x53\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x02\x9b\x02\x9d\x02\x7c\x01\x7c\x02\xac\x03\xab\x03\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[44];
    }
importlib__bootstrap_external_toplevel_consts_46_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 43,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Produce the data for a timestamp-based pyc.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_46_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_46_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_bytearray = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytearray",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_dumps = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dumps",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_46_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_bytearray._ascii.ob_base,
            & const_str_MAGIC_NUMBER._ascii.ob_base,
            &_Py_ID(extend),
            & const_str__pack_uint32._ascii.ob_base,
            & const_str_marshal._ascii.ob_base,
            & const_str_dumps._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str__code_to_timestamp_pyc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_code_to_timestamp_pyc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[89];
    }
importlib__bootstrap_external_toplevel_consts_46_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 88,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x14\x94\x5c\xd3\x0b\x22\x80\x44\xd8\x04\x08\x87\x4b\x81\x4b\x94\x0c\x98\x51\x93\x0f\xd4\x04\x20\xd8\x04\x08\x87\x4b\x81\x4b\x94\x0c\x98\x55\xd3\x10\x23\xd4\x04\x24\xd8\x04\x08\x87\x4b\x81\x4b\x94\x0c\x98\x5b\xd3\x10\x29\xd4\x04\x2a\xd8\x04\x08\x87\x4b\x81\x4b\x94\x07\x97\x0d\x91\x0d\x98\x64\xd3\x10\x23\xd4\x04\x24\xd8\x0b\x0f\x80\x4b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_mtime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mtime",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_46_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(code),
            & const_str_mtime._ascii.ob_base,
            & const_str_source_size._ascii.ob_base,
            &_Py_ID(data),
        },
    },
};
static
    struct _PyCode_DEF(264)
importlib__bootstrap_external_toplevel_consts_46 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 132,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_46_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_46_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 768,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 130,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_46_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__code_to_timestamp_pyc._ascii.ob_base,
    .co_qualname = & const_str__code_to_timestamp_pyc._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_46_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_external_toplevel_consts_47_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Produce the data for a hash-based pyc.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_47_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_47_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_47_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_bytearray._ascii.ob_base,
            & const_str_MAGIC_NUMBER._ascii.ob_base,
            &_Py_ID(extend),
            & const_str__pack_uint32._ascii.ob_base,
            &_Py_ID(len),
            & const_str_marshal._ascii.ob_base,
            & const_str_dumps._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__code_to_hash_pyc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_code_to_hash_pyc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[101];
    }
importlib__bootstrap_external_toplevel_consts_47_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 100,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x14\x94\x5c\xd3\x0b\x22\x80\x44\xd8\x0c\x0f\x90\x27\x98\x51\x91\x2c\xd1\x0c\x1e\x80\x45\xd8\x04\x08\x87\x4b\x81\x4b\x94\x0c\x98\x55\xd3\x10\x23\xd4\x04\x24\xdc\x0b\x0e\x88\x7b\xd3\x0b\x1b\x98\x71\xd2\x0b\x20\xd0\x04\x20\xd0\x0b\x20\xd8\x04\x08\x87\x4b\x81\x4b\x90\x0b\xd4\x04\x1c\xd8\x04\x08\x87\x4b\x81\x4b\x94\x07\x97\x0d\x91\x0d\x98\x64\xd3\x10\x23\xd4\x04\x24\xd8\x0b\x0f\x80\x4b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_checked = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "checked",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_47_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(code),
            & const_str_source_hash._ascii.ob_base,
            & const_str_checked._ascii.ob_base,
            &_Py_ID(data),
            &_Py_ID(flags),
        },
    },
};
static
    struct _PyCode_DEF(242)
importlib__bootstrap_external_toplevel_consts_47 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 121,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_47_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_47_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 778,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 131,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_47_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__code_to_hash_pyc._ascii.ob_base,
    .co_qualname = & const_str__code_to_hash_pyc._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_47_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x64\x01\x7c\x02\x64\x01\x7a\x03\x00\x00\x7a\x07\x00\x00\x7d\x04\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x28\x00\x00\x73\x02\x4a\x00\x82\x01\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[122];
    }
importlib__bootstrap_external_toplevel_consts_48_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 121,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x65\x63\x6f\x64\x65\x20\x62\x79\x74\x65\x73\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x69\x6e\x67\x20\x73\x6f\x75\x72\x63\x65\x20\x63\x6f\x64\x65\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x73\x74\x72\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x55\x6e\x69\x76\x65\x72\x73\x61\x6c\x20\x6e\x65\x77\x6c\x69\x6e\x65\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x69\x73\x20\x75\x73\x65\x64\x20\x69\x6e\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_48_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_48_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            Py_True,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_tokenize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tokenize",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_BytesIO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BytesIO",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_detect_encoding = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "detect_encoding",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
const_str_IncrementalNewlineDecoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalNewlineDecoder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_48_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_tokenize._ascii.ob_base,
            &_Py_ID(_io),
            & const_str_BytesIO._ascii.ob_base,
            &_Py_ID(readline),
            & const_str_detect_encoding._ascii.ob_base,
            & const_str_IncrementalNewlineDecoder._ascii.ob_base,
            &_Py_ID(decode),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_decode_source = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "decode_source",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[96];
    }
importlib__bootstrap_external_toplevel_consts_48_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 95,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf3\x0a\x00\x05\x14\xdc\x1c\x1f\x9f\x4b\x99\x4b\xa8\x0c\xd3\x1c\x35\xd7\x1c\x3e\xd1\x1c\x3e\xd0\x04\x19\xd8\x0f\x17\xd7\x0f\x27\xd1\x0f\x27\xd0\x28\x3d\xd3\x0f\x3e\x80\x48\xdc\x16\x19\xd7\x16\x33\xd1\x16\x33\xb0\x44\xb8\x24\xd3\x16\x3f\x80\x4f\xd8\x0b\x1a\xd7\x0b\x21\xd1\x0b\x21\xa0\x2c\xd7\x22\x35\xd1\x22\x35\xb0\x68\xb8\x71\xb1\x6b\xd3\x22\x42\xd3\x0b\x43\xd0\x04\x43",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_source_bytes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_bytes",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_source_bytes_readline = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_bytes_readline",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_newline_decoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "newline_decoder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_48_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_source_bytes._ascii.ob_base,
            & const_str_tokenize._ascii.ob_base,
            & const_str_source_bytes_readline._ascii.ob_base,
            &_Py_ID(encoding),
            & const_str_newline_decoder._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(220)
importlib__bootstrap_external_toplevel_consts_48 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 110,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_48_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_48_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 789,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 132,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_48_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_decode_source._ascii.ob_base,
    .co_qualname = & const_str_decode_source._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_48_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x7d\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x01\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[318];
    }
importlib__bootstrap_external_toplevel_consts_50_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 317,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x20\x73\x70\x65\x63\x20\x62\x61\x73\x65\x64\x20\x6f\x6e\x20\x61\x20\x66\x69\x6c\x65\x20\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x54\x6f\x20\x69\x6e\x64\x69\x63\x61\x74\x65\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x61\x20\x70\x61\x63\x6b\x61\x67\x65\x2c\x20\x73\x65\x74\x0a\x20\x20\x20\x20\x73\x75\x62\x6d\x6f\x64\x75\x6c\x65\x5f\x73\x65\x61\x72\x63\x68\x5f\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x73\x20\x74\x6f\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x70\x61\x74\x68\x73\x2e\x20\x20\x41\x6e\x0a\x20\x20\x20\x20\x65\x6d\x70\x74\x79\x20\x6c\x69\x73\x74\x20\x69\x73\x20\x73\x75\x66\x66\x69\x63\x69\x65\x6e\x74\x2c\x20\x74\x68\x6f\x75\x67\x68\x20\x69\x74\x73\x20\x6e\x6f\x74\x20\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x75\x73\x65\x66\x75\x6c\x20\x74\x6f\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x69\x6d\x70\x6f\x72\x74\x20\x73\x79\x73\x74\x65\x6d\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x20\x6d\x75\x73\x74\x20\x74\x61\x6b\x65\x20\x61\x20\x73\x70\x65\x63\x20\x61\x73\x20\x69\x74\x73\x20\x6f\x6e\x6c\x79\x20\x5f\x5f\x69\x6e\x69\x74\x5f\x5f\x28\x29\x20\x61\x72\x67\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_50_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_50_consts_0._ascii.ob_base,
            Py_None,
            &_Py_STR(anon_unknown),
            & const_str_get_filename._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_34_consts_2._object.ob_base.ob_base,
            Py_True,
            & const_str_is_package._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
const_str__get_supported_file_loaders = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_supported_file_loaders",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__POPULATE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_POPULATE",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_50_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            & const_str_hasattr._ascii.ob_base,
            & const_str_get_filename._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__path_abspath._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str_ModuleSpec._ascii.ob_base,
            & const_str__set_fileattr._ascii.ob_base,
            & const_str__get_supported_file_loaders._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str__POPULATE._ascii.ob_base,
            & const_str_is_package._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str__path_split._ascii.ob_base,
            &_Py_ID(append),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[362];
    }
importlib__bootstrap_external_toplevel_consts_50_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 361,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x18\x00\x08\x10\xd0\x07\x17\xf0\x08\x00\x14\x1f\x88\x08\xdc\x0b\x12\x90\x36\x98\x3e\xd4\x0b\x2a\xf0\x04\x03\x0d\x15\xd8\x1b\x21\xd7\x1b\x2e\xd1\x1b\x2e\xa8\x74\xd3\x1b\x34\x91\x08\xf4\x08\x00\x14\x17\x97\x3a\x91\x3a\x98\x68\xd3\x13\x27\x88\x08\xf0\x02\x03\x09\x11\xdc\x17\x24\xa0\x58\xd3\x17\x2e\x88\x48\xf4\x14\x00\x0c\x16\xd7\x0b\x20\xd1\x0b\x20\xa0\x14\xa0\x76\xb0\x68\xd0\x0b\x20\xd3\x0b\x3f\x80\x44\xd8\x19\x1d\x80\x44\xd4\x04\x16\xf0\x06\x00\x08\x0e\x80\x7e\xdc\x26\x41\xd6\x26\x43\xd1\x0c\x22\x88\x4c\x98\x28\xd8\x0f\x17\xd7\x0f\x20\xd1\x0f\x20\xa4\x15\xa0\x78\xa3\x1f\xd5\x0f\x31\xd9\x19\x25\xa0\x64\xa8\x48\xd3\x19\x35\x90\x06\xd8\x1e\x24\x90\x04\x94\x0b\xd9\x10\x15\xf0\x09\x00\x27\x44\x01\xf0\x0c\x00\x14\x18\xf0\x06\x00\x08\x22\xa4\x59\xd1\x07\x2e\xe4\x0b\x12\x90\x36\x98\x3c\xd4\x0b\x28\xf0\x02\x06\x0d\x39\xd8\x1d\x23\xd7\x1d\x2e\xd1\x1d\x2e\xa8\x74\xd3\x1d\x34\x90\x0a\xf1\x08\x00\x14\x1e\xd8\x36\x38\x90\x44\xd5\x14\x33\xe0\x2a\x44\x88\x04\xd4\x08\x27\xd8\x07\x0b\xd7\x07\x26\xd1\x07\x26\xa8\x22\xd2\x07\x2c\xd9\x0b\x13\xdc\x16\x21\xa0\x28\xd3\x16\x2b\xa8\x41\xd1\x16\x2e\x88\x47\xd8\x0c\x10\xd7\x0c\x2b\xd1\x0c\x2b\xd7\x0c\x32\xd1\x0c\x32\xb0\x37\xd4\x0c\x3b\xe0\x0b\x0f\x80\x4b\xf8\xf4\x5d\x01\x00\x14\x1f\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfb\xf4\x0c\x00\x10\x17\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfb\xf4\x38\x00\x14\x1f\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[53];
    }
importlib__bootstrap_external_toplevel_consts_50_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 52,
    },
    .ob_shash = -1,
    .ob_sval = "\x92\x11\x44\x18\x00\xba\x0b\x44\x27\x00\xc2\x3a\x11\x44\x36\x00\xc4\x18\x09\x44\x24\x03\xc4\x23\x01\x44\x24\x03\xc4\x27\x09\x44\x33\x03\xc4\x32\x01\x44\x33\x03\xc4\x36\x09\x45\x02\x03\xc5\x01\x01\x45\x02\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_loader_class = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader_class",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_suffixes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "suffixes",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_dirname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dirname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_50_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_location._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            & const_str_loader_class._ascii.ob_base,
            & const_str_suffixes._ascii.ob_base,
            & const_str_is_package._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(650)
importlib__bootstrap_external_toplevel_consts_50 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 325,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_50_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_50_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_50_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 2,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 806,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 133,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_50_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_spec_from_file_location._ascii.ob_base,
    .co_qualname = & const_str_spec_from_file_location._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_50_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x80\x21\x64\x02\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x72\x34\x09\x00\x7c\x02\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x6e\x21\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\x7c\x01\xac\x04\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x04\x64\x05\x7c\x04\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x80\x3f\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x31\x00\x00\x5c\x02\x00\x00\x7d\x05\x7d\x06\x7c\x01\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x21\x02\x00\x7c\x05\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x7c\x04\x5f\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x02\x04\x00\x79\x01\x7c\x03\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x28\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x72\x23\x09\x00\x7c\x02\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x07\x72\x0f\x67\x00\x7c\x04\x5f\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6e\x07\x7c\x03\x7c\x04\x5f\x10\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x6b\x28\x00\x00\x72\x2b\x7c\x01\x72\x29\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x07\x19\x00\x00\x00\x7d\x08\x7c\x04\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xde\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xed\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x65\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[68];
    }
importlib__bootstrap_external_toplevel_consts_51_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 67,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x48\x65\x6c\x70\x65\x72\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x66\x6f\x72\x20\x5f\x77\x61\x72\x6e\x69\x6e\x67\x73\x2e\x63\x0a\x0a\x20\x20\x20\x20\x53\x65\x65\x20\x47\x48\x23\x39\x37\x38\x35\x30\x20\x66\x6f\x72\x20\x64\x65\x74\x61\x69\x6c\x73\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[44];
    }
importlib__bootstrap_external_toplevel_consts_51_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 43,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Module globals is missing a __spec__.loader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[46];
    }
importlib__bootstrap_external_toplevel_consts_51_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 45,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Module globals; __loader__ != __spec__.loader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_51_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_51_consts_0._ascii.ob_base,
            Py_None,
            &_Py_ID(__loader__),
            &_Py_ID(__spec__),
            & importlib__bootstrap_external_toplevel_consts_51_consts_4._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_51_consts_6._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_51_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            &_Py_ID(dict),
            &_Py_ID(object),
            &_Py_ID(get),
            & const_str_ValueError._ascii.ob_base,
            &_Py_ID(getattr),
            & const_str_AttributeError._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_DeprecationWarning._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__bless_my_loader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_bless_my_loader",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[233];
    }
importlib__bootstrap_external_toplevel_consts_51_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 232,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x18\x00\x0c\x16\x90\x6e\xa4\x64\xd4\x0b\x2b\xd8\x0f\x13\xe4\x0e\x14\x8b\x68\x80\x47\xd8\x0d\x1b\xd7\x0d\x1f\xd1\x0d\x1f\xa0\x0c\xa8\x64\xd3\x0d\x33\x80\x46\xd8\x0b\x19\xd7\x0b\x1d\xd1\x0b\x1d\x98\x6a\xa8\x27\xd3\x0b\x32\x80\x44\xe0\x07\x0d\x80\x7e\xd8\x0b\x0f\x90\x37\x89\x3f\xf0\x06\x00\x14\x18\xd8\x0d\x11\x88\x5c\xdc\x12\x1c\xd0\x1d\x4a\xd3\x12\x4b\xd0\x0c\x4b\xe4\x12\x19\x98\x24\xa0\x08\xa8\x27\xd3\x12\x32\x80\x4b\xe0\x07\x12\x90\x77\xa0\x04\x90\x6f\xd1\x07\x25\xd8\x0b\x11\x88\x3e\xd8\x24\x2f\xb0\x37\xd1\x24\x3a\x95\x2e\xc4\x0a\x88\x43\xd9\x12\x15\xd0\x16\x43\xd3\x12\x44\xd0\x0c\x44\xdc\x08\x11\x8f\x0e\x89\x0e\xd8\x0c\x39\xdc\x0c\x1e\xf4\x05\x02\x09\x20\xf0\x06\x00\x17\x1d\x88\x0b\xe0\x0b\x16\xd0\x0b\x22\xd0\x04\x22\xd0\x0b\x22\xd8\x07\x0d\xd0\x07\x19\x98\x66\xa8\x0b\xd2\x1e\x33\xdc\x08\x11\x8f\x0e\x89\x0e\xd8\x0c\x3b\xdc\x0c\x1e\xf4\x05\x02\x09\x20\xf0\x06\x00\x10\x16\x88\x0d\xe0\x0b\x16\xd0\x04\x16",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_missing = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "missing",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_spec_loader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spec_loader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_51_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(module_globals),
            & const_str_missing._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            & const_str_spec_loader._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(396)
importlib__bootstrap_external_toplevel_consts_51 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 198,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_51_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_51_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 876,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 134,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_51_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__bless_my_loader._ascii.ob_base,
    .co_qualname = & const_str__bless_my_loader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_51_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x79\x01\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x02\x80\x12\x7c\x03\x7c\x01\x75\x00\x72\x01\x79\x01\x7c\x03\x80\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x05\x7c\x01\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x7c\x01\x64\x01\x66\x02\x76\x00\x72\x36\x7c\x02\x80\x18\x7c\x04\x7c\x01\x75\x00\x72\x06\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x6e\x05\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x02\x00\x7c\x05\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x7d\x04\x7c\x04\x80\x02\x4a\x00\x82\x01\x7c\x02\x81\x21\x7c\x02\x7c\x04\x6b\x37\x00\x00\x72\x1c\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x53\x00\x7c\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_WindowsRegistryFinder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "WindowsRegistryFinder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[63];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 62,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Meta path finder for modules declared in the Windows registry.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[60];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 59,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Software\\Python\\PythonCore\\{sys_version}\\Modules\\{fullname}",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[66];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 65,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Software\\Python\\PythonCore\\{sys_version}\\Modules\\{fullname}\\Debug",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_d.pyd",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_winreg = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "winreg",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_OpenKey = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "OpenKey",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_HKEY_CURRENT_USER = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HKEY_CURRENT_USER",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_HKEY_LOCAL_MACHINE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HKEY_LOCAL_MACHINE",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_winreg._ascii.ob_base,
            & const_str_OpenKey._ascii.ob_base,
            & const_str_HKEY_CURRENT_USER._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_HKEY_LOCAL_MACHINE._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__open_registry = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_open_registry",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[37];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 36,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "WindowsRegistryFinder._open_registry",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[72];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 71,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x09\x42\x01\xdc\x13\x19\x97\x3e\x91\x3e\xa4\x26\xd7\x22\x3a\xd1\x22\x3a\xb8\x43\xd3\x13\x40\xd0\x0c\x40\xf8\xdc\x0f\x16\xf2\x00\x01\x09\x42\x01\xdc\x13\x19\x97\x3e\x91\x3e\xa4\x26\xd7\x22\x3b\xd1\x22\x3b\xb8\x53\xd3\x13\x41\xd2\x0c\x41\xf0\x03\x01\x09\x42\x01\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x23\x26\x00\xa6\x2d\x41\x16\x03\xc1\x15\x01\x41\x16\x03",
};
static
    struct _PyCode_DEF(178)
importlib__bootstrap_external_toplevel_consts_52_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 89,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_52_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_52_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 938,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 135,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_1._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__open_registry._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_52_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_52_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x27\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "%d.%d",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_sys_version = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys_version",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_fullname._ascii.ob_base,
            & const_str_sys_version._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_52_consts_6_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & importlib__bootstrap_external_toplevel_consts_52_consts_6_consts_3._object.ob_base.ob_base,
            &_Py_STR(empty),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_DEBUG_BUILD = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "DEBUG_BUILD",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_REGISTRY_KEY_DEBUG = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "REGISTRY_KEY_DEBUG",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_REGISTRY_KEY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "REGISTRY_KEY",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_version_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "version_info",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_QueryValue = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "QueryValue",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_DEBUG_BUILD._ascii.ob_base,
            & const_str_REGISTRY_KEY_DEBUG._ascii.ob_base,
            & const_str_REGISTRY_KEY._ascii.ob_base,
            &_Py_ID(format),
            & const_str_sys._ascii.ob_base,
            & const_str_version_info._ascii.ob_base,
            & const_str__open_registry._ascii.ob_base,
            & const_str_winreg._ascii.ob_base,
            & const_str_QueryValue._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__search_registry = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_search_registry",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "WindowsRegistryFinder._search_registry",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[160];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 159,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x8f\x3f\x8a\x3f\xd8\x1b\x1e\xd7\x1b\x31\xd1\x1b\x31\x89\x4c\xe0\x1b\x1e\xd7\x1b\x2b\xd1\x1b\x2b\x88\x4c\xd8\x0e\x1a\xd7\x0e\x21\xd1\x0e\x21\xa8\x38\xd8\x2e\x35\xbc\x03\xd7\x38\x48\xd1\x38\x48\xc8\x12\xc8\x21\xd0\x38\x4c\xd1\x2e\x4c\xf0\x03\x00\x0f\x22\xf3\x00\x01\x0f\x4e\x01\x88\x03\xf0\x04\x04\x09\x18\xd8\x11\x14\xd7\x11\x23\xd1\x11\x23\xa0\x43\xd4\x11\x28\xa8\x44\xdc\x1b\x21\xd7\x1b\x2c\xd1\x1b\x2c\xa8\x54\xb0\x32\xd3\x1b\x36\x90\x08\xf7\x03\x00\x12\x29\xf0\x08\x00\x10\x18\x88\x0f\xf7\x09\x00\x12\x29\xf0\x08\x00\x10\x18\x88\x0f\xfb\xf4\x05\x00\x10\x17\xf2\x00\x01\x09\x18\xd9\x13\x17\xf0\x03\x01\x09\x18\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x0e\x11\x42\x0d\x00\xc1\x1f\x17\x42\x00\x03\xc1\x36\x08\x42\x0d\x00\xc2\x00\x05\x42\x0a\x07\xc2\x05\x03\x42\x0d\x00\xc2\x0a\x03\x42\x0d\x00\xc2\x0d\x09\x42\x19\x03\xc2\x18\x01\x42\x19\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_registry_key = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "registry_key",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_hkey = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hkey",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            & const_str_registry_key._ascii.ob_base,
            &_Py_ID(key),
            & const_str_hkey._ascii.ob_base,
            &_Py_ID(filepath),
        },
    },
};
static
    struct _PyCode_DEF(312)
importlib__bootstrap_external_toplevel_consts_52_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 156,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_52_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_52_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_52_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 945,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 136,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_52_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__search_registry._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_52_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_52_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x0d\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x6e\x0c\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x64\x02\x1a\x00\x7a\x06\x00\x00\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x09\x00\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x04\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x7f\x05\x53\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str__search_registry._ascii.ob_base,
            & const_str__path_stat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str__get_supported_file_loaders._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str_spec_from_loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "WindowsRegistryFinder.find_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[143];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 142,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x13\x16\xd7\x13\x27\xd1\x13\x27\xa8\x08\xd3\x13\x31\x88\x08\xd8\x0b\x13\xd0\x0b\x1b\xd8\x13\x17\xf0\x02\x03\x09\x18\xdc\x0c\x16\x90\x78\xd4\x0c\x20\xf4\x06\x00\x21\x3c\xd6\x20\x3d\xd1\x0c\x1c\x88\x46\x90\x48\xd8\x0f\x17\xd7\x0f\x20\xd1\x0f\x20\xa4\x15\xa0\x78\xa3\x1f\xd5\x0f\x31\xdc\x17\x21\xd7\x17\x32\xd1\x17\x32\xb0\x38\xd9\x33\x39\xb8\x28\xc0\x48\xd3\x33\x4d\xd8\x3a\x42\xf0\x05\x00\x18\x33\xf3\x00\x02\x18\x44\x01\x90\x04\xf0\x06\x00\x18\x1c\x92\x0b\xf1\x0b\x00\x21\x3e\xf8\xf4\x05\x00\x10\x17\xf2\x00\x01\x09\x18\xd9\x13\x17\xf0\x03\x01\x09\x18\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
importlib__bootstrap_external_toplevel_consts_52_consts_8_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x96\x0b\x41\x31\x00\xc1\x31\x09\x41\x3d\x03\xc1\x3c\x01\x41\x3d\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(target),
            &_Py_ID(filepath),
            & const_str_loader._ascii.ob_base,
            & const_str_suffixes._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(256)
importlib__bootstrap_external_toplevel_consts_52_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 128,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_44_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_52_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_52_consts_8_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 960,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 137,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_52_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_find_spec._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_52_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_52_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x80\x01\x79\x00\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x42\x00\x00\x5c\x02\x00\x00\x7d\x05\x7d\x06\x7c\x04\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x21\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x02\x00\x7c\x05\x7c\x01\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x04\xac\x01\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x07\x63\x02\x01\x00\x53\x00\x04\x00\x79\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_WindowsRegistryFinder._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52_consts_2._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52_consts_3._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52_consts_4._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52_consts_5.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52_consts_6.ob_base.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_52_consts_8.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__MS_WINDOWS = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_MS_WINDOWS",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_EXTENSION_SUFFIXES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "EXTENSION_SUFFIXES",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_52_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_REGISTRY_KEY._ascii.ob_base,
            & const_str_REGISTRY_KEY_DEBUG._ascii.ob_base,
            & const_str__MS_WINDOWS._ascii.ob_base,
            & const_str_EXTENSION_SUFFIXES._ascii.ob_base,
            & const_str_DEBUG_BUILD._ascii.ob_base,
            & const_str_staticmethod._ascii.ob_base,
            & const_str__open_registry._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            & const_str__search_registry._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[104];
    }
importlib__bootstrap_external_toplevel_consts_52_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 103,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x48\xf0\x06\x01\x09\x20\xf0\x03\x00\x05\x11\xf0\x08\x01\x09\x27\xf0\x03\x00\x05\x17\xf0\x06\x00\x14\x1f\xd2\x13\x41\xa0\x38\xd0\x2f\x41\xd0\x23\x41\x80\x4b\xe0\x05\x11\xf1\x02\x04\x05\x42\x01\xf3\x03\x00\x06\x12\xf0\x02\x04\x05\x42\x01\xf0\x0c\x00\x06\x11\xf1\x02\x0c\x05\x18\xf3\x03\x00\x06\x11\xf0\x02\x0c\x05\x18\xf0\x1c\x00\x06\x11\xf2\x02\x0d\x05\x1c\xf3\x03\x00\x06\x11\xf1\x02\x0d\x05\x1c",
};
static
    struct _PyCode_DEF(90)
importlib__bootstrap_external_toplevel_consts_52 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 45,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_52_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_52_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 926,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 138,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_WindowsRegistryFinder._ascii.ob_base,
    .co_qualname = & const_str_WindowsRegistryFinder._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_52_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x03\x5a\x05\x65\x06\x78\x01\x72\x04\x01\x00\x64\x04\x65\x07\x76\x00\x5a\x08\x65\x09\x64\x05\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0a\x65\x0b\x64\x06\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0c\x65\x0b\x64\x09\x64\x08\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0d\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__LoaderBasics = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LoaderBasics",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[84];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 83,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x42\x61\x73\x65\x20\x63\x6c\x61\x73\x73\x20\x6f\x66\x20\x63\x6f\x6d\x6d\x6f\x6e\x20\x63\x6f\x64\x65\x20\x6e\x65\x65\x64\x65\x64\x20\x62\x79\x20\x62\x6f\x74\x68\x20\x53\x6f\x75\x72\x63\x65\x4c\x6f\x61\x64\x65\x72\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x53\x6f\x75\x72\x63\x65\x6c\x65\x73\x73\x46\x69\x6c\x65\x4c\x6f\x61\x64\x65\x72\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[142];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 141,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6e\x63\x72\x65\x74\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x49\x6e\x73\x70\x65\x63\x74\x4c\x6f\x61\x64\x65\x72\x2e\x69\x73\x5f\x70\x61\x63\x6b\x61\x67\x65\x20\x62\x79\x20\x63\x68\x65\x63\x6b\x69\x6e\x67\x20\x69\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x62\x79\x20\x67\x65\x74\x5f\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x68\x61\x73\x20\x61\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x6f\x66\x20\x27\x5f\x5f\x69\x6e\x69\x74\x5f\x5f\x2e\x70\x79\x27\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_54_consts_2_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            &_Py_ID(__init__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__path_split._ascii.ob_base,
            & const_str_get_filename._ascii.ob_base,
            & const_str_rsplit._ascii.ob_base,
            & const_str_rpartition._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LoaderBasics.is_package",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[94];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 93,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x14\x1f\x98\x74\xd7\x1f\x30\xd1\x1f\x30\xb0\x18\xd3\x1f\x3a\xd3\x13\x3b\xb8\x41\xd1\x13\x3e\x88\x08\xd8\x18\x20\x9f\x0f\x99\x0f\xa8\x03\xa8\x51\xd3\x18\x2f\xb0\x01\xd1\x18\x32\x88\x0d\xd8\x14\x1c\xd7\x14\x27\xd1\x14\x27\xa8\x03\xd3\x14\x2c\xa8\x51\xd1\x14\x2f\x88\x09\xd8\x0f\x1c\xa0\x0a\xd1\x0f\x2a\xd2\x0f\x46\xa8\x79\xb8\x4a\xd1\x2f\x46\xd0\x08\x46",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_filename_base = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "filename_base",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_tail_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tail_name",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(filename),
            & const_str_filename_base._ascii.ob_base,
            & const_str_tail_name._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(166)
importlib__bootstrap_external_toplevel_consts_54_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 83,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_54_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_54_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 982,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 139,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_is_package._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_54_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x7d\x02\x7c\x02\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x64\x03\x19\x00\x00\x00\x7d\x03\x7c\x01\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x64\x04\x19\x00\x00\x00\x7d\x04\x7c\x03\x64\x05\x6b\x28\x00\x00\x78\x01\x72\x05\x01\x00\x7c\x04\x64\x05\x6b\x37\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[43];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 42,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Use default semantics for module creation.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_54_consts_3_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LoaderBasics.create_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\x81\x00",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_54_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_54_consts_3_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 990,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 140,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_create_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_54_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Execute the module.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cannot load module ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " when get_code() returns None",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_54_consts_4_consts_0._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_54_consts_4_consts_2._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_54_consts_4_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_get_code._ascii.ob_base,
            &_Py_ID(__name__),
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str_exec._ascii.ob_base,
            &_Py_ID(__dict__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LoaderBasics.exec_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[84];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 83,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7d\x89\x7d\x98\x56\x9f\x5f\x99\x5f\xd3\x0f\x2d\x88\x04\xd8\x0b\x0f\x88\x3c\xdc\x12\x1d\xd0\x20\x33\xb0\x46\xb7\x4f\xb1\x4f\xd0\x33\x46\xf0\x00\x01\x47\x01\x38\xf0\x00\x01\x1f\x38\xf3\x00\x01\x13\x39\xf0\x00\x01\x0d\x39\xe4\x08\x12\xd7\x08\x2c\xd1\x08\x2c\xac\x54\xb0\x34\xb8\x16\xbf\x1f\xb9\x1f\xd5\x08\x49",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(module),
            &_Py_ID(code),
        },
    },
};
static
    struct _PyCode_DEF(186)
importlib__bootstrap_external_toplevel_consts_54_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 93,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_54_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_54_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 993,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 141,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_exec_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_54_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x80\x19\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x03\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x01\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "This method is deprecated.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_54_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(_bootstrap),
            & const_str__load_module_shim._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LoaderBasics.load_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
importlib__bootstrap_external_toplevel_consts_54_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x10\x1a\xd7\x0f\x2b\xd1\x0f\x2b\xa8\x44\xb0\x28\xd3\x0f\x3b\xd0\x08\x3b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(46)
importlib__bootstrap_external_toplevel_consts_54_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_54_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_54_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1001,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 142,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_load_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_54_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__LoaderBasics._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_54_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_54_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_54_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_54_consts_4.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_54_consts_5.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_54_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_is_package._ascii.ob_base,
            & const_str_create_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib__bootstrap_external_toplevel_consts_54_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x01\x05\x1d\xf2\x06\x06\x05\x47\x01\xf2\x10\x01\x05\x39\xf2\x06\x06\x05\x4a\x01\xf3\x10\x03\x05\x3c",
};
static
    struct _PyCode_DEF(40)
importlib__bootstrap_external_toplevel_consts_54 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_54_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_54_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 977,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 143,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__LoaderBasics._ascii.ob_base,
    .co_qualname = & const_str__LoaderBasics._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x79\x06",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_SourceLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[166];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_1_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 165,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x6d\x65\x74\x68\x6f\x64\x20\x74\x68\x61\x74\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x6d\x6f\x64\x69\x66\x69\x63\x61\x74\x69\x6f\x6e\x20\x74\x69\x6d\x65\x20\x28\x61\x6e\x20\x69\x6e\x74\x29\x20\x66\x6f\x72\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x70\x61\x74\x68\x20\x28\x61\x20\x73\x74\x72\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x4f\x53\x45\x72\x72\x6f\x72\x20\x77\x68\x65\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x68\x61\x6e\x64\x6c\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_56_consts_1_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_path_mtime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_mtime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader.path_mtime",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x0f\x16\x88\x0d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
        },
    },
};
static
    struct _PyCode_DEF(14)
importlib__bootstrap_external_toplevel_consts_56_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_56_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1009,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 144,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_56_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_path_mtime._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_56_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_56_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[415];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 414,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x6d\x65\x74\x68\x6f\x64\x20\x72\x65\x74\x75\x72\x6e\x69\x6e\x67\x20\x61\x20\x6d\x65\x74\x61\x64\x61\x74\x61\x20\x64\x69\x63\x74\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x61\x74\x68\x20\x28\x61\x20\x73\x74\x72\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x50\x6f\x73\x73\x69\x62\x6c\x65\x20\x6b\x65\x79\x73\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x2d\x20\x27\x6d\x74\x69\x6d\x65\x27\x20\x28\x6d\x61\x6e\x64\x61\x74\x6f\x72\x79\x29\x20\x69\x73\x20\x74\x68\x65\x20\x6e\x75\x6d\x65\x72\x69\x63\x20\x74\x69\x6d\x65\x73\x74\x61\x6d\x70\x20\x6f\x66\x20\x6c\x61\x73\x74\x20\x73\x6f\x75\x72\x63\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6f\x64\x65\x20\x6d\x6f\x64\x69\x66\x69\x63\x61\x74\x69\x6f\x6e\x3b\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x2d\x20\x27\x73\x69\x7a\x65\x27\x20\x28\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x29\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x7a\x65\x20\x69\x6e\x20\x62\x79\x74\x65\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x63\x6f\x64\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x61\x6c\x6c\x6f\x77\x73\x20\x74\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x20\x74\x6f\x20\x72\x65\x61\x64\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x4f\x53\x45\x72\x72\x6f\x72\x20\x77\x68\x65\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x68\x61\x6e\x64\x6c\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_56_consts_2_consts_0._ascii.ob_base,
            & const_str_mtime._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_path_mtime._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_path_stats = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_stats",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader.path_stats",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x18\x00\x11\x18\x98\x14\x9f\x1f\x99\x1f\xa8\x14\xd3\x19\x2e\xd0\x0f\x2f\xd0\x08\x2f",
};
static
    struct _PyCode_DEF(40)
importlib__bootstrap_external_toplevel_consts_56_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_56_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1017,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 145,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_56_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_path_stats._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_56_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_56_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x69\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[229];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 228,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x6d\x65\x74\x68\x6f\x64\x20\x77\x68\x69\x63\x68\x20\x77\x72\x69\x74\x65\x73\x20\x64\x61\x74\x61\x20\x28\x62\x79\x74\x65\x73\x29\x20\x74\x6f\x20\x61\x20\x66\x69\x6c\x65\x20\x70\x61\x74\x68\x20\x28\x61\x20\x73\x74\x72\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x61\x6c\x6c\x6f\x77\x73\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x77\x72\x69\x74\x69\x6e\x67\x20\x6f\x66\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x70\x61\x74\x68\x20\x69\x73\x20\x6e\x65\x65\x64\x65\x64\x20\x69\x6e\x20\x6f\x72\x64\x65\x72\x20\x74\x6f\x20\x63\x6f\x72\x72\x65\x63\x74\x6c\x79\x20\x74\x72\x61\x6e\x73\x66\x65\x72\x20\x70\x65\x72\x6d\x69\x73\x73\x69\x6f\x6e\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_56_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_set_data = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "set_data",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_set_data._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__cache_bytecode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_cache_bytecode",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader._cache_bytecode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x10\x00\x10\x14\x8f\x7d\x89\x7d\x98\x5a\xa8\x14\xd3\x0f\x2e\xd0\x08\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_cache_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cache_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_source_path._ascii.ob_base,
            & const_str_cache_path._ascii.ob_base,
            &_Py_ID(data),
        },
    },
};
static
    struct _PyCode_DEF(38)
importlib__bootstrap_external_toplevel_consts_56_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_56_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1031,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 146,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_56_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__cache_bytecode._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_56_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_56_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[151];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 150,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x6d\x65\x74\x68\x6f\x64\x20\x77\x68\x69\x63\x68\x20\x77\x72\x69\x74\x65\x73\x20\x64\x61\x74\x61\x20\x28\x62\x79\x74\x65\x73\x29\x20\x74\x6f\x20\x61\x20\x66\x69\x6c\x65\x20\x70\x61\x74\x68\x20\x28\x61\x20\x73\x74\x72\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x61\x6c\x6c\x6f\x77\x73\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x77\x72\x69\x74\x69\x6e\x67\x20\x6f\x66\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_56_consts_4_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader.set_data",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            &_Py_ID(data),
        },
    },
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_56_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts_4_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1041,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 147,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_56_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_set_data._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_56_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[53];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 52,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Concrete implementation of InspectLoader.get_source.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source not available through get_data()",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_56_consts_5_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_5_consts_1._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_get_data = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_data",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_get_filename._ascii.ob_base,
            & const_str_get_data._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_decode_source._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader.get_source",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[91];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 90,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\xd7\x0f\x20\xd1\x0f\x20\xa0\x18\xd3\x0f\x2a\x88\x04\xf0\x02\x04\x09\x36\xd8\x1b\x1f\x9f\x3d\x99\x3d\xa8\x14\xd3\x1b\x2e\x88\x4c\xf4\x08\x00\x10\x1d\x98\x5c\xd3\x0f\x2a\xd0\x08\x2a\xf8\xf4\x07\x00\x10\x17\xf2\x00\x02\x09\x36\xdc\x12\x1d\xd0\x1e\x47\xd8\x23\x2b\xf4\x03\x01\x13\x2d\xd8\x32\x35\xf0\x03\x01\x0d\x36\xfb\xf0\x03\x02\x09\x36\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x93\x11\x2f\x00\xaf\x09\x41\x0b\x03\xb8\x0e\x41\x06\x03\xc1\x06\x05\x41\x0b\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            & const_str_source_bytes._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(156)
importlib__bootstrap_external_toplevel_consts_56_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 78,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_56_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_56_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1048,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 148,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_56_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(get_source),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_56_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_56_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x09\x00\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x13\x7d\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x01\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x04\x82\x02\x64\x03\x7d\x04\x7e\x04\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__optimize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_optimize",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_7 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__optimize._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[131];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 130,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x63\x6f\x6d\x70\x69\x6c\x65\x64\x20\x66\x72\x6f\x6d\x20\x73\x6f\x75\x72\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x27\x64\x61\x74\x61\x27\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x63\x61\x6e\x20\x62\x65\x20\x61\x6e\x79\x20\x6f\x62\x6a\x65\x63\x74\x20\x74\x79\x70\x65\x20\x74\x68\x61\x74\x20\x63\x6f\x6d\x70\x69\x6c\x65\x28\x29\x20\x73\x75\x70\x70\x6f\x72\x74\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_8_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(dont_inherit),
            &_Py_ID(optimize),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_56_consts_8_consts_0._ascii.ob_base,
            & const_str_exec._ascii.ob_base,
            Py_True,
            & importlib__bootstrap_external_toplevel_consts_56_consts_8_consts_3._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_compile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "compile",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(_bootstrap),
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str_compile._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_source_to_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "source_to_code",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader.source_to_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[44];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 43,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x10\x1a\xd7\x0f\x33\xd1\x0f\x33\xb4\x47\xb8\x54\xc0\x34\xc8\x16\xd8\x35\x39\xc0\x49\xf0\x03\x00\x10\x34\xf3\x00\x01\x10\x4f\x01\xf0\x00\x01\x09\x4f\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(data),
            &_Py_ID(path),
            & const_str__optimize._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(64)
importlib__bootstrap_external_toplevel_consts_56_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts_8_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_56_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 1058,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 149,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_56_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_source_to_code._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_56_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_56_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\x64\x01\x64\x02\x7c\x03\xac\x03\xab\x06\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[191];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 190,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6e\x63\x72\x65\x74\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x49\x6e\x73\x70\x65\x63\x74\x4c\x6f\x61\x64\x65\x72\x2e\x67\x65\x74\x5f\x63\x6f\x64\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x61\x64\x69\x6e\x67\x20\x6f\x66\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x20\x72\x65\x71\x75\x69\x72\x65\x73\x20\x70\x61\x74\x68\x5f\x73\x74\x61\x74\x73\x20\x74\x6f\x20\x62\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x2e\x20\x54\x6f\x20\x77\x72\x69\x74\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x62\x79\x74\x65\x63\x6f\x64\x65\x2c\x20\x73\x65\x74\x5f\x64\x61\x74\x61\x20\x6d\x75\x73\x74\x20\x61\x6c\x73\x6f\x20\x62\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_never = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "never",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_always = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "always",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_13 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "{} matches {}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_14 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_bytecode_path._ascii.ob_base,
            & const_str_source_path._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_15 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "code object from {}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_0._ascii.ob_base,
            Py_None,
            Py_False,
            Py_True,
            & const_str_mtime._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_45_consts_3._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & const_str_never._ascii.ob_base,
            & const_str_always._ascii.ob_base,
            &_Py_ID(size),
            & importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_13._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_14._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_9_consts_15._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_memoryview = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "memoryview",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_check_hash_based_pycs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "check_hash_based_pycs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__RAW_MAGIC_NUMBER = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_RAW_MAGIC_NUMBER",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_dont_write_bytecode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dont_write_bytecode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[27];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 27,
        },
        .ob_item = {
            & const_str_get_filename._ascii.ob_base,
            & const_str_cache_from_source._ascii.ob_base,
            & const_str_path_stats._ascii.ob_base,
            & const_str_int._ascii.ob_base,
            & const_str_get_data._ascii.ob_base,
            & const_str__classify_pyc._ascii.ob_base,
            & const_str_memoryview._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_check_hash_based_pycs._ascii.ob_base,
            & const_str_source_hash._ascii.ob_base,
            & const_str__RAW_MAGIC_NUMBER._ascii.ob_base,
            & const_str__validate_hash_pyc._ascii.ob_base,
            & const_str__validate_timestamp_pyc._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str__compile_bytecode._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_EOFError._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_NotImplementedError._ascii.ob_base,
            & const_str_source_to_code._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_dont_write_bytecode._ascii.ob_base,
            & const_str__code_to_hash_pyc._ascii.ob_base,
            & const_str__code_to_timestamp_pyc._ascii.ob_base,
            &_Py_ID(len),
            & const_str__cache_bytecode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceLoader.get_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[660];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 659,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0e\x00\x17\x1b\xd7\x16\x27\xd1\x16\x27\xa8\x08\xd3\x16\x31\x88\x0b\xd8\x17\x1b\x88\x0c\xd8\x17\x1b\x88\x0c\xd8\x16\x1a\x88\x0b\xd8\x15\x1a\x88\x0a\xd8\x17\x1b\x88\x0c\xf0\x02\x33\x09\x4a\x01\xdc\x1c\x2d\xa8\x6b\xd3\x1c\x3a\x88\x4d\xf0\x08\x2e\x0d\x4a\x01\xd8\x15\x19\x97\x5f\x91\x5f\xa0\x5b\xd3\x15\x31\x90\x02\xf4\x08\x00\x20\x23\xa0\x32\xa0\x67\xa1\x3b\xd3\x1f\x2f\x90\x0c\xf0\x02\x28\x11\x4a\x01\xd8\x1b\x1f\x9f\x3d\x99\x3d\xa8\x1d\xd3\x1b\x37\x90\x44\xf0\x0a\x00\x21\x29\xd8\x20\x2d\xf1\x05\x03\x23\x16\x90\x4b\xf0\x08\x1f\x15\x4a\x01\xdc\x20\x2d\xa8\x64\xb0\x48\xb8\x6b\xd3\x20\x4a\x98\x05\xdc\x25\x2f\xb0\x04\xd3\x25\x35\xb0\x62\xb0\x63\xd0\x25\x3a\x98\x0a\xd8\x25\x2a\xa8\x53\xa1\x5b\xb0\x41\xd1\x25\x35\x98\x0a\xd9\x1b\x25\xd8\x2b\x30\xb0\x34\xa9\x3c\xb8\x31\xd1\x2b\x3c\x98\x4c\xdc\x20\x24\xd7\x20\x3a\xd1\x20\x3a\xb8\x67\xd2\x20\x45\xd9\x21\x2d\xdc\x21\x25\xd7\x21\x3b\xd1\x21\x3b\xb8\x78\xd2\x21\x47\xd8\x2f\x33\xaf\x7d\xa9\x7d\xb8\x5b\xd3\x2f\x49\xa0\x0c\xdc\x2e\x32\xd7\x2e\x3e\xd1\x2e\x3e\xdc\x24\x35\xd8\x24\x30\xf3\x05\x03\x2f\x22\xa0\x0b\xf4\x08\x00\x21\x33\xb0\x34\xb8\x1b\xc0\x68\xd8\x33\x3e\xf5\x03\x01\x21\x40\x01\xf4\x06\x00\x1d\x34\xd8\x20\x24\xd8\x20\x2c\xd8\x20\x22\xa0\x36\xa1\x0a\xd8\x20\x28\xd8\x20\x2b\xf4\x0b\x06\x1d\x1e\xf4\x14\x00\x19\x23\xd7\x18\x33\xd1\x18\x33\xb0\x4f\xc0\x5d\xd8\x34\x3f\xf4\x03\x01\x19\x41\x01\xe4\x1f\x30\xb0\x1a\xc0\x28\xd8\x3f\x4c\xd8\x3d\x48\xf4\x05\x02\x20\x4a\x01\xf0\x00\x02\x19\x4a\x01\xf8\xf4\x0b\x00\x1d\x28\xac\x18\xd0\x1b\x32\xf2\x00\x01\x15\x1d\xd9\x18\x1c\xf0\x03\x01\x15\x1d\xfb\xf4\x3f\x00\x18\x1f\xf2\x00\x01\x11\x19\xd9\x14\x18\xf0\x03\x01\x11\x19\xfb\xf4\x0d\x00\x14\x1b\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfb\xf4\x0b\x00\x10\x23\xf2\x00\x01\x09\x21\xd8\x1c\x20\x8a\x4d\xf0\x03\x01\x09\x21\xfa\xf0\x64\x01\x00\x0c\x18\xd0\x0b\x1f\xd8\x1b\x1f\x9f\x3d\x99\x3d\xa8\x1b\xd3\x1b\x35\x88\x4c\xd8\x16\x1a\xd7\x16\x29\xd1\x16\x29\xa8\x2c\xb8\x0b\xd3\x16\x44\x88\x0b\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xd0\x24\x39\xb8\x3b\xd4\x08\x47\xdc\x10\x13\xd7\x10\x27\xd2\x10\x27\xa8\x4d\xd0\x2c\x45\xd8\x10\x1c\xd0\x10\x28\xd9\x0f\x19\xd8\x13\x1e\xd0\x13\x26\xdc\x22\x26\xd7\x22\x32\xd1\x22\x32\xd4\x33\x44\xd8\x33\x3f\xf3\x03\x01\x23\x41\x01\x90\x4b\xe4\x17\x28\xa8\x1b\xb0\x6b\xc0\x3c\xd3\x17\x50\x91\x04\xe4\x17\x2d\xa8\x6b\xb8\x3c\xdc\x2e\x31\xb0\x2c\xd3\x2e\x3f\xf3\x03\x01\x18\x41\x01\x90\x04\xf0\x04\x03\x0d\x15\xd8\x10\x14\xd7\x10\x24\xd1\x10\x24\xa0\x5b\xb0\x2d\xc0\x14\xd4\x10\x46\xf0\x06\x00\x10\x1b\xd0\x08\x1a\xf8\xf4\x05\x00\x14\x27\xf2\x00\x01\x0d\x15\xd8\x10\x14\xd8\x0f\x1a\xd0\x08\x1a\xf0\x05\x01\x0d\x15\xfa\xe0\x0f\x1a\xd0\x08\x1a",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[90];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 89,
    },
    .ob_shash = -1,
    .ob_sval = "\x9d\x0b\x45\x1a\x00\xa9\x11\x45\x0b\x00\xc1\x09\x11\x44\x3c\x00\xc1\x20\x42\x21\x44\x27\x00\xc4\x27\x0f\x44\x39\x03\xc4\x38\x01\x44\x39\x03\xc4\x3c\x09\x45\x08\x03\xc5\x07\x01\x45\x08\x03\xc5\x0b\x09\x45\x17\x03\xc5\x16\x01\x45\x17\x03\xc5\x1a\x0b\x45\x28\x03\xc5\x27\x01\x45\x28\x03\xc7\x3d\x13\x48\x12\x00\xc8\x12\x09\x48\x1f\x03\xc8\x1e\x01\x48\x1f\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_hash_based = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hash_based",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_check_source = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "check_source",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_st = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_bytes_data = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytes_data",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_code_object = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "code_object",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            & const_str_source_path._ascii.ob_base,
            & const_str_source_mtime._ascii.ob_base,
            & const_str_source_bytes._ascii.ob_base,
            & const_str_source_hash._ascii.ob_base,
            & const_str_hash_based._ascii.ob_base,
            & const_str_check_source._ascii.ob_base,
            & const_str_bytecode_path._ascii.ob_base,
            & const_str_st._ascii.ob_base,
            &_Py_ID(data),
            & const_str_exc_details._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_bytes_data._ascii.ob_base,
            & const_str_code_object._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib__bootstrap_external_toplevel_consts_56_consts_9_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "               ",
};
static
    struct _PyCode_DEF(1096)
importlib__bootstrap_external_toplevel_consts_56_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 548,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts_9_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_56_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_56_consts_9_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 22 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1066,
    .co_nlocalsplus = 15,
    .co_nlocals = 15,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 150,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_56_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_56_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_code._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_56_consts_9_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_56_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x01\x7d\x03\x64\x01\x7d\x04\x64\x01\x7d\x05\x64\x02\x7d\x06\x64\x03\x7d\x07\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x08\x09\x00\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x09\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x64\x04\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x09\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0a\x7c\x01\x7c\x08\x64\x05\x9c\x02\x7d\x0b\x09\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x7c\x01\x7c\x0b\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x0c\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x64\x06\x64\x01\x1a\x00\x7d\x0d\x7c\x0c\x64\x07\x7a\x01\x00\x00\x64\x08\x6b\x37\x00\x00\x7d\x06\x7c\x06\x72\x6a\x7c\x0c\x64\x09\x7a\x01\x00\x00\x64\x08\x6b\x37\x00\x00\x7d\x07\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\x6b\x37\x00\x00\x72\x61\x7c\x07\x73\x13\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\x6b\x28\x00\x00\x72\x4c\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x7c\x05\x7c\x01\x7c\x0b\xab\x04\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x12\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x7c\x03\x7c\x09\x64\x0c\x19\x00\x00\x00\x7c\x01\x7c\x0b\xab\x05\x00\x00\x00\x00\x00\x00\x01\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\x7c\x08\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x7c\x01\x7c\x08\x7c\x02\xac\x0e\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x33\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x24\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x24\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x24\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x15\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x01\x7d\x08\x59\x00\x6e\x04\x77\x00\x78\x03\x59\x00\x77\x01\x7c\x04\x80\x11\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x00\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0e\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0f\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x73\x6c\x7c\x08\x81\x6a\x7c\x03\x81\x68\x7c\x06\x72\x2a\x7c\x05\x80\x1a\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x2f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0e\x7c\x05\x7c\x07\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x0a\x6e\x16\x74\x31\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0e\x7c\x03\x74\x33\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x0a\x09\x00\x7c\x00\x6a\x35\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x08\x7c\x0a\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x0e\x53\x00\x23\x00\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x7c\x0e\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01\x7c\x0e\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_SourceLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_1.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_4.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_5.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & importlib__bootstrap_external_toplevel_consts_56_consts_7._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_8.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56_consts_9.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_56_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            & const_str_path_mtime._ascii.ob_base,
            & const_str_path_stats._ascii.ob_base,
            & const_str__cache_bytecode._ascii.ob_base,
            & const_str_set_data._ascii.ob_base,
            &_Py_ID(get_source),
            & const_str_source_to_code._ascii.ob_base,
            & const_str_get_code._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
importlib__bootstrap_external_toplevel_consts_56_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x04\x06\x05\x16\xf2\x10\x0c\x05\x30\xf2\x1c\x08\x05\x2f\xf2\x14\x04\x05\x0c\xf2\x0e\x08\x05\x2b\xf0\x14\x00\x37\x39\xf4\x00\x06\x05\x4f\x01\xf3\x10\x53\x01\x05\x1b",
};
static
    struct _PyCode_DEF(60)
importlib__bootstrap_external_toplevel_consts_56 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_56_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_56_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1007,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 151,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_SourceLoader._ascii.ob_base,
    .co_qualname = & const_str_SourceLoader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_56_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x64\x07\x9c\x01\x64\x08\x84\x02\x5a\x08\x64\x09\x84\x00\x5a\x09\x79\x0a",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_FileLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[104];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 103,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x42\x61\x73\x65\x20\x66\x69\x6c\x65\x20\x6c\x6f\x61\x64\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x77\x68\x69\x63\x68\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x73\x20\x74\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x20\x70\x72\x6f\x74\x6f\x63\x6f\x6c\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x74\x68\x61\x74\x0a\x20\x20\x20\x20\x72\x65\x71\x75\x69\x72\x65\x20\x66\x69\x6c\x65\x20\x73\x79\x73\x74\x65\x6d\x20\x75\x73\x61\x67\x65\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[76];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 75,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x61\x63\x68\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x66\x6f\x75\x6e\x64\x20\x62\x79\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x69\x6e\x64\x65\x72\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_58_consts_2_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x15\x1d\x88\x04\x8c\x09\xd8\x14\x18\x88\x04\x8d\x09",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
        },
    },
};
static
    struct _PyCode_DEF(32)
importlib__bootstrap_external_toplevel_consts_58_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_58_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_45_consts_3._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1157,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 152,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_58_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(__class__),
            &_Py_ID(__dict__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader.__eq__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x10\x14\x97\x0e\x91\x0e\xa0\x25\xa7\x2f\xa1\x2f\xd1\x10\x31\xf2\x00\x01\x11\x30\xd8\x10\x14\x97\x0d\x91\x0d\xa0\x15\xa7\x1e\xa1\x1e\xd1\x10\x2f\xf0\x03\x01\x09\x31",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_other._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(106)
importlib__bootstrap_external_toplevel_consts_58_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 53,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1163,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 153,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__eq__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_58_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x19\x01\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_hash = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hash",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_hash._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(path),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader.__hash__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x13\x90\x44\x97\x49\x91\x49\x8b\x7f\xa4\x14\xa0\x64\xa7\x69\xa1\x69\xa3\x1f\xd1\x0f\x30\xd0\x08\x30",
};
static
    struct _PyCode_DEF(88)
importlib__bootstrap_external_toplevel_consts_58_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 44,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1167,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 154,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__hash__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_58_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x0c\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[101];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 100,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4c\x6f\x61\x64\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x20\x66\x72\x6f\x6d\x20\x61\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x69\x73\x20\x64\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2e\x20\x20\x55\x73\x65\x20\x65\x78\x65\x63\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x29\x20\x69\x6e\x73\x74\x65\x61\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_58_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_super._ascii.ob_base,
            & const_str_FileLoader._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader.load_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf4\x14\x00\x10\x15\x94\x5a\xa0\x14\xd1\x0f\x32\xb0\x38\xd3\x0f\x3c\xd0\x08\x3c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(__class__),
        },
    },
};
static
    struct _PyCode_DEF(42)
importlib__bootstrap_external_toplevel_consts_58_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 21,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_58_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1170,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 155,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_load_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_58_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x8d\x0b\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[59];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 58,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the path to the source file as found by the finder.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_58_consts_6_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader.get_filename",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x10\x14\x8f\x79\x89\x79\xd0\x08\x18",
};
static
    struct _PyCode_DEF(26)
importlib__bootstrap_external_toplevel_consts_58_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_58_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1182,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 156,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_filename._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_58_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the data from path as raw bytes.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_58_consts_7_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_ExtensionFileLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_open_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "open_code",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_SourceLoader._ascii.ob_base,
            & const_str_ExtensionFileLoader._ascii.ob_base,
            &_Py_ID(_io),
            & const_str_open_code._ascii.ob_base,
            & const_str_str._ascii.ob_base,
            &_Py_ID(read),
            & const_str_FileIO._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader.get_data",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[107];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 106,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x15\x90\x64\x9c\x5c\xd4\x2b\x3e\xd0\x1c\x3f\xd4\x0b\x40\xdc\x11\x14\x97\x1d\x91\x1d\x9c\x73\xa0\x34\x9b\x79\xd4\x11\x29\xa8\x54\xd8\x17\x1b\x97\x79\x91\x79\x93\x7b\xf7\x03\x00\x12\x2a\xd1\x11\x29\xf4\x06\x00\x12\x15\x97\x1a\x91\x1a\x98\x44\xa0\x23\xd4\x11\x26\xa8\x24\xd8\x17\x1b\x97\x79\x91\x79\x93\x7b\xf7\x03\x00\x12\x27\xd1\x11\x26\xf7\x07\x00\x12\x2a\xd0\x11\x29\xfa\xf7\x06\x00\x12\x27\xd0\x11\x26\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\xb5\x10\x41\x3f\x03\xc1\x25\x10\x42\x0b\x03\xc1\x3f\x05\x42\x08\x07\xc2\x0b\x05\x42\x14\x07",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            &_Py_ID(file),
        },
    },
};
static
    struct _PyCode_DEF(302)
importlib__bootstrap_external_toplevel_consts_58_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 151,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_58_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_58_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1187,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 157,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_data._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_58_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\xab\x02\x00\x00\x00\x00\x00\x00\x72\x38\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x02\x7c\x02\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x63\x02\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x02\x7c\x02\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x63\x02\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x79\x01\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x79\x01\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_FileReader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileReader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_8_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_FileReader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_external_toplevel_consts_58_consts_8_consts_2._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_8_names_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "importlib.readers",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_58_consts_8_names_0._ascii.ob_base,
            & const_str_FileReader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_get_resource_reader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_resource_reader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileLoader.get_resource_reader",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib__bootstrap_external_toplevel_consts_58_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe5\x08\x30\xd9\x0f\x19\x98\x24\xd3\x0f\x1f\xd0\x08\x1f",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(module),
            & const_str_FileReader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(30)
importlib__bootstrap_external_toplevel_consts_58_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_58_consts_8_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1196,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 158,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_resource_reader._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_58_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x6d\x01\x7d\x02\x01\x00\x02\x00\x7c\x02\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_FileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_4.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_5.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_6.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_7.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58_consts_8.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            &_Py_ID(__eq__),
            &_Py_ID(__hash__),
            & const_str__check_name._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
            & const_str_get_filename._ascii.ob_base,
            & const_str_get_data._ascii.ob_base,
            & const_str_get_resource_reader._ascii.ob_base,
            &_Py_ID(__classcell__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[89];
    }
importlib__bootstrap_external_toplevel_consts_58_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 88,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x84\x00\xf1\x04\x01\x05\x22\xf2\x06\x04\x05\x19\xf2\x0c\x02\x05\x31\xf2\x08\x01\x05\x31\xf0\x06\x00\x06\x11\xf3\x02\x09\x05\x3d\xf3\x03\x00\x06\x11\xf0\x02\x09\x05\x3d\xf0\x16\x00\x06\x11\xf1\x02\x02\x05\x19\xf3\x03\x00\x06\x11\xf0\x02\x02\x05\x19\xf2\x08\x07\x05\x23\xf0\x12\x00\x06\x11\xf1\x02\x02\x05\x20\xf3\x03\x00\x06\x11\xf4\x02\x02\x05\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_58_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(__class__),
        },
    },
};
static
    struct _PyCode_DEF(100)
importlib__bootstrap_external_toplevel_consts_58 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 50,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_58_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1152,
    .co_nlocalsplus = 1,
    .co_nlocals = 0,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 159,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[64]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_FileLoader._ascii.ob_base,
    .co_qualname = & const_str_FileLoader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x65\x07\x88\x00\x66\x01\x64\x05\x84\x08\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x65\x07\x64\x06\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x09\x64\x07\x84\x00\x5a\x0a\x65\x07\x64\x08\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0b\x88\x00\x78\x01\x5a\x0c\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_SourceFileLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceFileLoader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[63];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 62,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Concrete implementation of SourceLoader using the file system.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the metadata for the path.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_2_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_mtime._ascii.ob_base,
            &_Py_ID(size),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_60_consts_2_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60_consts_2_consts_1._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_st_mtime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_mtime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_st_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_size",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__path_stat._ascii.ob_base,
            & const_str_st_mtime._ascii.ob_base,
            & const_str_st_size._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceFileLoader.path_stats",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0d\x17\x98\x04\xd3\x0d\x1d\x88\x02\xd8\x19\x1b\x9f\x1b\x99\x1b\xa8\x62\xaf\x6a\xa9\x6a\xd1\x0f\x39\xd0\x08\x39",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            & const_str_st._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(74)
importlib__bootstrap_external_toplevel_consts_60_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 37,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_60_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_60_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1206,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 160,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_60_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_path_stats._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_60_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_60_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x9c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__mode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_mode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_3_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__mode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_60_consts_3_consts_1._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__calc_mode._ascii.ob_base,
            & const_str_set_data._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceFileLoader._cache_bytecode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x19\x98\x2b\xd3\x0f\x26\x88\x04\xd8\x0f\x13\x8f\x7d\x89\x7d\x98\x5d\xa8\x44\xb8\x04\x88\x7d\xd3\x0f\x3d\xd0\x08\x3d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_source_path._ascii.ob_base,
            & const_str_bytecode_path._ascii.ob_base,
            &_Py_ID(data),
            &_Py_ID(mode),
        },
    },
};
static
    struct _PyCode_DEF(64)
importlib__bootstrap_external_toplevel_consts_60_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_60_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_60_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1211,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 161,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_60_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__cache_bytecode._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_60_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_60_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x03\x7c\x04\xac\x01\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Write bytes data to a file.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "could not create {!r}: {!r}",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "created {!r}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_60_consts_6_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60_consts_6_consts_1._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_60_consts_6_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_mkdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mkdir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_FileExistsError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileExistsError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str__path_split._ascii.ob_base,
            & const_str__path_isdir._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(reversed),
            & const_str__path_join._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str_mkdir._ascii.ob_base,
            & const_str_FileExistsError._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str__write_atomic._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourceFileLoader.set_data",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[252];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 251,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x1b\x26\xa0\x74\xd3\x1b\x2c\xd1\x08\x18\x88\x06\x90\x08\xd8\x15\x17\x88\x0a\xe1\x0e\x14\x9c\x5b\xa8\x16\xd4\x1d\x30\xdc\x1b\x26\xa0\x76\xd3\x1b\x2e\x89\x4c\x88\x46\x90\x44\xd8\x0c\x16\xd7\x0c\x1d\xd1\x0c\x1d\x98\x64\xd4\x0c\x23\xf1\x05\x00\x0f\x15\x9c\x5b\xa8\x16\xd5\x1d\x30\xf4\x08\x00\x15\x1d\x98\x5a\xd6\x14\x28\x88\x44\xdc\x15\x1f\xa0\x06\xa8\x04\xd3\x15\x2d\x88\x46\xf0\x02\x0a\x0d\x17\xdc\x10\x13\x97\x09\x91\x09\x98\x26\xd5\x10\x21\xf0\x07\x00\x15\x29\xf0\x1a\x06\x09\x2d\xdc\x0c\x19\x98\x24\xa0\x04\xa0\x65\xd4\x0c\x2c\xdc\x0c\x16\xd7\x0c\x27\xd1\x0c\x27\xa8\x0e\xb8\x04\xd5\x0c\x3d\xf8\xf4\x17\x00\x14\x23\xf2\x00\x02\x0d\x19\xe1\x10\x18\xdc\x13\x1a\xf2\x00\x05\x0d\x17\xf4\x06\x00\x11\x1b\xd7\x10\x2b\xd1\x10\x2b\xd0\x2c\x49\xd8\x2c\x32\xb0\x43\xf4\x03\x01\x11\x39\xe5\x10\x16\xfb\xf0\x0b\x05\x0d\x17\xfb\xf4\x12\x00\x10\x17\xf2\x00\x03\x09\x2d\xe4\x0c\x16\xd7\x0c\x27\xd1\x0c\x27\xd0\x28\x45\xc0\x74\xd8\x28\x2b\xf7\x03\x01\x0d\x2d\xf1\x00\x01\x0d\x2d\xfb\xf0\x05\x03\x09\x2d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[55];
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 54,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x26\x15\x42\x22\x02\xc1\x3e\x23\x43\x1a\x00\xc2\x22\x09\x43\x17\x05\xc2\x2d\x08\x43\x17\x05\xc2\x35\x17\x43\x12\x05\xc3\x12\x05\x43\x17\x05\xc3\x1a\x09\x44\x04\x03\xc3\x23\x17\x43\x3f\x03\xc3\x3f\x05\x44\x04\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            &_Py_ID(data),
            & const_str__mode._ascii.ob_base,
            &_Py_ID(parent),
            &_Py_ID(filename),
            & const_str_path_parts._ascii.ob_base,
            & const_str_part._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(526)
importlib__bootstrap_external_toplevel_consts_60_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 263,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_60_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_60_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_60_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1216,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 162,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_60_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_set_data._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_60_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_60_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x04\x7d\x05\x67\x00\x7d\x06\x7c\x04\x72\x38\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x73\x2d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x04\x7d\x07\x7c\x06\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x2d\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x24\x00\x00\x7d\x07\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x07\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x09\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x26\x04\x00\x09\x00\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\x7c\x03\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x57\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x22\x7d\x08\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x04\x7c\x08\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x02\x7d\x08\x7e\x08\x01\x00\x79\x02\x64\x02\x7d\x08\x7e\x08\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x21\x7d\x08\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x01\x7c\x08\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x02\x7d\x08\x7e\x08\x79\x02\x64\x02\x7d\x08\x7e\x08\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_SourceFileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60_consts_3.ob_base.ob_base,
            & const_int_438.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60_consts_3_consts_1._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60_consts_6.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_60_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_path_stats._ascii.ob_base,
            & const_str__cache_bytecode._ascii.ob_base,
            & const_str_set_data._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
importlib__bootstrap_external_toplevel_consts_60_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x48\xf2\x04\x03\x05\x3a\xf2\x0a\x03\x05\x3e\xf0\x0a\x00\x2d\x32\xf5\x00\x1c\x05\x2d",
};
static
    struct _PyCode_DEF(40)
importlib__bootstrap_external_toplevel_consts_60 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_60_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_60_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1202,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 163,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_SourceFileLoader._ascii.ob_base,
    .co_qualname = & const_str_SourceFileLoader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_60_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x64\x05\x9c\x01\x64\x06\x84\x02\x5a\x06\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_SourcelessFileLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourcelessFileLoader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[46];
    }
importlib__bootstrap_external_toplevel_consts_62_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 45,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Loader which handles sourceless file imports.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_62_consts_2_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_bytecode_path._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_62_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_45_consts_3._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            & importlib__bootstrap_external_toplevel_consts_62_consts_2_consts_3._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_62_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_get_filename._ascii.ob_base,
            & const_str_get_data._ascii.ob_base,
            & const_str__classify_pyc._ascii.ob_base,
            & const_str__compile_bytecode._ascii.ob_base,
            & const_str_memoryview._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_external_toplevel_consts_62_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourcelessFileLoader.get_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[96];
    }
importlib__bootstrap_external_toplevel_consts_62_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 95,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\xd7\x0f\x20\xd1\x0f\x20\xa0\x18\xd3\x0f\x2a\x88\x04\xd8\x0f\x13\x8f\x7d\x89\x7d\x98\x54\xd3\x0f\x22\x88\x04\xf0\x08\x00\x15\x1d\xd8\x14\x18\xf1\x05\x03\x17\x0a\x88\x0b\xf4\x08\x00\x09\x16\x90\x64\x98\x48\xa0\x6b\xd4\x08\x32\xdc\x0f\x20\xdc\x0c\x16\x90\x74\xd3\x0c\x1c\x98\x52\x98\x53\xd0\x0c\x21\xd8\x11\x19\xd8\x1a\x1e\xf4\x07\x04\x10\x0a\xf0\x00\x04\x09\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_62_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(data),
            & const_str_exc_details._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(158)
importlib__bootstrap_external_toplevel_consts_62_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 79,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_62_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_62_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1251,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 164,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_62_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_code._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_62_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_62_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x01\x7c\x02\x64\x01\x9c\x02\x7d\x04\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x01\x7c\x04\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x64\x00\x1a\x00\x7c\x01\x7c\x02\xac\x03\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
importlib__bootstrap_external_toplevel_consts_62_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return None as there is no source code.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_62_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_62_consts_3_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_62_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SourcelessFileLoader.get_source",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_external_toplevel_consts_62_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_62_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_62_consts_3_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1267,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 165,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(get_source),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_62_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_62_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_62_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_SourcelessFileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_62_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_62_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_62_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_62_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_get_code._ascii.ob_base,
            &_Py_ID(get_source),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib__bootstrap_external_toplevel_consts_62_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x37\xf2\x04\x0e\x05\x0a\xf3\x20\x02\x05\x14",
};
static
    struct _PyCode_DEF(28)
importlib__bootstrap_external_toplevel_consts_62 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_62_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_62_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1247,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 166,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_SourcelessFileLoader._ascii.ob_base,
    .co_qualname = & const_str_SourcelessFileLoader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_62_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[94];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 93,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4c\x6f\x61\x64\x65\x72\x20\x66\x6f\x72\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x6f\x72\x20\x69\x73\x20\x64\x65\x73\x69\x67\x6e\x65\x64\x20\x74\x6f\x20\x77\x6f\x72\x6b\x20\x77\x69\x74\x68\x20\x46\x69\x6c\x65\x46\x69\x6e\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x14\x18\x88\x04\x8c\x09\xd8\x14\x18\x88\x04\x8d\x09",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
            &_Py_ID(path),
        },
    },
};
static
    struct _PyCode_DEF(32)
importlib__bootstrap_external_toplevel_consts_64_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_45_consts_3._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1280,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 167,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_64_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_64_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.__eq__",
};
static
    struct _PyCode_DEF(106)
importlib__bootstrap_external_toplevel_consts_64_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 53,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1284,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 168,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__eq__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x19\x01\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.__hash__",
};
static
    struct _PyCode_DEF(88)
importlib__bootstrap_external_toplevel_consts_64_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 44,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_58_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1288,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 169,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__hash__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x0c\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[41];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 40,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Create an uninitialized extension module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "extension module {!r} loaded from {!r}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_64_consts_5_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_5_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_create_dynamic = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "create_dynamic",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(_bootstrap),
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_create_dynamic._ascii.ob_base,
            & const_str__verbose_message._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(path),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.create_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[65];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 64,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x11\x1b\xd7\x11\x35\xd1\x11\x35\xdc\x0c\x10\xd7\x0c\x1f\xd1\x0c\x1f\xa0\x14\xf3\x03\x01\x12\x27\x88\x06\xe4\x08\x12\xd7\x08\x23\xd1\x08\x23\xd0\x24\x4c\xd8\x19\x1d\x9f\x19\x99\x19\xa0\x44\xa7\x49\xa1\x49\xf4\x03\x01\x09\x2f\xe0\x0f\x15\x88\x0d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_spec._ascii.ob_base,
            &_Py_ID(module),
        },
    },
};
static
    struct _PyCode_DEF(164)
importlib__bootstrap_external_toplevel_consts_64_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 82,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_64_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_64_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1291,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 170,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_64_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_create_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_64_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x01\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Initialize an extension module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[41];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 40,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "extension module {!r} executed from {!r}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_64_consts_6_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_6_consts_1._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_exec_dynamic = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exec_dynamic",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(_bootstrap),
            & const_str__call_with_frames_removed._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_exec_dynamic._ascii.ob_base,
            & const_str__verbose_message._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(path),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.exec_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[55];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 54,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x12\xd7\x08\x2c\xd1\x08\x2c\xac\x54\xd7\x2d\x3e\xd1\x2d\x3e\xc0\x06\xd4\x08\x47\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xd0\x24\x4e\xd8\x19\x1d\x9f\x19\x99\x19\xa0\x44\xa7\x49\xa1\x49\xf5\x03\x01\x09\x2f",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(module),
        },
    },
};
static
    struct _PyCode_DEF(162)
importlib__bootstrap_external_toplevel_consts_64_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 81,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_64_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_64_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1299,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 171,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_64_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_exec_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_64_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[50];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 49,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if the extension module is a package.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(__init__),
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[50];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 49,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.is_package.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xf0\x00\x01\x13\x35\xd9\x21\x33\x90\x76\xf0\x03\x00\x14\x1d\xa0\x0a\xa8\x56\xd1\x20\x33\xd5\x13\x33\xd9\x21\x33\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x12\x15\x01",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_suffix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "suffix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_file_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "file_name",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            & const_str_suffix._ascii.ob_base,
            & const_str_file_name._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(46)
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1308,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 172,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x0c\x00\x00\x7d\x01\x89\x02\x64\x00\x7c\x01\x7a\x00\x00\x00\x6b\x28\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x0e\x04\x00\x79\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_any = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "any",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__path_split._ascii.ob_base,
            &_Py_ID(path),
            & const_str_any._ascii.ob_base,
            & const_str_EXTENSION_SUFFIXES._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.is_package",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xe4\x14\x1f\xa0\x04\xa7\x09\xa1\x09\xd3\x14\x2a\xa8\x31\xd1\x14\x2d\x88\x09\xdc\x0f\x12\xf3\x00\x01\x13\x35\xdd\x21\x33\xf3\x03\x01\x13\x35\xf3\x00\x01\x10\x35\xf0\x00\x01\x09\x35",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            & const_str_file_name._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(100)
importlib__bootstrap_external_toplevel_consts_64_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 50,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_64_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1305,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 173,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_64_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_is_package._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_64_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x02\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x8a\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x88\x02\x66\x01\x64\x02\x84\x08\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[64];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 63,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return None as an extension module cannot create a code object.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_64_consts_8_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.get_code",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_64_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_64_consts_8_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1311,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 174,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_code._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_62_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return None as extension modules have no source code.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_64_consts_9_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.get_source",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_64_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_64_consts_9_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1315,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 175,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(get_source),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_9_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_62_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib__bootstrap_external_toplevel_consts_64_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ExtensionFileLoader.get_filename",
};
static
    struct _PyCode_DEF(26)
importlib__bootstrap_external_toplevel_consts_64_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_58_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1319,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 176,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_filename._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_64_consts_10_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_58_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str_ExtensionFileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_4.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_5.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_6.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_7.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_8.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_9.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64_consts_10.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_64_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            &_Py_ID(__eq__),
            &_Py_ID(__hash__),
            & const_str_create_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str_is_package._ascii.ob_base,
            & const_str_get_code._ascii.ob_base,
            &_Py_ID(get_source),
            & const_str__check_name._ascii.ob_base,
            & const_str_get_filename._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[68];
    }
importlib__bootstrap_external_toplevel_consts_64_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 67,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x04\x05\x08\xf2\x0c\x02\x05\x19\xf2\x08\x02\x05\x31\xf2\x08\x01\x05\x31\xf2\x06\x06\x05\x16\xf2\x10\x04\x05\x2f\xf2\x0c\x04\x05\x35\xf2\x0c\x02\x05\x14\xf2\x08\x02\x05\x14\xf0\x08\x00\x06\x11\xf1\x02\x02\x05\x19\xf3\x03\x00\x06\x11\xf1\x02\x02\x05\x19",
};
static
    struct _PyCode_DEF(80)
importlib__bootstrap_external_toplevel_consts_64 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 40,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_64_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_64_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1272,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 177,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_ExtensionFileLoader._ascii.ob_base,
    .co_qualname = & const_str_ExtensionFileLoader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_64_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x64\x07\x84\x00\x5a\x09\x64\x08\x84\x00\x5a\x0a\x64\x09\x84\x00\x5a\x0b\x65\x0c\x64\x0a\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0d\x79\x0b",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__NamespacePath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[295];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 294,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x70\x72\x65\x73\x65\x6e\x74\x73\x20\x61\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x20\x70\x61\x63\x6b\x61\x67\x65\x27\x73\x20\x70\x61\x74\x68\x2e\x20\x20\x49\x74\x20\x75\x73\x65\x73\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x0a\x20\x20\x20\x20\x74\x6f\x20\x66\x69\x6e\x64\x20\x69\x74\x73\x20\x70\x61\x72\x65\x6e\x74\x20\x6d\x6f\x64\x75\x6c\x65\x2c\x20\x61\x6e\x64\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x72\x65\x20\x69\x74\x20\x6c\x6f\x6f\x6b\x73\x20\x75\x70\x20\x74\x68\x65\x20\x70\x61\x72\x65\x6e\x74\x27\x73\x0a\x20\x20\x20\x20\x5f\x5f\x70\x61\x74\x68\x5f\x5f\x2e\x20\x20\x57\x68\x65\x6e\x20\x74\x68\x69\x73\x20\x63\x68\x61\x6e\x67\x65\x73\x2c\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x27\x73\x20\x6f\x77\x6e\x20\x70\x61\x74\x68\x20\x69\x73\x20\x72\x65\x63\x6f\x6d\x70\x75\x74\x65\x64\x2c\x0a\x20\x20\x20\x20\x75\x73\x69\x6e\x67\x20\x70\x61\x74\x68\x5f\x66\x69\x6e\x64\x65\x72\x2e\x20\x20\x46\x6f\x72\x20\x74\x6f\x70\x2d\x6c\x65\x76\x65\x6c\x20\x6d\x6f\x64\x75\x6c\x65\x73\x2c\x20\x74\x68\x65\x20\x70\x61\x72\x65\x6e\x74\x20\x6d\x6f\x64\x75\x6c\x65\x27\x73\x20\x70\x61\x74\x68\x0a\x20\x20\x20\x20\x69\x73\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__get_parent_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_parent_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__last_parent_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_last_parent_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__epoch = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_epoch",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__last_epoch = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_last_epoch",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__path_finder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_finder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str__name._ascii.ob_base,
            & const_str__path._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str__get_parent_path._ascii.ob_base,
            & const_str__last_parent_path._ascii.ob_base,
            & const_str__epoch._ascii.ob_base,
            & const_str__last_epoch._ascii.ob_base,
            & const_str__path_finder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[59];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 58,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x15\x19\x88\x04\x8c\x0a\xd8\x15\x19\x88\x04\x8c\x0a\xdc\x21\x26\xa0\x74\xd7\x27\x3c\xd1\x27\x3c\xd3\x27\x3e\xd3\x21\x3f\x88\x04\xd4\x08\x1e\xd8\x1b\x1f\x9f\x3b\x99\x3b\x88\x04\xd4\x08\x18\xd8\x1c\x27\x88\x04\xd5\x08\x19",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_path_finder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_finder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
            &_Py_ID(path),
            & const_str_path_finder._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(140)
importlib__bootstrap_external_toplevel_consts_66_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 70,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1336,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 178,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[63];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 62,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Returns a tuple of (parent-module-name, parent-path-attr-name)",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_4_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(path),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_66_consts_4_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            &_Py_STR(empty),
            & importlib__bootstrap_external_toplevel_consts_66_consts_4_consts_3._object.ob_base.ob_base,
            &_Py_ID(__path__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__name._ascii.ob_base,
            & const_str_rpartition._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str__find_parent_path_names = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_find_parent_path_names",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath._find_parent_path_names",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x1a\x1e\x9f\x2a\x99\x2a\xd7\x1a\x2f\xd1\x1a\x2f\xb0\x03\xd3\x1a\x34\x89\x0f\x88\x06\x90\x03\x90\x52\xd8\x0b\x0e\x90\x22\x8a\x39\xe0\x13\x20\xf0\x06\x00\x10\x16\x90\x7a\xd0\x0f\x21\xd0\x08\x21",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_dot = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dot",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_me = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "me",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(parent),
            & const_str_dot._ascii.ob_base,
            & const_str_me._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(84)
importlib__bootstrap_external_toplevel_consts_66_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 42,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_66_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1343,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 179,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__find_parent_path_names._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x01\x7d\x02\x7d\x03\x7c\x02\x64\x02\x6b\x28\x00\x00\x72\x01\x79\x03\x7c\x01\x64\x04\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__find_parent_path_names._ascii.ob_base,
            &_Py_ID(getattr),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath._get_parent_path",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[46];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 45,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x2d\x31\xd7\x2d\x49\xd1\x2d\x49\xd3\x2d\x4b\xd1\x08\x2a\xd0\x08\x1a\x98\x4e\xdc\x0f\x16\x94\x73\x97\x7b\x91\x7b\xd0\x23\x35\xd1\x17\x36\xb8\x0e\xd3\x0f\x47\xd0\x08\x47",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_parent_module_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "parent_module_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_path_attr_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_attr_name",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_parent_module_name._ascii.ob_base,
            & const_str_path_attr_name._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(98)
importlib__bootstrap_external_toplevel_consts_66_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1353,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 180,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__get_parent_path._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_tuple._ascii.ob_base,
            & const_str__get_parent_path._ascii.ob_base,
            & const_str__last_parent_path._ascii.ob_base,
            & const_str__epoch._ascii.ob_base,
            & const_str__last_epoch._ascii.ob_base,
            & const_str__path_finder._ascii.ob_base,
            & const_str__name._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str__path._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__recalculate = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_recalculate",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath._recalculate",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[144];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 143,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x16\x1b\x98\x44\xd7\x1c\x31\xd1\x1c\x31\xd3\x1c\x33\xd3\x16\x34\x88\x0b\xd8\x0b\x16\x98\x24\xd7\x1a\x30\xd1\x1a\x30\xd2\x0b\x30\xb0\x44\xb7\x4b\xb1\x4b\xc0\x34\xd7\x43\x53\xd1\x43\x53\xd2\x34\x53\xd8\x13\x17\xd7\x13\x24\xd1\x13\x24\xa0\x54\xa7\x5a\xa1\x5a\xb0\x1b\xd3\x13\x3d\x88\x44\xf0\x06\x00\x10\x14\xd0\x0f\x1f\xa0\x44\xa7\x4b\xa1\x4b\xd0\x24\x37\xd8\x13\x17\xd7\x13\x32\xd2\x13\x32\xd8\x21\x25\xd7\x21\x40\xd1\x21\x40\x90\x44\x94\x4a\xd8\x25\x30\x88\x44\xd4\x0c\x22\xd8\x1f\x23\x9f\x7b\x99\x7b\x88\x44\xd4\x0c\x1c\xd8\x0f\x13\x8f\x7a\x89\x7a\xd0\x08\x19",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_parent_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "parent_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_parent_path._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(346)
importlib__bootstrap_external_toplevel_consts_66_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 173,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1357,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 181,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__recalculate._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x73\x19\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x5f\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x81\x29\x7c\x02\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x1d\x7c\x02\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x11\x7c\x02\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(iter),
            & const_str__recalculate._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[23];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 22,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x13\x90\x44\xd7\x14\x25\xd1\x14\x25\xd3\x14\x27\xd3\x0f\x28\xd0\x08\x28",
};
static
    struct _PyCode_DEF(52)
importlib__bootstrap_external_toplevel_consts_66_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 26,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1371,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 182,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__recalculate._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.__getitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[23];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 22,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\xd7\x0f\x20\xd1\x0f\x20\xd3\x0f\x22\xa0\x35\xd1\x0f\x29\xd0\x08\x29",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_index = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "index",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_index._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(40)
importlib__bootstrap_external_toplevel_consts_66_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1374,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 183,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getitem__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.__setitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x1c\x20\x88\x04\x8f\x0a\x89\x0a\x90\x35\xd2\x08\x19",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_index._ascii.ob_base,
            &_Py_ID(path),
        },
    },
};
static
    struct _PyCode_DEF(34)
importlib__bootstrap_external_toplevel_consts_66_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 17,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1377,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 184,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__setitem__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_9_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(len),
            & const_str__recalculate._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.__len__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[23];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 22,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x12\x90\x34\xd7\x13\x24\xd1\x13\x24\xd3\x13\x26\xd3\x0f\x27\xd0\x08\x27",
};
static
    struct _PyCode_DEF(52)
importlib__bootstrap_external_toplevel_consts_66_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 26,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1380,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 185,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__len__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_10_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_11_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath(",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_66_consts_11_consts_1._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[41],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x11\x20\xa0\x14\xa7\x1a\xa1\x1a\xa0\x0e\xa8\x61\xd0\x0f\x30\xd0\x08\x30",
};
static
    struct _PyCode_DEF(34)
importlib__bootstrap_external_toplevel_consts_66_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 17,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_66_consts_11_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1383,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 186,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_11_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x02\x9d\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.__contains__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[23];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 22,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\x90\x74\xd7\x17\x28\xd1\x17\x28\xd3\x17\x2a\xd0\x0f\x2a\xd0\x08\x2a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_12_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(item),
        },
    },
};
static
    struct _PyCode_DEF(38)
importlib__bootstrap_external_toplevel_consts_66_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1386,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 187,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_12_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__contains__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_12_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x76\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__path._ascii.ob_base,
            &_Py_ID(append),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_13_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespacePath.append",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib__bootstrap_external_toplevel_consts_66_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0a\x89\x0a\xd7\x08\x19\xd1\x08\x19\x98\x24\xd5\x08\x1f",
};
static
    struct _PyCode_DEF(58)
importlib__bootstrap_external_toplevel_consts_66_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 29,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1389,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 188,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_12_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(append),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_66_consts_13_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & const_str__NamespacePath._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_external_toplevel_consts_66_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_4.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_5.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_6.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_7.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_8.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_9.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_10.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_11.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_12.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66_consts_13.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_66_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str__epoch._ascii.ob_base,
            &_Py_ID(__init__),
            & const_str__find_parent_path_names._ascii.ob_base,
            & const_str__get_parent_path._ascii.ob_base,
            & const_str__recalculate._ascii.ob_base,
            &_Py_ID(__iter__),
            &_Py_ID(__getitem__),
            &_Py_ID(__setitem__),
            &_Py_ID(__len__),
            &_Py_ID(__repr__),
            &_Py_ID(__contains__),
            &_Py_ID(append),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[71];
    }
importlib__bootstrap_external_toplevel_consts_66_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 70,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x04\x05\x14\xf0\x10\x00\x0e\x0f\x80\x46\xf2\x04\x05\x05\x28\xf2\x0e\x08\x05\x22\xf2\x14\x02\x05\x48\x01\xf2\x08\x0c\x05\x1a\xf2\x1c\x01\x05\x29\xf2\x06\x01\x05\x2a\xf2\x06\x01\x05\x21\xf2\x06\x01\x05\x28\xf2\x06\x01\x05\x31\xf2\x06\x01\x05\x2b\xf3\x06\x01\x05\x20",
};
static
    struct _PyCode_DEF(86)
importlib__bootstrap_external_toplevel_consts_66 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 43,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_66_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_66_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1325,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 189,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__NamespacePath._ascii.ob_base,
    .co_qualname = & const_str__NamespacePath._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_66_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x64\x07\x84\x00\x5a\x09\x64\x08\x84\x00\x5a\x0a\x64\x09\x84\x00\x5a\x0b\x64\x0a\x84\x00\x5a\x0c\x64\x0b\x84\x00\x5a\x0d\x64\x0c\x84\x00\x5a\x0e\x64\x0d\x84\x00\x5a\x0f\x79\x0e",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__NamespacePath._ascii.ob_base,
            & const_str__path._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x15\x23\xa0\x44\xa8\x24\xb0\x0b\xd3\x15\x3c\x88\x04\x8d\x0a",
};
static
    struct _PyCode_DEF(40)
importlib__bootstrap_external_toplevel_consts_68_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_68_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1397,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 190,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\x7c\x03\xab\x03\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            Py_True,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.is_package",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_68_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_2_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1400,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 191,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_is_package._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_STR(empty),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.get_source",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x11",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_68_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_3_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1403,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 192,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(get_source),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_4_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(dont_inherit),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            Py_None,
            &_Py_STR(empty),
            &_Py_STR(anon_string),
            & const_str_exec._ascii.ob_base,
            Py_True,
            & importlib__bootstrap_external_toplevel_consts_68_consts_4_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_compile._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.get_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x16\x90\x72\x98\x3a\xa0\x76\xb8\x44\xd4\x0f\x41\xd0\x08\x41",
};
static
    struct _PyCode_DEF(32)
importlib__bootstrap_external_toplevel_consts_68_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_68_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1406,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 193,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_code._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x02\x64\x03\x64\x04\xac\x05\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.create_module",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_68_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_54_consts_3_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1409,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 194,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_create_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.exec_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c",
};
static
    struct _PyCode_DEF(4)
importlib__bootstrap_external_toplevel_consts_68_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1412,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 195,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_64_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_exec_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[99];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 98,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4c\x6f\x61\x64\x20\x61\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x69\x73\x20\x64\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2e\x20\x20\x55\x73\x65\x20\x65\x78\x65\x63\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x29\x20\x69\x6e\x73\x74\x65\x61\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_7_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "namespace module loaded with path {!r}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_68_consts_7_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_7_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str__path._ascii.ob_base,
            & const_str__load_module_shim._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.load_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[50];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 49,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x09\x13\xd7\x08\x23\xd1\x08\x23\xd0\x24\x4c\xd8\x24\x28\xa7\x4a\xa1\x4a\xf4\x03\x01\x09\x30\xf4\x06\x00\x10\x1a\xd7\x0f\x2b\xd1\x0f\x2b\xa8\x44\xb0\x28\xd3\x0f\x3b\xd0\x08\x3b",
};
static
    struct _PyCode_DEF(110)
importlib__bootstrap_external_toplevel_consts_68_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_68_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1415,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 196,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_load_module._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_NamespaceReader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceReader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_8_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_NamespaceReader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_external_toplevel_consts_68_consts_8_consts_2._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_58_consts_8_names_0._ascii.ob_base,
            & const_str_NamespaceReader._ascii.ob_base,
            & const_str__path._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NamespaceLoader.get_resource_reader",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib__bootstrap_external_toplevel_consts_68_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdd\x08\x35\xd9\x0f\x1e\x98\x74\x9f\x7a\x99\x7a\xd3\x0f\x2a\xd0\x08\x2a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(module),
            & const_str_NamespaceReader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(50)
importlib__bootstrap_external_toplevel_consts_68_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_8_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_68_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1427,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 197,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_68_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_get_resource_reader._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_68_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x6d\x01\x7d\x02\x01\x00\x02\x00\x7c\x02\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_NamespaceLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_1.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_4.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_5.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_6.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_7.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_8.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_68_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            & const_str_is_package._ascii.ob_base,
            &_Py_ID(get_source),
            & const_str_get_code._ascii.ob_base,
            & const_str_create_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
            & const_str_get_resource_reader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[44];
    }
importlib__bootstrap_external_toplevel_consts_68_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 43,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x02\x01\x05\x3d\xf2\x06\x01\x05\x14\xf2\x06\x01\x05\x12\xf2\x06\x01\x05\x42\x01\xf2\x06\x01\x05\x39\xf2\x06\x01\x05\x0d\xf2\x06\x0a\x05\x3c\xf3\x18\x02\x05\x2b",
};
static
    struct _PyCode_DEF(60)
importlib__bootstrap_external_toplevel_consts_68 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_68_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1396,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 198,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_NamespaceLoader._ascii.ob_base,
    .co_qualname = & const_str_NamespaceLoader._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_68_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x64\x07\x84\x00\x5a\x09\x64\x08\x84\x00\x5a\x0a\x79\x09",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_PathFinder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathFinder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[63];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 62,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Meta path finder for sys.path and package __path__ attributes.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[126];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 125,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x61\x6c\x6c\x20\x74\x68\x65\x20\x69\x6e\x76\x61\x6c\x69\x64\x61\x74\x65\x5f\x63\x61\x63\x68\x65\x73\x28\x29\x20\x6d\x65\x74\x68\x6f\x64\x20\x6f\x6e\x20\x61\x6c\x6c\x20\x70\x61\x74\x68\x20\x65\x6e\x74\x72\x79\x20\x66\x69\x6e\x64\x65\x72\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x6f\x72\x65\x64\x20\x69\x6e\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x5f\x69\x6d\x70\x6f\x72\x74\x65\x72\x5f\x63\x61\x63\x68\x65\x73\x20\x28\x77\x68\x65\x72\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x29\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_invalidate_caches = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "invalidate_caches",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_MetadataPathFinder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MetadataPathFinder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_MetadataPathFinder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_70_consts_2_consts_0._ascii.ob_base,
            Py_None,
            & const_str_invalidate_caches._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_external_toplevel_consts_70_consts_2_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_path_importer_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_importer_cache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_names_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "importlib.metadata",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_list._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_path_importer_cache._ascii.ob_base,
            &_Py_ID(items),
            & const_str__path_isabs._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            & const_str_invalidate_caches._ascii.ob_base,
            & const_str__NamespacePath._ascii.ob_base,
            & const_str__epoch._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70_consts_2_names_9._ascii.ob_base,
            & const_str_MetadataPathFinder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathFinder.invalidate_caches",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[125];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 124,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x08\x00\x1d\x21\xa4\x13\xd7\x21\x38\xd1\x21\x38\xd7\x21\x3e\xd1\x21\x3e\xd3\x21\x40\xd6\x1c\x41\x89\x4c\x88\x44\x90\x26\xf0\x06\x00\x10\x16\x88\x7e\xa4\x5b\xb0\x14\xd4\x25\x36\xdc\x14\x17\xd7\x14\x2b\xd1\x14\x2b\xa8\x44\xd1\x14\x31\xdc\x11\x18\x98\x16\xd0\x21\x34\xd5\x11\x35\xd8\x10\x16\xd7\x10\x28\xd1\x10\x28\xd5\x10\x2a\xf0\x0d\x00\x1d\x42\x01\xf4\x12\x00\x09\x17\xd7\x08\x1d\xd2\x08\x1d\xa0\x11\xd1\x08\x22\xd5\x08\x1d\xe5\x08\x39\xd8\x08\x1a\xd7\x08\x2c\xd1\x08\x2c\xd5\x08\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_finder._ascii.ob_base,
            & const_str_MetadataPathFinder._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(312)
importlib__bootstrap_external_toplevel_consts_70_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 156,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_70_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_70_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1442,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 199,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_70_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_invalidate_caches._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_70_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_70_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x41\x00\x00\x5c\x02\x00\x00\x7d\x00\x7d\x01\x7c\x01\x81\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x12\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x3d\x00\x8c\x25\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x32\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x43\x04\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x78\x01\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7a\x0d\x00\x00\x63\x02\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x64\x05\x6c\x09\x6d\x0a\x7d\x02\x01\x00\x7c\x02\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Search sys.path_hooks for a finder for 'path'.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys.path_hooks is empty",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_70_consts_3_consts_0._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_70_consts_3_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_path_hooks = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_hooks",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_path_hooks._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_ImportWarning._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__path_hooks = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_hooks",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathFinder._path_hooks",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[90];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 89,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x0c\x0f\x8f\x3e\x89\x3e\xd0\x0b\x25\xac\x63\xaf\x6e\xaa\x6e\xdc\x0c\x15\x8f\x4e\x89\x4e\xd0\x1b\x34\xb4\x6d\xd4\x0c\x44\xdc\x14\x17\x97\x4e\x94\x4e\x88\x44\xf0\x02\x03\x0d\x19\xd9\x17\x1b\x98\x44\x93\x7a\xd2\x10\x21\xf0\x05\x00\x15\x23\xf0\x0c\x00\x14\x18\xf8\xf4\x07\x00\x14\x1f\xf2\x00\x01\x0d\x19\xd9\x10\x18\xf0\x03\x01\x0d\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x0f\x07\x41\x1b\x02\xc1\x1b\x09\x41\x27\x05\xc1\x26\x01\x41\x27\x05",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(hook),
        },
    },
};
static
    struct _PyCode_DEF(212)
importlib__bootstrap_external_toplevel_consts_70_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 106,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_70_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_70_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_70_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1460,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 200,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_70_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_hooks._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_70_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_70_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x2a\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x73\x1a\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x0c\x00\x00\x7d\x01\x09\x00\x02\x00\x7c\x01\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x63\x02\x01\x00\x53\x00\x04\x00\x79\x01\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x1b\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[211];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 210,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x47\x65\x74\x20\x74\x68\x65\x20\x66\x69\x6e\x64\x65\x72\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x65\x6e\x74\x72\x79\x20\x66\x72\x6f\x6d\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x5f\x69\x6d\x70\x6f\x72\x74\x65\x72\x5f\x63\x61\x63\x68\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x65\x6e\x74\x72\x79\x20\x69\x73\x20\x6e\x6f\x74\x20\x69\x6e\x20\x74\x68\x65\x20\x63\x61\x63\x68\x65\x2c\x20\x66\x69\x6e\x64\x20\x74\x68\x65\x20\x61\x70\x70\x72\x6f\x70\x72\x69\x61\x74\x65\x20\x66\x69\x6e\x64\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x6e\x64\x20\x63\x61\x63\x68\x65\x20\x69\x74\x2e\x20\x49\x66\x20\x6e\x6f\x20\x66\x69\x6e\x64\x65\x72\x20\x69\x73\x20\x61\x76\x61\x69\x6c\x61\x62\x6c\x65\x2c\x20\x73\x74\x6f\x72\x65\x20\x4e\x6f\x6e\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_70_consts_4_consts_0._ascii.ob_base,
            &_Py_STR(empty),
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_FileNotFoundError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileNotFoundError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__os._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str_FileNotFoundError._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_path_importer_cache._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str__path_hooks._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str__path_importer_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_importer_cache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathFinder._path_importer_cache",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[129];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 128,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x10\x00\x0c\x10\x90\x32\x8a\x3a\xf0\x02\x05\x0d\x1c\xdc\x17\x1a\x97\x7a\x91\x7a\x93\x7c\x90\x04\xf0\x0a\x04\x09\x33\xdc\x15\x18\xd7\x15\x2c\xd1\x15\x2c\xa8\x54\xd1\x15\x32\x88\x46\xf0\x08\x00\x10\x16\x88\x0d\xf8\xf4\x13\x00\x14\x25\xf2\x00\x03\x0d\x1c\xf1\x06\x00\x18\x1c\xf0\x07\x03\x0d\x1c\xfb\xf4\x0c\x00\x10\x18\xf2\x00\x02\x09\x33\xd8\x15\x18\x97\x5f\x91\x5f\xa0\x54\xd3\x15\x2a\x88\x46\xd8\x2c\x32\x8c\x43\xd7\x0c\x23\xd1\x0c\x23\xa0\x44\xd2\x0c\x29\xd8\x0f\x15\x88\x0d\xf0\x07\x02\x09\x33\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x87\x14\x31\x00\x9c\x13\x41\x00\x00\xb1\x09\x3d\x03\xbc\x01\x3d\x03\xc1\x00\x2d\x41\x31\x03\xc1\x30\x01\x41\x31\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            &_Py_ID(path),
            & const_str_finder._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(232)
importlib__bootstrap_external_toplevel_consts_70_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 116,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_70_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_70_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_70_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1473,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 201,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_70_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__path_importer_cache._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_70_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_70_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x64\x01\x6b\x28\x00\x00\x72\x15\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x7c\x02\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x02\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x28\x01\x00\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x59\x00\x7c\x02\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[64];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 63,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Find the loader or namespace_path for this module/package name.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_6_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spec missing loader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_70_consts_6_consts_0._ascii.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_70_consts_6_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            & const_str__path_importer_cache._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(extend),
            &_Py_ID(_bootstrap),
            & const_str_ModuleSpec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__get_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_spec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathFinder._get_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[174];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 173,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x1a\x1c\x88\x0e\xdb\x15\x19\x88\x45\xdc\x13\x1d\x98\x65\xa4\x53\xd4\x13\x29\xd8\x10\x18\xd8\x15\x18\xd7\x15\x2d\xd1\x15\x2d\xa8\x65\xd3\x15\x34\x88\x46\xd8\x0f\x15\xd1\x0f\x21\xd8\x17\x1d\xd7\x17\x27\xd1\x17\x27\xa8\x08\xb0\x26\xd3\x17\x39\x90\x04\xd8\x13\x17\x90\x3c\xd8\x14\x1c\xd8\x13\x17\x97\x3b\x91\x3b\xd0\x13\x2a\xd8\x1b\x1f\x92\x4b\xd8\x1b\x1f\xd7\x1b\x3a\xd1\x1b\x3a\x90\x08\xd8\x13\x1b\xd0\x13\x23\xdc\x1a\x25\xd0\x26\x3b\xd3\x1a\x3c\xd0\x14\x3c\xf0\x0a\x00\x11\x1f\xd7\x10\x25\xd1\x10\x25\xa0\x68\xd5\x10\x2f\xf0\x23\x00\x16\x1a\xf4\x26\x00\x14\x1e\xd7\x13\x28\xd1\x13\x28\xa8\x18\xb0\x34\xd3\x13\x38\x88\x44\xd8\x2e\x3c\x88\x44\xd4\x0c\x2b\xd8\x13\x17\x88\x4b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_namespace_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "namespace_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_entry = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "entry",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_portions = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "portions",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(target),
            & const_str_namespace_path._ascii.ob_base,
            & const_str_entry._ascii.ob_base,
            & const_str_finder._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
            & const_str_portions._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(314)
importlib__bootstrap_external_toplevel_consts_70_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 157,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_70_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_70_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1495,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 202,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_70_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__get_spec._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_70_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_70_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x67\x00\x7d\x04\x7c\x02\x44\x00\x5d\x76\x00\x00\x7d\x05\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x14\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x80\x01\x8c\x28\x7c\x06\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x07\x80\x01\x8c\x3d\x7c\x07\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x04\x7c\x07\x63\x02\x01\x00\x53\x00\x7c\x07\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x80\x0b\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x04\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x78\x04\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x04\x7c\x07\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[142];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 141,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x72\x79\x20\x74\x6f\x20\x66\x69\x6e\x64\x20\x61\x20\x73\x70\x65\x63\x20\x66\x6f\x72\x20\x27\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x27\x20\x6f\x6e\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x20\x6f\x72\x20\x27\x70\x61\x74\x68\x27\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x73\x65\x61\x72\x63\x68\x20\x69\x73\x20\x62\x61\x73\x65\x64\x20\x6f\x6e\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x5f\x68\x6f\x6f\x6b\x73\x20\x61\x6e\x64\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x5f\x69\x6d\x70\x6f\x72\x74\x65\x72\x5f\x63\x61\x63\x68\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_70_consts_7_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(path),
            & const_str__get_spec._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            &_Py_ID(origin),
            & const_str__NamespacePath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathFinder.find_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[115];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 114,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0c\x00\x0c\x10\x88\x3c\xdc\x13\x16\x97\x38\x91\x38\x88\x44\xd8\x0f\x12\x8f\x7d\x89\x7d\x98\x58\xa0\x74\xa8\x56\xd3\x0f\x34\x88\x04\xd8\x0b\x0f\x88\x3c\xd8\x13\x17\xd8\x0d\x11\x8f\x5b\x89\x5b\xd0\x0d\x20\xd8\x1d\x21\xd7\x1d\x3c\xd1\x1d\x3c\x88\x4e\xd9\x0f\x1d\xf0\x06\x00\x1f\x23\x90\x04\x94\x0b\xdc\x32\x40\xc0\x18\xc8\x3e\xd0\x5b\x5e\xd7\x5b\x68\xd1\x5b\x68\xd3\x32\x69\x90\x04\xd4\x10\x2f\xd8\x17\x1b\x90\x0b\xe0\x17\x1b\xe0\x13\x17\x88\x4b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(target),
            & const_str_spec._ascii.ob_base,
            & const_str_namespace_path._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(214)
importlib__bootstrap_external_toplevel_consts_70_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 107,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_70_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_70_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1524,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 203,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_70_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_find_spec._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_70_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_70_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x80\x10\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\x7c\x03\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x80\x01\x79\x01\x7c\x04\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x34\x7c\x04\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x72\x25\x64\x01\x7c\x04\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x05\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x7c\x04\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x53\x00\x79\x01\x7c\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[289];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 288,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x46\x69\x6e\x64\x20\x64\x69\x73\x74\x72\x69\x62\x75\x74\x69\x6f\x6e\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x61\x6e\x20\x69\x74\x65\x72\x61\x62\x6c\x65\x20\x6f\x66\x20\x61\x6c\x6c\x20\x44\x69\x73\x74\x72\x69\x62\x75\x74\x69\x6f\x6e\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x73\x20\x63\x61\x70\x61\x62\x6c\x65\x20\x6f\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6c\x6f\x61\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x6d\x65\x74\x61\x64\x61\x74\x61\x20\x66\x6f\x72\x20\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x6d\x61\x74\x63\x68\x69\x6e\x67\x20\x60\x60\x63\x6f\x6e\x74\x65\x78\x74\x2e\x6e\x61\x6d\x65\x60\x60\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x28\x6f\x72\x20\x61\x6c\x6c\x20\x6e\x61\x6d\x65\x73\x20\x69\x66\x20\x60\x60\x4e\x6f\x6e\x65\x60\x60\x20\x69\x6e\x64\x69\x63\x61\x74\x65\x64\x29\x20\x61\x6c\x6f\x6e\x67\x20\x74\x68\x65\x20\x70\x61\x74\x68\x73\x20\x69\x6e\x20\x74\x68\x65\x20\x6c\x69\x73\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x66\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x60\x60\x63\x6f\x6e\x74\x65\x78\x74\x2e\x70\x61\x74\x68\x60\x60\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_70_consts_8_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & importlib__bootstrap_external_toplevel_consts_70_consts_2_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_find_distributions = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "find_distributions",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_70_consts_2_names_9._ascii.ob_base,
            & const_str_MetadataPathFinder._ascii.ob_base,
            & const_str_find_distributions._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathFinder.find_distributions",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
importlib__bootstrap_external_toplevel_consts_70_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf5\x14\x00\x09\x3a\xd8\x0f\x34\xd0\x0f\x21\xd7\x0f\x34\xd1\x0f\x34\xb0\x64\xd0\x0f\x45\xb8\x66\xd1\x0f\x45\xd0\x08\x45",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(args),
            & const_str_kwargs._ascii.ob_base,
            & const_str_MetadataPathFinder._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(50)
importlib__bootstrap_external_toplevel_consts_70_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_70_consts_8_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_70_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 15,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1548,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 204,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_70_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_find_distributions._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_70_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_70_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x6d\x01\x7d\x02\x01\x00\x02\x00\x7c\x02\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x69\x00\x7c\x01\xa4\x01\x8e\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_PathFinder._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70_consts_4.ob_base.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_70_consts_6.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70_consts_7.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70_consts_8.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_70_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_staticmethod._ascii.ob_base,
            & const_str_invalidate_caches._ascii.ob_base,
            & const_str__path_hooks._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            & const_str__path_importer_cache._ascii.ob_base,
            & const_str__get_spec._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_find_distributions._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[126];
    }
importlib__bootstrap_external_toplevel_consts_70_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 125,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x48\xe0\x05\x11\xf1\x02\x0f\x05\x2f\xf3\x03\x00\x06\x12\xf0\x02\x0f\x05\x2f\xf0\x22\x00\x06\x12\xf1\x02\x0a\x05\x18\xf3\x03\x00\x06\x12\xf0\x02\x0a\x05\x18\xf0\x18\x00\x06\x11\xf1\x02\x13\x05\x16\xf3\x03\x00\x06\x11\xf0\x02\x13\x05\x16\xf0\x2a\x00\x06\x11\xf2\x02\x1a\x05\x18\xf3\x03\x00\x06\x11\xf0\x02\x1a\x05\x18\xf0\x38\x00\x06\x11\xf2\x02\x15\x05\x18\xf3\x03\x00\x06\x11\xf0\x02\x15\x05\x18\xf0\x2e\x00\x06\x12\xf1\x02\x0a\x05\x46\x01\xf3\x03\x00\x06\x12\xf1\x02\x0a\x05\x46\x01",
};
static
    struct _PyCode_DEF(116)
importlib__bootstrap_external_toplevel_consts_70 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 58,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_70_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_70_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1438,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 205,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_PathFinder._ascii.ob_base,
    .co_qualname = & const_str_PathFinder._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_70_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x04\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x07\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x65\x07\x64\x09\x64\x06\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x09\x65\x07\x64\x0a\x64\x07\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0a\x65\x04\x64\x08\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0b\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_FileFinder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[173];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 172,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x46\x69\x6c\x65\x2d\x62\x61\x73\x65\x64\x20\x66\x69\x6e\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x49\x6e\x74\x65\x72\x61\x63\x74\x69\x6f\x6e\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x73\x79\x73\x74\x65\x6d\x20\x61\x72\x65\x20\x63\x61\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x70\x65\x72\x66\x6f\x72\x6d\x61\x6e\x63\x65\x2c\x20\x62\x65\x69\x6e\x67\x0a\x20\x20\x20\x20\x72\x65\x66\x72\x65\x73\x68\x65\x64\x20\x77\x68\x65\x6e\x20\x74\x68\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x74\x68\x65\x20\x66\x69\x6e\x64\x65\x72\x20\x69\x73\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x68\x61\x73\x20\x62\x65\x65\x6e\x20\x6d\x6f\x64\x69\x66\x69\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[155];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 154,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x49\x6e\x69\x74\x69\x61\x6c\x69\x7a\x65\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x73\x65\x61\x72\x63\x68\x20\x6f\x6e\x20\x61\x6e\x64\x20\x61\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x20\x6e\x75\x6d\x62\x65\x72\x20\x6f\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x32\x2d\x74\x75\x70\x6c\x65\x73\x20\x63\x6f\x6e\x74\x61\x69\x6e\x69\x6e\x67\x20\x74\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x73\x75\x66\x66\x69\x78\x65\x73\x20\x74\x68\x65\x20\x6c\x6f\x61\x64\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x63\x6f\x67\x6e\x69\x7a\x65\x73\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder.__init__.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xd0\x1a\x43\xb9\x28\xb0\x06\x98\x46\xa0\x46\xd4\x1b\x2b\xb9\x28\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x0e\x11\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            & const_str_suffix._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(38)
importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1577,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 206,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x08\x00\x00\x7d\x01\x7c\x01\x89\x02\x66\x02\x96\x01\x97\x01\x01\x00\x8c\x0a\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__loaders = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_loaders",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__path_mtime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_mtime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__path_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_cache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__relaxed_path_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_relaxed_path_cache",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(extend),
            & const_str__loaders._ascii.ob_base,
            & const_str__os._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            &_Py_ID(path),
            & const_str__path_abspath._ascii.ob_base,
            & const_str__path_mtime._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            & const_str__path_cache._ascii.ob_base,
            & const_str__relaxed_path_cache._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[113];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 112,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf0\x08\x00\x13\x15\x88\x07\xdb\x20\x2e\xd1\x0c\x1c\x88\x46\x90\x48\xd8\x0c\x13\x8f\x4e\x89\x4e\xd3\x1a\x43\xb9\x28\xd3\x1a\x43\xd5\x0c\x43\xf0\x03\x00\x21\x2f\xe0\x18\x1f\x88\x04\x8c\x0d\xe1\x0f\x13\x90\x74\x98\x73\x92\x7b\xdc\x18\x1b\x9f\x0a\x99\x0a\x9b\x0c\x88\x44\x8d\x49\xe4\x18\x25\xa0\x64\xd3\x18\x2b\x88\x44\x8c\x49\xd8\x1b\x1d\x88\x04\xd4\x08\x18\xdc\x1b\x1e\x9b\x35\x88\x04\xd4\x08\x18\xdc\x23\x26\xa3\x35\x88\x04\xd5\x08\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_loader_details = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader_details",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_loaders = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loaders",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            & const_str_loader_details._ascii.ob_base,
            & const_str_loaders._ascii.ob_base,
            & const_str_suffixes._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[7];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_2_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 6,
    },
    .ob_shash = -1,
    .ob_sval = "     @",
};
static
    struct _PyCode_DEF(268)
importlib__bootstrap_external_toplevel_consts_72_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 134,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_72_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1571,
    .co_nlocalsplus = 6,
    .co_nlocals = 5,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 207,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_72_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x05\x97\x00\x67\x00\x7d\x03\x7c\x02\x44\x00\x5d\x1f\x00\x00\x5c\x02\x00\x00\x8a\x05\x7d\x04\x7c\x03\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x88\x05\x66\x01\x64\x01\x84\x08\x7c\x04\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x21\x04\x00\x7c\x03\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x72\x05\x7c\x01\x64\x02\x6b\x28\x00\x00\x72\x1a\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6e\x10\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x79\x04",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Invalidate the directory mtime.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_72_consts_3_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_mtime._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder.invalidate_caches",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x1b\x1d\x88\x04\xd5\x08\x18",
};
static
    struct _PyCode_DEF(18)
importlib__bootstrap_external_toplevel_consts_72_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 9,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_72_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1588,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 208,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_invalidate_caches._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_33_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_spec_from_file_location._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder._get_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[39];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 38,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd9\x11\x1d\x98\x68\xa8\x04\xd3\x11\x2d\x88\x06\xdc\x0f\x26\xa0\x78\xb0\x14\xb8\x66\xd8\x42\x46\xf4\x03\x01\x10\x48\x01\xf0\x00\x01\x09\x48\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_loader_class._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            & const_str_smsl._ascii.ob_base,
            &_Py_ID(target),
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(50)
importlib__bootstrap_external_toplevel_consts_72_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_4_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_72_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 6,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1592,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 209,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_72_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__get_spec._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x7c\x01\x7c\x02\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x03\x7c\x06\x7c\x04\xac\x01\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[112];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 111,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x72\x79\x20\x74\x6f\x20\x66\x69\x6e\x64\x20\x61\x20\x73\x70\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x6d\x61\x74\x63\x68\x69\x6e\x67\x20\x73\x70\x65\x63\x2c\x20\x6f\x72\x20\x4e\x6f\x6e\x65\x20\x69\x66\x20\x6e\x6f\x74\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "trying {}",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "possible namespace for {}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_72_consts_6_consts_0._ascii.ob_base,
            Py_False,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            &_Py_ID(__init__),
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_72_consts_6_consts_7._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_24._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_6_consts_9._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__fill_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fill_cache",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[23];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 23,
        },
        .ob_item = {
            & const_str_rpartition._ascii.ob_base,
            & const_str__path_stat._ascii.ob_base,
            &_Py_ID(path),
            & const_str__os._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str_st_mtime._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str__path_mtime._ascii.ob_base,
            & const_str__fill_cache._ascii.ob_base,
            & const_str__relax_case._ascii.ob_base,
            & const_str__relaxed_path_cache._ascii.ob_base,
            & const_str_lower._ascii.ob_base,
            & const_str__path_cache._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str__loaders._ascii.ob_base,
            & const_str__path_isfile._ascii.ob_base,
            & const_str__get_spec._ascii.ob_base,
            & const_str__path_isdir._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str_ModuleSpec._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder.find_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[477];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 476,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x18\x1d\x88\x0c\xd8\x16\x1e\xd7\x16\x29\xd1\x16\x29\xa8\x23\xd3\x16\x2e\xa8\x71\xd1\x16\x31\x88\x0b\xf0\x02\x03\x09\x17\xdc\x14\x1e\x98\x74\x9f\x79\x99\x79\xd2\x1f\x38\xac\x43\xaf\x4a\xa9\x4a\xab\x4c\xd3\x14\x39\xd7\x14\x42\xd1\x14\x42\x88\x45\xf0\x06\x00\x0c\x11\x90\x44\xd7\x14\x24\xd1\x14\x24\xd2\x0b\x24\xd8\x0c\x10\xd7\x0c\x1c\xd1\x0c\x1c\xd4\x0c\x1e\xd8\x1f\x24\x88\x44\xd4\x0c\x1c\xe4\x0b\x16\x8c\x3d\xd8\x14\x18\xd7\x14\x2c\xd1\x14\x2c\x88\x45\xd8\x1b\x26\xd7\x1b\x2c\xd1\x1b\x2c\xd3\x1b\x2e\x89\x4c\xe0\x14\x18\xd7\x14\x24\xd1\x14\x24\x88\x45\xd8\x1b\x26\x88\x4c\xe0\x0b\x17\x98\x35\xd1\x0b\x20\xdc\x18\x22\xa0\x34\xa7\x39\xa1\x39\xa8\x6b\xd3\x18\x3a\x88\x49\xd8\x28\x2c\xaf\x0d\xac\x0d\xd1\x10\x24\x90\x06\x98\x0c\xd8\x20\x2a\xa8\x56\xd1\x20\x33\x90\x0d\xdc\x1c\x26\xa0\x79\xb0\x2d\xd3\x1c\x40\x90\x09\xdc\x13\x1f\xa0\x09\xd5\x13\x2a\xd8\x1b\x1f\x9f\x3e\x99\x3e\xa8\x2c\xb8\x08\xc0\x29\xc8\x69\xc8\x5b\xd0\x5a\x60\xd3\x1b\x61\xd2\x14\x61\xf0\x09\x00\x29\x36\xf4\x10\x00\x20\x2b\xa8\x39\xd3\x1f\x35\x90\x0c\xe0\x24\x28\xa7\x4d\xa4\x4d\xd1\x0c\x20\x88\x46\x90\x4c\xf0\x02\x03\x0d\x1c\xdc\x1c\x26\xa0\x74\xa7\x79\xa1\x79\xb0\x2b\xc0\x06\xd1\x32\x46\xd3\x1c\x47\x90\x09\xf4\x06\x00\x0d\x17\xd7\x0c\x27\xd1\x0c\x27\xa8\x0b\xb0\x59\xc8\x21\xd0\x0c\x27\xd4\x0c\x4c\xd8\x0f\x1b\x98\x66\xd1\x0f\x24\xa8\x05\xd2\x0f\x2d\xdc\x13\x1f\xa0\x09\xd5\x13\x2a\xd8\x1b\x1f\x9f\x3e\x99\x3e\xa8\x2c\xb8\x08\xc0\x29\xd8\x2a\x2e\xb0\x06\xf3\x03\x01\x1c\x38\xf2\x00\x01\x15\x38\xf0\x11\x00\x25\x32\xf1\x14\x00\x0c\x18\xdc\x0c\x16\xd7\x0c\x27\xd1\x0c\x27\xd0\x28\x43\xc0\x59\xd4\x0c\x4f\xdc\x13\x1d\xd7\x13\x28\xd1\x13\x28\xa8\x18\xb0\x34\xd3\x13\x38\x88\x44\xd8\x2f\x38\xa8\x6b\x88\x44\xd4\x0c\x2b\xd8\x13\x17\x88\x4b\xd8\x0f\x13\xf8\xf4\x51\x01\x00\x10\x17\xf2\x00\x01\x09\x17\xd8\x14\x16\x8b\x45\xf0\x03\x01\x09\x17\xfb\xf4\x38\x00\x14\x1e\xf2\x00\x01\x0d\x1c\xda\x17\x1b\xf0\x03\x01\x0d\x1c\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x98\x35\x46\x3e\x00\xc4\x28\x19\x47\x10\x02\xc6\x3e\x0b\x47\x0d\x03\xc7\x0c\x01\x47\x0d\x03\xc7\x10\x09\x47\x1d\x05\xc7\x1c\x01\x47\x1d\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_is_namespace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_namespace",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_tail_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tail_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_cache_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cache_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_base_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "base_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_init_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "init_filename",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_full_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "full_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(target),
            & const_str_is_namespace._ascii.ob_base,
            & const_str_tail_module._ascii.ob_base,
            & const_str_mtime._ascii.ob_base,
            & const_str_cache._ascii.ob_base,
            & const_str_cache_module._ascii.ob_base,
            & const_str_base_path._ascii.ob_base,
            & const_str_suffix._ascii.ob_base,
            & const_str_loader_class._ascii.ob_base,
            & const_str_init_filename._ascii.ob_base,
            & const_str_full_path._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_6_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "              ",
};
static
    struct _PyCode_DEF(960)
importlib__bootstrap_external_toplevel_consts_72_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 480,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_72_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_72_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 22 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 1597,
    .co_nlocalsplus = 14,
    .co_nlocals = 14,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 210,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_72_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_find_spec._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x03\x7c\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x64\x03\x19\x00\x00\x00\x7d\x04\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x78\x01\x73\x14\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x7c\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x17\x7c\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x7c\x00\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x72\x1d\x7c\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x04\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x6e\x0e\x7c\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x04\x7d\x07\x7c\x07\x7c\x06\x76\x00\x72\x69\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x39\x00\x00\x5c\x02\x00\x00\x7d\x09\x7d\x0a\x64\x05\x7c\x09\x7a\x00\x00\x00\x7d\x0b\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\x7c\x0b\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0c\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x23\x7c\x00\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x7c\x01\x7c\x0c\x7c\x08\x67\x01\x7c\x02\xab\x05\x00\x00\x00\x00\x00\x00\x63\x02\x01\x00\x53\x00\x04\x00\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x61\x00\x00\x5c\x02\x00\x00\x7d\x09\x7d\x0a\x09\x00\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x09\x7a\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0c\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x7c\x0c\x64\x03\xac\x08\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x07\x7c\x09\x7a\x00\x00\x00\x7c\x06\x76\x00\x73\x01\x8c\x40\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x4c\x7c\x00\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x7c\x01\x7c\x0c\x64\x06\x7c\x02\xab\x05\x00\x00\x00\x00\x00\x00\x63\x02\x01\x00\x53\x00\x04\x00\x7c\x03\x72\x36\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\x7f\x08\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x2b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0d\x7c\x08\x67\x01\x7c\x0d\x5f\x16\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x53\x00\x79\x06\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x06\x01\x00\x64\x04\x7d\x05\x59\x00\x90\x01\x8c\x7f\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x24\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x01\x00\x79\x06\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[69];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 68,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Fill the cache of potential modules and packages for this directory.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_72_consts_7_consts_0._ascii.ob_base,
            & const_str_win._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_listdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "listdir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_PermissionError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PermissionError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_NotADirectoryError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "NotADirectoryError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str__os._ascii.ob_base,
            & const_str_listdir._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str_FileNotFoundError._ascii.ob_base,
            & const_str_PermissionError._ascii.ob_base,
            & const_str_NotADirectoryError._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            & const_str__path_cache._ascii.ob_base,
            & const_str_partition._ascii.ob_base,
            & const_str_lower._ascii.ob_base,
            &_Py_ID(add),
            & const_str__CASE_INSENSITIVE_PLATFORMS._ascii.ob_base,
            & const_str__relaxed_path_cache._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder._fill_cache",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[261];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 260,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x79\x89\x79\x88\x04\xf0\x02\x05\x09\x1a\xdc\x17\x1a\x97\x7b\x91\x7b\xa0\x34\xd2\x23\x37\xac\x33\xaf\x3a\xa9\x3a\xab\x3c\xd3\x17\x38\x88\x48\xf4\x0e\x00\x10\x13\x8f\x7c\x89\x7c\xd7\x0f\x26\xd1\x0f\x26\xa0\x75\xd4\x0f\x2d\xdc\x1f\x22\xa0\x38\x9b\x7d\x88\x44\xd5\x0c\x1c\xf4\x0e\x00\x25\x28\xa3\x45\xd0\x0c\x21\xdb\x18\x20\x90\x04\xd8\x24\x28\xa7\x4e\xa1\x4e\xb0\x33\xd3\x24\x37\xd1\x10\x21\x90\x04\x90\x63\x98\x36\xd9\x13\x16\xd8\x22\x26\xa0\x16\xa0\x71\xa8\x16\xaf\x1c\xa9\x1c\xab\x1e\xd0\x28\x38\xd0\x1f\x39\x91\x48\xe0\x1f\x23\x90\x48\xd8\x10\x25\xd7\x10\x29\xd1\x10\x29\xa8\x28\xd5\x10\x33\xf0\x0d\x00\x19\x21\xf0\x0e\x00\x20\x35\x88\x44\xd4\x0c\x1c\xdc\x0b\x0e\x8f\x3c\x89\x3c\xd7\x0b\x22\xd1\x0b\x22\xd4\x23\x3e\xd4\x0b\x3f\xd9\x3d\x45\xd3\x27\x46\xb9\x58\xb0\x72\xa8\x02\xaf\x08\xa9\x08\xad\x0a\xb8\x58\xd1\x27\x46\x88\x44\xd5\x0c\x24\xf0\x03\x00\x0c\x40\x01\xf8\xf4\x2f\x00\x11\x22\xa4\x3f\xd4\x34\x46\xd0\x0f\x47\xf2\x00\x03\x09\x1a\xf0\x06\x00\x18\x1a\x8a\x48\xf0\x07\x03\x09\x1a\xfc\xf2\x30\x00\x28\x47\x01",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x8e\x2b\x44\x09\x00\xc3\x28\x17\x44\x25\x04\xc4\x09\x16\x44\x22\x03\xc4\x21\x01\x44\x22\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_contents = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "contents",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_lower_suffix_contents = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lower_suffix_contents",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_new_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "new_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_fn = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fn",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            & const_str_contents._ascii.ob_base,
            & const_str_lower_suffix_contents._ascii.ob_base,
            &_Py_ID(item),
            &_Py_ID(name),
            & const_str_dot._ascii.ob_base,
            & const_str_suffix._ascii.ob_base,
            & const_str_new_name._ascii.ob_base,
            & const_str_fn._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(596)
importlib__bootstrap_external_toplevel_consts_72_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 298,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_72_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_72_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1648,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 211,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_72_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__fill_cache._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_7_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x78\x01\x73\x14\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x73\x11\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x6e\x58\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x02\x44\x00\x5d\x42\x00\x00\x7d\x04\x7c\x04\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x05\x7d\x06\x7d\x07\x7c\x06\x72\x16\x7c\x05\x9b\x00\x64\x02\x7c\x07\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x03\x7d\x08\x6e\x02\x7c\x05\x7d\x08\x7c\x03\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x44\x04\x00\x7c\x03\x7c\x00\x5f\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x24\x7c\x02\x44\x00\x8f\x09\x63\x02\x68\x00\x63\x02\x5d\x12\x00\x00\x7d\x09\x7c\x09\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x92\x02\x8c\x14\x04\x00\x63\x02\x7d\x09\x7c\x00\x5f\x10\x00\x00\x00\x00\x00\x00\x00\x00\x79\x03\x79\x03\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x05\x01\x00\x67\x00\x7d\x02\x59\x00\x8c\xe8\x77\x00\x78\x03\x59\x00\x77\x01\x63\x02\x01\x00\x63\x02\x7d\x09\x77\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[277];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 276,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x63\x6c\x61\x73\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x77\x68\x69\x63\x68\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x61\x20\x63\x6c\x6f\x73\x75\x72\x65\x20\x74\x6f\x20\x75\x73\x65\x20\x6f\x6e\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x5f\x68\x6f\x6f\x6b\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x77\x68\x69\x63\x68\x20\x77\x69\x6c\x6c\x20\x72\x65\x74\x75\x72\x6e\x20\x61\x6e\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x20\x75\x73\x69\x6e\x67\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6c\x6f\x61\x64\x65\x72\x73\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x70\x61\x74\x68\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x61\x6c\x6c\x65\x64\x20\x6f\x6e\x20\x74\x68\x65\x20\x63\x6c\x6f\x73\x75\x72\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x63\x61\x6c\x6c\x65\x64\x20\x6f\x6e\x20\x74\x68\x65\x20\x63\x6c\x6f\x73\x75\x72\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x61\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x2c\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x61\x69\x73\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[46];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 45,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Path hook for importlib.machinery.FileFinder.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "only directories are supported",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__path_isdir._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str_path_hook_for_FileFinder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_hook_for_FileFinder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder.path_hook.<locals>.path_hook_for_FileFinder",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[42];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 41,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xe4\x13\x1e\x98\x74\xd4\x13\x24\xdc\x16\x21\xd0\x22\x42\xc8\x14\xd4\x16\x4e\xd0\x10\x4e\xd9\x13\x16\x90\x74\xd0\x13\x2d\x98\x6e\xd2\x13\x2d\xd0\x0c\x2d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_cls._ascii.ob_base,
            & const_str_loader_details._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[4];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 3,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x80\x80",
};
static
    struct _PyCode_DEF(70)
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1689,
    .co_nlocalsplus = 3,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 2,
    .co_version = 212,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_path_hook_for_FileFinder._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x02\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x0d\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x02\x00\x89\x01\x7c\x00\x67\x01\x89\x02\xa2\x01\xad\x06\x8e\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_0._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_path_hook = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_hook",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder.path_hook",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\xf9\x80\x00\xf5\x14\x04\x09\x2e\xf0\x0c\x00\x10\x28\xd0\x08\x27",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_loader_details._ascii.ob_base,
            & const_str_path_hook_for_FileFinder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[4];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_8_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 3,
    },
    .ob_shash = -1,
    .ob_sval = "`` ",
};
static
    struct _PyCode_DEF(22)
importlib__bootstrap_external_toplevel_consts_72_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 11,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1679,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 2,
    .co_nfreevars = 0,
    .co_version = 213,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_72_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_8_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_path_hook._ascii.ob_base,
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_8_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x87\x01\x97\x00\x88\x00\x88\x01\x66\x02\x64\x01\x84\x08\x7d\x02\x7c\x02\x53\x00",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_9_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder(",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_72_consts_9_consts_1._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[41],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FileFinder.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
importlib__bootstrap_external_toplevel_consts_72_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x11\x1c\x98\x54\x9f\x59\x99\x59\x98\x4d\xa8\x11\xd0\x0f\x2b\xd0\x08\x2b",
};
static
    struct _PyCode_DEF(34)
importlib__bootstrap_external_toplevel_consts_72_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 17,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts_9_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1697,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 214,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & importlib__bootstrap_external_toplevel_consts_72_consts_9_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x02\x9d\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_FileFinder._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_2.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_3.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_4.ob_base.ob_base,
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_72_consts_6.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_7.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_8.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_9.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_72_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            & const_str_invalidate_caches._ascii.ob_base,
            & const_str__get_spec._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str__fill_cache._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            & const_str_path_hook._ascii.ob_base,
            &_Py_ID(__repr__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[61];
    }
importlib__bootstrap_external_toplevel_consts_72_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 60,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x05\x05\x08\xf2\x0e\x0f\x05\x29\xf2\x22\x02\x05\x1e\xf2\x08\x03\x05\x48\x01\xf3\x0a\x31\x05\x14\xf2\x66\x01\x1d\x05\x47\x01\xf0\x3e\x00\x06\x11\xf1\x02\x0f\x05\x28\xf3\x03\x00\x06\x11\xf0\x02\x0f\x05\x28\xf3\x22\x01\x05\x2c",
};
static
    struct _PyCode_DEF(70)
importlib__bootstrap_external_toplevel_consts_72 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_72_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_72_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1562,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 215,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str_FileFinder._ascii.ob_base,
    .co_qualname = & const_str_FileFinder._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_72_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x0a\x64\x06\x84\x01\x5a\x07\x64\x07\x84\x00\x5a\x08\x65\x09\x64\x08\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0a\x64\x09\x84\x00\x5a\x0b\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_74_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__loader__),
            &_Py_ID(__spec__),
            & importlib__bootstrap_toplevel_consts_33_consts_4._object.ob_base.ob_base,
            &_Py_ID(__file__),
            & const_str___cached__._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_Exception = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Exception",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_74_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(get),
            & const_str_loader._ascii.ob_base,
            & const_str_SourcelessFileLoader._ascii.ob_base,
            & const_str_SourceFileLoader._ascii.ob_base,
            & const_str_spec_from_file_location._ascii.ob_base,
            & const_str__path_abspath._ascii.ob_base,
            & const_str_cached._ascii.ob_base,
            & const_str_Exception._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[171];
    }
importlib__bootstrap_external_toplevel_consts_74_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 170,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0d\x0f\x8f\x56\x89\x56\x90\x4c\xd3\x0d\x21\x80\x46\xd8\x0b\x0d\x8f\x36\x89\x36\x90\x2a\xd3\x0b\x1d\x80\x44\xd9\x0b\x11\xd9\x0b\x0f\xd8\x15\x19\x97\x5b\x91\x5b\x89\x46\xd8\x0d\x15\x98\x19\xd2\x0d\x22\xdc\x15\x29\xa8\x24\xb0\x08\xd3\x15\x39\x89\x46\xe4\x15\x25\xa0\x64\xa8\x48\xd3\x15\x35\x88\x46\xd9\x0b\x0f\xdc\x0f\x26\xa0\x74\xa8\x58\xb8\x66\xd4\x0f\x45\x88\x04\xd9\x0b\x14\xdc\x1a\x27\xa8\x09\xd3\x1a\x32\x88\x44\x8c\x4b\xf0\x02\x07\x05\x0d\xd8\x19\x1d\x88\x02\x88\x3a\x89\x0e\xd8\x1b\x21\x88\x02\x88\x3c\xd1\x08\x18\xd8\x19\x21\x88\x02\x88\x3a\x89\x0e\xd8\x1b\x24\x88\x02\x88\x3c\xd2\x08\x18\xf8\xdc\x0b\x14\xf2\x00\x02\x05\x0d\xe1\x08\x0c\xf0\x05\x02\x05\x0d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
importlib__bootstrap_external_toplevel_consts_74_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x35\x14\x42\x0a\x00\xc2\x0a\x09\x42\x16\x03\xc2\x15\x01\x42\x16\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_pathname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pathname",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_cpathname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cpathname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_74_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(ns),
            &_Py_ID(name),
            & const_str_pathname._ascii.ob_base,
            & const_str_cpathname._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(306)
importlib__bootstrap_external_toplevel_consts_74 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 153,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_74_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_74_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_74_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1703,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 216,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_74_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_ID(_fix_up_module),
    .co_qualname = &_Py_ID(_fix_up_module),
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_74_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x04\x73\x2d\x7c\x05\x72\x0d\x7c\x05\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x6e\x1e\x7c\x02\x7c\x03\x6b\x28\x00\x00\x72\x0d\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x6e\x0c\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x05\x73\x20\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\x7c\x04\xac\x03\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x03\x72\x10\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x05\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x09\x00\x7c\x05\x7c\x00\x64\x02\x3c\x00\x00\x00\x7c\x04\x7c\x00\x64\x01\x3c\x00\x00\x00\x7c\x02\x7c\x00\x64\x04\x3c\x00\x00\x00\x7c\x03\x7c\x00\x64\x05\x3c\x00\x00\x00\x79\x00\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[96];
    }
importlib__bootstrap_external_toplevel_consts_75_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 95,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x73\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x66\x69\x6c\x65\x2d\x62\x61\x73\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6c\x6f\x61\x64\x65\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x45\x61\x63\x68\x20\x69\x74\x65\x6d\x20\x69\x73\x20\x61\x20\x74\x75\x70\x6c\x65\x20\x28\x6c\x6f\x61\x64\x65\x72\x2c\x20\x73\x75\x66\x66\x69\x78\x65\x73\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_75_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_75_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_extension_suffixes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "extension_suffixes",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_75_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_ExtensionFileLoader._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_extension_suffixes._ascii.ob_base,
            & const_str_SourceFileLoader._ascii.ob_base,
            & const_str_SOURCE_SUFFIXES._ascii.ob_base,
            & const_str_SourcelessFileLoader._ascii.ob_base,
            & const_str_BYTECODE_SUFFIXES._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[58];
    }
importlib__bootstrap_external_toplevel_consts_75_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 57,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x12\x25\xa4\x64\xd7\x26\x3d\xd1\x26\x3d\xd3\x26\x3f\xd0\x11\x3f\x80\x4a\xdc\x0d\x1d\x9c\x7f\xd0\x0d\x2e\x80\x46\xdc\x0f\x23\xd4\x25\x36\xd0\x0f\x36\x80\x48\xd8\x0c\x16\x98\x06\xa0\x08\xd0\x0b\x29\xd0\x04\x29",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_extensions = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "extensions",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_bytecode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytecode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_75_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_extensions._ascii.ob_base,
            &_Py_ID(source),
            & const_str_bytecode._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(112)
importlib__bootstrap_external_toplevel_consts_75 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 56,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_75_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_75_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1728,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 217,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_75_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__get_supported_file_loaders._ascii.ob_base,
    .co_qualname = & const_str__get_supported_file_loaders._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_75_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x66\x02\x7d\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x7d\x01\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x7d\x02\x7c\x00\x7c\x01\x7c\x02\x67\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_76_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(_bootstrap),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str__set_bootstrap_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_set_bootstrap_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
importlib__bootstrap_external_toplevel_consts_76_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x11\x22\x81\x4a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__bootstrap_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_bootstrap_module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_76_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__bootstrap_module._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(8)
importlib__bootstrap_external_toplevel_consts_76 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 4,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_76_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1739,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 218,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_76_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__set_bootstrap_module._ascii.ob_base,
    .co_qualname = & const_str__set_bootstrap_module._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_76_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x61\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
importlib__bootstrap_external_toplevel_consts_77_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Install the path-based import components.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_77_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_77_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_77_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str__set_bootstrap_module._ascii.ob_base,
            & const_str__get_supported_file_loaders._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_path_hooks._ascii.ob_base,
            &_Py_ID(extend),
            & const_str_FileFinder._ascii.ob_base,
            & const_str_path_hook._ascii.ob_base,
            & const_str_meta_path._ascii.ob_base,
            &_Py_ID(append),
            & const_str_PathFinder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[72];
    }
importlib__bootstrap_external_toplevel_consts_77_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 71,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x04\x19\xd0\x1a\x2b\xd4\x04\x2c\xdc\x18\x33\xd3\x18\x35\xd0\x04\x15\xdc\x04\x07\x87\x4e\x81\x4e\xd7\x04\x19\xd1\x04\x19\x9c\x3a\xd7\x1b\x2f\xd1\x1b\x2f\xd0\x31\x42\xd0\x1b\x43\xd0\x1a\x44\xd4\x04\x45\xdc\x04\x07\x87\x4d\x81\x4d\xd7\x04\x18\xd1\x04\x18\x9c\x1a\xd5\x04\x24",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_supported_loaders = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "supported_loaders",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_77_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__bootstrap_module._ascii.ob_base,
            & const_str_supported_loaders._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(212)
importlib__bootstrap_external_toplevel_consts_77 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 106,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_77_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_consts_77_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1744,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 219,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_77_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = & const_str__install._ascii.ob_base,
    .co_qualname = & const_str__install._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_77_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x8e\x00\x67\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_78 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_int_438.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_80 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            Py_None,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_81 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts_82 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[83];
        }_object;
    }
importlib__bootstrap_external_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 83,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & const_str_win32._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            & importlib__bootstrap_external_toplevel_consts_6.ob_base.ob_base,
            &_Py_STR(empty),
            (PyObject *)&_Py_SINGLETON(strings).ascii[58],
            & importlib__bootstrap_external_toplevel_consts_9._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_10._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_11.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_12.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_13.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_14.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_15.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_16.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_17.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_18.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_19.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_20.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_21.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_22.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_23.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_24.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_25.ob_base.ob_base,
            & const_int_3531.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            &_Py_ID(little),
            & importlib__bootstrap_external_toplevel_consts_29.ob_base.ob_base,
            & const_str___pycache__._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_31._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_12._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_33._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_34._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_35._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_36.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_37.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_38.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_39.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_40.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_41.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_42.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_43.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_44.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_45.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_46.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_47.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_48.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_33_consts_5._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_50.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_51.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52.ob_base.ob_base,
            & const_str_WindowsRegistryFinder._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_54.ob_base.ob_base,
            & const_str__LoaderBasics._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_56.ob_base.ob_base,
            & const_str_SourceLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_58.ob_base.ob_base,
            & const_str_FileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_60.ob_base.ob_base,
            & const_str_SourceFileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_62.ob_base.ob_base,
            & const_str_SourcelessFileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_64.ob_base.ob_base,
            & const_str_ExtensionFileLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_66.ob_base.ob_base,
            & const_str__NamespacePath._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68.ob_base.ob_base,
            & const_str_NamespaceLoader._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_70.ob_base.ob_base,
            & const_str_PathFinder._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72.ob_base.ob_base,
            & const_str_FileFinder._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_74.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_75.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_76.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_77.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_78._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_80._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_81._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_82._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_all = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "all",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str__pathseps_with_colon = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_pathseps_with_colon",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
const_str__CASE_INSENSITIVE_PLATFORMS_BYTES_KEY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CASE_INSENSITIVE_PLATFORMS_BYTES_KEY",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str___code__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__code__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str_DEBUG_BYTECODE_SUFFIXES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "DEBUG_BYTECODE_SUFFIXES",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
const_str_OPTIMIZED_BYTECODE_SUFFIXES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "OPTIMIZED_BYTECODE_SUFFIXES",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__NamespaceLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_NamespaceLoader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[87];
        }_object;
    }
importlib__bootstrap_external_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 87,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            &_Py_ID(_bootstrap),
            & const_str__imp._ascii.ob_base,
            &_Py_ID(_io),
            & const_str_sys._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            & const_str_marshal._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str__MS_WINDOWS._ascii.ob_base,
            &_Py_ID(nt),
            & const_str__os._ascii.ob_base,
            & const_str_winreg._ascii.ob_base,
            &_Py_ID(posix),
            & const_str_path_separators._ascii.ob_base,
            & const_str_all._ascii.ob_base,
            & const_str_path_sep._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_path_sep_tuple._ascii.ob_base,
            &_Py_ID(join),
            & const_str__pathseps_with_colon._ascii.ob_base,
            & const_str__CASE_INSENSITIVE_PLATFORMS_STR_KEY._ascii.ob_base,
            & const_str__CASE_INSENSITIVE_PLATFORMS_BYTES_KEY._ascii.ob_base,
            & const_str__CASE_INSENSITIVE_PLATFORMS._ascii.ob_base,
            & const_str__make_relax_case._ascii.ob_base,
            & const_str__relax_case._ascii.ob_base,
            & const_str__pack_uint32._ascii.ob_base,
            & const_str__unpack_uint32._ascii.ob_base,
            & const_str__unpack_uint16._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str__path_split._ascii.ob_base,
            & const_str__path_stat._ascii.ob_base,
            & const_str__path_is_mode_type._ascii.ob_base,
            & const_str__path_isfile._ascii.ob_base,
            & const_str__path_isdir._ascii.ob_base,
            & const_str__path_isabs._ascii.ob_base,
            & const_str__path_abspath._ascii.ob_base,
            & const_str__write_atomic._ascii.ob_base,
            &_Py_ID(type),
            & const_str___code__._ascii.ob_base,
            & const_str__code_type._ascii.ob_base,
            & const_str_to_bytes._ascii.ob_base,
            & const_str_MAGIC_NUMBER._ascii.ob_base,
            & const_str_int._ascii.ob_base,
            & const_str_from_bytes._ascii.ob_base,
            & const_str__RAW_MAGIC_NUMBER._ascii.ob_base,
            & const_str__PYCACHE._ascii.ob_base,
            & const_str__OPT._ascii.ob_base,
            & const_str_SOURCE_SUFFIXES._ascii.ob_base,
            &_Py_ID(append),
            & const_str_extension_suffixes._ascii.ob_base,
            & const_str_EXTENSION_SUFFIXES._ascii.ob_base,
            & const_str_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_DEBUG_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_OPTIMIZED_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_cache_from_source._ascii.ob_base,
            & const_str_source_from_cache._ascii.ob_base,
            &_Py_ID(_get_sourcefile),
            & const_str__get_cached._ascii.ob_base,
            & const_str__calc_mode._ascii.ob_base,
            & const_str__check_name._ascii.ob_base,
            & const_str__classify_pyc._ascii.ob_base,
            & const_str__validate_timestamp_pyc._ascii.ob_base,
            & const_str__validate_hash_pyc._ascii.ob_base,
            & const_str__compile_bytecode._ascii.ob_base,
            & const_str__code_to_timestamp_pyc._ascii.ob_base,
            & const_str__code_to_hash_pyc._ascii.ob_base,
            & const_str_decode_source._ascii.ob_base,
            &_Py_ID(object),
            & const_str__POPULATE._ascii.ob_base,
            & const_str_spec_from_file_location._ascii.ob_base,
            & const_str__bless_my_loader._ascii.ob_base,
            & const_str_WindowsRegistryFinder._ascii.ob_base,
            & const_str__LoaderBasics._ascii.ob_base,
            & const_str_SourceLoader._ascii.ob_base,
            & const_str_FileLoader._ascii.ob_base,
            & const_str_SourceFileLoader._ascii.ob_base,
            & const_str_SourcelessFileLoader._ascii.ob_base,
            & const_str_ExtensionFileLoader._ascii.ob_base,
            & const_str__NamespacePath._ascii.ob_base,
            & const_str_NamespaceLoader._ascii.ob_base,
            & const_str__NamespaceLoader._ascii.ob_base,
            & const_str_PathFinder._ascii.ob_base,
            & const_str_FileFinder._ascii.ob_base,
            &_Py_ID(_fix_up_module),
            & const_str__get_supported_file_loaders._ascii.ob_base,
            & const_str__set_bootstrap_module._ascii.ob_base,
            & const_str__install._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[711];
    }
importlib__bootstrap_external_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 710,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x07\x01\x04\xf0\x2c\x00\x0e\x12\x80\x0a\xf3\x06\x00\x01\x0c\xdb\x00\x0a\xdb\x00\x0a\xdb\x00\x10\xdb\x00\x0e\xf0\x06\x00\x10\x13\x8f\x7c\x89\x7c\x98\x77\xd1\x0f\x26\x80\x0b\xd9\x03\x0e\xdb\x04\x14\xdc\x04\x11\xe3\x04\x17\xf1\x06\x00\x04\x0f\xd8\x17\x1b\x98\x53\x90\x6b\x81\x4f\xe0\x17\x1a\x90\x65\x80\x4f\xe1\x07\x0a\xd1\x0a\x34\xa1\x4f\xd3\x0a\x34\xd4\x07\x34\xd0\x00\x34\xd0\x07\x34\xd8\x0b\x1a\x98\x31\xd1\x0b\x1d\x80\x08\xd9\x11\x16\x90\x7f\xd3\x11\x27\x80\x0e\xd8\x12\x14\x97\x27\x91\x27\x98\x2f\xd3\x12\x2a\x80\x0f\xd9\x29\x38\xd3\x17\x39\xa9\x1f\xa0\x41\x98\x21\x98\x41\x98\x33\x9a\x07\xa8\x1f\xd1\x17\x39\xd0\x00\x14\xf0\x08\x00\x27\x2d\xd0\x00\x23\xd8\x28\x3a\xd0\x00\x25\xd8\x20\x45\xd8\x22\x45\xf1\x03\x01\x21\x46\x01\xd0\x00\x1b\xf2\x08\x0e\x01\x17\xf1\x20\x00\x0f\x1f\xd3\x0e\x20\x80\x0b\xf2\x06\x02\x01\x37\xf2\x0a\x03\x01\x2a\xf2\x0a\x03\x01\x2a\xf1\x0c\x00\x04\x0f\xf3\x02\x1b\x05\x2a\xf2\x3c\x03\x05\x3f\xf2\x0c\x05\x01\x22\xf2\x10\x07\x01\x1a\xf2\x14\x06\x01\x32\xf2\x12\x02\x01\x2e\xf2\x0a\x04\x01\x2e\xf1\x0e\x00\x04\x0f\xf3\x02\x05\x05\x52\x01\xf2\x10\x02\x05\x30\xf2\x0a\x07\x01\x14\xf3\x14\x17\x01\x0e\xf1\x34\x00\x0e\x12\x90\x2d\xd7\x12\x28\xd1\x12\x28\xd3\x0d\x29\x80\x0a\xf0\x70\x07\x00\x11\x15\x8f\x7f\x89\x7f\x98\x71\xa0\x28\xd3\x0f\x2b\xa8\x67\xd1\x0f\x35\x80\x0c\xe0\x14\x17\x97\x4e\x91\x4e\xa0\x3c\xb0\x18\xd3\x14\x3a\xd0\x00\x11\xe0\x0b\x18\x80\x08\xd8\x07\x0d\x80\x04\xe0\x13\x18\x90\x27\x80\x0f\xd9\x03\x0e\xd8\x04\x13\xd7\x04\x1a\xd1\x04\x1a\x98\x36\xd4\x04\x22\xe0\x15\x2c\x90\x54\xd7\x15\x2c\xd1\x15\x2c\xd3\x15\x2e\xd0\x00\x12\xe0\x15\x1b\x90\x48\xd0\x00\x11\xe0\x38\x49\xd0\x00\x49\xd0\x00\x17\xd0\x1a\x35\xf0\x04\x43\x01\x01\x30\xc0\x14\xf4\x00\x43\x01\x01\x30\xf2\x4c\x02\x25\x01\x40\x01\xf2\x50\x01\x10\x01\x47\x01\xf2\x26\x09\x01\x14\xf2\x18\x09\x01\x10\xf2\x18\x1c\x01\x1f\xf2\x3e\x1e\x01\x11\xf2\x42\x01\x19\x01\x4c\x01\xf2\x38\x15\x01\x0a\xf3\x30\x0a\x01\x39\xf3\x1a\x07\x01\x10\xf3\x14\x08\x01\x10\xf2\x16\x09\x01\x44\x01\xf1\x1c\x00\x0d\x13\x8b\x48\x80\x09\xf0\x06\x43\x01\x01\x10\xb8\x34\xd8\x37\x40\xf4\x03\x43\x01\x01\x10\xf2\x4c\x02\x2d\x01\x17\xf7\x64\x01\x30\x01\x1c\xf1\x00\x30\x01\x1c\xf7\x66\x01\x1b\x01\x3c\xf1\x00\x1b\x01\x3c\xf4\x3c\x4e\x02\x01\x1b\x90\x3d\xf4\x00\x4e\x02\x01\x1b\xf7\x62\x04\x2f\x01\x20\xf1\x00\x2f\x01\x20\xf4\x64\x01\x2a\x01\x2d\x90\x7a\xa0\x3c\xf4\x00\x2a\x01\x2d\xf4\x5a\x01\x16\x01\x14\x98\x3a\xa0\x7d\xf4\x00\x16\x01\x14\xf4\x32\x32\x01\x19\x98\x2a\xa0\x6d\xf4\x00\x32\x01\x19\xf7\x6a\x01\x41\x01\x01\x20\xf1\x00\x41\x01\x01\x20\xf7\x4e\x02\x21\x01\x2b\xf1\x00\x21\x01\x2b\xf0\x4a\x01\x00\x14\x23\xd0\x00\x10\xf7\x0a\x79\x01\x01\x46\x01\xf1\x00\x79\x01\x01\x46\x01\xf7\x78\x03\x48\x02\x01\x2c\xf1\x00\x48\x02\x01\x2c\xf3\x5a\x04\x16\x01\x0d\xf2\x32\x08\x01\x2a\xf2\x16\x02\x01\x23\xf3\x0a\x05\x01\x25\xf9\xf2\x7d\x34\x00\x18\x3a",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[7];
    }
importlib__bootstrap_external_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 6,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x34\x0c\x47\x37\x04",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib__bootstrap_external_toplevel_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
        },
    },
};
static
    struct _PyCode_DEF(1016)
importlib__bootstrap_external_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 508,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_external_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 220,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib__bootstrap_external_toplevel_consts_6_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & importlib__bootstrap_external_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x61\x01\x64\x02\x64\x01\x6c\x02\x5a\x02\x64\x02\x64\x01\x6c\x03\x5a\x03\x64\x02\x64\x01\x6c\x04\x5a\x04\x64\x02\x64\x01\x6c\x05\x5a\x05\x64\x02\x64\x01\x6c\x06\x5a\x06\x65\x04\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x6b\x28\x00\x00\x5a\x08\x65\x08\x72\x09\x64\x02\x64\x01\x6c\x09\x5a\x0a\x64\x02\x64\x01\x6c\x0b\x5a\x0b\x6e\x04\x64\x02\x64\x01\x6c\x0c\x5a\x0a\x65\x08\x72\x05\x64\x04\x64\x05\x67\x02\x5a\x0d\x6e\x03\x64\x05\x67\x01\x5a\x0d\x02\x00\x65\x0e\x64\x06\x84\x00\x65\x0d\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x02\x4a\x00\x82\x01\x65\x0d\x64\x02\x19\x00\x00\x00\x5a\x0f\x02\x00\x65\x10\x65\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x11\x64\x07\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x0d\x65\x0d\x44\x00\x8f\x00\x63\x02\x68\x00\x63\x02\x5d\x07\x00\x00\x7d\x00\x64\x08\x7c\x00\x9b\x00\x9d\x02\x92\x02\x8c\x09\x04\x00\x63\x02\x7d\x00\x5a\x13\x64\x09\x5a\x14\x64\x0a\x5a\x15\x65\x15\x65\x14\x7a\x00\x00\x00\x5a\x16\x64\x0b\x84\x00\x5a\x17\x02\x00\x65\x17\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x18\x64\x0c\x84\x00\x5a\x19\x64\x0d\x84\x00\x5a\x1a\x64\x0e\x84\x00\x5a\x1b\x65\x08\x72\x04\x64\x0f\x84\x00\x5a\x1c\x6e\x03\x64\x10\x84\x00\x5a\x1c\x64\x11\x84\x00\x5a\x1d\x64\x12\x84\x00\x5a\x1e\x64\x13\x84\x00\x5a\x1f\x64\x14\x84\x00\x5a\x20\x64\x15\x84\x00\x5a\x21\x65\x08\x72\x04\x64\x16\x84\x00\x5a\x22\x6e\x03\x64\x17\x84\x00\x5a\x22\x64\x18\x84\x00\x5a\x23\x64\x4e\x64\x19\x84\x01\x5a\x24\x02\x00\x65\x25\x65\x24\x6a\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x27\x64\x1a\x6a\x51\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x1b\x64\x1c\xab\x02\x00\x00\x00\x00\x00\x00\x64\x1d\x7a\x00\x00\x00\x5a\x29\x65\x2a\x6a\x57\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x29\x64\x1c\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x2c\x64\x1e\x5a\x2d\x64\x1f\x5a\x2e\x64\x20\x67\x01\x5a\x2f\x65\x08\x72\x11\x65\x2f\x6a\x61\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x21\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x02\x6a\x62\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x32\x64\x22\x67\x01\x5a\x33\x65\x33\x78\x01\x5a\x34\x5a\x35\x64\x4f\x64\x01\x64\x23\x9c\x01\x64\x24\x84\x03\x5a\x36\x64\x25\x84\x00\x5a\x37\x64\x26\x84\x00\x5a\x38\x64\x27\x84\x00\x5a\x39\x64\x28\x84\x00\x5a\x3a\x64\x29\x84\x00\x5a\x3b\x64\x2a\x84\x00\x5a\x3c\x64\x2b\x84\x00\x5a\x3d\x64\x2c\x84\x00\x5a\x3e\x64\x50\x64\x2d\x84\x01\x5a\x3f\x64\x51\x64\x2e\x84\x01\x5a\x40\x64\x52\x64\x2f\x84\x01\x5a\x41\x64\x30\x84\x00\x5a\x42\x02\x00\x65\x43\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x44\x64\x4f\x64\x01\x65\x44\x64\x31\x9c\x02\x64\x32\x84\x03\x5a\x45\x64\x33\x84\x00\x5a\x46\x02\x00\x47\x00\x64\x34\x84\x00\x64\x35\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x47\x02\x00\x47\x00\x64\x36\x84\x00\x64\x37\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x48\x02\x00\x47\x00\x64\x38\x84\x00\x64\x39\x65\x48\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x49\x02\x00\x47\x00\x64\x3a\x84\x00\x64\x3b\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x4a\x02\x00\x47\x00\x64\x3c\x84\x00\x64\x3d\x65\x4a\x65\x49\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x4b\x02\x00\x47\x00\x64\x3e\x84\x00\x64\x3f\x65\x4a\x65\x48\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x4c\x02\x00\x47\x00\x64\x40\x84\x00\x64\x41\x65\x4a\x65\x48\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x4d\x02\x00\x47\x00\x64\x42\x84\x00\x64\x43\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x4e\x02\x00\x47\x00\x64\x44\x84\x00\x64\x45\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x4f\x65\x4f\x5a\x50\x02\x00\x47\x00\x64\x46\x84\x00\x64\x47\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x51\x02\x00\x47\x00\x64\x48\x84\x00\x64\x49\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x52\x64\x4f\x64\x4a\x84\x01\x5a\x53\x64\x4b\x84\x00\x5a\x54\x64\x4c\x84\x00\x5a\x55\x64\x4d\x84\x00\x5a\x56\x79\x01\x63\x02\x01\x00\x63\x02\x7d\x00\x77\x00",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_importlib__bootstrap_external_toplevel(void)
{
    return Py_NewRef((PyObject *) &importlib__bootstrap_external_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[593];
    }
zipimport_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 592,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x66\x6f\x72\x20\x69\x6d\x70\x6f\x72\x74\x69\x6e\x67\x20\x50\x79\x74\x68\x6f\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x66\x72\x6f\x6d\x20\x5a\x69\x70\x20\x61\x72\x63\x68\x69\x76\x65\x73\x2e\x0a\x0a\x54\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x65\x78\x70\x6f\x72\x74\x73\x20\x74\x68\x72\x65\x65\x20\x6f\x62\x6a\x65\x63\x74\x73\x3a\x0a\x2d\x20\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x65\x72\x3a\x20\x61\x20\x63\x6c\x61\x73\x73\x3b\x20\x69\x74\x73\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x6f\x72\x20\x74\x61\x6b\x65\x73\x20\x61\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x61\x20\x5a\x69\x70\x20\x61\x72\x63\x68\x69\x76\x65\x2e\x0a\x2d\x20\x5a\x69\x70\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x3a\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x72\x61\x69\x73\x65\x64\x20\x62\x79\x20\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x65\x72\x20\x6f\x62\x6a\x65\x63\x74\x73\x2e\x20\x49\x74\x27\x73\x20\x61\x0a\x20\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x20\x6f\x66\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x2c\x20\x73\x6f\x20\x69\x74\x20\x63\x61\x6e\x20\x62\x65\x20\x63\x61\x75\x67\x68\x74\x20\x61\x73\x20\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x2c\x20\x74\x6f\x6f\x2e\x0a\x2d\x20\x5f\x7a\x69\x70\x5f\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x5f\x63\x61\x63\x68\x65\x3a\x20\x61\x20\x64\x69\x63\x74\x2c\x20\x6d\x61\x70\x70\x69\x6e\x67\x20\x61\x72\x63\x68\x69\x76\x65\x20\x70\x61\x74\x68\x73\x20\x74\x6f\x20\x7a\x69\x70\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x0a\x20\x20\x69\x6e\x66\x6f\x20\x64\x69\x63\x74\x73\x2c\x20\x61\x73\x20\x75\x73\x65\x64\x20\x69\x6e\x20\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x65\x72\x2e\x5f\x66\x69\x6c\x65\x73\x2e\x0a\x0a\x49\x74\x20\x69\x73\x20\x75\x73\x75\x61\x6c\x6c\x79\x20\x6e\x6f\x74\x20\x6e\x65\x65\x64\x65\x64\x20\x74\x6f\x20\x75\x73\x65\x20\x74\x68\x65\x20\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x20\x6d\x6f\x64\x75\x6c\x65\x20\x65\x78\x70\x6c\x69\x63\x69\x74\x6c\x79\x3b\x20\x69\x74\x20\x69\x73\x0a\x75\x73\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x62\x75\x69\x6c\x74\x69\x6e\x20\x69\x6d\x70\x6f\x72\x74\x20\x6d\x65\x63\x68\x61\x6e\x69\x73\x6d\x20\x66\x6f\x72\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x20\x69\x74\x65\x6d\x73\x20\x74\x68\x61\x74\x20\x61\x72\x65\x20\x70\x61\x74\x68\x73\x0a\x74\x6f\x20\x5a\x69\x70\x20\x61\x72\x63\x68\x69\x76\x65\x73\x2e\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__unpack_uint16._ascii.ob_base,
            & const_str__unpack_uint32._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_ZipImportError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ZipImportError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_zipimporter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_ZipImportError._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
zipimport_toplevel_consts_7_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen zipimport>",
};
static
    struct _PyCode_DEF(12)
zipimport_toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & zipimport_toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 34,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 221,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_ZipImportError._ascii.ob_base,
    .co_qualname = & const_str_ZipImportError._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
zipimport_toplevel_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x50\x4b\x05\x06",
};
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_65535 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 32767, 1 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_65535 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 65535 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[512];
    }
zipimport_toplevel_consts_11_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 511,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x65\x72\x28\x61\x72\x63\x68\x69\x76\x65\x70\x61\x74\x68\x29\x20\x2d\x3e\x20\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x65\x72\x20\x6f\x62\x6a\x65\x63\x74\x0a\x0a\x20\x20\x20\x20\x43\x72\x65\x61\x74\x65\x20\x61\x20\x6e\x65\x77\x20\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x20\x27\x61\x72\x63\x68\x69\x76\x65\x70\x61\x74\x68\x27\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x70\x61\x74\x68\x20\x74\x6f\x0a\x20\x20\x20\x20\x61\x20\x7a\x69\x70\x66\x69\x6c\x65\x2c\x20\x6f\x72\x20\x74\x6f\x20\x61\x20\x73\x70\x65\x63\x69\x66\x69\x63\x20\x70\x61\x74\x68\x20\x69\x6e\x73\x69\x64\x65\x20\x61\x20\x7a\x69\x70\x66\x69\x6c\x65\x2e\x20\x46\x6f\x72\x20\x65\x78\x61\x6d\x70\x6c\x65\x2c\x20\x69\x74\x20\x63\x61\x6e\x20\x62\x65\x0a\x20\x20\x20\x20\x27\x2f\x74\x6d\x70\x2f\x6d\x79\x69\x6d\x70\x6f\x72\x74\x2e\x7a\x69\x70\x27\x2c\x20\x6f\x72\x20\x27\x2f\x74\x6d\x70\x2f\x6d\x79\x69\x6d\x70\x6f\x72\x74\x2e\x7a\x69\x70\x2f\x6d\x79\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x27\x2c\x20\x69\x66\x20\x6d\x79\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x61\x0a\x20\x20\x20\x20\x76\x61\x6c\x69\x64\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x6e\x73\x69\x64\x65\x20\x74\x68\x65\x20\x61\x72\x63\x68\x69\x76\x65\x2e\x0a\x0a\x20\x20\x20\x20\x27\x5a\x69\x70\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x20\x69\x66\x20\x27\x61\x72\x63\x68\x69\x76\x65\x70\x61\x74\x68\x27\x20\x64\x6f\x65\x73\x6e\x27\x74\x20\x70\x6f\x69\x6e\x74\x20\x74\x6f\x20\x61\x20\x76\x61\x6c\x69\x64\x20\x5a\x69\x70\x0a\x20\x20\x20\x20\x61\x72\x63\x68\x69\x76\x65\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x27\x61\x72\x63\x68\x69\x76\x65\x27\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x20\x6f\x66\x20\x7a\x69\x70\x69\x6d\x70\x6f\x72\x74\x65\x72\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x20\x6f\x66\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x7a\x69\x70\x66\x69\x6c\x65\x20\x74\x61\x72\x67\x65\x74\x65\x64\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
zipimport_toplevel_consts_11_consts_2_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "expected str, not ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
zipimport_toplevel_consts_11_consts_2_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "archive path is empty",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
zipimport_toplevel_consts_11_consts_2_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "not a Zip file",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
zipimport_toplevel_consts_11_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            Py_None,
            & zipimport_toplevel_consts_11_consts_2_consts_1._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_2_consts_2._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
            & const_int_61440.ob_base,
            & const_int_32768.ob_base,
            & zipimport_toplevel_consts_11_consts_2_consts_6._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_alt_path_sep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "alt_path_sep",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str__zip_directory_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_zip_directory_cache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__read_directory = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_read_directory",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__files = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_files",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_archive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "archive",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "prefix",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
zipimport_toplevel_consts_11_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            & const_str_ZipImportError._ascii.ob_base,
            & const_str_alt_path_sep._ascii.ob_base,
            &_Py_ID(replace),
            & const_str_path_sep._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__path_stat._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str__path_split._ascii.ob_base,
            &_Py_ID(append),
            & const_str__zip_directory_cache._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str__read_directory._ascii.ob_base,
            & const_str__files._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str_prefix._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
zipimport_toplevel_consts_11_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[346];
    }
zipimport_toplevel_consts_11_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 345,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x24\xa4\x03\xd4\x0f\x24\xdc\x12\x1b\xd0\x1e\x30\xb4\x14\xb0\x64\xb3\x1a\xb0\x0e\xd0\x1c\x3f\xd3\x12\x40\xd0\x0c\x40\xd9\x0f\x13\xdc\x12\x20\xd0\x21\x38\xb8\x74\xd4\x12\x44\xd0\x0c\x44\xdd\x0b\x17\xd8\x13\x17\x97\x3c\x91\x3c\xa4\x0c\xac\x68\xd3\x13\x37\x88\x44\xe0\x11\x13\x88\x06\xd8\x0e\x12\xf0\x02\x0f\x0d\x16\xdc\x15\x28\xd7\x15\x33\xd1\x15\x33\xb0\x44\xd3\x15\x39\x90\x02\xf0\x16\x00\x15\x17\x97\x4a\x91\x4a\xa0\x18\xd1\x14\x29\xa8\x68\xd2\x13\x36\xe4\x1a\x28\xd0\x29\x39\xc0\x04\xd4\x1a\x45\xd0\x14\x45\xd8\x10\x15\xf0\x04\x04\x09\x2f\xdc\x14\x28\xa8\x14\xd1\x14\x2e\x88\x45\xf0\x08\x00\x17\x1c\x88\x04\x8c\x0b\xd8\x17\x1b\x88\x04\x8c\x0c\xe4\x16\x29\xd7\x16\x34\xd1\x16\x34\xb0\x66\xb9\x54\xb8\x72\xb8\x54\xb1\x6c\xd0\x16\x43\x88\x04\x8c\x0b\xd8\x0b\x0f\x8f\x3b\x8a\x3b\xd8\x0c\x10\x8f\x4b\x8a\x4b\x9c\x38\xd1\x0c\x23\x8e\x4b\xf0\x03\x00\x0c\x17\xf8\xf4\x31\x00\x15\x1c\x9c\x5a\xd0\x13\x28\xf2\x00\x07\x0d\x28\xf4\x06\x00\x25\x38\xd7\x24\x43\xd1\x24\x43\xc0\x44\xd3\x24\x49\xd1\x10\x21\x90\x07\x98\x18\xd8\x13\x1a\x98\x64\x92\x3f\xdc\x1a\x28\xd0\x29\x39\xc0\x04\xd4\x1a\x45\xd0\x14\x45\xd8\x17\x1e\x90\x04\xd8\x10\x16\x97\x0d\x91\x0d\x98\x68\xd6\x10\x27\xf0\x0f\x07\x0d\x28\xfa\xf0\x07\x00\x0f\x13\xf8\xf4\x28\x00\x10\x18\xf2\x00\x02\x09\x2f\xdc\x14\x23\xa0\x44\xd3\x14\x29\x88\x45\xd8\x29\x2e\xd4\x0c\x20\xa0\x14\xd3\x0c\x26\xf0\x05\x02\x09\x2f\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[38];
    }
zipimport_toplevel_consts_11_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 37,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x1b\x15\x43\x2c\x00\xc2\x11\x09\x44\x3f\x00\xc3\x2c\x41\x0c\x44\x3b\x03\xc4\x3a\x01\x44\x3b\x03\xc4\x3f\x1d\x45\x1f\x03\xc5\x1e\x01\x45\x1f\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_basename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "basename",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_files = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "files",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
zipimport_toplevel_consts_11_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            & const_str_prefix._ascii.ob_base,
            & const_str_st._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
            & const_str_basename._ascii.ob_base,
            & const_str_files._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(708)
zipimport_toplevel_consts_11_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 354,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_2_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_11_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 64,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 222,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & zipimport_toplevel_consts_11_consts_2_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x17\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x73\x0d\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x72\x1a\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x01\x67\x00\x7d\x02\x09\x00\x09\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7a\x01\x00\x00\x64\x05\x6b\x37\x00\x00\x72\x0d\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x09\x00\x09\x00\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x06\x7c\x06\x7c\x00\x5f\x12\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x5f\x13\x00\x00\x00\x00\x00\x00\x00\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x28\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x00\x64\x00\x64\x07\x85\x03\x19\x00\x00\x00\x8e\x00\x7c\x00\x5f\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x1a\x7c\x00\x78\x01\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x0d\x00\x00\x63\x02\x5f\x15\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x79\x00\x23\x00\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x40\x01\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x04\x7d\x05\x7c\x04\x7c\x01\x6b\x28\x00\x00\x72\x0d\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x04\x7d\x01\x7c\x02\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x6e\x04\x77\x00\x78\x03\x59\x00\x77\x01\x8c\xe5\x23\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x17\x01\x00\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x59\x00\x8c\xc4\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[108];
    }
zipimport_toplevel_consts_11_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 107,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x72\x65\x61\x74\x65\x20\x61\x20\x4d\x6f\x64\x75\x6c\x65\x53\x70\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x4e\x6f\x6e\x65\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_11_consts_4_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_is_package._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_11_consts_4_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_loader._ascii.ob_base,
            & const_str_is_package._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
zipimport_toplevel_consts_11_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_4_consts_0._ascii.ob_base,
            Py_None,
            & zipimport_toplevel_consts_11_consts_4_consts_2._object.ob_base.ob_base,
            Py_True,
            & zipimport_toplevel_consts_11_consts_4_consts_4._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__get_module_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_module_info",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__get_module_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_module_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__is_dir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_is_dir",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
zipimport_toplevel_consts_11_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str__get_module_info._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str_spec_from_loader._ascii.ob_base,
            & const_str__get_module_path._ascii.ob_base,
            & const_str__is_dir._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            & const_str_path_sep._ascii.ob_base,
            & const_str_ModuleSpec._ascii.ob_base,
            & const_str_submodule_search_locations._ascii.ob_base,
            &_Py_ID(append),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
zipimport_toplevel_consts_11_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.find_spec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[143];
    }
zipimport_toplevel_consts_11_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 142,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x17\x27\xa0\x74\xa8\x58\xd3\x16\x36\x88\x0b\xd8\x0b\x16\xd0\x0b\x22\xdc\x13\x1d\xd7\x13\x2e\xd1\x13\x2e\xa8\x78\xb8\x14\xc8\x2b\xd4\x13\x56\xd0\x0c\x56\xf4\x0e\x00\x17\x27\xa0\x74\xa8\x58\xd3\x16\x36\x88\x47\xdc\x0f\x16\x90\x74\x98\x57\xd4\x0f\x25\xf0\x08\x00\x1b\x1f\x9f\x2c\x99\x2c\x98\x1e\xac\x08\xa0\x7a\xb0\x27\xb0\x19\xd0\x17\x3b\x90\x04\xdc\x17\x21\xd7\x17\x2c\xd1\x17\x2c\xb0\x28\xc0\x34\xd8\x38\x3c\xf4\x03\x01\x18\x3e\x90\x04\xe0\x10\x14\xd7\x10\x2f\xd1\x10\x2f\xd7\x10\x36\xd1\x10\x36\xb0\x74\xd4\x10\x3c\xd8\x17\x1b\x90\x0b\xe0\x17\x1b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_module_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "module_info",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_modpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "modpath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
zipimport_toplevel_consts_11_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(target),
            & const_str_module_info._ascii.ob_base,
            & const_str_modpath._ascii.ob_base,
            &_Py_ID(path),
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(278)
zipimport_toplevel_consts_11_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 139,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_4_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 104,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 223,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_find_spec._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_4_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x81\x18\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x7c\x03\xac\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x72\x4b\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x7c\x04\x9b\x00\x9d\x03\x7d\x05\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x03\xac\x04\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x06\x53\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[167];
    }
zipimport_toplevel_consts_11_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 166,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x67\x65\x74\x5f\x63\x6f\x64\x65\x28\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x29\x20\x2d\x3e\x20\x63\x6f\x64\x65\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x20\x52\x61\x69\x73\x65\x20\x5a\x69\x70\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x63\x6f\x75\x6c\x64\x6e\x27\x74\x20\x62\x65\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_11_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__get_module_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_module_code",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_11_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__get_module_code._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
zipimport_toplevel_consts_11_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.get_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
zipimport_toplevel_consts_11_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x24\x34\xb0\x44\xb8\x28\xd3\x23\x43\xd1\x08\x20\x88\x04\x88\x69\x98\x17\xd8\x0f\x13\x88\x0b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_ispackage = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ispackage",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
zipimport_toplevel_consts_11_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(code),
            & const_str_ispackage._ascii.ob_base,
            & const_str_modpath._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(38)
zipimport_toplevel_consts_11_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_5_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 131,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 224,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_get_code._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_5_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x02\x7d\x03\x7d\x04\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[155];
    }
zipimport_toplevel_consts_11_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 154,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x67\x65\x74\x5f\x64\x61\x74\x61\x28\x70\x61\x74\x68\x6e\x61\x6d\x65\x29\x20\x2d\x3e\x20\x73\x74\x72\x69\x6e\x67\x20\x77\x69\x74\x68\x20\x66\x69\x6c\x65\x20\x64\x61\x74\x61\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x64\x61\x74\x61\x20\x61\x73\x73\x6f\x63\x69\x61\x74\x65\x64\x20\x77\x69\x74\x68\x20\x27\x70\x61\x74\x68\x6e\x61\x6d\x65\x27\x2e\x20\x52\x61\x69\x73\x65\x20\x4f\x53\x45\x72\x72\x6f\x72\x20\x69\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x77\x61\x73\x6e\x27\x74\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
zipimport_toplevel_consts_11_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_6_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_STR(empty),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__get_data = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_data",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
zipimport_toplevel_consts_11_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_alt_path_sep._ascii.ob_base,
            &_Py_ID(replace),
            & const_str_path_sep._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            &_Py_ID(len),
            & const_str__files._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str__get_data._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
zipimport_toplevel_consts_11_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.get_data",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[151];
    }
zipimport_toplevel_consts_11_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 150,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf5\x0c\x00\x0c\x18\xd8\x17\x1f\xd7\x17\x27\xd1\x17\x27\xac\x0c\xb4\x68\xd3\x17\x3f\x88\x48\xe0\x0e\x16\x88\x03\xd8\x0b\x13\xd7\x0b\x1e\xd1\x0b\x1e\x98\x74\x9f\x7c\x99\x7c\xac\x68\xd1\x1f\x36\xd4\x0b\x37\xd8\x12\x1a\x9c\x33\x98\x74\x9f\x7c\x99\x7c\xac\x68\xd1\x1f\x36\xd3\x1b\x37\xd0\x1b\x38\xd0\x12\x39\x88\x43\xf0\x04\x03\x09\x26\xd8\x18\x1c\x9f\x0b\x99\x0b\xa0\x43\xd1\x18\x28\x88\x49\xf4\x06\x00\x10\x19\x98\x14\x9f\x1c\x99\x1c\xa0\x79\xd3\x0f\x31\xd0\x08\x31\xf8\xf4\x05\x00\x10\x18\xf2\x00\x01\x09\x26\xdc\x12\x19\x98\x21\x98\x52\xa0\x13\xd3\x12\x25\xd0\x0c\x25\xf0\x03\x01\x09\x26\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
zipimport_toplevel_consts_11_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x25\x0f\x42\x0a\x00\xc2\x0a\x17\x42\x21\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_toc_entry = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "toc_entry",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
zipimport_toplevel_consts_11_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_pathname._ascii.ob_base,
            &_Py_ID(key),
            & const_str_toc_entry._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(328)
zipimport_toplevel_consts_11_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 164,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_6_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_11_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 141,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 225,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_get_data._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_6_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x1a\x7c\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x7d\x02\x7c\x01\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x1f\x7c\x01\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x1a\x00\x7d\x02\x09\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x19\x00\x00\x00\x7d\x03\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0e\x01\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[166];
    }
zipimport_toplevel_consts_11_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 165,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x67\x65\x74\x5f\x66\x69\x6c\x65\x6e\x61\x6d\x65\x28\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x29\x20\x2d\x3e\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x73\x74\x72\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x5a\x69\x70\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x66\x20\x69\x74\x20\x63\x6f\x75\x6c\x64\x6e\x27\x74\x20\x62\x65\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_11_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_7_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
zipimport_toplevel_consts_11_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.get_filename",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
zipimport_toplevel_consts_11_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x24\x34\xb0\x44\xb8\x28\xd3\x23\x43\xd1\x08\x20\x88\x04\x88\x69\x98\x17\xd8\x0f\x16\x88\x0e",
};
static
    struct _PyCode_DEF(38)
zipimport_toplevel_consts_11_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_7_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 162,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 226,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_get_filename._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_7_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x02\x7d\x03\x7d\x04\x7c\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[254];
    }
zipimport_toplevel_consts_11_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 253,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x67\x65\x74\x5f\x73\x6f\x75\x72\x63\x65\x28\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x29\x20\x2d\x3e\x20\x73\x6f\x75\x72\x63\x65\x20\x73\x74\x72\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x73\x6f\x75\x72\x63\x65\x20\x63\x6f\x64\x65\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x20\x52\x61\x69\x73\x65\x20\x5a\x69\x70\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x63\x6f\x75\x6c\x64\x6e\x27\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2c\x20\x72\x65\x74\x75\x72\x6e\x20\x4e\x6f\x6e\x65\x20\x69\x66\x20\x74\x68\x65\x20\x61\x72\x63\x68\x69\x76\x65\x20\x64\x6f\x65\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6f\x6e\x74\x61\x69\x6e\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2c\x20\x62\x75\x74\x20\x68\x61\x73\x20\x6e\x6f\x20\x73\x6f\x75\x72\x63\x65\x20\x66\x6f\x72\x20\x69\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
zipimport_toplevel_consts_11_consts_8_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "can't find module ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
zipimport_toplevel_consts_11_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_8_consts_0._ascii.ob_base,
            Py_None,
            & zipimport_toplevel_consts_11_consts_8_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_11._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_12._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
zipimport_toplevel_consts_11_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str__get_module_info._ascii.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
            & const_str__get_module_path._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str__files._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str__get_data._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            &_Py_ID(decode),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
zipimport_toplevel_consts_11_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.get_source",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[154];
    }
zipimport_toplevel_consts_11_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 153,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x0e\x1e\x98\x64\xa0\x48\xd3\x0d\x2d\x88\x02\xd8\x0b\x0d\x88\x3a\xdc\x12\x20\xd0\x23\x35\xb0\x68\xb0\x5c\xd0\x21\x42\xc8\x18\xd4\x12\x52\xd0\x0c\x52\xe4\x0f\x1f\xa0\x04\xa0\x68\xd3\x0f\x2f\x88\x04\xd9\x0b\x0d\xdc\x17\x2a\xd7\x17\x35\xd1\x17\x35\xb0\x64\xb8\x4d\xd3\x17\x4a\x89\x48\xe0\x1a\x1e\x98\x16\x98\x73\x90\x7c\x88\x48\xf0\x04\x04\x09\x18\xd8\x18\x1c\x9f\x0b\x99\x0b\xa0\x48\xd1\x18\x2d\x88\x49\xf4\x08\x00\x10\x19\x98\x14\x9f\x1c\x99\x1c\xa0\x79\xd3\x0f\x31\xd7\x0f\x38\xd1\x0f\x38\xd3\x0f\x3a\xd0\x08\x3a\xf8\xf4\x07\x00\x10\x18\xf2\x00\x02\x09\x18\xe1\x13\x17\xf0\x05\x02\x09\x18\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
zipimport_toplevel_consts_11_consts_8_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x0a\x0f\x41\x3d\x00\xc1\x3d\x09\x42\x09\x03\xc2\x08\x01\x42\x09\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_mi = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mi",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_fullpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fullpath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
zipimport_toplevel_consts_11_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            & const_str_mi._ascii.ob_base,
            &_Py_ID(path),
            & const_str_fullpath._ascii.ob_base,
            & const_str_toc_entry._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(280)
zipimport_toplevel_consts_11_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 140,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_8_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_11_consts_8_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 174,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 227,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = &_Py_ID(get_source),
    .co_qualname = & zipimport_toplevel_consts_11_consts_8_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x80\x10\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x01\x9b\x02\x9d\x02\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x02\x72\x17\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x6e\x05\x7c\x03\x9b\x00\x64\x05\x9d\x02\x7d\x04\x09\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x19\x00\x00\x00\x7d\x05\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[172];
    }
zipimport_toplevel_consts_11_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 171,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x69\x73\x5f\x70\x61\x63\x6b\x61\x67\x65\x28\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x29\x20\x2d\x3e\x20\x62\x6f\x6f\x6c\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x54\x72\x75\x65\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x62\x79\x20\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x20\x69\x73\x20\x61\x20\x70\x61\x63\x6b\x61\x67\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x20\x5a\x69\x70\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x63\x6f\x75\x6c\x64\x6e\x27\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_11_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_9_consts_0._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_8_consts_2._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_11_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__get_module_info._ascii.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
zipimport_toplevel_consts_11_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.is_package",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[48];
    }
zipimport_toplevel_consts_11_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 47,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x0e\x1e\x98\x64\xa0\x48\xd3\x0d\x2d\x88\x02\xd8\x0b\x0d\x88\x3a\xdc\x12\x20\xd0\x23\x35\xb0\x68\xb0\x5c\xd0\x21\x42\xc8\x18\xd4\x12\x52\xd0\x0c\x52\xd8\x0f\x11\x88\x09",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_11_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            & const_str_mi._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(66)
zipimport_toplevel_consts_11_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 33,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_9_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 200,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 228,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_is_package._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_9_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x80\x10\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x01\x9b\x02\x9d\x02\x7c\x01\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[321];
    }
zipimport_toplevel_consts_11_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 320,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x6c\x6f\x61\x64\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x29\x20\x2d\x3e\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4c\x6f\x61\x64\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x62\x79\x20\x27\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x27\x2e\x20\x27\x66\x75\x6c\x6c\x6e\x61\x6d\x65\x27\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x75\x6c\x6c\x79\x20\x71\x75\x61\x6c\x69\x66\x69\x65\x64\x20\x28\x64\x6f\x74\x74\x65\x64\x29\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x2e\x20\x49\x74\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6d\x6f\x64\x75\x6c\x65\x2c\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x73\x20\x5a\x69\x70\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x69\x74\x20\x63\x6f\x75\x6c\x64\x20\x6e\x6f\x74\x20\x62\x65\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x44\x65\x70\x72\x65\x63\x61\x74\x65\x64\x20\x73\x69\x6e\x63\x65\x20\x50\x79\x74\x68\x6f\x6e\x20\x33\x2e\x31\x30\x2e\x20\x55\x73\x65\x20\x65\x78\x65\x63\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x29\x20\x69\x6e\x73\x74\x65\x61\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[115];
    }
zipimport_toplevel_consts_11_consts_10_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 114,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimport.zipimporter.load_module() is deprecated and slated for removal in Python 3.12; use exec_module() instead",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
zipimport_toplevel_consts_11_consts_10_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Loaded module ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
zipimport_toplevel_consts_11_consts_10_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " not found in sys.modules",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
zipimport_toplevel_consts_11_consts_10_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "import {} # loaded from Zip {}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
zipimport_toplevel_consts_11_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_10_consts_0._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_10_consts_1._ascii.ob_base,
            &_Py_ID(__builtins__),
            & zipimport_toplevel_consts_11_consts_10_consts_3._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_10_consts_4._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_10_consts_5._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__module_type = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_module_type",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[24];
        }_object;
    }
zipimport_toplevel_consts_11_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 24,
        },
        .ob_item = {
            & const_str__warnings._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            & const_str_DeprecationWarning._ascii.ob_base,
            & const_str__get_module_code._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(get),
            &_Py_ID(isinstance),
            & const_str__module_type._ascii.ob_base,
            &_Py_ID(__loader__),
            & const_str__get_module_path._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            &_Py_ID(__path__),
            & const_str_hasattr._ascii.ob_base,
            &_Py_ID(__builtins__),
            &_Py_ID(_fix_up_module),
            &_Py_ID(__dict__),
            & const_str_exec._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
zipimport_toplevel_consts_11_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.load_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[323];
    }
zipimport_toplevel_consts_11_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 322,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x12\x01\x10\x43\x01\x88\x03\xe4\x08\x11\x8f\x0e\x89\x0e\x90\x73\xd4\x1c\x2e\xd4\x08\x2f\xdc\x23\x33\xb0\x44\xb8\x28\xd3\x23\x43\xd1\x08\x20\x88\x04\x88\x69\x98\x17\xdc\x0e\x11\x8f\x6b\x89\x6b\x8f\x6f\x89\x6f\x98\x68\xd3\x0e\x27\x88\x03\xd8\x0b\x0e\x88\x3b\x9c\x6a\xa8\x13\xac\x6c\xd4\x1e\x3b\xdc\x12\x1e\x98\x78\xd3\x12\x28\x88\x43\xd8\x24\x27\x8c\x43\x8f\x4b\x89\x4b\x98\x08\xd1\x0c\x21\xd8\x19\x1d\x88\x03\x8c\x0e\xf0\x04\x0e\x09\x12\xd9\x0f\x18\xf4\x06\x00\x18\x28\xa8\x04\xa8\x68\xd3\x17\x37\x90\x04\xdc\x1b\x2e\xd7\x1b\x39\xd1\x1b\x39\xb8\x24\xbf\x2c\xb9\x2c\xc8\x04\xd3\x1b\x4d\x90\x08\xd8\x20\x28\x98\x7a\x90\x03\x94\x0c\xe4\x13\x1a\x98\x33\xa0\x0e\xd4\x13\x2f\xdc\x23\x2f\x90\x03\xd4\x10\x20\xdc\x0c\x1f\xd7\x0c\x2e\xd1\x0c\x2e\xa8\x73\xaf\x7c\xa9\x7c\xb8\x58\xc0\x77\xd4\x0c\x4f\xdc\x0c\x10\x90\x14\x90\x73\x97\x7c\x91\x7c\xd4\x0c\x24\xf0\x0a\x03\x09\x56\x01\xdc\x12\x15\x97\x2b\x91\x2b\x98\x68\xd1\x12\x27\x88\x43\xf4\x06\x00\x09\x13\xd7\x08\x23\xd1\x08\x23\xd0\x24\x44\xc0\x68\xd0\x50\x57\xd4\x08\x58\xd8\x0f\x12\x88\x0a\xf8\xf0\x13\x02\x09\x12\xdc\x10\x13\x97\x0b\x91\x0b\x98\x48\xd0\x10\x25\xd8\x0c\x11\xfb\xf4\x08\x00\x10\x18\xf2\x00\x01\x09\x56\x01\xdc\x12\x1d\xa0\x0e\xa8\x78\xa8\x6c\xd0\x3a\x53\xd0\x1e\x54\xd3\x12\x55\xd0\x0c\x55\xf0\x03\x01\x09\x56\x01\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
zipimport_toplevel_consts_11_consts_10_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x04\x42\x04\x44\x35\x00\xc4\x09\x13\x45\x0c\x00\xc4\x35\x14\x45\x09\x03\xc5\x0c\x19\x45\x25\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
zipimport_toplevel_consts_11_consts_10_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(msg),
            &_Py_ID(code),
            & const_str_ispackage._ascii.ob_base,
            & const_str_modpath._ascii.ob_base,
            &_Py_ID(mod),
            &_Py_ID(path),
            & const_str_fullpath._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(720)
zipimport_toplevel_consts_11_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 360,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_10_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_11_consts_10_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 213,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 229,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_10_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_load_module._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_10_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x03\x7d\x04\x7d\x05\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x81\x10\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x1e\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x7c\x00\x7c\x06\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x09\x00\x7c\x04\x72\x34\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x07\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x67\x01\x7c\x06\x5f\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x0b\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x5f\x10\x00\x00\x00\x00\x00\x00\x00\x00\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x6a\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x05\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x74\x27\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x06\x6a\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x06\x74\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x2e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x01\x7c\x05\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x06\x53\x00\x23\x00\x01\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3d\x00\x82\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x28\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x10\x01\x00\x74\x2b\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x01\x9b\x02\x64\x04\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
zipimport_toplevel_consts_11_consts_11_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the ResourceReader for a module in a zip file.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_ZipReader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ZipReader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_11_consts_11_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_ZipReader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_11_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_11_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & zipimport_toplevel_consts_11_consts_11_consts_2._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_11_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_58_consts_8_names_0._ascii.ob_base,
            & const_str_ZipReader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
zipimport_toplevel_consts_11_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.get_resource_reader",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
zipimport_toplevel_consts_11_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe5\x08\x2f\xe1\x0f\x18\x98\x14\x98\x78\xd3\x0f\x28\xd0\x08\x28",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_11_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            & const_str_ZipReader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(32)
zipimport_toplevel_consts_11_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_11_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 256,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 230,
    .co_localsplusnames = & zipimport_toplevel_consts_11_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_get_resource_reader._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_11_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x6d\x01\x7d\x02\x01\x00\x02\x00\x7c\x02\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
zipimport_toplevel_consts_11_consts_12_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Reload the file data of the archive path.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_11_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & zipimport_toplevel_consts_11_consts_12_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
zipimport_toplevel_consts_11_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__read_directory._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            & const_str__files._ascii.ob_base,
            & const_str__zip_directory_cache._ascii.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
zipimport_toplevel_consts_11_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.invalidate_caches",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[85];
    }
zipimport_toplevel_consts_11_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 84,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x05\x09\x1d\xdc\x1a\x29\xa8\x24\xaf\x2c\xa9\x2c\xd3\x1a\x37\x88\x44\x8c\x4b\xd8\x31\x35\xb7\x1b\xb1\x1b\xd4\x0c\x20\xa0\x14\xa7\x1c\xa1\x1c\xd2\x0c\x2e\xf8\xdc\x0f\x1d\xf2\x00\x02\x09\x1d\xdc\x0c\x20\xd7\x0c\x24\xd1\x0c\x24\xa0\x54\xa7\x5c\xa1\x5c\xb0\x34\xd4\x0c\x38\xd8\x1a\x1c\x88\x44\x8e\x4b\xf0\x05\x02\x09\x1d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
zipimport_toplevel_consts_11_consts_12_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x37\x3a\x00\xba\x30\x41\x2d\x03\xc1\x2c\x01\x41\x2d\x03",
};
static
    struct _PyCode_DEF(224)
zipimport_toplevel_consts_11_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 112,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_12_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_11_consts_12_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 263,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 231,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_invalidate_caches._ascii.ob_base,
    .co_qualname = & zipimport_toplevel_consts_11_consts_12_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x79\x01\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x2a\x01\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x69\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
zipimport_toplevel_consts_11_consts_13_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<zipimporter object \"",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
zipimport_toplevel_consts_11_consts_13_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\">",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_11_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & zipimport_toplevel_consts_11_consts_13_consts_1._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_13_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_11_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_archive._ascii.ob_base,
            & const_str_path_sep._ascii.ob_base,
            & const_str_prefix._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
zipimport_toplevel_consts_11_consts_13_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimporter.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
zipimport_toplevel_consts_11_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x11\x26\xa0\x74\xa7\x7c\xa1\x7c\xa0\x6e\xb4\x58\xb0\x4a\xb8\x74\xbf\x7b\xb9\x7b\xb8\x6d\xc8\x32\xd0\x0f\x4e\xd0\x08\x4e",
};
static
    struct _PyCode_DEF(70)
zipimport_toplevel_consts_11_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts_13_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 273,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 232,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & zipimport_toplevel_consts_11_consts_13_qualname._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x02\x9d\x05\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
zipimport_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & const_str_zipimporter._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_1._ascii.ob_base,
            & zipimport_toplevel_consts_11_consts_2.ob_base.ob_base,
            Py_None,
            & zipimport_toplevel_consts_11_consts_4.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_5.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_6.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_7.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_8.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_9.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_10.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_11.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_12.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_13.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
zipimport_toplevel_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            & const_str_find_spec._ascii.ob_base,
            & const_str_get_code._ascii.ob_base,
            & const_str_get_data._ascii.ob_base,
            & const_str_get_filename._ascii.ob_base,
            &_Py_ID(get_source),
            & const_str_is_package._ascii.ob_base,
            & const_str_load_module._ascii.ob_base,
            & const_str_get_resource_reader._ascii.ob_base,
            & const_str_invalidate_caches._ascii.ob_base,
            &_Py_ID(__repr__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[66];
    }
zipimport_toplevel_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 65,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x0c\x05\x08\xf2\x22\x25\x05\x24\xf3\x50\x01\x19\x05\x1c\xf2\x36\x07\x05\x14\xf2\x14\x11\x05\x32\xf2\x2a\x09\x05\x17\xf2\x18\x16\x05\x3b\xf2\x34\x09\x05\x12\xf2\x1a\x28\x05\x13\xf2\x56\x01\x04\x05\x29\xf2\x0e\x07\x05\x1d\xf3\x14\x01\x05\x4f\x01",
};
static
    struct _PyCode_DEF(84)
zipimport_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 42,
    },
    .co_consts = & zipimport_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 46,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 233,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str_zipimporter._ascii.ob_base,
    .co_qualname = & const_str_zipimporter._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x0e\x64\x04\x84\x01\x5a\x05\x64\x05\x84\x00\x5a\x06\x64\x06\x84\x00\x5a\x07\x64\x07\x84\x00\x5a\x08\x64\x08\x84\x00\x5a\x09\x64\x09\x84\x00\x5a\x0a\x64\x0a\x84\x00\x5a\x0b\x64\x0b\x84\x00\x5a\x0c\x64\x0c\x84\x00\x5a\x0d\x64\x0d\x84\x00\x5a\x0e\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
zipimport_toplevel_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__init__.pyc",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_16 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_34._ascii.ob_base,
            Py_True,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_17 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_12._ascii.ob_base,
            Py_False,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_prefix._ascii.ob_base,
            & const_str_rpartition._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
zipimport_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\x8f\x3b\x89\x3b\x98\x18\xd7\x19\x2c\xd1\x19\x2c\xa8\x53\xd3\x19\x31\xb0\x21\xd1\x19\x34\xd1\x0b\x34\xd0\x04\x34",
};
static
    struct _PyCode_DEF(68)
zipimport_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 34,
    },
    .co_consts = & zipimport_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 291,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 234,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__get_module_path._ascii.ob_base,
    .co_qualname = & const_str__get_module_path._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\x7a\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_19_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_path_sep._ascii.ob_base,
            & const_str__files._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
zipimport_toplevel_consts_19_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x0f\x13\x94\x58\x89\x6f\x80\x47\xe0\x0b\x12\x90\x64\x97\x6b\x91\x6b\xd0\x0b\x21\xd0\x04\x21",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_dirpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dirpath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_19_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            & const_str_dirpath._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(48)
zipimport_toplevel_consts_19 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 24,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_19_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 295,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 235,
    .co_localsplusnames = & zipimport_toplevel_consts_19_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__is_dir._ascii.ob_base,
    .co_qualname = & const_str__is_dir._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_19_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x7d\x02\x7c\x02\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__zip_searchorder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_zip_searchorder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__get_module_path._ascii.ob_base,
            & const_str__zip_searchorder._ascii.ob_base,
            & const_str__files._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[64];
    }
zipimport_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 63,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0b\x1b\x98\x44\xa0\x28\xd3\x0b\x2b\x80\x44\xdf\x29\x39\xd1\x08\x25\x88\x06\x90\x0a\x98\x49\xd8\x13\x17\x98\x26\x91\x3d\x88\x08\xd8\x0b\x13\x90\x74\x97\x7b\x91\x7b\xd2\x0b\x22\xd8\x13\x1c\xd2\x0c\x1c\xf0\x07\x00\x2a\x3a\xf0\x08\x00\x0c\x10",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_isbytecode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isbytecode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
zipimport_toplevel_consts_20_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            & const_str_suffix._ascii.ob_base,
            & const_str_isbytecode._ascii.ob_base,
            & const_str_ispackage._ascii.ob_base,
            & const_str_fullpath._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(104)
zipimport_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 52,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 304,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 236,
    .co_localsplusnames = & zipimport_toplevel_consts_20_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__get_module_info._ascii.ob_base,
    .co_qualname = & const_str__get_module_info._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x1d\x00\x00\x5c\x03\x00\x00\x7d\x03\x7d\x04\x7d\x05\x7c\x02\x7c\x03\x7a\x00\x00\x00\x7d\x06\x7c\x06\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x73\x01\x8c\x1b\x7c\x05\x63\x02\x01\x00\x53\x00\x04\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
zipimport_toplevel_consts_21_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "can't open Zip file: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
zipimport_toplevel_consts_21_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "can't read Zip file: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
zipimport_toplevel_consts_21_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "not a Zip file: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
zipimport_toplevel_consts_21_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "corrupt Zip file: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
zipimport_toplevel_consts_21_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bad central directory size: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
zipimport_toplevel_consts_21_consts_13 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bad central directory offset: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
zipimport_toplevel_consts_21_consts_14 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bad central directory size or offset: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
zipimport_toplevel_consts_21_consts_16 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "EOF read where not expected",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
zipimport_toplevel_consts_21_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x50\x4b\x01\x02",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
zipimport_toplevel_consts_21_consts_27 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bad local header offset: ",
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_2048 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 2048 },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_ascii = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ascii",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
zipimport_toplevel_consts_21_consts_33 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimport: found {} names in {!r}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[34];
        }_object;
    }
zipimport_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 34,
        },
        .ob_item = {
            Py_None,
            & zipimport_toplevel_consts_21_consts_1._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & zipimport_toplevel_consts_21_consts_4._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 4],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & zipimport_toplevel_consts_21_consts_7._ascii.ob_base,
            & zipimport_toplevel_consts_21_consts_8._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 12],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 20],
            & zipimport_toplevel_consts_21_consts_12._ascii.ob_base,
            & zipimport_toplevel_consts_21_consts_13._ascii.ob_base,
            & zipimport_toplevel_consts_21_consts_14._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 46],
            & zipimport_toplevel_consts_21_consts_16._ascii.ob_base,
            & zipimport_toplevel_consts_21_consts_17.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 10],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 14],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 24],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 28],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 30],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 32],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 34],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 42],
            & zipimport_toplevel_consts_21_consts_27._ascii.ob_base,
            & const_int_2048.ob_base,
            & const_str_ascii._ascii.ob_base,
            &_Py_ID(latin1),
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & zipimport_toplevel_consts_21_consts_33._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_END_CENTRAL_DIR_SIZE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "END_CENTRAL_DIR_SIZE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_STRING_END_ARCHIVE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "STRING_END_ARCHIVE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_MAX_COMMENT_LEN = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MAX_COMMENT_LEN",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_UnicodeDecodeError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UnicodeDecodeError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_cp437_table = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cp437_table",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[26];
        }_object;
    }
zipimport_toplevel_consts_21_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 26,
        },
        .ob_item = {
            &_Py_ID(_io),
            & const_str_open_code._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
            &_Py_ID(tell),
            &_Py_ID(seek),
            & const_str_END_CENTRAL_DIR_SIZE._ascii.ob_base,
            &_Py_ID(read),
            &_Py_ID(len),
            & const_str_STRING_END_ARCHIVE._ascii.ob_base,
            & const_str_max._ascii.ob_base,
            & const_str_MAX_COMMENT_LEN._ascii.ob_base,
            & const_str_rfind._ascii.ob_base,
            & const_str__unpack_uint32._ascii.ob_base,
            & const_str_EOFError._ascii.ob_base,
            & const_str__unpack_uint16._ascii.ob_base,
            &_Py_ID(decode),
            & const_str_UnicodeDecodeError._ascii.ob_base,
            &_Py_ID(translate),
            & const_str_cp437_table._ascii.ob_base,
            &_Py_ID(replace),
            & const_str_path_sep._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__path_join._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[1494];
    }
zipimport_toplevel_consts_21_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 1493,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x03\x05\x50\x01\xdc\x0d\x10\x8f\x5d\x89\x5d\x98\x37\xd3\x0d\x23\x88\x02\xf2\x08\x00\x0a\x0c\xf0\x08\x00\x18\x1a\x97\x77\x91\x77\x93\x79\x88\x0c\xf0\x02\x6e\x01\x09\x22\xf0\x02\x05\x0d\x58\x01\xd8\x10\x12\x97\x07\x91\x07\xd4\x19\x2d\xd0\x18\x2d\xa8\x71\xd4\x10\x31\xd8\x22\x24\xa7\x27\xa1\x27\xa3\x29\x90\x0f\xd8\x19\x1b\x9f\x17\x99\x17\xd4\x21\x35\xd3\x19\x36\x90\x06\xf4\x06\x00\x10\x13\x90\x36\x8b\x7b\xd4\x1e\x32\xd2\x0f\x32\xdc\x16\x24\xd0\x27\x3c\xb8\x57\xb8\x4b\xd0\x25\x48\xc8\x77\xd4\x16\x57\xd0\x10\x57\xd8\x0f\x15\x90\x62\x90\x71\x88\x7a\xd4\x1d\x2f\xd2\x0f\x2f\xf0\x06\x05\x11\x37\xd8\x14\x16\x97\x47\x91\x47\x98\x41\x98\x71\x94\x4d\xd8\x20\x22\xa7\x07\xa1\x07\xa3\x09\x90\x49\xf4\x08\x00\x25\x28\xa8\x09\xb4\x4f\xd1\x28\x43\xdc\x28\x3c\xf1\x03\x01\x29\x3d\xd8\x3e\x3f\xf3\x03\x01\x25\x41\x01\xd0\x10\x21\xf0\x04\x05\x11\x37\xd8\x14\x16\x97\x47\x91\x47\xd0\x1c\x2d\xd4\x14\x2e\xd8\x1b\x1d\x9f\x37\x99\x37\x9b\x39\x90\x44\xf0\x08\x00\x17\x1b\x97\x6a\x91\x6a\xd4\x21\x33\xd3\x16\x34\x90\x03\xd8\x13\x16\x98\x11\x92\x37\xdc\x1a\x28\xd0\x2b\x3b\xb8\x47\xb8\x3b\xd0\x29\x47\xd8\x2e\x35\xf4\x03\x01\x1b\x37\xf0\x00\x01\x15\x37\xe0\x19\x1d\x98\x63\xa0\x23\xd4\x26\x3a\xd1\x22\x3a\xd0\x19\x3b\x90\x06\xdc\x13\x16\x90\x76\x93\x3b\xd4\x22\x36\xd2\x13\x36\xdc\x1a\x28\xd0\x2b\x3d\xb8\x67\xb8\x5b\xd0\x29\x49\xd8\x2e\x35\xf4\x03\x01\x1b\x37\xf0\x00\x01\x15\x37\xe0\x22\x2b\xac\x63\xb0\x24\xab\x69\xd1\x22\x37\xb8\x23\xd1\x22\x3d\x90\x0f\xe4\x1a\x28\xa8\x16\xb0\x02\xb0\x32\xa8\x1d\xd3\x1a\x37\x88\x4b\xdc\x1c\x2a\xa8\x36\xb0\x22\xb0\x52\xa8\x3d\xd3\x1c\x39\x88\x4d\xd8\x0f\x1e\xa0\x1b\xd2\x0f\x2c\xdc\x16\x24\xd0\x27\x43\xc0\x47\xc0\x3b\xd0\x25\x4f\xd0\x56\x5d\xd4\x16\x5e\xd0\x10\x5e\xd8\x0f\x1e\xa0\x1d\xd2\x0f\x2e\xdc\x16\x24\xd0\x27\x45\xc0\x67\xc0\x5b\xd0\x25\x51\xd0\x58\x5f\xd4\x16\x60\xd0\x10\x60\xd8\x0c\x1b\x98\x7b\xd1\x0c\x2a\x88\x4f\xd8\x19\x28\xa8\x3d\xd1\x19\x38\x88\x4a\xd8\x0f\x19\x98\x41\x8a\x7e\xdc\x16\x24\xd0\x27\x4d\xc8\x67\xc8\x5b\xd0\x25\x59\xd0\x60\x67\xd4\x16\x68\xd0\x10\x68\xe0\x14\x16\x88\x45\xe0\x14\x15\x88\x45\xf0\x02\x03\x0d\x58\x01\xd8\x10\x12\x97\x07\x91\x07\x98\x0f\xd4\x10\x28\xf0\x06\x00\x13\x17\xd8\x19\x1b\x9f\x17\x99\x17\xa0\x12\x9b\x1b\x90\x06\xdc\x13\x16\x90\x76\x93\x3b\xa0\x11\x92\x3f\xdc\x1a\x22\xd0\x23\x40\xd3\x1a\x41\xd0\x14\x41\xe0\x13\x19\x98\x22\x98\x31\x90\x3a\xa0\x1d\xd2\x13\x2e\xd9\x14\x19\xdc\x13\x16\x90\x76\x93\x3b\xa0\x22\xd2\x13\x24\xdc\x1a\x22\xd0\x23\x40\xd3\x1a\x41\xd0\x14\x41\xdc\x18\x26\xa0\x76\xa8\x61\xb0\x02\xa0\x7c\xd3\x18\x34\x90\x05\xdc\x1b\x29\xa8\x26\xb0\x12\xb0\x42\xa8\x2d\xd3\x1b\x38\x90\x08\xdc\x17\x25\xa0\x66\xa8\x52\xb0\x02\xa0\x6d\xd3\x17\x34\x90\x04\xdc\x17\x25\xa0\x66\xa8\x52\xb0\x02\xa0\x6d\xd3\x17\x34\x90\x04\xdc\x16\x24\xa0\x56\xa8\x42\xa8\x72\xa0\x5d\xd3\x16\x33\x90\x03\xdc\x1c\x2a\xa8\x36\xb0\x22\xb0\x52\xa8\x3d\xd3\x1c\x39\x90\x09\xdc\x1c\x2a\xa8\x36\xb0\x22\xb0\x52\xa8\x3d\xd3\x1c\x39\x90\x09\xdc\x1c\x2a\xa8\x36\xb0\x22\xb0\x52\xa8\x3d\xd3\x1c\x39\x90\x09\xdc\x1d\x2b\xa8\x46\xb0\x32\xb0\x62\xa8\x4d\xd3\x1d\x3a\x90\x0a\xdc\x1f\x2d\xa8\x66\xb0\x52\xb8\x02\xa8\x6d\xd3\x1f\x3c\x90\x0c\xdc\x1e\x2c\xa8\x56\xb0\x42\xb0\x72\xa8\x5d\xd3\x1e\x3b\x90\x0b\xd8\x1e\x27\xa8\x2a\xd1\x1e\x34\xb0\x7c\xd1\x1e\x43\x90\x0b\xd8\x13\x1e\xa0\x1d\xd2\x13\x2e\xdc\x1a\x28\xd0\x2b\x44\xc0\x57\xc0\x4b\xd0\x29\x50\xd0\x57\x5e\xd4\x1a\x5f\xd0\x14\x5f\xd8\x10\x1b\x98\x7a\xd1\x10\x29\x90\x0b\xf0\x04\x03\x11\x5c\x01\xd8\x1b\x1d\x9f\x37\x99\x37\xa0\x39\xd3\x1b\x2d\x90\x44\xf4\x06\x00\x14\x17\x90\x74\x93\x39\xa0\x09\xd2\x13\x29\xdc\x1a\x28\xd0\x2b\x40\xc0\x17\xc0\x0b\xd0\x29\x4c\xd0\x53\x5a\xd4\x1a\x5b\xd0\x14\x5b\xf0\x08\x04\x11\x5c\x01\xdc\x17\x1a\x98\x32\x9f\x37\x99\x37\xa0\x3b\xb0\x19\xd1\x23\x3a\xd3\x1b\x3b\xd3\x17\x3c\xc0\x0b\xc8\x69\xd1\x40\x57\xd2\x17\x57\xdc\x1e\x2c\xd0\x2f\x44\xc0\x57\xc0\x4b\xd0\x2d\x50\xd0\x57\x5e\xd4\x1e\x5f\xd0\x18\x5f\xf0\x03\x00\x18\x58\x01\xf0\x0a\x00\x14\x19\x98\x35\x92\x3d\xe0\x1b\x1f\x9f\x3b\x99\x3b\x9b\x3d\x91\x44\xf0\x06\x03\x15\x4c\x01\xd8\x1f\x23\x9f\x7b\x99\x7b\xa8\x37\xd3\x1f\x33\x98\x04\xf0\x08\x00\x18\x1c\x97\x7c\x91\x7c\xa0\x43\xac\x18\xd3\x17\x32\x90\x04\xdc\x17\x2a\xd7\x17\x35\xd1\x17\x35\xb0\x67\xb8\x74\xd3\x17\x44\x90\x04\xd8\x15\x19\x98\x38\xa0\x59\xb0\x09\xb8\x3b\xc8\x04\xc8\x64\xd0\x54\x57\xd0\x14\x58\x90\x01\xd8\x1e\x1f\x90\x05\x90\x64\x91\x0b\xd8\x10\x15\x98\x11\x91\x0a\x90\x05\xf1\x6d\x01\x00\x13\x17\xf0\x0c\x00\x15\x1a\xf0\x64\x01\x00\x0d\x0f\x8f\x47\x89\x47\x90\x4c\xd5\x0c\x21\xf7\x67\x03\x00\x0a\x0c\xf4\x68\x03\x00\x05\x0f\xd7\x04\x1f\xd1\x04\x1f\xd0\x20\x43\xc0\x55\xc8\x47\xd4\x04\x54\xd8\x0b\x10\x80\x4c\xf8\xf4\x71\x03\x00\x0c\x13\xf2\x00\x01\x05\x50\x01\xdc\x0e\x1c\xd0\x1f\x34\xb0\x57\xb0\x4b\xd0\x1d\x40\xc0\x77\xd4\x0e\x4f\xd0\x08\x4f\xf0\x03\x01\x05\x50\x01\xfb\xf4\x1a\x00\x14\x1b\xf2\x00\x01\x0d\x58\x01\xdc\x16\x24\xd0\x27\x3c\xb8\x57\xb8\x4b\xd0\x25\x48\xc8\x77\xd4\x16\x57\xd0\x10\x57\xf0\x03\x01\x0d\x58\x01\xfb\xf4\x14\x00\x18\x1f\xf2\x00\x02\x11\x37\xdc\x1a\x28\xd0\x2b\x40\xc0\x17\xc0\x0b\xd0\x29\x4c\xd8\x2e\x35\xf4\x03\x01\x1b\x37\xf0\x00\x01\x15\x37\xf0\x03\x02\x11\x37\xfb\xf4\x10\x00\x18\x1f\xf2\x00\x02\x11\x37\xdc\x1a\x28\xd0\x2b\x40\xc0\x17\xc0\x0b\xd0\x29\x4c\xd8\x2e\x35\xf4\x03\x01\x1b\x37\xf0\x00\x01\x15\x37\xf0\x03\x02\x11\x37\xfb\xf4\x3a\x00\x14\x1b\xf2\x00\x01\x0d\x58\x01\xdc\x16\x24\xd0\x27\x3c\xb8\x57\xb8\x4b\xd0\x25\x48\xc8\x77\xd4\x16\x57\xd0\x10\x57\xf0\x03\x01\x0d\x58\x01\xfb\xf4\x3a\x00\x18\x1f\xf2\x00\x01\x11\x5c\x01\xdc\x1a\x28\xd0\x2b\x40\xc0\x17\xc0\x0b\xd0\x29\x4c\xd0\x53\x5a\xd4\x1a\x5b\xd0\x14\x5b\xf0\x03\x01\x11\x5c\x01\xfb\xf4\x14\x00\x18\x1f\xf2\x00\x01\x11\x5c\x01\xdc\x1a\x28\xd0\x2b\x40\xc0\x17\xc0\x0b\xd0\x29\x4c\xd0\x53\x5a\xd4\x1a\x5b\xd0\x14\x5b\xf0\x03\x01\x11\x5c\x01\xfb\xf4\x14\x00\x1c\x2e\xf2\x00\x01\x15\x4c\x01\xd8\x1f\x23\x9f\x7b\x99\x7b\xa8\x38\xd3\x1f\x34\xd7\x1f\x3e\xd1\x1f\x3e\xbc\x7b\xd3\x1f\x4b\x9b\x04\xf0\x03\x01\x15\x4c\x01\xfb\xf0\x12\x00\x0d\x0f\x8f\x47\x89\x47\x90\x4c\xd5\x0c\x21\xfa\xf7\x67\x03\x00\x0a\x0c\x89\x12\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[223];
    }
zipimport_toplevel_consts_21_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 222,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x4f\x26\x00\x99\x11\x53\x3b\x03\xac\x3c\x50\x03\x02\xc1\x28\x2e\x53\x25\x02\xc2\x17\x22\x50\x20\x02\xc2\x39\x1a\x53\x25\x02\xc3\x14\x21\x50\x3d\x02\xc3\x35\x43\x12\x53\x25\x02\xc7\x08\x11\x51\x1a\x02\xc7\x19\x44\x0a\x53\x25\x02\xcb\x24\x11\x51\x37\x02\xcb\x35\x1e\x53\x25\x02\xcc\x14\x33\x52\x14\x02\xcd\x07\x17\x53\x25\x02\xcd\x1f\x11\x52\x31\x02\xcd\x30\x41\x02\x53\x25\x02\xce\x33\x11\x53\x3b\x03\xcf\x26\x1a\x50\x00\x03\xd0\x03\x1a\x50\x1d\x05\xd0\x1d\x03\x53\x25\x02\xd0\x20\x1a\x50\x3a\x05\xd0\x3a\x03\x53\x25\x02\xd0\x3d\x1a\x51\x17\x05\xd1\x17\x03\x53\x25\x02\xd1\x1a\x1a\x51\x34\x05\xd1\x34\x03\x53\x25\x02\xd1\x37\x1a\x52\x11\x05\xd2\x11\x03\x53\x25\x02\xd2\x14\x1a\x52\x2e\x05\xd2\x2e\x03\x53\x25\x02\xd2\x31\x2d\x53\x22\x05\xd3\x1e\x03\x53\x25\x02\xd3\x21\x01\x53\x22\x05\xd3\x22\x03\x53\x25\x02\xd3\x25\x13\x53\x38\x05\xd3\x38\x03\x53\x3b\x03\xd3\x3b\x05\x54\x05\x07",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_fp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fp",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_start_offset = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "start_offset",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_header_position = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "header_position",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_file_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "file_size",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_max_comment_start = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "max_comment_start",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_header_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "header_size",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_header_offset = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "header_offset",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_arc_offset = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "arc_offset",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_compress = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "compress",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_time = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "time",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_date = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "date",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_crc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "crc",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_data_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "data_size",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_name_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "name_size",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_extra_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "extra_size",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_comment_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "comment_size",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_file_offset = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "file_offset",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[27];
        }_object;
    }
zipimport_toplevel_consts_21_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 27,
        },
        .ob_item = {
            & const_str_archive._ascii.ob_base,
            & const_str_fp._ascii.ob_base,
            & const_str_start_offset._ascii.ob_base,
            & const_str_header_position._ascii.ob_base,
            &_Py_ID(buffer),
            & const_str_file_size._ascii.ob_base,
            & const_str_max_comment_start._ascii.ob_base,
            &_Py_ID(data),
            &_Py_ID(pos),
            & const_str_header_size._ascii.ob_base,
            & const_str_header_offset._ascii.ob_base,
            & const_str_arc_offset._ascii.ob_base,
            & const_str_files._ascii.ob_base,
            &_Py_ID(count),
            &_Py_ID(flags),
            & const_str_compress._ascii.ob_base,
            & const_str_time._ascii.ob_base,
            & const_str_date._ascii.ob_base,
            & const_str_crc._ascii.ob_base,
            & const_str_data_size._ascii.ob_base,
            & const_str_name_size._ascii.ob_base,
            & const_str_extra_size._ascii.ob_base,
            & const_str_comment_size._ascii.ob_base,
            & const_str_file_offset._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(path),
            (PyObject *)&_Py_SINGLETON(strings).ascii[116],
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
zipimport_toplevel_consts_21_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "                           ",
};
static
    struct _PyCode_DEF(2576)
zipimport_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 1288,
    },
    .co_consts = & zipimport_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_21_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 36 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 335,
    .co_nlocalsplus = 27,
    .co_nlocals = 27,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 237,
    .co_localsplusnames = & zipimport_toplevel_consts_21_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & zipimport_toplevel_consts_21_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__read_directory._ascii.ob_base,
    .co_qualname = & const_str__read_directory._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_21_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x35\x00\x01\x00\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x09\x00\x09\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x0b\x00\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x04\x64\x00\x64\x05\x1a\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\xc8\x09\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x0a\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x0a\x00\x00\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x09\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x07\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x64\x06\x6b\x02\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x07\x7c\x08\x7c\x08\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x1a\x00\x7d\x04\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x05\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x0a\x00\x00\x7c\x08\x7a\x00\x00\x00\x7d\x03\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x09\x64\x0a\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x09\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x0a\x64\x0b\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0a\x7c\x03\x7c\x09\x6b\x02\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0c\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x03\x7c\x0a\x6b\x02\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x03\x7c\x09\x7a\x17\x00\x00\x7d\x03\x7c\x03\x7c\x0a\x7a\x0a\x00\x00\x7d\x0b\x7c\x0b\x64\x06\x6b\x02\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0e\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x69\x00\x7d\x0c\x64\x06\x7d\x0d\x09\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0f\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x64\x05\x6b\x02\x00\x00\x72\x0b\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x10\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x04\x64\x00\x64\x05\x1a\x00\x64\x11\x6b\x37\x00\x00\x72\x02\x90\x01\x6e\xa4\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x64\x0f\x6b\x37\x00\x00\x72\x0b\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x10\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x12\x64\x13\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0e\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x13\x64\x09\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0f\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x09\x64\x14\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x10\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x14\x64\x0a\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x11\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x0a\x64\x0b\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x12\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x0b\x64\x15\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x13\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x15\x64\x16\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x16\x64\x17\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x14\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x17\x64\x18\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x15\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x18\x64\x19\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x16\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x1a\x64\x0f\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x17\x7c\x14\x7c\x15\x7a\x00\x00\x00\x7c\x16\x7a\x00\x00\x00\x7d\x09\x7c\x17\x7c\x0a\x6b\x44\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x1b\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x17\x7c\x0b\x7a\x0d\x00\x00\x7d\x17\x09\x00\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x14\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x18\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x18\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x14\x6b\x37\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x09\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x7c\x14\x7a\x0a\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x09\x7c\x14\x7a\x0a\x00\x00\x6b\x37\x00\x00\x72\x10\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x09\x00\x7c\x0e\x64\x1c\x7a\x01\x00\x00\x72\x11\x7c\x18\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x18\x6e\x12\x09\x00\x7c\x18\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x1d\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x18\x7c\x18\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x1f\x74\x2a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x18\x74\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x2e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x18\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x19\x7c\x19\x7c\x0f\x7c\x13\x7c\x05\x7c\x17\x7c\x10\x7c\x11\x7c\x12\x66\x08\x7d\x1a\x7c\x1a\x7c\x0c\x7c\x18\x3c\x00\x00\x00\x7c\x0d\x64\x20\x7a\x0d\x00\x00\x7d\x0d\x90\x01\x8c\xd8\x09\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x31\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x21\x7f\x0d\x7c\x00\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x7f\x0c\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x28\x01\x00\x7c\x18\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x1e\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x18\x59\x00\x90\x01\x8c\x71\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x90\x01\x8c\x38\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyCompactUnicodeObject _compact;
        uint16_t _data[257];
    }
zipimport_toplevel_consts_22 = {
    ._compact = {
        ._base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyUnicode_Type,
            },
            .length = 256,
            .hash = -1,
            .state = {
                .kind = 2,
                .compact = 1,
                .ascii = 0,
                .statically_allocated = 1,
            },
        },
        .utf8 = "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x3e\x3f\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x5b\x5c\x5d\x5e\x5f\x60\x61\x62\x63\x64\x65\x66\x67\x68\x69\x6a\x6b\x6c\x6d\x6e\x6f\x70\x71\x72\x73\x74\x75\x76\x77\x78\x79\x7a\x7b\x7c\x7d\x7e\x7f\xc3\x87\xc3\xbc\xc3\xa9\xc3\xa2\xc3\xa4\xc3\xa0\xc3\xa5\xc3\xa7\xc3\xaa\xc3\xab\xc3\xa8\xc3\xaf\xc3\xae\xc3\xac\xc3\x84\xc3\x85\xc3\x89\xc3\xa6\xc3\x86\xc3\xb4\xc3\xb6\xc3\xb2\xc3\xbb\xc3\xb9\xc3\xbf\xc3\x96\xc3\x9c\xc2\xa2\xc2\xa3\xc2\xa5\xe2\x82\xa7\xc6\x92\xc3\xa1\xc3\xad\xc3\xb3\xc3\xba\xc3\xb1\xc3\x91\xc2\xaa\xc2\xba\xc2\xbf\xe2\x8c\x90\xc2\xac\xc2\xbd\xc2\xbc\xc2\xa1\xc2\xab\xc2\xbb\xe2\x96\x91\xe2\x96\x92\xe2\x96\x93\xe2\x94\x82\xe2\x94\xa4\xe2\x95\xa1\xe2\x95\xa2\xe2\x95\x96\xe2\x95\x95\xe2\x95\xa3\xe2\x95\x91\xe2\x95\x97\xe2\x95\x9d\xe2\x95\x9c\xe2\x95\x9b\xe2\x94\x90\xe2\x94\x94\xe2\x94\xb4\xe2\x94\xac\xe2\x94\x9c\xe2\x94\x80\xe2\x94\xbc\xe2\x95\x9e\xe2\x95\x9f\xe2\x95\x9a\xe2\x95\x94\xe2\x95\xa9\xe2\x95\xa6\xe2\x95\xa0\xe2\x95\x90\xe2\x95\xac\xe2\x95\xa7\xe2\x95\xa8\xe2\x95\xa4\xe2\x95\xa5\xe2\x95\x99\xe2\x95\x98\xe2\x95\x92\xe2\x95\x93\xe2\x95\xab\xe2\x95\xaa\xe2\x94\x98\xe2\x94\x8c\xe2\x96\x88\xe2\x96\x84\xe2\x96\x8c\xe2\x96\x90\xe2\x96\x80\xce\xb1\xc3\x9f\xce\x93\xcf\x80\xce\xa3\xcf\x83\xc2\xb5\xcf\x84\xce\xa6\xce\x98\xce\xa9\xce\xb4\xe2\x88\x9e\xcf\x86\xce\xb5\xe2\x88\xa9\xe2\x89\xa1\xc2\xb1\xe2\x89\xa5\xe2\x89\xa4\xe2\x8c\xa0\xe2\x8c\xa1\xc3\xb7\xe2\x89\x88\xc2\xb0\xe2\x88\x99\xc2\xb7\xe2\x88\x9a\xe2\x81\xbf\xc2\xb2\xe2\x96\xa0\xc2\xa0",
        .utf8_length = 446,
    },
    ._data = {
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
        16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
        32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
        48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
        64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
        80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
        96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111,
        112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127,
        199, 252, 233, 226, 228, 224, 229, 231, 234, 235, 232, 239, 238, 236, 196, 197,
        201, 230, 198, 244, 246, 242, 251, 249, 255, 214, 220, 162, 163, 165, 8359, 402,
        225, 237, 243, 250, 241, 209, 170, 186, 191, 8976, 172, 189, 188, 161, 171, 187,
        9617, 9618, 9619, 9474, 9508, 9569, 9570, 9558, 9557, 9571, 9553, 9559, 9565, 9564, 9563, 9488,
        9492, 9524, 9516, 9500, 9472, 9532, 9566, 9567, 9562, 9556, 9577, 9574, 9568, 9552, 9580, 9575,
        9576, 9572, 9573, 9561, 9560, 9554, 9555, 9579, 9578, 9496, 9484, 9608, 9604, 9612, 9616, 9600,
        945, 223, 915, 960, 931, 963, 181, 964, 934, 920, 937, 948, 8734, 966, 949, 8745,
        8801, 177, 8805, 8804, 8992, 8993, 247, 8776, 176, 8729, 183, 8730, 8319, 178, 9632, 160,
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
zipimport_toplevel_consts_23_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimport: zlib UNAVAILABLE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
zipimport_toplevel_consts_23_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "can't decompress data; zlib not available",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_decompress = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "decompress",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_23_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_decompress._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
zipimport_toplevel_consts_23_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimport: zlib available",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
zipimport_toplevel_consts_23_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            Py_None,
            & zipimport_toplevel_consts_23_consts_1._ascii.ob_base,
            & zipimport_toplevel_consts_23_consts_2._ascii.ob_base,
            Py_True,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & zipimport_toplevel_consts_23_consts_5._object.ob_base.ob_base,
            Py_False,
            & zipimport_toplevel_consts_23_consts_7._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__importing_zlib = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_importing_zlib",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_zlib = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zlib",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
zipimport_toplevel_consts_23_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__importing_zlib._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
            & const_str_zlib._ascii.ob_base,
            & const_str_decompress._ascii.ob_base,
            & const_str_Exception._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str__get_decompress_func = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_decompress_func",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[130];
    }
zipimport_toplevel_consts_23_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 129,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe5\x07\x16\xf4\x06\x00\x09\x13\xd7\x08\x23\xd1\x08\x23\xd0\x24\x41\xd4\x08\x42\xdc\x0e\x1c\xd0\x1d\x48\xd3\x0e\x49\xd0\x08\x49\xe0\x16\x1a\x80\x4f\xf0\x02\x06\x05\x20\xde\x08\x23\xf0\x0a\x00\x1b\x20\x88\x0f\xe4\x04\x0e\xd7\x04\x1f\xd1\x04\x1f\xd0\x20\x3b\xd4\x04\x3c\xd8\x0b\x15\xd0\x04\x15\xf8\xf4\x0f\x00\x0c\x15\xf2\x00\x02\x05\x4a\x01\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xd0\x24\x41\xd4\x08\x42\xdc\x0e\x1c\xd0\x1d\x48\xd3\x0e\x49\xd0\x08\x49\xf0\x05\x02\x05\x4a\x01\xfb\xf0\x08\x00\x1b\x20\x89\x0f\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
zipimport_toplevel_consts_23_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\xaa\x06\x41\x0a\x00\xc1\x0a\x2a\x41\x34\x03\xc1\x34\x03\x41\x37\x00\xc1\x37\x04\x41\x3b\x03",
};
static
    struct _PyCode_DEF(252)
zipimport_toplevel_consts_23 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 126,
    },
    .co_consts = & zipimport_toplevel_consts_23_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_23_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_23_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 500,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 238,
    .co_localsplusnames = & zipimport_toplevel_consts_23_consts_5._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__get_decompress_func._ascii.ob_base,
    .co_qualname = & const_str__get_decompress_func._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_23_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x20\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x64\x03\x61\x00\x09\x00\x64\x04\x64\x05\x6c\x04\x6d\x05\x7d\x00\x01\x00\x09\x00\x64\x06\x61\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x53\x00\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x21\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x64\x06\x61\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
zipimport_toplevel_consts_24_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "negative data size",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
zipimport_toplevel_consts_24_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x50\x4b\x03\x04",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
zipimport_toplevel_consts_24_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bad local file header: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
zipimport_toplevel_consts_24_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zipimport: can't read data",
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_negative_15 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(-1, 1),
    .ob_digit = { 15 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
zipimport_toplevel_consts_24_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & zipimport_toplevel_consts_24_consts_2._ascii.ob_base,
            & zipimport_toplevel_consts_21_consts_4._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 30],
            & zipimport_toplevel_consts_21_consts_16._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 4],
            & zipimport_toplevel_consts_24_consts_8.ob_base.ob_base,
            & zipimport_toplevel_consts_24_consts_9._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 26],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 28],
            & zipimport_toplevel_consts_24_consts_12._ascii.ob_base,
            & zipimport_toplevel_consts_23_consts_2._ascii.ob_base,
            & const_int_negative_15.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
zipimport_toplevel_consts_24_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_ZipImportError._ascii.ob_base,
            &_Py_ID(_io),
            & const_str_open_code._ascii.ob_base,
            &_Py_ID(seek),
            & const_str_OSError._ascii.ob_base,
            &_Py_ID(read),
            &_Py_ID(len),
            & const_str_EOFError._ascii.ob_base,
            & const_str__unpack_uint16._ascii.ob_base,
            & const_str__get_decompress_func._ascii.ob_base,
            & const_str_Exception._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[436];
    }
zipimport_toplevel_consts_24_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 435,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x4d\x56\xd1\x04\x4a\x80\x48\x88\x68\x98\x09\xa0\x39\xa8\x6b\xb8\x34\xc0\x14\xc0\x73\xd8\x07\x10\x90\x31\x82\x7d\xdc\x0e\x1c\xd0\x1d\x31\xd3\x0e\x32\xd0\x08\x32\xe4\x09\x0c\x8f\x1d\x89\x1d\x90\x77\xd4\x09\x1f\xa0\x32\xf0\x04\x03\x09\x54\x01\xd8\x0c\x0e\x8f\x47\x89\x47\x90\x4b\xd4\x0c\x20\xf0\x06\x00\x12\x14\x97\x17\x91\x17\x98\x12\x93\x1b\x88\x06\xdc\x0b\x0e\x88\x76\x8b\x3b\x98\x22\xd2\x0b\x1c\xdc\x12\x1a\xd0\x1b\x38\xd3\x12\x39\xd0\x0c\x39\xe0\x0b\x11\x90\x22\x90\x31\x88\x3a\x98\x1d\xd2\x0b\x26\xe4\x12\x20\xd0\x23\x3a\xb8\x37\xb8\x2b\xd0\x21\x46\xc8\x57\xd4\x12\x55\xd0\x0c\x55\xe4\x14\x22\xa0\x36\xa8\x22\xa8\x52\xa0\x3d\xd3\x14\x31\x88\x09\xdc\x15\x23\xa0\x46\xa8\x32\xa8\x62\xa0\x4d\xd3\x15\x32\x88\x0a\xd8\x16\x18\x98\x39\x91\x6e\xa0\x7a\xd1\x16\x31\x88\x0b\xd8\x08\x13\x90\x7b\xd1\x08\x22\x88\x0b\xf0\x02\x03\x09\x54\x01\xd8\x0c\x0e\x8f\x47\x89\x47\x90\x4b\xd4\x0c\x20\xf0\x06\x00\x14\x16\x97\x37\x91\x37\x98\x39\xd3\x13\x25\x88\x08\xdc\x0b\x0e\x88\x78\x8b\x3d\x98\x49\xd2\x0b\x25\xdc\x12\x19\xd0\x1a\x36\xd3\x12\x37\xd0\x0c\x37\xf0\x03\x00\x0c\x26\xf7\x2f\x00\x0a\x20\xf0\x34\x00\x08\x10\x90\x31\x82\x7d\xe0\x0f\x17\x88\x0f\xf0\x06\x03\x05\x4a\x01\xdc\x15\x29\xd3\x15\x2b\x88\x0a\xf1\x06\x00\x0c\x16\x90\x68\xa0\x03\xd3\x0b\x24\xd0\x04\x24\xf8\xf4\x3f\x00\x10\x17\xf2\x00\x01\x09\x54\x01\xdc\x12\x20\xd0\x23\x38\xb8\x17\xb8\x0b\xd0\x21\x44\xc8\x37\xd4\x12\x53\xd0\x0c\x53\xf0\x03\x01\x09\x54\x01\xfb\xf4\x20\x00\x10\x17\xf2\x00\x01\x09\x54\x01\xdc\x12\x20\xd0\x23\x38\xb8\x17\xb8\x0b\xd0\x21\x44\xc8\x37\xd4\x12\x53\xd0\x0c\x53\xf0\x03\x01\x09\x54\x01\xfa\xf7\x29\x00\x0a\x20\xd0\x09\x1f\xfb\xf4\x42\x01\x00\x0c\x15\xf2\x00\x01\x05\x4a\x01\xdc\x0e\x1c\xd0\x1d\x48\xd3\x0e\x49\xd0\x08\x49\xf0\x03\x01\x05\x4a\x01\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[72];
    }
zipimport_toplevel_consts_24_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 71,
    },
    .ob_shash = -1,
    .ob_sval = "\xb1\x01\x45\x09\x03\xb3\x11\x44\x0f\x02\xc1\x04\x41\x2b\x45\x09\x03\xc2\x30\x11\x44\x2c\x02\xc3\x01\x2a\x45\x09\x03\xc3\x3c\x0a\x45\x15\x00\xc4\x0f\x1a\x44\x29\x05\xc4\x29\x03\x45\x09\x03\xc4\x2c\x1a\x45\x06\x05\xc5\x06\x03\x45\x09\x03\xc5\x09\x05\x45\x12\x07\xc5\x15\x15\x45\x2a\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_datapath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "datapath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_raw_data = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "raw_data",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
zipimport_toplevel_consts_24_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            & const_str_archive._ascii.ob_base,
            & const_str_toc_entry._ascii.ob_base,
            & const_str_datapath._ascii.ob_base,
            & const_str_compress._ascii.ob_base,
            & const_str_data_size._ascii.ob_base,
            & const_str_file_size._ascii.ob_base,
            & const_str_file_offset._ascii.ob_base,
            & const_str_time._ascii.ob_base,
            & const_str_date._ascii.ob_base,
            & const_str_crc._ascii.ob_base,
            & const_str_fp._ascii.ob_base,
            &_Py_ID(buffer),
            & const_str_name_size._ascii.ob_base,
            & const_str_extra_size._ascii.ob_base,
            & const_str_header_size._ascii.ob_base,
            & const_str_raw_data._ascii.ob_base,
            & const_str_decompress._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
zipimport_toplevel_consts_24_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "                 ",
};
static
    struct _PyCode_DEF(730)
zipimport_toplevel_consts_24 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 365,
    },
    .co_consts = & zipimport_toplevel_consts_24_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_24_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_24_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 25 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 521,
    .co_nlocalsplus = 17,
    .co_nlocals = 17,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 239,
    .co_localsplusnames = & zipimport_toplevel_consts_24_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & zipimport_toplevel_consts_24_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__get_data._ascii.ob_base,
    .co_qualname = & const_str__get_data._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_24_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x5c\x08\x00\x00\x7d\x02\x7d\x03\x7d\x04\x7d\x05\x7d\x06\x7d\x07\x7d\x08\x7d\x09\x7c\x04\x64\x01\x6b\x02\x00\x00\x72\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x0a\x09\x00\x7c\x0a\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x0a\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0b\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x64\x05\x6b\x37\x00\x00\x72\x0b\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x0b\x64\x00\x64\x07\x1a\x00\x64\x08\x6b\x37\x00\x00\x72\x10\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x04\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0b\x64\x0a\x64\x0b\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0c\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0b\x64\x0b\x64\x05\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0d\x64\x05\x7c\x0c\x7a\x00\x00\x00\x7c\x0d\x7a\x00\x00\x00\x7d\x0e\x7c\x06\x7c\x0e\x7a\x0d\x00\x00\x7d\x06\x09\x00\x7c\x0a\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x0a\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0f\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0f\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x04\x6b\x37\x00\x00\x72\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x09\x00\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x03\x64\x01\x6b\x28\x00\x00\x72\x02\x7f\x0f\x53\x00\x09\x00\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x10\x02\x00\x7c\x10\x7f\x0f\x64\x0e\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x04\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x11\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x9b\x02\x9d\x02\x7c\x00\xac\x04\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x5e\x78\x03\x59\x00\x77\x01\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0c\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_25_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_abs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abs",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_25_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_abs._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__eq_mtime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_eq_mtime",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
zipimport_toplevel_consts_25_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0e\x88\x72\x90\x42\x89\x77\x8b\x3c\x98\x31\xd1\x0b\x1c\xd0\x04\x1c",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_t1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "t1",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_t2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "t2",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_25_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_t1._ascii.ob_base,
            & const_str_t2._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(36)
zipimport_toplevel_consts_25 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 18,
    },
    .co_consts = & zipimport_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_25_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 567,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 240,
    .co_localsplusnames = & zipimport_toplevel_consts_25_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__eq_mtime._ascii.ob_base,
    .co_qualname = & const_str__eq_mtime._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_25_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7a\x0a\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x1a\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
zipimport_toplevel_consts_26_consts_11 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "compiled module ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
zipimport_toplevel_consts_26_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " is not a code object",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
zipimport_toplevel_consts_26_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_45_consts_3._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & const_str_never._ascii.ob_base,
            & const_str_always._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 12],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            & importlib__bootstrap_external_toplevel_consts_43_consts_4._ascii.ob_base,
            & zipimport_toplevel_consts_26_consts_11._ascii.ob_base,
            & zipimport_toplevel_consts_26_consts_12._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__get_pyc_source = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_pyc_source",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
const_str__get_mtime_and_size_of_source = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_mtime_and_size_of_source",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[18];
        }_object;
    }
zipimport_toplevel_consts_26_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 18,
        },
        .ob_item = {
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__classify_pyc._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_check_hash_based_pycs._ascii.ob_base,
            & const_str__get_pyc_source._ascii.ob_base,
            & const_str_source_hash._ascii.ob_base,
            & const_str__RAW_MAGIC_NUMBER._ascii.ob_base,
            & const_str__validate_hash_pyc._ascii.ob_base,
            & const_str__get_mtime_and_size_of_source._ascii.ob_base,
            & const_str__eq_mtime._ascii.ob_base,
            & const_str__unpack_uint32._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str_marshal._ascii.ob_base,
            & const_str_loads._ascii.ob_base,
            &_Py_ID(isinstance),
            & const_str__code_type._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__unmarshal_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_unmarshal_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[322];
    }
zipimport_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 321,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x10\x18\xd8\x10\x18\xf1\x05\x03\x13\x06\x80\x4b\xf4\x0a\x00\x0d\x20\xd7\x0c\x2d\xd1\x0c\x2d\xa8\x64\xb0\x48\xb8\x6b\xd3\x0c\x4a\x80\x45\xe0\x11\x16\x98\x13\x91\x1b\xa0\x01\xd1\x11\x21\x80\x4a\xd9\x07\x11\xd8\x17\x1c\x98\x74\x91\x7c\xa0\x71\xd1\x17\x28\x88\x0c\xdc\x0c\x10\xd7\x0c\x26\xd1\x0c\x26\xa8\x27\xd2\x0c\x31\xd9\x11\x1d\xa4\x14\xd7\x21\x3b\xd1\x21\x3b\xb8\x78\xd2\x21\x47\xdc\x1b\x2a\xa8\x34\xb0\x18\xd3\x1b\x3a\x88\x4c\xd8\x0f\x1b\xd0\x0f\x27\xdc\x1e\x22\xd7\x1e\x2e\xd1\x1e\x2e\xdc\x14\x27\xd7\x14\x39\xd1\x14\x39\xd8\x14\x20\xf3\x05\x03\x1f\x12\x90\x0b\xf4\x0a\x00\x11\x24\xd7\x10\x36\xd1\x10\x36\xd8\x14\x18\x98\x2b\xa0\x78\xb0\x1b\xf5\x03\x01\x11\x3e\xf4\x08\x00\x0d\x2a\xa8\x24\xb0\x08\xd3\x0c\x39\xf1\x03\x00\x09\x22\x88\x0c\x90\x6b\xf1\x06\x00\x0c\x18\xf4\x06\x00\x15\x1e\x9c\x6e\xa8\x54\xb0\x21\xb0\x42\xa8\x5a\xd3\x1e\x38\xb8\x2c\xd4\x14\x47\xdc\x14\x22\xa0\x34\xa8\x02\xa8\x32\xa0\x3b\xd3\x14\x2f\xb0\x3b\xd2\x14\x3e\xdc\x10\x1a\xd7\x10\x2b\xd1\x10\x2b\xd8\x16\x2c\xa8\x58\xa8\x4c\xd0\x14\x39\xf4\x03\x01\x11\x3b\xe0\x17\x1b\xe4\x0b\x12\x8f\x3d\x89\x3d\x98\x14\x98\x62\x98\x63\x98\x19\xd3\x0b\x23\x80\x44\xdc\x0b\x15\x90\x64\x9c\x4a\xd4\x0b\x27\xdc\x0e\x17\xd0\x1a\x2a\xa8\x38\xa8\x2c\xd0\x36\x4b\xd0\x18\x4c\xd3\x0e\x4d\xd0\x08\x4d\xd8\x0b\x0f\x80\x4b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
zipimport_toplevel_consts_26_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_pathname._ascii.ob_base,
            & const_str_fullpath._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(data),
            & const_str_exc_details._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_hash_based._ascii.ob_base,
            & const_str_check_source._ascii.ob_base,
            & const_str_source_bytes._ascii.ob_base,
            & const_str_source_hash._ascii.ob_base,
            & const_str_source_mtime._ascii.ob_base,
            & const_str_source_size._ascii.ob_base,
            &_Py_ID(code),
        },
    },
};
static
    struct _PyCode_DEF(604)
zipimport_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 302,
    },
    .co_consts = & zipimport_toplevel_consts_26_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 21 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 575,
    .co_nlocalsplus = 14,
    .co_nlocals = 14,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 241,
    .co_localsplusnames = & zipimport_toplevel_consts_26_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__unmarshal_code._ascii.ob_base,
    .co_qualname = & const_str__unmarshal_code._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x03\x7c\x02\x64\x01\x9c\x02\x7d\x05\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x03\x7c\x05\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x64\x02\x7a\x01\x00\x00\x64\x03\x6b\x37\x00\x00\x7d\x07\x7c\x07\x72\x7b\x7c\x06\x64\x04\x7a\x01\x00\x00\x64\x03\x6b\x37\x00\x00\x7d\x08\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x6b\x37\x00\x00\x72\xb3\x7c\x08\x73\x13\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x6b\x28\x00\x00\x72\x9e\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x09\x7c\x09\x81\x90\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0a\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x0a\x7c\x03\x7c\x05\xab\x04\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x53\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x0b\x7d\x0c\x7c\x0b\x72\x42\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x07\x64\x08\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x0b\xab\x02\x00\x00\x00\x00\x00\x00\x72\x11\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x08\x64\x09\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x0c\x6b\x37\x00\x00\x72\x19\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\x7c\x03\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x09\x64\x00\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0d\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x0f\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\x7c\x01\x9b\x02\x64\x0c\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x0d\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
zipimport_toplevel_consts_27_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_external_toplevel_consts_29.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[10]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[13]),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_27_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(replace),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str__normalize_line_endings = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_normalize_line_endings",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
zipimport_toplevel_consts_27_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0d\x13\x8f\x5e\x89\x5e\x98\x47\xa0\x55\xd3\x0d\x2b\x80\x46\xd8\x0d\x13\x8f\x5e\x89\x5e\x98\x45\xa0\x35\xd3\x0d\x29\x80\x46\xd8\x0b\x11\x80\x4d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
zipimport_toplevel_consts_27_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(source),
        },
    },
};
static
    struct _PyCode_DEF(78)
zipimport_toplevel_consts_27 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & zipimport_toplevel_consts_27_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_27_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 620,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 242,
    .co_localsplusnames = & zipimport_toplevel_consts_27_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__normalize_line_endings._ascii.ob_base,
    .co_qualname = & const_str__normalize_line_endings._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_27_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
zipimport_toplevel_consts_28_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & const_str_exec._ascii.ob_base,
            Py_True,
            & importlib__bootstrap_external_toplevel_consts_68_consts_4_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_28_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__normalize_line_endings._ascii.ob_base,
            & const_str_compile._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__compile_source = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_compile_source",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
zipimport_toplevel_consts_28_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0d\x24\xa0\x56\xd3\x0d\x2c\x80\x46\xdc\x0b\x12\x90\x36\x98\x38\xa0\x56\xb8\x24\xd4\x0b\x3f\xd0\x04\x3f",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_28_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_pathname._ascii.ob_base,
            &_Py_ID(source),
        },
    },
};
static
    struct _PyCode_DEF(54)
zipimport_toplevel_consts_28 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & zipimport_toplevel_consts_28_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_28_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 627,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 243,
    .co_localsplusnames = & zipimport_toplevel_consts_28_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__compile_source._ascii.ob_base,
    .co_qualname = & const_str__compile_source._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_28_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x64\x01\x64\x02\xac\x03\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_1980 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 1980 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
zipimport_toplevel_consts_29_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 9],
            & const_int_1980.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 5],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 15],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 31],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 11],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 63],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_mktime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mktime",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_29_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_time._ascii.ob_base,
            & const_str_mktime._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__parse_dostime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_parse_dostime",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[90];
    }
zipimport_toplevel_consts_29_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 89,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0b\x0f\x8f\x3b\x89\x3b\xd8\x09\x0a\x88\x61\x89\x16\x90\x34\x89\x0f\xd8\x09\x0a\x88\x61\x89\x16\x90\x33\x89\x0e\xd8\x08\x09\x88\x44\x89\x08\xd8\x08\x09\x88\x52\x89\x07\xd8\x09\x0a\x88\x61\x89\x16\x90\x34\x89\x0f\xd8\x09\x0a\x88\x54\x89\x18\x90\x51\x89\x0e\xd8\x08\x0a\x88\x42\x90\x02\xf0\x0f\x07\x18\x14\xf3\x00\x07\x0c\x15\xf0\x00\x07\x05\x15",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_29_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[100],
            (PyObject *)&_Py_SINGLETON(strings).ascii[116],
        },
    },
};
static
    struct _PyCode_DEF(122)
zipimport_toplevel_consts_29 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 61,
    },
    .co_consts = & zipimport_toplevel_consts_29_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_29_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 11,
    .co_firstlineno = 633,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 244,
    .co_localsplusnames = & zipimport_toplevel_consts_29_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__parse_dostime._ascii.ob_base,
    .co_qualname = & const_str__parse_dostime._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_29_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\x7a\x09\x00\x00\x64\x02\x7a\x00\x00\x00\x7c\x00\x64\x03\x7a\x09\x00\x00\x64\x04\x7a\x01\x00\x00\x7c\x00\x64\x05\x7a\x01\x00\x00\x7c\x01\x64\x06\x7a\x09\x00\x00\x7c\x01\x64\x03\x7a\x09\x00\x00\x64\x07\x7a\x01\x00\x00\x7c\x01\x64\x05\x7a\x01\x00\x00\x64\x08\x7a\x05\x00\x00\x64\x09\x64\x09\x64\x09\x66\x09\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
zipimport_toplevel_consts_30_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[99],
            (PyObject *)&_Py_SINGLETON(strings).ascii[111],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
zipimport_toplevel_consts_30_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & zipimport_toplevel_consts_30_consts_2._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 5],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 6],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            & importlib__bootstrap_external_toplevel_consts_81._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
zipimport_toplevel_consts_30_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__files._ascii.ob_base,
            & const_str__parse_dostime._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str_IndexError._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[127];
    }
zipimport_toplevel_consts_30_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 126,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x0c\x05\x14\xe0\x0f\x13\x90\x42\x90\x43\x88\x79\x98\x4a\xd1\x0f\x26\xd0\x08\x26\xd0\x0f\x26\xd8\x0f\x13\x90\x43\x90\x52\x88\x79\x88\x04\xd8\x14\x18\x97\x4b\x91\x4b\xa0\x04\xd1\x14\x25\x88\x09\xf0\x06\x00\x10\x19\x98\x11\x89\x7c\x88\x04\xd8\x0f\x18\x98\x11\x89\x7c\x88\x04\xd8\x1c\x25\xa0\x61\x99\x4c\xd0\x08\x19\xdc\x0f\x1d\x98\x64\xa0\x44\xd3\x0f\x29\xd0\x2b\x3c\xd0\x0f\x3c\xd0\x08\x3c\xf8\xdc\x0c\x14\x94\x6a\xa4\x29\xd0\x0b\x2c\xf2\x00\x01\x05\x14\xd9\x0f\x13\xf0\x03\x01\x05\x14\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
zipimport_toplevel_consts_30_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x39\x3c\x00\xbc\x14\x41\x13\x03\xc1\x12\x01\x41\x13\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_uncompressed_size = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "uncompressed_size",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
zipimport_toplevel_consts_30_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            & const_str_toc_entry._ascii.ob_base,
            & const_str_time._ascii.ob_base,
            & const_str_date._ascii.ob_base,
            & const_str_uncompressed_size._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(172)
zipimport_toplevel_consts_30 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 86,
    },
    .co_consts = & zipimport_toplevel_consts_30_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_30_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_30_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 646,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 245,
    .co_localsplusnames = & zipimport_toplevel_consts_30_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__get_mtime_and_size_of_source._ascii.ob_base,
    .co_qualname = & const_str__get_mtime_and_size_of_source._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_30_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x01\x64\x01\x64\x00\x1a\x00\x64\x02\x76\x00\x73\x02\x4a\x00\x82\x01\x7c\x01\x64\x00\x64\x01\x1a\x00\x7d\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x7c\x02\x64\x03\x19\x00\x00\x00\x7d\x03\x7c\x02\x64\x04\x19\x00\x00\x00\x7d\x04\x7c\x02\x64\x05\x19\x00\x00\x00\x7d\x05\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x05\x66\x02\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x03\x01\x00\x59\x00\x79\x06\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_31_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & zipimport_toplevel_consts_30_consts_2._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
zipimport_toplevel_consts_31_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__files._ascii.ob_base,
            & const_str__get_data._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[92];
    }
zipimport_toplevel_consts_31_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 91,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0f\x90\x02\x90\x03\x88\x39\x98\x0a\xd1\x0b\x22\xd0\x04\x22\xd0\x0b\x22\xd8\x0b\x0f\x90\x03\x90\x12\x88\x39\x80\x44\xf0\x04\x05\x05\x32\xd8\x14\x18\x97\x4b\x91\x4b\xa0\x04\xd1\x14\x25\x88\x09\xf4\x08\x00\x10\x19\x98\x14\x9f\x1c\x99\x1c\xa0\x79\xd3\x0f\x31\xd0\x08\x31\xf8\xf4\x07\x00\x0c\x14\xf2\x00\x01\x05\x14\xd9\x0f\x13\xf0\x03\x01\x05\x14\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
zipimport_toplevel_consts_31_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x90\x0f\x35\x00\xb5\x09\x41\x01\x03\xc1\x00\x01\x41\x01\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
zipimport_toplevel_consts_31_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            & const_str_toc_entry._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(136)
zipimport_toplevel_consts_31 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 68,
    },
    .co_consts = & zipimport_toplevel_consts_31_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_31_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_31_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 665,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 246,
    .co_localsplusnames = & zipimport_toplevel_consts_31_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__get_pyc_source._ascii.ob_base,
    .co_qualname = & const_str__get_pyc_source._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_31_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x64\x01\x64\x00\x1a\x00\x64\x02\x76\x00\x73\x02\x4a\x00\x82\x01\x7c\x01\x64\x00\x64\x01\x1a\x00\x7d\x01\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
zipimport_toplevel_consts_32_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "trying {}{}{}",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
zipimport_toplevel_consts_32_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "module load failed: ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
zipimport_toplevel_consts_32_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            Py_None,
            & zipimport_toplevel_consts_32_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & importlib__bootstrap_toplevel_consts_24._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & zipimport_toplevel_consts_32_consts_5._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
            & zipimport_toplevel_consts_11_consts_8_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
zipimport_toplevel_consts_32_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str__get_module_path._ascii.ob_base,
            & const_str__zip_searchorder._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__verbose_message._ascii.ob_base,
            & const_str_archive._ascii.ob_base,
            & const_str_path_sep._ascii.ob_base,
            & const_str__files._ascii.ob_base,
            & const_str__get_data._ascii.ob_base,
            & const_str__unmarshal_code._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__compile_source._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[291];
    }
zipimport_toplevel_consts_32_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 290,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0b\x1b\x98\x44\xa0\x28\xd3\x0b\x2b\x80\x44\xd8\x13\x17\x80\x4c\xdf\x29\x39\xd1\x08\x25\x88\x06\x90\x0a\x98\x49\xd8\x13\x17\x98\x26\x91\x3d\x88\x08\xdc\x08\x12\xd7\x08\x23\xd1\x08\x23\xa0\x4f\xb0\x54\xb7\x5c\xb1\x5c\xc4\x38\xc8\x58\xd0\x61\x62\xd5\x08\x63\xf0\x02\x14\x09\x2c\xd8\x18\x1c\x9f\x0b\x99\x0b\xa0\x48\xd1\x18\x2d\x88\x49\xf0\x08\x00\x17\x20\xa0\x01\x91\x6c\x88\x47\xdc\x13\x1c\x98\x54\x9f\x5c\x99\x5c\xa8\x39\xd3\x13\x35\x88\x44\xd8\x13\x17\x88\x44\xd9\x0f\x19\xf0\x02\x03\x11\x27\xdc\x1b\x2a\xa8\x34\xb0\x17\xb8\x28\xc0\x48\xc8\x64\xd3\x1b\x53\x91\x44\xf4\x08\x00\x18\x27\xa0\x77\xb0\x04\xd3\x17\x35\x90\x04\xd8\x0f\x13\x88\x7c\xf0\x06\x00\x11\x19\xd8\x16\x1f\xa0\x01\x91\x6c\x88\x47\xd8\x13\x17\x98\x19\xa0\x47\xd0\x13\x2b\xd2\x0c\x2b\xf0\x2f\x00\x2a\x3a\xf1\x32\x00\x0c\x18\xd8\x14\x28\xa8\x1c\xa8\x0e\xd0\x12\x37\x88\x43\xdc\x12\x20\xa0\x13\xa8\x38\xd4\x12\x34\xb8\x2c\xd0\x0c\x46\xe4\x12\x20\xd0\x23\x35\xb0\x68\xb0\x5c\xd0\x21\x42\xc8\x18\xd4\x12\x52\xd0\x0c\x52\xf8\xf4\x1f\x00\x18\x23\xf2\x00\x01\x11\x27\xd8\x23\x26\x95\x4c\xfb\xf0\x03\x01\x11\x27\xfb\xf4\x13\x00\x10\x18\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
zipimport_toplevel_consts_32_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x0a\x0f\x43\x22\x02\xc1\x39\x0f\x43\x0a\x02\xc3\x0a\x09\x43\x1f\x05\xc3\x13\x02\x43\x1a\x05\xc3\x1a\x05\x43\x1f\x05\xc3\x22\x09\x43\x2e\x05\xc3\x2d\x01\x43\x2e\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_import_error = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "import_error",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
zipimport_toplevel_consts_32_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_fullname._ascii.ob_base,
            &_Py_ID(path),
            & const_str_import_error._ascii.ob_base,
            & const_str_suffix._ascii.ob_base,
            & const_str_isbytecode._ascii.ob_base,
            & const_str_ispackage._ascii.ob_base,
            & const_str_fullpath._ascii.ob_base,
            & const_str_toc_entry._ascii.ob_base,
            & const_str_modpath._ascii.ob_base,
            &_Py_ID(data),
            &_Py_ID(code),
            & const_str_exc._ascii.ob_base,
            &_Py_ID(msg),
        },
    },
};
static
    struct _PyCode_DEF(482)
zipimport_toplevel_consts_32 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 241,
    },
    .co_consts = & zipimport_toplevel_consts_32_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_consts_32_names._object.ob_base.ob_base,
    .co_exceptiontable = & zipimport_toplevel_consts_32_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 22 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 680,
    .co_nlocalsplus = 14,
    .co_nlocals = 14,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 247,
    .co_localsplusnames = & zipimport_toplevel_consts_32_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = & const_str__get_module_code._ascii.ob_base,
    .co_qualname = & const_str__get_module_code._ascii.ob_base,
    .co_linetable = & zipimport_toplevel_consts_32_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x00\x7d\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x8d\x00\x00\x5c\x03\x00\x00\x7d\x04\x7d\x05\x7d\x06\x7c\x02\x7c\x04\x7a\x00\x00\x00\x7d\x07\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x64\x02\xac\x03\xab\x05\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x19\x00\x00\x00\x7d\x08\x7c\x08\x64\x04\x19\x00\x00\x00\x7d\x09\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0a\x64\x00\x7d\x0b\x7c\x05\x72\x11\x09\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x09\x7c\x07\x7c\x01\x7c\x0a\xab\x05\x00\x00\x00\x00\x00\x00\x7d\x0b\x6e\x0c\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x7c\x0a\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0b\x7c\x0b\x80\x01\x8c\x83\x7c\x08\x64\x04\x19\x00\x00\x00\x7d\x09\x7c\x0b\x7c\x06\x7c\x09\x66\x03\x63\x02\x01\x00\x53\x00\x04\x00\x7c\x03\x72\x13\x64\x05\x7c\x03\x9b\x00\x9d\x02\x7d\x0d\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x7c\x01\xac\x06\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x03\x82\x02\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x7c\x01\x9b\x02\x9d\x02\x7c\x01\xac\x06\xab\x02\x00\x00\x00\x00\x00\x00\x82\x01\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0c\x7d\x0c\x7c\x0c\x7d\x03\x59\x00\x64\x00\x7d\x0c\x7e\x0c\x8c\x45\x64\x00\x7d\x0c\x7e\x0c\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xd8\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[33];
        }_object;
    }
zipimport_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 33,
        },
        .ob_item = {
            & zipimport_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & zipimport_toplevel_consts_3._object.ob_base.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
            & const_str_zipimporter._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & zipimport_toplevel_consts_7.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 22],
            & zipimport_toplevel_consts_9.ob_base.ob_base,
            & const_int_65535.ob_base,
            & zipimport_toplevel_consts_11.ob_base.ob_base,
            & zipimport_toplevel_consts_12._ascii.ob_base,
            Py_True,
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_11._ascii.ob_base,
            Py_False,
            & zipimport_toplevel_consts_16._object.ob_base.ob_base,
            & zipimport_toplevel_consts_17._object.ob_base.ob_base,
            & zipimport_toplevel_consts_18.ob_base.ob_base,
            & zipimport_toplevel_consts_19.ob_base.ob_base,
            & zipimport_toplevel_consts_20.ob_base.ob_base,
            & zipimport_toplevel_consts_21.ob_base.ob_base,
            & zipimport_toplevel_consts_22._compact._base.ob_base,
            & zipimport_toplevel_consts_23.ob_base.ob_base,
            & zipimport_toplevel_consts_24.ob_base.ob_base,
            & zipimport_toplevel_consts_25.ob_base.ob_base,
            & zipimport_toplevel_consts_26.ob_base.ob_base,
            & zipimport_toplevel_consts_27.ob_base.ob_base,
            & zipimport_toplevel_consts_28.ob_base.ob_base,
            & zipimport_toplevel_consts_29.ob_base.ob_base,
            & zipimport_toplevel_consts_30.ob_base.ob_base,
            & zipimport_toplevel_consts_31.ob_base.ob_base,
            & zipimport_toplevel_consts_32.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__frozen_importlib = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_frozen_importlib",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[46];
        }_object;
    }
zipimport_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 46,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str__frozen_importlib_external._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str__unpack_uint16._ascii.ob_base,
            & const_str__unpack_uint32._ascii.ob_base,
            & const_str__frozen_importlib._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str__imp._ascii.ob_base,
            &_Py_ID(_io),
            & const_str_marshal._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_time._ascii.ob_base,
            & const_str__warnings._ascii.ob_base,
            &_Py_ID(__all__),
            & const_str_path_sep._ascii.ob_base,
            & const_str_path_separators._ascii.ob_base,
            & const_str_alt_path_sep._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_ZipImportError._ascii.ob_base,
            & const_str__zip_directory_cache._ascii.ob_base,
            &_Py_ID(type),
            & const_str__module_type._ascii.ob_base,
            & const_str_END_CENTRAL_DIR_SIZE._ascii.ob_base,
            & const_str_STRING_END_ARCHIVE._ascii.ob_base,
            & const_str_MAX_COMMENT_LEN._ascii.ob_base,
            & const_str__LoaderBasics._ascii.ob_base,
            & const_str_zipimporter._ascii.ob_base,
            & const_str__zip_searchorder._ascii.ob_base,
            & const_str__get_module_path._ascii.ob_base,
            & const_str__is_dir._ascii.ob_base,
            & const_str__get_module_info._ascii.ob_base,
            & const_str__read_directory._ascii.ob_base,
            & const_str_cp437_table._ascii.ob_base,
            & const_str__importing_zlib._ascii.ob_base,
            & const_str__get_decompress_func._ascii.ob_base,
            & const_str__get_data._ascii.ob_base,
            & const_str__eq_mtime._ascii.ob_base,
            & const_str__unmarshal_code._ascii.ob_base,
            & const_str___code__._ascii.ob_base,
            & const_str__code_type._ascii.ob_base,
            & const_str__normalize_line_endings._ascii.ob_base,
            & const_str__compile_source._ascii.ob_base,
            & const_str__parse_dostime._ascii.ob_base,
            & const_str__get_mtime_and_size_of_source._ascii.ob_base,
            & const_str__get_pyc_source._ascii.ob_base,
            & const_str__get_module_code._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[308];
    }
zipimport_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 307,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x0c\x01\x04\xf3\x20\x00\x01\x39\xdf\x00\x45\xdb\x00\x26\xdb\x00\x0b\xdb\x00\x0a\xdb\x00\x0e\xdb\x00\x0a\xdb\x00\x0b\xdb\x00\x10\xe0\x0b\x1b\x98\x5d\xd0\x0a\x2b\x80\x07\xf0\x06\x00\x0c\x1f\xd7\x0b\x27\xd1\x0b\x27\x80\x08\xd8\x0f\x22\xd7\x0f\x32\xd1\x0f\x32\xb0\x31\xb0\x32\xd0\x0f\x36\x80\x0c\xf4\x06\x01\x01\x09\x90\x5b\xf4\x00\x01\x01\x09\xf0\x08\x00\x18\x1a\xd0\x00\x14\xe1\x0f\x13\x90\x43\x8b\x79\x80\x0c\xe0\x17\x19\xd0\x00\x14\xd8\x15\x22\xd0\x00\x12\xd8\x12\x1f\x80\x0f\xf4\x04\x64\x03\x01\x4f\x01\xd0\x12\x25\xd7\x12\x33\xd1\x12\x33\xf4\x00\x64\x03\x01\x4f\x01\xf0\x5a\x07\x00\x06\x0e\x90\x0e\xd1\x05\x1e\xa0\x04\xa0\x64\xd0\x04\x2b\xd8\x05\x0d\x90\x0d\xd1\x05\x1d\x98\x75\xa0\x64\xd0\x04\x2b\xd8\x04\x19\xd8\x04\x19\xf0\x09\x05\x14\x02\xd0\x00\x10\xf2\x12\x01\x01\x35\xf2\x08\x06\x01\x22\xf2\x12\x06\x01\x10\xf2\x3e\x7b\x01\x01\x11\xf0\x4a\x04\x18\x05\x2f\xf0\x05\x00\x01\x0c\xf0\x3a\x00\x13\x18\x80\x0f\xf2\x0a\x12\x01\x16\xf2\x2a\x28\x01\x25\xf2\x5c\x01\x02\x01\x1d\xf2\x10\x26\x01\x10\xf1\x50\x01\x00\x0e\x12\x90\x2f\xd7\x12\x2a\xd1\x12\x2a\xd3\x0d\x2b\x80\x0a\xf2\x0a\x03\x01\x12\xf2\x0e\x02\x01\x40\x01\xf2\x0c\x08\x01\x15\xf2\x1a\x0d\x01\x14\xf2\x26\x0a\x01\x32\xf3\x1e\x20\x01\x53\x01",
};
static
    struct _PyCode_DEF(410)
zipimport_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 205,
    },
    .co_consts = & zipimport_toplevel_consts._object.ob_base.ob_base,
    .co_names = & zipimport_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 248,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & zipimport_toplevel_consts_7_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & zipimport_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x5a\x02\x64\x01\x64\x03\x6c\x01\x6d\x03\x5a\x03\x6d\x04\x5a\x04\x01\x00\x64\x01\x64\x02\x6c\x05\x5a\x06\x64\x01\x64\x02\x6c\x07\x5a\x07\x64\x01\x64\x02\x6c\x08\x5a\x08\x64\x01\x64\x02\x6c\x09\x5a\x09\x64\x01\x64\x02\x6c\x0a\x5a\x0a\x64\x01\x64\x02\x6c\x0b\x5a\x0b\x64\x01\x64\x02\x6c\x0c\x5a\x0c\x64\x04\x64\x05\x67\x02\x5a\x0d\x65\x02\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x0e\x65\x02\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x64\x02\x1a\x00\x5a\x10\x02\x00\x47\x00\x64\x07\x84\x00\x64\x04\x65\x11\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x12\x69\x00\x5a\x13\x02\x00\x65\x14\x65\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x15\x64\x08\x5a\x16\x64\x09\x5a\x17\x64\x0a\x5a\x18\x02\x00\x47\x00\x64\x0b\x84\x00\x64\x05\x65\x02\x6a\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x1a\x65\x0e\x64\x0c\x7a\x00\x00\x00\x64\x0d\x64\x0d\x66\x03\x65\x0e\x64\x0e\x7a\x00\x00\x00\x64\x0f\x64\x0d\x66\x03\x64\x10\x64\x11\x66\x04\x5a\x1b\x64\x12\x84\x00\x5a\x1c\x64\x13\x84\x00\x5a\x1d\x64\x14\x84\x00\x5a\x1e\x64\x15\x84\x00\x5a\x1f\x64\x16\x5a\x20\x64\x0f\x61\x21\x64\x17\x84\x00\x5a\x22\x64\x18\x84\x00\x5a\x23\x64\x19\x84\x00\x5a\x24\x64\x1a\x84\x00\x5a\x25\x02\x00\x65\x14\x65\x25\x6a\x4c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x27\x64\x1b\x84\x00\x5a\x28\x64\x1c\x84\x00\x5a\x29\x64\x1d\x84\x00\x5a\x2a\x64\x1e\x84\x00\x5a\x2b\x64\x1f\x84\x00\x5a\x2c\x64\x20\x84\x00\x5a\x2d\x79\x02",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_zipimport_toplevel(void)
{
    return Py_NewRef((PyObject *) &zipimport_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[52];
    }
abc_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 51,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Abstract Base Classes (ABCs) according to PEP 3119.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[586];
    }
abc_toplevel_consts_1_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 585,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x64\x65\x63\x6f\x72\x61\x74\x6f\x72\x20\x69\x6e\x64\x69\x63\x61\x74\x69\x6e\x67\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x73\x2e\x0a\x0a\x20\x20\x20\x20\x52\x65\x71\x75\x69\x72\x65\x73\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x6d\x65\x74\x61\x63\x6c\x61\x73\x73\x20\x69\x73\x20\x41\x42\x43\x4d\x65\x74\x61\x20\x6f\x72\x20\x64\x65\x72\x69\x76\x65\x64\x20\x66\x72\x6f\x6d\x20\x69\x74\x2e\x20\x20\x41\x0a\x20\x20\x20\x20\x63\x6c\x61\x73\x73\x20\x74\x68\x61\x74\x20\x68\x61\x73\x20\x61\x20\x6d\x65\x74\x61\x63\x6c\x61\x73\x73\x20\x64\x65\x72\x69\x76\x65\x64\x20\x66\x72\x6f\x6d\x20\x41\x42\x43\x4d\x65\x74\x61\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x0a\x20\x20\x20\x20\x69\x6e\x73\x74\x61\x6e\x74\x69\x61\x74\x65\x64\x20\x75\x6e\x6c\x65\x73\x73\x20\x61\x6c\x6c\x20\x6f\x66\x20\x69\x74\x73\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x61\x72\x65\x20\x6f\x76\x65\x72\x72\x69\x64\x64\x65\x6e\x2e\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x63\x61\x6c\x6c\x65\x64\x20\x75\x73\x69\x6e\x67\x20\x61\x6e\x79\x20\x6f\x66\x20\x74\x68\x65\x20\x6e\x6f\x72\x6d\x61\x6c\x0a\x20\x20\x20\x20\x27\x73\x75\x70\x65\x72\x27\x20\x63\x61\x6c\x6c\x20\x6d\x65\x63\x68\x61\x6e\x69\x73\x6d\x73\x2e\x20\x20\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x28\x29\x20\x6d\x61\x79\x20\x62\x65\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x64\x65\x63\x6c\x61\x72\x65\x0a\x20\x20\x20\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x66\x6f\x72\x20\x70\x72\x6f\x70\x65\x72\x74\x69\x65\x73\x20\x61\x6e\x64\x20\x64\x65\x73\x63\x72\x69\x70\x74\x6f\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x55\x73\x61\x67\x65\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6c\x61\x73\x73\x20\x43\x28\x6d\x65\x74\x61\x63\x6c\x61\x73\x73\x3d\x41\x42\x43\x4d\x65\x74\x61\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x40\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x66\x20\x6d\x79\x5f\x61\x62\x73\x74\x72\x61\x63\x74\x5f\x6d\x65\x74\x68\x6f\x64\x28\x73\x65\x6c\x66\x2c\x20\x61\x72\x67\x31\x2c\x20\x61\x72\x67\x32\x2c\x20\x61\x72\x67\x4e\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x2e\x2e\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
abc_toplevel_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & abc_toplevel_consts_1_consts_0._ascii.ob_base,
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(__isabstractmethod__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
abc_toplevel_consts_1_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen abc>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_abstractmethod = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abstractmethod",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
abc_toplevel_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x22\x00\x24\x28\x80\x47\xd4\x04\x20\xd8\x0b\x12\x80\x4e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_funcobj = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "funcobj",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_funcobj._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(20)
abc_toplevel_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 10,
    },
    .co_consts = & abc_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 7,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 249,
    .co_localsplusnames = & abc_toplevel_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_abstractmethod._ascii.ob_base,
    .co_qualname = & const_str_abstractmethod._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_abstractclassmethod = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abstractclassmethod",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[265];
    }
abc_toplevel_consts_2_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 264,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x64\x65\x63\x6f\x72\x61\x74\x6f\x72\x20\x69\x6e\x64\x69\x63\x61\x74\x69\x6e\x67\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x63\x6c\x61\x73\x73\x6d\x65\x74\x68\x6f\x64\x73\x2e\x0a\x0a\x20\x20\x20\x20\x44\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2c\x20\x75\x73\x65\x20\x27\x63\x6c\x61\x73\x73\x6d\x65\x74\x68\x6f\x64\x27\x20\x77\x69\x74\x68\x20\x27\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x27\x20\x69\x6e\x73\x74\x65\x61\x64\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6c\x61\x73\x73\x20\x43\x28\x41\x42\x43\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x40\x63\x6c\x61\x73\x73\x6d\x65\x74\x68\x6f\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x40\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x66\x20\x6d\x79\x5f\x61\x62\x73\x74\x72\x61\x63\x74\x5f\x63\x6c\x61\x73\x73\x6d\x65\x74\x68\x6f\x64\x28\x63\x6c\x73\x2c\x20\x2e\x2e\x2e\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x2e\x2e\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
abc_toplevel_consts_2_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(__isabstractmethod__),
            & const_str_super._ascii.ob_base,
            &_Py_ID(__init__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
abc_toplevel_consts_2_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abstractclassmethod.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
abc_toplevel_consts_2_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x28\x2c\x88\x08\xd4\x08\x25\xdc\x08\x0d\x89\x07\xd1\x08\x18\x98\x18\xd5\x08\x22",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_callable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "callable",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
abc_toplevel_consts_2_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_callable._ascii.ob_base,
            &_Py_ID(__class__),
        },
    },
};
static
    struct _PyCode_DEF(50)
abc_toplevel_consts_2_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_2_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_2_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 43,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 250,
    .co_localsplusnames = & abc_toplevel_consts_2_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & abc_toplevel_consts_2_consts_3_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_2_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x64\x01\x7c\x01\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x89\x02\x7c\x00\x8d\x09\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
abc_toplevel_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_abstractclassmethod._ascii.ob_base,
            & abc_toplevel_consts_2_consts_1._ascii.ob_base,
            Py_True,
            & abc_toplevel_consts_2_consts_3.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
abc_toplevel_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__isabstractmethod__),
            &_Py_ID(__init__),
            &_Py_ID(__classcell__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[27];
    }
abc_toplevel_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 26,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x84\x00\xf1\x02\x0a\x05\x08\xf0\x18\x00\x1c\x20\xd0\x04\x18\xf7\x04\x02\x05\x23\xf0\x00\x02\x05\x23",
};
static
    struct _PyCode_DEF(38)
abc_toplevel_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & abc_toplevel_consts_2_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 28,
    .co_nlocalsplus = 1,
    .co_nlocals = 0,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 251,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[64]),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_abstractclassmethod._ascii.ob_base,
    .co_qualname = & const_str_abstractclassmethod._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x88\x00\x66\x01\x64\x03\x84\x08\x5a\x05\x88\x00\x78\x01\x5a\x06\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_abstractstaticmethod = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abstractstaticmethod",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[264];
    }
abc_toplevel_consts_4_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 263,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x64\x65\x63\x6f\x72\x61\x74\x6f\x72\x20\x69\x6e\x64\x69\x63\x61\x74\x69\x6e\x67\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x73\x74\x61\x74\x69\x63\x6d\x65\x74\x68\x6f\x64\x73\x2e\x0a\x0a\x20\x20\x20\x20\x44\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2c\x20\x75\x73\x65\x20\x27\x73\x74\x61\x74\x69\x63\x6d\x65\x74\x68\x6f\x64\x27\x20\x77\x69\x74\x68\x20\x27\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x27\x20\x69\x6e\x73\x74\x65\x61\x64\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6c\x61\x73\x73\x20\x43\x28\x41\x42\x43\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x40\x73\x74\x61\x74\x69\x63\x6d\x65\x74\x68\x6f\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x40\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x66\x20\x6d\x79\x5f\x61\x62\x73\x74\x72\x61\x63\x74\x5f\x73\x74\x61\x74\x69\x63\x6d\x65\x74\x68\x6f\x64\x28\x2e\x2e\x2e\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x2e\x2e\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
abc_toplevel_consts_4_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abstractstaticmethod.__init__",
};
static
    struct _PyCode_DEF(50)
abc_toplevel_consts_4_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_2_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_2_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 63,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 252,
    .co_localsplusnames = & abc_toplevel_consts_2_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & abc_toplevel_consts_4_consts_3_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_2_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x64\x01\x7c\x01\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x89\x02\x7c\x00\x8d\x09\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
abc_toplevel_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_abstractstaticmethod._ascii.ob_base,
            & abc_toplevel_consts_4_consts_1._ascii.ob_base,
            Py_True,
            & abc_toplevel_consts_4_consts_3.ob_base.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(38)
abc_toplevel_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & abc_toplevel_consts_4_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 48,
    .co_nlocalsplus = 1,
    .co_nlocals = 0,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 253,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[64]),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_abstractstaticmethod._ascii.ob_base,
    .co_qualname = & const_str_abstractstaticmethod._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x88\x00\x66\x01\x64\x03\x84\x08\x5a\x05\x88\x00\x78\x01\x5a\x06\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_abstractproperty = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abstractproperty",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[250];
    }
abc_toplevel_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 249,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x64\x65\x63\x6f\x72\x61\x74\x6f\x72\x20\x69\x6e\x64\x69\x63\x61\x74\x69\x6e\x67\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x70\x72\x6f\x70\x65\x72\x74\x69\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x44\x65\x70\x72\x65\x63\x61\x74\x65\x64\x2c\x20\x75\x73\x65\x20\x27\x70\x72\x6f\x70\x65\x72\x74\x79\x27\x20\x77\x69\x74\x68\x20\x27\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x27\x20\x69\x6e\x73\x74\x65\x61\x64\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6c\x61\x73\x73\x20\x43\x28\x41\x42\x43\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x40\x70\x72\x6f\x70\x65\x72\x74\x79\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x40\x61\x62\x73\x74\x72\x61\x63\x74\x6d\x65\x74\x68\x6f\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x66\x20\x6d\x79\x5f\x61\x62\x73\x74\x72\x61\x63\x74\x5f\x70\x72\x6f\x70\x65\x72\x74\x79\x28\x73\x65\x6c\x66\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x2e\x2e\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
abc_toplevel_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_abstractproperty._ascii.ob_base,
            & abc_toplevel_consts_6_consts_1._ascii.ob_base,
            Py_True,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
abc_toplevel_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__isabstractmethod__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
abc_toplevel_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x0a\x05\x08\xf0\x18\x00\x1c\x20\xd1\x04\x18",
};
static
    struct _PyCode_DEF(20)
abc_toplevel_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 10,
    },
    .co_consts = & abc_toplevel_consts_6_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 68,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 254,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_abstractproperty._ascii.ob_base,
    .co_qualname = & const_str_abstractproperty._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_get_cache_token = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_cache_token",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__abc_init = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_init",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__abc_register = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_register",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__abc_instancecheck = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_instancecheck",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__abc_subclasscheck = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_subclasscheck",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__get_dump = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_dump",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__reset_registry = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_reset_registry",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__reset_caches = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_reset_caches",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
abc_toplevel_consts_9 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_get_cache_token._ascii.ob_base,
            & const_str__abc_init._ascii.ob_base,
            & const_str__abc_register._ascii.ob_base,
            & const_str__abc_instancecheck._ascii.ob_base,
            & const_str__abc_subclasscheck._ascii.ob_base,
            & const_str__get_dump._ascii.ob_base,
            & const_str__reset_registry._ascii.ob_base,
            & const_str__reset_caches._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_ABCMeta = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[657];
    }
abc_toplevel_consts_10_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 656,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4d\x65\x74\x61\x63\x6c\x61\x73\x73\x20\x66\x6f\x72\x20\x64\x65\x66\x69\x6e\x69\x6e\x67\x20\x41\x62\x73\x74\x72\x61\x63\x74\x20\x42\x61\x73\x65\x20\x43\x6c\x61\x73\x73\x65\x73\x20\x28\x41\x42\x43\x73\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x55\x73\x65\x20\x74\x68\x69\x73\x20\x6d\x65\x74\x61\x63\x6c\x61\x73\x73\x20\x74\x6f\x20\x63\x72\x65\x61\x74\x65\x20\x61\x6e\x20\x41\x42\x43\x2e\x20\x20\x41\x6e\x20\x41\x42\x43\x20\x63\x61\x6e\x20\x62\x65\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x64\x69\x72\x65\x63\x74\x6c\x79\x2c\x20\x61\x6e\x64\x20\x74\x68\x65\x6e\x20\x61\x63\x74\x73\x20\x61\x73\x20\x61\x20\x6d\x69\x78\x2d\x69\x6e\x20\x63\x6c\x61\x73\x73\x2e\x20\x20\x59\x6f\x75\x20\x63\x61\x6e\x20\x61\x6c\x73\x6f\x20\x72\x65\x67\x69\x73\x74\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x75\x6e\x72\x65\x6c\x61\x74\x65\x64\x20\x63\x6f\x6e\x63\x72\x65\x74\x65\x20\x63\x6c\x61\x73\x73\x65\x73\x20\x28\x65\x76\x65\x6e\x20\x62\x75\x69\x6c\x74\x2d\x69\x6e\x20\x63\x6c\x61\x73\x73\x65\x73\x29\x20\x61\x6e\x64\x20\x75\x6e\x72\x65\x6c\x61\x74\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x41\x42\x43\x73\x20\x61\x73\x20\x27\x76\x69\x72\x74\x75\x61\x6c\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x65\x73\x27\x20\x2d\x2d\x20\x74\x68\x65\x73\x65\x20\x61\x6e\x64\x20\x74\x68\x65\x69\x72\x20\x64\x65\x73\x63\x65\x6e\x64\x61\x6e\x74\x73\x20\x77\x69\x6c\x6c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x62\x65\x20\x63\x6f\x6e\x73\x69\x64\x65\x72\x65\x64\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x65\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x72\x65\x67\x69\x73\x74\x65\x72\x69\x6e\x67\x20\x41\x42\x43\x20\x62\x79\x20\x74\x68\x65\x20\x62\x75\x69\x6c\x74\x2d\x69\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x73\x73\x75\x62\x63\x6c\x61\x73\x73\x28\x29\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2c\x20\x62\x75\x74\x20\x74\x68\x65\x20\x72\x65\x67\x69\x73\x74\x65\x72\x69\x6e\x67\x20\x41\x42\x43\x20\x77\x6f\x6e\x27\x74\x20\x73\x68\x6f\x77\x20\x75\x70\x20\x69\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x74\x68\x65\x69\x72\x20\x4d\x52\x4f\x20\x28\x4d\x65\x74\x68\x6f\x64\x20\x52\x65\x73\x6f\x6c\x75\x74\x69\x6f\x6e\x20\x4f\x72\x64\x65\x72\x29\x20\x6e\x6f\x72\x20\x77\x69\x6c\x6c\x20\x6d\x65\x74\x68\x6f\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x73\x20\x64\x65\x66\x69\x6e\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x72\x65\x67\x69\x73\x74\x65\x72\x69\x6e\x67\x20\x41\x42\x43\x20\x62\x65\x20\x63\x61\x6c\x6c\x61\x62\x6c\x65\x20\x28\x6e\x6f\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x76\x65\x6e\x20\x76\x69\x61\x20\x73\x75\x70\x65\x72\x28\x29\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
abc_toplevel_consts_10_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_super._ascii.ob_base,
            &_Py_ID(__new__),
            & const_str__abc_init._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
abc_toplevel_consts_10_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta.__new__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[41];
    }
abc_toplevel_consts_10_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 40,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x12\x17\x91\x27\x91\x2f\xa0\x24\xa8\x04\xa8\x65\xb0\x59\xd1\x12\x49\xc0\x26\xd1\x12\x49\x88\x43\xdc\x0c\x15\x90\x63\x8c\x4e\xd8\x13\x16\x88\x4a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_mcls = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mcls",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_bases = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bases",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_namespace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "namespace",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
abc_toplevel_consts_10_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_mcls._ascii.ob_base,
            &_Py_ID(name),
            & const_str_bases._ascii.ob_base,
            & const_str_namespace._ascii.ob_base,
            & const_str_kwargs._ascii.ob_base,
            & const_str_cls._ascii.ob_base,
            &_Py_ID(__class__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
abc_toplevel_consts_10_consts_2_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x20\x20\x20\x20\x80",
};
static
    struct _PyCode_DEF(68)
abc_toplevel_consts_10_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 34,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 11,
    .co_argcount = 4,
    .co_posonlyargcount = 4,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 105,
    .co_nlocalsplus = 7,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 255,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & abc_toplevel_consts_10_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__new__),
    .co_qualname = & abc_toplevel_consts_10_consts_2_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x89\x06\x7c\x00\x8d\x04\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7c\x03\x66\x04\x69\x00\x7c\x04\xa4\x01\x8e\x01\x7d\x05\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[124];
    }
abc_toplevel_consts_10_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 123,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x67\x69\x73\x74\x65\x72\x20\x61\x20\x76\x69\x72\x74\x75\x61\x6c\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x20\x6f\x66\x20\x61\x6e\x20\x41\x42\x43\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x2c\x20\x74\x6f\x20\x61\x6c\x6c\x6f\x77\x20\x75\x73\x61\x67\x65\x20\x61\x73\x20\x61\x20\x63\x6c\x61\x73\x73\x20\x64\x65\x63\x6f\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & abc_toplevel_consts_10_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__abc_register._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_register = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "register",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
abc_toplevel_consts_10_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta.register",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
abc_toplevel_consts_10_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x14\x21\xa0\x13\xa0\x68\xd3\x13\x2f\xd0\x0c\x2f",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_subclass = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "subclass",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
abc_toplevel_consts_10_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_subclass._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(26)
abc_toplevel_consts_10_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & abc_toplevel_consts_10_consts_3_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 110,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 256,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_register._ascii.ob_base,
    .co_qualname = & abc_toplevel_consts_10_consts_3_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
abc_toplevel_consts_10_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Override for isinstance(instance, cls).",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & abc_toplevel_consts_10_consts_4_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__abc_instancecheck._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
abc_toplevel_consts_10_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta.__instancecheck__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
abc_toplevel_consts_10_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x13\x25\xa0\x63\xa8\x38\xd3\x13\x34\xd0\x0c\x34",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_instance = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "instance",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
abc_toplevel_consts_10_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_instance._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(26)
abc_toplevel_consts_10_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & abc_toplevel_consts_10_consts_4_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 117,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 257,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__instancecheck__),
    .co_qualname = & abc_toplevel_consts_10_consts_4_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
abc_toplevel_consts_10_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Override for issubclass(subclass, cls).",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & abc_toplevel_consts_10_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__abc_subclasscheck._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
abc_toplevel_consts_10_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta.__subclasscheck__",
};
static
    struct _PyCode_DEF(26)
abc_toplevel_consts_10_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & abc_toplevel_consts_10_consts_5_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 121,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 258,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasscheck__),
    .co_qualname = & abc_toplevel_consts_10_consts_5_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
abc_toplevel_consts_10_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Debug helper to print the ABC registry.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
abc_toplevel_consts_10_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Class: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
abc_toplevel_consts_10_consts_6_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Inv. counter: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
abc_toplevel_consts_10_consts_6_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_registry: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
abc_toplevel_consts_10_consts_6_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_cache: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
abc_toplevel_consts_10_consts_6_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_negative_cache: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
abc_toplevel_consts_10_consts_6_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_negative_cache_version: ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
abc_toplevel_consts_10_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & abc_toplevel_consts_10_consts_6_consts_0._ascii.ob_base,
            & abc_toplevel_consts_10_consts_6_consts_1._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & importlib__bootstrap_toplevel_consts_25_consts_3._object.ob_base.ob_base,
            & abc_toplevel_consts_10_consts_6_consts_4._ascii.ob_base,
            & abc_toplevel_consts_10_consts_6_consts_5._ascii.ob_base,
            & abc_toplevel_consts_10_consts_6_consts_6._ascii.ob_base,
            & abc_toplevel_consts_10_consts_6_consts_7._ascii.ob_base,
            & abc_toplevel_consts_10_consts_6_consts_8._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
abc_toplevel_consts_10_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_print._ascii.ob_base,
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            & const_str_get_cache_token._ascii.ob_base,
            & const_str__get_dump._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__dump_registry = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_dump_registry",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
abc_toplevel_consts_10_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta._dump_registry",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[159];
    }
abc_toplevel_consts_10_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 158,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0c\x11\x90\x47\x98\x43\x9f\x4e\x99\x4e\xd0\x1b\x2b\xa8\x31\xa8\x53\xd7\x2d\x3d\xd1\x2d\x3d\xd0\x2c\x3e\xd0\x12\x3f\xc0\x64\xd5\x0c\x4b\xdc\x0c\x11\x90\x4e\xa4\x3f\xd3\x23\x34\xd0\x22\x35\xd0\x12\x36\xb8\x54\xd5\x0c\x42\xe4\x2c\x35\xb0\x63\xab\x4e\xf1\x03\x01\x0d\x2a\x88\x5d\x98\x4a\xd0\x28\x3b\xd8\x0d\x28\xdc\x0c\x11\x90\x4f\xa0\x4d\xd0\x23\x34\xd0\x12\x35\xb8\x44\xd5\x0c\x41\xdc\x0c\x11\x90\x4c\xa0\x1a\xa0\x0e\xd0\x12\x2f\xb0\x64\xd5\x0c\x3b\xdc\x0c\x11\xd0\x14\x29\xd0\x2a\x3d\xd0\x29\x40\xd0\x12\x41\xc8\x04\xd5\x0c\x4d\xdc\x0c\x11\xd0\x14\x31\xd0\x32\x4d\xd0\x31\x50\xd0\x12\x51\xd8\x17\x1b\xf6\x03\x01\x0d\x1d",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__abc_registry = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_registry",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__abc_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_cache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__abc_negative_cache = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_negative_cache",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
const_str__abc_negative_cache_version = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_negative_cache_version",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
abc_toplevel_consts_10_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            &_Py_ID(file),
            & const_str__abc_registry._ascii.ob_base,
            & const_str__abc_cache._ascii.ob_base,
            & const_str__abc_negative_cache._ascii.ob_base,
            & const_str__abc_negative_cache_version._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(290)
abc_toplevel_consts_10_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 145,
    },
    .co_consts = & abc_toplevel_consts_10_consts_6_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 125,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 259,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__dump_registry._ascii.ob_base,
    .co_qualname = & abc_toplevel_consts_10_consts_6_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x02\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x04\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x02\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x04\x00\x00\x7d\x02\x7d\x03\x7d\x04\x7d\x05\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x02\x9b\x02\x9d\x02\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x7c\x03\x9b\x02\x9d\x02\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x7c\x04\x9b\x02\x9d\x02\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x7c\x05\x9b\x02\x9d\x02\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x09",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
abc_toplevel_consts_10_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Clear the registry (for debugging or testing).",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
abc_toplevel_consts_10_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & abc_toplevel_consts_10_consts_7_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__reset_registry._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__abc_registry_clear = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_registry_clear",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
abc_toplevel_consts_10_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta._abc_registry_clear",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
abc_toplevel_consts_10_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0c\x1b\x98\x43\xd5\x0c\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(26)
abc_toplevel_consts_10_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & abc_toplevel_consts_10_consts_7_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 137,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 260,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__abc_registry_clear._ascii.ob_base,
    .co_qualname = & abc_toplevel_consts_10_consts_7_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
abc_toplevel_consts_10_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Clear the caches (for debugging or testing).",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
abc_toplevel_consts_10_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & abc_toplevel_consts_10_consts_8_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_10_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__reset_caches._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__abc_caches_clear = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc_caches_clear",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
abc_toplevel_consts_10_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABCMeta._abc_caches_clear",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
abc_toplevel_consts_10_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0c\x19\x98\x23\xd5\x0c\x1e",
};
static
    struct _PyCode_DEF(26)
abc_toplevel_consts_10_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & abc_toplevel_consts_10_consts_8_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 141,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 261,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__abc_caches_clear._ascii.ob_base,
    .co_qualname = & abc_toplevel_consts_10_consts_8_qualname._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
abc_toplevel_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_ABCMeta._ascii.ob_base,
            & abc_toplevel_consts_10_consts_1._ascii.ob_base,
            & abc_toplevel_consts_10_consts_2.ob_base.ob_base,
            & abc_toplevel_consts_10_consts_3.ob_base.ob_base,
            & abc_toplevel_consts_10_consts_4.ob_base.ob_base,
            & abc_toplevel_consts_10_consts_5.ob_base.ob_base,
            & abc_toplevel_consts_10_consts_6.ob_base.ob_base,
            & abc_toplevel_consts_10_consts_7.ob_base.ob_base,
            & abc_toplevel_consts_10_consts_8.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
abc_toplevel_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__new__),
            & const_str_register._ascii.ob_base,
            &_Py_ID(__instancecheck__),
            &_Py_ID(__subclasscheck__),
            & const_str__dump_registry._ascii.ob_base,
            & const_str__abc_registry_clear._ascii.ob_base,
            & const_str__abc_caches_clear._ascii.ob_base,
            &_Py_ID(__classcell__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[44];
    }
abc_toplevel_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 43,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x84\x00\xf1\x02\x0b\x09\x0c\xf4\x18\x03\x09\x17\xf2\x0a\x05\x09\x30\xf2\x0e\x02\x09\x35\xf2\x08\x02\x09\x35\xf3\x08\x0a\x09\x1d\xf2\x18\x02\x09\x21\xf6\x08\x02\x09\x1f",
};
static
    struct _PyCode_DEF(72)
abc_toplevel_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 36,
    },
    .co_consts = & abc_toplevel_consts_10_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 92,
    .co_nlocalsplus = 1,
    .co_nlocals = 0,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 262,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[64]),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_ABCMeta._ascii.ob_base,
    .co_qualname = & const_str_ABCMeta._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x88\x00\x66\x01\x64\x02\x84\x08\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x09\x64\x06\x84\x01\x5a\x08\x64\x07\x84\x00\x5a\x09\x64\x08\x84\x00\x5a\x0a\x88\x00\x78\x01\x5a\x0b\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
abc_toplevel_consts_12 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_ABCMeta._ascii.ob_base,
            & const_str_get_cache_token._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_abc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abc",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[668];
    }
abc_toplevel_consts_14_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 667,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x63\x61\x6c\x63\x75\x6c\x61\x74\x65\x20\x74\x68\x65\x20\x73\x65\x74\x20\x6f\x66\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x6f\x66\x20\x61\x6e\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x63\x6c\x61\x73\x73\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x61\x20\x63\x6c\x61\x73\x73\x20\x68\x61\x73\x20\x68\x61\x64\x20\x6f\x6e\x65\x20\x6f\x66\x20\x69\x74\x73\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x20\x61\x66\x74\x65\x72\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x63\x6c\x61\x73\x73\x20\x77\x61\x73\x20\x63\x72\x65\x61\x74\x65\x64\x2c\x20\x74\x68\x65\x20\x6d\x65\x74\x68\x6f\x64\x20\x77\x69\x6c\x6c\x20\x6e\x6f\x74\x20\x62\x65\x20\x63\x6f\x6e\x73\x69\x64\x65\x72\x65\x64\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x20\x75\x6e\x74\x69\x6c\x0a\x20\x20\x20\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x69\x73\x20\x63\x61\x6c\x6c\x65\x64\x2e\x20\x41\x6c\x74\x65\x72\x6e\x61\x74\x69\x76\x65\x6c\x79\x2c\x20\x69\x66\x20\x61\x20\x6e\x65\x77\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x20\x68\x61\x73\x20\x62\x65\x65\x6e\x0a\x20\x20\x20\x20\x61\x64\x64\x65\x64\x20\x74\x6f\x20\x74\x68\x65\x20\x63\x6c\x61\x73\x73\x2c\x20\x69\x74\x20\x77\x69\x6c\x6c\x20\x6f\x6e\x6c\x79\x20\x62\x65\x20\x63\x6f\x6e\x73\x69\x64\x65\x72\x65\x64\x20\x61\x6e\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x6d\x65\x74\x68\x6f\x64\x20\x6f\x66\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x63\x6c\x61\x73\x73\x20\x61\x66\x74\x65\x72\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x69\x73\x20\x63\x61\x6c\x6c\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x20\x63\x61\x6c\x6c\x65\x64\x20\x62\x65\x66\x6f\x72\x65\x20\x61\x6e\x79\x20\x75\x73\x65\x20\x69\x73\x20\x6d\x61\x64\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x63\x6c\x61\x73\x73\x2c\x0a\x20\x20\x20\x20\x75\x73\x75\x61\x6c\x6c\x79\x20\x69\x6e\x20\x63\x6c\x61\x73\x73\x20\x64\x65\x63\x6f\x72\x61\x74\x6f\x72\x73\x20\x74\x68\x61\x74\x20\x61\x64\x64\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x74\x6f\x20\x74\x68\x65\x20\x73\x75\x62\x6a\x65\x63\x74\x20\x63\x6c\x61\x73\x73\x2e\x0a\x0a\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x63\x6c\x73\x2c\x20\x74\x6f\x20\x61\x6c\x6c\x6f\x77\x20\x75\x73\x61\x67\x65\x20\x61\x73\x20\x61\x20\x63\x6c\x61\x73\x73\x20\x64\x65\x63\x6f\x72\x61\x74\x6f\x72\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x63\x6c\x73\x20\x69\x73\x20\x6e\x6f\x74\x20\x61\x6e\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x20\x6f\x66\x20\x41\x42\x43\x4d\x65\x74\x61\x2c\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x68\x69\x6e\x67\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
abc_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & abc_toplevel_consts_14_consts_0._ascii.ob_base,
            &_Py_ID(__abstractmethods__),
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            Py_None,
            &_Py_ID(__isabstractmethod__),
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_frozenset = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "frozenset",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
abc_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_hasattr._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            &_Py_ID(__bases__),
            &_Py_ID(getattr),
            &_Py_ID(add),
            &_Py_ID(__dict__),
            &_Py_ID(items),
            & const_str_frozenset._ascii.ob_base,
            &_Py_ID(__abstractmethods__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str_update_abstractmethods = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "update_abstractmethods",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[179];
    }
abc_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 178,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x20\x00\x0c\x13\x90\x33\xd0\x18\x2d\xd4\x0b\x2e\xf0\x08\x00\x10\x13\x88\x0a\xe4\x10\x13\x93\x05\x80\x49\xf0\x06\x00\x11\x14\x97\x0d\x94\x0d\x88\x04\xdc\x14\x1b\x98\x44\xd0\x22\x37\xb8\x12\xd6\x14\x3c\x88\x44\xdc\x14\x1b\x98\x43\xa0\x14\xa0\x74\xd3\x14\x2c\x88\x45\xdc\x0f\x16\x90\x75\xd0\x1e\x34\xb0\x65\xd5\x0f\x3c\xd8\x10\x19\x97\x0d\x91\x0d\x98\x64\xd5\x10\x23\xf1\x07\x00\x15\x3d\xf0\x03\x00\x11\x1e\xf0\x0c\x00\x18\x1b\x97\x7c\x91\x7c\xd7\x17\x29\xd1\x17\x29\xd6\x17\x2b\x89\x0b\x88\x04\x88\x65\xdc\x0b\x12\x90\x35\xd0\x1a\x30\xb0\x25\xd5\x0b\x38\xd8\x0c\x15\x8f\x4d\x89\x4d\x98\x24\xd5\x0c\x1f\xf0\x05\x00\x18\x2c\xf4\x06\x00\x1f\x28\xa8\x09\xd3\x1e\x32\x80\x43\xd4\x04\x1b\xd8\x0b\x0e\x80\x4a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_abstracts = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abstracts",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_scls = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "scls",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
abc_toplevel_consts_14_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_abstracts._ascii.ob_base,
            & const_str_scls._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(374)
abc_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 187,
    },
    .co_consts = & abc_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 146,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 263,
    .co_localsplusnames = & abc_toplevel_consts_14_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_update_abstractmethods._ascii.ob_base,
    .co_qualname = & const_str_update_abstractmethods._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x73\x02\x7c\x00\x53\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x40\x00\x00\x7d\x02\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x01\x64\x02\xab\x03\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x2e\x00\x00\x7d\x03\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x03\x64\x03\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x04\x64\x05\xab\x03\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x1e\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x30\x04\x00\x8c\x42\x04\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x24\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x04\x64\x05\xab\x03\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x14\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x26\x04\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_ABC = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ABC",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[87];
    }
abc_toplevel_consts_15_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 86,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x48\x65\x6c\x70\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x74\x68\x61\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x61\x20\x73\x74\x61\x6e\x64\x61\x72\x64\x20\x77\x61\x79\x20\x74\x6f\x20\x63\x72\x65\x61\x74\x65\x20\x61\x6e\x20\x41\x42\x43\x20\x75\x73\x69\x6e\x67\x0a\x20\x20\x20\x20\x69\x6e\x68\x65\x72\x69\x74\x61\x6e\x63\x65\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
abc_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_ABC._ascii.ob_base,
            & abc_toplevel_consts_15_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
abc_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
abc_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x02\x05\x08\xf0\x06\x00\x11\x13\x81\x49",
};
static
    struct _PyCode_DEF(20)
abc_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 10,
    },
    .co_consts = & abc_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 184,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 264,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_ABC._ascii.ob_base,
    .co_qualname = & const_str_ABC._ascii.ob_base,
    .co_linetable = & abc_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
abc_toplevel_consts_17 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(metaclass),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
abc_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            & abc_toplevel_consts_0._ascii.ob_base,
            & abc_toplevel_consts_1.ob_base.ob_base,
            & abc_toplevel_consts_2.ob_base.ob_base,
            & const_str_abstractclassmethod._ascii.ob_base,
            & abc_toplevel_consts_4.ob_base.ob_base,
            & const_str_abstractstaticmethod._ascii.ob_base,
            & abc_toplevel_consts_6.ob_base.ob_base,
            & const_str_abstractproperty._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & abc_toplevel_consts_9._object.ob_base.ob_base,
            & abc_toplevel_consts_10.ob_base.ob_base,
            & const_str_ABCMeta._ascii.ob_base,
            & abc_toplevel_consts_12._object.ob_base.ob_base,
            & const_str_abc._ascii.ob_base,
            & abc_toplevel_consts_14.ob_base.ob_base,
            & abc_toplevel_consts_15.ob_base.ob_base,
            & const_str_ABC._ascii.ob_base,
            & abc_toplevel_consts_17._object.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str__abc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_abc",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__py_abc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_py_abc",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[24];
        }_object;
    }
abc_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 24,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_abstractmethod._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            & const_str_abstractclassmethod._ascii.ob_base,
            & const_str_staticmethod._ascii.ob_base,
            & const_str_abstractstaticmethod._ascii.ob_base,
            & const_str_property._ascii.ob_base,
            & const_str_abstractproperty._ascii.ob_base,
            & const_str__abc._ascii.ob_base,
            & const_str_get_cache_token._ascii.ob_base,
            & const_str__abc_init._ascii.ob_base,
            & const_str__abc_register._ascii.ob_base,
            & const_str__abc_instancecheck._ascii.ob_base,
            & const_str__abc_subclasscheck._ascii.ob_base,
            & const_str__get_dump._ascii.ob_base,
            & const_str__reset_registry._ascii.ob_base,
            & const_str__reset_caches._ascii.ob_base,
            &_Py_ID(type),
            & const_str_ABCMeta._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__py_abc._ascii.ob_base,
            &_Py_ID(__module__),
            & const_str_update_abstractmethods._ascii.ob_base,
            & const_str_ABC._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[132];
    }
abc_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 131,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x08\x00\x01\x3a\xf2\x06\x12\x01\x13\xf4\x2a\x11\x01\x23\x98\x2b\xf4\x00\x11\x01\x23\xf4\x28\x11\x01\x23\x98\x3c\xf4\x00\x11\x01\x23\xf4\x28\x0d\x01\x20\x90\x78\xf4\x00\x0d\x01\x20\xf0\x20\x3b\x01\x1f\xf7\x02\x02\x05\x36\xf7\x00\x02\x05\x36\xf3\x00\x02\x05\x36\xf4\x0e\x33\x05\x1f\x90\x24\xf4\x00\x33\x05\x1f\xf2\x6c\x01\x23\x01\x0f\xf4\x4c\x01\x04\x01\x13\x90\x47\xf6\x00\x04\x01\x13\xf8\xf0\x41\x03\x00\x08\x13\xf2\x00\x02\x01\x1f\xdf\x04\x30\xd8\x19\x1e\x80\x47\xd6\x04\x16\xf0\x05\x02\x01\x1f\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
abc_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\xa8\x14\x41\x17\x00\xc1\x17\x14\x41\x2e\x03\xc1\x2d\x01\x41\x2e\x03",
};
static
    struct _PyCode_DEF(226)
abc_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 113,
    },
    .co_consts = & abc_toplevel_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & abc_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 265,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & abc_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & abc_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x84\x00\x5a\x01\x02\x00\x47\x00\x64\x02\x84\x00\x64\x03\x65\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x03\x02\x00\x47\x00\x64\x04\x84\x00\x64\x05\x65\x04\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x05\x02\x00\x47\x00\x64\x06\x84\x00\x64\x07\x65\x06\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x07\x09\x00\x64\x08\x64\x09\x6c\x08\x6d\x09\x5a\x09\x6d\x0a\x5a\x0a\x6d\x0b\x5a\x0b\x6d\x0c\x5a\x0c\x6d\x0d\x5a\x0d\x6d\x0e\x5a\x0e\x6d\x0f\x5a\x0f\x6d\x10\x5a\x10\x01\x00\x02\x00\x47\x00\x64\x0a\x84\x00\x64\x0b\x65\x11\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x12\x64\x0e\x84\x00\x5a\x16\x02\x00\x47\x00\x64\x0f\x84\x00\x64\x10\x65\x12\xac\x11\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x17\x79\x12\x23\x00\x65\x13\x24\x00\x72\x12\x01\x00\x64\x08\x64\x0c\x6c\x14\x6d\x12\x5a\x12\x6d\x09\x5a\x09\x01\x00\x64\x0d\x65\x12\x5f\x15\x00\x00\x00\x00\x00\x00\x00\x00\x59\x00\x8c\x26\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_abc_toplevel(void)
{
    return Py_NewRef((PyObject *) &abc_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[159];
    }
codecs_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 158,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x63\x6f\x64\x65\x63\x73\x20\x2d\x2d\x20\x50\x79\x74\x68\x6f\x6e\x20\x43\x6f\x64\x65\x63\x20\x52\x65\x67\x69\x73\x74\x72\x79\x2c\x20\x41\x50\x49\x20\x61\x6e\x64\x20\x68\x65\x6c\x70\x65\x72\x73\x2e\x0a\x0a\x0a\x57\x72\x69\x74\x74\x65\x6e\x20\x62\x79\x20\x4d\x61\x72\x63\x2d\x41\x6e\x64\x72\x65\x20\x4c\x65\x6d\x62\x75\x72\x67\x20\x28\x6d\x61\x6c\x40\x6c\x65\x6d\x62\x75\x72\x67\x2e\x63\x6f\x6d\x29\x2e\x0a\x0a\x28\x63\x29\x20\x43\x6f\x70\x79\x72\x69\x67\x68\x74\x20\x43\x4e\x52\x49\x2c\x20\x41\x6c\x6c\x20\x52\x69\x67\x68\x74\x73\x20\x52\x65\x73\x65\x72\x76\x65\x64\x2e\x20\x4e\x4f\x20\x57\x41\x52\x52\x41\x4e\x54\x59\x2e\x0a\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[42],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
codecs_toplevel_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Failed to load the builtin codecs: %s",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_lookup = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lookup",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_EncodedFile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "EncodedFile",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_BOM = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_BOM_BE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_BE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_BOM_LE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_LE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_BOM32_BE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM32_BE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_BOM32_LE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM32_LE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_BOM64_BE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM64_BE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_BOM64_LE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM64_LE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_BOM_UTF8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_UTF8",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_BOM_UTF16 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_UTF16",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_BOM_UTF16_LE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_UTF16_LE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_BOM_UTF16_BE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_UTF16_BE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_BOM_UTF32 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_UTF32",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_BOM_UTF32_LE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_UTF32_LE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_BOM_UTF32_BE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BOM_UTF32_BE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_CodecInfo = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "CodecInfo",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_Codec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Codec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_IncrementalEncoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalEncoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_IncrementalDecoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalDecoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_StreamReader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_StreamWriter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_StreamReaderWriter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_StreamRecoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_getencoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getencoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_getdecoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getdecoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_getincrementalencoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getincrementalencoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_getincrementaldecoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getincrementaldecoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_getreader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getreader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_getwriter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getwriter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_iterencode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "iterencode",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_iterdecode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "iterdecode",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_strict_errors = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "strict_errors",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_ignore_errors = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ignore_errors",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_replace_errors = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "replace_errors",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str_xmlcharrefreplace_errors = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "xmlcharrefreplace_errors",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str_backslashreplace_errors = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "backslashreplace_errors",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_namereplace_errors = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "namereplace_errors",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_register_error = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "register_error",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_lookup_error = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lookup_error",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[44];
        }_object;
    }
codecs_toplevel_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 44,
        },
        .ob_item = {
            & const_str_register._ascii.ob_base,
            & const_str_lookup._ascii.ob_base,
            &_Py_ID(open),
            & const_str_EncodedFile._ascii.ob_base,
            & const_str_BOM._ascii.ob_base,
            & const_str_BOM_BE._ascii.ob_base,
            & const_str_BOM_LE._ascii.ob_base,
            & const_str_BOM32_BE._ascii.ob_base,
            & const_str_BOM32_LE._ascii.ob_base,
            & const_str_BOM64_BE._ascii.ob_base,
            & const_str_BOM64_LE._ascii.ob_base,
            & const_str_BOM_UTF8._ascii.ob_base,
            & const_str_BOM_UTF16._ascii.ob_base,
            & const_str_BOM_UTF16_LE._ascii.ob_base,
            & const_str_BOM_UTF16_BE._ascii.ob_base,
            & const_str_BOM_UTF32._ascii.ob_base,
            & const_str_BOM_UTF32_LE._ascii.ob_base,
            & const_str_BOM_UTF32_BE._ascii.ob_base,
            & const_str_CodecInfo._ascii.ob_base,
            & const_str_Codec._ascii.ob_base,
            & const_str_IncrementalEncoder._ascii.ob_base,
            & const_str_IncrementalDecoder._ascii.ob_base,
            & const_str_StreamReader._ascii.ob_base,
            & const_str_StreamWriter._ascii.ob_base,
            & const_str_StreamReaderWriter._ascii.ob_base,
            & const_str_StreamRecoder._ascii.ob_base,
            & const_str_getencoder._ascii.ob_base,
            & const_str_getdecoder._ascii.ob_base,
            & const_str_getincrementalencoder._ascii.ob_base,
            & const_str_getincrementaldecoder._ascii.ob_base,
            & const_str_getreader._ascii.ob_base,
            & const_str_getwriter._ascii.ob_base,
            &_Py_ID(encode),
            &_Py_ID(decode),
            & const_str_iterencode._ascii.ob_base,
            & const_str_iterdecode._ascii.ob_base,
            & const_str_strict_errors._ascii.ob_base,
            & const_str_ignore_errors._ascii.ob_base,
            & const_str_replace_errors._ascii.ob_base,
            & const_str_xmlcharrefreplace_errors._ascii.ob_base,
            & const_str_backslashreplace_errors._ascii.ob_base,
            & const_str_namereplace_errors._ascii.ob_base,
            & const_str_register_error._ascii.ob_base,
            & const_str_lookup_error._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[4];
    }
codecs_toplevel_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 3,
    },
    .ob_shash = -1,
    .ob_sval = "\xef\xbb\xbf",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
codecs_toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\xff\xfe",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
codecs_toplevel_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\xfe\xff",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
codecs_toplevel_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\xff\xfe\x00\x00",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
codecs_toplevel_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x00\x00\xfe\xff",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[49];
    }
codecs_toplevel_consts_12_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 48,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Codec details when looking up the codec registry",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_12_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(_is_text_encoding),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_incrementalencoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "incrementalencoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_incrementaldecoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "incrementaldecoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_streamwriter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "streamwriter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_streamreader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "streamreader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
codecs_toplevel_consts_12_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_tuple._ascii.ob_base,
            &_Py_ID(__new__),
            &_Py_ID(name),
            &_Py_ID(encode),
            &_Py_ID(decode),
            & const_str_incrementalencoder._ascii.ob_base,
            & const_str_incrementaldecoder._ascii.ob_base,
            & const_str_streamwriter._ascii.ob_base,
            & const_str_streamreader._ascii.ob_base,
            &_Py_ID(_is_text_encoding),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
codecs_toplevel_consts_12_consts_5_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen codecs>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
codecs_toplevel_consts_12_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "CodecInfo.__new__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[102];
    }
codecs_toplevel_consts_12_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 101,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x10\x15\x8f\x7d\x89\x7d\x98\x53\xa0\x36\xa8\x36\xb0\x3c\xc0\x1c\xd0\x22\x4e\xd3\x0f\x4f\x88\x04\xd8\x14\x18\x88\x04\x8c\x09\xd8\x16\x1c\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8c\x0b\xd8\x22\x34\x88\x04\xd4\x08\x1f\xd8\x22\x34\x88\x04\xd4\x08\x1f\xd8\x1c\x28\x88\x04\xd4\x08\x19\xd8\x1c\x28\x88\x04\xd4\x08\x19\xd8\x0b\x1c\xd0\x0b\x28\xd8\x25\x36\x88\x44\xd4\x0c\x22\xd8\x0f\x13\x88\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
codecs_toplevel_consts_12_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            &_Py_ID(encode),
            &_Py_ID(decode),
            & const_str_streamreader._ascii.ob_base,
            & const_str_streamwriter._ascii.ob_base,
            & const_str_incrementalencoder._ascii.ob_base,
            & const_str_incrementaldecoder._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(_is_text_encoding),
            &_Py_ID(self),
        },
    },
};
static
    struct _PyCode_DEF(174)
codecs_toplevel_consts_12_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 87,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_12_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 8,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 94,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 266,
    .co_localsplusnames = & codecs_toplevel_consts_12_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__new__),
    .co_qualname = & codecs_toplevel_consts_12_consts_5_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_12_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7c\x03\x7c\x04\x66\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x09\x7c\x07\x7c\x09\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x09\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x09\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x09\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x09\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x09\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x09\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\x81\x07\x7c\x08\x7c\x09\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
codecs_toplevel_consts_12_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<%s.%s object for encoding %s at %#x>",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_12_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & codecs_toplevel_consts_12_consts_6_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
codecs_toplevel_consts_12_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(__class__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(name),
            &_Py_ID(id),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
codecs_toplevel_consts_12_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "CodecInfo.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[59];
    }
codecs_toplevel_consts_12_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 58,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x36\xd8\x11\x15\x97\x1e\x91\x1e\xd7\x11\x2a\xd1\x11\x2a\xa8\x44\xaf\x4e\xa9\x4e\xd7\x2c\x47\xd1\x2c\x47\xd8\x11\x15\x97\x19\x91\x19\x9c\x42\x98\x74\x9b\x48\xf0\x03\x01\x11\x26\xf1\x03\x02\x10\x26\xf0\x00\x02\x09\x26",
};
static
    struct _PyCode_DEF(138)
codecs_toplevel_consts_12_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 69,
    },
    .co_consts = & codecs_toplevel_consts_12_consts_6_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_12_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 109,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 267,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & codecs_toplevel_consts_12_consts_6_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_12_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x66\x04\x7a\x06\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
codecs_toplevel_consts_12_consts_7 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            Py_None,
            Py_None,
            Py_None,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
codecs_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_CodecInfo._ascii.ob_base,
            & codecs_toplevel_consts_12_consts_1._ascii.ob_base,
            Py_True,
            Py_None,
            & codecs_toplevel_consts_12_consts_4._object.ob_base.ob_base,
            & codecs_toplevel_consts_12_consts_5.ob_base.ob_base,
            & codecs_toplevel_consts_12_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_12_consts_7._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
codecs_toplevel_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(_is_text_encoding),
            &_Py_ID(__new__),
            &_Py_ID(__repr__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[38];
    }
codecs_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 37,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd9\x04\x3a\xf0\x10\x00\x19\x1d\xd0\x04\x15\xe0\x45\x49\xd8\x3f\x43\xf0\x03\x0d\x05\x14\xe0\x1d\x21\xf4\x05\x0d\x05\x14\xf3\x1e\x03\x05\x26",
};
static
    struct _PyCode_DEF(44)
codecs_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & codecs_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 83,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 268,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_CodecInfo._ascii.ob_base,
    .co_qualname = & const_str_CodecInfo._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x09\x00\x09\x00\x64\x07\x64\x03\x64\x04\x9c\x01\x64\x05\x84\x03\x5a\x05\x64\x06\x84\x00\x5a\x06\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1082];
    }
codecs_toplevel_consts_14_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1081,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x44\x65\x66\x69\x6e\x65\x73\x20\x74\x68\x65\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x66\x6f\x72\x20\x73\x74\x61\x74\x65\x6c\x65\x73\x73\x20\x65\x6e\x63\x6f\x64\x65\x72\x73\x2f\x64\x65\x63\x6f\x64\x65\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x2e\x65\x6e\x63\x6f\x64\x65\x28\x29\x2f\x2e\x64\x65\x63\x6f\x64\x65\x28\x29\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x6d\x61\x79\x20\x75\x73\x65\x20\x64\x69\x66\x66\x65\x72\x65\x6e\x74\x20\x65\x72\x72\x6f\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x73\x63\x68\x65\x6d\x65\x73\x20\x62\x79\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x73\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x2e\x20\x54\x68\x65\x73\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x72\x69\x6e\x67\x20\x76\x61\x6c\x75\x65\x73\x20\x61\x72\x65\x20\x70\x72\x65\x64\x65\x66\x69\x6e\x65\x64\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x2d\x20\x72\x61\x69\x73\x65\x20\x61\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x65\x72\x72\x6f\x72\x20\x28\x6f\x72\x20\x61\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x69\x67\x6e\x6f\x72\x65\x27\x20\x2d\x20\x69\x67\x6e\x6f\x72\x65\x20\x74\x68\x65\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x20\x61\x6e\x64\x20\x63\x6f\x6e\x74\x69\x6e\x75\x65\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x6e\x65\x78\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x2d\x20\x72\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x61\x20\x73\x75\x69\x74\x61\x62\x6c\x65\x20\x72\x65\x70\x6c\x61\x63\x65\x6d\x65\x6e\x74\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x3b\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x50\x79\x74\x68\x6f\x6e\x20\x77\x69\x6c\x6c\x20\x75\x73\x65\x20\x74\x68\x65\x20\x6f\x66\x66\x69\x63\x69\x61\x6c\x20\x55\x2b\x46\x46\x46\x44\x20\x52\x45\x50\x4c\x41\x43\x45\x4d\x45\x4e\x54\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x43\x48\x41\x52\x41\x43\x54\x45\x52\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x62\x75\x69\x6c\x74\x69\x6e\x20\x55\x6e\x69\x63\x6f\x64\x65\x20\x63\x6f\x64\x65\x63\x73\x20\x6f\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x27\x3f\x27\x20\x6f\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x73\x75\x72\x72\x6f\x67\x61\x74\x65\x65\x73\x63\x61\x70\x65\x27\x20\x2d\x20\x72\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x70\x72\x69\x76\x61\x74\x65\x20\x63\x6f\x64\x65\x20\x70\x6f\x69\x6e\x74\x73\x20\x55\x2b\x44\x43\x6e\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x78\x6d\x6c\x63\x68\x61\x72\x72\x65\x66\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x2d\x20\x52\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x61\x70\x70\x72\x6f\x70\x72\x69\x61\x74\x65\x20\x58\x4d\x4c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x20\x72\x65\x66\x65\x72\x65\x6e\x63\x65\x20\x28\x6f\x6e\x6c\x79\x20\x66\x6f\x72\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x62\x61\x63\x6b\x73\x6c\x61\x73\x68\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x20\x2d\x20\x52\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x62\x61\x63\x6b\x73\x6c\x61\x73\x68\x65\x64\x20\x65\x73\x63\x61\x70\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x6e\x61\x6d\x65\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x20\x20\x20\x20\x20\x20\x2d\x20\x52\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x5c\x4e\x7b\x2e\x2e\x2e\x7d\x20\x65\x73\x63\x61\x70\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x28\x6f\x6e\x6c\x79\x20\x66\x6f\x72\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x73\x65\x74\x20\x6f\x66\x20\x61\x6c\x6c\x6f\x77\x65\x64\x20\x76\x61\x6c\x75\x65\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x65\x78\x74\x65\x6e\x64\x65\x64\x20\x76\x69\x61\x20\x72\x65\x67\x69\x73\x74\x65\x72\x5f\x65\x72\x72\x6f\x72\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[548];
    }
codecs_toplevel_consts_14_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 547,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x45\x6e\x63\x6f\x64\x65\x73\x20\x74\x68\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x69\x6e\x70\x75\x74\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x61\x20\x74\x75\x70\x6c\x65\x20\x28\x6f\x75\x74\x70\x75\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x62\x6a\x65\x63\x74\x2c\x20\x6c\x65\x6e\x67\x74\x68\x20\x63\x6f\x6e\x73\x75\x6d\x65\x64\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x65\x72\x72\x6f\x72\x73\x20\x64\x65\x66\x69\x6e\x65\x73\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x74\x6f\x20\x61\x70\x70\x6c\x79\x2e\x20\x49\x74\x20\x64\x65\x66\x61\x75\x6c\x74\x73\x20\x74\x6f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x6d\x65\x74\x68\x6f\x64\x20\x6d\x61\x79\x20\x6e\x6f\x74\x20\x73\x74\x6f\x72\x65\x20\x73\x74\x61\x74\x65\x20\x69\x6e\x20\x74\x68\x65\x20\x43\x6f\x64\x65\x63\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x20\x55\x73\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x20\x66\x6f\x72\x20\x63\x6f\x64\x65\x63\x73\x20\x77\x68\x69\x63\x68\x20\x68\x61\x76\x65\x20\x74\x6f\x20\x6b\x65\x65\x70\x20\x73\x74\x61\x74\x65\x20\x69\x6e\x20\x6f\x72\x64\x65\x72\x20\x74\x6f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6d\x61\x6b\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x65\x66\x66\x69\x63\x69\x65\x6e\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x65\x6e\x63\x6f\x64\x65\x72\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x62\x6c\x65\x20\x74\x6f\x20\x68\x61\x6e\x64\x6c\x65\x20\x7a\x65\x72\x6f\x20\x6c\x65\x6e\x67\x74\x68\x20\x69\x6e\x70\x75\x74\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x74\x75\x72\x6e\x20\x61\x6e\x20\x65\x6d\x70\x74\x79\x20\x6f\x62\x6a\x65\x63\x74\x20\x6f\x66\x20\x74\x68\x65\x20\x6f\x75\x74\x70\x75\x74\x20\x6f\x62\x6a\x65\x63\x74\x20\x74\x79\x70\x65\x20\x69\x6e\x20\x74\x68\x69\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x69\x74\x75\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_14_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_14_consts_2_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_14_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_NotImplementedError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
codecs_toplevel_consts_14_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Codec.encode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
codecs_toplevel_consts_14_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x22\x00\x0f\x22\xd0\x08\x21",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_14_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(input),
            &_Py_ID(errors),
        },
    },
};
static
    struct _PyCode_DEF(14)
codecs_toplevel_consts_14_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & codecs_toplevel_consts_14_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 138,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 269,
    .co_localsplusnames = & codecs_toplevel_consts_14_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(encode),
    .co_qualname = & codecs_toplevel_consts_14_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_14_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[755];
    }
codecs_toplevel_consts_14_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 754,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x44\x65\x63\x6f\x64\x65\x73\x20\x74\x68\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x69\x6e\x70\x75\x74\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x61\x20\x74\x75\x70\x6c\x65\x20\x28\x6f\x75\x74\x70\x75\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x62\x6a\x65\x63\x74\x2c\x20\x6c\x65\x6e\x67\x74\x68\x20\x63\x6f\x6e\x73\x75\x6d\x65\x64\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x69\x6e\x70\x75\x74\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x6e\x20\x6f\x62\x6a\x65\x63\x74\x20\x77\x68\x69\x63\x68\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x74\x68\x65\x20\x62\x66\x5f\x67\x65\x74\x72\x65\x61\x64\x62\x75\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x62\x75\x66\x66\x65\x72\x20\x73\x6c\x6f\x74\x2e\x20\x50\x79\x74\x68\x6f\x6e\x20\x73\x74\x72\x69\x6e\x67\x73\x2c\x20\x62\x75\x66\x66\x65\x72\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x61\x6e\x64\x20\x6d\x65\x6d\x6f\x72\x79\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6d\x61\x70\x70\x65\x64\x20\x66\x69\x6c\x65\x73\x20\x61\x72\x65\x20\x65\x78\x61\x6d\x70\x6c\x65\x73\x20\x6f\x66\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x73\x6c\x6f\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x65\x72\x72\x6f\x72\x73\x20\x64\x65\x66\x69\x6e\x65\x73\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x74\x6f\x20\x61\x70\x70\x6c\x79\x2e\x20\x49\x74\x20\x64\x65\x66\x61\x75\x6c\x74\x73\x20\x74\x6f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x6d\x65\x74\x68\x6f\x64\x20\x6d\x61\x79\x20\x6e\x6f\x74\x20\x73\x74\x6f\x72\x65\x20\x73\x74\x61\x74\x65\x20\x69\x6e\x20\x74\x68\x65\x20\x43\x6f\x64\x65\x63\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x20\x55\x73\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x20\x66\x6f\x72\x20\x63\x6f\x64\x65\x63\x73\x20\x77\x68\x69\x63\x68\x20\x68\x61\x76\x65\x20\x74\x6f\x20\x6b\x65\x65\x70\x20\x73\x74\x61\x74\x65\x20\x69\x6e\x20\x6f\x72\x64\x65\x72\x20\x74\x6f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6d\x61\x6b\x65\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x20\x65\x66\x66\x69\x63\x69\x65\x6e\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x72\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x62\x6c\x65\x20\x74\x6f\x20\x68\x61\x6e\x64\x6c\x65\x20\x7a\x65\x72\x6f\x20\x6c\x65\x6e\x67\x74\x68\x20\x69\x6e\x70\x75\x74\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x74\x75\x72\x6e\x20\x61\x6e\x20\x65\x6d\x70\x74\x79\x20\x6f\x62\x6a\x65\x63\x74\x20\x6f\x66\x20\x74\x68\x65\x20\x6f\x75\x74\x70\x75\x74\x20\x6f\x62\x6a\x65\x63\x74\x20\x74\x79\x70\x65\x20\x69\x6e\x20\x74\x68\x69\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x69\x74\x75\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_14_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_14_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
codecs_toplevel_consts_14_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Codec.decode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
codecs_toplevel_consts_14_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x2a\x00\x0f\x22\xd0\x08\x21",
};
static
    struct _PyCode_DEF(14)
codecs_toplevel_consts_14_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & codecs_toplevel_consts_14_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 157,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 270,
    .co_localsplusnames = & codecs_toplevel_consts_14_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(decode),
    .co_qualname = & codecs_toplevel_consts_14_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_14_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_14_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(strict),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
codecs_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_Codec._ascii.ob_base,
            & codecs_toplevel_consts_14_consts_1._ascii.ob_base,
            & codecs_toplevel_consts_14_consts_2.ob_base.ob_base,
            & codecs_toplevel_consts_14_consts_3.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
codecs_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(encode),
            &_Py_ID(decode),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
codecs_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x15\x05\x08\xf3\x2c\x11\x05\x22\xf4\x26\x15\x05\x22",
};
static
    struct _PyCode_DEF(32)
codecs_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & codecs_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 114,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 271,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Codec._ascii.ob_base,
    .co_qualname = & const_str_Codec._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x05\x64\x02\x84\x01\x5a\x04\x64\x05\x64\x03\x84\x01\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[233];
    }
codecs_toplevel_consts_16_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 232,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x41\x6e\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x20\x65\x6e\x63\x6f\x64\x65\x73\x20\x61\x6e\x20\x69\x6e\x70\x75\x74\x20\x69\x6e\x20\x6d\x75\x6c\x74\x69\x70\x6c\x65\x20\x73\x74\x65\x70\x73\x2e\x20\x54\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x63\x61\x6e\x0a\x20\x20\x20\x20\x62\x65\x20\x70\x61\x73\x73\x65\x64\x20\x70\x69\x65\x63\x65\x20\x62\x79\x20\x70\x69\x65\x63\x65\x20\x74\x6f\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x65\x28\x29\x20\x6d\x65\x74\x68\x6f\x64\x2e\x20\x54\x68\x65\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x0a\x20\x20\x20\x20\x72\x65\x6d\x65\x6d\x62\x65\x72\x73\x20\x74\x68\x65\x20\x73\x74\x61\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x70\x72\x6f\x63\x65\x73\x73\x20\x62\x65\x74\x77\x65\x65\x6e\x20\x63\x61\x6c\x6c\x73\x20\x74\x6f\x20\x65\x6e\x63\x6f\x64\x65\x28\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[245];
    }
codecs_toplevel_consts_16_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 244,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x43\x72\x65\x61\x74\x65\x73\x20\x61\x6e\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x20\x6d\x61\x79\x20\x75\x73\x65\x20\x64\x69\x66\x66\x65\x72\x65\x6e\x74\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x73\x63\x68\x65\x6d\x65\x73\x20\x62\x79\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x73\x20\x6b\x65\x79\x77\x6f\x72\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x2e\x20\x53\x65\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x64\x6f\x63\x73\x74\x72\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x6f\x72\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x70\x6f\x73\x73\x69\x62\x6c\x65\x20\x76\x61\x6c\x75\x65\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_16_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & codecs_toplevel_consts_16_consts_2_consts_0._ascii.ob_base,
            &_Py_STR(empty),
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_16_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(errors),
            &_Py_ID(buffer),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_16_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalEncoder.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
codecs_toplevel_consts_16_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x10\x00\x17\x1d\x88\x04\x8c\x0b\xd8\x16\x18\x88\x04\x8d\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_16_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(errors),
        },
    },
};
static
    struct _PyCode_DEF(32)
codecs_toplevel_consts_16_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & codecs_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_16_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 186,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 272,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_16_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_16_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[66];
    }
codecs_toplevel_consts_16_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 65,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x45\x6e\x63\x6f\x64\x65\x73\x20\x69\x6e\x70\x75\x74\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x69\x6e\x67\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_16_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_16_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
codecs_toplevel_consts_16_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalEncoder.encode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
codecs_toplevel_consts_16_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x08\x00\x0f\x22\xd0\x08\x21",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_16_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(input),
            &_Py_ID(final),
        },
    },
};
static
    struct _PyCode_DEF(14)
codecs_toplevel_consts_16_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & codecs_toplevel_consts_16_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 197,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 273,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(encode),
    .co_qualname = & codecs_toplevel_consts_16_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_16_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[59];
    }
codecs_toplevel_consts_16_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 58,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x73\x65\x74\x73\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x65\x72\x20\x74\x6f\x20\x74\x68\x65\x20\x69\x6e\x69\x74\x69\x61\x6c\x20\x73\x74\x61\x74\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_16_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_16_consts_4_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
codecs_toplevel_consts_16_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalEncoder.reset",
};
static
    struct _PyCode_DEF(4)
codecs_toplevel_consts_16_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & codecs_toplevel_consts_16_consts_4_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 203,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 274,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_16_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[59];
    }
codecs_toplevel_consts_16_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 58,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x73\x74\x61\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x65\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_16_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_16_consts_5_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_16_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalEncoder.getstate",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
codecs_toplevel_consts_16_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x10\x11",
};
static
    struct _PyCode_DEF(4)
codecs_toplevel_consts_16_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & codecs_toplevel_consts_16_consts_5_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 208,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 275,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(getstate),
    .co_qualname = & codecs_toplevel_consts_16_consts_5_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_16_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[109];
    }
codecs_toplevel_consts_16_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 108,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x53\x65\x74\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x73\x74\x61\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x65\x72\x2e\x20\x73\x74\x61\x74\x65\x20\x6d\x75\x73\x74\x20\x68\x61\x76\x65\x20\x62\x65\x65\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x62\x79\x20\x67\x65\x74\x73\x74\x61\x74\x65\x28\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_16_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_16_consts_6_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_16_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalEncoder.setstate",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_16_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_state._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(4)
codecs_toplevel_consts_16_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & codecs_toplevel_consts_16_consts_6_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 214,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 276,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(setstate),
    .co_qualname = & codecs_toplevel_consts_16_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_16_consts_9 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
codecs_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_IncrementalEncoder._ascii.ob_base,
            & codecs_toplevel_consts_16_consts_1._ascii.ob_base,
            & codecs_toplevel_consts_16_consts_2.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_3.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_4.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_5.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_6.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_9._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
codecs_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            &_Py_ID(encode),
            &_Py_ID(reset),
            &_Py_ID(getstate),
            &_Py_ID(setstate),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
codecs_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x04\x05\x08\xf3\x0a\x09\x05\x19\xf3\x16\x04\x05\x22\xf2\x0c\x03\x05\x0c\xf2\x0a\x04\x05\x11\xf3\x0c\x04\x05\x0c",
};
static
    struct _PyCode_DEF(50)
codecs_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & codecs_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 180,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 277,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_IncrementalEncoder._ascii.ob_base,
    .co_qualname = & const_str_IncrementalEncoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x08\x64\x02\x84\x01\x5a\x04\x64\x09\x64\x03\x84\x01\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str_BufferedIncrementalEncoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalEncoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[193];
    }
codecs_toplevel_consts_18_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 192,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x20\x6f\x66\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x20\x63\x61\x6e\x20\x62\x65\x20\x75\x73\x65\x64\x20\x61\x73\x20\x74\x68\x65\x20\x62\x61\x73\x65\x63\x6c\x61\x73\x73\x20\x66\x6f\x72\x20\x61\x6e\x0a\x20\x20\x20\x20\x69\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x20\x65\x6e\x63\x6f\x64\x65\x72\x20\x69\x66\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x65\x72\x20\x6d\x75\x73\x74\x20\x6b\x65\x65\x70\x20\x73\x6f\x6d\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x6f\x75\x74\x70\x75\x74\x20\x69\x6e\x20\x61\x0a\x20\x20\x20\x20\x62\x75\x66\x66\x65\x72\x20\x62\x65\x74\x77\x65\x65\x6e\x20\x63\x61\x6c\x6c\x73\x20\x74\x6f\x20\x65\x6e\x63\x6f\x64\x65\x28\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_18_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_IncrementalEncoder._ascii.ob_base,
            &_Py_ID(__init__),
            &_Py_ID(buffer),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
codecs_toplevel_consts_18_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalEncoder.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
codecs_toplevel_consts_18_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x08\x1a\xd7\x08\x23\xd1\x08\x23\xa0\x44\xa8\x26\xd4\x08\x31\xe0\x16\x18\x88\x04\x8d\x0b",
};
static
    struct _PyCode_DEF(62)
codecs_toplevel_consts_18_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 226,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 278,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_18_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__buffer_encode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_buffer_encode",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
codecs_toplevel_consts_18_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalEncoder._buffer_encode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
codecs_toplevel_consts_18_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x0f\x22\xd0\x08\x21",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_18_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(input),
            &_Py_ID(errors),
            &_Py_ID(final),
        },
    },
};
static
    struct _PyCode_DEF(14)
codecs_toplevel_consts_18_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 231,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 279,
    .co_localsplusnames = & codecs_toplevel_consts_18_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__buffer_encode._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_18_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_18_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(buffer),
            & const_str__buffer_encode._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
codecs_toplevel_consts_18_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalEncoder.encode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[64];
    }
codecs_toplevel_consts_18_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 63,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\x98\x55\xd1\x0f\x22\x88\x04\xd8\x1d\x21\xd7\x1d\x30\xd1\x1d\x30\xb0\x14\xb0\x74\xb7\x7b\xb1\x7b\xc0\x45\xd3\x1d\x4a\xd1\x08\x1a\x88\x16\x90\x18\xe0\x16\x1a\x98\x38\x98\x39\x90\x6f\x88\x04\x8c\x0b\xd8\x0f\x15\x88\x0d",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_result = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "result",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_consumed = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "consumed",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
codecs_toplevel_consts_18_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(input),
            &_Py_ID(final),
            &_Py_ID(data),
            & const_str_result._ascii.ob_base,
            & const_str_consumed._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(120)
codecs_toplevel_consts_18_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 60,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 236,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 280,
    .co_localsplusnames = & codecs_toplevel_consts_18_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(encode),
    .co_qualname = & codecs_toplevel_consts_18_consts_4_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7a\x00\x00\x00\x7d\x03\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x04\x7d\x05\x7c\x03\x7c\x05\x64\x00\x1a\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_18_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_IncrementalEncoder._ascii.ob_base,
            &_Py_ID(reset),
            &_Py_ID(buffer),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
codecs_toplevel_consts_18_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalEncoder.reset",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
codecs_toplevel_consts_18_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x08\x1a\xd7\x08\x20\xd1\x08\x20\xa0\x14\xd4\x08\x26\xd8\x16\x18\x88\x04\x8d\x0b",
};
static
    struct _PyCode_DEF(60)
codecs_toplevel_consts_18_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 244,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 281,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_18_consts_5_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_18_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(buffer),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
codecs_toplevel_consts_18_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalEncoder.getstate",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
codecs_toplevel_consts_18_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\x8f\x7b\x89\x7b\xd2\x0f\x1f\x98\x61\xd0\x08\x1f",
};
static
    struct _PyCode_DEF(34)
codecs_toplevel_consts_18_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 17,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 248,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 282,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(getstate),
    .co_qualname = & codecs_toplevel_consts_18_consts_6_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x78\x01\x73\x02\x01\x00\x64\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
codecs_toplevel_consts_18_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalEncoder.setstate",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[14];
    }
codecs_toplevel_consts_18_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 13,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x16\x1b\x92\x6b\x98\x72\x88\x04\x8d\x0b",
};
static
    struct _PyCode_DEF(26)
codecs_toplevel_consts_18_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 251,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 283,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(setstate),
    .co_qualname = & codecs_toplevel_consts_18_consts_7_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x78\x01\x73\x02\x01\x00\x64\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
codecs_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_BufferedIncrementalEncoder._ascii.ob_base,
            & codecs_toplevel_consts_18_consts_1._ascii.ob_base,
            & codecs_toplevel_consts_18_consts_2.ob_base.ob_base,
            & codecs_toplevel_consts_18_consts_3.ob_base.ob_base,
            & codecs_toplevel_consts_18_consts_4.ob_base.ob_base,
            & codecs_toplevel_consts_18_consts_5.ob_base.ob_base,
            & codecs_toplevel_consts_18_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_18_consts_7.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_9._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
codecs_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            & const_str__buffer_encode._ascii.ob_base,
            &_Py_ID(encode),
            &_Py_ID(reset),
            &_Py_ID(getstate),
            &_Py_ID(setstate),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[38];
    }
codecs_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 37,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x04\x05\x08\xf3\x0a\x03\x05\x19\xf2\x0a\x03\x05\x22\xf3\x0a\x06\x05\x16\xf2\x10\x02\x05\x19\xf2\x08\x01\x05\x20\xf3\x06\x01\x05\x22",
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & codecs_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 220,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 284,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_BufferedIncrementalEncoder._ascii.ob_base,
    .co_qualname = & const_str_BufferedIncrementalEncoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x09\x64\x02\x84\x01\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x0a\x64\x04\x84\x01\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x64\x07\x84\x00\x5a\x09\x79\x08",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[233];
    }
codecs_toplevel_consts_20_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 232,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x41\x6e\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x20\x64\x65\x63\x6f\x64\x65\x73\x20\x61\x6e\x20\x69\x6e\x70\x75\x74\x20\x69\x6e\x20\x6d\x75\x6c\x74\x69\x70\x6c\x65\x20\x73\x74\x65\x70\x73\x2e\x20\x54\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x63\x61\x6e\x0a\x20\x20\x20\x20\x62\x65\x20\x70\x61\x73\x73\x65\x64\x20\x70\x69\x65\x63\x65\x20\x62\x79\x20\x70\x69\x65\x63\x65\x20\x74\x6f\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x28\x29\x20\x6d\x65\x74\x68\x6f\x64\x2e\x20\x54\x68\x65\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x0a\x20\x20\x20\x20\x72\x65\x6d\x65\x6d\x62\x65\x72\x73\x20\x74\x68\x65\x20\x73\x74\x61\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x20\x70\x72\x6f\x63\x65\x73\x73\x20\x62\x65\x74\x77\x65\x65\x6e\x20\x63\x61\x6c\x6c\x73\x20\x74\x6f\x20\x64\x65\x63\x6f\x64\x65\x28\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[244];
    }
codecs_toplevel_consts_20_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 243,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x43\x72\x65\x61\x74\x65\x20\x61\x6e\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x20\x6d\x61\x79\x20\x75\x73\x65\x20\x64\x69\x66\x66\x65\x72\x65\x6e\x74\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x73\x63\x68\x65\x6d\x65\x73\x20\x62\x79\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x73\x20\x6b\x65\x79\x77\x6f\x72\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x2e\x20\x53\x65\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x64\x6f\x63\x73\x74\x72\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x6f\x72\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x70\x6f\x73\x73\x69\x62\x6c\x65\x20\x76\x61\x6c\x75\x65\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_20_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_20_consts_2_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_20_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_20_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalDecoder.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
codecs_toplevel_consts_20_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x10\x00\x17\x1d\x88\x04\x8d\x0b",
};
static
    struct _PyCode_DEF(18)
codecs_toplevel_consts_20_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 9,
    },
    .co_consts = & codecs_toplevel_consts_20_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_20_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 260,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 285,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_20_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_20_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[65];
    }
codecs_toplevel_consts_20_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 64,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x44\x65\x63\x6f\x64\x65\x20\x69\x6e\x70\x75\x74\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x69\x6e\x67\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_20_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_20_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
codecs_toplevel_consts_20_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalDecoder.decode",
};
static
    struct _PyCode_DEF(14)
codecs_toplevel_consts_20_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & codecs_toplevel_consts_20_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 270,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 286,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(decode),
    .co_qualname = & codecs_toplevel_consts_20_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_16_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[58];
    }
codecs_toplevel_consts_20_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 57,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x73\x65\x74\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x72\x20\x74\x6f\x20\x74\x68\x65\x20\x69\x6e\x69\x74\x69\x61\x6c\x20\x73\x74\x61\x74\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_20_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_20_consts_4_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
codecs_toplevel_consts_20_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalDecoder.reset",
};
static
    struct _PyCode_DEF(4)
codecs_toplevel_consts_20_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & codecs_toplevel_consts_20_consts_4_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 276,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 287,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_20_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[522];
    }
codecs_toplevel_consts_20_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 521,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x73\x74\x61\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x28\x62\x75\x66\x66\x65\x72\x65\x64\x5f\x69\x6e\x70\x75\x74\x2c\x20\x61\x64\x64\x69\x74\x69\x6f\x6e\x61\x6c\x5f\x73\x74\x61\x74\x65\x5f\x69\x6e\x66\x6f\x29\x20\x74\x75\x70\x6c\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x62\x75\x66\x66\x65\x72\x65\x64\x5f\x69\x6e\x70\x75\x74\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x62\x79\x74\x65\x73\x20\x6f\x62\x6a\x65\x63\x74\x20\x63\x6f\x6e\x74\x61\x69\x6e\x69\x6e\x67\x20\x62\x79\x74\x65\x73\x20\x74\x68\x61\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x77\x65\x72\x65\x20\x70\x61\x73\x73\x65\x64\x20\x74\x6f\x20\x64\x65\x63\x6f\x64\x65\x28\x29\x20\x74\x68\x61\x74\x20\x68\x61\x76\x65\x20\x6e\x6f\x74\x20\x79\x65\x74\x20\x62\x65\x65\x6e\x20\x63\x6f\x6e\x76\x65\x72\x74\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x64\x64\x69\x74\x69\x6f\x6e\x61\x6c\x5f\x73\x74\x61\x74\x65\x5f\x69\x6e\x66\x6f\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x6e\x6f\x6e\x2d\x6e\x65\x67\x61\x74\x69\x76\x65\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x65\x20\x73\x74\x61\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x72\x20\x57\x49\x54\x48\x4f\x55\x54\x20\x79\x65\x74\x20\x68\x61\x76\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x6f\x63\x65\x73\x73\x65\x64\x20\x74\x68\x65\x20\x63\x6f\x6e\x74\x65\x6e\x74\x73\x20\x6f\x66\x20\x62\x75\x66\x66\x65\x72\x65\x64\x5f\x69\x6e\x70\x75\x74\x2e\x20\x20\x49\x6e\x20\x74\x68\x65\x20\x69\x6e\x69\x74\x69\x61\x6c\x20\x73\x74\x61\x74\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x6e\x64\x20\x61\x66\x74\x65\x72\x20\x72\x65\x73\x65\x74\x28\x29\x2c\x20\x67\x65\x74\x73\x74\x61\x74\x65\x28\x29\x20\x6d\x75\x73\x74\x20\x72\x65\x74\x75\x72\x6e\x20\x28\x62\x22\x22\x2c\x20\x30\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_20_consts_5_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_20_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_20_consts_5_consts_0._ascii.ob_base,
            & codecs_toplevel_consts_20_consts_5_consts_1._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_20_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalDecoder.getstate",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
codecs_toplevel_consts_20_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x18\x00\x10\x18",
};
static
    struct _PyCode_DEF(4)
codecs_toplevel_consts_20_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & codecs_toplevel_consts_20_consts_5_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 281,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 288,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(getstate),
    .co_qualname = & codecs_toplevel_consts_20_consts_5_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_20_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[183];
    }
codecs_toplevel_consts_20_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 182,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x53\x65\x74\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x73\x74\x61\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x61\x74\x65\x20\x6d\x75\x73\x74\x20\x68\x61\x76\x65\x20\x62\x65\x65\x6e\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x62\x79\x20\x67\x65\x74\x73\x74\x61\x74\x65\x28\x29\x2e\x20\x20\x54\x68\x65\x20\x65\x66\x66\x65\x63\x74\x20\x6f\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x65\x74\x73\x74\x61\x74\x65\x28\x28\x62\x22\x22\x2c\x20\x30\x29\x29\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x65\x71\x75\x69\x76\x61\x6c\x65\x6e\x74\x20\x74\x6f\x20\x72\x65\x73\x65\x74\x28\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_20_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_20_consts_6_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_20_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IncrementalDecoder.setstate",
};
static
    struct _PyCode_DEF(4)
codecs_toplevel_consts_20_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & codecs_toplevel_consts_20_consts_6_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 295,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 289,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(setstate),
    .co_qualname = & codecs_toplevel_consts_20_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib__bootstrap_external_toplevel_consts_54_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
codecs_toplevel_consts_20_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_IncrementalDecoder._ascii.ob_base,
            & codecs_toplevel_consts_20_consts_1._ascii.ob_base,
            & codecs_toplevel_consts_20_consts_2.ob_base.ob_base,
            & codecs_toplevel_consts_20_consts_3.ob_base.ob_base,
            & codecs_toplevel_consts_20_consts_4.ob_base.ob_base,
            & codecs_toplevel_consts_20_consts_5.ob_base.ob_base,
            & codecs_toplevel_consts_20_consts_6.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_9._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
codecs_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            &_Py_ID(decode),
            &_Py_ID(reset),
            &_Py_ID(getstate),
            &_Py_ID(setstate),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
codecs_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x04\x05\x08\xf3\x0a\x08\x05\x1d\xf3\x14\x04\x05\x22\xf2\x0c\x03\x05\x0c\xf2\x0a\x0c\x05\x18\xf3\x1c\x06\x05\x0c",
};
static
    struct _PyCode_DEF(50)
codecs_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & codecs_toplevel_consts_20_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 254,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 290,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_IncrementalDecoder._ascii.ob_base,
    .co_qualname = & const_str_IncrementalDecoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x08\x64\x02\x84\x01\x5a\x04\x64\x09\x64\x03\x84\x01\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str_BufferedIncrementalDecoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalDecoder",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[175];
    }
codecs_toplevel_consts_22_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 174,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x20\x6f\x66\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x20\x63\x61\x6e\x20\x62\x65\x20\x75\x73\x65\x64\x20\x61\x73\x20\x74\x68\x65\x20\x62\x61\x73\x65\x63\x6c\x61\x73\x73\x20\x66\x6f\x72\x20\x61\x6e\x0a\x20\x20\x20\x20\x69\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x20\x64\x65\x63\x6f\x64\x65\x72\x20\x69\x66\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x72\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x62\x6c\x65\x20\x74\x6f\x20\x68\x61\x6e\x64\x6c\x65\x20\x69\x6e\x63\x6f\x6d\x70\x6c\x65\x74\x65\x0a\x20\x20\x20\x20\x62\x79\x74\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x73\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_22_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_empty),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_22_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_IncrementalDecoder._ascii.ob_base,
            &_Py_ID(__init__),
            &_Py_ID(buffer),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
codecs_toplevel_consts_22_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalDecoder.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
codecs_toplevel_consts_22_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x08\x1a\xd7\x08\x23\xd1\x08\x23\xa0\x44\xa8\x26\xd4\x08\x31\xe0\x16\x19\x88\x04\x8d\x0b",
};
static
    struct _PyCode_DEF(62)
codecs_toplevel_consts_22_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & codecs_toplevel_consts_22_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_22_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 309,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 291,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_22_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_22_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__buffer_decode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_buffer_decode",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
codecs_toplevel_consts_22_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalDecoder._buffer_decode",
};
static
    struct _PyCode_DEF(14)
codecs_toplevel_consts_22_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 314,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 292,
    .co_localsplusnames = & codecs_toplevel_consts_18_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__buffer_decode._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_22_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_22_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(buffer),
            & const_str__buffer_decode._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
codecs_toplevel_consts_22_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalDecoder.decode",
};
static
    struct _PyCode_DEF(120)
codecs_toplevel_consts_22_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 60,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_22_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 319,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 293,
    .co_localsplusnames = & codecs_toplevel_consts_18_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(decode),
    .co_qualname = & codecs_toplevel_consts_22_consts_4_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7a\x00\x00\x00\x7d\x03\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x04\x7d\x05\x7c\x03\x7c\x05\x64\x00\x1a\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_22_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_IncrementalDecoder._ascii.ob_base,
            &_Py_ID(reset),
            &_Py_ID(buffer),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
codecs_toplevel_consts_22_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalDecoder.reset",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
codecs_toplevel_consts_22_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x08\x1a\xd7\x08\x20\xd1\x08\x20\xa0\x14\xd4\x08\x26\xd8\x16\x19\x88\x04\x8d\x0b",
};
static
    struct _PyCode_DEF(60)
codecs_toplevel_consts_22_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & codecs_toplevel_consts_22_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_22_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 327,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 294,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_22_consts_5_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_22_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
codecs_toplevel_consts_22_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalDecoder.getstate",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
codecs_toplevel_consts_22_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x10\x14\x97\x0b\x91\x0b\x98\x51\xd0\x0f\x1f\xd0\x08\x1f",
};
static
    struct _PyCode_DEF(30)
codecs_toplevel_consts_22_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 331,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 295,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(getstate),
    .co_qualname = & codecs_toplevel_consts_22_consts_6_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_22_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
codecs_toplevel_consts_22_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIncrementalDecoder.setstate",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[14];
    }
codecs_toplevel_consts_22_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 13,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x16\x1b\x98\x41\x91\x68\x88\x04\x8d\x0b",
};
static
    struct _PyCode_DEF(24)
codecs_toplevel_consts_22_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_18_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 335,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 296,
    .co_localsplusnames = & codecs_toplevel_consts_16_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(setstate),
    .co_qualname = & codecs_toplevel_consts_22_consts_7_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_22_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x64\x01\x19\x00\x00\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
codecs_toplevel_consts_22_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_BufferedIncrementalDecoder._ascii.ob_base,
            & codecs_toplevel_consts_22_consts_1._ascii.ob_base,
            & codecs_toplevel_consts_22_consts_2.ob_base.ob_base,
            & codecs_toplevel_consts_22_consts_3.ob_base.ob_base,
            & codecs_toplevel_consts_22_consts_4.ob_base.ob_base,
            & codecs_toplevel_consts_22_consts_5.ob_base.ob_base,
            & codecs_toplevel_consts_22_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_22_consts_7.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_16_consts_9._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
codecs_toplevel_consts_22_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            & const_str__buffer_decode._ascii.ob_base,
            &_Py_ID(decode),
            &_Py_ID(reset),
            &_Py_ID(getstate),
            &_Py_ID(setstate),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[38];
    }
codecs_toplevel_consts_22_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 37,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x04\x05\x08\xf3\x0a\x03\x05\x1a\xf2\x0a\x03\x05\x22\xf3\x0a\x06\x05\x16\xf2\x10\x02\x05\x1a\xf2\x08\x02\x05\x20\xf3\x08\x02\x05\x1f",
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & codecs_toplevel_consts_22_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_22_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 303,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 297,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_BufferedIncrementalDecoder._ascii.ob_base,
    .co_qualname = & const_str_BufferedIncrementalDecoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x09\x64\x02\x84\x01\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x0a\x64\x04\x84\x01\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x64\x07\x84\x00\x5a\x09\x79\x08",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[888];
    }
codecs_toplevel_consts_24_consts_1_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 887,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x43\x72\x65\x61\x74\x65\x73\x20\x61\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x72\x65\x61\x6d\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x66\x69\x6c\x65\x2d\x6c\x69\x6b\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x6f\x70\x65\x6e\x20\x66\x6f\x72\x20\x77\x72\x69\x74\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x20\x6d\x61\x79\x20\x75\x73\x65\x20\x64\x69\x66\x66\x65\x72\x65\x6e\x74\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x63\x68\x65\x6d\x65\x73\x20\x62\x79\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x73\x20\x6b\x65\x79\x77\x6f\x72\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x2e\x20\x54\x68\x65\x73\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x73\x20\x61\x72\x65\x20\x70\x72\x65\x64\x65\x66\x69\x6e\x65\x64\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x2d\x20\x72\x61\x69\x73\x65\x20\x61\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x28\x6f\x72\x20\x61\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x69\x67\x6e\x6f\x72\x65\x27\x20\x2d\x20\x69\x67\x6e\x6f\x72\x65\x20\x74\x68\x65\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x20\x61\x6e\x64\x20\x63\x6f\x6e\x74\x69\x6e\x75\x65\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x6e\x65\x78\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x72\x65\x70\x6c\x61\x63\x65\x27\x2d\x20\x72\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x61\x20\x73\x75\x69\x74\x61\x62\x6c\x65\x20\x72\x65\x70\x6c\x61\x63\x65\x6d\x65\x6e\x74\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x78\x6d\x6c\x63\x68\x61\x72\x72\x65\x66\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x2d\x20\x52\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x61\x70\x70\x72\x6f\x70\x72\x69\x61\x74\x65\x20\x58\x4d\x4c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x20\x72\x65\x66\x65\x72\x65\x6e\x63\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x62\x61\x63\x6b\x73\x6c\x61\x73\x68\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x20\x2d\x20\x52\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x62\x61\x63\x6b\x73\x6c\x61\x73\x68\x65\x64\x20\x65\x73\x63\x61\x70\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x6e\x61\x6d\x65\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x20\x20\x20\x20\x20\x20\x2d\x20\x52\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x5c\x4e\x7b\x2e\x2e\x2e\x7d\x20\x65\x73\x63\x61\x70\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x73\x65\x74\x20\x6f\x66\x20\x61\x6c\x6c\x6f\x77\x65\x64\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x20\x76\x61\x6c\x75\x65\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x65\x78\x74\x65\x6e\x64\x65\x64\x20\x76\x69\x61\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x67\x69\x73\x74\x65\x72\x5f\x65\x72\x72\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_24_consts_1_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_stream = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "stream",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
codecs_toplevel_consts_24_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
codecs_toplevel_consts_24_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x2c\x00\x17\x1d\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8d\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_24_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_stream._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct _PyCode_DEF(32)
codecs_toplevel_consts_24_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 348,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 298,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_24_consts_1_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[63];
    }
codecs_toplevel_consts_24_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 62,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x57\x72\x69\x74\x65\x73\x20\x74\x68\x65\x20\x6f\x62\x6a\x65\x63\x74\x27\x73\x20\x63\x6f\x6e\x74\x65\x6e\x74\x73\x20\x65\x6e\x63\x6f\x64\x65\x64\x20\x74\x6f\x20\x73\x65\x6c\x66\x2e\x73\x74\x72\x65\x61\x6d\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_24_consts_2_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_24_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(encode),
            &_Py_ID(errors),
            & const_str_stream._ascii.ob_base,
            &_Py_ID(write),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
codecs_toplevel_consts_24_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.write",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[47];
    }
codecs_toplevel_consts_24_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 46,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x1a\x1e\x9f\x1b\x99\x1b\xa0\x56\xa8\x54\xaf\x5b\xa9\x5b\xd3\x19\x39\x89\x0e\x88\x04\x88\x68\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x19\xd1\x08\x19\x98\x24\xd5\x08\x1f",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_24_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(object),
            &_Py_ID(data),
            & const_str_consumed._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(120)
codecs_toplevel_consts_24_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 60,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 373,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 299,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(write),
    .co_qualname = & codecs_toplevel_consts_24_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[92];
    }
codecs_toplevel_consts_24_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 91,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x57\x72\x69\x74\x65\x73\x20\x74\x68\x65\x20\x63\x6f\x6e\x63\x61\x74\x65\x6e\x61\x74\x65\x64\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x73\x74\x72\x69\x6e\x67\x73\x20\x74\x6f\x20\x74\x68\x65\x20\x73\x74\x72\x65\x61\x6d\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x75\x73\x69\x6e\x67\x20\x2e\x77\x72\x69\x74\x65\x28\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_24_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & codecs_toplevel_consts_24_consts_3_consts_0._ascii.ob_base,
            &_Py_STR(empty),
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(write),
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_writelines = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "writelines",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
codecs_toplevel_consts_24_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.writelines",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
codecs_toplevel_consts_24_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x09\x0d\x8f\x0a\x89\x0a\x90\x32\x97\x37\x91\x37\x98\x34\x93\x3d\xd5\x08\x21",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_list._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(68)
codecs_toplevel_consts_24_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 34,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 380,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 300,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_writelines._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_24_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[307];
    }
codecs_toplevel_consts_24_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 306,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x52\x65\x73\x65\x74\x73\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x62\x75\x66\x66\x65\x72\x73\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x6b\x65\x65\x70\x69\x6e\x67\x20\x69\x6e\x74\x65\x72\x6e\x61\x6c\x20\x73\x74\x61\x74\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x43\x61\x6c\x6c\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x73\x68\x6f\x75\x6c\x64\x20\x65\x6e\x73\x75\x72\x65\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x64\x61\x74\x61\x20\x6f\x6e\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x75\x74\x70\x75\x74\x20\x69\x73\x20\x70\x75\x74\x20\x69\x6e\x74\x6f\x20\x61\x20\x63\x6c\x65\x61\x6e\x20\x73\x74\x61\x74\x65\x2c\x20\x74\x68\x61\x74\x20\x61\x6c\x6c\x6f\x77\x73\x20\x61\x70\x70\x65\x6e\x64\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x66\x20\x6e\x65\x77\x20\x66\x72\x65\x73\x68\x20\x64\x61\x74\x61\x20\x77\x69\x74\x68\x6f\x75\x74\x20\x68\x61\x76\x69\x6e\x67\x20\x74\x6f\x20\x72\x65\x73\x63\x61\x6e\x20\x74\x68\x65\x20\x77\x68\x6f\x6c\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x72\x65\x61\x6d\x20\x74\x6f\x20\x72\x65\x63\x6f\x76\x65\x72\x20\x73\x74\x61\x74\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_24_consts_4_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
codecs_toplevel_consts_24_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.reset",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
codecs_toplevel_consts_24_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x14\x00\x09\x0d",
};
static
    struct _PyCode_DEF(4)
codecs_toplevel_consts_24_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_4_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 387,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 301,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_24_consts_4_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_24_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
            &_Py_ID(seek),
            &_Py_ID(reset),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
codecs_toplevel_consts_24_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.seek",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
codecs_toplevel_consts_24_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x18\xd1\x08\x18\x98\x16\xa0\x16\xd4\x08\x28\xd8\x0b\x11\x90\x51\x8a\x3b\x98\x36\xa0\x51\x9a\x3b\xd8\x0c\x10\x8f\x4a\x89\x4a\x8d\x4c\xf0\x03\x00\x1c\x27\x88\x3b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_whence = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "whence",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_24_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(offset),
            & const_str_whence._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(116)
codecs_toplevel_consts_24_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 58,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 399,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 302,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(seek),
    .co_qualname = & codecs_toplevel_consts_24_consts_5_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x64\x01\x6b\x28\x00\x00\x72\x17\x7c\x01\x64\x01\x6b\x28\x00\x00\x72\x11\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x79\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[64];
    }
codecs_toplevel_consts_24_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 63,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x49\x6e\x68\x65\x72\x69\x74\x20\x61\x6c\x6c\x20\x6f\x74\x68\x65\x72\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x75\x6e\x64\x65\x72\x6c\x79\x69\x6e\x67\x20\x73\x74\x72\x65\x61\x6d\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_24_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_24_consts_6_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_24_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
codecs_toplevel_consts_24_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.__getattr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
codecs_toplevel_consts_24_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf1\x0a\x00\x10\x17\x90\x74\x97\x7b\x91\x7b\xa0\x44\xd3\x0f\x29\xd0\x08\x29",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_24_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
            &_Py_ID(getattr),
        },
    },
};
static
    struct _PyCode_DEF(40)
codecs_toplevel_consts_24_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_6_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 404,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 303,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getattr__),
    .co_qualname = & codecs_toplevel_consts_24_consts_6_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x7c\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_24_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
codecs_toplevel_consts_24_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\x88\x0b",
};
static
    struct _PyCode_DEF(6)
codecs_toplevel_consts_24_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 411,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 304,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & codecs_toplevel_consts_24_consts_7_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
            &_Py_ID(close),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
codecs_toplevel_consts_24_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.__exit__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
codecs_toplevel_consts_24_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x19\xd1\x08\x19\xd5\x08\x1b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_tb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tb",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_24_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(type),
            &_Py_ID(value),
            & const_str_tb._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_24_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 414,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 305,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & codecs_toplevel_consts_24_consts_8_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
codecs_toplevel_consts_24_consts_9_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "can't serialize %s",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & codecs_toplevel_consts_24_consts_9_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_24_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(__class__),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
codecs_toplevel_consts_24_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamWriter.__reduce_ex__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
codecs_toplevel_consts_24_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0e\x17\xd0\x18\x2c\xa8\x74\xaf\x7e\xa9\x7e\xd7\x2f\x46\xd1\x2f\x46\xd1\x18\x46\xd3\x0e\x47\xd0\x08\x47",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_24_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(proto),
        },
    },
};
static
    struct _PyCode_DEF(70)
codecs_toplevel_consts_24_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_9_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 417,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 306,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reduce_ex__),
    .co_qualname = & codecs_toplevel_consts_24_consts_9_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_24_consts_12 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
codecs_toplevel_consts_24_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_StreamWriter._ascii.ob_base,
            & codecs_toplevel_consts_24_consts_1.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_2.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_3.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_4.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_5.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_7.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_8.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_9.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_12._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
codecs_toplevel_consts_24_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            &_Py_ID(write),
            & const_str_writelines._ascii.ob_base,
            &_Py_ID(reset),
            &_Py_ID(seek),
            &_Py_ID(getattr),
            &_Py_ID(__getattr__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
            &_Py_ID(__reduce_ex__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[54];
    }
codecs_toplevel_consts_24_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 53,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf3\x04\x17\x05\x1d\xf2\x32\x05\x05\x20\xf2\x0e\x05\x05\x22\xf2\x0e\x0a\x05\x0d\xf3\x18\x03\x05\x19\xf0\x0c\x00\x1d\x24\xf3\x03\x05\x05\x2a\xf2\x0e\x01\x05\x14\xf2\x06\x01\x05\x1c\xf3\x06\x01\x05\x48\x01",
};
static
    struct _PyCode_DEF(74)
codecs_toplevel_consts_24 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 37,
    },
    .co_consts = & codecs_toplevel_consts_24_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 346,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 307,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_StreamWriter._ascii.ob_base,
    .co_qualname = & const_str_StreamWriter._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x0b\x64\x01\x84\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x0c\x64\x05\x84\x01\x5a\x07\x65\x08\x66\x01\x64\x06\x84\x01\x5a\x09\x64\x07\x84\x00\x5a\x0a\x64\x08\x84\x00\x5a\x0b\x64\x09\x84\x00\x5a\x0c\x79\x0a",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[654];
    }
codecs_toplevel_consts_26_consts_1_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 653,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x43\x72\x65\x61\x74\x65\x73\x20\x61\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x72\x65\x61\x6d\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x66\x69\x6c\x65\x2d\x6c\x69\x6b\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x6f\x70\x65\x6e\x20\x66\x6f\x72\x20\x72\x65\x61\x64\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x20\x6d\x61\x79\x20\x75\x73\x65\x20\x64\x69\x66\x66\x65\x72\x65\x6e\x74\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x63\x68\x65\x6d\x65\x73\x20\x62\x79\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x73\x20\x6b\x65\x79\x77\x6f\x72\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x2e\x20\x54\x68\x65\x73\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x73\x20\x61\x72\x65\x20\x70\x72\x65\x64\x65\x66\x69\x6e\x65\x64\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x2d\x20\x72\x61\x69\x73\x65\x20\x61\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x28\x6f\x72\x20\x61\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x69\x67\x6e\x6f\x72\x65\x27\x20\x2d\x20\x69\x67\x6e\x6f\x72\x65\x20\x74\x68\x65\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x20\x61\x6e\x64\x20\x63\x6f\x6e\x74\x69\x6e\x75\x65\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x6e\x65\x78\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x72\x65\x70\x6c\x61\x63\x65\x27\x2d\x20\x72\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x61\x20\x73\x75\x69\x74\x61\x62\x6c\x65\x20\x72\x65\x70\x6c\x61\x63\x65\x6d\x65\x6e\x74\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x27\x62\x61\x63\x6b\x73\x6c\x61\x73\x68\x72\x65\x70\x6c\x61\x63\x65\x27\x20\x2d\x20\x52\x65\x70\x6c\x61\x63\x65\x20\x77\x69\x74\x68\x20\x62\x61\x63\x6b\x73\x6c\x61\x73\x68\x65\x64\x20\x65\x73\x63\x61\x70\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x73\x3b\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x73\x65\x74\x20\x6f\x66\x20\x61\x6c\x6c\x6f\x77\x65\x64\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x20\x76\x61\x6c\x75\x65\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x65\x78\x74\x65\x6e\x64\x65\x64\x20\x76\x69\x61\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x67\x69\x73\x74\x65\x72\x5f\x65\x72\x72\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_26_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & codecs_toplevel_consts_26_consts_1_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_bytebuffer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytebuffer",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_charbuffertype = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "charbuffertype",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__empty_charbuffer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_empty_charbuffer",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_charbuffer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "charbuffer",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_linebuffer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "linebuffer",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
codecs_toplevel_consts_26_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
            &_Py_ID(errors),
            & const_str_bytebuffer._ascii.ob_base,
            & const_str_charbuffertype._ascii.ob_base,
            & const_str__empty_charbuffer._ascii.ob_base,
            & const_str_charbuffer._ascii.ob_base,
            & const_str_linebuffer._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
codecs_toplevel_consts_26_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
codecs_toplevel_consts_26_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x24\x00\x17\x1d\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8c\x0b\xd8\x1a\x1d\x88\x04\x8c\x0f\xd8\x21\x25\xd7\x21\x34\xd1\x21\x34\xd3\x21\x36\x88\x04\xd4\x08\x1e\xd8\x1a\x1e\xd7\x1a\x30\xd1\x1a\x30\x88\x04\x8c\x0f\xd8\x1a\x1e\x88\x04\x8d\x0f",
};
static
    struct _PyCode_DEF(136)
codecs_toplevel_consts_26_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 68,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_26_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 426,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 308,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_26_consts_1_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x00\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
codecs_toplevel_consts_26_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.decode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
codecs_toplevel_consts_26_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0e\x21\xd0\x08\x21",
};
static
    struct _PyCode_DEF(14)
codecs_toplevel_consts_26_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 451,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 309,
    .co_localsplusnames = & codecs_toplevel_consts_14_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(decode),
    .co_qualname = & codecs_toplevel_consts_26_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1261];
    }
codecs_toplevel_consts_26_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1260,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x44\x65\x63\x6f\x64\x65\x73\x20\x64\x61\x74\x61\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x73\x74\x72\x65\x61\x6d\x20\x73\x65\x6c\x66\x2e\x73\x74\x72\x65\x61\x6d\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x73\x75\x6c\x74\x69\x6e\x67\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x63\x68\x61\x72\x73\x20\x69\x6e\x64\x69\x63\x61\x74\x65\x73\x20\x74\x68\x65\x20\x6e\x75\x6d\x62\x65\x72\x20\x6f\x66\x20\x64\x65\x63\x6f\x64\x65\x64\x20\x63\x6f\x64\x65\x20\x70\x6f\x69\x6e\x74\x73\x20\x6f\x72\x20\x62\x79\x74\x65\x73\x20\x74\x6f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x74\x75\x72\x6e\x2e\x20\x72\x65\x61\x64\x28\x29\x20\x77\x69\x6c\x6c\x20\x6e\x65\x76\x65\x72\x20\x72\x65\x74\x75\x72\x6e\x20\x6d\x6f\x72\x65\x20\x64\x61\x74\x61\x20\x74\x68\x61\x6e\x20\x72\x65\x71\x75\x65\x73\x74\x65\x64\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x62\x75\x74\x20\x69\x74\x20\x6d\x69\x67\x68\x74\x20\x72\x65\x74\x75\x72\x6e\x20\x6c\x65\x73\x73\x2c\x20\x69\x66\x20\x74\x68\x65\x72\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x65\x6e\x6f\x75\x67\x68\x20\x61\x76\x61\x69\x6c\x61\x62\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x69\x7a\x65\x20\x69\x6e\x64\x69\x63\x61\x74\x65\x73\x20\x74\x68\x65\x20\x61\x70\x70\x72\x6f\x78\x69\x6d\x61\x74\x65\x20\x6d\x61\x78\x69\x6d\x75\x6d\x20\x6e\x75\x6d\x62\x65\x72\x20\x6f\x66\x20\x64\x65\x63\x6f\x64\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x62\x79\x74\x65\x73\x20\x6f\x72\x20\x63\x6f\x64\x65\x20\x70\x6f\x69\x6e\x74\x73\x20\x74\x6f\x20\x72\x65\x61\x64\x20\x66\x6f\x72\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x2e\x20\x54\x68\x65\x20\x64\x65\x63\x6f\x64\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x63\x61\x6e\x20\x6d\x6f\x64\x69\x66\x79\x20\x74\x68\x69\x73\x20\x73\x65\x74\x74\x69\x6e\x67\x20\x61\x73\x20\x61\x70\x70\x72\x6f\x70\x72\x69\x61\x74\x65\x2e\x20\x54\x68\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x76\x61\x6c\x75\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x2d\x31\x20\x69\x6e\x64\x69\x63\x61\x74\x65\x73\x20\x74\x6f\x20\x72\x65\x61\x64\x20\x61\x6e\x64\x20\x64\x65\x63\x6f\x64\x65\x20\x61\x73\x20\x6d\x75\x63\x68\x20\x61\x73\x20\x70\x6f\x73\x73\x69\x62\x6c\x65\x2e\x20\x20\x73\x69\x7a\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x69\x73\x20\x69\x6e\x74\x65\x6e\x64\x65\x64\x20\x74\x6f\x20\x70\x72\x65\x76\x65\x6e\x74\x20\x68\x61\x76\x69\x6e\x67\x20\x74\x6f\x20\x64\x65\x63\x6f\x64\x65\x20\x68\x75\x67\x65\x20\x66\x69\x6c\x65\x73\x20\x69\x6e\x20\x6f\x6e\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x65\x70\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x66\x69\x72\x73\x74\x6c\x69\x6e\x65\x20\x69\x73\x20\x74\x72\x75\x65\x2c\x20\x61\x6e\x64\x20\x61\x20\x55\x6e\x69\x63\x6f\x64\x65\x44\x65\x63\x6f\x64\x65\x45\x72\x72\x6f\x72\x20\x68\x61\x70\x70\x65\x6e\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x61\x66\x74\x65\x72\x20\x74\x68\x65\x20\x66\x69\x72\x73\x74\x20\x6c\x69\x6e\x65\x20\x74\x65\x72\x6d\x69\x6e\x61\x74\x6f\x72\x20\x69\x6e\x20\x74\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x6f\x6e\x6c\x79\x20\x74\x68\x65\x20\x66\x69\x72\x73\x74\x20\x6c\x69\x6e\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x2c\x20\x74\x68\x65\x20\x72\x65\x73\x74\x20\x6f\x66\x20\x74\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x6b\x65\x70\x74\x20\x75\x6e\x74\x69\x6c\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6e\x65\x78\x74\x20\x63\x61\x6c\x6c\x20\x74\x6f\x20\x72\x65\x61\x64\x28\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x6d\x65\x74\x68\x6f\x64\x20\x73\x68\x6f\x75\x6c\x64\x20\x75\x73\x65\x20\x61\x20\x67\x72\x65\x65\x64\x79\x20\x72\x65\x61\x64\x20\x73\x74\x72\x61\x74\x65\x67\x79\x2c\x20\x6d\x65\x61\x6e\x69\x6e\x67\x20\x74\x68\x61\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x69\x74\x20\x73\x68\x6f\x75\x6c\x64\x20\x72\x65\x61\x64\x20\x61\x73\x20\x6d\x75\x63\x68\x20\x64\x61\x74\x61\x20\x61\x73\x20\x69\x73\x20\x61\x6c\x6c\x6f\x77\x65\x64\x20\x77\x69\x74\x68\x69\x6e\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x66\x69\x6e\x69\x74\x69\x6f\x6e\x20\x6f\x66\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x73\x69\x7a\x65\x2c\x20\x65\x2e\x67\x2e\x20\x20\x69\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x65\x6e\x64\x69\x6e\x67\x73\x20\x6f\x72\x20\x73\x74\x61\x74\x65\x20\x6d\x61\x72\x6b\x65\x72\x73\x20\x61\x72\x65\x20\x61\x76\x61\x69\x6c\x61\x62\x6c\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x6e\x20\x74\x68\x65\x20\x73\x74\x72\x65\x61\x6d\x2c\x20\x74\x68\x65\x73\x65\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x20\x72\x65\x61\x64\x20\x74\x6f\x6f\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_26_consts_3_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(keepends),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
codecs_toplevel_consts_26_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & codecs_toplevel_consts_26_consts_3_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_True,
            & codecs_toplevel_consts_26_consts_3_consts_4._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_splitlines = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "splitlines",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
codecs_toplevel_consts_26_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_linebuffer._ascii.ob_base,
            & const_str__empty_charbuffer._ascii.ob_base,
            &_Py_ID(join),
            & const_str_charbuffer._ascii.ob_base,
            &_Py_ID(len),
            & const_str_stream._ascii.ob_base,
            &_Py_ID(read),
            & const_str_bytebuffer._ascii.ob_base,
            &_Py_ID(decode),
            &_Py_ID(errors),
            & const_str_UnicodeDecodeError._ascii.ob_base,
            &_Py_ID(start),
            & const_str_splitlines._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
codecs_toplevel_consts_26_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.read",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[409];
    }
codecs_toplevel_consts_26_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 408,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x38\x00\x0c\x10\x8f\x3f\x8a\x3f\xd8\x1e\x22\xd7\x1e\x34\xd1\x1e\x34\xd7\x1e\x39\xd1\x1e\x39\xb8\x24\xbf\x2f\xb9\x2f\xd3\x1e\x4a\x88\x44\x8c\x4f\xd8\x1e\x22\x88\x44\x8c\x4f\xe0\x0b\x10\x90\x31\x8a\x39\xf0\x06\x00\x15\x19\x88\x45\xf0\x06\x00\x0f\x13\xe0\x0f\x14\x98\x01\x8a\x7a\xdc\x13\x16\x90\x74\x97\x7f\x91\x7f\xd3\x13\x27\xa8\x35\xd2\x13\x30\xd8\x14\x19\xe0\x0f\x13\x90\x61\x8a\x78\xd8\x1a\x1e\x9f\x2b\x99\x2b\xd7\x1a\x2a\xd1\x1a\x2a\xd3\x1a\x2c\x91\x07\xe0\x1a\x1e\x9f\x2b\x99\x2b\xd7\x1a\x2a\xd1\x1a\x2a\xa8\x34\xd3\x1a\x30\x90\x07\xe0\x13\x17\x97\x3f\x91\x3f\xa0\x57\xd1\x13\x2c\x88\x44\xd9\x13\x17\xd8\x10\x15\xf0\x02\x0a\x0d\x1a\xd8\x29\x2d\xaf\x1b\xa9\x1b\xb0\x54\xb8\x34\xbf\x3b\xb9\x3b\xd3\x29\x47\xd1\x10\x26\x90\x08\x98\x2c\xf0\x16\x00\x1f\x23\xa0\x3c\xa0\x3d\xd0\x1e\x31\x88\x44\x8c\x4f\xe0\x0c\x10\x8f\x4f\x8a\x4f\x98\x78\xd1\x0c\x27\x8d\x4f\xe1\x13\x1a\xd8\x10\x15\xf0\x3f\x00\x0f\x13\xf0\x40\x01\x00\x0c\x11\x90\x31\x8a\x39\xe0\x15\x19\x97\x5f\x91\x5f\x88\x46\xd8\x1e\x22\xd7\x1e\x34\xd1\x1e\x34\x88\x44\x8c\x4f\xf0\x0a\x00\x10\x16\x88\x0d\xf0\x05\x00\x16\x1a\x97\x5f\x91\x5f\xa0\x56\xa0\x65\xd0\x15\x2c\x88\x46\xd8\x1e\x22\x9f\x6f\x99\x6f\xa8\x65\xa8\x66\xd0\x1e\x35\x88\x44\x8c\x4f\xd8\x0f\x15\x88\x0d\xf8\xf4\x31\x00\x14\x26\xf2\x00\x08\x0d\x1a\xd9\x13\x1c\xe0\x18\x1c\x9f\x0b\x99\x0b\xa0\x44\xa8\x1a\xa8\x23\xaf\x29\xa9\x29\xd0\x24\x34\xb0\x64\xb7\x6b\xb1\x6b\xd3\x18\x42\xf1\x03\x00\x15\x2b\x90\x48\x98\x6c\xe0\x1c\x24\xd7\x1c\x2f\xd1\x1c\x2f\xb8\x14\xd0\x1c\x2f\xd3\x1c\x3e\x90\x45\xdc\x17\x1a\x98\x35\x93\x7a\xa0\x31\x92\x7d\xd8\x18\x1d\xe0\x14\x19\xf4\x07\x00\x18\x25\xfb\xf0\x0b\x08\x0d\x1a\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
codecs_toplevel_consts_26_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x32\x1f\x44\x3d\x00\xc4\x3d\x09\x46\x20\x03\xc5\x06\x41\x10\x46\x1b\x03\xc6\x1b\x05\x46\x20\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_chars = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "chars",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_firstline = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "firstline",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_newdata = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "newdata",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_newchars = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "newchars",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_decodedbytes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "decodedbytes",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_lines = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lines",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
codecs_toplevel_consts_26_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(size),
            & const_str_chars._ascii.ob_base,
            & const_str_firstline._ascii.ob_base,
            & const_str_newdata._ascii.ob_base,
            &_Py_ID(data),
            & const_str_newchars._ascii.ob_base,
            & const_str_decodedbytes._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
            & const_str_lines._ascii.ob_base,
            & const_str_result._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(838)
codecs_toplevel_consts_26_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 419,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_26_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = & codecs_toplevel_consts_26_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 454,
    .co_nlocalsplus = 11,
    .co_nlocals = 11,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 310,
    .co_localsplusnames = & codecs_toplevel_consts_26_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(read),
    .co_qualname = & codecs_toplevel_consts_26_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x31\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x02\x6b\x02\x00\x00\x72\x02\x7c\x01\x7d\x02\x09\x00\x7c\x02\x64\x02\x6b\x5c\x00\x00\x72\x19\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x6b\x5c\x00\x00\x72\x01\x6e\x90\x7c\x01\x64\x02\x6b\x02\x00\x00\x72\x1b\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x6e\x1b\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7a\x00\x00\x00\x7d\x05\x7c\x05\x73\x01\x6e\x43\x09\x00\x7c\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x06\x7d\x07\x7c\x05\x7c\x07\x64\x01\x1a\x00\x7c\x00\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x78\x01\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7a\x0d\x00\x00\x63\x02\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x73\x01\x6e\x01\x8c\xae\x7c\x02\x64\x02\x6b\x02\x00\x00\x72\x1f\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x0a\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x53\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x02\x1a\x00\x7d\x0a\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x01\x1a\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x53\x00\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x5a\x7d\x08\x7c\x03\x72\x4d\x7c\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x64\x01\x7c\x08\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1a\x00\x7c\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x06\x7d\x07\x7c\x06\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xac\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x09\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x01\x00\x00\x00\x00\x00\x00\x64\x05\x6b\x1a\x00\x00\x72\x02\x82\x00\x82\x00\x59\x00\x64\x01\x7d\x08\x7e\x08\x8c\xca\x64\x01\x7d\x08\x7e\x08\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[178];
    }
codecs_toplevel_consts_26_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 177,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x52\x65\x61\x64\x20\x6f\x6e\x65\x20\x6c\x69\x6e\x65\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x73\x74\x72\x65\x61\x6d\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x63\x6f\x64\x65\x64\x20\x64\x61\x74\x61\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x69\x7a\x65\x2c\x20\x69\x66\x20\x67\x69\x76\x65\x6e\x2c\x20\x69\x73\x20\x70\x61\x73\x73\x65\x64\x20\x61\x73\x20\x73\x69\x7a\x65\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x74\x6f\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x61\x64\x28\x29\x20\x6d\x65\x74\x68\x6f\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_26_consts_5_consts_8 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_firstline._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_26_consts_5_consts_11 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(size),
            & const_str_chars._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_8000 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 8000 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
codecs_toplevel_consts_26_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & codecs_toplevel_consts_26_consts_5_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
            Py_False,
            & codecs_toplevel_consts_26_consts_3_consts_4._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 72],
            Py_True,
            & codecs_toplevel_consts_26_consts_5_consts_8._object.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[13],
            (PyObject *)&_Py_SINGLETON(bytes_characters[13]),
            & codecs_toplevel_consts_26_consts_5_consts_11._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & const_int_8000.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
codecs_toplevel_consts_26_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_linebuffer._ascii.ob_base,
            &_Py_ID(len),
            & const_str_charbuffer._ascii.ob_base,
            & const_str_splitlines._ascii.ob_base,
            & const_str__empty_charbuffer._ascii.ob_base,
            &_Py_ID(read),
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            &_Py_ID(bytes),
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
codecs_toplevel_consts_26_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.readline",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[572];
    }
codecs_toplevel_consts_26_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 571,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x16\x00\x0c\x10\x8f\x3f\x8a\x3f\xd8\x13\x17\x97\x3f\x91\x3f\xa0\x31\xd1\x13\x25\x88\x44\xd8\x10\x14\x97\x0f\x91\x0f\xa0\x01\xd0\x10\x22\xdc\x0f\x12\x90\x34\x97\x3f\x91\x3f\xd3\x0f\x23\xa0\x71\xd2\x0f\x28\xf0\x06\x00\x23\x27\xa7\x2f\xa1\x2f\xb0\x21\xd1\x22\x34\x90\x04\x94\x0f\xd8\x22\x26\x90\x04\x94\x0f\xd9\x13\x1b\xd8\x17\x1b\x97\x7f\x91\x7f\xb0\x05\x90\x7f\xd3\x17\x36\xb0\x71\xd1\x17\x39\x90\x04\xd8\x13\x17\x88\x4b\xe0\x13\x17\x92\x3a\x98\x32\x88\x08\xd8\x0f\x13\xd7\x0f\x25\xd1\x0f\x25\x88\x04\xe0\x0e\x12\xd8\x13\x17\x97\x39\x91\x39\x98\x58\xb0\x14\x90\x39\xd3\x13\x36\x88\x44\xd9\x0f\x13\xf4\x08\x00\x15\x1f\x98\x74\xa4\x53\xd4\x14\x29\xa8\x64\xaf\x6d\xa9\x6d\xb8\x44\xd4\x2e\x41\xdc\x14\x1e\x98\x74\xa4\x55\xd4\x14\x2b\xb0\x04\xb7\x0d\xb1\x0d\xb8\x65\xd4\x30\x44\xd8\x14\x18\x98\x44\x9f\x49\x99\x49\xa8\x31\xb0\x41\x98\x49\xd3\x1c\x36\xd1\x14\x36\x90\x44\xe0\x0c\x10\x90\x44\x89\x4c\x88\x44\xd8\x14\x18\x97\x4f\x91\x4f\xa8\x54\x90\x4f\xd3\x14\x32\x88\x45\xd9\x0f\x14\xdc\x13\x16\x90\x75\x93\x3a\xa0\x01\x92\x3e\xf0\x06\x00\x1c\x21\xa0\x11\x99\x38\x90\x44\xd8\x18\x1d\x98\x61\x98\x08\xdc\x17\x1a\x98\x35\x93\x7a\xa0\x41\x92\x7e\xe0\x18\x1d\x98\x62\x9b\x09\xa0\x54\xa7\x5f\xa1\x5f\xd1\x18\x34\x9b\x09\xd8\x2a\x2f\x98\x04\x9c\x0f\xd8\x2a\x2e\x98\x04\x9d\x0f\xf0\x06\x00\x2b\x30\xb0\x01\xa9\x28\xb0\x54\xb7\x5f\xb1\x5f\xd1\x2a\x44\x98\x04\x9c\x0f\xd9\x1b\x23\xd8\x1f\x23\x9f\x7f\x99\x7f\xb8\x05\x98\x7f\xd3\x1f\x3e\xb8\x71\xd1\x1f\x41\x98\x04\xd8\x14\x19\xf0\x26\x00\x10\x14\x88\x0b\xf0\x25\x00\x20\x25\xa0\x51\x99\x78\x90\x0c\xd8\x22\x27\xa8\x01\xa1\x28\xd7\x22\x35\xd1\x22\x35\xb8\x75\xd0\x22\x35\xd3\x22\x45\xc0\x61\xd1\x22\x48\x90\x0f\xd8\x13\x1f\xa0\x3f\xd2\x13\x32\xe0\x26\x2a\xd7\x26\x3c\xd1\x26\x3c\xd7\x26\x41\xd1\x26\x41\xc0\x25\xc8\x01\xc8\x02\xc0\x29\xd3\x26\x4c\xd8\x26\x2a\xa7\x6f\xa1\x6f\xf1\x03\x01\x27\x36\x90\x44\x94\x4f\xe1\x17\x1f\xd8\x1f\x2b\x98\x04\xf0\x06\x00\x15\x1a\xf0\x10\x00\x10\x14\x88\x0b\xf0\x13\x00\x20\x2f\x98\x04\xd8\x14\x19\xf0\x10\x00\x10\x14\x88\x0b\xf1\x0d\x00\x14\x18\x98\x34\xd0\x1b\x2b\xd9\x13\x17\xa1\x08\xd8\x1b\x1f\x9f\x3f\x99\x3f\xb0\x45\x98\x3f\xd3\x1b\x3a\xb8\x31\xd1\x1b\x3d\x90\x44\xd8\x10\x15\xf0\x06\x00\x10\x14\x88\x0b\xf0\x05\x00\x10\x18\x98\x24\x8a\x7f\xd8\x10\x18\x98\x41\x91\x0d\x90\x08\xf1\x5d\x01\x00\x0f\x13",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_readsize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "readsize",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_line0withend = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "line0withend",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_line0withoutend = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "line0withoutend",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
codecs_toplevel_consts_26_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(size),
            &_Py_ID(keepends),
            &_Py_ID(line),
            & const_str_readsize._ascii.ob_base,
            &_Py_ID(data),
            & const_str_lines._ascii.ob_base,
            & const_str_line0withend._ascii.ob_base,
            & const_str_line0withoutend._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(1062)
codecs_toplevel_consts_26_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 531,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_5_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_26_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 534,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 311,
    .co_localsplusnames = & codecs_toplevel_consts_26_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(readline),
    .co_qualname = & codecs_toplevel_consts_26_consts_5_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x68\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x7d\x03\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x3d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x28\x00\x00\x72\x1b\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x73\x15\x7c\x03\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xac\x05\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x7d\x03\x7c\x03\x53\x00\x7c\x01\x78\x01\x73\x02\x01\x00\x64\x06\x7d\x04\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x09\x00\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x64\x07\xac\x08\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x72\x58\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x11\x7c\x05\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\xab\x01\x00\x00\x00\x00\x00\x00\x73\x21\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x27\x7c\x05\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x72\x16\x7c\x05\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x02\xac\x0b\xab\x02\x00\x00\x00\x00\x00\x00\x7a\x0d\x00\x00\x7d\x05\x7c\x03\x7c\x05\x7a\x0d\x00\x00\x7d\x03\x7c\x03\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\xac\x05\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x72\xd9\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x44\x00\x00\x72\x6d\x7c\x06\x64\x01\x19\x00\x00\x00\x7d\x03\x7c\x06\x64\x01\x3d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x44\x00\x00\x72\x26\x7c\x06\x64\x0c\x78\x02\x78\x02\x19\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x0d\x00\x00\x63\x03\x63\x02\x3c\x00\x00\x00\x7c\x06\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6e\x17\x7c\x06\x64\x01\x19\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x73\x15\x7c\x03\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xac\x05\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x7d\x03\x09\x00\x7c\x03\x53\x00\x7c\x06\x64\x01\x19\x00\x00\x00\x7d\x07\x7c\x06\x64\x01\x19\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xac\x05\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x7d\x08\x7c\x07\x7c\x08\x6b\x37\x00\x00\x72\x3c\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x64\x02\x64\x03\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x72\x05\x7c\x07\x7d\x03\x09\x00\x7c\x03\x53\x00\x7c\x08\x7d\x03\x09\x00\x7c\x03\x53\x00\x7c\x05\x72\x02\x7c\x01\x81\x1c\x7c\x03\x72\x17\x7c\x02\x73\x15\x7c\x03\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xac\x05\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x7d\x03\x09\x00\x7c\x03\x53\x00\x7c\x04\x64\x0d\x6b\x02\x00\x00\x72\x05\x7c\x04\x64\x0e\x7a\x12\x00\x00\x7d\x04\x90\x01\x8c\x8b",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[340];
    }
codecs_toplevel_consts_26_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 339,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x52\x65\x61\x64\x20\x61\x6c\x6c\x20\x6c\x69\x6e\x65\x73\x20\x61\x76\x61\x69\x6c\x61\x62\x6c\x65\x20\x6f\x6e\x20\x74\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x73\x74\x72\x65\x61\x6d\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x6d\x20\x61\x73\x20\x61\x20\x6c\x69\x73\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x4c\x69\x6e\x65\x20\x62\x72\x65\x61\x6b\x73\x20\x61\x72\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x20\x75\x73\x69\x6e\x67\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x27\x73\x20\x64\x65\x63\x6f\x64\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6d\x65\x74\x68\x6f\x64\x20\x61\x6e\x64\x20\x61\x72\x65\x20\x69\x6e\x63\x6c\x75\x64\x65\x64\x20\x69\x6e\x20\x74\x68\x65\x20\x6c\x69\x73\x74\x20\x65\x6e\x74\x72\x69\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x69\x7a\x65\x68\x69\x6e\x74\x2c\x20\x69\x66\x20\x67\x69\x76\x65\x6e\x2c\x20\x69\x73\x20\x69\x67\x6e\x6f\x72\x65\x64\x20\x73\x69\x6e\x63\x65\x20\x74\x68\x65\x72\x65\x20\x69\x73\x20\x6e\x6f\x20\x65\x66\x66\x69\x63\x69\x65\x6e\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x77\x61\x79\x20\x74\x6f\x20\x66\x69\x6e\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x74\x72\x75\x65\x20\x65\x6e\x64\x2d\x6f\x66\x2d\x6c\x69\x6e\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_26_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_26_consts_6_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_26_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(read),
            & const_str_splitlines._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_readlines = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "readlines",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_26_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.readlines",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
codecs_toplevel_consts_26_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x18\x00\x10\x14\x8f\x79\x89\x79\x8b\x7b\x88\x04\xd8\x0f\x13\x8f\x7f\x89\x7f\x98\x78\xd3\x0f\x28\xd0\x08\x28",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_26_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(sizehint),
            &_Py_ID(keepends),
            &_Py_ID(data),
        },
    },
};
static
    struct _PyCode_DEF(68)
codecs_toplevel_consts_26_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 34,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_6_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_26_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 609,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 312,
    .co_localsplusnames = & codecs_toplevel_consts_26_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_readlines._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_26_consts_6_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[237];
    }
codecs_toplevel_consts_26_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 236,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x52\x65\x73\x65\x74\x73\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x62\x75\x66\x66\x65\x72\x73\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x6b\x65\x65\x70\x69\x6e\x67\x20\x69\x6e\x74\x65\x72\x6e\x61\x6c\x20\x73\x74\x61\x74\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x4e\x6f\x74\x65\x20\x74\x68\x61\x74\x20\x6e\x6f\x20\x73\x74\x72\x65\x61\x6d\x20\x72\x65\x70\x6f\x73\x69\x74\x69\x6f\x6e\x69\x6e\x67\x20\x73\x68\x6f\x75\x6c\x64\x20\x74\x61\x6b\x65\x20\x70\x6c\x61\x63\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x69\x73\x20\x70\x72\x69\x6d\x61\x72\x69\x6c\x79\x20\x69\x6e\x74\x65\x6e\x64\x65\x64\x20\x74\x6f\x20\x62\x65\x20\x61\x62\x6c\x65\x20\x74\x6f\x20\x72\x65\x63\x6f\x76\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x66\x72\x6f\x6d\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x20\x65\x72\x72\x6f\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_26_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & codecs_toplevel_consts_26_consts_7_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_26_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_bytebuffer._ascii.ob_base,
            & const_str__empty_charbuffer._ascii.ob_base,
            & const_str_charbuffer._ascii.ob_base,
            & const_str_linebuffer._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
codecs_toplevel_consts_26_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.reset",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
codecs_toplevel_consts_26_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x12\x00\x1b\x1e\x88\x04\x8c\x0f\xd8\x1a\x1e\xd7\x1a\x30\xd1\x1a\x30\x88\x04\x8c\x0f\xd8\x1a\x1e\x88\x04\x8d\x0f",
};
static
    struct _PyCode_DEF(66)
codecs_toplevel_consts_26_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 33,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_7_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_26_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 624,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 313,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_26_consts_7_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[113];
    }
codecs_toplevel_consts_26_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 112,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x53\x65\x74\x20\x74\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x73\x74\x72\x65\x61\x6d\x27\x73\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x70\x6f\x73\x69\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x73\x65\x74\x73\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x62\x75\x66\x66\x65\x72\x73\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x6b\x65\x65\x70\x69\x6e\x67\x20\x73\x74\x61\x74\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_26_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_26_consts_8_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
codecs_toplevel_consts_26_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.seek",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
codecs_toplevel_consts_26_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x09\x0d\x8f\x0b\x89\x0b\xd7\x08\x18\xd1\x08\x18\x98\x16\xa0\x16\xd4\x08\x28\xd8\x08\x0c\x8f\x0a\x89\x0a\x8d\x0c",
};
static
    struct _PyCode_DEF(92)
codecs_toplevel_consts_26_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 46,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_8_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 637,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 314,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(seek),
    .co_qualname = & codecs_toplevel_consts_26_consts_8_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[53];
    }
codecs_toplevel_consts_26_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 52,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " Return the next decoded line from the input stream.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_26_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_26_consts_9_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_StopIteration = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StopIteration",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_26_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(readline),
            & const_str_StopIteration._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
codecs_toplevel_consts_26_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.__next__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
codecs_toplevel_consts_26_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x10\x14\x8f\x7d\x89\x7d\x8b\x7f\x88\x04\xd9\x0b\x0f\xd8\x13\x17\x88\x4b\xdc\x0e\x1b\xd0\x08\x1b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_26_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(line),
        },
    },
};
static
    struct _PyCode_DEF(54)
codecs_toplevel_consts_26_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_9_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_26_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 645,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 315,
    .co_localsplusnames = & codecs_toplevel_consts_26_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__next__),
    .co_qualname = & codecs_toplevel_consts_26_consts_9_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x72\x02\x7c\x01\x53\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
codecs_toplevel_consts_26_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.__iter__",
};
static
    struct _PyCode_DEF(6)
codecs_toplevel_consts_26_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 653,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 316,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & codecs_toplevel_consts_26_consts_10_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
codecs_toplevel_consts_26_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.__getattr__",
};
static
    struct _PyCode_DEF(40)
codecs_toplevel_consts_26_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_6_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 656,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 317,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getattr__),
    .co_qualname = & codecs_toplevel_consts_26_consts_11_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x7c\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_26_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.__enter__",
};
static
    struct _PyCode_DEF(6)
codecs_toplevel_consts_26_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 663,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 318,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & codecs_toplevel_consts_26_consts_12_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
codecs_toplevel_consts_26_consts_13_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.__exit__",
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_26_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 666,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 319,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & codecs_toplevel_consts_26_consts_13_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
codecs_toplevel_consts_26_consts_14_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReader.__reduce_ex__",
};
static
    struct _PyCode_DEF(70)
codecs_toplevel_consts_26_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_9_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 669,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 320,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reduce_ex__),
    .co_qualname = & codecs_toplevel_consts_26_consts_14_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_26_consts_16 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
codecs_toplevel_consts_26_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            & const_str_StreamReader._ascii.ob_base,
            & codecs_toplevel_consts_26_consts_1.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_2.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_3.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_26_consts_5.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_7.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_8.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_9.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_10.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_11.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_12.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_13.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_14.ob_base.ob_base,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_26_consts_16._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_68_consts_2_consts._object.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_12._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
codecs_toplevel_consts_26_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            & const_str_str._ascii.ob_base,
            & const_str_charbuffertype._ascii.ob_base,
            &_Py_ID(__init__),
            &_Py_ID(decode),
            &_Py_ID(read),
            &_Py_ID(readline),
            & const_str_readlines._ascii.ob_base,
            &_Py_ID(reset),
            &_Py_ID(seek),
            &_Py_ID(__next__),
            &_Py_ID(__iter__),
            &_Py_ID(getattr),
            &_Py_ID(__getattr__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
            &_Py_ID(__reduce_ex__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[83];
    }
codecs_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 82,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x15\x18\x80\x4e\xf3\x04\x17\x05\x1f\xf3\x32\x01\x05\x22\xf3\x06\x4e\x01\x05\x16\xf3\x60\x02\x49\x01\x05\x14\xf3\x56\x02\x0d\x05\x29\xf2\x1e\x0b\x05\x1f\xf3\x1a\x06\x05\x15\xf2\x10\x06\x05\x1c\xf2\x10\x01\x05\x14\xf0\x08\x00\x1d\x24\xf3\x03\x05\x05\x2a\xf2\x0e\x01\x05\x14\xf2\x06\x01\x05\x1c\xf3\x06\x01\x05\x48\x01",
};
static
    struct _PyCode_DEF(110)
codecs_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & codecs_toplevel_consts_26_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 422,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 321,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_StreamReader._ascii.ob_base,
    .co_qualname = & const_str_StreamReader._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x65\x03\x5a\x04\x64\x0f\x64\x01\x84\x01\x5a\x05\x64\x0f\x64\x02\x84\x01\x5a\x06\x64\x10\x64\x03\x84\x01\x5a\x07\x64\x11\x64\x05\x84\x01\x5a\x08\x64\x11\x64\x06\x84\x01\x5a\x09\x64\x07\x84\x00\x5a\x0a\x64\x12\x64\x08\x84\x01\x5a\x0b\x64\x09\x84\x00\x5a\x0c\x64\x0a\x84\x00\x5a\x0d\x65\x0e\x66\x01\x64\x0b\x84\x01\x5a\x0f\x64\x0c\x84\x00\x5a\x10\x64\x0d\x84\x00\x5a\x11\x64\x0e\x84\x00\x5a\x12\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[258];
    }
codecs_toplevel_consts_28_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 257,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x57\x72\x69\x74\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x73\x20\x61\x6c\x6c\x6f\x77\x20\x77\x72\x61\x70\x70\x69\x6e\x67\x20\x73\x74\x72\x65\x61\x6d\x73\x20\x77\x68\x69\x63\x68\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x77\x6f\x72\x6b\x20\x69\x6e\x20\x62\x6f\x74\x68\x20\x72\x65\x61\x64\x20\x61\x6e\x64\x20\x77\x72\x69\x74\x65\x20\x6d\x6f\x64\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x64\x65\x73\x69\x67\x6e\x20\x69\x73\x20\x73\x75\x63\x68\x20\x74\x68\x61\x74\x20\x6f\x6e\x65\x20\x63\x61\x6e\x20\x75\x73\x65\x20\x74\x68\x65\x20\x66\x61\x63\x74\x6f\x72\x79\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x2e\x6c\x6f\x6f\x6b\x75\x70\x28\x29\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x74\x6f\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_unknown = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "unknown",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[339];
    }
codecs_toplevel_consts_28_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 338,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x43\x72\x65\x61\x74\x65\x73\x20\x61\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x57\x72\x69\x74\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x72\x65\x61\x6d\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x53\x74\x72\x65\x61\x6d\x2d\x6c\x69\x6b\x65\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x61\x64\x65\x72\x2c\x20\x57\x72\x69\x74\x65\x72\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x66\x61\x63\x74\x6f\x72\x79\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x73\x20\x6f\x72\x20\x63\x6c\x61\x73\x73\x65\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x2c\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x72\x65\x73\x70\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x45\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x69\x73\x20\x64\x6f\x6e\x65\x20\x69\x6e\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x77\x61\x79\x20\x61\x73\x20\x64\x65\x66\x69\x6e\x65\x64\x20\x66\x6f\x72\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x2f\x52\x65\x61\x64\x65\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_28_consts_3_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_reader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "reader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_writer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "writer",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_28_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
            & const_str_reader._ascii.ob_base,
            & const_str_writer._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_28_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[47];
    }
codecs_toplevel_consts_28_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 46,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x1a\x00\x17\x1d\x88\x04\x8c\x0b\xd9\x16\x1c\x98\x56\xa0\x56\xd3\x16\x2c\x88\x04\x8c\x0b\xd9\x16\x1c\x98\x56\xa0\x56\xd3\x16\x2c\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8d\x0b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_Reader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Reader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_Writer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Writer",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
codecs_toplevel_consts_28_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_stream._ascii.ob_base,
            & const_str_Reader._ascii.ob_base,
            & const_str_Writer._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct _PyCode_DEF(88)
codecs_toplevel_consts_28_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 44,
    },
    .co_consts = & codecs_toplevel_consts_28_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 687,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 322,
    .co_localsplusnames = & codecs_toplevel_consts_28_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_28_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x7c\x02\x7c\x01\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x7c\x03\x7c\x01\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            &_Py_ID(read),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
codecs_toplevel_consts_28_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.read",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
codecs_toplevel_consts_28_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x1f\xd1\x0f\x1f\xa0\x04\xd3\x0f\x25\xd0\x08\x25",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(size),
        },
    },
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_28_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 705,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 323,
    .co_localsplusnames = & codecs_toplevel_consts_28_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(read),
    .co_qualname = & codecs_toplevel_consts_28_consts_4_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            &_Py_ID(readline),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_28_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.readline",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
codecs_toplevel_consts_28_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x23\xd1\x0f\x23\xa0\x44\xd3\x0f\x29\xd0\x08\x29",
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_28_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 709,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 324,
    .co_localsplusnames = & codecs_toplevel_consts_28_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(readline),
    .co_qualname = & codecs_toplevel_consts_28_consts_6_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            & const_str_readlines._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
codecs_toplevel_consts_28_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.readlines",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
codecs_toplevel_consts_28_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x24\xd1\x0f\x24\xa0\x58\xd3\x0f\x2e\xd0\x08\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(sizehint),
        },
    },
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_28_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 713,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 325,
    .co_localsplusnames = & codecs_toplevel_consts_28_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_readlines._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_28_consts_7_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(next),
            & const_str_reader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_28_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.__next__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
codecs_toplevel_consts_28_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x10\x14\x90\x44\x97\x4b\x91\x4b\xd3\x0f\x20\xd0\x08\x20",
};
static
    struct _PyCode_DEF(44)
codecs_toplevel_consts_28_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_9_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 717,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 326,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__next__),
    .co_qualname = & codecs_toplevel_consts_28_consts_8_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_28_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.__iter__",
};
static
    struct _PyCode_DEF(6)
codecs_toplevel_consts_28_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 722,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 327,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & codecs_toplevel_consts_28_consts_9_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_writer._ascii.ob_base,
            &_Py_ID(write),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
codecs_toplevel_consts_28_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.write",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
codecs_toplevel_consts_28_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x20\xd1\x0f\x20\xa0\x14\xd3\x0f\x26\xd0\x08\x26",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_10_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(data),
        },
    },
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_28_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 725,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 328,
    .co_localsplusnames = & codecs_toplevel_consts_28_consts_10_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(write),
    .co_qualname = & codecs_toplevel_consts_28_consts_10_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_28_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_writer._ascii.ob_base,
            & const_str_writelines._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
codecs_toplevel_consts_28_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.writelines",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
codecs_toplevel_consts_28_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x25\xd1\x0f\x25\xa0\x64\xd3\x0f\x2b\xd0\x08\x2b",
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_28_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 729,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 329,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_writelines._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_28_consts_11_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_28_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            &_Py_ID(reset),
            & const_str_writer._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
codecs_toplevel_consts_28_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.reset",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[35];
    }
codecs_toplevel_consts_28_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 34,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x19\xd1\x08\x19\xd4\x08\x1b\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x19\xd1\x08\x19\xd5\x08\x1b",
};
static
    struct _PyCode_DEF(108)
codecs_toplevel_consts_28_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 54,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 733,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 330,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_28_consts_12_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
codecs_toplevel_consts_28_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
            &_Py_ID(seek),
            & const_str_reader._ascii.ob_base,
            &_Py_ID(reset),
            & const_str_writer._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
codecs_toplevel_consts_28_consts_13_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.seek",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[75];
    }
codecs_toplevel_consts_28_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 74,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x18\xd1\x08\x18\x98\x16\xa0\x16\xd4\x08\x28\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x19\xd1\x08\x19\xd4\x08\x1b\xd8\x0b\x11\x90\x51\x8a\x3b\x98\x36\xa0\x51\x9a\x3b\xd8\x0c\x10\x8f\x4b\x89\x4b\xd7\x0c\x1d\xd1\x0c\x1d\xd5\x0c\x1f\xf0\x03\x00\x1c\x27\x88\x3b",
};
static
    struct _PyCode_DEF(188)
codecs_toplevel_consts_28_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 94,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 738,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 331,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(seek),
    .co_qualname = & codecs_toplevel_consts_28_consts_13_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x64\x01\x6b\x28\x00\x00\x72\x21\x7c\x01\x64\x01\x6b\x28\x00\x00\x72\x1b\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x79\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
codecs_toplevel_consts_28_consts_14_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.__getattr__",
};
static
    struct _PyCode_DEF(40)
codecs_toplevel_consts_28_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_6_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 744,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 332,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getattr__),
    .co_qualname = & codecs_toplevel_consts_28_consts_14_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x7c\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
codecs_toplevel_consts_28_consts_15_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.__enter__",
};
static
    struct _PyCode_DEF(6)
codecs_toplevel_consts_28_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 753,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 333,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & codecs_toplevel_consts_28_consts_15_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_28_consts_16_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.__exit__",
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_28_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 756,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 334,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & codecs_toplevel_consts_28_consts_16_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
codecs_toplevel_consts_28_consts_17_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamReaderWriter.__reduce_ex__",
};
static
    struct _PyCode_DEF(70)
codecs_toplevel_consts_28_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_9_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 759,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 335,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reduce_ex__),
    .co_qualname = & codecs_toplevel_consts_28_consts_17_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_28_consts_19 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
codecs_toplevel_consts_28_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            & const_str_StreamReaderWriter._ascii.ob_base,
            & codecs_toplevel_consts_28_consts_1._ascii.ob_base,
            & const_str_unknown._ascii.ob_base,
            & codecs_toplevel_consts_28_consts_3.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_4.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_28_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_7.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_8.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_9.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_10.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_11.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_12.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_13.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_14.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_15.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_16.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_17.ob_base.ob_base,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_19._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_12._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[20];
        }_object;
    }
codecs_toplevel_consts_28_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 20,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(encoding),
            &_Py_ID(__init__),
            &_Py_ID(read),
            &_Py_ID(readline),
            & const_str_readlines._ascii.ob_base,
            &_Py_ID(__next__),
            &_Py_ID(__iter__),
            &_Py_ID(write),
            & const_str_writelines._ascii.ob_base,
            &_Py_ID(reset),
            &_Py_ID(seek),
            &_Py_ID(getattr),
            &_Py_ID(__getattr__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
            &_Py_ID(__reduce_ex__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[91];
    }
codecs_toplevel_consts_28_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 90,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x07\x05\x08\xf0\x12\x00\x10\x19\x80\x48\xf3\x04\x10\x05\x1d\xf3\x24\x02\x05\x26\xf3\x08\x02\x05\x2a\xf3\x08\x02\x05\x2f\xf2\x08\x03\x05\x21\xf2\x0a\x01\x05\x14\xf2\x06\x02\x05\x27\xf2\x08\x02\x05\x2c\xf2\x08\x03\x05\x1c\xf3\x0a\x04\x05\x20\xf0\x0e\x00\x1d\x24\xf3\x03\x05\x05\x2a\xf2\x12\x01\x05\x14\xf2\x06\x01\x05\x1c\xf3\x06\x01\x05\x48\x01",
};
static
    struct _PyCode_DEF(118)
codecs_toplevel_consts_28 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 59,
    },
    .co_consts = & codecs_toplevel_consts_28_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 674,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 336,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_StreamReaderWriter._ascii.ob_base,
    .co_qualname = & const_str_StreamReaderWriter._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x12\x64\x03\x84\x01\x5a\x05\x64\x13\x64\x04\x84\x01\x5a\x06\x64\x14\x64\x06\x84\x01\x5a\x07\x64\x14\x64\x07\x84\x01\x5a\x08\x64\x08\x84\x00\x5a\x09\x64\x09\x84\x00\x5a\x0a\x64\x0a\x84\x00\x5a\x0b\x64\x0b\x84\x00\x5a\x0c\x64\x0c\x84\x00\x5a\x0d\x64\x15\x64\x0d\x84\x01\x5a\x0e\x65\x0f\x66\x01\x64\x0e\x84\x01\x5a\x10\x64\x0f\x84\x00\x5a\x11\x64\x10\x84\x00\x5a\x12\x64\x11\x84\x00\x5a\x13\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[579];
    }
codecs_toplevel_consts_30_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 578,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x63\x6f\x64\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x73\x20\x74\x72\x61\x6e\x73\x6c\x61\x74\x65\x20\x64\x61\x74\x61\x20\x66\x72\x6f\x6d\x20\x6f\x6e\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x74\x6f\x20\x61\x6e\x6f\x74\x68\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x79\x20\x75\x73\x65\x20\x74\x68\x65\x20\x63\x6f\x6d\x70\x6c\x65\x74\x65\x20\x73\x65\x74\x20\x6f\x66\x20\x41\x50\x49\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x62\x79\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6f\x64\x65\x63\x73\x2e\x6c\x6f\x6f\x6b\x75\x70\x28\x29\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x74\x6f\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x20\x74\x68\x65\x69\x72\x20\x74\x61\x73\x6b\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x44\x61\x74\x61\x20\x77\x72\x69\x74\x74\x65\x6e\x20\x74\x6f\x20\x74\x68\x65\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x63\x6f\x64\x65\x72\x20\x69\x73\x20\x66\x69\x72\x73\x74\x20\x64\x65\x63\x6f\x64\x65\x64\x20\x69\x6e\x74\x6f\x20\x61\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x6e\x74\x65\x72\x6d\x65\x64\x69\x61\x74\x65\x20\x66\x6f\x72\x6d\x61\x74\x20\x28\x64\x65\x70\x65\x6e\x64\x69\x6e\x67\x20\x6f\x6e\x20\x74\x68\x65\x20\x22\x64\x65\x63\x6f\x64\x65\x22\x20\x63\x6f\x64\x65\x63\x29\x20\x61\x6e\x64\x20\x74\x68\x65\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x77\x72\x69\x74\x74\x65\x6e\x20\x74\x6f\x20\x74\x68\x65\x20\x75\x6e\x64\x65\x72\x6c\x79\x69\x6e\x67\x20\x73\x74\x72\x65\x61\x6d\x20\x75\x73\x69\x6e\x67\x20\x61\x6e\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x76\x69\x64\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x57\x72\x69\x74\x65\x72\x20\x63\x6c\x61\x73\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x6e\x20\x74\x68\x65\x20\x6f\x74\x68\x65\x72\x20\x64\x69\x72\x65\x63\x74\x69\x6f\x6e\x2c\x20\x64\x61\x74\x61\x20\x69\x73\x20\x72\x65\x61\x64\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x75\x6e\x64\x65\x72\x6c\x79\x69\x6e\x67\x20\x73\x74\x72\x65\x61\x6d\x20\x75\x73\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x20\x52\x65\x61\x64\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x20\x61\x6e\x64\x20\x74\x68\x65\x6e\x20\x65\x6e\x63\x6f\x64\x65\x64\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x74\x6f\x20\x74\x68\x65\x20\x63\x61\x6c\x6c\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[746];
    }
codecs_toplevel_consts_30_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 745,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x43\x72\x65\x61\x74\x65\x73\x20\x61\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x63\x6f\x64\x65\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x20\x77\x68\x69\x63\x68\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x73\x20\x61\x20\x74\x77\x6f\x2d\x77\x61\x79\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6f\x6e\x76\x65\x72\x73\x69\x6f\x6e\x3a\x20\x65\x6e\x63\x6f\x64\x65\x20\x61\x6e\x64\x20\x64\x65\x63\x6f\x64\x65\x20\x77\x6f\x72\x6b\x20\x6f\x6e\x20\x74\x68\x65\x20\x66\x72\x6f\x6e\x74\x65\x6e\x64\x20\x28\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x61\x74\x61\x20\x76\x69\x73\x69\x62\x6c\x65\x20\x74\x6f\x20\x2e\x72\x65\x61\x64\x28\x29\x20\x61\x6e\x64\x20\x2e\x77\x72\x69\x74\x65\x28\x29\x29\x20\x77\x68\x69\x6c\x65\x20\x52\x65\x61\x64\x65\x72\x20\x61\x6e\x64\x20\x57\x72\x69\x74\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x77\x6f\x72\x6b\x20\x6f\x6e\x20\x74\x68\x65\x20\x62\x61\x63\x6b\x65\x6e\x64\x20\x28\x74\x68\x65\x20\x64\x61\x74\x61\x20\x69\x6e\x20\x73\x74\x72\x65\x61\x6d\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x59\x6f\x75\x20\x63\x61\x6e\x20\x75\x73\x65\x20\x74\x68\x65\x73\x65\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x74\x6f\x20\x64\x6f\x20\x74\x72\x61\x6e\x73\x70\x61\x72\x65\x6e\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x74\x72\x61\x6e\x73\x63\x6f\x64\x69\x6e\x67\x73\x20\x66\x72\x6f\x6d\x20\x65\x2e\x67\x2e\x20\x6c\x61\x74\x69\x6e\x2d\x31\x20\x74\x6f\x20\x75\x74\x66\x2d\x38\x20\x61\x6e\x64\x20\x62\x61\x63\x6b\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x73\x74\x72\x65\x61\x6d\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x20\x66\x69\x6c\x65\x2d\x6c\x69\x6b\x65\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x63\x6f\x64\x65\x20\x61\x6e\x64\x20\x64\x65\x63\x6f\x64\x65\x20\x6d\x75\x73\x74\x20\x61\x64\x68\x65\x72\x65\x20\x74\x6f\x20\x74\x68\x65\x20\x43\x6f\x64\x65\x63\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x3b\x20\x52\x65\x61\x64\x65\x72\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x57\x72\x69\x74\x65\x72\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x66\x61\x63\x74\x6f\x72\x79\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x73\x20\x6f\x72\x20\x63\x6c\x61\x73\x73\x65\x73\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x20\x61\x6e\x64\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x73\x20\x72\x65\x73\x70\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x45\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x20\x69\x73\x20\x64\x6f\x6e\x65\x20\x69\x6e\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x77\x61\x79\x20\x61\x73\x20\x64\x65\x66\x69\x6e\x65\x64\x20\x66\x6f\x72\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x2f\x52\x65\x61\x64\x65\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_30_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_30_consts_3_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
codecs_toplevel_consts_30_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_stream._ascii.ob_base,
            &_Py_ID(encode),
            &_Py_ID(decode),
            & const_str_reader._ascii.ob_base,
            & const_str_writer._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_30_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[61];
    }
codecs_toplevel_consts_30_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 60,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x2a\x00\x17\x1d\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8c\x0b\xd9\x16\x1c\x98\x56\xa0\x56\xd3\x16\x2c\x88\x04\x8c\x0b\xd9\x16\x1c\x98\x56\xa0\x56\xd3\x16\x2c\x88\x04\x8c\x0b\xd8\x16\x1c\x88\x04\x8d\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
codecs_toplevel_consts_30_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_stream._ascii.ob_base,
            &_Py_ID(encode),
            &_Py_ID(decode),
            & const_str_Reader._ascii.ob_base,
            & const_str_Writer._ascii.ob_base,
            &_Py_ID(errors),
        },
    },
};
static
    struct _PyCode_DEF(116)
codecs_toplevel_consts_30_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 58,
    },
    .co_consts = & codecs_toplevel_consts_30_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 7,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 784,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 337,
    .co_localsplusnames = & codecs_toplevel_consts_30_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & codecs_toplevel_consts_30_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x7c\x04\x7c\x01\x7c\x06\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x7c\x05\x7c\x01\x7c\x06\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x00\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_30_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            &_Py_ID(read),
            &_Py_ID(encode),
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
codecs_toplevel_consts_30_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.read",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[53];
    }
codecs_toplevel_consts_30_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 52,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x1f\xd1\x0f\x1f\xa0\x04\xd3\x0f\x25\x88\x04\xd8\x1d\x21\x9f\x5b\x99\x5b\xa8\x14\xa8\x74\xaf\x7b\xa9\x7b\xd3\x1d\x3b\xd1\x08\x1a\x88\x04\x88\x6c\xd8\x0f\x13\x88\x0b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_bytesencoded = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytesencoded",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_30_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(size),
            &_Py_ID(data),
            & const_str_bytesencoded._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(122)
codecs_toplevel_consts_30_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 61,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 812,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 338,
    .co_localsplusnames = & codecs_toplevel_consts_30_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(read),
    .co_qualname = & codecs_toplevel_consts_30_consts_4_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_30_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            &_Py_ID(readline),
            &_Py_ID(encode),
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_30_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.readline",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[76];
    }
codecs_toplevel_consts_30_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 75,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0f\x88\x3c\xd8\x13\x17\x97\x3b\x91\x3b\xd7\x13\x27\xd1\x13\x27\xd3\x13\x29\x89\x44\xe0\x13\x17\x97\x3b\x91\x3b\xd7\x13\x27\xd1\x13\x27\xa8\x04\xd3\x13\x2d\x88\x44\xd8\x1d\x21\x9f\x5b\x99\x5b\xa8\x14\xa8\x74\xaf\x7b\xa9\x7b\xd3\x1d\x3b\xd1\x08\x1a\x88\x04\x88\x6c\xd8\x0f\x13\x88\x0b",
};
static
    struct _PyCode_DEF(180)
codecs_toplevel_consts_30_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 90,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 818,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 339,
    .co_localsplusnames = & codecs_toplevel_consts_30_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(readline),
    .co_qualname = & codecs_toplevel_consts_30_consts_6_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x80\x1b\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x6e\x1b\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_30_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            Py_True,
            & codecs_toplevel_consts_26_consts_3_consts_4._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
codecs_toplevel_consts_30_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            &_Py_ID(read),
            &_Py_ID(encode),
            &_Py_ID(errors),
            & const_str_splitlines._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
codecs_toplevel_consts_30_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.readlines",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
codecs_toplevel_consts_30_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x1f\xd1\x0f\x1f\xd3\x0f\x21\x88\x04\xd8\x1d\x21\x9f\x5b\x99\x5b\xa8\x14\xa8\x74\xaf\x7b\xa9\x7b\xd3\x1d\x3b\xd1\x08\x1a\x88\x04\x88\x6c\xd8\x0f\x13\x8f\x7f\x89\x7f\xa8\x04\x88\x7f\xd3\x0f\x2d\xd0\x08\x2d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_30_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(sizehint),
            &_Py_ID(data),
            & const_str_bytesencoded._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(152)
codecs_toplevel_consts_30_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 76,
    },
    .co_consts = & codecs_toplevel_consts_30_consts_7_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 827,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 340,
    .co_localsplusnames = & codecs_toplevel_consts_30_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_readlines._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_30_consts_7_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x02\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xac\x02\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_30_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(next),
            & const_str_reader._ascii.ob_base,
            &_Py_ID(encode),
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_30_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.__next__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
codecs_toplevel_consts_30_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x10\x14\x90\x44\x97\x4b\x91\x4b\xd3\x0f\x20\x88\x04\xd8\x1d\x21\x9f\x5b\x99\x5b\xa8\x14\xa8\x74\xaf\x7b\xa9\x7b\xd3\x1d\x3b\xd1\x08\x1a\x88\x04\x88\x6c\xd8\x0f\x13\x88\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_30_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(data),
            & const_str_bytesencoded._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(110)
codecs_toplevel_consts_30_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & codecs_toplevel_consts_26_consts_9_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 833,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 341,
    .co_localsplusnames = & codecs_toplevel_consts_30_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__next__),
    .co_qualname = & codecs_toplevel_consts_30_consts_8_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_30_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.__iter__",
};
static
    struct _PyCode_DEF(6)
codecs_toplevel_consts_30_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 840,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 342,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & codecs_toplevel_consts_30_consts_9_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_30_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(decode),
            &_Py_ID(errors),
            & const_str_writer._ascii.ob_base,
            &_Py_ID(write),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
codecs_toplevel_consts_30_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.write",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
codecs_toplevel_consts_30_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x1d\x21\x9f\x5b\x99\x5b\xa8\x14\xa8\x74\xaf\x7b\xa9\x7b\xd3\x1d\x3b\xd1\x08\x1a\x88\x04\x88\x6c\xd8\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x20\xd1\x0f\x20\xa0\x14\xd3\x0f\x26\xd0\x08\x26",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_bytesdecoded = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytesdecoded",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_30_consts_10_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(data),
            & const_str_bytesdecoded._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(118)
codecs_toplevel_consts_30_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 59,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 843,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 343,
    .co_localsplusnames = & codecs_toplevel_consts_30_consts_10_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(write),
    .co_qualname = & codecs_toplevel_consts_30_consts_10_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
codecs_toplevel_consts_30_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(join),
            &_Py_ID(decode),
            &_Py_ID(errors),
            & const_str_writer._ascii.ob_base,
            &_Py_ID(write),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
codecs_toplevel_consts_30_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.writelines",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[62];
    }
codecs_toplevel_consts_30_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 61,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x12\x8f\x78\x89\x78\x98\x04\x8b\x7e\x88\x04\xd8\x1d\x21\x9f\x5b\x99\x5b\xa8\x14\xa8\x74\xaf\x7b\xa9\x7b\xd3\x1d\x3b\xd1\x08\x1a\x88\x04\x88\x6c\xd8\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x20\xd1\x0f\x20\xa0\x14\xd3\x0f\x26\xd0\x08\x26",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_30_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_list._ascii.ob_base,
            &_Py_ID(data),
            & const_str_bytesdecoded._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(152)
codecs_toplevel_consts_30_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 76,
    },
    .co_consts = & codecs_toplevel_consts_22_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 848,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 344,
    .co_localsplusnames = & codecs_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_writelines._ascii.ob_base,
    .co_qualname = & codecs_toplevel_consts_30_consts_11_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
codecs_toplevel_consts_30_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.reset",
};
static
    struct _PyCode_DEF(108)
codecs_toplevel_consts_30_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 54,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_28_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 854,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 345,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reset),
    .co_qualname = & codecs_toplevel_consts_30_consts_12_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_28_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_30_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_reader._ascii.ob_base,
            &_Py_ID(seek),
            & const_str_writer._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
codecs_toplevel_consts_30_consts_13_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.seek",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
codecs_toplevel_consts_30_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x09\x0d\x8f\x0b\x89\x0b\xd7\x08\x18\xd1\x08\x18\x98\x16\xa0\x16\xd4\x08\x28\xd8\x08\x0c\x8f\x0b\x89\x0b\xd7\x08\x18\xd1\x08\x18\x98\x16\xa0\x16\xd5\x08\x28",
};
static
    struct _PyCode_DEF(116)
codecs_toplevel_consts_30_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 58,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 859,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 346,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(seek),
    .co_qualname = & codecs_toplevel_consts_30_consts_13_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
codecs_toplevel_consts_30_consts_14_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.__getattr__",
};
static
    struct _PyCode_DEF(40)
codecs_toplevel_consts_30_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_6_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 865,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 347,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getattr__),
    .co_qualname = & codecs_toplevel_consts_30_consts_14_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x7c\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
codecs_toplevel_consts_30_consts_15_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.__enter__",
};
static
    struct _PyCode_DEF(6)
codecs_toplevel_consts_30_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 872,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 348,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & codecs_toplevel_consts_30_consts_15_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
codecs_toplevel_consts_30_consts_16_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.__exit__",
};
static
    struct _PyCode_DEF(56)
codecs_toplevel_consts_30_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 875,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 349,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & codecs_toplevel_consts_30_consts_16_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
codecs_toplevel_consts_30_consts_17_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StreamRecoder.__reduce_ex__",
};
static
    struct _PyCode_DEF(70)
codecs_toplevel_consts_30_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & codecs_toplevel_consts_24_consts_9_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_24_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 878,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 350,
    .co_localsplusnames = & codecs_toplevel_consts_24_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reduce_ex__),
    .co_qualname = & codecs_toplevel_consts_30_consts_17_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
codecs_toplevel_consts_30_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            & const_str_StreamRecoder._ascii.ob_base,
            & codecs_toplevel_consts_30_consts_1._ascii.ob_base,
            & const_str_unknown._ascii.ob_base,
            & codecs_toplevel_consts_30_consts_3.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_4.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_30_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_7.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_8.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_9.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_10.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_11.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_12.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_13.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_14.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_15.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_16.ob_base.ob_base,
            & codecs_toplevel_consts_30_consts_17.ob_base.ob_base,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_28_consts_19._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
            & codecs_toplevel_consts_24_consts_12._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_data_encoding = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "data_encoding",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_file_encoding = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "file_encoding",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[21];
        }_object;
    }
codecs_toplevel_consts_30_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 21,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_data_encoding._ascii.ob_base,
            & const_str_file_encoding._ascii.ob_base,
            &_Py_ID(__init__),
            &_Py_ID(read),
            &_Py_ID(readline),
            & const_str_readlines._ascii.ob_base,
            &_Py_ID(__next__),
            &_Py_ID(__iter__),
            &_Py_ID(write),
            & const_str_writelines._ascii.ob_base,
            &_Py_ID(reset),
            &_Py_ID(seek),
            &_Py_ID(getattr),
            &_Py_ID(__getattr__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
            &_Py_ID(__reduce_ex__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[101];
    }
codecs_toplevel_consts_30_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 100,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x04\x0d\x05\x08\xf0\x1e\x00\x15\x1e\x80\x4d\xd8\x14\x1d\x80\x4d\xf0\x06\x00\x19\x21\xf3\x03\x1a\x05\x1d\xf3\x38\x04\x05\x14\xf3\x0c\x07\x05\x14\xf3\x12\x04\x05\x2e\xf2\x0c\x05\x05\x14\xf2\x0e\x01\x05\x14\xf2\x06\x03\x05\x27\xf2\x0a\x04\x05\x27\xf2\x0c\x03\x05\x1c\xf3\x0a\x04\x05\x29\xf0\x0e\x00\x1d\x24\xf3\x03\x05\x05\x2a\xf2\x0e\x01\x05\x14\xf2\x06\x01\x05\x1c\xf3\x06\x01\x05\x48\x01",
};
static
    struct _PyCode_DEF(124)
codecs_toplevel_consts_30 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 62,
    },
    .co_consts = & codecs_toplevel_consts_30_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_30_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 764,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 351,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_StreamRecoder._ascii.ob_base,
    .co_qualname = & const_str_StreamRecoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_30_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x02\x5a\x05\x09\x00\x64\x12\x64\x03\x84\x01\x5a\x06\x64\x13\x64\x04\x84\x01\x5a\x07\x64\x14\x64\x06\x84\x01\x5a\x08\x64\x14\x64\x07\x84\x01\x5a\x09\x64\x08\x84\x00\x5a\x0a\x64\x09\x84\x00\x5a\x0b\x64\x0a\x84\x00\x5a\x0c\x64\x0b\x84\x00\x5a\x0d\x64\x0c\x84\x00\x5a\x0e\x64\x15\x64\x0d\x84\x01\x5a\x0f\x65\x10\x66\x01\x64\x0e\x84\x01\x5a\x11\x64\x0f\x84\x00\x5a\x12\x64\x10\x84\x00\x5a\x13\x64\x11\x84\x00\x5a\x14\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1180];
    }
codecs_toplevel_consts_33_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1179,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x4f\x70\x65\x6e\x20\x61\x6e\x20\x65\x6e\x63\x6f\x64\x65\x64\x20\x66\x69\x6c\x65\x20\x75\x73\x69\x6e\x67\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x6d\x6f\x64\x65\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x20\x77\x72\x61\x70\x70\x65\x64\x20\x76\x65\x72\x73\x69\x6f\x6e\x20\x70\x72\x6f\x76\x69\x64\x69\x6e\x67\x20\x74\x72\x61\x6e\x73\x70\x61\x72\x65\x6e\x74\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x2f\x64\x65\x63\x6f\x64\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4e\x6f\x74\x65\x3a\x20\x54\x68\x65\x20\x77\x72\x61\x70\x70\x65\x64\x20\x76\x65\x72\x73\x69\x6f\x6e\x20\x77\x69\x6c\x6c\x20\x6f\x6e\x6c\x79\x20\x61\x63\x63\x65\x70\x74\x20\x74\x68\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x66\x6f\x72\x6d\x61\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x64\x65\x66\x69\x6e\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x73\x2c\x20\x69\x2e\x65\x2e\x20\x55\x6e\x69\x63\x6f\x64\x65\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x66\x6f\x72\x20\x6d\x6f\x73\x74\x20\x62\x75\x69\x6c\x74\x69\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x6f\x64\x65\x63\x73\x2e\x20\x4f\x75\x74\x70\x75\x74\x20\x69\x73\x20\x61\x6c\x73\x6f\x20\x63\x6f\x64\x65\x63\x20\x64\x65\x70\x65\x6e\x64\x65\x6e\x74\x20\x61\x6e\x64\x20\x77\x69\x6c\x6c\x20\x75\x73\x75\x61\x6c\x6c\x79\x20\x62\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x55\x6e\x69\x63\x6f\x64\x65\x20\x61\x73\x20\x77\x65\x6c\x6c\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x69\x73\x20\x6e\x6f\x74\x20\x4e\x6f\x6e\x65\x2c\x20\x74\x68\x65\x6e\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x75\x6e\x64\x65\x72\x6c\x79\x69\x6e\x67\x20\x65\x6e\x63\x6f\x64\x65\x64\x20\x66\x69\x6c\x65\x73\x20\x61\x72\x65\x20\x61\x6c\x77\x61\x79\x73\x20\x6f\x70\x65\x6e\x65\x64\x20\x69\x6e\x20\x62\x69\x6e\x61\x72\x79\x20\x6d\x6f\x64\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x66\x69\x6c\x65\x20\x6d\x6f\x64\x65\x20\x69\x73\x20\x27\x72\x27\x2c\x20\x6d\x65\x61\x6e\x69\x6e\x67\x20\x74\x6f\x20\x6f\x70\x65\x6e\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x69\x6e\x20\x72\x65\x61\x64\x20\x6d\x6f\x64\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x73\x70\x65\x63\x69\x66\x69\x65\x73\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x77\x68\x69\x63\x68\x20\x69\x73\x20\x74\x6f\x20\x62\x65\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x72\x72\x6f\x72\x73\x20\x6d\x61\x79\x20\x62\x65\x20\x67\x69\x76\x65\x6e\x20\x74\x6f\x20\x64\x65\x66\x69\x6e\x65\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x2e\x20\x49\x74\x20\x64\x65\x66\x61\x75\x6c\x74\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x74\x6f\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x77\x68\x69\x63\x68\x20\x63\x61\x75\x73\x65\x73\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x73\x20\x74\x6f\x20\x62\x65\x20\x72\x61\x69\x73\x65\x64\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x61\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x65\x72\x72\x6f\x72\x20\x6f\x63\x63\x75\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x62\x75\x66\x66\x65\x72\x69\x6e\x67\x20\x68\x61\x73\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x6d\x65\x61\x6e\x69\x6e\x67\x20\x61\x73\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x62\x75\x69\x6c\x74\x69\x6e\x20\x6f\x70\x65\x6e\x28\x29\x20\x41\x50\x49\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x74\x20\x64\x65\x66\x61\x75\x6c\x74\x73\x20\x74\x6f\x20\x2d\x31\x20\x77\x68\x69\x63\x68\x20\x6d\x65\x61\x6e\x73\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x62\x75\x66\x66\x65\x72\x20\x73\x69\x7a\x65\x20\x77\x69\x6c\x6c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x62\x65\x20\x75\x73\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x77\x72\x61\x70\x70\x65\x64\x20\x66\x69\x6c\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x61\x6e\x20\x65\x78\x74\x72\x61\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x2e\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x77\x68\x69\x63\x68\x20\x61\x6c\x6c\x6f\x77\x73\x20\x71\x75\x65\x72\x79\x69\x6e\x67\x20\x74\x68\x65\x20\x75\x73\x65\x64\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x2e\x20\x54\x68\x69\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x20\x69\x73\x20\x6f\x6e\x6c\x79\x20\x61\x76\x61\x69\x6c\x61\x62\x6c\x65\x20\x69\x66\x20\x61\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x77\x61\x73\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x61\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_33_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_33_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[98],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
codecs_toplevel_consts_33_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(builtins),
            &_Py_ID(open),
            & const_str_lookup._ascii.ob_base,
            & const_str_StreamReaderWriter._ascii.ob_base,
            & const_str_streamreader._ascii.ob_base,
            & const_str_streamwriter._ascii.ob_base,
            &_Py_ID(encoding),
            &_Py_ID(close),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[130];
    }
codecs_toplevel_consts_33_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 129,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x3e\x00\x08\x10\xd0\x07\x1b\xd8\x07\x0a\x90\x24\x81\x7f\xe0\x0f\x13\x90\x63\x89\x7a\x88\x04\xdc\x0b\x13\x8f\x3d\x89\x3d\x98\x18\xa0\x34\xa8\x19\xd3\x0b\x33\x80\x44\xd8\x07\x0f\xd0\x07\x17\xd8\x0f\x13\x88\x0b\xf0\x04\x08\x05\x0e\xdc\x0f\x15\x90\x68\xd3\x0f\x1f\x88\x04\xdc\x0e\x20\xa0\x14\xa0\x74\xd7\x27\x38\xd1\x27\x38\xb8\x24\xd7\x3a\x4b\xd1\x3a\x4b\xc8\x56\xd3\x0e\x54\x88\x03\xe0\x17\x1f\x88\x03\x8c\x0c\xd8\x0f\x12\x88\x0a\xf8\xf0\x02\x02\x05\x0e\xd8\x08\x0c\x8f\x0a\x89\x0a\x8c\x0c\xd8\x08\x0d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
codecs_toplevel_consts_33_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "\xa8\x35\x41\x1e\x00\xc1\x1e\x13\x41\x31\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_srw = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "srw",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
codecs_toplevel_consts_33_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(filename),
            &_Py_ID(mode),
            &_Py_ID(encoding),
            &_Py_ID(errors),
            &_Py_ID(buffering),
            &_Py_ID(file),
            & const_str_info._ascii.ob_base,
            & const_str_srw._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(232)
codecs_toplevel_consts_33 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 116,
    },
    .co_consts = & codecs_toplevel_consts_33_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_33_names._object.ob_base.ob_base,
    .co_exceptiontable = & codecs_toplevel_consts_33_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 883,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 352,
    .co_localsplusnames = & codecs_toplevel_consts_33_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(open),
    .co_qualname = &_Py_ID(open),
    .co_linetable = & codecs_toplevel_consts_33_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x81\x09\x64\x01\x7c\x01\x76\x01\x72\x05\x7c\x01\x64\x01\x7a\x00\x00\x00\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x04\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x02\x80\x02\x7c\x05\x53\x00\x09\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x06\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x04\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x02\x7c\x07\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x53\x00\x23\x00\x01\x00\x7c\x05\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x82\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[987];
    }
codecs_toplevel_consts_34_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 986,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x52\x65\x74\x75\x72\x6e\x20\x61\x20\x77\x72\x61\x70\x70\x65\x64\x20\x76\x65\x72\x73\x69\x6f\x6e\x20\x6f\x66\x20\x66\x69\x6c\x65\x20\x77\x68\x69\x63\x68\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x74\x72\x61\x6e\x73\x70\x61\x72\x65\x6e\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x74\x72\x61\x6e\x73\x6c\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x44\x61\x74\x61\x20\x77\x72\x69\x74\x74\x65\x6e\x20\x74\x6f\x20\x74\x68\x65\x20\x77\x72\x61\x70\x70\x65\x64\x20\x66\x69\x6c\x65\x20\x69\x73\x20\x64\x65\x63\x6f\x64\x65\x64\x20\x61\x63\x63\x6f\x72\x64\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x74\x6f\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x64\x61\x74\x61\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x74\x68\x65\x6e\x20\x65\x6e\x63\x6f\x64\x65\x64\x20\x74\x6f\x20\x74\x68\x65\x20\x75\x6e\x64\x65\x72\x6c\x79\x69\x6e\x67\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x69\x6c\x65\x20\x75\x73\x69\x6e\x67\x20\x66\x69\x6c\x65\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x2e\x20\x54\x68\x65\x20\x69\x6e\x74\x65\x72\x6d\x65\x64\x69\x61\x74\x65\x20\x64\x61\x74\x61\x20\x74\x79\x70\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x77\x69\x6c\x6c\x20\x75\x73\x75\x61\x6c\x6c\x79\x20\x62\x65\x20\x55\x6e\x69\x63\x6f\x64\x65\x20\x62\x75\x74\x20\x64\x65\x70\x65\x6e\x64\x73\x20\x6f\x6e\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x63\x6f\x64\x65\x63\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x42\x79\x74\x65\x73\x20\x72\x65\x61\x64\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x61\x72\x65\x20\x64\x65\x63\x6f\x64\x65\x64\x20\x75\x73\x69\x6e\x67\x20\x66\x69\x6c\x65\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x74\x68\x65\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x61\x73\x73\x65\x64\x20\x62\x61\x63\x6b\x20\x74\x6f\x20\x74\x68\x65\x20\x63\x61\x6c\x6c\x65\x72\x20\x65\x6e\x63\x6f\x64\x65\x64\x20\x75\x73\x69\x6e\x67\x20\x64\x61\x74\x61\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x66\x69\x6c\x65\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x69\x73\x20\x6e\x6f\x74\x20\x67\x69\x76\x65\x6e\x2c\x20\x69\x74\x20\x64\x65\x66\x61\x75\x6c\x74\x73\x20\x74\x6f\x20\x64\x61\x74\x61\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x72\x72\x6f\x72\x73\x20\x6d\x61\x79\x20\x62\x65\x20\x67\x69\x76\x65\x6e\x20\x74\x6f\x20\x64\x65\x66\x69\x6e\x65\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x2e\x20\x49\x74\x20\x64\x65\x66\x61\x75\x6c\x74\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x74\x6f\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x77\x68\x69\x63\x68\x20\x63\x61\x75\x73\x65\x73\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x73\x20\x74\x6f\x20\x62\x65\x20\x72\x61\x69\x73\x65\x64\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x61\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x65\x72\x72\x6f\x72\x20\x6f\x63\x63\x75\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x77\x72\x61\x70\x70\x65\x64\x20\x66\x69\x6c\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x74\x77\x6f\x20\x65\x78\x74\x72\x61\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x2e\x64\x61\x74\x61\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x2e\x66\x69\x6c\x65\x5f\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x77\x68\x69\x63\x68\x20\x72\x65\x66\x6c\x65\x63\x74\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x61\x72\x61\x6d\x65\x74\x65\x72\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x6e\x61\x6d\x65\x2e\x20\x54\x68\x65\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x6e\x74\x72\x6f\x73\x70\x65\x63\x74\x69\x6f\x6e\x20\x62\x79\x20\x50\x79\x74\x68\x6f\x6e\x20\x70\x72\x6f\x67\x72\x61\x6d\x73\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_34_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_34_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
codecs_toplevel_consts_34_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_lookup._ascii.ob_base,
            & const_str_StreamRecoder._ascii.ob_base,
            &_Py_ID(encode),
            &_Py_ID(decode),
            & const_str_streamreader._ascii.ob_base,
            & const_str_streamwriter._ascii.ob_base,
            & const_str_data_encoding._ascii.ob_base,
            & const_str_file_encoding._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[107];
    }
codecs_toplevel_consts_34_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 106,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x32\x00\x08\x15\xd0\x07\x1c\xd8\x18\x25\x88\x0d\xdc\x10\x16\x90\x7d\xd3\x10\x25\x80\x49\xdc\x10\x16\x90\x7d\xd3\x10\x25\x80\x49\xdc\x09\x16\x90\x74\x98\x59\xd7\x1d\x2d\xd1\x1d\x2d\xa8\x79\xd7\x2f\x3f\xd1\x2f\x3f\xd8\x17\x20\xd7\x17\x2d\xd1\x17\x2d\xa8\x79\xd7\x2f\x45\xd1\x2f\x45\xc0\x76\xf3\x03\x01\x0a\x4f\x01\x80\x42\xf0\x06\x00\x18\x25\x80\x42\xd4\x04\x14\xd8\x17\x24\x80\x42\xd4\x04\x14\xd8\x0b\x0d\x80\x49",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_data_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "data_info",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_file_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "file_info",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_sr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sr",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
codecs_toplevel_consts_34_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(file),
            & const_str_data_encoding._ascii.ob_base,
            & const_str_file_encoding._ascii.ob_base,
            &_Py_ID(errors),
            & const_str_data_info._ascii.ob_base,
            & const_str_file_info._ascii.ob_base,
            & const_str_sr._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(198)
codecs_toplevel_consts_34 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 99,
    },
    .co_consts = & codecs_toplevel_consts_34_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_34_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 932,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 353,
    .co_localsplusnames = & codecs_toplevel_consts_34_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_EncodedFile._ascii.ob_base,
    .co_qualname = & const_str_EncodedFile._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_34_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x80\x02\x7c\x01\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x04\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x06\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x01\x7c\x06\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x06\x5f\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[159];
    }
codecs_toplevel_consts_35_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 158,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x4c\x6f\x6f\x6b\x75\x70\x20\x75\x70\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x74\x73\x20\x65\x6e\x63\x6f\x64\x65\x72\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x61\x20\x4c\x6f\x6f\x6b\x75\x70\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_35_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_35_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_35_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_lookup._ascii.ob_base,
            &_Py_ID(encode),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
codecs_toplevel_consts_35_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x0c\x12\x90\x28\xd3\x0b\x1b\xd7\x0b\x22\xd1\x0b\x22\xd0\x04\x22",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_35_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(encoding),
        },
    },
};
static
    struct _PyCode_DEF(44)
codecs_toplevel_consts_35 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & codecs_toplevel_consts_35_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_35_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 970,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 354,
    .co_localsplusnames = & codecs_toplevel_consts_35_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getencoder._ascii.ob_base,
    .co_qualname = & const_str_getencoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_35_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[159];
    }
codecs_toplevel_consts_36_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 158,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x4c\x6f\x6f\x6b\x75\x70\x20\x75\x70\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x74\x73\x20\x64\x65\x63\x6f\x64\x65\x72\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x61\x20\x4c\x6f\x6f\x6b\x75\x70\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_36_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_36_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_36_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_lookup._ascii.ob_base,
            &_Py_ID(decode),
        },
    },
};
static
    struct _PyCode_DEF(44)
codecs_toplevel_consts_36 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & codecs_toplevel_consts_36_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_36_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 980,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 355,
    .co_localsplusnames = & codecs_toplevel_consts_35_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getdecoder._ascii.ob_base,
    .co_qualname = & const_str_getdecoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_35_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[248];
    }
codecs_toplevel_consts_37_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 247,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x4c\x6f\x6f\x6b\x75\x70\x20\x75\x70\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x74\x73\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x6f\x72\x20\x66\x61\x63\x74\x6f\x72\x79\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x61\x20\x4c\x6f\x6f\x6b\x75\x70\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x72\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x73\x20\x64\x6f\x65\x73\x6e\x27\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x20\x61\x6e\x20\x69\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x20\x65\x6e\x63\x6f\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_37_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_37_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_LookupError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LookupError",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_37_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_lookup._ascii.ob_base,
            & const_str_incrementalencoder._ascii.ob_base,
            & const_str_LookupError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[42];
    }
codecs_toplevel_consts_37_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 41,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x12\x00\x0f\x15\x90\x58\xd3\x0e\x1e\xd7\x0e\x31\xd1\x0e\x31\x80\x47\xd8\x07\x0e\x80\x7f\xdc\x0e\x19\x98\x28\xd3\x0e\x23\xd0\x08\x23\xd8\x0b\x12\x80\x4e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_encoder = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "encoder",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_37_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(encoding),
            & const_str_encoder._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(74)
codecs_toplevel_consts_37 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 37,
    },
    .co_consts = & codecs_toplevel_consts_37_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_37_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 990,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 356,
    .co_localsplusnames = & codecs_toplevel_consts_37_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getincrementalencoder._ascii.ob_base,
    .co_qualname = & const_str_getincrementalencoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_37_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x80\x0b\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[248];
    }
codecs_toplevel_consts_38_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 247,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x4c\x6f\x6f\x6b\x75\x70\x20\x75\x70\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x74\x73\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x6f\x72\x20\x66\x61\x63\x74\x6f\x72\x79\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x61\x20\x4c\x6f\x6f\x6b\x75\x70\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x72\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x73\x20\x64\x6f\x65\x73\x6e\x27\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x20\x61\x6e\x20\x69\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x20\x64\x65\x63\x6f\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_38_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_38_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
codecs_toplevel_consts_38_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_lookup._ascii.ob_base,
            & const_str_incrementaldecoder._ascii.ob_base,
            & const_str_LookupError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_38_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(encoding),
            &_Py_ID(decoder),
        },
    },
};
static
    struct _PyCode_DEF(74)
codecs_toplevel_consts_38 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 37,
    },
    .co_consts = & codecs_toplevel_consts_38_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_38_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1004,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 357,
    .co_localsplusnames = & codecs_toplevel_consts_38_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getincrementaldecoder._ascii.ob_base,
    .co_qualname = & const_str_getincrementaldecoder._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_37_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x80\x0b\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[181];
    }
codecs_toplevel_consts_39_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 180,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x4c\x6f\x6f\x6b\x75\x70\x20\x75\x70\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x74\x73\x20\x53\x74\x72\x65\x61\x6d\x52\x65\x61\x64\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x6f\x72\x20\x66\x61\x63\x74\x6f\x72\x79\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x61\x20\x4c\x6f\x6f\x6b\x75\x70\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_39_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_39_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_39_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_lookup._ascii.ob_base,
            & const_str_streamreader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
codecs_toplevel_consts_39_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x0c\x12\x90\x28\xd3\x0b\x1b\xd7\x0b\x28\xd1\x0b\x28\xd0\x04\x28",
};
static
    struct _PyCode_DEF(44)
codecs_toplevel_consts_39 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & codecs_toplevel_consts_39_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_39_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1018,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 358,
    .co_localsplusnames = & codecs_toplevel_consts_35_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getreader._ascii.ob_base,
    .co_qualname = & const_str_getreader._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_39_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[181];
    }
codecs_toplevel_consts_40_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 180,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x4c\x6f\x6f\x6b\x75\x70\x20\x75\x70\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x67\x69\x76\x65\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x74\x73\x20\x53\x74\x72\x65\x61\x6d\x57\x72\x69\x74\x65\x72\x20\x63\x6c\x61\x73\x73\x20\x6f\x72\x20\x66\x61\x63\x74\x6f\x72\x79\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x61\x20\x4c\x6f\x6f\x6b\x75\x70\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x63\x61\x73\x65\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x63\x61\x6e\x6e\x6f\x74\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_40_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_40_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_40_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_lookup._ascii.ob_base,
            & const_str_streamwriter._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(44)
codecs_toplevel_consts_40 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & codecs_toplevel_consts_40_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_40_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1028,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 359,
    .co_localsplusnames = & codecs_toplevel_consts_35_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getwriter._ascii.ob_base,
    .co_qualname = & const_str_getwriter._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_39_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[192];
    }
codecs_toplevel_consts_41_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 191,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x45\x6e\x63\x6f\x64\x69\x6e\x67\x20\x69\x74\x65\x72\x61\x74\x6f\x72\x2e\x0a\x0a\x20\x20\x20\x20\x45\x6e\x63\x6f\x64\x65\x73\x20\x74\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x73\x74\x72\x69\x6e\x67\x73\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x69\x74\x65\x72\x61\x74\x6f\x72\x20\x75\x73\x69\x6e\x67\x20\x61\x6e\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x65\x72\x72\x6f\x72\x73\x20\x61\x6e\x64\x20\x6b\x77\x61\x72\x67\x73\x20\x61\x72\x65\x20\x70\x61\x73\x73\x65\x64\x20\x74\x68\x72\x6f\x75\x67\x68\x20\x74\x6f\x20\x74\x68\x65\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x45\x6e\x63\x6f\x64\x65\x72\x0a\x20\x20\x20\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_41_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & codecs_toplevel_consts_41_consts_0._ascii.ob_base,
            &_Py_STR(empty),
            Py_True,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_41_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_getincrementalencoder._ascii.ob_base,
            &_Py_ID(encode),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[95];
    }
codecs_toplevel_consts_41_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 94,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf0\x12\x00\x0f\x2e\xd4\x0e\x23\xa0\x48\xd3\x0e\x2d\xa8\x66\xd1\x0e\x3f\xb8\x06\xd1\x0e\x3f\x80\x47\xdb\x11\x19\x88\x05\xd8\x11\x18\x97\x1e\x91\x1e\xa0\x05\xd3\x11\x26\x88\x06\xda\x0b\x11\xd8\x12\x18\x8b\x4c\xf0\x07\x00\x12\x1a\xf0\x08\x00\x0e\x15\x8f\x5e\x89\x5e\x98\x42\xa0\x04\xd3\x0d\x25\x80\x46\xd9\x07\x0d\xd8\x0e\x14\x8b\x0c\xf0\x03\x00\x08\x0e\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
codecs_toplevel_consts_41_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x2b\x41\x0e\x01\xae\x20\x41\x0e\x01",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_output = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "output",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
codecs_toplevel_consts_41_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_iterator._ascii.ob_base,
            &_Py_ID(encoding),
            &_Py_ID(errors),
            & const_str_kwargs._ascii.ob_base,
            & const_str_encoder._ascii.ob_base,
            &_Py_ID(input),
            & const_str_output._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(160)
codecs_toplevel_consts_41 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 80,
    },
    .co_consts = & codecs_toplevel_consts_41_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_41_names._object.ob_base.ob_base,
    .co_exceptiontable = & codecs_toplevel_consts_41_exceptiontable.ob_base.ob_base,
    .co_flags = 43,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1038,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 360,
    .co_localsplusnames = & codecs_toplevel_consts_41_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_iterencode._ascii.ob_base,
    .co_qualname = & const_str_iterencode._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_41_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x02\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x66\x01\x69\x00\x7c\x03\xa4\x01\x8e\x01\x7d\x04\x7c\x00\x44\x00\x5d\x1a\x00\x00\x7d\x05\x7c\x04\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x73\x01\x8c\x17\x7c\x06\x96\x01\x97\x01\x01\x00\x8c\x1c\x04\x00\x7c\x04\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x72\x05\x7c\x06\x96\x01\x97\x01\x01\x00\x79\x03\x79\x03\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[192];
    }
codecs_toplevel_consts_42_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 191,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x20\x20\x20\x20\x44\x65\x63\x6f\x64\x69\x6e\x67\x20\x69\x74\x65\x72\x61\x74\x6f\x72\x2e\x0a\x0a\x20\x20\x20\x20\x44\x65\x63\x6f\x64\x65\x73\x20\x74\x68\x65\x20\x69\x6e\x70\x75\x74\x20\x73\x74\x72\x69\x6e\x67\x73\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x69\x74\x65\x72\x61\x74\x6f\x72\x20\x75\x73\x69\x6e\x67\x20\x61\x6e\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x65\x72\x72\x6f\x72\x73\x20\x61\x6e\x64\x20\x6b\x77\x61\x72\x67\x73\x20\x61\x72\x65\x20\x70\x61\x73\x73\x65\x64\x20\x74\x68\x72\x6f\x75\x67\x68\x20\x74\x6f\x20\x74\x68\x65\x20\x49\x6e\x63\x72\x65\x6d\x65\x6e\x74\x61\x6c\x44\x65\x63\x6f\x64\x65\x72\x0a\x20\x20\x20\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_42_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & codecs_toplevel_consts_42_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            Py_True,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_42_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_getincrementaldecoder._ascii.ob_base,
            &_Py_ID(decode),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[95];
    }
codecs_toplevel_consts_42_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 94,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf0\x12\x00\x0f\x2e\xd4\x0e\x23\xa0\x48\xd3\x0e\x2d\xa8\x66\xd1\x0e\x3f\xb8\x06\xd1\x0e\x3f\x80\x47\xdb\x11\x19\x88\x05\xd8\x11\x18\x97\x1e\x91\x1e\xa0\x05\xd3\x11\x26\x88\x06\xda\x0b\x11\xd8\x12\x18\x8b\x4c\xf0\x07\x00\x12\x1a\xf0\x08\x00\x0e\x15\x8f\x5e\x89\x5e\x98\x43\xa0\x14\xd3\x0d\x26\x80\x46\xd9\x07\x0d\xd8\x0e\x14\x8b\x0c\xf0\x03\x00\x08\x0e\xf9",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
codecs_toplevel_consts_42_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_iterator._ascii.ob_base,
            &_Py_ID(encoding),
            &_Py_ID(errors),
            & const_str_kwargs._ascii.ob_base,
            &_Py_ID(decoder),
            &_Py_ID(input),
            & const_str_output._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(160)
codecs_toplevel_consts_42 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 80,
    },
    .co_consts = & codecs_toplevel_consts_42_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_42_names._object.ob_base.ob_base,
    .co_exceptiontable = & codecs_toplevel_consts_41_exceptiontable.ob_base.ob_base,
    .co_flags = 43,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1056,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 361,
    .co_localsplusnames = & codecs_toplevel_consts_42_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_iterdecode._ascii.ob_base,
    .co_qualname = & const_str_iterdecode._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_42_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x02\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x66\x01\x69\x00\x7c\x03\xa4\x01\x8e\x01\x7d\x04\x7c\x00\x44\x00\x5d\x1a\x00\x00\x7d\x05\x7c\x04\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x73\x01\x8c\x17\x7c\x06\x96\x01\x97\x01\x01\x00\x8c\x1c\x04\x00\x7c\x04\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x72\x05\x7c\x06\x96\x01\x97\x01\x01\x00\x79\x03\x79\x03\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[137];
    }
codecs_toplevel_consts_43_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 136,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x6d\x61\x6b\x65\x5f\x69\x64\x65\x6e\x74\x69\x74\x79\x5f\x64\x69\x63\x74\x28\x72\x6e\x67\x29\x20\x2d\x3e\x20\x64\x69\x63\x74\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x61\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x77\x68\x65\x72\x65\x20\x65\x6c\x65\x6d\x65\x6e\x74\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x72\x6e\x67\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x20\x61\x72\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6d\x61\x70\x70\x65\x64\x20\x74\x6f\x20\x74\x68\x65\x6d\x73\x65\x6c\x76\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_43_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & codecs_toplevel_consts_43_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_make_identity_dict = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "make_identity_dict",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
codecs_toplevel_consts_43_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf1\x10\x00\x1a\x1d\xd3\x0b\x1d\x99\x13\x90\x41\x88\x41\x88\x61\x89\x43\x98\x13\xd1\x0b\x1d\xd0\x04\x1d\xf9\xd2\x0b\x1d",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
codecs_toplevel_consts_43_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x85\x0a\x12\x04",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_rng = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rng",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_43_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_rng._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
        },
    },
};
static
    struct _PyCode_DEF(46)
codecs_toplevel_consts_43 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & codecs_toplevel_consts_43_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & codecs_toplevel_consts_43_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1076,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 362,
    .co_localsplusnames = & codecs_toplevel_consts_43_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_make_identity_dict._ascii.ob_base,
    .co_qualname = & const_str_make_identity_dict._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_43_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x44\x00\x8f\x01\x63\x02\x69\x00\x63\x02\x5d\x05\x00\x00\x7d\x01\x7c\x01\x7c\x01\x93\x02\x8c\x07\x04\x00\x63\x02\x7d\x01\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x01\x77\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[387];
    }
codecs_toplevel_consts_44_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 386,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x43\x72\x65\x61\x74\x65\x73\x20\x61\x6e\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x6d\x61\x70\x20\x66\x72\x6f\x6d\x20\x61\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x20\x6d\x61\x70\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x61\x20\x74\x61\x72\x67\x65\x74\x20\x6d\x61\x70\x70\x69\x6e\x67\x20\x69\x6e\x20\x74\x68\x65\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x20\x6d\x61\x70\x20\x6f\x63\x63\x75\x72\x73\x20\x6d\x75\x6c\x74\x69\x70\x6c\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x74\x69\x6d\x65\x73\x2c\x20\x74\x68\x65\x6e\x20\x74\x68\x61\x74\x20\x74\x61\x72\x67\x65\x74\x20\x69\x73\x20\x6d\x61\x70\x70\x65\x64\x20\x74\x6f\x20\x4e\x6f\x6e\x65\x20\x28\x75\x6e\x64\x65\x66\x69\x6e\x65\x64\x20\x6d\x61\x70\x70\x69\x6e\x67\x29\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x61\x75\x73\x69\x6e\x67\x20\x61\x6e\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x77\x68\x65\x6e\x20\x65\x6e\x63\x6f\x75\x6e\x74\x65\x72\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x63\x68\x61\x72\x6d\x61\x70\x20\x63\x6f\x64\x65\x63\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x64\x75\x72\x69\x6e\x67\x20\x74\x72\x61\x6e\x73\x6c\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4f\x6e\x65\x20\x65\x78\x61\x6d\x70\x6c\x65\x20\x77\x68\x65\x72\x65\x20\x74\x68\x69\x73\x20\x68\x61\x70\x70\x65\x6e\x73\x20\x69\x73\x20\x63\x70\x38\x37\x35\x2e\x70\x79\x20\x77\x68\x69\x63\x68\x20\x64\x65\x63\x6f\x64\x65\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6d\x75\x6c\x74\x69\x70\x6c\x65\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x20\x74\x6f\x20\x5c\x75\x30\x30\x31\x61\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_44_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & codecs_toplevel_consts_44_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
codecs_toplevel_consts_44_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(items),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_make_encoding_map = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "make_encoding_map",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[65];
    }
codecs_toplevel_consts_44_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 64,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x1a\x00\x09\x0b\x80\x41\xd8\x0f\x1b\xd7\x0f\x21\xd1\x0f\x21\xd6\x0f\x23\x89\x03\x88\x01\x88\x21\xd8\x0f\x10\x90\x41\x89\x76\xd8\x13\x14\x88\x41\x88\x61\x8a\x44\xe0\x13\x17\x88\x41\x88\x61\x8a\x44\xf0\x09\x00\x10\x24\xf0\x0a\x00\x0c\x0d\x80\x48",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_decoding_map = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "decoding_map",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_44_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_decoding_map._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[109],
            (PyObject *)&_Py_SINGLETON(strings).ascii[107],
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(88)
codecs_toplevel_consts_44 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 44,
    },
    .co_consts = & codecs_toplevel_consts_44_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_44_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1086,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 363,
    .co_localsplusnames = & codecs_toplevel_consts_44_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_make_encoding_map._ascii.ob_base,
    .co_qualname = & const_str_make_encoding_map._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_44_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x69\x00\x7d\x01\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x14\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x03\x7c\x01\x76\x01\x72\x06\x7c\x02\x7c\x01\x7c\x03\x3c\x00\x00\x00\x8c\x10\x64\x01\x7c\x01\x7c\x03\x3c\x00\x00\x00\x8c\x16\x04\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_xmlcharrefreplace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "xmlcharrefreplace",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_backslashreplace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "backslashreplace",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_namereplace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "namereplace",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
codecs_toplevel_consts_50 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            Py_None,
            &_Py_ID(strict),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
codecs_toplevel_consts_51 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(strict),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[53];
        }_object;
    }
codecs_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 53,
        },
        .ob_item = {
            & codecs_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & codecs_toplevel_consts_3._object.ob_base.ob_base,
            & codecs_toplevel_consts_4._ascii.ob_base,
            & codecs_toplevel_consts_5._object.ob_base.ob_base,
            & codecs_toplevel_consts_6.ob_base.ob_base,
            & codecs_toplevel_consts_7.ob_base.ob_base,
            & codecs_toplevel_consts_8.ob_base.ob_base,
            & codecs_toplevel_consts_9.ob_base.ob_base,
            & codecs_toplevel_consts_10.ob_base.ob_base,
            &_Py_ID(little),
            & codecs_toplevel_consts_12.ob_base.ob_base,
            & const_str_CodecInfo._ascii.ob_base,
            & codecs_toplevel_consts_14.ob_base.ob_base,
            & const_str_Codec._ascii.ob_base,
            & codecs_toplevel_consts_16.ob_base.ob_base,
            & const_str_IncrementalEncoder._ascii.ob_base,
            & codecs_toplevel_consts_18.ob_base.ob_base,
            & const_str_BufferedIncrementalEncoder._ascii.ob_base,
            & codecs_toplevel_consts_20.ob_base.ob_base,
            & const_str_IncrementalDecoder._ascii.ob_base,
            & codecs_toplevel_consts_22.ob_base.ob_base,
            & const_str_BufferedIncrementalDecoder._ascii.ob_base,
            & codecs_toplevel_consts_24.ob_base.ob_base,
            & const_str_StreamWriter._ascii.ob_base,
            & codecs_toplevel_consts_26.ob_base.ob_base,
            & const_str_StreamReader._ascii.ob_base,
            & codecs_toplevel_consts_28.ob_base.ob_base,
            & const_str_StreamReaderWriter._ascii.ob_base,
            & codecs_toplevel_consts_30.ob_base.ob_base,
            & const_str_StreamRecoder._ascii.ob_base,
            &_Py_ID(strict),
            & codecs_toplevel_consts_33.ob_base.ob_base,
            & codecs_toplevel_consts_34.ob_base.ob_base,
            & codecs_toplevel_consts_35.ob_base.ob_base,
            & codecs_toplevel_consts_36.ob_base.ob_base,
            & codecs_toplevel_consts_37.ob_base.ob_base,
            & codecs_toplevel_consts_38.ob_base.ob_base,
            & codecs_toplevel_consts_39.ob_base.ob_base,
            & codecs_toplevel_consts_40.ob_base.ob_base,
            & codecs_toplevel_consts_41.ob_base.ob_base,
            & codecs_toplevel_consts_42.ob_base.ob_base,
            & codecs_toplevel_consts_43.ob_base.ob_base,
            & codecs_toplevel_consts_44.ob_base.ob_base,
            &_Py_ID(ignore),
            &_Py_ID(replace),
            & const_str_xmlcharrefreplace._ascii.ob_base,
            & const_str_backslashreplace._ascii.ob_base,
            & const_str_namereplace._ascii.ob_base,
            & codecs_toplevel_consts_50._object.ob_base.ob_base,
            & codecs_toplevel_consts_51._object.ob_base.ob_base,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__codecs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_codecs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_why = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "why",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_SystemError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SystemError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__false = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_false",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_encodings = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "encodings",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[57];
        }_object;
    }
codecs_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 57,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            &_Py_ID(builtins),
            & const_str_sys._ascii.ob_base,
            & const_str__codecs._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_why._ascii.ob_base,
            & const_str_SystemError._ascii.ob_base,
            &_Py_ID(__all__),
            & const_str_BOM_UTF8._ascii.ob_base,
            & const_str_BOM_LE._ascii.ob_base,
            & const_str_BOM_UTF16_LE._ascii.ob_base,
            & const_str_BOM_BE._ascii.ob_base,
            & const_str_BOM_UTF16_BE._ascii.ob_base,
            & const_str_BOM_UTF32_LE._ascii.ob_base,
            & const_str_BOM_UTF32_BE._ascii.ob_base,
            &_Py_ID(byteorder),
            & const_str_BOM._ascii.ob_base,
            & const_str_BOM_UTF16._ascii.ob_base,
            & const_str_BOM_UTF32._ascii.ob_base,
            & const_str_BOM32_LE._ascii.ob_base,
            & const_str_BOM32_BE._ascii.ob_base,
            & const_str_BOM64_LE._ascii.ob_base,
            & const_str_BOM64_BE._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_CodecInfo._ascii.ob_base,
            & const_str_Codec._ascii.ob_base,
            &_Py_ID(object),
            & const_str_IncrementalEncoder._ascii.ob_base,
            & const_str_BufferedIncrementalEncoder._ascii.ob_base,
            & const_str_IncrementalDecoder._ascii.ob_base,
            & const_str_BufferedIncrementalDecoder._ascii.ob_base,
            & const_str_StreamWriter._ascii.ob_base,
            & const_str_StreamReader._ascii.ob_base,
            & const_str_StreamReaderWriter._ascii.ob_base,
            & const_str_StreamRecoder._ascii.ob_base,
            &_Py_ID(open),
            & const_str_EncodedFile._ascii.ob_base,
            & const_str_getencoder._ascii.ob_base,
            & const_str_getdecoder._ascii.ob_base,
            & const_str_getincrementalencoder._ascii.ob_base,
            & const_str_getincrementaldecoder._ascii.ob_base,
            & const_str_getreader._ascii.ob_base,
            & const_str_getwriter._ascii.ob_base,
            & const_str_iterencode._ascii.ob_base,
            & const_str_iterdecode._ascii.ob_base,
            & const_str_make_identity_dict._ascii.ob_base,
            & const_str_make_encoding_map._ascii.ob_base,
            & const_str_lookup_error._ascii.ob_base,
            & const_str_strict_errors._ascii.ob_base,
            & const_str_ignore_errors._ascii.ob_base,
            & const_str_replace_errors._ascii.ob_base,
            & const_str_xmlcharrefreplace_errors._ascii.ob_base,
            & const_str_backslashreplace_errors._ascii.ob_base,
            & const_str_namereplace_errors._ascii.ob_base,
            & const_str_LookupError._ascii.ob_base,
            & const_str__false._ascii.ob_base,
            & const_str_encodings._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[533];
    }
codecs_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 532,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x07\x01\x04\xf3\x12\x00\x01\x10\xdb\x00\x0a\xf0\x08\x03\x01\x45\x01\xdc\x04\x19\xf2\x08\x0d\x0b\x2d\x80\x07\xf0\x30\x00\x0c\x1b\x80\x08\xf0\x06\x00\x19\x24\xd0\x00\x23\x80\x06\x88\x1c\xf0\x06\x00\x19\x24\xd0\x00\x23\x80\x06\x88\x1c\xf0\x06\x00\x10\x23\x80\x0c\xf0\x06\x00\x10\x23\x80\x0c\xe0\x03\x06\x87\x3d\x81\x3d\x90\x48\xd2\x03\x1c\xf0\x06\x00\x17\x23\xd0\x04\x22\x80\x43\x88\x29\xf0\x06\x00\x11\x1d\x81\x49\xf0\x0a\x00\x17\x23\xd0\x04\x22\x80\x43\x88\x29\xf0\x06\x00\x11\x1d\x80\x49\xf0\x06\x00\x0c\x18\x80\x08\xd8\x0b\x17\x80\x08\xd8\x0b\x17\x80\x08\xd8\x0b\x17\x80\x08\xf4\x0a\x1d\x01\x26\x90\x05\xf4\x00\x1d\x01\x26\xf7\x3e\x40\x01\x01\x22\xf1\x00\x40\x01\x01\x22\xf4\x44\x02\x26\x01\x0c\x98\x16\xf4\x00\x26\x01\x0c\xf4\x50\x01\x20\x01\x22\xd0\x21\x33\xf4\x00\x20\x01\x22\xf4\x44\x01\x2f\x01\x0c\x98\x16\xf4\x00\x2f\x01\x0c\xf4\x62\x01\x22\x01\x1f\xd0\x21\x33\xf4\x00\x22\x01\x1f\xf4\x56\x01\x48\x01\x01\x48\x01\x90\x35\xf4\x00\x48\x01\x01\x48\x01\xf4\x58\x02\x78\x03\x01\x48\x01\x90\x35\xf4\x00\x78\x03\x01\x48\x01\xf7\x78\x07\x56\x01\x01\x48\x01\xf1\x00\x56\x01\x01\x48\x01\xf7\x74\x02\x73\x01\x01\x48\x01\xf1\x00\x73\x01\x01\x48\x01\xf3\x6e\x03\x2f\x01\x0e\xf3\x62\x01\x22\x01\x0e\xf2\x4c\x01\x08\x01\x23\xf2\x14\x08\x01\x23\xf2\x14\x0c\x01\x13\xf2\x1c\x0c\x01\x13\xf2\x1c\x08\x01\x29\xf2\x14\x08\x01\x29\xf3\x14\x10\x01\x15\xf3\x24\x10\x01\x15\xf2\x28\x08\x01\x1e\xf2\x14\x13\x01\x0d\xf0\x2e\x0e\x01\x1e\xd9\x14\x20\xa0\x18\xd3\x14\x2a\x80\x4d\xd9\x14\x20\xa0\x18\xd3\x14\x2a\x80\x4d\xd9\x15\x21\xa0\x29\xd3\x15\x2c\x80\x4e\xd9\x1f\x2b\xd0\x2c\x3f\xd3\x1f\x40\xd0\x04\x1c\xd9\x1e\x2a\xd0\x2b\x3d\xd3\x1e\x3e\xd0\x04\x1b\xd9\x19\x25\xa0\x6d\xd3\x19\x34\xd0\x04\x16\xf0\x18\x00\x0a\x0b\x80\x06\xd9\x03\x09\xdc\x04\x14\xf0\x03\x00\x04\x0a\xf8\xf0\x6f\x22\x00\x08\x13\xf2\x00\x01\x01\x45\x01\xd9\x0a\x15\xd0\x16\x3d\xc0\x03\xd1\x16\x43\xd3\x0a\x44\xd0\x04\x44\xfb\xf0\x03\x01\x01\x45\x01\xfb\xf0\x56\x22\x00\x08\x13\xf2\x00\x07\x01\x1e\xe0\x14\x18\x80\x4d\xd8\x14\x18\x80\x4d\xd8\x15\x19\x80\x4e\xd8\x1f\x23\xd0\x04\x1c\xd8\x1e\x22\xd0\x04\x1b\xd8\x19\x1d\xd2\x04\x16\xf0\x0f\x07\x01\x1e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[42];
    }
codecs_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 41,
    },
    .ob_shash = -1,
    .ob_sval = "\x8c\x05\x44\x15\x00\xc3\x1b\x30\x44\x2d\x00\xc4\x15\x05\x44\x2a\x03\xc4\x1a\x0b\x44\x25\x03\xc4\x25\x05\x44\x2a\x03\xc4\x2d\x11\x45\x01\x03\xc5\x00\x01\x45\x01\x03",
};
static
    struct _PyCode_DEF(648)
codecs_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 324,
    },
    .co_consts = & codecs_toplevel_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & codecs_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 364,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & codecs_toplevel_consts_12_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & codecs_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x5a\x01\x64\x01\x64\x02\x6c\x02\x5a\x02\x09\x00\x64\x01\x64\x03\x6c\x03\xad\x02\x01\x00\x67\x00\x64\x05\xa2\x01\x5a\x07\x64\x06\x5a\x08\x64\x07\x78\x01\x5a\x09\x5a\x0a\x64\x08\x78\x01\x5a\x0b\x5a\x0c\x64\x09\x5a\x0d\x64\x0a\x5a\x0e\x65\x02\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\x6b\x28\x00\x00\x72\x07\x65\x0a\x78\x01\x5a\x10\x5a\x11\x65\x0d\x5a\x12\x6e\x06\x65\x0c\x78\x01\x5a\x10\x5a\x11\x65\x0e\x5a\x12\x65\x0a\x5a\x13\x65\x0c\x5a\x14\x65\x0d\x5a\x15\x65\x0e\x5a\x16\x02\x00\x47\x00\x64\x0c\x84\x00\x64\x0d\x65\x17\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x18\x02\x00\x47\x00\x64\x0e\x84\x00\x64\x0f\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x19\x02\x00\x47\x00\x64\x10\x84\x00\x64\x11\x65\x1a\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x1b\x02\x00\x47\x00\x64\x12\x84\x00\x64\x13\x65\x1b\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x1c\x02\x00\x47\x00\x64\x14\x84\x00\x64\x15\x65\x1a\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x1d\x02\x00\x47\x00\x64\x16\x84\x00\x64\x17\x65\x1d\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x1e\x02\x00\x47\x00\x64\x18\x84\x00\x64\x19\x65\x19\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x1f\x02\x00\x47\x00\x64\x1a\x84\x00\x64\x1b\x65\x19\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x20\x02\x00\x47\x00\x64\x1c\x84\x00\x64\x1d\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x21\x02\x00\x47\x00\x64\x1e\x84\x00\x64\x1f\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x22\x64\x32\x64\x21\x84\x01\x5a\x23\x64\x33\x64\x22\x84\x01\x5a\x24\x64\x23\x84\x00\x5a\x25\x64\x24\x84\x00\x5a\x26\x64\x25\x84\x00\x5a\x27\x64\x26\x84\x00\x5a\x28\x64\x27\x84\x00\x5a\x29\x64\x28\x84\x00\x5a\x2a\x64\x34\x64\x29\x84\x01\x5a\x2b\x64\x34\x64\x2a\x84\x01\x5a\x2c\x64\x2b\x84\x00\x5a\x2d\x64\x2c\x84\x00\x5a\x2e\x09\x00\x02\x00\x65\x2f\x64\x20\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x30\x02\x00\x65\x2f\x64\x2d\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x31\x02\x00\x65\x2f\x64\x2e\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x32\x02\x00\x65\x2f\x64\x2f\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x33\x02\x00\x65\x2f\x64\x30\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x34\x02\x00\x65\x2f\x64\x31\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x35\x64\x01\x5a\x37\x65\x37\x72\x05\x64\x01\x64\x02\x6c\x38\x5a\x38\x79\x02\x79\x02\x23\x00\x65\x04\x24\x00\x72\x10\x5a\x05\x02\x00\x65\x06\x64\x04\x65\x05\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x64\x02\x5a\x05\x5b\x05\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x36\x24\x00\x72\x0f\x01\x00\x64\x02\x5a\x30\x64\x02\x5a\x31\x64\x02\x5a\x32\x64\x02\x5a\x33\x64\x02\x5a\x34\x64\x02\x5a\x35\x59\x00\x8c\x35\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_codecs_toplevel(void)
{
    return Py_NewRef((PyObject *) &codecs_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1474];
    }
io_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1473,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x68\x65\x20\x69\x6f\x20\x6d\x6f\x64\x75\x6c\x65\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x74\x68\x65\x20\x50\x79\x74\x68\x6f\x6e\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x73\x20\x74\x6f\x20\x73\x74\x72\x65\x61\x6d\x20\x68\x61\x6e\x64\x6c\x69\x6e\x67\x2e\x20\x54\x68\x65\x0a\x62\x75\x69\x6c\x74\x69\x6e\x20\x6f\x70\x65\x6e\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x69\x73\x20\x64\x65\x66\x69\x6e\x65\x64\x20\x69\x6e\x20\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x41\x74\x20\x74\x68\x65\x20\x74\x6f\x70\x20\x6f\x66\x20\x74\x68\x65\x20\x49\x2f\x4f\x20\x68\x69\x65\x72\x61\x72\x63\x68\x79\x20\x69\x73\x20\x74\x68\x65\x20\x61\x62\x73\x74\x72\x61\x63\x74\x20\x62\x61\x73\x65\x20\x63\x6c\x61\x73\x73\x20\x49\x4f\x42\x61\x73\x65\x2e\x20\x49\x74\x0a\x64\x65\x66\x69\x6e\x65\x73\x20\x74\x68\x65\x20\x62\x61\x73\x69\x63\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x74\x6f\x20\x61\x20\x73\x74\x72\x65\x61\x6d\x2e\x20\x4e\x6f\x74\x65\x2c\x20\x68\x6f\x77\x65\x76\x65\x72\x2c\x20\x74\x68\x61\x74\x20\x74\x68\x65\x72\x65\x20\x69\x73\x20\x6e\x6f\x0a\x73\x65\x70\x61\x72\x61\x74\x69\x6f\x6e\x20\x62\x65\x74\x77\x65\x65\x6e\x20\x72\x65\x61\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x77\x72\x69\x74\x69\x6e\x67\x20\x74\x6f\x20\x73\x74\x72\x65\x61\x6d\x73\x3b\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x73\x20\x61\x72\x65\x0a\x61\x6c\x6c\x6f\x77\x65\x64\x20\x74\x6f\x20\x72\x61\x69\x73\x65\x20\x61\x6e\x20\x4f\x53\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x74\x68\x65\x79\x20\x64\x6f\x20\x6e\x6f\x74\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x61\x20\x67\x69\x76\x65\x6e\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x45\x78\x74\x65\x6e\x64\x69\x6e\x67\x20\x49\x4f\x42\x61\x73\x65\x20\x69\x73\x20\x52\x61\x77\x49\x4f\x42\x61\x73\x65\x20\x77\x68\x69\x63\x68\x20\x64\x65\x61\x6c\x73\x20\x73\x69\x6d\x70\x6c\x79\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x72\x65\x61\x64\x69\x6e\x67\x20\x61\x6e\x64\x0a\x77\x72\x69\x74\x69\x6e\x67\x20\x6f\x66\x20\x72\x61\x77\x20\x62\x79\x74\x65\x73\x20\x74\x6f\x20\x61\x20\x73\x74\x72\x65\x61\x6d\x2e\x20\x46\x69\x6c\x65\x49\x4f\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x65\x73\x20\x52\x61\x77\x49\x4f\x42\x61\x73\x65\x20\x74\x6f\x20\x70\x72\x6f\x76\x69\x64\x65\x0a\x61\x6e\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x74\x6f\x20\x4f\x53\x20\x66\x69\x6c\x65\x73\x2e\x0a\x0a\x42\x75\x66\x66\x65\x72\x65\x64\x49\x4f\x42\x61\x73\x65\x20\x64\x65\x61\x6c\x73\x20\x77\x69\x74\x68\x20\x62\x75\x66\x66\x65\x72\x69\x6e\x67\x20\x6f\x6e\x20\x61\x20\x72\x61\x77\x20\x62\x79\x74\x65\x20\x73\x74\x72\x65\x61\x6d\x20\x28\x52\x61\x77\x49\x4f\x42\x61\x73\x65\x29\x2e\x20\x49\x74\x73\x0a\x73\x75\x62\x63\x6c\x61\x73\x73\x65\x73\x2c\x20\x42\x75\x66\x66\x65\x72\x65\x64\x57\x72\x69\x74\x65\x72\x2c\x20\x42\x75\x66\x66\x65\x72\x65\x64\x52\x65\x61\x64\x65\x72\x2c\x20\x61\x6e\x64\x20\x42\x75\x66\x66\x65\x72\x65\x64\x52\x57\x50\x61\x69\x72\x20\x62\x75\x66\x66\x65\x72\x0a\x73\x74\x72\x65\x61\x6d\x73\x20\x74\x68\x61\x74\x20\x61\x72\x65\x20\x72\x65\x61\x64\x61\x62\x6c\x65\x2c\x20\x77\x72\x69\x74\x61\x62\x6c\x65\x2c\x20\x61\x6e\x64\x20\x62\x6f\x74\x68\x20\x72\x65\x73\x70\x65\x63\x74\x69\x76\x65\x6c\x79\x2e\x0a\x42\x75\x66\x66\x65\x72\x65\x64\x52\x61\x6e\x64\x6f\x6d\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x61\x20\x62\x75\x66\x66\x65\x72\x65\x64\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x74\x6f\x20\x72\x61\x6e\x64\x6f\x6d\x20\x61\x63\x63\x65\x73\x73\x0a\x73\x74\x72\x65\x61\x6d\x73\x2e\x20\x42\x79\x74\x65\x73\x49\x4f\x20\x69\x73\x20\x61\x20\x73\x69\x6d\x70\x6c\x65\x20\x73\x74\x72\x65\x61\x6d\x20\x6f\x66\x20\x69\x6e\x2d\x6d\x65\x6d\x6f\x72\x79\x20\x62\x79\x74\x65\x73\x2e\x0a\x0a\x41\x6e\x6f\x74\x68\x65\x72\x20\x49\x4f\x42\x61\x73\x65\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x2c\x20\x54\x65\x78\x74\x49\x4f\x42\x61\x73\x65\x2c\x20\x64\x65\x61\x6c\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x61\x6e\x64\x20\x64\x65\x63\x6f\x64\x69\x6e\x67\x0a\x6f\x66\x20\x73\x74\x72\x65\x61\x6d\x73\x20\x69\x6e\x74\x6f\x20\x74\x65\x78\x74\x2e\x20\x54\x65\x78\x74\x49\x4f\x57\x72\x61\x70\x70\x65\x72\x2c\x20\x77\x68\x69\x63\x68\x20\x65\x78\x74\x65\x6e\x64\x73\x20\x69\x74\x2c\x20\x69\x73\x20\x61\x20\x62\x75\x66\x66\x65\x72\x65\x64\x20\x74\x65\x78\x74\x0a\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x74\x6f\x20\x61\x20\x62\x75\x66\x66\x65\x72\x65\x64\x20\x72\x61\x77\x20\x73\x74\x72\x65\x61\x6d\x20\x28\x60\x42\x75\x66\x66\x65\x72\x65\x64\x49\x4f\x42\x61\x73\x65\x60\x29\x2e\x20\x46\x69\x6e\x61\x6c\x6c\x79\x2c\x20\x53\x74\x72\x69\x6e\x67\x49\x4f\x0a\x69\x73\x20\x61\x6e\x20\x69\x6e\x2d\x6d\x65\x6d\x6f\x72\x79\x20\x73\x74\x72\x65\x61\x6d\x20\x66\x6f\x72\x20\x74\x65\x78\x74\x2e\x0a\x0a\x41\x72\x67\x75\x6d\x65\x6e\x74\x20\x6e\x61\x6d\x65\x73\x20\x61\x72\x65\x20\x6e\x6f\x74\x20\x70\x61\x72\x74\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x63\x61\x74\x69\x6f\x6e\x2c\x20\x61\x6e\x64\x20\x6f\x6e\x6c\x79\x20\x74\x68\x65\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x0a\x6f\x66\x20\x6f\x70\x65\x6e\x28\x29\x20\x61\x72\x65\x20\x69\x6e\x74\x65\x6e\x64\x65\x64\x20\x74\x6f\x20\x62\x65\x20\x75\x73\x65\x64\x20\x61\x73\x20\x6b\x65\x79\x77\x6f\x72\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x2e\x0a\x0a\x64\x61\x74\x61\x3a\x0a\x0a\x44\x45\x46\x41\x55\x4c\x54\x5f\x42\x55\x46\x46\x45\x52\x5f\x53\x49\x5a\x45\x0a\x0a\x20\x20\x20\x41\x6e\x20\x69\x6e\x74\x20\x63\x6f\x6e\x74\x61\x69\x6e\x69\x6e\x67\x20\x74\x68\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x62\x75\x66\x66\x65\x72\x20\x73\x69\x7a\x65\x20\x75\x73\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x27\x73\x20\x62\x75\x66\x66\x65\x72\x65\x64\x0a\x20\x20\x20\x49\x2f\x4f\x20\x63\x6c\x61\x73\x73\x65\x73\x2e\x20\x6f\x70\x65\x6e\x28\x29\x20\x75\x73\x65\x73\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x27\x73\x20\x62\x6c\x6b\x73\x69\x7a\x65\x20\x28\x61\x73\x20\x6f\x62\x74\x61\x69\x6e\x65\x64\x20\x62\x79\x20\x6f\x73\x2e\x73\x74\x61\x74\x29\x20\x69\x66\x0a\x20\x20\x20\x70\x6f\x73\x73\x69\x62\x6c\x65\x2e\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[236];
    }
io_toplevel_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 235,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Guido van Rossum <guido@python.org>, Mike Verdone <mike.verdone@gmail.com>, Mark Russell <mark.russell@zen.co.uk>, Antoine Pitrou <solipsis@pitrou.net>, Amaury Forgeot d'Arc <amauryfa@gmail.com>, Benjamin Peterson <benjamin@python.org>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_BlockingIOError = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BlockingIOError",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_IOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IOBase",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_RawIOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "RawIOBase",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_StringIO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StringIO",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_BufferedIOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedIOBase",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_BufferedReader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedReader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_BufferedWriter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedWriter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_BufferedRWPair = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedRWPair",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_BufferedRandom = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BufferedRandom",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_TextIOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "TextIOBase",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_UnsupportedOperation = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UnsupportedOperation",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_SEEK_SET = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SEEK_SET",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_SEEK_CUR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SEEK_CUR",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_SEEK_END = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SEEK_END",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_DEFAULT_BUFFER_SIZE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "DEFAULT_BUFFER_SIZE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_text_encoding = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "text_encoding",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
io_toplevel_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            & const_str_BlockingIOError._ascii.ob_base,
            &_Py_ID(open),
            & const_str_open_code._ascii.ob_base,
            & const_str_IOBase._ascii.ob_base,
            & const_str_RawIOBase._ascii.ob_base,
            & const_str_FileIO._ascii.ob_base,
            & const_str_BytesIO._ascii.ob_base,
            & const_str_StringIO._ascii.ob_base,
            & const_str_BufferedIOBase._ascii.ob_base,
            & const_str_BufferedReader._ascii.ob_base,
            & const_str_BufferedWriter._ascii.ob_base,
            & const_str_BufferedRWPair._ascii.ob_base,
            & const_str_BufferedRandom._ascii.ob_base,
            & const_str_TextIOBase._ascii.ob_base,
            &_Py_ID(TextIOWrapper),
            & const_str_UnsupportedOperation._ascii.ob_base,
            & const_str_SEEK_SET._ascii.ob_base,
            & const_str_SEEK_CUR._ascii.ob_base,
            & const_str_SEEK_END._ascii.ob_base,
            & const_str_DEFAULT_BUFFER_SIZE._ascii.ob_base,
            & const_str_text_encoding._ascii.ob_base,
            & const_str_IncrementalNewlineDecoder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
io_toplevel_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & const_str_DEFAULT_BUFFER_SIZE._ascii.ob_base,
            & const_str_BlockingIOError._ascii.ob_base,
            & const_str_UnsupportedOperation._ascii.ob_base,
            &_Py_ID(open),
            & const_str_open_code._ascii.ob_base,
            & const_str_FileIO._ascii.ob_base,
            & const_str_BytesIO._ascii.ob_base,
            & const_str_StringIO._ascii.ob_base,
            & const_str_BufferedReader._ascii.ob_base,
            & const_str_BufferedWriter._ascii.ob_base,
            & const_str_BufferedRWPair._ascii.ob_base,
            & const_str_BufferedRandom._ascii.ob_base,
            & const_str_IncrementalNewlineDecoder._ascii.ob_base,
            & const_str_text_encoding._ascii.ob_base,
            &_Py_ID(TextIOWrapper),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_io = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "io",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
io_toplevel_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_IOBase._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__IOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_IOBase",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
io_toplevel_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(_io),
            & const_str__IOBase._ascii.ob_base,
            &_Py_ID(__doc__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
io_toplevel_consts_9_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen io>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
io_toplevel_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd8\x0e\x11\x8f\x6b\x89\x6b\xd7\x0e\x21\xd1\x0e\x21\x81\x47",
};
static
    struct _PyCode_DEF(56)
io_toplevel_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & io_toplevel_consts_9_consts._object.ob_base.ob_base,
    .co_names = & io_toplevel_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 72,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 365,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & io_toplevel_consts_9_filename._ascii.ob_base,
    .co_name = & const_str_IOBase._ascii.ob_base,
    .co_qualname = & const_str_IOBase._ascii.ob_base,
    .co_linetable = & io_toplevel_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x65\x03\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
io_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_RawIOBase._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__RawIOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_RawIOBase",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
io_toplevel_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(_io),
            & const_str__RawIOBase._ascii.ob_base,
            &_Py_ID(__doc__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
io_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd8\x0e\x11\x8f\x6e\x89\x6e\xd7\x0e\x24\xd1\x0e\x24\x81\x47",
};
static
    struct _PyCode_DEF(56)
io_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & io_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & io_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 75,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 366,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & io_toplevel_consts_9_filename._ascii.ob_base,
    .co_name = & const_str_RawIOBase._ascii.ob_base,
    .co_qualname = & const_str_RawIOBase._ascii.ob_base,
    .co_linetable = & io_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x65\x03\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
io_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_BufferedIOBase._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__BufferedIOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_BufferedIOBase",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
io_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(_io),
            & const_str__BufferedIOBase._ascii.ob_base,
            &_Py_ID(__doc__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
io_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd8\x0e\x11\xd7\x0e\x21\xd1\x0e\x21\xd7\x0e\x29\xd1\x0e\x29\x81\x47",
};
static
    struct _PyCode_DEF(56)
io_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & io_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & io_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 78,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 367,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & io_toplevel_consts_9_filename._ascii.ob_base,
    .co_name = & const_str_BufferedIOBase._ascii.ob_base,
    .co_qualname = & const_str_BufferedIOBase._ascii.ob_base,
    .co_linetable = & io_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x65\x03\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
io_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_TextIOBase._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__TextIOBase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_TextIOBase",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
io_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(_io),
            & const_str__TextIOBase._ascii.ob_base,
            &_Py_ID(__doc__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
io_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd8\x0e\x11\x8f\x6f\x89\x6f\xd7\x0e\x25\xd1\x0e\x25\x81\x47",
};
static
    struct _PyCode_DEF(56)
io_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & io_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & io_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 81,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 368,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & io_toplevel_consts_9_filename._ascii.ob_base,
    .co_name = & const_str_TextIOBase._ascii.ob_base,
    .co_qualname = & const_str_TextIOBase._ascii.ob_base,
    .co_linetable = & io_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x65\x03\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
io_toplevel_consts_18 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(_WindowsConsoleIO),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
io_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            & io_toplevel_consts_0._ascii.ob_base,
            & io_toplevel_consts_1._ascii.ob_base,
            & io_toplevel_consts_2._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & io_toplevel_consts_5._object.ob_base.ob_base,
            & const_str_io._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & io_toplevel_consts_9.ob_base.ob_base,
            & const_str_IOBase._ascii.ob_base,
            & abc_toplevel_consts_17._object.ob_base.ob_base,
            & io_toplevel_consts_12.ob_base.ob_base,
            & const_str_RawIOBase._ascii.ob_base,
            & io_toplevel_consts_14.ob_base.ob_base,
            & const_str_BufferedIOBase._ascii.ob_base,
            & io_toplevel_consts_16.ob_base.ob_base,
            & const_str_TextIOBase._ascii.ob_base,
            & io_toplevel_consts_18._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str___author__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__author__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_klass = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "klass",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[37];
        }_object;
    }
io_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 37,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str___author__._ascii.ob_base,
            &_Py_ID(__all__),
            &_Py_ID(_io),
            & const_str_abc._ascii.ob_base,
            & const_str_DEFAULT_BUFFER_SIZE._ascii.ob_base,
            & const_str_BlockingIOError._ascii.ob_base,
            & const_str_UnsupportedOperation._ascii.ob_base,
            &_Py_ID(open),
            & const_str_open_code._ascii.ob_base,
            & const_str_FileIO._ascii.ob_base,
            & const_str_BytesIO._ascii.ob_base,
            & const_str_StringIO._ascii.ob_base,
            & const_str_BufferedReader._ascii.ob_base,
            & const_str_BufferedWriter._ascii.ob_base,
            & const_str_BufferedRWPair._ascii.ob_base,
            & const_str_BufferedRandom._ascii.ob_base,
            & const_str_IncrementalNewlineDecoder._ascii.ob_base,
            & const_str_text_encoding._ascii.ob_base,
            &_Py_ID(TextIOWrapper),
            &_Py_ID(__module__),
            & const_str_SEEK_SET._ascii.ob_base,
            & const_str_SEEK_CUR._ascii.ob_base,
            & const_str_SEEK_END._ascii.ob_base,
            & const_str__IOBase._ascii.ob_base,
            & const_str_ABCMeta._ascii.ob_base,
            & const_str_IOBase._ascii.ob_base,
            & const_str__RawIOBase._ascii.ob_base,
            & const_str_RawIOBase._ascii.ob_base,
            & const_str__BufferedIOBase._ascii.ob_base,
            & const_str_BufferedIOBase._ascii.ob_base,
            & const_str__TextIOBase._ascii.ob_base,
            & const_str_TextIOBase._ascii.ob_base,
            & const_str_register._ascii.ob_base,
            & const_str_klass._ascii.ob_base,
            &_Py_ID(_WindowsConsoleIO),
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[304];
    }
io_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 303,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x21\x01\x04\xf0\x48\x01\x05\x0f\x38\x80\x0a\xf2\x0e\x05\x0b\x50\x01\x80\x07\xf3\x10\x00\x01\x0b\xdb\x00\x0a\xf7\x04\x03\x01\x4a\x01\xf7\x00\x03\x01\x4a\x01\xf7\x00\x03\x01\x4a\x01\xf7\x00\x03\x01\x4a\x01\xf1\x00\x03\x01\x4a\x01\xf0\x0e\x00\x23\x27\xd0\x00\x14\xd4\x00\x1f\xf0\x06\x00\x0c\x0d\x80\x08\xd8\x0b\x0c\x80\x08\xd8\x0b\x0c\x80\x08\xf4\x0a\x01\x01\x22\x88\x53\x8f\x5b\x89\x5b\xa0\x43\xa7\x4b\xa1\x4b\xf5\x00\x01\x01\x22\xf4\x06\x01\x01\x25\x90\x03\x97\x0e\x91\x0e\xa0\x06\xf4\x00\x01\x01\x25\xf4\x06\x01\x01\x2a\x90\x53\xd7\x15\x28\xd1\x15\x28\xa8\x26\xf4\x00\x01\x01\x2a\xf4\x06\x01\x01\x26\x90\x13\x97\x1f\x91\x1f\xa0\x26\xf4\x00\x01\x01\x26\xf0\x06\x00\x01\x0a\xd7\x00\x12\xd1\x00\x12\x90\x36\xd4\x00\x1a\xe0\x0e\x15\x90\x7e\xa0\x7e\xb0\x7e\xd8\x0e\x1c\xf3\x03\x01\x0e\x1e\x80\x45\xe0\x04\x12\xd7\x04\x1b\xd1\x04\x1b\x98\x45\xd5\x04\x22\xf0\x05\x01\x0e\x1e\xf0\x08\x00\x0f\x17\x98\x0d\xd3\x0d\x26\x80\x45\xd8\x04\x0e\xd7\x04\x17\xd1\x04\x17\x98\x05\xd5\x04\x1e\xf0\x03\x00\x0e\x27\xe0\x04\x09\xf0\x04\x05\x01\x2a\xdd\x04\x25\xf0\x08\x00\x05\x0e\xd7\x04\x16\xd1\x04\x16\xd0\x17\x28\xd5\x04\x29\xf8\xf0\x07\x00\x08\x13\xf2\x00\x01\x01\x09\xd9\x04\x08\xf0\x03\x01\x01\x09\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
io_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc3\x2d\x06\x44\x05\x00\xc4\x05\x05\x44\x0d\x03\xc4\x0c\x01\x44\x0d\x03",
};
static
    struct _PyCode_DEF(544)
io_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 272,
    },
    .co_consts = & io_toplevel_consts._object.ob_base.ob_base,
    .co_names = & io_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & io_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 369,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & io_toplevel_consts_9_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & io_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x5a\x01\x67\x00\x64\x02\xa2\x01\x5a\x02\x64\x03\x64\x04\x6c\x03\x5a\x03\x64\x03\x64\x04\x6c\x04\x5a\x04\x64\x03\x64\x05\x6c\x03\x6d\x05\x5a\x05\x6d\x06\x5a\x06\x6d\x07\x5a\x07\x6d\x08\x5a\x08\x6d\x09\x5a\x09\x6d\x0a\x5a\x0a\x6d\x0b\x5a\x0b\x6d\x0c\x5a\x0c\x6d\x0d\x5a\x0d\x6d\x0e\x5a\x0e\x6d\x0f\x5a\x0f\x6d\x10\x5a\x10\x6d\x11\x5a\x11\x6d\x12\x5a\x12\x6d\x13\x5a\x13\x01\x00\x64\x06\x65\x07\x5f\x14\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x5a\x15\x64\x07\x5a\x16\x64\x08\x5a\x17\x02\x00\x47\x00\x64\x09\x84\x00\x64\x0a\x65\x03\x6a\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x04\x6a\x32\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x0b\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x1a\x02\x00\x47\x00\x64\x0c\x84\x00\x64\x0d\x65\x03\x6a\x36\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1a\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x1c\x02\x00\x47\x00\x64\x0e\x84\x00\x64\x0f\x65\x03\x6a\x3a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1a\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x1e\x02\x00\x47\x00\x64\x10\x84\x00\x64\x11\x65\x03\x6a\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1a\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x20\x65\x1c\x6a\x43\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x0b\x65\x0d\x65\x0e\x65\x10\x65\x0f\x66\x05\x44\x00\x5d\x13\x00\x00\x5a\x22\x65\x1e\x6a\x43\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x22\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x15\x04\x00\x65\x0c\x65\x13\x66\x02\x44\x00\x5d\x13\x00\x00\x5a\x22\x65\x20\x6a\x43\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x22\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x15\x04\x00\x5b\x22\x09\x00\x64\x03\x64\x12\x6c\x03\x6d\x23\x5a\x23\x01\x00\x65\x1c\x6a\x43\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x23\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x04\x23\x00\x65\x24\x24\x00\x72\x03\x01\x00\x59\x00\x79\x04\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_io_toplevel(void)
{
    return Py_NewRef((PyObject *) &io_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[107];
    }
_collections_abc_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 106,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x62\x73\x74\x72\x61\x63\x74\x20\x42\x61\x73\x65\x20\x43\x6c\x61\x73\x73\x65\x73\x20\x28\x41\x42\x43\x73\x29\x20\x66\x6f\x72\x20\x63\x6f\x6c\x6c\x65\x63\x74\x69\x6f\x6e\x73\x2c\x20\x61\x63\x63\x6f\x72\x64\x69\x6e\x67\x20\x74\x6f\x20\x50\x45\x50\x20\x33\x31\x31\x39\x2e\x0a\x0a\x55\x6e\x69\x74\x20\x74\x65\x73\x74\x73\x20\x61\x72\x65\x20\x69\x6e\x20\x74\x65\x73\x74\x5f\x63\x6f\x6c\x6c\x65\x63\x74\x69\x6f\x6e\x73\x2e\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_ABCMeta._ascii.ob_base,
            & const_str_abstractmethod._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
_collections_abc_toplevel_consts_5_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen _collections_abc>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str__f = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_f",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\x88\x24",
};
static
    struct _PyCode_DEF(4)
_collections_abc_toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 0 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 40,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 370,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__f._ascii.ob_base,
    .co_qualname = & const_str__f._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_Awaitable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Awaitable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_Coroutine = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Coroutine",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_AsyncIterable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncIterable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_AsyncIterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncIterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_AsyncGenerator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncGenerator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_Hashable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Hashable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_Iterable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Iterable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_Iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_Generator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Generator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_Reversible = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Reversible",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_Sized = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sized",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_Container = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Container",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_Callable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Callable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_Collection = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Collection",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_Set = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_MutableSet = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_Mapping = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_MutableMapping = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_MappingView = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MappingView",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_KeysView = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "KeysView",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_ItemsView = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ItemsView",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_ValuesView = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ValuesView",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_Sequence = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_MutableSequence = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_ByteString = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ByteString",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_Buffer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Buffer",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[26];
        }_object;
    }
_collections_abc_toplevel_consts_6 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 26,
        },
        .ob_item = {
            & const_str_Awaitable._ascii.ob_base,
            & const_str_Coroutine._ascii.ob_base,
            & const_str_AsyncIterable._ascii.ob_base,
            & const_str_AsyncIterator._ascii.ob_base,
            & const_str_AsyncGenerator._ascii.ob_base,
            & const_str_Hashable._ascii.ob_base,
            & const_str_Iterable._ascii.ob_base,
            & const_str_Iterator._ascii.ob_base,
            & const_str_Generator._ascii.ob_base,
            & const_str_Reversible._ascii.ob_base,
            & const_str_Sized._ascii.ob_base,
            & const_str_Container._ascii.ob_base,
            & const_str_Callable._ascii.ob_base,
            & const_str_Collection._ascii.ob_base,
            & const_str_Set._ascii.ob_base,
            & const_str_MutableSet._ascii.ob_base,
            & const_str_Mapping._ascii.ob_base,
            & const_str_MutableMapping._ascii.ob_base,
            & const_str_MappingView._ascii.ob_base,
            & const_str_KeysView._ascii.ob_base,
            & const_str_ItemsView._ascii.ob_base,
            & const_str_ValuesView._ascii.ob_base,
            & const_str_Sequence._ascii.ob_base,
            & const_str_MutableSequence._ascii.ob_base,
            & const_str_ByteString._ascii.ob_base,
            & const_str_Buffer._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
_collections_abc_toplevel_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "collections.abc",
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_1000 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 1000 },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
_collections_abc_toplevel_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\x9b\x35",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 35,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 88,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 371,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_lambda),
    .co_qualname = &_Py_STR(anon_lambda),
    .co_linetable = & _collections_abc_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x64\x00\x96\x00\x97\x01\x53\x00",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__coro = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_coro",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\x90\x34\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_14_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x02\x04\x01",
};
static
    struct _PyCode_DEF(12)
_collections_abc_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_14_exceptiontable.ob_base.ob_base,
    .co_flags = 131,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 90,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 372,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__coro._ascii.ob_base,
    .co_qualname = & const_str__coro._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str__ag = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ag",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\x95\x15\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_15_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x07\x09\x01",
};
static
    struct _PyCode_DEF(22)
_collections_abc_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 11,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_15_exceptiontable.ob_base.ob_base,
    .co_flags = 515,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 96,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 373,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__ag._ascii.ob_base,
    .co_qualname = & const_str__ag._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x64\x00\xad\x04\x96\x01\x97\x01\x01\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str___mro__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__mro__",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str___mro__._ascii.ob_base,
            &_Py_ID(__dict__),
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__check_methods = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_check_methods",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[81];
    }
_collections_abc_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 80,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0a\x0b\x8f\x29\x89\x29\x80\x43\xdb\x12\x19\x88\x06\xdb\x11\x14\x88\x41\xd8\x0f\x15\x98\x11\x9f\x1a\x99\x1a\xd2\x0f\x23\xd8\x13\x14\x97\x3a\x91\x3a\x98\x66\xd1\x13\x25\xd0\x13\x2d\xdc\x1b\x29\xd4\x14\x29\xd9\x10\x15\xf0\x09\x00\x12\x15\xf4\x0c\x00\x14\x22\xd2\x0c\x21\xf0\x0f\x00\x13\x1a\xf0\x10\x00\x0c\x10",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_methods = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "methods",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[67],
            & const_str_methods._ascii.ob_base,
            &_Py_ID(mro),
            &_Py_ID(method),
            (PyObject *)&_Py_SINGLETON(strings).ascii[66],
        },
    },
};
static
    struct _PyCode_DEF(152)
_collections_abc_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 76,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_68_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 104,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 374,
    .co_localsplusnames = & _collections_abc_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__check_methods._ascii.ob_base,
    .co_qualname = & const_str__check_methods._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x01\x44\x00\x5d\x39\x00\x00\x7d\x03\x7c\x02\x44\x00\x5d\x2b\x00\x00\x7d\x04\x7c\x03\x7c\x04\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x73\x01\x8c\x12\x7c\x04\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x19\x00\x00\x00\x80\x0a\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x63\x02\x01\x00\x63\x02\x01\x00\x53\x00\x01\x00\x8c\x32\x04\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x63\x02\x01\x00\x53\x00\x04\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_17_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Hashable.__hash__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
_collections_abc_toplevel_consts_17_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x10",
};
static
    struct _PyCode_DEF(4)
_collections_abc_toplevel_consts_17_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 120,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 375,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__hash__),
    .co_qualname = & _collections_abc_toplevel_consts_17_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_17_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_17_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__hash__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_17_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Hashable._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
_collections_abc_toplevel_consts_17_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Hashable.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
_collections_abc_toplevel_consts_17_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x28\x89\x3f\xdc\x13\x21\xa0\x21\xa0\x5a\xd3\x13\x30\xd0\x0c\x30\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_17_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[67],
        },
    },
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_17_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_17_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_17_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 124,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 376,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_17_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_17_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_Hashable._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_17_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_17_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_17_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__hash__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[46];
    }
_collections_abc_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 45,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x01\x05\x11\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x11\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x03\x05\x1e",
};
static
    struct _PyCode_DEF(48)
_collections_abc_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 24,
    },
    .co_consts = & _collections_abc_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_17_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 116,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 377,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Hashable._ascii.ob_base,
    .co_qualname = & const_str_Hashable._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_20_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Awaitable.__await__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
_collections_abc_toplevel_consts_20_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xe4\x08\x0d\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_20_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x06\x08\x01",
};
static
    struct _PyCode_DEF(20)
_collections_abc_toplevel_consts_20_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 10,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_20_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 135,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 378,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__await__),
    .co_qualname = & _collections_abc_toplevel_consts_20_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_20_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x64\x00\x96\x01\x97\x01\x01\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_20_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__await__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_20_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Awaitable._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_20_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Awaitable.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
_collections_abc_toplevel_consts_20_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x29\xd1\x0b\x1b\xdc\x13\x21\xa0\x21\xa0\x5b\xd3\x13\x31\xd0\x0c\x31\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_20_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_20_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_20_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 139,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 379,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_20_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_20_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_20_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_Awaitable._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_20_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_20_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_GenericAlias = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "GenericAlias",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__await__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
            & const_str_GenericAlias._ascii.ob_base,
            &_Py_ID(__class_getitem__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[59];
    }
_collections_abc_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 58,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x01\x05\x0e\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x0e\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf0\x02\x03\x05\x1e\xf1\x0a\x00\x19\x24\xa0\x4c\xd3\x18\x31\xd1\x04\x15",
};
static
    struct _PyCode_DEF(64)
_collections_abc_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & _collections_abc_toplevel_consts_20_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 131,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 380,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Awaitable._ascii.ob_base,
    .co_qualname = & const_str_Awaitable._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x02\x00\x65\x06\x65\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[100];
    }
_collections_abc_toplevel_consts_22_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 99,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x65\x6e\x64\x20\x61\x20\x76\x61\x6c\x75\x65\x20\x69\x6e\x74\x6f\x20\x74\x68\x65\x20\x63\x6f\x72\x6f\x75\x74\x69\x6e\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x6e\x65\x78\x74\x20\x79\x69\x65\x6c\x64\x65\x64\x20\x76\x61\x6c\x75\x65\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_22_consts_2_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_StopIteration._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_22_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Coroutine.send",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
_collections_abc_toplevel_consts_22_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x0f\x1c\xd0\x08\x1b",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_22_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & _collections_abc_toplevel_consts_22_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 152,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 381,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(send),
    .co_qualname = & _collections_abc_toplevel_consts_22_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[104];
    }
_collections_abc_toplevel_consts_22_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 103,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x61\x69\x73\x65\x20\x61\x6e\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x69\x6e\x20\x74\x68\x65\x20\x63\x6f\x72\x6f\x75\x74\x69\x6e\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x6e\x65\x78\x74\x20\x79\x69\x65\x6c\x64\x65\x64\x20\x76\x61\x6c\x75\x65\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_22_consts_4_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_with_traceback = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "with_traceback",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_with_traceback._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
_collections_abc_toplevel_consts_22_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Coroutine.throw",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[53];
    }
_collections_abc_toplevel_consts_22_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 52,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x0c\x0f\x88\x3b\xd8\x0f\x11\x88\x7a\xd8\x16\x19\x90\x09\xd9\x12\x15\x93\x25\x88\x43\xd8\x0b\x0d\x88\x3e\xd8\x12\x15\xd7\x12\x24\xd1\x12\x24\xa0\x52\xd3\x12\x28\x88\x43\xd8\x0e\x11\x88\x09",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_typ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "typ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_val = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "val",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_typ._ascii.ob_base,
            & const_str_val._ascii.ob_base,
            & const_str_tb._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(70)
_collections_abc_toplevel_consts_22_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & _collections_abc_toplevel_consts_22_consts_4_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 159,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 382,
    .co_localsplusnames = & _collections_abc_toplevel_consts_22_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(throw),
    .co_qualname = & _collections_abc_toplevel_consts_22_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x80\x0b\x7c\x03\x80\x02\x7c\x01\x82\x01\x02\x00\x7c\x01\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x03\x81\x11\x7c\x02\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
_collections_abc_toplevel_consts_22_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x61\x69\x73\x65\x20\x47\x65\x6e\x65\x72\x61\x74\x6f\x72\x45\x78\x69\x74\x20\x69\x6e\x73\x69\x64\x65\x20\x63\x6f\x72\x6f\x75\x74\x69\x6e\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
_collections_abc_toplevel_consts_22_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "coroutine ignored GeneratorExit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_22_consts_5_consts_0._ascii.ob_base,
            & _collections_abc_toplevel_consts_22_consts_5_consts_1._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_GeneratorExit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "GeneratorExit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(throw),
            & const_str_GeneratorExit._ascii.ob_base,
            & const_str_RuntimeError._ascii.ob_base,
            & const_str_StopIteration._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
_collections_abc_toplevel_consts_22_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Coroutine.close",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[60];
    }
_collections_abc_toplevel_consts_22_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 59,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x05\x09\x42\x01\xd8\x0c\x10\x8f\x4a\x89\x4a\x94\x7d\xd4\x0c\x25\xf4\x08\x00\x13\x1f\xd0\x1f\x40\xd3\x12\x41\xd0\x0c\x41\xf8\xf4\x07\x00\x11\x1e\x9c\x7d\xd0\x0f\x2d\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_22_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x22\x00\xa2\x0f\x34\x03\xb3\x01\x34\x03",
};
static
    struct _PyCode_DEF(110)
_collections_abc_toplevel_consts_22_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & _collections_abc_toplevel_consts_22_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_22_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 172,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 383,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(close),
    .co_qualname = & _collections_abc_toplevel_consts_22_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__await__),
            &_Py_ID(send),
            &_Py_ID(throw),
            &_Py_ID(close),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Coroutine._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_22_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Coroutine.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
_collections_abc_toplevel_consts_22_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x29\xd1\x0b\x1b\xdc\x13\x21\xa0\x21\xa0\x5b\xb0\x26\xb8\x27\xc0\x37\xd3\x13\x4b\xd0\x0c\x4b\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(60)
_collections_abc_toplevel_consts_22_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & _collections_abc_toplevel_consts_22_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 182,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 384,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_22_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0f\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x02\x64\x03\x64\x04\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_22_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_Coroutine._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_22_consts_2.ob_base.ob_base,
            Py_None,
            & _collections_abc_toplevel_consts_22_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_22_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_22_consts_6.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_22_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(send),
            &_Py_ID(throw),
            &_Py_ID(close),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[72];
    }
_collections_abc_toplevel_consts_22_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 71,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x04\x05\x1c\xf3\x03\x00\x06\x14\xf0\x02\x04\x05\x1c\xf0\x0c\x00\x06\x14\xf2\x02\x0a\x05\x12\xf3\x03\x00\x06\x14\xf0\x02\x0a\x05\x12\xf2\x18\x08\x05\x42\x01\xf0\x14\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x03\x05\x1e",
};
static
    struct _PyCode_DEF(72)
_collections_abc_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 36,
    },
    .co_consts = & _collections_abc_toplevel_consts_22_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 148,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 385,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Coroutine._ascii.ob_base,
    .co_qualname = & const_str_Coroutine._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x04\x64\x07\x64\x04\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x65\x08\x64\x06\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_24_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_AsyncIterator._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_24_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncIterable.__aiter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
_collections_abc_toplevel_consts_24_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x1c\x8b\x7f\xd0\x08\x1e",
};
static
    struct _PyCode_DEF(22)
_collections_abc_toplevel_consts_24_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 11,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_24_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 196,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 386,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__aiter__),
    .co_qualname = & _collections_abc_toplevel_consts_24_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_24_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_24_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__aiter__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_24_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_AsyncIterable._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
_collections_abc_toplevel_consts_24_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncIterable.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
_collections_abc_toplevel_consts_24_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x2d\xd1\x0b\x1f\xdc\x13\x21\xa0\x21\xa0\x5b\xd3\x13\x31\xd0\x0c\x31\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_24_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_24_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_24_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 200,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 387,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_24_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_24_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_24_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_AsyncIterable._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_24_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_24_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_24_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__aiter__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
            & const_str_GenericAlias._ascii.ob_base,
            &_Py_ID(__class_getitem__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[59];
    }
_collections_abc_toplevel_consts_24_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 58,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x01\x05\x1f\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x1f\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf0\x02\x03\x05\x1e\xf1\x0a\x00\x19\x24\xa0\x4c\xd3\x18\x31\xd1\x04\x15",
};
static
    struct _PyCode_DEF(64)
_collections_abc_toplevel_consts_24 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & _collections_abc_toplevel_consts_24_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_24_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 192,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 388,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_AsyncIterable._ascii.ob_base,
    .co_qualname = & const_str_AsyncIterable._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_24_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x02\x00\x65\x06\x65\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[65];
    }
_collections_abc_toplevel_consts_26_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 64,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the next item or raise StopAsyncIteration when exhausted.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_26_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_26_consts_2_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_StopAsyncIteration = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "StopAsyncIteration",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_26_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_StopAsyncIteration._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_26_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncIterator.__anext__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
_collections_abc_toplevel_consts_26_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf4\x06\x00\x0f\x21\xd0\x08\x20\xf9",
};
static
    struct _PyCode_DEF(22)
_collections_abc_toplevel_consts_26_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 11,
    },
    .co_consts = & _collections_abc_toplevel_consts_26_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_26_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_15_exceptiontable.ob_base.ob_base,
    .co_flags = 131,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 213,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 389,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__anext__),
    .co_qualname = & _collections_abc_toplevel_consts_26_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_26_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_26_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncIterator.__aiter__",
};
static
    struct _PyCode_DEF(6)
_collections_abc_toplevel_consts_26_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 218,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 390,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__aiter__),
    .co_qualname = & _collections_abc_toplevel_consts_26_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_26_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__anext__),
            &_Py_ID(__aiter__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_26_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_AsyncIterator._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
_collections_abc_toplevel_consts_26_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncIterator.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
_collections_abc_toplevel_consts_26_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x2d\xd1\x0b\x1f\xdc\x13\x21\xa0\x21\xa0\x5b\xb0\x2b\xd3\x13\x3e\xd0\x0c\x3e\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(56)
_collections_abc_toplevel_consts_26_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & _collections_abc_toplevel_consts_26_consts_4_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_26_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 221,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 391,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_26_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_26_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0d\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_26_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_AsyncIterator._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_26_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_26_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_26_consts_4.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_collections_abc_toplevel_consts_26_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__anext__),
            &_Py_ID(__aiter__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[51];
    }
_collections_abc_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 50,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x02\x05\x21\xf3\x03\x00\x06\x14\xf0\x02\x02\x05\x21\xf2\x08\x01\x05\x14\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x03\x05\x1e",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_26_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 209,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 392,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_AsyncIterator._ascii.ob_base,
    .co_qualname = & const_str_AsyncIterator._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x64\x03\x84\x00\x5a\x06\x65\x07\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[113];
    }
_collections_abc_toplevel_consts_28_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 112,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x6e\x65\x78\x74\x20\x69\x74\x65\x6d\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x61\x73\x79\x6e\x63\x68\x72\x6f\x6e\x6f\x75\x73\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x57\x68\x65\x6e\x20\x65\x78\x68\x61\x75\x73\x74\x65\x64\x2c\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x41\x73\x79\x6e\x63\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_28_consts_2_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_asend = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "asend",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_asend._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
_collections_abc_toplevel_consts_28_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncGenerator.__anext__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
_collections_abc_toplevel_consts_28_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf0\x08\x00\x16\x1a\x97\x5a\x91\x5a\xa0\x04\xd3\x15\x25\xd7\x0f\x25\xd0\x08\x25\xd0\x0f\x25\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_28_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x1e\x01\x97\x01\x1c\x04\x98\x05\x1e\x01",
};
static
    struct _PyCode_DEF(64)
_collections_abc_toplevel_consts_28_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & _collections_abc_toplevel_consts_28_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_28_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_28_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 131,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 232,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 393,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__anext__),
    .co_qualname = & _collections_abc_toplevel_consts_28_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_28_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x83\x00\x64\x01\x7b\x03\x00\x00\x96\x02\x97\x03\x86\x05\x05\x00\x53\x00\x37\x00\x8c\x04\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[118];
    }
_collections_abc_toplevel_consts_28_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 117,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x65\x6e\x64\x20\x61\x20\x76\x61\x6c\x75\x65\x20\x69\x6e\x74\x6f\x20\x74\x68\x65\x20\x61\x73\x79\x6e\x63\x68\x72\x6f\x6e\x6f\x75\x73\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x6e\x65\x78\x74\x20\x79\x69\x65\x6c\x64\x65\x64\x20\x76\x61\x6c\x75\x65\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x41\x73\x79\x6e\x63\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_28_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
_collections_abc_toplevel_consts_28_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncGenerator.asend",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
_collections_abc_toplevel_consts_28_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf4\x0a\x00\x0f\x21\xd0\x08\x20\xf9",
};
static
    struct _PyCode_DEF(22)
_collections_abc_toplevel_consts_28_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 11,
    },
    .co_consts = & _collections_abc_toplevel_consts_28_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_26_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_15_exceptiontable.ob_base.ob_base,
    .co_flags = 131,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 238,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 394,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_asend._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_28_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_28_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[122];
    }
_collections_abc_toplevel_consts_28_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 121,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x61\x69\x73\x65\x20\x61\x6e\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x69\x6e\x20\x74\x68\x65\x20\x61\x73\x79\x6e\x63\x68\x72\x6f\x6e\x6f\x75\x73\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x6e\x65\x78\x74\x20\x79\x69\x65\x6c\x64\x65\x64\x20\x76\x61\x6c\x75\x65\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x41\x73\x79\x6e\x63\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_28_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_athrow = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "athrow",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
_collections_abc_toplevel_consts_28_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncGenerator.athrow",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[57];
    }
_collections_abc_toplevel_consts_28_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 56,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf0\x0a\x00\x0c\x0f\x88\x3b\xd8\x0f\x11\x88\x7a\xd8\x16\x19\x90\x09\xd9\x12\x15\x93\x25\x88\x43\xd8\x0b\x0d\x88\x3e\xd8\x12\x15\xd7\x12\x24\xd1\x12\x24\xa0\x52\xd3\x12\x28\x88\x43\xd8\x0e\x11\x88\x09\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_28_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x23\x25\x01",
};
static
    struct _PyCode_DEF(78)
_collections_abc_toplevel_consts_28_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & _collections_abc_toplevel_consts_28_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_28_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 131,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 245,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 395,
    .co_localsplusnames = & _collections_abc_toplevel_consts_22_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_athrow._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_28_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_28_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x02\x80\x0b\x7c\x03\x80\x02\x7c\x01\x82\x01\x02\x00\x7c\x01\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x03\x81\x11\x7c\x02\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x82\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
_collections_abc_toplevel_consts_28_consts_6_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "asynchronous generator ignored GeneratorExit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_22_consts_5_consts_0._ascii.ob_base,
            Py_None,
            & _collections_abc_toplevel_consts_28_consts_6_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_athrow._ascii.ob_base,
            & const_str_GeneratorExit._ascii.ob_base,
            & const_str_RuntimeError._ascii.ob_base,
            & const_str_StopAsyncIteration._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_aclose = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "aclose",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
_collections_abc_toplevel_consts_28_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncGenerator.aclose",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[73];
    }
_collections_abc_toplevel_consts_28_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 72,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf0\x06\x05\x09\x4f\x01\xd8\x12\x16\x97\x2b\x91\x2b\x9c\x6d\xd3\x12\x2c\xd7\x0c\x2c\xd0\x0c\x2c\xf4\x08\x00\x13\x1f\xd0\x1f\x4d\xd3\x12\x4e\xd0\x0c\x4e\xf0\x09\x00\x0d\x2d\xf9\xdc\x10\x1d\xd4\x1f\x31\xd0\x0f\x32\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[48];
    }
_collections_abc_toplevel_consts_28_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 47,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x01\x41\x03\x01\x84\x18\x2e\x00\x9c\x01\x2c\x04\x9d\x04\x2e\x00\xa1\x0b\x41\x03\x01\xac\x01\x2e\x00\xae\x0f\x41\x00\x03\xbd\x02\x41\x03\x01\xbf\x01\x41\x00\x03\xc1\x00\x03\x41\x03\x01",
};
static
    struct _PyCode_DEF(138)
_collections_abc_toplevel_consts_28_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 69,
    },
    .co_consts = & _collections_abc_toplevel_consts_28_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_28_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_28_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 131,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 258,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 396,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_aclose._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_28_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_28_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x09\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x83\x00\x64\x01\x7b\x03\x00\x00\x96\x03\x97\x03\x86\x05\x05\x00\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x37\x00\x8c\x0f\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__aiter__),
            &_Py_ID(__anext__),
            & const_str_asend._ascii.ob_base,
            & const_str_athrow._ascii.ob_base,
            & const_str_aclose._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_AsyncGenerator._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
_collections_abc_toplevel_consts_28_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "AsyncGenerator.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
_collections_abc_toplevel_consts_28_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x2e\xd1\x0b\x20\xdc\x13\x21\xa0\x21\xa0\x5b\xb0\x2b\xd8\x22\x29\xa8\x38\xb0\x58\xf3\x03\x01\x14\x3f\xf0\x00\x01\x0d\x3f\xe4\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(62)
_collections_abc_toplevel_consts_28_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & _collections_abc_toplevel_consts_28_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_28_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 268,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 397,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_28_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_28_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x10\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x02\x64\x03\x64\x04\x64\x05\xab\x06\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_collections_abc_toplevel_consts_28_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_AsyncGenerator._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_28_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_28_consts_3.ob_base.ob_base,
            Py_None,
            & _collections_abc_toplevel_consts_28_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_28_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_28_consts_7.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
_collections_abc_toplevel_consts_28_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            &_Py_ID(__anext__),
            & const_str_abstractmethod._ascii.ob_base,
            & const_str_asend._ascii.ob_base,
            & const_str_athrow._ascii.ob_base,
            & const_str_aclose._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[79];
    }
_collections_abc_toplevel_consts_28_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 78,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xf2\x04\x04\x05\x26\xf0\x0c\x00\x06\x14\xf1\x02\x04\x05\x21\xf3\x03\x00\x06\x14\xf0\x02\x04\x05\x21\xf0\x0c\x00\x06\x14\xf2\x02\x0a\x05\x12\xf3\x03\x00\x06\x14\xf0\x02\x0a\x05\x12\xf2\x18\x08\x05\x4f\x01\xf0\x14\x00\x06\x11\xf1\x02\x04\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x04\x05\x1e",
};
static
    struct _PyCode_DEF(78)
_collections_abc_toplevel_consts_28 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & _collections_abc_toplevel_consts_28_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_28_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 228,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 398,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_AsyncGenerator._ascii.ob_base,
    .co_qualname = & const_str_AsyncGenerator._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_28_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x65\x05\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x05\x64\x08\x64\x05\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x65\x09\x64\x07\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0a\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_30_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Iterable.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
_collections_abc_toplevel_consts_30_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xe0\x0e\x13\xf9",
};
static
    struct _PyCode_DEF(12)
_collections_abc_toplevel_consts_30_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_14_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 283,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 399,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & _collections_abc_toplevel_consts_30_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_30_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_30_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__iter__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_30_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Iterable._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
_collections_abc_toplevel_consts_30_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Iterable.__subclasshook__",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_30_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_30_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_30_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 288,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 400,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_30_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_17_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_30_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_Iterable._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_30_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_30_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_30_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__iter__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
            & const_str_GenericAlias._ascii.ob_base,
            &_Py_ID(__class_getitem__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[59];
    }
_collections_abc_toplevel_consts_30_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 58,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x02\x05\x17\xf3\x03\x00\x06\x14\xf0\x02\x02\x05\x17\xf0\x08\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf0\x02\x03\x05\x1e\xf1\x0a\x00\x19\x24\xa0\x4c\xd3\x18\x31\xd1\x04\x15",
};
static
    struct _PyCode_DEF(64)
_collections_abc_toplevel_consts_30 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & _collections_abc_toplevel_consts_30_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_30_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 279,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 401,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Iterable._ascii.ob_base,
    .co_qualname = & const_str_Iterable._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_30_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x02\x00\x65\x06\x65\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[76];
    }
_collections_abc_toplevel_consts_32_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 75,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the next item from the iterator. When exhausted, raise StopIteration",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_32_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_32_consts_2_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_32_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Iterator.__next__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
_collections_abc_toplevel_consts_32_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x0f\x1c\xd0\x08\x1b",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_32_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & _collections_abc_toplevel_consts_32_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 301,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 402,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__next__),
    .co_qualname = & _collections_abc_toplevel_consts_32_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_32_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_32_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Iterator.__iter__",
};
static
    struct _PyCode_DEF(6)
_collections_abc_toplevel_consts_32_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 306,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 403,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & _collections_abc_toplevel_consts_32_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_24_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_32_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__iter__),
            &_Py_ID(__next__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_32_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Iterator._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
_collections_abc_toplevel_consts_32_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Iterator.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
_collections_abc_toplevel_consts_32_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x28\x89\x3f\xdc\x13\x21\xa0\x21\xa0\x5a\xb0\x1a\xd3\x13\x3c\xd0\x0c\x3c\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(56)
_collections_abc_toplevel_consts_32_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & _collections_abc_toplevel_consts_32_consts_4_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_32_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 309,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 404,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_32_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_32_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0d\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_32_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_Iterator._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_32_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_32_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_32_consts_4.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_collections_abc_toplevel_consts_32_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__next__),
            &_Py_ID(__iter__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[51];
    }
_collections_abc_toplevel_consts_32_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 50,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x02\x05\x1c\xf3\x03\x00\x06\x14\xf0\x02\x02\x05\x1c\xf2\x08\x01\x05\x14\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x03\x05\x1e",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_32 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_32_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_32_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 297,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 405,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Iterator._ascii.ob_base,
    .co_qualname = & const_str_Iterator._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_32_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x64\x03\x84\x00\x5a\x06\x65\x07\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_34_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Reversible.__reversed__",
};
static
    struct _PyCode_DEF(12)
_collections_abc_toplevel_consts_34_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_14_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 336,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 406,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reversed__),
    .co_qualname = & _collections_abc_toplevel_consts_34_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_30_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_34_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__reversed__),
            &_Py_ID(__iter__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_34_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Reversible._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
_collections_abc_toplevel_consts_34_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Reversible.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
_collections_abc_toplevel_consts_34_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x2a\xd1\x0b\x1c\xdc\x13\x21\xa0\x21\xa0\x5e\xb0\x5a\xd3\x13\x40\xd0\x0c\x40\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(56)
_collections_abc_toplevel_consts_34_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & _collections_abc_toplevel_consts_34_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_34_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 341,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 407,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_34_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_34_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0d\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_34_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_Reversible._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_34_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_34_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_34_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__reversed__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[46];
    }
_collections_abc_toplevel_consts_34_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 45,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x02\x05\x17\xf3\x03\x00\x06\x14\xf0\x02\x02\x05\x17\xf0\x08\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x03\x05\x1e",
};
static
    struct _PyCode_DEF(48)
_collections_abc_toplevel_consts_34 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 24,
    },
    .co_consts = & _collections_abc_toplevel_consts_34_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_34_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 332,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 408,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Reversible._ascii.ob_base,
    .co_qualname = & const_str_Reversible._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_34_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[95];
    }
_collections_abc_toplevel_consts_36_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 94,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x6e\x65\x78\x74\x20\x69\x74\x65\x6d\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x57\x68\x65\x6e\x20\x65\x78\x68\x61\x75\x73\x74\x65\x64\x2c\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_36_consts_2_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(send),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
_collections_abc_toplevel_consts_36_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Generator.__next__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
_collections_abc_toplevel_consts_36_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x10\x14\x8f\x79\x89\x79\x98\x14\x8b\x7f\xd0\x08\x1e",
};
static
    struct _PyCode_DEF(36)
_collections_abc_toplevel_consts_36_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 18,
    },
    .co_consts = & _collections_abc_toplevel_consts_36_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_36_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 352,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 409,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__next__),
    .co_qualname = & _collections_abc_toplevel_consts_36_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_36_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[100];
    }
_collections_abc_toplevel_consts_36_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 99,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x65\x6e\x64\x20\x61\x20\x76\x61\x6c\x75\x65\x20\x69\x6e\x74\x6f\x20\x74\x68\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x6e\x65\x78\x74\x20\x79\x69\x65\x6c\x64\x65\x64\x20\x76\x61\x6c\x75\x65\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_36_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_36_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Generator.send",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_36_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & _collections_abc_toplevel_consts_36_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 358,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 410,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(send),
    .co_qualname = & _collections_abc_toplevel_consts_36_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[104];
    }
_collections_abc_toplevel_consts_36_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 103,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x61\x69\x73\x65\x20\x61\x6e\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x69\x6e\x20\x74\x68\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x6e\x65\x78\x74\x20\x79\x69\x65\x6c\x64\x65\x64\x20\x76\x61\x6c\x75\x65\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x53\x74\x6f\x70\x49\x74\x65\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_36_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
_collections_abc_toplevel_consts_36_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Generator.throw",
};
static
    struct _PyCode_DEF(70)
_collections_abc_toplevel_consts_36_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & _collections_abc_toplevel_consts_36_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 365,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 411,
    .co_localsplusnames = & _collections_abc_toplevel_consts_22_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(throw),
    .co_qualname = & _collections_abc_toplevel_consts_36_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x80\x0b\x7c\x03\x80\x02\x7c\x01\x82\x01\x02\x00\x7c\x01\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x03\x81\x11\x7c\x02\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
_collections_abc_toplevel_consts_36_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x61\x69\x73\x65\x20\x47\x65\x6e\x65\x72\x61\x74\x6f\x72\x45\x78\x69\x74\x20\x69\x6e\x73\x69\x64\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
_collections_abc_toplevel_consts_36_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "generator ignored GeneratorExit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_36_consts_6_consts_0._ascii.ob_base,
            & _collections_abc_toplevel_consts_36_consts_6_consts_1._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
_collections_abc_toplevel_consts_36_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Generator.close",
};
static
    struct _PyCode_DEF(110)
_collections_abc_toplevel_consts_36_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & _collections_abc_toplevel_consts_36_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_22_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_22_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 378,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 412,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(close),
    .co_qualname = & _collections_abc_toplevel_consts_36_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_22_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__iter__),
            &_Py_ID(__next__),
            &_Py_ID(send),
            &_Py_ID(throw),
            &_Py_ID(close),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Generator._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_36_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Generator.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
_collections_abc_toplevel_consts_36_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x29\xd1\x0b\x1b\xdc\x13\x21\xa0\x21\xa0\x5a\xb0\x1a\xd8\x22\x28\xa8\x27\xb0\x37\xf3\x03\x01\x14\x3c\xf0\x00\x01\x0d\x3c\xe4\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(62)
_collections_abc_toplevel_consts_36_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & _collections_abc_toplevel_consts_36_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_36_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 388,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 413,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_36_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_36_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x10\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x02\x64\x03\x64\x04\x64\x05\xab\x06\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_collections_abc_toplevel_consts_36_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_Generator._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_36_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_36_consts_3.ob_base.ob_base,
            Py_None,
            & _collections_abc_toplevel_consts_36_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_36_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_36_consts_7.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
_collections_abc_toplevel_consts_36_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            &_Py_ID(__next__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(send),
            &_Py_ID(throw),
            &_Py_ID(close),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[79];
    }
_collections_abc_toplevel_consts_36_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 78,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xf2\x04\x04\x05\x1f\xf0\x0c\x00\x06\x14\xf1\x02\x04\x05\x1c\xf3\x03\x00\x06\x14\xf0\x02\x04\x05\x1c\xf0\x0c\x00\x06\x14\xf2\x02\x0a\x05\x12\xf3\x03\x00\x06\x14\xf0\x02\x0a\x05\x12\xf2\x18\x08\x05\x42\x01\xf0\x14\x00\x06\x11\xf1\x02\x04\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x04\x05\x1e",
};
static
    struct _PyCode_DEF(78)
_collections_abc_toplevel_consts_36 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & _collections_abc_toplevel_consts_36_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_36_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 348,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 414,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Generator._ascii.ob_base,
    .co_qualname = & const_str_Generator._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_36_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x65\x05\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x05\x64\x08\x64\x05\x84\x01\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x65\x09\x64\x07\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0a\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
_collections_abc_toplevel_consts_38_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sized.__len__",
};
static
    struct _PyCode_DEF(4)
_collections_abc_toplevel_consts_38_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 403,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 415,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__len__),
    .co_qualname = & _collections_abc_toplevel_consts_38_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_17_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_38_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__len__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_38_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Sized._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_38_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sized.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
_collections_abc_toplevel_consts_38_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x25\x89\x3c\xdc\x13\x21\xa0\x21\xa0\x59\xd3\x13\x2f\xd0\x0c\x2f\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_38_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_38_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_38_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 407,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 416,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_38_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_38_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_38_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_Sized._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_38_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_38_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_38_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__len__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct _PyCode_DEF(48)
_collections_abc_toplevel_consts_38 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 24,
    },
    .co_consts = & _collections_abc_toplevel_consts_38_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_38_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 399,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 417,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Sized._ascii.ob_base,
    .co_qualname = & const_str_Sized._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_40_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Container.__contains__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
_collections_abc_toplevel_consts_40_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0f\x14",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_40_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
        },
    },
};
static
    struct _PyCode_DEF(4)
_collections_abc_toplevel_consts_40_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_30_consts_4_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 418,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 418,
    .co_localsplusnames = & _collections_abc_toplevel_consts_40_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__contains__),
    .co_qualname = & _collections_abc_toplevel_consts_40_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_40_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_40_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__contains__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_40_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Container._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_40_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Container.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
_collections_abc_toplevel_consts_40_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x29\xd1\x0b\x1b\xdc\x13\x21\xa0\x21\xa0\x5e\xd3\x13\x34\xd0\x0c\x34\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_40_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_40_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_40_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 422,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 419,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_40_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_40_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_40_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_Container._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_40_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_40_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_40_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__contains__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
            & const_str_GenericAlias._ascii.ob_base,
            &_Py_ID(__class_getitem__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[59];
    }
_collections_abc_toplevel_consts_40_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 58,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x01\x05\x15\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x15\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf0\x02\x03\x05\x1e\xf1\x0a\x00\x19\x24\xa0\x4c\xd3\x18\x31\xd1\x04\x15",
};
static
    struct _PyCode_DEF(64)
_collections_abc_toplevel_consts_40 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & _collections_abc_toplevel_consts_40_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_40_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 414,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 420,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Container._ascii.ob_base,
    .co_qualname = & const_str_Container._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_40_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x02\x00\x65\x06\x65\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_42_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__len__),
            &_Py_ID(__iter__),
            &_Py_ID(__contains__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_42_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Collection._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
_collections_abc_toplevel_consts_42_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Collection.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
_collections_abc_toplevel_consts_42_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x2a\xd1\x0b\x1c\xdc\x13\x21\xa0\x21\xa0\x69\xb0\x1a\xb8\x5e\xd3\x13\x4c\xd0\x0c\x4c\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(58)
_collections_abc_toplevel_consts_42_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 29,
    },
    .co_consts = & _collections_abc_toplevel_consts_42_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_42_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 435,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 421,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_42_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_42_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0e\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x64\x02\x64\x03\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_42_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_Collection._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_42_consts_2.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_42_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
_collections_abc_toplevel_consts_42_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x10\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x03\x05\x1e",
};
static
    struct _PyCode_DEF(32)
_collections_abc_toplevel_consts_42 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & _collections_abc_toplevel_consts_42_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_42_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 431,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 422,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Collection._ascii.ob_base,
    .co_qualname = & const_str_Collection._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_42_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_44_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Buffer.__buffer__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_44_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0e\x21\xd0\x08\x21",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_44_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(flags),
        },
    },
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_44_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 2,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 446,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 423,
    .co_localsplusnames = & _collections_abc_toplevel_consts_44_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__buffer__),
    .co_qualname = & _collections_abc_toplevel_consts_44_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_44_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_44_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__buffer__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_44_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Buffer._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_44_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Buffer.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
_collections_abc_toplevel_consts_44_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x26\x89\x3d\xdc\x13\x21\xa0\x21\xa0\x5c\xd3\x13\x32\xd0\x0c\x32\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_44_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_44_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_44_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 450,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 424,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_44_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_44_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
_collections_abc_toplevel_consts_44_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_Buffer._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            &_Py_ID(flags),
            &_Py_ID(return),
            & _collections_abc_toplevel_consts_44_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_44_consts_5.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_44_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            & const_str_int._ascii.ob_base,
            & const_str_memoryview._ascii.ob_base,
            &_Py_ID(__buffer__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[60];
    }
_collections_abc_toplevel_consts_44_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 59,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf0\x02\x01\x05\x22\xa0\x03\xf0\x00\x01\x05\x22\xa8\x3a\xf2\x00\x01\x05\x22\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x22\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf1\x02\x03\x05\x1e",
};
static
    struct _PyCode_DEF(58)
_collections_abc_toplevel_consts_44 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 29,
    },
    .co_consts = & _collections_abc_toplevel_consts_44_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_44_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 442,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 425,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Buffer._ascii.ob_base,
    .co_qualname = & const_str_Buffer._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_44_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x65\x05\x64\x03\x65\x06\x66\x04\x64\x04\x84\x04\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x65\x08\x64\x05\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x06",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str__CallableGenericAlias = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CallableGenericAlias",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[253];
    }
_collections_abc_toplevel_consts_46_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 252,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x52\x65\x70\x72\x65\x73\x65\x6e\x74\x20\x60\x43\x61\x6c\x6c\x61\x62\x6c\x65\x5b\x61\x72\x67\x74\x79\x70\x65\x73\x2c\x20\x72\x65\x73\x75\x6c\x74\x74\x79\x70\x65\x5d\x60\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x73\x65\x74\x73\x20\x60\x60\x5f\x5f\x61\x72\x67\x73\x5f\x5f\x60\x60\x20\x74\x6f\x20\x61\x20\x74\x75\x70\x6c\x65\x20\x63\x6f\x6e\x74\x61\x69\x6e\x69\x6e\x67\x20\x74\x68\x65\x20\x66\x6c\x61\x74\x74\x65\x6e\x65\x64\x20\x60\x60\x61\x72\x67\x74\x79\x70\x65\x73\x60\x60\x0a\x20\x20\x20\x20\x66\x6f\x6c\x6c\x6f\x77\x65\x64\x20\x62\x79\x20\x60\x60\x72\x65\x73\x75\x6c\x74\x74\x79\x70\x65\x60\x60\x2e\x0a\x0a\x20\x20\x20\x20\x45\x78\x61\x6d\x70\x6c\x65\x3a\x20\x60\x60\x43\x61\x6c\x6c\x61\x62\x6c\x65\x5b\x5b\x69\x6e\x74\x2c\x20\x73\x74\x72\x5d\x2c\x20\x66\x6c\x6f\x61\x74\x5d\x60\x60\x20\x73\x65\x74\x73\x20\x60\x60\x5f\x5f\x61\x72\x67\x73\x5f\x5f\x60\x60\x20\x74\x6f\x0a\x20\x20\x20\x20\x60\x60\x28\x69\x6e\x74\x2c\x20\x73\x74\x72\x2c\x20\x66\x6c\x6f\x61\x74\x29\x60\x60\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
_collections_abc_toplevel_consts_46_consts_3_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Callable must be used as Callable[[arg, ...], result].",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[71];
    }
_collections_abc_toplevel_consts_46_consts_3_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 70,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Expected a list of types, an ellipsis, ParamSpec, or Concatenate. Got ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & _collections_abc_toplevel_consts_46_consts_3_consts_2._ascii.ob_base,
            & _collections_abc_toplevel_consts_46_consts_3_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__is_param_expr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_is_param_expr",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_tuple._ascii.ob_base,
            &_Py_ID(len),
            & const_str_TypeError._ascii.ob_base,
            & const_str_list._ascii.ob_base,
            & const_str__is_param_expr._ascii.ob_base,
            & const_str_super._ascii.ob_base,
            &_Py_ID(__new__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
_collections_abc_toplevel_consts_46_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CallableGenericAlias.__new__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[139];
    }
_collections_abc_toplevel_consts_46_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 138,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x10\x1a\x98\x34\xa4\x15\xd4\x10\x27\xac\x43\xb0\x04\xab\x49\xb8\x11\xaa\x4e\xdc\x12\x1b\xd8\x10\x48\xf3\x03\x01\x13\x4a\x01\xf0\x00\x01\x0d\x4a\x01\xe0\x1b\x1f\xd1\x08\x18\x88\x06\x90\x08\xdc\x0b\x15\x90\x66\x9c\x75\xa4\x64\x98\x6d\xd4\x0b\x2c\xd8\x13\x26\x90\x56\xd0\x13\x26\x98\x58\xd1\x13\x26\x89\x44\xdc\x11\x1f\xa0\x06\xd4\x11\x27\xdc\x12\x1b\xf0\x00\x01\x1f\x3e\xd8\x3e\x44\xb8\x58\xf0\x03\x01\x1d\x47\x01\xf3\x00\x01\x13\x48\x01\xf0\x00\x01\x0d\x48\x01\xe4\x0f\x14\x89\x77\x89\x7f\x98\x73\xa0\x46\xa8\x44\xd3\x0f\x31\xd0\x08\x31",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_t_args = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "t_args",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_t_result = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "t_result",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            &_Py_ID(origin),
            &_Py_ID(args),
            & const_str_t_args._ascii.ob_base,
            & const_str_t_result._ascii.ob_base,
            &_Py_ID(__class__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[7];
    }
_collections_abc_toplevel_consts_46_consts_3_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 6,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x20\x20\x20\x80",
};
static
    struct _PyCode_DEF(240)
_collections_abc_toplevel_consts_46_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 120,
    },
    .co_consts = & _collections_abc_toplevel_consts_46_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_46_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 469,
    .co_nlocalsplus = 6,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 426,
    .co_localsplusnames = & _collections_abc_toplevel_consts_46_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & _collections_abc_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__new__),
    .co_qualname = & _collections_abc_toplevel_consts_46_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_46_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x0e\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x73\x0b\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x5c\x02\x00\x00\x7d\x03\x7d\x04\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\xab\x02\x00\x00\x00\x00\x00\x00\x72\x08\x67\x00\x7c\x03\xa2\x01\x7c\x04\x91\x01\xad\x06\x7d\x02\x6e\x19\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x73\x0e\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x03\x9b\x00\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x89\x05\x7c\x00\x8d\x1d\x00\x00\x7c\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_46_consts_4_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "collections.abc.Callable[[",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
_collections_abc_toplevel_consts_46_consts_4_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "], ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & _collections_abc_toplevel_consts_46_consts_4_consts_3._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_30_consts_5_consts_6._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & _collections_abc_toplevel_consts_46_consts_4_consts_6._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[93],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__type_repr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_type_repr",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(len),
            &_Py_ID(__args__),
            & const_str__is_param_expr._ascii.ob_base,
            & const_str_super._ascii.ob_base,
            &_Py_ID(__repr__),
            &_Py_ID(join),
            & const_str__type_repr._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
_collections_abc_toplevel_consts_46_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CallableGenericAlias.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[140];
    }
_collections_abc_toplevel_consts_46_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 139,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x0b\x0e\x88\x74\x8f\x7d\x89\x7d\xd3\x0b\x1d\xa0\x11\xd2\x0b\x22\xa4\x7e\xb0\x64\xb7\x6d\xb1\x6d\xc0\x41\xd1\x36\x46\xd4\x27\x47\xdc\x13\x18\x91\x37\xd1\x13\x23\xd3\x13\x25\xd0\x0c\x25\xf0\x02\x01\x13\x15\xd8\x15\x19\x97\x59\x91\x59\xb0\x74\xb7\x7d\xb1\x7d\xc0\x53\xc0\x62\xd1\x37\x49\xd3\x1f\x4a\xd1\x37\x49\xb0\x21\xa4\x0a\xa8\x31\xa5\x0d\xd0\x37\x49\xd1\x1f\x4a\xd3\x15\x4b\xd0\x14\x4c\xc8\x43\xdc\x13\x1d\x98\x64\x9f\x6d\x99\x6d\xa8\x42\xd1\x1e\x2f\xd3\x13\x30\xd0\x12\x31\xb0\x11\xf0\x05\x02\x11\x34\xf0\x00\x02\x09\x35\xf9\xda\x1f\x4a",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[7];
    }
_collections_abc_toplevel_consts_46_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 6,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x1d\x12\x42\x12\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            (PyObject *)&_Py_SINGLETON(strings).ascii[97],
            &_Py_ID(__class__),
        },
    },
};
static
    struct _PyCode_DEF(302)
_collections_abc_toplevel_consts_46_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 151,
    },
    .co_consts = & _collections_abc_toplevel_consts_46_consts_4_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_46_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_46_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 481,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 427,
    .co_localsplusnames = & _collections_abc_toplevel_consts_46_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & _collections_abc_toplevel_consts_46_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_46_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x26\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0e\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x89\x02\x7c\x00\x8d\x11\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x53\x00\x64\x03\x64\x04\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x64\x05\x1a\x00\x44\x00\x8f\x01\x63\x02\x67\x00\x63\x02\x5d\x0d\x00\x00\x7d\x01\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x91\x02\x8c\x0f\x04\x00\x63\x02\x7d\x01\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x06\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x07\x9d\x05\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x01\x77\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__args__),
            &_Py_ID(len),
            & const_str__is_param_expr._ascii.ob_base,
            & const_str_list._ascii.ob_base,
            & const_str__CallableGenericAlias._ascii.ob_base,
            & const_str_Callable._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
_collections_abc_toplevel_consts_46_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CallableGenericAlias.__reduce__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[74];
    }
_collections_abc_toplevel_consts_46_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 73,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\x8f\x7d\x89\x7d\x88\x04\xdc\x10\x13\x90\x44\x93\x09\x98\x51\x92\x0e\xa4\x3e\xb0\x24\xb0\x71\xb1\x27\xd4\x23\x3a\xdc\x13\x17\x98\x04\x98\x53\x98\x62\x98\x09\x93\x3f\xa0\x44\xa8\x12\xa1\x48\xd0\x13\x2c\x88\x44\xdc\x0f\x24\xa4\x78\xb0\x14\xd0\x26\x36\xd0\x0f\x36\xd0\x08\x36",
};
static
    struct _PyCode_DEF(148)
_collections_abc_toplevel_consts_46_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 74,
    },
    .co_consts = & _collections_abc_toplevel_consts_46_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_46_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 488,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 428,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reduce__),
    .co_qualname = & _collections_abc_toplevel_consts_46_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_46_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x0e\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x02\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x13\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x00\x64\x03\x1a\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x03\x19\x00\x00\x00\x66\x02\x7d\x01\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x66\x02\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_tuple._ascii.ob_base,
            & const_str_super._ascii.ob_base,
            &_Py_ID(__getitem__),
            &_Py_ID(__args__),
            & const_str_list._ascii.ob_base,
            & const_str__CallableGenericAlias._ascii.ob_base,
            & const_str_Callable._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
_collections_abc_toplevel_consts_46_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_CallableGenericAlias.__getitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[111];
    }
_collections_abc_toplevel_consts_46_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 110,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf4\x0a\x00\x10\x1a\x98\x24\xa4\x05\xd4\x0f\x26\xd8\x14\x18\x90\x37\x88\x44\xe4\x13\x18\x91\x37\xd1\x13\x26\xa0\x74\xd3\x13\x2c\xd7\x13\x35\xd1\x13\x35\x88\x08\xf4\x06\x00\x10\x1a\x98\x28\xa0\x31\x99\x2b\xac\x05\xac\x74\xa0\x7d\xd4\x0f\x35\xd8\x17\x1f\xa0\x02\x91\x7c\x88\x48\xd8\x15\x1d\x98\x63\x98\x72\x90\x5d\x88\x46\xd8\x18\x1e\xa0\x08\xd0\x17\x29\x88\x48\xdc\x0f\x24\xa4\x58\xac\x75\xb0\x58\xab\x7f\xd3\x0f\x3f\xd0\x08\x3f",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_new_args = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "new_args",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(item),
            & const_str_new_args._ascii.ob_base,
            & const_str_t_result._ascii.ob_base,
            & const_str_t_args._ascii.ob_base,
            &_Py_ID(__class__),
        },
    },
};
static
    struct _PyCode_DEF(220)
_collections_abc_toplevel_consts_46_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 110,
    },
    .co_consts = & _collections_abc_toplevel_consts_46_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_46_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 494,
    .co_nlocalsplus = 6,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 429,
    .co_localsplusnames = & _collections_abc_toplevel_consts_46_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & _collections_abc_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getitem__),
    .co_qualname = & _collections_abc_toplevel_consts_46_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_46_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x03\x7c\x01\x66\x01\x7d\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x89\x05\x7c\x00\x8d\x0d\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x01\x19\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x0e\x7c\x02\x64\x02\x19\x00\x00\x00\x7d\x03\x7c\x02\x64\x00\x64\x02\x1a\x00\x7d\x04\x7c\x04\x7c\x03\x66\x02\x7d\x02\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
_collections_abc_toplevel_consts_46_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__CallableGenericAlias._ascii.ob_base,
            & _collections_abc_toplevel_consts_46_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_46_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_46_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_46_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_46_consts_6.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_46_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
            &_Py_ID(__new__),
            &_Py_ID(__repr__),
            &_Py_ID(__reduce__),
            &_Py_ID(__getitem__),
            &_Py_ID(__classcell__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
_collections_abc_toplevel_consts_46_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x84\x00\xf1\x02\x07\x05\x08\xf0\x12\x00\x11\x13\x80\x49\xf4\x04\x0a\x05\x32\xf4\x18\x05\x05\x35\xf2\x0e\x04\x05\x37\xf7\x0c\x0f\x05\x40\x01\xf0\x00\x0f\x05\x40\x01",
};
static
    struct _PyCode_DEF(64)
_collections_abc_toplevel_consts_46 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & _collections_abc_toplevel_consts_46_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_46_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 457,
    .co_nlocalsplus = 1,
    .co_nlocals = 0,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 430,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[64]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__CallableGenericAlias._ascii.ob_base,
    .co_qualname = & const_str__CallableGenericAlias._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_46_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x88\x00\x66\x01\x64\x03\x84\x08\x5a\x05\x88\x00\x66\x01\x64\x04\x84\x08\x5a\x06\x64\x05\x84\x00\x5a\x07\x88\x00\x66\x01\x64\x06\x84\x08\x5a\x08\x88\x00\x78\x01\x5a\x09\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[125];
    }
_collections_abc_toplevel_consts_48_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 124,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x68\x65\x63\x6b\x73\x20\x69\x66\x20\x6f\x62\x6a\x20\x6d\x61\x74\x63\x68\x65\x73\x20\x65\x69\x74\x68\x65\x72\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x74\x79\x70\x65\x73\x2c\x20\x60\x60\x2e\x2e\x2e\x60\x60\x2c\x20\x60\x60\x50\x61\x72\x61\x6d\x53\x70\x65\x63\x60\x60\x20\x6f\x72\x0a\x20\x20\x20\x20\x60\x60\x5f\x43\x6f\x6e\x63\x61\x74\x65\x6e\x61\x74\x65\x47\x65\x6e\x65\x72\x69\x63\x41\x6c\x69\x61\x73\x60\x60\x20\x66\x72\x6f\x6d\x20\x74\x79\x70\x69\x6e\x67\x2e\x70\x79\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_ParamSpec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ParamSpec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str__ConcatenateGenericAlias = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ConcatenateGenericAlias",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_48_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_ParamSpec._ascii.ob_base,
            & const_str__ConcatenateGenericAlias._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_typing = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "typing",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_48_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
_collections_abc_toplevel_consts_48_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_is_param_expr.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
_collections_abc_toplevel_consts_48_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xd0\x2d\x55\xc9\x75\xc0\x74\xa8\x63\xaf\x6c\xa9\x6c\xb8\x64\xd5\x2e\x42\xc9\x75\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_48_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x19\x1c\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_48_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(obj),
        },
    },
};
static
    struct _PyCode_DEF(60)
_collections_abc_toplevel_consts_48_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_48_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_48_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 521,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 431,
    .co_localsplusnames = & _collections_abc_toplevel_consts_48_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & _collections_abc_toplevel_consts_48_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_48_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x13\x00\x00\x7d\x01\x89\x02\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6b\x28\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x15\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_48_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_48_consts_0._ascii.ob_base,
            Py_True,
            & _collections_abc_toplevel_consts_48_consts_2._object.ob_base.ob_base,
            & const_str_typing._ascii.ob_base,
            & _collections_abc_toplevel_consts_48_consts_4.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_Ellipsis = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Ellipsis",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_48_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_Ellipsis._ascii.ob_base,
            &_Py_ID(isinstance),
            & const_str_list._ascii.ob_base,
            &_Py_ID(type),
            &_Py_ID(__module__),
            & const_str_any._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[74];
    }
_collections_abc_toplevel_consts_48_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 73,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf0\x08\x00\x08\x0b\x8c\x68\x81\x7f\xd8\x0f\x13\xdc\x07\x11\x90\x23\x94\x74\xd4\x07\x1c\xd8\x0f\x13\xdc\x0a\x0e\x88\x73\x8b\x29\x80\x43\xd8\x0c\x35\x80\x45\xd8\x0b\x0e\x8f\x3e\x89\x3e\x98\x58\xd1\x0b\x25\xd2\x0b\x55\xac\x23\xd3\x2d\x55\xc9\x75\xd3\x2d\x55\xd3\x2a\x55\xd0\x04\x55",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_names = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "names",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_48_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(obj),
            & const_str_names._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(156)
_collections_abc_toplevel_consts_48 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 78,
    },
    .co_consts = & _collections_abc_toplevel_consts_48_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_48_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 511,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 432,
    .co_localsplusnames = & _collections_abc_toplevel_consts_48_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__is_param_expr._ascii.ob_base,
    .co_qualname = & const_str__is_param_expr._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_48_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x89\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x01\x79\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x89\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x01\x79\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x89\x00\xab\x01\x00\x00\x00\x00\x00\x00\x8a\x00\x64\x02\x7d\x01\x89\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x6b\x28\x00\x00\x78\x01\x72\x14\x01\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x88\x00\x66\x01\x64\x04\x84\x08\x7c\x01\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[224];
    }
_collections_abc_toplevel_consts_49_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 223,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x72\x65\x70\x72\x28\x29\x20\x6f\x66\x20\x61\x6e\x20\x6f\x62\x6a\x65\x63\x74\x2c\x20\x73\x70\x65\x63\x69\x61\x6c\x2d\x63\x61\x73\x69\x6e\x67\x20\x74\x79\x70\x65\x73\x20\x28\x69\x6e\x74\x65\x72\x6e\x61\x6c\x20\x68\x65\x6c\x70\x65\x72\x29\x2e\x0a\x0a\x20\x20\x20\x20\x43\x6f\x70\x69\x65\x64\x20\x66\x72\x6f\x6d\x20\x3a\x6d\x6f\x64\x3a\x60\x74\x79\x70\x69\x6e\x67\x60\x20\x73\x69\x6e\x63\x65\x20\x63\x6f\x6c\x6c\x65\x63\x74\x69\x6f\x6e\x73\x2e\x61\x62\x63\x0a\x20\x20\x20\x20\x73\x68\x6f\x75\x6c\x64\x6e\x27\x74\x20\x64\x65\x70\x65\x6e\x64\x20\x6f\x6e\x20\x74\x68\x61\x74\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x20\x20\x20\x20\x28\x4b\x65\x65\x70\x20\x74\x68\x69\x73\x20\x72\x6f\x75\x67\x68\x6c\x79\x20\x69\x6e\x20\x73\x79\x6e\x63\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x74\x79\x70\x69\x6e\x67\x20\x76\x65\x72\x73\x69\x6f\x6e\x2e\x29\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
_collections_abc_toplevel_consts_49_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "...",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_49_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_49_consts_0._ascii.ob_base,
            &_Py_ID(builtins),
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & _collections_abc_toplevel_consts_49_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_FunctionType = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FunctionType",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_repr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "repr",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_49_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            &_Py_ID(type),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            & const_str_Ellipsis._ascii.ob_base,
            & const_str_FunctionType._ascii.ob_base,
            &_Py_ID(__name__),
            & const_str_repr._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[108];
    }
_collections_abc_toplevel_consts_49_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 107,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x08\x12\x90\x23\x94\x74\xd4\x07\x1c\xd8\x0b\x0e\x8f\x3e\x89\x3e\x98\x5a\xd2\x0b\x27\xd8\x13\x16\xd7\x13\x23\xd1\x13\x23\xd0\x0c\x23\xd8\x12\x15\x97\x2e\x91\x2e\xd0\x11\x21\xa0\x11\xa0\x33\xd7\x23\x33\xd1\x23\x33\xd0\x22\x34\xd0\x0f\x35\xd0\x08\x35\xd8\x07\x0a\x8c\x68\x81\x7f\xd8\x0f\x14\xdc\x07\x11\x90\x23\x94\x7c\xd4\x07\x24\xd8\x0f\x12\x8f\x7c\x89\x7c\xd0\x08\x1b\xdc\x0b\x0f\x90\x03\x8b\x39\xd0\x04\x14",
};
static
    struct _PyCode_DEF(238)
_collections_abc_toplevel_consts_49 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 119,
    },
    .co_consts = & _collections_abc_toplevel_consts_49_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_49_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 523,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 433,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__type_repr._ascii.ob_base,
    .co_qualname = & const_str__type_repr._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_49_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x36\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x0c\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x02\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x00\x9d\x03\x53\x00\x7c\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x01\x79\x03\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x0c\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_50_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Callable.__call__",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_50_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(args),
            & const_str_kwds._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(4)
_collections_abc_toplevel_consts_50_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_30_consts_4_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 15,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 545,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 434,
    .co_localsplusnames = & _collections_abc_toplevel_consts_50_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__call__),
    .co_qualname = & _collections_abc_toplevel_consts_50_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_40_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_50_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__call__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_50_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_Callable._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
_collections_abc_toplevel_consts_50_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Callable.__subclasshook__",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_50_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & _collections_abc_toplevel_consts_50_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_50_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 549,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 435,
    .co_localsplusnames = & _collections_abc_toplevel_consts_17_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & _collections_abc_toplevel_consts_50_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_17_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_50_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_Callable._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_50_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_50_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_50_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__call__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
            & const_str__CallableGenericAlias._ascii.ob_base,
            &_Py_ID(__class_getitem__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[60];
    }
_collections_abc_toplevel_consts_50_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 59,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x13\xf1\x02\x01\x05\x15\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x15\xf0\x06\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf0\x02\x03\x05\x1e\xf1\x0a\x00\x19\x24\xd0\x24\x39\xd3\x18\x3a\xd1\x04\x15",
};
static
    struct _PyCode_DEF(64)
_collections_abc_toplevel_consts_50 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & _collections_abc_toplevel_consts_50_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_50_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 541,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 436,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Callable._ascii.ob_base,
    .co_qualname = & const_str_Callable._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_50_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x02\x00\x65\x06\x65\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[347];
    }
_collections_abc_toplevel_consts_52_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 346,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x73\x65\x74\x20\x69\x73\x20\x61\x20\x66\x69\x6e\x69\x74\x65\x2c\x20\x69\x74\x65\x72\x61\x62\x6c\x65\x20\x63\x6f\x6e\x74\x61\x69\x6e\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x63\x6c\x61\x73\x73\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x63\x6f\x6e\x63\x72\x65\x74\x65\x20\x67\x65\x6e\x65\x72\x69\x63\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x73\x20\x6f\x66\x20\x61\x6c\x6c\x0a\x20\x20\x20\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x65\x78\x63\x65\x70\x74\x20\x66\x6f\x72\x20\x5f\x5f\x63\x6f\x6e\x74\x61\x69\x6e\x73\x5f\x5f\x2c\x20\x5f\x5f\x69\x74\x65\x72\x5f\x5f\x20\x61\x6e\x64\x20\x5f\x5f\x6c\x65\x6e\x5f\x5f\x2e\x0a\x0a\x20\x20\x20\x20\x54\x6f\x20\x6f\x76\x65\x72\x72\x69\x64\x65\x20\x74\x68\x65\x20\x63\x6f\x6d\x70\x61\x72\x69\x73\x6f\x6e\x73\x20\x28\x70\x72\x65\x73\x75\x6d\x61\x62\x6c\x79\x20\x66\x6f\x72\x20\x73\x70\x65\x65\x64\x2c\x20\x61\x73\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x73\x65\x6d\x61\x6e\x74\x69\x63\x73\x20\x61\x72\x65\x20\x66\x69\x78\x65\x64\x29\x2c\x20\x72\x65\x64\x65\x66\x69\x6e\x65\x20\x5f\x5f\x6c\x65\x5f\x5f\x20\x61\x6e\x64\x20\x5f\x5f\x67\x65\x5f\x5f\x2c\x0a\x20\x20\x20\x20\x74\x68\x65\x6e\x20\x74\x68\x65\x20\x6f\x74\x68\x65\x72\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x20\x77\x69\x6c\x6c\x20\x61\x75\x74\x6f\x6d\x61\x74\x69\x63\x61\x6c\x6c\x79\x20\x66\x6f\x6c\x6c\x6f\x77\x20\x73\x75\x69\x74\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            Py_False,
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Set._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
            &_Py_ID(len),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
_collections_abc_toplevel_consts_52_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__le__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[64];
    }
_collections_abc_toplevel_consts_52_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 63,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x13\xd4\x0f\x25\xdc\x13\x21\xd0\x0c\x21\xdc\x0b\x0e\x88\x74\x8b\x39\x94\x73\x98\x35\x93\x7a\xd2\x0b\x21\xd8\x13\x18\xdb\x14\x18\x88\x44\xd8\x0f\x13\x98\x35\xd2\x0f\x20\xd9\x17\x1c\xf0\x05\x00\x15\x19\xf0\x06\x00\x10\x14",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_elem = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "elem",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_other._ascii.ob_base,
            & const_str_elem._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(122)
_collections_abc_toplevel_consts_52_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 61,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 574,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 437,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__le__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x44\x00\x00\x72\x01\x79\x01\x7c\x00\x44\x00\x5d\x08\x00\x00\x7d\x02\x7c\x02\x7c\x01\x76\x01\x73\x01\x8c\x08\x01\x00\x79\x01\x04\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Set._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
            &_Py_ID(len),
            &_Py_ID(__le__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
_collections_abc_toplevel_consts_52_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__lt__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
_collections_abc_toplevel_consts_52_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x13\xd4\x0f\x25\xdc\x13\x21\xd0\x0c\x21\xdc\x0f\x12\x90\x34\x8b\x79\x9c\x33\x98\x75\x9b\x3a\xd1\x0f\x25\xd2\x0f\x3c\xa8\x24\xaf\x2b\xa9\x2b\xb0\x65\xd3\x2a\x3c\xd0\x08\x3c",
};
static
    struct _PyCode_DEF(130)
_collections_abc_toplevel_consts_52_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 65,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 584,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 438,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__lt__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x02\x00\x00\x78\x01\x72\x11\x01\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Set._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
            &_Py_ID(len),
            &_Py_ID(__ge__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
_collections_abc_toplevel_consts_52_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__gt__",
};
static
    struct _PyCode_DEF(130)
_collections_abc_toplevel_consts_52_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 65,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 589,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 439,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__gt__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x44\x00\x00\x78\x01\x72\x11\x01\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
_collections_abc_toplevel_consts_52_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__ge__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[64];
    }
_collections_abc_toplevel_consts_52_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 63,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x13\xd4\x0f\x25\xdc\x13\x21\xd0\x0c\x21\xdc\x0b\x0e\x88\x74\x8b\x39\x94\x73\x98\x35\x93\x7a\xd2\x0b\x21\xd8\x13\x18\xdb\x14\x19\x88\x44\xd8\x0f\x13\x98\x34\xd2\x0f\x1f\xd9\x17\x1c\xf0\x05\x00\x15\x1a\xf0\x06\x00\x10\x14",
};
static
    struct _PyCode_DEF(122)
_collections_abc_toplevel_consts_52_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 61,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 594,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 440,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__ge__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x02\x00\x00\x72\x01\x79\x01\x7c\x01\x44\x00\x5d\x08\x00\x00\x7d\x02\x7c\x02\x7c\x00\x76\x01\x73\x01\x8c\x08\x01\x00\x79\x01\x04\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
_collections_abc_toplevel_consts_52_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__eq__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
_collections_abc_toplevel_consts_52_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x13\xd4\x0f\x25\xdc\x13\x21\xd0\x0c\x21\xdc\x0f\x12\x90\x34\x8b\x79\x9c\x43\xa0\x05\x9b\x4a\xd1\x0f\x26\xd2\x0f\x3d\xa8\x34\xaf\x3b\xa9\x3b\xb0\x75\xd3\x2b\x3d\xd0\x08\x3d",
};
static
    struct _PyCode_DEF(130)
_collections_abc_toplevel_consts_52_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 65,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 604,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 441,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__eq__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x11\x01\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[189];
    }
_collections_abc_toplevel_consts_52_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 188,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6e\x73\x74\x72\x75\x63\x74\x20\x61\x6e\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x63\x6c\x61\x73\x73\x20\x66\x72\x6f\x6d\x20\x61\x6e\x79\x20\x69\x74\x65\x72\x61\x62\x6c\x65\x20\x69\x6e\x70\x75\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4d\x75\x73\x74\x20\x6f\x76\x65\x72\x72\x69\x64\x65\x20\x74\x68\x69\x73\x20\x6d\x65\x74\x68\x6f\x64\x20\x69\x66\x20\x74\x68\x65\x20\x63\x6c\x61\x73\x73\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x6f\x72\x20\x73\x69\x67\x6e\x61\x74\x75\x72\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x20\x61\x63\x63\x65\x70\x74\x20\x61\x6e\x20\x69\x74\x65\x72\x61\x62\x6c\x65\x20\x66\x6f\x72\x20\x61\x6e\x20\x69\x6e\x70\x75\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_52_consts_8_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__from_iterable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_from_iterable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
_collections_abc_toplevel_consts_52_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set._from_iterable",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[14];
    }
_collections_abc_toplevel_consts_52_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 13,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf1\x0e\x00\x10\x13\x90\x32\x8b\x77\x88\x0e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_it = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "it",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_it._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(18)
_collections_abc_toplevel_consts_52_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 9,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_8_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 609,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 442,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__from_iterable._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_8_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x02\x00\x7c\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
_collections_abc_toplevel_consts_52_consts_9_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__and__.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
_collections_abc_toplevel_consts_52_consts_9_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xd0\x22\x4d\xb1\x65\xa8\x55\xb8\x75\xc8\x04\xba\x7d\xa4\x35\xb1\x65\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_52_consts_9_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x09\x14\x01\x8d\x07\x14\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_9_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            &_Py_ID(value),
            &_Py_ID(self),
        },
    },
};
static
    struct _PyCode_DEF(44)
_collections_abc_toplevel_consts_52_consts_9_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_52_consts_9_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 621,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 443,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_9_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_9_consts_1_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_9_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x0b\x00\x00\x7d\x01\x7c\x01\x89\x02\x76\x00\x73\x01\x8c\x08\x7c\x01\x96\x01\x97\x01\x01\x00\x8c\x0d\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & _collections_abc_toplevel_consts_52_consts_9_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Iterable._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
            & const_str__from_iterable._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
_collections_abc_toplevel_consts_52_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__and__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
_collections_abc_toplevel_consts_52_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x0f\x19\x98\x25\xa4\x18\xd4\x0f\x2a\xdc\x13\x21\xd0\x0c\x21\xd8\x0f\x13\xd7\x0f\x22\xd1\x0f\x22\xd3\x22\x4d\xb1\x65\xd3\x22\x4d\xd3\x0f\x4d\xd0\x08\x4d",
};
static
    struct _PyCode_DEF(100)
_collections_abc_toplevel_consts_52_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 50,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_9_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 618,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 444,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__and__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_9_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x89\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x88\x00\x66\x01\x64\x01\x84\x08\x7c\x01\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[50];
    }
_collections_abc_toplevel_consts_52_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 49,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if two sets have a null intersection.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_52_consts_10_consts_0._ascii.ob_base,
            Py_False,
            Py_True,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_isdisjoint = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isdisjoint",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_52_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.isdisjoint",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
_collections_abc_toplevel_consts_52_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe3\x15\x1a\x88\x45\xd8\x0f\x14\x98\x04\x8a\x7d\xd9\x17\x1c\xf0\x05\x00\x16\x1b\xf0\x06\x00\x10\x14",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_10_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_other._ascii.ob_base,
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(30)
_collections_abc_toplevel_consts_52_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_10_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 625,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 445,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_10_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_isdisjoint._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_10_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x44\x00\x5d\x08\x00\x00\x7d\x02\x7c\x02\x7c\x00\x76\x00\x73\x01\x8c\x08\x01\x00\x79\x01\x04\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
_collections_abc_toplevel_consts_52_consts_11_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__or__.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
_collections_abc_toplevel_consts_52_consts_11_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xd0\x10\x35\x99\x4d\x90\x71\xb3\x31\xa8\x61\x94\x11\xb0\x31\x90\x11\x99\x4d\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_52_consts_11_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x13\x15\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_11_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            (PyObject *)&_Py_SINGLETON(strings).ascii[101],
        },
    },
};
static
    struct _PyCode_DEF(46)
_collections_abc_toplevel_consts_52_consts_11_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_52_consts_11_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 635,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 446,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_11_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_11_consts_1_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_11_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x0d\x00\x00\x7d\x01\x7c\x01\x44\x00\x5d\x06\x00\x00\x7d\x02\x7c\x02\x96\x01\x97\x01\x01\x00\x8c\x08\x04\x00\x8c\x0f\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & _collections_abc_toplevel_consts_52_consts_11_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
_collections_abc_toplevel_consts_52_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__or__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[50];
    }
_collections_abc_toplevel_consts_52_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 49,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x18\xd4\x0f\x2a\xdc\x13\x21\xd0\x0c\x21\xd9\x10\x35\x98\x54\xa0\x35\x99\x4d\xd3\x10\x35\x88\x05\xd8\x0f\x13\xd7\x0f\x22\xd1\x0f\x22\xa0\x35\xd3\x0f\x29\xd0\x08\x29",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_chain = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "chain",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_other._ascii.ob_base,
            & const_str_chain._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(102)
_collections_abc_toplevel_consts_52_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 51,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_11_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 632,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 447,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__or__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_11_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x64\x01\x84\x00\x7c\x00\x7c\x01\x66\x02\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
_collections_abc_toplevel_consts_52_consts_12_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__sub__.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
_collections_abc_toplevel_consts_52_consts_12_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xf0\x00\x01\x23\x3a\xb1\x64\xa8\x55\xd8\x26\x2b\xb0\x35\xd1\x26\x38\xf4\x03\x00\x24\x29\xb1\x64\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_52_consts_12_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x10\x13\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_12_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            &_Py_ID(value),
            & const_str_other._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(42)
_collections_abc_toplevel_consts_52_consts_12_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 21,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_52_consts_12_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 645,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 448,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_12_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_12_consts_1_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_12_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x0a\x00\x00\x7d\x01\x7c\x01\x89\x02\x76\x01\x72\x04\x7c\x01\x96\x01\x97\x01\x01\x00\x8c\x0c\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & _collections_abc_toplevel_consts_52_consts_12_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Set._ascii.ob_base,
            & const_str_Iterable._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
            & const_str__from_iterable._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
_collections_abc_toplevel_consts_52_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__sub__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[77];
    }
_collections_abc_toplevel_consts_52_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 76,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x0f\x19\x98\x25\xa4\x13\xd4\x0f\x25\xdc\x13\x1d\x98\x65\xa4\x58\xd4\x13\x2e\xdc\x17\x25\xd0\x10\x25\xd8\x14\x18\xd7\x14\x27\xd1\x14\x27\xa8\x05\xd3\x14\x2e\x88\x45\xd8\x0f\x13\xd7\x0f\x22\xd1\x0f\x22\xf3\x00\x01\x23\x3a\xb1\x64\xf3\x00\x01\x23\x3a\xf3\x00\x01\x10\x3a\xf0\x00\x01\x09\x3a",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
_collections_abc_toplevel_consts_52_consts_12_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = " `",
};
static
    struct _PyCode_DEF(166)
_collections_abc_toplevel_consts_52_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 83,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_12_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 640,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 449,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & _collections_abc_toplevel_consts_52_consts_12_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__sub__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_12_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x01\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x27\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\xab\x01\x00\x00\x00\x00\x00\x00\x8a\x01\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x88\x01\x66\x01\x64\x01\x84\x08\x7c\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
_collections_abc_toplevel_consts_52_consts_13_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__rsub__.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
_collections_abc_toplevel_consts_52_consts_13_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xf0\x00\x01\x23\x39\xb1\x65\xa8\x55\xd8\x26\x2b\xb0\x34\xd1\x26\x37\xf4\x03\x00\x24\x29\xb1\x65\xf9",
};
static
    struct _PyCode_DEF(42)
_collections_abc_toplevel_consts_52_consts_13_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 21,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_52_consts_12_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 653,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 450,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_9_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_13_consts_1_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_13_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x0a\x00\x00\x7d\x01\x7c\x01\x89\x02\x76\x01\x72\x04\x7c\x01\x96\x01\x97\x01\x01\x00\x8c\x0c\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & _collections_abc_toplevel_consts_52_consts_13_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
_collections_abc_toplevel_consts_52_consts_13_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__rsub__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[77];
    }
_collections_abc_toplevel_consts_52_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 76,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x0f\x19\x98\x25\xa4\x13\xd4\x0f\x25\xdc\x13\x1d\x98\x65\xa4\x58\xd4\x13\x2e\xdc\x17\x25\xd0\x10\x25\xd8\x14\x18\xd7\x14\x27\xd1\x14\x27\xa8\x05\xd3\x14\x2e\x88\x45\xd8\x0f\x13\xd7\x0f\x22\xd1\x0f\x22\xf3\x00\x01\x23\x39\xb1\x65\xf3\x00\x01\x23\x39\xf3\x00\x01\x10\x39\xf0\x00\x01\x09\x39",
};
static
    struct _PyCode_DEF(166)
_collections_abc_toplevel_consts_52_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 83,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_13_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 648,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 451,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__rsub__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_13_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x27\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x89\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x89\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x88\x00\x66\x01\x64\x01\x84\x08\x7c\x01\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
_collections_abc_toplevel_consts_52_consts_14_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set.__xor__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[64];
    }
_collections_abc_toplevel_consts_52_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 63,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x13\xd4\x0f\x25\xdc\x13\x1d\x98\x65\xa4\x58\xd4\x13\x2e\xdc\x17\x25\xd0\x10\x25\xd8\x14\x18\xd7\x14\x27\xd1\x14\x27\xa8\x05\xd3\x14\x2e\x88\x45\xd8\x10\x14\x90\x75\x91\x0c\xa0\x15\xa8\x14\xa1\x1c\xd1\x0f\x2e\xd0\x08\x2e",
};
static
    struct _PyCode_DEF(134)
_collections_abc_toplevel_consts_52_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 67,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 656,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 452,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__xor__),
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_14_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x27\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x7c\x01\x7a\x0a\x00\x00\x7c\x01\x7c\x00\x7a\x0a\x00\x00\x7a\x07\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[556];
    }
_collections_abc_toplevel_consts_52_consts_15_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 555,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6d\x70\x75\x74\x65\x20\x74\x68\x65\x20\x68\x61\x73\x68\x20\x76\x61\x6c\x75\x65\x20\x6f\x66\x20\x61\x20\x73\x65\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4e\x6f\x74\x65\x20\x74\x68\x61\x74\x20\x77\x65\x20\x64\x6f\x6e\x27\x74\x20\x64\x65\x66\x69\x6e\x65\x20\x5f\x5f\x68\x61\x73\x68\x5f\x5f\x3a\x20\x6e\x6f\x74\x20\x61\x6c\x6c\x20\x73\x65\x74\x73\x20\x61\x72\x65\x20\x68\x61\x73\x68\x61\x62\x6c\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x42\x75\x74\x20\x69\x66\x20\x79\x6f\x75\x20\x64\x65\x66\x69\x6e\x65\x20\x61\x20\x68\x61\x73\x68\x61\x62\x6c\x65\x20\x73\x65\x74\x20\x74\x79\x70\x65\x2c\x20\x69\x74\x73\x20\x5f\x5f\x68\x61\x73\x68\x5f\x5f\x20\x73\x68\x6f\x75\x6c\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x63\x61\x6c\x6c\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x63\x6f\x6d\x70\x61\x74\x69\x62\x6c\x65\x20\x5f\x5f\x65\x71\x5f\x5f\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x41\x6c\x6c\x20\x73\x65\x74\x73\x20\x6f\x75\x67\x68\x74\x20\x74\x6f\x20\x63\x6f\x6d\x70\x61\x72\x65\x20\x65\x71\x75\x61\x6c\x20\x69\x66\x20\x74\x68\x65\x79\x20\x63\x6f\x6e\x74\x61\x69\x6e\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6c\x65\x6d\x65\x6e\x74\x73\x2c\x20\x72\x65\x67\x61\x72\x64\x6c\x65\x73\x73\x20\x6f\x66\x20\x68\x6f\x77\x20\x74\x68\x65\x79\x20\x61\x72\x65\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x65\x64\x2c\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x67\x61\x72\x64\x6c\x65\x73\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x6f\x72\x64\x65\x72\x20\x6f\x66\x20\x74\x68\x65\x20\x65\x6c\x65\x6d\x65\x6e\x74\x73\x3b\x20\x73\x6f\x20\x74\x68\x65\x72\x65\x27\x73\x20\x6e\x6f\x74\x20\x6d\x75\x63\x68\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x72\x65\x65\x64\x6f\x6d\x20\x66\x6f\x72\x20\x5f\x5f\x65\x71\x5f\x5f\x20\x6f\x72\x20\x5f\x5f\x68\x61\x73\x68\x5f\x5f\x2e\x20\x20\x57\x65\x20\x6d\x61\x74\x63\x68\x20\x74\x68\x65\x20\x61\x6c\x67\x6f\x72\x69\x74\x68\x6d\x20\x75\x73\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x62\x79\x20\x74\x68\x65\x20\x62\x75\x69\x6c\x74\x2d\x69\x6e\x20\x66\x72\x6f\x7a\x65\x6e\x73\x65\x74\x20\x74\x79\x70\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[3];
    }
const_int_1927868237 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 3),
    .ob_digit = { 28493, 26065, 1 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_1927868237 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 854126413, 1 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_89869747 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 19891, 2742 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_89869747 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 89869747 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[3];
    }
const_int_3644798167 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 3),
    .ob_digit = { 13527, 12926, 3 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_3644798167 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 423572695, 3 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_69069 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 3533, 2 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_69069 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 69069 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_907133923 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 17379, 27683 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_907133923 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 907133923 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_590923713 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 18369, 18033 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_590923713 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 590923713 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_52_consts_15_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & const_int_1927868237.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            & const_int_89869747.ob_base,
            & const_int_3644798167.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 11],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 25],
            & const_int_69069.ob_base,
            & const_int_907133923.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & const_int_590923713.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_maxsize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "maxsize",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_maxsize._ascii.ob_base,
            &_Py_ID(len),
            & const_str_hash._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__hash = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_hash",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
_collections_abc_toplevel_consts_52_consts_15_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set._hash",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[200];
    }
_collections_abc_toplevel_consts_52_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 199,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x1e\x00\x0f\x12\x8f\x6b\x89\x6b\x88\x03\xd8\x0f\x10\x90\x33\x89\x77\x98\x11\x89\x7b\x88\x04\xdc\x0c\x0f\x90\x04\x8b\x49\x88\x01\xd8\x0c\x16\x98\x21\x98\x61\x99\x25\xd1\x0c\x20\x88\x01\xd8\x08\x09\x88\x54\x89\x09\x88\x01\xdb\x11\x15\x88\x41\xdc\x11\x15\x90\x61\x93\x17\x88\x42\xd8\x0c\x0d\x90\x22\x98\x02\x98\x62\x99\x08\x91\x2f\xa0\x48\xd1\x12\x2c\xb0\x1a\xd1\x11\x3b\xd1\x0c\x3b\x88\x41\xd8\x0c\x0d\x90\x14\x89\x49\x89\x41\xf0\x07\x00\x12\x16\xf0\x08\x00\x09\x0a\x88\x61\x90\x32\x89\x67\x98\x21\x98\x72\x99\x27\xd1\x0d\x22\xd1\x08\x22\x88\x01\xd8\x0c\x0d\x90\x05\x89\x49\x98\x09\xd1\x0c\x21\x88\x01\xd8\x08\x09\x88\x54\x89\x09\x88\x01\xd8\x0b\x0c\x88\x73\x8a\x37\xd8\x0c\x0d\x90\x14\x98\x01\x91\x18\x89\x4d\x88\x41\xd8\x0b\x0c\x90\x02\x8a\x37\xd8\x10\x19\x88\x41\xd8\x0f\x10\x88\x08",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_MAX = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MAX",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_MASK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MASK",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_hx = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hx",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts_15_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_MAX._ascii.ob_base,
            & const_str_MASK._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[110],
            (PyObject *)&_Py_SINGLETON(strings).ascii[104],
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
            & const_str_hx._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(276)
_collections_abc_toplevel_consts_52_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 138,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts_15_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 665,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 453,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_15_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__hash._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_52_consts_15_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x64\x01\x7c\x01\x7a\x05\x00\x00\x64\x02\x7a\x00\x00\x00\x7d\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x64\x03\x7c\x03\x64\x02\x7a\x00\x00\x00\x7a\x05\x00\x00\x7d\x04\x7c\x04\x7c\x02\x7a\x0e\x00\x00\x7d\x04\x7c\x00\x44\x00\x5d\x23\x00\x00\x7d\x05\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x04\x7c\x06\x7c\x06\x64\x04\x7a\x03\x00\x00\x7a\x0c\x00\x00\x64\x05\x7a\x0c\x00\x00\x64\x06\x7a\x05\x00\x00\x7a\x19\x00\x00\x7d\x04\x7c\x04\x7c\x02\x7a\x0e\x00\x00\x7d\x04\x8c\x25\x04\x00\x7c\x04\x7c\x04\x64\x07\x7a\x09\x00\x00\x7c\x04\x64\x08\x7a\x09\x00\x00\x7a\x0c\x00\x00\x7a\x19\x00\x00\x7d\x04\x7c\x04\x64\x09\x7a\x05\x00\x00\x64\x0a\x7a\x00\x00\x00\x7d\x04\x7c\x04\x7c\x02\x7a\x0e\x00\x00\x7d\x04\x7c\x04\x7c\x01\x6b\x44\x00\x00\x72\x08\x7c\x04\x7c\x02\x64\x02\x7a\x00\x00\x00\x7a\x17\x00\x00\x7d\x04\x7c\x04\x64\x0b\x6b\x28\x00\x00\x72\x02\x64\x0c\x7d\x04\x7c\x04\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
_collections_abc_toplevel_consts_52_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            & const_str_Set._ascii.ob_base,
            & _collections_abc_toplevel_consts_52_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_52_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_7.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_8.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_9.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_10.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_11.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_12.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_13.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_14.ob_base.ob_base,
            & _collections_abc_toplevel_consts_52_consts_15.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
_collections_abc_toplevel_consts_52_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
            &_Py_ID(__le__),
            &_Py_ID(__lt__),
            &_Py_ID(__gt__),
            &_Py_ID(__ge__),
            &_Py_ID(__eq__),
            & const_str_classmethod._ascii.ob_base,
            & const_str__from_iterable._ascii.ob_base,
            &_Py_ID(__and__),
            &_Py_ID(__rand__),
            & const_str_isdisjoint._ascii.ob_base,
            &_Py_ID(__or__),
            &_Py_ID(__ror__),
            &_Py_ID(__sub__),
            &_Py_ID(__rsub__),
            &_Py_ID(__xor__),
            &_Py_ID(__rxor__),
            & const_str__hash._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[117];
    }
_collections_abc_toplevel_consts_52_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 116,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x08\x05\x08\xf0\x14\x00\x11\x13\x80\x49\xf2\x04\x08\x05\x14\xf2\x14\x03\x05\x3d\xf2\x0a\x03\x05\x3d\xf2\x0a\x08\x05\x14\xf2\x14\x03\x05\x3e\xf0\x0a\x00\x06\x11\xf1\x02\x06\x05\x17\xf3\x03\x00\x06\x11\xf0\x02\x06\x05\x17\xf2\x10\x03\x05\x4e\x01\xf0\x0a\x00\x10\x17\x80\x48\xf2\x04\x05\x05\x14\xf2\x0e\x04\x05\x2a\xf0\x0c\x00\x0f\x15\x80\x47\xf2\x04\x06\x05\x3a\xf2\x10\x06\x05\x39\xf2\x10\x05\x05\x2f\xf0\x0e\x00\x10\x17\x80\x48\xf3\x04\x1f\x05\x11",
};
static
    struct _PyCode_DEF(120)
_collections_abc_toplevel_consts_52 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 60,
    },
    .co_consts = & _collections_abc_toplevel_consts_52_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_52_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 561,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 454,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Set._ascii.ob_base,
    .co_qualname = & const_str_Set._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_52_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x64\x07\x84\x00\x5a\x09\x65\x0a\x64\x08\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x0b\x64\x09\x84\x00\x5a\x0c\x65\x0c\x5a\x0d\x64\x0a\x84\x00\x5a\x0e\x64\x0b\x84\x00\x5a\x0f\x65\x0f\x5a\x10\x64\x0c\x84\x00\x5a\x11\x64\x0d\x84\x00\x5a\x12\x64\x0e\x84\x00\x5a\x13\x65\x13\x5a\x14\x64\x0f\x84\x00\x5a\x15\x79\x10",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[392];
    }
_collections_abc_toplevel_consts_54_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 391,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x6d\x75\x74\x61\x62\x6c\x65\x20\x73\x65\x74\x20\x69\x73\x20\x61\x20\x66\x69\x6e\x69\x74\x65\x2c\x20\x69\x74\x65\x72\x61\x62\x6c\x65\x20\x63\x6f\x6e\x74\x61\x69\x6e\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x63\x6c\x61\x73\x73\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x63\x6f\x6e\x63\x72\x65\x74\x65\x20\x67\x65\x6e\x65\x72\x69\x63\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x73\x20\x6f\x66\x20\x61\x6c\x6c\x0a\x20\x20\x20\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x65\x78\x63\x65\x70\x74\x20\x66\x6f\x72\x20\x5f\x5f\x63\x6f\x6e\x74\x61\x69\x6e\x73\x5f\x5f\x2c\x20\x5f\x5f\x69\x74\x65\x72\x5f\x5f\x2c\x20\x5f\x5f\x6c\x65\x6e\x5f\x5f\x2c\x0a\x20\x20\x20\x20\x61\x64\x64\x28\x29\x2c\x20\x61\x6e\x64\x20\x64\x69\x73\x63\x61\x72\x64\x28\x29\x2e\x0a\x0a\x20\x20\x20\x20\x54\x6f\x20\x6f\x76\x65\x72\x72\x69\x64\x65\x20\x74\x68\x65\x20\x63\x6f\x6d\x70\x61\x72\x69\x73\x6f\x6e\x73\x20\x28\x70\x72\x65\x73\x75\x6d\x61\x62\x6c\x79\x20\x66\x6f\x72\x20\x73\x70\x65\x65\x64\x2c\x20\x61\x73\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x73\x65\x6d\x61\x6e\x74\x69\x63\x73\x20\x61\x72\x65\x20\x66\x69\x78\x65\x64\x29\x2c\x20\x61\x6c\x6c\x20\x79\x6f\x75\x20\x68\x61\x76\x65\x20\x74\x6f\x20\x64\x6f\x20\x69\x73\x20\x72\x65\x64\x65\x66\x69\x6e\x65\x20\x5f\x5f\x6c\x65\x5f\x5f\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x74\x68\x65\x6e\x20\x74\x68\x65\x20\x6f\x74\x68\x65\x72\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x20\x77\x69\x6c\x6c\x20\x61\x75\x74\x6f\x6d\x61\x74\x69\x63\x61\x6c\x6c\x79\x20\x66\x6f\x6c\x6c\x6f\x77\x20\x73\x75\x69\x74\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
_collections_abc_toplevel_consts_54_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Add an element.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_54_consts_3_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_54_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.add",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_54_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & _collections_abc_toplevel_consts_54_consts_3_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 716,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 455,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(add),
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_3_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[57];
    }
_collections_abc_toplevel_consts_54_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 56,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Remove an element.  Do not raise an exception if absent.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_54_consts_4_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
_collections_abc_toplevel_consts_54_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.discard",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_54_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & _collections_abc_toplevel_consts_54_consts_4_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 721,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 456,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(discard),
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_4_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
_collections_abc_toplevel_consts_54_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Remove an element. If not a member, raise a KeyError.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_54_consts_5_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_KeyError._ascii.ob_base,
            &_Py_ID(discard),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_54_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.remove",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
_collections_abc_toplevel_consts_54_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x10\x98\x04\xd1\x0b\x1c\xdc\x12\x1a\x98\x35\x93\x2f\xd0\x0c\x21\xd8\x08\x0c\x8f\x0c\x89\x0c\x90\x55\xd5\x08\x1b",
};
static
    struct _PyCode_DEF(68)
_collections_abc_toplevel_consts_54_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 34,
    },
    .co_consts = & _collections_abc_toplevel_consts_54_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 726,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 457,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_remove._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x76\x01\x72\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[51];
    }
_collections_abc_toplevel_consts_54_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 50,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the popped value.  Raise KeyError if empty.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_54_consts_6_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(iter),
            &_Py_ID(next),
            & const_str_StopIteration._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            &_Py_ID(discard),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_54_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.pop",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[70];
    }
_collections_abc_toplevel_consts_54_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 69,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0d\x11\x90\x24\x8b\x5a\x88\x02\xf0\x02\x03\x09\x25\xdc\x14\x18\x98\x12\x93\x48\x88\x45\xf0\x06\x00\x09\x0d\x8f\x0c\x89\x0c\x90\x55\xd4\x08\x1b\xd8\x0f\x14\x88\x0c\xf8\xf4\x07\x00\x10\x1d\xf2\x00\x01\x09\x25\xdc\x12\x1a\xa0\x04\xd0\x0c\x24\xf0\x03\x01\x09\x25\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_54_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x8d\x0b\x2b\x00\xab\x11\x3c\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_it._ascii.ob_base,
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(126)
_collections_abc_toplevel_consts_54_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 63,
    },
    .co_consts = & _collections_abc_toplevel_consts_54_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_54_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 732,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 458,
    .co_localsplusnames = & _collections_abc_toplevel_consts_54_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_pop._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x08\x01\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
_collections_abc_toplevel_consts_54_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "This is slow (creates N new iterators!) but effective.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_54_consts_7_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_pop._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
_collections_abc_toplevel_consts_54_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.clear",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
_collections_abc_toplevel_consts_54_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x04\x09\x11\xd8\x12\x16\xd8\x10\x14\x97\x08\x91\x08\x94\x0a\xf0\x03\x00\x13\x17\xf8\xe4\x0f\x17\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_54_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x12\x14\x00\x94\x09\x20\x03\x9f\x01\x20\x03",
};
static
    struct _PyCode_DEF(70)
_collections_abc_toplevel_consts_54_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & _collections_abc_toplevel_consts_54_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_54_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 742,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 459,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(clear),
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x09\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x11\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(add),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
_collections_abc_toplevel_consts_54_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.__ior__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
_collections_abc_toplevel_consts_54_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdb\x15\x17\x88\x45\xd8\x0c\x10\x8f\x48\x89\x48\x90\x55\x8d\x4f\xf0\x03\x00\x16\x18\xe0\x0f\x13\x88\x0b",
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_54_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 750,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 460,
    .co_localsplusnames = & _collections_abc_toplevel_consts_54_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__ior__),
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_8_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x44\x00\x5d\x13\x00\x00\x7d\x02\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x15\x04\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(discard),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_54_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.__iand__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
_collections_abc_toplevel_consts_54_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x16\x1a\x98\x52\x94\x69\x88\x45\xd8\x0c\x10\x8f\x4c\x89\x4c\x98\x15\xd5\x0c\x1f\xf0\x03\x00\x17\x20\xe0\x0f\x13\x88\x0b",
};
static
    struct _PyCode_DEF(60)
_collections_abc_toplevel_consts_54_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 755,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 461,
    .co_localsplusnames = & _collections_abc_toplevel_consts_54_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iand__),
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_9_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x7c\x01\x7a\x0a\x00\x00\x44\x00\x5d\x13\x00\x00\x7d\x02\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x15\x04\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(clear),
            &_Py_ID(isinstance),
            & const_str_Set._ascii.ob_base,
            & const_str__from_iterable._ascii.ob_base,
            &_Py_ID(discard),
            &_Py_ID(add),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_54_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.__ixor__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[101];
    }
_collections_abc_toplevel_consts_54_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 100,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0d\x90\x14\x89\x3a\xd8\x0c\x10\x8f\x4a\x89\x4a\x8c\x4c\xf0\x12\x00\x10\x14\x88\x0b\xf4\x0f\x00\x14\x1e\x98\x62\xa4\x23\xd4\x13\x26\xd8\x15\x19\xd7\x15\x28\xd1\x15\x28\xa8\x12\xd3\x15\x2c\x90\x02\xdb\x19\x1b\x90\x05\xd8\x13\x18\x98\x44\x91\x3d\xd8\x14\x18\x97\x4c\x91\x4c\xa0\x15\xd5\x14\x27\xe0\x14\x18\x97\x48\x91\x48\x98\x55\x95\x4f\xf0\x09\x00\x1a\x1c\xf0\x0a\x00\x10\x14\x88\x0b",
};
static
    struct _PyCode_DEF(208)
_collections_abc_toplevel_consts_54_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 104,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 760,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 462,
    .co_localsplusnames = & _collections_abc_toplevel_consts_54_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__ixor__),
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_10_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x75\x00\x72\x12\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x53\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x11\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x44\x00\x5d\x29\x00\x00\x7d\x02\x7c\x02\x7c\x00\x76\x00\x72\x12\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x19\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x2b\x04\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(clear),
            &_Py_ID(discard),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_54_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSet.__isub__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[55];
    }
_collections_abc_toplevel_consts_54_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 54,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0d\x90\x14\x89\x3a\xd8\x0c\x10\x8f\x4a\x89\x4a\x8c\x4c\xf0\x08\x00\x10\x14\x88\x0b\xf3\x05\x00\x1a\x1c\x90\x05\xd8\x10\x14\x97\x0c\x91\x0c\x98\x55\xd5\x10\x23\xf0\x03\x00\x1a\x1c\xe0\x0f\x13\x88\x0b",
};
static
    struct _PyCode_DEF(98)
_collections_abc_toplevel_consts_54_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 773,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 463,
    .co_localsplusnames = & _collections_abc_toplevel_consts_54_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__isub__),
    .co_qualname = & _collections_abc_toplevel_consts_54_consts_11_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x75\x00\x72\x12\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x53\x00\x7c\x01\x44\x00\x5d\x13\x00\x00\x7d\x02\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x15\x04\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
_collections_abc_toplevel_consts_54_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_MutableSet._ascii.ob_base,
            & _collections_abc_toplevel_consts_54_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_54_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_7.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_8.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_9.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_10.ob_base.ob_base,
            & _collections_abc_toplevel_consts_54_consts_11.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
_collections_abc_toplevel_consts_54_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(add),
            &_Py_ID(discard),
            & const_str_remove._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            &_Py_ID(clear),
            &_Py_ID(__ior__),
            &_Py_ID(__iand__),
            &_Py_ID(__ixor__),
            &_Py_ID(__isub__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[88];
    }
_collections_abc_toplevel_consts_54_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 87,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x09\x05\x08\xf0\x16\x00\x11\x13\x80\x49\xe0\x05\x13\xf1\x02\x02\x05\x22\xf3\x03\x00\x06\x14\xf0\x02\x02\x05\x22\xf0\x08\x00\x06\x14\xf1\x02\x02\x05\x22\xf3\x03\x00\x06\x14\xf0\x02\x02\x05\x22\xf2\x08\x04\x05\x1c\xf2\x0c\x08\x05\x15\xf2\x14\x06\x05\x11\xf2\x10\x03\x05\x14\xf2\x0a\x03\x05\x14\xf2\x0a\x0b\x05\x14\xf3\x1a\x06\x05\x14",
};
static
    struct _PyCode_DEF(94)
_collections_abc_toplevel_consts_54 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 47,
    },
    .co_consts = & _collections_abc_toplevel_consts_54_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_54_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 702,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 464,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_MutableSet._ascii.ob_base,
    .co_qualname = & const_str_MutableSet._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_54_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x65\x05\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x05\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x64\x05\x84\x00\x5a\x08\x64\x06\x84\x00\x5a\x09\x64\x07\x84\x00\x5a\x0a\x64\x08\x84\x00\x5a\x0b\x64\x09\x84\x00\x5a\x0c\x64\x0a\x84\x00\x5a\x0d\x64\x0b\x84\x00\x5a\x0e\x79\x0c",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[199];
    }
_collections_abc_toplevel_consts_56_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 198,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x4d\x61\x70\x70\x69\x6e\x67\x20\x69\x73\x20\x61\x20\x67\x65\x6e\x65\x72\x69\x63\x20\x63\x6f\x6e\x74\x61\x69\x6e\x65\x72\x20\x66\x6f\x72\x20\x61\x73\x73\x6f\x63\x69\x61\x74\x69\x6e\x67\x20\x6b\x65\x79\x2f\x76\x61\x6c\x75\x65\x0a\x20\x20\x20\x20\x70\x61\x69\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x63\x6c\x61\x73\x73\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x63\x6f\x6e\x63\x72\x65\x74\x65\x20\x67\x65\x6e\x65\x72\x69\x63\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x73\x20\x6f\x66\x20\x61\x6c\x6c\x0a\x20\x20\x20\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x65\x78\x63\x65\x70\x74\x20\x66\x6f\x72\x20\x5f\x5f\x67\x65\x74\x69\x74\x65\x6d\x5f\x5f\x2c\x20\x5f\x5f\x69\x74\x65\x72\x5f\x5f\x2c\x20\x61\x6e\x64\x20\x5f\x5f\x6c\x65\x6e\x5f\x5f\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_56_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping.__getitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
_collections_abc_toplevel_consts_56_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0e\x16\x88\x0e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(key),
        },
    },
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_56_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 800,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 465,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getitem__),
    .co_qualname = & _collections_abc_toplevel_consts_56_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[61];
    }
_collections_abc_toplevel_consts_56_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 60,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_56_consts_6_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
_collections_abc_toplevel_consts_56_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping.get",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[38];
    }
_collections_abc_toplevel_consts_56_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 37,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x09\x1b\xd8\x13\x17\x98\x03\x91\x39\xd0\x0c\x1c\xf8\xdc\x0f\x17\xf2\x00\x01\x09\x1b\xd8\x13\x1a\x8a\x4e\xf0\x03\x01\x09\x1b\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_56_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x04\x07\x00\x87\x0b\x15\x03\x94\x01\x15\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(key),
            &_Py_ID(default),
        },
    },
};
static
    struct _PyCode_DEF(48)
_collections_abc_toplevel_consts_56_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 24,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_56_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 804,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 466,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(get),
    .co_qualname = & _collections_abc_toplevel_consts_56_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x53\x00\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x7c\x02\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            Py_True,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
_collections_abc_toplevel_consts_56_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping.__contains__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
_collections_abc_toplevel_consts_56_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x05\x09\x18\xd8\x0c\x10\x90\x13\x8a\x49\xf0\x08\x00\x14\x18\xf8\xf4\x07\x00\x10\x18\xf2\x00\x01\x09\x19\xd9\x13\x18\xf0\x03\x01\x09\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_56_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x05\x08\x00\x88\x09\x14\x03\x93\x01\x14\x03",
};
static
    struct _PyCode_DEF(46)
_collections_abc_toplevel_consts_56_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_56_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 811,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 467,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__contains__),
    .co_qualname = & _collections_abc_toplevel_consts_56_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x01\x00\x79\x01\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[59];
    }
_collections_abc_toplevel_consts_56_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 58,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "D.keys() -> a set-like object providing a view on D's keys",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_56_consts_8_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_KeysView._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
_collections_abc_toplevel_consts_56_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping.keys",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_56_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x17\x98\x04\x8b\x7e\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(24)
_collections_abc_toplevel_consts_56_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts_8_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 819,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 468,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(keys),
    .co_qualname = & _collections_abc_toplevel_consts_56_consts_8_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[61];
    }
_collections_abc_toplevel_consts_56_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 60,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "D.items() -> a set-like object providing a view on D's items",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_56_consts_9_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_ItemsView._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
_collections_abc_toplevel_consts_56_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping.items",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_56_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x18\x98\x14\x8b\x7f\xd0\x08\x1e",
};
static
    struct _PyCode_DEF(24)
_collections_abc_toplevel_consts_56_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts_9_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 823,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 469,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(items),
    .co_qualname = & _collections_abc_toplevel_consts_56_consts_9_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
_collections_abc_toplevel_consts_56_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "D.values() -> an object providing a view on D's values",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_56_consts_10_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_ValuesView._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_56_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping.values",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[14];
    }
_collections_abc_toplevel_consts_56_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 13,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x19\x98\x24\xd3\x0f\x1f\xd0\x08\x1f",
};
static
    struct _PyCode_DEF(24)
_collections_abc_toplevel_consts_56_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts_10_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 827,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 470,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(values),
    .co_qualname = & _collections_abc_toplevel_consts_56_consts_10_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Mapping._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
            &_Py_ID(dict),
            &_Py_ID(items),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_56_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Mapping.__eq__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
_collections_abc_toplevel_consts_56_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x17\xd4\x0f\x29\xdc\x13\x21\xd0\x0c\x21\xdc\x0f\x13\x90\x44\x97\x4a\x91\x4a\x93\x4c\xd3\x0f\x21\xa4\x54\xa8\x25\xaf\x2b\xa9\x2b\xab\x2d\xd3\x25\x38\xd1\x0f\x38\xd0\x08\x38",
};
static
    struct _PyCode_DEF(148)
_collections_abc_toplevel_consts_56_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 74,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 831,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 471,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__eq__),
    .co_qualname = & _collections_abc_toplevel_consts_56_consts_11_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
_collections_abc_toplevel_consts_56_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_Mapping._ascii.ob_base,
            & _collections_abc_toplevel_consts_56_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 64],
            & _collections_abc_toplevel_consts_56_consts_4.ob_base.ob_base,
            Py_None,
            & _collections_abc_toplevel_consts_56_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_56_consts_7.ob_base.ob_base,
            & _collections_abc_toplevel_consts_56_consts_8.ob_base.ob_base,
            & _collections_abc_toplevel_consts_56_consts_9.ob_base.ob_base,
            & _collections_abc_toplevel_consts_56_consts_10.ob_base.ob_base,
            & _collections_abc_toplevel_consts_56_consts_11.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
_collections_abc_toplevel_consts_56_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
            &_Py_ID(__abc_tpflags__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__getitem__),
            &_Py_ID(get),
            &_Py_ID(__contains__),
            &_Py_ID(keys),
            &_Py_ID(items),
            &_Py_ID(values),
            &_Py_ID(__eq__),
            &_Py_ID(__reversed__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[77];
    }
_collections_abc_toplevel_consts_56_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 76,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x05\x05\x08\xf0\x0e\x00\x11\x13\x80\x49\xf0\x06\x00\x17\x1d\x80\x4f\xe0\x05\x13\xf1\x02\x01\x05\x17\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x17\xf3\x06\x05\x05\x1b\xf2\x0e\x06\x05\x18\xf2\x10\x02\x05\x1e\xf2\x08\x02\x05\x1f\xf2\x08\x02\x05\x20\xf2\x08\x03\x05\x39\xf0\x0a\x00\x14\x18\x81\x4c",
};
static
    struct _PyCode_DEF(82)
_collections_abc_toplevel_consts_56 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 41,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 787,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 472,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Mapping._ascii.ob_base,
    .co_qualname = & const_str_Mapping._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x03\x5a\x05\x65\x06\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x64\x0c\x64\x06\x84\x01\x5a\x08\x64\x07\x84\x00\x5a\x09\x64\x08\x84\x00\x5a\x0a\x64\x09\x84\x00\x5a\x0b\x64\x0a\x84\x00\x5a\x0c\x64\x0b\x84\x00\x5a\x0d\x64\x05\x5a\x0e\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__mapping = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_mapping",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_58_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__mapping._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
_collections_abc_toplevel_consts_58_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MappingView.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
_collections_abc_toplevel_consts_58_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x18\x1f\x88\x04\x8d\x0d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_58_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(mapping),
        },
    },
};
static
    struct _PyCode_DEF(18)
_collections_abc_toplevel_consts_58_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 9,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_1._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 845,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 473,
    .co_localsplusnames = & _collections_abc_toplevel_consts_58_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & _collections_abc_toplevel_consts_58_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_58_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_58_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(len),
            & const_str__mapping._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_58_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MappingView.__len__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
_collections_abc_toplevel_consts_58_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x12\x90\x34\x97\x3d\x91\x3d\xd3\x0f\x21\xd0\x08\x21",
};
static
    struct _PyCode_DEF(44)
_collections_abc_toplevel_consts_58_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 848,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 474,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__len__),
    .co_qualname = & _collections_abc_toplevel_consts_58_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_58_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
_collections_abc_toplevel_consts_58_consts_4_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "{0.__class__.__name__}({0._mapping!r})",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_58_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & _collections_abc_toplevel_consts_58_consts_4_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_58_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(format),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
_collections_abc_toplevel_consts_58_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MappingView.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
_collections_abc_toplevel_consts_58_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x37\xd7\x0f\x3e\xd1\x0f\x3e\xb8\x74\xd3\x0f\x44\xd0\x08\x44",
};
static
    struct _PyCode_DEF(36)
_collections_abc_toplevel_consts_58_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 18,
    },
    .co_consts = & _collections_abc_toplevel_consts_58_consts_4_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 851,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 475,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & _collections_abc_toplevel_consts_58_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_58_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_58_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_MappingView._ascii.ob_base,
            & _collections_abc_toplevel_consts_58_consts_1._object.ob_base.ob_base,
            & _collections_abc_toplevel_consts_58_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_58_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_58_consts_4.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
_collections_abc_toplevel_consts_58_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            &_Py_ID(__init__),
            &_Py_ID(__len__),
            &_Py_ID(__repr__),
            & const_str_classmethod._ascii.ob_base,
            & const_str_GenericAlias._ascii.ob_base,
            &_Py_ID(__class_getitem__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
_collections_abc_toplevel_consts_58_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x1b\x80\x49\xf2\x04\x01\x05\x20\xf2\x06\x01\x05\x22\xf2\x06\x01\x05\x45\x01\xf1\x06\x00\x19\x24\xa0\x4c\xd3\x18\x31\xd1\x04\x15",
};
static
    struct _PyCode_DEF(50)
_collections_abc_toplevel_consts_58 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & _collections_abc_toplevel_consts_58_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 841,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 476,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_MappingView._ascii.ob_base,
    .co_qualname = & const_str_MappingView._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_58_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x02\x00\x65\x07\x65\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_60_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_set._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_60_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "KeysView._from_iterable",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
_collections_abc_toplevel_consts_60_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x12\x90\x32\x8b\x77\x88\x0e",
};
static
    struct _PyCode_DEF(24)
_collections_abc_toplevel_consts_60_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_60_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 861,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 477,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__from_iterable._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_60_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_60_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
_collections_abc_toplevel_consts_60_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "KeysView.__contains__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
_collections_abc_toplevel_consts_60_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x12\x90\x64\x97\x6d\x91\x6d\xd0\x0f\x23\xd0\x08\x23",
};
static
    struct _PyCode_DEF(30)
_collections_abc_toplevel_consts_60_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_1._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 865,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 478,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__contains__),
    .co_qualname = & _collections_abc_toplevel_consts_60_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_60_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_60_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "KeysView.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
_collections_abc_toplevel_consts_60_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xd8\x13\x17\x97\x3d\x91\x3d\xd7\x08\x20\xd2\x08\x20\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_60_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x10\x1a\x01\x92\x01\x18\x04\x93\x06\x1a\x01",
};
static
    struct _PyCode_DEF(56)
_collections_abc_toplevel_consts_60_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_1._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_60_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 868,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 479,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & _collections_abc_toplevel_consts_60_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_60_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x45\x00\x64\x00\x7b\x03\x00\x00\x96\x02\x97\x02\x86\x05\x05\x00\x01\x00\x79\x00\x37\x00\x8c\x05\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_60_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_KeysView._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_60_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_60_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_60_consts_4.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_collections_abc_toplevel_consts_60_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            & const_str_classmethod._ascii.ob_base,
            & const_str__from_iterable._ascii.ob_base,
            &_Py_ID(__contains__),
            &_Py_ID(__iter__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
_collections_abc_toplevel_consts_60_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x10\xf1\x02\x01\x05\x17\xf3\x03\x00\x06\x11\xf0\x02\x01\x05\x17\xf2\x06\x01\x05\x24\xf3\x06\x01\x05\x21",
};
static
    struct _PyCode_DEF(44)
_collections_abc_toplevel_consts_60 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & _collections_abc_toplevel_consts_60_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_60_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 857,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 480,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_KeysView._ascii.ob_base,
    .co_qualname = & const_str_KeysView._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_60_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x64\x03\x84\x00\x5a\x06\x64\x04\x84\x00\x5a\x07\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
_collections_abc_toplevel_consts_62_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ItemsView._from_iterable",
};
static
    struct _PyCode_DEF(24)
_collections_abc_toplevel_consts_62_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_60_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 879,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 481,
    .co_localsplusnames = & _collections_abc_toplevel_consts_52_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__from_iterable._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_62_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_60_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_62_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__mapping._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_62_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ItemsView.__contains__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[72];
    }
_collections_abc_toplevel_consts_62_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 71,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x15\x19\x89\x0a\x88\x03\x88\x55\xf0\x02\x05\x09\x2c\xd8\x10\x14\x97\x0d\x91\x0d\x98\x63\xd1\x10\x22\x88\x41\xf0\x08\x00\x14\x15\x98\x05\x90\x3a\xd2\x13\x2b\xa0\x11\xa0\x65\xa1\x1a\xd0\x0c\x2b\xf8\xf4\x07\x00\x10\x18\xf2\x00\x01\x09\x19\xd9\x13\x18\xf0\x03\x01\x09\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_62_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x87\x0f\x21\x00\xa1\x09\x2d\x03\xac\x01\x2d\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_62_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(item),
            &_Py_ID(key),
            &_Py_ID(value),
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(96)
_collections_abc_toplevel_consts_62_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 48,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_30_consts_4_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_62_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_62_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 883,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 482,
    .co_localsplusnames = & _collections_abc_toplevel_consts_62_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__contains__),
    .co_qualname = & _collections_abc_toplevel_consts_62_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_62_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x5c\x02\x00\x00\x7d\x02\x7d\x03\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x19\x00\x00\x00\x7d\x04\x7c\x04\x7c\x03\x75\x00\x78\x01\x73\x05\x01\x00\x7c\x04\x7c\x03\x6b\x28\x00\x00\x53\x00\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
_collections_abc_toplevel_consts_62_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ItemsView.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[41];
    }
_collections_abc_toplevel_consts_62_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 40,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xd8\x13\x17\x97\x3d\x94\x3d\x88\x43\xd8\x13\x16\x98\x04\x9f\x0d\x99\x0d\xa0\x63\xd1\x18\x2a\xd0\x12\x2b\xd3\x0c\x2b\xf1\x03\x00\x14\x21\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_62_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x26\x28\x01",
};
static
    struct _PyCode_DEF(84)
_collections_abc_toplevel_consts_62_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 42,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_1._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_62_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 892,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 483,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & _collections_abc_toplevel_consts_62_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_62_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x15\x00\x00\x7d\x01\x7c\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x66\x02\x96\x01\x97\x01\x01\x00\x8c\x17\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_62_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_ItemsView._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_62_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_62_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_62_consts_4.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
_collections_abc_toplevel_consts_62_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xe0\x05\x10\xf1\x02\x01\x05\x17\xf3\x03\x00\x06\x11\xf0\x02\x01\x05\x17\xf2\x06\x07\x05\x2c\xf3\x12\x02\x05\x2c",
};
static
    struct _PyCode_DEF(44)
_collections_abc_toplevel_consts_62 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & _collections_abc_toplevel_consts_62_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_60_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 875,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 484,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_ItemsView._ascii.ob_base,
    .co_qualname = & const_str_ItemsView._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_62_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x64\x03\x84\x00\x5a\x06\x64\x04\x84\x00\x5a\x07\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_64_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ValuesView.__contains__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[52];
    }
_collections_abc_toplevel_consts_64_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 51,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x13\x17\x97\x3d\x94\x3d\x88\x43\xd8\x10\x14\x97\x0d\x91\x0d\x98\x63\xd1\x10\x22\x88\x41\xd8\x0f\x10\x90\x45\x89\x7a\x98\x51\xa0\x25\x9b\x5a\xd9\x17\x1b\xf0\x07\x00\x14\x21\xf0\x08\x00\x10\x15",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_64_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(value),
            &_Py_ID(key),
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(90)
_collections_abc_toplevel_consts_64_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 45,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_1._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 904,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 485,
    .co_localsplusnames = & _collections_abc_toplevel_consts_64_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__contains__),
    .co_qualname = & _collections_abc_toplevel_consts_64_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_64_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x1c\x00\x00\x7d\x02\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x19\x00\x00\x00\x7d\x03\x7c\x03\x7c\x01\x75\x00\x73\x06\x7c\x03\x7c\x01\x6b\x28\x00\x00\x73\x01\x8c\x1c\x01\x00\x79\x01\x04\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_64_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ValuesView.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
_collections_abc_toplevel_consts_64_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xd8\x13\x17\x97\x3d\x94\x3d\x88\x43\xd8\x12\x16\x97\x2d\x91\x2d\xa0\x03\xd1\x12\x24\xd3\x0c\x24\xf1\x03\x00\x14\x21\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_64_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x24\x26\x01",
};
static
    struct _PyCode_DEF(80)
_collections_abc_toplevel_consts_64_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 40,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_58_consts_1._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_64_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 911,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 486,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & _collections_abc_toplevel_consts_64_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_64_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x13\x00\x00\x7d\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x15\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_64_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_ValuesView._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_64_consts_2.ob_base.ob_base,
            & _collections_abc_toplevel_consts_64_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_64_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__slots__),
            &_Py_ID(__contains__),
            &_Py_ID(__iter__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
_collections_abc_toplevel_consts_64_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe0\x10\x12\x80\x49\xf2\x04\x05\x05\x15\xf3\x0e\x02\x05\x25",
};
static
    struct _PyCode_DEF(28)
_collections_abc_toplevel_consts_64 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & _collections_abc_toplevel_consts_64_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_64_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 900,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 487,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_ValuesView._ascii.ob_base,
    .co_qualname = & const_str_ValuesView._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_64_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[236];
    }
_collections_abc_toplevel_consts_66_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 235,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x4d\x75\x74\x61\x62\x6c\x65\x4d\x61\x70\x70\x69\x6e\x67\x20\x69\x73\x20\x61\x20\x67\x65\x6e\x65\x72\x69\x63\x20\x63\x6f\x6e\x74\x61\x69\x6e\x65\x72\x20\x66\x6f\x72\x20\x61\x73\x73\x6f\x63\x69\x61\x74\x69\x6e\x67\x0a\x20\x20\x20\x20\x6b\x65\x79\x2f\x76\x61\x6c\x75\x65\x20\x70\x61\x69\x72\x73\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x63\x6c\x61\x73\x73\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x63\x6f\x6e\x63\x72\x65\x74\x65\x20\x67\x65\x6e\x65\x72\x69\x63\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x61\x74\x69\x6f\x6e\x73\x20\x6f\x66\x20\x61\x6c\x6c\x0a\x20\x20\x20\x20\x6d\x65\x74\x68\x6f\x64\x73\x20\x65\x78\x63\x65\x70\x74\x20\x66\x6f\x72\x20\x5f\x5f\x67\x65\x74\x69\x74\x65\x6d\x5f\x5f\x2c\x20\x5f\x5f\x73\x65\x74\x69\x74\x65\x6d\x5f\x5f\x2c\x20\x5f\x5f\x64\x65\x6c\x69\x74\x65\x6d\x5f\x5f\x2c\x0a\x20\x20\x20\x20\x5f\x5f\x69\x74\x65\x72\x5f\x5f\x2c\x20\x61\x6e\x64\x20\x5f\x5f\x6c\x65\x6e\x5f\x5f\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_66_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping.__setitem__",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(key),
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_66_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 930,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 488,
    .co_localsplusnames = & _collections_abc_toplevel_consts_66_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__setitem__),
    .co_qualname = & _collections_abc_toplevel_consts_66_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_66_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping.__delitem__",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_66_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 934,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 489,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__delitem__),
    .co_qualname = & _collections_abc_toplevel_consts_66_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_56_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[170];
    }
_collections_abc_toplevel_consts_66_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 169,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x2e\x70\x6f\x70\x28\x6b\x5b\x2c\x64\x5d\x29\x20\x2d\x3e\x20\x76\x2c\x20\x72\x65\x6d\x6f\x76\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6b\x65\x79\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x63\x6f\x72\x72\x65\x73\x70\x6f\x6e\x64\x69\x6e\x67\x20\x76\x61\x6c\x75\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x6b\x65\x79\x20\x69\x73\x20\x6e\x6f\x74\x20\x66\x6f\x75\x6e\x64\x2c\x20\x64\x20\x69\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x69\x66\x20\x67\x69\x76\x65\x6e\x2c\x20\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x4b\x65\x79\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_66_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str__MutableMapping__marker = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_MutableMapping__marker",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_KeyError._ascii.ob_base,
            & const_str__MutableMapping__marker._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
_collections_abc_toplevel_consts_66_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping.pop",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[68];
    }
_collections_abc_toplevel_consts_66_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 67,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x08\x09\x19\xd8\x14\x18\x98\x13\x91\x49\x88\x45\xf0\x0c\x00\x11\x15\x90\x53\x90\x09\xd8\x13\x18\x88\x4c\xf8\xf4\x0d\x00\x10\x18\xf2\x00\x03\x09\x1b\xd8\x0f\x16\x98\x24\x9f\x2d\x99\x2d\xd1\x0f\x27\xd8\x10\x15\xd8\x13\x1a\x8a\x4e\xf0\x07\x03\x09\x1b\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_66_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x05\x0c\x00\x8c\x1a\x29\x03\xa8\x01\x29\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(key),
            &_Py_ID(default),
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(88)
_collections_abc_toplevel_consts_66_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 44,
    },
    .co_consts = & _collections_abc_toplevel_consts_66_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_66_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_66_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 940,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 490,
    .co_localsplusnames = & _collections_abc_toplevel_consts_66_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_pop._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_66_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_66_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x7d\x03\x7c\x00\x7c\x01\x3d\x00\x7c\x03\x53\x00\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x14\x01\x00\x7c\x02\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x01\x82\x00\x7c\x02\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[132];
    }
_collections_abc_toplevel_consts_66_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 131,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x2e\x70\x6f\x70\x69\x74\x65\x6d\x28\x29\x20\x2d\x3e\x20\x28\x6b\x2c\x20\x76\x29\x2c\x20\x72\x65\x6d\x6f\x76\x65\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x73\x6f\x6d\x65\x20\x28\x6b\x65\x79\x2c\x20\x76\x61\x6c\x75\x65\x29\x20\x70\x61\x69\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x61\x73\x20\x61\x20\x32\x2d\x74\x75\x70\x6c\x65\x3b\x20\x62\x75\x74\x20\x72\x61\x69\x73\x65\x20\x4b\x65\x79\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x44\x20\x69\x73\x20\x65\x6d\x70\x74\x79\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_66_consts_6_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(next),
            &_Py_ID(iter),
            & const_str_StopIteration._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_popitem = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "popitem",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_66_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping.popitem",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[75];
    }
_collections_abc_toplevel_consts_66_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 74,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x03\x09\x25\xdc\x12\x16\x94\x74\x98\x44\x93\x7a\xd3\x12\x22\x88\x43\xf0\x06\x00\x11\x15\x90\x53\x91\x09\x88\x05\xd8\x0c\x10\x90\x13\x88\x49\xd8\x0f\x12\x90\x45\x88\x7a\xd0\x08\x19\xf8\xf4\x09\x00\x10\x1d\xf2\x00\x01\x09\x25\xdc\x12\x1a\xa0\x04\xd0\x0c\x24\xf0\x03\x01\x09\x25\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_66_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x14\x22\x00\xa2\x11\x33\x03",
};
static
    struct _PyCode_DEF(108)
_collections_abc_toplevel_consts_66_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 54,
    },
    .co_consts = & _collections_abc_toplevel_consts_66_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_66_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_66_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 954,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 491,
    .co_localsplusnames = & _collections_abc_toplevel_consts_66_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_popitem._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_66_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_66_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x7c\x00\x7c\x01\x3d\x00\x7c\x01\x7c\x02\x66\x02\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x08\x01\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
_collections_abc_toplevel_consts_66_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "D.clear() -> None.  Remove all items from D.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_66_consts_7_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_popitem._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
_collections_abc_toplevel_consts_66_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping.clear",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
_collections_abc_toplevel_consts_66_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x04\x09\x11\xd8\x12\x16\xd8\x10\x14\x97\x0c\x91\x0c\x94\x0e\xf0\x03\x00\x13\x17\xf8\xe4\x0f\x17\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct _PyCode_DEF(70)
_collections_abc_toplevel_consts_66_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & _collections_abc_toplevel_consts_66_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_66_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_54_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 966,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 492,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(clear),
    .co_qualname = & _collections_abc_toplevel_consts_66_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_66_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x09\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x11\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[339];
    }
_collections_abc_toplevel_consts_66_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 338,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x44\x2e\x75\x70\x64\x61\x74\x65\x28\x5b\x45\x2c\x20\x5d\x2a\x2a\x46\x29\x20\x2d\x3e\x20\x4e\x6f\x6e\x65\x2e\x20\x20\x55\x70\x64\x61\x74\x65\x20\x44\x20\x66\x72\x6f\x6d\x20\x6d\x61\x70\x70\x69\x6e\x67\x2f\x69\x74\x65\x72\x61\x62\x6c\x65\x20\x45\x20\x61\x6e\x64\x20\x46\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x45\x20\x70\x72\x65\x73\x65\x6e\x74\x20\x61\x6e\x64\x20\x68\x61\x73\x20\x61\x20\x2e\x6b\x65\x79\x73\x28\x29\x20\x6d\x65\x74\x68\x6f\x64\x2c\x20\x64\x6f\x65\x73\x3a\x20\x20\x20\x20\x20\x66\x6f\x72\x20\x6b\x20\x69\x6e\x20\x45\x2e\x6b\x65\x79\x73\x28\x29\x3a\x20\x44\x5b\x6b\x5d\x20\x3d\x20\x45\x5b\x6b\x5d\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x45\x20\x70\x72\x65\x73\x65\x6e\x74\x20\x61\x6e\x64\x20\x6c\x61\x63\x6b\x73\x20\x2e\x6b\x65\x79\x73\x28\x29\x20\x6d\x65\x74\x68\x6f\x64\x2c\x20\x64\x6f\x65\x73\x3a\x20\x20\x20\x20\x20\x66\x6f\x72\x20\x28\x6b\x2c\x20\x76\x29\x20\x69\x6e\x20\x45\x3a\x20\x44\x5b\x6b\x5d\x20\x3d\x20\x76\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x49\x6e\x20\x65\x69\x74\x68\x65\x72\x20\x63\x61\x73\x65\x2c\x20\x74\x68\x69\x73\x20\x69\x73\x20\x66\x6f\x6c\x6c\x6f\x77\x65\x64\x20\x62\x79\x3a\x20\x66\x6f\x72\x20\x6b\x2c\x20\x76\x20\x69\x6e\x20\x46\x2e\x69\x74\x65\x6d\x73\x28\x29\x3a\x20\x44\x5b\x6b\x5d\x20\x3d\x20\x76\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_66_consts_8_consts_0._ascii.ob_base,
            &_Py_ID(keys),
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Mapping._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            &_Py_ID(keys),
            &_Py_ID(items),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
_collections_abc_toplevel_consts_66_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping.update",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[131];
    }
_collections_abc_toplevel_consts_66_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 130,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x0c\x16\x90\x65\x9c\x57\xd4\x0b\x25\xdb\x17\x1c\x90\x03\xd8\x1c\x21\xa0\x23\x99\x4a\x90\x04\x90\x53\x92\x09\xf1\x03\x00\x18\x1d\xe4\x0d\x14\x90\x55\x98\x46\xd4\x0d\x23\xd8\x17\x1c\x97\x7a\x91\x7a\x96\x7c\x90\x03\xd8\x1c\x21\xa0\x23\x99\x4a\x90\x04\x90\x53\x92\x09\xf1\x03\x00\x18\x24\xf3\x06\x00\x1f\x24\x91\x0a\x90\x03\x90\x55\xd8\x1c\x21\x90\x04\x90\x53\x92\x09\xf0\x03\x00\x1f\x24\xe0\x1a\x1e\x9f\x2a\x99\x2a\x9e\x2c\x89\x4a\x88\x43\x90\x15\xd8\x18\x1d\x88\x44\x90\x13\x8a\x49\xf1\x03\x00\x1b\x27",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_other._ascii.ob_base,
            & const_str_kwds._ascii.ob_base,
            &_Py_ID(key),
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(240)
_collections_abc_toplevel_consts_66_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 120,
    },
    .co_consts = & _collections_abc_toplevel_consts_66_consts_8_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_66_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 11,
    .co_argcount = 2,
    .co_posonlyargcount = 2,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 974,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 493,
    .co_localsplusnames = & _collections_abc_toplevel_consts_66_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_update._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_66_consts_8_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_66_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x10\x7c\x01\x44\x00\x5d\x0a\x00\x00\x7d\x03\x7c\x01\x7c\x03\x19\x00\x00\x00\x7c\x00\x7c\x03\x3c\x00\x00\x00\x8c\x0c\x04\x00\x6e\x39\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x72\x1e\x7c\x01\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x0a\x00\x00\x7d\x03\x7c\x01\x7c\x03\x19\x00\x00\x00\x7c\x00\x7c\x03\x3c\x00\x00\x00\x8c\x0c\x04\x00\x6e\x0f\x7c\x01\x44\x00\x5d\x0a\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x7c\x04\x7c\x00\x7c\x03\x3c\x00\x00\x00\x8c\x0c\x04\x00\x7c\x02\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x0a\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x7c\x04\x7c\x00\x7c\x03\x3c\x00\x00\x00\x8c\x0c\x04\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[65];
    }
_collections_abc_toplevel_consts_66_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 64,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_66_consts_10_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
_collections_abc_toplevel_consts_66_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableMapping.setdefault",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[47];
    }
_collections_abc_toplevel_consts_66_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 46,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x09\x20\xd8\x13\x17\x98\x03\x91\x39\xd0\x0c\x1c\xf8\xdc\x0f\x17\xf2\x00\x01\x09\x20\xd8\x18\x1f\x88\x44\x90\x13\x8a\x49\xd8\x0f\x16\x88\x0e\xf0\x05\x01\x09\x20\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
_collections_abc_toplevel_consts_66_consts_10_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x04\x07\x00\x87\x0e\x19\x03\x98\x01\x19\x03",
};
static
    struct _PyCode_DEF(56)
_collections_abc_toplevel_consts_66_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & _collections_abc_toplevel_consts_66_consts_10_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_56_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_66_consts_10_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 992,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 494,
    .co_localsplusnames = & _collections_abc_toplevel_consts_56_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_setdefault._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_66_consts_10_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_66_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x53\x00\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x09\x01\x00\x7c\x02\x7c\x00\x7c\x01\x3c\x00\x00\x00\x59\x00\x7c\x02\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts_11 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            (PyObject *)& _Py_SINGLETON(tuple_empty),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
_collections_abc_toplevel_consts_66_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_MutableMapping._ascii.ob_base,
            & _collections_abc_toplevel_consts_66_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_66_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_66_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_66_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_66_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_66_consts_7.ob_base.ob_base,
            & _collections_abc_toplevel_consts_66_consts_8.ob_base.ob_base,
            Py_None,
            & _collections_abc_toplevel_consts_66_consts_10.ob_base.ob_base,
            & _collections_abc_toplevel_consts_66_consts_11._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
_collections_abc_toplevel_consts_66_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__setitem__),
            &_Py_ID(__delitem__),
            &_Py_ID(object),
            & const_str__MutableMapping__marker._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            & const_str_popitem._ascii.ob_base,
            &_Py_ID(clear),
            & const_str_update._ascii.ob_base,
            & const_str_setdefault._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[90];
    }
_collections_abc_toplevel_consts_66_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 89,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x06\x05\x08\xf0\x10\x00\x11\x13\x80\x49\xe0\x05\x13\xf1\x02\x01\x05\x17\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x17\xf0\x06\x00\x06\x14\xf1\x02\x01\x05\x17\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x17\xf1\x06\x00\x10\x16\x8b\x78\x80\x48\xe0\x1f\x27\xf3\x00\x0c\x05\x19\xf2\x1c\x0a\x05\x1a\xf2\x18\x06\x05\x11\xf3\x10\x10\x05\x1e\xf4\x24\x06\x05\x17",
};
static
    struct _PyCode_DEF(104)
_collections_abc_toplevel_consts_66 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 52,
    },
    .co_consts = & _collections_abc_toplevel_consts_66_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_66_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 919,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 495,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_MutableMapping._ascii.ob_base,
    .co_qualname = & const_str_MutableMapping._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_66_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x65\x05\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x05\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x02\x00\x65\x08\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x09\x65\x09\x66\x01\x64\x05\x84\x01\x5a\x0a\x64\x06\x84\x00\x5a\x0b\x64\x07\x84\x00\x5a\x0c\x64\x0b\x64\x08\x84\x01\x5a\x0d\x64\x0c\x64\x0a\x84\x01\x5a\x0e\x79\x09",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[139];
    }
_collections_abc_toplevel_consts_68_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 138,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x6c\x6c\x20\x74\x68\x65\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x20\x6f\x6e\x20\x61\x20\x72\x65\x61\x64\x2d\x6f\x6e\x6c\x79\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x43\x6f\x6e\x63\x72\x65\x74\x65\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x65\x73\x20\x6d\x75\x73\x74\x20\x6f\x76\x65\x72\x72\x69\x64\x65\x20\x5f\x5f\x6e\x65\x77\x5f\x5f\x20\x6f\x72\x20\x5f\x5f\x69\x6e\x69\x74\x5f\x5f\x2c\x0a\x20\x20\x20\x20\x5f\x5f\x67\x65\x74\x69\x74\x65\x6d\x5f\x5f\x2c\x20\x61\x6e\x64\x20\x5f\x5f\x6c\x65\x6e\x5f\x5f\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_IndexError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
_collections_abc_toplevel_consts_68_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence.__getitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_68_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0e\x18\xd0\x08\x18",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_68_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1018,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 496,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getitem__),
    .co_qualname = & _collections_abc_toplevel_consts_68_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_collections_abc_toplevel_consts_68_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
_collections_abc_toplevel_consts_68_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xd8\x0c\x0d\x88\x01\xf0\x02\x06\x09\x13\xd8\x12\x16\xd8\x14\x18\x98\x11\x91\x47\x90\x01\xd8\x16\x17\x92\x07\xd8\x10\x11\x90\x51\x91\x06\x90\x01\xf0\x07\x00\x13\x17\xf8\xf4\x08\x00\x10\x1a\xf2\x00\x01\x09\x13\xd9\x0c\x12\xf0\x03\x01\x09\x13\xfc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
_collections_abc_toplevel_consts_68_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x03\x25\x01\x86\x10\x16\x00\x96\x09\x22\x03\x9f\x02\x25\x01\xa1\x01\x22\x03\xa2\x03\x25\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(78)
_collections_abc_toplevel_consts_68_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & _collections_abc_toplevel_consts_68_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_68_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1022,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 497,
    .co_localsplusnames = & _collections_abc_toplevel_consts_68_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & _collections_abc_toplevel_consts_68_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x64\x01\x7d\x01\x09\x00\x09\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x7c\x02\x96\x02\x97\x01\x01\x00\x7c\x01\x64\x02\x7a\x0d\x00\x00\x7d\x01\x8c\x0f\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
_collections_abc_toplevel_consts_68_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence.__contains__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
_collections_abc_toplevel_consts_68_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdb\x11\x15\x88\x41\xd8\x0f\x10\x90\x45\x89\x7a\x98\x51\xa0\x25\x9b\x5a\xd9\x17\x1b\xf0\x05\x00\x12\x16\xf0\x06\x00\x10\x15",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(value),
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(40)
_collections_abc_toplevel_consts_68_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & _collections_abc_toplevel_consts_56_consts_7_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1032,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 498,
    .co_localsplusnames = & _collections_abc_toplevel_consts_68_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__contains__),
    .co_qualname = & _collections_abc_toplevel_consts_68_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x44\x00\x5d\x0d\x00\x00\x7d\x02\x7c\x02\x7c\x01\x75\x00\x73\x06\x7c\x02\x7c\x01\x6b\x28\x00\x00\x73\x01\x8c\x0d\x01\x00\x79\x01\x04\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_range = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "range",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(reversed),
            & const_str_range._ascii.ob_base,
            &_Py_ID(len),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
_collections_abc_toplevel_consts_68_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence.__reversed__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
_collections_abc_toplevel_consts_68_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xdc\x11\x19\x9c\x25\xa4\x03\xa0\x44\xa3\x09\xd3\x1a\x2a\xd6\x11\x2b\x88\x41\xd8\x12\x16\x90\x71\x91\x27\x8b\x4d\xf1\x03\x00\x12\x2c\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
_collections_abc_toplevel_consts_68_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x2b\x2d\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
        },
    },
};
static
    struct _PyCode_DEF(94)
_collections_abc_toplevel_consts_68_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 47,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_68_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1038,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 499,
    .co_localsplusnames = & _collections_abc_toplevel_consts_68_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reversed__),
    .co_qualname = & _collections_abc_toplevel_consts_68_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x09\x00\x00\x7d\x01\x7c\x00\x7c\x01\x19\x00\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x0b\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[231];
    }
_collections_abc_toplevel_consts_68_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 230,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x2e\x69\x6e\x64\x65\x78\x28\x76\x61\x6c\x75\x65\x2c\x20\x5b\x73\x74\x61\x72\x74\x2c\x20\x5b\x73\x74\x6f\x70\x5d\x5d\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x20\x2d\x2d\x20\x72\x65\x74\x75\x72\x6e\x20\x66\x69\x72\x73\x74\x20\x69\x6e\x64\x65\x78\x20\x6f\x66\x20\x76\x61\x6c\x75\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x73\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x74\x68\x65\x20\x76\x61\x6c\x75\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x70\x72\x65\x73\x65\x6e\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x53\x75\x70\x70\x6f\x72\x74\x69\x6e\x67\x20\x73\x74\x61\x72\x74\x20\x61\x6e\x64\x20\x73\x74\x6f\x70\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x69\x73\x20\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x2c\x20\x62\x75\x74\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x63\x6f\x6d\x6d\x65\x6e\x64\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_68_consts_9_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_max._ascii.ob_base,
            &_Py_ID(len),
            & const_str_IndexError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_68_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence.index",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[172];
    }
_collections_abc_toplevel_consts_68_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 171,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0e\x00\x0c\x11\xd0\x0b\x1c\xa0\x15\xa8\x11\xa2\x19\xdc\x14\x17\x9c\x03\x98\x44\x9b\x09\xa0\x45\xd1\x18\x29\xa8\x31\xd3\x14\x2d\x88\x45\xd8\x0b\x0f\xd0\x0b\x1b\xa0\x04\xa0\x71\xa2\x08\xd8\x0c\x10\x94\x43\x98\x04\x93\x49\xd1\x0c\x1d\x88\x44\xe0\x0c\x11\x88\x01\xd8\x0e\x12\x88\x6c\x98\x61\xa0\x24\x9a\x68\xf0\x02\x03\x0d\x16\xd8\x14\x18\x98\x11\x91\x47\x90\x01\xf0\x06\x00\x10\x11\x90\x45\x89\x7a\x98\x51\xa0\x25\x9a\x5a\xd8\x17\x18\x90\x08\xd8\x0c\x0d\x90\x11\x89\x46\x88\x41\xf0\x0f\x00\x0f\x13\x89\x6c\x98\x61\xa0\x24\x9b\x68\xf4\x10\x00\x0f\x19\xd0\x08\x18\xf8\xf4\x0b\x00\x14\x1e\xf2\x00\x01\x0d\x16\xd8\x10\x15\xf4\x08\x00\x0f\x19\xd0\x08\x18\xf0\x0b\x01\x0d\x16\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
_collections_abc_toplevel_consts_68_consts_9_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\xbf\x05\x41\x23\x00\xc1\x23\x09\x41\x34\x03\xc1\x33\x01\x41\x34\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_stop = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "stop",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(value),
            &_Py_ID(start),
            & const_str_stop._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(238)
_collections_abc_toplevel_consts_68_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 119,
    },
    .co_consts = & _collections_abc_toplevel_consts_68_consts_9_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_68_consts_9_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1042,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 500,
    .co_localsplusnames = & _collections_abc_toplevel_consts_68_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_index._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_68_consts_9_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x81\x1d\x7c\x02\x64\x01\x6b\x02\x00\x00\x72\x18\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x7a\x00\x00\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x03\x81\x13\x7c\x03\x64\x01\x6b\x02\x00\x00\x72\x0e\x7c\x03\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x0d\x00\x00\x7d\x03\x7c\x02\x7d\x04\x7c\x03\x81\x05\x7c\x04\x7c\x03\x6b\x02\x00\x00\x72\x1f\x09\x00\x7c\x00\x7c\x04\x19\x00\x00\x00\x7d\x05\x7c\x05\x7c\x01\x75\x00\x73\x05\x7c\x05\x7c\x01\x6b\x28\x00\x00\x72\x02\x7c\x04\x53\x00\x7c\x04\x64\x02\x7a\x0d\x00\x00\x7d\x04\x7c\x03\x80\x01\x8c\x19\x7c\x04\x7c\x03\x6b\x02\x00\x00\x72\x01\x8c\x1f\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x08\x01\x00\x59\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[67];
    }
_collections_abc_toplevel_consts_68_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 66,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S.count(value) -> integer -- return number of occurrences of value",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
_collections_abc_toplevel_consts_68_consts_10_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence.count.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
_collections_abc_toplevel_consts_68_consts_10_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xd0\x12\x3f\x99\x64\x98\x11\xa0\x61\xa8\x35\xa1\x6a\xb0\x41\xb8\x15\xb3\x4a\x94\x31\x99\x64\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_collections_abc_toplevel_consts_68_consts_10_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x0e\x19\x01\x92\x07\x19\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_10_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(54)
_collections_abc_toplevel_consts_68_consts_10_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & importlib__bootstrap_external_toplevel_consts_6_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & _collections_abc_toplevel_consts_68_consts_10_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1067,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 501,
    .co_localsplusnames = & _collections_abc_toplevel_consts_68_consts_10_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & _collections_abc_toplevel_consts_68_consts_10_consts_1_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_10_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x10\x00\x00\x7d\x01\x7c\x01\x89\x02\x75\x00\x73\x06\x7c\x01\x89\x02\x6b\x28\x00\x00\x73\x01\x8c\x0d\x64\x00\x96\x01\x97\x01\x01\x00\x8c\x12\x04\x00\x79\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_68_consts_10_consts_0._ascii.ob_base,
            & _collections_abc_toplevel_consts_68_consts_10_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_sum = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sum",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_sum._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
_collections_abc_toplevel_consts_68_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Sequence.count",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
_collections_abc_toplevel_consts_68_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xe4\x0f\x12\xd3\x12\x3f\x99\x64\xd3\x12\x3f\xd3\x0f\x3f\xd0\x08\x3f",
};
static
    struct _PyCode_DEF(44)
_collections_abc_toplevel_consts_68_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & _collections_abc_toplevel_consts_68_consts_10_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1065,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 502,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & _collections_abc_toplevel_consts_52_consts_12_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(count),
    .co_qualname = & _collections_abc_toplevel_consts_68_consts_10_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x01\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x88\x01\x66\x01\x64\x01\x84\x08\x7c\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts_11 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
_collections_abc_toplevel_consts_68_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str_Sequence._ascii.ob_base,
            & _collections_abc_toplevel_consts_68_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 32],
            & _collections_abc_toplevel_consts_68_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_68_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_68_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_68_consts_7.ob_base.ob_base,
            Py_None,
            & _collections_abc_toplevel_consts_68_consts_9.ob_base.ob_base,
            & _collections_abc_toplevel_consts_68_consts_10.ob_base.ob_base,
            & _collections_abc_toplevel_consts_68_consts_11._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
_collections_abc_toplevel_consts_68_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
            &_Py_ID(__abc_tpflags__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__getitem__),
            &_Py_ID(__iter__),
            &_Py_ID(__contains__),
            &_Py_ID(__reversed__),
            & const_str_index._ascii.ob_base,
            &_Py_ID(count),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[66];
    }
_collections_abc_toplevel_consts_68_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 65,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x04\x05\x08\xf0\x0c\x00\x11\x13\x80\x49\xf0\x06\x00\x17\x1d\x80\x4f\xe0\x05\x13\xf1\x02\x01\x05\x19\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x19\xf2\x06\x08\x05\x13\xf2\x14\x04\x05\x15\xf2\x0c\x02\x05\x1a\xf3\x08\x15\x05\x19\xf3\x2e\x02\x05\x40\x01",
};
static
    struct _PyCode_DEF(72)
_collections_abc_toplevel_consts_68 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 36,
    },
    .co_consts = & _collections_abc_toplevel_consts_68_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1006,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 503,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_Sequence._ascii.ob_base,
    .co_qualname = & const_str_Sequence._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x03\x5a\x05\x65\x06\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x64\x05\x84\x00\x5a\x08\x64\x06\x84\x00\x5a\x09\x64\x07\x84\x00\x5a\x0a\x64\x0b\x64\x09\x84\x01\x5a\x0b\x64\x0a\x84\x00\x5a\x0c\x79\x08",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str__DeprecateByteStringMeta = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DeprecateByteStringMeta",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
_collections_abc_toplevel_consts_70_consts_1_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "collections.abc.ByteString",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_1_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 14],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_1_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_remove._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            Py_None,
            & const_str_ByteString._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & _collections_abc_toplevel_consts_70_consts_1_consts_3._ascii.ob_base,
            & _collections_abc_toplevel_consts_70_consts_1_consts_4._object.ob_base.ob_base,
            & _collections_abc_toplevel_consts_70_consts_1_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__deprecated = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_deprecated",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(warnings),
            & const_str__deprecated._ascii.ob_base,
            & const_str_super._ascii.ob_base,
            &_Py_ID(__new__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
_collections_abc_toplevel_consts_70_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DeprecateByteStringMeta.__new__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[68];
    }
_collections_abc_toplevel_consts_70_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 67,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x0b\x0f\x90\x3c\xd2\x0b\x1f\xdb\x0c\x1b\xe0\x0c\x14\xd7\x0c\x20\xd1\x0c\x20\xd8\x10\x2c\xd8\x17\x1e\xf0\x05\x00\x0d\x21\xf4\x00\x03\x0d\x0e\xf4\x08\x00\x10\x15\x89\x77\x89\x7f\x98\x73\xa0\x44\xa8\x25\xb0\x19\xd1\x0f\x45\xb8\x66\xd1\x0f\x45\xd0\x08\x45",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            &_Py_ID(name),
            & const_str_bases._ascii.ob_base,
            & const_str_namespace._ascii.ob_base,
            & const_str_kwargs._ascii.ob_base,
            &_Py_ID(warnings),
            &_Py_ID(__class__),
        },
    },
};
static
    struct _PyCode_DEF(98)
_collections_abc_toplevel_consts_70_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & _collections_abc_toplevel_consts_70_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_70_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 11,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1075,
    .co_nlocalsplus = 7,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 504,
    .co_localsplusnames = & _collections_abc_toplevel_consts_70_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & abc_toplevel_consts_10_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__new__),
    .co_qualname = & _collections_abc_toplevel_consts_70_consts_1_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_70_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x7c\x01\x64\x01\x6b\x37\x00\x00\x72\x17\x64\x02\x64\x00\x6c\x00\x7d\x05\x7c\x05\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x64\x04\xac\x05\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x89\x06\x7c\x00\x8d\x0c\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7c\x03\x66\x04\x69\x00\x7c\x04\xa4\x01\x8e\x01\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & _collections_abc_toplevel_consts_70_consts_1_consts_3._ascii.ob_base,
            & _collections_abc_toplevel_consts_70_consts_1_consts_4._object.ob_base.ob_base,
            & _collections_abc_toplevel_consts_70_consts_1_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(warnings),
            & const_str__deprecated._ascii.ob_base,
            & const_str_super._ascii.ob_base,
            &_Py_ID(__instancecheck__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[43];
    }
_collections_abc_toplevel_consts_70_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 42,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_DeprecateByteStringMeta.__instancecheck__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[50];
    }
_collections_abc_toplevel_consts_70_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 49,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdb\x08\x17\xe0\x08\x10\xd7\x08\x1c\xd1\x08\x1c\xd8\x0c\x28\xd8\x13\x1a\xf0\x05\x00\x09\x1d\xf4\x00\x03\x09\x0a\xf4\x08\x00\x10\x15\x89\x77\xd1\x0f\x28\xa8\x18\xd3\x0f\x32\xd0\x08\x32",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_instance._ascii.ob_base,
            &_Py_ID(warnings),
            &_Py_ID(__class__),
        },
    },
};
static
    struct _PyCode_DEF(80)
_collections_abc_toplevel_consts_70_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 40,
    },
    .co_consts = & _collections_abc_toplevel_consts_70_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_70_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1085,
    .co_nlocalsplus = 4,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 505,
    .co_localsplusnames = & _collections_abc_toplevel_consts_70_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__instancecheck__),
    .co_qualname = & _collections_abc_toplevel_consts_70_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_70_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x64\x01\x64\x00\x6c\x00\x7d\x02\x7c\x02\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\xac\x04\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x89\x03\x7c\x00\x8d\x0d\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_70_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__DeprecateByteStringMeta._ascii.ob_base,
            & _collections_abc_toplevel_consts_70_consts_1.ob_base.ob_base,
            & _collections_abc_toplevel_consts_70_consts_2.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_collections_abc_toplevel_consts_70_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__new__),
            &_Py_ID(__instancecheck__),
            &_Py_ID(__classcell__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
_collections_abc_toplevel_consts_70_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x84\x00\xf4\x02\x08\x05\x46\x01\xf7\x14\x07\x05\x33\xf0\x00\x07\x05\x33",
};
static
    struct _PyCode_DEF(40)
_collections_abc_toplevel_consts_70 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & _collections_abc_toplevel_consts_70_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_70_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1074,
    .co_nlocalsplus = 1,
    .co_nlocals = 0,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 506,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[64]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__DeprecateByteStringMeta._ascii.ob_base,
    .co_qualname = & const_str__DeprecateByteStringMeta._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_70_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x88\x00\x66\x01\x64\x01\x84\x08\x5a\x03\x88\x00\x66\x01\x64\x02\x84\x08\x5a\x04\x88\x00\x78\x01\x5a\x05\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[78];
    }
_collections_abc_toplevel_consts_72_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 77,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x68\x69\x73\x20\x75\x6e\x69\x66\x69\x65\x73\x20\x62\x79\x74\x65\x73\x20\x61\x6e\x64\x20\x62\x79\x74\x65\x61\x72\x72\x61\x79\x2e\x0a\x0a\x20\x20\x20\x20\x58\x58\x58\x20\x53\x68\x6f\x75\x6c\x64\x20\x61\x64\x64\x20\x61\x6c\x6c\x20\x74\x68\x65\x69\x72\x20\x6d\x65\x74\x68\x6f\x64\x73\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_72_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_ByteString._ascii.ob_base,
            & _collections_abc_toplevel_consts_72_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            Py_None,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
_collections_abc_toplevel_consts_72_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x03\x05\x08\xf0\x0a\x00\x11\x13\x81\x49",
};
static
    struct _PyCode_DEF(20)
_collections_abc_toplevel_consts_72 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 10,
    },
    .co_consts = & _collections_abc_toplevel_consts_72_consts._object.ob_base.ob_base,
    .co_names = & abc_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1094,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 507,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_ByteString._ascii.ob_base,
    .co_qualname = & const_str_ByteString._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_72_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[175];
    }
_collections_abc_toplevel_consts_74_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 174,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x6c\x6c\x20\x74\x68\x65\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x20\x6f\x6e\x20\x61\x20\x72\x65\x61\x64\x2d\x77\x72\x69\x74\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x43\x6f\x6e\x63\x72\x65\x74\x65\x20\x73\x75\x62\x63\x6c\x61\x73\x73\x65\x73\x20\x6d\x75\x73\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x20\x5f\x5f\x6e\x65\x77\x5f\x5f\x20\x6f\x72\x20\x5f\x5f\x69\x6e\x69\x74\x5f\x5f\x2c\x0a\x20\x20\x20\x20\x5f\x5f\x67\x65\x74\x69\x74\x65\x6d\x5f\x5f\x2c\x20\x5f\x5f\x73\x65\x74\x69\x74\x65\x6d\x5f\x5f\x2c\x20\x5f\x5f\x64\x65\x6c\x69\x74\x65\x6d\x5f\x5f\x2c\x20\x5f\x5f\x6c\x65\x6e\x5f\x5f\x2c\x20\x61\x6e\x64\x20\x69\x6e\x73\x65\x72\x74\x28\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
_collections_abc_toplevel_consts_74_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.__setitem__",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_index._ascii.ob_base,
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_74_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1115,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 508,
    .co_localsplusnames = & _collections_abc_toplevel_consts_74_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__setitem__),
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
_collections_abc_toplevel_consts_74_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.__delitem__",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_74_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1119,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 509,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_66_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__delitem__),
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_68_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[52];
    }
_collections_abc_toplevel_consts_74_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 51,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S.insert(index, value) -- insert value before index",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_74_consts_5_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_74_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.insert",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
_collections_abc_toplevel_consts_74_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x0f\x19\xd0\x08\x18",
};
static
    struct _PyCode_DEF(14)
_collections_abc_toplevel_consts_74_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_68_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1123,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 510,
    .co_localsplusnames = & _collections_abc_toplevel_consts_74_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_insert._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[59];
    }
_collections_abc_toplevel_consts_74_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 58,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S.append(value) -- append value to the end of the sequence",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_74_consts_6_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_insert._ascii.ob_base,
            &_Py_ID(len),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_74_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.append",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
_collections_abc_toplevel_consts_74_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x08\x0c\x8f\x0b\x89\x0b\x94\x43\x98\x04\x93\x49\x98\x75\xd5\x08\x25",
};
static
    struct _PyCode_DEF(58)
_collections_abc_toplevel_consts_74_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 29,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_74_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1128,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 511,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(append),
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
_collections_abc_toplevel_consts_74_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S.clear() -> None -- remove all items from S",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_74_consts_7_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_pop._ascii.ob_base,
            & const_str_IndexError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
_collections_abc_toplevel_consts_74_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.clear",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
_collections_abc_toplevel_consts_74_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x04\x09\x11\xd8\x12\x16\xd8\x10\x14\x97\x08\x91\x08\x94\x0a\xf0\x03\x00\x13\x17\xf8\xe4\x0f\x19\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct _PyCode_DEF(70)
_collections_abc_toplevel_consts_74_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 35,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_74_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_54_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1132,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 512,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(clear),
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_7_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x09\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x11\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
_collections_abc_toplevel_consts_74_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S.reverse() -- reverse *IN PLACE*",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_74_consts_8_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(len),
            & const_str_range._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
_collections_abc_toplevel_consts_74_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.reverse",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[74];
    }
_collections_abc_toplevel_consts_74_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 73,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0c\x0f\x90\x04\x8b\x49\x88\x01\xdc\x11\x16\x90\x71\x98\x21\x91\x74\x96\x1b\x88\x41\xd8\x23\x27\xa8\x01\xa8\x21\xa9\x03\xa8\x41\xa9\x05\xa1\x3b\xb0\x04\xb0\x51\xb1\x07\xd0\x0c\x20\x88\x44\x90\x11\x89\x47\x90\x54\x98\x21\x98\x41\x99\x23\x98\x61\x99\x25\x92\x5b\xf1\x03\x00\x12\x1d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            (PyObject *)&_Py_SINGLETON(strings).ascii[110],
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
        },
    },
};
static
    struct _PyCode_DEF(122)
_collections_abc_toplevel_consts_74_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 61,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts_8_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_74_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1140,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 513,
    .co_localsplusnames = & _collections_abc_toplevel_consts_74_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(reverse),
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_8_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\x7a\x02\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x1f\x00\x00\x7d\x02\x7c\x00\x7c\x01\x7c\x02\x7a\x0a\x00\x00\x64\x02\x7a\x0a\x00\x00\x19\x00\x00\x00\x7c\x00\x7c\x02\x19\x00\x00\x00\x63\x02\x7c\x00\x7c\x02\x3c\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7a\x0a\x00\x00\x64\x02\x7a\x0a\x00\x00\x3c\x00\x00\x00\x8c\x21\x04\x00\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[78];
    }
_collections_abc_toplevel_consts_74_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 77,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S.extend(iterable) -- extend sequence by appending elements from the iterable",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_74_consts_9_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_list._ascii.ob_base,
            &_Py_ID(append),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_74_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.extend",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
_collections_abc_toplevel_consts_74_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x11\x90\x54\x89\x3e\xdc\x15\x19\x98\x26\x93\x5c\x88\x46\xdb\x11\x17\x88\x41\xd8\x0c\x10\x8f\x4b\x89\x4b\x98\x01\x8d\x4e\xf1\x03\x00\x12\x18",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(values),
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(82)
_collections_abc_toplevel_consts_74_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 41,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts_9_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_74_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1146,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 514,
    .co_localsplusnames = & _collections_abc_toplevel_consts_74_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(extend),
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_9_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x75\x00\x72\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x44\x00\x5d\x13\x00\x00\x7d\x02\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x15\x04\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[154];
    }
_collections_abc_toplevel_consts_74_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 153,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x2e\x70\x6f\x70\x28\x5b\x69\x6e\x64\x65\x78\x5d\x29\x20\x2d\x3e\x20\x69\x74\x65\x6d\x20\x2d\x2d\x20\x72\x65\x6d\x6f\x76\x65\x20\x61\x6e\x64\x20\x72\x65\x74\x75\x72\x6e\x20\x69\x74\x65\x6d\x20\x61\x74\x20\x69\x6e\x64\x65\x78\x20\x28\x64\x65\x66\x61\x75\x6c\x74\x20\x6c\x61\x73\x74\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x20\x49\x6e\x64\x65\x78\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x6c\x69\x73\x74\x20\x69\x73\x20\x65\x6d\x70\x74\x79\x20\x6f\x72\x20\x69\x6e\x64\x65\x78\x20\x69\x73\x20\x6f\x75\x74\x20\x6f\x66\x20\x72\x61\x6e\x67\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_74_consts_10_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
_collections_abc_toplevel_consts_74_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.pop",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
_collections_abc_toplevel_consts_74_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x0d\x11\x90\x15\x89\x4b\x88\x01\xd8\x0c\x10\x90\x15\x88\x4b\xd8\x0f\x10\x88\x08",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_10_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_index._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[118],
        },
    },
};
static
    struct _PyCode_DEF(22)
_collections_abc_toplevel_consts_74_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 11,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts_10_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1153,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 515,
    .co_localsplusnames = & _collections_abc_toplevel_consts_74_consts_10_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_pop._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_10_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x7c\x00\x7c\x01\x3d\x00\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[119];
    }
_collections_abc_toplevel_consts_74_consts_11_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 118,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x2e\x72\x65\x6d\x6f\x76\x65\x28\x76\x61\x6c\x75\x65\x29\x20\x2d\x2d\x20\x72\x65\x6d\x6f\x76\x65\x20\x66\x69\x72\x73\x74\x20\x6f\x63\x63\x75\x72\x72\x65\x6e\x63\x65\x20\x6f\x66\x20\x76\x61\x6c\x75\x65\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x52\x61\x69\x73\x65\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x74\x68\x65\x20\x76\x61\x6c\x75\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x70\x72\x65\x73\x65\x6e\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_74_consts_11_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_index._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_collections_abc_toplevel_consts_74_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.remove",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
_collections_abc_toplevel_consts_74_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x0d\x11\x90\x14\x97\x1a\x91\x1a\x98\x45\xd3\x11\x22\xd1\x0c\x23",
};
static
    struct _PyCode_DEF(40)
_collections_abc_toplevel_consts_74_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts_11_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_74_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1161,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 516,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_remove._ascii.ob_base,
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_11_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x3d\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(extend),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
_collections_abc_toplevel_consts_74_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MutableSequence.__iadd__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
_collections_abc_toplevel_consts_74_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0b\x89\x0b\x90\x46\xd4\x08\x1b\xd8\x0f\x13\x88\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts_12_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(values),
        },
    },
};
static
    struct _PyCode_DEF(40)
_collections_abc_toplevel_consts_74_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_74_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1167,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 517,
    .co_localsplusnames = & _collections_abc_toplevel_consts_74_consts_12_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iadd__),
    .co_qualname = & _collections_abc_toplevel_consts_74_consts_12_qualname._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
_collections_abc_toplevel_consts_74_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & const_str_MutableSequence._ascii.ob_base,
            & _collections_abc_toplevel_consts_74_consts_1._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_74_consts_3.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_4.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_6.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_7.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_8.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_9.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_10.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_11.ob_base.ob_base,
            & _collections_abc_toplevel_consts_74_consts_12.ob_base.ob_base,
            Py_None,
            & codecs_toplevel_consts_28_consts_19._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
_collections_abc_toplevel_consts_74_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__slots__),
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__setitem__),
            &_Py_ID(__delitem__),
            & const_str_insert._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(clear),
            &_Py_ID(reverse),
            &_Py_ID(extend),
            & const_str_pop._ascii.ob_base,
            & const_str_remove._ascii.ob_base,
            &_Py_ID(__iadd__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[108];
    }
_collections_abc_toplevel_consts_74_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 107,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x04\x05\x08\xf0\x0c\x00\x11\x13\x80\x49\xe0\x05\x13\xf1\x02\x01\x05\x19\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x19\xf0\x06\x00\x06\x14\xf1\x02\x01\x05\x19\xf3\x03\x00\x06\x14\xf0\x02\x01\x05\x19\xf0\x06\x00\x06\x14\xf1\x02\x02\x05\x19\xf3\x03\x00\x06\x14\xf0\x02\x02\x05\x19\xf2\x08\x02\x05\x26\xf2\x08\x06\x05\x11\xf2\x10\x04\x05\x38\xf2\x0c\x05\x05\x1b\xf3\x0e\x06\x05\x11\xf2\x10\x04\x05\x24\xf3\x0c\x02\x05\x14",
};
static
    struct _PyCode_DEF(112)
_collections_abc_toplevel_consts_74 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 56,
    },
    .co_consts = & _collections_abc_toplevel_consts_74_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_consts_74_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1106,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 518,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_MutableSequence._ascii.ob_base,
    .co_qualname = & const_str_MutableSequence._ascii.ob_base,
    .co_linetable = & _collections_abc_toplevel_consts_74_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x65\x05\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x05\x64\x04\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x65\x05\x64\x05\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x64\x06\x84\x00\x5a\x09\x64\x07\x84\x00\x5a\x0a\x64\x08\x84\x00\x5a\x0b\x64\x09\x84\x00\x5a\x0c\x64\x0e\x64\x0a\x84\x01\x5a\x0d\x64\x0b\x84\x00\x5a\x0e\x64\x0c\x84\x00\x5a\x0f\x79\x0d",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[76];
        }_object;
    }
_collections_abc_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 76,
        },
        .ob_item = {
            & _collections_abc_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & _collections_abc_toplevel_consts_2._object.ob_base.ob_base,
            Py_None,
            Py_Ellipsis,
            & _collections_abc_toplevel_consts_5.ob_base.ob_base,
            & _collections_abc_toplevel_consts_6._object.ob_base.ob_base,
            & _collections_abc_toplevel_consts_7._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & const_int_1000.ob_base,
            &_Py_STR(empty),
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            & _collections_abc_toplevel_consts_13.ob_base.ob_base,
            & _collections_abc_toplevel_consts_14.ob_base.ob_base,
            & _collections_abc_toplevel_consts_15.ob_base.ob_base,
            & _collections_abc_toplevel_consts_16.ob_base.ob_base,
            & _collections_abc_toplevel_consts_17.ob_base.ob_base,
            & const_str_Hashable._ascii.ob_base,
            & abc_toplevel_consts_17._object.ob_base.ob_base,
            & _collections_abc_toplevel_consts_20.ob_base.ob_base,
            & const_str_Awaitable._ascii.ob_base,
            & _collections_abc_toplevel_consts_22.ob_base.ob_base,
            & const_str_Coroutine._ascii.ob_base,
            & _collections_abc_toplevel_consts_24.ob_base.ob_base,
            & const_str_AsyncIterable._ascii.ob_base,
            & _collections_abc_toplevel_consts_26.ob_base.ob_base,
            & const_str_AsyncIterator._ascii.ob_base,
            & _collections_abc_toplevel_consts_28.ob_base.ob_base,
            & const_str_AsyncGenerator._ascii.ob_base,
            & _collections_abc_toplevel_consts_30.ob_base.ob_base,
            & const_str_Iterable._ascii.ob_base,
            & _collections_abc_toplevel_consts_32.ob_base.ob_base,
            & const_str_Iterator._ascii.ob_base,
            & _collections_abc_toplevel_consts_34.ob_base.ob_base,
            & const_str_Reversible._ascii.ob_base,
            & _collections_abc_toplevel_consts_36.ob_base.ob_base,
            & const_str_Generator._ascii.ob_base,
            & _collections_abc_toplevel_consts_38.ob_base.ob_base,
            & const_str_Sized._ascii.ob_base,
            & _collections_abc_toplevel_consts_40.ob_base.ob_base,
            & const_str_Container._ascii.ob_base,
            & _collections_abc_toplevel_consts_42.ob_base.ob_base,
            & const_str_Collection._ascii.ob_base,
            & _collections_abc_toplevel_consts_44.ob_base.ob_base,
            & const_str_Buffer._ascii.ob_base,
            & _collections_abc_toplevel_consts_46.ob_base.ob_base,
            & const_str__CallableGenericAlias._ascii.ob_base,
            & _collections_abc_toplevel_consts_48.ob_base.ob_base,
            & _collections_abc_toplevel_consts_49.ob_base.ob_base,
            & _collections_abc_toplevel_consts_50.ob_base.ob_base,
            & const_str_Callable._ascii.ob_base,
            & _collections_abc_toplevel_consts_52.ob_base.ob_base,
            & const_str_Set._ascii.ob_base,
            & _collections_abc_toplevel_consts_54.ob_base.ob_base,
            & const_str_MutableSet._ascii.ob_base,
            & _collections_abc_toplevel_consts_56.ob_base.ob_base,
            & const_str_Mapping._ascii.ob_base,
            & _collections_abc_toplevel_consts_58.ob_base.ob_base,
            & const_str_MappingView._ascii.ob_base,
            & _collections_abc_toplevel_consts_60.ob_base.ob_base,
            & const_str_KeysView._ascii.ob_base,
            & _collections_abc_toplevel_consts_62.ob_base.ob_base,
            & const_str_ItemsView._ascii.ob_base,
            & _collections_abc_toplevel_consts_64.ob_base.ob_base,
            & const_str_ValuesView._ascii.ob_base,
            & _collections_abc_toplevel_consts_66.ob_base.ob_base,
            & const_str_MutableMapping._ascii.ob_base,
            & _collections_abc_toplevel_consts_68.ob_base.ob_base,
            & const_str_Sequence._ascii.ob_base,
            & _collections_abc_toplevel_consts_70.ob_base.ob_base,
            & const_str__DeprecateByteStringMeta._ascii.ob_base,
            & _collections_abc_toplevel_consts_72.ob_base.ob_base,
            & const_str_ByteString._ascii.ob_base,
            & _collections_abc_toplevel_consts_74.ob_base.ob_base,
            & const_str_MutableSequence._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_EllipsisType = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "EllipsisType",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_bytes_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytes_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_bytearray_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytearray_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_dict_keyiterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dict_keyiterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_dict_valueiterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dict_valueiterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_dict_itemiterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dict_itemiterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_list_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "list_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_list_reverseiterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "list_reverseiterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_range_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "range_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_longrange_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "longrange_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_set_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "set_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_str_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "str_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_tuple_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tuple_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_zip = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zip",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_zip_iterator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "zip_iterator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_dict_keys = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dict_keys",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_dict_values = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dict_values",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_dict_items = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dict_items",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_mappingproxy = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mappingproxy",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_generator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "generator",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_coroutine = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "coroutine",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_async_generator = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "async_generator",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[85];
        }_object;
    }
_collections_abc_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 85,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_abc._ascii.ob_base,
            & const_str_ABCMeta._ascii.ob_base,
            & const_str_abstractmethod._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(type),
            & const_str_list._ascii.ob_base,
            & const_str_int._ascii.ob_base,
            & const_str_GenericAlias._ascii.ob_base,
            & const_str_EllipsisType._ascii.ob_base,
            & const_str__f._ascii.ob_base,
            & const_str_FunctionType._ascii.ob_base,
            &_Py_ID(__all__),
            &_Py_ID(__name__),
            &_Py_ID(iter),
            & const_str_bytes_iterator._ascii.ob_base,
            & const_str_bytearray._ascii.ob_base,
            & const_str_bytearray_iterator._ascii.ob_base,
            &_Py_ID(keys),
            & const_str_dict_keyiterator._ascii.ob_base,
            &_Py_ID(values),
            & const_str_dict_valueiterator._ascii.ob_base,
            &_Py_ID(items),
            & const_str_dict_itemiterator._ascii.ob_base,
            & const_str_list_iterator._ascii.ob_base,
            &_Py_ID(reversed),
            & const_str_list_reverseiterator._ascii.ob_base,
            & const_str_range._ascii.ob_base,
            & const_str_range_iterator._ascii.ob_base,
            & const_str_longrange_iterator._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            & const_str_set_iterator._ascii.ob_base,
            & const_str_str_iterator._ascii.ob_base,
            & const_str_tuple_iterator._ascii.ob_base,
            & const_str_zip._ascii.ob_base,
            & const_str_zip_iterator._ascii.ob_base,
            & const_str_dict_keys._ascii.ob_base,
            & const_str_dict_values._ascii.ob_base,
            & const_str_dict_items._ascii.ob_base,
            &_Py_ID(__dict__),
            & const_str_mappingproxy._ascii.ob_base,
            & const_str_generator._ascii.ob_base,
            & const_str__coro._ascii.ob_base,
            & const_str_coroutine._ascii.ob_base,
            &_Py_ID(close),
            & const_str__ag._ascii.ob_base,
            & const_str_async_generator._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_Hashable._ascii.ob_base,
            & const_str_Awaitable._ascii.ob_base,
            & const_str_Coroutine._ascii.ob_base,
            & const_str_register._ascii.ob_base,
            & const_str_AsyncIterable._ascii.ob_base,
            & const_str_AsyncIterator._ascii.ob_base,
            & const_str_AsyncGenerator._ascii.ob_base,
            & const_str_Iterable._ascii.ob_base,
            & const_str_Iterator._ascii.ob_base,
            & const_str_Reversible._ascii.ob_base,
            & const_str_Generator._ascii.ob_base,
            & const_str_Sized._ascii.ob_base,
            & const_str_Container._ascii.ob_base,
            & const_str_Collection._ascii.ob_base,
            & const_str_Buffer._ascii.ob_base,
            & const_str__CallableGenericAlias._ascii.ob_base,
            & const_str__is_param_expr._ascii.ob_base,
            & const_str__type_repr._ascii.ob_base,
            & const_str_Callable._ascii.ob_base,
            & const_str_Set._ascii.ob_base,
            & const_str_frozenset._ascii.ob_base,
            & const_str_MutableSet._ascii.ob_base,
            & const_str_Mapping._ascii.ob_base,
            & const_str_MappingView._ascii.ob_base,
            & const_str_KeysView._ascii.ob_base,
            & const_str_ItemsView._ascii.ob_base,
            & const_str_ValuesView._ascii.ob_base,
            & const_str_MutableMapping._ascii.ob_base,
            &_Py_ID(dict),
            & const_str_Sequence._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_str._ascii.ob_base,
            & const_str_memoryview._ascii.ob_base,
            & const_str__DeprecateByteStringMeta._ascii.ob_base,
            & const_str_ByteString._ascii.ob_base,
            &_Py_ID(bytes),
            & const_str_MutableSequence._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[1282];
    }
_collections_abc_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 1281,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x08\x03\x01\x04\xf7\x3e\x00\x01\x28\xdb\x00\x0a\xe1\x0f\x13\x90\x44\x98\x13\x91\x49\x8b\x7f\x80\x0c\xd9\x0f\x13\x90\x43\x8b\x79\x80\x0c\xda\x00\x0e\xd9\x0f\x13\x90\x42\x8b\x78\x80\x0c\xd8\x04\x06\xf2\x04\x09\x0b\x0d\x80\x07\xf0\x1e\x00\x0c\x1d\x80\x08\xf1\x12\x00\x12\x16\x91\x64\x98\x33\x93\x69\x93\x1f\x80\x0e\xd9\x15\x19\x99\x24\x99\x79\x9b\x7b\xd3\x1a\x2b\xd3\x15\x2c\xd0\x00\x12\xe1\x13\x17\x99\x04\x98\x52\x9f\x57\x99\x57\x9b\x59\x9b\x0f\xd3\x13\x28\xd0\x00\x10\xd9\x15\x19\x99\x24\x98\x72\x9f\x79\x99\x79\x9b\x7b\xd3\x1a\x2b\xd3\x15\x2c\xd0\x00\x12\xd9\x14\x18\x99\x14\x98\x62\x9f\x68\x99\x68\x9b\x6a\xd3\x19\x29\xd3\x14\x2a\xd0\x00\x11\xd9\x10\x14\x91\x54\x98\x22\x93\x58\x93\x0e\x80\x0d\xd9\x17\x1b\x99\x44\xa1\x18\xa8\x22\xa3\x1c\xd3\x1c\x2e\xd3\x17\x2f\xd0\x00\x14\xd9\x11\x15\x91\x64\x99\x35\xa0\x11\x9b\x38\x93\x6e\xd3\x11\x25\x80\x0e\xd9\x15\x19\x99\x24\x99\x75\xa0\x51\xa8\x24\xa1\x59\xd3\x1f\x2f\xd3\x1a\x30\xd3\x15\x31\xd0\x00\x12\xd9\x0f\x13\x91\x44\x99\x13\x9b\x15\x93\x4b\xd3\x0f\x20\x80\x0c\xd9\x0f\x13\x91\x44\x98\x12\x93\x48\x8b\x7e\x80\x0c\xd9\x11\x15\x91\x64\x98\x32\x93\x68\x93\x1e\x80\x0e\xd9\x0f\x13\x91\x44\x99\x13\x9b\x15\x93\x4b\xd3\x0f\x20\x80\x0c\xe1\x0c\x10\x90\x12\x97\x17\x91\x17\x93\x19\x8b\x4f\x80\x09\xd9\x0e\x12\x90\x32\x97\x39\x91\x39\x93\x3b\xd3\x0e\x1f\x80\x0b\xd9\x0d\x11\x90\x22\x97\x28\x91\x28\x93\x2a\xd3\x0d\x1d\x80\x0a\xe1\x0f\x13\x90\x44\x97\x4d\x91\x4d\xd3\x0f\x22\x80\x0c\xd9\x0c\x10\x92\x2f\xd3\x11\x24\xd3\x0c\x25\x80\x09\xe2\x00\x17\xd9\x08\x0d\x8b\x07\x80\x05\xd9\x0c\x10\x90\x15\x8b\x4b\x80\x09\xd8\x00\x05\x87\x0b\x81\x0b\x84\x0d\xd8\x04\x09\xe2\x00\x16\xd9\x06\x09\x83\x65\x80\x03\xd9\x12\x16\x90\x73\x93\x29\x80\x0f\xd8\x04\x07\xf2\x0a\x0a\x01\x10\xf4\x18\x0c\x01\x1e\x98\x17\xf5\x00\x0c\x01\x1e\xf4\x1e\x0e\x01\x32\x98\x27\xf5\x00\x0e\x01\x32\xf4\x22\x26\x01\x1e\x90\x09\xf4\x00\x26\x01\x1e\xf0\x52\x01\x00\x01\x0a\xd7\x00\x12\xd1\x00\x12\x90\x39\xd4\x00\x1d\xf4\x06\x0e\x01\x32\x98\x67\xf5\x00\x0e\x01\x32\xf4\x22\x10\x01\x1e\x90\x4d\xf4\x00\x10\x01\x1e\xf4\x26\x2d\x01\x1e\x90\x5d\xf4\x00\x2d\x01\x1e\xf0\x60\x01\x00\x01\x0f\xd7\x00\x17\xd1\x00\x17\x98\x0f\xd4\x00\x28\xf4\x06\x0f\x01\x32\x98\x17\xf5\x00\x0f\x01\x32\xf4\x24\x10\x01\x1e\x88\x78\xf4\x00\x10\x01\x1e\xf0\x26\x00\x01\x09\xd7\x00\x11\xd1\x00\x11\x90\x2e\xd4\x00\x21\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\xd0\x12\x24\xd4\x00\x25\xe0\x00\x08\xd7\x00\x11\xd1\x00\x11\xd0\x12\x22\xd4\x00\x23\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\xd0\x12\x24\xd4\x00\x25\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\xd0\x12\x23\xd4\x00\x24\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x2d\xd4\x00\x20\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\xd0\x12\x26\xd4\x00\x27\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x2e\xd4\x00\x21\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\xd0\x12\x24\xd4\x00\x25\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x2c\xd4\x00\x1f\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x2c\xd4\x00\x1f\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x2e\xd4\x00\x21\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x2c\xd4\x00\x1f\xf4\x06\x0d\x01\x1e\x90\x18\xf4\x00\x0d\x01\x1e\xf4\x20\x2d\x01\x1e\x90\x08\xf4\x00\x2d\x01\x1e\xf0\x60\x01\x00\x01\x0a\xd7\x00\x12\xd1\x00\x12\x90\x39\xd4\x00\x1d\xf4\x06\x0c\x01\x1e\x90\x67\xf5\x00\x0c\x01\x1e\xf4\x1e\x0e\x01\x32\x98\x27\xf5\x00\x0e\x01\x32\xf4\x22\x08\x01\x1e\x90\x15\x98\x08\xa0\x29\xf4\x00\x08\x01\x1e\xf4\x16\x0c\x01\x1e\x90\x77\xf5\x00\x0c\x01\x1e\xf4\x1e\x34\x01\x40\x01\x98\x4c\xf4\x00\x34\x01\x40\x01\xf2\x6c\x01\x0a\x01\x56\x01\xf2\x18\x0f\x01\x15\xf4\x24\x0e\x01\x3b\x98\x17\xf5\x00\x0e\x01\x3b\xf4\x28\x47\x02\x01\x11\x88\x2a\xf4\x00\x47\x02\x01\x11\xf0\x54\x04\x00\x01\x04\x87\x0c\x81\x0c\x88\x59\xd4\x00\x17\xf4\x06\x4d\x01\x01\x14\x90\x13\xf4\x00\x4d\x01\x01\x14\xf0\x60\x02\x00\x01\x0b\xd7\x00\x13\xd1\x00\x13\x90\x43\xd4\x00\x18\xf4\x0a\x31\x01\x18\x88\x6a\xf4\x00\x31\x01\x18\xf0\x66\x01\x00\x01\x08\xd7\x00\x10\xd1\x00\x10\x90\x1c\xd4\x00\x1e\xf4\x06\x0d\x01\x32\x90\x25\xf4\x00\x0d\x01\x32\xf4\x20\x0c\x01\x21\x88\x7b\x98\x43\xf4\x00\x0c\x01\x21\xf0\x1e\x00\x01\x09\xd7\x00\x11\xd1\x00\x11\x90\x29\xd4\x00\x1c\xf4\x06\x13\x01\x2c\x90\x0b\x98\x53\xf4\x00\x13\x01\x2c\xf0\x2c\x00\x01\x0a\xd7\x00\x12\xd1\x00\x12\x90\x3a\xd4\x00\x1e\xf4\x06\x0d\x01\x25\x90\x1b\x98\x6a\xf4\x00\x0d\x01\x25\xf0\x20\x00\x01\x0b\xd7\x00\x13\xd1\x00\x13\x90\x4b\xd4\x00\x20\xf4\x06\x4f\x01\x01\x17\x90\x57\xf4\x00\x4f\x01\x01\x17\xf0\x64\x02\x00\x01\x0f\xd7\x00\x17\xd1\x00\x17\x98\x04\xd4\x00\x1d\xf4\x0a\x3d\x01\x40\x01\x88\x7a\x98\x3a\xf4\x00\x3d\x01\x40\x01\xf0\x7e\x01\x00\x01\x09\xd7\x00\x11\xd1\x00\x11\x90\x25\xd4\x00\x18\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x23\xd4\x00\x16\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x25\xd4\x00\x18\xd8\x00\x08\xd7\x00\x11\xd1\x00\x11\x90\x2a\xd4\x00\x1d\xf4\x04\x12\x01\x33\x98\x77\xf4\x00\x12\x01\x33\xf4\x28\x06\x01\x13\x90\x18\xd0\x25\x3d\xf5\x00\x06\x01\x13\xf0\x10\x00\x01\x0b\xd7\x00\x13\xd1\x00\x13\x90\x45\xd4\x00\x1a\xd8\x00\x0a\xd7\x00\x13\xd1\x00\x13\x90\x49\xd4\x00\x1e\xf4\x06\x3f\x01\x14\x90\x68\xf4\x00\x3f\x01\x14\xf0\x44\x02\x00\x01\x10\xd7\x00\x18\xd1\x00\x18\x98\x14\xd4\x00\x1e\xd8\x00\x0f\xd7\x00\x18\xd1\x00\x18\x98\x19\xd5\x00\x23",
};
static
    struct _PyCode_DEF(2650)
_collections_abc_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 1325,
    },
    .co_consts = & _collections_abc_toplevel_consts._object.ob_base.ob_base,
    .co_names = & _collections_abc_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 519,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _collections_abc_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & _collections_abc_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x6d\x02\x5a\x02\x6d\x03\x5a\x03\x01\x00\x64\x01\x64\x03\x6c\x04\x5a\x04\x02\x00\x65\x05\x65\x06\x65\x07\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x08\x02\x00\x65\x05\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x64\x05\x84\x00\x5a\x0a\x02\x00\x65\x05\x65\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x0b\x5b\x0a\x67\x00\x64\x06\xa2\x01\x5a\x0c\x64\x07\x5a\x0d\x02\x00\x65\x05\x02\x00\x65\x0e\x64\x08\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x0f\x02\x00\x65\x05\x02\x00\x65\x0e\x02\x00\x65\x10\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x11\x02\x00\x65\x05\x02\x00\x65\x0e\x69\x00\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x13\x02\x00\x65\x05\x02\x00\x65\x0e\x69\x00\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x15\x02\x00\x65\x05\x02\x00\x65\x0e\x69\x00\x6a\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x17\x02\x00\x65\x05\x02\x00\x65\x0e\x67\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x18\x02\x00\x65\x05\x02\x00\x65\x0e\x02\x00\x65\x19\x67\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x1a\x02\x00\x65\x05\x02\x00\x65\x0e\x02\x00\x65\x1b\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x1c\x02\x00\x65\x05\x02\x00\x65\x0e\x02\x00\x65\x1b\x64\x09\x64\x0a\x7a\x03\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x1d\x02\x00\x65\x05\x02\x00\x65\x0e\x02\x00\x65\x1e\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x1f\x02\x00\x65\x05\x02\x00\x65\x0e\x64\x0b\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x20\x02\x00\x65\x05\x02\x00\x65\x0e\x64\x0c\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x21\x02\x00\x65\x05\x02\x00\x65\x0e\x02\x00\x65\x22\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x23\x02\x00\x65\x05\x69\x00\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x24\x02\x00\x65\x05\x69\x00\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x25\x02\x00\x65\x05\x69\x00\x6a\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x26\x02\x00\x65\x05\x65\x05\x6a\x4e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x28\x02\x00\x65\x05\x02\x00\x64\x0d\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x29\x64\x0e\x84\x00\x5a\x2a\x02\x00\x65\x2a\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x2a\x02\x00\x65\x05\x65\x2a\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x2b\x65\x2a\x6a\x59\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x5b\x2a\x64\x0f\x84\x00\x5a\x2d\x02\x00\x65\x2d\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x2d\x02\x00\x65\x05\x65\x2d\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x2e\x5b\x2d\x64\x10\x84\x00\x5a\x2f\x02\x00\x47\x00\x64\x11\x84\x00\x64\x12\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x30\x02\x00\x47\x00\x64\x14\x84\x00\x64\x15\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x31\x02\x00\x47\x00\x64\x16\x84\x00\x64\x17\x65\x31\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x32\x65\x32\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x2b\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x18\x84\x00\x64\x19\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x34\x02\x00\x47\x00\x64\x1a\x84\x00\x64\x1b\x65\x34\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x35\x02\x00\x47\x00\x64\x1c\x84\x00\x64\x1d\x65\x35\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x36\x65\x36\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x2e\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x1e\x84\x00\x64\x1f\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x37\x02\x00\x47\x00\x64\x20\x84\x00\x64\x21\x65\x37\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x38\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x0f\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x11\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x13\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x15\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x17\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x18\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1a\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1c\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1d\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1f\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x20\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x21\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x38\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x23\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x22\x84\x00\x64\x23\x65\x37\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x39\x02\x00\x47\x00\x64\x24\x84\x00\x64\x25\x65\x38\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x3a\x65\x3a\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x29\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x26\x84\x00\x64\x27\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x3b\x02\x00\x47\x00\x64\x28\x84\x00\x64\x29\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x3c\x02\x00\x47\x00\x64\x2a\x84\x00\x64\x2b\x65\x3b\x65\x37\x65\x3c\xab\x05\x00\x00\x00\x00\x00\x00\x5a\x3d\x02\x00\x47\x00\x64\x2c\x84\x00\x64\x2d\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x3e\x02\x00\x47\x00\x64\x2e\x84\x00\x64\x2f\x65\x08\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x3f\x64\x30\x84\x00\x5a\x40\x64\x31\x84\x00\x5a\x41\x02\x00\x47\x00\x64\x32\x84\x00\x64\x33\x65\x02\xac\x13\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x42\x02\x00\x47\x00\x64\x34\x84\x00\x64\x35\x65\x3d\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x43\x65\x43\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x44\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x36\x84\x00\x64\x37\x65\x43\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x45\x65\x45\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1e\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x38\x84\x00\x64\x39\x65\x3d\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x46\x65\x46\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x28\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x3a\x84\x00\x64\x3b\x65\x3b\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x47\x02\x00\x47\x00\x64\x3c\x84\x00\x64\x3d\x65\x47\x65\x43\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x48\x65\x48\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x24\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x3e\x84\x00\x64\x3f\x65\x47\x65\x43\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x49\x65\x49\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x26\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x40\x84\x00\x64\x41\x65\x47\x65\x3d\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x4a\x65\x4a\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x25\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x42\x84\x00\x64\x43\x65\x46\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x4b\x65\x4b\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x4c\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x44\x84\x00\x64\x45\x65\x39\x65\x3d\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x4d\x65\x4d\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x4e\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x4d\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x4f\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x4d\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1b\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x4d\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x50\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x46\x84\x00\x64\x47\x65\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x51\x02\x00\x47\x00\x64\x48\x84\x00\x64\x49\x65\x4d\x65\x51\xac\x13\xab\x04\x00\x00\x00\x00\x00\x00\x5a\x52\x65\x52\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x53\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x52\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x10\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x47\x00\x64\x4a\x84\x00\x64\x4b\x65\x4d\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x54\x65\x54\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x54\x6a\x67\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x10\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x03",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get__collections_abc_toplevel(void)
{
    return Py_NewRef((PyObject *) &_collections_abc_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[62];
    }
_sitebuiltins_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 61,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x54\x68\x65\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x75\x73\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x73\x69\x74\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x74\x6f\x20\x61\x64\x64\x20\x63\x75\x73\x74\x6f\x6d\x20\x62\x75\x69\x6c\x74\x69\x6e\x73\x2e\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_Quitter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Quitter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_eof = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "eof",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_sitebuiltins_toplevel_consts_3_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_eof._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
_sitebuiltins_toplevel_consts_3_consts_1_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen _sitebuiltins>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
_sitebuiltins_toplevel_consts_3_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Quitter.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
_sitebuiltins_toplevel_consts_3_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x14\x18\x88\x04\x8c\x09\xd8\x13\x16\x88\x04\x8d\x08",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
_sitebuiltins_toplevel_consts_3_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
            & const_str_eof._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(32)
_sitebuiltins_toplevel_consts_3_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_3_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 14,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 520,
    .co_localsplusnames = & _sitebuiltins_toplevel_consts_3_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & _sitebuiltins_toplevel_consts_3_consts_1_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_3_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
_sitebuiltins_toplevel_consts_3_consts_2_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Use ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
_sitebuiltins_toplevel_consts_3_consts_2_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "() or ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
_sitebuiltins_toplevel_consts_3_consts_2_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " to exit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_sitebuiltins_toplevel_consts_3_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & _sitebuiltins_toplevel_consts_3_consts_2_consts_1._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_3_consts_2_consts_2._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_3_consts_2_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
_sitebuiltins_toplevel_consts_3_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Quitter.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
_sitebuiltins_toplevel_consts_3_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x81\x00\xd8\x2b\x2f\xaf\x39\xab\x39\xb0\x64\xb7\x68\xb3\x68\xd0\x0f\x3f\xd0\x08\x3f",
};
static
    struct _PyCode_DEF(60)
_sitebuiltins_toplevel_consts_3_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 30,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_3_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_3_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 17,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 521,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & _sitebuiltins_toplevel_consts_3_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_3_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x01\x64\x02\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x01\x64\x03\x9d\x05\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_SystemExit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SystemExit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_sitebuiltins_toplevel_consts_3_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(stdin),
            &_Py_ID(close),
            & const_str_SystemExit._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
_sitebuiltins_toplevel_consts_3_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Quitter.__call__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[56];
    }
_sitebuiltins_toplevel_consts_3_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 55,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x03\x09\x11\xdc\x0c\x0f\x8f\x49\x89\x49\x8f\x4f\x89\x4f\xd4\x0c\x1d\xf4\x06\x00\x0f\x19\x98\x14\xd3\x0e\x1e\xd0\x08\x1e\xf8\xf0\x05\x01\x09\x11\xd8\x0c\x10\xdc\x0e\x18\x98\x14\xd3\x0e\x1e\xd0\x08\x1e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
_sitebuiltins_toplevel_consts_3_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x1e\x2b\x00\xab\x02\x39\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_sitebuiltins_toplevel_consts_3_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(code),
        },
    },
};
static
    struct _PyCode_DEF(120)
_sitebuiltins_toplevel_consts_3_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 60,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_3_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _sitebuiltins_toplevel_consts_3_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 19,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 522,
    .co_localsplusnames = & _sitebuiltins_toplevel_consts_3_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__call__),
    .co_qualname = & _sitebuiltins_toplevel_consts_3_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_3_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x23\x00\x01\x00\x59\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_sitebuiltins_toplevel_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_Quitter._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_3_consts_1.ob_base.ob_base,
            & _sitebuiltins_toplevel_consts_3_consts_2.ob_base.ob_base,
            Py_None,
            & _sitebuiltins_toplevel_consts_3_consts_4.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_sitebuiltins_toplevel_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            &_Py_ID(__repr__),
            &_Py_ID(__call__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
_sitebuiltins_toplevel_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x02\x02\x05\x17\xf2\x06\x01\x05\x40\x01\xf4\x04\x07\x05\x1f",
};
static
    struct _PyCode_DEF(32)
_sitebuiltins_toplevel_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_3_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 13,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 523,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_Quitter._ascii.ob_base,
    .co_qualname = & const_str_Quitter._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x05\x64\x04\x84\x01\x5a\x05\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__Printer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[111];
    }
_sitebuiltins_toplevel_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 110,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x69\x6e\x74\x65\x72\x61\x63\x74\x69\x76\x65\x20\x70\x72\x6f\x6d\x70\x74\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x66\x6f\x72\x20\x70\x72\x69\x6e\x74\x69\x6e\x67\x20\x74\x68\x65\x20\x6c\x69\x63\x65\x6e\x73\x65\x20\x74\x65\x78\x74\x2c\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x0a\x20\x20\x20\x20\x63\x6f\x6e\x74\x72\x69\x62\x75\x74\x6f\x72\x73\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x63\x6f\x70\x79\x72\x69\x67\x68\x74\x20\x6e\x6f\x74\x69\x63\x65\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_os = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "os",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__Printer__name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer__name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__Printer__data = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer__data",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__Printer__lines = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer__lines",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__Printer__filenames = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer__filenames",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str__Printer__name._ascii.ob_base,
            & const_str__Printer__data._ascii.ob_base,
            & const_str__Printer__lines._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(join),
            & const_str__Printer__filenames._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_sitebuiltins_toplevel_consts_5_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[92];
    }
_sitebuiltins_toplevel_consts_5_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 91,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdb\x08\x11\xd8\x16\x1a\x88\x04\x8c\x0b\xd8\x16\x1a\x88\x04\x8c\x0b\xd8\x17\x1b\x88\x04\x8c\x0c\xe1\x27\x2b\xf4\x03\x02\x1c\x33\xd9\x27\x2b\xa0\x03\xdb\x2c\x31\xa0\x08\xf0\x05\x00\x1d\x1f\x9f\x47\x99\x47\x9f\x4c\x99\x4c\xa8\x13\xa8\x68\xd5\x1c\x37\xe0\x2c\x31\xf0\x05\x00\x1d\x38\xd8\x27\x2b\xf2\x03\x02\x1c\x33\x88\x04\xd5\x08\x18\xf9\xf3\x00\x02\x1c\x33",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
_sitebuiltins_toplevel_consts_5_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x9f\x2a\x41\x13\x06",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_dirs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dirs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_dir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dir",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(name),
            &_Py_ID(data),
            & const_str_files._ascii.ob_base,
            & const_str_dirs._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_dir._ascii.ob_base,
            &_Py_ID(filename),
        },
    },
};
static
    struct _PyCode_DEF(178)
_sitebuiltins_toplevel_consts_5_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 89,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_5_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = & _sitebuiltins_toplevel_consts_5_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 35,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 524,
    .co_localsplusnames = & _sitebuiltins_toplevel_consts_5_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & _sitebuiltins_toplevel_consts_5_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_5_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x00\x6c\x00\x7d\x05\x7c\x01\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x44\x00\x8f\x06\x8f\x07\x63\x03\x67\x00\x63\x02\x5d\x25\x00\x00\x7d\x06\x7c\x03\x44\x00\x5d\x1e\x00\x00\x7d\x07\x7c\x05\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x07\xab\x02\x00\x00\x00\x00\x00\x00\x91\x03\x8c\x20\x04\x00\x8c\x27\x04\x00\x63\x03\x7d\x07\x7d\x06\x7c\x00\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x63\x02\x01\x00\x63\x03\x7d\x07\x7d\x06\x77\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            &_Py_STR(utf_8),
            & codecs_toplevel_consts_35_localsplusnames._object.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[10],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_split = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "split",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__Printer__linecnt = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer__linecnt",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str__Printer__lines._ascii.ob_base,
            & const_str__Printer__filenames._ascii.ob_base,
            &_Py_ID(open),
            &_Py_ID(read),
            & const_str_OSError._ascii.ob_base,
            & const_str__Printer__data._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            &_Py_ID(len),
            & const_str__Printer__linecnt._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str___setup = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__setup",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
_sitebuiltins_toplevel_consts_5_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer.__setup",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[144];
    }
_sitebuiltins_toplevel_consts_5_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 143,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\x8f\x3c\x8a\x3c\xd8\x0c\x12\xd8\x0f\x13\x88\x04\xd8\x18\x1c\xd7\x18\x28\xd4\x18\x28\x88\x48\xf0\x02\x05\x0d\x15\xdc\x15\x19\x98\x28\xa8\x57\xd5\x15\x35\xb8\x12\xd8\x1b\x1d\x9f\x37\x99\x37\x9b\x39\x90\x44\xf7\x03\x00\x16\x36\xe1\x10\x15\xf0\x09\x00\x19\x29\xf1\x0e\x00\x10\x14\xd8\x13\x17\x97\x3b\x91\x3b\x88\x44\xd8\x17\x1b\x97\x7a\x91\x7a\xa0\x24\xd3\x17\x27\x88\x04\x8c\x0c\xdc\x19\x1c\x98\x54\x9f\x5c\x99\x5c\xd3\x19\x2a\x88\x04\x8d\x0e\xf7\x11\x00\x16\x36\xd0\x15\x35\xfb\xf4\x06\x00\x14\x1b\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
_sitebuiltins_toplevel_consts_5_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\xa0\x0d\x42\x14\x02\xad\x11\x42\x08\x05\xbe\x08\x42\x14\x02\xc2\x08\x05\x42\x11\x09\xc2\x0d\x07\x42\x14\x02\xc2\x14\x09\x42\x20\x05\xc2\x1f\x01\x42\x20\x05",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(data),
            &_Py_ID(filename),
            & const_str_fp._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(326)
_sitebuiltins_toplevel_consts_5_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 163,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_5_consts_4_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_5_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _sitebuiltins_toplevel_consts_5_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 44,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 525,
    .co_localsplusnames = & _sitebuiltins_toplevel_consts_5_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = & const_str___setup._ascii.ob_base,
    .co_qualname = & _sitebuiltins_toplevel_consts_5_consts_4_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_5_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x01\x79\x00\x64\x00\x7d\x01\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x2a\x00\x00\x7d\x02\x09\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x01\xac\x02\xab\x02\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x03\x7c\x03\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x01\x00\x6e\x01\x04\x00\x7c\x01\x73\x0c\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x4b\x78\x03\x59\x00\x77\x01\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x83\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
_sitebuiltins_toplevel_consts_5_consts_5_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Type %s() to see the full %s text",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[10],
            & _sitebuiltins_toplevel_consts_5_consts_5_consts_2._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__Printer__setup = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer__setup",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_MAXLINES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MAXLINES",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__Printer__setup._ascii.ob_base,
            &_Py_ID(len),
            & const_str__Printer__lines._ascii.ob_base,
            & const_str_MAXLINES._ascii.ob_base,
            &_Py_ID(join),
            & const_str__Printer__name._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_sitebuiltins_toplevel_consts_5_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[74];
    }
_sitebuiltins_toplevel_consts_5_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 73,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0c\x89\x0c\x8c\x0e\xdc\x0b\x0e\x88\x74\x8f\x7c\x89\x7c\xd3\x0b\x1c\xa0\x04\xa7\x0d\xa1\x0d\xd2\x0b\x2d\xd8\x13\x17\x97\x39\x91\x39\x98\x54\x9f\x5c\x99\x5c\xd3\x13\x2a\xd0\x0c\x2a\xe0\x13\x36\xb8\x34\xbf\x3b\xb9\x3b\xb8\x2e\xc8\x11\xd1\x3a\x4a\xd1\x13\x4b\xd0\x0c\x4b",
};
static
    struct _PyCode_DEF(194)
_sitebuiltins_toplevel_consts_5_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 97,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_5_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_5_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 60,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 526,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & _sitebuiltins_toplevel_consts_5_consts_5_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_5_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x1a\x00\x00\x72\x1b\x64\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x64\x02\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x66\x01\x64\x03\x7a\x05\x00\x00\x7a\x06\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[49];
    }
_sitebuiltins_toplevel_consts_5_consts_6_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 48,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Hit Return for more, or q (and Return) to quit: ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_6_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_STR(empty),
            (PyObject *)&_Py_SINGLETON(strings).ascii[113],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            & _sitebuiltins_toplevel_consts_5_consts_6_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & _sitebuiltins_toplevel_consts_5_consts_6_consts_3._object.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[113],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__Printer__setup._ascii.ob_base,
            & const_str_range._ascii.ob_base,
            & const_str_MAXLINES._ascii.ob_base,
            & const_str_print._ascii.ob_base,
            & const_str__Printer__lines._ascii.ob_base,
            &_Py_ID(input),
            & const_str_IndexError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
_sitebuiltins_toplevel_consts_5_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Printer.__call__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[164];
    }
_sitebuiltins_toplevel_consts_5_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 163,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0c\x89\x0c\x8c\x0e\xd8\x11\x43\x88\x06\xd8\x11\x12\x88\x06\xd8\x0e\x0f\xf0\x02\x0d\x0d\x1a\xdc\x19\x1e\x98\x76\xa0\x76\xb0\x04\xb7\x0d\xb1\x0d\xd1\x27\x3d\xd6\x19\x3e\x90\x41\xdc\x14\x19\x98\x24\x9f\x2c\x99\x2c\xa0\x71\x99\x2f\xd5\x14\x2a\xf1\x03\x00\x1a\x3f\xf0\x0a\x00\x11\x17\x98\x24\x9f\x2d\x99\x2d\xd1\x10\x27\x90\x06\xd8\x16\x1a\x90\x03\xd8\x16\x19\x90\x6b\xdc\x1a\x1f\xa0\x06\x9b\x2d\x90\x43\xd8\x17\x1a\xa0\x29\xd1\x17\x2b\xd8\x1e\x22\x98\x03\xf0\x07\x00\x17\x1a\x91\x6b\xf0\x08\x00\x14\x17\x98\x23\x92\x3a\xd8\x14\x19\xf0\x1d\x00\x0f\x10\xf8\xf4\x08\x00\x14\x1e\xf2\x00\x01\x0d\x16\xd9\x10\x15\xf0\x03\x01\x0d\x16\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
_sitebuiltins_toplevel_consts_5_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x97\x36\x41\x3c\x00\xc1\x3c\x09\x42\x08\x03\xc2\x07\x01\x42\x08\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_prompt = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "prompt",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_prompt._ascii.ob_base,
            &_Py_ID(lineno),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            &_Py_ID(key),
        },
    },
};
static
    struct _PyCode_DEF(278)
_sitebuiltins_toplevel_consts_5_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 139,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_5_consts_6_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_5_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & _sitebuiltins_toplevel_consts_5_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 67,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 527,
    .co_localsplusnames = & _sitebuiltins_toplevel_consts_5_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__call__),
    .co_qualname = & _sitebuiltins_toplevel_consts_5_consts_6_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_5_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x7d\x01\x64\x02\x7d\x02\x09\x00\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x02\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x1a\x00\x00\x7d\x03\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x1c\x04\x00\x09\x00\x7c\x02\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x0d\x00\x00\x7d\x02\x64\x00\x7d\x04\x7c\x04\x80\x14\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x64\x03\x76\x01\x72\x02\x64\x00\x7d\x04\x7c\x04\x80\x01\x8c\x14\x7c\x04\x64\x04\x6b\x28\x00\x00\x72\x01\x79\x00\x8c\x66\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts_8 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            (PyObject *)& _Py_SINGLETON(tuple_empty),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str__Printer._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_5_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 23],
            & _sitebuiltins_toplevel_consts_5_consts_3.ob_base.ob_base,
            & _sitebuiltins_toplevel_consts_5_consts_4.ob_base.ob_base,
            & _sitebuiltins_toplevel_consts_5_consts_5.ob_base.ob_base,
            & _sitebuiltins_toplevel_consts_5_consts_6.ob_base.ob_base,
            Py_None,
            & _sitebuiltins_toplevel_consts_5_consts_8._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_sitebuiltins_toplevel_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_MAXLINES._ascii.ob_base,
            &_Py_ID(__init__),
            & const_str__Printer__setup._ascii.ob_base,
            &_Py_ID(__repr__),
            &_Py_ID(__call__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
_sitebuiltins_toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x01\x05\x2e\xf0\x06\x00\x10\x12\x80\x48\xf3\x04\x07\x05\x33\xf2\x12\x0e\x05\x2b\xf2\x20\x05\x05\x4c\x01\xf3\x0e\x12\x05\x1a",
};
static
    struct _PyCode_DEF(46)
_sitebuiltins_toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_5_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 29,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 528,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__Printer._ascii.ob_base,
    .co_qualname = & const_str__Printer._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x5a\x04\x64\x08\x64\x03\x84\x01\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__Helper = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Helper",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[308];
    }
_sitebuiltins_toplevel_consts_7_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 307,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x65\x66\x69\x6e\x65\x20\x74\x68\x65\x20\x62\x75\x69\x6c\x74\x69\x6e\x20\x27\x68\x65\x6c\x70\x27\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x69\x73\x20\x61\x20\x77\x72\x61\x70\x70\x65\x72\x20\x61\x72\x6f\x75\x6e\x64\x20\x70\x79\x64\x6f\x63\x2e\x68\x65\x6c\x70\x20\x74\x68\x61\x74\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x61\x20\x68\x65\x6c\x70\x66\x75\x6c\x20\x6d\x65\x73\x73\x61\x67\x65\x0a\x20\x20\x20\x20\x77\x68\x65\x6e\x20\x27\x68\x65\x6c\x70\x27\x20\x69\x73\x20\x74\x79\x70\x65\x64\x20\x61\x74\x20\x74\x68\x65\x20\x50\x79\x74\x68\x6f\x6e\x20\x69\x6e\x74\x65\x72\x61\x63\x74\x69\x76\x65\x20\x70\x72\x6f\x6d\x70\x74\x2e\x0a\x0a\x20\x20\x20\x20\x43\x61\x6c\x6c\x69\x6e\x67\x20\x68\x65\x6c\x70\x28\x29\x20\x61\x74\x20\x74\x68\x65\x20\x50\x79\x74\x68\x6f\x6e\x20\x70\x72\x6f\x6d\x70\x74\x20\x73\x74\x61\x72\x74\x73\x20\x61\x6e\x20\x69\x6e\x74\x65\x72\x61\x63\x74\x69\x76\x65\x20\x68\x65\x6c\x70\x20\x73\x65\x73\x73\x69\x6f\x6e\x2e\x0a\x20\x20\x20\x20\x43\x61\x6c\x6c\x69\x6e\x67\x20\x68\x65\x6c\x70\x28\x74\x68\x69\x6e\x67\x29\x20\x70\x72\x69\x6e\x74\x73\x20\x68\x65\x6c\x70\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x70\x79\x74\x68\x6f\x6e\x20\x6f\x62\x6a\x65\x63\x74\x20\x27\x74\x68\x69\x6e\x67\x27\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[73];
    }
_sitebuiltins_toplevel_consts_7_consts_2_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 72,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Type help() for interactive help, or help(object) for help about object.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_sitebuiltins_toplevel_consts_7_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & _sitebuiltins_toplevel_consts_7_consts_2_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
_sitebuiltins_toplevel_consts_7_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Helper.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
_sitebuiltins_toplevel_consts_7_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x01\x10\x38",
};
static
    struct _PyCode_DEF(4)
_sitebuiltins_toplevel_consts_7_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_7_consts_2_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 98,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 529,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & _sitebuiltins_toplevel_consts_7_consts_2_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_7_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_pydoc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pydoc",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_help = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "help",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
_sitebuiltins_toplevel_consts_7_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_pydoc._ascii.ob_base,
            & const_str_help._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
_sitebuiltins_toplevel_consts_7_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Helper.__call__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
_sitebuiltins_toplevel_consts_7_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdb\x08\x14\xd8\x0f\x19\x88\x75\x8f\x7a\x89\x7a\x98\x34\xd0\x0f\x28\xa0\x34\xd1\x0f\x28\xd0\x08\x28",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
_sitebuiltins_toplevel_consts_7_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(args),
            & const_str_kwds._ascii.ob_base,
            & const_str_pydoc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(46)
_sitebuiltins_toplevel_consts_7_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_7_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 15,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 101,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 530,
    .co_localsplusnames = & _sitebuiltins_toplevel_consts_7_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_ID(__call__),
    .co_qualname = & _sitebuiltins_toplevel_consts_7_consts_3_qualname._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_7_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x00\x6c\x00\x7d\x03\x02\x00\x7c\x03\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x69\x00\x7c\x02\xa4\x01\x8e\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
_sitebuiltins_toplevel_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__Helper._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_7_consts_1._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_7_consts_2.ob_base.ob_base,
            & _sitebuiltins_toplevel_consts_7_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_sitebuiltins_toplevel_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__repr__),
            &_Py_ID(__call__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
_sitebuiltins_toplevel_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x07\x05\x08\xf2\x12\x02\x05\x38\xf3\x06\x02\x05\x29",
};
static
    struct _PyCode_DEF(28)
_sitebuiltins_toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & _sitebuiltins_toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 88,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 531,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = & const_str__Helper._ascii.ob_base,
    .co_qualname = & const_str__Helper._ascii.ob_base,
    .co_linetable = & _sitebuiltins_toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
_sitebuiltins_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & _sitebuiltins_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & _sitebuiltins_toplevel_consts_3.ob_base.ob_base,
            & const_str_Quitter._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_5.ob_base.ob_base,
            & const_str__Printer._ascii.ob_base,
            & _sitebuiltins_toplevel_consts_7.ob_base.ob_base,
            & const_str__Helper._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
_sitebuiltins_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(object),
            & const_str_Quitter._ascii.ob_base,
            & const_str__Printer._ascii.ob_base,
            & const_str__Helper._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[53];
    }
_sitebuiltins_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 52,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x02\x01\x04\xf3\x14\x00\x01\x0b\xf4\x04\x0d\x01\x1f\x88\x66\xf4\x00\x0d\x01\x1f\xf4\x20\x38\x01\x1a\x88\x76\xf4\x00\x38\x01\x1a\xf4\x76\x01\x0f\x01\x29\x88\x66\xf5\x00\x0f\x01\x29",
};
static
    struct _PyCode_DEF(82)
_sitebuiltins_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 41,
    },
    .co_consts = & _sitebuiltins_toplevel_consts._object.ob_base.ob_base,
    .co_names = & _sitebuiltins_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 532,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & _sitebuiltins_toplevel_consts_3_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & _sitebuiltins_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x5a\x01\x02\x00\x47\x00\x64\x03\x84\x00\x64\x04\x65\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x03\x02\x00\x47\x00\x64\x05\x84\x00\x64\x06\x65\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x04\x02\x00\x47\x00\x64\x07\x84\x00\x64\x08\x65\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x05\x79\x02",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get__sitebuiltins_toplevel(void)
{
    return Py_NewRef((PyObject *) &_sitebuiltins_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[153];
    }
genericpath_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 152,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x50\x61\x74\x68\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x20\x63\x6f\x6d\x6d\x6f\x6e\x20\x74\x6f\x20\x6d\x6f\x72\x65\x20\x74\x68\x61\x6e\x20\x6f\x6e\x65\x20\x4f\x53\x0a\x44\x6f\x20\x6e\x6f\x74\x20\x75\x73\x65\x20\x64\x69\x72\x65\x63\x74\x6c\x79\x2e\x20\x20\x54\x68\x65\x20\x4f\x53\x20\x73\x70\x65\x63\x69\x66\x69\x63\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x69\x6d\x70\x6f\x72\x74\x20\x74\x68\x65\x20\x61\x70\x70\x72\x6f\x70\x72\x69\x61\x74\x65\x0a\x66\x75\x6e\x63\x74\x69\x6f\x6e\x73\x20\x66\x72\x6f\x6d\x20\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x74\x68\x65\x6d\x73\x65\x6c\x76\x65\x73\x2e\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_commonprefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "commonprefix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_exists = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exists",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_getatime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getatime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_getctime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getctime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_getmtime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getmtime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_getsize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getsize",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_isdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isdir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_isfile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isfile",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_islink = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "islink",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_samefile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "samefile",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_sameopenfile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sameopenfile",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_samestat = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "samestat",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_ALLOW_MISSING = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ALLOW_MISSING",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
genericpath_toplevel_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_commonprefix._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str_getatime._ascii.ob_base,
            & const_str_getctime._ascii.ob_base,
            & const_str_getmtime._ascii.ob_base,
            & const_str_getsize._ascii.ob_base,
            & const_str_isdir._ascii.ob_base,
            & const_str_isfile._ascii.ob_base,
            & const_str_islink._ascii.ob_base,
            & const_str_samefile._ascii.ob_base,
            & const_str_sameopenfile._ascii.ob_base,
            & const_str_samestat._ascii.ob_base,
            & const_str_ALLOW_MISSING._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[69];
    }
genericpath_toplevel_consts_4_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 68,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether a path exists.  Returns False for broken symbolic links",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
genericpath_toplevel_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & genericpath_toplevel_consts_4_consts_0._ascii.ob_base,
            Py_False,
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
genericpath_toplevel_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
genericpath_toplevel_consts_4_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen genericpath>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
genericpath_toplevel_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x08\x0a\x8f\x07\x89\x07\x90\x04\x8c\x0d\xf0\x06\x00\x0c\x10\xf8\xf4\x05\x00\x0d\x14\x94\x5a\xd0\x0b\x20\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
genericpath_toplevel_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x18\x00\x98\x0f\x2a\x03\xa9\x01\x2a\x03",
};
static
    struct _PyCode_DEF(90)
genericpath_toplevel_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 45,
    },
    .co_consts = & genericpath_toplevel_consts_4_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & genericpath_toplevel_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 16,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 533,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_exists._ascii.ob_base,
    .co_qualname = & const_str_exists._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
genericpath_toplevel_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether a path is a regular file",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & genericpath_toplevel_consts_5_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISREG = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISREG",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
genericpath_toplevel_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_S_ISREG._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
genericpath_toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x0d\x0f\x8f\x57\x89\x57\x90\x54\x8b\x5d\x88\x02\xf4\x06\x00\x0c\x10\x8f\x3c\x89\x3c\x98\x02\x9f\x0a\x99\x0a\xd3\x0b\x23\xd0\x04\x23\xf8\xf4\x05\x00\x0d\x14\x94\x5a\xd0\x0b\x20\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
genericpath_toplevel_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x36\x00\xb6\x0f\x41\x08\x03\xc1\x07\x01\x41\x08\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_st._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(150)
genericpath_toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 75,
    },
    .co_consts = & genericpath_toplevel_consts_5_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & genericpath_toplevel_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 27,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 534,
    .co_localsplusnames = & genericpath_toplevel_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_isfile._ascii.ob_base,
    .co_qualname = & const_str_isfile._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[61];
    }
genericpath_toplevel_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 60,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return true if the pathname refers to an existing directory.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & genericpath_toplevel_consts_6_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISDIR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISDIR",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
genericpath_toplevel_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_S_ISDIR._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
genericpath_toplevel_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x0d\x0f\x8f\x57\x89\x57\x90\x51\x8b\x5a\x88\x02\xf4\x06\x00\x0c\x10\x8f\x3c\x89\x3c\x98\x02\x9f\x0a\x99\x0a\xd3\x0b\x23\xd0\x04\x23\xf8\xf4\x05\x00\x0d\x14\x94\x5a\xd0\x0b\x20\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            & const_str_st._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(150)
genericpath_toplevel_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 75,
    },
    .co_consts = & genericpath_toplevel_consts_6_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & genericpath_toplevel_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 39,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 535,
    .co_localsplusnames = & genericpath_toplevel_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_isdir._ascii.ob_base,
    .co_qualname = & const_str_isdir._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
genericpath_toplevel_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether a path is a symbolic link",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & genericpath_toplevel_consts_7_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_lstat = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lstat",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISLNK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISLNK",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
genericpath_toplevel_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_lstat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_S_ISLNK._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[69];
    }
genericpath_toplevel_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 68,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x0d\x0f\x8f\x58\x89\x58\x90\x64\x8b\x5e\x88\x02\xf4\x06\x00\x0c\x10\x8f\x3c\x89\x3c\x98\x02\x9f\x0a\x99\x0a\xd3\x0b\x23\xd0\x04\x23\xf8\xf4\x05\x00\x0d\x14\x94\x5a\xa4\x1e\xd0\x0b\x30\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
genericpath_toplevel_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x36\x00\xb6\x14\x41\x0d\x03\xc1\x0c\x01\x41\x0d\x03",
};
static
    struct _PyCode_DEF(160)
genericpath_toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 80,
    },
    .co_consts = & genericpath_toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & genericpath_toplevel_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 51,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 536,
    .co_localsplusnames = & genericpath_toplevel_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_islink._ascii.ob_base,
    .co_qualname = & const_str_islink._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[50];
    }
genericpath_toplevel_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 49,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the size of a file, reported by os.stat().",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
genericpath_toplevel_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & genericpath_toplevel_consts_8_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
genericpath_toplevel_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_st_size._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
genericpath_toplevel_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0d\x8f\x37\x89\x37\x90\x38\xd3\x0b\x1c\xd7\x0b\x24\xd1\x0b\x24\xd0\x04\x24",
};
static
    struct _PyCode_DEF(64)
genericpath_toplevel_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & genericpath_toplevel_consts_8_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 60,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 537,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_39_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_getsize._ascii.ob_base,
    .co_qualname = & const_str_getsize._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[68];
    }
genericpath_toplevel_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 67,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the last modification time of a file, reported by os.stat().",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
genericpath_toplevel_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & genericpath_toplevel_consts_9_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
genericpath_toplevel_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_st_mtime._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
genericpath_toplevel_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0d\x8f\x37\x89\x37\x90\x38\xd3\x0b\x1c\xd7\x0b\x25\xd1\x0b\x25\xd0\x04\x25",
};
static
    struct _PyCode_DEF(64)
genericpath_toplevel_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & genericpath_toplevel_consts_9_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 65,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 538,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_39_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_getmtime._ascii.ob_base,
    .co_qualname = & const_str_getmtime._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[62];
    }
genericpath_toplevel_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 61,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the last access time of a file, reported by os.stat().",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
genericpath_toplevel_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & genericpath_toplevel_consts_10_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_st_atime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_atime",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
genericpath_toplevel_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_st_atime._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(64)
genericpath_toplevel_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & genericpath_toplevel_consts_10_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 70,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 539,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_39_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_getatime._ascii.ob_base,
    .co_qualname = & const_str_getatime._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[66];
    }
genericpath_toplevel_consts_11_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 65,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the metadata change time of a file, reported by os.stat().",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
genericpath_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & genericpath_toplevel_consts_11_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_st_ctime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_ctime",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
genericpath_toplevel_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_st_ctime._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(64)
genericpath_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & genericpath_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 75,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 540,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_39_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_getctime._ascii.ob_base,
    .co_qualname = & const_str_getctime._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[72];
    }
genericpath_toplevel_consts_12_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 71,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Given a list of pathnames, returns the longest common leading component",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
genericpath_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & genericpath_toplevel_consts_12_consts_0._ascii.ob_base,
            &_Py_STR(empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_min = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "min",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_enumerate = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "enumerate",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
genericpath_toplevel_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_list._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_map._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str_min._ascii.ob_base,
            & const_str_max._ascii.ob_base,
            & const_str_enumerate._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[114];
    }
genericpath_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 113,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe1\x0b\x0c\x90\x52\xf4\x0a\x00\x0c\x16\x90\x61\x98\x01\x91\x64\x9c\x54\xa4\x35\x98\x4d\xd4\x0b\x2a\xdc\x0c\x11\x94\x23\x94\x62\x97\x69\x91\x69\xa0\x11\xd3\x12\x23\xd3\x0c\x24\x88\x01\xdc\x09\x0c\x88\x51\x8b\x16\x80\x42\xdc\x09\x0c\x88\x51\x8b\x16\x80\x42\xdc\x10\x19\x98\x22\x96\x0d\x89\x04\x88\x01\x88\x31\xd8\x0b\x0c\x90\x02\x90\x31\x91\x05\x8b\x3a\xd8\x13\x15\x90\x62\x90\x71\x90\x36\x8a\x4d\xf0\x05\x00\x11\x1e\xf0\x06\x00\x0c\x0e\x80\x49",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_s1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "s1",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_s2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "s2",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
genericpath_toplevel_consts_12_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[109],
            & const_str_s1._ascii.ob_base,
            & const_str_s2._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            (PyObject *)&_Py_SINGLETON(strings).ascii[99],
        },
    },
};
static
    struct _PyCode_DEF(244)
genericpath_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 122,
    },
    .co_consts = & genericpath_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 81,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 541,
    .co_localsplusnames = & genericpath_toplevel_consts_12_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_commonprefix._ascii.ob_base,
    .co_qualname = & const_str_commonprefix._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x73\x01\x79\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\x19\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x23\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x14\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x7c\x04\x7c\x02\x7c\x03\x19\x00\x00\x00\x6b\x37\x00\x00\x73\x01\x8c\x0f\x7c\x01\x64\x03\x7c\x03\x1a\x00\x63\x02\x01\x00\x53\x00\x04\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
genericpath_toplevel_consts_13_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether two stat buffers reference the same file",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
genericpath_toplevel_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & genericpath_toplevel_consts_13_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_st_ino = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_ino",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_st_dev = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_dev",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_st_ino._ascii.ob_base,
            & const_str_st_dev._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
genericpath_toplevel_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0c\x0e\x8f\x49\x89\x49\x98\x12\x9f\x19\x99\x19\xd1\x0c\x22\xf2\x00\x01\x0d\x23\xd8\x0c\x0e\x8f\x49\x89\x49\x98\x12\x9f\x19\x99\x19\xd1\x0c\x22\xf0\x03\x01\x05\x24",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_13_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_s1._ascii.ob_base,
            & const_str_s2._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(106)
genericpath_toplevel_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 53,
    },
    .co_consts = & genericpath_toplevel_consts_13_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 99,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 542,
    .co_localsplusnames = & genericpath_toplevel_consts_13_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_samestat._ascii.ob_base,
    .co_qualname = & const_str_samestat._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x78\x01\x72\x19\x01\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[214];
    }
genericpath_toplevel_consts_14_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 213,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x65\x73\x74\x20\x77\x68\x65\x74\x68\x65\x72\x20\x74\x77\x6f\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x73\x20\x72\x65\x66\x65\x72\x65\x6e\x63\x65\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x61\x63\x74\x75\x61\x6c\x20\x66\x69\x6c\x65\x20\x6f\x72\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x69\x73\x20\x64\x65\x74\x65\x72\x6d\x69\x6e\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x64\x65\x76\x69\x63\x65\x20\x6e\x75\x6d\x62\x65\x72\x20\x61\x6e\x64\x20\x69\x2d\x6e\x6f\x64\x65\x20\x6e\x75\x6d\x62\x65\x72\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x72\x61\x69\x73\x65\x73\x20\x61\x6e\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x69\x66\x20\x61\x6e\x20\x6f\x73\x2e\x73\x74\x61\x74\x28\x29\x20\x63\x61\x6c\x6c\x20\x6f\x6e\x20\x65\x69\x74\x68\x65\x72\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x66\x61\x69\x6c\x73\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
genericpath_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & genericpath_toplevel_consts_14_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
genericpath_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_samestat._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[44];
    }
genericpath_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 43,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x0a\x0c\x8f\x17\x89\x17\x90\x12\x8b\x1b\x80\x42\xdc\x09\x0b\x8f\x17\x89\x17\x90\x12\x8b\x1b\x80\x42\xdc\x0b\x13\x90\x42\x98\x02\xd3\x0b\x1b\xd0\x04\x1b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_f1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "f1",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_f2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "f2",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
genericpath_toplevel_consts_14_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_f1._ascii.ob_base,
            & const_str_f2._ascii.ob_base,
            & const_str_s1._ascii.ob_base,
            & const_str_s2._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(110)
genericpath_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & genericpath_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 106,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 543,
    .co_localsplusnames = & genericpath_toplevel_consts_14_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_samefile._ascii.ob_base,
    .co_qualname = & const_str_samefile._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[59];
    }
genericpath_toplevel_consts_15_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 58,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether two open file objects reference the same file",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
genericpath_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & genericpath_toplevel_consts_15_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_fstat = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fstat",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
genericpath_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fstat._ascii.ob_base,
            & const_str_samestat._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[42];
    }
genericpath_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 41,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x09\x0b\x8f\x18\x89\x18\x90\x23\x8b\x1d\x80\x42\xdc\x09\x0b\x8f\x18\x89\x18\x90\x23\x8b\x1d\x80\x42\xdc\x0b\x13\x90\x42\x98\x02\xd3\x0b\x1b\xd0\x04\x1b",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_fp1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fp1",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_fp2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fp2",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
genericpath_toplevel_consts_15_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_fp1._ascii.ob_base,
            & const_str_fp2._ascii.ob_base,
            & const_str_s1._ascii.ob_base,
            & const_str_s2._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(110)
genericpath_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & genericpath_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 119,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 544,
    .co_localsplusnames = & genericpath_toplevel_consts_15_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_sameopenfile._ascii.ob_base,
    .co_qualname = & const_str_sameopenfile._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[165];
    }
genericpath_toplevel_consts_16_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 164,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x70\x6c\x69\x74\x20\x74\x68\x65\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x66\x72\x6f\x6d\x20\x61\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x45\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x69\x73\x20\x65\x76\x65\x72\x79\x74\x68\x69\x6e\x67\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x6c\x61\x73\x74\x20\x64\x6f\x74\x20\x74\x6f\x20\x74\x68\x65\x20\x65\x6e\x64\x2c\x20\x69\x67\x6e\x6f\x72\x69\x6e\x67\x0a\x20\x20\x20\x20\x6c\x65\x61\x64\x69\x6e\x67\x20\x64\x6f\x74\x73\x2e\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x22\x28\x72\x6f\x6f\x74\x2c\x20\x65\x78\x74\x29\x22\x3b\x20\x65\x78\x74\x20\x6d\x61\x79\x20\x62\x65\x20\x65\x6d\x70\x74\x79\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
genericpath_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & genericpath_toplevel_consts_16_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_rfind._ascii.ob_base,
            & const_str_max._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__splitext = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_splitext",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[163];
    }
genericpath_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 162,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0e\x00\x10\x11\x8f\x77\x89\x77\x90\x73\x8b\x7c\x80\x48\xd9\x07\x0d\xd8\x16\x17\x97\x67\x91\x67\x98\x66\x93\x6f\x88\x0b\xdc\x13\x16\x90\x78\xa0\x1b\xd3\x13\x2d\x88\x08\xe0\x0f\x10\x8f\x77\x89\x77\x90\x76\x8b\x7f\x80\x48\xd8\x07\x0f\x90\x28\xd2\x07\x1a\xe0\x18\x20\xa0\x31\x99\x0c\x88\x0d\xd8\x0e\x1b\x98\x68\xd2\x0e\x26\xd8\x0f\x10\x90\x1d\x98\x7d\xa8\x51\x99\x7f\xd0\x0f\x2f\xb0\x36\xd2\x0f\x39\xd8\x17\x18\x98\x19\x98\x28\x90\x7c\xa0\x51\xa0\x78\xa0\x79\xa0\x5c\xd0\x17\x31\xd0\x10\x31\xd8\x0c\x19\x98\x51\xd1\x0c\x1e\x88\x4d\xf0\x07\x00\x0f\x1c\x98\x68\xd3\x0e\x26\xf0\x0a\x00\x0c\x0d\x88\x61\x90\x02\x90\x11\x88\x65\x88\x38\x80\x4f",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_altsep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "altsep",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_extsep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "extsep",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_sepIndex = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sepIndex",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_altsepIndex = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "altsepIndex",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_dotIndex = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dotIndex",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_filenameIndex = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "filenameIndex",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
genericpath_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            & const_str_altsep._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
            & const_str_sepIndex._ascii.ob_base,
            & const_str_altsepIndex._ascii.ob_base,
            & const_str_dotIndex._ascii.ob_base,
            & const_str_filenameIndex._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(240)
genericpath_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 120,
    },
    .co_consts = & genericpath_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 133,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 545,
    .co_localsplusnames = & genericpath_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str__splitext._ascii.ob_base,
    .co_qualname = & const_str__splitext._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x02\x72\x1d\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x05\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x7c\x04\x6b\x44\x00\x00\x72\x2a\x7c\x04\x64\x01\x7a\x00\x00\x00\x7d\x07\x7c\x07\x7c\x06\x6b\x02\x00\x00\x72\x20\x7c\x00\x7c\x07\x7c\x07\x64\x01\x7a\x00\x00\x00\x1a\x00\x7c\x03\x6b\x37\x00\x00\x72\x0a\x7c\x00\x64\x02\x7c\x06\x1a\x00\x7c\x00\x7c\x06\x64\x02\x1a\x00\x66\x02\x53\x00\x7c\x07\x64\x01\x7a\x0d\x00\x00\x7d\x07\x7c\x07\x7c\x06\x6b\x02\x00\x00\x72\x01\x8c\x20\x7c\x00\x7c\x00\x64\x02\x64\x03\x1a\x00\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[60];
    }
genericpath_toplevel_consts_17_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 59,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "() argument must be str, bytes, or os.PathLike object, not ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
genericpath_toplevel_consts_17_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Can't mix strings and bytes in path components",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
genericpath_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            Py_False,
            Py_True,
            & genericpath_toplevel_consts_17_consts_3._ascii.ob_base,
            & genericpath_toplevel_consts_17_consts_4._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
genericpath_toplevel_consts_17_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            &_Py_ID(bytes),
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(__class__),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__check_arg_types = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_check_arg_types",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[130];
    }
genericpath_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 129,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x18\x1d\xd0\x04\x1d\x80\x46\x88\x58\xdb\x0d\x11\x88\x01\xdc\x0b\x15\x90\x61\x9c\x13\xd4\x0b\x1d\xd8\x15\x19\x89\x46\xdc\x0d\x17\x98\x01\x9c\x35\xd4\x0d\x21\xd8\x17\x1b\x89\x48\xe4\x12\x1b\x98\x78\x98\x6a\xf0\x00\x01\x29\x37\xd8\x37\x38\xb7\x7b\xb1\x7b\xd7\x37\x4b\xd1\x37\x4b\xd0\x36\x4e\xf0\x03\x01\x1d\x50\x01\xf3\x00\x01\x13\x51\x01\xd8\x56\x5a\xf0\x03\x01\x0d\x5b\x01\xf0\x0d\x00\x0e\x12\xf1\x10\x00\x08\x0e\x91\x28\xdc\x0e\x17\xd0\x18\x48\xd3\x0e\x49\xc8\x74\xd0\x08\x53\xf0\x03\x00\x13\x1b\x80\x76",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_funcname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "funcname",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_hasstr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hasstr",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_hasbytes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "hasbytes",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
genericpath_toplevel_consts_17_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_funcname._ascii.ob_base,
            &_Py_ID(args),
            & const_str_hasstr._ascii.ob_base,
            & const_str_hasbytes._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
        },
    },
};
static
    struct _PyCode_DEF(208)
genericpath_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 104,
    },
    .co_consts = & genericpath_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_17_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 156,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 546,
    .co_localsplusnames = & genericpath_toplevel_consts_17_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str__check_arg_types._ascii.ob_base,
    .co_qualname = & const_str__check_arg_types._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x78\x01\x7d\x02\x7d\x03\x7c\x01\x44\x00\x5d\x4c\x00\x00\x7d\x04\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x03\x64\x02\x7d\x02\x8c\x16\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x03\x64\x02\x7d\x03\x8c\x29\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x9b\x00\x64\x03\x7c\x04\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x82\x02\x04\x00\x7c\x02\x72\x0f\x7c\x03\x72\x0c\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x82\x02\x79\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[37];
    }
genericpath_toplevel_consts_18_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 36,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Special value for use in realpath().",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
genericpath_toplevel_consts_18_consts_2_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "os.path.ALLOW_MISSING",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_18_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & genericpath_toplevel_consts_18_consts_2_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
genericpath_toplevel_consts_18_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ALLOW_MISSING.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
genericpath_toplevel_consts_18_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x26",
};
static
    struct _PyCode_DEF(4)
genericpath_toplevel_consts_18_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & genericpath_toplevel_consts_18_consts_2_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 173,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 547,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & genericpath_toplevel_consts_18_consts_2_qualname._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_18_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
genericpath_toplevel_consts_18_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(__class__),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
genericpath_toplevel_consts_18_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ALLOW_MISSING.__reduce__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
genericpath_toplevel_consts_18_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\x8f\x7e\x89\x7e\xd7\x0f\x26\xd1\x0f\x26\xd0\x08\x26",
};
static
    struct _PyCode_DEF(46)
genericpath_toplevel_consts_18_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_18_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 175,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 548,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = &_Py_ID(__reduce__),
    .co_qualname = & genericpath_toplevel_consts_18_consts_3_qualname._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_18_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
genericpath_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_ALLOW_MISSING._ascii.ob_base,
            & genericpath_toplevel_consts_18_consts_1._ascii.ob_base,
            & genericpath_toplevel_consts_18_consts_2.ob_base.ob_base,
            & genericpath_toplevel_consts_18_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
genericpath_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__repr__),
            &_Py_ID(__reduce__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
genericpath_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x2e\xf2\x02\x01\x05\x27\xf3\x04\x01\x05\x27",
};
static
    struct _PyCode_DEF(28)
genericpath_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & genericpath_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 170,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 549,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = & const_str_ALLOW_MISSING._ascii.ob_base,
    .co_qualname = & const_str_ALLOW_MISSING._ascii.ob_base,
    .co_linetable = & genericpath_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[20];
        }_object;
    }
genericpath_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 20,
        },
        .ob_item = {
            & genericpath_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & genericpath_toplevel_consts_3._object.ob_base.ob_base,
            & genericpath_toplevel_consts_4.ob_base.ob_base,
            & genericpath_toplevel_consts_5.ob_base.ob_base,
            & genericpath_toplevel_consts_6.ob_base.ob_base,
            & genericpath_toplevel_consts_7.ob_base.ob_base,
            & genericpath_toplevel_consts_8.ob_base.ob_base,
            & genericpath_toplevel_consts_9.ob_base.ob_base,
            & genericpath_toplevel_consts_10.ob_base.ob_base,
            & genericpath_toplevel_consts_11.ob_base.ob_base,
            & genericpath_toplevel_consts_12.ob_base.ob_base,
            & genericpath_toplevel_consts_13.ob_base.ob_base,
            & genericpath_toplevel_consts_14.ob_base.ob_base,
            & genericpath_toplevel_consts_15.ob_base.ob_base,
            & genericpath_toplevel_consts_16.ob_base.ob_base,
            & genericpath_toplevel_consts_17.ob_base.ob_base,
            & genericpath_toplevel_consts_18.ob_base.ob_base,
            & const_str_ALLOW_MISSING._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[21];
        }_object;
    }
genericpath_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 21,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_os._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            &_Py_ID(__all__),
            & const_str_exists._ascii.ob_base,
            & const_str_isfile._ascii.ob_base,
            & const_str_isdir._ascii.ob_base,
            & const_str_islink._ascii.ob_base,
            & const_str_getsize._ascii.ob_base,
            & const_str_getmtime._ascii.ob_base,
            & const_str_getatime._ascii.ob_base,
            & const_str_getctime._ascii.ob_base,
            & const_str_commonprefix._ascii.ob_base,
            & const_str_samestat._ascii.ob_base,
            & const_str_samefile._ascii.ob_base,
            & const_str_sameopenfile._ascii.ob_base,
            & const_str__splitext._ascii.ob_base,
            & const_str__check_arg_types._ascii.ob_base,
            &_Py_ID(object),
            &_Py_ID(__new__),
            & const_str_ALLOW_MISSING._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[126];
    }
genericpath_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 125,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x04\x01\x04\xf3\x0a\x00\x01\x0a\xdb\x00\x0b\xf2\x04\x02\x0b\x28\x80\x07\xf2\x0e\x06\x01\x10\xf2\x16\x06\x01\x24\xf2\x18\x06\x01\x24\xf2\x18\x06\x01\x24\xf2\x12\x02\x01\x25\xf2\x0a\x02\x01\x26\xf2\x0a\x02\x01\x26\xf2\x0a\x02\x01\x26\xf2\x0c\x0e\x01\x0e\xf2\x24\x03\x01\x24\xf2\x0e\x08\x01\x1c\xf2\x1a\x04\x01\x1c\xf2\x1c\x15\x01\x14\xf2\x2e\x0b\x01\x54\x01\xf0\x1c\x00\x02\x08\x87\x1e\x81\x1e\xf7\x02\x05\x01\x27\xf0\x00\x05\x01\x27\xf3\x03\x00\x02\x10\xf1\x02\x05\x01\x27",
};
static
    struct _PyCode_DEF(166)
genericpath_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 83,
    },
    .co_consts = & genericpath_toplevel_consts._object.ob_base.ob_base,
    .co_names = & genericpath_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 550,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & genericpath_toplevel_consts_4_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & genericpath_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x5a\x01\x64\x01\x64\x02\x6c\x02\x5a\x02\x67\x00\x64\x03\xa2\x01\x5a\x03\x64\x04\x84\x00\x5a\x04\x64\x05\x84\x00\x5a\x05\x64\x06\x84\x00\x5a\x06\x64\x07\x84\x00\x5a\x07\x64\x08\x84\x00\x5a\x08\x64\x09\x84\x00\x5a\x09\x64\x0a\x84\x00\x5a\x0a\x64\x0b\x84\x00\x5a\x0b\x64\x0c\x84\x00\x5a\x0c\x64\x0d\x84\x00\x5a\x0d\x64\x0e\x84\x00\x5a\x0e\x64\x0f\x84\x00\x5a\x0f\x64\x10\x84\x00\x5a\x10\x64\x11\x84\x00\x5a\x11\x65\x12\x6a\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x47\x00\x64\x12\x84\x00\x64\x13\xab\x02\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x14\x79\x02",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_genericpath_toplevel(void)
{
    return Py_NewRef((PyObject *) &genericpath_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[145];
    }
ntpath_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 144,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6d\x6d\x6f\x6e\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x6d\x61\x6e\x69\x70\x75\x6c\x61\x74\x69\x6f\x6e\x73\x2c\x20\x57\x69\x6e\x64\x6f\x77\x73\x4e\x54\x2f\x39\x35\x20\x76\x65\x72\x73\x69\x6f\x6e\x2e\x0a\x0a\x49\x6e\x73\x74\x65\x61\x64\x20\x6f\x66\x20\x69\x6d\x70\x6f\x72\x74\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x64\x69\x72\x65\x63\x74\x6c\x79\x2c\x20\x69\x6d\x70\x6f\x72\x74\x20\x6f\x73\x20\x61\x6e\x64\x20\x72\x65\x66\x65\x72\x20\x74\x6f\x20\x74\x68\x69\x73\x0a\x6d\x6f\x64\x75\x6c\x65\x20\x61\x73\x20\x6f\x73\x2e\x70\x61\x74\x68\x2e\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
ntpath_toplevel_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "..",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
ntpath_toplevel_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".;C:\\bin",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_nul = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "nul",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_normcase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "normcase",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_isabs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isabs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_splitdrive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "splitdrive",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_splitroot = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "splitroot",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_splitext = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "splitext",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_lexists = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lexists",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_ismount = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ismount",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_expanduser = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "expanduser",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_expandvars = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "expandvars",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_normpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "normpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_abspath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abspath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_curdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "curdir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_pardir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pardir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_pathsep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pathsep",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_defpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "defpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_devnull = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "devnull",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_realpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "realpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str_supports_unicode_filenames = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "supports_unicode_filenames",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_relpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "relpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_commonpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "commonpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_isjunction = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isjunction",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[41];
        }_object;
    }
ntpath_toplevel_consts_11 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 41,
        },
        .ob_item = {
            & const_str_normcase._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            &_Py_ID(join),
            & const_str_splitdrive._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            & const_str_splitext._ascii.ob_base,
            & const_str_basename._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
            & const_str_commonprefix._ascii.ob_base,
            & const_str_getsize._ascii.ob_base,
            & const_str_getmtime._ascii.ob_base,
            & const_str_getatime._ascii.ob_base,
            & const_str_getctime._ascii.ob_base,
            & const_str_islink._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str_lexists._ascii.ob_base,
            & const_str_isdir._ascii.ob_base,
            & const_str_isfile._ascii.ob_base,
            & const_str_ismount._ascii.ob_base,
            & const_str_expanduser._ascii.ob_base,
            & const_str_expandvars._ascii.ob_base,
            & const_str_normpath._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pathsep._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            & const_str_altsep._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
            & const_str_devnull._ascii.ob_base,
            & const_str_realpath._ascii.ob_base,
            & const_str_supports_unicode_filenames._ascii.ob_base,
            & const_str_relpath._ascii.ob_base,
            & const_str_samefile._ascii.ob_base,
            & const_str_sameopenfile._ascii.ob_base,
            & const_str_samestat._ascii.ob_base,
            & const_str_commonpath._ascii.ob_base,
            & const_str_isjunction._ascii.ob_base,
            & const_str_ALLOW_MISSING._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
ntpath_toplevel_consts_12_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\\/",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
ntpath_toplevel_consts_12_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\/",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
ntpath_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & ntpath_toplevel_consts_12_consts_1.ob_base.ob_base,
            & ntpath_toplevel_consts_12_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
ntpath_toplevel_consts_12_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen ntpath>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__get_bothseps = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_bothseps",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
ntpath_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x0f\x15\xe0\x0f\x14",
};
static
    struct _PyCode_DEF(38)
ntpath_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & ntpath_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 36,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 551,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str__get_bothseps._ascii.ob_base,
    .co_qualname = & const_str__get_bothseps._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x01\x79\x01\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_LCMapStringEx = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LCMapStringEx",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_LOCALE_NAME_INVARIANT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LOCALE_NAME_INVARIANT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_LCMAP_LOWERCASE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LCMAP_LOWERCASE",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
ntpath_toplevel_consts_13 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_LCMapStringEx._ascii.ob_base,
            & const_str_LOCALE_NAME_INVARIANT._ascii.ob_base,
            & const_str_LCMAP_LOWERCASE._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[111];
    }
ntpath_toplevel_consts_14_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 110,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4e\x6f\x72\x6d\x61\x6c\x69\x7a\x65\x20\x63\x61\x73\x65\x20\x6f\x66\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4d\x61\x6b\x65\x73\x20\x61\x6c\x6c\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x73\x20\x6c\x6f\x77\x65\x72\x63\x61\x73\x65\x20\x61\x6e\x64\x20\x61\x6c\x6c\x20\x73\x6c\x61\x73\x68\x65\x73\x20\x69\x6e\x74\x6f\x20\x62\x61\x63\x6b\x73\x6c\x61\x73\x68\x65\x73\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_surrogateescape = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "surrogateescape",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
ntpath_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & ntpath_toplevel_consts_14_consts_0._ascii.ob_base,
            & const_str_surrogateescape._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_getfilesystemencoding = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getfilesystemencoding",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__LCMapStringEx = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LCMapStringEx",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str__LOCALE_NAME_INVARIANT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LOCALE_NAME_INVARIANT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__LCMAP_LOWERCASE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LCMAP_LOWERCASE",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
ntpath_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_sys._ascii.ob_base,
            & const_str_getfilesystemencoding._ascii.ob_base,
            &_Py_ID(decode),
            &_Py_ID(replace),
            & const_str__LCMapStringEx._ascii.ob_base,
            & const_str__LOCALE_NAME_INVARIANT._ascii.ob_base,
            & const_str__LCMAP_LOWERCASE._ascii.ob_base,
            &_Py_ID(encode),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[149];
    }
ntpath_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 148,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x0d\x0f\x8f\x49\x89\x49\x90\x61\x8b\x4c\x88\x01\xd9\x0f\x10\xd8\x13\x14\x88\x48\xdc\x0b\x15\x90\x61\x9c\x15\xd4\x0b\x1f\xdc\x17\x1a\xd7\x17\x30\xd1\x17\x30\xd3\x17\x32\x88\x48\xd8\x10\x11\x97\x08\x91\x08\x98\x18\xd0\x23\x34\xd3\x10\x35\xd7\x10\x3d\xd1\x10\x3d\xb8\x63\xc0\x34\xd3\x10\x48\x88\x41\xdc\x10\x1e\xd4\x1f\x35\xdc\x1f\x2f\xb0\x11\xf3\x03\x01\x11\x34\x88\x41\xe0\x13\x14\x97\x38\x91\x38\x98\x48\xd0\x26\x37\xd3\x13\x38\xd0\x0c\x38\xe4\x13\x21\xd4\x22\x38\xdc\x22\x32\xd8\x22\x23\xa7\x29\xa1\x29\xa8\x43\xb0\x14\xd3\x22\x36\xf3\x05\x02\x14\x38\xf0\x00\x02\x0d\x38",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_14_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            &_Py_ID(encoding),
        },
    },
};
static
    struct _PyCode_DEF(344)
ntpath_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 172,
    },
    .co_consts = & ntpath_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 52,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 552,
    .co_localsplusnames = & ntpath_toplevel_consts_14_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_normcase._ascii.ob_base,
    .co_qualname = & const_str_normcase._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x73\x02\x7c\x00\x53\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x5d\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
ntpath_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & ntpath_toplevel_consts_14_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_fsencode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fsencode",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_fsdecode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fsdecode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
ntpath_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_fsencode._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            &_Py_ID(replace),
            & const_str_lower._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[99];
    }
ntpath_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 98,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x0d\x0f\x8f\x49\x89\x49\x90\x61\x8b\x4c\x88\x01\xdc\x0b\x15\x90\x61\x9c\x15\xd4\x0b\x1f\xdc\x13\x15\x97\x3b\x91\x3b\x9c\x72\x9f\x7b\x99\x7b\xa8\x31\x9b\x7e\xd7\x1f\x35\xd1\x1f\x35\xb0\x63\xb8\x34\xd3\x1f\x40\xd7\x1f\x46\xd1\x1f\x46\xd3\x1f\x48\xd3\x13\x49\xd0\x0c\x49\xd8\x0f\x10\x8f\x79\x89\x79\x98\x13\x98\x64\xd3\x0f\x23\xd7\x0f\x29\xd1\x0f\x29\xd3\x0f\x2b\xd0\x08\x2b",
};
static
    struct _PyCode_DEF(280)
ntpath_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 140,
    },
    .co_consts = & ntpath_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 71,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 553,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_normcase._ascii.ob_base,
    .co_qualname = & const_str_normcase._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x46\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
ntpath_toplevel_consts_16_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether a path is absolute",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
ntpath_toplevel_consts_16_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = ":\\",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
ntpath_toplevel_consts_16_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ":\\",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
ntpath_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & ntpath_toplevel_consts_16_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            & ntpath_toplevel_consts_16_consts_3.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            & ntpath_toplevel_consts_16_consts_6._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_True,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
ntpath_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            &_Py_ID(replace),
            & const_str_startswith._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[111];
    }
ntpath_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 110,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x0a\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x07\x11\x90\x21\x94\x55\xd4\x07\x1b\xd8\x0e\x13\x88\x03\xd8\x11\x15\x88\x06\xd8\x14\x1a\x89\x09\xe0\x0e\x12\x88\x03\xd8\x11\x14\x88\x06\xd8\x14\x19\x88\x09\xd8\x08\x09\x88\x22\x88\x31\x88\x05\x8f\x0d\x89\x0d\x90\x66\x98\x63\xd3\x08\x22\x80\x41\xf0\x06\x00\x08\x09\x87\x7c\x81\x7c\x90\x43\xd4\x07\x18\x98\x41\x9f\x4c\x99\x4c\xa8\x19\xb0\x41\xd4\x1c\x36\xd8\x0f\x13\xd8\x0b\x10",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_colon_sep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "colon_sep",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
ntpath_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            &_Py_ID(sep),
            & const_str_altsep._ascii.ob_base,
            & const_str_colon_sep._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(218)
ntpath_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 109,
    },
    .co_consts = & ntpath_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 88,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 554,
    .co_localsplusnames = & ntpath_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_isabs._ascii.ob_base,
    .co_qualname = & const_str_isabs._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x07\x64\x01\x7d\x01\x64\x02\x7d\x02\x64\x03\x7d\x03\x6e\x06\x64\x04\x7d\x01\x64\x05\x7d\x02\x64\x06\x7d\x03\x7c\x00\x64\x07\x64\x08\x1a\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x73\x12\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x09\xab\x02\x00\x00\x00\x00\x00\x00\x72\x01\x79\x0a\x79\x0b",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
ntpath_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            & ntpath_toplevel_consts_12_consts_1.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[58]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            & ntpath_toplevel_consts_12_consts_2._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[58],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_BytesWarning = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "BytesWarning",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_genericpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "genericpath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
ntpath_toplevel_consts_17_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_splitroot._ascii.ob_base,
            & const_str_map._ascii.ob_base,
            & const_str_lower._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_BytesWarning._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            & const_str__check_arg_types._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[352];
    }
ntpath_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 351,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0b\x0d\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x0e\x13\x88\x03\xd8\x0f\x15\x88\x04\xd8\x10\x14\x89\x05\xe0\x0e\x12\x88\x03\xd8\x0f\x14\x88\x04\xd8\x10\x13\x88\x05\xf0\x02\x21\x05\x0e\xd9\x0f\x14\xd8\x0c\x10\x90\x12\x90\x21\x88\x48\x90\x73\x8a\x4e\xdc\x31\x3a\xb8\x34\xb3\x1f\xd1\x08\x2e\x88\x0c\x90\x6b\xa0\x3b\xdc\x11\x14\x94\x52\x97\x59\x91\x59\xa0\x05\xd6\x11\x26\x88\x41\xdc\x26\x2f\xb0\x01\xa3\x6c\xd1\x0c\x23\x88\x47\x90\x56\x98\x56\xd9\x0f\x15\xe1\x13\x1a\xa1\x2c\xd8\x23\x2a\x90\x4c\xd8\x1e\x24\x90\x0b\xd8\x1e\x24\x90\x0b\xd8\x10\x18\xd9\x11\x18\x98\x57\xa8\x0c\xd2\x1d\x34\xd8\x13\x1a\x97\x3d\x91\x3d\x93\x3f\xa0\x6c\xd7\x26\x38\xd1\x26\x38\xd3\x26\x3a\xd2\x13\x3a\xe0\x23\x2a\x90\x4c\xd8\x22\x28\x90\x4b\xd8\x22\x28\x90\x4b\xd8\x14\x1c\xe0\x1f\x26\x90\x0c\xe1\x0f\x1a\x98\x7b\xa8\x32\x99\x7f\xb0\x64\xd1\x1f\x3a\xd8\x1e\x29\xa8\x43\xd1\x1e\x2f\x90\x0b\xd8\x1a\x25\xa8\x06\xd1\x1a\x2e\x89\x4b\xf0\x2b\x00\x12\x27\xf1\x2e\x00\x0d\x18\xa1\x0b\xd9\x0c\x18\x98\x5c\xa8\x22\xa8\x23\xd0\x1d\x2e\xb0\x65\xb8\x64\xb1\x6c\xd1\x1d\x42\xd8\x13\x1f\xa0\x23\xd1\x13\x25\xa8\x0b\xd1\x13\x33\xd0\x0c\x33\xd8\x0f\x1b\x98\x6b\xd1\x0f\x29\xa8\x4b\xd1\x0f\x37\xd0\x08\x37\xf8\xdc\x0c\x15\x94\x7e\xa4\x7c\xd0\x0b\x34\xf2\x00\x02\x05\x0e\xdc\x08\x13\xd7\x08\x24\xd1\x08\x24\xa0\x56\xa8\x54\xd0\x08\x3a\xb0\x45\xd3\x08\x3a\xd8\x08\x0d\xf0\x05\x02\x05\x0e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
ntpath_toplevel_consts_17_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xb4\x42\x2f\x43\x2c\x00\xc3\x24\x07\x43\x2c\x00\xc3\x2c\x2d\x44\x19\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_paths = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "paths",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_seps = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "seps",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_colon = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "colon",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_result_drive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "result_drive",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_result_root = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "result_root",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_result_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "result_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_p_drive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "p_drive",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_p_root = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "p_root",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_p_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "p_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
ntpath_toplevel_consts_17_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_paths._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_seps._ascii.ob_base,
            & const_str_colon._ascii.ob_base,
            & const_str_result_drive._ascii.ob_base,
            & const_str_result_root._ascii.ob_base,
            & const_str_result_path._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            & const_str_p_drive._ascii.ob_base,
            & const_str_p_root._ascii.ob_base,
            & const_str_p_path._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(568)
ntpath_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 284,
    },
    .co_consts = & ntpath_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_17_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_17_exceptiontable.ob_base.ob_base,
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 108,
    .co_nlocalsplus = 12,
    .co_nlocals = 12,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 555,
    .co_localsplusnames = & ntpath_toplevel_consts_17_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_36_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = &_Py_ID(join),
    .co_qualname = &_Py_ID(join),
    .co_linetable = & ntpath_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x07\x64\x01\x7d\x02\x64\x02\x7d\x03\x64\x03\x7d\x04\x6e\x06\x64\x04\x7d\x02\x64\x05\x7d\x03\x64\x06\x7d\x04\x09\x00\x7c\x01\x73\x08\x7c\x00\x64\x00\x64\x07\x1a\x00\x7c\x02\x7a\x00\x00\x00\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x05\x7d\x06\x7d\x07\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x62\x00\x00\x7d\x08\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x09\x7d\x0a\x7d\x0b\x7c\x0a\x72\x0b\x7c\x09\x73\x02\x7c\x05\x73\x02\x7c\x09\x7d\x05\x7c\x0a\x7d\x06\x7c\x0b\x7d\x07\x8c\x1f\x7c\x09\x72\x2f\x7c\x09\x7c\x05\x6b\x37\x00\x00\x72\x2a\x7c\x09\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x07\x7c\x09\x7d\x05\x7c\x0a\x7d\x06\x7c\x0b\x7d\x07\x8c\x4e\x7c\x09\x7d\x05\x7c\x07\x72\x0c\x7c\x07\x64\x08\x19\x00\x00\x00\x7c\x03\x76\x01\x72\x05\x7c\x07\x7c\x02\x7a\x00\x00\x00\x7d\x07\x7c\x07\x7c\x0b\x7a\x00\x00\x00\x7d\x07\x8c\x64\x04\x00\x7c\x07\x72\x16\x7c\x06\x73\x14\x7c\x05\x72\x12\x7c\x05\x64\x08\x64\x00\x1a\x00\x7c\x04\x7c\x03\x7a\x00\x00\x00\x76\x01\x72\x08\x7c\x05\x7c\x02\x7a\x00\x00\x00\x7c\x07\x7a\x00\x00\x00\x53\x00\x7c\x05\x7c\x06\x7a\x00\x00\x00\x7c\x07\x7a\x00\x00\x00\x53\x00\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x19\x01\x00\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\x7c\x00\x67\x02\x7c\x01\xa2\x01\xad\x06\x8e\x00\x01\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[731];
    }
ntpath_toplevel_consts_18_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 730,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x70\x6c\x69\x74\x20\x61\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x69\x6e\x74\x6f\x20\x64\x72\x69\x76\x65\x2f\x55\x4e\x43\x20\x73\x68\x61\x72\x65\x70\x6f\x69\x6e\x74\x20\x61\x6e\x64\x20\x72\x65\x6c\x61\x74\x69\x76\x65\x20\x70\x61\x74\x68\x20\x73\x70\x65\x63\x69\x66\x69\x65\x72\x73\x2e\x0a\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x61\x20\x32\x2d\x74\x75\x70\x6c\x65\x20\x28\x64\x72\x69\x76\x65\x5f\x6f\x72\x5f\x75\x6e\x63\x2c\x20\x70\x61\x74\x68\x29\x3b\x20\x65\x69\x74\x68\x65\x72\x20\x70\x61\x72\x74\x20\x6d\x61\x79\x20\x62\x65\x20\x65\x6d\x70\x74\x79\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x79\x6f\x75\x20\x61\x73\x73\x69\x67\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x73\x75\x6c\x74\x20\x3d\x20\x73\x70\x6c\x69\x74\x64\x72\x69\x76\x65\x28\x70\x29\x0a\x20\x20\x20\x20\x49\x74\x20\x69\x73\x20\x61\x6c\x77\x61\x79\x73\x20\x74\x72\x75\x65\x20\x74\x68\x61\x74\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x65\x73\x75\x6c\x74\x5b\x30\x5d\x20\x2b\x20\x72\x65\x73\x75\x6c\x74\x5b\x31\x5d\x20\x3d\x3d\x20\x70\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x63\x6f\x6e\x74\x61\x69\x6e\x65\x64\x20\x61\x20\x64\x72\x69\x76\x65\x20\x6c\x65\x74\x74\x65\x72\x2c\x20\x64\x72\x69\x76\x65\x5f\x6f\x72\x5f\x75\x6e\x63\x20\x77\x69\x6c\x6c\x20\x63\x6f\x6e\x74\x61\x69\x6e\x20\x65\x76\x65\x72\x79\x74\x68\x69\x6e\x67\x0a\x20\x20\x20\x20\x75\x70\x20\x74\x6f\x20\x61\x6e\x64\x20\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x63\x6f\x6c\x6f\x6e\x2e\x20\x20\x65\x2e\x67\x2e\x20\x73\x70\x6c\x69\x74\x64\x72\x69\x76\x65\x28\x22\x63\x3a\x2f\x64\x69\x72\x22\x29\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x28\x22\x63\x3a\x22\x2c\x20\x22\x2f\x64\x69\x72\x22\x29\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x63\x6f\x6e\x74\x61\x69\x6e\x65\x64\x20\x61\x20\x55\x4e\x43\x20\x70\x61\x74\x68\x2c\x20\x74\x68\x65\x20\x64\x72\x69\x76\x65\x5f\x6f\x72\x5f\x75\x6e\x63\x20\x77\x69\x6c\x6c\x20\x63\x6f\x6e\x74\x61\x69\x6e\x20\x74\x68\x65\x20\x68\x6f\x73\x74\x20\x6e\x61\x6d\x65\x0a\x20\x20\x20\x20\x61\x6e\x64\x20\x73\x68\x61\x72\x65\x20\x75\x70\x20\x74\x6f\x20\x62\x75\x74\x20\x6e\x6f\x74\x20\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x74\x68\x65\x20\x66\x6f\x75\x72\x74\x68\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x20\x63\x68\x61\x72\x61\x63\x74\x65\x72\x2e\x0a\x20\x20\x20\x20\x65\x2e\x67\x2e\x20\x73\x70\x6c\x69\x74\x64\x72\x69\x76\x65\x28\x22\x2f\x2f\x68\x6f\x73\x74\x2f\x63\x6f\x6d\x70\x75\x74\x65\x72\x2f\x64\x69\x72\x22\x29\x20\x72\x65\x74\x75\x72\x6e\x73\x20\x28\x22\x2f\x2f\x68\x6f\x73\x74\x2f\x63\x6f\x6d\x70\x75\x74\x65\x72\x22\x2c\x20\x22\x2f\x64\x69\x72\x22\x29\x0a\x0a\x20\x20\x20\x20\x50\x61\x74\x68\x73\x20\x63\x61\x6e\x6e\x6f\x74\x20\x63\x6f\x6e\x74\x61\x69\x6e\x20\x62\x6f\x74\x68\x20\x61\x20\x64\x72\x69\x76\x65\x20\x6c\x65\x74\x74\x65\x72\x20\x61\x6e\x64\x20\x61\x20\x55\x4e\x43\x20\x70\x61\x74\x68\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & ntpath_toplevel_consts_18_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_splitroot._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
ntpath_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x26\x00\x19\x22\xa0\x21\x9b\x0c\xd1\x04\x15\x80\x45\x88\x34\x90\x14\xd8\x0b\x10\x90\x24\x98\x14\x91\x2b\xd0\x0b\x1d\xd0\x04\x1d",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_drive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "drive",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
ntpath_toplevel_consts_18_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            & const_str_drive._ascii.ob_base,
            & const_str_root._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(46)
ntpath_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & ntpath_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 157,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 556,
    .co_localsplusnames = & ntpath_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_splitdrive._ascii.ob_base,
    .co_qualname = & const_str_splitdrive._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x01\x7d\x02\x7d\x03\x7c\x01\x7c\x02\x7c\x03\x7a\x00\x00\x00\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[511];
    }
ntpath_toplevel_consts_19_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 510,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x70\x6c\x69\x74\x20\x61\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x69\x6e\x74\x6f\x20\x64\x72\x69\x76\x65\x2c\x20\x72\x6f\x6f\x74\x20\x61\x6e\x64\x20\x74\x61\x69\x6c\x2e\x20\x54\x68\x65\x20\x64\x72\x69\x76\x65\x20\x69\x73\x20\x64\x65\x66\x69\x6e\x65\x64\x0a\x20\x20\x20\x20\x65\x78\x61\x63\x74\x6c\x79\x20\x61\x73\x20\x69\x6e\x20\x73\x70\x6c\x69\x74\x64\x72\x69\x76\x65\x28\x29\x2e\x20\x4f\x6e\x20\x57\x69\x6e\x64\x6f\x77\x73\x2c\x20\x74\x68\x65\x20\x72\x6f\x6f\x74\x20\x6d\x61\x79\x20\x62\x65\x20\x61\x20\x73\x69\x6e\x67\x6c\x65\x20\x70\x61\x74\x68\x0a\x20\x20\x20\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x20\x6f\x72\x20\x61\x6e\x20\x65\x6d\x70\x74\x79\x20\x73\x74\x72\x69\x6e\x67\x2e\x20\x54\x68\x65\x20\x74\x61\x69\x6c\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x20\x61\x6e\x79\x74\x68\x69\x6e\x67\x20\x61\x66\x74\x65\x72\x20\x74\x68\x65\x20\x72\x6f\x6f\x74\x2e\x0a\x20\x20\x20\x20\x46\x6f\x72\x20\x65\x78\x61\x6d\x70\x6c\x65\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x2f\x2f\x73\x65\x72\x76\x65\x72\x2f\x73\x68\x61\x72\x65\x2f\x27\x29\x20\x3d\x3d\x20\x28\x27\x2f\x2f\x73\x65\x72\x76\x65\x72\x2f\x73\x68\x61\x72\x65\x27\x2c\x20\x27\x2f\x27\x2c\x20\x27\x27\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x43\x3a\x2f\x55\x73\x65\x72\x73\x2f\x42\x61\x72\x6e\x65\x79\x27\x29\x20\x3d\x3d\x20\x28\x27\x43\x3a\x27\x2c\x20\x27\x2f\x27\x2c\x20\x27\x55\x73\x65\x72\x73\x2f\x42\x61\x72\x6e\x65\x79\x27\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x43\x3a\x2f\x2f\x2f\x73\x70\x61\x6d\x2f\x2f\x2f\x68\x61\x6d\x27\x29\x20\x3d\x3d\x20\x28\x27\x43\x3a\x27\x2c\x20\x27\x2f\x27\x2c\x20\x27\x2f\x2f\x73\x70\x61\x6d\x2f\x2f\x2f\x68\x61\x6d\x27\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x57\x69\x6e\x64\x6f\x77\x73\x2f\x6e\x6f\x74\x65\x70\x61\x64\x27\x29\x20\x3d\x3d\x20\x28\x27\x27\x2c\x20\x27\x27\x2c\x20\x27\x57\x69\x6e\x64\x6f\x77\x73\x2f\x6e\x6f\x74\x65\x70\x61\x64\x27\x29\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
ntpath_toplevel_consts_19_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\\\\?\\UNC\\",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
ntpath_toplevel_consts_19_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\\\?\\UNC\\",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
ntpath_toplevel_consts_19_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            & ntpath_toplevel_consts_19_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[58]),
            & ntpath_toplevel_consts_19_consts_4.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[58],
            & ntpath_toplevel_consts_19_consts_9._ascii.ob_base,
            &_Py_STR(empty),
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_upper = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "upper",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_find = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "find",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
ntpath_toplevel_consts_19_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            &_Py_ID(replace),
            & const_str_upper._ascii.ob_base,
            & const_str_find._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[392];
    }
ntpath_toplevel_consts_19_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 391,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x16\x00\x09\x0b\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x07\x11\x90\x21\x94\x55\xd4\x07\x1b\xd8\x0e\x13\x88\x03\xd8\x11\x15\x88\x06\xd8\x10\x14\x88\x05\xd8\x15\x24\x88\x0a\xd8\x10\x13\x89\x05\xe0\x0e\x12\x88\x03\xd8\x11\x14\x88\x06\xd8\x10\x13\x88\x05\xd8\x15\x23\x88\x0a\xd8\x10\x12\x88\x05\xd8\x0c\x0d\x8f\x49\x89\x49\x90\x66\x98\x63\xd3\x0c\x22\x80\x45\xd8\x07\x0c\x88\x52\x88\x61\x80\x79\x90\x43\xd2\x07\x17\xd8\x0b\x10\x90\x11\x90\x31\x88\x3a\x98\x13\xd2\x0b\x1c\xf0\x06\x00\x1a\x1f\x98\x72\xa0\x01\x98\x19\x9f\x1f\x99\x1f\xd3\x19\x2a\xa8\x6a\xd2\x19\x38\x91\x41\xb8\x61\x88\x45\xd8\x14\x19\x97\x4a\x91\x4a\x98\x73\xa0\x45\xd3\x14\x2a\x88\x45\xd8\x0f\x14\x98\x02\x8a\x7b\xd8\x17\x18\x98\x25\xa0\x15\x90\x7f\xd0\x10\x26\xd8\x15\x1a\x97\x5a\x91\x5a\xa0\x03\xa0\x55\xa8\x51\xa1\x59\xd3\x15\x2f\x88\x46\xd8\x0f\x15\x98\x12\x8a\x7c\xd8\x17\x18\x98\x25\xa0\x15\x90\x7f\xd0\x10\x26\xd8\x13\x14\x90\x57\x90\x66\x90\x3a\x98\x71\xa0\x16\xa8\x06\xb0\x11\xa9\x0a\xd0\x1f\x33\xb0\x51\xb0\x76\xc0\x01\xb1\x7a\xb0\x7b\xb0\x5e\xd0\x13\x43\xd0\x0c\x43\xf0\x06\x00\x14\x19\x98\x21\x98\x42\x98\x51\x98\x25\xa0\x11\xa0\x31\xa0\x32\xa0\x15\xd0\x13\x26\xd0\x0c\x26\xd8\x09\x0e\x88\x71\x90\x11\x88\x1a\x90\x75\xd2\x09\x1c\xd8\x0b\x10\x90\x11\x90\x31\x88\x3a\x98\x13\xd2\x0b\x1c\xe0\x13\x14\x90\x52\x90\x61\x90\x35\x98\x21\x98\x41\x98\x61\x98\x26\xa0\x21\xa0\x41\xa0\x42\xa0\x25\xd0\x13\x27\xd0\x0c\x27\xf0\x06\x00\x14\x15\x90\x52\x90\x61\x90\x35\x98\x25\xa0\x11\xa0\x31\xa0\x32\xa0\x15\xd0\x13\x26\xd0\x0c\x26\xf0\x06\x00\x10\x15\x90\x65\x98\x51\x88\x7f\xd0\x08\x1e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_unc_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "unc_prefix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_empty = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "empty",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_normp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "normp",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_index2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "index2",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
ntpath_toplevel_consts_19_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            & const_str_altsep._ascii.ob_base,
            & const_str_colon._ascii.ob_base,
            & const_str_unc_prefix._ascii.ob_base,
            & const_str_empty._ascii.ob_base,
            & const_str_normp._ascii.ob_base,
            &_Py_ID(start),
            & const_str_index._ascii.ob_base,
            & const_str_index2._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(510)
ntpath_toplevel_consts_19 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 255,
    },
    .co_consts = & ntpath_toplevel_consts_19_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_19_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 180,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 557,
    .co_localsplusnames = & ntpath_toplevel_consts_19_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_splitroot._ascii.ob_base,
    .co_qualname = & const_str_splitroot._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_19_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x0b\x64\x01\x7d\x01\x64\x02\x7d\x02\x64\x03\x7d\x03\x64\x04\x7d\x04\x64\x05\x7d\x05\x6e\x0a\x64\x06\x7d\x01\x64\x07\x7d\x02\x64\x08\x7d\x03\x64\x09\x7d\x04\x64\x0a\x7d\x05\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x64\x0b\x64\x0c\x1a\x00\x7c\x01\x6b\x28\x00\x00\x72\x7c\x7c\x06\x64\x0c\x64\x0d\x1a\x00\x7c\x01\x6b\x28\x00\x00\x72\x69\x7c\x06\x64\x0b\x64\x0e\x1a\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6b\x28\x00\x00\x72\x02\x64\x0e\x6e\x01\x64\x0d\x7d\x07\x7c\x06\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x07\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x64\x0f\x6b\x28\x00\x00\x72\x05\x7c\x00\x7c\x05\x7c\x05\x66\x03\x53\x00\x7c\x06\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x08\x64\x0c\x7a\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x09\x7c\x09\x64\x0f\x6b\x28\x00\x00\x72\x05\x7c\x00\x7c\x05\x7c\x05\x66\x03\x53\x00\x7c\x00\x64\x0b\x7c\x09\x1a\x00\x7c\x00\x7c\x09\x7c\x09\x64\x0c\x7a\x00\x00\x00\x1a\x00\x7c\x00\x7c\x09\x64\x0c\x7a\x00\x00\x00\x64\x0b\x1a\x00\x66\x03\x53\x00\x7c\x05\x7c\x00\x64\x0b\x64\x0c\x1a\x00\x7c\x00\x64\x0c\x64\x0b\x1a\x00\x66\x03\x53\x00\x7c\x06\x64\x0c\x64\x0d\x1a\x00\x7c\x03\x6b\x28\x00\x00\x72\x21\x7c\x06\x64\x0d\x64\x10\x1a\x00\x7c\x01\x6b\x28\x00\x00\x72\x0e\x7c\x00\x64\x0b\x64\x0d\x1a\x00\x7c\x00\x64\x0d\x64\x10\x1a\x00\x7c\x00\x64\x10\x64\x0b\x1a\x00\x66\x03\x53\x00\x7c\x00\x64\x0b\x64\x0d\x1a\x00\x7c\x05\x7c\x00\x64\x0d\x64\x0b\x1a\x00\x66\x03\x53\x00\x7c\x05\x7c\x05\x7c\x00\x66\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[127];
    }
ntpath_toplevel_consts_20_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 126,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x70\x6c\x69\x74\x20\x61\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x20\x74\x75\x70\x6c\x65\x20\x28\x68\x65\x61\x64\x2c\x20\x74\x61\x69\x6c\x29\x20\x77\x68\x65\x72\x65\x20\x74\x61\x69\x6c\x20\x69\x73\x20\x65\x76\x65\x72\x79\x74\x68\x69\x6e\x67\x20\x61\x66\x74\x65\x72\x20\x74\x68\x65\x20\x66\x69\x6e\x61\x6c\x20\x73\x6c\x61\x73\x68\x2e\x0a\x20\x20\x20\x20\x45\x69\x74\x68\x65\x72\x20\x70\x61\x72\x74\x20\x6d\x61\x79\x20\x62\x65\x20\x65\x6d\x70\x74\x79\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
ntpath_toplevel_consts_20_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & ntpath_toplevel_consts_20_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
ntpath_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__get_bothseps._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            &_Py_ID(len),
            & const_str_rstrip._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[149];
    }
ntpath_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 148,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x09\x0b\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x0b\x18\x98\x11\xd3\x0b\x1b\x80\x44\xdc\x0e\x17\x98\x01\x8b\x6c\x81\x47\x80\x41\x80\x71\x88\x21\xe4\x08\x0b\x88\x41\x8b\x06\x80\x41\xd9\x0a\x0b\x90\x01\x90\x21\x90\x41\x91\x23\x91\x06\x98\x64\xd1\x10\x22\xd8\x08\x09\x88\x51\x89\x06\x88\x01\xf1\x03\x00\x0b\x0c\x90\x01\x90\x21\x90\x41\x91\x23\x91\x06\x98\x64\xd2\x10\x22\xe0\x11\x12\x90\x32\x90\x41\x90\x15\x98\x01\x98\x21\x98\x22\x98\x05\x88\x24\x80\x44\xd8\x0b\x0c\x88\x71\x89\x35\x90\x34\x97\x3b\x91\x3b\x98\x74\xd3\x13\x24\xd1\x0b\x24\xa0\x64\xd0\x0b\x2a\xd0\x04\x2a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
ntpath_toplevel_consts_20_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            & const_str_seps._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[100],
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            & const_str_head._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(248)
ntpath_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 124,
    },
    .co_consts = & ntpath_toplevel_consts_20_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 237,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 558,
    .co_localsplusnames = & ntpath_toplevel_consts_20_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_split._ascii.ob_base,
    .co_qualname = & const_str_split._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x02\x7d\x03\x7d\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x72\x1c\x7c\x00\x7c\x04\x64\x01\x7a\x0a\x00\x00\x19\x00\x00\x00\x7c\x01\x76\x01\x72\x12\x7c\x04\x64\x01\x7a\x17\x00\x00\x7d\x04\x7c\x04\x72\x0b\x7c\x00\x7c\x04\x64\x01\x7a\x0a\x00\x00\x19\x00\x00\x00\x7c\x01\x76\x01\x72\x01\x8c\x12\x7c\x00\x64\x02\x7c\x04\x1a\x00\x7c\x00\x7c\x04\x64\x02\x1a\x00\x7d\x06\x7d\x05\x7c\x02\x7c\x03\x7a\x00\x00\x00\x7c\x05\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x7c\x06\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
ntpath_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
ntpath_toplevel_consts_21_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_genericpath._ascii.ob_base,
            & const_str__splitext._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[72];
    }
ntpath_toplevel_consts_21_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 71,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x08\x0a\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x07\x11\x90\x21\x94\x55\xd4\x07\x1b\xdc\x0f\x1a\xd7\x0f\x24\xd1\x0f\x24\xa0\x51\xa8\x05\xa8\x74\xb0\x54\xd3\x0f\x3a\xd0\x08\x3a\xe4\x0f\x1a\xd7\x0f\x24\xd1\x0f\x24\xa0\x51\xa8\x04\xa8\x63\xb0\x33\xd3\x0f\x37\xd0\x08\x37",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_21_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
        },
    },
};
static
    struct _PyCode_DEF(172)
ntpath_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 86,
    },
    .co_consts = & ntpath_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 258,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 559,
    .co_localsplusnames = & ntpath_toplevel_consts_21_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_splitext._ascii.ob_base,
    .co_qualname = & const_str_splitext._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_21_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x18\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\x64\x02\x64\x03\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x04\x64\x05\x64\x06\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
ntpath_toplevel_consts_22_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Returns the final component of a pathname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_22_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & ntpath_toplevel_consts_22_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_22_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_split._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
ntpath_toplevel_consts_22_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x10\x90\x11\x8b\x38\x90\x41\x89\x3b\xd0\x04\x16",
};
static
    struct _PyCode_DEF(30)
ntpath_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & ntpath_toplevel_consts_22_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_22_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 269,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 560,
    .co_localsplusnames = & ntpath_toplevel_consts_21_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_basename._ascii.ob_base,
    .co_qualname = & const_str_basename._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[46];
    }
ntpath_toplevel_consts_23_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 45,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Returns the directory component of a pathname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_23_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & ntpath_toplevel_consts_23_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct _PyCode_DEF(30)
ntpath_toplevel_consts_23 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & ntpath_toplevel_consts_23_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_22_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 276,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 561,
    .co_localsplusnames = & ntpath_toplevel_consts_21_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_dirname._ascii.ob_base,
    .co_qualname = & const_str_dirname._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_st_reparse_tag = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_reparse_tag",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
ntpath_toplevel_consts_25_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether a path is a junction",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_25_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & ntpath_toplevel_consts_25_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str_IO_REPARSE_TAG_MOUNT_POINT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "IO_REPARSE_TAG_MOUNT_POINT",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
ntpath_toplevel_consts_25_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_lstat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_bool._ascii.ob_base,
            & const_str_st_reparse_tag._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_IO_REPARSE_TAG_MOUNT_POINT._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[78];
    }
ntpath_toplevel_consts_25_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 77,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x09\x19\xdc\x11\x13\x97\x18\x91\x18\x98\x24\x93\x1e\x88\x42\xf4\x06\x00\x10\x14\x90\x42\xd7\x14\x25\xd1\x14\x25\xac\x14\xd7\x29\x48\xd1\x29\x48\xd1\x14\x48\xd3\x0f\x49\xd0\x08\x49\xf8\xf4\x05\x00\x11\x18\x9c\x1a\xa4\x5e\xd0\x0f\x34\xf2\x00\x01\x09\x19\xd9\x13\x18\xf0\x03\x01\x09\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
ntpath_toplevel_consts_25_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x3d\x00\xbd\x14\x41\x14\x03\xc1\x13\x01\x41\x14\x03",
};
static
    struct _PyCode_DEF(174)
ntpath_toplevel_consts_25 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 87,
    },
    .co_consts = & ntpath_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_25_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_25_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 284,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 562,
    .co_localsplusnames = & genericpath_toplevel_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_isjunction._ascii.ob_base,
    .co_qualname = & const_str_isjunction._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_25_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_26_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
ntpath_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x0a\x8f\x09\x89\x09\x90\x24\x8c\x0f\xd8\x0f\x14",
};
static
    struct _PyCode_DEF(46)
ntpath_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & ntpath_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 292,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 563,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_isjunction._ascii.ob_base,
    .co_qualname = & const_str_isjunction._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[68];
    }
ntpath_toplevel_consts_27_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 67,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether a path exists.  Returns True for broken symbolic links",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
ntpath_toplevel_consts_27_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & ntpath_toplevel_consts_27_consts_0._ascii.ob_base,
            Py_False,
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
ntpath_toplevel_consts_27_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_lstat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[51];
    }
ntpath_toplevel_consts_27_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 50,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x0d\x0f\x8f\x58\x89\x58\x90\x64\x8b\x5e\x88\x02\xf0\x06\x00\x0c\x10\xf8\xf4\x05\x00\x0d\x14\x94\x5a\xd0\x0b\x20\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct _PyCode_DEF(90)
ntpath_toplevel_consts_27 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 45,
    },
    .co_consts = & ntpath_toplevel_consts_27_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_27_names._object.ob_base.ob_base,
    .co_exceptiontable = & genericpath_toplevel_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 300,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 564,
    .co_localsplusnames = & genericpath_toplevel_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_lexists._ascii.ob_base,
    .co_qualname = & const_str_lexists._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_27_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x79\x02\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__getvolumepathname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_getvolumepathname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_28 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__getvolumepathname._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[98];
    }
ntpath_toplevel_consts_29_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 97,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x65\x73\x74\x20\x77\x68\x65\x74\x68\x65\x72\x20\x61\x20\x70\x61\x74\x68\x20\x69\x73\x20\x61\x20\x6d\x6f\x75\x6e\x74\x20\x70\x6f\x69\x6e\x74\x20\x28\x61\x20\x64\x72\x69\x76\x65\x20\x72\x6f\x6f\x74\x2c\x20\x74\x68\x65\x20\x72\x6f\x6f\x74\x20\x6f\x66\x20\x61\x0a\x20\x20\x20\x20\x73\x68\x61\x72\x65\x2c\x20\x6f\x72\x20\x61\x20\x6d\x6f\x75\x6e\x74\x65\x64\x20\x76\x6f\x6c\x75\x6d\x65\x29",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
ntpath_toplevel_consts_29_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & ntpath_toplevel_consts_29_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_True,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
ntpath_toplevel_consts_29_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__get_bothseps._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            & const_str__getvolumepathname._ascii.ob_base,
            & const_str_rstrip._ascii.ob_base,
            & const_str_casefold._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[146];
    }
ntpath_toplevel_consts_29_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 145,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x0c\x0e\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x0b\x18\x98\x14\xd3\x0b\x1e\x80\x44\xdc\x0b\x12\x90\x34\x8b\x3d\x80\x44\xdc\x18\x21\xa0\x24\x9b\x0f\xd1\x04\x15\x80\x45\x88\x34\x90\x14\xd9\x07\x0c\x90\x15\x90\x71\x91\x18\x98\x54\xd1\x11\x21\xd8\x13\x17\x88\x78\x88\x0f\xd9\x07\x0b\x91\x44\xd8\x0f\x13\xe5\x07\x19\xd8\x0c\x10\x8f\x4b\x89\x4b\x98\x04\xd3\x0c\x1d\x88\x01\xdc\x0b\x1d\x98\x64\xd3\x0b\x23\xd7\x0b\x2a\xd1\x0b\x2a\xa8\x34\xd3\x0b\x30\x88\x01\xd8\x0f\x10\x8f\x7a\x89\x7a\x8b\x7c\x98\x71\x9f\x7a\x99\x7a\x9b\x7c\xd1\x0f\x2b\xd0\x08\x2b\xe0\x0f\x14",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
ntpath_toplevel_consts_29_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_seps._ascii.ob_base,
            & const_str_drive._ascii.ob_base,
            & const_str_root._ascii.ob_base,
            & const_str_rest._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
            (PyObject *)&_Py_SINGLETON(strings).ascii[121],
        },
    },
};
static
    struct _PyCode_DEF(318)
ntpath_toplevel_consts_29 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 159,
    },
    .co_consts = & ntpath_toplevel_consts_29_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_29_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 322,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 565,
    .co_localsplusnames = & ntpath_toplevel_consts_29_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_ismount._ascii.ob_base,
    .co_qualname = & const_str_ismount._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_29_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x02\x7d\x03\x7d\x04\x7c\x02\x72\x0a\x7c\x02\x64\x01\x19\x00\x00\x00\x7c\x01\x76\x00\x72\x03\x7c\x04\x0c\x00\x53\x00\x7c\x03\x72\x03\x7c\x04\x73\x01\x79\x02\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x72\x4c\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x05\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00\x79\x03",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[77];
    }
ntpath_toplevel_consts_30_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 76,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x78\x70\x61\x6e\x64\x20\x7e\x20\x61\x6e\x64\x20\x7e\x75\x73\x65\x72\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x73\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x75\x73\x65\x72\x20\x6f\x72\x20\x24\x48\x4f\x4d\x45\x20\x69\x73\x20\x75\x6e\x6b\x6e\x6f\x77\x6e\x2c\x20\x64\x6f\x20\x6e\x6f\x74\x68\x69\x6e\x67\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_USERPROFILE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "USERPROFILE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_HOMEPATH = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HOMEPATH",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_HOMEDRIVE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HOMEDRIVE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_USERNAME = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "USERNAME",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
ntpath_toplevel_consts_30_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & ntpath_toplevel_consts_30_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[126]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[126],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & const_str_USERPROFILE._ascii.ob_base,
            & const_str_HOMEPATH._ascii.ob_base,
            & const_str_HOMEDRIVE._ascii.ob_base,
            &_Py_STR(empty),
            & const_str_USERNAME._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
ntpath_toplevel_consts_30_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_startswith._ascii.ob_base,
            &_Py_ID(len),
            & const_str__get_bothseps._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            &_Py_ID(join),
            & const_str_fsdecode._ascii.ob_base,
            &_Py_ID(get),
            & const_str_basename._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
            & const_str_fsencode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[380];
    }
ntpath_toplevel_consts_30_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 379,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x08\x00\x0c\x0e\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x10\x14\x89\x05\xe0\x10\x13\x88\x05\xd8\x0b\x0f\x8f\x3f\x89\x3f\x98\x35\xd4\x0b\x21\xd8\x0f\x13\x88\x0b\xd8\x0b\x0c\x8c\x63\x90\x24\x8b\x69\x80\x71\x80\x41\xd8\x0a\x0b\x88\x61\x8a\x25\x90\x44\x98\x11\x91\x47\xa4\x3d\xb0\x14\xd3\x23\x36\xd1\x14\x36\xd8\x08\x09\x88\x51\x89\x06\x88\x01\xf0\x03\x00\x0b\x0c\x88\x61\x8a\x25\x90\x44\x98\x11\x91\x47\xa4\x3d\xb0\x14\xd3\x23\x36\xd2\x14\x36\xf0\x06\x00\x08\x15\x9c\x02\x9f\x0a\x99\x0a\xd1\x07\x22\xdc\x13\x15\x97\x3a\x91\x3a\x98\x6d\xd1\x13\x2c\x89\x08\xd8\x0d\x17\x9c\x32\x9f\x3a\x99\x3a\xd1\x0d\x25\xd8\x0f\x13\x88\x0b\xf0\x04\x03\x09\x17\xdc\x14\x16\x97\x4a\x91\x4a\x98\x7b\xd1\x14\x2b\x88\x45\xf4\x06\x00\x14\x18\x98\x05\x9c\x72\x9f\x7a\x99\x7a\xa8\x2a\xd1\x1f\x35\xd3\x13\x36\x88\x08\xe0\x07\x08\x88\x41\x82\x76\xd8\x16\x1a\x98\x31\x98\x51\x90\x69\x88\x0b\xdc\x0b\x15\x90\x6b\xa4\x35\xd4\x0b\x29\xdc\x1a\x1c\x9f\x2b\x99\x2b\xa0\x6b\xd3\x1a\x32\x88\x4b\xdc\x17\x19\x97\x7a\x91\x7a\x97\x7e\x91\x7e\xa0\x6a\xd3\x17\x31\x88\x0c\xe0\x0b\x16\x98\x2c\xd2\x0b\x26\xf0\x0c\x00\x10\x1c\x9c\x78\xa8\x08\xd3\x1f\x31\xd2\x0f\x31\xd8\x17\x1b\x90\x0b\xdc\x17\x1b\x9c\x47\xa0\x48\xd3\x1c\x2d\xa8\x7b\xd3\x17\x3b\x88\x48\xe4\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xdc\x13\x15\x97\x3b\x91\x3b\x98\x78\xd3\x13\x28\x88\x08\xe0\x0b\x13\x90\x64\x98\x31\x98\x32\x90\x68\xd1\x0b\x1e\xd0\x04\x1e\xf8\xf4\x2f\x00\x10\x18\xf2\x00\x01\x09\x17\xd8\x14\x16\x8a\x45\xf0\x03\x01\x09\x17\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
ntpath_toplevel_consts_30_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x36\x13\x46\x0b\x00\xc6\x0b\x0b\x46\x19\x03\xc6\x18\x01\x46\x19\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_tilde = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tilde",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_userhome = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "userhome",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_target_user = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "target_user",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_current_user = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "current_user",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
ntpath_toplevel_consts_30_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_tilde._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            (PyObject *)&_Py_SINGLETON(strings).ascii[110],
            & const_str_userhome._ascii.ob_base,
            & const_str_drive._ascii.ob_base,
            & const_str_target_user._ascii.ob_base,
            & const_str_current_user._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(824)
ntpath_toplevel_consts_30 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 412,
    },
    .co_consts = & ntpath_toplevel_consts_30_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_30_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_30_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 351,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 566,
    .co_localsplusnames = & ntpath_toplevel_consts_30_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_expanduser._ascii.ob_base,
    .co_qualname = & const_str_expanduser._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_30_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x03\x64\x01\x7d\x01\x6e\x02\x64\x02\x7d\x01\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x73\x02\x7c\x00\x53\x00\x64\x03\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7d\x02\x7c\x02\x7c\x03\x6b\x02\x00\x00\x72\x2b\x7c\x00\x7c\x02\x19\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x76\x01\x72\x1b\x7c\x02\x64\x03\x7a\x0d\x00\x00\x7d\x02\x7c\x02\x7c\x03\x6b\x02\x00\x00\x72\x11\x7c\x00\x7c\x02\x19\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x76\x01\x72\x01\x8c\x1b\x64\x04\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x14\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x19\x00\x00\x00\x7d\x04\x6e\x45\x64\x05\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x02\x7c\x00\x53\x00\x09\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x19\x00\x00\x00\x7d\x05\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x19\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x02\x64\x03\x6b\x37\x00\x00\x72\x73\x7c\x00\x64\x03\x7c\x02\x1a\x00\x7d\x06\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x15\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x06\x7c\x07\x6b\x37\x00\x00\x72\x25\x7c\x07\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x02\x7c\x00\x53\x00\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x15\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x7c\x00\x7c\x02\x64\x09\x1a\x00\x7a\x00\x00\x00\x53\x00\x23\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x07\x7d\x05\x59\x00\x8c\xcf\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
ntpath_toplevel_consts_31 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "'[^']*'?|%(%|[^%]*%?)|\\$(\\$|[-\\w]+|\\{[^}]*\\}?)",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[103];
    }
ntpath_toplevel_consts_32_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 102,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x78\x70\x61\x6e\x64\x20\x73\x68\x65\x6c\x6c\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x66\x6f\x72\x6d\x73\x20\x24\x76\x61\x72\x2c\x20\x24\x7b\x76\x61\x72\x7d\x20\x61\x6e\x64\x20\x25\x76\x61\x72\x25\x2e\x0a\x0a\x20\x20\x20\x20\x55\x6e\x6b\x6e\x6f\x77\x6e\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x73\x20\x61\x72\x65\x20\x6c\x65\x66\x74\x20\x75\x6e\x63\x68\x61\x6e\x67\x65\x64\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_environb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "environb",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
ntpath_toplevel_consts_32_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_lastindex = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lastindex",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
ntpath_toplevel_consts_32_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_lastindex._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fsencode._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
ntpath_toplevel_consts_32_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "expandvars.<locals>.repl",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[221];
    }
ntpath_toplevel_consts_32_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 220,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x14\x15\x97\x4b\x91\x4b\x88\x09\xd8\x0b\x14\xd0\x0b\x1c\xd8\x13\x14\x90\x51\x91\x34\x88\x4b\xd8\x0f\x10\x90\x19\x89\x7c\x88\x04\xd8\x0b\x14\x98\x01\x8a\x3e\xd8\x0f\x13\x90\x77\x8a\x7f\xd8\x17\x1b\x90\x0b\xd8\x13\x17\x97\x3d\x91\x3d\xa0\x17\xd4\x13\x29\xd8\x17\x18\x98\x11\x91\x74\x90\x0b\xd8\x13\x17\x98\x03\x98\x12\x90\x39\x89\x44\xe0\x0f\x13\x90\x76\x8a\x7e\xd8\x17\x1b\x90\x0b\xd8\x0f\x13\x8f\x7f\x89\x7f\x98\x75\xd4\x0f\x25\xd8\x17\x1b\x97\x7d\x91\x7d\xa0\x56\xd4\x17\x2c\xd8\x1b\x1c\x98\x51\x99\x34\x90\x4b\xd8\x17\x1b\x98\x41\x98\x62\x90\x7a\x90\x04\xf0\x04\x06\x09\x18\xd8\x0f\x16\x88\x7f\xdc\x17\x19\x97\x7b\x91\x7b\xa4\x32\xa7\x3a\xa1\x3a\xac\x62\xaf\x6b\xa9\x6b\xb8\x24\xd3\x2e\x3f\xd1\x23\x40\xd3\x17\x41\xd0\x10\x41\xe0\x17\x1e\x98\x74\x91\x7d\xd0\x10\x24\xf8\xdc\x0f\x17\xf2\x00\x01\x09\x18\xd8\x13\x14\x90\x51\x91\x34\x8a\x4b\xf0\x03\x01\x09\x18\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
ntpath_toplevel_consts_32_consts_12_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x36\x3a\x42\x36\x00\xc2\x31\x04\x42\x36\x00\xc2\x36\x0e\x43\x07\x03\xc3\x06\x01\x43\x07\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_brace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "brace",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_dollar = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dollar",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_percent = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "percent",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_rbrace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rbrace",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
ntpath_toplevel_consts_32_consts_12_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[109],
            & const_str_lastindex._ascii.ob_base,
            &_Py_ID(name),
            & const_str_brace._ascii.ob_base,
            & const_str_dollar._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            & const_str_percent._ascii.ob_base,
            & const_str_rbrace._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
ntpath_toplevel_consts_32_consts_12_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x20\x80\x80\x80\x80\x80",
};
static
    struct _PyCode_DEF(404)
ntpath_toplevel_consts_32_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 202,
    },
    .co_consts = & ntpath_toplevel_consts_32_consts_12_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_32_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_32_consts_12_exceptiontable.ob_base.ob_base,
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 447,
    .co_nlocalsplus = 8,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 5,
    .co_version = 567,
    .co_localsplusnames = & ntpath_toplevel_consts_32_consts_12_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & ntpath_toplevel_consts_32_consts_12_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = &_Py_ID(repl),
    .co_qualname = & ntpath_toplevel_consts_32_consts_12_qualname._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_32_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x05\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x80\x05\x7c\x00\x64\x01\x19\x00\x00\x00\x53\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x7c\x01\x64\x02\x6b\x28\x00\x00\x72\x23\x7c\x02\x89\x06\x6b\x28\x00\x00\x72\x02\x7c\x02\x53\x00\x7c\x02\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x06\xab\x01\x00\x00\x00\x00\x00\x00\x73\x05\x7c\x00\x64\x01\x19\x00\x00\x00\x53\x00\x7c\x02\x64\x00\x64\x03\x1a\x00\x7d\x02\x6e\x33\x7c\x02\x89\x04\x6b\x28\x00\x00\x72\x02\x7c\x02\x53\x00\x7c\x02\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x03\xab\x01\x00\x00\x00\x00\x00\x00\x72\x1b\x7c\x02\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x07\xab\x01\x00\x00\x00\x00\x00\x00\x73\x05\x7c\x00\x64\x01\x19\x00\x00\x00\x53\x00\x7c\x02\x64\x02\x64\x03\x1a\x00\x7d\x02\x09\x00\x89\x05\x80\x39\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x89\x05\x7c\x02\x19\x00\x00\x00\x53\x00\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x08\x01\x00\x7c\x00\x64\x01\x19\x00\x00\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
ntpath_toplevel_consts_32_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & ntpath_toplevel_consts_32_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[36]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[37]),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_characters[123]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[125]),
            & const_str_environb._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[36],
            (PyObject *)&_Py_SINGLETON(strings).ascii[37],
            (PyObject *)&_Py_SINGLETON(strings).ascii[123],
            (PyObject *)&_Py_SINGLETON(strings).ascii[125],
            & ntpath_toplevel_consts_32_consts_12.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__varsubb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_varsubb",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_re = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "re",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__varpattern = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_varpattern",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_ASCII = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ASCII",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_sub = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sub",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__varsub = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_varsub",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
ntpath_toplevel_consts_32_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str__varsubb._ascii.ob_base,
            & const_str_re._ascii.ob_base,
            & const_str_compile._ascii.ob_base,
            & const_str__varpattern._ascii.ob_base,
            &_Py_ID(encode),
            & const_str_ASCII._ascii.ob_base,
            & const_str_sub._ascii.ob_base,
            &_Py_ID(getattr),
            & const_str__varsub._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[237];
    }
ntpath_toplevel_consts_32_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 236,
    },
    .ob_shash = -1,
    .ob_sval = "\xfc\x80\x00\xf4\x08\x00\x0c\x0e\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xe4\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x0b\x0f\x90\x74\xd1\x0b\x1b\xa0\x04\xa8\x44\xd1\x20\x30\xd8\x13\x17\x88\x4b\xdd\x0f\x17\xdb\x0c\x15\xd8\x17\x19\x97\x7a\x91\x7a\xa4\x2b\xd7\x22\x34\xd1\x22\x34\xd3\x22\x36\xb8\x02\xbf\x08\xb9\x08\xd3\x17\x41\xd7\x17\x45\xd1\x17\x45\x88\x48\xdc\x0e\x16\x88\x03\xd8\x12\x16\x88\x07\xd8\x10\x14\x88\x05\xd8\x11\x15\x88\x06\xd8\x11\x15\x88\x06\xdc\x12\x19\x9c\x22\x98\x6a\xa8\x24\xd3\x12\x2f\x89\x07\xe0\x0b\x0e\x90\x64\x89\x3f\x98\x73\xa8\x24\x99\x7f\xd8\x13\x17\x88\x4b\xdd\x0f\x16\xdb\x0c\x15\xd8\x16\x18\x97\x6a\x91\x6a\xa4\x1b\xa8\x62\xaf\x68\xa9\x68\xd3\x16\x37\xd7\x16\x3b\xd1\x16\x3b\x88\x47\xdc\x0e\x15\x88\x03\xd8\x12\x15\x88\x07\xd8\x10\x13\x88\x05\xd8\x11\x14\x88\x06\xd8\x11\x14\x88\x06\xdc\x12\x14\x97\x2a\x91\x2a\x88\x07\xf7\x04\x19\x05\x18\xf0\x00\x19\x05\x18\xf1\x36\x00\x0c\x0f\x88\x74\x90\x54\x8b\x3f\xd0\x04\x1a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
ntpath_toplevel_consts_32_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_re._ascii.ob_base,
            & const_str_sub._ascii.ob_base,
            &_Py_ID(repl),
            & const_str_brace._ascii.ob_base,
            & const_str_dollar._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            & const_str_percent._ascii.ob_base,
            & const_str_rbrace._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
ntpath_toplevel_consts_32_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "    @@@@@",
};
static
    struct _PyCode_DEF(522)
ntpath_toplevel_consts_32 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 261,
    },
    .co_consts = & ntpath_toplevel_consts_32_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_32_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 416,
    .co_nlocalsplus = 9,
    .co_nlocals = 4,
    .co_ncellvars = 5,
    .co_nfreevars = 0,
    .co_version = 568,
    .co_localsplusnames = & ntpath_toplevel_consts_32_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & ntpath_toplevel_consts_32_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_expandvars._ascii.ob_base,
    .co_qualname = & const_str_expandvars._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_32_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x04\x87\x05\x87\x06\x87\x07\x87\x08\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x6c\x64\x01\x7c\x00\x76\x01\x72\x06\x64\x02\x7c\x00\x76\x01\x72\x02\x7c\x00\x53\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x73\x3c\x64\x03\x64\x04\x6c\x05\x7d\x01\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x61\x04\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x02\x8a\x07\x64\x05\x8a\x04\x64\x06\x8a\x08\x64\x01\x8a\x05\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x64\x04\xab\x03\x00\x00\x00\x00\x00\x00\x8a\x06\x6e\x5c\x64\x08\x7c\x00\x76\x01\x72\x06\x64\x09\x7c\x00\x76\x01\x72\x02\x7c\x00\x53\x00\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x73\x2e\x64\x03\x64\x04\x6c\x05\x7d\x01\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x61\x0c\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x09\x8a\x07\x64\x0a\x8a\x04\x64\x0b\x8a\x08\x64\x08\x8a\x05\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8a\x06\x88\x04\x88\x05\x88\x06\x88\x07\x88\x08\x66\x05\x64\x0c\x84\x08\x7d\x03\x02\x00\x7c\x02\x7c\x03\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 5,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__path_normpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_normpath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_33 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_normpath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[49];
    }
ntpath_toplevel_consts_34_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 48,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Normalize path, eliminating double slashes, etc.",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
ntpath_toplevel_consts_34_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "..",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
ntpath_toplevel_consts_34_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & ntpath_toplevel_consts_34_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            & ntpath_toplevel_consts_34_consts_4.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & ntpath_toplevel_consts_2._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
ntpath_toplevel_consts_34_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            &_Py_ID(replace),
            & const_str_splitroot._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            &_Py_ID(len),
            &_Py_ID(append),
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[308];
    }
ntpath_toplevel_consts_34_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 307,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x11\x8f\x79\x89\x79\x98\x14\x8b\x7f\x88\x04\xdc\x0b\x15\x90\x64\x9c\x45\xd4\x0b\x22\xd8\x12\x17\x88\x43\xd8\x15\x19\x88\x46\xd8\x15\x19\x88\x46\xd8\x15\x1a\x89\x46\xe0\x12\x16\x88\x43\xd8\x15\x18\x88\x46\xd8\x15\x18\x88\x46\xd8\x15\x19\x88\x46\xd8\x0f\x13\x8f\x7c\x89\x7c\x98\x46\xa0\x43\xd3\x0f\x28\x88\x04\xdc\x1c\x25\xa0\x64\x9b\x4f\xd1\x08\x19\x88\x05\x88\x74\x90\x54\xd8\x11\x16\x98\x14\x91\x1c\x88\x06\xd8\x10\x14\x97\x0a\x91\x0a\x98\x33\x93\x0f\x88\x05\xd8\x0c\x0d\x88\x01\xd8\x0e\x0f\x94\x23\x90\x65\x93\x2a\x8a\x6e\xd8\x13\x18\x98\x11\x92\x38\x98\x75\xa0\x51\x99\x78\xa8\x36\xd2\x1f\x31\xd8\x14\x19\x98\x21\x91\x48\xd8\x11\x16\x90\x71\x91\x18\x98\x56\xd2\x11\x23\xd8\x13\x14\x90\x71\x92\x35\x98\x55\xa0\x31\xa0\x51\xa1\x33\x99\x5a\xa8\x36\xd2\x1d\x31\xd8\x18\x1d\x98\x61\xa0\x01\x99\x63\xa0\x21\xa0\x41\xa1\x23\x98\x67\x98\x0e\xd8\x14\x15\x98\x11\x91\x46\x91\x41\xd8\x15\x16\x98\x21\x92\x56\xa1\x04\xd8\x18\x1d\x98\x61\x99\x08\xe0\x14\x15\x98\x11\x91\x46\x91\x41\xe0\x10\x11\x90\x51\x91\x06\x90\x01\xf0\x19\x00\x0f\x10\x94\x23\x90\x65\x93\x2a\x8b\x6e\xf1\x1c\x00\x10\x16\x99\x65\xd8\x0c\x11\x8f\x4c\x89\x4c\x98\x16\xd4\x0c\x20\xd8\x0f\x15\x98\x03\x9f\x08\x99\x08\xa0\x15\x9b\x0f\xd1\x0f\x27\xd0\x08\x27",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_comps = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "comps",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
ntpath_toplevel_consts_34_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(sep),
            & const_str_altsep._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            & const_str_drive._ascii.ob_base,
            & const_str_root._ascii.ob_base,
            & const_str_prefix._ascii.ob_base,
            & const_str_comps._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
        },
    },
};
static
    struct _PyCode_DEF(524)
ntpath_toplevel_consts_34 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 262,
    },
    .co_consts = & ntpath_toplevel_consts_34_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_34_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 484,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 569,
    .co_localsplusnames = & ntpath_toplevel_consts_34_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_normpath._ascii.ob_base,
    .co_qualname = & const_str_normpath._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_34_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x09\x64\x01\x7d\x01\x64\x02\x7d\x02\x64\x03\x7d\x03\x64\x04\x7d\x04\x6e\x08\x64\x05\x7d\x01\x64\x06\x7d\x02\x64\x07\x7d\x03\x64\x08\x7d\x04\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x05\x7d\x06\x7d\x00\x7c\x05\x7c\x06\x7a\x00\x00\x00\x7d\x07\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x08\x64\x09\x7d\x09\x7c\x09\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x02\x00\x00\x72\x5f\x7c\x08\x7c\x09\x19\x00\x00\x00\x72\x08\x7c\x08\x7c\x09\x19\x00\x00\x00\x7c\x03\x6b\x28\x00\x00\x72\x04\x7c\x08\x7c\x09\x3d\x00\x6e\x3f\x7c\x08\x7c\x09\x19\x00\x00\x00\x7c\x04\x6b\x28\x00\x00\x72\x32\x7c\x09\x64\x09\x6b\x44\x00\x00\x72\x1c\x7c\x08\x7c\x09\x64\x0a\x7a\x0a\x00\x00\x19\x00\x00\x00\x7c\x04\x6b\x37\x00\x00\x72\x11\x7c\x08\x7c\x09\x64\x0a\x7a\x0a\x00\x00\x7c\x09\x64\x0a\x7a\x00\x00\x00\x85\x02\x3d\x00\x7c\x09\x64\x0a\x7a\x17\x00\x00\x7d\x09\x6e\x16\x7c\x09\x64\x09\x6b\x28\x00\x00\x72\x06\x7c\x06\x72\x04\x7c\x08\x7c\x09\x3d\x00\x6e\x0b\x7c\x09\x64\x0a\x7a\x0d\x00\x00\x7d\x09\x6e\x05\x7c\x09\x64\x0a\x7a\x0d\x00\x00\x7d\x09\x7c\x09\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x02\x00\x00\x72\x01\x8c\x5f\x7c\x07\x73\x13\x7c\x08\x73\x11\x7c\x08\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x07\x7c\x01\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__getfullpathname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_getfullpathname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_35 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__getfullpathname._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
ntpath_toplevel_consts_36_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the absolute version of a path.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
ntpath_toplevel_consts_36_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & ntpath_toplevel_consts_36_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_getcwdb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getcwdb",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
ntpath_toplevel_consts_36_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str__getfullpathname._ascii.ob_base,
            & const_str_normpath._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_getcwdb._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[240];
    }
ntpath_toplevel_consts_36_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 239,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x04\x09\x11\xdc\x13\x23\xa4\x48\xa8\x54\xa3\x4e\xd3\x13\x33\xd0\x0c\x33\xf8\xdc\x10\x17\x9c\x1a\xd0\x0f\x24\xf2\x00\x02\x09\x11\xe1\x0c\x10\xf0\x05\x02\x09\x11\xfa\xf4\x06\x00\x10\x12\x8f\x79\x89\x79\x98\x14\x8b\x7f\x88\x04\xdc\x0f\x14\x90\x54\x8c\x7b\xdc\x0f\x19\x98\x24\xa4\x05\xd4\x0f\x26\xd8\x16\x1b\x90\x03\xdc\x19\x1b\x9f\x1a\x99\x1a\x91\x06\xe0\x16\x1a\x90\x03\xdc\x19\x1b\x9f\x19\x99\x19\x90\x06\xdc\x20\x29\xa8\x24\xa3\x0f\xd1\x0c\x1d\x88\x45\x90\x34\x98\x14\xe1\x0f\x14\x99\x04\xf0\x02\x04\x11\x2e\xdc\x1b\x1f\xd4\x20\x30\xb0\x15\xb8\x14\xb1\x1c\xd3\x20\x3e\xc0\x04\xd3\x1b\x45\x90\x44\xf4\x0c\x00\x10\x18\x98\x04\x8b\x7e\xd0\x08\x1d\xf8\xf4\x0b\x00\x19\x20\xa4\x1a\xd0\x17\x2c\xf2\x00\x02\x11\x2e\xe0\x1b\x20\xa0\x33\x99\x3b\xa8\x14\xd1\x1b\x2d\x91\x44\xf4\x06\x00\x10\x18\x98\x04\x8b\x7e\xd0\x08\x1d\xf0\x0b\x02\x11\x2e\xfa\xf4\x08\x00\x18\x1c\x99\x46\x9b\x48\xa0\x64\xd3\x17\x2b\x90\x04\xdc\x0f\x17\x98\x04\x8b\x7e\xd0\x08\x1d",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
ntpath_toplevel_consts_36_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x13\x16\x00\x96\x0f\x28\x03\xa7\x01\x28\x03\xc2\x14\x18\x42\x37\x00\xc2\x37\x17\x43\x1b\x03\xc3\x1a\x01\x43\x1b\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
ntpath_toplevel_consts_36_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(sep),
            & const_str_getcwd._ascii.ob_base,
            & const_str_drive._ascii.ob_base,
            & const_str_root._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(500)
ntpath_toplevel_consts_36 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 250,
    },
    .co_consts = & ntpath_toplevel_consts_36_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_36_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_36_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 538,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 570,
    .co_localsplusnames = & ntpath_toplevel_consts_36_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_abspath._ascii.ob_base,
    .co_qualname = & const_str_abspath._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_36_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x04\x77\x00\x78\x03\x59\x00\x77\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\xa4\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x13\x64\x01\x7d\x01\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x6e\x12\x64\x02\x7d\x01\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x03\x7d\x04\x7d\x00\x7c\x03\x73\x02\x7c\x04\x72\x4b\x09\x00\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x04\x7a\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x15\x01\x00\x7c\x03\x7c\x01\x7a\x00\x00\x00\x7c\x00\x7a\x00\x00\x00\x7d\x00\x59\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x7c\x02\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_37_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & ntpath_toplevel_consts_36_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
ntpath_toplevel_consts_37_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_getcwdb._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            &_Py_ID(join),
            & const_str_normpath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[76];
    }
ntpath_toplevel_consts_37_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 75,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x11\x8f\x79\x89\x79\x98\x14\x8b\x7f\x88\x04\xdc\x0f\x14\x90\x54\x8c\x7b\xdc\x0f\x19\x98\x24\xa4\x05\xd4\x0f\x26\xdc\x16\x18\x97\x6a\x91\x6a\x93\x6c\x91\x03\xe4\x16\x18\x97\x69\x91\x69\x93\x6b\x90\x03\xdc\x13\x17\x98\x03\x98\x54\x93\x3f\x88\x44\xdc\x0f\x17\x98\x04\x8b\x7e\xd0\x08\x1d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_37_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(cwd),
        },
    },
};
static
    struct _PyCode_DEF(226)
ntpath_toplevel_consts_37 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 113,
    },
    .co_consts = & ntpath_toplevel_consts_37_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_37_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 526,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 571,
    .co_localsplusnames = & ntpath_toplevel_consts_37_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_abspath._ascii.ob_base,
    .co_qualname = & const_str_abspath._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_37_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x45\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x15\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x6e\x14\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__getfinalpathname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_getfinalpathname",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_readlink = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "readlink",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_38 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__getfinalpathname._ascii.ob_base,
            & const_str_readlink._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_4390 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 4390 },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_4392 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 4392 },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_4393 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 4393 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
ntpath_toplevel_consts_39_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 5],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 21],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 32],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 50],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 67],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 87],
            & const_int_4390.ob_base,
            & const_int_4392.ob_base,
            & const_int_4393.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_39_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & ntpath_toplevel_consts_39_consts_1._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__nt_readlink = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_nt_readlink",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_winerror = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "winerror",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
ntpath_toplevel_consts_39_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_set._ascii.ob_base,
            & const_str_normcase._ascii.ob_base,
            &_Py_ID(add),
            & const_str__nt_readlink._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            & const_str_islink._ascii.ob_base,
            & const_str_normpath._ascii.ob_base,
            &_Py_ID(join),
            & const_str_dirname._ascii.ob_base,
            & const_str_winerror._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__readlink_deep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_readlink_deep",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[208];
    }
ntpath_toplevel_consts_39_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 207,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x1e\x00\x1c\x4c\x01\xd0\x08\x18\xe4\x0f\x12\x8b\x75\x88\x04\xdc\x0e\x16\x90\x74\x8b\x6e\xa0\x44\xd1\x0e\x28\xd8\x0c\x10\x8f\x48\x89\x48\x94\x58\x98\x64\x93\x5e\xd4\x0c\x24\xf0\x02\x13\x0d\x16\xd8\x1b\x1f\x90\x08\xdc\x17\x23\xa0\x44\xd3\x17\x29\x90\x04\xf4\x06\x00\x18\x1d\x98\x54\x94\x7b\xf4\x08\x00\x1c\x22\xa0\x28\xd4\x1b\x2b\xd8\x1f\x27\x98\x04\xd8\x18\x1d\xf0\x12\x00\x10\x14\x88\x0b\xf4\x11\x00\x1c\x24\xa4\x44\xac\x17\xb0\x18\xd3\x29\x3a\xb8\x44\xd3\x24\x41\xd3\x1b\x42\x90\x44\xf4\x1d\x00\x0f\x17\x90\x74\x8b\x6e\xa0\x44\xd2\x0e\x28\xf0\x2c\x00\x10\x14\x88\x0b\xf8\xf0\x0f\x00\x14\x21\xf2\x00\x03\x0d\x16\xd8\x13\x15\x97\x3b\x91\x3b\xd0\x22\x32\xd1\x13\x32\xdb\x14\x19\xf0\x0a\x00\x10\x14\x88\x0b\xf0\x09\x00\x11\x16\xfb\xdc\x13\x1d\xf2\x00\x02\x0d\x16\xe0\x10\x15\xd8\x0f\x13\x88\x0b\xf0\x07\x02\x0d\x16\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[42];
    }
ntpath_toplevel_consts_39_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 41,
    },
    .ob_shash = -1,
    .ob_sval = "\xb5\x25\x42\x0b\x00\xc1\x1d\x1e\x42\x0b\x00\xc2\x0b\x05\x42\x35\x03\xc2\x10\x0e\x42\x25\x03\xc2\x24\x01\x42\x25\x03\xc2\x25\x0c\x42\x35\x03\xc2\x34\x01\x42\x35\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_ignored_error = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ignored_error",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_allowed_winerror = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "allowed_winerror",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_seen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "seen",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_old_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "old_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_ex = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ex",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
ntpath_toplevel_consts_39_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_ignored_error._ascii.ob_base,
            & const_str_allowed_winerror._ascii.ob_base,
            & const_str_seen._ascii.ob_base,
            & const_str_old_path._ascii.ob_base,
            & const_str_ex._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(368)
ntpath_toplevel_consts_39 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 184,
    },
    .co_consts = & ntpath_toplevel_consts_39_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_39_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_39_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 572,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 572,
    .co_localsplusnames = & ntpath_toplevel_consts_39_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str__readlink_deep._ascii.ob_base,
    .co_qualname = & const_str__readlink_deep._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_39_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x03\x76\x01\x72\x6f\x7c\x03\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7c\x00\x7d\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x2e\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x73\x05\x7c\x04\x7d\x00\x09\x00\x7c\x00\x53\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x03\x76\x01\x72\x01\x8c\x6f\x7c\x00\x53\x00\x23\x00\x7c\x01\x24\x00\x72\x1a\x7d\x05\x7c\x05\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x76\x00\x72\x06\x59\x00\x64\x00\x7d\x05\x7e\x05\x7c\x00\x53\x00\x82\x00\x64\x00\x7d\x05\x7e\x05\x77\x01\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x04\x01\x00\x59\x00\x7c\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_1920 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 1920 },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_1921 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 1921 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
ntpath_toplevel_consts_40_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 5],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 21],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 32],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 50],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 53],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 65],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 67],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 87],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 123],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 161],
            & const_int_1920.ob_base,
            & const_int_1921.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_40_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_ignored_error._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
ntpath_toplevel_consts_40_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & ntpath_toplevel_consts_40_consts_1._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & ntpath_toplevel_consts_40_consts_3._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
ntpath_toplevel_consts_40_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__getfinalpathname._ascii.ob_base,
            &_Py_ID(join),
            & const_str_winerror._ascii.ob_base,
            & const_str__readlink_deep._ascii.ob_base,
            & const_str_split._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
const_str__getfinalpathname_nonstrict = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_getfinalpathname_nonstrict",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[229];
    }
ntpath_toplevel_consts_40_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 228,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x24\x00\x1c\x58\x01\xd0\x08\x18\xf0\x08\x00\x10\x14\x90\x42\x90\x51\x88\x78\x88\x04\xd9\x0e\x12\xf0\x02\x17\x0d\x3a\xdc\x17\x28\xa8\x14\xd3\x17\x2e\x90\x04\xd9\x2b\x2f\x94\x74\x98\x44\xa0\x24\xd3\x17\x27\xd0\x10\x39\xb0\x54\xd0\x10\x39\xf0\x2c\x00\x10\x14\x88\x0b\xf8\xf0\x2b\x00\x14\x21\xf2\x00\x14\x0d\x3a\xd8\x13\x15\x97\x3b\x91\x3b\xd0\x26\x36\xd1\x13\x36\xd8\x14\x19\xf0\x02\x0a\x11\x19\xf4\x08\x00\x20\x2e\xa8\x64\xd8\x3c\x49\xf4\x03\x01\x20\x4b\x01\x90\x48\xe0\x17\x1f\xa0\x34\xd2\x17\x27\xd9\x37\x3b\x9c\x74\xa0\x48\xa8\x64\xd4\x1f\x33\xc0\x18\xd5\x18\x49\xf0\x03\x00\x18\x28\xf8\xe0\x17\x24\xf2\x00\x02\x11\x19\xe1\x14\x18\xf0\x05\x02\x11\x19\xfa\xf4\x06\x00\x1e\x23\xa0\x34\x9b\x5b\x91\x0a\x90\x04\x90\x64\xf1\x08\x00\x14\x18\xa1\x04\xd8\x1b\x1f\xa0\x24\x99\x3b\xd5\x14\x26\xd9\x2b\x2f\x94\x74\x98\x44\xa0\x24\xd4\x17\x27\xb0\x54\x95\x04\xfb\xf0\x29\x14\x0d\x3a\xfa\xf2\x09\x00\x0f\x13\xf8",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[78];
    }
ntpath_toplevel_consts_40_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 77,
    },
    .ob_shash = -1,
    .ob_sval = "\x8b\x18\x28\x00\xa4\x01\x28\x00\xa8\x05\x42\x26\x03\xad\x0f\x42\x21\x03\xbd\x21\x41\x25\x02\xc1\x1e\x01\x42\x26\x03\xc1\x24\x01\x42\x21\x03\xc1\x25\x05\x41\x2d\x05\xc1\x2a\x02\x42\x21\x03\xc1\x2c\x01\x41\x2d\x05\xc1\x2d\x19\x42\x21\x03\xc2\x06\x01\x42\x26\x03\xc2\x0c\x10\x42\x21\x03\xc2\x21\x05\x42\x26\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_new_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "new_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
ntpath_toplevel_consts_40_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_ignored_error._ascii.ob_base,
            & const_str_allowed_winerror._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
            & const_str_ex._ascii.ob_base,
            & const_str_new_path._ascii.ob_base,
            &_Py_ID(name),
        },
    },
};
static
    struct _PyCode_DEF(346)
ntpath_toplevel_consts_40 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 173,
    },
    .co_consts = & ntpath_toplevel_consts_40_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_40_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_40_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 614,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 573,
    .co_localsplusnames = & ntpath_toplevel_consts_40_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str__getfinalpathname_nonstrict._ascii.ob_base,
    .co_qualname = & const_str__getfinalpathname_nonstrict._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_40_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x02\x7c\x00\x64\x00\x64\x02\x1a\x00\x7d\x03\x7c\x00\x72\x1c\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x03\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x53\x00\x7c\x03\x53\x00\x23\x00\x7c\x01\x24\x00\x72\x79\x7d\x04\x7c\x04\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x76\x01\x72\x01\x82\x00\x09\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xac\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x7c\x00\x6b\x37\x00\x00\x72\x15\x7c\x03\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x6e\x01\x7c\x05\x63\x02\x59\x00\x64\x00\x7d\x04\x7e\x04\x53\x00\x6e\x0b\x23\x00\x7c\x01\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x04\x77\x00\x78\x03\x59\x00\x77\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x00\x7d\x06\x7c\x00\x72\x0c\x7c\x06\x73\x0a\x7c\x00\x7c\x03\x7a\x00\x00\x00\x63\x02\x59\x00\x64\x00\x7d\x04\x7e\x04\x53\x00\x7c\x03\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x03\xab\x02\x00\x00\x00\x00\x00\x00\x6e\x01\x7c\x06\x7d\x03\x59\x00\x64\x00\x7d\x04\x7e\x04\x6e\x08\x64\x00\x7d\x04\x7e\x04\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x7c\x00\x72\x01\x8c\xa2\x8c\x87",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
ntpath_toplevel_consts_43_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\\\\?\\",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
ntpath_toplevel_consts_43_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\\\\",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
ntpath_toplevel_consts_43_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\\\\.\\NUL",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
ntpath_toplevel_consts_43_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\\\?\\",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
ntpath_toplevel_consts_43_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\\\.\\NUL",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
ntpath_toplevel_consts_43_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            Py_None,
            & ntpath_toplevel_consts_43_consts_1.ob_base.ob_base,
            & ntpath_toplevel_consts_19_consts_4.ob_base.ob_base,
            & ntpath_toplevel_consts_43_consts_3.ob_base.ob_base,
            & ntpath_toplevel_consts_43_consts_4.ob_base.ob_base,
            & ntpath_toplevel_consts_43_consts_5._ascii.ob_base,
            & ntpath_toplevel_consts_19_consts_9._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_22_consts_6._ascii.ob_base,
            & ntpath_toplevel_consts_43_consts_8._ascii.ob_base,
            Py_True,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & ntpath_toplevel_consts_40_consts_3._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[21];
        }_object;
    }
ntpath_toplevel_consts_43_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 21,
        },
        .ob_item = {
            & const_str_normpath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_os._ascii.ob_base,
            & const_str_getcwdb._ascii.ob_base,
            & const_str_normcase._ascii.ob_base,
            & const_str_fsencode._ascii.ob_base,
            & const_str_devnull._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_ALLOW_MISSING._ascii.ob_base,
            & const_str_FileNotFoundError._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            &_Py_ID(join),
            & const_str__getfinalpathname._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_str._ascii.ob_base,
            & const_str_winerror._ascii.ob_base,
            & const_str__getfinalpathname_nonstrict._ascii.ob_base,
            &_Py_ID(len),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[465];
    }
ntpath_toplevel_consts_43_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 464,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x17\x98\x04\x8b\x7e\x88\x04\xdc\x0b\x15\x90\x64\x9c\x45\xd4\x0b\x22\xd8\x15\x1f\x88\x46\xd8\x19\x28\x88\x4a\xd8\x1d\x24\x88\x4e\xdc\x12\x14\x97\x2a\x91\x2a\x93\x2c\x88\x43\xe4\x0f\x17\x98\x04\x8b\x7e\xa4\x18\xac\x22\xaf\x2b\xa9\x2b\xb4\x67\xd3\x2a\x3e\xd3\x21\x3f\xd2\x0f\x3f\xd8\x17\x24\xe0\x15\x1e\x88\x46\xd8\x19\x27\x88\x4a\xd8\x1d\x23\x88\x4e\xdc\x12\x14\x97\x29\x91\x29\x93\x2b\x88\x43\xe4\x0f\x17\x98\x04\x8b\x7e\xa4\x18\xac\x27\xd3\x21\x32\xd2\x0f\x32\xd8\x17\x23\xd8\x15\x19\x97\x5f\x91\x5f\xa0\x56\xd3\x15\x2c\x88\x0a\xe0\x0b\x11\x94\x5d\xd1\x0b\x22\xdc\x1c\x2d\x88\x4d\xd8\x15\x19\x89\x46\xd9\x0d\x13\xd8\x1c\x1e\x89\x4d\xe4\x1c\x23\x88\x4d\xe1\x0f\x19\xa4\x25\xa8\x04\xa4\x2b\xdc\x13\x17\x98\x03\x98\x54\x93\x3f\x88\x44\xf0\x02\x0e\x09\x4c\x01\xdc\x13\x24\xa0\x54\xd3\x13\x2a\x88\x44\xd8\x1f\x20\xd0\x0c\x1c\xf1\x20\x00\x10\x1a\x98\x64\x9f\x6f\x99\x6f\xa8\x66\xd4\x1e\x35\xf0\x06\x00\x10\x14\x8f\x7f\x89\x7f\x98\x7a\xd4\x0f\x2a\xd8\x18\x26\xa8\x14\xac\x63\xb0\x2a\xab\x6f\xd0\x2e\x3e\xd0\x29\x3f\xd1\x18\x3f\x91\x05\xe0\x18\x1c\x9c\x53\xa0\x16\x9b\x5b\x98\x5c\xd0\x18\x2a\x90\x05\xf0\x04\x0b\x0d\x21\xdc\x13\x24\xa0\x55\xd3\x13\x2b\xa8\x74\xd2\x13\x33\xd8\x1b\x20\x90\x44\xf0\x14\x00\x10\x14\x88\x0b\x88\x74\x88\x0b\xf8\xf4\x47\x01\x00\x10\x1a\xf2\x00\x07\x09\x22\xf1\x0a\x00\x10\x16\xdc\x16\x1d\x9c\x63\xa0\x22\x9b\x67\xd3\x16\x26\xa8\x44\xd0\x10\x30\xdc\x13\x1b\x98\x44\x93\x3e\x8d\x44\xfb\xd8\x0f\x1c\xf2\x00\x03\x09\x4c\x01\xd8\x1f\x21\x9f\x7b\x99\x7b\xd0\x0c\x1c\xdc\x13\x2e\xa8\x74\xd8\x3d\x4a\xf4\x03\x01\x14\x4c\x01\x8d\x44\xfb\xf0\x05\x03\x09\x4c\x01\xfb\xf4\x24\x00\x14\x1e\xf2\x00\x03\x0d\x15\xf3\x06\x00\x11\x15\xf0\x0c\x00\x10\x14\x88\x0b\xfb\xf4\x0b\x00\x14\x1b\xf2\x00\x04\x0d\x21\xf0\x06\x00\x14\x16\x97\x3b\x91\x3b\xd0\x22\x32\xd2\x13\x32\xd8\x1b\x20\x90\x44\xfb\xd8\x0f\x13\x88\x0b\xfb\xf0\x0b\x04\x0d\x21\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[67];
    }
ntpath_toplevel_consts_43_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 66,
    },
    .ob_shash = -1,
    .ob_sval = "\xc3\x22\x0d\x45\x08\x00\xc4\x34\x10\x46\x26\x00\xc5\x08\x09\x46\x23\x03\xc5\x11\x22\x45\x38\x03\xc5\x38\x08\x46\x23\x03\xc6\x00\x19\x46\x1e\x03\xc6\x1e\x05\x46\x23\x03\xc6\x26\x09\x47\x1d\x03\xc6\x35\x0c\x47\x1d\x03\xc7\x01\x11\x47\x18\x03\xc7\x18\x05\x47\x1d\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_new_unc_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "new_unc_prefix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_had_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "had_prefix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_initial_winerror = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "initial_winerror",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_spath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
ntpath_toplevel_consts_43_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(strict),
            & const_str_prefix._ascii.ob_base,
            & const_str_unc_prefix._ascii.ob_base,
            & const_str_new_unc_prefix._ascii.ob_base,
            &_Py_ID(cwd),
            & const_str_had_prefix._ascii.ob_base,
            & const_str_ignored_error._ascii.ob_base,
            & const_str_initial_winerror._ascii.ob_base,
            & const_str_ex._ascii.ob_base,
            & const_str_spath._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(960)
ntpath_toplevel_consts_43 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 480,
    },
    .co_consts = & ntpath_toplevel_consts_43_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_43_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_43_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 664,
    .co_nlocalsplus = 11,
    .co_nlocals = 11,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 574,
    .co_localsplusnames = & ntpath_toplevel_consts_43_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_realpath._ascii.ob_base,
    .co_qualname = & const_str_realpath._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_43_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x49\x64\x01\x7d\x02\x64\x02\x7d\x03\x64\x03\x7d\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x72\x37\x79\x04\x64\x05\x7d\x02\x64\x06\x7d\x03\x64\x07\x7d\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x72\x01\x79\x08\x7c\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x01\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x09\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x64\x09\x7d\x01\x6e\x0b\x7c\x01\x72\x03\x64\x0a\x7d\x07\x6e\x06\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x06\x73\x17\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x0c\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x09\x00\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x64\x0b\x7d\x08\x7c\x06\x73\x55\x7c\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x72\x44\x7c\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x72\x12\x7c\x04\x7c\x00\x74\x29\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x1a\x00\x7a\x00\x00\x00\x7d\x0a\x6e\x0e\x7c\x00\x74\x29\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x1a\x00\x7d\x0a\x09\x00\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x6b\x28\x00\x00\x72\x02\x7c\x0a\x7d\x00\x7c\x00\x53\x00\x7c\x00\x53\x00\x23\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x2c\x7d\x09\x7c\x01\x72\x15\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x82\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x59\x00\x64\x00\x7d\x09\x7e\x09\x8c\x89\x64\x00\x7d\x09\x7e\x09\x77\x01\x7c\x07\x24\x00\x72\x23\x7d\x09\x7c\x09\x6a\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x74\x27\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x07\xac\x0c\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x59\x00\x64\x00\x7d\x09\x7e\x09\x8c\xaf\x64\x00\x7d\x09\x7e\x09\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0b\x7d\x09\x59\x00\x64\x00\x7d\x09\x7e\x09\x7c\x00\x53\x00\x64\x00\x7d\x09\x7e\x09\x77\x01\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x1c\x7d\x09\x7c\x09\x6a\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7f\x08\x6b\x28\x00\x00\x72\x02\x7c\x0a\x7d\x00\x59\x00\x64\x00\x7d\x09\x7e\x09\x7c\x00\x53\x00\x64\x00\x7d\x09\x7e\x09\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_44_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_abspath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
ntpath_toplevel_consts_44_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x16\x90\x74\x8b\x7d\xd0\x08\x1c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_44_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(strict),
        },
    },
};
static
    struct _PyCode_DEF(24)
ntpath_toplevel_consts_44 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_44_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 569,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 575,
    .co_localsplusnames = & ntpath_toplevel_consts_44_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_realpath._ascii.ob_base,
    .co_qualname = & const_str_realpath._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_44_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
ntpath_toplevel_consts_46_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return a relative version of a path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
ntpath_toplevel_consts_46_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "no path specified",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
ntpath_toplevel_consts_46_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path is on mount ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
ntpath_toplevel_consts_46_consts_10 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ", start on mount ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
ntpath_toplevel_consts_46_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & ntpath_toplevel_consts_46_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            & ntpath_toplevel_consts_34_consts_4.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & ntpath_toplevel_consts_2._ascii.ob_base,
            Py_None,
            & ntpath_toplevel_consts_46_consts_8._ascii.ob_base,
            & ntpath_toplevel_consts_46_consts_9._ascii.ob_base,
            & ntpath_toplevel_consts_46_consts_10._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & const_str_relpath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
ntpath_toplevel_consts_46_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_ValueError._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str_normpath._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            & const_str_normcase._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            & const_str_zip._ascii.ob_base,
            &_Py_ID(len),
            &_Py_ID(join),
            & const_str_TypeError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_BytesWarning._ascii.ob_base,
            & const_str_DeprecationWarning._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            & const_str__check_arg_types._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[437];
    }
ntpath_toplevel_consts_46_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 436,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0d\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x0e\x13\x88\x03\xd8\x11\x15\x88\x06\xd8\x11\x16\x89\x06\xe0\x0e\x12\x88\x03\xd8\x11\x14\x88\x06\xd8\x11\x15\x88\x06\xe0\x07\x0c\x80\x7d\xd8\x10\x16\x88\x05\xe1\x0b\x0f\xdc\x0e\x18\xd0\x19\x2c\xd3\x0e\x2d\xd0\x08\x2d\xe4\x0c\x0e\x8f\x49\x89\x49\x90\x65\xd3\x0c\x1c\x80\x45\xf0\x02\x18\x05\x0e\xdc\x14\x1b\x9c\x48\xa0\x55\x9b\x4f\xd3\x14\x2c\x88\x09\xdc\x13\x1a\x9c\x38\xa0\x44\x9b\x3e\xd3\x13\x2a\x88\x08\xdc\x25\x2e\xa8\x79\xd3\x25\x39\xd1\x08\x22\x88\x0b\x90\x51\x98\x0a\xdc\x23\x2c\xa8\x58\xd3\x23\x36\xd1\x08\x20\x88\x0a\x90\x41\x90\x79\xdc\x0b\x13\x90\x4b\xd3\x0b\x20\xa4\x48\xa8\x5a\xd3\x24\x38\xd2\x0b\x38\xdd\x12\x1c\xda\x10\x1a\x99\x4b\xf0\x03\x01\x1e\x29\xf3\x00\x01\x13\x2a\xf0\x00\x01\x0d\x2a\xf0\x06\x00\x22\x2c\xd7\x21\x31\xd1\x21\x31\xb0\x23\xd4\x21\x36\xd3\x15\x3c\xd1\x21\x36\x98\x41\xba\x21\x92\x61\xd0\x21\x36\x88\x0a\xd0\x15\x3c\xd8\x20\x29\xa7\x0f\xa1\x0f\xb0\x03\xd4\x20\x34\xd3\x14\x3a\xd1\x20\x34\x98\x31\xba\x01\x92\x51\xd0\x20\x34\x88\x09\xd0\x14\x3a\xe0\x0c\x0d\x88\x01\xdc\x16\x19\x98\x2a\xa0\x69\xd6\x16\x30\x89\x46\x88\x42\x90\x02\xdc\x0f\x17\x98\x02\x8b\x7c\x9c\x78\xa8\x02\x9b\x7c\xd2\x0f\x2b\xd9\x10\x15\xd8\x0c\x0d\x90\x11\x89\x46\x89\x41\xf0\x07\x00\x17\x31\xf0\x0a\x00\x15\x1b\x90\x38\x9c\x73\xa0\x3a\x9b\x7f\xa8\x71\xd1\x1f\x30\xd1\x13\x31\xb0\x49\xb8\x61\xb8\x62\xb0\x4d\xd1\x13\x41\x88\x08\xd9\x0f\x17\xd8\x13\x19\x88\x4d\xdc\x0f\x13\x90\x58\x88\x7f\xd0\x08\x1e\xf9\xf2\x19\x00\x16\x3d\xf9\xda\x14\x3a\xf8\xf4\x18\x00\x0d\x16\x94\x7a\xa4\x3e\xb4\x3c\xd4\x41\x53\xd0\x0b\x54\xf2\x00\x02\x05\x0e\xdc\x08\x13\xd7\x08\x24\xd1\x08\x24\xa0\x59\xb0\x04\xb0\x65\xd4\x08\x3c\xd8\x08\x0d\xf0\x05\x02\x05\x0e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
ntpath_toplevel_consts_46_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x1a\x42\x01\x45\x2c\x00\xc3\x1b\x07\x45\x22\x04\xc3\x23\x04\x45\x22\x04\xc3\x27\x15\x45\x2c\x00\xc3\x3c\x07\x45\x27\x04\xc4\x04\x04\x45\x27\x04\xc4\x08\x41\x11\x45\x2c\x00\xc5\x1a\x07\x45\x2c\x00\xc5\x22\x0a\x45\x2c\x00\xc5\x2c\x37\x46\x23\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_start_abs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "start_abs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_path_abs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_abs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_start_drive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "start_drive",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_start_rest = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "start_rest",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_path_drive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_drive",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_path_rest = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_rest",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_start_list = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "start_list",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_path_list = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_list",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_e1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "e1",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_e2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "e2",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_rel_list = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rel_list",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
ntpath_toplevel_consts_46_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(start),
            &_Py_ID(sep),
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            & const_str_start_abs._ascii.ob_base,
            & const_str_path_abs._ascii.ob_base,
            & const_str_start_drive._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            & const_str_start_rest._ascii.ob_base,
            & const_str_path_drive._ascii.ob_base,
            & const_str_path_rest._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
            & const_str_start_list._ascii.ob_base,
            & const_str_path_list._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            & const_str_e1._ascii.ob_base,
            & const_str_e2._ascii.ob_base,
            & const_str_rel_list._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
ntpath_toplevel_consts_46_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "                   ",
};
static
    struct _PyCode_DEF(844)
ntpath_toplevel_consts_46 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 422,
    },
    .co_consts = & ntpath_toplevel_consts_46_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_46_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_46_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 26 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 738,
    .co_nlocalsplus = 19,
    .co_nlocals = 19,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 576,
    .co_localsplusnames = & ntpath_toplevel_consts_46_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & ntpath_toplevel_consts_46_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_relpath._ascii.ob_base,
    .co_qualname = & const_str_relpath._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_46_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x07\x64\x01\x7d\x02\x64\x02\x7d\x03\x64\x03\x7d\x04\x6e\x06\x64\x04\x7d\x02\x64\x05\x7d\x03\x64\x06\x7d\x04\x7c\x01\x80\x02\x7c\x03\x7d\x01\x7c\x00\x73\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x07\x7d\x08\x7d\x09\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x0a\x7d\x08\x7d\x0b\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x11\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\x7c\x0a\x9b\x02\x64\x0a\x7c\x07\x9b\x02\x9d\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x09\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x8f\x0c\x63\x02\x67\x00\x63\x02\x5d\x07\x00\x00\x7d\x0c\x7c\x0c\x73\x01\x8c\x06\x7c\x0c\x91\x02\x8c\x09\x04\x00\x7d\x0d\x7d\x0c\x7c\x0b\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x8f\x0c\x63\x02\x67\x00\x63\x02\x5d\x07\x00\x00\x7d\x0c\x7c\x0c\x73\x01\x8c\x06\x7c\x0c\x91\x02\x8c\x09\x04\x00\x7d\x0e\x7d\x0c\x64\x0b\x7d\x0f\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x7c\x0e\xab\x02\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x23\x00\x00\x5c\x02\x00\x00\x7d\x10\x7d\x11\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x10\xab\x01\x00\x00\x00\x00\x00\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x11\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x02\x01\x00\x6e\x07\x7c\x0f\x64\x0c\x7a\x0d\x00\x00\x7d\x0f\x8c\x25\x04\x00\x7c\x04\x67\x01\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x0f\x7a\x0a\x00\x00\x7a\x05\x00\x00\x7c\x0e\x7c\x0f\x64\x07\x1a\x00\x7a\x00\x00\x00\x7d\x12\x7c\x12\x73\x02\x7c\x03\x53\x00\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x12\x8e\x00\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x0c\x77\x00\x63\x02\x01\x00\x63\x02\x7d\x0c\x77\x00\x23\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x66\x05\x24\x00\x72\x19\x01\x00\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\x7c\x00\x7c\x01\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[69];
    }
ntpath_toplevel_consts_47_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 68,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Given a sequence of path names, returns the longest common sub-path.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
ntpath_toplevel_consts_47_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "commonpath() arg is an empty sequence",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
ntpath_toplevel_consts_47_consts_10 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Can't mix absolute and relative paths",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
ntpath_toplevel_consts_47_consts_11 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Paths don't have the same drive",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
ntpath_toplevel_consts_47_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & ntpath_toplevel_consts_47_consts_0._ascii.ob_base,
            & ntpath_toplevel_consts_47_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_Py_SINGLETON(bytes_characters[92]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & ntpath_toplevel_consts_47_consts_10._ascii.ob_base,
            & ntpath_toplevel_consts_47_consts_11._ascii.ob_base,
            Py_None,
            & const_str_commonpath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[20];
        }_object;
    }
ntpath_toplevel_consts_47_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 20,
        },
        .ob_item = {
            & const_str_ValueError._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_map._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_splitroot._ascii.ob_base,
            &_Py_ID(replace),
            & const_str_lower._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            &_Py_ID(len),
            & const_str_min._ascii.ob_base,
            & const_str_max._ascii.ob_base,
            & const_str_enumerate._ascii.ob_base,
            &_Py_ID(join),
            & const_str_TypeError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            & const_str__check_arg_types._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[558];
    }
ntpath_toplevel_consts_47_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 557,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf1\x06\x00\x0c\x11\xdc\x0e\x18\xd0\x19\x40\xd3\x0e\x41\xd0\x08\x41\xe4\x0c\x11\x94\x23\x94\x62\x97\x69\x91\x69\xa0\x15\xd3\x12\x27\xd3\x0c\x28\x80\x45\xdc\x07\x11\x90\x25\x98\x01\x91\x28\x9c\x45\xd4\x07\x22\xd8\x0e\x13\x88\x03\xd8\x11\x15\x88\x06\xd8\x11\x15\x89\x06\xe0\x0e\x12\x88\x03\xd8\x11\x14\x88\x06\xd8\x11\x14\x88\x06\xf0\x04\x1e\x05\x0e\xd9\x4a\x4f\xd3\x16\x50\xc9\x25\xc0\x51\x94\x79\xa0\x11\xa7\x19\xa1\x19\xa8\x36\xb0\x33\xd3\x21\x37\xd7\x21\x3d\xd1\x21\x3d\xd3\x21\x3f\xd5\x17\x40\xc8\x25\x88\x0b\xd0\x16\x50\xd9\x33\x3e\xd5\x16\x3f\xb1\x3b\xa9\x07\xa8\x01\xa8\x31\xa8\x61\x90\x71\x97\x77\x91\x77\x98\x73\x95\x7c\xb0\x3b\x88\x0b\xd2\x16\x3f\xe4\x0b\x0e\xa1\x1b\xd5\x0f\x2d\xa1\x1b\x91\x67\x90\x61\x98\x11\x98\x41\x92\x01\xa0\x1b\xd3\x0f\x2d\xd3\x0b\x2e\xb0\x21\xd2\x0b\x33\xdc\x12\x1c\xd0\x1d\x44\xd3\x12\x45\xd0\x0c\x45\xf4\x0a\x00\x0c\x0f\xa1\x1b\xd5\x0f\x2d\xa1\x1b\x91\x67\x90\x61\x98\x11\x98\x41\x92\x01\xa0\x1b\xd3\x0f\x2d\xd3\x0b\x2e\xb0\x21\xd2\x0b\x33\xdc\x12\x1c\xd0\x1d\x3e\xd3\x12\x3f\xd0\x0c\x3f\xe4\x1c\x25\xa0\x65\xa8\x41\xa1\x68\xd7\x26\x36\xd1\x26\x36\xb0\x76\xb8\x73\xd3\x26\x43\xd3\x1c\x44\xd1\x08\x19\x88\x05\x88\x74\x90\x54\xd8\x11\x15\x97\x1a\x91\x1a\x98\x43\x93\x1f\x88\x06\xd9\x1d\x23\xd3\x11\x39\x99\x56\x98\x01\xa2\x71\xa8\x51\xb0\x26\xab\x5b\x92\x21\x98\x56\x88\x06\xd0\x11\x39\xe1\x44\x4f\xd4\x16\x50\xc1\x4b\xb8\x71\xa1\x31\xd3\x17\x3a\xa1\x31\x98\x61\xaa\x01\xa8\x61\xb0\x36\xab\x6b\x9a\x01\xa0\x31\xd3\x17\x3a\xc0\x4b\x88\x0b\xd1\x16\x50\xdc\x0d\x10\x90\x1b\xd3\x0d\x1d\x88\x02\xdc\x0d\x10\x90\x1b\xd3\x0d\x1d\x88\x02\xdc\x14\x1d\x98\x62\x96\x4d\x89\x44\x88\x41\x88\x71\xd8\x0f\x10\x90\x42\x90\x71\x91\x45\x8b\x7a\xd8\x19\x1f\xa0\x02\xa0\x11\x98\x1a\x90\x06\xd9\x10\x15\xf0\x07\x00\x15\x22\xf0\x0a\x00\x16\x1c\x98\x48\x9c\x53\xa0\x12\x9b\x57\xd0\x15\x25\x88\x46\xe0\x0f\x14\x90\x74\x89\x7c\x98\x63\x9f\x68\x99\x68\xa0\x76\xd3\x1e\x2e\xd1\x0f\x2e\xd0\x08\x2e\xf9\xf2\x35\x00\x17\x51\x01\xf9\xdc\x16\x3f\xf9\xe4\x0f\x2d\xf9\xf4\x0c\x00\x10\x2e\xf9\xf2\x0a\x00\x12\x3a\xf9\xe2\x17\x3a\xf9\xd3\x16\x50\xf8\xf4\x16\x00\x0d\x16\x94\x7e\xd0\x0b\x26\xf2\x00\x02\x05\x0e\xdc\x08\x13\xd7\x08\x24\xd1\x08\x24\xa0\x5c\xd0\x08\x3a\xb0\x45\xd3\x08\x3a\xd8\x08\x0d\xf0\x05\x02\x05\x0e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[146];
    }
ntpath_toplevel_consts_47_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 145,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x12\x04\x48\x06\x00\xc1\x16\x30\x47\x1c\x04\xc2\x06\x08\x48\x06\x00\xc2\x0e\x1c\x47\x21\x08\xc2\x2a\x0f\x48\x06\x00\xc2\x39\x0d\x47\x28\x0c\xc3\x06\x22\x48\x06\x00\xc3\x28\x0d\x47\x2f\x0c\xc3\x35\x41\x0e\x48\x06\x00\xc5\x03\x07\x47\x36\x04\xc5\x0b\x05\x47\x36\x04\xc5\x11\x04\x47\x36\x04\xc5\x15\x07\x48\x06\x00\xc5\x1c\x09\x48\x00\x06\xc5\x25\x07\x47\x3b\x0c\xc5\x2d\x05\x47\x3b\x0c\xc5\x33\x04\x47\x3b\x0c\xc5\x37\x05\x48\x00\x06\xc5\x3c\x32\x48\x06\x00\xc6\x2f\x2c\x48\x06\x00\xc7\x1c\x1f\x48\x06\x00\xc7\x3b\x05\x48\x00\x06\xc8\x00\x06\x48\x06\x00\xc8\x06\x27\x48\x2d\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_drivesplits = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "drivesplits",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_split_paths = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "split_paths",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_common = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "common",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[18];
        }_object;
    }
ntpath_toplevel_consts_47_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 18,
        },
        .ob_item = {
            & const_str_paths._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_altsep._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            & const_str_drivesplits._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[100],
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            & const_str_split_paths._ascii.ob_base,
            & const_str_drive._ascii.ob_base,
            & const_str_root._ascii.ob_base,
            &_Py_ID(path),
            & const_str_common._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[99],
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            & const_str_s1._ascii.ob_base,
            & const_str_s2._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
ntpath_toplevel_consts_47_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "                  ",
};
static
    struct _PyCode_DEF(1120)
ntpath_toplevel_consts_47 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 560,
    },
    .co_consts = & ntpath_toplevel_consts_47_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_47_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_47_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 28 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 10,
    .co_firstlineno = 794,
    .co_nlocalsplus = 18,
    .co_nlocals = 18,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 577,
    .co_localsplusnames = & ntpath_toplevel_consts_47_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & ntpath_toplevel_consts_47_localspluskinds.ob_base.ob_base,
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_commonpath._ascii.ob_base,
    .co_qualname = & const_str_commonpath._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_47_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x73\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\x19\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x07\x64\x03\x7d\x01\x64\x04\x7d\x02\x64\x05\x7d\x03\x6e\x06\x64\x06\x7d\x01\x64\x07\x7d\x02\x64\x08\x7d\x03\x09\x00\x7c\x00\x44\x00\x8f\x04\x63\x02\x67\x00\x63\x02\x5d\x2b\x00\x00\x7d\x04\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x91\x02\x8c\x2d\x04\x00\x7d\x05\x7d\x04\x7c\x05\x44\x00\x8f\x06\x8f\x07\x8f\x04\x63\x04\x67\x00\x63\x02\x5d\x17\x00\x00\x5c\x03\x00\x00\x7d\x06\x7d\x07\x7d\x04\x7c\x04\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x91\x02\x8c\x19\x04\x00\x7d\x08\x7d\x07\x7d\x06\x7d\x04\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x44\x00\x8f\x06\x8f\x07\x8f\x04\x63\x04\x68\x00\x63\x02\x5d\x08\x00\x00\x5c\x03\x00\x00\x7d\x06\x7d\x07\x7d\x04\x7c\x07\x92\x02\x8c\x0a\x04\x00\x63\x04\x7d\x04\x7d\x07\x7d\x06\xab\x01\x00\x00\x00\x00\x00\x00\x64\x09\x6b\x37\x00\x00\x72\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x44\x00\x8f\x06\x8f\x07\x8f\x04\x63\x04\x68\x00\x63\x02\x5d\x08\x00\x00\x5c\x03\x00\x00\x7d\x06\x7d\x07\x7d\x04\x7c\x06\x92\x02\x8c\x0a\x04\x00\x63\x04\x7d\x04\x7d\x07\x7d\x06\xab\x01\x00\x00\x00\x00\x00\x00\x64\x09\x6b\x37\x00\x00\x72\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\x19\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x09\x7d\x0a\x7d\x0b\x7c\x0b\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0c\x7c\x0c\x44\x00\x8f\x0d\x63\x02\x67\x00\x63\x02\x5d\x0d\x00\x00\x7d\x0d\x7c\x0d\x73\x01\x8c\x06\x7c\x0d\x7c\x03\x6b\x37\x00\x00\x73\x01\x8c\x0c\x7c\x0d\x91\x02\x8c\x0f\x04\x00\x7d\x0c\x7d\x0d\x7c\x08\x44\x00\x8f\x0e\x8f\x0d\x63\x03\x67\x00\x63\x02\x5d\x1b\x00\x00\x7d\x0e\x7c\x0e\x44\x00\x8f\x0d\x63\x02\x67\x00\x63\x02\x5d\x0d\x00\x00\x7d\x0d\x7c\x0d\x73\x01\x8c\x06\x7c\x0d\x7c\x03\x6b\x37\x00\x00\x73\x01\x8c\x0c\x7c\x0d\x91\x02\x8c\x0f\x04\x00\x63\x02\x7d\x0d\x91\x02\x8c\x1d\x04\x00\x7d\x08\x7d\x0e\x7d\x0d\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0f\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x10\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0f\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x14\x00\x00\x5c\x02\x00\x00\x7d\x11\x7d\x0d\x7c\x0d\x7c\x10\x7c\x11\x19\x00\x00\x00\x6b\x37\x00\x00\x73\x01\x8c\x0f\x7c\x0c\x64\x0c\x7c\x11\x1a\x00\x7d\x0c\x01\x00\x6e\x0f\x04\x00\x7c\x0c\x64\x0c\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0f\xab\x01\x00\x00\x00\x00\x00\x00\x1a\x00\x7d\x0c\x7c\x09\x7c\x0a\x7a\x00\x00\x00\x7c\x01\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x04\x77\x00\x63\x02\x01\x00\x63\x04\x7d\x04\x7d\x07\x7d\x06\x77\x00\x63\x02\x01\x00\x63\x04\x7d\x04\x7d\x07\x7d\x06\x77\x00\x63\x02\x01\x00\x63\x04\x7d\x04\x7d\x07\x7d\x06\x77\x00\x63\x02\x01\x00\x63\x02\x7d\x0d\x77\x00\x63\x02\x01\x00\x63\x02\x7d\x0d\x77\x00\x63\x02\x01\x00\x63\x03\x7d\x0d\x7d\x0e\x77\x00\x23\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x18\x01\x00\x74\x25\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\x67\x01\x7c\x00\xa2\x01\xad\x06\x8e\x00\x01\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_48 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_isdir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_49 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_isfile._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__path_islink = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_islink",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_50 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_islink._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__path_exists = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_exists",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_51 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_exists._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__path_isdevdrive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_path_isdevdrive",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
ntpath_toplevel_consts_52 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__path_isdevdrive._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[65];
    }
ntpath_toplevel_consts_53_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 64,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Determines whether the specified path is on a Windows Dev Drive.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
ntpath_toplevel_consts_53_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & ntpath_toplevel_consts_53_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
ntpath_toplevel_consts_53_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__path_isdevdrive._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_isdevdrive = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isdevdrive",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[41];
    }
ntpath_toplevel_consts_53_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 40,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x09\x19\xdc\x13\x23\xa4\x47\xa8\x44\xa3\x4d\xd3\x13\x32\xd0\x0c\x32\xf8\xdc\x0f\x16\xf2\x00\x01\x09\x19\xd9\x13\x18\xf0\x03\x01\x09\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
ntpath_toplevel_consts_53_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x13\x16\x00\x96\x09\x22\x03\xa1\x01\x22\x03",
};
static
    struct _PyCode_DEF(74)
ntpath_toplevel_consts_53 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 37,
    },
    .co_consts = & ntpath_toplevel_consts_53_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_53_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_consts_53_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 864,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 578,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_isdevdrive._ascii.ob_base,
    .co_qualname = & const_str_isdevdrive._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_53_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
ntpath_toplevel_consts_54_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x06\x00\x10\x15",
};
static
    struct _PyCode_DEF(4)
ntpath_toplevel_consts_54 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & ntpath_toplevel_consts_53_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 859,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 579,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = & const_str_isdevdrive._ascii.ob_base,
    .co_qualname = & const_str_isdevdrive._ascii.ob_base,
    .co_linetable = & ntpath_toplevel_consts_54_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[56];
        }_object;
    }
ntpath_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 56,
        },
        .ob_item = {
            & ntpath_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & ntpath_toplevel_consts_2._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            (PyObject *)&_Py_SINGLETON(strings).ascii[59],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            & ntpath_toplevel_consts_6._ascii.ob_base,
            & const_str_nul._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & codecs_toplevel_consts_3._object.ob_base.ob_base,
            & ntpath_toplevel_consts_11._object.ob_base.ob_base,
            & ntpath_toplevel_consts_12.ob_base.ob_base,
            & ntpath_toplevel_consts_13._object.ob_base.ob_base,
            & ntpath_toplevel_consts_14.ob_base.ob_base,
            & ntpath_toplevel_consts_15.ob_base.ob_base,
            & ntpath_toplevel_consts_16.ob_base.ob_base,
            & ntpath_toplevel_consts_17.ob_base.ob_base,
            & ntpath_toplevel_consts_18.ob_base.ob_base,
            & ntpath_toplevel_consts_19.ob_base.ob_base,
            & ntpath_toplevel_consts_20.ob_base.ob_base,
            & ntpath_toplevel_consts_21.ob_base.ob_base,
            & ntpath_toplevel_consts_22.ob_base.ob_base,
            & ntpath_toplevel_consts_23.ob_base.ob_base,
            & const_str_st_reparse_tag._ascii.ob_base,
            & ntpath_toplevel_consts_25.ob_base.ob_base,
            & ntpath_toplevel_consts_26.ob_base.ob_base,
            & ntpath_toplevel_consts_27.ob_base.ob_base,
            & ntpath_toplevel_consts_28._object.ob_base.ob_base,
            & ntpath_toplevel_consts_29.ob_base.ob_base,
            & ntpath_toplevel_consts_30.ob_base.ob_base,
            & ntpath_toplevel_consts_31._ascii.ob_base,
            & ntpath_toplevel_consts_32.ob_base.ob_base,
            & ntpath_toplevel_consts_33._object.ob_base.ob_base,
            & ntpath_toplevel_consts_34.ob_base.ob_base,
            & ntpath_toplevel_consts_35._object.ob_base.ob_base,
            & ntpath_toplevel_consts_36.ob_base.ob_base,
            & ntpath_toplevel_consts_37.ob_base.ob_base,
            & ntpath_toplevel_consts_38._object.ob_base.ob_base,
            & ntpath_toplevel_consts_39.ob_base.ob_base,
            & ntpath_toplevel_consts_40.ob_base.ob_base,
            Py_False,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & ntpath_toplevel_consts_43.ob_base.ob_base,
            & ntpath_toplevel_consts_44.ob_base.ob_base,
            Py_True,
            & ntpath_toplevel_consts_46.ob_base.ob_base,
            & ntpath_toplevel_consts_47.ob_base.ob_base,
            & ntpath_toplevel_consts_48._object.ob_base.ob_base,
            & ntpath_toplevel_consts_49._object.ob_base.ob_base,
            & ntpath_toplevel_consts_50._object.ob_base.ob_base,
            & ntpath_toplevel_consts_51._object.ob_base.ob_base,
            & ntpath_toplevel_consts_52._object.ob_base.ob_base,
            & ntpath_toplevel_consts_53.ob_base.ob_base,
            & ntpath_toplevel_consts_54.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__winapi = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_winapi",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_stat_result = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "stat_result",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[69];
        }_object;
    }
ntpath_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 69,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pathsep._ascii.ob_base,
            & const_str_altsep._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            & const_str_devnull._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            &_Py_ID(__all__),
            & const_str__get_bothseps._ascii.ob_base,
            & const_str__winapi._ascii.ob_base,
            & const_str_LCMapStringEx._ascii.ob_base,
            & const_str__LCMapStringEx._ascii.ob_base,
            & const_str_LOCALE_NAME_INVARIANT._ascii.ob_base,
            & const_str__LOCALE_NAME_INVARIANT._ascii.ob_base,
            & const_str_LCMAP_LOWERCASE._ascii.ob_base,
            & const_str__LCMAP_LOWERCASE._ascii.ob_base,
            & const_str_normcase._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            &_Py_ID(join),
            & const_str_splitdrive._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            & const_str_splitext._ascii.ob_base,
            & const_str__splitext._ascii.ob_base,
            & const_str_basename._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            & const_str_stat_result._ascii.ob_base,
            & const_str_isjunction._ascii.ob_base,
            & const_str_lexists._ascii.ob_base,
            &_Py_ID(nt),
            & const_str__getvolumepathname._ascii.ob_base,
            & const_str_ismount._ascii.ob_base,
            & const_str_expanduser._ascii.ob_base,
            & const_str__varpattern._ascii.ob_base,
            & const_str__varsub._ascii.ob_base,
            & const_str__varsubb._ascii.ob_base,
            & const_str_expandvars._ascii.ob_base,
            & const_str__path_normpath._ascii.ob_base,
            & const_str_normpath._ascii.ob_base,
            & const_str__getfullpathname._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str__getfinalpathname._ascii.ob_base,
            & const_str_readlink._ascii.ob_base,
            & const_str__nt_readlink._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str__readlink_deep._ascii.ob_base,
            & const_str__getfinalpathname_nonstrict._ascii.ob_base,
            & const_str_realpath._ascii.ob_base,
            & const_str_supports_unicode_filenames._ascii.ob_base,
            & const_str_relpath._ascii.ob_base,
            & const_str_commonpath._ascii.ob_base,
            & const_str__path_isdir._ascii.ob_base,
            & const_str_isdir._ascii.ob_base,
            & const_str__path_isfile._ascii.ob_base,
            & const_str_isfile._ascii.ob_base,
            & const_str__path_islink._ascii.ob_base,
            & const_str_islink._ascii.ob_base,
            & const_str__path_exists._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str__path_isdevdrive._ascii.ob_base,
            & const_str_isdevdrive._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[515];
    }
ntpath_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 514,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x04\x04\x01\x04\xf0\x12\x00\x0a\x0d\x80\x06\xd8\x09\x0d\x80\x06\xd8\x09\x0c\x80\x06\xd8\x06\x0a\x80\x03\xd8\x0a\x0d\x80\x07\xd8\x09\x0c\x80\x06\xd8\x0a\x15\x80\x07\xd8\x0a\x0f\x80\x07\xe3\x00\x09\xdb\x00\x0a\xdb\x00\x0b\xdb\x00\x12\xdc\x00\x19\xf2\x06\x07\x0b\x1c\x80\x07\xf2\x12\x04\x01\x15\xf0\x14\x21\x01\x2c\xf7\x02\x03\x05\x2d\xf1\x00\x03\x05\x2d\xf2\x0a\x11\x05\x38\xf2\x48\x01\x10\x01\x11\xf2\x28\x2b\x01\x0e\xf2\x62\x01\x14\x01\x1e\xf2\x2e\x31\x01\x1f\xf2\x72\x01\x0d\x01\x2b\xf2\x2a\x05\x01\x38\xf0\x0c\x00\x14\x1f\xd7\x13\x28\xd1\x13\x28\xd7\x13\x30\xd1\x13\x30\x80\x08\xd4\x00\x10\xf2\x0a\x02\x01\x17\xf2\x0e\x02\x01\x17\xf1\x0e\x00\x04\x0b\x88\x32\x8f\x3e\x89\x3e\xd0\x1b\x2b\xd4\x03\x2c\xf3\x02\x06\x05\x4a\x01\xf2\x10\x03\x05\x15\xf2\x10\x06\x01\x10\xf0\x24\x03\x01\x1e\xdd\x04\x25\xf2\x06\x11\x01\x15\xf2\x3a\x2d\x01\x1f\xf0\x7a\x01\x00\x0f\x40\x01\x80\x0b\xd8\x0a\x0e\x80\x07\xd8\x0b\x0f\x80\x08\xf2\x04\x3a\x01\x1b\xf0\x40\x02\x26\x01\x28\xdd\x04\x2d\xf0\x52\x01\x29\x01\x1e\xdd\x04\x23\xf2\x1e\x19\x05\x1e\xf0\x36\x67\x02\x01\x14\xdf\x04\x3e\xf0\x0c\x00\x2c\x33\xf3\x00\x28\x05\x14\xf0\x54\x01\x00\x39\x40\x01\xf3\x00\x30\x05\x14\xf0\x64\x01\x00\x22\x27\xf4\x00\x44\x01\x05\x14\xf0\x50\x02\x00\x1e\x22\xd0\x00\x1a\xf3\x04\x2b\x01\x0e\xf2\x70\x01\x2e\x01\x0e\xf0\x62\x01\x0a\x01\x09\xf5\x08\x00\x05\x28\xdd\x04\x29\xdd\x04\x29\xdd\x04\x29\xf0\x0c\x0d\x01\x19\xdd\x04\x23\xf3\x0e\x05\x05\x19\xf8\xf0\x75\x18\x00\x08\x13\xf2\x00\x09\x01\x2c\xf4\x02\x08\x05\x2c\xf0\x03\x09\x01\x2c\xfb\xf0\x74\x07\x00\x08\x13\xf2\x00\x01\x01\x1e\xd8\x19\x1d\xd2\x04\x16\xf0\x03\x01\x01\x1e\xfb\xf0\x46\x05\x00\x08\x13\xf2\x00\x23\x01\x28\xf4\x02\x22\x05\x28\xf0\x03\x23\x01\x28\xfb\xf0\x54\x01\x00\x08\x13\xf2\x00\x0a\x01\x1e\xf4\x02\x09\x05\x1e\xf0\x03\x0a\x01\x1e\xfb\xf0\x54\x01\x00\x08\x13\xf2\x00\x03\x01\x1d\xe0\x21\x26\xf6\x00\x01\x05\x1d\xf0\x05\x03\x01\x1d\xfb\xf0\x78\x08\x00\x08\x13\xf2\x00\x02\x01\x09\xe1\x04\x08\xf0\x05\x02\x01\x09\xfb\xf0\x0e\x00\x08\x13\xf2\x00\x04\x01\x15\xf4\x02\x03\x05\x15\xf0\x03\x04\x01\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[126];
    }
ntpath_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 125,
    },
    .ob_shash = -1,
    .ob_sval = "\xb0\x0d\x43\x3a\x00\xc2\x0e\x06\x44\x08\x00\xc2\x24\x06\x44\x15\x00\xc2\x2b\x06\x44\x23\x00\xc2\x35\x08\x44\x31\x00\xc3\x17\x18\x45\x02\x00\xc3\x30\x06\x45\x0d\x00\xc3\x3a\x08\x44\x05\x03\xc4\x04\x01\x44\x05\x03\xc4\x08\x07\x44\x12\x03\xc4\x11\x01\x44\x12\x03\xc4\x15\x08\x44\x20\x03\xc4\x1f\x01\x44\x20\x03\xc4\x23\x08\x44\x2e\x03\xc4\x2d\x01\x44\x2e\x03\xc4\x31\x0b\x44\x3f\x03\xc4\x3e\x01\x44\x3f\x03\xc5\x02\x05\x45\x0a\x03\xc5\x09\x01\x45\x0a\x03\xc5\x0d\x08\x45\x18\x03\xc5\x17\x01\x45\x18\x03",
};
static
    struct _PyCode_DEF(694)
ntpath_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 347,
    },
    .co_consts = & ntpath_toplevel_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & ntpath_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 580,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & ntpath_toplevel_consts_12_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & ntpath_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x5a\x01\x64\x02\x5a\x02\x64\x01\x5a\x03\x64\x03\x5a\x04\x64\x04\x5a\x05\x64\x05\x5a\x06\x64\x06\x5a\x07\x64\x07\x5a\x08\x64\x08\x64\x09\x6c\x09\x5a\x09\x64\x08\x64\x09\x6c\x0a\x5a\x0a\x64\x08\x64\x09\x6c\x0b\x5a\x0b\x64\x08\x64\x09\x6c\x0c\x5a\x0c\x64\x08\x64\x0a\x6c\x0c\xad\x02\x01\x00\x67\x00\x64\x0b\xa2\x01\x5a\x0d\x64\x0c\x84\x00\x5a\x0e\x09\x00\x64\x08\x64\x0d\x6c\x0f\x6d\x10\x5a\x11\x6d\x12\x5a\x13\x6d\x14\x5a\x15\x01\x00\x64\x0e\x84\x00\x5a\x16\x64\x10\x84\x00\x5a\x18\x64\x11\x84\x00\x5a\x19\x64\x12\x84\x00\x5a\x1a\x64\x13\x84\x00\x5a\x1b\x64\x14\x84\x00\x5a\x1c\x64\x15\x84\x00\x5a\x1d\x65\x0c\x6a\x3c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x1d\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x16\x84\x00\x5a\x1f\x64\x17\x84\x00\x5a\x20\x02\x00\x65\x21\x65\x09\x6a\x44\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x18\xab\x02\x00\x00\x00\x00\x00\x00\x72\x04\x64\x19\x84\x00\x5a\x23\x6e\x03\x64\x1a\x84\x00\x5a\x23\x64\x1b\x84\x00\x5a\x24\x09\x00\x64\x08\x64\x1c\x6c\x25\x6d\x26\x5a\x26\x01\x00\x64\x1d\x84\x00\x5a\x27\x64\x1e\x84\x00\x5a\x28\x64\x1f\x5a\x29\x64\x09\x61\x2a\x64\x09\x61\x2b\x64\x20\x84\x00\x5a\x2c\x09\x00\x64\x08\x64\x21\x6c\x25\x6d\x2d\x5a\x2e\x01\x00\x09\x00\x64\x08\x64\x23\x6c\x25\x6d\x2f\x5a\x2f\x01\x00\x64\x24\x84\x00\x5a\x30\x09\x00\x64\x08\x64\x26\x6c\x25\x6d\x31\x5a\x31\x6d\x32\x5a\x33\x01\x00\x65\x34\x66\x01\x64\x27\x84\x01\x5a\x35\x65\x34\x66\x01\x64\x28\x84\x01\x5a\x36\x64\x29\x64\x2a\x9c\x01\x64\x2b\x84\x02\x5a\x37\x64\x2d\x5a\x38\x64\x37\x64\x2e\x84\x01\x5a\x39\x64\x2f\x84\x00\x5a\x3a\x09\x00\x64\x08\x64\x30\x6c\x25\x6d\x3b\x5a\x3c\x01\x00\x64\x08\x64\x31\x6c\x25\x6d\x3d\x5a\x3e\x01\x00\x64\x08\x64\x32\x6c\x25\x6d\x3f\x5a\x40\x01\x00\x64\x08\x64\x33\x6c\x25\x6d\x41\x5a\x42\x01\x00\x09\x00\x64\x08\x64\x34\x6c\x25\x6d\x43\x5a\x43\x01\x00\x64\x35\x84\x00\x5a\x44\x79\x09\x23\x00\x65\x17\x24\x00\x72\x06\x01\x00\x64\x0f\x84\x00\x5a\x16\x59\x00\x8c\xc7\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x17\x24\x00\x72\x05\x01\x00\x64\x09\x5a\x26\x59\x00\x8c\x7d\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x17\x24\x00\x72\x06\x01\x00\x64\x22\x84\x00\x5a\x2e\x59\x00\x8c\x75\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x17\x24\x00\x72\x06\x01\x00\x64\x25\x84\x00\x5a\x30\x59\x00\x8c\x79\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x17\x24\x00\x72\x09\x01\x00\x64\x29\x64\x2a\x9c\x01\x64\x2c\x84\x02\x5a\x37\x59\x00\x8c\x71\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x17\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x5a\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x17\x24\x00\x72\x06\x01\x00\x64\x36\x84\x00\x5a\x44\x59\x00\x79\x09\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_ntpath_toplevel(void)
{
    return Py_NewRef((PyObject *) &ntpath_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[474];
    }
posixpath_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 473,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6d\x6d\x6f\x6e\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x20\x6f\x6e\x20\x50\x6f\x73\x69\x78\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x73\x2e\x0a\x0a\x49\x6e\x73\x74\x65\x61\x64\x20\x6f\x66\x20\x69\x6d\x70\x6f\x72\x74\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x64\x69\x72\x65\x63\x74\x6c\x79\x2c\x20\x69\x6d\x70\x6f\x72\x74\x20\x6f\x73\x20\x61\x6e\x64\x20\x72\x65\x66\x65\x72\x20\x74\x6f\x0a\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x61\x73\x20\x6f\x73\x2e\x70\x61\x74\x68\x2e\x20\x20\x54\x68\x65\x20\x22\x6f\x73\x2e\x70\x61\x74\x68\x22\x20\x6e\x61\x6d\x65\x20\x69\x73\x20\x61\x6e\x20\x61\x6c\x69\x61\x73\x20\x66\x6f\x72\x20\x74\x68\x69\x73\x0a\x6d\x6f\x64\x75\x6c\x65\x20\x6f\x6e\x20\x50\x6f\x73\x69\x78\x20\x73\x79\x73\x74\x65\x6d\x73\x3b\x20\x6f\x6e\x20\x6f\x74\x68\x65\x72\x20\x73\x79\x73\x74\x65\x6d\x73\x20\x28\x65\x2e\x67\x2e\x20\x57\x69\x6e\x64\x6f\x77\x73\x29\x2c\x0a\x6f\x73\x2e\x70\x61\x74\x68\x20\x70\x72\x6f\x76\x69\x64\x65\x73\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x20\x69\x6e\x20\x61\x20\x6d\x61\x6e\x6e\x65\x72\x20\x73\x70\x65\x63\x69\x66\x69\x63\x20\x74\x6f\x20\x74\x68\x61\x74\x0a\x70\x6c\x61\x74\x66\x6f\x72\x6d\x2c\x20\x61\x6e\x64\x20\x69\x73\x20\x61\x6e\x20\x61\x6c\x69\x61\x73\x20\x74\x6f\x20\x61\x6e\x6f\x74\x68\x65\x72\x20\x6d\x6f\x64\x75\x6c\x65\x20\x28\x65\x2e\x67\x2e\x20\x6e\x74\x70\x61\x74\x68\x29\x2e\x0a\x0a\x53\x6f\x6d\x65\x20\x6f\x66\x20\x74\x68\x69\x73\x20\x63\x61\x6e\x20\x61\x63\x74\x75\x61\x6c\x6c\x79\x20\x62\x65\x20\x75\x73\x65\x66\x75\x6c\x20\x6f\x6e\x20\x6e\x6f\x6e\x2d\x50\x6f\x73\x69\x78\x20\x73\x79\x73\x74\x65\x6d\x73\x20\x74\x6f\x6f\x2c\x20\x65\x2e\x67\x2e\x0a\x66\x6f\x72\x20\x6d\x61\x6e\x69\x70\x75\x6c\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x63\x6f\x6d\x70\x6f\x6e\x65\x6e\x74\x20\x6f\x66\x20\x55\x52\x4c\x73\x2e\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
posixpath_toplevel_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "/bin:/usr/bin",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
posixpath_toplevel_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "/dev/null",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[41];
        }_object;
    }
posixpath_toplevel_consts_10 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 41,
        },
        .ob_item = {
            & const_str_normcase._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            &_Py_ID(join),
            & const_str_splitdrive._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            & const_str_splitext._ascii.ob_base,
            & const_str_basename._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
            & const_str_commonprefix._ascii.ob_base,
            & const_str_getsize._ascii.ob_base,
            & const_str_getmtime._ascii.ob_base,
            & const_str_getatime._ascii.ob_base,
            & const_str_getctime._ascii.ob_base,
            & const_str_islink._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str_lexists._ascii.ob_base,
            & const_str_isdir._ascii.ob_base,
            & const_str_isfile._ascii.ob_base,
            & const_str_ismount._ascii.ob_base,
            & const_str_expanduser._ascii.ob_base,
            & const_str_expandvars._ascii.ob_base,
            & const_str_normpath._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str_samefile._ascii.ob_base,
            & const_str_sameopenfile._ascii.ob_base,
            & const_str_samestat._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pathsep._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            & const_str_altsep._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
            & const_str_devnull._ascii.ob_base,
            & const_str_realpath._ascii.ob_base,
            & const_str_supports_unicode_filenames._ascii.ob_base,
            & const_str_relpath._ascii.ob_base,
            & const_str_commonpath._ascii.ob_base,
            & const_str_isjunction._ascii.ob_base,
            & const_str_ALLOW_MISSING._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
posixpath_toplevel_consts_11_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen posixpath>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__get_sep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_sep",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
posixpath_toplevel_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x0f\x13\xe0\x0f\x12",
};
static
    struct _PyCode_DEF(38)
posixpath_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & posixpath_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 41,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 581,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str__get_sep._ascii.ob_base,
    .co_qualname = & const_str__get_sep._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x01\x79\x01\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[55];
    }
posixpath_toplevel_consts_12_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 54,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Normalize case of pathname.  Has no effect under Posix",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
posixpath_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & posixpath_toplevel_consts_12_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
posixpath_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0d\x8f\x39\x89\x39\x90\x51\x8b\x3c\xd0\x04\x17",
};
static
    struct _PyCode_DEF(44)
posixpath_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & posixpath_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 52,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 582,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_normcase._ascii.ob_base,
    .co_qualname = & const_str_normcase._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
posixpath_toplevel_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & ntpath_toplevel_consts_16_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__get_sep._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
posixpath_toplevel_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x0a\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x0a\x12\x90\x31\x8b\x2b\x80\x43\xd8\x0b\x0c\x8f\x3c\x89\x3c\x98\x03\xd3\x0b\x1c\xd0\x04\x1c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
posixpath_toplevel_consts_13_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            &_Py_ID(sep),
        },
    },
};
static
    struct _PyCode_DEF(100)
posixpath_toplevel_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 50,
    },
    .co_consts = & posixpath_toplevel_consts_13_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 60,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 583,
    .co_localsplusnames = & posixpath_toplevel_consts_13_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_isabs._ascii.ob_base,
    .co_qualname = & const_str_isabs._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[231];
    }
posixpath_toplevel_consts_14_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 230,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4a\x6f\x69\x6e\x20\x74\x77\x6f\x20\x6f\x72\x20\x6d\x6f\x72\x65\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x63\x6f\x6d\x70\x6f\x6e\x65\x6e\x74\x73\x2c\x20\x69\x6e\x73\x65\x72\x74\x69\x6e\x67\x20\x27\x2f\x27\x20\x61\x73\x20\x6e\x65\x65\x64\x65\x64\x2e\x0a\x20\x20\x20\x20\x49\x66\x20\x61\x6e\x79\x20\x63\x6f\x6d\x70\x6f\x6e\x65\x6e\x74\x20\x69\x73\x20\x61\x6e\x20\x61\x62\x73\x6f\x6c\x75\x74\x65\x20\x70\x61\x74\x68\x2c\x20\x61\x6c\x6c\x20\x70\x72\x65\x76\x69\x6f\x75\x73\x20\x70\x61\x74\x68\x20\x63\x6f\x6d\x70\x6f\x6e\x65\x6e\x74\x73\x0a\x20\x20\x20\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x64\x69\x73\x63\x61\x72\x64\x65\x64\x2e\x20\x20\x41\x6e\x20\x65\x6d\x70\x74\x79\x20\x6c\x61\x73\x74\x20\x70\x61\x72\x74\x20\x77\x69\x6c\x6c\x20\x72\x65\x73\x75\x6c\x74\x20\x69\x6e\x20\x61\x20\x70\x61\x74\x68\x20\x74\x68\x61\x74\x0a\x20\x20\x20\x20\x65\x6e\x64\x73\x20\x77\x69\x74\x68\x20\x61\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & posixpath_toplevel_consts_14_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
posixpath_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__get_sep._ascii.ob_base,
            & const_str_map._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_BytesWarning._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            & const_str__check_arg_types._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[182];
    }
posixpath_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 181,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x09\x0b\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x0a\x12\x90\x31\x8b\x2b\x80\x43\xd8\x0b\x0c\x80\x44\xf0\x02\x0c\x05\x0e\xd9\x0f\x10\xd8\x0c\x10\x90\x12\x90\x21\x88\x48\x90\x73\x8a\x4e\xdc\x11\x14\x94\x52\x97\x59\x91\x59\xa0\x01\xd6\x11\x22\x88\x41\xd8\x0f\x10\x8f\x7c\x89\x7c\x98\x43\xd4\x0f\x20\xd8\x17\x18\x91\x04\xd9\x15\x19\x98\x54\x9f\x5d\x99\x5d\xa8\x33\xd4\x1d\x2f\xd8\x10\x14\x98\x01\x91\x09\x91\x04\xe0\x10\x14\x98\x03\x98\x61\x99\x07\x91\x0f\x91\x04\xf1\x0d\x00\x12\x23\xf0\x14\x00\x0c\x10\x80\x4b\xf8\xf4\x07\x00\x0d\x16\x94\x7e\xa4\x7c\xd0\x0b\x34\xf2\x00\x02\x05\x0e\xdc\x08\x13\xd7\x08\x24\xd1\x08\x24\xa0\x56\xa8\x51\xd0\x08\x33\xb0\x11\xd3\x08\x33\xd8\x08\x0d\xf0\x05\x02\x05\x0e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
posixpath_toplevel_consts_14_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\xa4\x41\x1e\x42\x05\x00\xc2\x05\x2d\x42\x32\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
posixpath_toplevel_consts_14_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[97],
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            &_Py_ID(path),
            (PyObject *)&_Py_SINGLETON(strings).ascii[98],
        },
    },
};
static
    struct _PyCode_DEF(362)
posixpath_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 181,
    },
    .co_consts = & posixpath_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_consts_14_exceptiontable.ob_base.ob_base,
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 71,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 584,
    .co_localsplusnames = & posixpath_toplevel_consts_14_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = &_Py_ID(join),
    .co_qualname = &_Py_ID(join),
    .co_linetable = & posixpath_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x7d\x03\x09\x00\x7c\x01\x73\x08\x7c\x03\x64\x01\x64\x02\x1a\x00\x7c\x02\x7a\x00\x00\x00\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x37\x00\x00\x7d\x04\x7c\x04\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x72\x03\x7c\x04\x7d\x03\x8c\x17\x7c\x03\x72\x11\x7c\x03\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x72\x06\x7c\x03\x7c\x04\x7a\x0d\x00\x00\x7d\x03\x8c\x30\x7c\x03\x7c\x02\x7c\x04\x7a\x00\x00\x00\x7a\x0d\x00\x00\x7d\x03\x8c\x39\x04\x00\x09\x00\x7c\x03\x53\x00\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x19\x01\x00\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x00\x67\x02\x7c\x01\xa2\x01\xad\x06\x8e\x00\x01\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[129];
    }
posixpath_toplevel_consts_15_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 128,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x70\x6c\x69\x74\x20\x61\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x2e\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x75\x70\x6c\x65\x20\x22\x28\x68\x65\x61\x64\x2c\x20\x74\x61\x69\x6c\x29\x22\x20\x77\x68\x65\x72\x65\x20\x22\x74\x61\x69\x6c\x22\x20\x69\x73\x0a\x20\x20\x20\x20\x65\x76\x65\x72\x79\x74\x68\x69\x6e\x67\x20\x61\x66\x74\x65\x72\x20\x74\x68\x65\x20\x66\x69\x6e\x61\x6c\x20\x73\x6c\x61\x73\x68\x2e\x20\x20\x45\x69\x74\x68\x65\x72\x20\x70\x61\x72\x74\x20\x6d\x61\x79\x20\x62\x65\x20\x65\x6d\x70\x74\x79\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & posixpath_toplevel_consts_15_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
posixpath_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__get_sep._ascii.ob_base,
            & const_str_rfind._ascii.ob_base,
            &_Py_ID(len),
            & const_str_rstrip._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[108];
    }
posixpath_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 107,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x09\x0b\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x0a\x12\x90\x31\x8b\x2b\x80\x43\xd8\x08\x09\x8f\x07\x89\x07\x90\x03\x8b\x0c\x90\x71\xd1\x08\x18\x80\x41\xd8\x11\x12\x90\x32\x90\x41\x90\x15\x98\x01\x98\x21\x98\x22\x98\x05\x88\x24\x80\x44\xd9\x07\x0b\x90\x04\x98\x03\x9c\x43\xa0\x04\x9b\x49\x99\x0d\xd2\x10\x25\xd8\x0f\x13\x8f\x7b\x89\x7b\x98\x33\xd3\x0f\x1f\x88\x04\xd8\x0b\x0f\x90\x14\x88\x3a\xd0\x04\x15",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
posixpath_toplevel_consts_15_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            & const_str_head._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(206)
posixpath_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 103,
    },
    .co_consts = & posixpath_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 100,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 585,
    .co_localsplusnames = & posixpath_toplevel_consts_15_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_split._ascii.ob_base,
    .co_qualname = & const_str_split._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x7a\x00\x00\x00\x7d\x02\x7c\x00\x64\x02\x7c\x02\x1a\x00\x7c\x00\x7c\x02\x64\x02\x1a\x00\x7d\x04\x7d\x03\x7c\x03\x72\x22\x7c\x03\x7c\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x05\x00\x00\x6b\x37\x00\x00\x72\x11\x7c\x03\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x7c\x04\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
posixpath_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[69];
    }
posixpath_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 68,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x08\x0a\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x07\x11\x90\x21\x94\x55\xd4\x07\x1b\xd8\x0e\x12\x88\x03\xd8\x11\x15\x89\x06\xe0\x0e\x11\x88\x03\xd8\x11\x14\x88\x06\xdc\x0b\x16\xd7\x0b\x20\xd1\x0b\x20\xa0\x11\xa0\x43\xa8\x14\xa8\x76\xd3\x0b\x36\xd0\x04\x36",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            & const_str_extsep._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(142)
posixpath_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 71,
    },
    .co_consts = & posixpath_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 117,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 586,
    .co_localsplusnames = & posixpath_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_splitext._ascii.ob_base,
    .co_qualname = & const_str_splitext._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x05\x64\x01\x7d\x01\x64\x02\x7d\x02\x6e\x04\x64\x03\x7d\x01\x64\x04\x7d\x02\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x64\x00\x7c\x02\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[75];
    }
posixpath_toplevel_consts_17_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 74,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x70\x6c\x69\x74\x20\x61\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x69\x6e\x74\x6f\x20\x64\x72\x69\x76\x65\x20\x61\x6e\x64\x20\x70\x61\x74\x68\x2e\x20\x4f\x6e\x20\x50\x6f\x73\x69\x78\x2c\x20\x64\x72\x69\x76\x65\x20\x69\x73\x20\x61\x6c\x77\x61\x79\x73\x0a\x20\x20\x20\x20\x65\x6d\x70\x74\x79\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & posixpath_toplevel_consts_17_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
posixpath_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x09\x0b\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xd8\x0b\x0c\x88\x52\x88\x61\x88\x35\x90\x21\x88\x38\x80\x4f",
};
static
    struct _PyCode_DEF(58)
posixpath_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 29,
    },
    .co_consts = & posixpath_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 131,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 587,
    .co_localsplusnames = & ntpath_toplevel_consts_21_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_splitdrive._ascii.ob_base,
    .co_qualname = & const_str_splitdrive._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x64\x01\x64\x02\x1a\x00\x7c\x00\x66\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[422];
    }
posixpath_toplevel_consts_18_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 421,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x53\x70\x6c\x69\x74\x20\x61\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x69\x6e\x74\x6f\x20\x64\x72\x69\x76\x65\x2c\x20\x72\x6f\x6f\x74\x20\x61\x6e\x64\x20\x74\x61\x69\x6c\x2e\x20\x4f\x6e\x20\x50\x6f\x73\x69\x78\x2c\x20\x64\x72\x69\x76\x65\x20\x69\x73\x20\x61\x6c\x77\x61\x79\x73\x0a\x20\x20\x20\x20\x65\x6d\x70\x74\x79\x3b\x20\x74\x68\x65\x20\x72\x6f\x6f\x74\x20\x6d\x61\x79\x20\x62\x65\x20\x65\x6d\x70\x74\x79\x2c\x20\x61\x20\x73\x69\x6e\x67\x6c\x65\x20\x73\x6c\x61\x73\x68\x2c\x20\x6f\x72\x20\x74\x77\x6f\x20\x73\x6c\x61\x73\x68\x65\x73\x2e\x20\x54\x68\x65\x20\x74\x61\x69\x6c\x0a\x20\x20\x20\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x20\x61\x6e\x79\x74\x68\x69\x6e\x67\x20\x61\x66\x74\x65\x72\x20\x74\x68\x65\x20\x72\x6f\x6f\x74\x2e\x20\x46\x6f\x72\x20\x65\x78\x61\x6d\x70\x6c\x65\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x20\x3d\x3d\x20\x28\x27\x27\x2c\x20\x27\x27\x2c\x20\x27\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x2f\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x20\x3d\x3d\x20\x28\x27\x27\x2c\x20\x27\x2f\x27\x2c\x20\x27\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x2f\x2f\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x20\x3d\x3d\x20\x28\x27\x27\x2c\x20\x27\x2f\x2f\x27\x2c\x20\x27\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x73\x70\x6c\x69\x74\x72\x6f\x6f\x74\x28\x27\x2f\x2f\x2f\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x20\x3d\x3d\x20\x28\x27\x27\x2c\x20\x27\x2f\x27\x2c\x20\x27\x2f\x2f\x66\x6f\x6f\x2f\x62\x61\x72\x27\x29\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
posixpath_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & posixpath_toplevel_consts_18_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            &_Py_STR(empty),
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[144];
    }
posixpath_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 143,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x14\x00\x09\x0b\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x07\x11\x90\x21\x94\x55\xd4\x07\x1b\xd8\x0e\x12\x88\x03\xd8\x10\x13\x89\x05\xe0\x0e\x11\x88\x03\xd8\x10\x12\x88\x05\xd8\x07\x08\x88\x12\x88\x21\x80\x75\x90\x03\x82\x7c\xe0\x0f\x14\x90\x65\x98\x51\x88\x7f\xd0\x08\x1e\xd8\x09\x0a\x88\x31\x88\x51\x88\x16\x90\x33\x8a\x1d\x98\x21\x98\x41\x98\x61\x98\x26\xa0\x43\x9a\x2d\xe0\x0f\x14\x90\x63\x98\x31\x98\x51\x98\x52\x98\x35\xd0\x0f\x20\xd0\x08\x20\xf0\x08\x00\x10\x15\x90\x61\x98\x02\x98\x11\x90\x65\x98\x51\x98\x71\x98\x72\x98\x55\xd0\x0f\x22\xd0\x08\x22",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_18_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            & const_str_empty._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(190)
posixpath_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 95,
    },
    .co_consts = & posixpath_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 138,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 588,
    .co_localsplusnames = & posixpath_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_splitroot._ascii.ob_base,
    .co_qualname = & const_str_splitroot._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x05\x64\x01\x7d\x01\x64\x02\x7d\x02\x6e\x04\x64\x03\x7d\x01\x64\x04\x7d\x02\x7c\x00\x64\x05\x64\x06\x1a\x00\x7c\x01\x6b\x37\x00\x00\x72\x05\x7c\x02\x7c\x02\x7c\x00\x66\x03\x53\x00\x7c\x00\x64\x06\x64\x07\x1a\x00\x7c\x01\x6b\x37\x00\x00\x73\x08\x7c\x00\x64\x07\x64\x08\x1a\x00\x7c\x01\x6b\x28\x00\x00\x72\x08\x7c\x02\x7c\x01\x7c\x00\x64\x06\x64\x05\x1a\x00\x66\x03\x53\x00\x7c\x02\x7c\x00\x64\x05\x64\x07\x1a\x00\x7c\x00\x64\x07\x64\x05\x1a\x00\x66\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_19_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & ntpath_toplevel_consts_22_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_19_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__get_sep._ascii.ob_base,
            & const_str_rfind._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[54];
    }
posixpath_toplevel_consts_19_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 53,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x0a\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x0a\x12\x90\x31\x8b\x2b\x80\x43\xd8\x08\x09\x8f\x07\x89\x07\x90\x03\x8b\x0c\x90\x71\xd1\x08\x18\x80\x41\xd8\x0b\x0c\x88\x51\x88\x52\x88\x35\x80\x4c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_19_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
        },
    },
};
static
    struct _PyCode_DEF(116)
posixpath_toplevel_consts_19 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 58,
    },
    .co_consts = & posixpath_toplevel_consts_19_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_19_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 169,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 589,
    .co_localsplusnames = & posixpath_toplevel_consts_19_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_basename._ascii.ob_base,
    .co_qualname = & const_str_basename._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_19_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x7a\x00\x00\x00\x7d\x02\x7c\x00\x7c\x02\x64\x02\x1a\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_20_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & ntpath_toplevel_consts_23_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[91];
    }
posixpath_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 90,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x0a\x8f\x09\x89\x09\x90\x21\x8b\x0c\x80\x41\xdc\x0a\x12\x90\x31\x8b\x2b\x80\x43\xd8\x08\x09\x8f\x07\x89\x07\x90\x03\x8b\x0c\x90\x71\xd1\x08\x18\x80\x41\xd8\x0b\x0c\x88\x52\x88\x61\x88\x35\x80\x44\xd9\x07\x0b\x90\x04\x98\x03\x9c\x43\xa0\x04\x9b\x49\x99\x0d\xd2\x10\x25\xd8\x0f\x13\x8f\x7b\x89\x7b\x98\x33\xd3\x0f\x1f\x88\x04\xd8\x0b\x0f\x80\x4b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_20_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            & const_str_head._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(192)
posixpath_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 96,
    },
    .co_consts = & posixpath_toplevel_consts_20_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 179,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 590,
    .co_localsplusnames = & posixpath_toplevel_consts_20_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_dirname._ascii.ob_base,
    .co_qualname = & const_str_dirname._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x7a\x00\x00\x00\x7d\x02\x7c\x00\x64\x02\x7c\x02\x1a\x00\x7d\x03\x7c\x03\x72\x22\x7c\x03\x7c\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x05\x00\x00\x6b\x37\x00\x00\x72\x11\x7c\x03\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[82];
    }
posixpath_toplevel_consts_21_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 81,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x54\x65\x73\x74\x20\x77\x68\x65\x74\x68\x65\x72\x20\x61\x20\x70\x61\x74\x68\x20\x69\x73\x20\x61\x20\x6a\x75\x6e\x63\x74\x69\x6f\x6e\x0a\x20\x20\x20\x20\x4a\x75\x6e\x63\x74\x69\x6f\x6e\x73\x20\x61\x72\x65\x20\x6e\x6f\x74\x20\x61\x20\x70\x61\x72\x74\x20\x6f\x66\x20\x70\x6f\x73\x69\x78\x20\x73\x65\x6d\x61\x6e\x74\x69\x63\x73",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
posixpath_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & posixpath_toplevel_consts_21_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
posixpath_toplevel_consts_21_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x05\x07\x87\x49\x81\x49\x88\x64\x84\x4f\xd8\x0b\x10",
};
static
    struct _PyCode_DEF(46)
posixpath_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & posixpath_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 192,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 591,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_isjunction._ascii.ob_base,
    .co_qualname = & const_str_isjunction._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_21_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[49];
    }
posixpath_toplevel_consts_22_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 48,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x03\x05\x15\xdc\x08\x0a\x8f\x08\x89\x08\x90\x14\x8c\x0e\xf0\x06\x00\x0c\x10\xf8\xf4\x05\x00\x0d\x14\x94\x5a\xd0\x0b\x20\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct _PyCode_DEF(90)
posixpath_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 45,
    },
    .co_consts = & ntpath_toplevel_consts_27_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_27_names._object.ob_base.ob_base,
    .co_exceptiontable = & genericpath_toplevel_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 201,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 592,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_lexists._ascii.ob_base,
    .co_qualname = & const_str_lexists._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[37];
    }
posixpath_toplevel_consts_23_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 36,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Test whether a path is a mount point",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
posixpath_toplevel_consts_23_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & posixpath_toplevel_consts_23_consts_0._ascii.ob_base,
            Py_False,
            & ntpath_toplevel_consts_34_consts_4.ob_base.ob_base,
            & ntpath_toplevel_consts_2._ascii.ob_base,
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
posixpath_toplevel_consts_23_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_lstat._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_S_ISLNK._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            &_Py_ID(join),
            & const_str_realpath._ascii.ob_base,
            & const_str_st_dev._ascii.ob_base,
            & const_str_st_ino._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[228];
    }
posixpath_toplevel_consts_23_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 227,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x08\x05\x19\xdc\x0d\x0f\x8f\x58\x89\x58\x90\x64\x8b\x5e\x88\x02\xf4\x0c\x00\x0c\x10\x8f\x3c\x89\x3c\x98\x02\x9f\x0a\x99\x0a\xd4\x0b\x23\xd8\x13\x18\xe4\x0b\x0d\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xdc\x11\x15\x90\x64\x98\x45\xd3\x11\x22\x89\x06\xe4\x11\x15\x90\x64\x98\x44\xd3\x11\x21\x88\x06\xdc\x0d\x15\x90\x66\xd3\x0d\x1d\x80\x46\xf0\x02\x03\x05\x15\xdc\x0d\x0f\x8f\x58\x89\x58\x90\x66\xd3\x0d\x1d\x88\x02\xf0\x08\x00\x0c\x0e\x8f\x39\x89\x39\x80\x44\xd8\x0b\x0d\x8f\x39\x89\x39\x80\x44\xd8\x07\x0b\x88\x74\x82\x7c\xd8\x0f\x13\xd8\x0b\x0d\x8f\x39\x89\x39\x80\x44\xd8\x0b\x0d\x8f\x39\x89\x39\x80\x44\xd8\x07\x0b\x88\x74\x82\x7c\xd8\x0f\x13\xd8\x0b\x10\xf8\xf4\x37\x00\x0d\x14\x94\x5a\xd0\x0b\x20\xf2\x00\x02\x05\x15\xe1\x0f\x14\xf0\x05\x02\x05\x15\xfb\xf4\x20\x00\x0d\x14\x94\x5a\xd0\x0b\x20\xf2\x00\x01\x05\x15\xd9\x0f\x14\xf0\x03\x01\x05\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
posixpath_toplevel_consts_23_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x15\x43\x13\x00\xc2\x01\x15\x43\x28\x00\xc3\x13\x0f\x43\x25\x03\xc3\x24\x01\x43\x25\x03\xc3\x28\x0f\x43\x3a\x03\xc3\x39\x01\x43\x3a\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_dev1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dev1",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_dev2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dev2",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_ino1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ino1",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_ino2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ino2",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
posixpath_toplevel_consts_23_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_s1._ascii.ob_base,
            &_Py_ID(parent),
            & const_str_s2._ascii.ob_base,
            & const_str_dev1._ascii.ob_base,
            & const_str_dev2._ascii.ob_base,
            & const_str_ino1._ascii.ob_base,
            & const_str_ino2._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(506)
posixpath_toplevel_consts_23 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 253,
    },
    .co_consts = & posixpath_toplevel_consts_23_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_23_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_consts_23_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 213,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 593,
    .co_localsplusnames = & posixpath_toplevel_consts_23_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_ismount._ascii.ob_base,
    .co_qualname = & const_str_ismount._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_23_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x01\x79\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x0d\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x6e\x0c\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x01\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x03\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x04\x7c\x05\x6b\x37\x00\x00\x72\x01\x79\x04\x7c\x01\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x03\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x06\x7c\x07\x6b\x28\x00\x00\x72\x01\x79\x04\x79\x01\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[80];
    }
posixpath_toplevel_consts_24_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 79,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x78\x70\x61\x6e\x64\x20\x7e\x20\x61\x6e\x64\x20\x7e\x75\x73\x65\x72\x20\x63\x6f\x6e\x73\x74\x72\x75\x63\x74\x69\x6f\x6e\x73\x2e\x20\x20\x49\x66\x20\x75\x73\x65\x72\x20\x6f\x72\x20\x24\x48\x4f\x4d\x45\x20\x69\x73\x20\x75\x6e\x6b\x6e\x6f\x77\x6e\x2c\x0a\x20\x20\x20\x20\x64\x6f\x20\x6e\x6f\x74\x68\x69\x6e\x67\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_HOME = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HOME",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_vxworks = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "vxworks",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
posixpath_toplevel_consts_24_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & posixpath_toplevel_consts_24_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[126]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[126],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & const_str_HOME._ascii.ob_base,
            Py_None,
            & const_str_vxworks._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_pwd = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pwd",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_getpwuid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getpwuid",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_getuid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getuid",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_pw_dir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pw_dir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_getpwnam = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getpwnam",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[21];
        }_object;
    }
posixpath_toplevel_consts_24_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 21,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_startswith._ascii.ob_base,
            & const_str__get_sep._ascii.ob_base,
            & const_str_find._ascii.ob_base,
            &_Py_ID(len),
            & const_str_environ._ascii.ob_base,
            & const_str_pwd._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_getpwuid._ascii.ob_base,
            & const_str_getuid._ascii.ob_base,
            & const_str_pw_dir._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            & const_str_getpwnam._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str_fsencode._ascii.ob_base,
            & const_str_rstrip._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[427];
    }
posixpath_toplevel_consts_24_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 426,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x0c\x0e\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x10\x14\x89\x05\xe0\x10\x13\x88\x05\xd8\x0b\x0f\x8f\x3f\x89\x3f\x98\x35\xd4\x0b\x21\xd8\x0f\x13\x88\x0b\xdc\x0a\x12\x90\x34\x8b\x2e\x80\x43\xd8\x08\x0c\x8f\x09\x89\x09\x90\x23\x90\x71\xd3\x08\x19\x80\x41\xd8\x07\x08\x88\x31\x82\x75\xdc\x0c\x0f\x90\x04\x8b\x49\x88\x01\xd8\x07\x08\x88\x41\x82\x76\xd8\x0b\x11\x9c\x12\x9f\x1a\x99\x1a\xd1\x0b\x23\xf0\x02\x04\x0d\x1c\xdb\x10\x1a\xf0\x08\x05\x0d\x1c\xd8\x1b\x1e\x9f\x3c\x99\x3c\xac\x02\xaf\x09\xa9\x09\xab\x0b\xd3\x1b\x34\xd7\x1b\x3b\xd1\x1b\x3b\x91\x08\xf4\x0c\x00\x18\x1a\x97\x7a\x91\x7a\xa0\x26\xd1\x17\x29\x89\x48\xf0\x04\x04\x09\x18\xdb\x0c\x16\xf0\x08\x00\x10\x14\x90\x41\x90\x61\x88\x79\x88\x04\xdc\x0b\x15\x90\x64\x9c\x45\xd4\x0b\x22\xdc\x13\x15\x97\x3b\x91\x3b\x98\x74\xd3\x13\x24\x88\x44\xf0\x02\x05\x09\x18\xd8\x14\x17\x97\x4c\x91\x4c\xa0\x14\xd3\x14\x26\x88\x45\xf0\x0a\x00\x14\x19\x97\x3c\x91\x3c\x88\x08\xe0\x07\x0f\xd0\x07\x17\x9c\x43\x9f\x4c\x99\x4c\xa8\x49\xd2\x1c\x35\xd8\x0f\x13\x88\x0b\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xdc\x13\x15\x97\x3b\x91\x3b\x98\x78\xd3\x13\x28\x88\x08\xd8\x0f\x13\x89\x04\xe0\x0f\x12\x88\x04\xd8\x0f\x17\x8f\x7f\x89\x7f\x98\x74\xd3\x0f\x24\x80\x48\xd8\x0c\x14\x90\x74\x98\x41\x98\x42\x90\x78\xd1\x0c\x1f\xd2\x0b\x28\xa0\x44\xd0\x04\x28\xf8\xf4\x49\x01\x00\x14\x1f\xf2\x00\x02\x0d\x1c\xe0\x17\x1b\x92\x0b\xf0\x05\x02\x0d\x1c\xfb\xf4\x0a\x00\x14\x1c\xf2\x00\x03\x0d\x1c\xf0\x06\x00\x18\x1c\x92\x0b\xf0\x07\x03\x0d\x1c\xfb\xf4\x12\x00\x10\x1b\xf2\x00\x02\x09\x18\xe0\x13\x17\x8a\x4b\xf0\x05\x02\x09\x18\xfb\xf4\x10\x00\x10\x18\xf2\x00\x03\x09\x18\xf0\x06\x00\x14\x18\x8a\x4b\xf0\x07\x03\x09\x18\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[73];
    }
posixpath_toplevel_consts_24_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 72,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x03\x04\x45\x35\x00\xc2\x08\x2d\x46\x06\x00\xc3\x0b\x04\x46\x17\x00\xc3\x3a\x11\x46\x28\x00\xc5\x35\x0b\x46\x03\x03\xc6\x02\x01\x46\x03\x03\xc6\x06\x0b\x46\x14\x03\xc6\x13\x01\x46\x14\x03\xc6\x17\x0b\x46\x25\x03\xc6\x24\x01\x46\x25\x03\xc6\x28\x0b\x46\x36\x03\xc6\x35\x01\x46\x36\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_pwent = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pwent",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
posixpath_toplevel_consts_24_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_tilde._ascii.ob_base,
            &_Py_ID(sep),
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            & const_str_pwd._ascii.ob_base,
            & const_str_userhome._ascii.ob_base,
            &_Py_ID(name),
            & const_str_pwent._ascii.ob_base,
            & const_str_root._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(882)
posixpath_toplevel_consts_24 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 441,
    },
    .co_consts = & posixpath_toplevel_consts_24_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_24_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_consts_24_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 256,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 594,
    .co_localsplusnames = & posixpath_toplevel_consts_24_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_expanduser._ascii.ob_base,
    .co_qualname = & const_str_expanduser._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_24_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x03\x64\x01\x7d\x01\x6e\x02\x64\x02\x7d\x01\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x73\x02\x7c\x00\x53\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x64\x04\x6b\x02\x00\x00\x72\x0b\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x64\x03\x6b\x28\x00\x00\x72\x5a\x64\x05\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x34\x09\x00\x64\x04\x64\x06\x6c\x09\x7d\x04\x09\x00\x7c\x04\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x6e\x61\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x19\x00\x00\x00\x7d\x05\x6e\x4d\x09\x00\x64\x04\x64\x06\x6c\x09\x7d\x04\x7c\x00\x64\x03\x7c\x03\x1a\x00\x7d\x06\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x15\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x09\x00\x7c\x04\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x07\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x80\x15\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x6b\x28\x00\x00\x72\x02\x7c\x00\x53\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x18\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x26\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x64\x08\x7d\x08\x6e\x02\x64\x09\x7d\x08\x7c\x05\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x7c\x00\x7c\x03\x64\x06\x1a\x00\x7a\x00\x00\x00\x78\x01\x73\x02\x01\x00\x7c\x08\x53\x00\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x7c\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x7c\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x7c\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x7c\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
posixpath_toplevel_consts_25 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\$(\\w+|\\{[^}]*\\}?)",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[91];
    }
posixpath_toplevel_consts_26_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 90,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x78\x70\x61\x6e\x64\x20\x73\x68\x65\x6c\x6c\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x73\x20\x6f\x66\x20\x66\x6f\x72\x6d\x20\x24\x76\x61\x72\x20\x61\x6e\x64\x20\x24\x7b\x76\x61\x72\x7d\x2e\x20\x20\x55\x6e\x6b\x6e\x6f\x77\x6e\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x73\x0a\x20\x20\x20\x20\x61\x72\x65\x20\x6c\x65\x66\x74\x20\x75\x6e\x63\x68\x61\x6e\x67\x65\x64\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_26_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
posixpath_toplevel_consts_26_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_startswith._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fsencode._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[150];
    }
posixpath_toplevel_consts_26_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 149,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x0f\x10\x90\x11\x89\x74\x88\x04\xd8\x0b\x0f\x8f\x3f\x89\x3f\x98\x35\xd4\x0b\x21\xd8\x13\x17\x97\x3d\x91\x3d\xa0\x13\xd4\x13\x25\xd8\x17\x18\x98\x11\x91\x74\x90\x0b\xd8\x13\x17\x98\x01\x98\x22\x90\x3a\x88\x44\xf0\x02\x08\x09\x19\xd8\x0f\x16\x88\x7f\xdc\x18\x1a\x9f\x0b\x99\x0b\xa4\x42\xa7\x4a\xa1\x4a\xac\x72\xaf\x7b\xa9\x7b\xb8\x34\xd3\x2f\x40\xd1\x24\x41\xd3\x18\x42\x90\x05\xf0\x0c\x00\x14\x19\x88\x4c\xf0\x09\x00\x19\x20\xa0\x04\x99\x0d\x91\x05\xf0\x08\x00\x14\x19\x88\x4c\xf8\xf4\x07\x00\x10\x18\xf2\x00\x01\x09\x18\xd8\x13\x14\x90\x51\x91\x34\x8a\x4b\xf0\x03\x01\x09\x18\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
posixpath_toplevel_consts_26_consts_10_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\xb4\x3b\x41\x39\x00\xc1\x31\x05\x41\x39\x00\xc1\x39\x0e\x42\x0a\x03\xc2\x09\x01\x42\x0a\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
posixpath_toplevel_consts_26_consts_10_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[109],
            &_Py_ID(name),
            &_Py_ID(value),
            &_Py_ID(end),
            & const_str_environ._ascii.ob_base,
            &_Py_ID(start),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[7];
    }
posixpath_toplevel_consts_26_consts_10_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 6,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x20\x80\x80\x80",
};
static
    struct _PyCode_DEF(282)
posixpath_toplevel_consts_26_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 141,
    },
    .co_consts = & posixpath_toplevel_consts_26_consts_10_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_26_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_consts_26_consts_10_exceptiontable.ob_base.ob_base,
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 347,
    .co_nlocalsplus = 6,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 3,
    .co_version = 595,
    .co_localsplusnames = & posixpath_toplevel_consts_26_consts_10_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & posixpath_toplevel_consts_26_consts_10_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = &_Py_ID(repl),
    .co_qualname = & ntpath_toplevel_consts_32_consts_12_qualname._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_26_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x03\x97\x00\x7c\x00\x64\x01\x19\x00\x00\x00\x7d\x01\x7c\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x05\xab\x01\x00\x00\x00\x00\x00\x00\x72\x1b\x7c\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x03\xab\x01\x00\x00\x00\x00\x00\x00\x73\x05\x7c\x00\x64\x02\x19\x00\x00\x00\x53\x00\x7c\x01\x64\x01\x64\x03\x1a\x00\x7d\x01\x09\x00\x89\x04\x80\x3b\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x53\x00\x89\x04\x7c\x01\x19\x00\x00\x00\x7d\x02\x09\x00\x7c\x02\x53\x00\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x08\x01\x00\x7c\x00\x64\x02\x19\x00\x00\x00\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
posixpath_toplevel_consts_26_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & posixpath_toplevel_consts_26_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[36]),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_characters[123]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[125]),
            & const_str_environb._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[36],
            (PyObject *)&_Py_SINGLETON(strings).ascii[123],
            (PyObject *)&_Py_SINGLETON(strings).ascii[125],
            & posixpath_toplevel_consts_26_consts_10.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[199];
    }
posixpath_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 198,
    },
    .ob_shash = -1,
    .ob_sval = "\xfa\x80\x00\xf4\x06\x00\x0c\x0e\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xe4\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x0b\x0f\x90\x74\xd1\x0b\x1b\xd8\x13\x17\x88\x4b\xdd\x0f\x17\xdb\x0c\x15\xd8\x17\x19\x97\x7a\x91\x7a\xa4\x2b\xd7\x22\x34\xd1\x22\x34\xd3\x22\x36\xb8\x02\xbf\x08\xb9\x08\xd3\x17\x41\xd7\x17\x45\xd1\x17\x45\x88\x48\xdc\x0e\x16\x88\x03\xd8\x10\x14\x88\x05\xd8\x0e\x12\x88\x03\xdc\x12\x19\x9c\x22\x98\x6a\xa8\x24\xd3\x12\x2f\x89\x07\xe0\x0b\x0e\x90\x64\x89\x3f\xd8\x13\x17\x88\x4b\xdd\x0f\x16\xdb\x0c\x15\xd8\x16\x18\x97\x6a\x91\x6a\xa4\x1b\xa8\x62\xaf\x68\xa9\x68\xd3\x16\x37\xd7\x16\x3b\xd1\x16\x3b\x88\x47\xdc\x0e\x15\x88\x03\xd8\x10\x13\x88\x05\xd8\x0e\x11\x88\x03\xdc\x12\x14\x97\x2a\x91\x2a\x88\x07\xf6\x04\x0e\x05\x19\xf1\x20\x00\x0c\x0f\x88\x74\x90\x54\x8b\x3f\xd0\x04\x1a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
posixpath_toplevel_consts_26_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_re._ascii.ob_base,
            & const_str_sub._ascii.ob_base,
            &_Py_ID(repl),
            &_Py_ID(end),
            & const_str_environ._ascii.ob_base,
            &_Py_ID(start),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
posixpath_toplevel_consts_26_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "    @@@",
};
static
    struct _PyCode_DEF(482)
posixpath_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 241,
    },
    .co_consts = & posixpath_toplevel_consts_26_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_32_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 321,
    .co_nlocalsplus = 7,
    .co_nlocals = 4,
    .co_ncellvars = 3,
    .co_nfreevars = 0,
    .co_version = 596,
    .co_localsplusnames = & posixpath_toplevel_consts_26_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & posixpath_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_expandvars._ascii.ob_base,
    .co_qualname = & const_str_expandvars._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x04\x87\x05\x87\x06\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x64\x64\x01\x7c\x00\x76\x01\x72\x02\x7c\x00\x53\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x73\x3c\x64\x02\x64\x03\x6c\x05\x7d\x01\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x61\x04\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x04\x8a\x06\x64\x05\x8a\x04\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x64\x03\xab\x03\x00\x00\x00\x00\x00\x00\x8a\x05\x6e\x54\x64\x07\x7c\x00\x76\x01\x72\x02\x7c\x00\x53\x00\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x73\x2e\x64\x02\x64\x03\x6c\x05\x7d\x01\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x61\x0c\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x08\x8a\x06\x64\x09\x8a\x04\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x8a\x05\x88\x04\x88\x05\x88\x06\x66\x03\x64\x0a\x84\x08\x7d\x03\x02\x00\x7c\x02\x7c\x03\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
posixpath_toplevel_consts_28_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & ntpath_toplevel_consts_34_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_empty),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            & ntpath_toplevel_consts_34_consts_4.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            &_Py_STR(empty),
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & ntpath_toplevel_consts_2._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
posixpath_toplevel_consts_28_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_splitroot._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            &_Py_ID(append),
            & const_str_pop._ascii.ob_base,
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[223];
    }
posixpath_toplevel_consts_28_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 222,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x11\x8f\x79\x89\x79\x98\x14\x8b\x7f\x88\x04\xdc\x0b\x15\x90\x64\x9c\x45\xd4\x0b\x22\xd8\x12\x16\x88\x43\xd8\x14\x17\x88\x45\xd8\x12\x16\x88\x43\xd8\x15\x1a\x89\x46\xe0\x12\x15\x88\x43\xd8\x14\x16\x88\x45\xd8\x12\x15\x88\x43\xd8\x15\x19\x88\x46\xd8\x0b\x0f\x90\x35\x8a\x3d\xd8\x13\x16\x88\x4a\xdc\x23\x2c\xa8\x54\xa3\x3f\xd1\x08\x20\x88\x01\x88\x3f\x98\x44\xd8\x10\x14\x97\x0a\x91\x0a\x98\x33\x93\x0f\x88\x05\xd8\x14\x16\x88\x09\xdb\x14\x19\x88\x44\xd8\x0f\x13\x98\x05\x98\x73\x90\x7c\xd1\x0f\x23\xd8\x10\x18\xd8\x10\x14\x98\x06\x92\x0e\xa1\x7f\xb9\x79\xd9\x12\x1b\xa0\x09\xa8\x22\xa1\x0d\xb0\x16\xd2\x20\x37\xd8\x10\x19\xd7\x10\x20\xd1\x10\x20\xa0\x14\xd5\x10\x26\xda\x11\x1a\xd8\x10\x19\x97\x0d\x91\x0d\x95\x0f\xf0\x0f\x00\x15\x1a\xf0\x10\x00\x11\x1a\x88\x05\xd8\x0f\x1e\xa0\x13\xa7\x18\xa1\x18\xa8\x25\xa3\x1f\xd1\x0f\x30\x88\x04\xd8\x0f\x13\x8a\x7b\x90\x73\xd0\x08\x1a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_dotdot = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dotdot",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_initial_slashes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "initial_slashes",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_new_comps = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "new_comps",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_comp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "comp",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
posixpath_toplevel_consts_28_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(sep),
            & const_str_empty._ascii.ob_base,
            & const_str_dot._ascii.ob_base,
            & const_str_dotdot._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            & const_str_initial_slashes._ascii.ob_base,
            & const_str_comps._ascii.ob_base,
            & const_str_new_comps._ascii.ob_base,
            & const_str_comp._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(388)
posixpath_toplevel_consts_28 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 194,
    },
    .co_consts = & posixpath_toplevel_consts_28_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_28_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 374,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 597,
    .co_localsplusnames = & posixpath_toplevel_consts_28_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_normpath._ascii.ob_base,
    .co_qualname = & const_str_normpath._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_28_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x09\x64\x01\x7d\x01\x64\x02\x7d\x02\x64\x03\x7d\x03\x64\x04\x7d\x04\x6e\x08\x64\x05\x7d\x01\x64\x06\x7d\x02\x64\x07\x7d\x03\x64\x08\x7d\x04\x7c\x00\x7c\x02\x6b\x28\x00\x00\x72\x02\x7c\x03\x53\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x05\x7d\x06\x7d\x00\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x67\x00\x7d\x08\x7c\x07\x44\x00\x5d\x41\x00\x00\x7d\x09\x7c\x09\x7c\x02\x7c\x03\x66\x02\x76\x00\x72\x01\x8c\x0a\x7c\x09\x7c\x04\x6b\x37\x00\x00\x73\x0e\x7c\x06\x73\x02\x7c\x08\x72\x0a\x7c\x08\x72\x1a\x7c\x08\x64\x09\x19\x00\x00\x00\x7c\x04\x6b\x28\x00\x00\x72\x12\x7c\x08\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x2f\x7c\x08\x73\x01\x8c\x32\x7c\x08\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x43\x04\x00\x7c\x08\x7d\x07\x7c\x06\x7c\x01\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x7d\x00\x7c\x00\x78\x01\x73\x02\x01\x00\x7c\x03\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
posixpath_toplevel_consts_29_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return an absolute path.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
posixpath_toplevel_consts_29_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & posixpath_toplevel_consts_29_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[76];
    }
posixpath_toplevel_consts_29_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 75,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0d\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x0b\x10\x90\x14\x8c\x3b\xdc\x0b\x15\x90\x64\x9c\x45\xd4\x0b\x22\xdc\x12\x14\x97\x2a\x91\x2a\x93\x2c\x89\x43\xe4\x12\x14\x97\x29\x91\x29\x93\x2b\x88\x43\xdc\x0f\x13\x90\x43\x98\x14\x8b\x7f\x88\x04\xdc\x0b\x13\x90\x44\x8b\x3e\xd0\x04\x19",
};
static
    struct _PyCode_DEF(226)
posixpath_toplevel_consts_29 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 113,
    },
    .co_consts = & posixpath_toplevel_consts_29_consts._object.ob_base.ob_base,
    .co_names = & ntpath_toplevel_consts_37_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 405,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 598,
    .co_localsplusnames = & ntpath_toplevel_consts_37_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_abspath._ascii.ob_base,
    .co_qualname = & const_str_abspath._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_29_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x45\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x15\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x6e\x14\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[109];
    }
posixpath_toplevel_consts_32_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 108,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x63\x61\x6e\x6f\x6e\x69\x63\x61\x6c\x20\x70\x61\x74\x68\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x2c\x20\x65\x6c\x69\x6d\x69\x6e\x61\x74\x69\x6e\x67\x20\x61\x6e\x79\x0a\x73\x79\x6d\x62\x6f\x6c\x69\x63\x20\x6c\x69\x6e\x6b\x73\x20\x65\x6e\x63\x6f\x75\x6e\x74\x65\x72\x65\x64\x20\x69\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_32_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & posixpath_toplevel_consts_32_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__joinrealpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_joinrealpath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_32_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__joinrealpath._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[55];
    }
posixpath_toplevel_consts_32_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 54,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x10\x12\x8f\x79\x89\x79\x98\x18\xd3\x0f\x22\x80\x48\xdc\x0f\x1c\x98\x58\xa0\x62\xa0\x71\x98\x5c\xa8\x38\xb0\x56\xb8\x52\xd3\x0f\x40\x81\x48\x80\x44\x88\x22\xdc\x0b\x12\x90\x34\x8b\x3d\xd0\x04\x18",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
const_str_ok = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ok",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
posixpath_toplevel_consts_32_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(filename),
            &_Py_ID(strict),
            &_Py_ID(path),
            & const_str_ok._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(106)
posixpath_toplevel_consts_32 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 53,
    },
    .co_consts = & posixpath_toplevel_consts_32_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_32_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 420,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 599,
    .co_localsplusnames = & posixpath_toplevel_consts_32_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_realpath._ascii.ob_base,
    .co_qualname = & const_str_realpath._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_32_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\x64\x02\x1a\x00\x7c\x00\x7c\x01\x69\x00\xab\x04\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
posixpath_toplevel_consts_33_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            & ntpath_toplevel_consts_34_consts_4.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & ntpath_toplevel_consts_2._ascii.ob_base,
            (PyObject *)& _Py_SINGLETON(tuple_empty),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_False,
            Py_True,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
posixpath_toplevel_consts_33_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_os._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str_ALLOW_MISSING._ascii.ob_base,
            & const_str_FileNotFoundError._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            & const_str_partition._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            &_Py_ID(join),
            & const_str_lstat._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_S_ISLNK._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
            & const_str__joinrealpath._ascii.ob_base,
            & const_str_readlink._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[424];
    }
posixpath_toplevel_consts_33_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 423,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x0e\x12\x88\x03\xd8\x11\x15\x88\x06\xd8\x11\x16\x89\x06\xe0\x0e\x11\x88\x03\xd8\x11\x14\x88\x06\xd8\x11\x15\x88\x06\xdc\x11\x13\x97\x19\x91\x19\x88\x06\xd8\x07\x0d\x94\x1d\xd1\x07\x1e\xdc\x18\x29\x89\x0d\xd9\x09\x0f\xd8\x18\x1a\x89\x0d\xe4\x18\x1f\x88\x0d\xe0\x0f\x13\x80\x48\xe4\x07\x0c\x88\x54\x84\x7b\xd8\x0f\x13\x90\x41\x90\x42\x88\x78\x88\x04\xd8\x0f\x12\x88\x04\xe2\x0a\x0e\xd8\x18\x1c\x9f\x0e\x99\x0e\xa0\x73\xd3\x18\x2b\x89\x0d\x88\x04\x88\x61\x90\x14\xd9\x0f\x13\x90\x74\x98\x76\x92\x7e\xe0\x0c\x14\xd8\x0b\x0f\x90\x36\x8a\x3e\xe1\x0f\x13\xdc\x1d\x22\xa0\x34\x9b\x5b\x91\x0a\x90\x04\x90\x64\xd8\x13\x17\x98\x36\x92\x3e\xdc\x1b\x1f\xa0\x04\xa0\x66\xa8\x66\xd3\x1b\x35\x91\x44\xe0\x17\x1d\x90\x04\xd8\x0c\x14\xdc\x12\x16\x90\x74\x98\x54\xd3\x12\x22\x88\x07\xf0\x02\x05\x09\x2f\xdc\x11\x13\x97\x18\x91\x18\x98\x27\xd3\x11\x22\x88\x42\xf4\x08\x00\x17\x1b\x97\x6c\x91\x6c\xa0\x32\xa7\x3a\xa1\x3a\xd3\x16\x2e\x88\x47\xd9\x0f\x16\xd8\x13\x1a\x88\x44\xd8\x0c\x14\xe0\x0b\x12\x90\x64\x89\x3f\xe0\x13\x17\x98\x07\x91\x3d\x88\x44\xd8\x0f\x13\xd0\x0f\x1f\xe0\x10\x18\xe1\x0f\x15\xe4\x10\x12\x97\x07\x91\x07\x98\x07\xd5\x10\x20\xf4\x06\x00\x18\x1c\x98\x47\xa0\x54\xd3\x17\x2a\xa8\x45\xd0\x17\x31\xd0\x10\x31\xd8\x18\x1c\x88\x04\x88\x57\x89\x0d\xdc\x13\x20\xa0\x14\xa4\x72\xa7\x7b\xa1\x7b\xb0\x37\xd3\x27\x3b\xb8\x56\xc0\x54\xd3\x13\x4a\x89\x08\x88\x04\x88\x62\xd9\x0f\x11\xdc\x13\x17\x98\x04\x98\x64\xd3\x13\x23\xa0\x55\xd0\x13\x2a\xd0\x0c\x2a\xd8\x18\x1c\x88\x04\x88\x57\x89\x0d\xf3\x55\x01\x00\x0b\x0f\xf0\x58\x01\x00\x0c\x10\x90\x14\x88\x3a\xd0\x04\x15\xf8\xf0\x37\x00\x10\x1d\xf2\x00\x01\x09\x1c\xd8\x16\x1b\x8a\x47\xf0\x03\x01\x09\x1c\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
posixpath_toplevel_consts_33_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x34\x15\x45\x25\x00\xc5\x25\x07\x45\x2f\x03\xc5\x2e\x01\x45\x2f\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_maxlinks = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "maxlinks",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_newpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "newpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_is_link = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_link",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
posixpath_toplevel_consts_33_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_rest._ascii.ob_base,
            &_Py_ID(strict),
            & const_str_seen._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            & const_str_getcwd._ascii.ob_base,
            & const_str_ignored_error._ascii.ob_base,
            & const_str_maxlinks._ascii.ob_base,
            &_Py_ID(name),
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            & const_str_newpath._ascii.ob_base,
            & const_str_st._ascii.ob_base,
            & const_str_is_link._ascii.ob_base,
            & const_str_ok._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
posixpath_toplevel_consts_33_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "                ",
};
static
    struct _PyCode_DEF(740)
posixpath_toplevel_consts_33 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 370,
    },
    .co_consts = & posixpath_toplevel_consts_33_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_33_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_consts_33_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 22 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 429,
    .co_nlocalsplus = 16,
    .co_nlocals = 16,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 600,
    .co_localsplusnames = & posixpath_toplevel_consts_33_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & posixpath_toplevel_consts_33_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str__joinrealpath._ascii.ob_base,
    .co_qualname = & const_str__joinrealpath._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_33_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x07\x64\x01\x7d\x04\x64\x02\x7d\x05\x64\x03\x7d\x06\x6e\x16\x64\x04\x7d\x04\x64\x05\x7d\x05\x64\x06\x7d\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x02\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x07\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x6e\x0b\x7c\x02\x72\x03\x64\x07\x7d\x08\x6e\x06\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x64\x00\x7d\x09\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x72\x07\x7c\x01\x64\x08\x64\x00\x1a\x00\x7d\x01\x7c\x04\x7d\x00\x7c\x01\x90\x01\x72\x02\x7c\x01\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x0a\x7d\x0b\x7d\x01\x7c\x0a\x72\x05\x7c\x0a\x7c\x05\x6b\x28\x00\x00\x72\x01\x8c\x20\x7c\x0a\x7c\x06\x6b\x28\x00\x00\x72\x26\x7c\x00\x72\x21\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x00\x7d\x0a\x7c\x0a\x7c\x06\x6b\x28\x00\x00\x72\x10\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x06\x7c\x06\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x00\x6e\x02\x7c\x06\x7d\x00\x8c\x4b\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x0a\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x0c\x09\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0d\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0e\x7c\x0e\x73\x03\x7c\x0c\x7d\x00\x8c\x91\x7c\x0c\x7c\x03\x76\x00\x72\x2e\x7c\x03\x7c\x0c\x19\x00\x00\x00\x7d\x00\x7c\x00\x81\x01\x8c\x9d\x7c\x02\x72\x16\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x0e\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x64\x09\x66\x02\x53\x00\x64\x00\x7c\x03\x7c\x0c\x3c\x00\x00\x00\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x03\xab\x04\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x00\x7d\x0f\x7c\x0f\x73\x0e\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x64\x09\x66\x02\x53\x00\x7c\x00\x7c\x03\x7c\x0c\x3c\x00\x00\x00\x7c\x01\x72\x02\x90\x01\x8c\x02\x7c\x00\x64\x0a\x66\x02\x53\x00\x23\x00\x7c\x08\x24\x00\x72\x05\x01\x00\x64\x09\x7d\x0e\x59\x00\x8c\x86\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
posixpath_toplevel_consts_35_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & ntpath_toplevel_consts_46_consts_0._ascii.ob_base,
            & ntpath_toplevel_consts_46_consts_8._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            & ntpath_toplevel_consts_34_consts_4.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            & ntpath_toplevel_consts_2._ascii.ob_base,
            Py_None,
            & const_str_relpath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
posixpath_toplevel_consts_35_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            & const_str_ValueError._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_abspath._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            &_Py_ID(len),
            & const_str_commonprefix._ascii.ob_base,
            &_Py_ID(join),
            & const_str_TypeError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_BytesWarning._ascii.ob_base,
            & const_str_DeprecationWarning._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            & const_str__check_arg_types._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[307];
    }
posixpath_toplevel_consts_35_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 306,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf1\x06\x00\x0c\x10\xdc\x0e\x18\xd0\x19\x2c\xd3\x0e\x2d\xd0\x08\x2d\xe4\x0b\x0d\x8f\x39\x89\x39\x90\x54\x8b\x3f\x80\x44\xdc\x07\x11\x90\x24\x9c\x05\xd4\x07\x1e\xd8\x11\x15\x88\x06\xd8\x0e\x12\x88\x03\xd8\x11\x16\x89\x06\xe0\x11\x14\x88\x06\xd8\x0e\x11\x88\x03\xd8\x11\x15\x88\x06\xe0\x07\x0c\x80\x7d\xd8\x10\x16\x89\x05\xe4\x10\x12\x97\x09\x91\x09\x98\x25\xd3\x10\x20\x88\x05\xf0\x04\x0c\x05\x0e\xdc\x21\x28\xa8\x15\xa3\x1e\xd7\x21\x35\xd1\x21\x35\xb0\x63\xd4\x21\x3a\xd3\x15\x40\xd1\x21\x3a\x98\x41\xba\x61\x92\x61\xd0\x21\x3a\x88\x0a\xd0\x15\x40\xdc\x20\x27\xa8\x04\xa3\x0d\xd7\x20\x33\xd1\x20\x33\xb0\x43\xd4\x20\x38\xd3\x14\x3e\xd1\x20\x38\x98\x31\xba\x41\x92\x51\xd0\x20\x38\x88\x09\xd0\x14\x3e\xe4\x0c\x0f\x94\x0c\x98\x6a\xa8\x29\xd0\x1d\x34\xd3\x10\x35\xd3\x0c\x36\x88\x01\xe0\x14\x1a\x90\x38\x9c\x73\xa0\x3a\x9b\x7f\xa8\x71\xd1\x1f\x30\xd1\x13\x31\xb0\x49\xb8\x61\xb8\x62\xb0\x4d\xd1\x13\x41\x88\x08\xd9\x0f\x17\xd8\x13\x19\x88\x4d\xdc\x0f\x13\x90\x58\x88\x7f\xd0\x08\x1e\xf9\xf2\x11\x00\x16\x41\x01\xf9\xda\x14\x3e\xf8\xf4\x10\x00\x0d\x16\x94\x7e\xa4\x7c\xd4\x35\x47\xd0\x0b\x48\xf2\x00\x02\x05\x0e\xdc\x08\x13\xd7\x08\x24\xd1\x08\x24\xa0\x59\xb0\x04\xb0\x65\xd4\x08\x3c\xd8\x08\x0d\xf0\x05\x02\x05\x0e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[61];
    }
posixpath_toplevel_consts_35_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 60,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x1b\x1c\x43\x33\x00\xc1\x37\x07\x43\x29\x04\xc1\x3f\x04\x43\x29\x04\xc2\x03\x1e\x43\x33\x00\xc2\x21\x07\x43\x2e\x04\xc2\x29\x04\x43\x2e\x04\xc2\x2d\x33\x43\x33\x00\xc3\x21\x07\x43\x33\x00\xc3\x29\x0a\x43\x33\x00\xc3\x33\x32\x44\x25\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
posixpath_toplevel_consts_35_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(start),
            & const_str_curdir._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pardir._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
            & const_str_start_list._ascii.ob_base,
            & const_str_path_list._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            & const_str_rel_list._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(592)
posixpath_toplevel_consts_35 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 296,
    },
    .co_consts = & posixpath_toplevel_consts_35_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_35_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_consts_35_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 501,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 601,
    .co_localsplusnames = & posixpath_toplevel_consts_35_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_relpath._ascii.ob_base,
    .co_qualname = & const_str_relpath._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_35_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x73\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x07\x64\x02\x7d\x02\x64\x03\x7d\x03\x64\x04\x7d\x04\x6e\x06\x64\x05\x7d\x02\x64\x06\x7d\x03\x64\x07\x7d\x04\x7c\x01\x80\x03\x7c\x02\x7d\x01\x6e\x15\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x8f\x05\x63\x02\x67\x00\x63\x02\x5d\x07\x00\x00\x7d\x05\x7c\x05\x73\x01\x8c\x06\x7c\x05\x91\x02\x8c\x09\x04\x00\x7d\x06\x7d\x05\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x8f\x05\x63\x02\x67\x00\x63\x02\x5d\x07\x00\x00\x7d\x05\x7c\x05\x73\x01\x8c\x06\x7c\x05\x91\x02\x8c\x09\x04\x00\x7d\x07\x7d\x05\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x07\x67\x02\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x04\x67\x01\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x08\x7a\x0a\x00\x00\x7a\x05\x00\x00\x7c\x07\x7c\x08\x64\x08\x1a\x00\x7a\x00\x00\x00\x7d\x09\x7c\x09\x73\x02\x7c\x02\x53\x00\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x8e\x00\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x05\x77\x00\x63\x02\x01\x00\x63\x02\x7d\x05\x77\x00\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x66\x04\x24\x00\x72\x19\x01\x00\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\x7c\x00\x7c\x01\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
posixpath_toplevel_consts_36_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "commonpath.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
posixpath_toplevel_consts_36_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xd0\x18\x35\xa9\x75\xa8\x21\x98\x11\x98\x32\x98\x41\x98\x15\xa0\x23\x9d\x1c\xa9\x75\xf9",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
posixpath_toplevel_consts_36_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            &_Py_ID(sep),
        },
    },
};
static
    struct _PyCode_DEF(46)
posixpath_toplevel_consts_36_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & zipimport_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = & importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 560,
    .co_nlocalsplus = 3,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 602,
    .co_localsplusnames = & posixpath_toplevel_consts_36_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & posixpath_toplevel_consts_36_consts_7_qualname._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_36_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x0c\x00\x00\x7d\x01\x7c\x01\x64\x00\x64\x01\x1a\x00\x89\x02\x6b\x28\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x0e\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
posixpath_toplevel_consts_36_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & ntpath_toplevel_consts_47_consts_0._ascii.ob_base,
            & ntpath_toplevel_consts_47_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_Py_SINGLETON(bytes_characters[47]),
            (PyObject *)&_Py_SINGLETON(bytes_characters[46]),
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & posixpath_toplevel_consts_36_consts_7.ob_base.ob_base,
            & ntpath_toplevel_consts_47_consts_10._ascii.ob_base,
            Py_None,
            & const_str_commonpath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
posixpath_toplevel_consts_36_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            & const_str_ValueError._ascii.ob_base,
            & const_str_tuple._ascii.ob_base,
            & const_str_map._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_split._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            & const_str_min._ascii.ob_base,
            & const_str_max._ascii.ob_base,
            & const_str_enumerate._ascii.ob_base,
            &_Py_ID(join),
            & const_str_TypeError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            & const_str__check_arg_types._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[374];
    }
posixpath_toplevel_consts_36_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 373,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf1\x06\x00\x0c\x11\xdc\x0e\x18\xd0\x19\x40\xd3\x0e\x41\xd0\x08\x41\xe4\x0c\x11\x94\x23\x94\x62\x97\x69\x91\x69\xa0\x15\xd3\x12\x27\xd3\x0c\x28\x80\x45\xdc\x07\x11\x90\x25\x98\x01\x91\x28\x9c\x45\xd4\x07\x22\xd8\x0e\x12\x88\x03\xd8\x11\x15\x89\x06\xe0\x0e\x11\x88\x03\xd8\x11\x14\x88\x06\xf0\x04\x15\x05\x0e\xd9\x33\x38\xd3\x16\x39\xb1\x35\xa8\x34\x90\x74\x97\x7a\x91\x7a\xa0\x23\x95\x7f\xb0\x35\x88\x0b\xd0\x16\x39\xf0\x04\x03\x09\x50\x01\xdc\x15\x18\xd3\x18\x35\xa9\x75\xd3\x18\x35\xd3\x15\x35\x89\x46\x88\x45\xf1\x08\x00\x45\x01\x50\x01\xd4\x16\x50\xc1\x4b\xb8\x71\xa1\x31\xd3\x17\x3a\xa1\x31\x98\x61\xaa\x01\xa8\x61\xb0\x36\xab\x6b\x9a\x01\xa0\x31\xd3\x17\x3a\xc0\x4b\x88\x0b\xd1\x16\x50\xdc\x0d\x10\x90\x1b\xd3\x0d\x1d\x88\x02\xdc\x0d\x10\x90\x1b\xd3\x0d\x1d\x88\x02\xd8\x11\x13\x88\x06\xdc\x14\x1d\x98\x62\x96\x4d\x89\x44\x88\x41\x88\x71\xd8\x0f\x10\x90\x42\x90\x71\x91\x45\x8b\x7a\xd8\x19\x1b\x98\x42\x98\x51\x98\x16\x90\x06\xd9\x10\x15\xf0\x07\x00\x15\x22\xf1\x0a\x00\x19\x1e\x91\x13\xa0\x33\xa0\x72\xa8\x01\xa0\x37\x88\x06\xd8\x0f\x15\x98\x03\x9f\x08\x99\x08\xa0\x16\xd3\x18\x28\xd1\x0f\x28\xd0\x08\x28\xf9\xf2\x23\x00\x17\x3a\xf8\xf4\x08\x00\x10\x1a\xf2\x00\x01\x09\x50\x01\xdc\x12\x1c\xd0\x1d\x44\xd3\x12\x45\xc8\x34\xd0\x0c\x4f\xf0\x03\x01\x09\x50\x01\xfc\xf2\x06\x00\x18\x3b\xf9\xd3\x16\x50\xf8\xf4\x16\x00\x0d\x16\x94\x7e\xd0\x0b\x26\xf2\x00\x02\x05\x0e\xdc\x08\x13\xd7\x08\x24\xd1\x08\x24\xa0\x5c\xd0\x08\x3a\xb0\x45\xd3\x08\x3a\xd8\x08\x0d\xf0\x05\x02\x05\x0e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[109];
    }
posixpath_toplevel_consts_36_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 108,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x0f\x04\x44\x2c\x00\xc1\x13\x18\x44\x03\x04\xc1\x2b\x02\x44\x2c\x00\xc1\x2e\x16\x44\x08\x00\xc2\x04\x05\x44\x2c\x00\xc2\x09\x09\x44\x26\x06\xc2\x12\x07\x44\x21\x0c\xc2\x1a\x05\x44\x21\x0c\xc2\x20\x04\x44\x21\x0c\xc2\x24\x05\x44\x26\x06\xc2\x29\x34\x44\x2c\x00\xc3\x1e\x24\x44\x2c\x00\xc4\x03\x05\x44\x2c\x00\xc4\x08\x16\x44\x1e\x03\xc4\x1e\x03\x44\x2c\x00\xc4\x21\x05\x44\x26\x06\xc4\x26\x06\x44\x2c\x00\xc4\x2c\x27\x45\x13\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
posixpath_toplevel_consts_36_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_paths._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            &_Py_ID(path),
            & const_str_split_paths._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            (PyObject *)&_Py_SINGLETON(strings).ascii[99],
            & const_str_s1._ascii.ob_base,
            & const_str_s2._ascii.ob_base,
            & const_str_common._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[105],
            & const_str_prefix._ascii.ob_base,
            &_Py_ID(sep),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[14];
    }
posixpath_toplevel_consts_36_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 13,
    },
    .ob_shash = -1,
    .ob_sval = "            @",
};
static
    struct _PyCode_DEF(684)
posixpath_toplevel_consts_36 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 342,
    },
    .co_consts = & posixpath_toplevel_consts_36_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_consts_36_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_consts_36_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 22 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 542,
    .co_nlocalsplus = 13,
    .co_nlocals = 12,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 603,
    .co_localsplusnames = & posixpath_toplevel_consts_36_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & posixpath_toplevel_consts_36_localspluskinds.ob_base.ob_base,
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_commonpath._ascii.ob_base,
    .co_qualname = & const_str_commonpath._ascii.ob_base,
    .co_linetable = & posixpath_toplevel_consts_36_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x0c\x97\x00\x7c\x00\x73\x0b\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x02\x19\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x05\x64\x03\x8a\x0c\x64\x04\x7d\x01\x6e\x04\x64\x05\x8a\x0c\x64\x06\x7d\x01\x09\x00\x7c\x00\x44\x00\x8f\x02\x63\x02\x67\x00\x63\x02\x5d\x13\x00\x00\x7d\x02\x7c\x02\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x91\x02\x8c\x15\x04\x00\x7d\x03\x7d\x02\x09\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x88\x0c\x66\x01\x64\x07\x84\x08\x7c\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x01\x00\x00\x7d\x04\x7c\x03\x44\x00\x8f\x05\x8f\x06\x63\x03\x67\x00\x63\x02\x5d\x1b\x00\x00\x7d\x05\x7c\x05\x44\x00\x8f\x06\x63\x02\x67\x00\x63\x02\x5d\x0d\x00\x00\x7d\x06\x7c\x06\x73\x01\x8c\x06\x7c\x06\x7c\x01\x6b\x37\x00\x00\x73\x01\x8c\x0c\x7c\x06\x91\x02\x8c\x0f\x04\x00\x63\x02\x7d\x06\x91\x02\x8c\x1d\x04\x00\x7d\x03\x7d\x05\x7d\x06\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x07\x7d\x09\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x14\x00\x00\x5c\x02\x00\x00\x7d\x0a\x7d\x06\x7c\x06\x7c\x08\x7c\x0a\x19\x00\x00\x00\x6b\x37\x00\x00\x73\x01\x8c\x0f\x7c\x07\x64\x09\x7c\x0a\x1a\x00\x7d\x09\x01\x00\x6e\x01\x04\x00\x7c\x04\x72\x02\x89\x0c\x6e\x04\x89\x0c\x64\x09\x64\x02\x1a\x00\x7d\x0b\x7c\x0b\x89\x0c\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x53\x00\x63\x02\x01\x00\x63\x02\x7d\x02\x77\x00\x23\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0d\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\xab\x01\x00\x00\x00\x00\x00\x00\x64\x09\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01\x63\x02\x01\x00\x63\x02\x7d\x06\x77\x00\x63\x02\x01\x00\x63\x03\x7d\x06\x7d\x05\x77\x00\x23\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x18\x01\x00\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\x67\x01\x7c\x00\xa2\x01\xad\x06\x8e\x00\x01\x00\x82\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[38];
        }_object;
    }
posixpath_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 38,
        },
        .ob_item = {
            & posixpath_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & ntpath_toplevel_consts_2._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            (PyObject *)&_Py_SINGLETON(strings).ascii[58],
            & posixpath_toplevel_consts_5._ascii.ob_base,
            Py_None,
            & posixpath_toplevel_consts_7._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & codecs_toplevel_consts_3._object.ob_base.ob_base,
            & posixpath_toplevel_consts_10._object.ob_base.ob_base,
            & posixpath_toplevel_consts_11.ob_base.ob_base,
            & posixpath_toplevel_consts_12.ob_base.ob_base,
            & posixpath_toplevel_consts_13.ob_base.ob_base,
            & posixpath_toplevel_consts_14.ob_base.ob_base,
            & posixpath_toplevel_consts_15.ob_base.ob_base,
            & posixpath_toplevel_consts_16.ob_base.ob_base,
            & posixpath_toplevel_consts_17.ob_base.ob_base,
            & posixpath_toplevel_consts_18.ob_base.ob_base,
            & posixpath_toplevel_consts_19.ob_base.ob_base,
            & posixpath_toplevel_consts_20.ob_base.ob_base,
            & posixpath_toplevel_consts_21.ob_base.ob_base,
            & posixpath_toplevel_consts_22.ob_base.ob_base,
            & posixpath_toplevel_consts_23.ob_base.ob_base,
            & posixpath_toplevel_consts_24.ob_base.ob_base,
            & posixpath_toplevel_consts_25._ascii.ob_base,
            & posixpath_toplevel_consts_26.ob_base.ob_base,
            & ntpath_toplevel_consts_33._object.ob_base.ob_base,
            & posixpath_toplevel_consts_28.ob_base.ob_base,
            & posixpath_toplevel_consts_29.ob_base.ob_base,
            Py_False,
            & codecs_toplevel_consts_14_consts_5._object.ob_base.ob_base,
            & posixpath_toplevel_consts_32.ob_base.ob_base,
            & posixpath_toplevel_consts_33.ob_base.ob_base,
            & const_str_darwin._ascii.ob_base,
            & posixpath_toplevel_consts_35.ob_base.ob_base,
            & posixpath_toplevel_consts_36.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[44];
        }_object;
    }
posixpath_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 44,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pathsep._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            & const_str_altsep._ascii.ob_base,
            & const_str_devnull._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_genericpath._ascii.ob_base,
            &_Py_ID(__all__),
            & const_str__get_sep._ascii.ob_base,
            & const_str_normcase._ascii.ob_base,
            & const_str_isabs._ascii.ob_base,
            &_Py_ID(join),
            & const_str_split._ascii.ob_base,
            & const_str_splitext._ascii.ob_base,
            & const_str__splitext._ascii.ob_base,
            & const_str_splitdrive._ascii.ob_base,
            & const_str_splitroot._ascii.ob_base,
            & const_str_basename._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
            & const_str_isjunction._ascii.ob_base,
            & const_str_lexists._ascii.ob_base,
            & const_str_ismount._ascii.ob_base,
            & const_str_expanduser._ascii.ob_base,
            & const_str__varpattern._ascii.ob_base,
            & const_str__varsub._ascii.ob_base,
            & const_str__varsubb._ascii.ob_base,
            & const_str_expandvars._ascii.ob_base,
            &_Py_ID(posix),
            & const_str__path_normpath._ascii.ob_base,
            & const_str_normpath._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str_realpath._ascii.ob_base,
            & const_str__joinrealpath._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str_supports_unicode_filenames._ascii.ob_base,
            & const_str_relpath._ascii.ob_base,
            & const_str_commonpath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[274];
    }
posixpath_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 273,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x0a\x01\x04\xf0\x1e\x00\x0a\x0d\x80\x06\xd8\x09\x0d\x80\x06\xd8\x09\x0c\x80\x06\xd8\x06\x09\x80\x03\xd8\x0a\x0d\x80\x07\xd8\x0a\x19\x80\x07\xd8\x09\x0d\x80\x06\xd8\x0a\x15\x80\x07\xe3\x00\x09\xdb\x00\x0a\xdb\x00\x0b\xdb\x00\x12\xdc\x00\x19\xf2\x04\x07\x0b\x38\x80\x07\xf2\x14\x04\x01\x13\xf2\x16\x02\x01\x18\xf2\x10\x04\x01\x1d\xf2\x16\x15\x01\x10\xf2\x3a\x09\x01\x16\xf2\x22\x08\x01\x37\xf0\x12\x00\x14\x1f\xd7\x13\x28\xd1\x13\x28\xd7\x13\x30\xd1\x13\x30\x80\x08\xd4\x00\x10\xf2\x0a\x04\x01\x14\xf2\x0e\x1a\x01\x23\xf2\x3e\x05\x01\x11\xf2\x14\x08\x01\x10\xf2\x1a\x04\x01\x11\xf2\x12\x06\x01\x10\xf2\x18\x1f\x01\x11\xf2\x56\x01\x36\x01\x29\xf0\x7a\x01\x00\x0f\x24\x80\x0b\xd8\x0a\x0e\x80\x07\xd8\x0b\x0f\x80\x08\xf2\x04\x2a\x01\x1b\xf0\x62\x01\x20\x01\x1b\xdd\x04\x30\xf2\x44\x01\x09\x01\x1a\xf0\x1e\x00\x22\x27\xf4\x00\x05\x01\x19\xf2\x12\x43\x01\x01\x16\xf0\x4c\x02\x00\x1f\x22\x9f\x6c\x99\x6c\xa8\x68\xd1\x1e\x36\xd0\x00\x1a\xf3\x04\x21\x01\x0e\xf3\x52\x01\x23\x01\x0e\xf8\xf0\x53\x05\x00\x08\x13\xf2\x00\x1d\x01\x1b\xf4\x02\x1c\x05\x1b\xf0\x03\x1d\x01\x1b\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
posixpath_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x3b\x06\x42\x24\x00\xc2\x24\x08\x42\x2f\x03\xc2\x2e\x01\x42\x2f\x03",
};
static
    struct _PyCode_DEF(356)
posixpath_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 178,
    },
    .co_consts = & posixpath_toplevel_consts._object.ob_base.ob_base,
    .co_names = & posixpath_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & posixpath_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 604,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & posixpath_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & posixpath_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x5a\x01\x64\x02\x5a\x02\x64\x01\x5a\x03\x64\x03\x5a\x04\x64\x04\x5a\x05\x64\x05\x5a\x06\x64\x06\x5a\x07\x64\x07\x5a\x08\x64\x08\x64\x06\x6c\x09\x5a\x09\x64\x08\x64\x06\x6c\x0a\x5a\x0a\x64\x08\x64\x06\x6c\x0b\x5a\x0b\x64\x08\x64\x06\x6c\x0c\x5a\x0c\x64\x08\x64\x09\x6c\x0c\xad\x02\x01\x00\x67\x00\x64\x0a\xa2\x01\x5a\x0d\x64\x0b\x84\x00\x5a\x0e\x64\x0c\x84\x00\x5a\x0f\x64\x0d\x84\x00\x5a\x10\x64\x0e\x84\x00\x5a\x11\x64\x0f\x84\x00\x5a\x12\x64\x10\x84\x00\x5a\x13\x65\x0c\x6a\x28\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x13\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x11\x84\x00\x5a\x15\x64\x12\x84\x00\x5a\x16\x64\x13\x84\x00\x5a\x17\x64\x14\x84\x00\x5a\x18\x64\x15\x84\x00\x5a\x19\x64\x16\x84\x00\x5a\x1a\x64\x17\x84\x00\x5a\x1b\x64\x18\x84\x00\x5a\x1c\x64\x19\x5a\x1d\x64\x06\x61\x1e\x64\x06\x61\x1f\x64\x1a\x84\x00\x5a\x20\x09\x00\x64\x08\x64\x1b\x6c\x21\x6d\x22\x5a\x23\x01\x00\x64\x1d\x84\x00\x5a\x25\x64\x1e\x64\x1f\x9c\x01\x64\x20\x84\x02\x5a\x26\x64\x21\x84\x00\x5a\x27\x65\x0a\x6a\x50\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x22\x6b\x28\x00\x00\x5a\x29\x64\x25\x64\x23\x84\x01\x5a\x2a\x64\x24\x84\x00\x5a\x2b\x79\x06\x23\x00\x65\x24\x24\x00\x72\x06\x01\x00\x64\x1c\x84\x00\x5a\x23\x59\x00\x8c\x2d\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_posixpath_toplevel(void)
{
    return Py_NewRef((PyObject *) &posixpath_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1103];
    }
os_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1102,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x4f\x53\x20\x72\x6f\x75\x74\x69\x6e\x65\x73\x20\x66\x6f\x72\x20\x4e\x54\x20\x6f\x72\x20\x50\x6f\x73\x69\x78\x20\x64\x65\x70\x65\x6e\x64\x69\x6e\x67\x20\x6f\x6e\x20\x77\x68\x61\x74\x20\x73\x79\x73\x74\x65\x6d\x20\x77\x65\x27\x72\x65\x20\x6f\x6e\x2e\x0a\x0a\x54\x68\x69\x73\x20\x65\x78\x70\x6f\x72\x74\x73\x3a\x0a\x20\x20\x2d\x20\x61\x6c\x6c\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x73\x20\x66\x72\x6f\x6d\x20\x70\x6f\x73\x69\x78\x20\x6f\x72\x20\x6e\x74\x2c\x20\x65\x2e\x67\x2e\x20\x75\x6e\x6c\x69\x6e\x6b\x2c\x20\x73\x74\x61\x74\x2c\x20\x65\x74\x63\x2e\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x70\x61\x74\x68\x20\x69\x73\x20\x65\x69\x74\x68\x65\x72\x20\x70\x6f\x73\x69\x78\x70\x61\x74\x68\x20\x6f\x72\x20\x6e\x74\x70\x61\x74\x68\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x6e\x61\x6d\x65\x20\x69\x73\x20\x65\x69\x74\x68\x65\x72\x20\x27\x70\x6f\x73\x69\x78\x27\x20\x6f\x72\x20\x27\x6e\x74\x27\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x63\x75\x72\x64\x69\x72\x20\x69\x73\x20\x61\x20\x73\x74\x72\x69\x6e\x67\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x28\x61\x6c\x77\x61\x79\x73\x20\x27\x2e\x27\x29\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x70\x61\x72\x64\x69\x72\x20\x69\x73\x20\x61\x20\x73\x74\x72\x69\x6e\x67\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x65\x20\x70\x61\x72\x65\x6e\x74\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x28\x61\x6c\x77\x61\x79\x73\x20\x27\x2e\x2e\x27\x29\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x73\x65\x70\x20\x69\x73\x20\x74\x68\x65\x20\x28\x6f\x72\x20\x61\x20\x6d\x6f\x73\x74\x20\x63\x6f\x6d\x6d\x6f\x6e\x29\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x20\x28\x27\x2f\x27\x20\x6f\x72\x20\x27\x5c\x5c\x27\x29\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x65\x78\x74\x73\x65\x70\x20\x69\x73\x20\x74\x68\x65\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x20\x28\x61\x6c\x77\x61\x79\x73\x20\x27\x2e\x27\x29\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x61\x6c\x74\x73\x65\x70\x20\x69\x73\x20\x74\x68\x65\x20\x61\x6c\x74\x65\x72\x6e\x61\x74\x65\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x20\x28\x4e\x6f\x6e\x65\x20\x6f\x72\x20\x27\x2f\x27\x29\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x70\x61\x74\x68\x73\x65\x70\x20\x69\x73\x20\x74\x68\x65\x20\x63\x6f\x6d\x70\x6f\x6e\x65\x6e\x74\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x20\x75\x73\x65\x64\x20\x69\x6e\x20\x24\x50\x41\x54\x48\x20\x65\x74\x63\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x6c\x69\x6e\x65\x73\x65\x70\x20\x69\x73\x20\x74\x68\x65\x20\x6c\x69\x6e\x65\x20\x73\x65\x70\x61\x72\x61\x74\x6f\x72\x20\x69\x6e\x20\x74\x65\x78\x74\x20\x66\x69\x6c\x65\x73\x20\x28\x27\x5c\x72\x27\x20\x6f\x72\x20\x27\x5c\x6e\x27\x20\x6f\x72\x20\x27\x5c\x72\x5c\x6e\x27\x29\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x64\x65\x66\x70\x61\x74\x68\x20\x69\x73\x20\x74\x68\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x73\x65\x61\x72\x63\x68\x20\x70\x61\x74\x68\x20\x66\x6f\x72\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x73\x0a\x20\x20\x2d\x20\x6f\x73\x2e\x64\x65\x76\x6e\x75\x6c\x6c\x20\x69\x73\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x70\x61\x74\x68\x20\x6f\x66\x20\x74\x68\x65\x20\x6e\x75\x6c\x6c\x20\x64\x65\x76\x69\x63\x65\x20\x28\x27\x2f\x64\x65\x76\x2f\x6e\x75\x6c\x6c\x27\x2c\x20\x65\x74\x63\x2e\x29\x0a\x0a\x50\x72\x6f\x67\x72\x61\x6d\x73\x20\x74\x68\x61\x74\x20\x69\x6d\x70\x6f\x72\x74\x20\x61\x6e\x64\x20\x75\x73\x65\x20\x27\x6f\x73\x27\x20\x73\x74\x61\x6e\x64\x20\x61\x20\x62\x65\x74\x74\x65\x72\x20\x63\x68\x61\x6e\x63\x65\x20\x6f\x66\x20\x62\x65\x69\x6e\x67\x0a\x70\x6f\x72\x74\x61\x62\x6c\x65\x20\x62\x65\x74\x77\x65\x65\x6e\x20\x64\x69\x66\x66\x65\x72\x65\x6e\x74\x20\x70\x6c\x61\x74\x66\x6f\x72\x6d\x73\x2e\x20\x20\x4f\x66\x20\x63\x6f\x75\x72\x73\x65\x2c\x20\x74\x68\x65\x79\x20\x6d\x75\x73\x74\x20\x74\x68\x65\x6e\x0a\x6f\x6e\x6c\x79\x20\x75\x73\x65\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x73\x20\x74\x68\x61\x74\x20\x61\x72\x65\x20\x64\x65\x66\x69\x6e\x65\x64\x20\x62\x79\x20\x61\x6c\x6c\x20\x70\x6c\x61\x74\x66\x6f\x72\x6d\x73\x20\x28\x65\x2e\x67\x2e\x2c\x20\x75\x6e\x6c\x69\x6e\x6b\x0a\x61\x6e\x64\x20\x6f\x70\x65\x6e\x64\x69\x72\x29\x2c\x20\x61\x6e\x64\x20\x6c\x65\x61\x76\x65\x20\x61\x6c\x6c\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x6d\x61\x6e\x69\x70\x75\x6c\x61\x74\x69\x6f\x6e\x20\x74\x6f\x20\x6f\x73\x2e\x70\x61\x74\x68\x0a\x28\x65\x2e\x67\x2e\x2c\x20\x73\x70\x6c\x69\x74\x20\x61\x6e\x64\x20\x6a\x6f\x69\x6e\x29\x2e\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__check_methods._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_linesep = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "linesep",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_get_exec_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_exec_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_fdopen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fdopen",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[18];
        }_object;
    }
os_toplevel_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 18,
        },
        .ob_item = {
            & const_str_altsep._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pathsep._ascii.ob_base,
            & const_str_linesep._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(path),
            & const_str_devnull._ascii.ob_base,
            & const_str_SEEK_SET._ascii.ob_base,
            & const_str_SEEK_CUR._ascii.ob_base,
            & const_str_SEEK_END._ascii.ob_base,
            & const_str_fsencode._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            & const_str_get_exec_path._ascii.ob_base,
            & const_str_fdopen._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(globals),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
os_toplevel_consts_5_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen os>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__exists = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_exists",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
os_toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\x94\x37\x93\x39\xd0\x0b\x1c\xd0\x04\x1c",
};
static
    struct _PyCode_DEF(26)
os_toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 41,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 605,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__exists._ascii.ob_base,
    .co_qualname = & const_str__exists._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x76\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_list._ascii.ob_base,
            &_Py_ID(__all__),
            & const_str_AttributeError._ascii.ob_base,
            & const_str_dir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str__get_exports_list = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_exports_list",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[73];
    }
os_toplevel_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 72,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x03\x05\x37\xdc\x0f\x13\x90\x46\x97\x4e\x91\x4e\xd3\x0f\x23\xd0\x08\x23\xf8\xdc\x0b\x19\xf2\x00\x01\x05\x37\xdc\x1b\x1e\x98\x76\x9c\x3b\xd3\x0f\x36\x99\x3b\x90\x61\xa8\x21\xa8\x41\xa9\x24\xb0\x23\xab\x2b\x92\x01\x99\x3b\xf9\xd4\x0f\x36\xd2\x08\x36\xf0\x03\x01\x05\x37\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
os_toplevel_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x14\x17\x00\x97\x16\x41\x0b\x03\xad\x0d\x41\x00\x06\xbb\x04\x41\x00\x06\xbf\x09\x41\x0b\x03\xc1\x0a\x01\x41\x0b\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(module),
            (PyObject *)&_Py_SINGLETON(strings).ascii[110],
        },
    },
};
static
    struct _PyCode_DEF(156)
os_toplevel_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 78,
    },
    .co_consts = & os_toplevel_consts_6_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 44,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 606,
    .co_localsplusnames = & os_toplevel_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__get_exports_list._ascii.ob_base,
    .co_qualname = & const_str__get_exports_list._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x2b\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x8f\x01\x63\x02\x67\x00\x63\x02\x5d\x0d\x00\x00\x7d\x01\x7c\x01\x64\x01\x19\x00\x00\x00\x64\x02\x6b\x37\x00\x00\x73\x01\x8c\x0c\x7c\x01\x91\x02\x8c\x0f\x04\x00\x6e\x05\x63\x02\x01\x00\x63\x02\x7d\x01\x77\x00\x63\x02\x7d\x01\x63\x02\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__exit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_exit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_10 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__exit._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__have_functions = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_have_functions",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_12 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__have_functions._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
os_toplevel_consts_14 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0d\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
os_toplevel_consts_15 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "no os specific module found",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
os_toplevel_consts_16 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "os.path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
os_toplevel_consts_17 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pathsep._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
            & const_str_altsep._ascii.ob_base,
            & const_str_devnull._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__globals = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_globals",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str__set = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_set",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_19_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__globals._ascii.ob_base,
            & const_str__have_functions._ascii.ob_base,
            & const_str__set._ascii.ob_base,
            &_Py_ID(add),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str__add = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_add",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
os_toplevel_consts_19_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0c\x0e\x94\x28\x89\x4e\xa0\x13\xac\x0f\xd1\x21\x37\xdc\x0c\x10\x8f\x48\x89\x48\x94\x58\x98\x62\x91\x5c\xd5\x0c\x22\xf0\x03\x00\x22\x38\x88\x4e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_19_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_str._ascii.ob_base,
            & const_str_fn._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(96)
os_toplevel_consts_19 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 48,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_19_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 104,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 607,
    .co_localsplusnames = & os_toplevel_consts_19_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__add._ascii.ob_base,
    .co_qualname = & const_str__add._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_19_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x26\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x1d\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x79\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_HAVE_FACCESSAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FACCESSAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_FCHMODAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FCHMODAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_chmod = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "chmod",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_FCHOWNAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FCHOWNAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_chown = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "chown",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_HAVE_FSTATAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FSTATAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_HAVE_FUTIMESAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FUTIMESAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_utime = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "utime",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_HAVE_LINKAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_LINKAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_link = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "link",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_HAVE_MKDIRAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_MKDIRAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_MKFIFOAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_MKFIFOAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_mkfifo = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mkfifo",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_HAVE_MKNODAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_MKNODAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_mknod = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mknod",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_HAVE_OPENAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_OPENAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_HAVE_READLINKAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_READLINKAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_RENAMEAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_RENAMEAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_rename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rename",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_HAVE_SYMLINKAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_SYMLINKAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_symlink = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "symlink",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_UNLINKAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_UNLINKAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_rmdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rmdir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_HAVE_UTIMENSAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_UTIMENSAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_HAVE_FCHDIR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FCHDIR",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_chdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "chdir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_HAVE_FCHMOD = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FCHMOD",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_HAVE_FCHOWN = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FCHOWN",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_HAVE_FDOPENDIR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FDOPENDIR",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_scandir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "scandir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_HAVE_FEXECVE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FEXECVE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_execve = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execve",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_HAVE_FTRUNCATE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FTRUNCATE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_FUTIMENS = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FUTIMENS",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_HAVE_FUTIMES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FUTIMES",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_HAVE_FPATHCONF = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FPATHCONF",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_pathconf = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pathconf",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_statvfs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "statvfs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_fstatvfs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fstatvfs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_FSTATVFS = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_FSTATVFS",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_HAVE_LCHFLAGS = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_LCHFLAGS",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_chflags = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "chflags",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_HAVE_LCHMOD = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_LCHMOD",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_lchown = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lchown",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_HAVE_LCHOWN = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_LCHOWN",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_HAVE_LUTIMES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_LUTIMES",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_HAVE_LSTAT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "HAVE_LSTAT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_MS_WINDOWS = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "MS_WINDOWS",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[396];
    }
os_toplevel_consts_79_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 395,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x6d\x61\x6b\x65\x64\x69\x72\x73\x28\x6e\x61\x6d\x65\x20\x5b\x2c\x20\x6d\x6f\x64\x65\x3d\x30\x6f\x37\x37\x37\x5d\x5b\x2c\x20\x65\x78\x69\x73\x74\x5f\x6f\x6b\x3d\x46\x61\x6c\x73\x65\x5d\x29\x0a\x0a\x20\x20\x20\x20\x53\x75\x70\x65\x72\x2d\x6d\x6b\x64\x69\x72\x3b\x20\x63\x72\x65\x61\x74\x65\x20\x61\x20\x6c\x65\x61\x66\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x61\x6e\x64\x20\x61\x6c\x6c\x20\x69\x6e\x74\x65\x72\x6d\x65\x64\x69\x61\x74\x65\x20\x6f\x6e\x65\x73\x2e\x20\x20\x57\x6f\x72\x6b\x73\x20\x6c\x69\x6b\x65\x0a\x20\x20\x20\x20\x6d\x6b\x64\x69\x72\x2c\x20\x65\x78\x63\x65\x70\x74\x20\x74\x68\x61\x74\x20\x61\x6e\x79\x20\x69\x6e\x74\x65\x72\x6d\x65\x64\x69\x61\x74\x65\x20\x70\x61\x74\x68\x20\x73\x65\x67\x6d\x65\x6e\x74\x20\x28\x6e\x6f\x74\x20\x6a\x75\x73\x74\x20\x74\x68\x65\x20\x72\x69\x67\x68\x74\x6d\x6f\x73\x74\x29\x0a\x20\x20\x20\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x63\x72\x65\x61\x74\x65\x64\x20\x69\x66\x20\x69\x74\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x20\x65\x78\x69\x73\x74\x2e\x20\x49\x66\x20\x74\x68\x65\x20\x74\x61\x72\x67\x65\x74\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x61\x6c\x72\x65\x61\x64\x79\x0a\x20\x20\x20\x20\x65\x78\x69\x73\x74\x73\x2c\x20\x72\x61\x69\x73\x65\x20\x61\x6e\x20\x4f\x53\x45\x72\x72\x6f\x72\x20\x69\x66\x20\x65\x78\x69\x73\x74\x5f\x6f\x6b\x20\x69\x73\x20\x46\x61\x6c\x73\x65\x2e\x20\x4f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x6e\x6f\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x69\x73\x0a\x20\x20\x20\x20\x72\x61\x69\x73\x65\x64\x2e\x20\x20\x54\x68\x69\x73\x20\x69\x73\x20\x72\x65\x63\x75\x72\x73\x69\x76\x65\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_exist_ok = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exist_ok",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_79_consts_1 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_exist_ok._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_79_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & os_toplevel_consts_79_consts_0._ascii.ob_base,
            & os_toplevel_consts_79_consts_1._object.ob_base.ob_base,
            & const_str_ASCII._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_makedirs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "makedirs",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
os_toplevel_consts_79_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_split._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str_makedirs._ascii.ob_base,
            & const_str_FileExistsError._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_mkdir._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_isdir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[189];
    }
os_toplevel_consts_79_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 188,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x14\x00\x12\x16\x97\x1a\x91\x1a\x98\x44\xd3\x11\x21\x81\x4a\x80\x44\x88\x24\xd9\x0b\x0f\xdc\x15\x19\x97\x5a\x91\x5a\xa0\x04\xd3\x15\x25\x89\x0a\x88\x04\x88\x64\xd9\x07\x0b\x91\x04\x9c\x54\x9f\x5b\x99\x5b\xa8\x14\xd4\x1d\x2e\xf0\x02\x04\x09\x11\xdc\x0c\x14\x90\x54\xa0\x48\xd5\x0c\x2d\xf4\x08\x00\x10\x16\x88\x04\xdc\x0b\x15\x90\x64\x9c\x45\xd4\x0b\x22\xdc\x13\x18\x9c\x16\xa0\x17\xd3\x13\x29\x88\x44\xd8\x0b\x0f\x90\x34\x8a\x3c\xd8\x0c\x12\xf0\x02\x06\x05\x12\xdc\x08\x0d\x88\x64\x90\x44\xd5\x08\x19\xf8\xf4\x13\x00\x10\x1f\xf2\x00\x02\x09\x11\xe1\x0c\x10\xf0\x05\x02\x09\x11\xfb\xf4\x14\x00\x0c\x13\xf2\x00\x04\x05\x12\xf1\x06\x00\x10\x18\x9c\x74\x9f\x7a\x99\x7a\xa8\x24\xd4\x1f\x2f\xd8\x0c\x11\xf1\x03\x00\x20\x30\xf0\x07\x04\x05\x12\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
os_toplevel_consts_79_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x0d\x0d\x42\x14\x00\xc2\x07\x0c\x42\x23\x00\xc2\x14\x09\x42\x20\x03\xc2\x1f\x01\x42\x20\x03\xc2\x23\x21\x43\x07\x03\xc3\x06\x01\x43\x07\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_cdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cdir",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
os_toplevel_consts_79_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(mode),
            & const_str_exist_ok._ascii.ob_base,
            & const_str_head._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
            & const_str_cdir._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(404)
os_toplevel_consts_79 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 202,
    },
    .co_consts = & os_toplevel_consts_79_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_79_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_79_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 200,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 608,
    .co_localsplusnames = & os_toplevel_consts_79_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_makedirs._ascii.ob_base,
    .co_qualname = & const_str_makedirs._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_79_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x7c\x04\x73\x18\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x7c\x03\x72\x51\x7c\x04\x72\x4f\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x73\x3a\x09\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x02\xac\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x10\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x04\x7c\x05\x6b\x28\x00\x00\x72\x01\x79\x03\x09\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x03\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x45\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x1b\x01\x00\x7c\x02\x72\x15\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x82\x00\x59\x00\x79\x03\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[429];
    }
os_toplevel_consts_80_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 428,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x72\x65\x6d\x6f\x76\x65\x64\x69\x72\x73\x28\x6e\x61\x6d\x65\x29\x0a\x0a\x20\x20\x20\x20\x53\x75\x70\x65\x72\x2d\x72\x6d\x64\x69\x72\x3b\x20\x72\x65\x6d\x6f\x76\x65\x20\x61\x20\x6c\x65\x61\x66\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x61\x6e\x64\x20\x61\x6c\x6c\x20\x65\x6d\x70\x74\x79\x20\x69\x6e\x74\x65\x72\x6d\x65\x64\x69\x61\x74\x65\x0a\x20\x20\x20\x20\x6f\x6e\x65\x73\x2e\x20\x20\x57\x6f\x72\x6b\x73\x20\x6c\x69\x6b\x65\x20\x72\x6d\x64\x69\x72\x20\x65\x78\x63\x65\x70\x74\x20\x74\x68\x61\x74\x2c\x20\x69\x66\x20\x74\x68\x65\x20\x6c\x65\x61\x66\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x0a\x20\x20\x20\x20\x73\x75\x63\x63\x65\x73\x73\x66\x75\x6c\x6c\x79\x20\x72\x65\x6d\x6f\x76\x65\x64\x2c\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x63\x6f\x72\x72\x65\x73\x70\x6f\x6e\x64\x69\x6e\x67\x20\x74\x6f\x20\x72\x69\x67\x68\x74\x6d\x6f\x73\x74\x20\x70\x61\x74\x68\x0a\x20\x20\x20\x20\x73\x65\x67\x6d\x65\x6e\x74\x73\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x70\x72\x75\x6e\x65\x64\x20\x61\x77\x61\x79\x20\x75\x6e\x74\x69\x6c\x20\x65\x69\x74\x68\x65\x72\x20\x74\x68\x65\x20\x77\x68\x6f\x6c\x65\x20\x70\x61\x74\x68\x20\x69\x73\x0a\x20\x20\x20\x20\x63\x6f\x6e\x73\x75\x6d\x65\x64\x20\x6f\x72\x20\x61\x6e\x20\x65\x72\x72\x6f\x72\x20\x6f\x63\x63\x75\x72\x73\x2e\x20\x20\x45\x72\x72\x6f\x72\x73\x20\x64\x75\x72\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x6c\x61\x74\x74\x65\x72\x20\x70\x68\x61\x73\x65\x20\x61\x72\x65\x0a\x20\x20\x20\x20\x69\x67\x6e\x6f\x72\x65\x64\x20\x2d\x2d\x20\x74\x68\x65\x79\x20\x67\x65\x6e\x65\x72\x61\x6c\x6c\x79\x20\x6d\x65\x61\x6e\x20\x74\x68\x61\x74\x20\x61\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x77\x61\x73\x20\x6e\x6f\x74\x20\x65\x6d\x70\x74\x79\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_80_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_80_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_80_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_rmdir._ascii.ob_base,
            &_Py_ID(path),
            & const_str_split._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_removedirs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "removedirs",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[121];
    }
os_toplevel_consts_80_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 120,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x16\x00\x05\x0a\x88\x24\x84\x4b\xdc\x11\x15\x97\x1a\x91\x1a\x98\x44\xd3\x11\x21\x81\x4a\x80\x44\x88\x24\xd9\x0b\x0f\xdc\x15\x19\x97\x5a\x91\x5a\xa0\x04\xd3\x15\x25\x89\x0a\x88\x04\x88\x64\xd9\x0a\x0e\x91\x34\xf0\x02\x03\x09\x12\xdc\x0c\x11\x90\x24\x8c\x4b\xf4\x06\x00\x16\x1a\x97\x5a\x91\x5a\xa0\x04\xd3\x15\x25\x89\x0a\x88\x04\x88\x64\xf1\x0b\x00\x0b\x0f\x93\x34\x88\x24\x90\x34\x88\x24\xf8\xf4\x06\x00\x10\x17\xf2\x00\x01\x09\x12\xd9\x0c\x11\xf0\x03\x01\x09\x12\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
os_toplevel_consts_80_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x03\x0b\x41\x2f\x00\xc1\x2f\x09\x41\x3b\x03\xc1\x3a\x01\x41\x3b\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_80_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_head._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(252)
os_toplevel_consts_80 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 126,
    },
    .co_consts = & os_toplevel_consts_80_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_80_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_80_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 232,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 609,
    .co_localsplusnames = & os_toplevel_consts_80_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_removedirs._ascii.ob_base,
    .co_qualname = & const_str_removedirs._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_80_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x7c\x02\x73\x18\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x7c\x01\x72\x2e\x7c\x02\x72\x2b\x09\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x7c\x01\x72\x04\x7c\x02\x72\x01\x8c\x29\x79\x01\x79\x01\x79\x01\x79\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[573];
    }
os_toplevel_consts_81_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 572,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x72\x65\x6e\x61\x6d\x65\x73\x28\x6f\x6c\x64\x2c\x20\x6e\x65\x77\x29\x0a\x0a\x20\x20\x20\x20\x53\x75\x70\x65\x72\x2d\x72\x65\x6e\x61\x6d\x65\x3b\x20\x63\x72\x65\x61\x74\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x61\x73\x20\x6e\x65\x63\x65\x73\x73\x61\x72\x79\x20\x61\x6e\x64\x20\x64\x65\x6c\x65\x74\x65\x20\x61\x6e\x79\x20\x6c\x65\x66\x74\x0a\x20\x20\x20\x20\x65\x6d\x70\x74\x79\x2e\x20\x20\x57\x6f\x72\x6b\x73\x20\x6c\x69\x6b\x65\x20\x72\x65\x6e\x61\x6d\x65\x2c\x20\x65\x78\x63\x65\x70\x74\x20\x63\x72\x65\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x61\x6e\x79\x20\x69\x6e\x74\x65\x72\x6d\x65\x64\x69\x61\x74\x65\x0a\x20\x20\x20\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x6e\x65\x65\x64\x65\x64\x20\x74\x6f\x20\x6d\x61\x6b\x65\x20\x74\x68\x65\x20\x6e\x65\x77\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x67\x6f\x6f\x64\x20\x69\x73\x20\x61\x74\x74\x65\x6d\x70\x74\x65\x64\x0a\x20\x20\x20\x20\x66\x69\x72\x73\x74\x2e\x20\x20\x41\x66\x74\x65\x72\x20\x74\x68\x65\x20\x72\x65\x6e\x61\x6d\x65\x2c\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x63\x6f\x72\x72\x65\x73\x70\x6f\x6e\x64\x69\x6e\x67\x20\x74\x6f\x20\x72\x69\x67\x68\x74\x6d\x6f\x73\x74\x0a\x20\x20\x20\x20\x70\x61\x74\x68\x20\x73\x65\x67\x6d\x65\x6e\x74\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x6f\x6c\x64\x20\x6e\x61\x6d\x65\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x70\x72\x75\x6e\x65\x64\x20\x75\x6e\x74\x69\x6c\x20\x65\x69\x74\x68\x65\x72\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x77\x68\x6f\x6c\x65\x20\x70\x61\x74\x68\x20\x69\x73\x20\x63\x6f\x6e\x73\x75\x6d\x65\x64\x20\x6f\x72\x20\x61\x20\x6e\x6f\x6e\x65\x6d\x70\x74\x79\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x0a\x20\x20\x20\x20\x4e\x6f\x74\x65\x3a\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x63\x61\x6e\x20\x66\x61\x69\x6c\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x6e\x65\x77\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x73\x74\x72\x75\x63\x74\x75\x72\x65\x20\x6d\x61\x64\x65\x0a\x20\x20\x20\x20\x69\x66\x20\x79\x6f\x75\x20\x6c\x61\x63\x6b\x20\x70\x65\x72\x6d\x69\x73\x73\x69\x6f\x6e\x73\x20\x6e\x65\x65\x64\x65\x64\x20\x74\x6f\x20\x75\x6e\x6c\x69\x6e\x6b\x20\x74\x68\x65\x20\x6c\x65\x61\x66\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x6f\x72\x0a\x20\x20\x20\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_81_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_81_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
os_toplevel_consts_81_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_split._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str_makedirs._ascii.ob_base,
            & const_str_rename._ascii.ob_base,
            & const_str_removedirs._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_renames = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "renames",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[117];
    }
os_toplevel_consts_81_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 116,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x1e\x00\x12\x16\x97\x1a\x91\x1a\x98\x43\x93\x1f\x81\x4a\x80\x44\x88\x24\xd9\x07\x0b\x91\x04\x9c\x54\x9f\x5b\x99\x5b\xa8\x14\xd4\x1d\x2e\xdc\x08\x10\x90\x14\x8c\x0e\xdc\x04\x0a\x88\x33\x90\x03\xd4\x04\x14\xdc\x11\x15\x97\x1a\x91\x1a\x98\x43\x93\x1f\x81\x4a\x80\x44\x88\x24\xd9\x07\x0b\x91\x04\xf0\x02\x03\x09\x11\xdc\x0c\x16\x90\x74\xd5\x0c\x1c\xf0\x05\x00\x11\x15\x80\x74\xf8\xf4\x06\x00\x10\x17\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
os_toplevel_consts_81_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x26\x0b\x41\x34\x00\xc1\x34\x09\x42\x00\x03\xc1\x3f\x01\x42\x00\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_81_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_old._ascii.ob_base,
            & const_str_new._ascii.ob_base,
            & const_str_head._ascii.ob_base,
            & const_str_tail._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(262)
os_toplevel_consts_81 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 131,
    },
    .co_consts = & os_toplevel_consts_81_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_81_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_81_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 254,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 610,
    .co_localsplusnames = & os_toplevel_consts_81_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_renames._ascii.ob_base,
    .co_qualname = & const_str_renames._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_81_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x02\x72\x22\x7c\x03\x72\x20\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x73\x0b\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x02\x72\x10\x7c\x03\x72\x0d\x09\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01\x79\x01\x79\x01\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_82 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_makedirs._ascii.ob_base,
            & const_str_removedirs._ascii.ob_base,
            & const_str_renames._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[2877];
    }
os_toplevel_consts_83_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2876,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x74\x72\x65\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x0a\x20\x20\x20\x20\x46\x6f\x72\x20\x65\x61\x63\x68\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x6e\x20\x74\x68\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x74\x72\x65\x65\x20\x72\x6f\x6f\x74\x65\x64\x20\x61\x74\x20\x74\x6f\x70\x20\x28\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x74\x6f\x70\x0a\x20\x20\x20\x20\x69\x74\x73\x65\x6c\x66\x2c\x20\x62\x75\x74\x20\x65\x78\x63\x6c\x75\x64\x69\x6e\x67\x20\x27\x2e\x27\x20\x61\x6e\x64\x20\x27\x2e\x2e\x27\x29\x2c\x20\x79\x69\x65\x6c\x64\x73\x20\x61\x20\x33\x2d\x74\x75\x70\x6c\x65\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x64\x69\x72\x70\x61\x74\x68\x2c\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x2c\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x73\x0a\x0a\x20\x20\x20\x20\x64\x69\x72\x70\x61\x74\x68\x20\x69\x73\x20\x61\x20\x73\x74\x72\x69\x6e\x67\x2c\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x74\x68\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x2e\x20\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x20\x69\x73\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x0a\x20\x20\x20\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x69\x6e\x20\x64\x69\x72\x70\x61\x74\x68\x20\x28\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x73\x79\x6d\x6c\x69\x6e\x6b\x73\x20\x74\x6f\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x2c\x0a\x20\x20\x20\x20\x61\x6e\x64\x20\x65\x78\x63\x6c\x75\x64\x69\x6e\x67\x20\x27\x2e\x27\x20\x61\x6e\x64\x20\x27\x2e\x2e\x27\x29\x2e\x0a\x20\x20\x20\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x73\x20\x69\x73\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x6e\x6f\x6e\x2d\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x66\x69\x6c\x65\x73\x20\x69\x6e\x20\x64\x69\x72\x70\x61\x74\x68\x2e\x0a\x20\x20\x20\x20\x4e\x6f\x74\x65\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x73\x20\x69\x6e\x20\x74\x68\x65\x20\x6c\x69\x73\x74\x73\x20\x61\x72\x65\x20\x6a\x75\x73\x74\x20\x6e\x61\x6d\x65\x73\x2c\x20\x77\x69\x74\x68\x20\x6e\x6f\x20\x70\x61\x74\x68\x20\x63\x6f\x6d\x70\x6f\x6e\x65\x6e\x74\x73\x2e\x0a\x20\x20\x20\x20\x54\x6f\x20\x67\x65\x74\x20\x61\x20\x66\x75\x6c\x6c\x20\x70\x61\x74\x68\x20\x28\x77\x68\x69\x63\x68\x20\x62\x65\x67\x69\x6e\x73\x20\x77\x69\x74\x68\x20\x74\x6f\x70\x29\x20\x74\x6f\x20\x61\x20\x66\x69\x6c\x65\x20\x6f\x72\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x6e\x0a\x20\x20\x20\x20\x64\x69\x72\x70\x61\x74\x68\x2c\x20\x64\x6f\x20\x6f\x73\x2e\x70\x61\x74\x68\x2e\x6a\x6f\x69\x6e\x28\x64\x69\x72\x70\x61\x74\x68\x2c\x20\x6e\x61\x6d\x65\x29\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x61\x72\x67\x20\x27\x74\x6f\x70\x64\x6f\x77\x6e\x27\x20\x69\x73\x20\x74\x72\x75\x65\x20\x6f\x72\x20\x6e\x6f\x74\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x2c\x20\x74\x68\x65\x20\x74\x72\x69\x70\x6c\x65\x20\x66\x6f\x72\x20\x61\x0a\x20\x20\x20\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x67\x65\x6e\x65\x72\x61\x74\x65\x64\x20\x62\x65\x66\x6f\x72\x65\x20\x74\x68\x65\x20\x74\x72\x69\x70\x6c\x65\x73\x20\x66\x6f\x72\x20\x61\x6e\x79\x20\x6f\x66\x20\x69\x74\x73\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x0a\x20\x20\x20\x20\x28\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x61\x72\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x65\x64\x20\x74\x6f\x70\x20\x64\x6f\x77\x6e\x29\x2e\x20\x20\x49\x66\x20\x74\x6f\x70\x64\x6f\x77\x6e\x20\x69\x73\x20\x66\x61\x6c\x73\x65\x2c\x20\x74\x68\x65\x20\x74\x72\x69\x70\x6c\x65\x0a\x20\x20\x20\x20\x66\x6f\x72\x20\x61\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x67\x65\x6e\x65\x72\x61\x74\x65\x64\x20\x61\x66\x74\x65\x72\x20\x74\x68\x65\x20\x74\x72\x69\x70\x6c\x65\x73\x20\x66\x6f\x72\x20\x61\x6c\x6c\x20\x6f\x66\x20\x69\x74\x73\x0a\x20\x20\x20\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x28\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x61\x72\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x65\x64\x20\x62\x6f\x74\x74\x6f\x6d\x20\x75\x70\x29\x2e\x0a\x0a\x20\x20\x20\x20\x57\x68\x65\x6e\x20\x74\x6f\x70\x64\x6f\x77\x6e\x20\x69\x73\x20\x74\x72\x75\x65\x2c\x20\x74\x68\x65\x20\x63\x61\x6c\x6c\x65\x72\x20\x63\x61\x6e\x20\x6d\x6f\x64\x69\x66\x79\x20\x74\x68\x65\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x20\x6c\x69\x73\x74\x20\x69\x6e\x2d\x70\x6c\x61\x63\x65\x0a\x20\x20\x20\x20\x28\x65\x2e\x67\x2e\x2c\x20\x76\x69\x61\x20\x64\x65\x6c\x20\x6f\x72\x20\x73\x6c\x69\x63\x65\x20\x61\x73\x73\x69\x67\x6e\x6d\x65\x6e\x74\x29\x2c\x20\x61\x6e\x64\x20\x77\x61\x6c\x6b\x20\x77\x69\x6c\x6c\x20\x6f\x6e\x6c\x79\x20\x72\x65\x63\x75\x72\x73\x65\x20\x69\x6e\x74\x6f\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x77\x68\x6f\x73\x65\x20\x6e\x61\x6d\x65\x73\x20\x72\x65\x6d\x61\x69\x6e\x20\x69\x6e\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x3b\x20\x74\x68\x69\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x70\x72\x75\x6e\x65\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x73\x65\x61\x72\x63\x68\x2c\x20\x6f\x72\x20\x74\x6f\x20\x69\x6d\x70\x6f\x73\x65\x20\x61\x20\x73\x70\x65\x63\x69\x66\x69\x63\x20\x6f\x72\x64\x65\x72\x20\x6f\x66\x20\x76\x69\x73\x69\x74\x69\x6e\x67\x2e\x20\x20\x4d\x6f\x64\x69\x66\x79\x69\x6e\x67\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x20\x77\x68\x65\x6e\x0a\x20\x20\x20\x20\x74\x6f\x70\x64\x6f\x77\x6e\x20\x69\x73\x20\x66\x61\x6c\x73\x65\x20\x68\x61\x73\x20\x6e\x6f\x20\x65\x66\x66\x65\x63\x74\x20\x6f\x6e\x20\x74\x68\x65\x20\x62\x65\x68\x61\x76\x69\x6f\x72\x20\x6f\x66\x20\x6f\x73\x2e\x77\x61\x6c\x6b\x28\x29\x2c\x20\x73\x69\x6e\x63\x65\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x69\x6e\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x20\x68\x61\x76\x65\x20\x61\x6c\x72\x65\x61\x64\x79\x20\x62\x65\x65\x6e\x20\x67\x65\x6e\x65\x72\x61\x74\x65\x64\x20\x62\x79\x20\x74\x68\x65\x20\x74\x69\x6d\x65\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x0a\x20\x20\x20\x20\x69\x74\x73\x65\x6c\x66\x20\x69\x73\x20\x67\x65\x6e\x65\x72\x61\x74\x65\x64\x2e\x20\x4e\x6f\x20\x6d\x61\x74\x74\x65\x72\x20\x74\x68\x65\x20\x76\x61\x6c\x75\x65\x20\x6f\x66\x20\x74\x6f\x70\x64\x6f\x77\x6e\x2c\x20\x74\x68\x65\x20\x6c\x69\x73\x74\x20\x6f\x66\x0a\x20\x20\x20\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x69\x73\x20\x72\x65\x74\x72\x69\x65\x76\x65\x64\x20\x62\x65\x66\x6f\x72\x65\x20\x74\x68\x65\x20\x74\x75\x70\x6c\x65\x73\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x61\x6e\x64\x20\x69\x74\x73\x0a\x20\x20\x20\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x61\x72\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x42\x79\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x65\x72\x72\x6f\x72\x73\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x6f\x73\x2e\x73\x63\x61\x6e\x64\x69\x72\x28\x29\x20\x63\x61\x6c\x6c\x20\x61\x72\x65\x20\x69\x67\x6e\x6f\x72\x65\x64\x2e\x20\x20\x49\x66\x0a\x20\x20\x20\x20\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x61\x72\x67\x20\x27\x6f\x6e\x65\x72\x72\x6f\x72\x27\x20\x69\x73\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x2c\x20\x69\x74\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x20\x61\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x3b\x20\x69\x74\x0a\x20\x20\x20\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x63\x61\x6c\x6c\x65\x64\x20\x77\x69\x74\x68\x20\x6f\x6e\x65\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x2c\x20\x61\x6e\x20\x4f\x53\x45\x72\x72\x6f\x72\x20\x69\x6e\x73\x74\x61\x6e\x63\x65\x2e\x20\x20\x49\x74\x20\x63\x61\x6e\x0a\x20\x20\x20\x20\x72\x65\x70\x6f\x72\x74\x20\x74\x68\x65\x20\x65\x72\x72\x6f\x72\x20\x74\x6f\x20\x63\x6f\x6e\x74\x69\x6e\x75\x65\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x77\x61\x6c\x6b\x2c\x20\x6f\x72\x20\x72\x61\x69\x73\x65\x20\x74\x68\x65\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x0a\x20\x20\x20\x20\x74\x6f\x20\x61\x62\x6f\x72\x74\x20\x74\x68\x65\x20\x77\x61\x6c\x6b\x2e\x20\x20\x4e\x6f\x74\x65\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x69\x73\x20\x61\x76\x61\x69\x6c\x61\x62\x6c\x65\x20\x61\x73\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x61\x74\x74\x72\x69\x62\x75\x74\x65\x20\x6f\x66\x20\x74\x68\x65\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x0a\x20\x20\x20\x20\x42\x79\x20\x64\x65\x66\x61\x75\x6c\x74\x2c\x20\x6f\x73\x2e\x77\x61\x6c\x6b\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x20\x66\x6f\x6c\x6c\x6f\x77\x20\x73\x79\x6d\x62\x6f\x6c\x69\x63\x20\x6c\x69\x6e\x6b\x73\x20\x74\x6f\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x6f\x6e\x0a\x20\x20\x20\x20\x73\x79\x73\x74\x65\x6d\x73\x20\x74\x68\x61\x74\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x74\x68\x65\x6d\x2e\x20\x20\x49\x6e\x20\x6f\x72\x64\x65\x72\x20\x74\x6f\x20\x67\x65\x74\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x61\x6c\x69\x74\x79\x2c\x20\x73\x65\x74\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x27\x66\x6f\x6c\x6c\x6f\x77\x6c\x69\x6e\x6b\x73\x27\x20\x74\x6f\x20\x74\x72\x75\x65\x2e\x0a\x0a\x20\x20\x20\x20\x43\x61\x75\x74\x69\x6f\x6e\x3a\x20\x20\x69\x66\x20\x79\x6f\x75\x20\x70\x61\x73\x73\x20\x61\x20\x72\x65\x6c\x61\x74\x69\x76\x65\x20\x70\x61\x74\x68\x6e\x61\x6d\x65\x20\x66\x6f\x72\x20\x74\x6f\x70\x2c\x20\x64\x6f\x6e\x27\x74\x20\x63\x68\x61\x6e\x67\x65\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x77\x6f\x72\x6b\x69\x6e\x67\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x62\x65\x74\x77\x65\x65\x6e\x20\x72\x65\x73\x75\x6d\x70\x74\x69\x6f\x6e\x73\x20\x6f\x66\x20\x77\x61\x6c\x6b\x2e\x20\x20\x77\x61\x6c\x6b\x20\x6e\x65\x76\x65\x72\x0a\x20\x20\x20\x20\x63\x68\x61\x6e\x67\x65\x73\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x2c\x20\x61\x6e\x64\x20\x61\x73\x73\x75\x6d\x65\x73\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x63\x6c\x69\x65\x6e\x74\x20\x64\x6f\x65\x73\x6e\x27\x74\x0a\x20\x20\x20\x20\x65\x69\x74\x68\x65\x72\x2e\x0a\x0a\x20\x20\x20\x20\x45\x78\x61\x6d\x70\x6c\x65\x3a\x0a\x0a\x20\x20\x20\x20\x69\x6d\x70\x6f\x72\x74\x20\x6f\x73\x0a\x20\x20\x20\x20\x66\x72\x6f\x6d\x20\x6f\x73\x2e\x70\x61\x74\x68\x20\x69\x6d\x70\x6f\x72\x74\x20\x6a\x6f\x69\x6e\x2c\x20\x67\x65\x74\x73\x69\x7a\x65\x0a\x20\x20\x20\x20\x66\x6f\x72\x20\x72\x6f\x6f\x74\x2c\x20\x64\x69\x72\x73\x2c\x20\x66\x69\x6c\x65\x73\x20\x69\x6e\x20\x6f\x73\x2e\x77\x61\x6c\x6b\x28\x27\x70\x79\x74\x68\x6f\x6e\x2f\x4c\x69\x62\x2f\x78\x6d\x6c\x27\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x69\x6e\x74\x28\x72\x6f\x6f\x74\x2c\x20\x22\x63\x6f\x6e\x73\x75\x6d\x65\x73\x20\x22\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x69\x6e\x74\x28\x73\x75\x6d\x28\x67\x65\x74\x73\x69\x7a\x65\x28\x6a\x6f\x69\x6e\x28\x72\x6f\x6f\x74\x2c\x20\x6e\x61\x6d\x65\x29\x29\x20\x66\x6f\x72\x20\x6e\x61\x6d\x65\x20\x69\x6e\x20\x66\x69\x6c\x65\x73\x29\x2c\x20\x65\x6e\x64\x3d\x22\x20\x22\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x69\x6e\x74\x28\x22\x62\x79\x74\x65\x73\x20\x69\x6e\x22\x2c\x20\x6c\x65\x6e\x28\x66\x69\x6c\x65\x73\x29\x2c\x20\x22\x6e\x6f\x6e\x2d\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x66\x69\x6c\x65\x73\x22\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x66\x20\x27\x5f\x5f\x70\x79\x63\x61\x63\x68\x65\x5f\x5f\x27\x20\x69\x6e\x20\x64\x69\x72\x73\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x69\x72\x73\x2e\x72\x65\x6d\x6f\x76\x65\x28\x27\x5f\x5f\x70\x79\x63\x61\x63\x68\x65\x5f\x5f\x27\x29\x20\x20\x23\x20\x64\x6f\x6e\x27\x74\x20\x76\x69\x73\x69\x74\x20\x5f\x5f\x70\x79\x63\x61\x63\x68\x65\x5f\x5f\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
os_toplevel_consts_83_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "os.walk",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_83_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(follow_symlinks),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
os_toplevel_consts_83_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & os_toplevel_consts_83_consts_0._ascii.ob_base,
            & os_toplevel_consts_83_consts_1._ascii.ob_base,
            Py_None,
            Py_False,
            Py_True,
            & os_toplevel_consts_83_consts_5._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_audit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "audit",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str__walk_symlinks_as_files = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_walk_symlinks_as_files",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_is_dir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_dir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_is_junction = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_junction",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_is_symlink = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_symlink",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[20];
        }_object;
    }
os_toplevel_consts_83_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 20,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_audit._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(path),
            & const_str_islink._ascii.ob_base,
            &_Py_ID(join),
            & const_str_pop._ascii.ob_base,
            &_Py_ID(isinstance),
            & const_str_tuple._ascii.ob_base,
            & const_str_scandir._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            &_Py_ID(next),
            & const_str_StopIteration._ascii.ob_base,
            & const_str__walk_symlinks_as_files._ascii.ob_base,
            & const_str_is_dir._ascii.ob_base,
            & const_str_is_junction._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(name),
            & const_str_is_symlink._ascii.ob_base,
            &_Py_ID(reversed),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_walk = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "walk",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[598];
    }
os_toplevel_consts_83_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 597,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf4\x78\x01\x00\x05\x08\x87\x49\x81\x49\x88\x69\x98\x13\x98\x67\xa0\x77\xb0\x0b\xd4\x04\x3c\xe4\x0d\x13\x90\x43\x8b\x5b\x88\x4d\x80\x45\xdc\x13\x17\x97\x3b\x91\x3b\xa4\x04\xa7\x09\xa1\x09\x88\x44\x80\x46\xd9\x0a\x0f\xd8\x0e\x13\x8f\x69\x89\x69\x8b\x6b\x88\x03\xdc\x0b\x15\x90\x63\x9c\x35\xd4\x0b\x21\xd8\x12\x15\x8a\x49\xd8\x0c\x14\xe0\x0f\x11\x88\x04\xd8\x12\x14\x88\x07\xd8\x14\x16\x88\x09\xf0\x0e\x05\x09\x15\xdc\x19\x20\xa0\x13\x9b\x1c\x88\x4a\xf0\x0c\x00\x10\x15\x88\x04\xda\x0d\x17\xd8\x12\x16\xf0\x02\x09\x11\x1a\xf0\x02\x03\x15\x1e\xdc\x20\x24\xa0\x5a\xd3\x20\x30\x99\x05\xf0\x12\x08\x11\x23\xd8\x17\x22\xd4\x26\x3d\xd1\x17\x3d\xd8\x21\x26\xa7\x1c\xa1\x1c\xb8\x65\xa0\x1c\xd3\x21\x44\xd2\x21\x60\xc8\x55\xd7\x4d\x5e\xd1\x4d\x5e\xd3\x4d\x60\xd0\x49\x60\x99\x06\xe0\x21\x26\xa7\x1c\xa1\x1c\xa3\x1e\x98\x06\xf1\x0c\x00\x14\x1a\xd8\x14\x18\x97\x4b\x91\x4b\xa0\x05\xa7\x0a\xa1\x0a\xd5\x14\x2b\xe0\x14\x1b\x97\x4e\x91\x4e\xa0\x35\xa7\x3a\xa1\x3a\xd4\x14\x2e\xe1\x17\x1e\xa1\x36\xf1\x06\x00\x18\x23\xd8\x24\x28\x99\x09\xf0\x04\x06\x19\x2f\xd8\x29\x2e\xd7\x29\x39\xd1\x29\x39\xd3\x29\x3b\x98\x4a\xf0\x0c\x00\x29\x33\xa0\x4e\x98\x09\xe1\x17\x20\xd8\x18\x21\xd7\x18\x28\xd1\x18\x28\xa8\x15\xaf\x1a\xa9\x1a\xd4\x18\x34\xf0\x57\x01\x00\x13\x17\xf8\xf0\x31\x00\x0b\x10\xf8\xf4\x22\x00\x10\x17\xf2\x00\x03\x09\x15\xd8\x0f\x16\xd0\x0f\x22\xd9\x10\x17\x98\x05\x94\x0e\xdd\x0c\x14\xfb\xf0\x07\x03\x09\x15\xfb\xf4\x16\x00\x1c\x29\xf2\x00\x01\x15\x1e\xd9\x18\x1d\xf0\x03\x01\x15\x1e\xfb\xe4\x17\x1e\xf2\x00\x04\x11\x1a\xd8\x17\x1e\xd0\x17\x2a\xd9\x18\x1f\xa0\x05\x9c\x0e\xd8\x1b\x1f\x90\x44\xdc\x14\x19\xfb\xf0\x09\x04\x11\x1a\xfb\xf4\x16\x00\x18\x1f\xf2\x00\x03\x11\x23\xf0\x06\x00\x1e\x23\x92\x46\xf0\x07\x03\x11\x23\xfb\xf4\x24\x00\x20\x27\xf2\x00\x04\x19\x2f\xf0\x08\x00\x2a\x2f\x9a\x4a\xf0\x09\x04\x19\x2f\xfa\xf7\x49\x01\x00\x0e\x18\x8f\x5a\x89\x5a\xfa\xf1\x5a\x01\x00\x0c\x10\xd9\x0c\x14\xe1\x0b\x12\xe0\x12\x15\x90\x74\x98\x57\xd0\x12\x24\xd2\x0c\x24\xe4\x1b\x23\xa0\x44\x9e\x3e\x90\x07\xd9\x1b\x1f\xa0\x03\xa0\x57\xd3\x1b\x2d\x90\x08\xf1\x0a\x00\x14\x1f\xa1\x66\xa8\x58\xd5\x26\x36\xd8\x14\x19\x97\x4c\x91\x4c\xa0\x18\xd5\x14\x2a\xf1\x0f\x00\x1c\x2a\xf0\x14\x00\x0d\x12\x8f\x4c\x89\x4c\x98\x23\x98\x74\xa0\x57\xd0\x19\x2d\xd4\x0c\x2e\xe4\x1c\x24\xa0\x59\xd6\x1c\x2f\x90\x08\xd8\x10\x15\x97\x0c\x91\x0c\x98\x58\xd5\x10\x26\xf0\x03\x00\x1d\x30\xf3\x6f\x02\x00\x0b\x10\xfb",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[231];
    }
os_toplevel_consts_83_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 230,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x41\x33\x49\x11\x01\xc1\x36\x0b\x45\x0a\x00\xc2\x01\x04\x49\x11\x01\xc2\x05\x02\x47\x07\x03\xc2\x09\x0b\x45\x2b\x02\xc2\x14\x01\x47\x07\x03\xc2\x16\x3e\x46\x1c\x02\xc3\x14\x41\x02\x47\x07\x03\xc4\x17\x10\x46\x2d\x02\xc4\x27\x21\x47\x07\x03\xc5\x08\x02\x49\x11\x01\xc5\x0a\x09\x45\x28\x03\xc5\x13\x0a\x45\x23\x03\xc5\x1d\x06\x49\x11\x01\xc5\x23\x05\x45\x28\x03\xc5\x28\x03\x49\x11\x01\xc5\x2b\x09\x45\x37\x05\xc5\x34\x01\x45\x3a\x02\xc5\x35\x01\x47\x07\x03\xc5\x36\x01\x45\x37\x05\xc5\x37\x03\x45\x3a\x02\xc5\x3a\x09\x46\x19\x05\xc6\x03\x0c\x46\x14\x05\xc6\x0f\x05\x47\x07\x03\xc6\x14\x05\x46\x19\x05\xc6\x19\x03\x47\x07\x03\xc6\x1c\x0b\x46\x2a\x05\xc6\x27\x02\x47\x07\x03\xc6\x29\x01\x46\x2a\x05\xc6\x2a\x03\x47\x07\x03\xc6\x2d\x0b\x46\x3b\x05\xc6\x38\x02\x47\x07\x03\xc6\x3a\x01\x46\x3b\x05\xc6\x3b\x03\x47\x07\x03\xc6\x3e\x09\x49\x11\x01\xc7\x07\x05\x47\x10\x07\xc7\x0c\x35\x49\x11\x01\xc8\x02\x41\x0b\x49\x11\x01",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_topdown = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "topdown",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_onerror = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "onerror",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_followlinks = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "followlinks",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_stack = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "stack",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_nondirs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "nondirs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_walk_dirs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "walk_dirs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_scandir_it = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "scandir_it",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_error = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "error",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_cont = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cont",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_walk_into = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "walk_into",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
os_toplevel_consts_83_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            &_Py_ID(top),
            & const_str_topdown._ascii.ob_base,
            & const_str_onerror._ascii.ob_base,
            & const_str_followlinks._ascii.ob_base,
            & const_str_stack._ascii.ob_base,
            & const_str_islink._ascii.ob_base,
            &_Py_ID(join),
            & const_str_dirs._ascii.ob_base,
            & const_str_nondirs._ascii.ob_base,
            & const_str_walk_dirs._ascii.ob_base,
            & const_str_scandir_it._ascii.ob_base,
            & const_str_error._ascii.ob_base,
            & const_str_cont._ascii.ob_base,
            & const_str_entry._ascii.ob_base,
            & const_str_is_dir._ascii.ob_base,
            & const_str_walk_into._ascii.ob_base,
            & const_str_is_symlink._ascii.ob_base,
            & const_str_dirname._ascii.ob_base,
            & const_str_new_path._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(1190)
os_toplevel_consts_83 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 595,
    },
    .co_consts = & os_toplevel_consts_83_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_83_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_83_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 26 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 286,
    .co_nlocalsplus = 19,
    .co_nlocals = 19,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 611,
    .co_localsplusnames = & os_toplevel_consts_83_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & ntpath_toplevel_consts_46_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_walk._ascii.ob_base,
    .co_qualname = & const_str_walk._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_83_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x7c\x01\x7c\x02\x7c\x03\xab\x05\x00\x00\x00\x00\x00\x00\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x67\x01\x7d\x04\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x7d\x05\x7c\x04\x72\xff\x7c\x04\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x05\x7c\x00\x96\x01\x97\x01\x01\x00\x8c\x27\x67\x00\x7d\x07\x67\x00\x7d\x08\x67\x00\x7d\x09\x09\x00\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0a\x64\x03\x7d\x0c\x7c\x0a\x35\x00\x01\x00\x09\x00\x09\x00\x09\x00\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0d\x09\x00\x09\x00\x7c\x03\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x26\x7c\x0d\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xac\x05\xab\x01\x00\x00\x00\x00\x00\x00\x78\x01\x72\x11\x01\x00\x7c\x0d\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x0c\x00\x7d\x0e\x6e\x10\x7c\x0d\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x0e\x7c\x0e\x72\x1c\x7c\x07\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x1b\x7c\x08\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x73\x38\x7c\x0e\x72\x36\x7c\x03\x72\x03\x64\x04\x7d\x0f\x6e\x14\x09\x00\x7c\x0d\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x10\x7c\x10\x0c\x00\x7d\x0f\x7c\x0f\x72\x1b\x7c\x09\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\xc1\x79\x02\x79\x02\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x15\x7d\x0b\x7c\x02\x81\x08\x02\x00\x7c\x02\x7c\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x02\x7d\x0b\x7e\x0b\x90\x01\x8c\x1b\x64\x02\x7d\x0b\x7e\x0b\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x48\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x16\x7d\x0b\x7c\x02\x81\x08\x02\x00\x7c\x02\x7c\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x04\x7d\x0c\x59\x00\x64\x02\x7d\x0b\x7e\x0b\x6e\x2a\x64\x02\x7d\x0b\x7e\x0b\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x03\x7d\x0e\x59\x00\x8c\xd5\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x03\x7d\x10\x59\x00\x8c\x93\x77\x00\x78\x03\x59\x00\x77\x01\x64\x02\x64\x02\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x0c\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x6e\x03\x78\x03\x59\x00\x77\x01\x7c\x0c\x72\x02\x90\x01\x8c\x8f\x7c\x01\x72\x3d\x7c\x00\x7c\x07\x7c\x08\x66\x03\x96\x01\x97\x01\x01\x00\x74\x27\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x27\x00\x00\x7d\x11\x02\x00\x7c\x06\x7c\x00\x7c\x11\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x12\x7c\x03\x73\x09\x02\x00\x7c\x05\x7c\x12\xab\x01\x00\x00\x00\x00\x00\x00\x72\x01\x8c\x17\x7c\x04\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x12\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x29\x04\x00\x6e\x35\x7c\x04\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x07\x7c\x08\x66\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x27\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x13\x00\x00\x7d\x12\x7c\x04\x6a\x21\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x12\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x15\x04\x00\x7c\x04\x72\x02\x90\x02\x8c\x05\x90\x01\x8c\x09\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_85 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(follow_symlinks),
            &_Py_ID(dir_fd),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1305];
    }
os_toplevel_consts_86_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1304,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x74\x72\x65\x65\x20\x67\x65\x6e\x65\x72\x61\x74\x6f\x72\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x62\x65\x68\x61\x76\x65\x73\x20\x65\x78\x61\x63\x74\x6c\x79\x20\x6c\x69\x6b\x65\x20\x77\x61\x6c\x6b\x28\x29\x2c\x20\x65\x78\x63\x65\x70\x74\x20\x74\x68\x61\x74\x20\x69\x74\x20\x79\x69\x65\x6c\x64\x73\x20\x61\x20\x34\x2d\x74\x75\x70\x6c\x65\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x69\x72\x70\x61\x74\x68\x2c\x20\x64\x69\x72\x6e\x61\x6d\x65\x73\x2c\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x73\x2c\x20\x64\x69\x72\x66\x64\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x60\x64\x69\x72\x70\x61\x74\x68\x60\x2c\x20\x60\x64\x69\x72\x6e\x61\x6d\x65\x73\x60\x20\x61\x6e\x64\x20\x60\x66\x69\x6c\x65\x6e\x61\x6d\x65\x73\x60\x20\x61\x72\x65\x20\x69\x64\x65\x6e\x74\x69\x63\x61\x6c\x20\x74\x6f\x20\x77\x61\x6c\x6b\x28\x29\x20\x6f\x75\x74\x70\x75\x74\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x6e\x64\x20\x60\x64\x69\x72\x66\x64\x60\x20\x69\x73\x20\x61\x20\x66\x69\x6c\x65\x20\x64\x65\x73\x63\x72\x69\x70\x74\x6f\x72\x20\x72\x65\x66\x65\x72\x72\x69\x6e\x67\x20\x74\x6f\x20\x74\x68\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x60\x64\x69\x72\x70\x61\x74\x68\x60\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x61\x64\x76\x61\x6e\x74\x61\x67\x65\x20\x6f\x66\x20\x66\x77\x61\x6c\x6b\x28\x29\x20\x6f\x76\x65\x72\x20\x77\x61\x6c\x6b\x28\x29\x20\x69\x73\x20\x74\x68\x61\x74\x20\x69\x74\x27\x73\x20\x73\x61\x66\x65\x20\x61\x67\x61\x69\x6e\x73\x74\x20\x73\x79\x6d\x6c\x69\x6e\x6b\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x72\x61\x63\x65\x73\x20\x28\x77\x68\x65\x6e\x20\x66\x6f\x6c\x6c\x6f\x77\x5f\x73\x79\x6d\x6c\x69\x6e\x6b\x73\x20\x69\x73\x20\x46\x61\x6c\x73\x65\x29\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x49\x66\x20\x64\x69\x72\x5f\x66\x64\x20\x69\x73\x20\x6e\x6f\x74\x20\x4e\x6f\x6e\x65\x2c\x20\x69\x74\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x20\x61\x20\x66\x69\x6c\x65\x20\x64\x65\x73\x63\x72\x69\x70\x74\x6f\x72\x20\x6f\x70\x65\x6e\x20\x74\x6f\x20\x61\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x61\x6e\x64\x20\x74\x6f\x70\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x20\x72\x65\x6c\x61\x74\x69\x76\x65\x3b\x20\x74\x6f\x70\x20\x77\x69\x6c\x6c\x20\x74\x68\x65\x6e\x20\x62\x65\x20\x72\x65\x6c\x61\x74\x69\x76\x65\x20\x74\x6f\x20\x74\x68\x61\x74\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x28\x64\x69\x72\x5f\x66\x64\x20\x69\x73\x20\x61\x6c\x77\x61\x79\x73\x20\x73\x75\x70\x70\x6f\x72\x74\x65\x64\x20\x66\x6f\x72\x20\x66\x77\x61\x6c\x6b\x2e\x29\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x43\x61\x75\x74\x69\x6f\x6e\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x53\x69\x6e\x63\x65\x20\x66\x77\x61\x6c\x6b\x28\x29\x20\x79\x69\x65\x6c\x64\x73\x20\x66\x69\x6c\x65\x20\x64\x65\x73\x63\x72\x69\x70\x74\x6f\x72\x73\x2c\x20\x74\x68\x6f\x73\x65\x20\x61\x72\x65\x20\x6f\x6e\x6c\x79\x20\x76\x61\x6c\x69\x64\x20\x75\x6e\x74\x69\x6c\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6e\x65\x78\x74\x20\x69\x74\x65\x72\x61\x74\x69\x6f\x6e\x20\x73\x74\x65\x70\x2c\x20\x73\x6f\x20\x79\x6f\x75\x20\x73\x68\x6f\x75\x6c\x64\x20\x64\x75\x70\x28\x29\x20\x74\x68\x65\x6d\x20\x69\x66\x20\x79\x6f\x75\x20\x77\x61\x6e\x74\x20\x74\x6f\x20\x6b\x65\x65\x70\x20\x74\x68\x65\x6d\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x6f\x72\x20\x61\x20\x6c\x6f\x6e\x67\x65\x72\x20\x70\x65\x72\x69\x6f\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x45\x78\x61\x6d\x70\x6c\x65\x3a\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x69\x6d\x70\x6f\x72\x74\x20\x6f\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x66\x6f\x72\x20\x72\x6f\x6f\x74\x2c\x20\x64\x69\x72\x73\x2c\x20\x66\x69\x6c\x65\x73\x2c\x20\x72\x6f\x6f\x74\x66\x64\x20\x69\x6e\x20\x6f\x73\x2e\x66\x77\x61\x6c\x6b\x28\x27\x70\x79\x74\x68\x6f\x6e\x2f\x4c\x69\x62\x2f\x78\x6d\x6c\x27\x29\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x69\x6e\x74\x28\x72\x6f\x6f\x74\x2c\x20\x22\x63\x6f\x6e\x73\x75\x6d\x65\x73\x22\x2c\x20\x65\x6e\x64\x3d\x22\x22\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x69\x6e\x74\x28\x73\x75\x6d\x28\x6f\x73\x2e\x73\x74\x61\x74\x28\x6e\x61\x6d\x65\x2c\x20\x64\x69\x72\x5f\x66\x64\x3d\x72\x6f\x6f\x74\x66\x64\x29\x2e\x73\x74\x5f\x73\x69\x7a\x65\x20\x66\x6f\x72\x20\x6e\x61\x6d\x65\x20\x69\x6e\x20\x66\x69\x6c\x65\x73\x29\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x64\x3d\x22\x22\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x70\x72\x69\x6e\x74\x28\x22\x62\x79\x74\x65\x73\x20\x69\x6e\x22\x2c\x20\x6c\x65\x6e\x28\x66\x69\x6c\x65\x73\x29\x2c\x20\x22\x6e\x6f\x6e\x2d\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x66\x69\x6c\x65\x73\x22\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x69\x66\x20\x27\x5f\x5f\x70\x79\x63\x61\x63\x68\x65\x5f\x5f\x27\x20\x69\x6e\x20\x64\x69\x72\x73\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x64\x69\x72\x73\x2e\x72\x65\x6d\x6f\x76\x65\x28\x27\x5f\x5f\x70\x79\x63\x61\x63\x68\x65\x5f\x5f\x27\x29\x20\x20\x23\x20\x64\x6f\x6e\x27\x74\x20\x76\x69\x73\x69\x74\x20\x5f\x5f\x70\x79\x63\x61\x63\x68\x65\x5f\x5f\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
os_toplevel_consts_86_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "os.fwalk",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_86_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & os_toplevel_consts_86_consts_0._ascii.ob_base,
            & os_toplevel_consts_86_consts_1._ascii.ob_base,
            Py_True,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__fwalk_walk = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fwalk_walk",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__fwalk = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fwalk",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__fwalk_close = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fwalk_close",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
os_toplevel_consts_86_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_audit._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            & const_str__fwalk_walk._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str__fwalk._ascii.ob_base,
            & const_str_pop._ascii.ob_base,
            & const_str__fwalk_close._ascii.ob_base,
            &_Py_ID(close),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_fwalk = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fwalk",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[199];
    }
os_toplevel_consts_86_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 198,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf4\x42\x01\x00\x09\x0c\x8f\x09\x89\x09\x90\x2a\x98\x63\xa0\x37\xa8\x47\xb0\x5f\xc0\x66\xd4\x08\x4d\xdc\x0e\x14\x90\x53\x8b\x6b\x88\x03\xdc\x12\x1d\xa0\x04\xa0\x66\xa8\x63\xb0\x33\xb8\x04\xd0\x1f\x3d\xd0\x11\x3e\xd0\x10\x3f\x88\x05\xdc\x12\x1c\x98\x53\xa4\x25\xd3\x12\x28\x88\x07\xf0\x02\x08\x09\x21\xd9\x12\x17\xdc\x1b\x21\xa0\x25\xa8\x17\xb0\x27\xb8\x37\xc0\x4f\xd3\x1b\x54\xd7\x10\x54\xd0\x10\x54\xf2\x03\x00\x13\x18\xf1\x08\x00\x13\x18\xd8\x20\x25\xa7\x09\xa1\x09\xa3\x0b\x91\x0d\x90\x06\x98\x05\xd8\x13\x19\x9c\x5c\xd2\x13\x29\xdc\x14\x19\x98\x25\x94\x4c\xf4\x07\x00\x13\x18\xf0\x07\x00\x11\x55\x01\xf9\xf1\x06\x00\x13\x18\xd8\x20\x25\xa7\x09\xa1\x09\xa3\x0b\x91\x0d\x90\x06\x98\x05\xd8\x13\x19\x9c\x5c\xd2\x13\x29\xdc\x14\x19\x98\x25\x94\x4c\xf4\x07\x00\x13\x18\xfc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[61];
    }
os_toplevel_consts_86_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 60,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x41\x04\x43\x05\x01\xc1\x07\x14\x42\x13\x00\xc1\x1b\x01\x42\x11\x04\xc1\x1c\x06\x42\x13\x00\xc1\x23\x2b\x43\x05\x01\xc2\x0f\x02\x43\x05\x01\xc2\x11\x01\x42\x13\x00\xc2\x13\x2c\x43\x02\x03\xc3\x00\x02\x43\x02\x03\xc3\x02\x03\x43\x05\x01",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_isbytes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isbytes",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_action = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "action",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
os_toplevel_consts_86_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(top),
            & const_str_topdown._ascii.ob_base,
            & const_str_onerror._ascii.ob_base,
            &_Py_ID(follow_symlinks),
            &_Py_ID(dir_fd),
            & const_str_stack._ascii.ob_base,
            & const_str_isbytes._ascii.ob_base,
            & const_str_action._ascii.ob_base,
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(398)
os_toplevel_consts_86 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 199,
    },
    .co_consts = & os_toplevel_consts_86_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_86_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_86_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 2,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 444,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 612,
    .co_localsplusnames = & os_toplevel_consts_86_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_fwalk._ascii.ob_base,
    .co_qualname = & const_str_fwalk._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_86_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x7c\x01\x7c\x02\x7c\x03\x7c\x04\xab\x06\x00\x00\x00\x00\x00\x00\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x04\x7c\x00\x7c\x00\x64\x03\x66\x05\x66\x02\x67\x01\x7d\x05\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x09\x00\x7c\x05\x72\x1a\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x06\x7c\x01\x7c\x02\x7c\x03\xab\x05\x00\x00\x00\x00\x00\x00\x45\x00\x64\x03\x7b\x03\x00\x00\x96\x03\x97\x02\x86\x05\x05\x00\x01\x00\x7c\x05\x72\x01\x8c\x1a\x7c\x05\x72\x2b\x7c\x05\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x07\x7d\x08\x7c\x07\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x72\x0b\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x72\x01\x8c\x2a\x79\x03\x79\x03\x37\x00\x8c\x35\x23\x00\x7c\x05\x72\x2b\x7c\x05\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x07\x7d\x08\x7c\x07\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x72\x0b\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x72\x01\x8c\x2a\x77\x00\x77\x00\x78\x03\x59\x00\x77\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_87_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(dir_fd),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_87_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_False,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_87_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__fwalk_walk._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
os_toplevel_consts_87_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fwalk.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[46];
    }
os_toplevel_consts_87_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 45,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xf0\x00\x02\x19\x28\xe1\x1c\x26\x90\x44\xf4\x03\x00\x12\x1d\x98\x75\xa0\x65\xa8\x57\xb0\x74\xa9\x5e\xb8\x54\xc0\x34\xd0\x1e\x48\xd4\x10\x49\xd9\x1c\x26\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
os_toplevel_consts_87_consts_6_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x1a\x1d\x01",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_topfd = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "topfd",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_toppath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "toppath",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_87_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            &_Py_ID(name),
            & const_str_topfd._ascii.ob_base,
            & const_str_toppath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
os_toplevel_consts_87_consts_6_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x80\x80",
};
static
    struct _PyCode_DEF(62)
os_toplevel_consts_87_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & os_toplevel_consts_87_consts_6_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_87_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_87_consts_6_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 562,
    .co_nlocalsplus = 4,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 2,
    .co_version = 613,
    .co_localsplusnames = & os_toplevel_consts_87_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_87_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & os_toplevel_consts_87_consts_6_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_87_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x02\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x14\x00\x00\x7d\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x89\x02\x89\x03\x7c\x01\x7a\x00\x00\x00\x7c\x01\x64\x01\x66\x05\x66\x02\x96\x01\x97\x01\x01\x00\x8c\x16\x04\x00\x79\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[51];
    }
os_toplevel_consts_87_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 50,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xf0\x00\x02\x19\x43\x01\xe1\x23\x41\x91\x4b\x90\x44\x98\x25\xf4\x03\x00\x12\x1d\x98\x75\xa0\x65\xa8\x57\xb0\x74\xa9\x5e\xb8\x54\xc0\x35\xd0\x1e\x49\xd4\x10\x4a\xd9\x23\x41\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
os_toplevel_consts_87_consts_8_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x1d\x20\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_87_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            &_Py_ID(name),
            & const_str_entry._ascii.ob_base,
            & const_str_topfd._ascii.ob_base,
            & const_str_toppath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
os_toplevel_consts_87_consts_8_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x20\x20\x80\x80",
};
static
    struct _PyCode_DEF(68)
os_toplevel_consts_87_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 34,
    },
    .co_consts = & os_toplevel_consts_87_consts_6_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_87_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_87_consts_8_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 566,
    .co_nlocalsplus = 5,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 2,
    .co_version = 614,
    .co_localsplusnames = & os_toplevel_consts_87_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_87_consts_8_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & os_toplevel_consts_87_consts_6_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_87_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x02\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x17\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x89\x03\x89\x04\x7c\x01\x7a\x00\x00\x00\x7c\x01\x7c\x02\x66\x05\x66\x02\x96\x01\x97\x01\x01\x00\x8c\x19\x04\x00\x79\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
os_toplevel_consts_87_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            Py_None,
            Py_False,
            & os_toplevel_consts_85._object.ob_base.ob_base,
            & os_toplevel_consts_83_consts_5._object.ob_base.ob_base,
            & os_toplevel_consts_87_consts_4._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & os_toplevel_consts_87_consts_6.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            & os_toplevel_consts_87_consts_8.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__fwalk_yield = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fwalk_yield",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_O_RDONLY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "O_RDONLY",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_O_NONBLOCK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "O_NONBLOCK",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[24];
        }_object;
    }
os_toplevel_consts_87_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 24,
        },
        .ob_item = {
            & const_str_pop._ascii.ob_base,
            & const_str__fwalk_close._ascii.ob_base,
            &_Py_ID(close),
            & const_str__fwalk_yield._ascii.ob_base,
            & const_str__fwalk_walk._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            &_Py_ID(open),
            & const_str_O_RDONLY._ascii.ob_base,
            & const_str_O_NONBLOCK._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            &_Py_ID(append),
            & const_str_st._ascii.ob_base,
            & const_str_S_ISDIR._ascii.ob_base,
            & const_str_st_mode._ascii.ob_base,
            &_Py_ID(path),
            & const_str_samestat._ascii.ob_base,
            & const_str_scandir._ascii.ob_base,
            &_Py_ID(name),
            & const_str_fsencode._ascii.ob_base,
            & const_str_is_dir._ascii.ob_base,
            & const_str_is_symlink._ascii.ob_base,
            &_Py_ID(join),
            &_Py_ID(extend),
            & const_str_zip._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[597];
    }
os_toplevel_consts_87_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 596,
    },
    .ob_shash = -1,
    .ob_sval = "\xf9\xe8\x00\xf8\x80\x00\xf0\x0a\x00\x19\x1e\x9f\x09\x99\x09\x9b\x0b\x89\x0d\x88\x06\x90\x05\xd8\x0b\x11\x94\x5c\xd2\x0b\x21\xdc\x0c\x11\x90\x25\x8c\x4c\xd8\x0c\x12\xd8\x0d\x13\x94\x7c\xd2\x0d\x23\xd8\x12\x17\x8a\x4b\xd8\x0c\x12\xd8\x0f\x15\x9c\x1b\xd2\x0f\x24\xd0\x08\x24\xd0\x0f\x24\xd8\x31\x36\xd1\x08\x2e\x88\x06\x90\x05\x90\x77\xa0\x07\xa8\x15\xf0\x02\x0e\x09\x13\xd9\x13\x22\xf0\x06\x00\x14\x19\x90\x3d\xdc\x1e\x22\xa0\x37\xb8\x45\xc8\x25\xd4\x1e\x50\x91\x47\xe0\x1e\x23\x9f\x6a\x99\x6a\xb8\x15\x98\x6a\xd3\x1e\x3f\x90\x47\xdc\x14\x18\x98\x17\xa4\x28\xac\x5a\xd1\x22\x37\xc0\x05\xd4\x14\x46\x88\x45\xf0\x0e\x00\x09\x0e\x8f\x0c\x89\x0c\x94\x6c\xa0\x45\xd0\x15\x2a\xd4\x08\x2b\xd9\x0f\x1e\xd9\x0f\x15\x9c\x62\x9f\x6a\x99\x6a\xa8\x17\xaf\x1f\xa9\x1f\xd4\x1e\x39\xd8\x10\x16\xdc\x13\x17\x97\x3d\x91\x3d\xa0\x17\xac\x24\xa8\x75\xab\x2b\xd4\x13\x36\xd8\x10\x16\xe4\x15\x1c\x98\x55\x93\x5e\x88\x0a\xd8\x0f\x11\x88\x04\xd8\x12\x14\x88\x07\xd9\x1a\x21\xa1\x5f\x91\x24\xb8\x22\x88\x07\xdb\x15\x1f\x88\x45\xd8\x13\x18\x97\x3a\x91\x3a\x88\x44\xd9\x0f\x16\xdc\x17\x1f\xa0\x04\x93\x7e\x90\x04\xf0\x02\x0d\x0d\x19\xd8\x13\x18\x97\x3c\x91\x3c\x94\x3e\xd8\x14\x18\x97\x4b\x91\x4b\xa0\x04\xd4\x14\x25\xd8\x17\x1e\xd0\x17\x2a\xd8\x18\x1f\x9f\x0e\x99\x0e\xa0\x75\xd5\x18\x2d\xe0\x14\x1b\x97\x4e\x91\x4e\xa0\x34\xd4\x14\x28\xf8\xf0\x15\x00\x16\x20\xf1\x26\x00\x0c\x13\xd8\x12\x19\x98\x34\xa0\x17\xa8\x25\xd0\x12\x2f\xd3\x0c\x2f\xe0\x0c\x11\x8f\x4c\x89\x4c\x9c\x2c\xa8\x17\xb0\x24\xb8\x07\xc0\x15\xd0\x28\x47\xd0\x19\x48\xd4\x0c\x49\xe4\x12\x16\x97\x29\x91\x29\x98\x47\xa0\x57\xa8\x52\xa8\x61\xa0\x5b\xd3\x12\x31\x88\x07\xd8\x0b\x12\x88\x3f\xd8\x0c\x11\x8f\x4c\x89\x4c\xf4\x00\x02\x19\x28\xe0\x1c\x20\xa1\x14\xa0\x32\xa0\x14\x9a\x4a\xf3\x05\x02\x19\x28\xf5\x00\x02\x0d\x28\xf0\x08\x00\x0d\x12\x8f\x4c\x89\x4c\xf4\x00\x02\x19\x43\x01\xe4\x23\x26\xa0\x74\xa9\x44\xa8\x62\xa8\x44\xa1\x7a\xb0\x37\xb9\x34\xb8\x52\xb8\x34\xb1\x3d\xd4\x23\x41\xf3\x05\x02\x19\x43\x01\xf5\x00\x02\x0d\x43\x01\xf8\xf4\x5f\x01\x00\x10\x17\xf2\x00\x05\x09\x13\xd9\x0f\x15\xd8\x10\x15\xd8\x0f\x16\xd0\x0f\x22\xd9\x10\x17\x98\x03\x94\x0c\xdc\x0c\x12\xfb\xf0\x0b\x05\x09\x13\xfb\xf4\x38\x00\x14\x1b\xf2\x00\x06\x0d\x19\xf0\x02\x05\x11\x19\xe0\x17\x1c\xd7\x17\x27\xd1\x17\x27\xd4\x17\x29\xd8\x18\x1f\x9f\x0e\x99\x0e\xa0\x74\xd4\x18\x2c\xf9\xdc\x17\x1e\xf2\x00\x01\x11\x19\xd9\x14\x18\xf0\x03\x01\x11\x19\xfd\xf0\x0b\x06\x0d\x19\xfc",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[124];
    }
os_toplevel_consts_87_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 123,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x41\x0a\x49\x20\x01\xc1\x0f\x3e\x47\x3b\x00\xc2\x0d\x42\x10\x49\x20\x01\xc4\x1e\x41\x06\x48\x1e\x02\xc5\x24\x42\x17\x49\x20\x01\xc7\x3b\x09\x48\x1b\x03\xc8\x04\x0d\x48\x16\x03\xc8\x11\x05\x49\x20\x01\xc8\x16\x05\x48\x1b\x03\xc8\x1b\x03\x49\x20\x01\xc8\x1e\x09\x49\x1d\x05\xc8\x28\x21\x49\x0a\x04\xc9\x09\x01\x49\x1d\x05\xc9\x0a\x09\x49\x16\x07\xc9\x13\x02\x49\x1d\x05\xc9\x15\x01\x49\x16\x07\xc9\x16\x03\x49\x1d\x05\xc9\x19\x03\x49\x20\x01\xc9\x1c\x01\x49\x1d\x05\xc9\x1d\x03\x49\x20\x01",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_isroot = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isroot",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_dirfd = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dirfd",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_topname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "topname",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_orig_st = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "orig_st",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_err = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "err",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_entries = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "entries",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[20];
        }_object;
    }
os_toplevel_consts_87_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 20,
        },
        .ob_item = {
            & const_str_stack._ascii.ob_base,
            & const_str_isbytes._ascii.ob_base,
            & const_str_topdown._ascii.ob_base,
            & const_str_onerror._ascii.ob_base,
            &_Py_ID(follow_symlinks),
            & const_str_action._ascii.ob_base,
            &_Py_ID(value),
            & const_str_isroot._ascii.ob_base,
            & const_str_dirfd._ascii.ob_base,
            & const_str_topname._ascii.ob_base,
            & const_str_entry._ascii.ob_base,
            & const_str_orig_st._ascii.ob_base,
            & const_str_err._ascii.ob_base,
            & const_str_scandir_it._ascii.ob_base,
            & const_str_dirs._ascii.ob_base,
            & const_str_nondirs._ascii.ob_base,
            & const_str_entries._ascii.ob_base,
            &_Py_ID(name),
            & const_str_topfd._ascii.ob_base,
            & const_str_toppath._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
os_toplevel_consts_87_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "                  @@",
};
static
    struct _PyCode_DEF(1220)
os_toplevel_consts_87 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 610,
    },
    .co_consts = & os_toplevel_consts_87_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_87_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_87_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 30 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 10,
    .co_firstlineno = 496,
    .co_nlocalsplus = 20,
    .co_nlocals = 18,
    .co_ncellvars = 2,
    .co_nfreevars = 0,
    .co_version = 615,
    .co_localsplusnames = & os_toplevel_consts_87_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_87_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__fwalk._ascii.ob_base,
    .co_qualname = & const_str__fwalk._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_87_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x12\x87\x13\x4b\x00\x01\x00\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x05\x7d\x06\x7c\x05\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x72\x0c\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x7c\x05\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x72\x05\x7c\x06\x96\x01\x97\x01\x01\x00\x79\x00\x7c\x05\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x73\x02\x4a\x00\x82\x01\x7c\x06\x5c\x05\x00\x00\x7d\x07\x7d\x08\x8a\x13\x7d\x09\x7d\x0a\x09\x00\x7c\x04\x73\x23\x7c\x0a\x80\x0f\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x64\x01\x7c\x08\xac\x02\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x0b\x6e\x12\x7c\x0a\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xac\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0b\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x07\x00\x00\x7c\x08\xac\x04\xab\x03\x00\x00\x00\x00\x00\x00\x8a\x12\x7c\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x89\x12\x66\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x73\x42\x7c\x07\x72\x20\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7f\x0b\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x79\x00\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7f\x0b\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x89\x12\xab\x01\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x01\x79\x00\x74\x21\x00\x00\x00\x00\x00\x00\x00\x00\x89\x12\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0d\x67\x00\x7d\x0e\x67\x00\x7d\x0f\x7c\x02\x73\x02\x7c\x04\x72\x02\x64\x00\x6e\x01\x67\x00\x7d\x10\x7c\x0d\x44\x00\x5d\x62\x00\x00\x7d\x0a\x7c\x0a\x6a\x22\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x11\x7c\x01\x72\x0b\x74\x25\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x11\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x11\x09\x00\x7c\x0a\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x72\x25\x7c\x0e\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x11\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x10\x81\x23\x7c\x10\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x11\x7c\x0f\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x11\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x64\x04\x00\x7c\x02\x72\x09\x89\x13\x7c\x0e\x7c\x0f\x89\x12\x66\x04\x96\x01\x97\x01\x01\x00\x6e\x1b\x7c\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x89\x13\x7c\x0e\x7c\x0f\x89\x12\x66\x04\x66\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x13\x89\x13\x64\x00\x64\x05\x1a\x00\xab\x02\x00\x00\x00\x00\x00\x00\x8a\x13\x7c\x10\x80\x22\x7c\x00\x6a\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x88\x12\x88\x13\x66\x02\x64\x06\x84\x08\x7c\x0e\x64\x00\x64\x00\x64\x07\x85\x03\x19\x00\x00\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x7c\x00\x6a\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x88\x12\x88\x13\x66\x02\x64\x08\x84\x08\x74\x2f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0e\x64\x00\x64\x00\x64\x07\x85\x03\x19\x00\x00\x00\x7c\x10\x64\x00\x64\x00\x64\x07\x85\x03\x19\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x17\x7d\x0c\x7c\x07\x72\x01\x82\x00\x7c\x03\x81\x08\x02\x00\x7c\x03\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x00\x7d\x0c\x7e\x0c\x79\x00\x64\x00\x7d\x0c\x7e\x0c\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x36\x01\x00\x09\x00\x7c\x0a\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x72\x11\x7c\x0f\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x11\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x0f\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x04\x77\x00\x78\x03\x59\x00\x77\x01\x59\x00\x90\x01\x8c\x5b\x77\x00\x78\x03\x59\x00\x77\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 4,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[113];
    }
os_toplevel_consts_89_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 112,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x65\x78\x65\x63\x6c\x28\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x29\x0a\x0a\x20\x20\x20\x20\x45\x78\x65\x63\x75\x74\x65\x20\x74\x68\x65\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x20\x66\x69\x6c\x65\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x6c\x69\x73\x74\x20\x61\x72\x67\x73\x2c\x20\x72\x65\x70\x6c\x61\x63\x69\x6e\x67\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_89_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_89_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_execv = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execv",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_89_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_execv._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_execl = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execl",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
os_toplevel_consts_89_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x05\x0a\x88\x24\x90\x04\xd5\x04\x15",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_89_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(file),
            &_Py_ID(args),
        },
    },
};
static
    struct _PyCode_DEF(28)
os_toplevel_consts_89 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & os_toplevel_consts_89_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_89_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 572,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 616,
    .co_localsplusnames = & os_toplevel_consts_89_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_execl._ascii.ob_base,
    .co_qualname = & const_str_execl._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_89_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[139];
    }
os_toplevel_consts_90_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 138,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x65\x78\x65\x63\x6c\x65\x28\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x2c\x20\x65\x6e\x76\x29\x0a\x0a\x20\x20\x20\x20\x45\x78\x65\x63\x75\x74\x65\x20\x74\x68\x65\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x20\x66\x69\x6c\x65\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x6c\x69\x73\x74\x20\x61\x72\x67\x73\x20\x61\x6e\x64\x0a\x20\x20\x20\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x20\x65\x6e\x76\x2c\x20\x72\x65\x70\x6c\x61\x63\x69\x6e\x67\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_90_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & os_toplevel_consts_90_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_90_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_execve._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_execle = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execle",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
os_toplevel_consts_90_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x0b\x0f\x88\x72\x89\x28\x80\x43\xdc\x04\x0a\x88\x34\x90\x14\x90\x63\x90\x72\x90\x19\x98\x43\xd5\x04\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_90_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(file),
            &_Py_ID(args),
            &_Py_ID(env),
        },
    },
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_90 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & os_toplevel_consts_90_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_90_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 579,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 617,
    .co_localsplusnames = & os_toplevel_consts_90_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_execle._ascii.ob_base,
    .co_qualname = & const_str_execle._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_90_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x64\x01\x19\x00\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x64\x02\x64\x01\x1a\x00\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[150];
    }
os_toplevel_consts_91_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 149,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x65\x78\x65\x63\x6c\x70\x28\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x29\x0a\x0a\x20\x20\x20\x20\x45\x78\x65\x63\x75\x74\x65\x20\x74\x68\x65\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x0a\x20\x20\x20\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x6c\x69\x73\x74\x20\x61\x72\x67\x73\x2c\x20\x72\x65\x70\x6c\x61\x63\x69\x6e\x67\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_91_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_91_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_execvp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execvp",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_91_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_execvp._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_execlp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execlp",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
os_toplevel_consts_91_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0a\x00\x05\x0b\x88\x34\x90\x14\xd5\x04\x16",
};
static
    struct _PyCode_DEF(28)
os_toplevel_consts_91 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & os_toplevel_consts_91_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_91_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 587,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 618,
    .co_localsplusnames = & os_toplevel_consts_89_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_execlp._ascii.ob_base,
    .co_qualname = & const_str_execlp._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_91_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[180];
    }
os_toplevel_consts_92_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 179,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x65\x78\x65\x63\x6c\x70\x65\x28\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x2c\x20\x65\x6e\x76\x29\x0a\x0a\x20\x20\x20\x20\x45\x78\x65\x63\x75\x74\x65\x20\x74\x68\x65\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x0a\x20\x20\x20\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x6c\x69\x73\x74\x20\x61\x72\x67\x73\x20\x61\x6e\x64\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x20\x65\x6e\x76\x2c\x20\x72\x65\x70\x6c\x61\x63\x69\x6e\x67\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x0a\x20\x20\x20\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_92_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & os_toplevel_consts_92_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_execvpe = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execvpe",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_92_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_execvpe._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_execlpe = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execlpe",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[32];
    }
os_toplevel_consts_92_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 31,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0c\x00\x0b\x0f\x88\x72\x89\x28\x80\x43\xdc\x04\x0b\x88\x44\x90\x24\x90\x73\x98\x02\x90\x29\x98\x53\xd5\x04\x21",
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_92 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & os_toplevel_consts_92_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_92_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 594,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 619,
    .co_localsplusnames = & os_toplevel_consts_90_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_execlpe._ascii.ob_base,
    .co_qualname = & const_str_execlpe._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_92_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x64\x01\x19\x00\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x64\x02\x64\x01\x1a\x00\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[193];
    }
os_toplevel_consts_93_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 192,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x65\x78\x65\x63\x76\x70\x28\x66\x69\x6c\x65\x2c\x20\x61\x72\x67\x73\x29\x0a\x0a\x20\x20\x20\x20\x45\x78\x65\x63\x75\x74\x65\x20\x74\x68\x65\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x0a\x20\x20\x20\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x6c\x69\x73\x74\x20\x61\x72\x67\x73\x2c\x20\x72\x65\x70\x6c\x61\x63\x69\x6e\x67\x20\x74\x68\x65\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x20\x20\x20\x20\x61\x72\x67\x73\x20\x6d\x61\x79\x20\x62\x65\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x72\x20\x74\x75\x70\x6c\x65\x20\x6f\x66\x20\x73\x74\x72\x69\x6e\x67\x73\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_93_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_93_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__execvpe = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_execvpe",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_93_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__execvpe._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
os_toplevel_consts_93_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0c\x00\x05\x0d\x88\x54\x90\x34\xd5\x04\x18",
};
static
    struct _PyCode_DEF(28)
os_toplevel_consts_93 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & os_toplevel_consts_93_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_93_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 603,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 620,
    .co_localsplusnames = & os_toplevel_consts_89_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_execvp._ascii.ob_base,
    .co_qualname = & const_str_execvp._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_93_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[223];
    }
os_toplevel_consts_94_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 222,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x65\x78\x65\x63\x76\x70\x65\x28\x66\x69\x6c\x65\x2c\x20\x61\x72\x67\x73\x2c\x20\x65\x6e\x76\x29\x0a\x0a\x20\x20\x20\x20\x45\x78\x65\x63\x75\x74\x65\x20\x74\x68\x65\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x0a\x20\x20\x20\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x6c\x69\x73\x74\x20\x61\x72\x67\x73\x20\x61\x6e\x64\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x20\x65\x6e\x76\x2c\x20\x72\x65\x70\x6c\x61\x63\x69\x6e\x67\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x63\x75\x72\x72\x65\x6e\x74\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x20\x20\x20\x20\x61\x72\x67\x73\x20\x6d\x61\x79\x20\x62\x65\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x72\x20\x74\x75\x70\x6c\x65\x20\x6f\x66\x20\x73\x74\x72\x69\x6e\x67\x73\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_94_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_94_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
os_toplevel_consts_94_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x05\x0d\x88\x54\x90\x34\x98\x13\xd5\x04\x1d",
};
static
    struct _PyCode_DEF(30)
os_toplevel_consts_94 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & os_toplevel_consts_94_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_93_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 611,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 621,
    .co_localsplusnames = & os_toplevel_consts_90_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_execvpe._ascii.ob_base,
    .co_qualname = & const_str_execvpe._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_94_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
os_toplevel_consts_95 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_execl._ascii.ob_base,
            & const_str_execle._ascii.ob_base,
            & const_str_execlp._ascii.ob_base,
            & const_str_execlpe._ascii.ob_base,
            & const_str_execvp._ascii.ob_base,
            & const_str_execvpe._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_96_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(nt),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
os_toplevel_consts_96_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_execve._ascii.ob_base,
            & const_str_execv._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            &_Py_ID(path),
            & const_str_dirname._ascii.ob_base,
            & const_str_get_exec_path._ascii.ob_base,
            &_Py_ID(name),
            & const_str_fsencode._ascii.ob_base,
            & const_str_map._ascii.ob_base,
            &_Py_ID(join),
            & const_str_FileNotFoundError._ascii.ob_base,
            & const_str_NotADirectoryError._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[227];
    }
os_toplevel_consts_96_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 226,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x07\x0a\x80\x7f\xdc\x14\x1a\x88\x09\xd8\x13\x17\x98\x13\x90\x2b\x89\x07\xe4\x14\x19\x88\x09\xd8\x13\x17\x90\x27\x88\x07\xdc\x0e\x15\x88\x03\xe4\x07\x0b\x87\x7c\x81\x7c\x90\x44\xd4\x07\x19\xd9\x08\x11\x90\x24\xd0\x08\x21\x98\x17\xd3\x08\x21\xd8\x08\x0e\xd8\x10\x14\x80\x49\xdc\x10\x1d\x98\x63\xd3\x10\x22\x80\x49\xdc\x07\x0b\x88\x74\x82\x7c\xdc\x0f\x17\x98\x04\x8b\x7e\x88\x04\xdc\x14\x17\x9c\x08\xa0\x29\xd3\x14\x2c\x88\x09\xdb\x0f\x18\x88\x03\xdc\x13\x17\x97\x39\x91\x39\x98\x53\xa0\x24\xd3\x13\x27\x88\x08\xf0\x02\x07\x09\x1e\xd9\x0c\x15\x90\x68\xd0\x0c\x29\xa0\x17\xd4\x0c\x29\xf0\x07\x00\x10\x19\xf0\x14\x00\x08\x11\xd0\x07\x1c\xd8\x0e\x17\x88\x0f\xd8\x0a\x12\x80\x4e\xf8\xf4\x11\x00\x11\x22\xd4\x23\x35\xd0\x0f\x36\xf2\x00\x01\x09\x19\xd8\x17\x18\x8d\x48\xfb\xdc\x0f\x16\xf2\x00\x03\x09\x1e\xd8\x17\x18\x88\x48\xd8\x0f\x18\xd0\x0f\x20\xd8\x1c\x1d\x90\x09\xff\xf8\xf0\x07\x03\x09\x1e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
os_toplevel_consts_96_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x09\x09\x42\x1a\x02\xc2\x1a\x0f\x43\x0c\x05\xc2\x29\x02\x42\x30\x05\xc2\x30\x0c\x43\x0c\x05\xc2\x3c\x06\x43\x07\x05\xc3\x07\x05\x43\x0c\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_exec_func = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exec_func",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_argrest = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "argrest",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_saved_exc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "saved_exc",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
os_toplevel_consts_96_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(file),
            &_Py_ID(args),
            &_Py_ID(env),
            & const_str_exec_func._ascii.ob_base,
            & const_str_argrest._ascii.ob_base,
            & const_str_saved_exc._ascii.ob_base,
            & const_str_path_list._ascii.ob_base,
            & const_str_dir._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[101],
            &_Py_ID(last_exc),
        },
    },
};
static
    struct _PyCode_DEF(414)
os_toplevel_consts_96 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 207,
    },
    .co_consts = & os_toplevel_consts_96_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_96_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_96_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 622,
    .co_nlocalsplus = 11,
    .co_nlocals = 11,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 622,
    .co_localsplusnames = & os_toplevel_consts_96_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__execvpe._ascii.ob_base,
    .co_qualname = & const_str__execvpe._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_96_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x81\x0b\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x01\x7c\x02\x66\x02\x7d\x04\x6e\x0f\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x01\x66\x01\x7d\x04\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0a\x02\x00\x7c\x03\x7c\x00\x67\x01\x7c\x04\xa2\x01\xad\x06\x8e\x00\x01\x00\x79\x00\x64\x00\x7d\x05\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x37\x00\x00\x72\x1b\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x44\x00\x5d\x22\x00\x00\x7d\x07\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x08\x09\x00\x02\x00\x7c\x03\x7c\x08\x67\x01\x7c\x04\xa2\x01\xad\x06\x8e\x00\x01\x00\x8c\x24\x04\x00\x7c\x05\x81\x02\x7c\x05\x82\x01\x7f\x0a\x82\x01\x23\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x0c\x7d\x09\x7c\x09\x7d\x0a\x59\x00\x64\x00\x7d\x09\x7e\x09\x8c\x41\x64\x00\x7d\x09\x7e\x09\x77\x01\x74\x18\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x10\x7d\x09\x7c\x09\x7d\x0a\x7c\x05\x80\x02\x7c\x09\x7d\x05\x59\x00\x64\x00\x7d\x09\x7e\x09\x8c\x58\x64\x00\x7d\x09\x7e\x09\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[244];
    }
os_toplevel_consts_97_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 243,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x73\x65\x71\x75\x65\x6e\x63\x65\x20\x6f\x66\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x74\x68\x61\x74\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x6e\x61\x6d\x65\x64\x20\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x20\x28\x73\x69\x6d\x69\x6c\x61\x72\x20\x74\x6f\x20\x61\x20\x73\x68\x65\x6c\x6c\x29\x20\x77\x68\x65\x6e\x20\x6c\x61\x75\x6e\x63\x68\x69\x6e\x67\x20\x61\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x0a\x20\x20\x20\x20\x2a\x65\x6e\x76\x2a\x20\x6d\x75\x73\x74\x20\x62\x65\x20\x61\x6e\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x20\x64\x69\x63\x74\x20\x6f\x72\x20\x4e\x6f\x6e\x65\x2e\x20\x20\x49\x66\x20\x2a\x65\x6e\x76\x2a\x20\x69\x73\x20\x4e\x6f\x6e\x65\x2c\x0a\x20\x20\x20\x20\x6f\x73\x2e\x65\x6e\x76\x69\x72\x6f\x6e\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x75\x73\x65\x64\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_PATH = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PATH",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
os_toplevel_consts_97_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "PATH",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[43];
    }
os_toplevel_consts_97_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 42,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "env cannot contain 'PATH' and b'PATH' keys",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
os_toplevel_consts_97_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & os_toplevel_consts_97_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            &_Py_ID(ignore),
            & const_str_PATH._ascii.ob_base,
            & os_toplevel_consts_97_consts_5.ob_base.ob_base,
            & os_toplevel_consts_97_consts_6._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_catch_warnings = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "catch_warnings",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_simplefilter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "simplefilter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str_supports_bytes_environ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "supports_bytes_environ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
os_toplevel_consts_97_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            &_Py_ID(warnings),
            & const_str_environ._ascii.ob_base,
            & const_str_catch_warnings._ascii.ob_base,
            & const_str_simplefilter._ascii.ob_base,
            & const_str_BytesWarning._ascii.ob_base,
            &_Py_ID(get),
            & const_str_TypeError._ascii.ob_base,
            & const_str_supports_bytes_environ._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_fsdecode._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            & const_str_split._ascii.ob_base,
            & const_str_pathsep._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[229];
    }
os_toplevel_consts_97_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 228,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf3\x14\x00\x05\x14\xe0\x07\x0a\x80\x7b\xdc\x0e\x15\x88\x03\xf0\x08\x00\x0a\x12\xd7\x09\x20\xd1\x09\x20\xd5\x09\x22\xd8\x08\x10\xd7\x08\x1d\xd1\x08\x1d\x98\x68\xac\x0c\xd4\x08\x35\xf0\x04\x03\x09\x1d\xd8\x18\x1b\x9f\x07\x99\x07\xa0\x06\x9b\x0f\x88\x49\xf5\x08\x00\x0c\x22\xf0\x02\x08\x0d\x27\xd8\x1d\x20\xa0\x17\x99\x5c\x90\x0a\xf0\x08\x00\x14\x1d\xd0\x13\x28\xdc\x1a\x24\xd8\x18\x44\xf3\x03\x01\x1b\x46\x01\xf0\x00\x01\x15\x46\x01\xe0\x1c\x26\x90\x09\xe0\x0f\x18\xd0\x0f\x24\xac\x1a\xb0\x49\xbc\x75\xd4\x29\x45\xdc\x1c\x24\xa0\x59\xd3\x1c\x2f\x90\x09\xf7\x29\x00\x0a\x23\xf0\x2c\x00\x08\x11\xd0\x07\x18\xdc\x14\x1b\x88\x09\xd8\x0b\x14\x8f\x3f\x89\x3f\x9c\x37\xd3\x0b\x23\xd0\x04\x23\xf8\xf4\x27\x00\x10\x19\xf2\x00\x01\x09\x1d\xd8\x18\x1c\x8a\x49\xf0\x03\x01\x09\x1d\xfb\xf4\x0c\x00\x15\x1d\x9c\x69\xd0\x13\x28\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfa\xf7\x17\x00\x0a\x23\xd0\x09\x22\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[83];
    }
os_toplevel_consts_97_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 82,
    },
    .ob_shash = -1,
    .ob_sval = "\x9d\x17\x43\x09\x03\xb5\x11\x42\x23\x02\xc1\x06\x06\x43\x09\x03\xc1\x0d\x05\x42\x34\x02\xc1\x12\x2c\x43\x09\x03\xc2\x23\x0b\x42\x31\x05\xc2\x2e\x02\x43\x09\x03\xc2\x30\x01\x42\x31\x05\xc2\x31\x03\x43\x09\x03\xc2\x34\x0f\x43\x06\x05\xc3\x03\x02\x43\x09\x03\xc3\x05\x01\x43\x06\x05\xc3\x06\x03\x43\x09\x03\xc3\x09\x05\x43\x12\x07",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_path_listb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_listb",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_97_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(env),
            &_Py_ID(warnings),
            & const_str_path_list._ascii.ob_base,
            & const_str_path_listb._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(426)
os_toplevel_consts_97 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 213,
    },
    .co_consts = & os_toplevel_consts_97_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_97_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_97_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 654,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 623,
    .co_localsplusnames = & os_toplevel_consts_97_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_get_exec_path._ascii.ob_base,
    .co_qualname = & const_str_get_exec_path._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_97_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x7d\x01\x7c\x00\x80\x06\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x01\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x7c\x01\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x72\x32\x09\x00\x7c\x00\x64\x05\x19\x00\x00\x00\x7d\x03\x7c\x02\x81\x0b\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x03\x7d\x02\x7c\x02\x81\x1b\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x0b\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x02\x64\x02\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7f\x02\x80\x06\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x05\x01\x00\x64\x02\x7d\x02\x59\x00\x8c\x6a\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x64\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x4c\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_98 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_MutableMapping._ascii.ob_base,
            & const_str_Mapping._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__Environ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_encodekey = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "encodekey",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_decodekey = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "decodekey",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_encodevalue = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "encodevalue",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_decodevalue = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "decodevalue",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__data = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_data",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_99_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_encodekey._ascii.ob_base,
            & const_str_decodekey._ascii.ob_base,
            & const_str_encodevalue._ascii.ob_base,
            & const_str_decodevalue._ascii.ob_base,
            & const_str__data._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
os_toplevel_consts_99_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[40];
    }
os_toplevel_consts_99_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 39,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x19\x22\x88\x04\x8c\x0e\xd8\x19\x22\x88\x04\x8c\x0e\xd8\x1b\x26\x88\x04\xd4\x08\x18\xd8\x1b\x26\x88\x04\xd4\x08\x18\xd8\x15\x19\x88\x04\x8d\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
os_toplevel_consts_99_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(data),
            & const_str_encodekey._ascii.ob_base,
            & const_str_decodekey._ascii.ob_base,
            & const_str_encodevalue._ascii.ob_base,
            & const_str_decodevalue._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(74)
os_toplevel_consts_99_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 37,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 6,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 702,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 624,
    .co_localsplusnames = & os_toplevel_consts_99_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & os_toplevel_consts_99_consts_1_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_99_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__data._ascii.ob_base,
            & const_str_encodekey._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
            & const_str_decodevalue._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
os_toplevel_consts_99_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__getitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[79];
    }
os_toplevel_consts_99_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 78,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x04\x09\x2a\xd8\x14\x18\x97\x4a\x91\x4a\x98\x74\x9f\x7e\x99\x7e\xa8\x63\xd3\x1f\x32\xd1\x14\x33\x88\x45\xf0\x08\x00\x10\x14\xd7\x0f\x1f\xd1\x0f\x1f\xa0\x05\xd3\x0f\x26\xd0\x08\x26\xf8\xf4\x07\x00\x10\x18\xf2\x00\x02\x09\x2a\xe4\x12\x1a\x98\x33\x93\x2d\xa0\x54\xd0\x0c\x29\xf0\x05\x02\x09\x2a\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
os_toplevel_consts_99_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x1e\x31\x00\xb1\x16\x41\x07\x03",
};
static
    struct _PyCode_DEF(148)
os_toplevel_consts_99_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 74,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_99_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 709,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 625,
    .co_localsplusnames = & _collections_abc_toplevel_consts_66_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getitem__),
    .co_qualname = & os_toplevel_consts_99_consts_2_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x19\x00\x00\x00\x7d\x02\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0d\x01\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_putenv = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "putenv",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_99_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_encodekey._ascii.ob_base,
            & const_str_encodevalue._ascii.ob_base,
            & const_str_putenv._ascii.ob_base,
            & const_str__data._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
os_toplevel_consts_99_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__setitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[56];
    }
os_toplevel_consts_99_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 55,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0e\x12\x8f\x6e\x89\x6e\x98\x53\xd3\x0e\x21\x88\x03\xd8\x10\x14\xd7\x10\x20\xd1\x10\x20\xa0\x15\xd3\x10\x27\x88\x05\xdc\x08\x0e\x88\x73\x90\x45\xd4\x08\x1a\xd8\x1a\x1f\x88\x04\x8f\x0a\x89\x0a\x90\x33\x8a\x0f",
};
static
    struct _PyCode_DEF(126)
os_toplevel_consts_99_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 63,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 717,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 626,
    .co_localsplusnames = & _collections_abc_toplevel_consts_66_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__setitem__),
    .co_qualname = & os_toplevel_consts_99_consts_3_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_unsetenv = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "unsetenv",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_99_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_encodekey._ascii.ob_base,
            & const_str_unsetenv._ascii.ob_base,
            & const_str__data._ascii.ob_base,
            & const_str_KeyError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
os_toplevel_consts_99_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__delitem__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[69];
    }
os_toplevel_consts_99_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 68,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x15\x19\x97\x5e\x91\x5e\xa0\x43\xd3\x15\x28\x88\x0a\xdc\x08\x10\x90\x1a\xd4\x08\x1c\xf0\x02\x04\x09\x2a\xd8\x10\x14\x97\x0a\x91\x0a\x98\x3a\xd1\x10\x26\xf8\xdc\x0f\x17\xf2\x00\x02\x09\x2a\xe4\x12\x1a\x98\x33\x93\x2d\xa0\x54\xd0\x0c\x29\xf0\x05\x02\x09\x2a\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[10];
    }
os_toplevel_consts_99_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 9,
    },
    .ob_shash = -1,
    .ob_sval = "\x9e\x0d\x2c\x00\xac\x16\x41\x02\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_encodedkey = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "encodedkey",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_99_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(key),
            & const_str_encodedkey._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(138)
os_toplevel_consts_99_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 69,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_99_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 723,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 627,
    .co_localsplusnames = & os_toplevel_consts_99_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__delitem__),
    .co_qualname = & os_toplevel_consts_99_consts_4_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x3d\x00\x79\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x0d\x01\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x00\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_99_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_list._ascii.ob_base,
            & const_str__data._ascii.ob_base,
            & const_str_decodekey._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
os_toplevel_consts_99_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[46];
    }
os_toplevel_consts_99_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 45,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xe4\x0f\x13\x90\x44\x97\x4a\x91\x4a\xd3\x0f\x1f\x88\x04\xdb\x13\x17\x88\x43\xd8\x12\x16\x97\x2e\x91\x2e\xa0\x13\xd3\x12\x25\xd3\x0c\x25\xf1\x03\x00\x14\x18\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
os_toplevel_consts_99_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x31\x33\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_99_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(keys),
            &_Py_ID(key),
        },
    },
};
static
    struct _PyCode_DEF(106)
os_toplevel_consts_99_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 53,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_99_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 35,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 732,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 628,
    .co_localsplusnames = & os_toplevel_consts_99_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & os_toplevel_consts_99_consts_5_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x44\x00\x5d\x15\x00\x00\x7d\x02\x7c\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x96\x01\x97\x01\x01\x00\x8c\x17\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_99_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(len),
            & const_str__data._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
os_toplevel_consts_99_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__len__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
os_toplevel_consts_99_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x12\x90\x34\x97\x3a\x91\x3a\x8b\x7f\xd0\x08\x1e",
};
static
    struct _PyCode_DEF(44)
os_toplevel_consts_99_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 738,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 629,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__len__),
    .co_qualname = & os_toplevel_consts_99_consts_6_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_99_consts_7_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_42_consts_4._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_99_consts_7_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_decodekey._ascii.ob_base,
            & const_str_decodevalue._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[37];
    }
os_toplevel_consts_99_consts_7_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 36,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__repr__.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
os_toplevel_consts_99_consts_7_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\xe8\x00\xf8\x80\x00\xf0\x00\x03\x24\x0a\xe1\x1e\x30\x91\x0a\x90\x03\x90\x55\xf0\x03\x00\x10\x14\x8f\x7e\x89\x7e\x98\x63\xd3\x0f\x22\xd0\x0e\x25\xa0\x52\xa8\x04\xd7\x28\x38\xd1\x28\x38\xb8\x15\xd3\x28\x3f\xd0\x27\x42\xd4\x0c\x43\xd9\x1e\x30\xf9",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
os_toplevel_consts_99_consts_7_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x32\x35\x01",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_99_consts_7_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            &_Py_ID(key),
            &_Py_ID(value),
            &_Py_ID(self),
        },
    },
};
static
    struct _PyCode_DEF(110)
os_toplevel_consts_99_consts_7_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 55,
    },
    .co_consts = & os_toplevel_consts_99_consts_7_consts_2_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_7_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_99_consts_7_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 742,
    .co_nlocalsplus = 4,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 630,
    .co_localsplusnames = & os_toplevel_consts_99_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & os_toplevel_consts_99_consts_7_consts_2_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_7_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x2c\x00\x00\x5c\x02\x00\x00\x7d\x01\x7d\x02\x89\x03\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x02\x64\x00\x89\x03\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x03\x96\x01\x97\x01\x01\x00\x8c\x2e\x04\x00\x79\x01\xad\x03\x77\x01",
    ._co_firsttraceable = 3,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
os_toplevel_consts_99_consts_7_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "environ({",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
os_toplevel_consts_99_consts_7_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "})",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_99_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_30_consts_5_consts_6._ascii.ob_base,
            & os_toplevel_consts_99_consts_7_consts_2.ob_base.ob_base,
            & os_toplevel_consts_99_consts_7_consts_3._ascii.ob_base,
            & os_toplevel_consts_99_consts_7_consts_4._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_99_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(join),
            & const_str__data._ascii.ob_base,
            &_Py_ID(items),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
os_toplevel_consts_99_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[62];
    }
os_toplevel_consts_99_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 61,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x1a\x1e\x9f\x29\x99\x29\xf3\x00\x03\x24\x0a\xe0\x1e\x22\x9f\x6a\x99\x6a\xd7\x1e\x2e\xd1\x1e\x2e\xd4\x1e\x30\xf3\x05\x03\x24\x0a\xf3\x00\x03\x1b\x0a\x88\x0f\xf0\x08\x00\x12\x1c\x98\x4f\xd0\x1b\x2c\xa8\x43\xd0\x0f\x30\xd0\x08\x30",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_formatted_items = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "formatted_items",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_99_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_formatted_items._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(116)
os_toplevel_consts_99_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 58,
    },
    .co_consts = & os_toplevel_consts_99_consts_7_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 741,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 1,
    .co_nfreevars = 0,
    .co_version = 631,
    .co_localsplusnames = & os_toplevel_consts_99_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_26_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & os_toplevel_consts_99_consts_7_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x97\x00\x64\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x88\x00\x66\x01\x64\x02\x84\x08\x89\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x64\x03\x7c\x01\x9b\x00\x64\x04\x9d\x03\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_99_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(dict),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
os_toplevel_consts_99_consts_8_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.copy",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
os_toplevel_consts_99_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x13\x90\x44\x8b\x7a\xd0\x08\x19",
};
static
    struct _PyCode_DEF(24)
os_toplevel_consts_99_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 748,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 632,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(copy),
    .co_qualname = & os_toplevel_consts_99_consts_8_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
os_toplevel_consts_99_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.setdefault",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
os_toplevel_consts_99_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0e\x90\x64\x89\x3f\xd8\x18\x1d\x88\x44\x90\x13\x89\x49\xd8\x0f\x13\x90\x43\x89\x79\xd0\x08\x18",
};
static
    struct _PyCode_DEF(30)
os_toplevel_consts_99_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 751,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 633,
    .co_localsplusnames = & _collections_abc_toplevel_consts_66_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_setdefault._ascii.ob_base,
    .co_qualname = & os_toplevel_consts_99_consts_9_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x76\x01\x72\x05\x7c\x02\x7c\x00\x7c\x01\x3c\x00\x00\x00\x7c\x00\x7c\x01\x19\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_99_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_update._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
os_toplevel_consts_99_consts_10_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__ior__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
os_toplevel_consts_99_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\x8f\x0b\x89\x0b\x90\x45\xd4\x08\x1a\xd8\x0f\x13\x88\x0b",
};
static
    struct _PyCode_DEF(40)
os_toplevel_consts_99_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 756,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 634,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_58_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__ior__),
    .co_qualname = & os_toplevel_consts_99_consts_10_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_99_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_Mapping._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
            &_Py_ID(dict),
            & const_str_update._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
os_toplevel_consts_99_consts_11_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__or__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
os_toplevel_consts_99_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x17\xd4\x0f\x29\xdc\x13\x21\xd0\x0c\x21\xdc\x0e\x12\x90\x34\x8b\x6a\x88\x03\xd8\x08\x0b\x8f\x0a\x89\x0a\x90\x35\xd4\x08\x19\xd8\x0f\x12\x88\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_99_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_other._ascii.ob_base,
            & const_str_new._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(106)
os_toplevel_consts_99_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 53,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 760,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 635,
    .co_localsplusnames = & os_toplevel_consts_99_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__or__),
    .co_qualname = & os_toplevel_consts_99_consts_11_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
os_toplevel_consts_99_consts_12_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Environ.__ror__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
os_toplevel_consts_99_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x17\xd4\x0f\x29\xdc\x13\x21\xd0\x0c\x21\xdc\x0e\x12\x90\x35\x8b\x6b\x88\x03\xd8\x08\x0b\x8f\x0a\x89\x0a\x90\x34\xd4\x08\x18\xd8\x0f\x12\x88\x0a",
};
static
    struct _PyCode_DEF(106)
os_toplevel_consts_99_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 53,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 767,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 636,
    .co_localsplusnames = & os_toplevel_consts_99_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__ror__),
    .co_qualname = & os_toplevel_consts_99_consts_12_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x06\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
os_toplevel_consts_99_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & const_str__Environ._ascii.ob_base,
            & os_toplevel_consts_99_consts_1.ob_base.ob_base,
            & os_toplevel_consts_99_consts_2.ob_base.ob_base,
            & os_toplevel_consts_99_consts_3.ob_base.ob_base,
            & os_toplevel_consts_99_consts_4.ob_base.ob_base,
            & os_toplevel_consts_99_consts_5.ob_base.ob_base,
            & os_toplevel_consts_99_consts_6.ob_base.ob_base,
            & os_toplevel_consts_99_consts_7.ob_base.ob_base,
            & os_toplevel_consts_99_consts_8.ob_base.ob_base,
            & os_toplevel_consts_99_consts_9.ob_base.ob_base,
            & os_toplevel_consts_99_consts_10.ob_base.ob_base,
            & os_toplevel_consts_99_consts_11.ob_base.ob_base,
            & os_toplevel_consts_99_consts_12.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
os_toplevel_consts_99_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            &_Py_ID(__getitem__),
            &_Py_ID(__setitem__),
            &_Py_ID(__delitem__),
            &_Py_ID(__iter__),
            &_Py_ID(__len__),
            &_Py_ID(__repr__),
            &_Py_ID(copy),
            & const_str_setdefault._ascii.ob_base,
            &_Py_ID(__ior__),
            &_Py_ID(__or__),
            &_Py_ID(__ror__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
os_toplevel_consts_99_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x02\x05\x05\x1a\xf2\x0e\x06\x05\x27\xf2\x10\x04\x05\x20\xf2\x0c\x07\x05\x2a\xf2\x12\x04\x05\x26\xf2\x0c\x01\x05\x1f\xf2\x06\x05\x05\x31\xf2\x0e\x01\x05\x1a\xf2\x06\x03\x05\x19\xf2\x0a\x02\x05\x14\xf2\x08\x05\x05\x13\xf3\x0e\x05\x05\x13",
};
static
    struct _PyCode_DEF(84)
os_toplevel_consts_99 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 42,
    },
    .co_consts = & os_toplevel_consts_99_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_99_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 701,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 637,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__Environ._ascii.ob_base,
    .co_qualname = & const_str__Environ._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_99_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x64\x07\x84\x00\x5a\x09\x64\x08\x84\x00\x5a\x0a\x64\x09\x84\x00\x5a\x0b\x64\x0a\x84\x00\x5a\x0c\x64\x0b\x84\x00\x5a\x0d\x64\x0c\x84\x00\x5a\x0e\x79\x0d",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
os_toplevel_consts_101_consts_2_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "str expected, not %s",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_101_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_101_consts_2_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_101_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_check_str = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "check_str",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
os_toplevel_consts_101_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_createenviron.<locals>.check_str",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
os_toplevel_consts_101_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x13\x1d\x98\x65\xa4\x53\xd4\x13\x29\xdc\x16\x1f\xd0\x20\x36\xbc\x14\xb8\x65\xbb\x1b\xd7\x39\x4d\xd1\x39\x4d\xd1\x20\x4d\xd3\x16\x4e\xd0\x10\x4e\xd8\x13\x18\x88\x4c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_101_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(104)
os_toplevel_consts_101_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 52,
    },
    .co_consts = & os_toplevel_consts_101_consts_2_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_101_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 777,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 638,
    .co_localsplusnames = & os_toplevel_consts_101_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_check_str._ascii.ob_base,
    .co_qualname = & os_toplevel_consts_101_consts_2_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_101_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x21\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_101_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_upper._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
os_toplevel_consts_101_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_createenviron.<locals>.encodekey",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[23];
    }
os_toplevel_consts_101_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 22,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd9\x13\x19\x98\x23\x93\x3b\xd7\x13\x24\xd1\x13\x24\xd3\x13\x26\xd0\x0c\x26",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_101_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(key),
            &_Py_ID(encode),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
os_toplevel_consts_101_consts_3_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\x20\x80",
};
static
    struct _PyCode_DEF(48)
os_toplevel_consts_101_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 24,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_101_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 783,
    .co_nlocalsplus = 2,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 639,
    .co_localsplusnames = & os_toplevel_consts_101_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_101_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_encodekey._ascii.ob_base,
    .co_qualname = & os_toplevel_consts_101_consts_3_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_101_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x02\x00\x89\x01\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_101_consts_4_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_101_consts_2_consts_1._ascii.ob_base,
            & const_str_surrogateescape._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
os_toplevel_consts_101_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            &_Py_ID(__name__),
            &_Py_ID(encode),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
os_toplevel_consts_101_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_createenviron.<locals>.encode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[59];
    }
os_toplevel_consts_101_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 58,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xdc\x13\x1d\x98\x65\xa4\x53\xd4\x13\x29\xdc\x16\x1f\xd0\x20\x36\xbc\x14\xb8\x65\xbb\x1b\xd7\x39\x4d\xd1\x39\x4d\xd1\x20\x4d\xd3\x16\x4e\xd0\x10\x4e\xd8\x13\x18\x97\x3c\x91\x3c\xa0\x08\xd0\x2a\x3b\xd3\x13\x3c\xd0\x0c\x3c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_101_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(value),
            &_Py_ID(encoding),
        },
    },
};
static
    struct _PyCode_DEF(138)
os_toplevel_consts_101_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 69,
    },
    .co_consts = & os_toplevel_consts_101_consts_4_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_101_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 791,
    .co_nlocalsplus = 2,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 640,
    .co_localsplusnames = & os_toplevel_consts_101_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_101_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(encode),
    .co_qualname = & os_toplevel_consts_101_consts_4_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_101_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x21\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_101_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & const_str_surrogateescape._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_101_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(decode),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
os_toplevel_consts_101_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_createenviron.<locals>.decode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
os_toplevel_consts_101_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xd8\x13\x18\x97\x3c\x91\x3c\xa0\x08\xd0\x2a\x3b\xd3\x13\x3c\xd0\x0c\x3c",
};
static
    struct _PyCode_DEF(40)
os_toplevel_consts_101_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & os_toplevel_consts_101_consts_5_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_101_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 795,
    .co_nlocalsplus = 2,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 1,
    .co_version = 641,
    .co_localsplusnames = & os_toplevel_consts_101_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_101_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(decode),
    .co_qualname = & os_toplevel_consts_101_consts_5_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_101_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x01\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
os_toplevel_consts_101_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(nt),
            & os_toplevel_consts_101_consts_2.ob_base.ob_base,
            & os_toplevel_consts_101_consts_3.ob_base.ob_base,
            & os_toplevel_consts_101_consts_4.ob_base.ob_base,
            & os_toplevel_consts_101_consts_5.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
os_toplevel_consts_101_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_str._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            &_Py_ID(items),
            & const_str_sys._ascii.ob_base,
            & const_str_getfilesystemencoding._ascii.ob_base,
            & const_str__Environ._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__createenviron = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_createenviron",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[133];
    }
os_toplevel_consts_101_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 132,
    },
    .ob_shash = -1,
    .ob_sval = "\xf9\x80\x00\xdc\x07\x0b\x88\x74\x82\x7c\xf2\x04\x03\x09\x19\xf0\x08\x00\x12\x1b\x88\x06\xdc\x11\x14\x88\x06\xf4\x02\x01\x09\x27\xe0\x0f\x11\x88\x04\xdc\x1a\x21\x9f\x2d\x99\x2d\x9e\x2f\x89\x4a\x88\x43\x90\x15\xd8\x23\x28\x88\x44\x91\x19\x98\x33\x93\x1e\xd2\x0c\x20\xf1\x03\x00\x1b\x2a\xf4\x08\x00\x14\x17\xd7\x13\x2c\xd1\x13\x2c\xd3\x13\x2e\x88\x08\xf4\x02\x03\x09\x3d\xf4\x08\x01\x09\x3d\xe0\x14\x1a\x88\x09\xdc\x0f\x16\x88\x04\xdc\x0b\x13\x90\x44\xd8\x08\x11\x90\x36\xd8\x08\x0e\x90\x06\xf3\x05\x02\x0c\x18\xf0\x00\x02\x05\x18",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
os_toplevel_consts_101_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_check_str._ascii.ob_base,
            &_Py_ID(decode),
            & const_str_encodekey._ascii.ob_base,
            &_Py_ID(data),
            &_Py_ID(key),
            &_Py_ID(value),
            &_Py_ID(encode),
            &_Py_ID(encoding),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[9];
    }
os_toplevel_consts_101_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 8,
    },
    .ob_shash = -1,
    .ob_sval = "      @@",
};
static
    struct _PyCode_DEF(246)
os_toplevel_consts_101 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 123,
    },
    .co_consts = & os_toplevel_consts_101_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_101_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 15 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 774,
    .co_nlocalsplus = 8,
    .co_nlocals = 6,
    .co_ncellvars = 2,
    .co_nfreevars = 0,
    .co_version = 642,
    .co_localsplusnames = & os_toplevel_consts_101_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_101_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__createenviron._ascii.ob_base,
    .co_qualname = & const_str__createenviron._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_101_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x06\x87\x07\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x3a\x64\x02\x84\x00\x7d\x00\x7c\x00\x8a\x06\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x88\x06\x66\x01\x64\x03\x84\x08\x7d\x02\x69\x00\x7d\x03\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x10\x00\x00\x5c\x02\x00\x00\x7d\x04\x7d\x05\x7c\x05\x7c\x03\x02\x00\x7c\x02\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x8c\x12\x04\x00\x6e\x26\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x8a\x07\x88\x07\x66\x01\x64\x04\x84\x08\x8a\x06\x88\x07\x66\x01\x64\x05\x84\x08\x7d\x01\x89\x06\x7d\x02\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x02\x7c\x01\x89\x06\x7c\x01\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[170];
    }
os_toplevel_consts_102_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 169,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x47\x65\x74\x20\x61\x6e\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x2c\x20\x72\x65\x74\x75\x72\x6e\x20\x4e\x6f\x6e\x65\x20\x69\x66\x20\x69\x74\x20\x64\x6f\x65\x73\x6e\x27\x74\x20\x65\x78\x69\x73\x74\x2e\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x73\x65\x63\x6f\x6e\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x63\x61\x6e\x20\x73\x70\x65\x63\x69\x66\x79\x20\x61\x6e\x20\x61\x6c\x74\x65\x72\x6e\x61\x74\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x2e\x0a\x20\x20\x20\x20\x6b\x65\x79\x2c\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x20\x61\x72\x65\x20\x73\x74\x72\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_102_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_102_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_102_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_environ._ascii.ob_base,
            &_Py_ID(get),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_getenv = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getenv",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
os_toplevel_consts_102_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x08\x00\x0c\x13\x8f\x3b\x89\x3b\x90\x73\x98\x47\xd3\x0b\x24\xd0\x04\x24",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_102_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(key),
            &_Py_ID(default),
        },
    },
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_102 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & os_toplevel_consts_102_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_102_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 808,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 643,
    .co_localsplusnames = & os_toplevel_consts_102_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getenv._ascii.ob_base,
    .co_qualname = & const_str_getenv._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_102_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_103 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_getenv._ascii.ob_base,
            & const_str_supports_bytes_environ._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
os_toplevel_consts_104_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bytes expected, not %s",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_104_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_104_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_104_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__check_bytes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_check_bytes",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[45];
    }
os_toplevel_consts_104_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 44,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x25\xa4\x15\xd4\x0f\x27\xdc\x12\x1b\xd0\x1c\x34\xb4\x74\xb8\x45\xb3\x7b\xd7\x37\x4b\xd1\x37\x4b\xd1\x1c\x4b\xd3\x12\x4c\xd0\x0c\x4c\xd8\x0f\x14\x88\x0c",
};
static
    struct _PyCode_DEF(104)
os_toplevel_consts_104 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 52,
    },
    .co_consts = & os_toplevel_consts_104_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_104_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 818,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 644,
    .co_localsplusnames = & os_toplevel_consts_101_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__check_bytes._ascii.ob_base,
    .co_qualname = & const_str__check_bytes._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_104_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x21\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[180];
    }
os_toplevel_consts_105_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 179,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x47\x65\x74\x20\x61\x6e\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x2c\x20\x72\x65\x74\x75\x72\x6e\x20\x4e\x6f\x6e\x65\x20\x69\x66\x20\x69\x74\x20\x64\x6f\x65\x73\x6e\x27\x74\x20\x65\x78\x69\x73\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x65\x20\x6f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x73\x65\x63\x6f\x6e\x64\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x63\x61\x6e\x20\x73\x70\x65\x63\x69\x66\x79\x20\x61\x6e\x20\x61\x6c\x74\x65\x72\x6e\x61\x74\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x6b\x65\x79\x2c\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x61\x6e\x64\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x20\x61\x72\x65\x20\x62\x79\x74\x65\x73\x2e",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_105_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_105_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_105_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_environb._ascii.ob_base,
            &_Py_ID(get),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_getenvb = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getenvb",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
os_toplevel_consts_105_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x08\x00\x10\x18\x8f\x7c\x89\x7c\x98\x43\xa0\x17\xd3\x0f\x29\xd0\x08\x29",
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_105 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & os_toplevel_consts_105_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_105_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 829,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 645,
    .co_localsplusnames = & os_toplevel_consts_102_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_getenvb._ascii.ob_base,
    .co_qualname = & const_str_getenvb._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_105_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_106 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_environb._ascii.ob_base,
            & const_str_getenvb._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[280];
    }
os_toplevel_consts_107_consts_1_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 279,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x6e\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x28\x61\x6e\x20\x6f\x73\x2e\x50\x61\x74\x68\x4c\x69\x6b\x65\x2c\x20\x62\x79\x74\x65\x73\x2c\x20\x6f\x72\x20\x73\x74\x72\x29\x20\x74\x6f\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x73\x79\x73\x74\x65\x6d\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x77\x69\x74\x68\x20\x27\x73\x75\x72\x72\x6f\x67\x61\x74\x65\x65\x73\x63\x61\x70\x65\x27\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x65\x72\x2c\x20\x72\x65\x74\x75\x72\x6e\x20\x62\x79\x74\x65\x73\x20\x75\x6e\x63\x68\x61\x6e\x67\x65\x64\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x4f\x6e\x20\x57\x69\x6e\x64\x6f\x77\x73\x2c\x20\x75\x73\x65\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x65\x72\x20\x69\x66\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x73\x79\x73\x74\x65\x6d\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x69\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x27\x6d\x62\x63\x73\x27\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x74\x68\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_107_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_107_consts_1_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_107_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            &_Py_ID(encode),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
os_toplevel_consts_107_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fscodec.<locals>.fsencode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[48];
    }
os_toplevel_consts_107_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 47,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf4\x0c\x00\x14\x1a\x98\x28\xd3\x13\x23\x88\x08\xdc\x0b\x15\x90\x68\xa4\x03\xd4\x0b\x24\xd8\x13\x1b\x97\x3f\x91\x3f\xa0\x38\xa8\x56\xd3\x13\x34\xd0\x0c\x34\xe0\x13\x1b\x88\x4f",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_107_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(filename),
            &_Py_ID(encoding),
            &_Py_ID(errors),
        },
    },
};
static
    struct _PyCode_DEF(98)
os_toplevel_consts_107_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & os_toplevel_consts_107_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_107_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 841,
    .co_nlocalsplus = 3,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 2,
    .co_version = 646,
    .co_localsplusnames = & os_toplevel_consts_107_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_fsencode._ascii.ob_base,
    .co_qualname = & os_toplevel_consts_107_consts_1_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_107_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x02\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x12\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\x89\x02\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[280];
    }
os_toplevel_consts_107_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 279,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x65\x63\x6f\x64\x65\x20\x66\x69\x6c\x65\x6e\x61\x6d\x65\x20\x28\x61\x6e\x20\x6f\x73\x2e\x50\x61\x74\x68\x4c\x69\x6b\x65\x2c\x20\x62\x79\x74\x65\x73\x2c\x20\x6f\x72\x20\x73\x74\x72\x29\x20\x66\x72\x6f\x6d\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x73\x79\x73\x74\x65\x6d\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x77\x69\x74\x68\x20\x27\x73\x75\x72\x72\x6f\x67\x61\x74\x65\x65\x73\x63\x61\x70\x65\x27\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x65\x72\x2c\x20\x72\x65\x74\x75\x72\x6e\x20\x73\x74\x72\x20\x75\x6e\x63\x68\x61\x6e\x67\x65\x64\x2e\x20\x4f\x6e\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x57\x69\x6e\x64\x6f\x77\x73\x2c\x20\x75\x73\x65\x20\x27\x73\x74\x72\x69\x63\x74\x27\x20\x65\x72\x72\x6f\x72\x20\x68\x61\x6e\x64\x6c\x65\x72\x20\x69\x66\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x20\x73\x79\x73\x74\x65\x6d\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x20\x69\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x27\x6d\x62\x63\x73\x27\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x74\x68\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x65\x6e\x63\x6f\x64\x69\x6e\x67\x29\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_107_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_107_consts_2_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_107_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(bytes),
            &_Py_ID(decode),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
os_toplevel_consts_107_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fscodec.<locals>.fsdecode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[48];
    }
os_toplevel_consts_107_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 47,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf4\x0c\x00\x14\x1a\x98\x28\xd3\x13\x23\x88\x08\xdc\x0b\x15\x90\x68\xa4\x05\xd4\x0b\x26\xd8\x13\x1b\x97\x3f\x91\x3f\xa0\x38\xa8\x56\xd3\x13\x34\xd0\x0c\x34\xe0\x13\x1b\x88\x4f",
};
static
    struct _PyCode_DEF(98)
os_toplevel_consts_107_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 49,
    },
    .co_consts = & os_toplevel_consts_107_consts_2_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_107_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 853,
    .co_nlocalsplus = 3,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 2,
    .co_version = 647,
    .co_localsplusnames = & os_toplevel_consts_107_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_fsdecode._ascii.ob_base,
    .co_qualname = & os_toplevel_consts_107_consts_2_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_107_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x02\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x12\x7c\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\x89\x02\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_107_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_107_consts_1.ob_base.ob_base,
            & os_toplevel_consts_107_consts_2.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
const_str_getfilesystemencodeerrors = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getfilesystemencodeerrors",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_107_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_getfilesystemencoding._ascii.ob_base,
            & const_str_getfilesystemencodeerrors._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str__fscodec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fscodec",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[55];
    }
os_toplevel_consts_107_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 54,
    },
    .ob_shash = -1,
    .ob_sval = "\xf9\x80\x00\xdc\x0f\x12\xd7\x0f\x28\xd1\x0f\x28\xd3\x0f\x2a\x80\x48\xdc\x0d\x10\xd7\x0d\x2a\xd1\x0d\x2a\xd3\x0d\x2c\x80\x46\xf5\x04\x0a\x05\x1c\xf5\x18\x0a\x05\x1c\xf0\x18\x00\x0c\x14\x90\x58\xd0\x0b\x1d\xd0\x04\x1d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_107_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_fsencode._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            &_Py_ID(encoding),
            &_Py_ID(errors),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[5];
    }
os_toplevel_consts_107_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 4,
    },
    .ob_shash = -1,
    .ob_sval = "  @@",
};
static
    struct _PyCode_DEF(118)
os_toplevel_consts_107 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 59,
    },
    .co_consts = & os_toplevel_consts_107_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_107_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 837,
    .co_nlocalsplus = 4,
    .co_nlocals = 2,
    .co_ncellvars = 2,
    .co_nfreevars = 0,
    .co_version = 648,
    .co_localsplusnames = & os_toplevel_consts_107_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_107_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__fscodec._ascii.ob_base,
    .co_qualname = & const_str__fscodec._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_107_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x02\x87\x03\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x8a\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x8a\x03\x88\x02\x88\x03\x66\x02\x64\x01\x84\x08\x7d\x00\x88\x02\x88\x03\x66\x02\x64\x02\x84\x08\x7d\x01\x7c\x00\x7c\x01\x66\x02\x53\x00",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_fork = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fork",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_spawnv = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnv",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_P_WAIT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "P_WAIT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_P_NOWAIT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "P_NOWAIT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_P_NOWAITO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "P_NOWAITO",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_111 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_P_WAIT._ascii.ob_base,
            & const_str_P_NOWAIT._ascii.ob_base,
            & const_str_P_NOWAITO._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
os_toplevel_consts_112_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "argv must be a tuple or a list",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[35];
    }
os_toplevel_consts_112_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 34,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "argv first element cannot be empty",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_112_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_112_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & os_toplevel_consts_112_consts_3._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 127],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_waitpid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "waitpid",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_WIFSTOPPED = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "WIFSTOPPED",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str_waitstatus_to_exitcode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "waitstatus_to_exitcode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
os_toplevel_consts_112_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_tuple._ascii.ob_base,
            & const_str_list._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_fork._ascii.ob_base,
            & const_str__exit._ascii.ob_base,
            & const_str_P_NOWAIT._ascii.ob_base,
            & const_str_waitpid._ascii.ob_base,
            & const_str_WIFSTOPPED._ascii.ob_base,
            & const_str_waitstatus_to_exitcode._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__spawnvef = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_spawnvef",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[165];
    }
os_toplevel_consts_112_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 164,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x19\x98\x24\xa4\x15\xac\x04\xa0\x0d\xd4\x0f\x2e\xdc\x12\x1b\xd0\x1c\x3c\xd3\x12\x3d\xd0\x0c\x3d\xd9\x0f\x13\x98\x34\xa0\x01\x9a\x37\xdc\x12\x1c\xd0\x1d\x41\xd3\x12\x42\xd0\x0c\x42\xdc\x0e\x12\x8b\x66\x88\x03\xd9\x0f\x12\xf0\x04\x06\x0d\x1b\xd8\x13\x16\x90\x3b\xd9\x14\x18\x98\x14\x98\x74\xd5\x14\x24\xe1\x14\x18\x98\x14\x98\x74\xa0\x53\xd5\x14\x29\xf0\x05\x00\x15\x25\xf0\x0e\x00\x10\x14\x94\x78\xd2\x0f\x1f\xd8\x17\x1a\x90\x0a\xd8\x12\x13\xdc\x1c\x23\xa0\x43\xa8\x11\x9b\x4f\x91\x09\x90\x04\x90\x63\xdc\x13\x1d\x98\x63\x94\x3f\xd8\x14\x1c\xe4\x17\x2d\xa8\x63\xd3\x17\x32\xd0\x10\x32\xf8\xf0\x17\x01\x0d\x1b\xdc\x10\x15\x90\x63\x96\x0a\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
os_toplevel_consts_112_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x01\x16\x42\x0b\x00\xc2\x0b\x0d\x42\x1a\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_wpid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "wpid",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_sts = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sts",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
os_toplevel_consts_112_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(mode),
            &_Py_ID(file),
            &_Py_ID(args),
            &_Py_ID(env),
            &_Py_ID(func),
            &_Py_ID(pid),
            & const_str_wpid._ascii.ob_base,
            & const_str_sts._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(314)
os_toplevel_consts_112 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 157,
    },
    .co_consts = & os_toplevel_consts_112_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_112_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_112_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 5,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 882,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 649,
    .co_localsplusnames = & os_toplevel_consts_112_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__spawnvef._ascii.ob_base,
    .co_qualname = & const_str__spawnvef._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_112_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\xab\x02\x00\x00\x00\x00\x00\x00\x73\x0b\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x72\x05\x7c\x02\x64\x02\x19\x00\x00\x00\x73\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x73\x19\x09\x00\x7c\x03\x80\x0a\x02\x00\x7c\x04\x7c\x01\x7c\x02\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x0b\x02\x00\x7c\x04\x7c\x01\x7c\x02\x7c\x03\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x79\x00\x7c\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x72\x02\x7c\x05\x53\x00\x09\x00\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x06\x7d\x07\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x72\x01\x8c\x1c\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x01\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x79\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[278];
    }
os_toplevel_consts_113_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 277,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x76\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x61\x72\x67\x73\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x20\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_113_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_113_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_113_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__spawnvef._ascii.ob_base,
            & const_str_execv._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
os_toplevel_consts_113_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x10\x19\x98\x14\x98\x74\xa0\x54\xa8\x34\xb4\x15\xd3\x0f\x37\xd0\x08\x37",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_113_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(mode),
            &_Py_ID(file),
            &_Py_ID(args),
        },
    },
};
static
    struct _PyCode_DEF(40)
os_toplevel_consts_113 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & os_toplevel_consts_113_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_113_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 909,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 650,
    .co_localsplusnames = & os_toplevel_consts_113_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnv._ascii.ob_base,
    .co_qualname = & const_str_spawnv._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_113_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x64\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[315];
    }
os_toplevel_consts_114_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 314,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x76\x65\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x61\x72\x67\x73\x2c\x20\x65\x6e\x76\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x20\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x0a\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_114_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_114_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_114_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__spawnvef._ascii.ob_base,
            & const_str_execve._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_spawnve = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnve",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
os_toplevel_consts_114_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x10\x19\x98\x14\x98\x74\xa0\x54\xa8\x33\xb4\x06\xd3\x0f\x37\xd0\x08\x37",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_114_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(mode),
            &_Py_ID(file),
            &_Py_ID(args),
            &_Py_ID(env),
        },
    },
};
static
    struct _PyCode_DEF(40)
os_toplevel_consts_114 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & os_toplevel_consts_114_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_114_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 918,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 651,
    .co_localsplusnames = & os_toplevel_consts_114_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnve._ascii.ob_base,
    .co_qualname = & const_str_spawnve._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_114_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7c\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[313];
    }
os_toplevel_consts_115_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 312,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x76\x70\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x61\x72\x67\x73\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x6c\x6f\x6f\x6b\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x0a\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_115_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & os_toplevel_consts_115_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_115_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__spawnvef._ascii.ob_base,
            & const_str_execvp._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_spawnvp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnvp",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
os_toplevel_consts_115_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x10\x19\x98\x14\x98\x74\xa0\x54\xa8\x34\xb4\x16\xd3\x0f\x38\xd0\x08\x38",
};
static
    struct _PyCode_DEF(40)
os_toplevel_consts_115 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & os_toplevel_consts_115_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_115_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 930,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 652,
    .co_localsplusnames = & os_toplevel_consts_113_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnvp._ascii.ob_base,
    .co_qualname = & const_str_spawnvp._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_115_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x64\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[349];
    }
os_toplevel_consts_116_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 348,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x76\x70\x65\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x61\x72\x67\x73\x2c\x20\x65\x6e\x76\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x6c\x6f\x6f\x6b\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x0a\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x73\x75\x70\x70\x6c\x69\x65\x64\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_116_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_116_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_116_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__spawnvef._ascii.ob_base,
            & const_str_execvpe._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_spawnvpe = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnvpe",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
os_toplevel_consts_116_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x10\x19\x98\x14\x98\x74\xa0\x54\xa8\x33\xb4\x07\xd3\x0f\x38\xd0\x08\x38",
};
static
    struct _PyCode_DEF(40)
os_toplevel_consts_116 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & os_toplevel_consts_116_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_116_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 940,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 653,
    .co_localsplusnames = & os_toplevel_consts_114_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnvpe._ascii.ob_base,
    .co_qualname = & const_str_spawnvpe._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_116_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7c\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_117 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_spawnv._ascii.ob_base,
            & const_str_spawnve._ascii.ob_base,
            & const_str_spawnvp._ascii.ob_base,
            & const_str_spawnvpe._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[279];
    }
os_toplevel_consts_118_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 278,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x6c\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x20\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_118_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_118_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_118_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_spawnv._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_spawnl = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnl",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
os_toplevel_consts_118_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x10\x16\x90\x64\x98\x44\xa0\x24\xd3\x0f\x27\xd0\x08\x27",
};
static
    struct _PyCode_DEF(28)
os_toplevel_consts_118 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & os_toplevel_consts_118_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_118_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 958,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 654,
    .co_localsplusnames = & os_toplevel_consts_113_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnl._ascii.ob_base,
    .co_qualname = & const_str_spawnl._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_118_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[315];
    }
os_toplevel_consts_119_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 314,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x6c\x65\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x2c\x20\x65\x6e\x76\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x20\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x0a\x73\x75\x70\x70\x6c\x69\x65\x64\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_119_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & os_toplevel_consts_119_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_119_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_spawnve._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_spawnle = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnle",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
os_toplevel_consts_119_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x10\x00\x0f\x13\x90\x32\x89\x68\x88\x03\xdc\x0f\x16\x90\x74\x98\x54\xa0\x34\xa8\x03\xa8\x12\xa0\x39\xa8\x63\xd3\x0f\x32\xd0\x08\x32",
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_119 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & os_toplevel_consts_119_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_119_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 967,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 655,
    .co_localsplusnames = & os_toplevel_consts_114_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnle._ascii.ob_base,
    .co_qualname = & const_str_spawnle._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_119_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x64\x01\x19\x00\x00\x00\x7d\x03\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x64\x02\x64\x01\x1a\x00\x7c\x03\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[344];
    }
os_toplevel_consts_123_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 343,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x6c\x70\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x6c\x6f\x6f\x6b\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x0a\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x73\x75\x70\x70\x6c\x69\x65\x64\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_123_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_123_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_123_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_spawnvp._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_spawnlp = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnlp",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
os_toplevel_consts_123_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x10\x17\x90\x74\x98\x54\xa0\x34\xd3\x0f\x28\xd0\x08\x28",
};
static
    struct _PyCode_DEF(28)
os_toplevel_consts_123 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & os_toplevel_consts_123_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_123_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 985,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 656,
    .co_localsplusnames = & os_toplevel_consts_113_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnlp._ascii.ob_base,
    .co_qualname = & const_str_spawnlp._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_123_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[350];
    }
os_toplevel_consts_124_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 349,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x73\x70\x61\x77\x6e\x6c\x70\x65\x28\x6d\x6f\x64\x65\x2c\x20\x66\x69\x6c\x65\x2c\x20\x2a\x61\x72\x67\x73\x2c\x20\x65\x6e\x76\x29\x20\x2d\x3e\x20\x69\x6e\x74\x65\x67\x65\x72\x0a\x0a\x45\x78\x65\x63\x75\x74\x65\x20\x66\x69\x6c\x65\x20\x28\x77\x68\x69\x63\x68\x20\x69\x73\x20\x6c\x6f\x6f\x6b\x65\x64\x20\x66\x6f\x72\x20\x61\x6c\x6f\x6e\x67\x20\x24\x50\x41\x54\x48\x29\x20\x77\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x66\x72\x6f\x6d\x0a\x61\x72\x67\x73\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x70\x72\x6f\x63\x65\x73\x73\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x73\x75\x70\x70\x6c\x69\x65\x64\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x4e\x4f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x69\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x2e\x0a\x49\x66\x20\x6d\x6f\x64\x65\x20\x3d\x3d\x20\x50\x5f\x57\x41\x49\x54\x20\x72\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x63\x65\x73\x73\x27\x73\x20\x65\x78\x69\x74\x20\x63\x6f\x64\x65\x20\x69\x66\x20\x69\x74\x20\x65\x78\x69\x74\x73\x20\x6e\x6f\x72\x6d\x61\x6c\x6c\x79\x3b\x0a\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x72\x65\x74\x75\x72\x6e\x20\x2d\x53\x49\x47\x2c\x20\x77\x68\x65\x72\x65\x20\x53\x49\x47\x20\x69\x73\x20\x74\x68\x65\x20\x73\x69\x67\x6e\x61\x6c\x20\x74\x68\x61\x74\x20\x6b\x69\x6c\x6c\x65\x64\x20\x69\x74\x2e\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_124_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & os_toplevel_consts_124_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_124_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_spawnvpe._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_spawnlpe = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "spawnlpe",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
os_toplevel_consts_124_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x10\x00\x0f\x13\x90\x32\x89\x68\x88\x03\xdc\x0f\x17\x98\x04\x98\x64\xa0\x44\xa8\x13\xa8\x22\xa0\x49\xa8\x73\xd3\x0f\x33\xd0\x08\x33",
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_124 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & os_toplevel_consts_124_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_124_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 995,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 657,
    .co_localsplusnames = & os_toplevel_consts_114_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_spawnlpe._ascii.ob_base,
    .co_qualname = & const_str_spawnlpe._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_124_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x64\x01\x19\x00\x00\x00\x7d\x03\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x64\x02\x64\x01\x1a\x00\x7c\x03\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
os_toplevel_consts_128_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "invalid cmd type (%s, expected string)",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_128_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            (PyObject *)&_Py_SINGLETON(strings).ascii[119],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
os_toplevel_consts_128_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "invalid mode %r",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[44];
    }
os_toplevel_consts_128_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 43,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "popen() does not support unbuffered streams",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_shell = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "shell",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_128_consts_8 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_shell._ascii.ob_base,
            &_Py_ID(text),
            &_Py_ID(stdout),
            &_Py_ID(bufsize),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_128_consts_9 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_shell._ascii.ob_base,
            &_Py_ID(text),
            &_Py_ID(stdin),
            &_Py_ID(bufsize),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
os_toplevel_consts_128_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_128_consts_1._ascii.ob_base,
            & os_toplevel_consts_128_consts_2._object.ob_base.ob_base,
            & os_toplevel_consts_128_consts_3._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & os_toplevel_consts_128_consts_5._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            Py_True,
            & os_toplevel_consts_128_consts_8._object.ob_base.ob_base,
            & os_toplevel_consts_128_consts_9._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_subprocess = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "subprocess",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_Popen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Popen",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_PIPE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PIPE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__wrap_close = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap_close",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
os_toplevel_consts_128_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            & const_str_ValueError._ascii.ob_base,
            & const_str_subprocess._ascii.ob_base,
            & const_str_Popen._ascii.ob_base,
            & const_str_PIPE._ascii.ob_base,
            & const_str__wrap_close._ascii.ob_base,
            &_Py_ID(stdout),
            &_Py_ID(stdin),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_popen = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "popen",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[207];
    }
os_toplevel_consts_128_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 206,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x19\x98\x23\x9c\x73\xd4\x0f\x23\xdc\x12\x1b\xd0\x1c\x44\xc4\x74\xc8\x43\xc3\x79\xd1\x1c\x50\xd3\x12\x51\xd0\x0c\x51\xd8\x0b\x0f\x90\x7a\xd1\x0b\x21\xdc\x12\x1c\xd0\x1d\x2e\xb0\x14\xd1\x1d\x35\xd3\x12\x36\xd0\x0c\x36\xd8\x0b\x14\x98\x01\x8a\x3e\x98\x59\xd0\x1d\x2e\xdc\x12\x1c\xd0\x1d\x4a\xd3\x12\x4b\xd0\x0c\x4b\xdb\x08\x19\xd8\x0b\x0f\x90\x33\x8a\x3b\xd8\x13\x1d\xd7\x13\x23\xd1\x13\x23\xa0\x43\xd8\x2a\x2e\xb0\x54\xd8\x2b\x35\xaf\x3f\xa9\x3f\xd8\x2c\x35\xf0\x07\x00\x14\x24\xf3\x00\x03\x14\x37\x88\x44\xf4\x08\x00\x14\x1f\x98\x74\x9f\x7b\x99\x7b\xa8\x44\xd3\x13\x31\xd0\x0c\x31\xe0\x13\x1d\xd7\x13\x23\xd1\x13\x23\xa0\x43\xd8\x2a\x2e\xb0\x54\xd8\x2a\x34\xaf\x2f\xa9\x2f\xd8\x2c\x35\xf0\x07\x00\x14\x24\xf3\x00\x03\x14\x37\x88\x44\xf4\x08\x00\x14\x1f\x98\x74\x9f\x7a\x99\x7a\xa8\x34\xd3\x13\x30\xd0\x0c\x30",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_cmd = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "cmd",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_proc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "proc",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_128_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_cmd._ascii.ob_base,
            &_Py_ID(mode),
            &_Py_ID(buffering),
            & const_str_subprocess._ascii.ob_base,
            & const_str_proc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(386)
os_toplevel_consts_128 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 193,
    },
    .co_consts = & os_toplevel_consts_128_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_128_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1013,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 658,
    .co_localsplusnames = & os_toplevel_consts_128_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_popen._ascii.ob_base,
    .co_qualname = & const_str_popen._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_128_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x17\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x01\x64\x02\x76\x01\x72\x0e\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7c\x01\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x02\x64\x04\x6b\x28\x00\x00\x73\x02\x7c\x02\x80\x0b\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x64\x04\x64\x00\x6c\x05\x7d\x03\x7c\x01\x64\x06\x6b\x28\x00\x00\x72\x36\x7c\x03\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x07\x64\x07\x7c\x03\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xac\x08\xab\x05\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x03\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x07\x64\x07\x7c\x03\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xac\x09\xab\x05\x00\x00\x00\x00\x00\x00\x7d\x04\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__stream = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_stream",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__proc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_proc",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_129_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__stream._ascii.ob_base,
            & const_str__proc._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
os_toplevel_consts_129_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap_close.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
os_toplevel_consts_129_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x1b\x21\x88\x44\x8c\x4c\xd8\x19\x1d\x88\x44\x8d\x4a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_129_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_stream._ascii.ob_base,
            & const_str_proc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(32)
os_toplevel_consts_129_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_129_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1036,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 659,
    .co_localsplusnames = & os_toplevel_consts_129_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & os_toplevel_consts_129_consts_1_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_129_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_ID(nt),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_wait = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "wait",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_129_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__stream._ascii.ob_base,
            &_Py_ID(close),
            & const_str__proc._ascii.ob_base,
            & const_str_wait._ascii.ob_base,
            &_Py_ID(name),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
os_toplevel_consts_129_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap_close.close",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[68];
    }
os_toplevel_consts_129_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 67,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0c\x10\x8f\x4c\x89\x4c\xd7\x0c\x1e\xd1\x0c\x1e\xd4\x0c\x20\xd8\x19\x1d\x9f\x1a\x99\x1a\x9f\x1f\x99\x1f\xd3\x19\x2a\x88\x4a\xd8\x0f\x19\x98\x51\x8a\x7f\xd8\x17\x1b\xdc\x0f\x13\x90\x74\x8a\x7c\xd8\x17\x21\xd0\x10\x21\xe0\x17\x21\xa0\x51\x91\x7f\xd0\x10\x26",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_returncode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "returncode",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_129_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_returncode._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(150)
os_toplevel_consts_129_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 75,
    },
    .co_consts = & os_toplevel_consts_129_consts_2_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_129_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1039,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 660,
    .co_localsplusnames = & os_toplevel_consts_129_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(close),
    .co_qualname = & os_toplevel_consts_129_consts_2_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x64\x01\x6b\x28\x00\x00\x72\x01\x79\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x28\x00\x00\x72\x02\x7c\x01\x53\x00\x7c\x01\x64\x03\x7a\x03\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
os_toplevel_consts_129_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap_close.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[8];
    }
os_toplevel_consts_129_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 7,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x13\x17\x88\x4b",
};
static
    struct _PyCode_DEF(6)
os_toplevel_consts_129_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1048,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 661,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & os_toplevel_consts_129_consts_3_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_129_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(close),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
os_toplevel_consts_129_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap_close.__exit__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
os_toplevel_consts_129_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0c\x10\x8f\x4a\x89\x4a\x8d\x4c",
};
static
    struct _PyCode_DEF(36)
os_toplevel_consts_129_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 18,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_129_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1050,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 662,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & os_toplevel_consts_129_consts_4_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_129_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(getattr),
            & const_str__stream._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
os_toplevel_consts_129_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap_close.__getattr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[20];
    }
os_toplevel_consts_129_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 19,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x13\x1a\x98\x34\x9f\x3c\x99\x3c\xa8\x14\xd3\x13\x2e\xd0\x0c\x2e",
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_129_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_129_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1052,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 663,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_14_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getattr__),
    .co_qualname = & os_toplevel_consts_129_consts_5_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_129_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(iter),
            & const_str__stream._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
os_toplevel_consts_129_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_wrap_close.__iter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
os_toplevel_consts_129_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x13\x17\x98\x04\x9f\x0c\x99\x0c\xd3\x13\x25\xd0\x0c\x25",
};
static
    struct _PyCode_DEF(44)
os_toplevel_consts_129_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 22,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_129_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1054,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 664,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__iter__),
    .co_qualname = & os_toplevel_consts_129_consts_6_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
os_toplevel_consts_129_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str__wrap_close._ascii.ob_base,
            & os_toplevel_consts_129_consts_1.ob_base.ob_base,
            & os_toplevel_consts_129_consts_2.ob_base.ob_base,
            & os_toplevel_consts_129_consts_3.ob_base.ob_base,
            & os_toplevel_consts_129_consts_4.ob_base.ob_base,
            & os_toplevel_consts_129_consts_5.ob_base.ob_base,
            & os_toplevel_consts_129_consts_6.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
os_toplevel_consts_129_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            &_Py_ID(close),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
            &_Py_ID(__getattr__),
            &_Py_ID(__iter__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
os_toplevel_consts_129_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x02\x02\x09\x1e\xf2\x06\x08\x09\x27\xf2\x12\x01\x09\x18\xf2\x04\x01\x09\x19\xf2\x04\x01\x09\x2f\xf3\x04\x01\x09\x26",
};
static
    struct _PyCode_DEF(48)
os_toplevel_consts_129 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 24,
    },
    .co_consts = & os_toplevel_consts_129_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_129_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1035,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 665,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__wrap_close._ascii.ob_base,
    .co_qualname = & const_str__wrap_close._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x64\x06\x84\x00\x5a\x08\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[39];
    }
os_toplevel_consts_132_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 38,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "invalid fd type (%s, expected integer)",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_132_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_132_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_Py_SINGLETON(strings).ascii[98],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
os_toplevel_consts_132_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_int._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(type),
            & const_str_io._ascii.ob_base,
            & const_str_text_encoding._ascii.ob_base,
            &_Py_ID(open),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[93];
    }
os_toplevel_consts_132_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 92,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0b\x15\x90\x62\x9c\x23\xd4\x0b\x1e\xdc\x0e\x17\xd0\x18\x40\xc4\x34\xc8\x02\xc3\x38\xd1\x18\x4b\xd3\x0e\x4c\xd0\x08\x4c\xdb\x04\x0d\xd8\x07\x0a\x90\x24\x81\x7f\xd8\x13\x15\xd7\x13\x23\xd1\x13\x23\xa0\x48\xd3\x13\x2d\x88\x08\xd8\x0b\x12\x88\x32\x8f\x37\x89\x37\x90\x32\x90\x74\x98\x59\xa8\x08\xd0\x0b\x42\xb0\x34\xd2\x0b\x42\xb8\x36\xd1\x0b\x42\xd0\x04\x42",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
os_toplevel_consts_132_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(fd),
            &_Py_ID(mode),
            &_Py_ID(buffering),
            &_Py_ID(encoding),
            &_Py_ID(args),
            & const_str_kwargs._ascii.ob_base,
            & const_str_io._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(180)
os_toplevel_consts_132 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 90,
    },
    .co_consts = & os_toplevel_consts_132_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_132_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 15,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 1060,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 666,
    .co_localsplusnames = & os_toplevel_consts_132_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_fdopen._ascii.ob_base,
    .co_qualname = & const_str_fdopen._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_132_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x73\x17\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x64\x02\x64\x00\x6c\x04\x7d\x06\x64\x03\x7c\x01\x76\x01\x72\x11\x7c\x06\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x02\x00\x7c\x06\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\x7c\x02\x7c\x03\x67\x04\x7c\x04\xa2\x01\xad\x06\x69\x00\x7c\x05\xa4\x01\x8e\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[354];
    }
os_toplevel_consts_133_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 353,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x61\x20\x70\x61\x74\x68\x2d\x6c\x69\x6b\x65\x20\x6f\x62\x6a\x65\x63\x74\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x73\x74\x72\x20\x6f\x72\x20\x62\x79\x74\x65\x73\x20\x69\x73\x20\x70\x61\x73\x73\x65\x64\x20\x69\x6e\x2c\x20\x69\x74\x20\x69\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x75\x6e\x63\x68\x61\x6e\x67\x65\x64\x2e\x20\x4f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x6f\x73\x2e\x50\x61\x74\x68\x4c\x69\x6b\x65\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x69\x73\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x67\x65\x74\x20\x74\x68\x65\x20\x70\x61\x74\x68\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x61\x74\x69\x6f\x6e\x2e\x20\x49\x66\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x70\x61\x74\x68\x20\x72\x65\x70\x72\x65\x73\x65\x6e\x74\x61\x74\x69\x6f\x6e\x20\x69\x73\x20\x6e\x6f\x74\x20\x73\x74\x72\x20\x6f\x72\x20\x62\x79\x74\x65\x73\x2c\x20\x54\x79\x70\x65\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x20\x49\x66\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x70\x72\x6f\x76\x69\x64\x65\x64\x20\x70\x61\x74\x68\x20\x69\x73\x20\x6e\x6f\x74\x20\x73\x74\x72\x2c\x20\x62\x79\x74\x65\x73\x2c\x20\x6f\x72\x20\x6f\x73\x2e\x50\x61\x74\x68\x4c\x69\x6b\x65\x2c\x20\x54\x79\x70\x65\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[48];
    }
os_toplevel_consts_133_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 47,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "expected str, bytes or os.PathLike object, not ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[56];
    }
os_toplevel_consts_133_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 55,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "expected {}.__fspath__() to return str or bytes, not {}",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_133_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & os_toplevel_consts_133_consts_0._ascii.ob_base,
            &_Py_ID(__fspath__),
            & os_toplevel_consts_133_consts_2._ascii.ob_base,
            & os_toplevel_consts_133_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
os_toplevel_consts_133_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(isinstance),
            & const_str_str._ascii.ob_base,
            &_Py_ID(bytes),
            &_Py_ID(type),
            &_Py_ID(__fspath__),
            & const_str_AttributeError._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            &_Py_ID(__name__),
            &_Py_ID(format),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__fspath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_fspath",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[192];
    }
os_toplevel_consts_133_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 191,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x08\x12\x90\x24\x9c\x13\x9c\x65\x98\x0c\xd4\x07\x25\xd8\x0f\x13\x88\x0b\xf4\x08\x00\x11\x15\x90\x54\x93\x0a\x80\x49\xf0\x02\x07\x05\x39\xd8\x14\x1d\xd7\x14\x28\xd1\x14\x28\xa8\x14\xd3\x14\x2e\x88\x09\xf4\x0e\x00\x08\x12\x90\x29\x9c\x63\xa4\x35\x98\x5c\xd4\x07\x2a\xd8\x0f\x18\xd0\x08\x18\xe4\x0e\x17\xf0\x00\x01\x19\x21\xdf\x21\x27\xa1\x16\xa8\x09\xd7\x28\x3a\xd1\x28\x3a\xdc\x28\x2c\xa8\x59\xab\x0f\xd7\x28\x40\xd1\x28\x40\xf3\x03\x01\x22\x42\x01\xf3\x03\x02\x0f\x43\x01\xf0\x00\x02\x09\x43\x01\xf8\xf4\x13\x00\x0c\x1a\xf2\x00\x05\x05\x39\xdc\x0b\x12\x90\x39\x98\x6c\xd4\x0b\x2b\xd8\x0c\x11\xe4\x12\x1b\xf0\x00\x01\x1d\x23\xd8\x25\x2e\xd7\x25\x37\xd1\x25\x37\xf1\x03\x01\x1d\x38\xf3\x00\x01\x13\x39\xf0\x00\x01\x0d\x39\xf0\x09\x05\x05\x39\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
os_toplevel_consts_133_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "\xa5\x11\x42\x06\x00\xc2\x06\x2f\x42\x35\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_path_type = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_type",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_path_repr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_repr",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_133_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str_path_type._ascii.ob_base,
            & const_str_path_repr._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(368)
os_toplevel_consts_133 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 184,
    },
    .co_consts = & os_toplevel_consts_133_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_133_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_consts_133_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 1071,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 667,
    .co_localsplusnames = & os_toplevel_consts_133_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__fspath._ascii.ob_base,
    .co_qualname = & const_str__fspath._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_133_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\xab\x02\x00\x00\x00\x00\x00\x00\x72\x02\x7c\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x7c\x01\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\xab\x02\x00\x00\x00\x00\x00\x00\x72\x02\x7c\x02\x53\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x26\x01\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x72\x01\x82\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x7c\x01\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_PathLike = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathLike",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[68];
    }
os_toplevel_consts_135_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 67,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Abstract base class for implementing the file system path protocol.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[58];
    }
os_toplevel_consts_135_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 57,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the file system path representation of the object.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
os_toplevel_consts_135_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & os_toplevel_consts_135_consts_2_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
os_toplevel_consts_135_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathLike.__fspath__",
};
static
    struct _PyCode_DEF(14)
os_toplevel_consts_135_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 7,
    },
    .co_consts = & os_toplevel_consts_135_consts_2_consts._object.ob_base.ob_base,
    .co_names = & codecs_toplevel_consts_14_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1111,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 668,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__fspath__),
    .co_qualname = & os_toplevel_consts_135_consts_2_qualname._ascii.ob_base,
    .co_linetable = & codecs_toplevel_consts_18_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x82\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_135_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__fspath__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_135_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_PathLike._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            & const_str_NotImplemented._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
os_toplevel_consts_135_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PathLike.__subclasshook__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
os_toplevel_consts_135_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0e\x94\x28\x89\x3f\xdc\x13\x21\xa0\x28\xa8\x4c\xd3\x13\x39\xd0\x0c\x39\xdc\x0f\x1d\xd0\x08\x1d",
};
static
    struct _PyCode_DEF(54)
os_toplevel_consts_135_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & os_toplevel_consts_135_consts_3_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_135_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1116,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 669,
    .co_localsplusnames = & abc_toplevel_consts_10_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__subclasshook__),
    .co_qualname = & os_toplevel_consts_135_consts_3_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_135_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
os_toplevel_consts_135_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_PathLike._ascii.ob_base,
            & os_toplevel_consts_135_consts_1._ascii.ob_base,
            & os_toplevel_consts_135_consts_2.ob_base.ob_base,
            & os_toplevel_consts_135_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
os_toplevel_consts_135_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_abc._ascii.ob_base,
            & const_str_abstractmethod._ascii.ob_base,
            &_Py_ID(__fspath__),
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(__subclasshook__),
            & const_str_GenericAlias._ascii.ob_base,
            &_Py_ID(__class_getitem__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
os_toplevel_consts_135_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x4d\xe0\x05\x08\xd7\x05\x17\xd1\x05\x17\xf1\x02\x02\x05\x22\xf3\x03\x00\x06\x18\xf0\x02\x02\x05\x22\xf0\x08\x00\x06\x11\xf1\x02\x03\x05\x1e\xf3\x03\x00\x06\x11\xf0\x02\x03\x05\x1e\xf1\x0a\x00\x19\x24\xa0\x4c\xd3\x18\x31\xd1\x04\x15",
};
static
    struct _PyCode_DEF(84)
os_toplevel_consts_135 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 42,
    },
    .co_consts = & os_toplevel_consts_135_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_135_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1107,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 670,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_PathLike._ascii.ob_base,
    .co_qualname = & const_str_PathLike._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_135_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x06\x65\x07\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x02\x00\x65\x07\x65\x09\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x0a\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__AddedDllDirectory = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_AddedDllDirectory",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__cookie = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_cookie",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str__remove_dll_directory = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_remove_dll_directory",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_137_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(path),
            & const_str__cookie._ascii.ob_base,
            & const_str__remove_dll_directory._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
os_toplevel_consts_137_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_AddedDllDirectory.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
os_toplevel_consts_137_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x18\x1c\x88\x44\x8c\x49\xd8\x1b\x21\x88\x44\x8c\x4c\xd8\x29\x3d\x88\x44\xd5\x0c\x26",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_remove_dll_directory = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "remove_dll_directory",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_137_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(path),
            &_Py_ID(cookie),
            & const_str_remove_dll_directory._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(46)
os_toplevel_consts_137_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_137_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1127,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 671,
    .co_localsplusnames = & os_toplevel_consts_137_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & os_toplevel_consts_137_consts_1_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_137_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_137_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__remove_dll_directory._ascii.ob_base,
            & const_str__cookie._ascii.ob_base,
            &_Py_ID(path),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
os_toplevel_consts_137_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_AddedDllDirectory.close",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
os_toplevel_consts_137_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0c\x10\xd7\x0c\x26\xd1\x0c\x26\xa0\x74\xa7\x7c\xa1\x7c\xd4\x0c\x34\xd8\x18\x1c\x88\x44\x8d\x49",
};
static
    struct _PyCode_DEF(72)
os_toplevel_consts_137_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 36,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_137_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1131,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 672,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(close),
    .co_qualname = & os_toplevel_consts_137_consts_2_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_137_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
os_toplevel_consts_137_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_AddedDllDirectory.__enter__",
};
static
    struct _PyCode_DEF(6)
os_toplevel_consts_137_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 3,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1134,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 673,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & os_toplevel_consts_137_consts_3_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
os_toplevel_consts_137_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_AddedDllDirectory.__exit__",
};
static
    struct _PyCode_DEF(36)
os_toplevel_consts_137_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 18,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_129_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1136,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 674,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & os_toplevel_consts_137_consts_4_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_129_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
os_toplevel_consts_137_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<AddedDllDirectory({!r})>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
os_toplevel_consts_137_consts_5_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<AddedDllDirectory()>",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_137_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & os_toplevel_consts_137_consts_5_consts_1._ascii.ob_base,
            & os_toplevel_consts_137_consts_5_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_137_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(format),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
os_toplevel_consts_137_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_AddedDllDirectory.__repr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[34];
    }
os_toplevel_consts_137_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 33,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\x8f\x79\x8a\x79\xd8\x17\x32\xd7\x17\x39\xd1\x17\x39\xb8\x24\xbf\x29\xb9\x29\xd3\x17\x44\xd0\x10\x44\xd8\x13\x2a",
};
static
    struct _PyCode_DEF(82)
os_toplevel_consts_137_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 41,
    },
    .co_consts = & os_toplevel_consts_137_consts_5_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_137_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1138,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 675,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_ID(__repr__),
    .co_qualname = & os_toplevel_consts_137_consts_5_qualname._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_137_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x1b\x64\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
os_toplevel_consts_137_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__AddedDllDirectory._ascii.ob_base,
            & os_toplevel_consts_137_consts_1.ob_base.ob_base,
            & os_toplevel_consts_137_consts_2.ob_base.ob_base,
            & os_toplevel_consts_137_consts_3.ob_base.ob_base,
            & os_toplevel_consts_137_consts_4.ob_base.ob_base,
            & os_toplevel_consts_137_consts_5.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
os_toplevel_consts_137_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__init__),
            &_Py_ID(close),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
            &_Py_ID(__repr__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[28];
    }
os_toplevel_consts_137_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 27,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x02\x03\x09\x3e\xf2\x08\x02\x09\x1d\xf2\x06\x01\x09\x18\xf2\x04\x01\x09\x19\xf3\x04\x03\x09\x2b",
};
static
    struct _PyCode_DEF(42)
os_toplevel_consts_137 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 21,
    },
    .co_consts = & os_toplevel_consts_137_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_137_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 1126,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 676,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str__AddedDllDirectory._ascii.ob_base,
    .co_qualname = & const_str__AddedDllDirectory._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_137_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x64\x05\x84\x00\x5a\x07\x79\x06",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[336];
    }
os_toplevel_consts_139_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 335,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x64\x64\x20\x61\x20\x70\x61\x74\x68\x20\x74\x6f\x20\x74\x68\x65\x20\x44\x4c\x4c\x20\x73\x65\x61\x72\x63\x68\x20\x70\x61\x74\x68\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x54\x68\x69\x73\x20\x73\x65\x61\x72\x63\x68\x20\x70\x61\x74\x68\x20\x69\x73\x20\x75\x73\x65\x64\x20\x77\x68\x65\x6e\x20\x72\x65\x73\x6f\x6c\x76\x69\x6e\x67\x20\x64\x65\x70\x65\x6e\x64\x65\x6e\x63\x69\x65\x73\x20\x66\x6f\x72\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x28\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x74\x73\x65\x6c\x66\x20\x69\x73\x20\x72\x65\x73\x6f\x6c\x76\x65\x64\x20\x74\x68\x72\x6f\x75\x67\x68\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x29\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x61\x6e\x64\x20\x61\x6c\x73\x6f\x20\x62\x79\x20\x63\x74\x79\x70\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x52\x65\x6d\x6f\x76\x65\x20\x74\x68\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x62\x79\x20\x63\x61\x6c\x6c\x69\x6e\x67\x20\x63\x6c\x6f\x73\x65\x28\x29\x20\x6f\x6e\x20\x74\x68\x65\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x6f\x62\x6a\x65\x63\x74\x20\x6f\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x75\x73\x69\x6e\x67\x20\x69\x74\x20\x69\x6e\x20\x61\x20\x77\x69\x74\x68\x20\x73\x74\x61\x74\x65\x6d\x65\x6e\x74\x2e\x0a\x20\x20\x20\x20\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_139_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & os_toplevel_consts_139_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str__add_dll_directory = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_add_dll_directory",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
os_toplevel_consts_139_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(nt),
            & const_str__add_dll_directory._ascii.ob_base,
            & const_str__AddedDllDirectory._ascii.ob_base,
            & const_str__remove_dll_directory._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_add_dll_directory = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "add_dll_directory",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[54];
    }
os_toplevel_consts_139_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 53,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf3\x14\x00\x09\x12\xd8\x11\x26\x90\x12\xd7\x11\x26\xd1\x11\x26\xa0\x74\xd3\x11\x2c\x88\x06\xdc\x0f\x21\xd8\x0c\x10\xd8\x0c\x12\xd8\x0c\x0e\xd7\x0c\x24\xd1\x0c\x24\xf3\x07\x04\x10\x0a\xf0\x00\x04\x09\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_139_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(path),
            &_Py_ID(nt),
            &_Py_ID(cookie),
        },
    },
};
static
    struct _PyCode_DEF(92)
os_toplevel_consts_139 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 46,
    },
    .co_consts = & os_toplevel_consts_139_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_consts_139_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1143,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 677,
    .co_localsplusnames = & os_toplevel_consts_139_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = & const_str_add_dll_directory._ascii.ob_base,
    .co_qualname = & const_str_add_dll_directory._ascii.ob_base,
    .co_linetable = & os_toplevel_consts_139_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x7d\x01\x02\x00\x7c\x01\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\x7c\x01\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_511 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 511 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_140 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_int_511.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_141 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_True,
            Py_None,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_142 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            Py_True,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
os_toplevel_consts_144 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
os_toplevel_consts_145 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[146];
        }_object;
    }
os_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 146,
        },
        .ob_item = {
            & os_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & os_toplevel_consts_3._object.ob_base.ob_base,
            & os_toplevel_consts_4._object.ob_base.ob_base,
            & os_toplevel_consts_5.ob_base.ob_base,
            & os_toplevel_consts_6.ob_base.ob_base,
            &_Py_ID(posix),
            (PyObject *)&_Py_SINGLETON(strings).ascii[10],
            & codecs_toplevel_consts_3._object.ob_base.ob_base,
            & os_toplevel_consts_10._object.ob_base.ob_base,
            & const_str__exit._ascii.ob_base,
            & os_toplevel_consts_12._object.ob_base.ob_base,
            &_Py_ID(nt),
            & os_toplevel_consts_14._ascii.ob_base,
            & os_toplevel_consts_15._ascii.ob_base,
            & os_toplevel_consts_16._ascii.ob_base,
            & os_toplevel_consts_17._object.ob_base.ob_base,
            & const_str__have_functions._ascii.ob_base,
            & os_toplevel_consts_19.ob_base.ob_base,
            & const_str_HAVE_FACCESSAT._ascii.ob_base,
            &_Py_ID(access),
            & const_str_HAVE_FCHMODAT._ascii.ob_base,
            & const_str_chmod._ascii.ob_base,
            & const_str_HAVE_FCHOWNAT._ascii.ob_base,
            & const_str_chown._ascii.ob_base,
            & const_str_HAVE_FSTATAT._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_HAVE_FUTIMESAT._ascii.ob_base,
            & const_str_utime._ascii.ob_base,
            & const_str_HAVE_LINKAT._ascii.ob_base,
            & const_str_link._ascii.ob_base,
            & const_str_HAVE_MKDIRAT._ascii.ob_base,
            & const_str_mkdir._ascii.ob_base,
            & const_str_HAVE_MKFIFOAT._ascii.ob_base,
            & const_str_mkfifo._ascii.ob_base,
            & const_str_HAVE_MKNODAT._ascii.ob_base,
            & const_str_mknod._ascii.ob_base,
            & const_str_HAVE_OPENAT._ascii.ob_base,
            &_Py_ID(open),
            & const_str_HAVE_READLINKAT._ascii.ob_base,
            & const_str_readlink._ascii.ob_base,
            & const_str_HAVE_RENAMEAT._ascii.ob_base,
            & const_str_rename._ascii.ob_base,
            & const_str_HAVE_SYMLINKAT._ascii.ob_base,
            & const_str_symlink._ascii.ob_base,
            & const_str_HAVE_UNLINKAT._ascii.ob_base,
            &_Py_ID(unlink),
            & const_str_rmdir._ascii.ob_base,
            & const_str_HAVE_UTIMENSAT._ascii.ob_base,
            & const_str_HAVE_FCHDIR._ascii.ob_base,
            & const_str_chdir._ascii.ob_base,
            & const_str_HAVE_FCHMOD._ascii.ob_base,
            & const_str_HAVE_FCHOWN._ascii.ob_base,
            & const_str_HAVE_FDOPENDIR._ascii.ob_base,
            & const_str_listdir._ascii.ob_base,
            & const_str_scandir._ascii.ob_base,
            & const_str_HAVE_FEXECVE._ascii.ob_base,
            & const_str_execve._ascii.ob_base,
            & const_str_HAVE_FTRUNCATE._ascii.ob_base,
            &_Py_ID(truncate),
            & const_str_HAVE_FUTIMENS._ascii.ob_base,
            & const_str_HAVE_FUTIMES._ascii.ob_base,
            & const_str_HAVE_FPATHCONF._ascii.ob_base,
            & const_str_pathconf._ascii.ob_base,
            & const_str_statvfs._ascii.ob_base,
            & const_str_fstatvfs._ascii.ob_base,
            & const_str_HAVE_FSTATVFS._ascii.ob_base,
            & const_str_HAVE_LCHFLAGS._ascii.ob_base,
            & const_str_chflags._ascii.ob_base,
            & const_str_HAVE_LCHMOD._ascii.ob_base,
            & const_str_lchown._ascii.ob_base,
            & const_str_HAVE_LCHOWN._ascii.ob_base,
            & const_str_HAVE_LUTIMES._ascii.ob_base,
            & const_str_HAVE_LSTAT._ascii.ob_base,
            & const_str_MS_WINDOWS._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            Py_False,
            & os_toplevel_consts_79.ob_base.ob_base,
            & os_toplevel_consts_80.ob_base.ob_base,
            & os_toplevel_consts_81.ob_base.ob_base,
            & os_toplevel_consts_82._object.ob_base.ob_base,
            & os_toplevel_consts_83.ob_base.ob_base,
            & const_str_walk._ascii.ob_base,
            & os_toplevel_consts_85._object.ob_base.ob_base,
            & os_toplevel_consts_86.ob_base.ob_base,
            & os_toplevel_consts_87.ob_base.ob_base,
            & const_str_fwalk._ascii.ob_base,
            & os_toplevel_consts_89.ob_base.ob_base,
            & os_toplevel_consts_90.ob_base.ob_base,
            & os_toplevel_consts_91.ob_base.ob_base,
            & os_toplevel_consts_92.ob_base.ob_base,
            & os_toplevel_consts_93.ob_base.ob_base,
            & os_toplevel_consts_94.ob_base.ob_base,
            & os_toplevel_consts_95._object.ob_base.ob_base,
            & os_toplevel_consts_96.ob_base.ob_base,
            & os_toplevel_consts_97.ob_base.ob_base,
            & os_toplevel_consts_98._object.ob_base.ob_base,
            & os_toplevel_consts_99.ob_base.ob_base,
            & const_str__Environ._ascii.ob_base,
            & os_toplevel_consts_101.ob_base.ob_base,
            & os_toplevel_consts_102.ob_base.ob_base,
            & os_toplevel_consts_103._object.ob_base.ob_base,
            & os_toplevel_consts_104.ob_base.ob_base,
            & os_toplevel_consts_105.ob_base.ob_base,
            & os_toplevel_consts_106._object.ob_base.ob_base,
            & os_toplevel_consts_107.ob_base.ob_base,
            & const_str_fork._ascii.ob_base,
            & const_str_spawnv._ascii.ob_base,
            & const_str_execv._ascii.ob_base,
            & os_toplevel_consts_111._object.ob_base.ob_base,
            & os_toplevel_consts_112.ob_base.ob_base,
            & os_toplevel_consts_113.ob_base.ob_base,
            & os_toplevel_consts_114.ob_base.ob_base,
            & os_toplevel_consts_115.ob_base.ob_base,
            & os_toplevel_consts_116.ob_base.ob_base,
            & os_toplevel_consts_117._object.ob_base.ob_base,
            & os_toplevel_consts_118.ob_base.ob_base,
            & os_toplevel_consts_119.ob_base.ob_base,
            & const_str_spawnl._ascii.ob_base,
            & const_str_spawnle._ascii.ob_base,
            & const_str_spawnvp._ascii.ob_base,
            & os_toplevel_consts_123.ob_base.ob_base,
            & os_toplevel_consts_124.ob_base.ob_base,
            & const_str_spawnlp._ascii.ob_base,
            & const_str_spawnlpe._ascii.ob_base,
            & const_str_vxworks._ascii.ob_base,
            & os_toplevel_consts_128.ob_base.ob_base,
            & os_toplevel_consts_129.ob_base.ob_base,
            & const_str__wrap_close._ascii.ob_base,
            & const_str_popen._ascii.ob_base,
            & os_toplevel_consts_132.ob_base.ob_base,
            & os_toplevel_consts_133.ob_base.ob_base,
            & const_str_fspath._ascii.ob_base,
            & os_toplevel_consts_135.ob_base.ob_base,
            & const_str_PathLike._ascii.ob_base,
            & os_toplevel_consts_137.ob_base.ob_base,
            & const_str__AddedDllDirectory._ascii.ob_base,
            & os_toplevel_consts_139.ob_base.ob_base,
            & os_toplevel_consts_140._object.ob_base.ob_base,
            & os_toplevel_consts_141._object.ob_base.ob_base,
            & os_toplevel_consts_142._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
            & os_toplevel_consts_144._object.ob_base.ob_base,
            & os_toplevel_consts_145._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__collections_abc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_collections_abc",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__names = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_names",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_posixpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "posixpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_ntpath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ntpath",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_supports_dir_fd = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "supports_dir_fd",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str_supports_effective_ids = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "supports_effective_ids",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_supports_fd = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "supports_fd",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str_supports_follow_symlinks = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "supports_follow_symlinks",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[110];
        }_object;
    }
os_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 110,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_abc._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_st._ascii.ob_base,
            & const_str__collections_abc._ascii.ob_base,
            & const_str__check_methods._ascii.ob_base,
            &_Py_ID(type),
            & const_str_list._ascii.ob_base,
            & const_str_int._ascii.ob_base,
            & const_str_GenericAlias._ascii.ob_base,
            & const_str_builtin_module_names._ascii.ob_base,
            & const_str__names._ascii.ob_base,
            &_Py_ID(__all__),
            & const_str__exists._ascii.ob_base,
            & const_str__get_exports_list._ascii.ob_base,
            &_Py_ID(name),
            & const_str_linesep._ascii.ob_base,
            &_Py_ID(posix),
            & const_str__exit._ascii.ob_base,
            &_Py_ID(append),
            & const_str_ImportError._ascii.ob_base,
            & const_str_posixpath._ascii.ob_base,
            &_Py_ID(path),
            & const_str__have_functions._ascii.ob_base,
            &_Py_ID(extend),
            &_Py_ID(nt),
            & const_str_ntpath._ascii.ob_base,
            &_Py_ID(modules),
            & os_toplevel_consts_16._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            & const_str_pardir._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_pathsep._ascii.ob_base,
            & const_str_defpath._ascii.ob_base,
            & const_str_extsep._ascii.ob_base,
            & const_str_altsep._ascii.ob_base,
            & const_str_devnull._ascii.ob_base,
            &_Py_ID(globals),
            & const_str__globals._ascii.ob_base,
            & const_str__add._ascii.ob_base,
            & const_str_set._ascii.ob_base,
            & const_str__set._ascii.ob_base,
            & const_str_supports_dir_fd._ascii.ob_base,
            & const_str_supports_effective_ids._ascii.ob_base,
            &_Py_ID(add),
            & const_str_supports_fd._ascii.ob_base,
            & const_str_supports_follow_symlinks._ascii.ob_base,
            & const_str_SEEK_SET._ascii.ob_base,
            & const_str_SEEK_CUR._ascii.ob_base,
            & const_str_SEEK_END._ascii.ob_base,
            & const_str_makedirs._ascii.ob_base,
            & const_str_removedirs._ascii.ob_base,
            & const_str_renames._ascii.ob_base,
            &_Py_ID(object),
            & const_str__walk_symlinks_as_files._ascii.ob_base,
            & const_str_walk._ascii.ob_base,
            &_Py_ID(open),
            & const_str_scandir._ascii.ob_base,
            & const_str_fwalk._ascii.ob_base,
            & const_str__fwalk_walk._ascii.ob_base,
            & const_str__fwalk_yield._ascii.ob_base,
            & const_str__fwalk_close._ascii.ob_base,
            & const_str__fwalk._ascii.ob_base,
            & const_str_execl._ascii.ob_base,
            & const_str_execle._ascii.ob_base,
            & const_str_execlp._ascii.ob_base,
            & const_str_execlpe._ascii.ob_base,
            & const_str_execvp._ascii.ob_base,
            & const_str_execvpe._ascii.ob_base,
            & const_str__execvpe._ascii.ob_base,
            & const_str_get_exec_path._ascii.ob_base,
            & const_str_MutableMapping._ascii.ob_base,
            & const_str_Mapping._ascii.ob_base,
            & const_str__Environ._ascii.ob_base,
            & const_str__createenviron._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            & const_str_getenv._ascii.ob_base,
            & const_str_supports_bytes_environ._ascii.ob_base,
            & const_str__check_bytes._ascii.ob_base,
            & const_str__data._ascii.ob_base,
            &_Py_ID(bytes),
            & const_str_environb._ascii.ob_base,
            & const_str_getenvb._ascii.ob_base,
            & const_str__fscodec._ascii.ob_base,
            & const_str_fsencode._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            & const_str_P_WAIT._ascii.ob_base,
            & const_str_P_NOWAIT._ascii.ob_base,
            & const_str_P_NOWAITO._ascii.ob_base,
            & const_str__spawnvef._ascii.ob_base,
            & const_str_spawnv._ascii.ob_base,
            & const_str_spawnve._ascii.ob_base,
            & const_str_spawnvp._ascii.ob_base,
            & const_str_spawnvpe._ascii.ob_base,
            & const_str_spawnl._ascii.ob_base,
            & const_str_spawnle._ascii.ob_base,
            & const_str_spawnlp._ascii.ob_base,
            & const_str_spawnlpe._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str_popen._ascii.ob_base,
            & const_str__wrap_close._ascii.ob_base,
            & const_str_fdopen._ascii.ob_base,
            & const_str__fspath._ascii.ob_base,
            & const_str_fspath._ascii.ob_base,
            &_Py_ID(__name__),
            & const_str_ABC._ascii.ob_base,
            & const_str_PathLike._ascii.ob_base,
            & const_str__AddedDllDirectory._ascii.ob_base,
            & const_str_add_dll_directory._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[1542];
    }
os_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 1541,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x15\x01\x04\xf3\x30\x00\x01\x0b\xdb\x00\x0a\xdb\x00\x11\xe5\x00\x2b\xe1\x0f\x13\x90\x44\x98\x13\x91\x49\x8b\x7f\x80\x0c\xe0\x09\x0c\xd7\x09\x21\xd1\x09\x21\x80\x06\xf2\x06\x03\x0b\x15\x80\x07\xf2\x0a\x01\x01\x1d\xf2\x06\x04\x01\x37\xf0\x10\x00\x04\x0b\x88\x66\xd1\x03\x14\xd8\x0b\x12\x80\x44\xd8\x0e\x12\x80\x47\xdc\x04\x17\xf0\x02\x04\x05\x0d\xdd\x08\x1f\xd8\x08\x0f\x8f\x0e\x89\x0e\x90\x77\xd4\x08\x1f\xf3\x06\x00\x05\x1d\xf0\x04\x03\x05\x0d\xdd\x08\x29\xf3\x08\x00\x05\x11\xd8\x04\x0b\x87\x4e\x81\x4e\xd1\x13\x24\xa0\x55\xd3\x13\x2b\xd4\x04\x2c\xd9\x08\x0d\xe0\x05\x09\x88\x56\x81\x5e\xd8\x0b\x0f\x80\x44\xd8\x0e\x14\x80\x47\xdc\x04\x14\xf0\x02\x04\x05\x0d\xdd\x08\x1c\xd8\x08\x0f\x8f\x0e\x89\x0e\x90\x77\xd4\x08\x1f\xf3\x06\x00\x05\x1a\xe3\x04\x0d\xd8\x04\x0b\x87\x4e\x81\x4e\xd1\x13\x24\xa0\x52\xd3\x13\x28\xd4\x04\x29\xd8\x08\x0a\xf0\x04\x03\x05\x0d\xde\x08\x26\xf1\x0a\x00\x0b\x16\xd0\x16\x33\xd3\x0a\x34\xd0\x04\x34\xe0\x19\x1d\x80\x03\x87\x0b\x81\x0b\x88\x49\xd1\x00\x16\xf7\x02\x01\x01\x0d\xf7\x00\x01\x01\x0d\xf3\x00\x01\x01\x0d\xf0\x06\x00\x05\x0b\xf1\x06\x00\x04\x0b\xd0\x0b\x1c\xd5\x03\x1d\xd9\x0f\x16\x8b\x79\x80\x48\xf2\x02\x02\x05\x23\xf1\x08\x00\x0c\x0f\x8b\x35\x80\x44\xd9\x04\x08\xd0\x09\x19\x98\x48\xd4\x04\x25\xd9\x04\x08\x88\x1f\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1f\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1e\x98\x46\xd4\x04\x23\xd9\x04\x08\xd0\x09\x19\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1d\x98\x46\xd4\x04\x23\xd9\x04\x08\x88\x1e\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1f\x98\x48\xd4\x04\x25\xd9\x04\x08\x88\x1e\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1d\x98\x46\xd4\x04\x23\xd9\x04\x08\xd0\x09\x1a\x98\x4a\xd4\x04\x27\xd9\x04\x08\x88\x1f\x98\x48\xd4\x04\x25\xd9\x04\x08\xd0\x09\x19\x98\x49\xd4\x04\x26\xd9\x04\x08\x88\x1f\x98\x48\xd4\x04\x25\xd9\x04\x08\x88\x1f\x98\x47\xd4\x04\x24\xd9\x04\x08\xd0\x09\x19\x98\x47\xd4\x04\x24\xd8\x16\x1a\x80\x4f\xe1\x0b\x0e\x8b\x35\x80\x44\xd9\x04\x08\xd0\x09\x19\x98\x48\xd4\x04\x25\xd8\x1d\x21\xd0\x04\x1a\xe1\x0b\x0e\x8b\x35\x80\x44\xd9\x04\x08\x88\x1d\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1d\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1d\x98\x47\xd4\x04\x24\xd9\x04\x08\xd0\x09\x19\x98\x49\xd4\x04\x26\xd9\x04\x08\xd0\x09\x19\x98\x49\xd4\x04\x26\xd9\x04\x08\x88\x1e\x98\x48\xd4\x04\x25\xd8\x04\x08\x87\x48\x81\x48\x88\x54\x84\x4e\xd9\x04\x08\xd0\x09\x19\x98\x4a\xd4\x04\x27\xd9\x04\x08\x88\x1f\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1e\x98\x47\xd4\x04\x24\xd9\x04\x08\xd0\x09\x19\x98\x4a\xd4\x04\x27\xd9\x07\x0e\x88\x79\xd4\x07\x19\x99\x67\xa0\x6a\xd4\x1e\x31\xd9\x08\x0c\x88\x5f\x98\x69\xd4\x08\x28\xd8\x12\x16\x80\x4b\xe1\x0b\x0e\x8b\x35\x80\x44\xd9\x04\x08\xd0\x09\x19\x98\x48\xd4\x04\x25\xf1\x2c\x00\x05\x09\x88\x1f\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1e\x98\x46\xd4\x04\x23\xd9\x04\x08\x88\x1f\x98\x49\xd4\x04\x26\xd9\x04\x08\x88\x1d\x98\x47\xd4\x04\x24\xd9\x07\x0e\x88\x78\xd4\x07\x18\xd9\x08\x0c\x88\x5d\x98\x47\xd4\x08\x24\xd9\x04\x08\x88\x1d\x98\x46\xd4\x04\x23\xd9\x04\x08\x88\x1e\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1c\x98\x46\xd4\x04\x23\xd9\x04\x08\x88\x1e\x98\x46\xd4\x04\x23\xd9\x04\x08\xd0\x09\x19\x98\x47\xd4\x04\x24\xd9\x04\x08\x88\x1c\x98\x46\xd4\x04\x23\xd8\x1f\x23\xd0\x04\x1c\xe0\x08\x0c\xd8\x08\x17\xd8\x08\x10\xd8\x08\x0c\xf0\x0c\x00\x0c\x0d\x80\x08\xd8\x0b\x0c\x80\x08\xd8\x0b\x0c\x80\x08\xf3\x0a\x1e\x01\x12\xf2\x40\x01\x14\x01\x26\xf2\x2c\x18\x01\x11\xf0\x34\x00\x01\x08\x87\x0e\x81\x0e\xd2\x0f\x34\xd4\x00\x35\xf1\x08\x00\x1b\x21\x9b\x28\xd0\x00\x17\xf3\x04\x58\x02\x01\x27\xf0\x74\x04\x00\x01\x08\x87\x0e\x81\x0e\x88\x76\xd4\x00\x16\xe0\x04\x08\x88\x24\x80\x3c\x90\x3f\xd2\x03\x22\xa8\x07\xb0\x14\xa0\x7f\xb8\x2b\xd2\x27\x45\xf0\x04\x2d\x05\x21\xc0\x65\xd0\x54\x58\xf4\x00\x2d\x05\x21\xf0\x60\x01\x00\x13\x14\x80\x4b\xd8\x13\x14\x80\x4c\xd8\x13\x14\x80\x4c\xf2\x04\x48\x01\x05\x43\x01\xf0\x54\x02\x00\x05\x0c\x87\x4e\x81\x4e\x90\x37\xd4\x04\x1b\xf2\x04\x05\x01\x16\xf2\x0e\x06\x01\x21\xf2\x10\x05\x01\x17\xf2\x0e\x07\x01\x22\xf2\x12\x06\x01\x19\xf2\x10\x07\x01\x1e\xf0\x12\x00\x01\x08\x87\x0e\x81\x0e\xd2\x0f\x47\xd4\x00\x48\xf3\x04\x1d\x01\x13\xf3\x40\x01\x29\x01\x24\xf7\x5a\x01\x00\x01\x35\xf4\x04\x47\x01\x01\x13\x88\x7e\xf4\x00\x47\x01\x01\x13\xf2\x52\x02\x1b\x01\x18\xf1\x3c\x00\x0b\x19\xd3\x0a\x1a\x80\x07\xd8\x04\x12\xf3\x06\x04\x01\x25\xf0\x0c\x00\x1b\x1f\xa0\x24\x99\x2c\xd0\x00\x16\xd8\x00\x07\x87\x0e\x81\x0e\xd0\x0f\x33\xd4\x00\x34\xe1\x03\x19\xf2\x02\x03\x05\x15\xf1\x0c\x00\x10\x18\x98\x07\x9f\x0d\x99\x0d\xd8\x08\x14\x90\x65\xd8\x08\x14\x90\x65\xf3\x05\x02\x10\x1d\x80\x48\xf0\x06\x00\x09\x15\xf3\x04\x04\x05\x2a\xf0\x0c\x00\x05\x0c\x87\x4e\x81\x4e\xd0\x13\x2a\xd4\x04\x2b\xf2\x04\x1c\x01\x1e\xf1\x3c\x00\x16\x1e\x93\x5a\xd1\x00\x12\x80\x08\x88\x28\xd8\x04\x0c\xf1\x06\x00\x04\x0b\x88\x36\x84\x3f\x99\x37\xa0\x38\xd4\x1b\x2c\xb1\x17\xb8\x17\xd4\x31\x41\xe0\x0d\x0e\x80\x46\xd8\x1b\x1c\xd0\x04\x1c\x80\x48\x88\x79\xe0\x04\x0b\x87\x4e\x81\x4e\xd2\x13\x36\xd4\x04\x37\xf2\x0c\x19\x05\x33\xf2\x36\x07\x05\x38\xf2\x12\x08\x05\x38\xf2\x18\x08\x05\x39\xf2\x14\x08\x05\x39\xf0\x16\x00\x05\x0c\x87\x4e\x81\x4e\xd2\x13\x3f\xd4\x04\x40\xf1\x06\x00\x04\x0b\x88\x38\xd4\x03\x14\xf2\x08\x07\x05\x28\xf2\x12\x09\x05\x33\xf0\x18\x00\x05\x0c\x87\x4e\x81\x4e\x90\x48\x98\x69\xd0\x13\x28\xd4\x04\x29\xf1\x06\x00\x04\x0b\x88\x39\xd4\x03\x15\xf2\x06\x08\x05\x29\xf2\x14\x09\x05\x34\xf0\x18\x00\x05\x0c\x87\x4e\x81\x4e\x90\x49\x98\x7a\xd0\x13\x2a\xd4\x04\x2b\xf0\x08\x00\x04\x07\x87\x3c\x81\x3c\x90\x39\xd2\x03\x1c\xf3\x04\x13\x05\x31\xf7\x2c\x14\x05\x26\xf1\x00\x14\x05\x26\xf0\x2c\x00\x05\x0c\x87\x4e\x81\x4e\x90\x37\xd4\x04\x1b\xf3\x06\x06\x01\x43\x01\xf2\x16\x1b\x01\x43\x01\xf1\x3e\x00\x08\x0f\x88\x78\xd4\x07\x18\xd8\x0d\x14\x80\x46\xd8\x16\x1e\x80\x46\x84\x4f\xf4\x06\x0f\x01\x32\x88\x73\x8f\x77\x89\x77\xf4\x00\x0f\x01\x32\xf0\x24\x00\x04\x08\x88\x34\x82\x3c\xf7\x02\x0f\x05\x2b\xf1\x00\x0f\x05\x2b\xf3\x22\x10\x05\x0a\xf0\x25\x00\x04\x10\xf8\xf0\x55\x21\x00\x0c\x17\xf2\x00\x01\x05\x0d\xda\x08\x0c\xf0\x03\x01\x05\x0d\xfb\xf0\x0c\x00\x0c\x17\xf2\x00\x01\x05\x0d\xda\x08\x0c\xf0\x03\x01\x05\x0d\xfb\xf0\x1c\x00\x0c\x17\xf2\x00\x01\x05\x0d\xda\x08\x0c\xf0\x03\x01\x05\x0d\xfb\xf0\x14\x00\x0c\x17\xf2\x00\x01\x05\x0d\xda\x08\x0c\xf0\x03\x01\x05\x0d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[73];
    }
os_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 72,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x04\x17\x53\x1f\x00\xc1\x20\x06\x53\x2b\x00\xc2\x11\x17\x53\x37\x00\xc3\x09\x06\x54\x03\x00\xd3\x1f\x05\x53\x28\x03\xd3\x27\x01\x53\x28\x03\xd3\x2b\x05\x53\x34\x03\xd3\x33\x01\x53\x34\x03\xd3\x37\x05\x54\x00\x03\xd3\x3f\x01\x54\x00\x03\xd4\x03\x05\x54\x0c\x03\xd4\x0b\x01\x54\x0c\x03",
};
static
    struct _PyCode_DEF(2590)
os_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 1295,
    },
    .co_consts = & os_toplevel_consts._object.ob_base.ob_base,
    .co_names = & os_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & os_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 678,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & os_toplevel_consts_5_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & os_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x5a\x01\x64\x01\x64\x02\x6c\x02\x5a\x02\x64\x01\x64\x02\x6c\x03\x5a\x04\x64\x01\x64\x03\x6c\x05\x6d\x06\x5a\x06\x01\x00\x02\x00\x65\x07\x65\x08\x65\x09\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x0a\x65\x02\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5a\x0c\x67\x00\x64\x04\xa2\x01\x5a\x0d\x64\x05\x84\x00\x5a\x0e\x64\x06\x84\x00\x5a\x0f\x64\x07\x65\x0c\x76\x00\x72\x49\x64\x07\x5a\x10\x64\x08\x5a\x11\x64\x01\x64\x09\x6c\x12\xad\x02\x01\x00\x09\x00\x64\x01\x64\x0a\x6c\x12\x6d\x13\x5a\x13\x01\x00\x65\x0d\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x64\x02\x6c\x16\x5a\x17\x09\x00\x64\x01\x64\x0c\x6c\x12\x6d\x18\x5a\x18\x01\x00\x64\x01\x64\x02\x6c\x12\x5a\x12\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x65\x0f\x65\x12\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x5b\x12\x6e\x55\x64\x0d\x65\x0c\x76\x00\x72\x49\x64\x0d\x5a\x10\x64\x0e\x5a\x11\x64\x01\x64\x09\x6c\x1a\xad\x02\x01\x00\x09\x00\x64\x01\x64\x0a\x6c\x1a\x6d\x13\x5a\x13\x01\x00\x65\x0d\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x64\x02\x6c\x1b\x5a\x17\x64\x01\x64\x02\x6c\x1a\x5a\x1a\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x65\x0f\x65\x1a\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x5b\x1a\x09\x00\x64\x01\x64\x0c\x6c\x1a\x6d\x18\x5a\x18\x01\x00\x6e\x08\x02\x00\x65\x15\x64\x0f\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x65\x17\x65\x02\x6a\x38\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x10\x3c\x00\x00\x00\x64\x01\x64\x11\x6c\x1d\x6d\x1e\x5a\x1e\x6d\x1f\x5a\x1f\x6d\x20\x5a\x20\x6d\x21\x5a\x21\x6d\x22\x5a\x22\x6d\x23\x5a\x23\x6d\x24\x5a\x24\x6d\x25\x5a\x25\x01\x00\x5b\x0c\x02\x00\x65\x0e\x64\x12\xab\x01\x00\x00\x00\x00\x00\x00\x90\x01\x72\xc3\x02\x00\x65\x26\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x27\x64\x13\x84\x00\x5a\x28\x02\x00\x65\x29\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x2a\x02\x00\x65\x28\x64\x14\x64\x15\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x16\x64\x17\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x18\x64\x19\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x1a\x64\x1b\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x1c\x64\x1d\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x1e\x64\x1f\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x20\x64\x21\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x22\x64\x23\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x24\x64\x25\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x26\x64\x27\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x28\x64\x29\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x2a\x64\x2b\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x2c\x64\x2d\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x2e\x64\x2f\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x2e\x64\x30\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x31\x64\x1d\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x65\x2a\x5a\x2b\x02\x00\x65\x29\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x2a\x02\x00\x65\x28\x64\x14\x64\x15\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x65\x2a\x5a\x2c\x02\x00\x65\x29\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x2a\x02\x00\x65\x28\x64\x32\x64\x33\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x34\x64\x17\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x35\x64\x19\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x36\x64\x37\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x36\x64\x38\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x39\x64\x3a\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x65\x2a\x6a\x5b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x3b\x64\x3c\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x3d\x64\x1d\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x3e\x64\x1d\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x3f\x64\x40\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x0e\x64\x41\xab\x01\x00\x00\x00\x00\x00\x00\x72\x11\x02\x00\x65\x0e\x64\x42\xab\x01\x00\x00\x00\x00\x00\x00\x72\x09\x02\x00\x65\x28\x64\x43\x64\x41\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x65\x2a\x5a\x2e\x02\x00\x65\x29\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x2a\x02\x00\x65\x28\x64\x14\x64\x15\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x18\x64\x19\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x1a\x64\x1b\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x44\x64\x45\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x46\x64\x17\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x0e\x64\x47\xab\x01\x00\x00\x00\x00\x00\x00\x72\x09\x02\x00\x65\x28\x64\x48\x64\x19\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x1e\x64\x1f\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x49\x64\x1d\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x4a\x64\x1b\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x1a\x64\x1b\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x31\x64\x1d\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x28\x64\x4b\x64\x1b\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x65\x2a\x5a\x2f\x5b\x2a\x5b\x18\x5b\x27\x5b\x28\x64\x01\x5a\x30\x64\x4c\x5a\x31\x64\x4d\x5a\x32\x64\x8c\x64\x4f\x84\x01\x5a\x33\x64\x50\x84\x00\x5a\x34\x64\x51\x84\x00\x5a\x35\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x64\x52\xa2\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x36\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x37\x64\x8d\x64\x53\x84\x01\x5a\x38\x65\x0d\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x54\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x39\x65\x03\x68\x02\x65\x2b\x6b\x1a\x00\x00\x72\x29\x65\x3a\x65\x03\x68\x02\x65\x2e\x6b\x1a\x00\x00\x72\x22\x64\x8e\x64\x4e\x64\x02\x64\x55\x9c\x02\x64\x56\x84\x03\x5a\x3b\x64\x01\x5a\x3c\x64\x4c\x5a\x3d\x64\x4d\x5a\x3e\x64\x57\x84\x00\x5a\x3f\x65\x0d\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x58\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x59\x84\x00\x5a\x40\x64\x5a\x84\x00\x5a\x41\x64\x5b\x84\x00\x5a\x42\x64\x5c\x84\x00\x5a\x43\x64\x5d\x84\x00\x5a\x44\x64\x5e\x84\x00\x5a\x45\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x64\x5f\xa2\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x8f\x64\x60\x84\x01\x5a\x46\x64\x8f\x64\x61\x84\x01\x5a\x47\x64\x01\x64\x62\x6c\x05\x6d\x48\x5a\x48\x6d\x49\x5a\x49\x01\x00\x02\x00\x47\x00\x64\x63\x84\x00\x64\x64\x65\x48\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x4a\x64\x65\x84\x00\x5a\x4b\x02\x00\x65\x4b\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x4c\x5b\x4b\x64\x8f\x64\x66\x84\x01\x5a\x4d\x65\x10\x64\x0d\x6b\x37\x00\x00\x5a\x4e\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x67\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x4e\x72\x2f\x64\x68\x84\x00\x5a\x4f\x02\x00\x65\x4a\x65\x4c\x6a\xa0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x4f\x65\x51\x65\x4f\x65\x51\xab\x05\x00\x00\x00\x00\x00\x00\x5a\x52\x5b\x4f\x64\x8f\x64\x69\x84\x01\x5a\x53\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x6a\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x6b\x84\x00\x5a\x54\x02\x00\x65\x54\xab\x00\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x5a\x55\x5a\x56\x5b\x54\x02\x00\x65\x0e\x64\x6c\xab\x01\x00\x00\x00\x00\x00\x00\x72\x4b\x02\x00\x65\x0e\x64\x6d\xab\x01\x00\x00\x00\x00\x00\x00\x73\x43\x02\x00\x65\x0e\x64\x6e\xab\x01\x00\x00\x00\x00\x00\x00\x72\x3b\x64\x01\x5a\x57\x64\x4c\x78\x01\x5a\x58\x5a\x59\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x64\x6f\xa2\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x70\x84\x00\x5a\x5a\x64\x71\x84\x00\x5a\x5b\x64\x72\x84\x00\x5a\x5c\x64\x73\x84\x00\x5a\x5d\x64\x74\x84\x00\x5a\x5e\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x64\x75\xa2\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x0e\x64\x6d\xab\x01\x00\x00\x00\x00\x00\x00\x72\x19\x64\x76\x84\x00\x5a\x5f\x64\x77\x84\x00\x5a\x60\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x78\x64\x79\x67\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x65\x0e\x64\x7a\xab\x01\x00\x00\x00\x00\x00\x00\x72\x19\x64\x7b\x84\x00\x5a\x61\x64\x7c\x84\x00\x5a\x62\x65\x0d\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x7d\x64\x7e\x67\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x65\x02\x6a\xc6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x7f\x6b\x37\x00\x00\x72\x1f\x64\x90\x64\x80\x84\x01\x5a\x64\x02\x00\x47\x00\x64\x81\x84\x00\x64\x82\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x65\x65\x0d\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x83\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x91\x64\x84\x84\x01\x5a\x66\x64\x85\x84\x00\x5a\x67\x02\x00\x65\x0e\x64\x86\xab\x01\x00\x00\x00\x00\x00\x00\x73\x09\x65\x67\x5a\x68\x64\x86\x65\x68\x5f\x69\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x47\x00\x64\x87\x84\x00\x64\x88\x65\x01\x6a\xd4\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x6b\x65\x10\x64\x0d\x6b\x28\x00\x00\x72\x0e\x02\x00\x47\x00\x64\x89\x84\x00\x64\x8a\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x6c\x64\x8b\x84\x00\x5a\x6d\x79\x02\x79\x02\x23\x00\x65\x15\x24\x00\x72\x04\x01\x00\x59\x00\x90\x04\x8c\x8c\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x15\x24\x00\x72\x04\x01\x00\x59\x00\x90\x04\x8c\x8d\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x15\x24\x00\x72\x04\x01\x00\x59\x00\x90\x04\x8c\x57\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x65\x15\x24\x00\x72\x04\x01\x00\x59\x00\x90\x04\x8c\x33\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_os_toplevel(void)
{
    return Py_NewRef((PyObject *) &os_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[2999];
    }
site_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2998,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x70\x70\x65\x6e\x64\x20\x6d\x6f\x64\x75\x6c\x65\x20\x73\x65\x61\x72\x63\x68\x20\x70\x61\x74\x68\x73\x20\x66\x6f\x72\x20\x74\x68\x69\x72\x64\x2d\x70\x61\x72\x74\x79\x20\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x2e\x0a\x0a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x0a\x2a\x20\x54\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x61\x75\x74\x6f\x6d\x61\x74\x69\x63\x61\x6c\x6c\x79\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x20\x64\x75\x72\x69\x6e\x67\x20\x69\x6e\x69\x74\x69\x61\x6c\x69\x7a\x61\x74\x69\x6f\x6e\x2e\x20\x2a\x0a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x2a\x0a\x0a\x54\x68\x69\x73\x20\x77\x69\x6c\x6c\x20\x61\x70\x70\x65\x6e\x64\x20\x73\x69\x74\x65\x2d\x73\x70\x65\x63\x69\x66\x69\x63\x20\x70\x61\x74\x68\x73\x20\x74\x6f\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x73\x65\x61\x72\x63\x68\x20\x70\x61\x74\x68\x2e\x20\x20\x4f\x6e\x0a\x55\x6e\x69\x78\x20\x28\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x4d\x61\x63\x20\x4f\x53\x58\x29\x2c\x20\x69\x74\x20\x73\x74\x61\x72\x74\x73\x20\x77\x69\x74\x68\x20\x73\x79\x73\x2e\x70\x72\x65\x66\x69\x78\x20\x61\x6e\x64\x0a\x73\x79\x73\x2e\x65\x78\x65\x63\x5f\x70\x72\x65\x66\x69\x78\x20\x28\x69\x66\x20\x64\x69\x66\x66\x65\x72\x65\x6e\x74\x29\x20\x61\x6e\x64\x20\x61\x70\x70\x65\x6e\x64\x73\x0a\x6c\x69\x62\x2f\x70\x79\x74\x68\x6f\x6e\x3c\x76\x65\x72\x73\x69\x6f\x6e\x3e\x2f\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x2e\x0a\x4f\x6e\x20\x6f\x74\x68\x65\x72\x20\x70\x6c\x61\x74\x66\x6f\x72\x6d\x73\x20\x28\x73\x75\x63\x68\x20\x61\x73\x20\x57\x69\x6e\x64\x6f\x77\x73\x29\x2c\x20\x69\x74\x20\x74\x72\x69\x65\x73\x20\x65\x61\x63\x68\x20\x6f\x66\x20\x74\x68\x65\x0a\x70\x72\x65\x66\x69\x78\x65\x73\x20\x64\x69\x72\x65\x63\x74\x6c\x79\x2c\x20\x61\x73\x20\x77\x65\x6c\x6c\x20\x61\x73\x20\x77\x69\x74\x68\x20\x6c\x69\x62\x2f\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x61\x70\x70\x65\x6e\x64\x65\x64\x2e\x20\x20\x54\x68\x65\x0a\x72\x65\x73\x75\x6c\x74\x69\x6e\x67\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x2c\x20\x69\x66\x20\x74\x68\x65\x79\x20\x65\x78\x69\x73\x74\x2c\x20\x61\x72\x65\x20\x61\x70\x70\x65\x6e\x64\x65\x64\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x2c\x20\x61\x6e\x64\x0a\x61\x6c\x73\x6f\x20\x69\x6e\x73\x70\x65\x63\x74\x65\x64\x20\x66\x6f\x72\x20\x70\x61\x74\x68\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x20\x66\x69\x6c\x65\x73\x2e\x0a\x0a\x49\x66\x20\x61\x20\x66\x69\x6c\x65\x20\x6e\x61\x6d\x65\x64\x20\x22\x70\x79\x76\x65\x6e\x76\x2e\x63\x66\x67\x22\x20\x65\x78\x69\x73\x74\x73\x20\x6f\x6e\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x61\x62\x6f\x76\x65\x20\x73\x79\x73\x2e\x65\x78\x65\x63\x75\x74\x61\x62\x6c\x65\x2c\x0a\x73\x79\x73\x2e\x70\x72\x65\x66\x69\x78\x20\x61\x6e\x64\x20\x73\x79\x73\x2e\x65\x78\x65\x63\x5f\x70\x72\x65\x66\x69\x78\x20\x61\x72\x65\x20\x73\x65\x74\x20\x74\x6f\x20\x74\x68\x61\x74\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x61\x6e\x64\x0a\x69\x74\x20\x69\x73\x20\x61\x6c\x73\x6f\x20\x63\x68\x65\x63\x6b\x65\x64\x20\x66\x6f\x72\x20\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x28\x73\x79\x73\x2e\x62\x61\x73\x65\x5f\x70\x72\x65\x66\x69\x78\x20\x61\x6e\x64\x0a\x73\x79\x73\x2e\x62\x61\x73\x65\x5f\x65\x78\x65\x63\x5f\x70\x72\x65\x66\x69\x78\x20\x77\x69\x6c\x6c\x20\x61\x6c\x77\x61\x79\x73\x20\x62\x65\x20\x74\x68\x65\x20\x22\x72\x65\x61\x6c\x22\x20\x70\x72\x65\x66\x69\x78\x65\x73\x20\x6f\x66\x20\x74\x68\x65\x20\x50\x79\x74\x68\x6f\x6e\x0a\x69\x6e\x73\x74\x61\x6c\x6c\x61\x74\x69\x6f\x6e\x29\x2e\x20\x49\x66\x20\x22\x70\x79\x76\x65\x6e\x76\x2e\x63\x66\x67\x22\x20\x28\x61\x20\x62\x6f\x6f\x74\x73\x74\x72\x61\x70\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x20\x66\x69\x6c\x65\x29\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x0a\x74\x68\x65\x20\x6b\x65\x79\x20\x22\x69\x6e\x63\x6c\x75\x64\x65\x2d\x73\x79\x73\x74\x65\x6d\x2d\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x22\x20\x73\x65\x74\x20\x74\x6f\x20\x61\x6e\x79\x74\x68\x69\x6e\x67\x20\x6f\x74\x68\x65\x72\x20\x74\x68\x61\x6e\x20\x22\x66\x61\x6c\x73\x65\x22\x0a\x28\x63\x61\x73\x65\x2d\x69\x6e\x73\x65\x6e\x73\x69\x74\x69\x76\x65\x29\x2c\x20\x74\x68\x65\x20\x73\x79\x73\x74\x65\x6d\x2d\x6c\x65\x76\x65\x6c\x20\x70\x72\x65\x66\x69\x78\x65\x73\x20\x77\x69\x6c\x6c\x20\x73\x74\x69\x6c\x6c\x20\x61\x6c\x73\x6f\x20\x62\x65\x0a\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x3b\x20\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x74\x68\x65\x79\x20\x77\x6f\x6e\x27\x74\x2e\x0a\x0a\x41\x6c\x6c\x20\x6f\x66\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x69\x6e\x67\x20\x73\x69\x74\x65\x2d\x73\x70\x65\x63\x69\x66\x69\x63\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x2c\x20\x69\x66\x20\x74\x68\x65\x79\x20\x65\x78\x69\x73\x74\x2c\x20\x61\x72\x65\x0a\x61\x70\x70\x65\x6e\x64\x65\x64\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x2c\x20\x61\x6e\x64\x20\x61\x6c\x73\x6f\x20\x69\x6e\x73\x70\x65\x63\x74\x65\x64\x20\x66\x6f\x72\x20\x70\x61\x74\x68\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x0a\x66\x69\x6c\x65\x73\x2e\x0a\x0a\x41\x20\x70\x61\x74\x68\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x20\x66\x69\x6c\x65\x20\x69\x73\x20\x61\x20\x66\x69\x6c\x65\x20\x77\x68\x6f\x73\x65\x20\x6e\x61\x6d\x65\x20\x68\x61\x73\x20\x74\x68\x65\x20\x66\x6f\x72\x6d\x0a\x3c\x70\x61\x63\x6b\x61\x67\x65\x3e\x2e\x70\x74\x68\x3b\x20\x69\x74\x73\x20\x63\x6f\x6e\x74\x65\x6e\x74\x73\x20\x61\x72\x65\x20\x61\x64\x64\x69\x74\x69\x6f\x6e\x61\x6c\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x28\x6f\x6e\x65\x20\x70\x65\x72\x20\x6c\x69\x6e\x65\x29\x0a\x74\x6f\x20\x62\x65\x20\x61\x64\x64\x65\x64\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x2e\x20\x20\x4e\x6f\x6e\x2d\x65\x78\x69\x73\x74\x69\x6e\x67\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x28\x6f\x72\x0a\x6e\x6f\x6e\x2d\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x29\x20\x61\x72\x65\x20\x6e\x65\x76\x65\x72\x20\x61\x64\x64\x65\x64\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x3b\x20\x6e\x6f\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x61\x64\x64\x65\x64\x20\x74\x6f\x0a\x73\x79\x73\x2e\x70\x61\x74\x68\x20\x6d\x6f\x72\x65\x20\x74\x68\x61\x6e\x20\x6f\x6e\x63\x65\x2e\x20\x20\x42\x6c\x61\x6e\x6b\x20\x6c\x69\x6e\x65\x73\x20\x61\x6e\x64\x20\x6c\x69\x6e\x65\x73\x20\x62\x65\x67\x69\x6e\x6e\x69\x6e\x67\x20\x77\x69\x74\x68\x0a\x27\x23\x27\x20\x61\x72\x65\x20\x73\x6b\x69\x70\x70\x65\x64\x2e\x20\x4c\x69\x6e\x65\x73\x20\x73\x74\x61\x72\x74\x69\x6e\x67\x20\x77\x69\x74\x68\x20\x27\x69\x6d\x70\x6f\x72\x74\x27\x20\x61\x72\x65\x20\x65\x78\x65\x63\x75\x74\x65\x64\x2e\x0a\x0a\x46\x6f\x72\x20\x65\x78\x61\x6d\x70\x6c\x65\x2c\x20\x73\x75\x70\x70\x6f\x73\x65\x20\x73\x79\x73\x2e\x70\x72\x65\x66\x69\x78\x20\x61\x6e\x64\x20\x73\x79\x73\x2e\x65\x78\x65\x63\x5f\x70\x72\x65\x66\x69\x78\x20\x61\x72\x65\x20\x73\x65\x74\x20\x74\x6f\x0a\x2f\x75\x73\x72\x2f\x6c\x6f\x63\x61\x6c\x20\x61\x6e\x64\x20\x74\x68\x65\x72\x65\x20\x69\x73\x20\x61\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x2f\x75\x73\x72\x2f\x6c\x6f\x63\x61\x6c\x2f\x6c\x69\x62\x2f\x70\x79\x74\x68\x6f\x6e\x32\x2e\x35\x2f\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x0a\x77\x69\x74\x68\x20\x74\x68\x72\x65\x65\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x2c\x20\x66\x6f\x6f\x2c\x20\x62\x61\x72\x20\x61\x6e\x64\x20\x73\x70\x61\x6d\x2c\x20\x61\x6e\x64\x20\x74\x77\x6f\x20\x70\x61\x74\x68\x0a\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x20\x66\x69\x6c\x65\x73\x2c\x20\x66\x6f\x6f\x2e\x70\x74\x68\x20\x61\x6e\x64\x20\x62\x61\x72\x2e\x70\x74\x68\x2e\x20\x20\x41\x73\x73\x75\x6d\x65\x20\x66\x6f\x6f\x2e\x70\x74\x68\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x20\x74\x68\x65\x0a\x66\x6f\x6c\x6c\x6f\x77\x69\x6e\x67\x3a\x0a\x0a\x20\x20\x23\x20\x66\x6f\x6f\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x0a\x20\x20\x66\x6f\x6f\x0a\x20\x20\x62\x61\x72\x0a\x20\x20\x62\x6c\x65\x74\x63\x68\x0a\x0a\x61\x6e\x64\x20\x62\x61\x72\x2e\x70\x74\x68\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x3a\x0a\x0a\x20\x20\x23\x20\x62\x61\x72\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x0a\x20\x20\x62\x61\x72\x0a\x0a\x54\x68\x65\x6e\x20\x74\x68\x65\x20\x66\x6f\x6c\x6c\x6f\x77\x69\x6e\x67\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x61\x72\x65\x20\x61\x64\x64\x65\x64\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x2c\x20\x69\x6e\x20\x74\x68\x69\x73\x20\x6f\x72\x64\x65\x72\x3a\x0a\x0a\x20\x20\x2f\x75\x73\x72\x2f\x6c\x6f\x63\x61\x6c\x2f\x6c\x69\x62\x2f\x70\x79\x74\x68\x6f\x6e\x32\x2e\x35\x2f\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x2f\x62\x61\x72\x0a\x20\x20\x2f\x75\x73\x72\x2f\x6c\x6f\x63\x61\x6c\x2f\x6c\x69\x62\x2f\x70\x79\x74\x68\x6f\x6e\x32\x2e\x35\x2f\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x2f\x66\x6f\x6f\x0a\x0a\x4e\x6f\x74\x65\x20\x74\x68\x61\x74\x20\x62\x6c\x65\x74\x63\x68\x20\x69\x73\x20\x6f\x6d\x69\x74\x74\x65\x64\x20\x62\x65\x63\x61\x75\x73\x65\x20\x69\x74\x20\x64\x6f\x65\x73\x6e\x27\x74\x20\x65\x78\x69\x73\x74\x3b\x20\x62\x61\x72\x20\x70\x72\x65\x63\x65\x64\x65\x73\x20\x66\x6f\x6f\x0a\x62\x65\x63\x61\x75\x73\x65\x20\x62\x61\x72\x2e\x70\x74\x68\x20\x63\x6f\x6d\x65\x73\x20\x61\x6c\x70\x68\x61\x62\x65\x74\x69\x63\x61\x6c\x6c\x79\x20\x62\x65\x66\x6f\x72\x65\x20\x66\x6f\x6f\x2e\x70\x74\x68\x3b\x20\x61\x6e\x64\x20\x73\x70\x61\x6d\x20\x69\x73\x0a\x6f\x6d\x69\x74\x74\x65\x64\x20\x62\x65\x63\x61\x75\x73\x65\x20\x69\x74\x20\x69\x73\x20\x6e\x6f\x74\x20\x6d\x65\x6e\x74\x69\x6f\x6e\x65\x64\x20\x69\x6e\x20\x65\x69\x74\x68\x65\x72\x20\x70\x61\x74\x68\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x20\x66\x69\x6c\x65\x2e\x0a\x0a\x54\x68\x65\x20\x72\x65\x61\x64\x6c\x69\x6e\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x61\x6c\x73\x6f\x20\x61\x75\x74\x6f\x6d\x61\x74\x69\x63\x61\x6c\x6c\x79\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x65\x64\x20\x74\x6f\x20\x65\x6e\x61\x62\x6c\x65\x0a\x63\x6f\x6d\x70\x6c\x65\x74\x69\x6f\x6e\x20\x66\x6f\x72\x20\x73\x79\x73\x74\x65\x6d\x73\x20\x74\x68\x61\x74\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x69\x74\x2e\x20\x20\x54\x68\x69\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x6f\x76\x65\x72\x72\x69\x64\x64\x65\x6e\x20\x69\x6e\x0a\x73\x69\x74\x65\x63\x75\x73\x74\x6f\x6d\x69\x7a\x65\x2c\x20\x75\x73\x65\x72\x63\x75\x73\x74\x6f\x6d\x69\x7a\x65\x20\x6f\x72\x20\x50\x59\x54\x48\x4f\x4e\x53\x54\x41\x52\x54\x55\x50\x2e\x20\x20\x53\x74\x61\x72\x74\x69\x6e\x67\x20\x50\x79\x74\x68\x6f\x6e\x20\x69\x6e\x0a\x69\x73\x6f\x6c\x61\x74\x65\x64\x20\x6d\x6f\x64\x65\x20\x28\x2d\x49\x29\x20\x64\x69\x73\x61\x62\x6c\x65\x73\x20\x61\x75\x74\x6f\x6d\x61\x74\x69\x63\x20\x72\x65\x61\x64\x6c\x69\x6e\x65\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x41\x66\x74\x65\x72\x20\x74\x68\x65\x73\x65\x20\x6f\x70\x65\x72\x61\x74\x69\x6f\x6e\x73\x2c\x20\x61\x6e\x20\x61\x74\x74\x65\x6d\x70\x74\x20\x69\x73\x20\x6d\x61\x64\x65\x20\x74\x6f\x20\x69\x6d\x70\x6f\x72\x74\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x0a\x6e\x61\x6d\x65\x64\x20\x73\x69\x74\x65\x63\x75\x73\x74\x6f\x6d\x69\x7a\x65\x2c\x20\x77\x68\x69\x63\x68\x20\x63\x61\x6e\x20\x70\x65\x72\x66\x6f\x72\x6d\x20\x61\x72\x62\x69\x74\x72\x61\x72\x79\x20\x61\x64\x64\x69\x74\x69\x6f\x6e\x61\x6c\x0a\x73\x69\x74\x65\x2d\x73\x70\x65\x63\x69\x66\x69\x63\x20\x63\x75\x73\x74\x6f\x6d\x69\x7a\x61\x74\x69\x6f\x6e\x73\x2e\x20\x20\x49\x66\x20\x74\x68\x69\x73\x20\x69\x6d\x70\x6f\x72\x74\x20\x66\x61\x69\x6c\x73\x20\x77\x69\x74\x68\x20\x61\x6e\x0a\x49\x6d\x70\x6f\x72\x74\x45\x72\x72\x6f\x72\x20\x65\x78\x63\x65\x70\x74\x69\x6f\x6e\x2c\x20\x69\x74\x20\x69\x73\x20\x73\x69\x6c\x65\x6e\x74\x6c\x79\x20\x69\x67\x6e\x6f\x72\x65\x64\x2e\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & importlib__bootstrap_toplevel_consts_25_consts_3._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
site_toplevel_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_verbose._ascii.ob_base,
            & const_str_print._ascii.ob_base,
            &_Py_ID(stderr),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
site_toplevel_consts_3_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen site>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__trace = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_trace",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[35];
    }
site_toplevel_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 34,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x07\x0a\x87\x79\x81\x79\xd7\x07\x18\xd2\x07\x18\xdc\x08\x0d\x88\x67\x9c\x43\x9f\x4a\x99\x4a\xd6\x08\x27\xf0\x03\x00\x08\x19",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
site_toplevel_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(message),
        },
    },
};
static
    struct _PyCode_DEF(112)
site_toplevel_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 56,
    },
    .co_consts = & site_toplevel_consts_3_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 92,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 679,
    .co_localsplusnames = & site_toplevel_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str__trace._ascii.ob_base,
    .co_qualname = & const_str__trace._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x1c\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
site_toplevel_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(join),
            & const_str_abspath._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_normcase._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_makepath = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "makepath",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[92];
    }
site_toplevel_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 91,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0a\x0c\x8f\x27\x89\x27\x8f\x2c\x89\x2c\x98\x05\xd0\x0a\x1e\x80\x43\xf0\x02\x03\x05\x0d\xdc\x0e\x10\x8f\x67\x89\x67\x8f\x6f\x89\x6f\x98\x63\xd3\x0e\x22\x88\x03\xf0\x06\x00\x0c\x0f\x94\x02\x97\x07\x91\x07\xd7\x10\x20\xd1\x10\x20\xa0\x13\xd3\x10\x25\xd0\x0b\x25\xd0\x04\x25\xf8\xf4\x05\x00\x0c\x13\xf2\x00\x01\x05\x0d\xd9\x08\x0c\xf0\x03\x01\x05\x0d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
site_toplevel_consts_4_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x9e\x1f\x41\x1e\x00\xc1\x1e\x09\x41\x2a\x03\xc1\x29\x01\x41\x2a\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_paths._ascii.ob_base,
            & const_str_dir._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(218)
site_toplevel_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 109,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_4_exceptiontable.ob_base.ob_base,
    .co_flags = 7,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 97,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 680,
    .co_localsplusnames = & site_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_makepath._ascii.ob_base,
    .co_qualname = & const_str_makepath._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x8e\x00\x7d\x01\x09\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x66\x02\x53\x00\x23\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x2c\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[70];
    }
site_toplevel_consts_5_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 69,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set all module __file__ and __cached__ attributes to an absolute path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_5_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__frozen_importlib._ascii.ob_base,
            & const_str__frozen_importlib_external._ascii.ob_base,
        },
    },
};
// TODO: The above tuple should be a frozenset
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & site_toplevel_consts_5_consts_0._ascii.ob_base,
            Py_None,
            & site_toplevel_consts_5_consts_2._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
site_toplevel_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            & const_str_set._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(values),
            &_Py_ID(__loader__),
            &_Py_ID(__module__),
            & const_str_AttributeError._ascii.ob_base,
            &_Py_ID(__spec__),
            & const_str_loader._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_abspath._ascii.ob_base,
            &_Py_ID(__file__),
            & const_str_OSError._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            & const_str___cached__._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_abs_paths = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "abs_paths",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[241];
    }
site_toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 240,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0d\x10\x94\x13\x97\x1b\x91\x1b\xd7\x11\x23\xd1\x11\x23\xd3\x11\x25\xd6\x0d\x26\x88\x01\xd8\x18\x1c\x88\x0d\xf0\x02\x06\x09\x15\xd8\x1c\x1d\x9f\x4c\x99\x4c\xd7\x1c\x33\xd1\x1c\x33\x88\x4d\xf0\x0c\x00\x0c\x19\xd0\x20\x53\xd1\x0b\x53\xd8\x0c\x14\xf0\x02\x03\x09\x11\xdc\x19\x1b\x9f\x17\x99\x17\x9f\x1f\x99\x1f\xa8\x11\xaf\x1a\xa9\x1a\xd3\x19\x34\x88\x41\x8c\x4a\xf0\x06\x03\x09\x11\xdc\x1b\x1d\x9f\x37\x99\x37\x9f\x3f\x99\x3f\xa8\x31\xaf\x3c\xa9\x3c\xd3\x1b\x38\x88\x41\x8d\x4c\xf1\x21\x00\x0e\x27\xf8\xf4\x08\x00\x10\x1e\xf2\x00\x04\x09\x15\xf0\x02\x03\x0d\x15\xd8\x20\x21\xa7\x0a\xa1\x0a\xd7\x20\x31\xd1\x20\x31\xd7\x20\x3c\xd1\x20\x3c\x91\x0d\xf8\xdc\x13\x21\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfc\xf0\x07\x04\x09\x15\xfb\xf4\x12\x00\x11\x1f\xa4\x07\xac\x19\xd0\x0f\x33\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfb\xf4\x08\x00\x11\x1f\xa4\x07\xac\x19\xd0\x0f\x33\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[90];
    }
site_toplevel_consts_5_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 89,
    },
    .ob_shash = -1,
    .ob_sval = "\xae\x16\x42\x2a\x02\xc1\x0a\x2e\x43\x2a\x02\xc1\x39\x2e\x44\x04\x02\xc2\x2a\x09\x43\x27\x05\xc2\x34\x20\x43\x15\x04\xc3\x14\x01\x43\x27\x05\xc3\x15\x09\x43\x21\x07\xc3\x1e\x02\x43\x27\x05\xc3\x20\x01\x43\x21\x07\xc3\x21\x03\x43\x27\x05\xc3\x26\x01\x43\x27\x05\xc3\x2a\x14\x44\x01\x05\xc4\x00\x01\x44\x01\x05\xc4\x04\x14\x44\x1b\x05\xc4\x1a\x01\x44\x1b\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_loader_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader_module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_5_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[109],
            & const_str_loader_module._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(572)
site_toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 286,
    },
    .co_consts = & site_toplevel_consts_5_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_5_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 106,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 681,
    .co_localsplusnames = & site_toplevel_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_abs_paths._ascii.ob_base,
    .co_qualname = & const_str_abs_paths._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x7e\x00\x00\x7d\x00\x64\x01\x7d\x01\x09\x00\x7c\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x01\x64\x02\x76\x01\x72\x01\x8c\x21\x09\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x09\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x8c\x80\x04\x00\x79\x01\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x34\x01\x00\x09\x00\x7c\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x6e\x0f\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x6e\x04\x77\x00\x78\x03\x59\x00\x77\x01\x59\x00\x8c\xa2\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x88\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x74\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x66\x03\x24\x00\x72\x03\x01\x00\x59\x00\x8c\xf2\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[76];
    }
site_toplevel_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 75,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x52\x65\x6d\x6f\x76\x65\x20\x64\x75\x70\x6c\x69\x63\x61\x74\x65\x20\x65\x6e\x74\x72\x69\x65\x73\x20\x66\x72\x6f\x6d\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x20\x61\x6c\x6f\x6e\x67\x20\x77\x69\x74\x68\x20\x6d\x61\x6b\x69\x6e\x67\x20\x74\x68\x65\x6d\x0a\x20\x20\x20\x20\x61\x62\x73\x6f\x6c\x75\x74\x65",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & site_toplevel_consts_6_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
site_toplevel_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_set._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(path),
            & const_str_makepath._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(add),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_removeduppaths = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "removeduppaths",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[98];
    }
site_toplevel_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 97,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x09\x0b\x80\x41\xdc\x12\x15\x93\x25\x80\x4b\xdc\x0f\x12\x8f\x78\x8c\x78\x88\x03\xf4\x08\x00\x18\x20\xa0\x03\x93\x7d\x89\x0c\x88\x03\x88\x57\xd8\x0b\x12\x98\x2b\xd2\x0b\x25\xd8\x0c\x0d\x8f\x48\x89\x48\x90\x53\x8c\x4d\xd8\x0c\x17\x8f\x4f\x89\x4f\x98\x47\xd5\x0c\x24\xf0\x0f\x00\x10\x18\xf0\x10\x00\x13\x14\x84\x43\x87\x48\x81\x48\x89\x51\x80\x4b\xd8\x0b\x16\xd0\x04\x16",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_known_paths = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "known_paths",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_dircase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dircase",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
site_toplevel_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[76],
            & const_str_known_paths._ascii.ob_base,
            & const_str_dir._ascii.ob_base,
            & const_str_dircase._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(216)
site_toplevel_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 108,
    },
    .co_consts = & site_toplevel_consts_6_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 129,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 682,
    .co_localsplusnames = & site_toplevel_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_removeduppaths._ascii.ob_base,
    .co_qualname = & const_str_removeduppaths._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x67\x00\x7d\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x37\x00\x00\x7d\x02\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x03\x7c\x01\x76\x01\x73\x01\x8c\x16\x7c\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x39\x04\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x01\x1b\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[70];
    }
site_toplevel_consts_7_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 69,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return a set containing all existing file system items from sys.path.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
site_toplevel_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & site_toplevel_consts_7_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
site_toplevel_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_set._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(path),
            & const_str_os._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str_makepath._ascii.ob_base,
            &_Py_ID(add),
            & const_str_TypeError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__init_pathinfo = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_init_pathinfo",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[97];
    }
site_toplevel_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 96,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x08\x0b\x8b\x05\x80\x41\xdc\x10\x13\x97\x08\x94\x08\x88\x04\xf0\x02\x05\x09\x15\xdc\x0f\x11\x8f\x77\x89\x77\x8f\x7e\x89\x7e\x98\x64\xd4\x0f\x23\xdc\x1e\x26\xa0\x74\x9b\x6e\x91\x0b\x90\x01\x90\x38\xd8\x10\x11\x97\x05\x91\x05\x90\x68\x94\x0f\xf8\xf0\x09\x00\x11\x19\xf0\x0e\x00\x0c\x0d\x80\x48\xf8\xf4\x05\x00\x10\x19\xf2\x00\x01\x09\x15\xd9\x0c\x14\xf0\x03\x01\x09\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
site_toplevel_consts_7_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x9f\x3e\x41\x21\x02\xc1\x21\x09\x41\x2d\x05\xc1\x2c\x01\x41\x2d\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_itemcase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "itemcase",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
site_toplevel_consts_7_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            (PyObject *)&_Py_SINGLETON(strings).ascii[100],
            &_Py_ID(item),
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            & const_str_itemcase._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(224)
site_toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 112,
    },
    .co_consts = & site_toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 148,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 683,
    .co_localsplusnames = & site_toplevel_consts_7_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str__init_pathinfo._ascii.ob_base,
    .co_qualname = & const_str__init_pathinfo._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x41\x00\x00\x7d\x01\x09\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x72\x1f\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x02\x7d\x03\x7c\x00\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x43\x04\x00\x7c\x00\x53\x00\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x51\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[215];
    }
site_toplevel_consts_8_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 214,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x50\x72\x6f\x63\x65\x73\x73\x20\x61\x20\x2e\x70\x74\x68\x20\x66\x69\x6c\x65\x20\x77\x69\x74\x68\x69\x6e\x20\x74\x68\x65\x20\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x46\x6f\x72\x20\x65\x61\x63\x68\x20\x6c\x69\x6e\x65\x20\x69\x6e\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x2c\x20\x65\x69\x74\x68\x65\x72\x20\x63\x6f\x6d\x62\x69\x6e\x65\x20\x69\x74\x20\x77\x69\x74\x68\x20\x73\x69\x74\x65\x64\x69\x72\x20\x74\x6f\x20\x61\x20\x70\x61\x74\x68\x0a\x20\x20\x20\x20\x20\x20\x20\x61\x6e\x64\x20\x61\x64\x64\x20\x74\x68\x61\x74\x20\x74\x6f\x20\x6b\x6e\x6f\x77\x6e\x5f\x70\x61\x74\x68\x73\x2c\x20\x6f\x72\x20\x65\x78\x65\x63\x75\x74\x65\x20\x69\x74\x20\x69\x66\x20\x69\x74\x20\x73\x74\x61\x72\x74\x73\x20\x77\x69\x74\x68\x20\x27\x69\x6d\x70\x6f\x72\x74\x20\x27\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_st_flags = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_flags",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_st_file_attributes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "st_file_attributes",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[28];
    }
site_toplevel_consts_8_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 27,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Skipping hidden .pth file: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
site_toplevel_consts_8_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Processing .pth file: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
site_toplevel_consts_8_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "utf-8-sig",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
site_toplevel_consts_8_consts_10 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Cannot read ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
site_toplevel_consts_8_consts_11 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " as UTF-8. Using fallback encoding ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
site_toplevel_consts_8_consts_15_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x69\x6d\x70\x6f\x72\x74\x09",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_8_consts_15 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_25_consts_1_1._ascii.ob_base,
            & site_toplevel_consts_8_consts_15_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
site_toplevel_consts_8_consts_16 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Error processing line ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
site_toplevel_consts_8_consts_18 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " of ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
site_toplevel_consts_8_consts_19 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x3a\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[3];
    }
site_toplevel_consts_8_consts_21 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 2,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "  ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
site_toplevel_consts_8_consts_22 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x0a\x52\x65\x6d\x61\x69\x6e\x64\x65\x72\x20\x6f\x66\x20\x66\x69\x6c\x65\x20\x69\x67\x6e\x6f\x72\x65\x64",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[23];
        }_object;
    }
site_toplevel_consts_8_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 23,
        },
        .ob_item = {
            & site_toplevel_consts_8_consts_0._ascii.ob_base,
            Py_None,
            Py_True,
            Py_False,
            & const_str_st_flags._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & const_str_st_file_attributes._ascii.ob_base,
            & site_toplevel_consts_8_consts_7._ascii.ob_base,
            & site_toplevel_consts_8_consts_8._ascii.ob_base,
            & site_toplevel_consts_8_consts_9._ascii.ob_base,
            & site_toplevel_consts_8_consts_10._ascii.ob_base,
            & site_toplevel_consts_8_consts_11._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_Py_SINGLETON(strings).ascii[35],
            &_Py_STR(empty),
            & site_toplevel_consts_8_consts_15._object.ob_base.ob_base,
            & site_toplevel_consts_8_consts_16._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[100],
            & site_toplevel_consts_8_consts_18._ascii.ob_base,
            & site_toplevel_consts_8_consts_19._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_25_consts_3._object.ob_base.ob_base,
            & site_toplevel_consts_8_consts_21._ascii.ob_base,
            & site_toplevel_consts_8_consts_22._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_UF_HIDDEN = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UF_HIDDEN",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_FILE_ATTRIBUTE_HIDDEN = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_HIDDEN",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_getencoding = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getencoding",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_strip = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "strip",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_format_exception = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "format_exception",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[34];
        }_object;
    }
site_toplevel_consts_8_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 34,
        },
        .ob_item = {
            & const_str__init_pathinfo._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(join),
            & const_str_lstat._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            &_Py_ID(getattr),
            & const_str_stat._ascii.ob_base,
            & const_str_UF_HIDDEN._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_HIDDEN._ascii.ob_base,
            & const_str__trace._ascii.ob_base,
            & const_str_io._ascii.ob_base,
            & const_str_open_code._ascii.ob_base,
            &_Py_ID(read),
            &_Py_ID(decode),
            & const_str_UnicodeDecodeError._ascii.ob_base,
            &_Py_ID(locale),
            & const_str_getencoding._ascii.ob_base,
            & const_str_enumerate._ascii.ob_base,
            & const_str_splitlines._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_strip._ascii.ob_base,
            & const_str_exec._ascii.ob_base,
            & const_str_rstrip._ascii.ob_base,
            & const_str_makepath._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(add),
            & const_str_Exception._ascii.ob_base,
            & const_str_print._ascii.ob_base,
            &_Py_ID(stderr),
            &_Py_ID(traceback),
            & const_str_format_exception._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_addpackage = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "addpackage",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[665];
    }
site_toplevel_consts_8_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 664,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x08\x13\xd0\x07\x1a\xdc\x16\x24\xd3\x16\x26\x88\x0b\xd8\x10\x14\x89\x05\xe0\x10\x15\x88\x05\xdc\x0f\x11\x8f\x77\x89\x77\x8f\x7c\x89\x7c\x98\x47\xa0\x54\xd3\x0f\x2a\x80\x48\xf0\x02\x03\x05\x0f\xdc\x0d\x0f\x8f\x58\x89\x58\x90\x68\xd3\x0d\x1f\x88\x02\xf4\x06\x00\x0a\x11\x90\x12\x90\x5a\xa0\x11\xd3\x09\x23\xa4\x64\xa7\x6e\xa1\x6e\xd2\x09\x34\xdc\x09\x10\x90\x12\xd0\x15\x29\xa8\x31\xd3\x09\x2d\xb4\x04\xd7\x30\x4a\xd1\x30\x4a\xd2\x09\x4a\xdc\x08\x0e\xd0\x11\x2c\xa8\x58\xa8\x4c\xd0\x0f\x39\xd4\x08\x3a\xd8\x08\x0e\xdc\x04\x0a\xd0\x0d\x23\xa0\x48\xa0\x3c\xd0\x0b\x30\xd4\x04\x31\xf0\x02\x04\x05\x0f\xdc\x0d\x0f\x8f\x5c\x89\x5c\x98\x28\xd4\x0d\x23\xa0\x71\xd8\x1a\x1b\x9f\x26\x99\x26\x9b\x28\x88\x4b\xf7\x03\x00\x0e\x24\xf0\x0a\x0a\x05\x44\x01\xf0\x06\x00\x17\x22\xd7\x16\x28\xd1\x16\x28\xa8\x1b\xd3\x16\x35\x88\x0b\xf4\x12\x00\x14\x1d\x98\x5b\xd7\x1d\x33\xd1\x1d\x33\xd3\x1d\x35\xb0\x71\xd6\x13\x39\x89\x07\x88\x01\x88\x34\xd8\x0b\x0f\x8f\x3f\x89\x3f\x98\x33\xd4\x0b\x1f\xd8\x0c\x14\xd8\x0b\x0f\x8f\x3a\x89\x3a\x8b\x3c\x98\x32\xd2\x0b\x1d\xd8\x0c\x14\xf0\x02\x11\x09\x12\xd8\x0f\x13\x8f\x7f\x89\x7f\xd0\x1f\x36\xd4\x0f\x37\xdc\x10\x14\x90\x54\x94\x0a\xd8\x10\x18\xd8\x13\x17\x97\x3b\x91\x3b\x93\x3d\x88\x44\xdc\x1b\x23\xa0\x47\xa8\x54\xd3\x1b\x32\x89\x4c\x88\x43\x90\x17\xd8\x0f\x16\x98\x6b\xd1\x0f\x29\xac\x62\xaf\x67\xa9\x67\xaf\x6e\xa9\x6e\xb8\x53\xd4\x2e\x41\xdc\x10\x13\x97\x08\x91\x08\x97\x0f\x91\x0f\xa0\x03\xd4\x10\x24\xd8\x10\x1b\x97\x0f\x91\x0f\xa0\x07\xd4\x10\x28\xf8\xf0\x1b\x00\x14\x3a\xf1\x2e\x00\x08\x0d\xd8\x16\x1a\x88\x0b\xd8\x0b\x16\xd0\x04\x16\xf8\xf4\x65\x01\x00\x0c\x13\xf2\x00\x01\x05\x0f\xd9\x08\x0e\xf0\x03\x01\x05\x0f\xfa\xf7\x10\x00\x0e\x24\xd1\x0d\x23\xfb\xe4\x0b\x12\xf2\x00\x01\x05\x0f\xd9\x08\x0e\xf0\x03\x01\x05\x0f\xfb\xf4\x0e\x00\x0c\x1e\xf2\x00\x06\x05\x44\x01\xf3\x06\x00\x09\x16\xd8\x16\x21\xd7\x16\x28\xd1\x16\x28\xa8\x16\xd7\x29\x3b\xd1\x29\x3b\xd3\x29\x3d\xd3\x16\x3e\x88\x0b\xdc\x08\x0e\x90\x1c\x98\x68\x98\x5c\xf0\x00\x01\x2a\x2a\xd8\x2a\x30\xd7\x2a\x3c\xd1\x2a\x3c\xd3\x2a\x3e\xd0\x29\x41\xf0\x03\x01\x10\x43\x01\xf7\x00\x01\x09\x44\x01\xf0\x0b\x06\x05\x44\x01\xfb\xf4\x2c\x00\x10\x19\xf2\x00\x08\x09\x12\xdc\x0c\x11\xd0\x14\x2a\xa8\x31\xa8\x51\xa8\x25\xa8\x74\xb0\x48\xb0\x3a\xb8\x53\xd0\x12\x41\xdc\x17\x1a\x97\x7a\x91\x7a\xf5\x03\x01\x0d\x23\xe3\x0c\x1c\xd8\x1a\x23\xd7\x1a\x34\xd1\x1a\x34\xb0\x53\xd6\x1a\x39\x90\x06\xd8\x1c\x22\xd7\x1c\x2d\xd1\x1c\x2d\xd6\x1c\x2f\x90\x44\xdc\x14\x19\x98\x24\x98\x74\x99\x29\xac\x23\xaf\x2a\xa9\x2a\xd6\x14\x35\xf1\x03\x00\x1d\x30\xf0\x03\x00\x1b\x3a\xf4\x06\x00\x0d\x12\xd0\x12\x2f\xb4\x63\xb7\x6a\xb1\x6a\xd5\x0c\x41\xde\x0c\x11\xfb\xf0\x11\x08\x09\x12\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[111];
    }
site_toplevel_consts_8_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 110,
    },
    .ob_shash = -1,
    .ob_sval = "\xb3\x15\x47\x00\x00\xc2\x22\x15\x47\x1c\x00\xc2\x37\x11\x47\x0f\x03\xc3\x08\x08\x47\x1c\x00\xc3\x11\x11\x47\x2b\x00\xc4\x29\x1c\x48\x3d\x02\xc5\x06\x41\x32\x48\x3d\x02\xc7\x00\x09\x47\x0c\x03\xc7\x0b\x01\x47\x0c\x03\xc7\x0f\x05\x47\x19\x07\xc7\x14\x08\x47\x1c\x00\xc7\x1c\x09\x47\x28\x03\xc7\x27\x01\x47\x28\x03\xc7\x2b\x41\x0b\x48\x3a\x03\xc8\x39\x01\x48\x3a\x03\xc8\x3d\x09\x4b\x1d\x05\xc9\x06\x42\x0b\x4b\x18\x05\xcb\x18\x05\x4b\x1d\x05",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_sitedir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sitedir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_pth_content = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pth_content",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_record = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "record",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
site_toplevel_consts_8_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            & const_str_sitedir._ascii.ob_base,
            &_Py_ID(name),
            & const_str_known_paths._ascii.ob_base,
            &_Py_ID(reset),
            & const_str_fullname._ascii.ob_base,
            & const_str_st._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[102],
            & const_str_pth_content._ascii.ob_base,
            &_Py_ID(locale),
            (PyObject *)&_Py_SINGLETON(strings).ascii[110],
            &_Py_ID(line),
            & const_str_dir._ascii.ob_base,
            & const_str_dircase._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
            &_Py_ID(traceback),
            & const_str_record._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(1472)
site_toplevel_consts_8 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 736,
    },
    .co_consts = & site_toplevel_consts_8_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_8_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_8_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 25 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 161,
    .co_nlocalsplus = 16,
    .co_nlocals = 16,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 684,
    .co_localsplusnames = & site_toplevel_consts_8_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & posixpath_toplevel_consts_33_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_addpackage._ascii.ob_base,
    .co_qualname = & const_str_addpackage._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_8_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x80\x0d\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x64\x02\x7d\x03\x6e\x02\x64\x03\x7d\x03\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x09\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x64\x04\x64\x05\xab\x03\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x01\x00\x00\x73\x1e\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x64\x06\x64\x05\xab\x03\x00\x00\x00\x00\x00\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x01\x00\x00\x72\x0f\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x7c\x04\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x7c\x04\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x06\x7c\x06\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x7f\x07\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x74\x25\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x64\x0c\xab\x02\x00\x00\x00\x00\x00\x00\x44\x00\x5d\xbb\x00\x00\x5c\x02\x00\x00\x7d\x09\x7d\x0a\x7c\x0a\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x72\x01\x8c\x18\x7c\x0a\x6a\x2b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x64\x0e\x6b\x28\x00\x00\x72\x01\x8c\x2c\x09\x00\x7c\x0a\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0f\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0c\x74\x2d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x4a\x7c\x0a\x6a\x2f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x0a\x74\x31\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x0a\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x0b\x7d\x0c\x7c\x0c\x7c\x02\x76\x01\x72\x4f\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x72\x30\x74\x34\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x37\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0b\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x02\x6a\x39\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0c\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\xbd\x04\x00\x7c\x03\x72\x02\x64\x01\x7d\x02\x7c\x02\x53\x00\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x90\x01\x8c\x09\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x46\x01\x00\x64\x05\x64\x01\x6c\x10\x7d\x08\x7f\x07\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\x6a\x23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x07\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\x7c\x04\x9b\x02\x64\x0b\x7c\x08\x6a\x23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x90\x01\x8c\x57\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x3a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x97\x7d\x0d\x74\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x10\x7c\x09\x64\x11\x9b\x04\x64\x12\x7c\x04\x9b\x00\x64\x13\x9d\x05\x74\x34\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x14\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x64\x05\x64\x01\x6c\x20\x7d\x0e\x7c\x0e\x6a\x43\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x35\x00\x00\x7d\x0f\x7c\x0f\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x20\x00\x00\x7d\x0a\x74\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x15\x7c\x0a\x7a\x00\x00\x00\x74\x34\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x14\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x22\x04\x00\x8c\x37\x04\x00\x74\x3d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x16\x74\x34\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x3e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x14\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x01\x7d\x0d\x7e\x0d\x01\x00\x90\x01\x8c\x1e\x64\x01\x7d\x0d\x7e\x0d\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[85];
    }
site_toplevel_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 84,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x64\x64\x20\x27\x73\x69\x74\x65\x64\x69\x72\x27\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x20\x69\x66\x20\x6d\x69\x73\x73\x69\x6e\x67\x20\x61\x6e\x64\x20\x68\x61\x6e\x64\x6c\x65\x20\x2e\x70\x74\x68\x20\x66\x69\x6c\x65\x73\x20\x69\x6e\x0a\x20\x20\x20\x20\x27\x73\x69\x74\x65\x64\x69\x72\x27",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
site_toplevel_consts_9_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Adding directory: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
site_toplevel_consts_9_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".pth",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
site_toplevel_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & site_toplevel_consts_9_consts_0._ascii.ob_base,
            & site_toplevel_consts_9_consts_1._ascii.ob_base,
            Py_None,
            Py_True,
            Py_False,
            & site_toplevel_consts_9_consts_5._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
site_toplevel_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & const_str__trace._ascii.ob_base,
            & const_str__init_pathinfo._ascii.ob_base,
            & const_str_makepath._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(append),
            &_Py_ID(add),
            & const_str_os._ascii.ob_base,
            & const_str_listdir._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            & const_str_startswith._ascii.ob_base,
            & const_str_sorted._ascii.ob_base,
            & const_str_addpackage._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_addsitedir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "addsitedir",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[234];
    }
site_toplevel_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 233,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x05\x0b\xd0\x0d\x1f\xa0\x07\x98\x7b\xd0\x0b\x2b\xd4\x04\x2c\xd8\x07\x12\xd0\x07\x1a\xdc\x16\x24\xd3\x16\x26\x88\x0b\xd8\x10\x14\x89\x05\xe0\x10\x15\x88\x05\xdc\x1b\x23\xa0\x47\xd3\x1b\x2c\xd1\x04\x18\x80\x47\x88\x5b\xd8\x0b\x16\x98\x2b\xd1\x0b\x25\xdc\x08\x0b\x8f\x08\x89\x08\x8f\x0f\x89\x0f\x98\x07\xd4\x08\x20\xd8\x08\x13\x8f\x0f\x89\x0f\x98\x0b\xd4\x08\x24\xf0\x02\x03\x05\x0f\xdc\x10\x12\x97\x0a\x91\x0a\x98\x37\xd3\x10\x23\x88\x05\xf1\x06\x00\x1f\x24\xf3\x00\x01\x0d\x44\x01\x99\x65\x90\x64\xd8\x10\x14\x97\x0d\x91\x0d\x98\x66\xd4\x10\x25\xa8\x64\xaf\x6f\xa9\x6f\xb8\x63\xd4\x2e\x42\xf2\x03\x00\x0e\x12\x98\x65\x80\x45\xf0\x00\x01\x0d\x44\x01\xe4\x10\x16\x90\x75\x96\x0d\x88\x04\xdc\x08\x12\x90\x37\x98\x44\xa0\x2b\xd5\x08\x2e\xf0\x03\x00\x11\x1e\xe1\x07\x0c\xd8\x16\x1a\x88\x0b\xd8\x0b\x16\xd0\x04\x16\xf8\xf4\x11\x00\x0c\x13\xf2\x00\x01\x05\x0f\xd9\x08\x0e\xf0\x03\x01\x05\x0f\xfc\xf2\x04\x01\x0d\x44\x01",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[25];
    }
site_toplevel_consts_9_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 24,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x23\x15\x43\x0c\x00\xc1\x3c\x2b\x43\x1b\x04\xc3\x0c\x09\x43\x18\x03\xc3\x17\x01\x43\x18\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_sitedircase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sitedircase",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
site_toplevel_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_sitedir._ascii.ob_base,
            & const_str_known_paths._ascii.ob_base,
            &_Py_ID(reset),
            & const_str_sitedircase._ascii.ob_base,
            & const_str_names._ascii.ob_base,
            &_Py_ID(name),
        },
    },
};
static
    struct _PyCode_DEF(448)
site_toplevel_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 224,
    },
    .co_consts = & site_toplevel_consts_9_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_9_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 227,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 685,
    .co_localsplusnames = & site_toplevel_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_addsitedir._ascii.ob_base,
    .co_qualname = & const_str_addsitedir._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7c\x00\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x80\x0d\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x64\x03\x7d\x02\x6e\x02\x64\x04\x7d\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x02\x00\x00\x7d\x00\x7d\x03\x7c\x03\x7c\x01\x76\x01\x72\x30\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x04\x44\x00\x8f\x05\x63\x02\x67\x00\x63\x02\x5d\x26\x00\x00\x7d\x05\x7c\x05\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x01\x00\x00\x00\x00\x00\x00\x72\x13\x7c\x05\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x73\x02\x7c\x05\x91\x02\x8c\x28\x04\x00\x7d\x04\x7d\x05\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x0f\x00\x00\x7d\x05\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x05\x7c\x01\xab\x03\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x11\x04\x00\x7c\x02\x72\x02\x64\x02\x7d\x01\x7c\x01\x53\x00\x23\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x02\x77\x00\x78\x03\x59\x00\x77\x01\x63\x02\x01\x00\x63\x02\x7d\x05\x77\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[301];
    }
site_toplevel_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 300,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x68\x65\x63\x6b\x20\x69\x66\x20\x75\x73\x65\x72\x20\x73\x69\x74\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x73\x61\x66\x65\x20\x66\x6f\x72\x20\x69\x6e\x63\x6c\x75\x73\x69\x6f\x6e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x74\x65\x73\x74\x73\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x63\x6f\x6d\x6d\x61\x6e\x64\x20\x6c\x69\x6e\x65\x20\x66\x6c\x61\x67\x20\x28\x69\x6e\x63\x6c\x75\x64\x69\x6e\x67\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x20\x76\x61\x72\x29\x2c\x0a\x20\x20\x20\x20\x70\x72\x6f\x63\x65\x73\x73\x20\x75\x69\x64\x2f\x67\x69\x64\x20\x65\x71\x75\x61\x6c\x20\x74\x6f\x20\x65\x66\x66\x65\x63\x74\x69\x76\x65\x20\x75\x69\x64\x2f\x67\x69\x64\x2e\x0a\x0a\x20\x20\x20\x20\x4e\x6f\x6e\x65\x3a\x20\x44\x69\x73\x61\x62\x6c\x65\x64\x20\x66\x6f\x72\x20\x73\x65\x63\x75\x72\x69\x74\x79\x20\x72\x65\x61\x73\x6f\x6e\x73\x0a\x20\x20\x20\x20\x46\x61\x6c\x73\x65\x3a\x20\x44\x69\x73\x61\x62\x6c\x65\x64\x20\x62\x79\x20\x75\x73\x65\x72\x20\x28\x63\x6f\x6d\x6d\x61\x6e\x64\x20\x6c\x69\x6e\x65\x20\x6f\x70\x74\x69\x6f\x6e\x29\x0a\x20\x20\x20\x20\x54\x72\x75\x65\x3a\x20\x53\x61\x66\x65\x20\x61\x6e\x64\x20\x65\x6e\x61\x62\x6c\x65\x64\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_geteuid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "geteuid",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_getgid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getgid",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_getegid = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getegid",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
site_toplevel_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & site_toplevel_consts_10_consts_0._ascii.ob_base,
            Py_False,
            & const_str_getuid._ascii.ob_base,
            & const_str_geteuid._ascii.ob_base,
            Py_None,
            & const_str_getgid._ascii.ob_base,
            & const_str_getegid._ascii.ob_base,
            Py_True,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_no_user_site = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "no_user_site",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
site_toplevel_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_no_user_site._ascii.ob_base,
            & const_str_hasattr._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_geteuid._ascii.ob_base,
            & const_str_getuid._ascii.ob_base,
            & const_str_getegid._ascii.ob_base,
            & const_str_getgid._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str_check_enableusersite = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "check_enableusersite",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[108];
    }
site_toplevel_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 107,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x14\x00\x08\x0b\x87\x79\x81\x79\xd7\x07\x1d\xd2\x07\x1d\xd8\x0f\x14\xe4\x07\x0e\x8c\x72\x90\x38\xd4\x07\x1c\xa4\x17\xac\x12\xa8\x59\xd4\x21\x37\xe4\x0b\x0d\x8f\x3a\x89\x3a\x8b\x3c\x9c\x32\x9f\x39\x99\x39\x9b\x3b\xd2\x0b\x26\xd8\x13\x17\xdc\x07\x0e\x8c\x72\x90\x38\xd4\x07\x1c\xa4\x17\xac\x12\xa8\x59\xd4\x21\x37\xe4\x0b\x0d\x8f\x3a\x89\x3a\x8b\x3c\x9c\x32\x9f\x39\x99\x39\x9b\x3b\xd2\x0b\x26\xd8\x13\x17\xe0\x0b\x0f",
};
static
    struct _PyCode_DEF(354)
site_toplevel_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 177,
    },
    .co_consts = & site_toplevel_consts_10_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 253,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 686,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_check_enableusersite._ascii.ob_base,
    .co_qualname = & const_str_check_enableusersite._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x01\x79\x01\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x72\x3a\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x72\x2a\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x01\x79\x04\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x02\x00\x00\x00\x00\x00\x00\x72\x3a\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x72\x2a\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x01\x79\x04\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_PYTHONUSERBASE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PYTHONUSERBASE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_emscripten = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "emscripten",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_wasi = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "wasi",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_11_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_emscripten._ascii.ob_base,
            & const_str_vxworks._ascii.ob_base,
            & const_str_wasi._ascii.ob_base,
        },
    },
};
// TODO: The above tuple should be a frozenset
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
site_toplevel_consts_11_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_expanduser._ascii.ob_base,
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_joinuser = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "joinuser",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
site_toplevel_consts_11_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_getuserbase.<locals>.joinuser",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
site_toplevel_consts_11_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x11\x8f\x77\x89\x77\xd7\x0f\x21\xd1\x0f\x21\xa4\x22\xa7\x27\xa1\x27\xa7\x2c\xa1\x2c\xb0\x04\xd0\x22\x35\xd3\x0f\x36\xd0\x08\x36",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
site_toplevel_consts_11_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(args),
        },
    },
};
static
    struct _PyCode_DEF(116)
site_toplevel_consts_11_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 58,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_11_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 23,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 294,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 687,
    .co_localsplusnames = & site_toplevel_consts_11_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_joinuser._ascii.ob_base,
    .co_qualname = & site_toplevel_consts_11_consts_3_qualname._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_11_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x8e\x00\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_APPDATA = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "APPDATA",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_Python = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Python",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_Library = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Library",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
site_toplevel_consts_11_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".local",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
site_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            Py_None,
            & const_str_PYTHONUSERBASE._ascii.ob_base,
            & site_toplevel_consts_11_consts_2._object.ob_base.ob_base,
            & site_toplevel_consts_11_consts_3.ob_base.ob_base,
            &_Py_ID(nt),
            & const_str_APPDATA._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[126],
            & const_str_Python._ascii.ob_base,
            & const_str_darwin._ascii.ob_base,
            & const_str_Library._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_52_consts_6_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & site_toplevel_consts_11_consts_12._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str__framework = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_framework",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
site_toplevel_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            &_Py_ID(get),
            & const_str_sys._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            &_Py_ID(name),
            & const_str__framework._ascii.ob_base,
            & const_str_version_info._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__getuserbase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_getuserbase",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[181];
    }
site_toplevel_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 180,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0f\x11\x8f\x7a\x89\x7a\x8f\x7e\x89\x7e\xd0\x1e\x2e\xb0\x04\xd3\x0f\x35\x80\x48\xd9\x07\x0f\xd8\x0f\x17\x88\x0f\xf4\x06\x00\x08\x0b\x87\x7c\x81\x7c\xd0\x17\x38\xd1\x07\x38\xd8\x0f\x13\xf2\x04\x01\x05\x37\xf4\x06\x00\x08\x0a\x87\x77\x81\x77\x90\x24\x82\x7f\xdc\x0f\x11\x8f\x7a\x89\x7a\x8f\x7e\x89\x7e\x98\x69\xd3\x0f\x28\xd2\x0f\x2f\xa8\x43\x88\x04\xd9\x0f\x17\x98\x04\x98\x68\xd3\x0f\x27\xd0\x08\x27\xe4\x07\x0a\x87\x7c\x81\x7c\x90\x78\xd2\x07\x1f\xa4\x43\xa7\x4e\xa2\x4e\xd9\x0f\x17\x98\x03\x98\x59\xac\x03\xaf\x0e\xa9\x0e\xd8\x18\x1f\xa4\x23\xd7\x22\x32\xd1\x22\x32\xb0\x32\xb0\x41\xd0\x22\x36\xd1\x18\x36\xf3\x03\x01\x10\x38\xf0\x00\x01\x09\x38\xf1\x06\x00\x0c\x14\x90\x43\x98\x18\xd3\x0b\x22\xd0\x04\x22",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_env_base = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "env_base",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_env_base._ascii.ob_base,
            & const_str_joinuser._ascii.ob_base,
            &_Py_ID(base),
        },
    },
};
static
    struct _PyCode_DEF(422)
site_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 211,
    },
    .co_consts = & site_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 285,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 688,
    .co_localsplusnames = & site_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str__getuserbase._ascii.ob_base,
    .co_qualname = & const_str__getuserbase._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x72\x02\x7c\x00\x53\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x76\x00\x72\x01\x79\x00\x64\x03\x84\x00\x7d\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x6b\x28\x00\x00\x72\x2c\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x01\x00\x00\x00\x00\x00\x00\x78\x01\x73\x02\x01\x00\x64\x06\x7d\x02\x02\x00\x7c\x01\x7c\x02\x64\x07\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x6b\x28\x00\x00\x72\x3d\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x2d\x02\x00\x7c\x01\x64\x06\x64\x09\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0a\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x00\x64\x0b\x1a\x00\x7a\x06\x00\x00\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00\x02\x00\x7c\x01\x64\x06\x64\x0c\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
site_toplevel_consts_12_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\Python",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
site_toplevel_consts_12_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\\site-packages",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
site_toplevel_consts_12_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "/lib/python/site-packages",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
site_toplevel_consts_12_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "/lib/python",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
site_toplevel_consts_12_consts_11 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "/site-packages",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
site_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(nt),
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            &_Py_STR(empty),
            & site_toplevel_consts_12_consts_4._ascii.ob_base,
            & site_toplevel_consts_12_consts_5._ascii.ob_base,
            & const_str_darwin._ascii.ob_base,
            & site_toplevel_consts_12_consts_7._ascii.ob_base,
            & site_toplevel_consts_12_consts_8._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & site_toplevel_consts_12_consts_11._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_winver = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "winver",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
site_toplevel_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str_version_info._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(name),
            & const_str_winver._ascii.ob_base,
            &_Py_ID(replace),
            & const_str_platform._ascii.ob_base,
            & const_str__framework._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__get_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_path",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[131];
    }
site_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 130,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x0e\x11\xd7\x0e\x1e\xd1\x0e\x1e\x80\x47\xe4\x07\x09\x87\x77\x81\x77\x90\x24\x82\x7f\xdc\x14\x17\x97\x4a\x91\x4a\xd7\x14\x26\xd1\x14\x26\xa0\x73\xa8\x42\xd3\x14\x2f\x88\x09\xd8\x12\x1a\x90\x1a\x98\x38\xa0\x49\xa0\x3b\xa8\x6f\xd0\x0f\x3e\xd0\x08\x3e\xe4\x07\x0a\x87\x7c\x81\x7c\x90\x78\xd2\x07\x1f\xa4\x43\xa7\x4e\xa2\x4e\xd8\x12\x1a\x90\x1a\xd0\x1b\x34\xd0\x0f\x35\xd0\x08\x35\xe0\x0e\x16\x88\x5a\x90\x7b\xa0\x37\xa8\x31\xa1\x3a\xa0\x2c\xa8\x61\xb0\x07\xb8\x01\xb1\x0a\xa8\x7c\xb8\x3e\xd0\x0b\x4a\xd0\x04\x4a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_userbase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "userbase",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_ver_nodot = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ver_nodot",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_12_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_userbase._ascii.ob_base,
            &_Py_ID(version),
            & const_str_ver_nodot._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(266)
site_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 133,
    },
    .co_consts = & site_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 309,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 689,
    .co_localsplusnames = & site_toplevel_consts_12_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str__get_path._ascii.ob_base,
    .co_qualname = & const_str__get_path._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x28\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x9b\x00\x64\x04\x7c\x02\x9b\x00\x64\x05\x9d\x04\x53\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\x6b\x28\x00\x00\x72\x15\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x05\x7c\x00\x9b\x00\x64\x07\x9d\x02\x53\x00\x7c\x00\x9b\x00\x64\x08\x7c\x01\x64\x09\x19\x00\x00\x00\x9b\x00\x64\x02\x7c\x01\x64\x0a\x19\x00\x00\x00\x9b\x00\x64\x0b\x9d\x06\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[204];
    }
site_toplevel_consts_13_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 203,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x60\x75\x73\x65\x72\x20\x62\x61\x73\x65\x60\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x70\x61\x74\x68\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x60\x75\x73\x65\x72\x20\x62\x61\x73\x65\x60\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x63\x61\x6e\x20\x62\x65\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x73\x74\x6f\x72\x65\x20\x64\x61\x74\x61\x2e\x20\x49\x66\x20\x74\x68\x65\x20\x67\x6c\x6f\x62\x61\x6c\x0a\x20\x20\x20\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x20\x60\x60\x55\x53\x45\x52\x5f\x42\x41\x53\x45\x60\x60\x20\x69\x73\x20\x6e\x6f\x74\x20\x69\x6e\x69\x74\x69\x61\x6c\x69\x7a\x65\x64\x20\x79\x65\x74\x2c\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x77\x69\x6c\x6c\x20\x61\x6c\x73\x6f\x20\x73\x65\x74\x0a\x20\x20\x20\x20\x69\x74\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
site_toplevel_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & site_toplevel_consts_13_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_USER_BASE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "USER_BASE",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_USER_BASE._ascii.ob_base,
            & const_str__getuserbase._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_getuserbase = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getuserbase",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
site_toplevel_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x08\x11\xd0\x07\x18\xdc\x14\x20\x93\x4e\x88\x09\xdc\x0b\x14\xd0\x04\x14",
};
static
    struct _PyCode_DEF(46)
site_toplevel_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 23,
    },
    .co_consts = & site_toplevel_consts_13_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 322,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 690,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_getuserbase._ascii.ob_base,
    .co_qualname = & const_str_getuserbase._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x0a\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x61\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[163];
    }
site_toplevel_consts_14_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 162,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x75\x73\x65\x72\x2d\x73\x70\x65\x63\x69\x66\x69\x63\x20\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x70\x61\x74\x68\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x74\x68\x65\x20\x67\x6c\x6f\x62\x61\x6c\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x20\x60\x60\x55\x53\x45\x52\x5f\x53\x49\x54\x45\x60\x60\x20\x69\x73\x20\x6e\x6f\x74\x20\x69\x6e\x69\x74\x69\x61\x6c\x69\x7a\x65\x64\x20\x79\x65\x74\x2c\x20\x74\x68\x69\x73\x0a\x20\x20\x20\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x77\x69\x6c\x6c\x20\x61\x6c\x73\x6f\x20\x73\x65\x74\x20\x69\x74\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & site_toplevel_consts_14_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_USER_SITE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "USER_SITE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_ENABLE_USER_SITE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ENABLE_USER_SITE",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
site_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_getuserbase._ascii.ob_base,
            & const_str_USER_SITE._ascii.ob_base,
            & const_str_ENABLE_USER_SITE._ascii.ob_base,
            & const_str__get_path._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_getusersitepackages = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getusersitepackages",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[56];
    }
site_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 55,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x10\x1b\x8b\x7d\x80\x48\xe4\x07\x10\xd0\x07\x18\xd8\x0b\x13\xd0\x0b\x1b\xd8\x1f\x24\xd0\x0c\x1c\xf4\x08\x00\x0c\x15\xd0\x04\x14\xf4\x05\x00\x19\x22\xa0\x28\xd3\x18\x2b\x88\x49\xe4\x0b\x14\xd0\x04\x14",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
site_toplevel_consts_14_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_userbase._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(88)
site_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 44,
    },
    .co_consts = & site_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 335,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 691,
    .co_localsplusnames = & site_toplevel_consts_14_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_getusersitepackages._ascii.ob_base,
    .co_qualname = & const_str_getusersitepackages._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x80\x15\x7c\x00\x80\x08\x64\x01\x61\x02\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x61\x01\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[135];
    }
site_toplevel_consts_15_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 134,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x64\x64\x20\x61\x20\x70\x65\x72\x20\x75\x73\x65\x72\x20\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x20\x74\x6f\x20\x73\x79\x73\x2e\x70\x61\x74\x68\x0a\x0a\x20\x20\x20\x20\x45\x61\x63\x68\x20\x75\x73\x65\x72\x20\x68\x61\x73\x20\x69\x74\x73\x20\x6f\x77\x6e\x20\x70\x79\x74\x68\x6f\x6e\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x77\x69\x74\x68\x20\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x69\x6e\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x68\x6f\x6d\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
site_toplevel_consts_15_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Processing user site-packages",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & site_toplevel_consts_15_consts_0._ascii.ob_base,
            & site_toplevel_consts_15_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
site_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__trace._ascii.ob_base,
            & const_str_getusersitepackages._ascii.ob_base,
            & const_str_ENABLE_USER_SITE._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_isdir._ascii.ob_base,
            & const_str_addsitedir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str_addusersitepackages = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "addusersitepackages",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[56];
    }
site_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 55,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x05\x0b\xd0\x0b\x2a\xd4\x04\x2b\xdc\x10\x23\xd3\x10\x25\x80\x49\xe5\x07\x17\x9c\x42\x9f\x47\x99\x47\x9f\x4d\x99\x4d\xa8\x29\xd4\x1c\x34\xdc\x08\x12\x90\x39\x98\x6b\xd4\x08\x2a\xd8\x0b\x16\xd0\x04\x16",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_user_site = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "user_site",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_15_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_known_paths._ascii.ob_base,
            & const_str_user_site._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(146)
site_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 73,
    },
    .co_consts = & site_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 352,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 692,
    .co_localsplusnames = & site_toplevel_consts_15_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_addusersitepackages._ascii.ob_base,
    .co_qualname = & const_str_addusersitepackages._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x72\x2b\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0c\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[287];
    }
site_toplevel_consts_16_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 286,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x73\x20\x61\x20\x6c\x69\x73\x74\x20\x63\x6f\x6e\x74\x61\x69\x6e\x69\x6e\x67\x20\x61\x6c\x6c\x20\x67\x6c\x6f\x62\x61\x6c\x20\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x46\x6f\x72\x20\x65\x61\x63\x68\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x70\x72\x65\x73\x65\x6e\x74\x20\x69\x6e\x20\x60\x60\x70\x72\x65\x66\x69\x78\x65\x73\x60\x60\x20\x28\x6f\x72\x20\x74\x68\x65\x20\x67\x6c\x6f\x62\x61\x6c\x20\x60\x60\x50\x52\x45\x46\x49\x58\x45\x53\x60\x60\x29\x2c\x0a\x20\x20\x20\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x77\x69\x6c\x6c\x20\x66\x69\x6e\x64\x20\x69\x74\x73\x20\x60\x73\x69\x74\x65\x2d\x70\x61\x63\x6b\x61\x67\x65\x73\x60\x20\x73\x75\x62\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x64\x65\x70\x65\x6e\x64\x69\x6e\x67\x20\x6f\x6e\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x73\x79\x73\x74\x65\x6d\x20\x65\x6e\x76\x69\x72\x6f\x6e\x6d\x65\x6e\x74\x2c\x20\x61\x6e\x64\x20\x77\x69\x6c\x6c\x20\x72\x65\x74\x75\x72\x6e\x20\x61\x20\x6c\x69\x73\x74\x20\x6f\x66\x20\x66\x75\x6c\x6c\x20\x70\x61\x74\x68\x73\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_lib = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "lib",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
site_toplevel_consts_16_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "python%d.%d",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
site_toplevel_consts_16_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "site-packages",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_Lib = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Lib",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
site_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & site_toplevel_consts_16_consts_0._ascii.ob_base,
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[47],
            & const_str_lib._ascii.ob_base,
            & site_toplevel_consts_16_consts_4._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & site_toplevel_consts_16_consts_6._ascii.ob_base,
            & const_str_Lib._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_PREFIXES = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "PREFIXES",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_platlibdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "platlibdir",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
site_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_set._ascii.ob_base,
            & const_str_PREFIXES._ascii.ob_base,
            &_Py_ID(add),
            & const_str_os._ascii.ob_base,
            &_Py_ID(sep),
            & const_str_sys._ascii.ob_base,
            & const_str_platlibdir._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(path),
            &_Py_ID(join),
            & const_str_version_info._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_getsitepackages = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "getsitepackages",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[236];
    }
site_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 235,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0e\x00\x14\x16\x80\x4c\xdc\x0b\x0e\x8b\x35\x80\x44\xe0\x07\x0f\xd0\x07\x17\xdc\x13\x1b\x88\x08\xe3\x12\x1a\x88\x06\xd9\x0f\x15\x98\x16\xa0\x34\x99\x1e\xd8\x0c\x14\xd8\x08\x0c\x8f\x08\x89\x08\x90\x16\xd4\x08\x18\xe4\x0b\x0d\x8f\x36\x89\x36\x90\x53\x8a\x3d\xdc\x17\x1a\x97\x7e\x91\x7e\xd0\x16\x26\x88\x47\xdc\x0f\x12\x8f\x7e\x89\x7e\xa0\x15\xd2\x0f\x26\xd8\x10\x17\x97\x0e\x91\x0e\x98\x75\xd4\x10\x25\xe3\x1a\x21\x90\x06\xdc\x17\x19\x97\x77\x91\x77\x97\x7c\x91\x7c\xa0\x46\xa8\x46\xd8\x24\x31\xb4\x43\xd7\x34\x44\xd1\x34\x44\xc0\x52\xc0\x61\xd0\x34\x48\xd1\x24\x48\xd8\x24\x33\xf3\x05\x02\x18\x35\x90\x04\xf0\x06\x00\x11\x1d\xd7\x10\x23\xd1\x10\x23\xa0\x44\xd5\x10\x29\xf1\x09\x00\x1b\x22\xf0\x0c\x00\x0d\x19\xd7\x0c\x1f\xd1\x0c\x1f\xa0\x06\xd4\x0c\x27\xd8\x0c\x18\xd7\x0c\x1f\xd1\x0c\x1f\xa4\x02\xa7\x07\xa1\x07\xa7\x0c\xa1\x0c\xa8\x56\xb0\x55\xb8\x4f\xd3\x20\x4c\xd5\x0c\x4d\xf0\x23\x00\x13\x1b\xf0\x24\x00\x0c\x18\xd0\x04\x17",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_prefixes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "prefixes",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_sitepackages = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sitepackages",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_libdirs = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "libdirs",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_libdir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "libdir",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
site_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_prefixes._ascii.ob_base,
            & const_str_sitepackages._ascii.ob_base,
            & const_str_seen._ascii.ob_base,
            & const_str_prefix._ascii.ob_base,
            & const_str_libdirs._ascii.ob_base,
            & const_str_libdir._ascii.ob_base,
            &_Py_ID(path),
        },
    },
};
static
    struct _PyCode_DEF(540)
site_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 270,
    },
    .co_consts = & site_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 17 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 10,
    .co_firstlineno = 367,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 693,
    .co_localsplusnames = & site_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_getsitepackages._ascii.ob_base,
    .co_qualname = & const_str_getsitepackages._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x67\x00\x7d\x01\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x80\x06\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x00\x7c\x00\x44\x00\x5d\xf2\x00\x00\x7d\x03\x7c\x03\x72\x04\x7c\x03\x7c\x02\x76\x00\x72\x01\x8c\x0a\x7c\x02\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x6b\x28\x00\x00\x72\x84\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x01\x7d\x04\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x6b\x37\x00\x00\x72\x11\x7c\x04\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x44\x00\x5d\x49\x00\x00\x7d\x05\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x05\x64\x04\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x05\x1a\x00\x7a\x06\x00\x00\x64\x06\xab\x04\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x4b\x04\x00\x8c\xb2\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x13\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x07\x64\x06\xab\x03\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\xf4\x04\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[30];
    }
site_toplevel_consts_17_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 29,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Add site-packages to sys.path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
site_toplevel_consts_17_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Processing global site-packages",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & site_toplevel_consts_17_consts_0._ascii.ob_base,
            & site_toplevel_consts_17_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
site_toplevel_consts_17_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__trace._ascii.ob_base,
            & const_str_getsitepackages._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_isdir._ascii.ob_base,
            & const_str_addsitedir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str_addsitepackages = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "addsitepackages",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[61];
    }
site_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 60,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x04\x0a\xd0\x0b\x2c\xd4\x04\x2d\xdc\x13\x22\xa0\x38\xd6\x13\x2c\x88\x07\xdc\x0b\x0d\x8f\x37\x89\x37\x8f\x3d\x89\x3d\x98\x17\xd5\x0b\x21\xdc\x0c\x16\x90\x77\xa0\x0b\xd5\x0c\x2c\xf0\x05\x00\x14\x2d\xf0\x08\x00\x0c\x17\xd0\x04\x16",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_17_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_known_paths._ascii.ob_base,
            & const_str_prefixes._ascii.ob_base,
            & const_str_sitedir._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(148)
site_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 74,
    },
    .co_consts = & site_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_17_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 8 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 400,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 694,
    .co_localsplusnames = & site_toplevel_consts_17_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_addsitepackages._ascii.ob_base,
    .co_qualname = & const_str_addsitepackages._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x2e\x00\x00\x7d\x02\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x8c\x23\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x30\x04\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[174];
    }
site_toplevel_consts_18_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 173,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x44\x65\x66\x69\x6e\x65\x20\x6e\x65\x77\x20\x62\x75\x69\x6c\x74\x69\x6e\x73\x20\x27\x71\x75\x69\x74\x27\x20\x61\x6e\x64\x20\x27\x65\x78\x69\x74\x27\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x73\x65\x20\x61\x72\x65\x20\x6f\x62\x6a\x65\x63\x74\x73\x20\x77\x68\x69\x63\x68\x20\x6d\x61\x6b\x65\x20\x74\x68\x65\x20\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x20\x65\x78\x69\x74\x20\x77\x68\x65\x6e\x20\x63\x61\x6c\x6c\x65\x64\x2e\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x72\x65\x70\x72\x20\x6f\x66\x20\x65\x61\x63\x68\x20\x6f\x62\x6a\x65\x63\x74\x20\x63\x6f\x6e\x74\x61\x69\x6e\x73\x20\x61\x20\x68\x69\x6e\x74\x20\x61\x74\x20\x68\x6f\x77\x20\x69\x74\x20\x77\x6f\x72\x6b\x73\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
site_toplevel_consts_18_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Ctrl-Z plus Return",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
site_toplevel_consts_18_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Ctrl-D (i.e. EOF)",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_quit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "quit",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_exit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exit",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
site_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & site_toplevel_consts_18_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[92],
            & site_toplevel_consts_18_consts_2._ascii.ob_base,
            & site_toplevel_consts_18_consts_3._ascii.ob_base,
            & const_str_quit._ascii.ob_base,
            & const_str_exit._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__sitebuiltins = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_sitebuiltins",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
site_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            &_Py_ID(sep),
            & const_str__sitebuiltins._ascii.ob_base,
            & const_str_Quitter._ascii.ob_base,
            &_Py_ID(builtins),
            & const_str_quit._ascii.ob_base,
            & const_str_exit._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_setquit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "setquit",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[66];
    }
site_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 65,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x0e\x00\x08\x0a\x87\x76\x81\x76\x90\x14\x82\x7e\xd8\x0e\x22\x89\x03\xe0\x0e\x21\x88\x03\xe4\x14\x21\xd7\x14\x29\xd1\x14\x29\xa8\x26\xb0\x23\xd3\x14\x36\x84\x48\x84\x4d\xdc\x14\x21\xd7\x14\x29\xd1\x14\x29\xa8\x26\xb0\x23\xd3\x14\x36\x84\x48\x85\x4d",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
site_toplevel_consts_18_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_eof._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(176)
site_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 88,
    },
    .co_consts = & site_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 409,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 695,
    .co_localsplusnames = & site_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_setquit._ascii.ob_base,
    .co_qualname = & const_str_setquit._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x03\x64\x02\x7d\x00\x6e\x02\x64\x03\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x79\x06",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
site_toplevel_consts_19_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Set 'copyright' and 'credits' in builtins",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_copyright = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "copyright",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_credits = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "credits",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[192];
    }
site_toplevel_consts_19_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 191,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x20\x20\x20\x54\x68\x61\x6e\x6b\x73\x20\x74\x6f\x20\x43\x57\x49\x2c\x20\x43\x4e\x52\x49\x2c\x20\x42\x65\x4f\x70\x65\x6e\x2c\x20\x5a\x6f\x70\x65\x20\x43\x6f\x72\x70\x6f\x72\x61\x74\x69\x6f\x6e\x2c\x20\x74\x68\x65\x20\x50\x79\x74\x68\x6f\x6e\x20\x53\x6f\x66\x74\x77\x61\x72\x65\x0a\x20\x20\x20\x20\x46\x6f\x75\x6e\x64\x61\x74\x69\x6f\x6e\x2c\x20\x61\x6e\x64\x20\x61\x20\x63\x61\x73\x74\x20\x6f\x66\x20\x74\x68\x6f\x75\x73\x61\x6e\x64\x73\x20\x66\x6f\x72\x20\x73\x75\x70\x70\x6f\x72\x74\x69\x6e\x67\x20\x50\x79\x74\x68\x6f\x6e\x0a\x20\x20\x20\x20\x64\x65\x76\x65\x6c\x6f\x70\x6d\x65\x6e\x74\x2e\x20\x20\x53\x65\x65\x20\x77\x77\x77\x2e\x70\x79\x74\x68\x6f\x6e\x2e\x6f\x72\x67\x20\x66\x6f\x72\x20\x6d\x6f\x72\x65\x20\x69\x6e\x66\x6f\x72\x6d\x61\x74\x69\x6f\x6e\x2e",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
site_toplevel_consts_19_consts_7 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LICENSE.txt",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_LICENSE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LICENSE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_license = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "license",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
site_toplevel_consts_19_consts_10 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "See https://www.python.org/psf/license/",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
site_toplevel_consts_19_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & site_toplevel_consts_19_consts_0._ascii.ob_base,
            & const_str_copyright._ascii.ob_base,
            & const_str_credits._ascii.ob_base,
            & site_toplevel_consts_19_consts_3._ascii.ob_base,
            & const_str__stdlib_dir._ascii.ob_base,
            Py_None,
            &_Py_ID(__file__),
            & site_toplevel_consts_19_consts_7._ascii.ob_base,
            & const_str_LICENSE._ascii.ob_base,
            & const_str_license._ascii.ob_base,
            & site_toplevel_consts_19_consts_10._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
site_toplevel_consts_19_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            & const_str__sitebuiltins._ascii.ob_base,
            & const_str__Printer._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_copyright._ascii.ob_base,
            &_Py_ID(builtins),
            & const_str_credits._ascii.ob_base,
            &_Py_ID(getattr),
            & const_str_hasattr._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_dirname._ascii.ob_base,
            &_Py_ID(__file__),
            &_Py_ID(extend),
            &_Py_ID(join),
            & const_str_pardir._ascii.ob_base,
            & const_str_curdir._ascii.ob_base,
            & const_str_license._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_setcopyright = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "setcopyright",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[207];
    }
site_toplevel_consts_19_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 206,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x19\x26\xd7\x19\x2f\xd1\x19\x2f\xb0\x0b\xbc\x53\xbf\x5d\xb9\x5d\xd3\x19\x4b\x84\x48\xd4\x04\x16\xdc\x17\x24\xd7\x17\x2d\xd1\x17\x2d\xa8\x69\xf0\x00\x03\x3a\x3e\xf3\x00\x03\x18\x3f\x84\x48\xd4\x04\x14\xf0\x08\x00\x13\x15\x90\x62\x88\x34\x80\x45\xf4\x06\x00\x0c\x13\x94\x33\x98\x0d\xa0\x74\xd3\x0b\x2c\x80\x44\xd9\x0b\x0f\x94\x47\x9c\x42\xa0\x0a\xd4\x14\x2b\xdc\x0f\x11\x8f\x77\x89\x77\x8f\x7f\x89\x7f\x9c\x72\x9f\x7b\x99\x7b\xd3\x0f\x2b\x88\x04\xd9\x07\x0b\xd8\x08\x0d\x8f\x0c\x89\x0c\x90\x6d\xa0\x59\xd0\x15\x2f\xd4\x08\x30\xd8\x08\x0c\x8f\x0b\x89\x0b\x94\x52\x97\x57\x91\x57\x97\x5c\x91\x5c\xa0\x24\xac\x02\xaf\x09\xa9\x09\xd3\x15\x32\xb0\x44\xbc\x22\xbf\x29\xb9\x29\xd0\x14\x44\xd4\x08\x45\xdc\x17\x24\xd7\x17\x2d\xd1\x17\x2d\xd8\x08\x11\xd8\x08\x31\xd8\x08\x0d\x88\x74\xf3\x07\x03\x18\x15\x84\x48\xd5\x04\x14",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_here = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "here",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_19_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_files._ascii.ob_base,
            & const_str_dirs._ascii.ob_base,
            & const_str_here._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(588)
site_toplevel_consts_19 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 294,
    },
    .co_consts = & site_toplevel_consts_19_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_19_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 425,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 696,
    .co_localsplusnames = & site_toplevel_consts_19_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_setcopyright._ascii.ob_base,
    .co_qualname = & const_str_setcopyright._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_19_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x05\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x67\x00\x7d\x01\x7d\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x64\x05\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x73\x3d\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x72\x2d\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x72\x61\x7c\x00\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\x64\x08\x67\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x10\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\x64\x0a\x7c\x00\x7c\x01\xab\x04\x00\x00\x00\x00\x00\x00\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x10\x00\x00\x00\x00\x00\x00\x00\x00\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
site_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__sitebuiltins._ascii.ob_base,
            & const_str__Helper._ascii.ob_base,
            &_Py_ID(builtins),
            & const_str_help._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_sethelper = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sethelper",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
site_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x14\x21\xd7\x14\x29\xd1\x14\x29\xd3\x14\x2b\x84\x48\x85\x4d",
};
static
    struct _PyCode_DEF(62)
site_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 447,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 697,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_sethelper._ascii.ob_base,
    .co_qualname = & const_str_sethelper._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[363];
    }
site_toplevel_consts_21_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 362,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x6e\x61\x62\x6c\x65\x20\x64\x65\x66\x61\x75\x6c\x74\x20\x72\x65\x61\x64\x6c\x69\x6e\x65\x20\x63\x6f\x6e\x66\x69\x67\x75\x72\x61\x74\x69\x6f\x6e\x20\x6f\x6e\x20\x69\x6e\x74\x65\x72\x61\x63\x74\x69\x76\x65\x20\x70\x72\x6f\x6d\x70\x74\x73\x2c\x20\x62\x79\x0a\x20\x20\x20\x20\x72\x65\x67\x69\x73\x74\x65\x72\x69\x6e\x67\x20\x61\x20\x73\x79\x73\x2e\x5f\x5f\x69\x6e\x74\x65\x72\x61\x63\x74\x69\x76\x65\x68\x6f\x6f\x6b\x5f\x5f\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x74\x68\x65\x20\x72\x65\x61\x64\x6c\x69\x6e\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x63\x61\x6e\x20\x62\x65\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2c\x20\x74\x68\x65\x20\x68\x6f\x6f\x6b\x20\x77\x69\x6c\x6c\x20\x73\x65\x74\x20\x74\x68\x65\x20\x54\x61\x62\x20\x6b\x65\x79\x0a\x20\x20\x20\x20\x61\x73\x20\x63\x6f\x6d\x70\x6c\x65\x74\x69\x6f\x6e\x20\x6b\x65\x79\x20\x61\x6e\x64\x20\x72\x65\x67\x69\x73\x74\x65\x72\x20\x7e\x2f\x2e\x70\x79\x74\x68\x6f\x6e\x5f\x68\x69\x73\x74\x6f\x72\x79\x20\x61\x73\x20\x68\x69\x73\x74\x6f\x72\x79\x20\x66\x69\x6c\x65\x2e\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x63\x61\x6e\x20\x62\x65\x20\x6f\x76\x65\x72\x72\x69\x64\x64\x65\x6e\x20\x69\x6e\x20\x74\x68\x65\x20\x73\x69\x74\x65\x63\x75\x73\x74\x6f\x6d\x69\x7a\x65\x20\x6f\x72\x20\x75\x73\x65\x72\x63\x75\x73\x74\x6f\x6d\x69\x7a\x65\x20\x6d\x6f\x64\x75\x6c\x65\x2c\x0a\x20\x20\x20\x20\x6f\x72\x20\x69\x6e\x20\x61\x20\x50\x59\x54\x48\x4f\x4e\x53\x54\x41\x52\x54\x55\x50\x20\x66\x69\x6c\x65\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_libedit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "libedit",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
site_toplevel_consts_21_consts_1_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bind ^I rl_complete",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
site_toplevel_consts_21_consts_1_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "tab: complete",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
site_toplevel_consts_21_consts_1_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".python_history",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
const_str_write_history_file = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "write_history_file",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_21_consts_1_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_write_history_file._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_write_history = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "write_history",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[68];
    }
site_toplevel_consts_21_consts_1_consts_9_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 67,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "enablerlcompleter.<locals>.register_readline.<locals>.write_history",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
site_toplevel_consts_21_consts_1_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xf0\x02\x05\x11\x19\xd8\x14\x1c\xd7\x14\x2f\xd1\x14\x2f\xb0\x07\xd5\x14\x38\xf8\xdc\x17\x1e\xf2\x00\x03\x11\x19\xf1\x06\x00\x15\x19\xf0\x07\x03\x11\x19\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[13];
    }
site_toplevel_consts_21_consts_1_consts_9_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 12,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x11\x15\x00\x95\x09\x21\x03\xa0\x01\x21\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_history = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "history",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_21_consts_1_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_history._ascii.ob_base,
            &_Py_ID(readline),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
site_toplevel_consts_21_consts_1_consts_9_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x80",
};
static
    struct _PyCode_DEF(72)
site_toplevel_consts_21_consts_1_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 36,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_21_consts_1_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_21_consts_1_consts_9_exceptiontable.ob_base.ob_base,
    .co_flags = 19,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 497,
    .co_nlocalsplus = 2,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 2,
    .co_version = 698,
    .co_localsplusnames = & site_toplevel_consts_21_consts_1_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & site_toplevel_consts_21_consts_1_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_write_history._ascii.ob_base,
    .co_qualname = & site_toplevel_consts_21_consts_1_consts_9_qualname._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_21_consts_1_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x02\x97\x00\x09\x00\x89\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
site_toplevel_consts_21_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_ID(__doc__),
            &_Py_STR(empty),
            & const_str_libedit._ascii.ob_base,
            & site_toplevel_consts_21_consts_1_consts_5._ascii.ob_base,
            & site_toplevel_consts_21_consts_1_consts_6._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[126],
            & site_toplevel_consts_21_consts_1_consts_8._ascii.ob_base,
            & site_toplevel_consts_21_consts_1_consts_9.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_atexit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "atexit",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_rlcompleter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "rlcompleter",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_parse_and_bind = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "parse_and_bind",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_read_init_file = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "read_init_file",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str_get_current_history_length = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_current_history_length",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_read_history_file = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "read_history_file",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
site_toplevel_consts_21_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & const_str_atexit._ascii.ob_base,
            &_Py_ID(readline),
            & const_str_rlcompleter._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(getattr),
            & const_str_parse_and_bind._ascii.ob_base,
            & const_str_read_init_file._ascii.ob_base,
            & const_str_OSError._ascii.ob_base,
            & const_str_get_current_history_length._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            &_Py_ID(join),
            & const_str_expanduser._ascii.ob_base,
            & const_str_read_history_file._ascii.ob_base,
            & const_str_register._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_register_readline = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "register_readline",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
site_toplevel_consts_21_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "enablerlcompleter.<locals>.register_readline",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[255];
    }
site_toplevel_consts_21_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 254,
    },
    .ob_shash = -1,
    .ob_sval = "\xf9\x80\x00\xdb\x08\x15\xf0\x02\x04\x09\x13\xdb\x0c\x1b\xdb\x0c\x1e\xf4\x0c\x00\x18\x1f\x98\x78\xa8\x19\xb0\x42\xd3\x17\x37\x88\x0c\xd8\x0b\x17\xd0\x0b\x23\xa8\x09\xb0\x5c\xd1\x28\x41\xd8\x0c\x14\xd7\x0c\x23\xd1\x0c\x23\xd0\x24\x39\xd5\x0c\x3a\xe0\x0c\x14\xd7\x0c\x23\xd1\x0c\x23\xa0\x4f\xd4\x0c\x34\xf0\x04\x07\x09\x11\xd8\x0c\x14\xd7\x0c\x23\xd1\x0c\x23\xd4\x0c\x25\xf0\x10\x00\x0c\x14\xd7\x0b\x2e\xd1\x0b\x2e\xd3\x0b\x30\xb0\x41\xd2\x0b\x35\xf4\x0c\x00\x17\x19\x97\x67\x91\x67\x97\x6c\x91\x6c\xa4\x32\xa7\x37\xa1\x37\xd7\x23\x35\xd1\x23\x35\xb0\x63\xd3\x23\x3a\xd8\x23\x34\xf3\x03\x01\x17\x36\x88\x47\xf0\x04\x03\x0d\x15\xd8\x10\x18\xd7\x10\x2a\xd1\x10\x2a\xa8\x37\xd4\x10\x33\xf5\x08\x06\x0d\x19\xf0\x10\x00\x0d\x13\x8f\x4f\x89\x4f\x98\x4d\xd5\x0c\x2a\xf0\x2b\x00\x0c\x36\xf8\xf4\x29\x00\x10\x1b\xf2\x00\x01\x09\x13\xd9\x0c\x12\xf0\x03\x01\x09\x13\xfb\xf4\x1a\x00\x10\x17\xf2\x00\x05\x09\x11\xf1\x0a\x00\x0d\x11\xf0\x0b\x05\x09\x11\xfb\xf4\x22\x00\x14\x1b\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[54];
    }
site_toplevel_consts_21_consts_1_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 53,
    },
    .ob_shash = -1,
    .ob_sval = "\x88\x08\x43\x12\x00\xc1\x07\x10\x43\x21\x00\xc2\x28\x11\x43\x30\x00\xc3\x12\x09\x43\x1e\x03\xc3\x1d\x01\x43\x1e\x03\xc3\x21\x09\x43\x2d\x03\xc3\x2c\x01\x43\x2d\x03\xc3\x30\x09\x43\x3c\x03\xc3\x3b\x01\x43\x3c\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_readline_doc = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "readline_doc",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
site_toplevel_consts_21_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_atexit._ascii.ob_base,
            & const_str_rlcompleter._ascii.ob_base,
            & const_str_readline_doc._ascii.ob_base,
            & const_str_write_history._ascii.ob_base,
            & const_str_history._ascii.ob_base,
            &_Py_ID(readline),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[7];
    }
site_toplevel_consts_21_consts_1_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 6,
    },
    .ob_shash = -1,
    .ob_sval = "    @@",
};
static
    struct _PyCode_DEF(510)
site_toplevel_consts_21_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 255,
    },
    .co_consts = & site_toplevel_consts_21_consts_1_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_21_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_21_consts_1_exceptiontable.ob_base.ob_base,
    .co_flags = 19,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 459,
    .co_nlocalsplus = 6,
    .co_nlocals = 4,
    .co_ncellvars = 2,
    .co_nfreevars = 0,
    .co_version = 699,
    .co_localsplusnames = & site_toplevel_consts_21_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & site_toplevel_consts_21_consts_1_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_register_readline._ascii.ob_base,
    .co_qualname = & site_toplevel_consts_21_consts_1_qualname._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_21_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x04\x87\x05\x97\x00\x64\x01\x64\x00\x6c\x00\x7d\x00\x09\x00\x64\x01\x64\x00\x6c\x01\x8a\x05\x64\x01\x64\x00\x6c\x02\x7d\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x89\x05\x64\x02\x64\x03\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x81\x16\x64\x04\x7c\x02\x76\x00\x72\x12\x89\x05\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x11\x89\x05\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x89\x05\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x89\x05\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x67\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x07\xab\x01\x00\x00\x00\x00\x00\x00\x64\x08\xab\x02\x00\x00\x00\x00\x00\x00\x8a\x04\x09\x00\x89\x05\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x88\x04\x88\x05\x66\x02\x64\x09\x84\x08\x7d\x03\x7c\x00\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x79\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x95\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x42\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & site_toplevel_consts_21_consts_0._ascii.ob_base,
            & site_toplevel_consts_21_consts_1.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str___interactivehook__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__interactivehook__",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_21_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            & const_str___interactivehook__._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_enablerlcompleter = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "enablerlcompleter",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
site_toplevel_consts_21_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf2\x12\x2e\x05\x2b\xf0\x60\x01\x00\x1f\x30\x84\x43\xd5\x04\x1b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
site_toplevel_consts_21_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_register_readline._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(32)
site_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & site_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 450,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 700,
    .co_localsplusnames = & site_toplevel_consts_21_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_enablerlcompleter._ascii.ob_base,
    .co_qualname = & const_str_enablerlcompleter._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_21_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x84\x00\x7d\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str___PYVENV_LAUNCHER__ = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__PYVENV_LAUNCHER__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
site_toplevel_consts_22_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pyvenv.cfg",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_22_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_isfile._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
site_toplevel_consts_22_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "venv.<locals>.<genexpr>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[47];
    }
site_toplevel_consts_22_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 46,
    },
    .ob_shash = -1,
    .ob_sval = "\xe8\x00\xf8\x80\x00\xf0\x00\x06\x09\x0a\xf1\x02\x03\x26\x0e\x98\x18\xf4\x08\x00\x10\x12\x8f\x77\x89\x77\x8f\x7e\x89\x7e\x98\x68\xd4\x0f\x27\xf4\x09\x00\x0d\x15\xf1\x00\x03\x26\x0e\xf9",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_conffile = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "conffile",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_22_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib__bootstrap_external_toplevel_consts_6_localsplusnames_0._ascii.ob_base,
            & const_str_conffile._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(94)
site_toplevel_consts_22_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 47,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_22_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = & _collections_abc_toplevel_consts_68_consts_7_exceptiontable.ob_base.ob_base,
    .co_flags = 51,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 522,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 701,
    .co_localsplusnames = & site_toplevel_consts_22_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_genexpr),
    .co_qualname = & site_toplevel_consts_22_consts_4_qualname._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_22_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x4b\x00\x01\x00\x97\x00\x7c\x00\x5d\x25\x00\x00\x7d\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x72\x04\x7c\x01\x96\x01\x97\x01\x01\x00\x8c\x27\x04\x00\x79\x00\xad\x03\x77\x01",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
site_toplevel_consts_22_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "include-system-site-packages",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_home = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "home",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
site_toplevel_consts_22_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            Py_None,
            & const_str_darwin._ascii.ob_base,
            & const_str___PYVENV_LAUNCHER__._ascii.ob_base,
            & site_toplevel_consts_22_consts_3._ascii.ob_base,
            & site_toplevel_consts_22_consts_4.ob_base.ob_base,
            &_Py_ID(true),
            &_Py_STR(utf_8),
            & codecs_toplevel_consts_35_localsplusnames._object.ob_base.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[61],
            & site_toplevel_consts_22_consts_9._ascii.ob_base,
            & const_str_home._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__base_executable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_base_executable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_executable = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "executable",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__home = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_home",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_exec_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exec_prefix",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
site_toplevel_consts_22_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            & const_str_environ._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_platform._ascii.ob_base,
            & const_str__base_executable._ascii.ob_base,
            & const_str_executable._ascii.ob_base,
            &_Py_ID(path),
            & const_str_dirname._ascii.ob_base,
            & const_str_abspath._ascii.ob_base,
            & const_str__home._ascii.ob_base,
            &_Py_ID(next),
            &_Py_ID(join),
            &_Py_ID(open),
            & const_str_partition._ascii.ob_base,
            & const_str_strip._ascii.ob_base,
            & const_str_lower._ascii.ob_base,
            & const_str_prefix._ascii.ob_base,
            & const_str_exec_prefix._ascii.ob_base,
            & const_str_addsitepackages._ascii.ob_base,
            & const_str_PREFIXES._ascii.ob_base,
            & const_str_insert._ascii.ob_base,
            & const_str_ENABLE_USER_SITE._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_venv = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "venv",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[433];
    }
site_toplevel_consts_22_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 432,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x06\x00\x0b\x0d\x8f\x2a\x89\x2a\x80\x43\xdc\x07\x0a\x87\x7c\x81\x7c\x90\x78\xd2\x07\x1f\xd0\x24\x39\xb8\x53\xd1\x24\x40\xdc\x2c\x2e\xaf\x4a\xa9\x4a\xd0\x37\x4c\xd1\x2c\x4d\xd0\x08\x4d\x88\x0a\x94\x53\xd5\x15\x29\xe4\x15\x18\x97\x5e\x91\x5e\x88\x0a\xdc\x0e\x10\x8f\x67\x89\x67\x8f\x6f\x89\x6f\x9c\x62\x9f\x67\x99\x67\x9f\x6f\x99\x6f\xa8\x6a\xd3\x1e\x39\xd3\x0e\x3a\x80\x47\xdc\x12\x14\x97\x27\x91\x27\x97\x2f\x91\x2f\xa0\x27\xd3\x12\x2a\x80\x4b\xd8\x10\x14\x84\x43\x84\x49\xd8\x14\x20\x80\x4d\xdc\x15\x19\xf1\x02\x06\x09\x0a\xe4\x10\x12\x97\x07\x91\x07\x97\x0c\x91\x0c\x98\x57\xa0\x6d\xd3\x10\x34\xdc\x10\x12\x97\x07\x91\x07\x97\x0c\x91\x0c\x98\x5b\xa8\x2d\xd3\x10\x38\xf1\x05\x03\x26\x0e\xf3\x03\x06\x09\x0a\xf0\x0e\x00\x09\x0d\xf3\x11\x09\x16\x06\x80\x4e\xf2\x16\x00\x08\x16\xd8\x17\x25\x88\x0c\xd8\x16\x1c\x88\x0b\xf4\x06\x00\x0e\x12\x90\x2c\xa8\x17\xd5\x0d\x31\xb0\x51\xdb\x18\x19\x90\x04\xd8\x13\x16\x98\x24\x92\x3b\xd8\x24\x28\xa7\x4e\xa1\x4e\xb0\x33\xd3\x24\x37\x91\x4d\x90\x43\x98\x11\x98\x45\xd8\x1a\x1d\x9f\x29\x99\x29\x9b\x2b\xd7\x1a\x2b\xd1\x1a\x2b\xd3\x1a\x2d\x90\x43\xd8\x1c\x21\x9f\x4b\x99\x4b\x9b\x4d\x90\x45\xd8\x17\x1a\xd0\x1e\x3c\xd2\x17\x3c\xd8\x26\x2b\xa7\x6b\xa1\x6b\xa3\x6d\x99\x0b\xd8\x19\x1c\xa0\x06\x9b\x1d\xd8\x24\x29\x9c\x03\x9d\x09\xf1\x11\x00\x19\x1a\xf7\x03\x00\x0e\x32\xf0\x16\x00\x28\x33\xd0\x08\x32\x8c\x03\x8c\x0a\x94\x53\x94\x5f\xf4\x06\x00\x09\x18\x98\x0b\xa4\x63\xa7\x6a\xa1\x6a\xa0\x5c\xd4\x08\x32\xf0\x08\x00\x0c\x17\x98\x26\xd2\x0b\x20\xdc\x0c\x14\x8f\x4f\x89\x4f\x98\x41\x9c\x73\x9f\x7a\x99\x7a\xd4\x0c\x2a\xf0\x0a\x00\x0c\x17\xd0\x04\x16\xf4\x07\x00\x19\x1c\x9f\x0a\x99\x0a\x90\x7c\x88\x48\xd8\x1f\x24\xd0\x0c\x1c\xe0\x0b\x16\xd0\x04\x16\xf7\x31\x00\x0e\x32\xd0\x0d\x31\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
site_toplevel_consts_22_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\xc4\x24\x0a\x48\x15\x03\xc4\x2f\x41\x1e\x48\x15\x03\xc6\x0e\x0d\x48\x15\x03\xc8\x15\x05\x48\x1e\x07",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_exe_dir = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exe_dir",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_site_prefix = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "site_prefix",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_conf_basename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "conf_basename",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_candidate_conf = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "candidate_conf",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_virtual_conf = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "virtual_conf",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_system_site = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "system_site",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[14];
        }_object;
    }
site_toplevel_consts_22_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 14,
        },
        .ob_item = {
            & const_str_known_paths._ascii.ob_base,
            &_Py_ID(env),
            & const_str_executable._ascii.ob_base,
            & const_str_exe_dir._ascii.ob_base,
            & const_str_site_prefix._ascii.ob_base,
            & const_str_conf_basename._ascii.ob_base,
            & const_str_candidate_conf._ascii.ob_base,
            & const_str_virtual_conf._ascii.ob_base,
            & const_str_system_site._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[102],
            &_Py_ID(line),
            &_Py_ID(key),
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(1090)
site_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 545,
    },
    .co_consts = & site_toplevel_consts_22_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_22_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_22_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 22 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 509,
    .co_nlocalsplus = 14,
    .co_nlocals = 14,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 702,
    .co_localsplusnames = & site_toplevel_consts_22_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_external_toplevel_consts_72_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_venv._ascii.ob_base,
    .co_qualname = & const_str_venv._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_22_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x6b\x28\x00\x00\x72\x23\x64\x02\x7c\x01\x76\x00\x72\x1f\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\x78\x01\x7d\x02\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6e\x10\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x64\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x7d\x05\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x84\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x05\xab\x02\x00\x00\x00\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x7c\x05\xab\x02\x00\x00\x00\x00\x00\x00\x66\x02\x44\x00\xab\x00\x00\x00\x00\x00\x00\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x06\x90\x01\x72\x00\x7c\x06\x7d\x07\x64\x05\x7d\x08\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x64\x06\xac\x07\xab\x02\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x09\x7c\x09\x44\x00\x5d\x71\x00\x00\x7d\x0a\x64\x08\x7c\x0a\x76\x00\x73\x01\x8c\x08\x7c\x0a\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x0b\x7d\x0c\x7d\x0d\x7c\x0b\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x0b\x7c\x0d\x6a\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x0d\x7c\x0b\x64\x09\x6b\x28\x00\x00\x72\x11\x7c\x0d\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x8c\x61\x7c\x0b\x64\x0a\x6b\x28\x00\x00\x73\x01\x8c\x67\x7c\x0d\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x09\x00\x00\x00\x00\x00\x00\x00\x00\x8c\x73\x04\x00\x09\x00\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x78\x01\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x10\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x11\x00\x00\x00\x00\x00\x00\x00\x00\x74\x25\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x08\x64\x05\x6b\x28\x00\x00\x72\x26\x74\x26\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x29\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x01\x61\x13\x64\x0c\x61\x15\x7c\x00\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x7a\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
site_toplevel_consts_23_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Run custom site specific code, if available.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_sitecustomize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sitecustomize",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[58];
    }
site_toplevel_consts_23_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 57,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x73\x69\x74\x65\x63\x75\x73\x74\x6f\x6d\x69\x7a\x65\x3b\x20\x73\x65\x74\x20\x50\x59\x54\x48\x4f\x4e\x56\x45\x52\x42\x4f\x53\x45\x20\x66\x6f\x72\x20\x74\x72\x61\x63\x65\x62\x61\x63\x6b\x3a\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
site_toplevel_consts_23_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & site_toplevel_consts_23_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & const_str_sitecustomize._ascii.ob_base,
            & site_toplevel_consts_23_consts_4._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_42_consts_4._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[10],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_exc_info = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "exc_info",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
site_toplevel_consts_23_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_sitecustomize._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(name),
            & const_str_Exception._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_verbose._ascii.ob_base,
            &_Py_ID(excepthook),
            & const_str_exc_info._ascii.ob_base,
            &_Py_ID(stderr),
            &_Py_ID(write),
            &_Py_ID(__class__),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_execsitecustomize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execsitecustomize",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[153];
    }
site_toplevel_consts_23_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 152,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x04\x0f\x05\x2f\xf0\x02\x06\x09\x16\xdc\x0c\x20\xf8\xdc\x0f\x1a\xf2\x00\x04\x09\x16\xd8\x0f\x12\x8f\x78\x89\x78\x98\x3f\xd2\x0f\x2a\xd8\x10\x14\xe0\x10\x15\xf4\x05\x00\x11\x15\xfb\xf0\x05\x04\x09\x16\xfb\xf4\x0a\x00\x0c\x15\xf2\x00\x07\x05\x2f\xdc\x0b\x0e\x8f\x39\x89\x39\xd7\x0b\x1c\xd2\x0b\x1c\xdc\x0c\x0f\x8f\x4e\x89\x4e\x9c\x43\x9f\x4c\x99\x4c\x9b\x4e\xd2\x0c\x2b\xe4\x0c\x0f\x8f\x4a\x89\x4a\xd7\x0c\x1c\xd2\x0c\x1c\xf0\x06\x00\x12\x15\x97\x1d\x91\x1d\xd7\x11\x27\xd3\x11\x27\xaa\x13\xf0\x05\x02\x11\x2e\xf7\x03\x03\x0d\x2f\xf1\x00\x03\x0d\x2f\xf4\x05\x00\x0d\x2c\xfb\xf0\x05\x07\x05\x2f\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[42];
    }
site_toplevel_consts_23_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 41,
    },
    .ob_shash = -1,
    .ob_sval = "\x83\x04\x08\x00\x88\x09\x2c\x03\x91\x11\x27\x03\xa2\x04\x2f\x00\xa7\x05\x2c\x03\xac\x03\x2f\x00\xaf\x09\x43\x00\x03\xb8\x41\x39\x42\x3b\x03\xc2\x3b\x05\x43\x00\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_23_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_sitecustomize._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
            & const_str_err._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(390)
site_toplevel_consts_23 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 195,
    },
    .co_consts = & site_toplevel_consts_23_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_23_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_23_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 564,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 703,
    .co_localsplusnames = & site_toplevel_consts_23_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_execsitecustomize._ascii.ob_base,
    .co_qualname = & const_str_execsitecustomize._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_23_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x09\x00\x64\x01\x64\x02\x6c\x00\x7d\x00\x79\x02\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x1b\x7d\x01\x7c\x01\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x6b\x28\x00\x00\x72\x01\x6e\x01\x82\x00\x59\x00\x64\x02\x7d\x01\x7e\x01\x79\x02\x64\x02\x7d\x01\x7e\x01\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x88\x7d\x02\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x25\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x8e\x00\x01\x00\x6e\x3f\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x02\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x01\x64\x05\x7c\x02\x9b\x01\x64\x06\x9d\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x02\x7d\x02\x7e\x02\x79\x02\x59\x00\x64\x02\x7d\x02\x7e\x02\x79\x02\x64\x02\x7d\x02\x7e\x02\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
site_toplevel_consts_24_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Run custom user specific code, if available.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_usercustomize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "usercustomize",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[58];
    }
site_toplevel_consts_24_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 57,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x45\x72\x72\x6f\x72\x20\x69\x6e\x20\x75\x73\x65\x72\x63\x75\x73\x74\x6f\x6d\x69\x7a\x65\x3b\x20\x73\x65\x74\x20\x50\x59\x54\x48\x4f\x4e\x56\x45\x52\x42\x4f\x53\x45\x20\x66\x6f\x72\x20\x74\x72\x61\x63\x65\x62\x61\x63\x6b\x3a\x0a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
site_toplevel_consts_24_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & site_toplevel_consts_24_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & const_str_usercustomize._ascii.ob_base,
            & site_toplevel_consts_24_consts_4._ascii.ob_base,
            & importlib__bootstrap_external_toplevel_consts_42_consts_4._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[10],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
site_toplevel_consts_24_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_usercustomize._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(name),
            & const_str_Exception._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_verbose._ascii.ob_base,
            &_Py_ID(excepthook),
            & const_str_exc_info._ascii.ob_base,
            &_Py_ID(stderr),
            &_Py_ID(write),
            &_Py_ID(__class__),
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
const_str_execusercustomize = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "execusercustomize",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_24_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_usercustomize._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
            & const_str_err._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(390)
site_toplevel_consts_24 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 195,
    },
    .co_consts = & site_toplevel_consts_24_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_24_names._object.ob_base.ob_base,
    .co_exceptiontable = & site_toplevel_consts_23_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 584,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 704,
    .co_localsplusnames = & site_toplevel_consts_24_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_execusercustomize._ascii.ob_base,
    .co_qualname = & const_str_execusercustomize._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_23_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x09\x00\x64\x01\x64\x02\x6c\x00\x7d\x00\x79\x02\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x1b\x7d\x01\x7c\x01\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\x6b\x28\x00\x00\x72\x01\x6e\x01\x82\x00\x59\x00\x64\x02\x7d\x01\x7e\x01\x79\x02\x64\x02\x7d\x01\x7e\x01\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x88\x7d\x02\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x25\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x8e\x00\x01\x00\x6e\x3f\x74\x08\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x15\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x02\x6a\x16\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x01\x64\x05\x7c\x02\x9b\x01\x64\x06\x9d\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x02\x7d\x02\x7e\x02\x79\x02\x59\x00\x64\x02\x7d\x02\x7e\x02\x79\x02\x64\x02\x7d\x02\x7e\x02\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[208];
    }
site_toplevel_consts_25_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 207,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x64\x64\x20\x73\x74\x61\x6e\x64\x61\x72\x64\x20\x73\x69\x74\x65\x2d\x73\x70\x65\x63\x69\x66\x69\x63\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x69\x65\x73\x20\x74\x6f\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x73\x65\x61\x72\x63\x68\x20\x70\x61\x74\x68\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x69\x73\x20\x63\x61\x6c\x6c\x65\x64\x20\x61\x75\x74\x6f\x6d\x61\x74\x69\x63\x61\x6c\x6c\x79\x20\x77\x68\x65\x6e\x20\x74\x68\x69\x73\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2c\x0a\x20\x20\x20\x20\x75\x6e\x6c\x65\x73\x73\x20\x74\x68\x65\x20\x70\x79\x74\x68\x6f\x6e\x20\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x20\x77\x61\x73\x20\x73\x74\x61\x72\x74\x65\x64\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x2d\x53\x20\x66\x6c\x61\x67\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_25_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & site_toplevel_consts_25_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_isolated = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "isolated",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
site_toplevel_consts_25_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(path),
            & const_str_removeduppaths._ascii.ob_base,
            & const_str_abs_paths._ascii.ob_base,
            & const_str_venv._ascii.ob_base,
            & const_str_ENABLE_USER_SITE._ascii.ob_base,
            & const_str_check_enableusersite._ascii.ob_base,
            & const_str_addusersitepackages._ascii.ob_base,
            & const_str_addsitepackages._ascii.ob_base,
            & const_str_setquit._ascii.ob_base,
            & const_str_setcopyright._ascii.ob_base,
            & const_str_sethelper._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_isolated._ascii.ob_base,
            & const_str_enablerlcompleter._ascii.ob_base,
            & const_str_execsitecustomize._ascii.ob_base,
            & const_str_execusercustomize._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_main = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "main",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[144];
    }
site_toplevel_consts_25_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 143,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x10\x00\x11\x14\x97\x08\x91\x08\x99\x11\x90\x0b\x80\x49\xdc\x12\x20\xd3\x12\x22\x80\x4b\xd8\x07\x10\x94\x43\x97\x48\x91\x48\xd2\x07\x1c\xf4\x06\x00\x09\x12\x8c\x0b\xe4\x12\x16\x90\x7b\xd3\x12\x23\x80\x4b\xdc\x07\x17\xd0\x07\x1f\xdc\x1b\x2f\xd3\x1b\x31\xd0\x08\x18\xdc\x12\x25\xa0\x6b\xd3\x12\x32\x80\x4b\xdc\x12\x21\xa0\x2b\xd3\x12\x2e\x80\x4b\xdc\x04\x0b\x84\x49\xdc\x04\x10\x84\x4e\xdc\x04\x0d\x84\x4b\xdc\x0b\x0e\x8f\x39\x89\x39\xd7\x0b\x1d\xd2\x0b\x1d\xdc\x08\x19\xd4\x08\x1b\xdc\x04\x15\xd4\x04\x17\xdd\x07\x17\xdc\x08\x19\xd5\x08\x1b\xf0\x03\x00\x08\x18",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_orig_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "orig_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
site_toplevel_consts_25_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_orig_path._ascii.ob_base,
            & const_str_known_paths._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(404)
site_toplevel_consts_25 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 202,
    },
    .co_consts = & site_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_25_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 604,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 705,
    .co_localsplusnames = & site_toplevel_consts_25_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_main._ascii.ob_base,
    .co_qualname = & const_str_main._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_25_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x64\x01\x1a\x00\x7d\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x0a\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x80\x0a\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x61\x05\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x01\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x73\x0a\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x1f\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x72\x0b\x74\x21\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[435];
    }
site_toplevel_consts_26_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 434,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x20\x20\x20\x20\x25\x73\x20\x5b\x2d\x2d\x75\x73\x65\x72\x2d\x62\x61\x73\x65\x5d\x20\x5b\x2d\x2d\x75\x73\x65\x72\x2d\x73\x69\x74\x65\x5d\x0a\x0a\x20\x20\x20\x20\x57\x69\x74\x68\x6f\x75\x74\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x70\x72\x69\x6e\x74\x20\x73\x6f\x6d\x65\x20\x75\x73\x65\x66\x75\x6c\x20\x69\x6e\x66\x6f\x72\x6d\x61\x74\x69\x6f\x6e\x0a\x20\x20\x20\x20\x57\x69\x74\x68\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x70\x72\x69\x6e\x74\x20\x74\x68\x65\x20\x76\x61\x6c\x75\x65\x20\x6f\x66\x20\x55\x53\x45\x52\x5f\x42\x41\x53\x45\x20\x61\x6e\x64\x2f\x6f\x72\x20\x55\x53\x45\x52\x5f\x53\x49\x54\x45\x20\x73\x65\x70\x61\x72\x61\x74\x65\x64\x0a\x20\x20\x20\x20\x62\x79\x20\x27\x25\x73\x27\x2e\x0a\x0a\x20\x20\x20\x20\x45\x78\x69\x74\x20\x63\x6f\x64\x65\x73\x20\x77\x69\x74\x68\x20\x2d\x2d\x75\x73\x65\x72\x2d\x62\x61\x73\x65\x20\x6f\x72\x20\x2d\x2d\x75\x73\x65\x72\x2d\x73\x69\x74\x65\x3a\x0a\x20\x20\x20\x20\x20\x20\x30\x20\x2d\x20\x75\x73\x65\x72\x20\x73\x69\x74\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x65\x6e\x61\x62\x6c\x65\x64\x0a\x20\x20\x20\x20\x20\x20\x31\x20\x2d\x20\x75\x73\x65\x72\x20\x73\x69\x74\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x64\x69\x73\x61\x62\x6c\x65\x64\x20\x62\x79\x20\x75\x73\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x32\x20\x2d\x20\x75\x73\x65\x72\x20\x73\x69\x74\x65\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x69\x73\x20\x64\x69\x73\x61\x62\x6c\x65\x64\x20\x62\x79\x20\x73\x75\x70\x65\x72\x20\x75\x73\x65\x72\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x6f\x72\x20\x66\x6f\x72\x20\x73\x65\x63\x75\x72\x69\x74\x79\x20\x72\x65\x61\x73\x6f\x6e\x73\x0a\x20\x20\x20\x20\x20\x3e\x32\x20\x2d\x20\x75\x6e\x6b\x6e\x6f\x77\x6e\x20\x65\x72\x72\x6f\x72\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
site_toplevel_consts_26_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "sys.path = [",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
site_toplevel_consts_26_consts_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "    ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
site_toplevel_consts_26_consts_7_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "doesn't exist",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_26_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & const_str_exists._ascii.ob_base,
            & site_toplevel_consts_26_consts_7_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
site_toplevel_consts_26_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_isdir._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
site_toplevel_consts_26_consts_7_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_script.<locals>.exists",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
site_toplevel_consts_26_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\xd0\x0f\x1f\xa4\x42\xa7\x47\xa1\x47\xa7\x4d\xa1\x4d\xb0\x24\xd4\x24\x37\xd8\x17\x1f\xe0\x17\x26",
};
static
    struct _PyCode_DEF(72)
site_toplevel_consts_26_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 36,
    },
    .co_consts = & site_toplevel_consts_26_consts_7_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_26_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 19,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 661,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 706,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str_exists._ascii.ob_base,
    .co_qualname = & site_toplevel_consts_26_consts_7_qualname._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_26_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x81\x20\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x72\x01\x79\x01\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
site_toplevel_consts_26_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "USER_BASE: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
site_toplevel_consts_26_consts_11 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "USER_SITE: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
site_toplevel_consts_26_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ENABLE_USER_SITE: ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
site_toplevel_consts_26_consts_14 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "--user-base",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
site_toplevel_consts_26_consts_15 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "--user-site",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[20];
        }_object;
    }
site_toplevel_consts_26_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 20,
        },
        .ob_item = {
            Py_None,
            & site_toplevel_consts_26_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & site_toplevel_consts_26_consts_3._ascii.ob_base,
            & site_toplevel_consts_26_consts_4._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[44],
            (PyObject *)&_Py_SINGLETON(strings).ascii[93],
            & site_toplevel_consts_26_consts_7.ob_base.ob_base,
            & site_toplevel_consts_26_consts_8._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_29_consts_8._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[41],
            & site_toplevel_consts_26_consts_11._ascii.ob_base,
            & site_toplevel_consts_26_consts_12._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & site_toplevel_consts_26_consts_14._ascii.ob_base,
            & site_toplevel_consts_26_consts_15._ascii.ob_base,
            Py_False,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 10],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_textwrap = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "textwrap",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_dedent = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "dedent",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[16];
        }_object;
    }
site_toplevel_consts_26_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 16,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(argv),
            & const_str_getuserbase._ascii.ob_base,
            & const_str_getusersitepackages._ascii.ob_base,
            & const_str_print._ascii.ob_base,
            &_Py_ID(path),
            & const_str_ENABLE_USER_SITE._ascii.ob_base,
            & const_str_exit._ascii.ob_base,
            &_Py_ID(append),
            & const_str_USER_BASE._ascii.ob_base,
            & const_str_USER_SITE._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_pathsep._ascii.ob_base,
            &_Py_ID(join),
            & const_str_textwrap._ascii.ob_base,
            & const_str_dedent._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str__script = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_script",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[362];
    }
site_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 361,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x02\x0d\x0c\x08\x80\x44\xf4\x1c\x00\x0c\x0f\x8f\x38\x89\x38\x90\x41\x90\x42\x88\x3c\x80\x44\xd9\x0b\x0f\xdc\x14\x1f\x93\x4d\x88\x09\xdc\x14\x27\xd3\x14\x29\x88\x09\xdc\x08\x0d\x88\x6e\xd4\x08\x1d\xdc\x13\x16\x97\x38\x94\x38\x88\x43\xdd\x0c\x11\x9a\x73\xd0\x12\x24\xd5\x0c\x25\xf0\x03\x00\x14\x1c\xe4\x08\x0d\x88\x63\x8c\x0a\xf2\x02\x04\x09\x27\xf4\x0a\x00\x09\x0e\x90\x0b\x98\x49\x98\x3d\xa8\x02\xa9\x36\xb0\x29\xd3\x2b\x3c\xd0\x2a\x3d\xb8\x51\xd0\x0e\x3f\xd4\x08\x40\xdc\x08\x0d\x90\x0b\x98\x49\x98\x3d\xa8\x02\xa9\x36\xb0\x29\xd3\x2b\x3c\xd0\x2a\x3d\xb8\x51\xd0\x0e\x3f\xd4\x08\x40\xdc\x08\x0d\xd0\x10\x22\xd4\x23\x33\xd0\x22\x36\xd0\x0e\x37\xd4\x08\x38\xdc\x08\x0b\x8f\x08\x89\x08\x90\x11\x8c\x0b\xe0\x0d\x0f\x80\x46\xd8\x07\x14\x98\x04\xd1\x07\x1c\xd8\x08\x0e\x8f\x0d\x89\x0d\x94\x69\xd4\x08\x20\xd8\x07\x14\x98\x04\xd1\x07\x1c\xd8\x08\x0e\x8f\x0d\x89\x0d\x94\x69\xd4\x08\x20\xe1\x07\x0d\xdc\x08\x0d\x8c\x62\x8f\x6a\x89\x6a\x8f\x6f\x89\x6f\x98\x66\xd3\x0e\x25\xd4\x08\x26\xdd\x0b\x1b\xdc\x0c\x0f\x8f\x48\x89\x48\x90\x51\x8d\x4b\xdc\x0d\x1d\xa0\x15\xd1\x0d\x26\xdc\x0c\x0f\x8f\x48\x89\x48\x90\x51\x8d\x4b\xdc\x0d\x1d\xd0\x0d\x25\xdc\x0c\x0f\x8f\x48\x89\x48\x90\x51\x8d\x4b\xe4\x0c\x0f\x8f\x48\x89\x48\x90\x51\x8d\x4b\xe3\x08\x17\xdc\x08\x0d\x88\x68\x8f\x6f\x89\x6f\x98\x64\xa4\x63\xa7\x68\xa1\x68\xa8\x71\xa1\x6b\xb4\x32\xb7\x3a\xb1\x3a\xd0\x25\x3e\xd1\x1e\x3e\xd3\x0e\x3f\xd4\x08\x40\xdc\x08\x0b\x8f\x08\x89\x08\x90\x12\x8d\x0c",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_user_base = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "user_base",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
site_toplevel_consts_26_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_help._ascii.ob_base,
            &_Py_ID(args),
            & const_str_user_base._ascii.ob_base,
            & const_str_user_site._ascii.ob_base,
            & const_str_dir._ascii.ob_base,
            & const_str_exists._ascii.ob_base,
            &_Py_ID(buffer),
            & const_str_textwrap._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(964)
site_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 482,
    },
    .co_consts = & site_toplevel_consts_26_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 638,
    .co_nlocalsplus = 8,
    .co_nlocals = 8,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 707,
    .co_localsplusnames = & site_toplevel_consts_26_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_13_localspluskinds.ob_base.ob_base,
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = & const_str__script._ascii.ob_base,
    .co_qualname = & const_str__script._ascii.ob_base,
    .co_linetable = & site_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x64\x00\x1a\x00\x7d\x01\x7c\x01\x73\xa8\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x11\x00\x00\x7d\x04\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x04\x9b\x02\x64\x05\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x13\x04\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x06\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x07\x84\x00\x7d\x05\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x7c\x02\x9b\x02\x64\x09\x02\x00\x7c\x05\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x0a\x9d\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0b\x7c\x03\x9b\x02\x64\x09\x02\x00\x7c\x05\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x0a\x9d\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0c\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x67\x00\x7d\x06\x64\x0e\x7c\x01\x76\x00\x72\x15\x7c\x06\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x64\x0f\x7c\x01\x76\x00\x72\x15\x7c\x06\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x14\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x06\x72\x94\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x72\x16\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x64\x10\x75\x00\x72\x16\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x74\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x80\x16\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x11\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x12\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00\x64\x0d\x64\x00\x6c\x0e\x7d\x07\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x0d\x19\x00\x00\x00\x74\x16\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x18\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x66\x02\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x13\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[29];
        }_object;
    }
site_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 29,
        },
        .ob_item = {
            & site_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & site_toplevel_consts_3.ob_base.ob_base,
            & site_toplevel_consts_4.ob_base.ob_base,
            & site_toplevel_consts_5.ob_base.ob_base,
            & site_toplevel_consts_6.ob_base.ob_base,
            & site_toplevel_consts_7.ob_base.ob_base,
            & site_toplevel_consts_8.ob_base.ob_base,
            & site_toplevel_consts_9.ob_base.ob_base,
            & site_toplevel_consts_10.ob_base.ob_base,
            & site_toplevel_consts_11.ob_base.ob_base,
            & site_toplevel_consts_12.ob_base.ob_base,
            & site_toplevel_consts_13.ob_base.ob_base,
            & site_toplevel_consts_14.ob_base.ob_base,
            & site_toplevel_consts_15.ob_base.ob_base,
            & site_toplevel_consts_16.ob_base.ob_base,
            & site_toplevel_consts_17.ob_base.ob_base,
            & site_toplevel_consts_18.ob_base.ob_base,
            & site_toplevel_consts_19.ob_base.ob_base,
            & site_toplevel_consts_20.ob_base.ob_base,
            & site_toplevel_consts_21.ob_base.ob_base,
            & site_toplevel_consts_22.ob_base.ob_base,
            & site_toplevel_consts_23.ob_base.ob_base,
            & site_toplevel_consts_24.ob_base.ob_base,
            & site_toplevel_consts_25.ob_base.ob_base,
            & site_toplevel_consts_26.ob_base.ob_base,
            &_Py_ID(__main__),
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_no_site = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "no_site",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[40];
        }_object;
    }
site_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 40,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_sys._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(builtins),
            & const_str__sitebuiltins._ascii.ob_base,
            & const_str_io._ascii.ob_base,
            & const_str_stat._ascii.ob_base,
            & const_str_prefix._ascii.ob_base,
            & const_str_exec_prefix._ascii.ob_base,
            & const_str_PREFIXES._ascii.ob_base,
            & const_str_ENABLE_USER_SITE._ascii.ob_base,
            & const_str_USER_SITE._ascii.ob_base,
            & const_str_USER_BASE._ascii.ob_base,
            & const_str__trace._ascii.ob_base,
            & const_str_makepath._ascii.ob_base,
            & const_str_abs_paths._ascii.ob_base,
            & const_str_removeduppaths._ascii.ob_base,
            & const_str__init_pathinfo._ascii.ob_base,
            & const_str_addpackage._ascii.ob_base,
            & const_str_addsitedir._ascii.ob_base,
            & const_str_check_enableusersite._ascii.ob_base,
            & const_str__getuserbase._ascii.ob_base,
            & const_str__get_path._ascii.ob_base,
            & const_str_getuserbase._ascii.ob_base,
            & const_str_getusersitepackages._ascii.ob_base,
            & const_str_addusersitepackages._ascii.ob_base,
            & const_str_getsitepackages._ascii.ob_base,
            & const_str_addsitepackages._ascii.ob_base,
            & const_str_setquit._ascii.ob_base,
            & const_str_setcopyright._ascii.ob_base,
            & const_str_sethelper._ascii.ob_base,
            & const_str_enablerlcompleter._ascii.ob_base,
            & const_str_venv._ascii.ob_base,
            & const_str_execsitecustomize._ascii.ob_base,
            & const_str_execusercustomize._ascii.ob_base,
            & const_str_main._ascii.ob_base,
            &_Py_ID(flags),
            & const_str_no_site._ascii.ob_base,
            & const_str__script._ascii.ob_base,
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[240];
    }
site_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 239,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x45\x01\x01\x04\xf3\x4e\x02\x00\x01\x0b\xdb\x00\x09\xdb\x00\x0f\xdb\x00\x14\xdb\x00\x09\xdb\x00\x0b\xf0\x06\x00\x0d\x10\x8f\x4a\x89\x4a\x98\x03\x9f\x0f\x99\x0f\xd0\x0b\x28\x80\x08\xf0\x06\x00\x14\x18\xd0\x00\x10\xf0\x0a\x00\x0d\x11\x80\x09\xd8\x0c\x10\x80\x09\xf2\x06\x02\x01\x28\xf2\x0a\x06\x01\x26\xf2\x12\x14\x01\x11\xf2\x2e\x10\x01\x17\xf2\x26\x0a\x01\x0d\xf2\x1a\x3f\x01\x17\xf3\x44\x02\x17\x01\x17\xf2\x34\x16\x01\x10\xf2\x40\x01\x14\x01\x23\xf2\x30\x0a\x01\x4b\x01\xf2\x1a\x0a\x01\x15\xf2\x1a\x0f\x01\x15\xf2\x22\x0d\x01\x17\xf3\x1e\x1f\x01\x18\xf3\x42\x01\x07\x01\x17\xf2\x12\x0d\x01\x37\xf2\x20\x13\x01\x15\xf2\x2c\x01\x01\x2c\xf2\x06\x39\x01\x30\xf2\x76\x01\x34\x01\x17\xf2\x6e\x01\x11\x01\x2f\xf2\x28\x11\x01\x2f\xf2\x28\x1b\x01\x1c\xf0\x3e\x00\x08\x0b\x87\x79\x81\x79\xd7\x07\x18\xd2\x07\x18\xd9\x04\x08\x84\x46\xf2\x04\x34\x01\x15\xf0\x6c\x01\x00\x04\x0c\x88\x7a\xd2\x03\x19\xd9\x04\x0b\x85\x49\xf0\x03\x00\x04\x1a",
};
static
    struct _PyCode_DEF(350)
site_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 175,
    },
    .co_consts = & site_toplevel_consts._object.ob_base.ob_base,
    .co_names = & site_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 708,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & site_toplevel_consts_3_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & site_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x5a\x01\x64\x01\x64\x02\x6c\x02\x5a\x02\x64\x01\x64\x02\x6c\x03\x5a\x03\x64\x01\x64\x02\x6c\x04\x5a\x04\x64\x01\x64\x02\x6c\x05\x5a\x05\x64\x01\x64\x02\x6c\x06\x5a\x06\x65\x01\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x65\x01\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x67\x02\x61\x09\x64\x02\x61\x0a\x64\x02\x61\x0b\x64\x02\x61\x0c\x64\x03\x84\x00\x5a\x0d\x64\x04\x84\x00\x5a\x0e\x64\x05\x84\x00\x5a\x0f\x64\x06\x84\x00\x5a\x10\x64\x07\x84\x00\x5a\x11\x64\x08\x84\x00\x5a\x12\x64\x1c\x64\x09\x84\x01\x5a\x13\x64\x0a\x84\x00\x5a\x14\x64\x0b\x84\x00\x5a\x15\x64\x0c\x84\x00\x5a\x16\x64\x0d\x84\x00\x5a\x17\x64\x0e\x84\x00\x5a\x18\x64\x0f\x84\x00\x5a\x19\x64\x1c\x64\x10\x84\x01\x5a\x1a\x64\x1c\x64\x11\x84\x01\x5a\x1b\x64\x12\x84\x00\x5a\x1c\x64\x13\x84\x00\x5a\x1d\x64\x14\x84\x00\x5a\x1e\x64\x15\x84\x00\x5a\x1f\x64\x16\x84\x00\x5a\x20\x64\x17\x84\x00\x5a\x21\x64\x18\x84\x00\x5a\x22\x64\x19\x84\x00\x5a\x23\x65\x01\x6a\x48\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x4a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x73\x07\x02\x00\x65\x23\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x64\x1a\x84\x00\x5a\x26\x65\x27\x64\x1b\x6b\x28\x00\x00\x72\x08\x02\x00\x65\x26\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02\x79\x02",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_site_toplevel(void)
{
    return Py_NewRef((PyObject *) &site_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[112];
    }
stat_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 111,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x43\x6f\x6e\x73\x74\x61\x6e\x74\x73\x2f\x66\x75\x6e\x63\x74\x69\x6f\x6e\x73\x20\x66\x6f\x72\x20\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x69\x6e\x67\x20\x72\x65\x73\x75\x6c\x74\x73\x20\x6f\x66\x20\x6f\x73\x2e\x73\x74\x61\x74\x28\x29\x20\x61\x6e\x64\x20\x6f\x73\x2e\x6c\x73\x74\x61\x74\x28\x29\x2e\x0a\x0a\x53\x75\x67\x67\x65\x73\x74\x65\x64\x20\x75\x73\x61\x67\x65\x3a\x20\x66\x72\x6f\x6d\x20\x73\x74\x61\x74\x20\x69\x6d\x70\x6f\x72\x74\x20\x2a\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[78];
    }
stat_toplevel_consts_11_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 77,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x6f\x72\x74\x69\x6f\x6e\x20\x6f\x66\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x27\x73\x20\x6d\x6f\x64\x65\x20\x74\x68\x61\x74\x20\x63\x61\x6e\x20\x62\x65\x20\x73\x65\x74\x20\x62\x79\x0a\x20\x20\x20\x20\x6f\x73\x2e\x63\x68\x6d\x6f\x64\x28\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_4095 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 4095 },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & stat_toplevel_consts_11_consts_0._ascii.ob_base,
            & const_int_4095.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
stat_toplevel_consts_11_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen stat>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IMODE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IMODE",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
stat_toplevel_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x0c\x10\x90\x26\x89\x3d\xd0\x04\x18",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(mode),
        },
    },
};
static
    struct _PyCode_DEF(12)
stat_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & stat_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 21,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 709,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_IMODE._ascii.ob_base,
    .co_qualname = & const_str_S_IMODE._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x64\x01\x7a\x01\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[77];
    }
stat_toplevel_consts_12_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 76,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x70\x6f\x72\x74\x69\x6f\x6e\x20\x6f\x66\x20\x74\x68\x65\x20\x66\x69\x6c\x65\x27\x73\x20\x6d\x6f\x64\x65\x20\x74\x68\x61\x74\x20\x64\x65\x73\x63\x72\x69\x62\x65\x73\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x66\x69\x6c\x65\x20\x74\x79\x70\x65\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & stat_toplevel_consts_12_consts_0._ascii.ob_base,
            & const_int_61440.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_S_IFMT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFMT",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[15];
    }
stat_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 14,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x0c\x10\x90\x28\x89\x3f\xd0\x04\x1a",
};
static
    struct _PyCode_DEF(12)
stat_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 6,
    },
    .co_consts = & stat_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 27,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 710,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_IFMT._ascii.ob_base,
    .co_qualname = & const_str_S_IFMT._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x64\x01\x7a\x01\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_8192 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 8192 },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_24576 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 24576 },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_4096 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 4096 },
};
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_40960 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 8192, 1 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_40960 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 40960 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_49152 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 16384, 1 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_49152 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 49152 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[41];
    }
stat_toplevel_consts_20_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 40,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a directory.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_20_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & stat_toplevel_consts_20_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IFDIR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFDIR",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_20_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFDIR._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
stat_toplevel_consts_20_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x11\x90\x24\x8b\x3c\x9c\x37\xd1\x0b\x22\xd0\x04\x22",
};
static
    struct _PyCode_DEF(38)
stat_toplevel_consts_20 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & stat_toplevel_consts_20_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_20_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 50,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 711,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISDIR._ascii.ob_base,
    .co_qualname = & const_str_S_ISDIR._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[61];
    }
stat_toplevel_consts_21_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 60,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a character special device file.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & stat_toplevel_consts_21_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IFCHR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFCHR",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_21_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFCHR._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISCHR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISCHR",
};
static
    struct _PyCode_DEF(38)
stat_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & stat_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 54,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 712,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISCHR._ascii.ob_base,
    .co_qualname = & const_str_S_ISCHR._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[57];
    }
stat_toplevel_consts_22_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 56,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a block special device file.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_22_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & stat_toplevel_consts_22_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IFBLK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFBLK",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_22_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFBLK._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISBLK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISBLK",
};
static
    struct _PyCode_DEF(38)
stat_toplevel_consts_22 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & stat_toplevel_consts_22_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_22_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 58,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 713,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISBLK._ascii.ob_base,
    .co_qualname = & const_str_S_ISBLK._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[44];
    }
stat_toplevel_consts_23_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 43,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a regular file.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_23_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & stat_toplevel_consts_23_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IFREG = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFREG",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_23_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFREG._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(38)
stat_toplevel_consts_23 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & stat_toplevel_consts_23_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_23_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 62,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 714,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISREG._ascii.ob_base,
    .co_qualname = & const_str_S_ISREG._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[49];
    }
stat_toplevel_consts_24_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 48,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a FIFO (named pipe).",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_24_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & stat_toplevel_consts_24_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IFIFO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFIFO",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_24_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFIFO._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_ISFIFO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISFIFO",
};
static
    struct _PyCode_DEF(38)
stat_toplevel_consts_24 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & stat_toplevel_consts_24_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_24_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 66,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 715,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISFIFO._ascii.ob_base,
    .co_qualname = & const_str_S_ISFIFO._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[45];
    }
stat_toplevel_consts_25_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 44,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a symbolic link.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_25_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & stat_toplevel_consts_25_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IFLNK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFLNK",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_25_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFLNK._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(38)
stat_toplevel_consts_25 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & stat_toplevel_consts_25_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_25_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 70,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 716,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISLNK._ascii.ob_base,
    .co_qualname = & const_str_S_ISLNK._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_20_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
stat_toplevel_consts_26_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a socket.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
stat_toplevel_consts_26_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & stat_toplevel_consts_26_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_IFSOCK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFSOCK",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_26_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFSOCK._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_ISSOCK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISSOCK",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
stat_toplevel_consts_26_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x11\x90\x24\x8b\x3c\x9c\x38\xd1\x0b\x23\xd0\x04\x23",
};
static
    struct _PyCode_DEF(38)
stat_toplevel_consts_26 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 19,
    },
    .co_consts = & stat_toplevel_consts_26_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_26_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 74,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 717,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISSOCK._ascii.ob_base,
    .co_qualname = & const_str_S_ISSOCK._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_26_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6b\x28\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
stat_toplevel_consts_27_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a door.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_27_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & stat_toplevel_consts_27_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_ISDOOR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISDOOR",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
stat_toplevel_consts_27_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x10",
};
static
    struct _PyCode_DEF(4)
stat_toplevel_consts_27 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & stat_toplevel_consts_27_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 78,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 718,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISDOOR._ascii.ob_base,
    .co_qualname = & const_str_S_ISDOOR._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_27_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[43];
    }
stat_toplevel_consts_28_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 42,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from an event port.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_28_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & stat_toplevel_consts_28_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_ISPORT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISPORT",
};
static
    struct _PyCode_DEF(4)
stat_toplevel_consts_28 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & stat_toplevel_consts_28_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 82,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 719,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISPORT._ascii.ob_base,
    .co_qualname = & const_str_S_ISPORT._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_27_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
stat_toplevel_consts_29_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return True if mode is from a whiteout.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
stat_toplevel_consts_29_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & stat_toplevel_consts_29_consts_0._ascii.ob_base,
            Py_False,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISWHT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISWHT",
};
static
    struct _PyCode_DEF(4)
stat_toplevel_consts_29 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & stat_toplevel_consts_29_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 86,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 720,
    .co_localsplusnames = & stat_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_S_ISWHT._ascii.ob_base,
    .co_qualname = & const_str_S_ISWHT._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_27_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_1024 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 1024 },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_512 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 512 },
};
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_448 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 448 },
};
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_65536 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 0, 2 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_65536 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 65536 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_131072 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 0, 4 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_131072 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 131072 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_262144 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 0, 8 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_262144 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 262144 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_1048576 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 0, 32 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_1048576 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 1048576 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
#if PYLONG_BITS_IN_DIGIT == 15
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[2];
    }
const_int_2097152 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 2),
    .ob_digit = { 0, 64 },
};
#elif PYLONG_BITS_IN_DIGIT == 30
static
    struct {
        PyObject ob_base;
        uintptr_t lv_tag;
        digit ob_digit[1];
    }
const_int_2097152 = {
    .ob_base = {
        .ob_refcnt = _Py_IMMORTAL_REFCNT,
        .ob_type = &PyLong_Type,
    },
    .lv_tag = TAG_FROM_SIGN_AND_SIZE(1, 1),
    .ob_digit = { 2097152 },
};
#else
#error "PYLONG_BITS_IN_DIGIT should be 15 or 30"
#endif
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[60];
    }
stat_toplevel_consts_58_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 59,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Convert a file's mode to a string of the form '-rwxrwxrwx'.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
stat_toplevel_consts_58_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & stat_toplevel_consts_58_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[45],
            &_Py_STR(empty),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[16];
    }
const_str__filemode_table = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 15,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_filemode_table",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
stat_toplevel_consts_58_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__filemode_table._ascii.ob_base,
            &_Py_ID(append),
            &_Py_ID(join),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_filemode = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "filemode",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[89];
    }
stat_toplevel_consts_58_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 88,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0d\x80\x44\xdf\x11\x20\x88\x05\xdb\x19\x1e\x89\x49\x88\x43\x90\x14\xd8\x0f\x13\x90\x63\x89\x7a\x98\x53\xd3\x0f\x20\xd8\x10\x14\x97\x0b\x91\x0b\x98\x44\xd4\x10\x21\xd9\x10\x15\xf0\x07\x00\x1a\x1f\xf0\x0a\x00\x0d\x11\x8f\x4b\x89\x4b\x98\x03\xd5\x0c\x1c\xf0\x0d\x00\x12\x21\xf0\x0e\x00\x0c\x0e\x8f\x37\x89\x37\x90\x34\x8b\x3d\xd0\x04\x18",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_perm = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "perm",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_table = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "table",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[4];
    }
const_str_bit = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 3,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "bit",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_char = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "char",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
stat_toplevel_consts_58_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            &_Py_ID(mode),
            & const_str_perm._ascii.ob_base,
            & const_str_table._ascii.ob_base,
            & const_str_bit._ascii.ob_base,
            & const_str_char._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(170)
stat_toplevel_consts_58 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 85,
    },
    .co_consts = & stat_toplevel_consts_58_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_consts_58_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 156,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 721,
    .co_localsplusnames = & stat_toplevel_consts_58_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = & const_str_filemode._ascii.ob_base,
    .co_qualname = & const_str_filemode._ascii.ob_base,
    .co_linetable = & stat_toplevel_consts_58_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x67\x00\x7d\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x38\x00\x00\x7d\x02\x7c\x02\x44\x00\x5d\x20\x00\x00\x5c\x02\x00\x00\x7d\x03\x7d\x04\x7c\x00\x7c\x03\x7a\x01\x00\x00\x7c\x03\x6b\x28\x00\x00\x73\x01\x8c\x0f\x7c\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x01\x00\x8c\x27\x04\x00\x7c\x01\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x8c\x3a\x04\x00\x64\x02\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[61];
        }_object;
    }
stat_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 61,
        },
        .ob_item = {
            & stat_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 3],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 4],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 5],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 6],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 7],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 8],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 9],
            & stat_toplevel_consts_11.ob_base.ob_base,
            & stat_toplevel_consts_12.ob_base.ob_base,
            & const_int_16384.ob_base,
            & const_int_8192.ob_base,
            & const_int_24576.ob_base,
            & const_int_32768.ob_base,
            & const_int_4096.ob_base,
            & const_int_40960.ob_base,
            & const_int_49152.ob_base,
            & stat_toplevel_consts_20.ob_base.ob_base,
            & stat_toplevel_consts_21.ob_base.ob_base,
            & stat_toplevel_consts_22.ob_base.ob_base,
            & stat_toplevel_consts_23.ob_base.ob_base,
            & stat_toplevel_consts_24.ob_base.ob_base,
            & stat_toplevel_consts_25.ob_base.ob_base,
            & stat_toplevel_consts_26.ob_base.ob_base,
            & stat_toplevel_consts_27.ob_base.ob_base,
            & stat_toplevel_consts_28.ob_base.ob_base,
            & stat_toplevel_consts_29.ob_base.ob_base,
            & const_int_2048.ob_base,
            & const_int_1024.ob_base,
            & const_int_512.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 256],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 128],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 64],
            & const_int_448.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 56],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 32],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 16],
            & const_int_65536.ob_base,
            & const_int_131072.ob_base,
            & const_int_262144.ob_base,
            & const_int_1048576.ob_base,
            & const_int_2097152.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[108],
            (PyObject *)&_Py_SINGLETON(strings).ascii[115],
            (PyObject *)&_Py_SINGLETON(strings).ascii[45],
            (PyObject *)&_Py_SINGLETON(strings).ascii[98],
            (PyObject *)&_Py_SINGLETON(strings).ascii[100],
            (PyObject *)&_Py_SINGLETON(strings).ascii[99],
            (PyObject *)&_Py_SINGLETON(strings).ascii[112],
            (PyObject *)&_Py_SINGLETON(strings).ascii[114],
            (PyObject *)&_Py_SINGLETON(strings).ascii[119],
            (PyObject *)&_Py_SINGLETON(strings).ascii[83],
            (PyObject *)&_Py_SINGLETON(strings).ascii[120],
            (PyObject *)&_Py_SINGLETON(strings).ascii[116],
            (PyObject *)&_Py_SINGLETON(strings).ascii[84],
            & stat_toplevel_consts_58.ob_base.ob_base,
            & codecs_toplevel_consts_3._object.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_ST_MODE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_MODE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_ST_INO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_INO",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_ST_DEV = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_DEV",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_ST_NLINK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_NLINK",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_ST_UID = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_UID",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_ST_GID = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_GID",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_ST_SIZE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_SIZE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_ST_ATIME = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_ATIME",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_ST_MTIME = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_MTIME",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_ST_CTIME = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ST_CTIME",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_IFDOOR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFDOOR",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_IFPORT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFPORT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IFWHT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IFWHT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISUID = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISUID",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISGID = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISGID",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ENFMT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ENFMT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_ISVTX = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_ISVTX",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IREAD = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IREAD",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_S_IWRITE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IWRITE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IEXEC = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IEXEC",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IRWXU = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IRWXU",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IRUSR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IRUSR",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IWUSR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IWUSR",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IXUSR = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IXUSR",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IRWXG = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IRWXG",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IRGRP = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IRGRP",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IWGRP = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IWGRP",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IXGRP = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IXGRP",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IRWXO = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IRWXO",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IROTH = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IROTH",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IWOTH = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IWOTH",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_S_IXOTH = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "S_IXOTH",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_UF_NODUMP = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UF_NODUMP",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_UF_IMMUTABLE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UF_IMMUTABLE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_UF_APPEND = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UF_APPEND",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_UF_OPAQUE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UF_OPAQUE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_UF_NOUNLINK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UF_NOUNLINK",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_UF_COMPRESSED = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "UF_COMPRESSED",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_SF_ARCHIVED = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SF_ARCHIVED",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_SF_IMMUTABLE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SF_IMMUTABLE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_SF_APPEND = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SF_APPEND",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_SF_NOUNLINK = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SF_NOUNLINK",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_SF_SNAPSHOT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "SF_SNAPSHOT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str_FILE_ATTRIBUTE_ARCHIVE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_ARCHIVE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
const_str_FILE_ATTRIBUTE_COMPRESSED = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_COMPRESSED",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_FILE_ATTRIBUTE_DEVICE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_DEVICE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str_FILE_ATTRIBUTE_DIRECTORY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_DIRECTORY",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str_FILE_ATTRIBUTE_ENCRYPTED = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_ENCRYPTED",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
const_str_FILE_ATTRIBUTE_INTEGRITY_STREAM = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_INTEGRITY_STREAM",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_FILE_ATTRIBUTE_NORMAL = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_NORMAL",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[35];
    }
const_str_FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 34,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_NOT_CONTENT_INDEXED",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
const_str_FILE_ATTRIBUTE_NO_SCRUB_DATA = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_NO_SCRUB_DATA",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str_FILE_ATTRIBUTE_OFFLINE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_OFFLINE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
const_str_FILE_ATTRIBUTE_READONLY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_READONLY",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
const_str_FILE_ATTRIBUTE_REPARSE_POINT = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_REPARSE_POINT",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[27];
    }
const_str_FILE_ATTRIBUTE_SPARSE_FILE = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 26,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_SPARSE_FILE",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
const_str_FILE_ATTRIBUTE_SYSTEM = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_SYSTEM",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str_FILE_ATTRIBUTE_TEMPORARY = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_TEMPORARY",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
const_str_FILE_ATTRIBUTE_VIRTUAL = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "FILE_ATTRIBUTE_VIRTUAL",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str__stat = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_stat",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[85];
        }_object;
    }
stat_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 85,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_ST_MODE._ascii.ob_base,
            & const_str_ST_INO._ascii.ob_base,
            & const_str_ST_DEV._ascii.ob_base,
            & const_str_ST_NLINK._ascii.ob_base,
            & const_str_ST_UID._ascii.ob_base,
            & const_str_ST_GID._ascii.ob_base,
            & const_str_ST_SIZE._ascii.ob_base,
            & const_str_ST_ATIME._ascii.ob_base,
            & const_str_ST_MTIME._ascii.ob_base,
            & const_str_ST_CTIME._ascii.ob_base,
            & const_str_S_IMODE._ascii.ob_base,
            & const_str_S_IFMT._ascii.ob_base,
            & const_str_S_IFDIR._ascii.ob_base,
            & const_str_S_IFCHR._ascii.ob_base,
            & const_str_S_IFBLK._ascii.ob_base,
            & const_str_S_IFREG._ascii.ob_base,
            & const_str_S_IFIFO._ascii.ob_base,
            & const_str_S_IFLNK._ascii.ob_base,
            & const_str_S_IFSOCK._ascii.ob_base,
            & const_str_S_IFDOOR._ascii.ob_base,
            & const_str_S_IFPORT._ascii.ob_base,
            & const_str_S_IFWHT._ascii.ob_base,
            & const_str_S_ISDIR._ascii.ob_base,
            & const_str_S_ISCHR._ascii.ob_base,
            & const_str_S_ISBLK._ascii.ob_base,
            & const_str_S_ISREG._ascii.ob_base,
            & const_str_S_ISFIFO._ascii.ob_base,
            & const_str_S_ISLNK._ascii.ob_base,
            & const_str_S_ISSOCK._ascii.ob_base,
            & const_str_S_ISDOOR._ascii.ob_base,
            & const_str_S_ISPORT._ascii.ob_base,
            & const_str_S_ISWHT._ascii.ob_base,
            & const_str_S_ISUID._ascii.ob_base,
            & const_str_S_ISGID._ascii.ob_base,
            & const_str_S_ENFMT._ascii.ob_base,
            & const_str_S_ISVTX._ascii.ob_base,
            & const_str_S_IREAD._ascii.ob_base,
            & const_str_S_IWRITE._ascii.ob_base,
            & const_str_S_IEXEC._ascii.ob_base,
            & const_str_S_IRWXU._ascii.ob_base,
            & const_str_S_IRUSR._ascii.ob_base,
            & const_str_S_IWUSR._ascii.ob_base,
            & const_str_S_IXUSR._ascii.ob_base,
            & const_str_S_IRWXG._ascii.ob_base,
            & const_str_S_IRGRP._ascii.ob_base,
            & const_str_S_IWGRP._ascii.ob_base,
            & const_str_S_IXGRP._ascii.ob_base,
            & const_str_S_IRWXO._ascii.ob_base,
            & const_str_S_IROTH._ascii.ob_base,
            & const_str_S_IWOTH._ascii.ob_base,
            & const_str_S_IXOTH._ascii.ob_base,
            & const_str_UF_NODUMP._ascii.ob_base,
            & const_str_UF_IMMUTABLE._ascii.ob_base,
            & const_str_UF_APPEND._ascii.ob_base,
            & const_str_UF_OPAQUE._ascii.ob_base,
            & const_str_UF_NOUNLINK._ascii.ob_base,
            & const_str_UF_COMPRESSED._ascii.ob_base,
            & const_str_UF_HIDDEN._ascii.ob_base,
            & const_str_SF_ARCHIVED._ascii.ob_base,
            & const_str_SF_IMMUTABLE._ascii.ob_base,
            & const_str_SF_APPEND._ascii.ob_base,
            & const_str_SF_NOUNLINK._ascii.ob_base,
            & const_str_SF_SNAPSHOT._ascii.ob_base,
            & const_str__filemode_table._ascii.ob_base,
            & const_str_filemode._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_ARCHIVE._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_COMPRESSED._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_DEVICE._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_DIRECTORY._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_ENCRYPTED._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_HIDDEN._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_INTEGRITY_STREAM._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_NORMAL._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_NOT_CONTENT_INDEXED._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_NO_SCRUB_DATA._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_OFFLINE._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_READONLY._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_REPARSE_POINT._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_SPARSE_FILE._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_SYSTEM._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_TEMPORARY._ascii.ob_base,
            & const_str_FILE_ATTRIBUTE_VIRTUAL._ascii.ob_base,
            & const_str__stat._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[710];
    }
stat_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 709,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x03\x01\x04\xf0\x0e\x00\x0c\x0d\x80\x07\xd8\x0b\x0c\x80\x06\xd8\x0b\x0c\x80\x06\xd8\x0b\x0c\x80\x08\xd8\x0b\x0c\x80\x06\xd8\x0b\x0c\x80\x06\xd8\x0b\x0c\x80\x07\xd8\x0b\x0c\x80\x08\xd8\x0b\x0c\x80\x08\xd8\x0b\x0c\x80\x08\xf2\x08\x04\x01\x19\xf2\x0c\x04\x01\x1b\xf0\x12\x00\x0c\x14\x80\x07\xd8\x0b\x13\x80\x07\xd8\x0b\x13\x80\x07\xd8\x0b\x13\x80\x07\xd8\x0b\x13\x80\x07\xd8\x0b\x13\x80\x07\xd8\x0b\x13\x80\x08\xe0\x0b\x0c\x80\x08\xd8\x0b\x0c\x80\x08\xd8\x0a\x0b\x80\x07\xf2\x08\x02\x01\x23\xf2\x08\x02\x01\x23\xf2\x08\x02\x01\x23\xf2\x08\x02\x01\x23\xf2\x08\x02\x01\x23\xf2\x08\x02\x01\x23\xf2\x08\x02\x01\x24\xf2\x08\x02\x01\x11\xf2\x08\x02\x01\x11\xf2\x08\x02\x01\x11\xf0\x0c\x00\x0b\x11\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x11\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0b\x11\x80\x08\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xd8\x0a\x10\x80\x07\xf0\x08\x00\x10\x1a\x80\x09\xd8\x0f\x19\x80\x0c\xd8\x0f\x19\x80\x09\xd8\x0f\x19\x80\x09\xd8\x0f\x19\x80\x0b\xd8\x10\x1a\x80\x0d\xd8\x0f\x19\x80\x09\xd8\x0f\x19\x80\x0b\xd8\x0f\x19\x80\x0c\xd8\x0f\x19\x80\x09\xd8\x0f\x19\x80\x0b\xd8\x0f\x19\x80\x0b\xf0\x08\x00\x07\x0e\x90\x73\xd0\x05\x1b\xd8\x06\x0e\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xf0\x0d\x06\x05\x1d\xf0\x10\x00\x07\x0e\x90\x73\xd0\x05\x1b\xd0\x04\x1d\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd0\x04\x1d\xd8\x06\x0d\x88\x67\x81\x6f\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xf0\x05\x02\x05\x1d\xf0\x08\x00\x07\x0e\x90\x73\xd0\x05\x1b\xd0\x04\x1d\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd0\x04\x1d\xd8\x06\x0d\x88\x67\x81\x6f\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xf0\x05\x02\x05\x1d\xf0\x08\x00\x07\x0e\x90\x73\xd0\x05\x1b\xd0\x04\x1d\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd0\x04\x1d\xd8\x06\x0d\x88\x67\x81\x6f\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xd8\x06\x0d\x90\x73\xd0\x05\x1b\xf0\x05\x02\x05\x1d\xf0\x2f\x1a\x13\x02\x80\x0f\xf2\x38\x0a\x01\x19\xf0\x20\x00\x1a\x1c\xd0\x00\x16\xd8\x1c\x20\xd0\x00\x19\xd8\x18\x1a\xd0\x00\x15\xd8\x1b\x1d\xd0\x00\x18\xd8\x1b\x20\xd0\x00\x18\xd8\x18\x19\xd0\x00\x15\xd8\x22\x27\xd0\x00\x1f\xd8\x18\x1b\xd0\x00\x15\xd8\x25\x29\xd0\x00\x22\xd8\x1f\x25\xd0\x00\x1c\xd8\x19\x1d\xd0\x00\x16\xd8\x1a\x1b\xd0\x00\x17\xd8\x1f\x23\xd0\x00\x1c\xd8\x1d\x20\xd0\x00\x1a\xd8\x18\x19\xd0\x00\x15\xd8\x1b\x1e\xd0\x00\x18\xd8\x19\x1e\xd0\x00\x16\xf0\x08\x03\x01\x09\xdd\x04\x17\xf8\xd8\x07\x12\xf2\x00\x01\x01\x09\xd9\x04\x08\xf0\x03\x01\x01\x09\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
stat_toplevel_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xc4\x0a\x05\x44\x10\x00\xc4\x10\x05\x44\x18\x03\xc4\x17\x01\x44\x18\x03",
};
static
    struct _PyCode_DEF(566)
stat_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 283,
    },
    .co_consts = & stat_toplevel_consts._object.ob_base.ob_base,
    .co_names = & stat_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = & stat_toplevel_exceptiontable.ob_base.ob_base,
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 13,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 722,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & stat_toplevel_consts_11_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & stat_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x5a\x01\x64\x02\x5a\x02\x64\x03\x5a\x03\x64\x04\x5a\x04\x64\x05\x5a\x05\x64\x06\x5a\x06\x64\x07\x5a\x07\x64\x08\x5a\x08\x64\x09\x5a\x09\x64\x0a\x5a\x0a\x64\x0b\x84\x00\x5a\x0b\x64\x0c\x84\x00\x5a\x0c\x64\x0d\x5a\x0d\x64\x0e\x5a\x0e\x64\x0f\x5a\x0f\x64\x10\x5a\x10\x64\x11\x5a\x11\x64\x12\x5a\x12\x64\x13\x5a\x13\x64\x01\x5a\x14\x64\x01\x5a\x15\x64\x01\x5a\x16\x64\x14\x84\x00\x5a\x17\x64\x15\x84\x00\x5a\x18\x64\x16\x84\x00\x5a\x19\x64\x17\x84\x00\x5a\x1a\x64\x18\x84\x00\x5a\x1b\x64\x19\x84\x00\x5a\x1c\x64\x1a\x84\x00\x5a\x1d\x64\x1b\x84\x00\x5a\x1e\x64\x1c\x84\x00\x5a\x1f\x64\x1d\x84\x00\x5a\x20\x64\x1e\x5a\x21\x64\x1f\x5a\x22\x65\x22\x5a\x23\x64\x20\x5a\x24\x64\x21\x5a\x25\x64\x22\x5a\x26\x64\x23\x5a\x27\x64\x24\x5a\x28\x64\x21\x5a\x29\x64\x22\x5a\x2a\x64\x23\x5a\x2b\x64\x25\x5a\x2c\x64\x26\x5a\x2d\x64\x27\x5a\x2e\x64\x09\x5a\x2f\x64\x08\x5a\x30\x64\x05\x5a\x31\x64\x03\x5a\x32\x64\x02\x5a\x33\x64\x02\x5a\x34\x64\x03\x5a\x35\x64\x05\x5a\x36\x64\x09\x5a\x37\x64\x27\x5a\x38\x64\x26\x5a\x39\x64\x10\x5a\x3a\x64\x28\x5a\x3b\x64\x29\x5a\x3c\x64\x2a\x5a\x3d\x64\x2b\x5a\x3e\x64\x2c\x5a\x3f\x65\x12\x64\x2d\x66\x02\x65\x13\x64\x2e\x66\x02\x65\x10\x64\x2f\x66\x02\x65\x0f\x64\x30\x66\x02\x65\x0d\x64\x31\x66\x02\x65\x0e\x64\x32\x66\x02\x65\x11\x64\x33\x66\x02\x66\x07\x65\x29\x64\x34\x66\x02\x66\x01\x65\x2a\x64\x35\x66\x02\x66\x01\x65\x2b\x65\x21\x7a\x07\x00\x00\x64\x2e\x66\x02\x65\x21\x64\x36\x66\x02\x65\x2b\x64\x37\x66\x02\x66\x03\x65\x2d\x64\x34\x66\x02\x66\x01\x65\x2e\x64\x35\x66\x02\x66\x01\x65\x2f\x65\x22\x7a\x07\x00\x00\x64\x2e\x66\x02\x65\x22\x64\x36\x66\x02\x65\x2f\x64\x37\x66\x02\x66\x03\x65\x31\x64\x34\x66\x02\x66\x01\x65\x32\x64\x35\x66\x02\x66\x01\x65\x33\x65\x24\x7a\x07\x00\x00\x64\x38\x66\x02\x65\x24\x64\x39\x66\x02\x65\x33\x64\x37\x66\x02\x66\x03\x66\x0a\x5a\x40\x64\x3a\x84\x00\x5a\x41\x64\x26\x5a\x42\x64\x1e\x5a\x43\x64\x23\x5a\x44\x64\x27\x5a\x45\x64\x0d\x5a\x46\x64\x03\x5a\x47\x64\x10\x5a\x48\x64\x22\x5a\x49\x64\x0e\x5a\x4a\x64\x29\x5a\x4b\x64\x11\x5a\x4c\x64\x02\x5a\x4d\x64\x1f\x5a\x4e\x64\x20\x5a\x4f\x64\x05\x5a\x50\x64\x21\x5a\x51\x64\x28\x5a\x52\x09\x00\x64\x01\x64\x3b\x6c\x53\xad\x02\x01\x00\x79\x3c\x23\x00\x65\x54\x24\x00\x72\x03\x01\x00\x59\x00\x79\x3c\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_stat_toplevel(void)
{
    return Py_NewRef((PyObject *) &stat_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[46];
    }
importlib_util_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 45,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Utility code for constructing importers, etc.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str_Loader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Loader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_Loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_module_from_spec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__resolve_name._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_spec_from_loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_6 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__find_spec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_7 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_MAGIC_NUMBER._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_8 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__RAW_MAGIC_NUMBER._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_9 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_cache_from_source._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_10 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_decode_source._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_11 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_source_from_cache._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[67];
    }
importlib_util_toplevel_consts_15_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 66,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Return the hash of *source_bytes* as used in hash-based pyc files.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_15_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib_util_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str_source_hash._ascii.ob_base,
            & const_str__RAW_MAGIC_NUMBER._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib_util_toplevel_consts_15_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen importlib.util>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[23];
    }
importlib_util_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 22,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x0f\xd7\x0b\x1b\xd1\x0b\x1b\xd4\x1c\x2d\xa8\x7c\xd3\x0b\x3c\xd0\x04\x3c",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_15_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_source_bytes._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(54)
importlib_util_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 27,
    },
    .co_consts = & importlib_util_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 19,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 723,
    .co_localsplusnames = & importlib_util_toplevel_consts_15_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str_source_hash._ascii.ob_base,
    .co_qualname = & const_str_source_hash._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[26];
    }
importlib_util_toplevel_consts_16_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 25,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "no package specified for ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
importlib_util_toplevel_consts_16_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " (required for relative module names)",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib_util_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & importlib__bootstrap_toplevel_consts_50_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & importlib_util_toplevel_consts_16_consts_2._ascii.ob_base,
            & importlib_util_toplevel_consts_16_consts_3._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib_util_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_startswith._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_repr._ascii.ob_base,
            & const_str__resolve_name._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_resolve_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "resolve_name",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[120];
    }
importlib_util_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 119,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x0b\x0f\x8f\x3f\x89\x3f\x98\x33\xd4\x0b\x1f\xd8\x0f\x13\x88\x0b\xd9\x0d\x14\xdc\x0e\x19\xd0\x1c\x35\xb4\x64\xb8\x34\xb3\x6a\xb0\x5c\xf0\x00\x01\x42\x01\x41\x01\xf0\x00\x01\x1b\x41\x01\xf3\x00\x01\x0f\x42\x01\xf0\x00\x01\x09\x42\x01\xe0\x0c\x0d\x80\x45\xdb\x15\x19\x88\x09\xd8\x0b\x14\x98\x03\xd2\x0b\x1b\xd9\x0c\x11\xd8\x08\x0d\x90\x11\x89\x0a\x89\x05\xf0\x07\x00\x16\x1a\xf4\x08\x00\x0c\x19\x98\x14\x98\x65\x98\x66\x98\x1c\xa0\x77\xb0\x05\xd3\x0b\x36\xd0\x04\x36",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_character = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "character",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib_util_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_package._ascii.ob_base,
            &_Py_ID(level),
            & const_str_character._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(166)
importlib_util_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 83,
    },
    .co_consts = & importlib_util_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 10 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 24,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 724,
    .co_localsplusnames = & importlib_util_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str_resolve_name._ascii.ob_base,
    .co_qualname = & const_str_resolve_name._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x73\x02\x7c\x00\x53\x00\x7c\x01\x73\x18\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x9b\x00\x64\x03\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x64\x04\x7d\x02\x7c\x00\x44\x00\x5d\x0e\x00\x00\x7d\x03\x7c\x03\x64\x01\x6b\x37\x00\x00\x72\x02\x01\x00\x6e\x07\x7c\x02\x64\x05\x7a\x0d\x00\x00\x7d\x02\x8c\x10\x04\x00\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\x64\x06\x1a\x00\x7c\x01\x7c\x02\xab\x03\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[648];
    }
importlib_util_toplevel_consts_17_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 647,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x73\x70\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x46\x69\x72\x73\x74\x2c\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x20\x69\x73\x20\x63\x68\x65\x63\x6b\x65\x64\x20\x74\x6f\x20\x73\x65\x65\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x77\x61\x73\x20\x61\x6c\x72\x65\x61\x64\x79\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x20\x49\x66\x0a\x20\x20\x20\x20\x73\x6f\x2c\x20\x74\x68\x65\x6e\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x5b\x6e\x61\x6d\x65\x5d\x2e\x5f\x5f\x73\x70\x65\x63\x5f\x5f\x20\x69\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x2e\x20\x49\x66\x20\x74\x68\x61\x74\x20\x68\x61\x70\x70\x65\x6e\x73\x20\x74\x6f\x20\x62\x65\x0a\x20\x20\x20\x20\x73\x65\x74\x20\x74\x6f\x20\x4e\x6f\x6e\x65\x2c\x20\x74\x68\x65\x6e\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x20\x49\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x69\x6e\x0a\x20\x20\x20\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x2c\x20\x74\x68\x65\x6e\x20\x73\x79\x73\x2e\x6d\x65\x74\x61\x5f\x70\x61\x74\x68\x20\x69\x73\x20\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x61\x20\x73\x75\x69\x74\x61\x62\x6c\x65\x20\x73\x70\x65\x63\x20\x77\x69\x74\x68\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x76\x61\x6c\x75\x65\x20\x6f\x66\x20\x27\x70\x61\x74\x68\x27\x20\x67\x69\x76\x65\x6e\x20\x74\x6f\x20\x74\x68\x65\x20\x66\x69\x6e\x64\x65\x72\x73\x2e\x20\x4e\x6f\x6e\x65\x20\x69\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x69\x66\x20\x6e\x6f\x20\x73\x70\x65\x63\x20\x63\x6f\x75\x6c\x64\x0a\x20\x20\x20\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x0a\x20\x20\x20\x20\x44\x6f\x74\x74\x65\x64\x20\x6e\x61\x6d\x65\x73\x20\x64\x6f\x20\x6e\x6f\x74\x20\x68\x61\x76\x65\x20\x74\x68\x65\x69\x72\x20\x70\x61\x72\x65\x6e\x74\x20\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x69\x6d\x70\x6c\x69\x63\x69\x74\x6c\x79\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x20\x59\x6f\x75\x20\x77\x69\x6c\x6c\x0a\x20\x20\x20\x20\x6d\x6f\x73\x74\x20\x6c\x69\x6b\x65\x6c\x79\x20\x6e\x65\x65\x64\x20\x74\x6f\x20\x65\x78\x70\x6c\x69\x63\x69\x74\x6c\x79\x20\x69\x6d\x70\x6f\x72\x74\x20\x61\x6c\x6c\x20\x70\x61\x72\x65\x6e\x74\x20\x70\x61\x63\x6b\x61\x67\x65\x73\x20\x69\x6e\x20\x74\x68\x65\x20\x70\x72\x6f\x70\x65\x72\x0a\x20\x20\x20\x20\x6f\x72\x64\x65\x72\x20\x66\x6f\x72\x20\x61\x20\x73\x75\x62\x6d\x6f\x64\x75\x6c\x65\x20\x74\x6f\x20\x67\x65\x74\x20\x74\x68\x65\x20\x63\x6f\x72\x72\x65\x63\x74\x20\x73\x70\x65\x63\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[18];
    }
importlib_util_toplevel_consts_17_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 17,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".__spec__ is None",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
importlib_util_toplevel_consts_17_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".__spec__ is not set",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib_util_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_17_consts_0._ascii.ob_base,
            Py_None,
            & importlib_util_toplevel_consts_17_consts_2._ascii.ob_base,
            & importlib_util_toplevel_consts_17_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib_util_toplevel_consts_17_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str__find_spec._ascii.ob_base,
            &_Py_ID(__spec__),
            & const_str_ValueError._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str__find_spec_from_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_find_spec_from_path",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[137];
    }
importlib_util_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 136,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x1e\x00\x08\x0c\x94\x33\x97\x3b\x91\x3b\xd1\x07\x1e\xdc\x0f\x19\x98\x24\xa0\x04\xd3\x0f\x25\xd0\x08\x25\xe4\x11\x14\x97\x1b\x91\x1b\x98\x54\xd1\x11\x22\x88\x06\xd8\x0b\x11\x88\x3e\xd8\x13\x17\xf0\x02\x07\x09\x18\xd8\x13\x19\x97\x3f\x91\x3f\x88\x44\xf0\x08\x00\x10\x14\x88\x7c\xdc\x16\x20\xa0\x44\xa0\x36\xd0\x29\x3a\xd0\x21\x3b\xd3\x16\x3c\xd0\x10\x3c\xd8\x13\x17\x88\x4b\xf8\xf4\x0b\x00\x10\x1e\xf2\x00\x01\x09\x46\x01\xdc\x12\x1c\xa0\x04\x98\x76\xd0\x25\x39\xd0\x1d\x3a\xd3\x12\x3b\xc0\x14\xd0\x0c\x45\xf0\x03\x01\x09\x46\x01\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[12];
    }
importlib_util_toplevel_consts_17_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 11,
    },
    .ob_shash = -1,
    .ob_sval = "\xb6\x0c\x41\x14\x00\xc1\x14\x19\x41\x2d\x03",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib_util_toplevel_consts_17_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(name),
            &_Py_ID(path),
            &_Py_ID(module),
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(224)
importlib_util_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 112,
    },
    .co_consts = & importlib_util_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_17_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib_util_toplevel_consts_17_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 9 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 39,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 725,
    .co_localsplusnames = & importlib_util_toplevel_consts_17_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str__find_spec_from_path._ascii.ob_base,
    .co_qualname = & const_str__find_spec_from_path._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x0c\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x19\x00\x00\x00\x7d\x02\x7c\x02\x80\x01\x79\x01\x09\x00\x7c\x02\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x80\x0e\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x9b\x00\x64\x02\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x03\x53\x00\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x10\x01\x00\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x9b\x00\x64\x03\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x64\x01\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[688];
    }
importlib_util_toplevel_consts_18_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 687,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x65\x74\x75\x72\x6e\x20\x74\x68\x65\x20\x73\x70\x65\x63\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x2e\x0a\x0a\x20\x20\x20\x20\x46\x69\x72\x73\x74\x2c\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x20\x69\x73\x20\x63\x68\x65\x63\x6b\x65\x64\x20\x74\x6f\x20\x73\x65\x65\x20\x69\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x77\x61\x73\x20\x61\x6c\x72\x65\x61\x64\x79\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x20\x49\x66\x0a\x20\x20\x20\x20\x73\x6f\x2c\x20\x74\x68\x65\x6e\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x5b\x6e\x61\x6d\x65\x5d\x2e\x5f\x5f\x73\x70\x65\x63\x5f\x5f\x20\x69\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x2e\x20\x49\x66\x20\x74\x68\x61\x74\x20\x68\x61\x70\x70\x65\x6e\x73\x20\x74\x6f\x20\x62\x65\x0a\x20\x20\x20\x20\x73\x65\x74\x20\x74\x6f\x20\x4e\x6f\x6e\x65\x2c\x20\x74\x68\x65\x6e\x20\x56\x61\x6c\x75\x65\x45\x72\x72\x6f\x72\x20\x69\x73\x20\x72\x61\x69\x73\x65\x64\x2e\x20\x49\x66\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x6e\x6f\x74\x20\x69\x6e\x0a\x20\x20\x20\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x2c\x20\x74\x68\x65\x6e\x20\x73\x79\x73\x2e\x6d\x65\x74\x61\x5f\x70\x61\x74\x68\x20\x69\x73\x20\x73\x65\x61\x72\x63\x68\x65\x64\x20\x66\x6f\x72\x20\x61\x20\x73\x75\x69\x74\x61\x62\x6c\x65\x20\x73\x70\x65\x63\x20\x77\x69\x74\x68\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x76\x61\x6c\x75\x65\x20\x6f\x66\x20\x27\x70\x61\x74\x68\x27\x20\x67\x69\x76\x65\x6e\x20\x74\x6f\x20\x74\x68\x65\x20\x66\x69\x6e\x64\x65\x72\x73\x2e\x20\x4e\x6f\x6e\x65\x20\x69\x73\x20\x72\x65\x74\x75\x72\x6e\x65\x64\x20\x69\x66\x20\x6e\x6f\x20\x73\x70\x65\x63\x20\x63\x6f\x75\x6c\x64\x0a\x20\x20\x20\x20\x62\x65\x20\x66\x6f\x75\x6e\x64\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x74\x68\x65\x20\x6e\x61\x6d\x65\x20\x69\x73\x20\x66\x6f\x72\x20\x73\x75\x62\x6d\x6f\x64\x75\x6c\x65\x20\x28\x63\x6f\x6e\x74\x61\x69\x6e\x73\x20\x61\x20\x64\x6f\x74\x29\x2c\x20\x74\x68\x65\x20\x70\x61\x72\x65\x6e\x74\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x0a\x20\x20\x20\x20\x61\x75\x74\x6f\x6d\x61\x74\x69\x63\x61\x6c\x6c\x79\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x54\x68\x65\x20\x6e\x61\x6d\x65\x20\x61\x6e\x64\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x20\x77\x6f\x72\x6b\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x61\x73\x20\x69\x6d\x70\x6f\x72\x74\x6c\x69\x62\x2e\x69\x6d\x70\x6f\x72\x74\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x29\x2e\x0a\x20\x20\x20\x20\x49\x6e\x20\x6f\x74\x68\x65\x72\x20\x77\x6f\x72\x64\x73\x2c\x20\x72\x65\x6c\x61\x74\x69\x76\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x73\x20\x28\x77\x69\x74\x68\x20\x6c\x65\x61\x64\x69\x6e\x67\x20\x64\x6f\x74\x73\x29\x20\x77\x6f\x72\x6b\x2e\x0a\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_18_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            &_Py_ID(fromlist),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib_util_toplevel_consts_18_consts_5 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__path__ attribute not found on ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib_util_toplevel_consts_18_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " while trying to find ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib_util_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_18_consts_0._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            &_Py_ID(__path__),
            & importlib_util_toplevel_consts_18_consts_4._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_18_consts_5._ascii.ob_base,
            & importlib_util_toplevel_consts_18_consts_6._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_4_localsplusnames._object.ob_base.ob_base,
            Py_None,
            & importlib_util_toplevel_consts_17_consts_2._ascii.ob_base,
            & importlib_util_toplevel_consts_17_consts_3._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[12];
        }_object;
    }
importlib_util_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 12,
        },
        .ob_item = {
            & const_str_startswith._ascii.ob_base,
            & const_str_resolve_name._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str_rpartition._ascii.ob_base,
            &_Py_ID(__import__),
            &_Py_ID(__path__),
            & const_str_AttributeError._ascii.ob_base,
            & const_str_ModuleNotFoundError._ascii.ob_base,
            & const_str__find_spec._ascii.ob_base,
            &_Py_ID(__spec__),
            & const_str_ValueError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[285];
    }
importlib_util_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 284,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x22\x00\x2f\x33\xaf\x6f\xa9\x6f\xb8\x63\xd4\x2e\x42\x8c\x7c\x98\x44\xa0\x27\xd4\x0f\x2a\xc8\x04\x80\x48\xd8\x07\x0f\x94\x73\x97\x7b\x91\x7b\xd1\x07\x22\xd8\x16\x1e\xd7\x16\x29\xd1\x16\x29\xa8\x23\xd3\x16\x2e\xa8\x71\xd1\x16\x31\x88\x0b\xd9\x0b\x16\xdc\x15\x1f\xa0\x0b\xb0\x7a\xb0\x6c\xd4\x15\x43\x88\x46\xf0\x02\x05\x0d\x50\x01\xd8\x1e\x24\x9f\x6f\x99\x6f\x91\x0b\xf0\x0c\x00\x1b\x1f\x88\x4b\xdc\x0f\x19\x98\x28\xa0\x4b\xd3\x0f\x30\xd0\x08\x30\xe4\x11\x14\x97\x1b\x91\x1b\x98\x58\xd1\x11\x26\x88\x06\xd8\x0b\x11\x88\x3e\xd8\x13\x17\xf0\x02\x07\x09\x18\xd8\x13\x19\x97\x3f\x91\x3f\x88\x44\xf0\x08\x00\x10\x14\x88\x7c\xdc\x16\x20\xa0\x44\xa0\x36\xd0\x29\x3a\xd0\x21\x3b\xd3\x16\x3c\xd0\x10\x3c\xd8\x13\x17\x88\x4b\xf8\xf4\x25\x00\x14\x22\xf2\x00\x03\x0d\x50\x01\xdc\x16\x29\xd8\x16\x36\xb0\x7b\xb0\x6f\xf0\x00\x01\x46\x01\x2c\xd8\x2c\x34\xa8\x3c\xf0\x03\x01\x15\x39\xd8\x3f\x47\xf4\x05\x02\x17\x49\x01\xe0\x4e\x4f\xf0\x05\x02\x11\x50\x01\xfb\xf0\x03\x03\x0d\x50\x01\xfb\xf4\x1a\x00\x10\x1e\xf2\x00\x01\x09\x46\x01\xdc\x12\x1c\xa0\x04\x98\x76\xd0\x25\x39\xd0\x1d\x3a\xd3\x12\x3b\xc0\x14\xd0\x0c\x45\xf0\x03\x01\x09\x46\x01\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[37];
    }
importlib_util_toplevel_consts_18_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 36,
    },
    .ob_shash = -1,
    .ob_sval = "\xc1\x17\x0c\x42\x27\x00\xc2\x09\x0c\x43\x0c\x00\xc2\x27\x09\x43\x09\x03\xc2\x30\x14\x43\x04\x03\xc3\x04\x05\x43\x09\x03\xc3\x0c\x19\x43\x25\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_parent_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "parent_name",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib_util_toplevel_consts_18_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(name),
            & const_str_package._ascii.ob_base,
            & const_str_fullname._ascii.ob_base,
            & const_str_parent_name._ascii.ob_base,
            &_Py_ID(parent),
            & const_str_parent_path._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[101],
            &_Py_ID(module),
            & const_str_spec._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(464)
importlib_util_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 232,
    },
    .co_consts = & importlib_util_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib_util_toplevel_consts_18_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 16 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 70,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 726,
    .co_localsplusnames = & importlib_util_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str_find_spec._ascii.ob_base,
    .co_qualname = & const_str_find_spec._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0c\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x6e\x01\x7c\x00\x7d\x02\x7c\x02\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x40\x7c\x02\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\x7d\x03\x7c\x03\x72\x1c\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x03\x67\x01\xac\x04\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x09\x00\x7c\x04\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x6e\x02\x64\x08\x7d\x05\x74\x13\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x7c\x05\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x19\x00\x00\x00\x7d\x07\x7c\x07\x80\x01\x79\x08\x09\x00\x7c\x07\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x80\x0e\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x9b\x00\x64\x09\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x08\x53\x00\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x19\x7d\x06\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x7c\x03\x9b\x02\x64\x06\x7c\x02\x9b\x02\x9d\x04\x7c\x02\xac\x07\xab\x02\x00\x00\x00\x00\x00\x00\x7c\x06\x82\x02\x64\x08\x7d\x06\x7e\x06\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x10\x01\x00\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x9b\x00\x64\x0a\x9d\x02\xab\x01\x00\x00\x00\x00\x00\x00\x64\x08\x82\x02\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[44];
    }
const_str__incompatible_extension_module_restrictions = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 43,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_incompatible_extension_module_restrictions",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[1384];
    }
importlib_util_toplevel_consts_19_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 1383,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x41\x20\x63\x6f\x6e\x74\x65\x78\x74\x20\x6d\x61\x6e\x61\x67\x65\x72\x20\x74\x68\x61\x74\x20\x63\x61\x6e\x20\x74\x65\x6d\x70\x6f\x72\x61\x72\x69\x6c\x79\x20\x73\x6b\x69\x70\x20\x74\x68\x65\x20\x63\x6f\x6d\x70\x61\x74\x69\x62\x69\x6c\x69\x74\x79\x20\x63\x68\x65\x63\x6b\x2e\x0a\x0a\x20\x20\x20\x20\x4e\x4f\x54\x45\x3a\x20\x54\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x69\x73\x20\x6d\x65\x61\x6e\x74\x20\x74\x6f\x20\x61\x63\x63\x6f\x6d\x6d\x6f\x64\x61\x74\x65\x20\x61\x6e\x20\x75\x6e\x75\x73\x75\x61\x6c\x20\x63\x61\x73\x65\x3b\x20\x6f\x6e\x65\x0a\x20\x20\x20\x20\x77\x68\x69\x63\x68\x20\x69\x73\x20\x6c\x69\x6b\x65\x6c\x79\x20\x74\x6f\x20\x65\x76\x65\x6e\x74\x75\x61\x6c\x6c\x79\x20\x67\x6f\x20\x61\x77\x61\x79\x2e\x20\x20\x54\x68\x65\x72\x65\x27\x73\x20\x69\x73\x20\x61\x20\x70\x72\x65\x74\x74\x79\x20\x67\x6f\x6f\x64\x0a\x20\x20\x20\x20\x63\x68\x61\x6e\x63\x65\x20\x74\x68\x69\x73\x20\x69\x73\x20\x6e\x6f\x74\x20\x77\x68\x61\x74\x20\x79\x6f\x75\x20\x77\x65\x72\x65\x20\x6c\x6f\x6f\x6b\x69\x6e\x67\x20\x66\x6f\x72\x2e\x0a\x0a\x20\x20\x20\x20\x57\x41\x52\x4e\x49\x4e\x47\x3a\x20\x55\x73\x69\x6e\x67\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x74\x6f\x20\x64\x69\x73\x61\x62\x6c\x65\x20\x74\x68\x65\x20\x63\x68\x65\x63\x6b\x20\x63\x61\x6e\x20\x6c\x65\x61\x64\x20\x74\x6f\x0a\x20\x20\x20\x20\x75\x6e\x65\x78\x70\x65\x63\x74\x65\x64\x20\x62\x65\x68\x61\x76\x69\x6f\x72\x20\x61\x6e\x64\x20\x65\x76\x65\x6e\x20\x63\x72\x61\x73\x68\x65\x73\x2e\x20\x20\x49\x74\x20\x73\x68\x6f\x75\x6c\x64\x20\x6f\x6e\x6c\x79\x20\x62\x65\x20\x75\x73\x65\x64\x20\x64\x75\x72\x69\x6e\x67\x0a\x20\x20\x20\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x20\x64\x65\x76\x65\x6c\x6f\x70\x6d\x65\x6e\x74\x2e\x0a\x0a\x20\x20\x20\x20\x49\x66\x20\x22\x64\x69\x73\x61\x62\x6c\x65\x5f\x63\x68\x65\x63\x6b\x22\x20\x69\x73\x20\x54\x72\x75\x65\x20\x74\x68\x65\x6e\x20\x74\x68\x65\x20\x63\x6f\x6d\x70\x61\x74\x69\x62\x69\x6c\x69\x74\x79\x20\x63\x68\x65\x63\x6b\x20\x77\x69\x6c\x6c\x20\x6e\x6f\x74\x0a\x20\x20\x20\x20\x68\x61\x70\x70\x65\x6e\x20\x77\x68\x69\x6c\x65\x20\x74\x68\x65\x20\x63\x6f\x6e\x74\x65\x78\x74\x20\x6d\x61\x6e\x61\x67\x65\x72\x20\x69\x73\x20\x61\x63\x74\x69\x76\x65\x2e\x20\x20\x4f\x74\x68\x65\x72\x77\x69\x73\x65\x20\x74\x68\x65\x20\x63\x68\x65\x63\x6b\x0a\x20\x20\x20\x20\x2a\x77\x69\x6c\x6c\x2a\x20\x68\x61\x70\x70\x65\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x4e\x6f\x72\x6d\x61\x6c\x6c\x79\x2c\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x73\x20\x74\x68\x61\x74\x20\x64\x6f\x20\x6e\x6f\x74\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x6d\x75\x6c\x74\x69\x70\x6c\x65\x20\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x73\x0a\x20\x20\x20\x20\x6d\x61\x79\x20\x6e\x6f\x74\x20\x62\x65\x20\x69\x6d\x70\x6f\x72\x74\x65\x64\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x2e\x20\x20\x54\x68\x61\x74\x20\x69\x6d\x70\x6c\x69\x65\x73\x20\x6d\x6f\x64\x75\x6c\x65\x73\x0a\x20\x20\x20\x20\x74\x68\x61\x74\x20\x64\x6f\x20\x6e\x6f\x74\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x20\x6d\x75\x6c\x74\x69\x2d\x70\x68\x61\x73\x65\x20\x69\x6e\x69\x74\x20\x6f\x72\x20\x74\x68\x61\x74\x20\x65\x78\x70\x6c\x69\x63\x69\x74\x6c\x79\x20\x6f\x66\x20\x6f\x75\x74\x2e\x0a\x0a\x20\x20\x20\x20\x4c\x69\x6b\x65\x77\x69\x73\x65\x20\x66\x6f\x72\x20\x6d\x6f\x64\x75\x6c\x65\x73\x20\x69\x6d\x70\x6f\x72\x74\x20\x69\x6e\x20\x61\x20\x73\x75\x62\x69\x6e\x74\x65\x72\x70\x65\x74\x65\x72\x20\x77\x69\x74\x68\x20\x69\x74\x73\x20\x6f\x77\x6e\x20\x47\x49\x4c\x0a\x20\x20\x20\x20\x77\x68\x65\x6e\x20\x74\x68\x65\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x61\x20\x70\x65\x72\x2d\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x20\x47\x49\x4c\x2e\x20\x20\x54\x68\x69\x73\x0a\x20\x20\x20\x20\x69\x6d\x70\x6c\x69\x65\x73\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x64\x6f\x65\x73\x20\x6e\x6f\x74\x20\x68\x61\x76\x65\x20\x61\x20\x50\x79\x5f\x6d\x6f\x64\x5f\x6d\x75\x6c\x74\x69\x70\x6c\x65\x5f\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x73\x20\x73\x6c\x6f\x74\x0a\x20\x20\x20\x20\x73\x65\x74\x20\x74\x6f\x20\x50\x79\x5f\x4d\x4f\x44\x5f\x50\x45\x52\x5f\x49\x4e\x54\x45\x52\x50\x52\x45\x54\x45\x52\x5f\x47\x49\x4c\x5f\x53\x55\x50\x50\x4f\x52\x54\x45\x44\x2e\x0a\x0a\x20\x20\x20\x20\x49\x6e\x20\x62\x6f\x74\x68\x20\x63\x61\x73\x65\x73\x2c\x20\x74\x68\x69\x73\x20\x63\x6f\x6e\x74\x65\x78\x74\x20\x6d\x61\x6e\x61\x67\x65\x72\x20\x6d\x61\x79\x20\x62\x65\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x74\x65\x6d\x70\x6f\x72\x61\x72\x69\x6c\x79\x0a\x20\x20\x20\x20\x64\x69\x73\x61\x62\x6c\x65\x20\x74\x68\x65\x20\x63\x68\x65\x63\x6b\x20\x66\x6f\x72\x20\x63\x6f\x6d\x70\x61\x74\x69\x62\x6c\x65\x20\x65\x78\x74\x65\x6e\x73\x69\x6f\x6e\x20\x6d\x6f\x64\x75\x6c\x65\x73\x2e\x0a\x0a\x20\x20\x20\x20\x59\x6f\x75\x20\x63\x61\x6e\x20\x67\x65\x74\x20\x74\x68\x65\x20\x73\x61\x6d\x65\x20\x65\x66\x66\x65\x63\x74\x20\x61\x73\x20\x74\x68\x69\x73\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x62\x79\x20\x69\x6d\x70\x6c\x65\x6d\x65\x6e\x74\x69\x6e\x67\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x62\x61\x73\x69\x63\x20\x69\x6e\x74\x65\x72\x66\x61\x63\x65\x20\x6f\x66\x20\x6d\x75\x6c\x74\x69\x2d\x70\x68\x61\x73\x65\x20\x69\x6e\x69\x74\x20\x28\x50\x45\x50\x20\x34\x38\x39\x29\x20\x61\x6e\x64\x20\x6c\x79\x69\x6e\x67\x20\x61\x62\x6f\x75\x74\x0a\x20\x20\x20\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x66\x6f\x72\x20\x6d\x75\x6c\x74\x69\x70\x6c\x65\x20\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x73\x20\x28\x6f\x72\x20\x70\x65\x72\x2d\x69\x6e\x74\x65\x72\x70\x72\x65\x74\x65\x72\x20\x47\x49\x4c\x29\x2e\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_disable_check = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "disable_check",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_19_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_bool._ascii.ob_base,
            & const_str_disable_check._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[53];
    }
importlib_util_toplevel_consts_19_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 52,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_incompatible_extension_module_restrictions.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[16];
    }
importlib_util_toplevel_consts_19_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 15,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x1d\x21\xa0\x2d\xd3\x1d\x30\x88\x04\xd5\x08\x1a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_19_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_disable_check._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(36)
importlib_util_toplevel_consts_19_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 18,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_19_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 1,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 151,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 727,
    .co_localsplusnames = & importlib_util_toplevel_consts_19_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib_util_toplevel_consts_19_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_19_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[40];
    }
const_str__override_multi_interp_extensions_check = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 39,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_override_multi_interp_extensions_check",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib_util_toplevel_consts_19_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__imp._ascii.ob_base,
            & const_str__override_multi_interp_extensions_check._ascii.ob_base,
            & const_str_override._ascii.ob_base,
            & const_str_old._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
importlib_util_toplevel_consts_19_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_incompatible_extension_module_restrictions.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib_util_toplevel_consts_19_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x13\x17\xd7\x13\x3f\xd1\x13\x3f\xc0\x04\xc7\x0d\xc1\x0d\xd3\x13\x4e\x88\x04\x8c\x08\xd8\x0f\x13\x88\x0b",
};
static
    struct _PyCode_DEF(78)
importlib_util_toplevel_consts_19_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 39,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_19_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 154,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 728,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & importlib_util_toplevel_consts_19_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_19_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib_util_toplevel_consts_19_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_old._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str__override_multi_interp_extensions_check._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[53];
    }
importlib_util_toplevel_consts_19_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 52,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_incompatible_extension_module_restrictions.__exit__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[31];
    }
importlib_util_toplevel_consts_19_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 30,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0e\x12\x8f\x68\x89\x68\x88\x03\xd8\x0c\x10\x88\x48\xdc\x08\x0c\xd7\x08\x34\xd1\x08\x34\xb0\x53\xd5\x08\x39",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib_util_toplevel_consts_19_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(args),
            & const_str_old._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(74)
importlib_util_toplevel_consts_19_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 37,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_19_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 158,
    .co_nlocalsplus = 3,
    .co_nlocals = 3,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 729,
    .co_localsplusnames = & importlib_util_toplevel_consts_19_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & importlib_util_toplevel_consts_19_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_19_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x00\x60\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib_util_toplevel_consts_19_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -1],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_19_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_disable_check._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[53];
    }
importlib_util_toplevel_consts_19_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 52,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_incompatible_extension_module_restrictions.override",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[22];
    }
importlib_util_toplevel_consts_19_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 21,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe0\x15\x19\xd7\x15\x27\xd2\x15\x27\x88\x72\xd0\x08\x2e\xa8\x51\xd0\x08\x2e",
};
static
    struct _PyCode_DEF(34)
importlib_util_toplevel_consts_19_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 17,
    },
    .co_consts = & importlib_util_toplevel_consts_19_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_19_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 163,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 730,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str_override._ascii.ob_base,
    .co_qualname = & importlib_util_toplevel_consts_19_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_19_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x02\x64\x01\x53\x00\x64\x02\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
importlib_util_toplevel_consts_19_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str__incompatible_extension_module_restrictions._ascii.ob_base,
            & importlib_util_toplevel_consts_19_consts_1._ascii.ob_base,
            & importlib_util_toplevel_consts_19_consts_2.ob_base.ob_base,
            & importlib_util_toplevel_consts_19_consts_3.ob_base.ob_base,
            & importlib_util_toplevel_consts_19_consts_4.ob_base.ob_base,
            & importlib_util_toplevel_consts_19_consts_5.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
importlib_util_toplevel_consts_19_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__init__),
            &_Py_ID(__enter__),
            &_Py_ID(__exit__),
            & const_str_property._ascii.ob_base,
            & const_str_override._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[43];
    }
importlib_util_toplevel_consts_19_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 42,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf1\x02\x1d\x05\x08\xf2\x3e\x01\x05\x31\xf2\x06\x02\x05\x14\xf2\x08\x03\x05\x3a\xf0\x0a\x00\x06\x0e\xf1\x02\x01\x05\x2f\xf3\x03\x00\x06\x0e\xf1\x02\x01\x05\x2f",
};
static
    struct _PyCode_DEF(50)
importlib_util_toplevel_consts_19 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib_util_toplevel_consts_19_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_19_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 119,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 731,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str__incompatible_extension_module_restrictions._ascii.ob_base,
    .co_qualname = & const_str__incompatible_extension_module_restrictions._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_19_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x65\x07\x64\x05\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x08\x79\x06",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__LazyModule = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LazyModule",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[76];
    }
importlib_util_toplevel_consts_21_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 75,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "A subclass of the module type which triggers loading upon attribute access.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[57];
    }
importlib_util_toplevel_consts_21_consts_2_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 56,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Trigger the load of the module and return the attribute.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_is_loading = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "is_loading",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib_util_toplevel_consts_21_consts_2_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "module object for ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[47];
    }
importlib_util_toplevel_consts_21_consts_2_consts_9 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 46,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " substituted in sys.modules during a lazy load",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib_util_toplevel_consts_21_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_21_consts_2_consts_0._ascii.ob_base,
            &_Py_ID(__spec__),
            & const_str_lock._ascii.ob_base,
            &_Py_ID(__class__),
            & const_str_is_loading._ascii.ob_base,
            Py_None,
            Py_True,
            &_Py_ID(__dict__),
            & importlib_util_toplevel_consts_21_consts_2_consts_8._ascii.ob_base,
            & importlib_util_toplevel_consts_21_consts_2_consts_9._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_types = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "types",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_ModuleType = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "ModuleType",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[17];
        }_object;
    }
importlib_util_toplevel_consts_21_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 17,
        },
        .ob_item = {
            &_Py_ID(object),
            &_Py_ID(__getattribute__),
            & const_str_loader_state._ascii.ob_base,
            & const_str__LazyModule._ascii.ob_base,
            &_Py_ID(name),
            &_Py_ID(items),
            &_Py_ID(id),
            & const_str_loader._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str_ValueError._ascii.ob_base,
            & const_str_update._ascii.ob_base,
            & const_str_types._ascii.ob_base,
            & const_str_ModuleType._ascii.ob_base,
            &_Py_ID(__class__),
            &_Py_ID(getattr),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib_util_toplevel_consts_21_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LazyModule.__getattribute__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[388];
    }
importlib_util_toplevel_consts_21_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 387,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x13\x19\xd7\x13\x2a\xd1\x13\x2a\xa8\x34\xb0\x1a\xd3\x13\x3c\x88\x08\xd8\x17\x1f\xd7\x17\x2c\xd1\x17\x2c\x88\x0c\xd8\x0d\x19\x98\x26\xd3\x0d\x21\xf4\x06\x00\x10\x16\xd7\x0f\x26\xd1\x0f\x26\xa0\x74\xa8\x5b\xd3\x0f\x39\xbc\x5b\xd2\x0f\x48\xf0\x0a\x00\x14\x20\xa0\x0c\xd2\x13\x2d\xdc\x1b\x21\xd7\x1b\x32\xd1\x1b\x32\xb0\x34\xb8\x14\xd3\x1b\x3e\xf7\x13\x00\x0e\x22\xd1\x0d\x21\xf0\x14\x00\x2e\x32\x90\x0c\x98\x5c\xd1\x10\x2a\xe4\x1b\x21\xd7\x1b\x32\xd1\x1b\x32\xb0\x34\xb8\x1a\xd3\x1b\x44\x90\x08\xf0\x0c\x00\x21\x29\xa7\x0d\xa1\x0d\x90\x0d\xf0\x06\x00\x1e\x2a\xa8\x2a\xd1\x1d\x35\x90\x0a\xd8\x1c\x24\x90\x09\xd8\x20\x22\x90\x0d\xd8\x22\x2b\xa7\x2f\xa1\x2f\xd6\x22\x33\x91\x4a\x90\x43\x98\x15\xf0\x06\x00\x18\x1b\xa0\x2a\xd1\x17\x2c\xd8\x2d\x32\x98\x0d\xa0\x63\xd2\x18\x2a\xdc\x19\x1b\x98\x49\xa0\x63\x99\x4e\xd3\x19\x2b\xac\x72\xb0\x2a\xb8\x53\xb1\x2f\xd3\x2f\x42\xd3\x19\x42\xd8\x2d\x32\x98\x0d\xa0\x63\xd2\x18\x2a\xf0\x0d\x00\x23\x34\xf0\x0e\x00\x11\x19\x97\x0f\x91\x0f\xd7\x10\x2b\xd1\x10\x2b\xa8\x44\xd4\x10\x31\xf0\x06\x00\x14\x21\xa4\x43\xa7\x4b\xa1\x4b\xd1\x13\x2f\xdc\x17\x19\x98\x24\x93\x78\xa4\x32\xa4\x63\xa7\x6b\xa1\x6b\xb0\x2d\xd1\x26\x40\xd3\x23\x41\xd2\x17\x41\xdc\x1e\x28\xd0\x2b\x3d\xb8\x6d\xd0\x3d\x4e\xf0\x00\x02\x4f\x01\x31\xf0\x00\x02\x2a\x31\xf3\x00\x02\x1f\x32\xf0\x00\x02\x19\x32\xf0\x0a\x00\x11\x19\x97\x0f\x91\x0f\xa0\x0d\xd4\x10\x2e\xe4\x21\x26\xd7\x21\x31\xd1\x21\x31\x90\x04\x94\x0e\xf7\x57\x01\x00\x0e\x22\xf4\x5a\x01\x00\x10\x17\x90\x74\x98\x54\xd3\x0f\x22\xd0\x08\x22\xf7\x5b\x01\x00\x0e\x22\xd0\x0d\x21\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
importlib_util_toplevel_consts_21_consts_2_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\xa8\x38\x45\x3d\x03\xc1\x2a\x41\x2d\x45\x3d\x03\xc3\x18\x42\x11\x45\x3d\x03\xc5\x3d\x05\x46\x06\x07",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_attr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "attr",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_original_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "original_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_attrs_then = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "attrs_then",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_attrs_now = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "attrs_now",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_attrs_updated = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "attrs_updated",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib_util_toplevel_consts_21_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_attr._ascii.ob_base,
            &_Py_ID(__spec__),
            & const_str_loader_state._ascii.ob_base,
            &_Py_ID(__dict__),
            & const_str_original_name._ascii.ob_base,
            & const_str_attrs_then._ascii.ob_base,
            & const_str_attrs_now._ascii.ob_base,
            & const_str_attrs_updated._ascii.ob_base,
            &_Py_ID(key),
            &_Py_ID(value),
        },
    },
};
static
    struct _PyCode_DEF(786)
importlib_util_toplevel_consts_21_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 393,
    },
    .co_consts = & importlib_util_toplevel_consts_21_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_21_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = & importlib_util_toplevel_consts_21_consts_2_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 18 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 172,
    .co_nlocalsplus = 11,
    .co_nlocals = 11,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 732,
    .co_localsplusnames = & importlib_util_toplevel_consts_21_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_ID(__getattribute__),
    .co_qualname = & importlib_util_toplevel_consts_21_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_21_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x02\x7c\x02\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x03\x7c\x03\x64\x02\x19\x00\x00\x00\x35\x00\x01\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x75\x00\x90\x01\x72\x23\x7c\x03\x64\x04\x19\x00\x00\x00\x72\x1f\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x63\x02\x64\x05\x64\x05\x64\x05\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x53\x00\x64\x06\x7c\x03\x64\x04\x3c\x00\x00\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x07\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x04\x7c\x02\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x03\x64\x07\x19\x00\x00\x00\x7d\x06\x7c\x04\x7d\x07\x69\x00\x7d\x08\x7c\x07\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x44\x00\x5d\x32\x00\x00\x5c\x02\x00\x00\x7d\x09\x7d\x0a\x7c\x09\x7c\x06\x76\x01\x72\x06\x7c\x0a\x7c\x08\x7c\x09\x3c\x00\x00\x00\x8c\x10\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\x7c\x09\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x09\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x73\x01\x8c\x2e\x7c\x0a\x7c\x08\x7c\x09\x3c\x00\x00\x00\x8c\x34\x04\x00\x7c\x02\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x05\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x00\x72\x37\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x6b\x37\x00\x00\x72\x0f\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x64\x08\x7c\x05\x9b\x02\x64\x09\x9d\x03\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x04\x6a\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x08\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x64\x05\x64\x05\x64\x05\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x74\x21\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x15\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[48];
    }
importlib_util_toplevel_consts_21_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 47,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Trigger the load and then perform the deletion.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_21_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_21_consts_3_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_delattr = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "delattr",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_21_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(__getattribute__),
            & const_str_delattr._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
importlib_util_toplevel_consts_21_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LazyModule.__delattr__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[29];
    }
importlib_util_toplevel_consts_21_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 28,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x09\x0d\xd7\x08\x1d\xd1\x08\x1d\x98\x64\xd4\x08\x23\xdc\x08\x0f\x90\x04\x90\x64\xd5\x08\x1b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_21_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_attr._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(62)
importlib_util_toplevel_consts_21_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & importlib_util_toplevel_consts_21_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_21_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 223,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 733,
    .co_localsplusnames = & importlib_util_toplevel_consts_21_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_ID(__delattr__),
    .co_qualname = & importlib_util_toplevel_consts_21_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_21_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib_util_toplevel_consts_21_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__LazyModule._ascii.ob_base,
            & importlib_util_toplevel_consts_21_consts_1._ascii.ob_base,
            & importlib_util_toplevel_consts_21_consts_2.ob_base.ob_base,
            & importlib_util_toplevel_consts_21_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
importlib_util_toplevel_consts_21_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            &_Py_ID(__getattribute__),
            &_Py_ID(__delattr__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[17];
    }
importlib_util_toplevel_consts_21_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 16,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x55\xf2\x04\x31\x05\x23\xf3\x66\x01\x05\x05\x1c",
};
static
    struct _PyCode_DEF(28)
importlib_util_toplevel_consts_21 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 14,
    },
    .co_consts = & importlib_util_toplevel_consts_21_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_21_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 168,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 734,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str__LazyModule._ascii.ob_base,
    .co_qualname = & const_str__LazyModule._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_21_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_LazyLoader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LazyLoader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[76];
    }
importlib_util_toplevel_consts_23_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 75,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "A loader that creates a module which defers loading until attribute access.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[33];
    }
importlib_util_toplevel_consts_23_consts_2_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 32,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "loader must define exec_module()",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & const_str_exec_module._ascii.ob_base,
            & importlib_util_toplevel_consts_23_consts_2_consts_2._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_hasattr._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
const_str___check_eager_loader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "__check_eager_loader",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
importlib_util_toplevel_consts_23_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LazyLoader.__check_eager_loader",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
importlib_util_toplevel_consts_23_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0f\x16\x90\x76\x98\x7d\xd4\x0f\x2d\xdc\x12\x1b\xd0\x1c\x3e\xd3\x12\x3f\xd0\x0c\x3f\xf0\x03\x00\x10\x2e",
};
static
    struct _PyCode_DEF(50)
importlib_util_toplevel_consts_23_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 25,
    },
    .co_consts = & importlib_util_toplevel_consts_23_consts_2_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_23_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 235,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 735,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_33_consts_4._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str___check_eager_loader._ascii.ob_base,
    .co_qualname = & importlib_util_toplevel_consts_23_consts_2_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_23_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x73\x0b\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[63];
    }
importlib_util_toplevel_consts_23_consts_3_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 62,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Construct a callable which returns the eager loader made lazy.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[37];
    }
importlib_util_toplevel_consts_23_consts_3_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 36,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LazyLoader.factory.<locals>.<lambda>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib_util_toplevel_consts_23_consts_3_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\xf8\x80\x00\xa1\x73\xa9\x36\xb0\x34\xd0\x2b\x42\xb8\x36\xd1\x2b\x42\xd4\x27\x43",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_3_consts_1_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(args),
            & const_str_kwargs._ascii.ob_base,
            & const_str_cls._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(32)
importlib_util_toplevel_consts_23_consts_3_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 16,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 31,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 244,
    .co_nlocalsplus = 4,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 2,
    .co_version = 736,
    .co_localsplusnames = & importlib_util_toplevel_consts_23_consts_3_consts_1_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & os_toplevel_consts_87_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_lambda),
    .co_qualname = & importlib_util_toplevel_consts_23_consts_3_consts_1_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_23_consts_3_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x95\x02\x97\x00\x02\x00\x89\x02\x02\x00\x89\x03\x7c\x00\x69\x00\x7c\x01\xa4\x01\x8e\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 1,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_23_consts_3_consts_0._ascii.ob_base,
            & importlib_util_toplevel_consts_23_consts_3_consts_1.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
const_str__LazyLoader__check_eager_loader = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_LazyLoader__check_eager_loader",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str__LazyLoader__check_eager_loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
importlib_util_toplevel_consts_23_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LazyLoader.factory",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[26];
    }
importlib_util_toplevel_consts_23_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 25,
    },
    .ob_shash = -1,
    .ob_sval = "\xf9\x80\x00\xf0\x06\x00\x09\x0c\xd7\x08\x20\xd1\x08\x20\xa0\x16\xd4\x08\x28\xdc\x0f\x43\xd0\x08\x43",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_3_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_cls._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[3];
    }
importlib_util_toplevel_consts_23_consts_3_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 2,
    },
    .ob_shash = -1,
    .ob_sval = "``",
};
static
    struct _PyCode_DEF(52)
importlib_util_toplevel_consts_23_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 26,
    },
    .co_consts = & importlib_util_toplevel_consts_23_consts_3_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_23_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 240,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 2,
    .co_nfreevars = 0,
    .co_version = 737,
    .co_localsplusnames = & importlib_util_toplevel_consts_23_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib_util_toplevel_consts_23_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_ID(factory),
    .co_qualname = & importlib_util_toplevel_consts_23_consts_3_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_23_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x87\x00\x87\x01\x97\x00\x89\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x89\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x88\x00\x88\x01\x66\x02\x64\x01\x84\x08\x53\x00",
    ._co_firsttraceable = 2,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str__LazyLoader__check_eager_loader._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
importlib_util_toplevel_consts_23_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LazyLoader.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
importlib_util_toplevel_consts_23_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x08\x0c\xd7\x08\x21\xd1\x08\x21\xa0\x26\xd4\x08\x29\xd8\x16\x1c\x88\x04\x8d\x0b",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_4_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_loader._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(52)
importlib_util_toplevel_consts_23_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 26,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_23_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 246,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 738,
    .co_localsplusnames = & importlib_util_toplevel_consts_23_consts_4_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & importlib_util_toplevel_consts_23_consts_4_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_23_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_5_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_loader._ascii.ob_base,
            & const_str_create_module._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
importlib_util_toplevel_consts_23_consts_5_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LazyLoader.create_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
importlib_util_toplevel_consts_23_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0f\x13\x8f\x7b\x89\x7b\xd7\x0f\x28\xd1\x0f\x28\xa8\x14\xd3\x0f\x2e\xd0\x08\x2e",
};
static
    struct _PyCode_DEF(56)
importlib_util_toplevel_consts_23_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 28,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_23_consts_5_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 250,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 739,
    .co_localsplusnames = & importlib__bootstrap_external_toplevel_consts_54_consts_3_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str_create_module._ascii.ob_base,
    .co_qualname = & importlib_util_toplevel_consts_23_consts_5_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_23_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib_util_toplevel_consts_23_consts_6_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Make the module load lazily.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_6_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_23_consts_6_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            &_Py_ID(__dict__),
            &_Py_ID(__class__),
            & const_str_lock._ascii.ob_base,
            Py_False,
            & const_str_is_loading._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_6_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(threading),
            & const_str_loader._ascii.ob_base,
            &_Py_ID(__spec__),
            &_Py_ID(__loader__),
            &_Py_ID(__dict__),
            &_Py_ID(copy),
            &_Py_ID(__class__),
            & const_str_RLock._ascii.ob_base,
            & const_str_loader_state._ascii.ob_base,
            & const_str__LazyModule._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[23];
    }
importlib_util_toplevel_consts_23_consts_6_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 22,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "LazyLoader.exec_module",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[129];
    }
importlib_util_toplevel_consts_23_consts_6_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 128,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf3\x08\x00\x09\x19\xd8\x21\x25\xa7\x1b\xa1\x1b\x88\x06\x8f\x0f\x89\x0f\xd4\x08\x1e\xd8\x1c\x20\x9f\x4b\x99\x4b\x88\x06\xd4\x08\x19\xf0\x0a\x00\x18\x1a\x88\x0c\xd8\x23\x29\xa7\x3f\xa1\x3f\xd7\x23\x37\xd1\x23\x37\xd3\x23\x39\x88\x0c\x90\x5a\xd1\x08\x20\xd8\x24\x2a\xd7\x24\x34\xd1\x24\x34\x88\x0c\x90\x5b\xd1\x08\x21\xd8\x1f\x28\x9f\x7f\x99\x7f\xd3\x1f\x30\x88\x0c\x90\x56\xd1\x08\x1c\xd8\x25\x2a\x88\x0c\x90\x5c\xd1\x08\x22\xd8\x27\x33\x88\x06\x8f\x0f\x89\x0f\xd4\x08\x24\xdc\x1b\x26\x88\x06\xd5\x08\x18",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
importlib_util_toplevel_consts_23_consts_6_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(self),
            &_Py_ID(module),
            &_Py_ID(threading),
            & const_str_loader_state._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(296)
importlib_util_toplevel_consts_23_consts_6 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 148,
    },
    .co_consts = & importlib_util_toplevel_consts_23_consts_6_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_23_consts_6_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 7 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 253,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 740,
    .co_localsplusnames = & importlib_util_toplevel_consts_23_consts_6_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str_exec_module._ascii.ob_base,
    .co_qualname = & importlib_util_toplevel_consts_23_consts_6_qualname._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_23_consts_6_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x7d\x02\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x69\x00\x7d\x03\x7c\x01\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x03\x3c\x00\x00\x00\x7c\x01\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x04\x3c\x00\x00\x00\x7c\x02\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x64\x05\x3c\x00\x00\x00\x64\x06\x7c\x03\x64\x07\x3c\x00\x00\x00\x7c\x03\x7c\x01\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x5f\x08\x00\x00\x00\x00\x00\x00\x00\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x5f\x06\x00\x00\x00\x00\x00\x00\x00\x00\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[8];
        }_object;
    }
importlib_util_toplevel_consts_23_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 8,
        },
        .ob_item = {
            & const_str_LazyLoader._ascii.ob_base,
            & importlib_util_toplevel_consts_23_consts_1._ascii.ob_base,
            & importlib_util_toplevel_consts_23_consts_2.ob_base.ob_base,
            & importlib_util_toplevel_consts_23_consts_3.ob_base.ob_base,
            & importlib_util_toplevel_consts_23_consts_4.ob_base.ob_base,
            & importlib_util_toplevel_consts_23_consts_5.ob_base.ob_base,
            & importlib_util_toplevel_consts_23_consts_6.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
importlib_util_toplevel_consts_23_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
            & const_str_staticmethod._ascii.ob_base,
            & const_str__LazyLoader__check_eager_loader._ascii.ob_base,
            & const_str_classmethod._ascii.ob_base,
            &_Py_ID(factory),
            &_Py_ID(__init__),
            & const_str_create_module._ascii.ob_base,
            & const_str_exec_module._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[63];
    }
importlib_util_toplevel_consts_23_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 62,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xe1\x04\x55\xe0\x05\x11\xf1\x02\x02\x05\x40\x01\xf3\x03\x00\x06\x12\xf0\x02\x02\x05\x40\x01\xf0\x08\x00\x06\x11\xf1\x02\x03\x05\x44\x01\xf3\x03\x00\x06\x11\xf0\x02\x03\x05\x44\x01\xf2\x0a\x02\x05\x1d\xf2\x08\x01\x05\x2f\xf3\x06\x11\x05\x27",
};
static
    struct _PyCode_DEF(66)
importlib_util_toplevel_consts_23 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 33,
    },
    .co_consts = & importlib_util_toplevel_consts_23_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_consts_23_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 231,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 741,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = & const_str_LazyLoader._ascii.ob_base,
    .co_qualname = & const_str_LazyLoader._ascii.ob_base,
    .co_linetable = & importlib_util_toplevel_consts_23_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x65\x04\x64\x02\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x05\x65\x06\x64\x03\x84\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5a\x07\x64\x04\x84\x00\x5a\x08\x64\x05\x84\x00\x5a\x09\x64\x06\x84\x00\x5a\x0a\x79\x07",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[26];
        }_object;
    }
importlib_util_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 26,
        },
        .ob_item = {
            & importlib_util_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & importlib_util_toplevel_consts_2._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_3._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_4._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_5._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_6._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_7._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_8._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_9._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_10._object.ob_base.ob_base,
            & importlib_util_toplevel_consts_11._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_72_consts_4_names._object.ob_base.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & importlib_util_toplevel_consts_15.ob_base.ob_base,
            & importlib_util_toplevel_consts_16.ob_base.ob_base,
            & importlib_util_toplevel_consts_17.ob_base.ob_base,
            & importlib_util_toplevel_consts_18.ob_base.ob_base,
            & importlib_util_toplevel_consts_19.ob_base.ob_base,
            & const_str__incompatible_extension_module_restrictions._ascii.ob_base,
            & importlib_util_toplevel_consts_21.ob_base.ob_base,
            & const_str__LazyModule._ascii.ob_base,
            & importlib_util_toplevel_consts_23.ob_base.ob_base,
            & const_str_LazyLoader._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[26];
        }_object;
    }
importlib_util_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 26,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str__abc._ascii.ob_base,
            & const_str_Loader._ascii.ob_base,
            &_Py_ID(_bootstrap),
            & const_str_module_from_spec._ascii.ob_base,
            & const_str__resolve_name._ascii.ob_base,
            & const_str_spec_from_loader._ascii.ob_base,
            & const_str__find_spec._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str_MAGIC_NUMBER._ascii.ob_base,
            & const_str__RAW_MAGIC_NUMBER._ascii.ob_base,
            & const_str_cache_from_source._ascii.ob_base,
            & const_str_decode_source._ascii.ob_base,
            & const_str_source_from_cache._ascii.ob_base,
            & const_str_spec_from_file_location._ascii.ob_base,
            & const_str__imp._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_types._ascii.ob_base,
            & const_str_source_hash._ascii.ob_base,
            & const_str_resolve_name._ascii.ob_base,
            & const_str__find_spec_from_path._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str__incompatible_extension_module_restrictions._ascii.ob_base,
            & const_str_ModuleType._ascii.ob_base,
            & const_str__LazyModule._ascii.ob_base,
            & const_str_LazyLoader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[114];
    }
importlib_util_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 113,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xd9\x00\x33\xdd\x00\x18\xdd\x00\x28\xdd\x00\x25\xdd\x00\x28\xdd\x00\x22\xdd\x00\x2d\xdd\x00\x32\xdd\x00\x32\xdd\x00\x2e\xdd\x00\x32\xdd\x00\x38\xe3\x00\x0b\xdb\x00\x0a\xdb\x00\x0c\xf2\x06\x02\x01\x3d\xf2\x0a\x0c\x01\x37\xf3\x1e\x1c\x01\x18\xf3\x3e\x2a\x01\x18\xf7\x62\x01\x2e\x01\x2f\xf1\x00\x2e\x01\x2f\xf4\x62\x01\x3c\x01\x1c\x90\x25\xd7\x12\x22\xd1\x12\x22\xf4\x00\x3c\x01\x1c\xf4\x7e\x01\x27\x01\x27\x90\x16\xf5\x00\x27\x01\x27",
};
static
    struct _PyCode_DEF(276)
importlib_util_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 138,
    },
    .co_consts = & importlib_util_toplevel_consts._object.ob_base.ob_base,
    .co_names = & importlib_util_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 742,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib_util_toplevel_consts_15_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & importlib_util_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x6d\x02\x5a\x02\x01\x00\x64\x01\x64\x03\x6c\x03\x6d\x04\x5a\x04\x01\x00\x64\x01\x64\x04\x6c\x03\x6d\x05\x5a\x05\x01\x00\x64\x01\x64\x05\x6c\x03\x6d\x06\x5a\x06\x01\x00\x64\x01\x64\x06\x6c\x03\x6d\x07\x5a\x07\x01\x00\x64\x01\x64\x07\x6c\x08\x6d\x09\x5a\x09\x01\x00\x64\x01\x64\x08\x6c\x08\x6d\x0a\x5a\x0a\x01\x00\x64\x01\x64\x09\x6c\x08\x6d\x0b\x5a\x0b\x01\x00\x64\x01\x64\x0a\x6c\x08\x6d\x0c\x5a\x0c\x01\x00\x64\x01\x64\x0b\x6c\x08\x6d\x0d\x5a\x0d\x01\x00\x64\x01\x64\x0c\x6c\x08\x6d\x0e\x5a\x0e\x01\x00\x64\x0d\x64\x0e\x6c\x0f\x5a\x0f\x64\x0d\x64\x0e\x6c\x10\x5a\x10\x64\x0d\x64\x0e\x6c\x11\x5a\x11\x64\x0f\x84\x00\x5a\x12\x64\x10\x84\x00\x5a\x13\x64\x19\x64\x11\x84\x01\x5a\x14\x64\x19\x64\x12\x84\x01\x5a\x15\x02\x00\x47\x00\x64\x13\x84\x00\x64\x14\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x16\x02\x00\x47\x00\x64\x15\x84\x00\x64\x16\x65\x11\x6a\x2e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x18\x02\x00\x47\x00\x64\x17\x84\x00\x64\x18\x65\x02\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x19\x79\x0e",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_importlib_util_toplevel(void)
{
    return Py_NewRef((PyObject *) &importlib_util_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[58];
    }
importlib_machinery_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 57,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "The machinery of importlib: finders, loaders, hooks, etc.",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_ModuleSpec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_3 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_BuiltinImporter._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_4 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_FrozenImporter._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
importlib_machinery_toplevel_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_SOURCE_SUFFIXES._ascii.ob_base,
            & const_str_DEBUG_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_OPTIMIZED_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_EXTENSION_SUFFIXES._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_6 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_WindowsRegistryFinder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_7 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_PathFinder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_8 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_FileFinder._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_9 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_SourceFileLoader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_10 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_SourcelessFileLoader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_11 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_ExtensionFileLoader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_12 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_NamespaceLoader._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[66];
    }
importlib_machinery_toplevel_consts_13_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 65,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Returns a list of all recognized module suffixes for this process",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
importlib_machinery_toplevel_consts_13_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & importlib_machinery_toplevel_consts_13_consts_0._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
importlib_machinery_toplevel_consts_13_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_SOURCE_SUFFIXES._ascii.ob_base,
            & const_str_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_EXTENSION_SUFFIXES._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
importlib_machinery_toplevel_consts_13_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen importlib.machinery>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_all_suffixes = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "all_suffixes",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
importlib_machinery_toplevel_consts_13_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe4\x0b\x1a\xd4\x1d\x2e\xd1\x0b\x2e\xd4\x31\x43\xd1\x0b\x43\xd0\x04\x43",
};
static
    struct _PyCode_DEF(42)
importlib_machinery_toplevel_consts_13 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 21,
    },
    .co_consts = & importlib_machinery_toplevel_consts_13_consts._object.ob_base.ob_base,
    .co_names = & importlib_machinery_toplevel_consts_13_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 18,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 743,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib_machinery_toplevel_consts_13_filename._ascii.ob_base,
    .co_name = & const_str_all_suffixes._ascii.ob_base,
    .co_qualname = & const_str_all_suffixes._ascii.ob_base,
    .co_linetable = & importlib_machinery_toplevel_consts_13_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x7a\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[15];
        }_object;
    }
importlib_machinery_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 15,
        },
        .ob_item = {
            & importlib_machinery_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 1],
            & importlib_machinery_toplevel_consts_2._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_3._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_4._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_5._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_6._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_7._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_8._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_9._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_10._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_11._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_12._object.ob_base.ob_base,
            & importlib_machinery_toplevel_consts_13.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[19];
        }_object;
    }
importlib_machinery_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 19,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            &_Py_ID(_bootstrap),
            & const_str_ModuleSpec._ascii.ob_base,
            & const_str_BuiltinImporter._ascii.ob_base,
            & const_str_FrozenImporter._ascii.ob_base,
            & const_str__bootstrap_external._ascii.ob_base,
            & const_str_SOURCE_SUFFIXES._ascii.ob_base,
            & const_str_DEBUG_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_OPTIMIZED_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_BYTECODE_SUFFIXES._ascii.ob_base,
            & const_str_EXTENSION_SUFFIXES._ascii.ob_base,
            & const_str_WindowsRegistryFinder._ascii.ob_base,
            & const_str_PathFinder._ascii.ob_base,
            & const_str_FileFinder._ascii.ob_base,
            & const_str_SourceFileLoader._ascii.ob_base,
            & const_str_SourcelessFileLoader._ascii.ob_base,
            & const_str_ExtensionFileLoader._ascii.ob_base,
            & const_str_NamespaceLoader._ascii.ob_base,
            & const_str_all_suffixes._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[57];
    }
importlib_machinery_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 56,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xd9\x00\x3f\xe5\x00\x22\xdd\x00\x27\xdd\x00\x26\xf7\x02\x02\x01\x29\xf5\x00\x02\x01\x29\xf5\x06\x00\x01\x37\xdd\x00\x2b\xdd\x00\x2b\xdd\x00\x31\xdd\x00\x35\xdd\x00\x34\xdd\x00\x30\xf3\x06\x02\x01\x44\x01",
};
static
    struct _PyCode_DEF(162)
importlib_machinery_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 81,
    },
    .co_consts = & importlib_machinery_toplevel_consts._object.ob_base.ob_base,
    .co_names = & importlib_machinery_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 744,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & importlib_machinery_toplevel_consts_13_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & importlib_machinery_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x6d\x02\x5a\x02\x01\x00\x64\x01\x64\x03\x6c\x01\x6d\x03\x5a\x03\x01\x00\x64\x01\x64\x04\x6c\x01\x6d\x04\x5a\x04\x01\x00\x64\x01\x64\x05\x6c\x05\x6d\x06\x5a\x06\x6d\x07\x5a\x07\x6d\x08\x5a\x08\x6d\x09\x5a\x09\x6d\x0a\x5a\x0a\x01\x00\x64\x01\x64\x06\x6c\x05\x6d\x0b\x5a\x0b\x01\x00\x64\x01\x64\x07\x6c\x05\x6d\x0c\x5a\x0c\x01\x00\x64\x01\x64\x08\x6c\x05\x6d\x0d\x5a\x0d\x01\x00\x64\x01\x64\x09\x6c\x05\x6d\x0e\x5a\x0e\x01\x00\x64\x01\x64\x0a\x6c\x05\x6d\x0f\x5a\x0f\x01\x00\x64\x01\x64\x0b\x6c\x05\x6d\x10\x5a\x10\x01\x00\x64\x01\x64\x0c\x6c\x05\x6d\x11\x5a\x11\x01\x00\x64\x0d\x84\x00\x5a\x12\x79\x0e",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_importlib_machinery_toplevel(void)
{
    return Py_NewRef((PyObject *) &importlib_machinery_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[347];
    }
runpy_toplevel_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 346,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x72\x75\x6e\x70\x79\x2e\x70\x79\x20\x2d\x20\x6c\x6f\x63\x61\x74\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x75\x6e\x6e\x69\x6e\x67\x20\x50\x79\x74\x68\x6f\x6e\x20\x63\x6f\x64\x65\x20\x75\x73\x69\x6e\x67\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x0a\x0a\x50\x72\x6f\x76\x69\x64\x65\x73\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x66\x6f\x72\x20\x6c\x6f\x63\x61\x74\x69\x6e\x67\x20\x61\x6e\x64\x20\x72\x75\x6e\x6e\x69\x6e\x67\x20\x50\x79\x74\x68\x6f\x6e\x20\x73\x63\x72\x69\x70\x74\x73\x20\x75\x73\x69\x6e\x67\x20\x74\x68\x65\x20\x50\x79\x74\x68\x6f\x6e\x0a\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x20\x69\x6e\x73\x74\x65\x61\x64\x20\x6f\x66\x20\x74\x68\x65\x20\x6e\x61\x74\x69\x76\x65\x20\x66\x69\x6c\x65\x73\x79\x73\x74\x65\x6d\x2e\x0a\x0a\x54\x68\x69\x73\x20\x61\x6c\x6c\x6f\x77\x73\x20\x50\x79\x74\x68\x6f\x6e\x20\x63\x6f\x64\x65\x20\x74\x6f\x20\x70\x6c\x61\x79\x20\x6e\x69\x63\x65\x6c\x79\x20\x77\x69\x74\x68\x20\x6e\x6f\x6e\x2d\x66\x69\x6c\x65\x73\x79\x73\x74\x65\x6d\x20\x62\x61\x73\x65\x64\x20\x50\x45\x50\x20\x33\x30\x32\x0a\x69\x6d\x70\x6f\x72\x74\x65\x72\x73\x20\x77\x68\x65\x6e\x20\x6c\x6f\x63\x61\x74\x69\x6e\x67\x20\x73\x75\x70\x70\x6f\x72\x74\x20\x73\x63\x72\x69\x70\x74\x73\x20\x61\x73\x20\x77\x65\x6c\x6c\x20\x61\x73\x20\x77\x68\x65\x6e\x20\x69\x6d\x70\x6f\x72\x74\x69\x6e\x67\x20\x6d\x6f\x64\x75\x6c\x65\x73\x2e\x0a",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_run_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "run_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_run_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "run_path",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str__TempModule = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_TempModule",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[68];
    }
runpy_toplevel_consts_5_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 67,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Temporarily replace a module in sys.modules with an empty namespace",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_mod_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mod_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str__saved_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_saved_module",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
runpy_toplevel_consts_5_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_mod_name._ascii.ob_base,
            & const_str_ModuleType._ascii.ob_base,
            &_Py_ID(module),
            & const_str__saved_module._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
runpy_toplevel_consts_5_consts_2_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen runpy>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
runpy_toplevel_consts_5_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_TempModule.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[30];
    }
runpy_toplevel_consts_5_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 29,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x18\x20\x88\x04\x8c\x0d\xdc\x16\x20\xa0\x18\xd3\x16\x2a\x88\x04\x8c\x0b\xd8\x1d\x1f\x88\x04\xd5\x08\x1a",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
runpy_toplevel_consts_5_consts_2_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            &_Py_ID(self),
            & const_str_mod_name._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(64)
runpy_toplevel_consts_5_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 32,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_5_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 28,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 745,
    .co_localsplusnames = & runpy_toplevel_consts_5_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & runpy_toplevel_consts_5_consts_2_qualname._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_5_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x67\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
runpy_toplevel_consts_5_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_mod_name._ascii.ob_base,
            & const_str__saved_module._ascii.ob_base,
            &_Py_ID(append),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str_KeyError._ascii.ob_base,
            &_Py_ID(module),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
runpy_toplevel_consts_5_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_TempModule.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[91];
    }
runpy_toplevel_consts_5_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 90,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x13\x17\x97\x3d\x91\x3d\x88\x08\xf0\x02\x03\x09\x11\xd8\x0c\x10\xd7\x0c\x1e\xd1\x0c\x1e\xd7\x0c\x25\xd1\x0c\x25\xa4\x63\xa7\x6b\xa1\x6b\xb0\x28\xd1\x26\x3b\xd4\x0c\x3c\xf0\x06\x00\x21\x25\xa7\x0b\xa1\x0b\x8c\x03\x8f\x0b\x89\x0b\x90\x48\xd1\x08\x1d\xd8\x0f\x13\x88\x0b\xf8\xf4\x07\x00\x10\x18\xf2\x00\x01\x09\x11\xd9\x0c\x10\xf0\x03\x01\x09\x11\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
runpy_toplevel_consts_5_consts_3_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x8e\x2c\x41\x19\x00\xc1\x19\x09\x41\x25\x03\xc1\x24\x01\x41\x25\x03",
};
static
    struct _PyCode_DEF(208)
runpy_toplevel_consts_5_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 104,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_5_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = & runpy_toplevel_consts_5_consts_3_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 6 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 33,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 746,
    .co_localsplusnames = & runpy_toplevel_consts_5_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & runpy_toplevel_consts_5_consts_3_qualname._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_5_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x01\x09\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x7c\x00\x53\x00\x23\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x8c\x2a\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
runpy_toplevel_consts_5_consts_4_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str__saved_module._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str_mod_name._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
runpy_toplevel_consts_5_consts_4_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_TempModule.__exit__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[77];
    }
runpy_toplevel_consts_5_consts_4_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 76,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\xd7\x0b\x1d\xd2\x0b\x1d\xd8\x29\x2d\xd7\x29\x3b\xd1\x29\x3b\xb8\x41\xd1\x29\x3e\x8c\x43\x8f\x4b\x89\x4b\x98\x04\x9f\x0d\x99\x0d\xd1\x0c\x26\xf0\x06\x00\x1e\x20\x88\x04\xd5\x08\x1a\xf4\x03\x00\x11\x14\x97\x0b\x91\x0b\x98\x44\x9f\x4d\x99\x4d\xd0\x10\x2a\xd8\x1d\x1f\x88\x04\xd5\x08\x1a",
};
static
    struct _PyCode_DEF(196)
runpy_toplevel_consts_5_consts_4 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 98,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_5_consts_4_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 42,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 747,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & runpy_toplevel_consts_5_consts_4_qualname._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_5_consts_4_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x72\x32\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x00\x00\x00\x67\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x3d\x00\x67\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
runpy_toplevel_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__TempModule._ascii.ob_base,
            & runpy_toplevel_consts_5_consts_1._ascii.ob_base,
            & runpy_toplevel_consts_5_consts_2.ob_base.ob_base,
            & runpy_toplevel_consts_5_consts_3.ob_base.ob_base,
            & runpy_toplevel_consts_5_consts_4.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[21];
    }
runpy_toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 20,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xd9\x04\x4d\xf2\x02\x03\x05\x20\xf2\x0a\x07\x05\x14\xf3\x12\x05\x05\x20",
};
static
    struct _PyCode_DEF(34)
runpy_toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 17,
    },
    .co_consts = & runpy_toplevel_consts_5_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 26,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 748,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__TempModule._ascii.ob_base,
    .co_qualname = & const_str__TempModule._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x64\x04\x84\x00\x5a\x06\x79\x05",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str__ModifiedArgv0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModifiedArgv0",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str__saved_value = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_saved_value",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__sentinel = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_sentinel",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
runpy_toplevel_consts_7_consts_1_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(value),
            &_Py_ID(object),
            & const_str__saved_value._ascii.ob_base,
            & const_str__sentinel._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
runpy_toplevel_consts_7_consts_1_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModifiedArgv0.__init__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[27];
    }
runpy_toplevel_consts_7_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 26,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x15\x1a\x88\x04\x8c\x0a\xdc\x2d\x33\xab\x58\xd0\x08\x35\x88\x04\xd4\x08\x19\x98\x44\x9d\x4e",
};
static
    struct _PyCode_DEF(62)
runpy_toplevel_consts_7_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 31,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_7_consts_1_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 50,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 749,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = &_Py_ID(__init__),
    .co_qualname = & runpy_toplevel_consts_7_consts_1_qualname._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_7_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x01\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x78\x01\x7c\x00\x5f\x02\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x03\x00\x00\x00\x00\x00\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[31];
    }
runpy_toplevel_consts_7_consts_2_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 30,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Already preserving saved value",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
runpy_toplevel_consts_7_consts_2_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            & runpy_toplevel_consts_7_consts_2_consts_1._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
runpy_toplevel_consts_7_consts_2_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str__saved_value._ascii.ob_base,
            & const_str__sentinel._ascii.ob_base,
            & const_str_RuntimeError._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(argv),
            &_Py_ID(value),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
runpy_toplevel_consts_7_consts_2_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModifiedArgv0.__enter__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[66];
    }
runpy_toplevel_consts_7_consts_2_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 65,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x0b\x0f\xd7\x0b\x1c\xd1\x0b\x1c\xa0\x44\xa7\x4e\xa1\x4e\xd1\x0b\x32\xdc\x12\x1e\xd0\x1f\x3f\xd3\x12\x40\xd0\x0c\x40\xdc\x1c\x1f\x9f\x48\x99\x48\xa0\x51\x99\x4b\x88\x04\xd4\x08\x19\xd8\x16\x1a\x97\x6a\x91\x6a\x8c\x03\x8f\x08\x89\x08\x90\x11\x8a\x0b",
};
static
    struct _PyCode_DEF(180)
runpy_toplevel_consts_7_consts_2 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 90,
    },
    .co_consts = & runpy_toplevel_consts_7_consts_2_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_7_consts_2_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 54,
    .co_nlocalsplus = 1,
    .co_nlocals = 1,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 750,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_7_consts_2_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_characters[32]),
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = &_Py_ID(__enter__),
    .co_qualname = & runpy_toplevel_consts_7_consts_2_qualname._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_7_consts_2_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x75\x01\x72\x0b\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x19\x00\x00\x00\x7c\x00\x5f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x02\x3c\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
runpy_toplevel_consts_7_consts_3_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__sentinel._ascii.ob_base,
            &_Py_ID(value),
            & const_str__saved_value._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(argv),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
runpy_toplevel_consts_7_consts_3_qualname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_ModifiedArgv0.__exit__",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[33];
    }
runpy_toplevel_consts_7_consts_3_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 32,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x15\x19\x97\x5e\x91\x5e\x88\x04\x8c\x0a\xd8\x16\x1a\xd7\x16\x27\xd1\x16\x27\x8c\x03\x8f\x08\x89\x08\x90\x11\x8a\x0b",
};
static
    struct _PyCode_DEF(96)
runpy_toplevel_consts_7_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 48,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_16_consts_2_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_7_consts_3_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 7,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 60,
    .co_nlocalsplus = 2,
    .co_nlocals = 2,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 751,
    .co_localsplusnames = & importlib__bootstrap_toplevel_consts_30_consts_5_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_14_consts_2_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = &_Py_ID(__exit__),
    .co_qualname = & runpy_toplevel_consts_7_consts_3_qualname._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_7_consts_3_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x5f\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x3c\x00\x00\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
runpy_toplevel_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str__ModifiedArgv0._ascii.ob_base,
            & runpy_toplevel_consts_7_consts_1.ob_base.ob_base,
            & runpy_toplevel_consts_7_consts_2.ob_base.ob_base,
            & runpy_toplevel_consts_7_consts_3.ob_base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[18];
    }
runpy_toplevel_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 17,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xf2\x02\x02\x05\x36\xf2\x08\x04\x05\x21\xf3\x0c\x02\x05\x28",
};
static
    struct _PyCode_DEF(30)
runpy_toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 15,
    },
    .co_consts = & runpy_toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & importlib__bootstrap_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 49,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 752,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__ModifiedArgv0._ascii.ob_base,
    .co_qualname = & const_str__ModifiedArgv0._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x84\x00\x5a\x03\x64\x02\x84\x00\x5a\x04\x64\x03\x84\x00\x5a\x05\x79\x04",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[42];
    }
runpy_toplevel_consts_9_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 41,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Helper to run code in nominated namespace",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
runpy_toplevel_consts_9_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__file__),
            & const_str___cached__._ascii.ob_base,
            &_Py_ID(__doc__),
            &_Py_ID(__loader__),
            &_Py_ID(__package__),
            &_Py_ID(__spec__),
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
runpy_toplevel_consts_9_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & runpy_toplevel_consts_9_consts_0._ascii.ob_base,
            Py_None,
            & runpy_toplevel_consts_9_consts_2._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
runpy_toplevel_consts_9_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_update._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            &_Py_ID(origin),
            & const_str_cached._ascii.ob_base,
            &_Py_ID(parent),
            & const_str_exec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str__run_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_run_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[145];
    }
runpy_toplevel_consts_9_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 144,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x08\x14\xd0\x07\x1f\xd8\x08\x13\xd7\x08\x1a\xd1\x08\x1a\x98\x3c\xd4\x08\x28\xd8\x07\x0f\xd0\x07\x17\xd8\x11\x15\x88\x06\xd8\x10\x1b\x88\x05\xd8\x11\x15\x89\x06\xe0\x11\x19\x97\x1f\x91\x1f\x88\x06\xd8\x10\x18\x97\x0f\x91\x0f\x88\x05\xd8\x11\x19\x97\x1f\x91\x1f\x88\x06\xd8\x0b\x13\xd0\x0b\x1b\xd8\x17\x1f\x97\x7f\x91\x7f\x88\x48\xd8\x04\x0f\xd7\x04\x16\xd1\x04\x16\xa0\x28\xd8\x22\x27\xd8\x24\x2a\xd8\x21\x25\xd8\x24\x2a\xd8\x25\x2d\xd8\x22\x2a\xf0\x0d\x00\x05\x17\xf4\x00\x06\x05\x2c\xf4\x0e\x00\x05\x09\x88\x14\x88\x7b\xd4\x04\x1b\xd8\x0b\x16\xd0\x04\x16",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_run_globals = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "run_globals",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_init_globals = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "init_globals",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_mod_spec = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mod_spec",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_pkg_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pkg_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_script_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "script_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[6];
    }
const_str_fname = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 5,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "fname",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
runpy_toplevel_consts_9_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            &_Py_ID(code),
            & const_str_run_globals._ascii.ob_base,
            & const_str_init_globals._ascii.ob_base,
            & const_str_mod_name._ascii.ob_base,
            & const_str_mod_spec._ascii.ob_base,
            & const_str_pkg_name._ascii.ob_base,
            & const_str_script_name._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_fname._ascii.ob_base,
            & const_str_cached._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(234)
runpy_toplevel_consts_9 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 117,
    },
    .co_consts = & runpy_toplevel_consts_9_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_9_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 7,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 19 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 65,
    .co_nlocalsplus = 10,
    .co_nlocals = 10,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 753,
    .co_localsplusnames = & runpy_toplevel_consts_9_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__run_code._ascii.ob_base,
    .co_qualname = & const_str__run_code._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_9_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x81\x11\x7c\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x80\x07\x64\x01\x7d\x07\x7c\x06\x7d\x08\x64\x01\x7d\x09\x6e\x32\x7c\x04\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x07\x7c\x04\x6a\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x04\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x09\x7c\x05\x80\x0c\x7c\x04\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x01\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x7c\x08\x7c\x09\x64\x01\x7c\x07\x7c\x05\x7c\x04\xac\x02\xab\x07\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x01\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[54];
    }
runpy_toplevel_consts_10_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 53,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Helper to run code in new namespace with sys modified",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
runpy_toplevel_consts_10_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & runpy_toplevel_consts_10_consts_0._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
runpy_toplevel_consts_10_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            &_Py_ID(origin),
            & const_str__TempModule._ascii.ob_base,
            & const_str__ModifiedArgv0._ascii.ob_base,
            &_Py_ID(module),
            &_Py_ID(__dict__),
            & const_str__run_code._ascii.ob_base,
            &_Py_ID(copy),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str__run_module_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_run_module_code",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[135];
    }
runpy_toplevel_consts_10_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 134,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x08\x00\x1c\x24\xd0\x1b\x2b\x89\x4b\xb0\x18\xb7\x1f\xb1\x1f\x80\x45\xdc\x09\x14\x90\x58\xd4\x09\x1e\xa0\x2b\xac\x7e\xb8\x65\xd5\x2f\x44\xd8\x16\x21\xd7\x16\x28\xd1\x16\x28\xd7\x16\x31\xd1\x16\x31\x88\x0b\xdc\x08\x11\x90\x24\x98\x0b\xa0\x5c\xd8\x12\x1a\x98\x48\xa0\x68\xb0\x0b\xf4\x03\x01\x09\x3d\xf7\x05\x00\x30\x45\x01\xd7\x09\x1e\xf0\x0c\x00\x0c\x17\xd7\x0b\x1b\xd1\x0b\x1b\xd3\x0b\x1d\xd0\x04\x1d\xf7\x0d\x00\x30\x45\x01\xd0\x2f\x44\xfa\xd7\x09\x1e\xf0\x0c\x00\x0c\x17\xd7\x0b\x1b\xd1\x0b\x1b\xd3\x0b\x1d\xd0\x04\x1d\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[35];
    }
runpy_toplevel_consts_10_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 34,
    },
    .ob_shash = -1,
    .ob_sval = "\x9c\x0c\x41\x3c\x03\xa8\x28\x41\x30\x05\xc1\x10\x08\x41\x3c\x03\xc1\x30\x05\x41\x39\x09\xc1\x35\x07\x41\x3c\x03\xc1\x3c\x05\x42\x14\x07",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_temp_module = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "temp_module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_mod_globals = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "mod_globals",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
runpy_toplevel_consts_10_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            &_Py_ID(code),
            & const_str_init_globals._ascii.ob_base,
            & const_str_mod_name._ascii.ob_base,
            & const_str_mod_spec._ascii.ob_base,
            & const_str_pkg_name._ascii.ob_base,
            & const_str_script_name._ascii.ob_base,
            & const_str_fname._ascii.ob_base,
            & const_str_temp_module._ascii.ob_base,
            & const_str_mod_globals._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(302)
runpy_toplevel_consts_10 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 151,
    },
    .co_consts = & runpy_toplevel_consts_10_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_10_names._object.ob_base.ob_base,
    .co_exceptiontable = & runpy_toplevel_consts_10_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 6,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 20 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 11,
    .co_firstlineno = 91,
    .co_nlocalsplus = 9,
    .co_nlocals = 9,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 754,
    .co_localsplusnames = & runpy_toplevel_consts_10_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_61_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__run_module_code._ascii.ob_base,
    .co_qualname = & const_str__run_module_code._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_10_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x03\x80\x02\x7c\x05\x6e\x0b\x7c\x03\x6a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x07\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x7c\x07\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x08\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x08\x7c\x01\x7c\x02\x7c\x03\x7c\x04\x7c\x05\xab\x07\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7f\x08\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x21\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x7f\x08\x6a\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x53\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[36];
    }
runpy_toplevel_consts_11_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 35,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Relative module names not supported",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
runpy_toplevel_consts_11_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_warn._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[155];
    }
runpy_toplevel_consts_11_consts_6 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 154,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "{mod_name!r} found in sys.modules after import of package {pkg_name!r}, but prior to execution of {mod_name!r}; this may result in unpredictable behaviour",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
runpy_toplevel_consts_11_consts_7 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_mod_name._ascii.ob_base,
            & const_str_pkg_name._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[59];
    }
runpy_toplevel_consts_11_consts_8 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 58,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Error while finding module specification for {!r} ({}: {})",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
runpy_toplevel_consts_11_consts_10 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ". Try using '",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
runpy_toplevel_consts_11_consts_12 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "' instead of '",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[22];
    }
runpy_toplevel_consts_11_consts_13 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 21,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "' as the module name.",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
runpy_toplevel_consts_11_consts_14 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "No module named %s",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
runpy_toplevel_consts_11_consts_16 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = ".__main__",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[38];
    }
runpy_toplevel_consts_11_consts_17 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 37,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Cannot use package as __main__ module",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[46];
    }
runpy_toplevel_consts_11_consts_19 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 45,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " is a package and cannot be directly executed",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[49];
    }
runpy_toplevel_consts_11_consts_20 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 48,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "%r is a namespace package and cannot be executed",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[32];
    }
runpy_toplevel_consts_11_consts_21 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 31,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "No code object available for %s",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[22];
        }_object;
    }
runpy_toplevel_consts_11_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 22,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            & runpy_toplevel_consts_11_consts_2._ascii.ob_base,
            &_Py_ID(__path__),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & runpy_toplevel_consts_11_consts_5._object.ob_base.ob_base,
            & runpy_toplevel_consts_11_consts_6._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_7._object.ob_base.ob_base,
            & runpy_toplevel_consts_11_consts_8._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_46_consts_5_consts_12._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_10._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + -3],
            & runpy_toplevel_consts_11_consts_12._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_13._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_14._ascii.ob_base,
            &_Py_ID(__main__),
            & runpy_toplevel_consts_11_consts_16._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_17._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_55_consts_3._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_19._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_20._ascii.ob_base,
            & runpy_toplevel_consts_11_consts_21._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
const_str_RuntimeWarning = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "RuntimeWarning",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[5];
    }
const_str_util = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 4,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "util",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__get_module_details = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_module_details",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[26];
        }_object;
    }
runpy_toplevel_consts_11_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 26,
        },
        .ob_item = {
            & const_str_startswith._ascii.ob_base,
            & const_str_rpartition._ascii.ob_base,
            &_Py_ID(__import__),
            & const_str_ImportError._ascii.ob_base,
            &_Py_ID(name),
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(get),
            & const_str_hasattr._ascii.ob_base,
            &_Py_ID(warnings),
            & const_str_warn._ascii.ob_base,
            &_Py_ID(format),
            & const_str_RuntimeWarning._ascii.ob_base,
            &_Py_ID(importlib),
            & const_str_util._ascii.ob_base,
            & const_str_find_spec._ascii.ob_base,
            & const_str_AttributeError._ascii.ob_base,
            & const_str_TypeError._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
            & const_str_endswith._ascii.ob_base,
            &_Py_ID(type),
            &_Py_ID(__name__),
            & const_str_submodule_search_locations._ascii.ob_base,
            & const_str__get_module_details._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            & const_str_get_code._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[656];
    }
runpy_toplevel_consts_11_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 655,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xd8\x07\x0f\xd7\x07\x1a\xd1\x07\x1a\x98\x33\xd4\x07\x1f\xd9\x0e\x13\xd0\x14\x39\xd3\x0e\x3a\xd0\x08\x3a\xd8\x15\x1d\xd7\x15\x28\xd1\x15\x28\xa8\x13\xd3\x15\x2d\x81\x4e\x80\x48\x88\x61\x90\x11\xd9\x07\x0f\xf0\x04\x08\x09\x16\xdc\x0c\x16\x90\x78\xd4\x0c\x20\xf4\x12\x00\x14\x17\x97\x3b\x91\x3b\x97\x3f\x91\x3f\xa0\x38\xd3\x13\x2c\x88\x08\xd8\x0b\x13\xd0\x0b\x1f\xac\x07\xb0\x08\xb8\x2a\xd4\x28\x45\xdd\x0c\x25\xf0\x02\x03\x13\x1c\xf7\x06\x00\x1d\x23\x99\x46\xa8\x48\xb8\x78\x98\x46\xd3\x1c\x48\xf0\x07\x00\x0d\x10\xf1\x08\x00\x0d\x11\x94\x1e\xa0\x03\xd3\x11\x24\xd4\x0c\x25\xf0\x04\x0a\x05\x49\x01\xdc\x0f\x18\x8f\x7e\x89\x7e\xd7\x0f\x27\xd1\x0f\x27\xa8\x08\xd3\x0f\x31\x88\x04\xf0\x14\x00\x08\x0c\x80\x7c\xd9\x0e\x13\xd0\x14\x28\xa8\x38\xd1\x14\x33\xd3\x0e\x34\xd0\x08\x34\xd8\x07\x0b\xd7\x07\x26\xd1\x07\x26\xd0\x07\x32\xd8\x0b\x13\x90\x7a\xd2\x0b\x21\xa0\x58\xd7\x25\x36\xd1\x25\x36\xb0\x7b\xd4\x25\x43\xd9\x12\x17\xd0\x18\x3f\xd3\x12\x40\xd0\x0c\x40\xf0\x02\x07\x09\x47\x01\xd8\x1c\x24\xa0\x7b\xd1\x1c\x32\x88\x4d\xdc\x13\x26\xa0\x7d\xb0\x65\xd3\x13\x3c\xd0\x0c\x3c\xf0\x0c\x00\x0e\x12\x8f\x5b\x89\x5b\x80\x46\xd8\x07\x0d\x80\x7e\xd9\x0e\x13\xd0\x14\x46\xd8\x43\x4b\xf1\x03\x01\x15\x4c\x01\xf3\x00\x01\x0f\x4d\x01\xf0\x00\x01\x09\x4d\x01\xf0\x04\x03\x05\x26\xd8\x0f\x15\x8f\x7f\x89\x7f\x98\x78\xd3\x0f\x28\x88\x04\xf0\x06\x00\x08\x0c\x80\x7c\xd9\x0e\x13\xd0\x14\x35\xb8\x08\xd1\x14\x40\xd3\x0e\x41\xd0\x08\x41\xd8\x0b\x13\x90\x54\x98\x34\xd0\x0b\x1f\xd0\x04\x1f\xf8\xf4\x67\x01\x00\x10\x1b\xf2\x00\x06\x09\x16\xf0\x08\x00\x10\x11\x8f\x76\x89\x76\x88\x7e\xa0\x21\xa7\x26\xa1\x26\xa8\x48\xd2\x22\x34\xd8\x18\x20\xd7\x18\x2b\xd1\x18\x2b\xa8\x41\xaf\x46\xa9\x46\xb0\x53\xa9\x4c\xd4\x18\x39\xd8\x10\x15\xff\xf9\xf0\x0d\x06\x09\x16\xfb\xf4\x26\x00\x0d\x18\x9c\x1e\xac\x19\xb4\x4a\xd0\x0b\x3f\xf2\x00\x08\x05\x49\x01\xf0\x08\x00\x0f\x4b\x01\x88\x03\xd8\x0b\x13\xd7\x0b\x1c\xd1\x0b\x1c\x98\x55\xd4\x0b\x23\xd8\x0c\x0f\x90\x6d\xa0\x48\xa8\x53\xa8\x62\xa0\x4d\xa0\x3f\xf0\x00\x01\x33\x18\xd8\x18\x20\x90\x7a\xd0\x21\x36\xf0\x03\x01\x15\x38\xf1\x00\x01\x0d\x39\x88\x43\xe1\x0e\x13\x90\x43\x97\x4a\x91\x4a\x98\x78\xac\x14\xa8\x62\xab\x18\xd7\x29\x3a\xd1\x29\x3a\xb8\x42\xd3\x14\x3f\xd3\x0e\x40\xc0\x62\xd0\x08\x48\xfb\xf0\x11\x08\x05\x49\x01\xfb\xf0\x22\x00\x10\x15\xf2\x00\x04\x09\x47\x01\xd8\x0f\x17\x9c\x73\x9f\x7b\x99\x7b\xd1\x0f\x2a\xd8\x10\x15\xd9\x12\x17\xda\x39\x3a\xba\x48\xf0\x03\x01\x19\x46\x01\xf3\x00\x01\x13\x47\x01\xf0\x00\x01\x0d\x47\x01\xfb\xf0\x07\x04\x09\x47\x01\xfb\xf4\x16\x00\x0c\x17\xf2\x00\x01\x05\x26\xd9\x0e\x13\x94\x46\x98\x31\x93\x49\xd3\x0e\x1e\xa0\x41\xd0\x08\x25\xfb\xf0\x03\x01\x05\x26\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[97];
    }
runpy_toplevel_consts_11_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 96,
    },
    .ob_shash = -1,
    .ob_sval = "\xb2\x0b\x44\x3a\x00\xc2\x15\x1f\x46\x0b\x00\xc3\x2c\x10\x47\x3b\x00\xc4\x17\x11\x48\x29\x00\xc4\x3a\x09\x46\x08\x03\xc5\x03\x3a\x46\x03\x03\xc6\x03\x05\x46\x08\x03\xc6\x0b\x19\x47\x38\x03\xc6\x24\x41\x0f\x47\x33\x03\xc7\x33\x05\x47\x38\x03\xc7\x3b\x05\x48\x26\x03\xc8\x00\x21\x48\x21\x03\xc8\x21\x05\x48\x26\x03\xc8\x29\x09\x49\x09\x03\xc8\x32\x12\x49\x04\x03\xc9\x04\x05\x49\x09\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_existing = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "existing",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[14];
    }
const_str_pkg_main_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 13,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pkg_main_name",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[13];
        }_object;
    }
runpy_toplevel_consts_11_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 13,
        },
        .ob_item = {
            & const_str_mod_name._ascii.ob_base,
            & const_str_error._ascii.ob_base,
            & const_str_pkg_name._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[95],
            (PyObject *)&_Py_SINGLETON(strings).ascii[101],
            & const_str_existing._ascii.ob_base,
            & const_str_warn._ascii.ob_base,
            &_Py_ID(msg),
            & const_str_spec._ascii.ob_base,
            & const_str_ex._ascii.ob_base,
            & const_str_pkg_main_name._ascii.ob_base,
            & const_str_loader._ascii.ob_base,
            &_Py_ID(code),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[14];
    }
runpy_toplevel_consts_11_localspluskinds = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 13,
    },
    .ob_shash = -1,
    .ob_sval = "             ",
};
static
    struct _PyCode_DEF(1176)
runpy_toplevel_consts_11 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 588,
    },
    .co_consts = & runpy_toplevel_consts_11_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_11_names._object.ob_base.ob_base,
    .co_exceptiontable = & runpy_toplevel_consts_11_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 22 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 9,
    .co_firstlineno = 105,
    .co_nlocalsplus = 13,
    .co_nlocals = 13,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 755,
    .co_localsplusnames = & runpy_toplevel_consts_11_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & runpy_toplevel_consts_11_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__get_module_details._ascii.ob_base,
    .co_qualname = & const_str__get_module_details._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_11_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x00\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x72\x08\x02\x00\x7c\x01\x64\x02\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x6a\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x02\x7d\x03\x7d\x03\x7c\x02\x72\x63\x09\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x7c\x05\x81\x36\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x64\x03\xab\x02\x00\x00\x00\x00\x00\x00\x73\x2a\x64\x04\x64\x05\x6c\x09\x6d\x0a\x7d\x06\x01\x00\x64\x06\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x7c\x02\xac\x07\xab\x02\x00\x00\x00\x00\x00\x00\x7d\x07\x02\x00\x7c\x06\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x07\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x1a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x1f\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x08\x7c\x08\x80\x0b\x02\x00\x7c\x01\x64\x0e\x7c\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x08\x6a\x2c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x30\x7c\x00\x64\x0f\x6b\x28\x00\x00\x73\x11\x7c\x00\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x10\xab\x01\x00\x00\x00\x00\x00\x00\x72\x08\x02\x00\x7c\x01\x64\x11\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x09\x00\x7c\x00\x64\x10\x7a\x00\x00\x00\x7d\x0a\x74\x2f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x0a\x7c\x01\xab\x02\x00\x00\x00\x00\x00\x00\x53\x00\x7c\x08\x6a\x30\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x0b\x7c\x0b\x80\x0b\x02\x00\x7c\x01\x64\x14\x7c\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x09\x00\x7c\x0b\x6a\x33\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x0c\x7c\x0c\x80\x0b\x02\x00\x7c\x01\x64\x15\x7c\x00\x7a\x06\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x7c\x00\x7c\x08\x7c\x0c\x66\x03\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x45\x7d\x04\x7c\x04\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x81\x2d\x7c\x04\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\x6b\x37\x00\x00\x72\x1f\x7c\x02\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x7a\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x73\x01\x82\x00\x59\x00\x64\x00\x7d\x04\x7e\x04\x90\x01\x8c\x46\x64\x00\x7d\x04\x7e\x04\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x74\x20\x00\x00\x00\x00\x00\x00\x00\x00\x74\x22\x00\x00\x00\x00\x00\x00\x00\x00\x74\x24\x00\x00\x00\x00\x00\x00\x00\x00\x66\x04\x24\x00\x72\x54\x7d\x09\x64\x08\x7d\x07\x7c\x00\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x09\xab\x01\x00\x00\x00\x00\x00\x00\x72\x0f\x7c\x07\x64\x0a\x7c\x00\x64\x00\x64\x0b\x1a\x00\x9b\x00\x64\x0c\x7c\x00\x9b\x00\x64\x0d\x9d\x05\x7a\x0d\x00\x00\x7d\x07\x02\x00\x7c\x01\x7c\x07\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x29\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x01\x00\x00\x00\x00\x00\x00\x6a\x2a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x09\xab\x03\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x09\x82\x02\x64\x00\x7d\x09\x7e\x09\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x7c\x01\x24\x00\x72\x26\x7d\x04\x7c\x00\x74\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x76\x01\x72\x01\x82\x00\x02\x00\x7c\x01\x7c\x04\x9b\x01\x64\x12\x7c\x00\x9b\x02\x64\x13\x9d\x04\xab\x01\x00\x00\x00\x00\x00\x00\x82\x01\x64\x00\x7d\x04\x7e\x04\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x17\x7d\x04\x02\x00\x7c\x01\x74\x17\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x04\xab\x01\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x04\x82\x02\x64\x00\x7d\x04\x7e\x04\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[7];
    }
const_str__Error = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 6,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_Error",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[67];
    }
runpy_toplevel_consts_12_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 66,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Error that _run_module_as_main() should report without a traceback",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
runpy_toplevel_consts_12_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__Error._ascii.ob_base,
            & runpy_toplevel_consts_12_consts_1._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
runpy_toplevel_consts_12_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            &_Py_ID(__name__),
            &_Py_ID(__module__),
            &_Py_ID(__qualname__),
            &_Py_ID(__doc__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
runpy_toplevel_consts_12_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xda\x04\x4c",
};
static
    struct _PyCode_DEF(16)
runpy_toplevel_consts_12 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 8,
    },
    .co_consts = & runpy_toplevel_consts_12_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 166,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 756,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__Error._ascii.ob_base,
    .co_qualname = & const_str__Error._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_12_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[454];
    }
runpy_toplevel_consts_14_consts_0 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 453,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "\x52\x75\x6e\x73\x20\x74\x68\x65\x20\x64\x65\x73\x69\x67\x6e\x61\x74\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x6e\x20\x74\x68\x65\x20\x5f\x5f\x6d\x61\x69\x6e\x5f\x5f\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x4e\x6f\x74\x65\x20\x74\x68\x61\x74\x20\x74\x68\x65\x20\x65\x78\x65\x63\x75\x74\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x20\x77\x69\x6c\x6c\x20\x68\x61\x76\x65\x20\x66\x75\x6c\x6c\x20\x61\x63\x63\x65\x73\x73\x20\x74\x6f\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x5f\x5f\x6d\x61\x69\x6e\x5f\x5f\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x2e\x20\x49\x66\x20\x74\x68\x69\x73\x20\x69\x73\x20\x6e\x6f\x74\x20\x64\x65\x73\x69\x72\x61\x62\x6c\x65\x2c\x20\x74\x68\x65\x20\x72\x75\x6e\x5f\x6d\x6f\x64\x75\x6c\x65\x28\x29\x0a\x20\x20\x20\x20\x20\x20\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x73\x68\x6f\x75\x6c\x64\x20\x62\x65\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x72\x75\x6e\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x63\x6f\x64\x65\x20\x69\x6e\x20\x61\x20\x66\x72\x65\x73\x68\x20\x6e\x61\x6d\x65\x73\x70\x61\x63\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x41\x74\x20\x74\x68\x65\x20\x76\x65\x72\x79\x20\x6c\x65\x61\x73\x74\x2c\x20\x74\x68\x65\x73\x65\x20\x76\x61\x72\x69\x61\x62\x6c\x65\x73\x20\x69\x6e\x20\x5f\x5f\x6d\x61\x69\x6e\x5f\x5f\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x6f\x76\x65\x72\x77\x72\x69\x74\x74\x65\x6e\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x5f\x5f\x6e\x61\x6d\x65\x5f\x5f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x5f\x5f\x66\x69\x6c\x65\x5f\x5f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x5f\x5f\x63\x61\x63\x68\x65\x64\x5f\x5f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x5f\x5f\x6c\x6f\x61\x64\x65\x72\x5f\x5f\x0a\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x5f\x5f\x70\x61\x63\x6b\x61\x67\x65\x5f\x5f\x0a\x20\x20\x20\x20",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
runpy_toplevel_consts_14_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & runpy_toplevel_consts_14_consts_0._ascii.ob_base,
            &_Py_ID(__main__),
            & importlib__bootstrap_external_toplevel_consts_42_consts_4._ascii.ob_base,
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
const_str__get_main_module_details = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_main_module_details",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
runpy_toplevel_consts_14_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str__get_module_details._ascii.ob_base,
            & const_str__Error._ascii.ob_base,
            & const_str__get_main_module_details._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            & const_str_executable._ascii.ob_base,
            & const_str_exit._ascii.ob_base,
            &_Py_ID(modules),
            &_Py_ID(__dict__),
            &_Py_ID(origin),
            &_Py_ID(argv),
            & const_str__run_code._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__run_module_as_main = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_run_module_as_main",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[165];
    }
runpy_toplevel_consts_14_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 164,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x1c\x07\x05\x16\xd9\x0b\x15\x98\x18\xa0\x5a\xd2\x19\x2f\xdc\x27\x3a\xb8\x38\xc4\x56\xd3\x27\x4c\xd1\x0c\x24\x88\x48\x90\x68\xa1\x04\xe4\x27\x3f\xc4\x06\xd3\x27\x47\xd1\x0c\x24\x88\x48\x90\x68\xa0\x04\xf4\x08\x00\x14\x17\x97\x3b\x91\x3b\x98\x7a\xd1\x13\x2a\xd7\x13\x33\xd1\x13\x33\x80\x4c\xd9\x07\x11\xd8\x16\x1e\x97\x6f\x91\x6f\x8c\x03\x8f\x08\x89\x08\x90\x11\x89\x0b\xdc\x0b\x14\x90\x54\x98\x3c\xa8\x14\xd8\x15\x1f\xa0\x18\xf3\x03\x01\x0c\x2b\xf0\x00\x01\x05\x2b\xf8\xf4\x0d\x00\x0c\x12\xf2\x00\x02\x05\x16\xdc\x1a\x1d\x9f\x2e\x9b\x2e\xa9\x23\xd0\x0e\x2e\x88\x03\xdc\x08\x0b\x8f\x08\x89\x08\x90\x13\x8f\x0d\x89\x0d\xfb\xf0\x05\x02\x05\x16\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
runpy_toplevel_consts_14_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\x82\x2f\x41\x3c\x00\xc1\x3c\x09\x42\x39\x03\xc2\x05\x2a\x42\x34\x03\xc2\x34\x05\x42\x39\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_alter_argv = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "alter_argv",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_main_globals = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "main_globals",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
runpy_toplevel_consts_14_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & const_str_mod_name._ascii.ob_base,
            & const_str_alter_argv._ascii.ob_base,
            & const_str_mod_spec._ascii.ob_base,
            &_Py_ID(code),
            & const_str_exc._ascii.ob_base,
            &_Py_ID(msg),
            & const_str_main_globals._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(376)
runpy_toplevel_consts_14 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 188,
    },
    .co_consts = & runpy_toplevel_consts_14_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_14_names._object.ob_base.ob_base,
    .co_exceptiontable = & runpy_toplevel_consts_14_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 2,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 14 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 173,
    .co_nlocalsplus = 7,
    .co_nlocals = 7,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 757,
    .co_localsplusnames = & runpy_toplevel_consts_14_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_9_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__run_module_as_main._ascii.ob_base,
    .co_qualname = & const_str__run_module_as_main._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_14_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x09\x00\x7c\x01\x73\x05\x7c\x00\x64\x01\x6b\x37\x00\x00\x72\x15\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x00\x7d\x02\x7d\x03\x6e\x13\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x00\x7d\x02\x7d\x03\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\x6a\x0e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x06\x7c\x01\x72\x1d\x7f\x02\x6a\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x12\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x3c\x00\x00\x00\x74\x15\x00\x00\x00\x00\x00\x00\x00\x00\x7f\x03\x7c\x06\x64\x03\x64\x01\x7f\x02\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00\x23\x00\x74\x02\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x34\x7d\x04\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x9b\x01\x64\x02\x7c\x04\x9b\x01\x9d\x03\x7d\x05\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x59\x00\x64\x03\x7d\x04\x7e\x04\x8c\x83\x64\x03\x7d\x04\x7e\x04\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyCompactUnicodeObject _compact;
        uint16_t _data[801];
    }
runpy_toplevel_consts_15_consts_0 = {
    ._compact = {
        ._base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyUnicode_Type,
            },
            .length = 800,
            .hash = -1,
            .state = {
                .kind = 2,
                .compact = 1,
                .ascii = 0,
                .statically_allocated = 1,
            },
        },
        .utf8 = "\x45\x78\x65\x63\x75\x74\x65\x20\x61\x20\x6d\x6f\x64\x75\x6c\x65\x27\x73\x20\x63\x6f\x64\x65\x20\x77\x69\x74\x68\x6f\x75\x74\x20\x69\x6d\x70\x6f\x72\x74\x69\x6e\x67\x20\x69\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x6d\x6f\x64\x5f\x6e\x61\x6d\x65\x20\x2d\x2d\x20\x61\x6e\x20\x61\x62\x73\x6f\x6c\x75\x74\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6e\x61\x6d\x65\x20\x6f\x72\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x6e\x61\x6d\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x4f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x69\x6e\x69\x74\x5f\x67\x6c\x6f\x62\x61\x6c\x73\x20\x2d\x2d\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x70\x72\x65\x2d\x70\x6f\x70\x75\x6c\x61\x74\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\xe2\x80\x99\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x67\x6c\x6f\x62\x61\x6c\x73\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x62\x65\x66\x6f\x72\x65\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x20\x69\x73\x20\x65\x78\x65\x63\x75\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x72\x75\x6e\x5f\x6e\x61\x6d\x65\x20\x2d\x2d\x20\x69\x66\x20\x6e\x6f\x74\x20\x4e\x6f\x6e\x65\x2c\x20\x74\x68\x69\x73\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x73\x65\x74\x74\x69\x6e\x67\x20\x5f\x5f\x6e\x61\x6d\x65\x5f\x5f\x3b\x0a\x20\x20\x20\x20\x20\x20\x20\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x2c\x20\x5f\x5f\x6e\x61\x6d\x65\x5f\x5f\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x73\x65\x74\x20\x74\x6f\x20\x6d\x6f\x64\x5f\x6e\x61\x6d\x65\x20\x2b\x20\x27\x5f\x5f\x6d\x61\x69\x6e\x5f\x5f\x27\x20\x69\x66\x20\x74\x68\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x6e\x61\x6d\x65\x64\x20\x6d\x6f\x64\x75\x6c\x65\x20\x69\x73\x20\x61\x20\x70\x61\x63\x6b\x61\x67\x65\x20\x61\x6e\x64\x20\x74\x6f\x20\x6a\x75\x73\x74\x20\x6d\x6f\x64\x5f\x6e\x61\x6d\x65\x20\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x61\x6c\x74\x65\x72\x5f\x73\x79\x73\x20\x2d\x2d\x20\x69\x66\x20\x54\x72\x75\x65\x2c\x20\x73\x79\x73\x2e\x61\x72\x67\x76\x5b\x30\x5d\x20\x69\x73\x20\x75\x70\x64\x61\x74\x65\x64\x20\x77\x69\x74\x68\x20\x74\x68\x65\x20\x76\x61\x6c\x75\x65\x20\x6f\x66\x0a\x20\x20\x20\x20\x20\x20\x20\x5f\x5f\x66\x69\x6c\x65\x5f\x5f\x20\x61\x6e\x64\x20\x73\x79\x73\x2e\x6d\x6f\x64\x75\x6c\x65\x73\x5b\x5f\x5f\x6e\x61\x6d\x65\x5f\x5f\x5d\x20\x69\x73\x20\x75\x70\x64\x61\x74\x65\x64\x20\x77\x69\x74\x68\x20\x61\x20\x74\x65\x6d\x70\x6f\x72\x61\x72\x79\x0a\x20\x20\x20\x20\x20\x20\x20\x6d\x6f\x64\x75\x6c\x65\x20\x6f\x62\x6a\x65\x63\x74\x20\x66\x6f\x72\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\x20\x62\x65\x69\x6e\x67\x20\x65\x78\x65\x63\x75\x74\x65\x64\x2e\x20\x42\x6f\x74\x68\x20\x61\x72\x65\x0a\x20\x20\x20\x20\x20\x20\x20\x72\x65\x73\x74\x6f\x72\x65\x64\x20\x74\x6f\x20\x74\x68\x65\x69\x72\x20\x6f\x72\x69\x67\x69\x6e\x61\x6c\x20\x76\x61\x6c\x75\x65\x73\x20\x62\x65\x66\x6f\x72\x65\x20\x74\x68\x65\x20\x66\x75\x6e\x63\x74\x69\x6f\x6e\x20\x72\x65\x74\x75\x72\x6e\x73\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x69\x6e\x67\x20\x6d\x6f\x64\x75\x6c\x65\x20\x67\x6c\x6f\x62\x61\x6c\x73\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x2e\x0a\x20\x20\x20\x20",
        .utf8_length = 802,
    },
    ._data = {
        69, 120, 101, 99, 117, 116, 101, 32, 97, 32, 109, 111, 100, 117, 108, 101,
        39, 115, 32, 99, 111, 100, 101, 32, 119, 105, 116, 104, 111, 117, 116, 32,
        105, 109, 112, 111, 114, 116, 105, 110, 103, 32, 105, 116, 46, 10, 10, 32,
        32, 32, 32, 32, 32, 32, 109, 111, 100, 95, 110, 97, 109, 101, 32, 45,
        45, 32, 97, 110, 32, 97, 98, 115, 111, 108, 117, 116, 101, 32, 109, 111,
        100, 117, 108, 101, 32, 110, 97, 109, 101, 32, 111, 114, 32, 112, 97, 99,
        107, 97, 103, 101, 32, 110, 97, 109, 101, 46, 10, 10, 32, 32, 32, 32,
        32, 32, 32, 79, 112, 116, 105, 111, 110, 97, 108, 32, 97, 114, 103, 117,
        109, 101, 110, 116, 115, 58, 10, 32, 32, 32, 32, 32, 32, 32, 105, 110,
        105, 116, 95, 103, 108, 111, 98, 97, 108, 115, 32, 45, 45, 32, 100, 105,
        99, 116, 105, 111, 110, 97, 114, 121, 32, 117, 115, 101, 100, 32, 116, 111,
        32, 112, 114, 101, 45, 112, 111, 112, 117, 108, 97, 116, 101, 32, 116, 104,
        101, 32, 109, 111, 100, 117, 108, 101, 8217, 115, 10, 32, 32, 32, 32, 32,
        32, 32, 103, 108, 111, 98, 97, 108, 115, 32, 100, 105, 99, 116, 105, 111,
        110, 97, 114, 121, 32, 98, 101, 102, 111, 114, 101, 32, 116, 104, 101, 32,
        99, 111, 100, 101, 32, 105, 115, 32, 101, 120, 101, 99, 117, 116, 101, 100,
        46, 10, 10, 32, 32, 32, 32, 32, 32, 32, 114, 117, 110, 95, 110, 97,
        109, 101, 32, 45, 45, 32, 105, 102, 32, 110, 111, 116, 32, 78, 111, 110,
        101, 44, 32, 116, 104, 105, 115, 32, 119, 105, 108, 108, 32, 98, 101, 32,
        117, 115, 101, 100, 32, 102, 111, 114, 32, 115, 101, 116, 116, 105, 110, 103,
        32, 95, 95, 110, 97, 109, 101, 95, 95, 59, 10, 32, 32, 32, 32, 32,
        32, 32, 111, 116, 104, 101, 114, 119, 105, 115, 101, 44, 32, 95, 95, 110,
        97, 109, 101, 95, 95, 32, 119, 105, 108, 108, 32, 98, 101, 32, 115, 101,
        116, 32, 116, 111, 32, 109, 111, 100, 95, 110, 97, 109, 101, 32, 43, 32,
        39, 95, 95, 109, 97, 105, 110, 95, 95, 39, 32, 105, 102, 32, 116, 104,
        101, 10, 32, 32, 32, 32, 32, 32, 32, 110, 97, 109, 101, 100, 32, 109,
        111, 100, 117, 108, 101, 32, 105, 115, 32, 97, 32, 112, 97, 99, 107, 97,
        103, 101, 32, 97, 110, 100, 32, 116, 111, 32, 106, 117, 115, 116, 32, 109,
        111, 100, 95, 110, 97, 109, 101, 32, 111, 116, 104, 101, 114, 119, 105, 115,
        101, 46, 10, 10, 32, 32, 32, 32, 32, 32, 32, 97, 108, 116, 101, 114,
        95, 115, 121, 115, 32, 45, 45, 32, 105, 102, 32, 84, 114, 117, 101, 44,
        32, 115, 121, 115, 46, 97, 114, 103, 118, 91, 48, 93, 32, 105, 115, 32,
        117, 112, 100, 97, 116, 101, 100, 32, 119, 105, 116, 104, 32, 116, 104, 101,
        32, 118, 97, 108, 117, 101, 32, 111, 102, 10, 32, 32, 32, 32, 32, 32,
        32, 95, 95, 102, 105, 108, 101, 95, 95, 32, 97, 110, 100, 32, 115, 121,
        115, 46, 109, 111, 100, 117, 108, 101, 115, 91, 95, 95, 110, 97, 109, 101,
        95, 95, 93, 32, 105, 115, 32, 117, 112, 100, 97, 116, 101, 100, 32, 119,
        105, 116, 104, 32, 97, 32, 116, 101, 109, 112, 111, 114, 97, 114, 121, 10,
        32, 32, 32, 32, 32, 32, 32, 109, 111, 100, 117, 108, 101, 32, 111, 98,
        106, 101, 99, 116, 32, 102, 111, 114, 32, 116, 104, 101, 32, 109, 111, 100,
        117, 108, 101, 32, 98, 101, 105, 110, 103, 32, 101, 120, 101, 99, 117, 116,
        101, 100, 46, 32, 66, 111, 116, 104, 32, 97, 114, 101, 10, 32, 32, 32,
        32, 32, 32, 32, 114, 101, 115, 116, 111, 114, 101, 100, 32, 116, 111, 32,
        116, 104, 101, 105, 114, 32, 111, 114, 105, 103, 105, 110, 97, 108, 32, 118,
        97, 108, 117, 101, 115, 32, 98, 101, 102, 111, 114, 101, 32, 116, 104, 101,
        32, 102, 117, 110, 99, 116, 105, 111, 110, 32, 114, 101, 116, 117, 114, 110,
        115, 46, 10, 10, 32, 32, 32, 32, 32, 32, 32, 82, 101, 116, 117, 114,
        110, 115, 32, 116, 104, 101, 32, 114, 101, 115, 117, 108, 116, 105, 110, 103,
        32, 109, 111, 100, 117, 108, 101, 32, 103, 108, 111, 98, 97, 108, 115, 32,
        100, 105, 99, 116, 105, 111, 110, 97, 114, 121, 46, 10, 32, 32, 32, 32,
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
runpy_toplevel_consts_15_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & runpy_toplevel_consts_15_consts_0._compact._base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
runpy_toplevel_consts_15_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str__get_module_details._ascii.ob_base,
            & const_str__run_module_code._ascii.ob_base,
            & const_str__run_code._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[74];
    }
runpy_toplevel_consts_15_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 73,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf4\x2a\x00\x20\x33\xb0\x38\xd3\x1f\x3c\xd1\x04\x1c\x80\x48\x88\x68\x98\x04\xd8\x07\x0f\xd0\x07\x17\xd8\x13\x1b\x88\x08\xd9\x07\x10\xdc\x0f\x1f\xa0\x04\xa0\x6c\xb0\x48\xb8\x68\xd3\x0f\x47\xd0\x08\x47\xf4\x06\x00\x10\x19\x98\x14\x98\x72\xa0\x3c\xb0\x18\xb8\x38\xd3\x0f\x44\xd0\x08\x44",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_run_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "run_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_alter_sys = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "alter_sys",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
runpy_toplevel_consts_15_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_mod_name._ascii.ob_base,
            & const_str_init_globals._ascii.ob_base,
            & const_str_run_name._ascii.ob_base,
            & const_str_alter_sys._ascii.ob_base,
            & const_str_mod_spec._ascii.ob_base,
            &_Py_ID(code),
        },
    },
};
static
    struct _PyCode_DEF(102)
runpy_toplevel_consts_15 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 51,
    },
    .co_consts = & runpy_toplevel_consts_15_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_15_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 4,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 13 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 7,
    .co_firstlineno = 201,
    .co_nlocalsplus = 6,
    .co_nlocals = 6,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 758,
    .co_localsplusnames = & runpy_toplevel_consts_15_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_30_consts_4_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str_run_module._ascii.ob_base,
    .co_qualname = & const_str_run_module._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_15_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x00\x7d\x04\x7d\x05\x7c\x02\x80\x02\x7c\x00\x7d\x02\x7c\x03\x72\x0e\x74\x03\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x7c\x01\x7c\x02\x7c\x04\xab\x04\x00\x00\x00\x00\x00\x00\x53\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x69\x00\x7c\x01\x7c\x02\x7c\x04\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
runpy_toplevel_consts_16_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "can't find ",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
runpy_toplevel_consts_16_consts_3 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = " module in ",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
runpy_toplevel_consts_16_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            Py_None,
            &_Py_ID(__main__),
            & runpy_toplevel_consts_16_consts_2._ascii.ob_base,
            & runpy_toplevel_consts_16_consts_3._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
runpy_toplevel_consts_16_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_sys._ascii.ob_base,
            &_Py_ID(modules),
            & const_str__get_module_details._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str_str._ascii.ob_base,
            &_Py_ID(path),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[150];
    }
runpy_toplevel_consts_16_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 149,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x0a\x00\x11\x1b\x80\x49\xdc\x11\x14\x97\x1b\x91\x1b\x98\x59\xd1\x11\x27\x80\x4a\xdc\x08\x0b\x8f\x0b\x89\x0b\x90\x49\xd0\x08\x1e\xf0\x02\x08\x05\x2c\xdc\x0f\x22\xa0\x39\xd3\x0f\x2d\xf0\x0e\x00\x22\x2c\x8c\x03\x8f\x0b\x89\x0b\x90\x49\xd2\x08\x1e\xf8\xf4\x0d\x00\x0c\x17\xf2\x00\x04\x05\x0e\xd8\x0b\x14\x9c\x03\x98\x43\x9b\x08\xd1\x0b\x20\xda\x12\x17\xda\x1f\x28\xac\x23\xaf\x28\xa9\x28\xb0\x31\xaa\x2b\xf0\x03\x01\x19\x37\xf3\x00\x01\x13\x38\xd8\x3d\x40\xf0\x03\x01\x0d\x41\x01\xe0\x08\x0d\xfb\xf0\x09\x04\x05\x0e\xfb\xf0\x0c\x00\x22\x2c\x8c\x03\x8f\x0b\x89\x0b\x90\x49\xd2\x08\x1e\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
runpy_toplevel_consts_16_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\xa8\x0a\x41\x06\x00\xc1\x06\x09\x42\x02\x03\xc1\x0f\x2e\x41\x3d\x03\xc1\x3d\x05\x42\x02\x03\xc2\x02\x03\x42\x05\x00\xc2\x05\x15\x42\x1a\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_main_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "main_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
const_str_saved_main = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "saved_main",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
runpy_toplevel_consts_16_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            & const_str_error._ascii.ob_base,
            & const_str_main_name._ascii.ob_base,
            & const_str_saved_main._ascii.ob_base,
            & const_str_exc._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(314)
runpy_toplevel_consts_16 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 157,
    },
    .co_consts = & runpy_toplevel_consts_16_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_16_names._object.ob_base.ob_base,
    .co_exceptiontable = & runpy_toplevel_consts_16_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 12 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 8,
    .co_firstlineno = 231,
    .co_nlocalsplus = 4,
    .co_nlocals = 4,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 759,
    .co_localsplusnames = & runpy_toplevel_consts_16_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_3_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__get_main_module_details._ascii.ob_base,
    .co_qualname = & const_str__get_main_module_details._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_16_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x7d\x01\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x19\x00\x00\x00\x7d\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3d\x00\x09\x00\x74\x05\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x53\x00\x23\x00\x74\x06\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x33\x7d\x03\x7c\x01\x74\x09\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x76\x00\x72\x20\x02\x00\x7c\x00\x64\x02\x7c\x01\x9b\x02\x64\x03\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0a\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x19\x00\x00\x00\x9b\x02\x9d\x04\xab\x01\x00\x00\x00\x00\x00\x00\x7c\x03\x82\x02\x82\x00\x64\x00\x7d\x03\x7e\x03\x77\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x7c\x02\x74\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x01\x3c\x00\x00\x00\x77\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_read_code = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "read_code",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
runpy_toplevel_consts_17_consts_2 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_read_code._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
runpy_toplevel_consts_17_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_None,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & runpy_toplevel_consts_17_consts_2._object.ob_base.ob_base,
            & const_str_exec._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[8];
    }
const_str_pkgutil = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 7,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "pkgutil",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[9];
        }_object;
    }
runpy_toplevel_consts_17_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 9,
        },
        .ob_item = {
            & const_str_pkgutil._ascii.ob_base,
            & const_str_read_code._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(path),
            & const_str_abspath._ascii.ob_base,
            & const_str_io._ascii.ob_base,
            & const_str_open_code._ascii.ob_base,
            & const_str_compile._ascii.ob_base,
            &_Py_ID(read),
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
const_str__get_code_from_file = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "_get_code_from_file",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[125];
    }
runpy_toplevel_consts_17_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 124,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xe5\x04\x21\xdc\x10\x12\x97\x07\x91\x07\x97\x0f\x91\x0f\xa0\x05\xd3\x10\x26\x80\x49\xdc\x09\x0b\x8f\x1c\x89\x1c\x90\x69\xd4\x09\x20\xa0\x41\xd9\x0f\x18\x98\x11\x8b\x7c\x88\x04\xf7\x03\x00\x0a\x21\xe0\x07\x0b\x80\x7c\xe4\x0d\x0f\x8f\x5c\x89\x5c\x98\x29\xd4\x0d\x24\xa8\x01\xdc\x13\x1a\x98\x31\x9f\x36\x99\x36\x9b\x38\xa0\x55\xa8\x46\xd3\x13\x33\x88\x44\xf7\x03\x00\x0e\x25\xe0\x0b\x0f\x80\x4b\x88\x34\x80\x4b\xf7\x0d\x00\x0a\x21\xd0\x09\x20\xfa\xf7\x08\x00\x0e\x25\xe0\x0b\x0f\x80\x4b\xfa",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[24];
    }
runpy_toplevel_consts_17_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 23,
    },
    .ob_shash = -1,
    .ob_sval = "\xbb\x09\x42\x0b\x03\xc1\x23\x1c\x42\x17\x03\xc2\x0b\x05\x42\x14\x07\xc2\x17\x05\x42\x21\x07",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_code_path = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "code_path",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[5];
        }_object;
    }
runpy_toplevel_consts_17_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 5,
        },
        .ob_item = {
            & const_str_fname._ascii.ob_base,
            & const_str_read_code._ascii.ob_base,
            & const_str_code_path._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[102],
            &_Py_ID(code),
        },
    },
};
static
    struct _PyCode_DEF(328)
runpy_toplevel_consts_17 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 164,
    },
    .co_consts = & runpy_toplevel_consts_17_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_17_names._object.ob_base.ob_base,
    .co_exceptiontable = & runpy_toplevel_consts_17_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 1,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 11 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 6,
    .co_firstlineno = 250,
    .co_nlocalsplus = 5,
    .co_nlocals = 5,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 760,
    .co_localsplusnames = & runpy_toplevel_consts_17_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_7_consts_5_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str__get_code_from_file._ascii.ob_base,
    .co_qualname = & const_str__get_code_from_file._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_17_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x01\x64\x02\x6c\x00\x6d\x01\x7d\x01\x01\x00\x74\x04\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x09\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x02\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x03\x02\x00\x7c\x01\x7c\x03\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x04\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7f\x04\x80\x3b\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x0c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x03\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x03\x6a\x11\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x7c\x00\x64\x03\xab\x03\x00\x00\x00\x00\x00\x00\x7d\x04\x64\x00\x64\x00\x64\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x7c\x04\x53\x00\x7c\x04\x53\x00\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x8c\x48\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x7c\x04\x53\x00\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyCompactUnicodeObject _compact;
        uint16_t _data[531];
    }
runpy_toplevel_consts_18_consts_0 = {
    ._compact = {
        ._base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyUnicode_Type,
            },
            .length = 530,
            .hash = -1,
            .state = {
                .kind = 2,
                .compact = 1,
                .ascii = 0,
                .statically_allocated = 1,
            },
        },
        .utf8 = "\x45\x78\x65\x63\x75\x74\x65\x20\x63\x6f\x64\x65\x20\x6c\x6f\x63\x61\x74\x65\x64\x20\x61\x74\x20\x74\x68\x65\x20\x73\x70\x65\x63\x69\x66\x69\x65\x64\x20\x66\x69\x6c\x65\x73\x79\x73\x74\x65\x6d\x20\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x70\x61\x74\x68\x5f\x6e\x61\x6d\x65\x20\x2d\x2d\x20\x66\x69\x6c\x65\x73\x79\x73\x74\x65\x6d\x20\x6c\x6f\x63\x61\x74\x69\x6f\x6e\x20\x6f\x66\x20\x61\x20\x50\x79\x74\x68\x6f\x6e\x20\x73\x63\x72\x69\x70\x74\x2c\x20\x7a\x69\x70\x66\x69\x6c\x65\x2c\x0a\x20\x20\x20\x20\x20\x20\x20\x6f\x72\x20\x64\x69\x72\x65\x63\x74\x6f\x72\x79\x20\x63\x6f\x6e\x74\x61\x69\x6e\x69\x6e\x67\x20\x61\x20\x74\x6f\x70\x20\x6c\x65\x76\x65\x6c\x20\x5f\x5f\x6d\x61\x69\x6e\x5f\x5f\x2e\x70\x79\x20\x73\x63\x72\x69\x70\x74\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x4f\x70\x74\x69\x6f\x6e\x61\x6c\x20\x61\x72\x67\x75\x6d\x65\x6e\x74\x73\x3a\x0a\x20\x20\x20\x20\x20\x20\x20\x69\x6e\x69\x74\x5f\x67\x6c\x6f\x62\x61\x6c\x73\x20\x2d\x2d\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x70\x72\x65\x2d\x70\x6f\x70\x75\x6c\x61\x74\x65\x20\x74\x68\x65\x20\x6d\x6f\x64\x75\x6c\x65\xe2\x80\x99\x73\x0a\x20\x20\x20\x20\x20\x20\x20\x67\x6c\x6f\x62\x61\x6c\x73\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x20\x62\x65\x66\x6f\x72\x65\x20\x74\x68\x65\x20\x63\x6f\x64\x65\x20\x69\x73\x20\x65\x78\x65\x63\x75\x74\x65\x64\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x72\x75\x6e\x5f\x6e\x61\x6d\x65\x20\x2d\x2d\x20\x69\x66\x20\x6e\x6f\x74\x20\x4e\x6f\x6e\x65\x2c\x20\x74\x68\x69\x73\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x75\x73\x65\x64\x20\x74\x6f\x20\x73\x65\x74\x20\x5f\x5f\x6e\x61\x6d\x65\x5f\x5f\x3b\x0a\x20\x20\x20\x20\x20\x20\x20\x6f\x74\x68\x65\x72\x77\x69\x73\x65\x2c\x20\x27\x3c\x72\x75\x6e\x5f\x70\x61\x74\x68\x3e\x27\x20\x77\x69\x6c\x6c\x20\x62\x65\x20\x75\x73\x65\x64\x20\x66\x6f\x72\x20\x5f\x5f\x6e\x61\x6d\x65\x5f\x5f\x2e\x0a\x0a\x20\x20\x20\x20\x20\x20\x20\x52\x65\x74\x75\x72\x6e\x73\x20\x74\x68\x65\x20\x72\x65\x73\x75\x6c\x74\x69\x6e\x67\x20\x6d\x6f\x64\x75\x6c\x65\x20\x67\x6c\x6f\x62\x61\x6c\x73\x20\x64\x69\x63\x74\x69\x6f\x6e\x61\x72\x79\x2e\x0a\x20\x20\x20\x20",
        .utf8_length = 532,
    },
    ._data = {
        69, 120, 101, 99, 117, 116, 101, 32, 99, 111, 100, 101, 32, 108, 111, 99,
        97, 116, 101, 100, 32, 97, 116, 32, 116, 104, 101, 32, 115, 112, 101, 99,
        105, 102, 105, 101, 100, 32, 102, 105, 108, 101, 115, 121, 115, 116, 101, 109,
        32, 108, 111, 99, 97, 116, 105, 111, 110, 46, 10, 10, 32, 32, 32, 32,
        32, 32, 32, 112, 97, 116, 104, 95, 110, 97, 109, 101, 32, 45, 45, 32,
        102, 105, 108, 101, 115, 121, 115, 116, 101, 109, 32, 108, 111, 99, 97, 116,
        105, 111, 110, 32, 111, 102, 32, 97, 32, 80, 121, 116, 104, 111, 110, 32,
        115, 99, 114, 105, 112, 116, 44, 32, 122, 105, 112, 102, 105, 108, 101, 44,
        10, 32, 32, 32, 32, 32, 32, 32, 111, 114, 32, 100, 105, 114, 101, 99,
        116, 111, 114, 121, 32, 99, 111, 110, 116, 97, 105, 110, 105, 110, 103, 32,
        97, 32, 116, 111, 112, 32, 108, 101, 118, 101, 108, 32, 95, 95, 109, 97,
        105, 110, 95, 95, 46, 112, 121, 32, 115, 99, 114, 105, 112, 116, 46, 10,
        10, 32, 32, 32, 32, 32, 32, 32, 79, 112, 116, 105, 111, 110, 97, 108,
        32, 97, 114, 103, 117, 109, 101, 110, 116, 115, 58, 10, 32, 32, 32, 32,
        32, 32, 32, 105, 110, 105, 116, 95, 103, 108, 111, 98, 97, 108, 115, 32,
        45, 45, 32, 100, 105, 99, 116, 105, 111, 110, 97, 114, 121, 32, 117, 115,
        101, 100, 32, 116, 111, 32, 112, 114, 101, 45, 112, 111, 112, 117, 108, 97,
        116, 101, 32, 116, 104, 101, 32, 109, 111, 100, 117, 108, 101, 8217, 115, 10,
        32, 32, 32, 32, 32, 32, 32, 103, 108, 111, 98, 97, 108, 115, 32, 100,
        105, 99, 116, 105, 111, 110, 97, 114, 121, 32, 98, 101, 102, 111, 114, 101,
        32, 116, 104, 101, 32, 99, 111, 100, 101, 32, 105, 115, 32, 101, 120, 101,
        99, 117, 116, 101, 100, 46, 10, 10, 32, 32, 32, 32, 32, 32, 32, 114,
        117, 110, 95, 110, 97, 109, 101, 32, 45, 45, 32, 105, 102, 32, 110, 111,
        116, 32, 78, 111, 110, 101, 44, 32, 116, 104, 105, 115, 32, 119, 105, 108,
        108, 32, 98, 101, 32, 117, 115, 101, 100, 32, 116, 111, 32, 115, 101, 116,
        32, 95, 95, 110, 97, 109, 101, 95, 95, 59, 10, 32, 32, 32, 32, 32,
        32, 32, 111, 116, 104, 101, 114, 119, 105, 115, 101, 44, 32, 39, 60, 114,
        117, 110, 95, 112, 97, 116, 104, 62, 39, 32, 119, 105, 108, 108, 32, 98,
        101, 32, 117, 115, 101, 100, 32, 102, 111, 114, 32, 95, 95, 110, 97, 109,
        101, 95, 95, 46, 10, 10, 32, 32, 32, 32, 32, 32, 32, 82, 101, 116,
        117, 114, 110, 115, 32, 116, 104, 101, 32, 114, 101, 115, 117, 108, 116, 105,
        110, 103, 32, 109, 111, 100, 117, 108, 101, 32, 103, 108, 111, 98, 97, 108,
        115, 32, 100, 105, 99, 116, 105, 111, 110, 97, 114, 121, 46, 10, 32, 32,
        32, 32,
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[11];
    }
runpy_toplevel_consts_18_consts_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 10,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<run_path>",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
const_str_get_importer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "get_importer",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
runpy_toplevel_consts_18_consts_5 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_get_importer._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
runpy_toplevel_consts_18_consts_6 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_pkg_name._ascii.ob_base,
            & const_str_script_name._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[7];
        }_object;
    }
runpy_toplevel_consts_18_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 7,
        },
        .ob_item = {
            & runpy_toplevel_consts_18_consts_0._compact._base.ob_base,
            Py_None,
            & runpy_toplevel_consts_18_consts_2._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).ascii[46],
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            & runpy_toplevel_consts_18_consts_5._object.ob_base.ob_base,
            & runpy_toplevel_consts_18_consts_6._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[21];
        }_object;
    }
runpy_toplevel_consts_18_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 21,
        },
        .ob_item = {
            & const_str_rpartition._ascii.ob_base,
            & const_str_pkgutil._ascii.ob_base,
            & const_str_get_importer._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            & const_str_fsdecode._ascii.ob_base,
            &_Py_ID(isinstance),
            &_Py_ID(type),
            & const_str__get_code_from_file._ascii.ob_base,
            & const_str__run_module_code._ascii.ob_base,
            & const_str_sys._ascii.ob_base,
            &_Py_ID(path),
            & const_str_insert._ascii.ob_base,
            & const_str__get_main_module_details._ascii.ob_base,
            & const_str__TempModule._ascii.ob_base,
            & const_str__ModifiedArgv0._ascii.ob_base,
            &_Py_ID(module),
            &_Py_ID(__dict__),
            & const_str__run_code._ascii.ob_base,
            &_Py_ID(copy),
            & const_str_remove._ascii.ob_base,
            & const_str_ValueError._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[380];
    }
runpy_toplevel_consts_18_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 379,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xf0\x1e\x00\x08\x10\xd0\x07\x17\xd8\x13\x1f\x88\x08\xd8\x0f\x17\xd7\x0f\x22\xd1\x0f\x22\xa0\x33\xd3\x0f\x27\xa8\x01\xd1\x0f\x2a\x80\x48\xdd\x04\x24\xd9\x0f\x1b\x98\x49\xd3\x0f\x26\x80\x48\xdc\x10\x12\x97\x0b\x91\x0b\x98\x49\xd3\x10\x26\x80\x49\xdc\x07\x11\x90\x28\x9c\x44\xa0\x14\x9b\x4a\xd4\x07\x27\xf4\x06\x00\x10\x23\xa0\x39\xd3\x0f\x2d\x88\x04\xdc\x0f\x1f\xa0\x04\xa0\x6c\xb0\x48\xd8\x29\x31\xb8\x79\xf4\x03\x01\x10\x4a\x01\xf0\x00\x01\x09\x4a\x01\xf4\x0a\x00\x09\x0c\x8f\x08\x89\x08\x8f\x0f\x89\x0f\x98\x01\x98\x39\xd4\x08\x25\xf0\x02\x11\x09\x15\xf4\x0e\x00\x28\x40\x01\xd3\x27\x41\xd1\x0c\x24\x88\x48\x90\x68\xa0\x04\xdc\x11\x1c\x98\x58\xd4\x11\x26\xa8\x2b\xdc\x11\x1f\xa0\x09\xd5\x11\x2a\xd8\x1e\x29\xd7\x1e\x30\xd1\x1e\x30\xd7\x1e\x39\xd1\x1e\x39\x90\x0b\xdc\x17\x20\xa0\x14\xa0\x7b\xb0\x4c\xd8\x24\x2c\xa8\x68\xb8\x08\xf3\x03\x01\x18\x42\x01\xdf\x42\x46\xc1\x24\xc3\x26\xf7\x07\x00\x12\x2b\xd0\x11\x2a\xf7\x03\x00\x12\x27\xd0\x11\x26\xf0\x0c\x03\x0d\x15\xdc\x10\x13\x97\x08\x91\x08\x97\x0f\x91\x0f\xa0\x09\xd5\x10\x2a\xf8\xdc\x13\x1d\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfa\xf7\x0f\x00\x12\x2b\xd0\x11\x2a\xfa\xd0\x11\x2a\xf7\x03\x00\x12\x27\xd7\x11\x26\xd1\x11\x26\xfa\xf0\x0c\x03\x0d\x15\xdc\x10\x13\x97\x08\x91\x08\x97\x0f\x91\x0f\xa0\x09\xd5\x10\x2a\xf8\xdc\x13\x1d\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfb\xf0\x05\x03\x0d\x15\xdc\x10\x13\x97\x08\x91\x08\x97\x0f\x91\x0f\xa0\x09\xd5\x10\x2a\xf8\xdc\x13\x1d\xf2\x00\x01\x0d\x15\xd9\x10\x14\xf0\x03\x01\x0d\x15\xfd",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[139];
    }
runpy_toplevel_consts_18_exceptiontable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 138,
    },
    .ob_shash = -1,
    .ob_sval = "\xc2\x0d\x19\x45\x3a\x00\xc2\x26\x0c\x44\x3e\x03\xc2\x32\x34\x44\x28\x05\xc3\x26\x09\x44\x3e\x03\xc3\x2f\x09\x45\x3a\x00\xc3\x39\x1f\x44\x19\x02\xc4\x19\x09\x44\x25\x05\xc4\x24\x01\x44\x25\x05\xc4\x28\x05\x44\x31\x09\xc4\x2d\x07\x44\x3e\x03\xc4\x35\x09\x45\x3a\x00\xc4\x3e\x05\x45\x07\x07\xc5\x03\x07\x45\x3a\x00\xc5\x0b\x1f\x45\x2b\x00\xc5\x2b\x09\x45\x37\x03\xc5\x36\x01\x45\x37\x03\xc5\x3a\x01\x46\x2b\x03\xc5\x3c\x1f\x46\x1c\x04\xc6\x1b\x01\x46\x2b\x03\xc6\x1c\x09\x46\x28\x07\xc6\x25\x02\x46\x2b\x03\xc6\x27\x01\x46\x28\x07\xc6\x28\x03\x46\x2b\x03",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[10];
    }
const_str_path_name = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 9,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "path_name",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[9];
    }
const_str_importer = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 8,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "importer",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[11];
        }_object;
    }
runpy_toplevel_consts_18_localsplusnames = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 11,
        },
        .ob_item = {
            & const_str_path_name._ascii.ob_base,
            & const_str_init_globals._ascii.ob_base,
            & const_str_run_name._ascii.ob_base,
            & const_str_pkg_name._ascii.ob_base,
            & const_str_get_importer._ascii.ob_base,
            & const_str_importer._ascii.ob_base,
            &_Py_ID(code),
            & const_str_mod_name._ascii.ob_base,
            & const_str_mod_spec._ascii.ob_base,
            & const_str_temp_module._ascii.ob_base,
            & const_str_mod_globals._ascii.ob_base,
        },
    },
};
static
    struct _PyCode_DEF(860)
runpy_toplevel_consts_18 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 430,
    },
    .co_consts = & runpy_toplevel_consts_18_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_18_names._object.ob_base.ob_base,
    .co_exceptiontable = & runpy_toplevel_consts_18_exceptiontable.ob_base.ob_base,
    .co_flags = 3,
    .co_argcount = 3,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 21 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 10,
    .co_firstlineno = 262,
    .co_nlocalsplus = 11,
    .co_nlocals = 11,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 761,
    .co_localsplusnames = & runpy_toplevel_consts_18_localsplusnames._object.ob_base.ob_base,
    .co_localspluskinds = & importlib__bootstrap_toplevel_consts_46_consts_6_localspluskinds.ob_base.ob_base,
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = & const_str_run_path._ascii.ob_base,
    .co_qualname = & const_str_run_path._ascii.ob_base,
    .co_linetable = & runpy_toplevel_consts_18_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x7c\x02\x80\x02\x64\x02\x7d\x02\x7c\x02\x6a\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x03\xab\x01\x00\x00\x00\x00\x00\x00\x64\x04\x19\x00\x00\x00\x7d\x03\x64\x04\x64\x05\x6c\x01\x6d\x02\x7d\x04\x01\x00\x02\x00\x7c\x04\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x05\x74\x07\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x00\x74\x0b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x05\x74\x0d\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\xab\x02\x00\x00\x00\x00\x00\x00\x72\x1b\x74\x0f\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x7d\x06\x74\x11\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x01\x7c\x02\x7c\x03\x7c\x00\xac\x06\xab\x05\x00\x00\x00\x00\x00\x00\x53\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x17\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x04\x7c\x00\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x19\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x5c\x03\x00\x00\x7d\x07\x7d\x08\x7d\x06\x74\x1b\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x02\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x7d\x09\x74\x1d\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x35\x00\x01\x00\x7c\x09\x6a\x1e\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x20\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7d\x0a\x74\x23\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x06\x7c\x0a\x7c\x01\x7c\x02\x7c\x08\x7c\x03\xab\x06\x00\x00\x00\x00\x00\x00\x6a\x25\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x00\x00\x00\x00\x00\x00\x00\x63\x02\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x63\x02\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x09\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x53\x00\x23\x00\x74\x28\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x53\x00\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x6e\x03\x78\x03\x59\x00\x77\x01\x09\x00\x64\x01\x64\x01\x64\x01\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x6e\x0c\x23\x00\x31\x00\x73\x01\x77\x02\x01\x00\x59\x00\x01\x00\x01\x00\x6e\x03\x78\x03\x59\x00\x77\x01\x09\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x01\x23\x00\x74\x28\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x79\x01\x77\x00\x78\x03\x59\x00\x77\x01\x23\x00\x09\x00\x74\x12\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x6a\x27\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x7c\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x77\x00\x23\x00\x74\x28\x00\x00\x00\x00\x00\x00\x00\x00\x24\x00\x72\x03\x01\x00\x59\x00\x77\x00\x77\x00\x78\x03\x59\x00\x77\x01\x78\x03\x59\x00\x77\x01",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[34];
    }
runpy_toplevel_consts_21 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 33,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "No module specified for execution",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
runpy_toplevel_consts_25 = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_None,
            Py_None,
            Py_False,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[27];
        }_object;
    }
runpy_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 27,
        },
        .ob_item = {
            & runpy_toplevel_consts_0._ascii.ob_base,
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 0],
            Py_None,
            & const_str_run_module._ascii.ob_base,
            & const_str_run_path._ascii.ob_base,
            & runpy_toplevel_consts_5.ob_base.ob_base,
            & const_str__TempModule._ascii.ob_base,
            & runpy_toplevel_consts_7.ob_base.ob_base,
            & const_str__ModifiedArgv0._ascii.ob_base,
            & runpy_toplevel_consts_9.ob_base.ob_base,
            & runpy_toplevel_consts_10.ob_base.ob_base,
            & runpy_toplevel_consts_11.ob_base.ob_base,
            & runpy_toplevel_consts_12.ob_base.ob_base,
            & const_str__Error._ascii.ob_base,
            & runpy_toplevel_consts_14.ob_base.ob_base,
            & runpy_toplevel_consts_15.ob_base.ob_base,
            & runpy_toplevel_consts_16.ob_base.ob_base,
            & runpy_toplevel_consts_17.ob_base.ob_base,
            & runpy_toplevel_consts_18.ob_base.ob_base,
            &_Py_ID(__main__),
            (PyObject *)&_PyLong_SMALL_INTS[_PY_NSMALLNEGINTS + 2],
            & runpy_toplevel_consts_21._ascii.ob_base,
            & importlib__bootstrap_toplevel_consts_25_consts_3._object.ob_base.ob_base,
            & codecs_toplevel_consts_12_consts_7._object.ob_base.ob_base,
            & importlib__bootstrap_external_toplevel_consts_82._object.ob_base.ob_base,
            & runpy_toplevel_consts_25._object.ob_base.ob_base,
            & importlib__bootstrap_toplevel_consts_44_consts_10._object.ob_base.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
runpy_toplevel_names_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "importlib.machinery",
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[15];
    }
runpy_toplevel_names_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 14,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "importlib.util",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[29];
        }_object;
    }
runpy_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 29,
        },
        .ob_item = {
            &_Py_ID(__doc__),
            & const_str_sys._ascii.ob_base,
            & runpy_toplevel_names_2._ascii.ob_base,
            &_Py_ID(importlib),
            & runpy_toplevel_names_4._ascii.ob_base,
            & const_str_io._ascii.ob_base,
            & const_str_os._ascii.ob_base,
            &_Py_ID(__all__),
            &_Py_ID(type),
            & const_str_ModuleType._ascii.ob_base,
            &_Py_ID(object),
            & const_str__TempModule._ascii.ob_base,
            & const_str__ModifiedArgv0._ascii.ob_base,
            & const_str__run_code._ascii.ob_base,
            & const_str__run_module_code._ascii.ob_base,
            & const_str_ImportError._ascii.ob_base,
            & const_str__get_module_details._ascii.ob_base,
            & const_str_Exception._ascii.ob_base,
            & const_str__Error._ascii.ob_base,
            & const_str__run_module_as_main._ascii.ob_base,
            & const_str_run_module._ascii.ob_base,
            & const_str__get_main_module_details._ascii.ob_base,
            & const_str__get_code_from_file._ascii.ob_base,
            & const_str_run_path._ascii.ob_base,
            &_Py_ID(__name__),
            &_Py_ID(len),
            &_Py_ID(argv),
            & const_str_print._ascii.ob_base,
            &_Py_ID(stderr),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[247];
    }
runpy_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 246,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xf1\x02\x07\x01\x04\xf3\x18\x00\x01\x0b\xdb\x00\x1a\xdb\x00\x15\xdb\x00\x09\xdb\x00\x09\xf0\x06\x00\x05\x11\x90\x2a\xf0\x03\x02\x0b\x02\x80\x07\xf1\x0a\x00\x0e\x12\x90\x23\x8b\x59\x80\x0a\xf4\x04\x15\x01\x20\x90\x26\xf4\x00\x15\x01\x20\xf4\x2e\x0d\x01\x28\x90\x56\xf4\x00\x0d\x01\x28\xf0\x20\x00\x2f\x33\xd8\x26\x2a\xd8\x29\x2d\xf3\x05\x18\x01\x17\xf0\x34\x00\x29\x2d\xd8\x2c\x30\xd8\x2f\x33\xf3\x05\x0b\x01\x1e\xf0\x1c\x00\x29\x34\xf3\x00\x3b\x01\x20\xf4\x7a\x01\x01\x01\x4d\x01\x88\x59\xf4\x00\x01\x01\x4d\x01\xf3\x0e\x1a\x01\x2b\xf0\x38\x00\x27\x2b\xd8\x28\x2d\xf3\x03\x1c\x01\x45\x01\xf0\x3c\x00\x24\x2f\xf3\x00\x10\x01\x2c\xf2\x26\x0a\x01\x10\xf3\x18\x30\x01\x15\xf0\x66\x01\x00\x04\x0c\x88\x7a\xd2\x03\x19\xe1\x07\x0a\x88\x33\x8f\x38\x89\x38\x83\x7d\x90\x71\xd2\x07\x18\xd9\x08\x0d\xd0\x0e\x31\xb8\x03\xbf\x0a\xb9\x0a\xd6\x08\x43\xe0\x0c\x0f\x8f\x48\x89\x48\x90\x51\x88\x4b\xd9\x08\x1b\x98\x43\x9f\x48\x99\x48\xa0\x51\x99\x4b\xd5\x08\x28\xf0\x0d\x00\x04\x1a",
};
static
    struct _PyCode_DEF(384)
runpy_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 192,
    },
    .co_consts = & runpy_toplevel_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 5 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 5,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 762,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & runpy_toplevel_consts_5_consts_2_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & runpy_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x64\x02\x6c\x01\x5a\x01\x64\x01\x64\x02\x6c\x02\x5a\x03\x64\x01\x64\x02\x6c\x04\x5a\x03\x64\x01\x64\x02\x6c\x05\x5a\x05\x64\x01\x64\x02\x6c\x06\x5a\x06\x64\x03\x64\x04\x67\x02\x5a\x07\x02\x00\x65\x08\x65\x01\xab\x01\x00\x00\x00\x00\x00\x00\x5a\x09\x02\x00\x47\x00\x64\x05\x84\x00\x64\x06\x65\x0a\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x0b\x02\x00\x47\x00\x64\x07\x84\x00\x64\x08\x65\x0a\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x0c\x09\x00\x09\x00\x09\x00\x64\x17\x64\x09\x84\x01\x5a\x0d\x09\x00\x09\x00\x09\x00\x64\x17\x64\x0a\x84\x01\x5a\x0e\x65\x0f\x66\x01\x64\x0b\x84\x01\x5a\x10\x02\x00\x47\x00\x64\x0c\x84\x00\x64\x0d\x65\x11\xab\x03\x00\x00\x00\x00\x00\x00\x5a\x12\x64\x18\x64\x0e\x84\x01\x5a\x13\x09\x00\x09\x00\x64\x19\x64\x0f\x84\x01\x5a\x14\x65\x0f\x66\x01\x64\x10\x84\x01\x5a\x15\x64\x11\x84\x00\x5a\x16\x64\x1a\x64\x12\x84\x01\x5a\x17\x65\x18\x64\x13\x6b\x28\x00\x00\x72\x4d\x02\x00\x65\x19\x65\x01\x6a\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x64\x14\x6b\x02\x00\x00\x72\x15\x02\x00\x65\x1b\x64\x15\x65\x01\x6a\x38\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xac\x16\xab\x02\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02\x65\x01\x6a\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x3d\x00\x02\x00\x65\x13\x65\x01\x6a\x34\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\x19\x00\x00\x00\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02\x79\x02",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_runpy_toplevel(void)
{
    return Py_NewRef((PyObject *) &runpy_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_TestFrozenUtf8_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "TestFrozenUtf8_1",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
__hello___toplevel_consts_1_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_TestFrozenUtf8_1._ascii.ob_base,
            (PyObject *)&_Py_SINGLETON(strings).latin1[54],
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[19];
    }
__hello___toplevel_consts_1_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 18,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen __hello__>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
__hello___toplevel_consts_1_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xda\x04\x10",
};
static
    struct _PyCode_DEF(16)
__hello___toplevel_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 8,
    },
    .co_consts = & __hello___toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 3,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 763,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __hello___toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_TestFrozenUtf8_1._ascii.ob_base,
    .co_qualname = & const_str_TestFrozenUtf8_1._ascii.ob_base,
    .co_linetable = & __hello___toplevel_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_TestFrozenUtf8_2 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "TestFrozenUtf8_2",
};
static
    struct {
        PyCompactUnicodeObject _compact;
        uint16_t _data[2];
    }
__hello___toplevel_consts_3_consts_1 = {
    ._compact = {
        ._base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyUnicode_Type,
            },
            .length = 1,
            .hash = -1,
            .state = {
                .kind = 2,
                .compact = 1,
                .ascii = 0,
                .statically_allocated = 1,
            },
        },
        .utf8 = "\xcf\x80",
        .utf8_length = 2,
    },
    ._data = {
        960,
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
__hello___toplevel_consts_3_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_TestFrozenUtf8_2._ascii.ob_base,
            & __hello___toplevel_consts_3_consts_1._compact._base.ob_base,
            Py_None,
        },
    },
};
static
    struct _PyCode_DEF(16)
__hello___toplevel_consts_3 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 8,
    },
    .co_consts = & __hello___toplevel_consts_3_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 6,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 764,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __hello___toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_TestFrozenUtf8_2._ascii.ob_base,
    .co_qualname = & const_str_TestFrozenUtf8_2._ascii.ob_base,
    .co_linetable = & __hello___toplevel_consts_1_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[17];
    }
const_str_TestFrozenUtf8_4 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 16,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "TestFrozenUtf8_4",
};
static
    struct {
        PyCompactUnicodeObject _compact;
        uint32_t _data[2];
    }
__hello___toplevel_consts_5_consts_1 = {
    ._compact = {
        ._base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyUnicode_Type,
            },
            .length = 1,
            .hash = -1,
            .state = {
                .kind = 4,
                .compact = 1,
                .ascii = 0,
                .statically_allocated = 1,
            },
        },
        .utf8 = "\xf0\x9f\x98\x80",
        .utf8_length = 4,
    },
    ._data = {
        128512,
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
__hello___toplevel_consts_5_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_TestFrozenUtf8_4._ascii.ob_base,
            & __hello___toplevel_consts_5_consts_1._compact._base.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
__hello___toplevel_consts_5_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\x84\x00\xda\x04\x14",
};
static
    struct _PyCode_DEF(16)
__hello___toplevel_consts_5 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 8,
    },
    .co_consts = & __hello___toplevel_consts_5_consts._object.ob_base.ob_base,
    .co_names = & runpy_toplevel_consts_12_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 1 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 1,
    .co_firstlineno = 9,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 765,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __hello___toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_TestFrozenUtf8_4._ascii.ob_base,
    .co_qualname = & const_str_TestFrozenUtf8_4._ascii.ob_base,
    .co_linetable = & __hello___toplevel_consts_5_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x65\x00\x5a\x01\x64\x00\x5a\x02\x64\x01\x5a\x03\x79\x02",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[13];
    }
__hello___toplevel_consts_7_consts_1 = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 12,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "Hello world!",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
__hello___toplevel_consts_7_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            Py_None,
            & __hello___toplevel_consts_7_consts_1._ascii.ob_base,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[1];
        }_object;
    }
__hello___toplevel_consts_7_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 1,
        },
        .ob_item = {
            & const_str_print._ascii.ob_base,
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[11];
    }
__hello___toplevel_consts_7_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 10,
    },
    .ob_shash = -1,
    .ob_sval = "\x80\x00\xdc\x04\x09\x88\x2e\xd5\x04\x19",
};
static
    struct _PyCode_DEF(26)
__hello___toplevel_consts_7 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & __hello___toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & __hello___toplevel_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 12,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 766,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __hello___toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_main._ascii.ob_base,
    .co_qualname = & const_str_main._ascii.ob_base,
    .co_linetable = & __hello___toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[10];
        }_object;
    }
__hello___toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 10,
        },
        .ob_item = {
            Py_True,
            & __hello___toplevel_consts_1.ob_base.ob_base,
            & const_str_TestFrozenUtf8_1._ascii.ob_base,
            & __hello___toplevel_consts_3.ob_base.ob_base,
            & const_str_TestFrozenUtf8_2._ascii.ob_base,
            & __hello___toplevel_consts_5.ob_base.ob_base,
            & const_str_TestFrozenUtf8_4._ascii.ob_base,
            & __hello___toplevel_consts_7.ob_base.ob_base,
            &_Py_ID(__main__),
            Py_None,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[12];
    }
const_str_initialized = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 11,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "initialized",
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[6];
        }_object;
    }
__hello___toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 6,
        },
        .ob_item = {
            & const_str_initialized._ascii.ob_base,
            & const_str_TestFrozenUtf8_1._ascii.ob_base,
            & const_str_TestFrozenUtf8_2._ascii.ob_base,
            & const_str_TestFrozenUtf8_4._ascii.ob_base,
            & const_str_main._ascii.ob_base,
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[66];
    }
__hello___toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 65,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xd8\x0e\x12\x80\x0b\xf7\x04\x01\x01\x11\xf1\x00\x01\x01\x11\xf7\x06\x01\x01\x11\xf1\x00\x01\x01\x11\xf7\x06\x01\x01\x15\xf1\x00\x01\x01\x15\xf2\x06\x01\x01\x1a\xf0\x06\x00\x04\x0c\x88\x7a\xd2\x03\x19\xd9\x04\x08\x85\x46\xf0\x03\x00\x04\x1a",
};
static
    struct _PyCode_DEF(100)
__hello___toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 50,
    },
    .co_consts = & __hello___toplevel_consts._object.ob_base.ob_base,
    .co_names = & __hello___toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 4 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 4,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 767,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __hello___toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & __hello___toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x02\x00\x47\x00\x64\x01\x84\x00\x64\x02\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x01\x02\x00\x47\x00\x64\x03\x84\x00\x64\x04\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x02\x02\x00\x47\x00\x64\x05\x84\x00\x64\x06\xab\x02\x00\x00\x00\x00\x00\x00\x5a\x03\x64\x07\x84\x00\x5a\x04\x65\x05\x64\x08\x6b\x28\x00\x00\x72\x08\x02\x00\x65\x04\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x09\x79\x09",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get___hello___toplevel(void)
{
    return Py_NewRef((PyObject *) &__hello___toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[20];
    }
__phello___toplevel_consts_1_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 19,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen __phello__>",
};
static
    struct _PyCode_DEF(26)
__phello___toplevel_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & __hello___toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & __hello___toplevel_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 3,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 768,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __phello___toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_main._ascii.ob_base,
    .co_qualname = & const_str_main._ascii.ob_base,
    .co_linetable = & __hello___toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
__phello___toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_True,
            & __phello___toplevel_consts_1.ob_base.ob_base,
            &_Py_ID(__main__),
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
__phello___toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            & const_str_initialized._ascii.ob_base,
            & const_str_main._ascii.ob_base,
            &_Py_ID(__name__),
        },
    },
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[36];
    }
__phello___toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 35,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xd8\x0e\x12\x80\x0b\xf2\x04\x01\x01\x1a\xf0\x06\x00\x04\x0c\x88\x7a\xd2\x03\x19\xd9\x04\x08\x85\x46\xf0\x03\x00\x04\x1a",
};
static
    struct _PyCode_DEF(40)
__phello___toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & __phello___toplevel_consts._object.ob_base.ob_base,
    .co_names = & __phello___toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 769,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __phello___toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & __phello___toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x84\x00\x5a\x01\x65\x02\x64\x02\x6b\x28\x00\x00\x72\x08\x02\x00\x65\x01\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x03\x79\x03",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get___phello___toplevel(void)
{
    return Py_NewRef((PyObject *) &__phello___toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[24];
    }
__phello___ham_toplevel_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 23,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen __phello__.ham>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[6];
    }
__phello___ham_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 5,
    },
    .ob_shash = -1,
    .ob_sval = "\xf1\x03\x01\x01\x01",
};
static
    struct _PyCode_DEF(4)
__phello___ham_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 0 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 770,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __phello___ham_toplevel_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & __phello___ham_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x00",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get___phello___ham_toplevel(void)
{
    return Py_NewRef((PyObject *) &__phello___ham_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[29];
    }
__phello___ham_eggs_toplevel_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 28,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen __phello__.ham.eggs>",
};
static
    struct _PyCode_DEF(4)
__phello___ham_eggs_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 2,
    },
    .co_consts = & importlib__bootstrap_toplevel_consts_1_consts._object.ob_base.ob_base,
    .co_names = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 0 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 0,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 771,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __phello___ham_eggs_toplevel_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & __phello___ham_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x79\x00",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get___phello___ham_eggs_toplevel(void)
{
    return Py_NewRef((PyObject *) &__phello___ham_eggs_toplevel);
}

static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[25];
    }
__phello___spam_toplevel_consts_1_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 24,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen __phello__.spam>",
};
static
    struct _PyCode_DEF(26)
__phello___spam_toplevel_consts_1 = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 13,
    },
    .co_consts = & __hello___toplevel_consts_7_consts._object.ob_base.ob_base,
    .co_names = & __hello___toplevel_consts_7_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 3,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 3,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 772,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __phello___spam_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = & const_str_main._ascii.ob_base,
    .co_qualname = & const_str_main._ascii.ob_base,
    .co_linetable = & __hello___toplevel_consts_7_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x74\x01\x00\x00\x00\x00\x00\x00\x00\x00\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x00",
    ._co_firsttraceable = 0,
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[4];
        }_object;
    }
__phello___spam_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 4,
        },
        .ob_item = {
            Py_True,
            & __phello___spam_toplevel_consts_1.ob_base.ob_base,
            &_Py_ID(__main__),
            Py_None,
        },
    },
};
static
    struct _PyCode_DEF(40)
__phello___spam_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 20,
    },
    .co_consts = & __phello___spam_toplevel_consts._object.ob_base.ob_base,
    .co_names = & __phello___toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 2 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 2,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 773,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & __phello___spam_toplevel_consts_1_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & __phello___toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x64\x01\x84\x00\x5a\x01\x65\x02\x64\x02\x6b\x28\x00\x00\x72\x08\x02\x00\x65\x01\xab\x00\x00\x00\x00\x00\x00\x00\x01\x00\x79\x03\x79\x03",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get___phello___spam_toplevel(void)
{
    return Py_NewRef((PyObject *) &__phello___spam_toplevel);
}

static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[3];
        }_object;
    }
frozen_only_toplevel_consts = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 3,
        },
        .ob_item = {
            Py_True,
            & __hello___toplevel_consts_7_consts_1._ascii.ob_base,
            Py_None,
        },
    },
};
static
    struct {
        PyGC_Head _gc_head;
        struct {
            PyObject_VAR_HEAD
            PyObject *ob_item[2];
        }_object;
    }
frozen_only_toplevel_names = {
    ._object = {
        .ob_base = {
            .ob_base = {
                .ob_refcnt = _Py_IMMORTAL_REFCNT,
                .ob_type = &PyTuple_Type,
            },
            .ob_size = 2,
        },
        .ob_item = {
            & const_str_initialized._ascii.ob_base,
            & const_str_print._ascii.ob_base,
        },
    },
};
static
    struct {
        PyASCIIObject _ascii;
        uint8_t _data[21];
    }
frozen_only_toplevel_filename = {
    ._ascii = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyUnicode_Type,
        },
        .length = 20,
        .hash = -1,
        .state = {
            .kind = 1,
            .compact = 1,
            .ascii = 1,
            .statically_allocated = 1,
        },
    },
    ._data = "<frozen frozen_only>",
};
static
    struct {
        PyObject_VAR_HEAD
        Py_hash_t ob_shash;
        char ob_sval[19];
    }
frozen_only_toplevel_linetable = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyBytes_Type,
        },
        .ob_size = 18,
    },
    .ob_shash = -1,
    .ob_sval = "\xf0\x03\x01\x01\x01\xd8\x0e\x12\x80\x0b\xd9\x00\x05\x80\x6e\xd5\x00\x15",
};
static
    struct _PyCode_DEF(24)
frozen_only_toplevel = {
    .ob_base = {
        .ob_base = {
            .ob_refcnt = _Py_IMMORTAL_REFCNT,
            .ob_type = &PyCode_Type,
        },
        .ob_size = 12,
    },
    .co_consts = & frozen_only_toplevel_consts._object.ob_base.ob_base,
    .co_names = & frozen_only_toplevel_names._object.ob_base.ob_base,
    .co_exceptiontable = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_flags = 0,
    .co_argcount = 0,
    .co_posonlyargcount = 0,
    .co_kwonlyargcount = 0,
    .co_framesize = 3 + FRAME_SPECIALS_SIZE,
    .co_stacksize = 3,
    .co_firstlineno = 1,
    .co_nlocalsplus = 0,
    .co_nlocals = 0,
    .co_ncellvars = 0,
    .co_nfreevars = 0,
    .co_version = 774,
    .co_localsplusnames = (PyObject *)& _Py_SINGLETON(tuple_empty),
    .co_localspluskinds = (PyObject *)&_Py_SINGLETON(bytes_empty),
    .co_filename = & frozen_only_toplevel_filename._ascii.ob_base,
    .co_name = &_Py_STR(anon_module),
    .co_qualname = &_Py_STR(anon_module),
    .co_linetable = & frozen_only_toplevel_linetable.ob_base.ob_base,
    ._co_cached = NULL,
    .co_code_adaptive = "\x97\x00\x64\x00\x5a\x00\x02\x00\x65\x01\x64\x01\xab\x01\x00\x00\x00\x00\x00\x00\x01\x00\x79\x02",
    ._co_firsttraceable = 0,
};

PyObject *
_Py_get_frozen_only_toplevel(void)
{
    return Py_NewRef((PyObject *) &frozen_only_toplevel);
}

void
_Py_Deepfreeze_Fini(void) {
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_20_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_25);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_26_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_27_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_27);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_28);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_29);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_33);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_34);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_37);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_38);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_39);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_40);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_41);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_42);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_43);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_48_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_48_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_48);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_50);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_51);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_52);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_55);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_56);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_57);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_59);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_60);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_61);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_62);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_63);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_64);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel_consts_65);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_11_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_11_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_17_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_19);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_23);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_24);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_25);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_36);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_37);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_38);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_39);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_40);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_41_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_41_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_41);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_42);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_43);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_44);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_45);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_46);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_47);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_48);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_50);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_51);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_62_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_62_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_62);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_74);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_75);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_76);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_77);
    _PyStaticCode_Fini((PyCodeObject *)&importlib__bootstrap_external_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_19);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_23);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_24);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_25);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_27);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_28);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_29);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_30);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_31);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel_consts_32);
    _PyStaticCode_Fini((PyCodeObject *)&zipimport_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_2_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_4_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&abc_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_12_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_12_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_14_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_14_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_16_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_16_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_16_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_16_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_16_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_18_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_18_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_18_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_18_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_18_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_18_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_20_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_20_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_20_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_20_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_20_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_22_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_22_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_22_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_22_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_22_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_22_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_24);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_28);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_30);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_33);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_34);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_35);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_36);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_37);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_38);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_39);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_40);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_41);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_42);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_43);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel_consts_44);
    _PyStaticCode_Fini((PyCodeObject *)&codecs_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&io_toplevel_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&io_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&io_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&io_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&io_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_17_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_17_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_20_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_20_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_24_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_24_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_24);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_26_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_26_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_26_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_28);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_30_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_30_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_30);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_32_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_32_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_32_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_32);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_34_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_34_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_34);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_36);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_38_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_38_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_38);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_40_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_40_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_40);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_42_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_42);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_44_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_44_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_44);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_46);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_48_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_48);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_49);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_50_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_50_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_50);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_9_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_11_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_12_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_13_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_52);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_54);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_56);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_58_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_58_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_58_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_58);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_60_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_60_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_60_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_60);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_62_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_62_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_62_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_62);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_64_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_64_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_64);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_66);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_10_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_68);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_70_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_70_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_70);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_72);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel_consts_74);
    _PyStaticCode_Fini((PyCodeObject *)&_collections_abc_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_3_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_3_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_3_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_7_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_7_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&_sitebuiltins_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_18_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_18_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&genericpath_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_19);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_23);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_25);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_27);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_29);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_30);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_32_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_32);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_34);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_36);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_37);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_39);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_40);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_43);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_44);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_46);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_47);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_53);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel_consts_54);
    _PyStaticCode_Fini((PyCodeObject *)&ntpath_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_19);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_23);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_24);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_26_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_28);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_29);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_32);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_33);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_35);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_36_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel_consts_36);
    _PyStaticCode_Fini((PyCodeObject *)&posixpath_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_19);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_79);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_80);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_81);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_83);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_86);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_87_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_87_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_87);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_89);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_90);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_91);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_92);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_93);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_94);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_96);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_97);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_7_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_99);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_101_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_101_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_101_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_101_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_101);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_102);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_104);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_105);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_107_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_107_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_107);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_112);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_113);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_114);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_115);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_116);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_118);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_119);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_123);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_124);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_128);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_129_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_129_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_129_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_129_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_129_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_129_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_129);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_132);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_133);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_135_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_135_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_135);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_137_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_137_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_137_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_137_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_137_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_137);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel_consts_139);
    _PyStaticCode_Fini((PyCodeObject *)&os_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_8);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_11_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_19);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_21_consts_1_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_21_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_22_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_23);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_24);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_25);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_26_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&site_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_20);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_22);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_23);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_24);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_25);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_26);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_27);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_28);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_29);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel_consts_58);
    _PyStaticCode_Fini((PyCodeObject *)&stat_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_19);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_21_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_21_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_21);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_3_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_6);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel_consts_23);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_util_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_machinery_toplevel_consts_13);
    _PyStaticCode_Fini((PyCodeObject *)&importlib_machinery_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_5_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_5_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_5_consts_4);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_7_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_7_consts_2);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_7_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_9);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_10);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_11);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_12);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_14);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_15);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_16);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_17);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel_consts_18);
    _PyStaticCode_Fini((PyCodeObject *)&runpy_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&__hello___toplevel_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&__hello___toplevel_consts_3);
    _PyStaticCode_Fini((PyCodeObject *)&__hello___toplevel_consts_5);
    _PyStaticCode_Fini((PyCodeObject *)&__hello___toplevel_consts_7);
    _PyStaticCode_Fini((PyCodeObject *)&__hello___toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&__phello___toplevel_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&__phello___toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&__phello___ham_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&__phello___ham_eggs_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&__phello___spam_toplevel_consts_1);
    _PyStaticCode_Fini((PyCodeObject *)&__phello___spam_toplevel);
    _PyStaticCode_Fini((PyCodeObject *)&frozen_only_toplevel);
}
int
_Py_Deepfreeze_Init(void) {
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_20_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_25) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_26_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_27_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_27) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_28) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_29) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_30) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_33) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_34) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_37) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_38) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_39) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_40) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_41) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_42) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_43) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_44) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_46) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_48_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_48_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_48) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_50) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_51) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_52) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_55) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_56) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_57) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_59) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_60) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_61) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_62) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_63) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_64) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel_consts_65) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_11_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_11_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_17_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_19) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_23) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_24) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_25) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_36) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_37) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_38) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_39) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_40) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_41_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_41_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_41) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_42) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_43) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_44) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_45) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_46) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_47) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_48) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_50) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_51) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_52) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_54) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_56) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_58) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_60) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_62_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_62_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_62) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_7_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_64) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_66) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_68) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_70) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_2_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_8_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_72) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_74) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_75) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_76) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel_consts_77) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib__bootstrap_external_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_19) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_23) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_24) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_25) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_27) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_28) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_29) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_30) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_31) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel_consts_32) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&zipimport_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_2_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_4_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&abc_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_12_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_12_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_14_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_14_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_16_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_16_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_16_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_16_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_16_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_18_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_18_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_18_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_18_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_18_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_18_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_20_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_20_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_20_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_20_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_20_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_22_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_22_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_22_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_22_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_22_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_22_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_24) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_28) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_30) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_33) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_34) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_35) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_36) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_37) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_38) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_39) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_40) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_41) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_42) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_43) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel_consts_44) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&codecs_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&io_toplevel_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&io_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&io_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&io_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&io_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_17_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_17_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_20_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_20_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_22_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_24_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_24_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_24) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_26_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_26_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_26_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_28_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_28) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_30_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_30_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_30) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_32_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_32_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_32_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_32) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_34_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_34_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_34) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_36_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_36) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_38_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_38_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_38) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_40_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_40_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_40) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_42_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_42) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_44_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_44_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_44) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_46_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_46) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_48_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_48) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_49) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_50_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_50_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_50) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_9_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_11_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_12_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_13_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_52) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_54) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_56) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_58_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_58_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_58_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_58) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_60_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_60_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_60_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_60) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_62_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_62_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_62_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_62) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_64_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_64_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_64) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_66) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_10_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_68) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_70_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_70_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_70) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_72) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel_consts_74) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_collections_abc_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_3_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_3_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_3_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_5_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_7_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_7_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&_sitebuiltins_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_18_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_18_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&genericpath_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_19) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_23) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_25) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_27) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_29) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_30) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_32_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_32) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_34) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_36) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_37) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_39) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_40) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_43) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_44) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_46) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_47) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_53) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel_consts_54) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&ntpath_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_19) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_23) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_24) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_26_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_28) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_29) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_32) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_33) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_35) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_36_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel_consts_36) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&posixpath_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_19) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_79) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_80) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_81) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_83) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_86) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_87_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_87_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_87) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_89) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_90) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_91) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_92) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_93) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_94) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_96) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_97) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_7_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_99) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_101_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_101_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_101_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_101_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_101) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_102) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_104) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_105) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_107_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_107_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_107) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_112) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_113) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_114) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_115) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_116) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_118) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_119) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_123) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_124) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_128) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_129_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_129_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_129_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_129_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_129_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_129_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_129) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_132) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_133) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_135_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_135_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_135) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_137_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_137_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_137_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_137_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_137_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_137) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel_consts_139) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&os_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_8) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_11_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_19) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_21_consts_1_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_21_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_22_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_23) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_24) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_25) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_26_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&site_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_20) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_22) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_23) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_24) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_25) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_26) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_27) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_28) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_29) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel_consts_58) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&stat_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_19_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_19) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_21_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_21_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_21) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_3_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_23_consts_6) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel_consts_23) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_util_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_machinery_toplevel_consts_13) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&importlib_machinery_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_5_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_5_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_5_consts_4) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_7_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_7_consts_2) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_7_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_9) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_10) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_11) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_12) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_14) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_15) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_16) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_17) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel_consts_18) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&runpy_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__hello___toplevel_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__hello___toplevel_consts_3) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__hello___toplevel_consts_5) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__hello___toplevel_consts_7) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__hello___toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__phello___toplevel_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__phello___toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__phello___ham_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__phello___ham_eggs_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__phello___spam_toplevel_consts_1) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&__phello___spam_toplevel) < 0) {
        return -1;
    }
    if (_PyStaticCode_Init((PyCodeObject *)&frozen_only_toplevel) < 0) {
        return -1;
    }
    return 0;
}

uint32_t _Py_next_func_version = 775;

